/**
 * SECTION:applet-calendar-clock
 * @short_description: 时钟的显示
 *
 * 时钟表盘的描画，时针分针秒针的描画，时间刻度的
 * 描画，表针的动态显示以及时间的实时更新。
 *
 */


#include <math.h>

#include "common.h"
#include "applet-calendar-clock.h"
#include "applet-calendar-private.h"

#define CLOCK_DEFAULT_WIDTH (140)
#define CLOCK_DEFAULT_HEIGHT (140)
#define DEFAULT_REPAINT_INTERVAL (500)

#define SEC_IN_A_MIN (60)
#define SEC_IN_AN_HOUR (60 * SEC_IN_A_MIN)
#define SEC_IN_A_DAY (24 * SEC_IN_AN_HOUR)

struct _AppletCalendarClockPrivate
{
    StWidget *clockDrawArea;

    GSettings *settings;
    gboolean need_show_seconds;
    gboolean is_use_24h;

    StWidget *timeLabel;
    StWidget *weekLabel;

    gboolean needRepaint;
};

static void applet_calendar_clock_repaint (StDrawingArea *area, gpointer user_data);

static void load_settings (AppletCalendarClock *self);
static void apply_settings (AppletCalendarClock *self);
static void update_datetime (AppletCalendarClock *self);

static void gsetting_value_changed (GSettings *settings, gchar *key, gpointer user_data);

G_DEFINE_TYPE_WITH_PRIVATE (AppletCalendarClock, applet_calendar_clock, G_TYPE_OBJECT)

static void
applet_calendar_clock_finalize (GObject *calendarclock)
{
    AppletCalendarClockPrivate *priv = APPLET_CALENDAR_CLOCK (calendarclock)->priv;

    priv->needRepaint = FALSE;

    g_object_unref (priv->clockDrawArea);
    g_object_unref (priv->timeLabel);
    g_object_unref (priv->weekLabel);
    G_OBJECT_CLASS (applet_calendar_clock_parent_class)->finalize (calendarclock);
}

static gboolean
callRepaint (gpointer user_data)
{
    AppletCalendarClockPrivate *priv = APPLET_CALENDAR_CLOCK (user_data)->priv;

    update_datetime (APPLET_CALENDAR_CLOCK (user_data));

    GDateTime *now = g_date_time_new_now_local ();
    st_label_set_text (ST_LABEL (priv->weekLabel), g_date_time_format (now, "%A"));
    g_date_time_unref (now);

    st_drawing_area_queue_repaint ((StDrawingArea *)priv->clockDrawArea);
    return priv->needRepaint;
}

static void
applet_calendar_clock_class_init (AppletCalendarClockClass *klass)
{
    GObjectClass *objcls = G_OBJECT_CLASS (klass);

    objcls->finalize = applet_calendar_clock_finalize;
}

static void
applet_calendar_clock_init (AppletCalendarClock *self)
{
    AppletCalendarClockPrivate *priv;
    priv = self->priv = applet_calendar_clock_get_instance_private (self);

    priv->settings = g_settings_new (GSETTING_SCHEMA_NAME);
    load_settings (self);
    g_signal_connect (priv->settings, "changed::" SHOW_SEC_KEY, G_CALLBACK (gsetting_value_changed), self);
    g_signal_connect (priv->settings, "changed::" USE_24H_KEY, G_CALLBACK (gsetting_value_changed), self);

    priv->clockDrawArea = g_object_new (ST_TYPE_DRAWING_AREA, NULL);
    st_widget_add_style_class_name (ST_WIDGET (priv->clockDrawArea), "calendar-clock");
    clutter_actor_set_size (CLUTTER_ACTOR (priv->clockDrawArea), CLOCK_DEFAULT_WIDTH, CLOCK_DEFAULT_HEIGHT);
    clutter_actor_set_x_align (CLUTTER_ACTOR (priv->clockDrawArea), CLUTTER_ACTOR_ALIGN_CENTER);
    g_signal_connect (priv->clockDrawArea, "repaint", G_CALLBACK (applet_calendar_clock_repaint), self);

    self->actor = st_box_layout_new ();
    st_box_layout_set_vertical (ST_BOX_LAYOUT (self->actor), TRUE);

    priv->timeLabel = st_label_new ("");
    st_widget_set_style_class_name (ST_WIDGET (priv->timeLabel), "datemenu-date-label");
    clutter_actor_set_x_align (CLUTTER_ACTOR (priv->timeLabel), CLUTTER_ACTOR_ALIGN_CENTER);

    priv->weekLabel = st_label_new ("");
    st_widget_set_style_class_name (ST_WIDGET (priv->weekLabel), "datemenu-date-label");
    clutter_actor_set_x_align (CLUTTER_ACTOR (priv->weekLabel), CLUTTER_ACTOR_ALIGN_CENTER);
    priv->needRepaint = TRUE;

    _add_actor (self->actor, priv->clockDrawArea);
    _add_actor (self->actor, priv->timeLabel);
    _add_actor (self->actor, priv->weekLabel);

    clutter_actor_set_reactive (CLUTTER_ACTOR (self->actor), TRUE);

    apply_settings (self);

    g_timeout_add (DEFAULT_REPAINT_INTERVAL, callRepaint, self);
}

/**
 * applet_calendar_clock_new:
 *
 * 新建时钟
 *
 * Returns: 新建时钟对象
 */
AppletCalendarClock *
applet_calendar_clock_new (void)
{
    return g_object_new (APPLET_TYPE_CALENDAR_CLOCK, NULL);
}

static void
applet_calendar_clock_drawOuterCircle (cairo_t *context, gfloat clockX, gfloat clockY, gfloat clockRadius)
{
    cairo_save (context);
    //    cairo_set_source_rgba(context, 100.0f, 140.0f, 230.0f, 0.7);
    cairo_arc (context, clockX, clockY, clockRadius, 0, M_PI * 2);
    cairo_stroke (context);
    cairo_restore (context);
}

static void
applet_calendar_clock_drawCircle (cairo_t *context, gfloat clockX, gfloat clockY, gfloat clockRadius)
{
    cairo_save (context);
    //    cairo_set_source_rgba(context, 100.0f, 140.0f, 230.0f, 0.9);
    cairo_arc (context, clockX, clockY, clockRadius, 0, M_PI * 2);
    cairo_stroke (context);
    cairo_arc (context, clockX, clockY, clockRadius - 1, 0, M_PI * 2);
    cairo_stroke (context);
    cairo_restore (context);
}

static void
applet_calendar_clock_drawTick (cairo_t *context,
                                gfloat clockX,
                                gfloat clockY,
                                gfloat clockRadius,
                                gfloat angle,
                                int cnt)
{
    int tickWidth = 0;
    cairo_save (context);
    cairo_set_line_width (context, 1);
    if (0 == (cnt % 5))
    {
        tickWidth = 5;
        //        cairo_set_source_rgba(context, 100.0f, 140.0f, 230.0f, 0.9);
    }
    else
    {
        //        clockRadius = clockRadius - 1;
        tickWidth = 2;
        //        cairo_set_source_rgba(context, 100.0f, 140.0f, 230.0f, 0.7);
    }
    cairo_move_to (
        context, clockX + cos (angle) * (clockRadius - tickWidth), clockY + sin (angle) * (clockRadius - tickWidth));
    cairo_line_to (context, clockX + cos (angle) * (clockRadius), clockY + sin (angle) * (clockRadius));
    cairo_stroke (context);
    cairo_restore (context);
}

static void
applet_calendar_clock_drawTicks (cairo_t *context, gfloat clockX, gfloat clockY, gfloat clockRadius)
{
    gfloat angle = 0;
    int cnt = 0;
    for (angle = 0, cnt = 0; angle < M_PI * 2; angle = angle + M_PI / 30, cnt++)
    {
        applet_calendar_clock_drawTick (context, clockX, clockY, clockRadius, angle, cnt);
    }
}

static void
applet_calendar_clock_drawText (cairo_t *context, gfloat clockX, gfloat clockY, gfloat clockRadius)
{
    cairo_save (context);
    //    context.textAlign = "center";
    //    context.textBaseline = "middle";
    cairo_set_font_size (context, 8);
    //    cairo_set_source_rgba(context, 100.0f, 150.0f, 230.0f, 0.9);
    int i = 0;
    gfloat angle = 0;
    gchar buffer[5] = {0};

    for (angle = 0; angle < M_PI * 2; angle = angle + M_PI / 6)
    {
        cairo_move_to (context,
                       clockX + cos (angle - M_PI_2) * (clockRadius - 16) - 3,
                       clockY + sin (angle - M_PI_2) * (clockRadius - 18) + 3);
        if (angle == 0)
        {
            cairo_show_text (context, "12");
        }
        else
        {
            g_snprintf (buffer, sizeof (buffer), "%d", i);
            cairo_show_text (context, buffer);
        }
        i++;
    }
    cairo_restore (context);
}

static void
applet_calendar_clock_drawArm (cairo_t *context,
                               gfloat clockX,
                               gfloat clockY,
                               gfloat clockRadius,
                               gfloat progress,
                               gfloat armThickness,
                               gfloat armLength,
                               const gchar *armColor)
{
    gfloat armRadians = (2 * M_PI * progress) - (M_PI_2);
    gfloat targetX = clockX + cos (armRadians) * (armLength * clockRadius);
    gfloat targetY = clockY + sin (armRadians) * (armLength * clockRadius);
    cairo_save (context);
    cairo_set_line_width (context, armThickness);
    if (0 == g_strcmp0 (armColor, "second"))
    {
        //        cairo_set_source_rgb(context, 50.0f, 60.0f, 40.0f);
        gfloat secLeftArm = (0.5 + progress) * 2 * M_PI - (M_PI_2); //\short part of second's hand
        gfloat targetx2 = clockX + cos (secLeftArm) * (0.2 * clockRadius);
        gfloat targety2 = clockY + sin (secLeftArm) * (0.2 * clockRadius);
        cairo_move_to (context, clockX, clockY);
        cairo_line_to (context, targetx2, targety2);
        cairo_stroke (context);

        cairo_move_to (context, clockX, clockY); // Start at the center
        cairo_line_to (context, targetX, targetY); // long part of second
        cairo_stroke (context);
        cairo_restore (context);
    }
    else
    { // for hour and minute's hand
        //        cairo_set_source_rgba(context, 51.0f, 51.0f, 51.0f, 1.0f);
        cairo_move_to (context, clockX, clockY); // Start at the center
        cairo_line_to (context, targetX, targetY); // Draw a line outwards
        cairo_stroke (context);
        cairo_restore (context);
    }
}

static void
applet_calendar_clock_repaint (StDrawingArea *area, gpointer user_data)
{
    GDateTime *now = g_date_time_new_now_local ();
    gint h = g_date_time_get_hour (now);
    gint m = g_date_time_get_minute (now);
    gint s = g_date_time_get_second (now);
    g_date_time_unref (now);

    guint width;
    guint height;
    st_drawing_area_get_surface_size (area, &width, &height);
    //    g_print("%s --> (width, height) is (%d, %d)!\n", __func__, width, height);

    cairo_t *context = st_drawing_area_get_context (area);
    gfloat scaleX = 1;
    gfloat scaleY = 1;
    gfloat clockX = (((gfloat)width) / 2.0f) / scaleX;
    gfloat clockY = (((gfloat)height) / 2.0f) / scaleY;
    gfloat clockRadius = MIN (clockX, clockY) - 2;

    //    g_print("%s --> (x, y, r) is (%f, %f, %f)!\n", __func__, clockX, clockY, clockRadius);
    cairo_scale (context, scaleX, scaleY);
    applet_calendar_clock_drawOuterCircle (context, clockX, clockY, clockRadius);
    applet_calendar_clock_drawOuterCircle (context, clockX, clockY, clockRadius - 4);
    applet_calendar_clock_drawCircle (context, clockX, clockY, clockRadius - 4);
    applet_calendar_clock_drawText (context, clockX, clockY, clockRadius);
    applet_calendar_clock_drawTicks (context, clockX, clockY, clockRadius - 6);
    applet_calendar_clock_drawArm (
        context, clockX, clockY, clockRadius, ((gfloat)h) / 12 + ((gfloat)m) / 720, 3, 0.7, "hour");
    applet_calendar_clock_drawArm (context, clockX, clockY, clockRadius, ((gfloat)m) / 60, 2, 0.8, "minute");
    applet_calendar_clock_drawArm (context, clockX, clockY, clockRadius, ((gfloat)s) / 60, 1.5, 0.85, "second");
}

static void
load_settings (AppletCalendarClock *self)
{
    AppletCalendarClockPrivate *priv = self->priv;
    priv->need_show_seconds = g_settings_get_boolean (priv->settings, SHOW_SEC_KEY);
    priv->is_use_24h = g_settings_get_boolean (priv->settings, USE_24H_KEY);
}

static void
apply_settings (AppletCalendarClock *self)
{
    update_datetime (self);
}

static void
update_datetime (AppletCalendarClock *self)
{
    AppletCalendarClockPrivate *priv = self->priv;

    GDateTime *now = g_date_time_new_now_local ();
    gchar *datetimeformat =
        g_strdup_printf ("%s:%%M%s", priv->is_use_24h ? "%H" : "%p %l", priv->need_show_seconds ? ":%S" : "");
    gchar *timestrtmp = g_date_time_format (now, datetimeformat);
    st_label_set_text (ST_LABEL (priv->timeLabel), timestrtmp);

    g_free (timestrtmp);
    g_free (datetimeformat);
    g_date_time_unref (now);
}

static void
gsetting_value_changed (GSettings *settings, gchar *key, gpointer user_data)
{
    load_settings (APPLET_CALENDAR_CLOCK (user_data));
    apply_settings (APPLET_CALENDAR_CLOCK (user_data));
}
