/**
 * SECTION:applet-calendar-item
 * @short_description: 日历项显示
 *
 * 当选项为月份时，点击显示该月份的天数；
 * 当选项为年时，显示该年的12个月份；
 * 当选项为年的区间时，显示该区间的年份。
 *
 */
#include "common.h"
#include "applet-calendar-item.h"

struct _AppletCalendarItemPrivate
{
    CalendarItemOption option;
    int data;

    StLabel *label;
    StLabel *lunar_label;
};

G_DEFINE_TYPE_WITH_PRIVATE (AppletCalendarItem, applet_calendar_item, ST_TYPE_WIDGET)

static void
applet_calendar_item_get_preferred_width (ClutterActor *actor,
                                          gfloat        for_height,
                                          gfloat       *min_width,
                                          gfloat       *nat_width)
{
    StThemeNode *theme_node = st_widget_get_theme_node ((StWidget *)actor);
    st_theme_node_adjust_for_height (theme_node, &for_height);

    gfloat min_label_width, nat_label_width;
    gfloat min_lunar_width, nat_lunar_width = 0;

    AppletCalendarItem *self = APPLET_CALENDAR_ITEM (actor);
    clutter_actor_get_preferred_width ((ClutterActor *)self->priv->label,
                                       for_height,
                                       &min_label_width,
                                       &nat_label_width);
    if (self->priv->option == OPT_DAY_WITH_LUNAR)
    {
        clutter_actor_get_preferred_width ((ClutterActor *)self->priv->lunar_label,
                                           for_height,
                                           &min_lunar_width,
                                           &nat_lunar_width);
    }

    if (min_width)
        *min_width = MAX(min_label_width, min_lunar_width);
    if (nat_width)
        *nat_width = MAX(nat_label_width, nat_lunar_width);

    st_theme_node_adjust_preferred_width (theme_node, min_width, nat_width);
}

static void
applet_calendar_item_get_preferred_height (ClutterActor *actor,
                                           gfloat        for_width,
                                           gfloat       *min_height,
                                           gfloat       *nat_height)
{
    StThemeNode *theme_node = st_widget_get_theme_node ((StWidget *)actor);
    st_theme_node_adjust_for_width (theme_node, &for_width);

    gfloat min_label_height, nat_label_height;
    gfloat min_lunar_height, nat_lunar_height = 0;

    AppletCalendarItem *self = APPLET_CALENDAR_ITEM (actor);
    clutter_actor_get_preferred_height ((ClutterActor *)self->priv->label,
                                        for_width,
                                        &min_label_height,
                                        &nat_label_height);
    if (self->priv->option == OPT_DAY_WITH_LUNAR)
    {
        clutter_actor_get_preferred_height ((ClutterActor *)self->priv->lunar_label,
                                            for_width,
                                            &min_lunar_height,
                                            &nat_lunar_height);
    }

    if (min_height)
        *min_height = min_label_height + min_lunar_height;
    if (nat_height)
        *nat_height = nat_label_height + nat_lunar_height;

    st_theme_node_adjust_preferred_height (theme_node, min_height, nat_height);
}

static void
applet_calendar_item_allocate (ClutterActor *actor,
                               const ClutterActorBox *box,
                               ClutterAllocationFlags flags)
{
    AppletCalendarItem *self = APPLET_CALENDAR_ITEM (actor);

    StThemeNode *theme_node = st_widget_get_theme_node ((StWidget *)actor);

    clutter_actor_set_allocation (actor, box, flags);

    gfloat min_label_width, nat_label_width;
    gfloat min_lunar_width, nat_lunar_width = 0;
    gfloat min_label_height, nat_label_height;
    gfloat min_lunar_height, nat_lunar_height = 0;

    clutter_actor_get_preferred_size ((ClutterActor *)self->priv->label,
                                       &min_label_width,
                                       &min_label_height,
                                       &nat_label_width,
                                       &nat_label_height);
    if (self->priv->option == OPT_DAY_WITH_LUNAR)
    {
        clutter_actor_get_preferred_size ((ClutterActor *)self->priv->lunar_label,
                                          &min_lunar_width,
                                          &min_lunar_height,
                                          &nat_lunar_width,
                                          &nat_lunar_height);
    }

    ClutterActorBox content_box, child_box;
    st_theme_node_get_content_box (theme_node, box, &content_box);

    gfloat avail_width = content_box.x2 - content_box.x1;
    gfloat avail_height = content_box.y2 - content_box.y1;
    gfloat hh = nat_label_height + nat_lunar_height;

    child_box.x1 = (int)(0.5 + (avail_width - nat_label_width) / 2);
    child_box.x2 = child_box.x1 + nat_label_width;
    child_box.y1 = (int)(0.5 + (avail_height - hh) / 2);
    child_box.y2 = child_box.y1 + nat_label_height;
    clutter_actor_allocate ((ClutterActor *)self->priv->label, &child_box, flags);

    if (self->priv->option == OPT_DAY_WITH_LUNAR)
    {
        child_box.x1 = (int)(0.5 + (avail_width - nat_lunar_width) / 2);
        child_box.x2 = child_box.x1 + nat_lunar_width;
        child_box.y1 = child_box.y2;
        child_box.y2 = child_box.y1 + nat_lunar_height;
        clutter_actor_allocate ((ClutterActor *)self->priv->lunar_label, &child_box, flags);
    }
}

static void
applet_calendar_item_finalize (GObject *calendaritem)
{
    AppletCalendarItemPrivate *priv = APPLET_CALENDAR_ITEM (calendaritem)->priv;
    priv->option = OPT_INVALID;
    priv->data = 0;
    G_OBJECT_CLASS (applet_calendar_item_parent_class)->finalize (calendaritem);
}

static void
applet_calendar_item_class_init (AppletCalendarItemClass *klass)
{
    GObjectClass *objcls = G_OBJECT_CLASS (klass);

    objcls->finalize = applet_calendar_item_finalize;

    ClutterActorClass *actor_class = CLUTTER_ACTOR_CLASS (klass);
    actor_class->get_preferred_width = applet_calendar_item_get_preferred_width;
    actor_class->get_preferred_height = applet_calendar_item_get_preferred_height;
    actor_class->allocate = applet_calendar_item_allocate;
}

static void
applet_calendar_item_init (AppletCalendarItem *self)
{
    AppletCalendarItemPrivate *priv = self->priv = applet_calendar_item_get_instance_private (self);
    priv->option = OPT_INVALID;
    priv->data = -1;

    priv->label = (StLabel *)st_label_new ("");
    _add_actor (self, priv->label);
    clutter_actor_set_reactive ((ClutterActor *)self, TRUE);
    st_widget_set_track_hover ((StWidget *)self, TRUE);
}

StWidget *
applet_calendar_item_new_with_label_month (const gchar *text, int month)
{
    StWidget *item = g_object_new (APPLET_TYPE_CALENDAR_ITEM, NULL);
    AppletCalendarItemPrivate *priv = APPLET_CALENDAR_ITEM (item)->priv;
    priv->option = OPT_MONTH;
    priv->data = month;
    st_label_set_text (priv->label, text);
    return item;
}

StWidget *
applet_calendar_item_new_with_label_year (const gchar *text, int year)
{
    StWidget *item = g_object_new (APPLET_TYPE_CALENDAR_ITEM, NULL);
    AppletCalendarItemPrivate *priv = APPLET_CALENDAR_ITEM (item)->priv;
    priv->option = OPT_YEAR;
    priv->data = year;
    st_label_set_text (priv->label, text);
    return item;
}

StWidget *
applet_calendar_item_new_with_label_day_of_month (const gchar *text, int day_of_month)
{
    StWidget *item = g_object_new (APPLET_TYPE_CALENDAR_ITEM, NULL);
    AppletCalendarItemPrivate *priv = APPLET_CALENDAR_ITEM (item)->priv;
    priv->option = OPT_DAY;
    priv->data = day_of_month;
    st_label_set_text (priv->label, text);
    return item;
}

StWidget *
applet_calendar_item_new_with_label_day_and_lunar (const gchar *text,
                                                   int day_of_month,
                                                   const gchar *lunar)
{
    StWidget *item = g_object_new (APPLET_TYPE_CALENDAR_ITEM, NULL);
    AppletCalendarItemPrivate *priv = APPLET_CALENDAR_ITEM (item)->priv;
    priv->option = OPT_DAY_WITH_LUNAR;
    priv->data = day_of_month;
    st_label_set_text (priv->label, text);
    priv->lunar_label = (StLabel *)st_label_new (lunar);
    st_widget_set_style_class_name ((StWidget *)priv->lunar_label, "calendar-lunar");
    _add_actor (item, priv->lunar_label);
    return item;
}

/**
 * applet_calendar_item_get_option:
 * @item: AppletCalendarItem对象
 *
 * Return value:返回当前item的用途，值为枚举值（无效/年/月/日）的一项。
 */
CalendarItemOption
applet_calendar_item_get_option (AppletCalendarItem *item)
{
    AppletCalendarItemPrivate *priv = item->priv;
    return priv->option;
}

/**
 * applet_calendar_item_get_item_data:
 * @item: AppletCalendarItem对象
 *
 * Return value:当前的年份、月份或者是日期。
 */
int
applet_calendar_item_get_item_data (AppletCalendarItem *item)
{
    AppletCalendarItemPrivate *priv = item->priv;
    return priv->data;
}
