/**
 * SECTION:st-applet-calendar
 * @short_description: 日期的显示
 *
 * 按照用户点击需求显示具体的年月日：
 *      当选项为月份时，点击显示该月份的天数；
 *      当选项为年时，显示该年的12个月份；
 *      当选项为年的区间时，显示该区间的年份。
 *
 */

#include <glib.h>
#include "st.h"

#include "common.h"
#include "cdos-lunar.h"

#include "applet-calendar.h"
#include "applet-calendar-item.h"

#define CALENDAR_PARAM_COMMON (G_PARAM_STATIC_NAME | G_PARAM_STATIC_NICK | G_PARAM_STATIC_BLURB)
#define CALENDAR_PARAM_READWRITE (G_PARAM_READWRITE | CALENDAR_PARAM_COMMON)
#define CALENDAR_PARAM_READABLE (G_PARAM_READABLE | CALENDAR_PARAM_COMMON)
#define CALENDAR_PARAM_WRITABLE (G_PARAM_WRITABLE | CALENDAR_PARAM_COMMON)

#define SEC_IN_A_MIN (60)
#define SEC_IN_AN_HOUR (60 * SEC_IN_A_MIN)
#define SEC_IN_A_DAY (24 * SEC_IN_AN_HOUR)

#define ACTOR_DEFAULT_WIDTH (168)
#define ACTOR_DEFAULT_HEIGHT (198)
#define TOP_BOX_DEFAULT_HEIGHT (30)
#define TOP_BOX_DEFAULT_WIDTH ACTOR_DEFAULT_WIDTH
#define MAIN_BOX_DEFAULT_WIDTH ACTOR_DEFAULT_WIDTH
#define MAIN_BOX_DEFAULT_HEIGHT (ACTOR_DEFAULT_HEIGHT - TOP_BOX_DEFAULT_HEIGHT)

#define ANIMATION_DURATION_TIME (600)
#define ANIMATION_DELAY_TIME (200)
#define ANIMATION_MODE_OUT (CLUTTER_EASE_OUT_SINE)
#define ANIMATION_MODE_IN (CLUTTER_EASE_IN_SINE)

struct _AppletCalendarPrivate
{
    StWidget *top_box;
    StWidget *main_box;
    StWidget *swipe_box[3];

    StWidget *view_type_switch;

    gint view_type;

    gint selected_year;
    gint selected_month;
    gint selected_day_of_month;

    gint day_month[6][7];
    gint day[6][7];
    const gchar *lunar_day[6][7];

    gint focus_unit;

    guint year_before : 1;

    gint week_start;
    gboolean show_week_num;

};

static const guint month_length[2][13] = {
    {0, 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31},
    {0, 31, 29, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31}
};
/*
static const guint days_in_months[2][14] = {
    {0, 0, 31, 59, 90, 120, 151, 181, 212, 243, 273, 304, 334, 365},
    {0, 0, 31, 60, 91, 121, 152, 182, 213, 244, 274, 305, 335, 366}
};
*/

static char *default_abbreviated_dayname[7] = {0};
static char *default_monthname[12] = {0};

static gboolean
leap (guint year)
{
    return ((((year % 4) == 0) && ((year % 100) != 0)) || ((year % 400) == 0));
}

static guint
day_of_week (guint year, guint mm, guint dd)
{
    GDateTime *dt;
    guint days;

    dt = g_date_time_new_local (year, mm, dd, 1, 1, 1);
    if (NULL == dt)
    {
        return 0;
    }

    days = g_date_time_get_day_of_week (dt);
    g_date_time_unref (dt);

    return days - 1;
}

static guint
week_of_year (guint year, guint mm, guint dd)
{
    GDateTime *dt;
    guint week;

    dt = g_date_time_new_local (year, mm, dd, 1, 1, 1);
    if (NULL == dt)
    {
        return 0;
    }

    week = g_date_time_get_week_of_year (dt);
    g_date_time_unref (dt);

    return week;
}

static gboolean
is_work_day (guint dayofweek)
{
    guint days = dayofweek % 7;
    return (0 != days && 6 != days);
}

enum
{
    VIEWTYPE_DAY = 0,
    VIEWTYPE_MONTH,
    VIEWTYPE_YEAR
};

enum
{
    MONTH_PREV,
    MONTH_CURRENT,
    MONTH_NEXT
};

enum
{
    MONTH_CHANGED_SIGNAL,
    DAY_SELECTED_SIGNAL,
    DAY_SELECTED_DOUBLE_CLICK_SIGNAL,
    PREV_MONTH_SIGNAL,
    NEXT_MONTH_SIGNAL,
    LAST_SIGNAL
};

enum
{
    PROP_0,
    PROP_YEAR,
    PROP_MONTH,
    PROP_DAY,
    PROP_SHOW_WEEK_NUMBERS,
};

G_DEFINE_TYPE_WITH_PRIVATE (AppletCalendar, applet_calendar, G_TYPE_OBJECT)

static guint applet_calendar_signals[LAST_SIGNAL] = {0};

// static void applet_calendar_set_property(GObject * obj,
//                                       guint prop_id,
//                                       const GValue * value);

// static void applet_calendar_get_property(GObject * obj,
//                                       guint prop_id,
//                                       GValue * value,
//                                       GParamSpec * pspec);

static void build_header (AppletCalendar *calendar);

static void calendar_compute_days (AppletCalendar *calendar);

static void set_date (AppletCalendar *calendar, GDateTime *date, gboolean isForceLoad, GTimeSpan changed);

static void on_prev_button_clicked (StButton *button, int clickedBtn, gpointer user_data);
static void on_next_button_clicked (StButton *button, int clickedBtn, gpointer user_data);
static void on_date_label_clicked (StButton *button, int clickedBtn, gpointer user_data);
static gboolean on_calendar_item_clicked (StButton *button, ClutterEvent *event, gpointer user_data);

static void update_day_view (AppletCalendar *calendar, gboolean isForceLoad, GTimeSpan changed);
static void update_month_view (AppletCalendar *calendar, gboolean isForceLoad, GTimeSpan changed);
static void update_year_view (AppletCalendar *calendar, gboolean isForceLoad, GTimeSpan changed);

static void
set_sensitive (StWidget *actor, gboolean sensitive)
{
    g_return_if_fail (NULL != actor && ST_IS_WIDGET (actor));

    clutter_actor_set_reactive (CLUTTER_ACTOR (actor), sensitive);
    st_widget_set_can_focus (ST_WIDGET (actor), sensitive);
    if (sensitive)
    {
        st_widget_remove_style_pseudo_class (ST_WIDGET (actor), "insensitive");
    }
    else
    {
        st_widget_remove_style_pseudo_class (ST_WIDGET (actor), "hover");
        st_widget_remove_style_pseudo_class (ST_WIDGET (actor), "active");
        st_widget_add_style_pseudo_class (ST_WIDGET (actor), "insensitive");
    }
}

static void
applet_calendar_class_init (AppletCalendarClass *klass)
{
    // g_print("%s --> enter!\n", __func__);
    GObjectClass *objClass = G_OBJECT_CLASS (klass);

    //    objClass->get_property = applet_calendar_get_property;
    //    objClass->set_property = applet_calendar_set_property;

    /**
     * CdosCalendar:year:
     *
     * The selected year.
     * This property gets initially set to the current year.
     */
    //    g_object_class_install_property(objClass,
    //                                    PROP_YEAR,
    //                                    g_param_spec_int("year",
    //                                                     "Year",
    //                                                     "The selected year",
    //                                                     0, G_MAXINT >> 9, 0,
    //                                                     CALENDAR_PARAM_READWRITE));

    /**
     * CdosCalendar:month:
     *
     * The selected month (as a number between 1 and 12).
     * This property gets initially set to the current month.
     */
    //    g_object_class_install_property(objClass,
    //                                    PROP_MONTH,
    //                                    g_param_spec_int("month",
    //                                                     "Month",
    //                                                     "The selected month (as a number between 1 and 12)",
    //                                                     1, 12, 0,
    //                                                     CALENDAR_PARAM_READWRITE));

    /**
     * CdosCalendar:day:
     *
     * The selected day (as a number between 1 and 31, or 0
     * to unselect the currently selected day).
     * This property gets initially set to the current day.
     */
    //    g_object_class_install_property(objClass,
    //                                    PROP_DAY,
    //                                    g_param_spec_int("day",
    //                                                     "Day",
    //                                                     "The selected day (as a number between 1 and 31, or 0 to
    //                                                     unselect the currently selected day)",
    //                                                     0, 31, 0,
    //                                                     CALENDAR_PARAM_READWRITE));

    /**
     * CdosCalendar:show-week-numbers:
     *
     * Determines whether week numbers are displayed.
     *
     */
    //    g_object_class_install_property(objClass,
    //                                    PROP_SHOW_WEEK_NUMBERS,
    //                                    g_param_spec_boolean("show-week-numbers",
    //                                                         "Show Week Numbers",
    //                                                         "If TRUE, week numbers are displayed",
    //                                                         FALSE,
    //                                                         CALENDAR_PARAM_READWRITE));

    /**
     * CdosCalendar:month-changed:
     * @calendar: the object which received the signal.
     *
     * Emitted when the user clicks a button to change the selected month on a
     * calendar.
     */
    applet_calendar_signals[MONTH_CHANGED_SIGNAL] = g_signal_new ("month-changed",
                                                                  G_OBJECT_CLASS_TYPE (objClass),
                                                                  G_SIGNAL_RUN_FIRST,
                                                                  G_STRUCT_OFFSET (AppletCalendarClass, month_changed),
                                                                  NULL,
                                                                  NULL,
                                                                  g_cclosure_marshal_VOID__VOID,
                                                                  G_TYPE_NONE,
                                                                  0);

    /**
     * CdosCalendar:day-selected:
     * @calendar: the object which received the signal.
     *
     * Emitted when the user selects a day.
     */
    applet_calendar_signals[DAY_SELECTED_SIGNAL] = g_signal_new ("day-selected",
                                                                 G_OBJECT_CLASS_TYPE (objClass),
                                                                 G_SIGNAL_RUN_FIRST,
                                                                 G_STRUCT_OFFSET (AppletCalendarClass, day_selected),
                                                                 NULL,
                                                                 NULL,
                                                                 g_cclosure_marshal_VOID__VOID,
                                                                 G_TYPE_NONE,
                                                                 0);

    /**
     * CdosCalendar:day-selected-double-click:
     * @calendar: the object which received the signal.
     *
     * Emitted when the user double-clicks a day.
     */
    applet_calendar_signals[DAY_SELECTED_DOUBLE_CLICK_SIGNAL] =
        g_signal_new ("day-selected-double-click",
                      G_OBJECT_CLASS_TYPE (objClass),
                      G_SIGNAL_RUN_FIRST,
                      G_STRUCT_OFFSET (AppletCalendarClass, day_selected_double_click),
                      NULL,
                      NULL,
                      g_cclosure_marshal_VOID__VOID,
                      G_TYPE_NONE,
                      0);

    /**
     * CdosCalendar:prev-month:
     * @calendar: the object which received the signal.
     *
     * Emitted when the user switched to the previous month.
     */
    applet_calendar_signals[PREV_MONTH_SIGNAL] = g_signal_new ("prev-month",
                                                               G_OBJECT_CLASS_TYPE (objClass),
                                                               G_SIGNAL_RUN_FIRST,
                                                               G_STRUCT_OFFSET (AppletCalendarClass, prev_month),
                                                               NULL,
                                                               NULL,
                                                               g_cclosure_marshal_VOID__VOID,
                                                               G_TYPE_NONE,
                                                               0);

    /**
     * CdosCalendar:next-month:
     * @calendar: the object which received the signal.
     *
     * Emitted when the user switched to the next month.
     */
    applet_calendar_signals[NEXT_MONTH_SIGNAL] = g_signal_new ("next-month",
                                                               G_OBJECT_CLASS_TYPE (objClass),
                                                               G_SIGNAL_RUN_FIRST,
                                                               G_STRUCT_OFFSET (AppletCalendarClass, next_month),
                                                               NULL,
                                                               NULL,
                                                               g_cclosure_marshal_VOID__VOID,
                                                               G_TYPE_NONE,
                                                               0);
    // g_print("%s --> end!\n", __func__);
}

static void
applet_calendar_init (AppletCalendar *self)
{
    // g_print("%s --> enter!\n", __func__);
    AppletCalendarPrivate *priv;
    time_t tmp_time;
    time_t secs;
    struct tm *tm;
    char buffer[255];
    int i = 0;

    priv = self->priv = applet_calendar_get_instance_private (self);

    if (!default_abbreviated_dayname[0])
    {
        for (i = 0; i < 7; ++i)
        {
            tmp_time = (i + 4) * SEC_IN_A_DAY;
            strftime (buffer, sizeof (buffer), "%a", localtime (&tmp_time));
            default_abbreviated_dayname[i] = g_locale_to_utf8 (buffer, -1, NULL, NULL, NULL);
        }
    }
    if (!default_monthname[0])
    {
        for (i = 0; i < 12; ++i)
        {
            tmp_time = i * 32 * SEC_IN_A_DAY;
            strftime (buffer, sizeof (buffer), "%B", localtime (&tmp_time));
            default_monthname[i] = g_locale_to_utf8 (buffer, -1, NULL, NULL, NULL);
        }
    }

    secs = time (NULL);
    tm = localtime (&secs);

    priv->selected_year = tm->tm_year + 1900;
    priv->selected_month = tm->tm_mon + 1;
    priv->selected_day_of_month = tm->tm_mday;

    priv->focus_unit = -1;
    priv->year_before = 1;

    priv->week_start = 0;

    priv->show_week_num = FALSE;

    //calendar_compute_days (self);

    priv->view_type = VIEWTYPE_DAY;

    self->actor = st_box_layout_new ();
    st_widget_set_style_class_name (self->actor, "calendar");
    st_box_layout_set_vertical (ST_BOX_LAYOUT (self->actor), TRUE);

    priv->top_box = st_box_layout_new ();
    clutter_actor_set_x_align (CLUTTER_ACTOR (priv->top_box), CLUTTER_ACTOR_ALIGN_START);
    st_widget_set_style_class_name (priv->top_box, "calendar-top");
    st_box_layout_set_vertical (ST_BOX_LAYOUT (priv->top_box), FALSE);

    priv->main_box = st_box_layout_new ();
    st_widget_set_style_class_name (priv->main_box, "calendar-main-container");
    st_box_layout_set_vertical (ST_BOX_LAYOUT (priv->main_box), TRUE);
    clutter_actor_set_x_align (CLUTTER_ACTOR (priv->main_box), CLUTTER_ACTOR_ALIGN_CENTER);

    priv->swipe_box[VIEWTYPE_DAY] = NULL;
    priv->swipe_box[VIEWTYPE_MONTH] = NULL;
    priv->swipe_box[VIEWTYPE_YEAR] = NULL;

    _add_actor (self->actor, priv->top_box);
    _add_actor (self->actor, priv->main_box);
    clutter_actor_set_reactive (CLUTTER_ACTOR (self->actor), TRUE);

    g_object_set (priv->top_box,
                  "x-expand", TRUE,
                  "x-align", CLUTTER_ACTOR_ALIGN_FILL,
                  "y-align", CLUTTER_ACTOR_ALIGN_CENTER,
                  NULL);

    build_header (self);
    update_day_view (self, FALSE, 0);
    // g_print("%s --> end!\n", __func__);
}

/**
 * applet_calendar_new:
 *
 * 新建日历
 *
 * Returns: 新建日历对象
 */
AppletCalendar *
applet_calendar_new (void)
{
    return g_object_new (APPLET_TYPE_CALENDAR, NULL);
}

static void
calendar_compute_days (AppletCalendar *calendar)
{
    // g_print("%s --> enter!\n", __func__);
    AppletCalendarPrivate *priv = calendar->priv;
    gint month = 1;
    gint year = 1970;
    gint ndays_in_month = 31;
    gint ndays_in_prevmonth = 31;
    gint first_day = 4;
    gint row = 0;
    gint col = 0;
    gint day = 1;

    year = priv->selected_year;
    month = priv->selected_month;

    ndays_in_month = month_length[leap (year)][month];

    first_day = day_of_week (year, month, 1);
    first_day = (first_day + 7 - priv->week_start) % 7;
    if (0 == first_day)
    {
        first_day = 7;
    }

    if (1 < month)
    {
        ndays_in_prevmonth = month_length[leap (year)][month - 1];
    }
    else
    {
        ndays_in_prevmonth = month_length[leap (year - 1)][12];
    }

    day = ndays_in_prevmonth - first_day + 1;
    for (col = 0; col < first_day; ++col)
    {
        priv->day[0][col] = day;
        priv->lunar_day[0][col] = cdos_sun_to_lunar_festival (month == 1 ? year - 1 : year,
                                                              month == 1 ? 12 : month - 1,
                                                              day);
        priv->day_month[0][col] = MONTH_PREV;
        ++day;
    }

    row = first_day / 7;
    col = first_day % 7;
    for (day = 1; day <= ndays_in_month; ++day)
    {
        priv->day[row][col] = day;
        priv->lunar_day[row][col] = cdos_sun_to_lunar_festival (year, month, day);
        priv->day_month[row][col] = MONTH_CURRENT;

        ++col;
        if (7 == col)
        {
            ++row;
            col = 0;
        }
    }

    day = 1;
    for (; row <= 5; ++row)
    {
        for (; col <= 6; ++col)
        {
            priv->day[row][col] = day;
            priv->lunar_day[row][col] = cdos_sun_to_lunar_festival (month == 12 ? year + 1 : year,
                                                                    month == 12 ? 1 : month + 1,
                                                                    day);
            priv->day_month[row][col] = MONTH_NEXT;
            ++day;
        }
        col = 0;
    }
    // g_print("%s --> end!\n", __func__);
}

void
applet_calendar_update (AppletCalendar *self)
{
    g_return_if_fail (NULL != self && APPLET_IS_CALENDAR (self));
    GDateTime *newDate = g_date_time_new_now_local ();
    set_date (self, newDate, FALSE, 0);
    g_date_time_unref (newDate);
}

static void
flags_on_days_clicked (StButton *button, int clicked, gpointer user_data)
{
    AppletCalendarPrivate *priv = APPLET_CALENDAR (user_data)->priv;
    //priv->view_type = VIEWTYPE_DAY;
    //on_date_label_clicked (NULL, clicked, user_data);
}

static void
flags_on_months_clicked (StButton *button, int clicked, gpointer user_data)
{
    AppletCalendarPrivate *priv = APPLET_CALENDAR (user_data)->priv;
    //priv->view_type = VIEWTYPE_MONTH;
    //on_date_label_clicked (NULL, clicked, user_data);
}

static void
flags_on_years_clicked (StButton *button, int clicked, gpointer user_data)
{
    AppletCalendarPrivate *priv = APPLET_CALENDAR (user_data)->priv;
    //priv->view_type = VIEWTYPE_YEAR;
}

static void
build_header (AppletCalendar *calendar)
{
    // g_print("%s --> enter!\n", __func__);
    AppletCalendarPrivate *priv = calendar->priv;
    StWidget *top_boxArr = st_box_layout_new ();

    st_box_layout_set_vertical (ST_BOX_LAYOUT (top_boxArr), FALSE);

    clutter_actor_destroy_all_children (CLUTTER_ACTOR (priv->top_box));

    priv->view_type_switch = st_button_new_with_label ("");
    st_widget_set_style_class_name (priv->view_type_switch, "calendar-month-label");
    _add_actor (priv->top_box, priv->view_type_switch);
    g_signal_connect (priv->view_type_switch, "clicked", G_CALLBACK (on_date_label_clicked), calendar);
    g_object_set (priv->view_type_switch,
                  "x-expand", TRUE,
                  "x-align", CLUTTER_ACTOR_ALIGN_START,
                  "y-align", CLUTTER_ACTOR_ALIGN_CENTER,
                  NULL);

    _add_actor (priv->top_box, top_boxArr);
    g_object_set (top_boxArr,
                  "x-expand", TRUE,
                  "x-align", CLUTTER_ACTOR_ALIGN_START,
                  "y-align", CLUTTER_ACTOR_ALIGN_CENTER,
                  NULL);

    StWidget *back = st_button_new ();
    st_widget_set_style_class_name (ST_WIDGET (back), "calendar-change-month-back");
    _add_actor (top_boxArr, back);
    g_signal_connect (back, "clicked", G_CALLBACK (on_prev_button_clicked), calendar);
    g_object_set (back,
                  "x-expand", FALSE,
                  "x-align", CLUTTER_ACTOR_ALIGN_END,
                  "y-align", CLUTTER_ACTOR_ALIGN_CENTER,
                  NULL);

    StWidget *forward = st_button_new ();
    st_widget_set_style_class_name (ST_WIDGET (forward), "calendar-change-month-forward");
    _add_actor (top_boxArr, forward);
    g_signal_connect (forward, "clicked", G_CALLBACK (on_next_button_clicked), calendar);
    g_object_set (forward,
                  "x-expand", FALSE,
                  "x-align", CLUTTER_ACTOR_ALIGN_END,
                  "y-align", CLUTTER_ACTOR_ALIGN_CENTER,
                  NULL);

    //暂时先不添加
    /* StWidget *flags = st_box_layout_new ();
    st_box_layout_set_vertical (ST_BOX_LAYOUT (flags), FALSE);
    _add_actor (priv->top_box, flags);
    g_object_set (flags,
                  "x-expand", TRUE,
                  "x-align", CLUTTER_ACTOR_ALIGN_END,
                  "y-align", CLUTTER_ACTOR_ALIGN_CENTER,
                  NULL);

    StWidget *days = st_button_new_with_label ("Days");
    st_widget_set_style_class_name (ST_WIDGET (days), "calendar-flag-days");
    _add_actor (flags, days);
    g_signal_connect (days, "clicked", G_CALLBACK (flags_on_days_clicked), calendar);
    g_object_set (days,
                  "x-expand", FALSE,
                  "x-align", CLUTTER_ACTOR_ALIGN_END,
                  "y-align", CLUTTER_ACTOR_ALIGN_CENTER,
                  NULL);
    
    StWidget *months = st_button_new_with_label ("Months");
    st_widget_set_style_class_name (ST_WIDGET (months), "calendar-flag-months");
    _add_actor (flags, months);
    g_signal_connect (months, "clicked", G_CALLBACK (flags_on_months_clicked), calendar);
    g_object_set (months,
                  "x-expand", FALSE,
                  "x-align", CLUTTER_ACTOR_ALIGN_END,
                  "y-align", CLUTTER_ACTOR_ALIGN_CENTER,
                  NULL);

    StWidget *years = st_button_new_with_label ("Years");
    st_widget_set_style_class_name (ST_WIDGET (years), "calendar-flag-years");
    _add_actor (flags, years);
    g_signal_connect (years, "clicked", G_CALLBACK (flags_on_years_clicked), calendar);
    g_object_set (years,
                  "x-expand", FALSE,
                  "x-align", CLUTTER_ACTOR_ALIGN_END,
                  "y-align", CLUTTER_ACTOR_ALIGN_CENTER,
                  NULL);
    */
}

static void
applyDateBrowseAction (AppletCalendar *calendar, int yearChange, int monthChange)
{
    AppletCalendarPrivate *priv = calendar->priv;
    /*
    StWidget *swipebox = NULL;
    switch (priv->view_type)
    {
        case VIEWTYPE_DAY:
            swipebox = priv->swipe_box[0];
            break;
        case VIEWTYPE_MONTH:
            swipebox = priv->swipe_box[1];
            break;
        case VIEWTYPE_YEAR:
            swipebox = priv->swipe_box[2];
            break;
        default:
            g_warn_if_reached ();
            return;
    }
    */

    gint oldYear = priv->selected_year;
    gint oldMonth = priv->selected_month;
    gint oldDayOfMonth = priv->selected_day_of_month;
    gint newMonth = oldMonth + monthChange;
    if (12 < newMonth)
    {
        yearChange = yearChange + 1;
        newMonth = 1;
    }
    else if (1 > newMonth)
    {
        yearChange = yearChange - 1;
        newMonth = 12;
    }

    gint newYear = oldYear + yearChange;
    gint newDayOfMonth = oldDayOfMonth;
    int newMonthLength = month_length[leap (newYear)][newMonth];
    if (newMonthLength < newDayOfMonth)
    {
        newDayOfMonth = newMonthLength;
    }

    GDateTime *oldDate = g_date_time_new_local (oldYear, oldMonth, oldDayOfMonth, 0, 0, 0);
    GDateTime *newDate = g_date_time_new_local (newYear, newMonth, newDayOfMonth, 0, 0, 0);

    set_date (calendar, newDate, FALSE, g_date_time_difference (oldDate, newDate));

    g_date_time_unref (newDate);
    g_date_time_unref (oldDate);
    // TODO: need to add tweener animations here.
}

static void
on_prev_button_clicked (StButton *button, int clickedBtn, gpointer user_data)
{
    AppletCalendarPrivate *priv = APPLET_CALENDAR (user_data)->priv;
    switch (priv->view_type)
    {
        case VIEWTYPE_DAY:
            applyDateBrowseAction (APPLET_CALENDAR (user_data), 0, -1);
            break;
        case VIEWTYPE_MONTH:
            applyDateBrowseAction (APPLET_CALENDAR (user_data), -1, 0);
            break;
        case VIEWTYPE_YEAR:
            applyDateBrowseAction (APPLET_CALENDAR (user_data), -1 * 10, 0);
            break;
        default:
            break;
    }
}

static void
on_next_button_clicked (StButton *button, int clickedBtn, gpointer user_data)
{
    AppletCalendarPrivate *priv = APPLET_CALENDAR (user_data)->priv;
    switch (priv->view_type)
    {
        case VIEWTYPE_DAY:
            applyDateBrowseAction (APPLET_CALENDAR (user_data), 0, +1);
            break;
        case VIEWTYPE_MONTH:
            applyDateBrowseAction (APPLET_CALENDAR (user_data), +1, 0);
            break;
        case VIEWTYPE_YEAR:
            applyDateBrowseAction (APPLET_CALENDAR (user_data), +1 * 10, 0);
            break;
        default:
            break;
    }
}

static void
on_date_label_clicked (StButton *button, int clickedBtn, gpointer user_data)
{
    AppletCalendarPrivate *priv = APPLET_CALENDAR (user_data)->priv;
    switch (priv->view_type)
    {
        case VIEWTYPE_DAY:
        {
            gfloat x, y;
            clutter_actor_get_position ((ClutterActor *)priv->swipe_box[VIEWTYPE_DAY], &x, &y);
            // g_print("%s:%d --> (x, y) is (%.03f, %.03f)!\n", __func__, __LINE__, x, y);
        }
            clutter_actor_save_easing_state ((ClutterActor *)priv->swipe_box[VIEWTYPE_DAY]);
            clutter_actor_set_scale ((ClutterActor *)priv->swipe_box[VIEWTYPE_DAY], 0.2, 0.2);
            clutter_actor_set_opacity ((ClutterActor *)priv->swipe_box[VIEWTYPE_DAY], 0);
            clutter_actor_set_easing_duration ((ClutterActor *)priv->swipe_box[VIEWTYPE_DAY], ANIMATION_DURATION_TIME);
            clutter_actor_set_easing_mode ((ClutterActor *)priv->swipe_box[VIEWTYPE_DAY], ANIMATION_MODE_OUT);
            clutter_actor_restore_easing_state ((ClutterActor *)priv->swipe_box[VIEWTYPE_DAY]);

            update_month_view (APPLET_CALENDAR (user_data), TRUE, 0);
            clutter_actor_set_scale ((ClutterActor *)priv->swipe_box[VIEWTYPE_MONTH], 2, 2);
            clutter_actor_set_opacity ((ClutterActor *)priv->swipe_box[VIEWTYPE_MONTH], 0);
            clutter_actor_save_easing_state ((ClutterActor *)priv->swipe_box[VIEWTYPE_MONTH]);
            clutter_actor_set_scale ((ClutterActor *)priv->swipe_box[VIEWTYPE_MONTH], 1, 1);
            clutter_actor_set_opacity ((ClutterActor *)priv->swipe_box[VIEWTYPE_MONTH], 255);
            clutter_actor_set_easing_duration ((ClutterActor *)priv->swipe_box[VIEWTYPE_MONTH], ANIMATION_DURATION_TIME);
            clutter_actor_set_easing_delay ((ClutterActor *)priv->swipe_box[VIEWTYPE_MONTH], ANIMATION_DELAY_TIME);
            clutter_actor_set_easing_mode ((ClutterActor *)priv->swipe_box[VIEWTYPE_MONTH], ANIMATION_MODE_OUT);
            clutter_actor_restore_easing_state ((ClutterActor *)priv->swipe_box[VIEWTYPE_MONTH]);
            break;
        case VIEWTYPE_MONTH:
        {
            gfloat x, y;
            clutter_actor_get_position ((ClutterActor *)priv->swipe_box[VIEWTYPE_MONTH], &x, &y);
            // g_print("%s:%d --> (x, y) is (%.03f, %.03f)!\n", __func__, __LINE__, x, y);
        }
            clutter_actor_save_easing_state ((ClutterActor *)priv->swipe_box[VIEWTYPE_MONTH]);
            clutter_actor_set_scale ((ClutterActor *)priv->swipe_box[VIEWTYPE_MONTH], 0.2, 0.2);
            clutter_actor_set_opacity ((ClutterActor *)priv->swipe_box[VIEWTYPE_MONTH], 0);
            clutter_actor_set_easing_duration ((ClutterActor *)priv->swipe_box[VIEWTYPE_MONTH], ANIMATION_DURATION_TIME);
            clutter_actor_set_easing_mode ((ClutterActor *)priv->swipe_box[VIEWTYPE_MONTH], ANIMATION_MODE_OUT);
            clutter_actor_restore_easing_state ((ClutterActor *)priv->swipe_box[VIEWTYPE_MONTH]);

            update_year_view (APPLET_CALENDAR (user_data), TRUE, 0);
            clutter_actor_set_scale ((ClutterActor *)priv->swipe_box[VIEWTYPE_YEAR], 2, 2);
            clutter_actor_set_opacity ((ClutterActor *)priv->swipe_box[VIEWTYPE_YEAR], 0);
            clutter_actor_save_easing_state ((ClutterActor *)priv->swipe_box[VIEWTYPE_YEAR]);
            clutter_actor_set_scale ((ClutterActor *)priv->swipe_box[VIEWTYPE_YEAR], 1, 1);
            clutter_actor_set_opacity ((ClutterActor *)priv->swipe_box[VIEWTYPE_YEAR], 255);
            clutter_actor_set_easing_duration ((ClutterActor *)priv->swipe_box[VIEWTYPE_YEAR], ANIMATION_DURATION_TIME);
            clutter_actor_set_easing_delay ((ClutterActor *)priv->swipe_box[VIEWTYPE_YEAR], ANIMATION_DELAY_TIME);
            clutter_actor_set_easing_mode ((ClutterActor *)priv->swipe_box[VIEWTYPE_YEAR], ANIMATION_MODE_OUT);
            clutter_actor_restore_easing_state ((ClutterActor *)priv->swipe_box[VIEWTYPE_YEAR]);

            set_sensitive (ST_WIDGET (priv->view_type_switch), FALSE);
            break;
        default:
            // default: do nothing!
            return;
    }
}

static gboolean
isTheSameYear (GDateTime *date1, GDateTime *date2)
{
    gint year1 = g_date_time_get_year (date1);
    gint year2 = g_date_time_get_year (date2);
    return (year1 == year2);
}

static gboolean
isTheSameMonth (GDateTime *date1, GDateTime *date2)
{
    gint month1 = g_date_time_get_month (date1);
    gint month2 = g_date_time_get_month (date2);
    return (isTheSameYear (date1, date2) && (month1 == month2));
}

static gboolean
isTheSameDate (GDateTime *date1, GDateTime *date2)
{
    gint day1 = g_date_time_get_day_of_month (date1);
    gint day2 = g_date_time_get_day_of_month (date2);
    return (isTheSameMonth (date1, date2) && (day1 == day2));
}

static void
set_date (AppletCalendar *calendar, GDateTime *date, gboolean isForceLoad, GTimeSpan changed)
{
    AppletCalendarPrivate *priv = calendar->priv;
    GDateTime *selectedDate =
        g_date_time_new_local (priv->selected_year, priv->selected_month, priv->selected_day_of_month, 1, 1, 1);
    if (!isTheSameDate (date, selectedDate))
    {
        g_date_time_get_ymd (date, &(priv->selected_year), &priv->selected_month, &(priv->selected_day_of_month));
        switch (priv->view_type)
        {
            case VIEWTYPE_DAY:
                update_day_view (calendar, isForceLoad, changed);
                break;
            case VIEWTYPE_MONTH:
                update_month_view (calendar, isForceLoad, changed);
                break;
            case VIEWTYPE_YEAR:
                update_year_view (calendar, isForceLoad, changed);
                break;
            default:
                break;
        }
    }
    else
    {
        if (isForceLoad)
        {
            update_day_view (calendar, isForceLoad, changed);
        }
    }
    g_date_time_unref (selectedDate);
}

static void
update_day_view (AppletCalendar *calendar, gboolean isForceLoad, GTimeSpan changed)
{
    // g_print("%s --> enter!\n", __func__);
    calendar_compute_days (calendar);
    AppletCalendarPrivate *priv = calendar->priv;
    StWidget *swipebox;
    gint rowCount = 7;
    gint offsetCols = (priv->show_week_num ? 1 : 0);
    gint colCount = 7 + offsetCols;
    int row = 0;
    int col = 0;
    char buffer[32];
    int selected_year = priv->selected_year;
    int selected_month = priv->selected_month;
    GDateTime *nowDate = g_date_time_new_now_local ();

    priv->view_type = VIEWTYPE_DAY;

    clutter_actor_destroy_all_children (CLUTTER_ACTOR (priv->main_box));

    priv->swipe_box[priv->view_type] = st_box_layout_new ();
    st_box_layout_set_vertical (ST_BOX_LAYOUT (priv->swipe_box[priv->view_type]), FALSE);
    swipebox = priv->swipe_box[priv->view_type];

    _add_actor (priv->main_box, priv->swipe_box[priv->view_type]);
    clutter_actor_set_name (CLUTTER_ACTOR (swipebox), "calendar-table");

    GDateTime *selectedDate =
        g_date_time_new_local (priv->selected_year, priv->selected_month, priv->selected_day_of_month, 1, 1, 1);
    gchar *dateLabelText = g_date_time_format (selectedDate, "%Y %b");
    st_button_set_label (ST_BUTTON (priv->view_type_switch), dateLabelText);
    g_date_time_unref (selectedDate);
    g_free (dateLabelText);

    for (col = 0; col < colCount; ++col)
    {
        StWidget *daylayout = st_box_layout_new ();
        st_box_layout_set_vertical (ST_BOX_LAYOUT (daylayout), TRUE);
        for (row = 0; row < rowCount; ++row)
        {
            if (priv->show_week_num && (0 == col))
            {
                if (0 == row)
                {
                    StWidget *weeknumbtn = st_button_new_with_label ("");
                    st_widget_add_style_class_name (weeknumbtn, "calendar-day-base");
                    st_widget_add_style_class_name (weeknumbtn, "calendar-day");
                    st_widget_add_style_class_name (weeknumbtn, "calendar-week-number");
                    _add_actor (daylayout, weeknumbtn);
                }
                else
                {
                    int day = priv->day[row - 1][6];
                    int month = selected_month;
                    int year = selected_year;
                    switch (priv->day_month[row - 1][6])
                    {
                        case MONTH_PREV:
                            --month;
                            if (1 > month)
                            {
                                month = 12;
                                --year;
                            }
                            break;
                        case MONTH_CURRENT:
                            break;
                        case MONTH_NEXT:
                            ++month;
                            if (12 < month)
                            {
                                month = 1;
                                ++year;
                            }
                            break;
                        default:
                            g_assert (FALSE);
                            break;
                    }

                    int week = week_of_year (year, month, day);
                    g_snprintf (buffer, sizeof (buffer), "%02d", week);
                    StWidget *weeknumbtn = st_button_new_with_label (buffer);
                    st_widget_add_style_class_name (weeknumbtn, "calendar-day-base");
                    st_widget_add_style_class_name (weeknumbtn, "calendar-day");
                    st_widget_add_style_class_name (weeknumbtn, "calendar-week-number");
                    _add_actor (daylayout, weeknumbtn);
                } // end: if (0 == row)
            }
            else
            {
                if (0 == row)
                {
                    StWidget *weekTitleItem = st_button_new_with_label (
                        default_abbreviated_dayname[(col - offsetCols + priv->week_start) % 7]);
                    st_widget_add_style_class_name (weekTitleItem, "calendar-day-base");
                    st_widget_add_style_class_name (weekTitleItem, "calendar-day");
                    st_widget_add_style_class_name (weekTitleItem, "calendar-week-number");
                    _add_actor (daylayout, weekTitleItem);
                }
                else
                {
                    g_snprintf (buffer, sizeof (buffer), "%d", priv->day[row - 1][col - offsetCols]);
                    StWidget *daybtn =
                        applet_calendar_item_new_with_label_day_and_lunar (buffer,
                                priv->day[row - 1][col - offsetCols],
                                priv->lunar_day[row - 1][col - offsetCols]);

                    st_widget_add_style_class_name (ST_WIDGET (daybtn), "calendar-day-base");
                    st_widget_add_style_class_name (ST_WIDGET (daybtn), "calendar-day");
                    if (is_work_day (col - offsetCols + priv->week_start))
                    {
                        st_widget_add_style_class_name (ST_WIDGET (daybtn), "calendar-work-day");
                    }
                    else
                    {
                        st_widget_add_style_class_name (ST_WIDGET (daybtn), "calendar-nonwork-day");
                    }

                    switch (priv->day_month[row - 1][col - offsetCols])
                    {
                        case MONTH_CURRENT:
                        {
                            GDateTime *currentDate = g_date_time_new_local (
                                selected_year, selected_month, priv->day[row - 1][col - offsetCols], 0, 0, 0);
                            if (isTheSameDate (nowDate, currentDate))
                            {
                                st_widget_add_style_class_name (ST_WIDGET (daybtn), "calendar-today");
                            }
                            g_date_time_unref (currentDate);
                        }
                        break;
                        case MONTH_PREV:
                        case MONTH_NEXT:
                            st_widget_add_style_class_name (ST_WIDGET (daybtn), "calendar-other-month-day");
                            break;
                        default:
                            break;
                    }
                    _add_actor (daylayout, daybtn);
                }
            } // end: if (priv->show_week_num && (0 == col))
        } // end: for row
        _add_actor (swipebox, daylayout);
    } // end: for col
    g_date_time_unref (nowDate);
    // g_print("%s --> end!\n", __func__);
}

static void
update_month_view (AppletCalendar *calendar, gboolean isForceLoad, GTimeSpan changed)
{
    AppletCalendarPrivate *priv = calendar->priv;
    StWidget *swipebox = NULL;
    int row = 0;
    int col = 0;
    int rowCount = 3;
    int colCount = 4;
    GDateTime *nowDate = g_date_time_new_now_local ();
    int selected_year = priv->selected_year;

    priv->view_type = VIEWTYPE_MONTH;

    clutter_actor_destroy_all_children (CLUTTER_ACTOR (priv->main_box));

    priv->swipe_box[priv->view_type] = st_box_layout_new ();
    st_box_layout_set_vertical (ST_BOX_LAYOUT (priv->swipe_box[priv->view_type]), TRUE);
    swipebox = priv->swipe_box[priv->view_type];
    st_widget_set_style (swipebox, "padding-top:33px;padding-bottom:33px;");
    clutter_actor_insert_child_at_index (CLUTTER_ACTOR (priv->main_box), CLUTTER_ACTOR (priv->swipe_box[priv->view_type]), -1);
    clutter_actor_set_name (CLUTTER_ACTOR (swipebox), "calendar-table");
    GDateTime *selectedDate =
        g_date_time_new_local (priv->selected_year, priv->selected_month, priv->selected_day_of_month, 1, 1, 1);
    gchar *dateLabelText = g_date_time_format (selectedDate, "%Y");
    st_button_set_label (ST_BUTTON (priv->view_type_switch), dateLabelText);
    g_date_time_unref (selectedDate);
    g_free (dateLabelText);

    for (row = 0; row < rowCount; ++row)
    {
        StWidget *monthlayout = st_box_layout_new ();
        st_box_layout_set_vertical (ST_BOX_LAYOUT (monthlayout), FALSE);
        for (col = 0; col < colCount; ++col)
        {
            int index = row * colCount + col;
            StWidget *monthButton = applet_calendar_item_new_with_label_month (default_monthname[index], index + 1);
            g_signal_connect (monthButton, "button-release-event", G_CALLBACK (on_calendar_item_clicked), calendar);

            st_widget_add_style_class_name (ST_WIDGET (monthButton), "calendar-month-base");
            st_widget_add_style_class_name (ST_WIDGET (monthButton), "calendar-month");
            GDateTime *currentDate = g_date_time_new_local (selected_year, index + 1, 1, 0, 0, 0);
            if (isTheSameMonth (nowDate, currentDate))
            {
                st_widget_add_style_class_name (ST_WIDGET (monthButton), "calendar-current-month");
            }
            g_date_time_unref (currentDate);
            _add_actor (monthlayout, monthButton);
        }
        _add_actor (swipebox, monthlayout);
    }
}

static void
update_year_view (AppletCalendar *calendar, gboolean isForceLoad, GTimeSpan changed)
{
    AppletCalendarPrivate *priv = calendar->priv;
    StWidget *swipebox = NULL;
    int row = 0;
    int col = 0;
    int rowCount = 3;
    int colCount = 4;
    char buffer[32];
    GDateTime *nowDate = g_date_time_new_now_local ();
    int nowYear = g_date_time_get_year (nowDate);
    int selected_year = priv->selected_year;

    int beginYear = ((int)selected_year / 10) * 10 - 1;
    int endYear = ((int)selected_year / 10 + 1) * 10;
    int iterYear = beginYear;

    priv->view_type = VIEWTYPE_YEAR;

    clutter_actor_destroy_all_children (CLUTTER_ACTOR (priv->main_box));

    priv->swipe_box[priv->view_type] = st_box_layout_new ();
    st_box_layout_set_vertical (ST_BOX_LAYOUT (priv->swipe_box[priv->view_type]), TRUE);
    swipebox = priv->swipe_box[priv->view_type];
    st_widget_set_style (swipebox, "padding-top:33px;padding-bottom:33px;");

    clutter_actor_insert_child_at_index(CLUTTER_ACTOR (priv->main_box), CLUTTER_ACTOR (priv->swipe_box[priv->view_type]), -1);
    clutter_actor_set_name (CLUTTER_ACTOR (swipebox), "calendar-table");

    g_snprintf (buffer, sizeof (buffer), "%d-%d", beginYear, endYear);
    st_button_set_label (ST_BUTTON (priv->view_type_switch), buffer);

    for (row = 0; row < rowCount; ++row)
    {
        StWidget *yearLayout = st_box_layout_new ();
        st_box_layout_set_vertical (ST_BOX_LAYOUT (yearLayout), FALSE);
        for (col = 0; col < colCount; ++col)
        {
            int index = row * colCount + col;
            iterYear = beginYear + index;
            g_snprintf (buffer, sizeof (buffer), "%d", iterYear);
            StWidget *yearButton = applet_calendar_item_new_with_label_year (buffer, iterYear);
            g_signal_connect (yearButton, "button-release-event", G_CALLBACK (on_calendar_item_clicked), calendar);

            st_widget_add_style_class_name (ST_WIDGET (yearButton), "calendar-year-base");
            st_widget_add_style_class_name (ST_WIDGET (yearButton), "calendar-year");
            if (nowYear == iterYear)
            {
                st_widget_add_style_class_name (ST_WIDGET (yearButton), "calendar-current-year");
            }
            else if (((int)(iterYear / 10)) != ((int)(selected_year / 10)))
            {
                st_widget_add_style_class_name (ST_WIDGET (yearButton), "calendar-other-x-year");
            }
            _add_actor (yearLayout, yearButton);
        }
        _add_actor (swipebox, yearLayout);
    }
}

static gboolean
on_calendar_item_clicked (StButton *button, ClutterEvent *event, gpointer user_data)
{
    int old_month, old_day_of_month;
    AppletCalendarPrivate *priv = APPLET_CALENDAR (user_data)->priv;
    int item_opt = applet_calendar_item_get_option (APPLET_CALENDAR_ITEM (button));
    int item_data = applet_calendar_item_get_item_data (APPLET_CALENDAR_ITEM (button));

    int new_year = priv->selected_year;
    int new_month = old_month = priv->selected_month;
    int new_day_of_month = old_day_of_month = priv->selected_day_of_month;

    set_sensitive (ST_WIDGET (priv->view_type_switch), TRUE);

    switch (priv->view_type)
    {
        case VIEWTYPE_DAY:
            break;
        case VIEWTYPE_MONTH:
            g_return_val_if_fail (OPT_MONTH == item_opt, TRUE);
            clutter_actor_save_easing_state ((ClutterActor *)priv->swipe_box[VIEWTYPE_MONTH]);
            clutter_actor_set_scale ((ClutterActor *)priv->swipe_box[VIEWTYPE_MONTH], 2, 2);
            clutter_actor_set_opacity ((ClutterActor *)priv->swipe_box[VIEWTYPE_MONTH], 0);
            clutter_actor_set_easing_duration ((ClutterActor *)priv->swipe_box[VIEWTYPE_MONTH], ANIMATION_DURATION_TIME);
            clutter_actor_set_easing_mode ((ClutterActor *)priv->swipe_box[VIEWTYPE_MONTH], ANIMATION_MODE_OUT);
            clutter_actor_restore_easing_state ((ClutterActor *)priv->swipe_box[VIEWTYPE_MONTH]);

            new_month = item_data;
            update_day_view (APPLET_CALENDAR (user_data), FALSE, 0);
            break;
        case VIEWTYPE_YEAR:
            g_return_val_if_fail (OPT_YEAR == item_opt, TRUE);
            clutter_actor_save_easing_state ((ClutterActor *)priv->swipe_box[VIEWTYPE_YEAR]);
            clutter_actor_set_scale ((ClutterActor *)priv->swipe_box[VIEWTYPE_YEAR], 2, 2);
            clutter_actor_set_opacity ((ClutterActor *)priv->swipe_box[VIEWTYPE_YEAR], 0);
            clutter_actor_set_easing_duration ((ClutterActor *)priv->swipe_box[VIEWTYPE_YEAR], ANIMATION_DURATION_TIME);
            clutter_actor_set_easing_mode ((ClutterActor *)priv->swipe_box[VIEWTYPE_YEAR], ANIMATION_MODE_OUT);
            clutter_actor_restore_easing_state ((ClutterActor *)priv->swipe_box[VIEWTYPE_YEAR]);

            new_year = item_data;
            update_month_view (APPLET_CALENDAR (user_data), FALSE, 0);
            break;
        default:
            break;
    }

    int new_month_length = month_length[leap (new_year)][new_month];
    if (new_month_length < new_day_of_month)
    {
        new_day_of_month = new_month_length;
    }

    GDateTime *new_date = g_date_time_new_local (new_year, new_month, new_day_of_month, 0, 0, 0);
    set_date (APPLET_CALENDAR (user_data), new_date, FALSE, 0);
    g_date_time_unref (new_date);

    clutter_actor_set_scale ((ClutterActor *)priv->swipe_box[priv->view_type], 0.2, 0.2);
    clutter_actor_set_opacity ((ClutterActor *)priv->swipe_box[priv->view_type], 0);
    clutter_actor_save_easing_state ((ClutterActor *)priv->swipe_box[priv->view_type]);
    clutter_actor_set_scale ((ClutterActor *)priv->swipe_box[priv->view_type], 1, 1);
    clutter_actor_set_opacity ((ClutterActor *)priv->swipe_box[priv->view_type], 255);
    clutter_actor_set_easing_duration ((ClutterActor *)priv->swipe_box[priv->view_type], ANIMATION_DURATION_TIME);
    //    clutter_actor_set_easing_delay(priv->swipe_box[priv->view_type], ANIMATION_DELAY_TIME);
    clutter_actor_set_easing_mode ((ClutterActor *)priv->swipe_box[priv->view_type], ANIMATION_MODE_OUT);
    clutter_actor_restore_easing_state ((ClutterActor *)priv->swipe_box[priv->view_type]);
    return TRUE;
}
