#include "config.h"
#include "common.h"
#include <clutter/clutter.h>
#include <gio/gio.h>
#include <meta/meta-workspace-manager.h>
#include "applet-feedback.h"

#define FEEDBACK_CONFIG_FILENAME "/etc/desktop/feedback"
#define FEEDBACK_CONFIG_NFS_GROUP     "Desktop Feedback"
#define FEEDBACK_COMPANY_NAME "CompanyName"
#define FEEDBACK_TEL "Telephone"
#define FEEDBACK_MAIL "Mail"
#define FEEDBACK_SERVICE_ACCOUNT_URI "ServiceAccountUri"

struct _CdosFeedbackPrivate
{
    guint orientation;
    CdosPopupMenu *menu;
    StWidget *feedback_box;

    GKeyFile *feedback_config_file;
    gchar *feedback_config_filename;
    gchar *feedback_company_name;
    gchar *feedback_tel;
    gchar *feedback_mail;
    gchar *feedback_image_file;
}; 

G_DEFINE_TYPE_WITH_PRIVATE (CdosFeedback, cdos_feedback, CDOS_TYPE_ICONAPPLET);


static gboolean
should_enable_feedback (void)
{
    GSettings *settings = NULL;
    gboolean value = TRUE;

    settings = g_settings_new ("org.cdos");
    value = g_settings_get_boolean (settings, "enable-feedback");

    return value;
}

static void
free_strings (CdosFeedback *self)
{
    CdosFeedbackPrivate *priv = self->priv;
    g_free (priv->feedback_company_name);
    g_free (priv->feedback_tel);
    g_free (priv->feedback_mail);
    g_free (priv->feedback_image_file);
}

static void
cdos_feedback_dispose (GObject *gobject)
{
    CdosFeedback *self = CDOS_FEEDBACK (gobject);
    CdosFeedbackPrivate *priv = self->priv;

    if (priv->feedback_config_file)
    {
        g_key_file_unref (priv->feedback_config_file);
        priv->feedback_config_file = NULL;
    }

    G_OBJECT_CLASS (cdos_feedback_parent_class)->dispose (gobject);
}

static void
cdos_feedback_finalize (GObject *gobject)
{
    CdosFeedback *self = CDOS_FEEDBACK (gobject);
    CdosFeedbackPrivate *priv = self->priv;

    free_strings (self);
    g_free (priv->feedback_config_filename);

    G_OBJECT_CLASS (cdos_feedback_parent_class)->finalize (gobject);
}

static void
cdos_feedback_new_for_config_file (CdosFeedback *self)
{
    GError *error = NULL;
    CdosFeedbackPrivate *priv = self->priv;

    priv->feedback_config_file = g_key_file_new ();
    priv->feedback_config_filename = g_strdup (FEEDBACK_CONFIG_FILENAME);
    if (priv->feedback_config_file == NULL || priv->feedback_config_filename == NULL)
        return;

    if (!g_key_file_load_from_file (priv->feedback_config_file, 
                                    priv->feedback_config_filename, 
                                    G_KEY_FILE_KEEP_COMMENTS, 
                                    &error))
    {
        g_warning ("Load background config file failed: %s", error->message);
        g_error_free (error);
        g_object_unref (priv->feedback_config_file);
        priv->feedback_config_file = NULL;
        return;
    }
}

static void
cdos_feedback_get_info_from_config_file (CdosFeedback *self,const gchar *group)
{
    CdosFeedbackPrivate *priv = self->priv;
    if (priv->feedback_config_file == NULL)
        return;

    free_strings (self);
    priv->feedback_company_name = g_key_file_get_locale_string (priv->feedback_config_file, 
                                                                        group, 
                                                                        FEEDBACK_COMPANY_NAME,
                                                                        NULL,
                                                                        NULL);
    priv->feedback_tel = g_key_file_get_string (priv->feedback_config_file, 
                                                group, 
                                                FEEDBACK_TEL, 
                                                NULL);
    priv->feedback_mail = g_key_file_get_string (priv->feedback_config_file, 
                                                    group, 
                                                    FEEDBACK_MAIL, 
                                                    NULL);
    priv->feedback_image_file = g_key_file_get_string (priv->feedback_config_file, 
                                                            group, 
                                                            FEEDBACK_SERVICE_ACCOUNT_URI, 
                                                            NULL);
}

static void
cdos_feedback_create_title_box (CdosFeedback *self)
{
    CdosFeedbackPrivate *priv = self->priv;

    cdos_applet_set_popup_menu_active ((CdosApplet *)self, TRUE);
    CdosIconApplet *iconapplet = (CdosIconApplet *)self;
    cdos_iconapplet_set_icon_name ((CdosIconApplet *)iconapplet, "found-feedback");
    cdos_iconapplet_set_icon_size ((CdosIconApplet *)iconapplet, 16);
    cdos_applet_set_applet_tooltip ((CdosApplet *)iconapplet, _("Contact Us"));
    priv->menu = cdos_applet_get_popup_menu (CDOS_APPLET (self));

    // menu_box 
    StBoxLayout *menu_box = cdos_popup_menu_base_get_box (CDOS_POPUP_MENU_BASE (priv->menu));
    st_box_layout_set_vertical (ST_BOX_LAYOUT (menu_box), TRUE);

    // 标题 - item
    StWidget *title_box = st_box_layout_new ();
    st_box_layout_set_vertical (ST_BOX_LAYOUT (title_box), FALSE);
    st_widget_set_style_class_name (ST_WIDGET (title_box), "feedback-title-box");
    _add_actor (menu_box, title_box);

    StWidget *title_label = st_label_new (_("Contact Us"));
    st_widget_set_style_class_name (ST_WIDGET (title_label), "feedback-title-font");
    clutter_actor_set_x_expand (CLUTTER_ACTOR (title_label), TRUE);
    clutter_actor_set_x_align (CLUTTER_ACTOR (title_label), ST_ALIGN_MIDDLE);
    _add_actor (title_box, title_label);

    // 方德 + 厂商的反馈box 都放在这个 feedback_box 里面。
    StWidget *feedback_box = st_box_layout_new ();
    st_box_layout_set_vertical (ST_BOX_LAYOUT (feedback_box), FALSE);
    _add_actor (menu_box, feedback_box);

    priv->feedback_box = feedback_box;
}

static void
cdos_feedback_create_contant_box (CdosFeedback *self, const gchar *group)
{
    GFile *file;
    GIcon *gicon;
    CdosFeedbackPrivate *priv = self->priv;

    StWidget *nfs_box = st_box_layout_new ();
    st_box_layout_set_vertical (ST_BOX_LAYOUT (nfs_box), TRUE);
    _add_actor (priv->feedback_box, nfs_box);

    //  联系我们 - item
    StWidget *contact_item_box = st_box_layout_new ();
    st_box_layout_set_vertical (ST_BOX_LAYOUT (contact_item_box), FALSE);
    st_widget_set_style_class_name (ST_WIDGET (contact_item_box), "feedback-context-box");
    _add_actor (nfs_box, contact_item_box);

    StWidget *contact_item_label;
    if (g_strcmp0 (group,FEEDBACK_CONFIG_NFS_GROUP) == 0)
        contact_item_label = st_label_new (_("If you need os technical support and service, please contact us:"));
    else 
        contact_item_label = st_label_new (_("If you need hardware technical support and services, please contact us:"));
    clutter_actor_set_x_expand (CLUTTER_ACTOR (contact_item_label), TRUE);
    clutter_actor_set_x_align (CLUTTER_ACTOR (contact_item_label), ST_ALIGN_START);
    _add_actor (contact_item_box, contact_item_label);

    // 公司 - item
    StWidget *ltd_item_box = st_box_layout_new ();
    st_box_layout_set_vertical (ST_BOX_LAYOUT (ltd_item_box), FALSE);
    st_widget_set_style_class_name (ST_WIDGET (ltd_item_box), "feedback-context-box");
    _add_actor (nfs_box, ltd_item_box);

    StWidget *ltd_item_label = st_label_new (priv->feedback_company_name);
    clutter_actor_set_x_expand (CLUTTER_ACTOR (ltd_item_label), TRUE);
    clutter_actor_set_x_align (CLUTTER_ACTOR (ltd_item_label), ST_ALIGN_START);
    _add_actor (ltd_item_box, ltd_item_label);

    // 联系电话 - item
    StWidget *tel_item_box = st_box_layout_new ();
    st_widget_set_style_class_name (ST_WIDGET (tel_item_box), "feedback-context-box");
    st_box_layout_set_vertical (ST_BOX_LAYOUT (tel_item_box), FALSE);
    _add_actor (nfs_box, tel_item_box);

    StIcon *tel_item_icon = ST_ICON (st_icon_new ());
    st_icon_set_icon_size (ST_ICON (tel_item_icon), 16);
    clutter_actor_set_x_expand (CLUTTER_ACTOR (tel_item_icon), FALSE);
    clutter_actor_set_x_align (CLUTTER_ACTOR (tel_item_icon), ST_ALIGN_START);
    st_icon_set_icon_name (ST_ICON (tel_item_icon), "found-feedback-tel");
    _add_actor (tel_item_box, tel_item_icon);

    StWidget *tel_item_label = st_label_new (g_strdup_printf ("%s%s", _("Tel:"), priv->feedback_tel));
    st_widget_set_style_class_name (ST_WIDGET (tel_item_label), "feedback-context-label");
    clutter_actor_set_x_expand (CLUTTER_ACTOR (tel_item_label), TRUE);
    clutter_actor_set_x_align (CLUTTER_ACTOR (tel_item_label), ST_ALIGN_START);
    _add_actor (tel_item_box, tel_item_label);

    // 联系邮件 - item
    StWidget *mail_item_box = st_box_layout_new ();
    st_box_layout_set_vertical (ST_BOX_LAYOUT (mail_item_box), FALSE);
    st_widget_set_style_class_name (ST_WIDGET (mail_item_box), "feedback-context-box");
    _add_actor (nfs_box, mail_item_box);

    StIcon *mail_item_icon = ST_ICON (st_icon_new ());
    st_icon_set_icon_size (ST_ICON (mail_item_icon), 16);
    clutter_actor_set_x_expand (CLUTTER_ACTOR (mail_item_icon), FALSE);
    clutter_actor_set_x_align (CLUTTER_ACTOR (mail_item_icon), ST_ALIGN_START);
    st_icon_set_icon_name (ST_ICON (mail_item_icon), "found-feedback-mail");
    _add_actor (mail_item_box, mail_item_icon);

    StWidget *mail_item_label = st_label_new (g_strdup_printf ("%s%s", _("E-Mail:"), priv->feedback_mail));
    st_widget_set_style_class_name (ST_WIDGET (mail_item_label), "feedback-context-label");
    clutter_actor_set_x_expand (CLUTTER_ACTOR (mail_item_label), TRUE);
    clutter_actor_set_x_align (CLUTTER_ACTOR (mail_item_label), ST_ALIGN_START);
    _add_actor (mail_item_box, mail_item_label);

    // 技术服务号 - 图片
    StWidget *pic_item_box = st_box_layout_new ();
    st_box_layout_set_vertical (ST_BOX_LAYOUT (pic_item_box), FALSE);
    st_widget_set_style_class_name (ST_WIDGET (pic_item_box), "feedback-pic-box");
    _add_actor (nfs_box, pic_item_box);

    StIcon *pic_item_icon = ST_ICON (st_icon_new ());
    st_icon_set_icon_size (ST_ICON (pic_item_icon), 128);
    clutter_actor_set_x_expand (CLUTTER_ACTOR (pic_item_icon), FALSE);
    clutter_actor_set_x_align (CLUTTER_ACTOR (pic_item_icon), ST_ALIGN_MIDDLE);
    file = g_file_new_for_path (priv->feedback_image_file);
    gicon = g_file_icon_new (file);
    st_icon_set_gicon (ST_ICON (pic_item_icon), gicon);
    g_object_unref (gicon);
    g_object_unref (file);
    _add_actor (pic_item_box, pic_item_icon);

    // 技术服务号 - item
    StWidget *ac_item_box = st_box_layout_new ();
    st_box_layout_set_vertical (ST_BOX_LAYOUT (ac_item_box), FALSE);
    st_widget_set_style_class_name (ST_WIDGET (ac_item_box), "feedback-ac-box");
    _add_actor (nfs_box, ac_item_box);

    StWidget *ac_item_label = st_label_new (_("Product Service"));
    clutter_actor_set_x_expand (CLUTTER_ACTOR (ac_item_label), TRUE);
    clutter_actor_set_x_align (CLUTTER_ACTOR (ac_item_label), ST_ALIGN_MIDDLE);
    _add_actor (ac_item_box, ac_item_label);
}

static void 
cdos_feedback_class_init( CdosFeedbackClass *kclass)
{
    GObjectClass *gobject_class = G_OBJECT_CLASS (kclass);
    gobject_class->finalize = cdos_feedback_finalize;
    gobject_class->dispose = cdos_feedback_dispose;
}

static void
cdos_feedback_init(CdosFeedback *self)
{
    CdosFeedbackPrivate *priv;
    priv = self->priv = cdos_feedback_get_instance_private (self);
}

static CdosFeedback *
cdos_feedback_new (guint orientation, guint panel_height)
{
    gsize len;
    int i;
    CdosFeedback *self  = g_object_new(CDOS_TYPE_FEEDBACK, "orientation", orientation, NULL);
    CdosFeedbackPrivate *priv = self->priv;

    cdos_feedback_new_for_config_file (self); 
    cdos_feedback_create_title_box (self);

    // 循环加入每个厂商的反馈：
    if (priv->feedback_config_file)
    {
        gchar **group = g_key_file_get_groups (priv->feedback_config_file, &len);
        for (i = 0; i < len; i++)
        {
            cdos_feedback_get_info_from_config_file (self, *(group + i));
            cdos_feedback_create_contant_box (self,*(group + i));
        }
        g_strfreev (group);
    }

    return self;
}

CdosApplet * 
applet_feedback_main (MetaData *meta,guint orientation, guint panel_height)
{
    CdosFeedback *self = NULL;
    // 通过gsettings值来决定是否创建
    if (should_enable_feedback())
    {
        self = cdos_feedback_new (orientation, panel_height);
    }
    
    return CDOS_APPLET (self);
}