#include <math.h>
#include <gio/gio.h>
#include "applets-group.h"
#include "cdos-popup-separator-menu-item.h"
#include "cdos-util.h"

#define APPLET_GROUP_COLUMN 3

struct _AppletGroupPrivate
{
    CdosPopupMenu         *menu;
    guint                  orientation;
    CdosPopupMenuItem     *setting_item;
    CdosPopupBaseMenuItem *separator;
    GList                 *applets;
    StWidget              *table;
    ClutterGridLayout     *grid_layout;
    StBoxLayout           *box;
    guint                  count;
    GSettings             *settings;
    gint                   row_size;
};

G_DEFINE_TYPE_WITH_PRIVATE (AppletGroup, applet_group, CDOS_TYPE_ICONAPPLET);

static AppletGroup *applet_group = NULL;

static gboolean applet_group_has_applet (AppletGroup *self, CdosApplet *applet);

static void applet_group_table_add_child (AppletGroup *self, ClutterActor *actor);

static void
applet_group_dispose (GObject *object)
{
    AppletGroup *self = APPLET_GROUP (object);
    AppletGroupPrivate *priv = self->priv;
    if (priv->table)
    {
        clutter_actor_remove_all_children ((ClutterActor *)priv->table);
        priv->table = NULL;
    }
    applet_group = NULL;
    G_OBJECT_CLASS (applet_group_parent_class)->dispose (object);
}

static gboolean
applet_group_setting_item_on_activate (ClutterActor *actor, gpointer event, gboolean keep, gpointer data)
{
    cdos_util_spawn_command_line("found-systray-manager");
    return TRUE;
}

static gboolean
applet_group_has_applet (AppletGroup *self, CdosApplet *applet)
{
    AppletGroupPrivate *priv = self->priv;
    if (priv->applets)
    {
        GList *applets = priv->applets;
        while (applets != NULL)
        {
            CdosApplet *applet_tmp = applets->data;
            if (applet_tmp && applet_tmp == applet)
            {
                return TRUE;
            }
            applets = applets->next;
        }
    }
    return FALSE;
}

static void
applet_group_table_add_child (AppletGroup *self, ClutterActor *actor)
{
    AppletGroupPrivate *priv = self->priv;
    gint row_size, row, col, x;
    row_size = (int)ceil (sqrt (priv->count));

    if (row_size <= 3)
    {
        x = priv->count - priv->row_size * priv->row_size;
        if (x < row_size)
        {
            row = x - 1;
            col = row_size - 1;
        }
        else
        {
            col = priv->count % row_size - 1;
            row = row_size - 1;
        }
        if (col < 0)
        {
            col = col + row_size;
            priv->row_size = row_size;
        }
    }
    else
    {
        row_size = APPLET_GROUP_COLUMN;
        col = (priv->count - 1) % row_size;
        row = (int)(priv->count / row_size);
    }
    clutter_grid_layout_attach (priv->grid_layout, actor, col, row, 1, 1);
}

void
applet_group_add_applet (AppletGroup *self, CdosApplet *applet)
{
    g_return_if_fail (APPLET_IS_GROUP (self));
    g_return_if_fail (CDOS_IS_APPLET (applet));

    AppletGroupPrivate *priv = self->priv;
    ClutterActor *actor = CLUTTER_ACTOR (cdos_applet_get_actor (applet));
    if (applet_group_has_applet (self, applet) == TRUE)
    {
        return;
    }
    priv->applets = g_list_append (priv->applets, applet);
    priv->count++;

    //StBin *applet_icon_box = cdos_iconapplet_get_applet_icon_box (CDOS_ICONAPPLET (applet));
    st_widget_add_style_class_name (ST_WIDGET (actor), "applet-group-item");

    applet_group_table_add_child (self, actor);

    clutter_actor_show ((ClutterActor *)priv->table);
}

void
applet_group_remove_all_applets (AppletGroup *self)
{
    g_return_if_fail (APPLET_IS_GROUP (self));
    AppletGroupPrivate *priv = self->priv;
    if (priv->applets)
    {
        GList *applets = priv->applets;
        while (applets)
        {
            CdosApplet *applet = applets->data;
            if (applet)
            {
                ClutterActor *actor = CLUTTER_ACTOR (applet);
                //StBin *applet_icon_box = cdos_iconapplet_get_applet_icon_box (CDOS_ICONAPPLET (applet));
                st_widget_remove_style_class_name (ST_WIDGET (actor), "applet-group-item");
                g_object_ref (actor);
                clutter_actor_remove_child (CLUTTER_ACTOR (priv->table), actor);
            }
            applets = applets->next;
        }
        g_list_free (priv->applets);
        priv->applets = NULL;
    }
    priv->count = 0;
    priv->row_size = 0;
    clutter_actor_hide ((ClutterActor *)priv->table);
}

static AppletGroup *
applet_group_new (MetaData *metadata, guint orientation)
{
    AppletGroup *group;
    AppletGroupPrivate *priv = NULL;
    group = g_object_new (APPLET_TYPE_GROUP, "orientation", orientation, NULL);
    priv = group->priv;

    priv->orientation = orientation;
    return group;
}

AppletGroup *
applet_group_main (MetaData *metadata, guint orientation, gfloat panel_height)
{
    if (applet_group == NULL)
    {
        applet_group = applet_group_new (metadata, orientation);
    }
    return applet_group;
}

static void
applet_group_class_init (AppletGroupClass *kclass)
{
    GObjectClass *object_class = G_OBJECT_CLASS (kclass);
    object_class->dispose = applet_group_dispose;
}

static void
applet_group_init (AppletGroup *self)
{
    AppletGroupPrivate *priv;
    priv = self->priv = applet_group_get_instance_private (self);
    cdos_applet_set_popup_menu_active (CDOS_APPLET (self), TRUE);

    CdosIconApplet *iconapplet = (CdosIconApplet *)self;
    cdos_iconapplet_set_icon_name ((CdosIconApplet *)iconapplet, "found-systray-group");
    cdos_iconapplet_set_icon_size ((CdosIconApplet *)iconapplet, 16);
    cdos_applet_set_applet_tooltip ((CdosApplet *)iconapplet, _("running applets"));

    priv->menu = cdos_applet_get_popup_menu ((CdosApplet *)self);
    st_widget_set_style (ST_WIDGET (priv->menu->actor), "min-width:90px;");

    priv->table = g_object_new (ST_TYPE_WIDGET, NULL);
    clutter_actor_hide ((ClutterActor *)priv->table);
    priv->grid_layout = (ClutterGridLayout *)clutter_grid_layout_new ();
    clutter_actor_set_layout_manager ((ClutterActor *)priv->table, (ClutterLayoutManager *)priv->grid_layout);
    clutter_grid_layout_set_row_homogeneous (priv->grid_layout, TRUE);
    clutter_grid_layout_set_column_homogeneous (priv->grid_layout, TRUE);

    // clutter_actor_add_child ((ClutterActor *)priv->box, CLUTTER_ACTOR (priv->table));
    cdos_popup_menu_base_add_actor ((CdosPopupMenuBase *)priv->menu, (ClutterActor *)priv->table);

    priv->count = 0;
    priv->row_size = 0;

    priv->separator = (CdosPopupBaseMenuItem *)cdos_popup_separator_menu_item_new ();
    cdos_popup_menu_base_add_menu_item (CDOS_POPUP_MENU_BASE (priv->menu), priv->separator, -1);

    priv->setting_item = cdos_popup_menu_item_new ();
    cdos_popup_menu_item_set_label (priv->setting_item, _("Settings"));
    cdos_popup_menu_item_set_label_align (priv->setting_item, ST_ALIGN_MIDDLE, 0);
    g_signal_connect ((CdosPopupBaseMenuItem *)priv->setting_item,
                      "activate",
                      G_CALLBACK (applet_group_setting_item_on_activate),
                      self);
    cdos_popup_menu_base_add_menu_item (CDOS_POPUP_MENU_BASE (priv->menu), (CdosPopupBaseMenuItem *)priv->setting_item, -1);
}
