/**
 * Copyright (c) 2016-2018 CPU and Fundamental Software Research Center, CAS
 *
 * This software is published by the license of CPU-OS License，you can use and
 * distibuted this software under this License. See CPU-OS Liense for more detail.
 *
 * You should have receive a copy of CPU-OS License. If not, please contact us
 * by email <support_os@cpu-os.ac.cn>
 *
 **/
//
// Created by zhangtao on 18-1-18.
// Email: zhangtao@cpu-os.ac.cn
//
#include <syslog.h>
#include "common.h"

#include "applet-applications-manager.h"
#include "applet-applications-category.h"
#include "applet-menu-helper.h"
#include "applet-applications.h"
#include "applet-menu.h"
#include "applet-menu-private.h"
#include "cdos-app-usage.h"
#include "cdos-recently-parser.h"
#include "cdos-app-system.h"
#include "applet-application-button.h"

#define APPLET_APPLICATIONS_MANAGER_GET_PRIVATE(obj)      \
(G_TYPE_INSTANCE_GET_PRIVATE((obj), APPLET_TYPE_APPLICATIONS_MANAGER, AppletApplicationsManagerPrivate))

G_DEFINE_TYPE(AppletApplicationsManager, applet_applications_manager, G_TYPE_OBJECT);

struct _AppletApplicationsManagerPrivate
{
    StBoxLayout     *box;
    StScrollView    *scroll_box;
    StBoxLayout     *applications_list;

    GList                   *categories;
    AppletApplications      *apps;
    CdosRecentlyParser    *newly_installed;
    AppletApplicationsCategory *newly_installed_category;
    CdosAppUsage            *frequently_used;
    AppletApplicationsCategory *frequently_used_category;

    gulong frequently_delete_id;
    gulong newly_delete_id;

    GSettings       *settings;
    ApplicationsGroupType group_type;
    gboolean        show_expand_button;
    guint            all_close_id;
};

static void applet_applications_manager_dispose (GObject *object);
static void applet_applications_manager_finalize (GObject *object);
static void
applet_applications_manager_load_data (AppletApplicationsManager *self);

static void
app_system_on_installed_changed (AppletApplications *apps, AppletApplicationsManager *self);

static AppletApplicationsCategory *
applet_applications_add_category (AppletApplicationsManager *self, const gchar *title, gint index);
static void
update_icon_theme_changed (StTextureCache *texture_cache, AppletApplicationsManager *self);
static void
newly_installed_on_changed (CdosRecentlyParser *recently_installed, gpointer user_data);
static void
applet_applications_newly_installed (AppletApplicationsManager *self);
static void
applet_applications_frequently_used (AppletApplicationsManager *self);
static void
usage_changed_on_changed (CdosAppUsage *frequently_used, gpointer user_data);
static void
stage_on_key_focus_changed  (GObject    *gobject,
                            GParamSpec  *pspec,
                            gpointer    user_data);
static void
scroll_bar_on_changed (GObject    *object,
                      GParamSpec  *pspec,
                      gpointer user_data);
static void
menu_settings_on_changed (GSettings *settings,
                         gchar *key,
                         gpointer    user_data);
static void
applet_applications_manager_init (AppletApplicationsManager *self)
{
    AppletApplicationsManagerPrivate *priv;
    priv = self->priv = APPLET_APPLICATIONS_MANAGER_GET_PRIVATE (self);
    priv->show_expand_button = FALSE;

    priv->settings = g_settings_new (CDOS_MENU_SCHEMA);
    //priv->group_type = g_settings_get_int (priv->settings, MENU_APPLICATIONS_CATEGORY_TYPE);
    priv->group_type = APPLICATIONS_GROUP_BY_LETTER;
    g_signal_connect (priv->settings,
                     "changed",
                     G_CALLBACK (menu_settings_on_changed),
                     self);

    priv->box = ST_BOX_LAYOUT (st_box_layout_new ());
    st_box_layout_set_vertical (priv->box, FALSE);
    st_widget_set_style_class_name ((StWidget *)priv->box, "menu-apps-box");

    priv->scroll_box = g_object_new(ST_TYPE_SCROLL_VIEW,
                                "style-class" , "vfade menu-applications-scrollbox",
                                "x-fill", TRUE,
                                "y-fill", FALSE,
                                "y-align", CLUTTER_ACTOR_ALIGN_START,
                                NULL);

    st_scroll_view_set_policy (priv->scroll_box, GTK_POLICY_NEVER, GTK_POLICY_AUTOMATIC);
    st_scroll_view_set_auto_scrolling (priv->scroll_box, TRUE);
    priv->applications_list = ST_BOX_LAYOUT (st_box_layout_new ());
    st_widget_set_style_class_name ((StWidget *)priv->applications_list, "menu-apps-list");

    ClutterActor *scroll_bar;
    StAdjustment *adjustment;
    scroll_bar = st_scroll_view_get_vscroll_bar (priv->scroll_box);
    adjustment = st_scroll_bar_get_adjustment ((StScrollBar  *)scroll_bar);
    g_signal_connect (adjustment,
                     "notify::value",
                     G_CALLBACK (scroll_bar_on_changed),
                     self);
    st_box_layout_set_vertical (priv->applications_list, TRUE);
    menu_container_add_actor ((ClutterActor *)priv->box, (ClutterActor *)priv->scroll_box,
                             "x-expand", TRUE,
                             //"x-fill", TRUE,
                             NULL);
    /*menu_container_add_actor ((ClutterActor *)priv->scroll_box, (ClutterActor *)priv->applications_list,
                             "x-expand", TRUE,
                             //"x-fill", TRUE,
                             NULL);*/
    st_scroll_view_add_child (priv->scroll_box, (ClutterActor *)priv->applications_list);

    CdosGlobal *global = cdos_global_get();
    ClutterStage *global_stage = cdos_global_get_stage (global);
    g_signal_connect (global_stage,
                     "notify::key-focus",
                     G_CALLBACK (stage_on_key_focus_changed),
                     self);
    priv->apps = applet_applications_get_default ();
    g_signal_connect (priv->apps,
                     "app-list-changed",
                     G_CALLBACK (app_system_on_installed_changed),
                     self);
    StTextureCache *text_cache;
    text_cache = st_texture_cache_get_default ();
    g_signal_connect (text_cache,
                      "icon-theme-changed",
                      G_CALLBACK (update_icon_theme_changed),
                      self);

    priv->newly_installed = cdos_recently_parser_new ();
    g_signal_connect (priv->newly_installed,
                      "recent-changed",
                      G_CALLBACK (newly_installed_on_changed),
                      self);
    priv->frequently_used = cdos_app_usage_get_default ();
    g_signal_connect (priv->frequently_used,
                     "usage-changed",
                     G_CALLBACK (usage_changed_on_changed),
                     self);
    applet_applications_newly_installed (self);
    applet_applications_frequently_used (self);

    applet_applications_manager_load_data (self);
}

static void
applet_applications_manager_class_init (AppletApplicationsManagerClass *klass)
{

    GObjectClass *gobject_class = G_OBJECT_CLASS (klass);
    g_type_class_add_private (klass, sizeof (AppletApplicationsManagerPrivate));

    gobject_class->dispose = applet_applications_manager_dispose;
    gobject_class->finalize = applet_applications_manager_finalize;

}

static void
applet_applications_manager_dispose (GObject *object)
{
    AppletApplicationsManager *self = APPLET_APPLICATIONS_MANAGER (object);
    AppletApplicationsManagerPrivate *priv = self->priv;

    CdosGlobal *global = cdos_global_get();
    ClutterStage *global_stage = cdos_global_get_stage (global);
    g_signal_handlers_disconnect_by_data (global_stage, self);
    if (priv->all_close_id > 0)
    {
        g_source_remove (priv->all_close_id);
        priv->all_close_id = 0;
    }
    if (priv->categories)
    {
        g_list_free_full (priv->categories, (GDestroyNotify )clutter_actor_destroy);
        priv->categories = NULL;
    }
    if (priv->apps)
    {
        g_clear_object (&priv->apps);
    }
    if (priv->box)
    {
        clutter_actor_destroy ((ClutterActor *)priv->box);
        priv->box = NULL;
    }
    g_clear_object (&priv->settings);
    g_clear_object (&priv->newly_installed);
    g_clear_object (&priv->frequently_used);
    G_OBJECT_CLASS (applet_applications_manager_parent_class)->dispose (object);
}

static void
applet_applications_manager_finalize (GObject *object)
{
    G_OBJECT_CLASS (applet_applications_manager_parent_class)->finalize (object);
}

static void
update_icon_theme_changed (StTextureCache *texture_cache, AppletApplicationsManager *self)
{
    applet_applications_manager_load_data (self);
}

static void
applet_applications_manager_load_data (AppletApplicationsManager *self)
{
    AppletApplicationsManagerPrivate *priv = self->priv;
    gchar *icon_name;
    if (priv->categories)
    {
        g_list_free_full (priv->categories, (GDestroyNotify )clutter_actor_destroy);
        priv->categories = NULL;
    }
    GList *groups = applet_applications_get_groups (priv->apps, priv->group_type);
    GList *g = groups;
    while (groups != NULL)
    {
        gchar *letter = groups->data;
        GList *source_data = NULL;
        if (priv->group_type == APPLICATIONS_GROUP_BY_LETTER)
        {
            source_data = applet_applications_get_by_letter (priv->apps, letter);
        }
        else if (priv->group_type == APPLICATIONS_GROUP_BY_CATEGORY)
        {
            source_data = applet_applications_get_by_category (priv->apps, letter);
        }

        if (source_data)
        {
            AppletApplicationsCategory *category = applet_applications_add_category (self, letter, -1);
            if (priv->group_type == APPLICATIONS_GROUP_BY_CATEGORY)
            {
                applet_category_base_show_expand_button ((AppletCategoryBase *)category, TRUE);
                applet_category_base_set_expand ((AppletCategoryBase *)category, FALSE);
                applet_category_base_set_default_show_num ((AppletCategoryBase *)category, 0);
                applet_category_base_set_head_sensitive ((AppletCategoryBase *)category, TRUE);
                icon_name = applet_applications_get_category_icon_name (priv->apps, letter);
                applet_category_base_set_icon_name ((AppletCategoryBase *)category, icon_name);
                g_free (icon_name);
            }
            applet_category_base_set_source_data ((AppletCategoryBase *)category, source_data);
            g_list_free (source_data);
        }
        groups = groups->next;
    }
    g_list_free_full (g, g_free);
}

static void
app_system_on_installed_changed (AppletApplications *apps, AppletApplicationsManager *self)
{
    applet_applications_manager_load_data (self);
}

static AppletApplicationsCategory *
applet_applications_add_category (AppletApplicationsManager *self, const gchar *title, gint index)
{
    AppletApplicationsManagerPrivate *priv = self->priv;
    AppletApplicationsCategory *category = applet_applications_category_new (title);
    clutter_actor_set_x_expand ((ClutterActor *)category, TRUE);
    applet_category_base_set_head_sensitive ((AppletCategoryBase *)category, FALSE);
    if (index >= 0)
    {
        priv->categories = g_list_insert (priv->categories, category, index);
        clutter_actor_insert_child_at_index ((ClutterActor *)priv->applications_list, (ClutterActor *)category, index);
    } else {
        priv->categories = g_list_append (priv->categories, category);
        //applet_category_base_set_source_data ((AppletCategoryBase *)category, source_data);
        _add_actor (priv->applications_list, category);
    }
    return category;
}

static void
remove_newly_installed_category (AppletApplicationsManager *self)
{
    AppletApplicationsManagerPrivate *priv = self->priv;
    if (priv->newly_installed_category)
    {
        clutter_actor_remove_child ((ClutterActor *)priv->applications_list,
                                    (ClutterActor *)priv->newly_installed_category);
        priv->newly_installed_category = NULL;
    }
}

static gboolean
newly_remove_all_app (AppletApplicationsManager *self)
{
    AppletApplicationsManagerPrivate *priv = self->priv;
    GList * source_list = cdos_recently_parser_get_app_list (priv->newly_installed);
    CdosApp *app;
    const gchar *id;
    while (source_list)
    {
        app = source_list->data;
        id = cdos_app_get_id (app);
        cdos_recently_parser_remove_app (priv->newly_installed, id);
        source_list = source_list->next;
    }
    g_list_free (source_list);
    applet_applications_newly_installed (self);
    return FALSE;
}

static void
newly_remove_item (CdosPopupBaseMenuItem *item,
             ClutterEvent          *event,
             gboolean               keepMenu,
             AppletApplicationsManager *self)
{
    AppletApplicationsManagerPrivate *priv = self->priv;
    g_signal_handler_block (priv->newly_installed_category, priv->newly_delete_id);
    g_idle_add (newly_remove_all_app, self);   
}

static void
newly_category_on_relayout_finish (ClutterActor *category, AppletApplicationsManager *self)
{
    GList * button_list = applet_category_base_get_app_buttons ((AppletCategoryBase *)category);

    GList * l = button_list;
    while (l != NULL)
    {
        applet_application_button_add_context_menu_item ((AppletApplicationButton*)l->data, _("Delete list"), 2, (GCallback)newly_remove_item, self);
        l = l->next;
    }
}

static void
applet_applications_newly_installed (AppletApplicationsManager *self)
{
    AppletApplicationsManagerPrivate *priv = self->priv;
    AppletApplicationsCategory *category;
    GList *recent_app_list = NULL;
    CdosAppSystem *appsys = cdos_app_system_get_default ();
    recent_app_list = cdos_recently_parser_get_app_list (priv->newly_installed);

    const gchar *title = _("Newly installed applications");
    if (recent_app_list != NULL)
    {
        if (priv->newly_installed_category == NULL)
        {
            category = applet_applications_add_category (self, title, 0);
            priv->categories = g_list_remove (priv->categories, category);
            applet_category_base_show_expand_button ((AppletCategoryBase *)category, FALSE);
            applet_category_base_set_button_layout_mode ((AppletCategoryBase *)category, BUTTON_LAYOUT_H);
            applet_category_base_set_default_show_num ((AppletCategoryBase *)category, 5);
            applet_category_base_set_source_data ((AppletCategoryBase *)category, recent_app_list);
            priv->newly_installed_category = category;

            priv->newly_delete_id = g_signal_connect (priv->newly_installed_category,
                                                      "relayout-finish",
                                                      G_CALLBACK (newly_category_on_relayout_finish),
                                                      self);
        } else {
            applet_category_base_source_data_changed ((AppletCategoryBase *)priv->newly_installed_category,
                                                     recent_app_list,
                                                     TRUE);
        }
        g_list_free (recent_app_list);
    } else {
        remove_newly_installed_category (self);
    }
}

static void
newly_installed_on_changed (CdosRecentlyParser *recently_installed, gpointer user_data)
{
    AppletApplicationsManager *self = APPLET_APPLICATIONS_MANAGER (user_data);
    ApplicationsGroupType group_type = applet_applications_manager_get_group_type (self);
    // 默认排序时，显示"最近添加";所有分类时，不显示"最近添加"。
    if (group_type == APPLICATIONS_GROUP_BY_LETTER)
        applet_applications_newly_installed (self); 
    else
        remove_newly_installed_category (self);
}

static void
remove_frequently_used_category (AppletApplicationsManager *self)
{
    AppletApplicationsManagerPrivate *priv = self->priv;
    if (priv->frequently_used_category)
    {
        clutter_actor_remove_child ((ClutterActor *)priv->applications_list,
                                    (ClutterActor *)priv->frequently_used_category);
        priv->frequently_used_category = NULL;
    }
}

gboolean
frequently_remove_all_app (AppletApplicationsManager *self)
{
    AppletApplicationsManagerPrivate *priv = self->priv;
    GList * list = cdos_app_usage_get_list (priv->frequently_used);
    GList * tmp_list = g_list_copy (list);
    GList * l;
    for (l = tmp_list; l != NULL; l = l->next)
    {
        cdos_app_usage_remove_app (priv->frequently_used, (CdosApp *)l->data);
    }
    g_list_free (tmp_list);
    applet_applications_frequently_used (self);
    return FALSE;
}

static void
frequently_remove_item (CdosPopupBaseMenuItem *item,
				                                ClutterEvent          *event,
                                                gboolean               keepMenu,
                                              AppletApplicationsManager *self)
{
    AppletApplicationsManagerPrivate *priv = self->priv;
    g_signal_handler_block (priv->frequently_used_category, priv->frequently_delete_id);
    g_idle_add (frequently_remove_all_app, self);   
}

static  void
frequently_category_on_relayout_finish (ClutterActor *category, AppletApplicationsManager *self)
{
    GList * button_list = applet_category_base_get_app_buttons ((AppletCategoryBase *)category);

    GList *l = button_list;
    while (l != NULL)
    {
        applet_application_button_add_context_menu_item ((AppletApplicationButton*)l->data, _("Delete list"), 2, (GCallback)frequently_remove_item, self);
        l = l->next;
    }
}

static void
applet_applications_frequently_used (AppletApplicationsManager *self)
{
    AppletApplicationsManagerPrivate *priv = self->priv;
    AppletApplicationsCategory *category;
    GList *source_data;

    source_data = cdos_app_usage_get_list (priv->frequently_used);
    const gchar *title = _("Frequently used applications");
    if (source_data != NULL)
    {
        if (priv->frequently_used_category == NULL)
        {
            guint index = priv->newly_installed_category != NULL ? 1 : 0;
            category = applet_applications_add_category (self, title, index);
            priv->categories = g_list_remove (priv->categories, category);
            applet_category_base_show_expand_button ((AppletCategoryBase *)category, FALSE);
            applet_category_base_set_button_layout_mode ((AppletCategoryBase *)category, BUTTON_LAYOUT_H);
            applet_category_base_set_default_show_num ((AppletCategoryBase *)category, 5);
            applet_category_base_set_source_data ((AppletCategoryBase *)category, source_data);
            priv->frequently_used_category = category;

            priv->frequently_delete_id = g_signal_connect (priv->frequently_used_category,
                          "relayout-finish",
                          G_CALLBACK (frequently_category_on_relayout_finish),
                          self);
        } else {
            applet_category_base_source_data_changed ((AppletCategoryBase *)priv->frequently_used_category,
                                                     source_data,
                                                     TRUE);
        }
    } else {
        remove_frequently_used_category (self);
    }
}

static void
usage_changed_on_changed (CdosAppUsage *frequently_used, gpointer user_data)
{
    AppletApplicationsManager *self = APPLET_APPLICATIONS_MANAGER (user_data);
    ApplicationsGroupType group_type = applet_applications_manager_get_group_type (self);
    // 默认排序时，显示"最常使用";所有分类时，不显示"最常使用"。
    if (group_type == APPLICATIONS_GROUP_BY_LETTER)
        applet_applications_frequently_used (self);        
    else
        remove_frequently_used_category (self);
}

static void
stage_on_key_focus_changed  (GObject    *gobject,
                            GParamSpec  *pspec,
                            gpointer    user_data)
{
    AppletApplicationsManager *self = APPLET_APPLICATIONS_MANAGER (user_data);
    AppletApplicationsManagerPrivate *priv = self->priv;
    ClutterActor *focused;
    focused = clutter_stage_get_key_focus (CLUTTER_STAGE (gobject));
    if (APPLET_IS_APPLICATION_BUTTON (focused))
    {
        applet_helper_scroll_to_focused_actor (priv->scroll_box);
    }
}

static void
all_context_menu_close_delay (AppletApplicationsManager *self)
{
    AppletApplicationsManagerPrivate *priv = self->priv;
    GList *categories = priv->categories;
    if (priv->newly_installed_category)
    {
        applet_category_base_close_all_popup_menu ((AppletCategoryBase *)priv->newly_installed_category);
    }
    if (priv->frequently_used_category)
    {
        applet_category_base_close_all_popup_menu ((AppletCategoryBase *)priv->frequently_used_category);
    }
    while (categories)
    {
        AppletApplicationsCategory *category = categories->data;
        applet_category_base_close_all_popup_menu ((AppletCategoryBase *)category);
        categories = categories->next;
    }
    g_clear_handle_id (&priv->all_close_id, g_source_remove);
    return FALSE;
}

static void
scroll_bar_on_changed (GObject     *object,
                      GParamSpec  *pspec,
                      gpointer user_data)
{
    AppletApplicationsManager *self = APPLET_APPLICATIONS_MANAGER (user_data);
    AppletApplicationsManagerPrivate *priv = self->priv;
    if (priv->all_close_id > 0)
    {
        g_source_remove (priv->all_close_id);
        priv->all_close_id = 0;
        return;
    }
    priv->all_close_id = g_idle_add (G_CALLBACK (all_context_menu_close_delay), self);
}

static void
menu_settings_on_changed (GSettings *settings,
                         gchar *key,
                         gpointer    user_data)
{
    AppletApplicationsManager *self = APPLET_APPLICATIONS_MANAGER (user_data);
    AppletApplicationsManagerPrivate *priv = self->priv;
    /* if (g_strcmp0 (key, MENU_APPLICATIONS_CATEGORY_TYPE) == 0)
    {
        priv->group_type = g_settings_get_int (priv->settings, MENU_APPLICATIONS_CATEGORY_TYPE);
        applet_applications_manager_load_data (self);
    } */
}

AppletApplicationsManager *
applet_applications_manager_new (void)
{
    AppletApplicationsManager *applications_manager ;
    applications_manager = g_object_new (APPLET_TYPE_APPLICATIONS_MANAGER, NULL);
    return applications_manager;
}

StBoxLayout *
applet_applications_manager_get_box (AppletApplicationsManager *self)
{
    g_return_val_if_fail (APPLET_IS_APPLICATIONS_MANAGER(self), NULL);
    AppletApplicationsManagerPrivate *priv = self->priv;
    return priv->box;
}

void
applet_applications_manager_start_key_focus (AppletApplicationsManager *self)
{
    g_return_if_fail (APPLET_IS_APPLICATIONS_MANAGER(self));
    AppletApplicationsManagerPrivate *priv = self->priv;
    AppletApplicationsCategory *category;
    GList *children;
    children = clutter_actor_get_children ((ClutterActor *)priv->applications_list);
    category = g_list_nth_data (children, 0);
    applet_category_base_buttons_navigate_focus ((AppletCategoryBase *)category, 0);
    g_list_free (children);
}

void
applet_applications_manager_set_group_type (AppletApplicationsManager *self, 
                                            ApplicationsGroupType group_type)
{
    g_return_if_fail (APPLET_IS_APPLICATIONS_MANAGER(self));
    AppletApplicationsManagerPrivate *priv = self->priv;
    if (group_type != priv->group_type)
    {
        priv->group_type = group_type;
        applet_applications_manager_load_data (self);
    }
}

ApplicationsGroupType
applet_applications_manager_get_group_type (AppletApplicationsManager *self)
{
    g_return_val_if_fail (APPLET_IS_APPLICATIONS_MANAGER(self), APPLICATIONS_GROUP_BY_LETTER);
    AppletApplicationsManagerPrivate *priv = self->priv;
    return priv->group_type;
}

void
applet_applications_manager_set_expand_button_visible (AppletApplicationsManager *self, gboolean shown)
{
    g_return_if_fail (APPLET_IS_APPLICATIONS_MANAGER(self));
    AppletApplicationsManagerPrivate *priv = self->priv;
    priv->show_expand_button = TRUE;
}

void
applet_application_mananger_set_frequently_used_visible (AppletApplicationsManager *self, gboolean show)
{
    g_return_if_fail (APPLET_IS_APPLICATIONS_MANAGER(self));
    if (show)
    {
        applet_applications_newly_installed (self);
        applet_applications_frequently_used (self);
    } else {
        remove_newly_installed_category (self);
        remove_frequently_used_category (self);
    }
}

gboolean 
applet_application_manager_get_is_scrolling (AppletApplicationsManager *self)
{
    g_return_if_fail (APPLET_IS_APPLICATIONS_MANAGER(self));
    AppletApplicationsManagerPrivate *priv = self->priv;
    gboolean is_scrolling = FALSE;
    is_scrolling = st_scroll_view_is_scrolling (priv->scroll_box);
    return is_scrolling;
}
