/**
 * Copyright (c) 2016-2018 CPU and Fundamental Software Research Center, CAS
 *
 * This software is published by the license of CPU-OS License，you can use and
 * distibuted this software under this License. See CPU-OS Liense for more detail.
 *
 * You should have receive a copy of CPU-OS License. If not, please contact us
 * by email <support_os@cpu-os.ac.cn>
 *
 **/
//
// Created by zhangtao on 18-1-18.
// Email: zhangtao@cpu-os.ac.cn
//
#include <string.h>
#include <syslog.h>
#include <glib/gi18n.h>
#include "applet-applications.h"
#include "cdos-app-system.h"
#include "cdos-app.h"
#include "cdos-pinyin.h"
#include <gmenu-tree.h>

#define APPLET_APPLICATIONS_GET_PRIVATE(obj)                                                                           \
    (G_TYPE_INSTANCE_GET_PRIVATE ((obj), APPLET_TYPE_APPLICATIONS, AppletApplicationsPrivate))

G_DEFINE_TYPE (AppletApplications, applet_applications, G_TYPE_OBJECT);

static AppletApplications *applications = NULL;

enum {
  APP_LIST_CHANGED,
  LAST_SIGNAL
};

static guint signals[LAST_SIGNAL] = {0};


struct _AppletApplicationsPrivate
{
    CdosAppSystem *app_system;
    GSList *all_apps;
    GList *app_letters[128];
    GList *categories;
    GHashTable *apps_category;
};

static const gchar default_letter = '#';

const char *not_show_list[] = {"blueman-adapters.desktop",
                               "blueman-manager.desktop",
                               "openjdk-8-policytool.desktop",
                               "ydm.desktop",
                               "tracker-preferences.desktop",
                               "im-config.desktop",
                               "tracker"
                               };

static void applet_applications_dispose (GObject *object);
static void applet_applications_finalize (GObject *object);

static void applet_applications_letters_add_app (AppletApplications *self, CdosApp *app);
static void applet_applications_categories_add_app (AppletApplications *self, CdosApp *app);
static void app_system_on_installed_changed (CdosAppSystem *app_system, CdosApp *app, gpointer user_data);
static void app_system_on_remove_changed (CdosAppSystem *app_system, const gchar *id, gpointer user_data);
static void applet_applications_load_data (AppletApplications *self);
/*
static gboolean
is_cdos_settings_app (CdosApp *app);
*/
static gboolean should_show_app (CdosApp *app);

static void
applet_applications_init (AppletApplications *self)
{
    AppletApplicationsPrivate *priv;
    priv = self->priv = APPLET_APPLICATIONS_GET_PRIVATE (self);

    priv->all_apps = NULL;
    priv->apps_category = g_hash_table_new_full (g_str_hash, g_str_equal, NULL, (GDestroyNotify)g_list_free);

    priv->app_system = cdos_app_system_get_default ();

    g_signal_connect (priv->app_system, "installed-changed", G_CALLBACK (app_system_on_installed_changed), self);

    g_signal_connect (priv->app_system, "removed-changed", G_CALLBACK (app_system_on_remove_changed), self);
}

static void
applet_applications_class_init (AppletApplicationsClass *klass)
{
    GObjectClass *gobject_class = G_OBJECT_CLASS (klass);
    g_type_class_add_private (klass, sizeof (AppletApplicationsPrivate));

    gobject_class->dispose = applet_applications_dispose;
    gobject_class->finalize = applet_applications_finalize;

    signals[APP_LIST_CHANGED] = g_signal_new ("app-list-changed",
                                APPLET_TYPE_APPLICATIONS,
                                G_SIGNAL_RUN_LAST,
                                0,
                                NULL, NULL,
                                g_cclosure_marshal_VOID__VOID,
                                G_TYPE_NONE, 0);
}

static void
applet_applications_dispose (GObject *object)
{
    AppletApplicationsPrivate *priv = APPLET_APPLICATIONS (object)->priv;
    g_signal_handlers_disconnect_by_data (priv->app_system, APPLET_APPLICATIONS (object));
    if (priv->all_apps)
    {
        g_slist_free (priv->all_apps);
        priv->all_apps = NULL;
    }
    for (int i = 0; i < G_N_ELEMENTS (priv->app_letters); i++)
    {
        if (priv->app_letters[i])
        {
            g_list_free (priv->app_letters[i]);
            priv->app_letters[i] = NULL;
        }
    }
    if (priv->apps_category)
    {
        g_hash_table_destroy (priv->apps_category);
    }
    G_OBJECT_CLASS (applet_applications_parent_class)->dispose (object);
}

static void
applet_applications_finalize (GObject *object)
{
    G_OBJECT_CLASS (applet_applications_parent_class)->finalize (object);
}

static void
applet_applications_load_data (AppletApplications *self)
{
    AppletApplicationsPrivate *priv = self->priv;
    if (priv->all_apps)
    {
        // g_slist_free_full (priv->all_apps, g_object_unref);
        g_slist_free (priv->all_apps);
        priv->all_apps = NULL;
        for (int i = 0; i < G_N_ELEMENTS (priv->app_letters); i++)
        {
            if (priv->app_letters[i])
            {
                g_list_free (priv->app_letters[i]);
                priv->app_letters[i] = NULL;
            }
        }
        g_hash_table_remove_all (priv->apps_category);
    }
    if (priv->categories)
    {
        g_list_free (priv->categories);
        priv->categories = NULL;
    }
    CdosAppSystem *system = cdos_app_system_get_default ();
    GSList *apps = cdos_app_system_get_all (system);
    // priv->all_apps = apps;
    while (apps != NULL)
    {
        CdosApp *app = apps->data;
        if (app)
        {
            // if (is_cdos_settings_app (app))
            if (!should_show_app (app))
            {
                apps = apps->next;
                continue;
            }
            priv->all_apps = g_slist_prepend (priv->all_apps, app);
            applet_applications_letters_add_app (self, app);
            applet_applications_categories_add_app (self, app);
        }
        apps = apps->next;
    }
    priv->all_apps = g_slist_reverse (priv->all_apps);

    g_slist_free (apps);
}

static gchar
get_first_letter (const gchar *pinyin)
{
    gchar out = 0;
    if (pinyin)
        out = pinyin[0];
    if (out >= 'a' && out <= 'z')
        out = out - ('a' - 'A');
    return out;
}

static gboolean
is_cdos_settings_app (CdosApp *app)
{
    GDesktopAppInfo *app_info = cdos_app_get_app_info (app);
    // 设置类程序，包括一些mate不能在CDOS显示
    if (app_info)
    {
        const char *filename = g_desktop_app_info_get_filename (app_info);

        if (g_str_match_string ("cdos-settings", filename, FALSE) || g_str_match_string ("cdos-sound", filename, FALSE)
            || g_str_match_string ("cdos-display", filename, FALSE)
            || g_str_match_string ("cdos-network", filename, FALSE))
        {
            return TRUE;
        }
    }

    if (!g_desktop_app_info_get_show_in (app_info, NULL))
    {
        return TRUE;
    }

    return FALSE;
}

static gboolean
is_not_show_app (CdosApp *app)
{
    const char *filename;
    int i;

    filename = cdos_app_get_id (app);

    if (filename)
    {
        for (i = 0; i < G_N_ELEMENTS (not_show_list); i++)
        {
            if (g_strrstr (filename, not_show_list[i]))
            {
                return TRUE;
            }
        }
    }

    return FALSE;
}

static gboolean
should_show_app (CdosApp *app)
{
    //过滤掉系统设置中的应用，如：电源管理、声音、网络、显示等
    if (is_cdos_settings_app (app))
    {
        return FALSE;
    }

    if (is_not_show_app (app))
    {
        return FALSE;
    }

    return TRUE;
}

static void
applet_applications_letters_add_app (AppletApplications *self, CdosApp *app)
{
    AppletApplicationsPrivate *priv = self->priv;
    const char *name = cdos_app_get_name (app);
    const gchar *pinyin = cdos_utf8_to_pinyin_fast (name);
    if (!pinyin)
        pinyin = name;
    gchar key = get_first_letter (pinyin);

    gint index_value = key - 'A';
    if (index_value < 0 || index_value > 25)
    {
        key = default_letter;
    }
    if (key == 0)
    {
        key = default_letter;
    }
    GList **group_apps = NULL;
    if (key != default_letter)
        group_apps = &priv->app_letters[key - 'A'];
    else
        group_apps = &priv->app_letters[G_N_ELEMENTS (priv->app_letters) - 1];

    GList *iter = *group_apps;
    while (iter)
    {
        CdosApp *app1 = iter->data;
        const char *name1 = cdos_app_get_name (app1);
        int a = g_utf8_collate (name, name1);
        if (a <= 0)
            break;
        else
            iter = iter->next;
    }
    *group_apps = g_list_insert_before (*group_apps, iter, app);
}

static GMenuTreeDirectory *
categories_get_directory_by_name (AppletApplications *self, const gchar *directory_name)
{
    AppletApplicationsPrivate *priv = self->priv;
    GMenuTreeDirectory *iter = NULL;
    const gchar *iter_name;
    GList *directories = priv->categories;

    while (directories != NULL)
    {
        iter = directories->data;
        iter_name = gmenu_tree_directory_get_name (iter);
        if (iter_name && g_strcmp0 (directory_name, iter_name) == 0)
        {
            return iter;
        }
        directories = directories->next;
    }
    return NULL;
}

static void
applet_applications_categories_add_app (AppletApplications *self, CdosApp *app)
{
    AppletApplicationsPrivate *priv = self->priv;
    GDesktopAppInfo *app_info = cdos_app_get_app_info (app);
    const gchar *categories = NULL;
    gboolean contains;
    GList *apps;
    if (app_info)
    {
        GMenuTreeEntry *entry = cdos_app_get_tree_entry (app);
        GMenuTreeDirectory *directory = gmenu_tree_entry_get_parent (entry);
        if (directory)
        {
            categories = gmenu_tree_directory_get_name (directory);
            if (categories_get_directory_by_name (self, categories) == NULL)
            {
                priv->categories = g_list_append (priv->categories, directory);
            }
        } else {
            categories = _("Others");
        }
    } else {
        categories = _("Others");
    }
    apps = g_hash_table_lookup (priv->apps_category, categories);
    contains = apps != NULL;
    apps = g_list_append (apps, app);
    if (contains)
    {
        g_hash_table_steal (priv->apps_category, categories);
        g_hash_table_replace (priv->apps_category, (gpointer)categories, apps);
    }
    else
    {
        g_hash_table_insert (priv->apps_category, (gpointer)categories, apps);
    }
}

static void
app_system_on_installed_changed (CdosAppSystem *app_system, CdosApp *app, gpointer user_data)
{  
    AppletApplications *self;
    self = (AppletApplications *)user_data;
    applet_applications_load_data (self);
    g_signal_emit (self, signals[APP_LIST_CHANGED], 0);
}

static void 
app_system_on_remove_changed (CdosAppSystem *app_system, const gchar *id, gpointer user_data)
{
    AppletApplications *self;
    self = (AppletApplications *)user_data;
    applet_applications_load_data (self);
    g_signal_emit (self, signals[APP_LIST_CHANGED], 0); 
}

static gint
groups_sort (gconstpointer a, gconstpointer b)
{
    gint x = strcoll (a, b);
    return x;
}

static GList *
app_system_get_categories (AppletApplications *self)
{
    //AppletApplicationsPrivate *priv = self->priv;
    CdosAppSystem *appsys;
    GMenuTree *tree;
    GMenuTreeDirectory *root;
    GMenuTreeIter *iter;
    GMenuTreeItemType nextType;
    GMenuTreeDirectory *dir;
    const gchar *menu_id;
    GList *groups = NULL;

    appsys = cdos_app_system_get_default ();
    tree = cdos_app_system_get_tree (appsys);
    root = gmenu_tree_get_root_directory (tree);
    iter = gmenu_tree_directory_iter (root);

    while ((nextType = gmenu_tree_iter_next (iter)) != GMENU_TREE_ITEM_INVALID)
    {
        if (nextType == GMENU_TREE_ITEM_DIRECTORY)
        {
            dir = gmenu_tree_iter_get_directory (iter);
            menu_id = gmenu_tree_directory_get_name (dir);
            if (gmenu_tree_directory_get_is_nodisplay (dir) || g_strcmp0 (menu_id, "Preferences") == 0)
                continue;
            groups = g_list_append (groups, g_strdup (menu_id));
        }
    }
    return groups;
}

GList *
applet_applications_get_groups (AppletApplications *self, ApplicationsGroupType group_type)
{
    g_return_val_if_fail (APPLET_IS_APPLICATIONS (self), NULL);
    AppletApplicationsPrivate *priv = self->priv;
    GList *groups = NULL;
    if (group_type == APPLICATIONS_GROUP_BY_LETTER)
    {
        gint i;
        gchar buf[2];
        
        buf[0] = default_letter;
        buf[1] = '\0';
        groups = g_list_append (groups, g_strdup (buf));
        for (i = 'A'; i <= 'Z'; i++)
        {
            buf[0] = i;
            buf[1] = '\0';
            groups = g_list_append (groups, g_strdup (buf));
        }
    }
    else if (group_type == APPLICATIONS_GROUP_BY_CATEGORY)
    {
        groups = app_system_get_categories (self);
    }
    return groups;
}

AppletApplications *
applet_applications_get_default (void)
{
    if (applications == NULL)
    {
        applications = g_object_new (APPLET_TYPE_APPLICATIONS, NULL);

        applet_applications_load_data (applications);
    }
    return applications;
}

GList *
applet_applications_get_by_letter (AppletApplications *self, const gchar *letter)
{
    g_return_val_if_fail (APPLET_IS_APPLICATIONS (self), NULL);
    g_return_val_if_fail (letter != NULL, NULL);
    AppletApplicationsPrivate *priv = self->priv;
    GList *apps = NULL;
    char key = letter[0];
    if (key == default_letter)
        key = G_N_ELEMENTS(priv->app_letters) - 1;
    else if (key >= 'A' && key <= 'Z')
        key = key - 'A';
    else
        return NULL;

    apps = priv->app_letters[(int)key];
    if (apps)
        return g_list_copy (apps);
    else
        return NULL;
}

GList *
applet_applications_get_by_category (AppletApplications *self, const gchar *category)
{
    g_return_val_if_fail (APPLET_IS_APPLICATIONS (self), NULL);
    AppletApplicationsPrivate *priv = self->priv;
    GList *apps = NULL;
    apps = g_hash_table_lookup (priv->apps_category, category);
    if (apps)
        return g_list_copy_deep (apps, (GCopyFunc)g_object_ref, NULL);
    return NULL;
}

gchar *
applet_applications_get_category_icon_name (AppletApplications *self, const gchar *category)
{
    g_return_val_if_fail (APPLET_IS_APPLICATIONS (self), NULL);
    AppletApplicationsPrivate *priv = self->priv;
    GMenuTreeDirectory *directory;
    GIcon *icon = NULL;
    directory = categories_get_directory_by_name (self, category);
    if (directory)
    {
        icon = gmenu_tree_directory_get_icon (directory);
    }
    if (icon)
    {
        return g_icon_to_string (icon);
    }
    return g_strdup ("applications-other");
}

GList *
applet_applications_search (AppletApplications *self, const gchar *keyword)
{
    g_return_val_if_fail (APPLET_IS_APPLICATIONS (self), NULL);
    g_return_val_if_fail (keyword != NULL, NULL);

    AppletApplicationsPrivate *priv = self->priv;
    GSList *all_apps = priv->all_apps;
    GList *results = NULL;
    const gchar * const *keywords;
    // const gchar *description, *generic_name, *filename, *categories, *exec;

    while (all_apps != NULL)
    {
        gboolean match = FALSE;
        CdosApp *app = all_apps->data;
        const gchar *name = cdos_app_get_name (app);

        /* GDesktopAppInfo *app_info = cdos_app_get_app_info (app);
        description = cdos_app_get_description (app);
        categories = g_desktop_app_info_get_categories (app_info);
        generic_name = g_desktop_app_info_get_generic_name (app_info);
        filename = g_desktop_app_info_get_filename (app_info);
        exec = g_desktop_app_info_get_string (app_info, "Exec");*/

        if (name == NULL)
            continue;
        match = cdos_pinyin_match (name, keyword, 1);
        /*
        if (!match)
        {
            keywords = g_desktop_app_info_get_keywords (app_info);
            int i;
            for (i = 0; keywords && keywords[i]; i++)
            {
                const gchar *tmp = keywords[i];
                match = cdos_pinyin_match (tmp, keyword, 0);
                if (match)
                    break;
            }
        }

        if (!match && description)
            match = cdos_pinyin_match (description, keyword, 1);
        if (!match && filename)
            match = cdos_pinyin_match (filename, keyword, 1);
        if (!match && categories)
            match = cdos_pinyin_match (categories, keyword, 1);
        if (!match && generic_name)
            match = cdos_pinyin_match (generic_name, keyword, 1);
        if (!match && exec)
            match = cdos_pinyin_match (exec, keyword, 0);
        */
        //查找到指定字符串
        if (match)
        {
            results = g_list_prepend (results, app);
        }
        all_apps = all_apps->next;
    }
    results = g_list_reverse (results);
    return results;
}
