/**
 * Copyright (c) 2016-2018 CPU and Fundamental Software Research Center, CAS
 *
 * This software is published by the license of CPU-OS License，you can use and
 * distibuted this software under this License. See CPU-OS Liense for more detail.
 *
 * You should have receive a copy of CPU-OS License. If not, please contact us
 * by email <support_os@cpu-os.ac.cn>
 *
 **/
//
// Created by zhangtao on 18-1-25.
// Email: zhangtao@cpu-os.ac.cn
//
#include <math.h>
#include "common.h"
#include <syslog.h>
#include "cdos-app-system.h"
#include "cdos-drag-drop.h"
#include "cdos-app.h"
#include "cdos-app-system.h"
#include "cdos-app-usage.h"

#include "applet-favorites-manager.h"
#include "applet-favorites-category.h"
#include "applet-menu-helper.h"
#include "applet-menu-types.h"
#include "applet-menu-private.h"
#include "applet-favorites.h"
#include "applet-menu.h"
#include "applet-favorites.h"
#include "applet-application-button.h"

#define APPLET_FAVORITES_MANAGER_GET_PRIVATE(obj)      \
(G_TYPE_INSTANCE_GET_PRIVATE((obj), APPLET_TYPE_FAVORITES_MANAGER, AppletFavoritesManagerPrivate))

G_DEFINE_TYPE(AppletFavoritesManager, applet_favorites_manager, G_TYPE_OBJECT);

struct _AppletFavoritesManagerPrivate
{
    StBoxLayout     *box;
    StScrollView    *scroll_box;
    StBoxLayout     *favorites_box;
    StBoxLayout     *list_box;
    ClutterActor    *groups_none;
    ClutterDropAction   *drop_action;
    StWidget        *preview_holder;
    gboolean        over_category;

    GList           *categories;
    GSettings       *settings;
    AppletFavorites *favorites;
    CdosAppSystem *app_system;
    CdosAppUsage *app_usage;
};

static void applet_favorites_manager_dispose (GObject *object);
static void applet_favorites_manager_finalize (GObject *object);

static void applet_favorites_manager_relayout (AppletFavoritesManager *self);
static void applet_favorites_manager_add_category (AppletFavoritesManager *self,
                                                  AppletFavoritesCategory *category,
                                                  FavoritesSection section);
static void applet_favorites_manager_remove_category (AppletFavoritesManager *self,
                                                     AppletFavoritesCategory *category,
                                                     gboolean destroy);
static AppletFavoritesCategory *
applet_favorites_manager_create_category (AppletFavoritesManager *self,
                                         GroupData *group_data,
                                         gint index);
static void menu_settings_on_changed (GSettings *settings,
                                     gchar *key,
                                     AppletFavoritesManager *self);
static void
favorites_on_app_added (AppletFavorites *favorites,
                       GroupData *group_data,
                       const gchar *app_id,
                       AppletFavoritesManager *self);
static void
favorites_on_app_removed (AppletFavorites *favorites,
                         GroupData *group_data,
                         const gchar *app_id,
                         AppletFavoritesManager *self);
static void
favorites_on_app_moved (AppletFavorites *favorites,
                       GroupData *source_group_data,
                       GroupData *dest_group_data,
                       const gchar *app_id,
                       AppletFavoritesManager *self);
static void
favorites_on_group_added (AppletFavorites *favorites,
                         GroupData *group_data,
                         AppletFavoritesManager *self);
static void
favorites_on_group_removed (AppletFavorites *favorites,
                         GroupData *group_data,
                         AppletFavoritesManager *self);
static void
favorites_on_group_label_changed (AppletFavorites *favorites,
                                 GroupData *group_data,
                                 AppletFavoritesManager *self);
static void
favorites_on_group_moved (AppletFavorites *favorites,
                         AppletFavoritesManager *self);
static void
category_on_title_changed (AppletCategoryBase *category,
                          const gchar *new_title,
                          AppletFavoritesManager *self);

static void
box_add_drop_action (AppletFavoritesManager *self, StBoxLayout *box, ClutterDropAction *drop_action);

static gboolean
action_on_can_drop (ClutterDropAction *action,
               ClutterActor      *actor,
               gfloat             event_x,
               gfloat             event_y,
               gpointer           user_data);
static void
action_on_drop (ClutterDropAction *action,
               ClutterActor      *actor,
               gfloat             event_x,
               gfloat             event_y,
               gpointer           user_data);
static void
action_on_drop_cancel (ClutterDropAction *action,
               ClutterActor      *actor,
               gfloat             event_x,
               gfloat             event_y,
               gpointer           user_data);
static gchar *
get_data_from_drag_action (const gchar *key);
static gboolean
main_box_on_key_press (ClutterActor *actor, ClutterEvent *event, gpointer user_data);
static void
applet_favorites_manager_hide_preview_holder (AppletFavoritesManager *self);
static gboolean
applet_favorites_manager_is_accept_drop (AppletFavoritesManager *self, gfloat y);
static void
stage_on_key_focus_changed  (GObject    *gobject,
                            GParamSpec  *pspec,
                            gpointer    user_data);
static void
scroll_bar_on_changed (GObject     *object,
                      GParamSpec  *pspec,
                      gpointer user_data);

static GroupData *
create_new_group_at_position (AppletFavoritesManager *self,
                              const gchar *group_label,
                              FavoritesSection section,
                              gint position)
{
    AppletFavoritesManagerPrivate *priv;
    GroupData *group_data = NULL;

    priv = self->priv;
    group_data = applet_favorites_get_group_by_label (priv->favorites, group_label);
    /*
    if (!applet_favorites_query_group_exist_by_label (priv->favorites, group_label))
    */
    if (!group_data)
    {
        group_data = applet_favorites_add_group_at_position (priv->favorites,
                                                             group_label,
                                                             position);

        applet_favorites_set_group_label_translatable (priv->favorites,
                                                       group_data->group_name,
                                                       "yes");

        applet_favorites_set_group_section (priv->favorites,
                                            group_data->group_name,
                                            section);
        GList *l;
        if (position == 1)
            l = cdos_app_usage_get_list (priv->app_usage);
        while (l)
        {
            const gchar *app_id = NULL;
            if (position == 1)
            {
                CdosApp *app = l->data;
                app_id = cdos_app_get_id (app);
            }
            if (app_id)
                applet_favorites_add_app (priv->favorites, group_data->group_name, app_id, FALSE);
            l = l->next;
        }
    }

    return group_data;
}

static void
applet_favorites_manager_frequently_used (AppletFavoritesManager *self)
{
    AppletFavoritesManagerPrivate *priv = self->priv;
    if (applet_favorites_query_group_exist_by_label (priv->favorites,
                                                     "Frequently used applications"))
    {
        applet_favorites_remove_group_by_label (priv->favorites,
                                                "Frequently used applications");
    }
    create_new_group_at_position (self,
                                  "Frequently used applications",
                                  FAVORITES_SECTION_SECONDARY,
                                  1);
}

static void
on_usage_changed_cb (CdosAppUsage *app_usage,
                     gpointer user_data)
{
    AppletFavoritesManager *self = (AppletFavoritesManager *) user_data;
    applet_favorites_manager_frequently_used (self);
}

static void
applet_favorites_manager_init (AppletFavoritesManager *self)
{
    AppletFavoritesManagerPrivate *priv;
    priv = self->priv = APPLET_FAVORITES_MANAGER_GET_PRIVATE (self);
    priv->settings = g_settings_new (CDOS_MENU_SCHEMA);

    StTextureCache *texture_cache;
    texture_cache = st_texture_cache_get_default();

    g_signal_connect (priv->settings,
                      "changed",
                      G_CALLBACK (menu_settings_on_changed),
                      self);

    priv->box = ST_BOX_LAYOUT (st_box_layout_new ());
    st_box_layout_set_vertical (priv->box, TRUE);
    st_widget_set_style_class_name ((StWidget *)priv->box, "menu-favorites-box");

    priv->groups_none = g_object_new (ST_TYPE_WIDGET,
                                      "x-expand", TRUE,
                                      "x-align", CLUTTER_ACTOR_ALIGN_FILL,
                                      "y-align", CLUTTER_ACTOR_ALIGN_START,
                                      "style-class", "menu-results-none",
                                      NULL);
    StLabel *label = ST_LABEL(st_label_new (_("Your favorites is empty,you can add some frequently used applications.")));
    menu_container_add_actor (priv->groups_none, (ClutterActor *)label,
                             "x-expand", TRUE,
                             "x-align", CLUTTER_ACTOR_ALIGN_FILL,
                             NULL);
    _add_actor (priv->box, priv->groups_none);
    clutter_actor_hide (priv->groups_none);

    priv->scroll_box = g_object_new (ST_TYPE_SCROLL_VIEW,
                                "style-class" , "vfade menu-favorites-scrollbox",
								"x-fill", TRUE,
								"y-fill", FALSE,
								"y-align", CLUTTER_ACTOR_ALIGN_START,
                                "x-align", CLUTTER_ACTOR_ALIGN_FILL,
                                NULL);

    st_scroll_view_set_policy (priv->scroll_box, GTK_POLICY_NEVER, GTK_POLICY_AUTOMATIC);
    ClutterActor *scroll_bar;
    StAdjustment *adjustment;
    scroll_bar = st_scroll_view_get_vscroll_bar (priv->scroll_box);
    adjustment = st_scroll_bar_get_adjustment ((StScrollBar *)scroll_bar);
    g_signal_connect (adjustment,
                     "notify::value",
                     G_CALLBACK (scroll_bar_on_changed),
                     self);

    priv->favorites_box = ST_BOX_LAYOUT (st_box_layout_new ());
    clutter_actor_set_reactive ((ClutterActor *)priv->favorites_box, TRUE);
    st_widget_set_style_class_name ((StWidget *)priv->favorites_box, "menu-favorites-list");
    priv->drop_action = (ClutterDropAction *) clutter_drop_action_new ();
    box_add_drop_action (self, priv->favorites_box, priv->drop_action);

    menu_container_add_actor ((ClutterActor *)priv->box, (ClutterActor *)priv->scroll_box,
                             "x-expand", TRUE,
                             "x-align", CLUTTER_ACTOR_ALIGN_FILL,
                             NULL);
    /*menu_container_add_actor ((ClutterActor *)priv->box, (ClutterActor *)priv->favorites_box,
                             "x-expand", TRUE,
                             "x-align", CLUTTER_ACTOR_ALIGN_FILL,
                             NULL);*/
    st_scroll_view_add_child (priv->scroll_box, (ClutterActor *)priv->favorites_box);

    priv->list_box = ST_BOX_LAYOUT (st_box_layout_new ());
    st_box_layout_set_vertical (priv->list_box, TRUE);
    st_widget_set_style_class_name ((StWidget *)priv->list_box, "menu-favorites-list-box");
    menu_container_add_actor ((ClutterActor *)priv->favorites_box, (ClutterActor *)priv->list_box,
                             "x-expand", FALSE,
                             "y-align", CLUTTER_ACTOR_ALIGN_START,
                             "x-align", CLUTTER_ACTOR_ALIGN_FILL,
                             NULL);
    /*
    CdosGlobal *global = cdos_global_get();
    ClutterStage *global_stage = cdos_global_get_stage (global);
    g_signal_connect (global_stage,
					 "notify::key-focus",
					 G_CALLBACK (stage_on_key_focus_changed),
					 self);
    */
    priv->favorites = applet_favorites_get_default ();
    priv->app_system = cdos_app_system_get_default ();
    priv->app_usage = cdos_app_usage_get_default ();
    applet_favorites_manager_frequently_used (self);
    GList *groups = applet_favorites_get_groups (priv->favorites);
    if (groups)
    {
        gint index;
        for (index = 0; groups != NULL; groups = groups->next)
        {
            GroupData *group = groups->data;
            AppletFavoritesCategory *c = applet_favorites_manager_create_category (self, group, index);
            if (c != NULL)
            {
                index++;
            }
        }
    }
    if (priv->categories == NULL)
    {
        clutter_actor_show (priv->groups_none);
    }

    g_signal_connect (priv->favorites,
                     "app-added",
                     G_CALLBACK (favorites_on_app_added),
                     self);
    g_signal_connect (priv->favorites,
                     "app-removed",
                     G_CALLBACK (favorites_on_app_removed),
                     self);
    g_signal_connect (priv->favorites,
                     "app-moved",
                     G_CALLBACK (favorites_on_app_moved),
                     self);
    g_signal_connect (priv->favorites,
                     "group-added",
                     G_CALLBACK (favorites_on_group_added),
                     self);
    g_signal_connect (priv->favorites,
                     "group-removed",
                     G_CALLBACK (favorites_on_group_removed),
                     self);
    g_signal_connect (priv->favorites,
                     "group-label-changed",
                     G_CALLBACK (favorites_on_group_label_changed),
                     self);
    g_signal_connect (priv->favorites,
                     "group-moved",
                     G_CALLBACK (favorites_on_group_moved),
                     self);
    /*
    StBoxLayout *main_box = applet_menu_get_main_box ();
    g_signal_connect (main_box,
					 "key-press-event",
					 G_CALLBACK (main_box_on_key_press),
					 self);
    */
    g_signal_connect (priv->app_usage,
                      "usage-changed",
                      G_CALLBACK (on_usage_changed_cb),
                      self);
    g_signal_connect (texture_cache,
                      "icon-theme-changed",
                      G_CALLBACK (on_usage_changed_cb),
                      self);
    priv->preview_holder = st_box_layout_new ();
    st_widget_set_style_class_name ((StWidget *)priv->preview_holder, "menu-preview-holder");
    priv->over_category = FALSE;
}

static void
applet_favorites_manager_class_init (AppletFavoritesManagerClass *klass)
{
    GObjectClass *gobject_class = G_OBJECT_CLASS (klass);
    g_type_class_add_private (klass, sizeof (AppletFavoritesManagerPrivate));

    gobject_class->dispose = applet_favorites_manager_dispose;
    gobject_class->finalize = applet_favorites_manager_finalize;

}

static void
applet_favorites_manager_dispose (GObject *object)
{
    AppletFavoritesManager *self = APPLET_FAVORITES_MANAGER (object);
    AppletFavoritesManagerPrivate *priv = self->priv;
    StBoxLayout *main_box = applet_menu_get_main_box ();
    g_signal_handlers_disconnect_by_data ((ClutterActor *)main_box, self);
    CdosGlobal *global = cdos_global_get();
    ClutterStage *global_stage = cdos_global_get_stage (global);
    g_signal_handlers_disconnect_by_data (global_stage, self);

    clutter_actor_clear_actions ((ClutterActor *)priv->favorites_box);
    priv->drop_action = NULL;
    if (priv->settings)
    {
        g_clear_object (&priv->settings);
    }
    if (priv->favorites)
    {
        g_clear_object (&priv->favorites);
    }
    if (priv->categories)
    {
        g_list_free_full (priv->categories, (GDestroyNotify )clutter_actor_destroy);
    }
    if (priv->box)
    {
        clutter_actor_destroy ((ClutterActor *)priv->box);
    }
    if (priv->preview_holder)
    {
        clutter_actor_destroy ((ClutterActor *)priv->preview_holder);
    }
    G_OBJECT_CLASS (applet_favorites_manager_parent_class)->dispose (object);
}

static void
applet_favorites_manager_finalize (GObject *object)
{
    G_OBJECT_CLASS (applet_favorites_manager_parent_class)->finalize (object);
}

static void
applet_favorites_manager_relayout (AppletFavoritesManager *self)
{
    AppletFavoritesManagerPrivate *priv = self->priv;
    gint index = 0;
    GList *children = NULL;
    children = priv->categories;
    //remove all children, but don't destroy them;
    while (children != NULL)
    {
        AppletFavoritesCategory *category = children->data;
        applet_favorites_manager_remove_category (self, category, FALSE);
        children = children->next;
    }
    /*
    if (clutter_actor_get_parent ((ClutterActor *)priv->preview_holder))
    {
        applet_favorites_manager_remove_category (self, (ClutterActor *)priv->preview_holder, FALSE);
    }
    */
    //add all child;
    children = NULL;
    children = priv->categories;
    while (children != NULL)
    {
        AppletFavoritesCategory *category = children->data;
        applet_favorites_manager_add_category (self, category, FAVORITES_SECTION_NONE);
        index ++;
        children = children->next;
    }
}

static void
category_on_delete (ClutterActor *category, AppletFavoritesManager *self)
{
    AppletFavoritesManagerPrivate *priv = self->priv;
    const gchar* group_name = applet_category_base_get_name ((AppletCategoryBase *)category);
    applet_favorites_remove_group (priv->favorites, group_name);
}

static void
applet_favorites_manager_add_category (AppletFavoritesManager *self,
                                       AppletFavoritesCategory *category,
                                       FavoritesSection section)
{
    AppletFavoritesManagerPrivate *priv = self->priv;
    StBoxLayout *actor = NULL;
    const gchar *title;
    if (section == FAVORITES_SECTION_NONE)
    {
        gchar *s = g_object_get_data ((GObject *)category, "section");
        section = (int)g_strtod (s, NULL);
    }

    actor = priv->list_box;
    title = applet_category_base_get_title ((AppletCategoryBase *) category);

    //设置“最近添加”和“最常使用”在快捷面板的最上面显示
    if (g_strcmp0 (title, _("Newly installed applications")) == 0)
    {
        if (!g_list_find (priv->categories, category))
        {
            priv->categories = g_list_insert (priv->categories, category, 0);
        }
        menu_container_add_actor_at_index ((ClutterActor *)actor,
                                           (ClutterActor *)category,
                                           0,
                                           "x-expand", TRUE,
                                           "x-align", CLUTTER_ACTOR_ALIGN_FILL,
                                           "y-align", CLUTTER_ACTOR_ALIGN_START,
                                           NULL);
        clutter_actor_set_child_at_index ((ClutterActor *)actor, (ClutterActor *)category, 0);
        //“最近添加”分组标题不可编辑修改
        applet_category_base_set_can_edit_title ((AppletCategoryBase *) category, FALSE);
        applet_category_base_set_can_delete ((AppletCategoryBase *) category, FALSE);
        applet_category_base_set_app_button_can_drop ((AppletCategoryBase *) category, TRUE);
    }
    else if (g_strcmp0 (title, _("Frequently used applications")) == 0)
    {
        if (!g_list_find (priv->categories, category))
            priv->categories = g_list_insert (priv->categories, category, 1);
        menu_container_add_actor_at_index ((ClutterActor *)actor,
                                           (ClutterActor *)category,
                                           1,
                                           "x-expand", TRUE,
                                           "x-align", CLUTTER_ACTOR_ALIGN_FILL,
                                           "y-align", CLUTTER_ACTOR_ALIGN_START,
                                           NULL);
        clutter_actor_set_child_at_index ((ClutterActor *)actor, (ClutterActor *)category, 0);
        //“最常使用”分组标题不可编辑修改
        applet_category_base_set_can_edit_title ((AppletCategoryBase *) category, FALSE);
        applet_category_base_set_can_delete ((AppletCategoryBase *) category, FALSE);
        applet_category_base_set_app_button_can_drop ((AppletCategoryBase *) category, TRUE);
    } else {
        if (!g_list_find (priv->categories, category))
        {
            priv->categories = g_list_append (priv->categories, category);
        }
        menu_container_add_actor ((ClutterActor *)actor, (ClutterActor *)category,
                                  "x-expand", TRUE,
                                  "x-align", CLUTTER_ACTOR_ALIGN_FILL,
                                  "y-align", CLUTTER_ACTOR_ALIGN_START,
                                  NULL);
        g_signal_connect (category,
                          "delete",
                          G_CALLBACK (category_on_delete),
                          self);
    }
}

static void
applet_favorites_manager_remove_category (AppletFavoritesManager *self,
                                          AppletFavoritesCategory *category,
                                          gboolean destroy)
{
    AppletFavoritesManagerPrivate *priv = self->priv;
    ClutterActor *parent = clutter_actor_get_parent ((ClutterActor *)category);
    if (destroy)
    {
        priv->categories = g_list_remove (priv->categories, category);
        //clutter_actor_destroy ((ClutterActor *)category);
    } else {
        g_object_ref (category);
    }
    if (parent)
        clutter_actor_remove_child ((ClutterActor *)parent, (ClutterActor *)category);
}

static AppletFavoritesCategory *
applet_favorites_manager_get_category_by_name (AppletFavoritesManager *self, const gchar *group_name)
{
    AppletFavoritesManagerPrivate *priv = self->priv;
    GList *categories = priv->categories;
    while (categories != NULL)
    {
        AppletFavoritesCategory *category = categories->data;
        const gchar *name = applet_category_base_get_name ((AppletCategoryBase *)category);
        if (group_name && g_strcmp0 (name, group_name) == 0)
        {
            return category;
        }
        categories = categories->next;
    }
    return NULL;
}

static AppletFavoritesCategory *
applet_favorites_manager_create_category (AppletFavoritesManager *self,
                                         GroupData *group_data,
                                         gint index)
{
    AppletFavoritesManagerPrivate *priv = self->priv;
    AppletFavoritesCategory *category = NULL;
    GList *new_apps = NULL;
    const gchar *group_name = group_data->group_name;
    new_apps = applet_favorites_get_group_apps (priv->favorites, group_name);

    if (new_apps)
    {
        category = applet_favorites_category_new ((const gchar *)group_name,
                                                  _(group_data->group_label));
        g_object_set_data ((GObject *)category, "favorites-manager", self);
        g_object_set_data_full ((GObject *)category, "section", g_strdup_printf ("%d", group_data->section), g_free);
        //priv->categories = g_list_append (priv->categories, category);
        //The index of last one;
        /*
        if (index == -1)
        {
            index = g_list_length (priv->categories) - 1;
        }
        */
        applet_favorites_manager_add_category (self, category, FAVORITES_SECTION_NONE);
        applet_category_base_set_source_data ((AppletCategoryBase *)category, new_apps);
        g_signal_connect (category,
                         "title-changed",
                         G_CALLBACK (category_on_title_changed),
                         self);
        g_list_free (new_apps);
    }
    return category;
}

static void
menu_settings_on_changed (GSettings *settings,
                         gchar *key,
                         AppletFavoritesManager *self)
{
    AppletFavoritesManagerPrivate *priv = self->priv;
}

static void
favorites_on_app_added (AppletFavorites *favorites,
                       GroupData *group_data,
                       const gchar *app_id,
                       AppletFavoritesManager *self)
{
    GList *new_apps = NULL;
    AppletFavoritesCategory *category = NULL;
    const gchar *group_name = group_data->group_name;

    category = applet_favorites_manager_get_category_by_name (self, group_name);
    if (category)
    {
        new_apps = applet_favorites_get_group_apps (favorites, group_name);
        applet_category_base_source_data_changed ((AppletCategoryBase *)category, new_apps, TRUE);
        g_list_free (new_apps);
    } else {
        applet_favorites_manager_create_category (self, group_data, -1);
    }
}

static void
favorites_on_app_removed (AppletFavorites *favorites,
                         GroupData *group_data,
                         const gchar *app_id,
                         AppletFavoritesManager *self)
{
    GList *new_apps;
    AppletFavoritesCategory *category;
    const gchar *group_name = group_data->group_name;
    category = applet_favorites_manager_get_category_by_name (self, group_name);
    new_apps = applet_favorites_get_group_apps (favorites, group_name);
    if (new_apps && category)
    {
        applet_category_base_source_data_changed ((AppletCategoryBase *)category, new_apps, TRUE);
    }
    g_list_free (new_apps);
}

static void
favorites_on_app_moved (AppletFavorites *favorites,
                       GroupData *source_group_data,
                       GroupData *dest_group_data,
                       const gchar *app_id,
                       AppletFavoritesManager *self)
{
    favorites_on_app_removed (favorites, source_group_data, app_id, self);
    if (source_group_data != dest_group_data)
    {
        favorites_on_app_added (favorites, dest_group_data, app_id, self);
    }
}

static void
favorites_on_group_added (AppletFavorites *favorites,
                         GroupData *group_data,
                         AppletFavoritesManager *self)
{
    AppletFavoritesManagerPrivate *priv = self->priv;
    if (priv->categories == NULL)
    {
        clutter_actor_hide (priv->groups_none);
    }
}

static gboolean
category_delay_destroy (gpointer user_data)
{
    AppletFavoritesCategory *category = (AppletFavoritesCategory *)user_data;
    clutter_actor_destroy ((ClutterActor *)category);
    return FALSE;
}

static void
favorites_on_group_removed (AppletFavorites *favorites,
                         GroupData *group_data,
                         AppletFavoritesManager *self)
{
    AppletFavoritesManagerPrivate *priv = self->priv;
    AppletFavoritesCategory *category;
    gchar *group_name = group_data->group_name;
    category = applet_favorites_manager_get_category_by_name (self, group_name);
    if (category)
    {
        //异步延迟销毁，否则在拖拽过程尚未结束，就进入析构，导致鼠标事件丢失。
        g_object_ref (category);
        applet_favorites_manager_remove_category (self, category, TRUE);
        applet_favorites_manager_relayout (self);
        g_idle_add ((GSourceFunc )category_delay_destroy, category);
    }
    if (priv->categories == NULL)
    {
        clutter_actor_show (priv->groups_none);
    }
}
static void
favorites_on_group_label_changed (AppletFavorites *favorites,
                                 GroupData *group_data,
                                 AppletFavoritesManager *self)
{

    AppletFavoritesCategory *category;
    gchar *group_name = group_data->group_name;
    category = applet_favorites_manager_get_category_by_name (self, group_name);
    applet_category_base_set_title ((AppletCategoryBase *)category, group_data->group_label);
}

static void
favorites_on_group_moved (AppletFavorites *favorites,
                         AppletFavoritesManager *self)
{
    AppletFavoritesManagerPrivate *priv = self->priv;
    GList *groups = applet_favorites_get_groups (priv->favorites);
    GList *new_categories = NULL;
    gint index;
    for (index = 0; groups != NULL; groups = groups->next, index++)
    {
        GroupData *group_data = groups->data;
        gchar *group_name;
        AppletFavoritesCategory *category;
        group_name = group_data->group_name;
        category = applet_favorites_manager_get_category_by_name (self, group_name);
        g_object_set_data_full ((GObject *)category, "section", g_strdup_printf ("%d", group_data->section), g_free);
        new_categories = g_list_append (new_categories, category);
    }
    g_list_free (priv->categories);
    priv->categories = new_categories;
    applet_favorites_manager_relayout (self);
}

static void
category_on_title_changed (AppletCategoryBase *category,
                          const gchar *new_title,
                          AppletFavoritesManager *self)
{
    AppletFavoritesManagerPrivate *priv = self->priv;
    const gchar* group_name = applet_category_base_get_name ((AppletCategoryBase *)category);
    applet_favorites_set_group_label (priv->favorites, group_name, new_title);
}

static void
box_add_drop_action (AppletFavoritesManager *self, StBoxLayout *box, ClutterDropAction *drop_action)
{
    clutter_actor_add_action ((ClutterActor *)box, (ClutterAction *)drop_action);
    g_signal_connect (drop_action,
                    "can-drop",
                    G_CALLBACK (action_on_can_drop),
                    self);
    g_signal_connect (drop_action,
                    "drop",
                    G_CALLBACK (action_on_drop),
                    self);
    g_signal_connect (drop_action,
                    "drop-cancel",
                    G_CALLBACK (action_on_drop_cancel),
                    self);
}

static gboolean
action_on_can_drop (ClutterDropAction *action,
               ClutterActor      *actor,
               gfloat             event_x,
               gfloat             event_y,
               gpointer           user_data)
{
    AppletFavoritesManager *self = APPLET_FAVORITES_MANAGER (user_data);
    gchar             *drag_data = NULL;
    gfloat  my;
    gboolean accept;
    ClutterDragAction *dragging_action =  cdos_get_current_drag_action ();

    drag_data = get_data_from_drag_action ("desktop-id");
    if (drag_data == NULL)
    {
        drag_data = get_data_from_drag_action ("favorites-group-name");
    }
    if (drag_data)
    {
        clutter_drag_action_get_motion_coords (dragging_action, NULL, &my);
        accept = applet_favorites_manager_is_accept_drop (self, my);
        return  accept;
    }
    return FALSE;
}

static void
action_on_drop (ClutterDropAction *action,
               ClutterActor      *actor,
               gfloat             event_x,
               gfloat             event_y,
               gpointer           user_data)
{
    AppletFavoritesManager *self = APPLET_FAVORITES_MANAGER (user_data);
    GroupData *group;
    gchar             *drag_data = NULL;
    gboolean          result;
    FavoritesSection  section = FAVORITES_SECTION_PRIMARY;
    drag_data = get_data_from_drag_action ("desktop-id");
    AppletFavorites *favorites = applet_favorites_get_default ();

    if (drag_data)
    {
        group = applet_favorites_add_group (favorites, FAVORITES_GROUP_DEFAULT_LABEL);
        result = applet_favorites_remove_app (favorites, NULL, drag_data);
        if (result)
        {
            applet_favorites_set_group_section (favorites, group->group_name, section);
            applet_favorites_add_app (favorites, group->group_name, drag_data, TRUE);
        }
    }
    if (drag_data == NULL)
    {
        drag_data = get_data_from_drag_action ("favorites-group-name");
        if (drag_data)
        {
            applet_favorites_group_move (favorites, drag_data, NULL, section, MOVE_AFTER);
        }
    }
    applet_favorites_manager_hide_preview_holder (self);
}

static void
action_on_drop_cancel (ClutterDropAction *action,
               ClutterActor      *actor,
               gfloat             event_x,
               gfloat             event_y,
               gpointer           user_data)
{
    AppletFavoritesManager *self = APPLET_FAVORITES_MANAGER (user_data);
    applet_favorites_manager_hide_preview_holder (self);
}

static void
applet_favorites_manager_hide_preview_holder (AppletFavoritesManager *self)
{
    AppletFavoritesManagerPrivate *priv = self->priv;

    ClutterActor *parent = clutter_actor_get_parent ((ClutterActor *)priv->preview_holder);
    g_object_set_data ((GObject *)priv->preview_holder, "target", NULL);
    g_object_set_data ((GObject *)priv->preview_holder, "section", NULL);
    if (parent == NULL)
    {
        return ;
    }
    applet_favorites_manager_relayout (self);
}

static gfloat
get_children_height (AppletFavoritesManager *self, ClutterActor *actor)
{
    GList *children, *l;
    children = l = clutter_actor_get_children (actor);
    AppletFavoritesManagerPrivate *priv = self->priv;
    gfloat ch = 0;
    while (children != NULL)
    {
        ClutterActor *actor = children->data;
        if (actor != (ClutterActor *)priv->preview_holder)
        {
            gfloat h = clutter_actor_get_height (actor);
            ch += h;
        }
        children = children->next;
    }
    g_list_free (l);
    return ch;
}

static gboolean
applet_favorites_manager_is_accept_drop (AppletFavoritesManager *self, gfloat y)
{
    AppletFavoritesManagerPrivate *priv = self->priv;
    gfloat h, h1, py;
    h = get_children_height (self, (ClutterActor *)priv->list_box);
    clutter_actor_get_transformed_position ((ClutterActor *)priv->list_box, NULL, &py);

    //10作为修正值；
    if (ceil (y) + 10 >= h + py)
    {
        return TRUE;
    }
    return FALSE;
}

static gchar *
get_data_from_drag_action (const gchar *key)
{
    ClutterDragAction *dragging_action;
    gchar             *value = NULL;
    dragging_action =  cdos_get_current_drag_action ();
    if (dragging_action)
    {
        value = g_object_get_data ((GObject *)dragging_action, key);
    }
    return value;
}

static void
get_app_button_col_and_position (AppletFavoritesManager *self,
                                AppletApplicationButton *button,
                                StBoxLayout             *list_actor,
                                gint *col_index,
                                gint *app_position)
{
    GList   *categories, *l;
    gint    col = -1;
    categories = l = clutter_actor_get_children ((ClutterActor *)list_actor);
    while (categories != NULL)
    {
        AppletFavoritesCategory *category = categories->data;
        GList *buttons = applet_category_base_get_app_buttons ((AppletCategoryBase *)category);
        gint   index = -1;
        gboolean get = FALSE;
        gint    col_n = applet_category_base_get_default_column_num ((AppletCategoryBase *)category);
        while (buttons != NULL)
        {
            AppletApplicationButton *btn = buttons->data;
            index ++;
            if (index % col_n == 0)
            {
                col ++;
            }
            if (button == btn)
            {
                if (index % col_n == 0 && buttons->next == NULL)
                {
                    //only one element in this column;
                    *app_position = 3;
                }
                else if (index % col_n == 0)
                {
                    //at the first position of this column
                    *app_position = 1;
                }
                else if ((index + 1) % col_n == 0 || buttons->next == NULL)
                {
                    //at the last position of this column
                    *app_position = 2;
                } else {
                    *app_position = 0;
                }

                get = TRUE;
                break ;
            }
            buttons = buttons->next;
        }
        if (get)
        {
            *col_index = col;
            return ;
        }
        categories = categories->next;
    }
    g_list_free (l);
}

static void
set_app_button_focus (AppletFavoritesManager *self,
                      StBoxLayout            *list_actor,
                      gint col_index,
                      guint symbol)
{
    GList *categories, *l;
    gint col = -1;

    categories = l = clutter_actor_get_children ((ClutterActor *)list_actor);
    while (categories != NULL)
    {
        AppletFavoritesCategory *category = categories->data;
        GList *buttons = applet_category_base_get_app_buttons ((AppletCategoryBase *)category);
        gint   index = -1;
        gint    col_n = applet_category_base_get_default_column_num ((AppletCategoryBase *)category);

        while (buttons != NULL)
        {
            AppletApplicationButton *button = buttons->data;
            index ++;
            if (index % col_n == 0)
            {
                col ++;
            }
            if (symbol == CLUTTER_KEY_Right)
            {
                //focus the first one of this column;
                if (col == col_index)
                {
                    clutter_actor_grab_key_focus ((ClutterActor *)button);
                    return ;
                }
            } else if (symbol == CLUTTER_KEY_Left) {
                //focus the last one of this column;
                if (col == col_index && ((index + 1) % col_n == 0 || buttons->next == NULL))
                {
                    clutter_actor_grab_key_focus ((ClutterActor *)button);
                    return ;
                }
            }
            buttons = buttons->next;
        }
        categories = categories->next;
    }
    g_list_free (l);
}

static gboolean
main_box_on_key_press (ClutterActor *actor, ClutterEvent *event, gpointer user_data)
{
    AppletFavoritesManager *self = APPLET_FAVORITES_MANAGER (user_data);
    AppletFavoritesManagerPrivate *priv = self->priv;
    guint symbol;
    ClutterActor *focused;
    ClutterStage *global_stage;
    MenuPane pane = applet_menu_get_current_pane ();
    if (pane != MENU_PANE_FAVORITES)
    {
        return FALSE;
    }

    symbol = clutter_event_get_key_symbol (event);
    CdosGlobal *global = cdos_global_get();
    global_stage = cdos_global_get_stage (global);
    focused = clutter_stage_get_key_focus (CLUTTER_STAGE (global_stage));

    gint col_index, app_position;
    if (symbol == CLUTTER_KEY_Left)
    {
        //暂时留着
    }
    else if (symbol == CLUTTER_KEY_Right)
    {
        //暂时留着
    }

    return FALSE;
}

static void
stage_on_key_focus_changed  (GObject    *gobject,
                            GParamSpec  *pspec,
                            gpointer    user_data)
{
    AppletFavoritesManager *self = APPLET_FAVORITES_MANAGER (user_data);
    AppletFavoritesManagerPrivate *priv = self->priv;
    ClutterActor *focused;
    focused = clutter_stage_get_key_focus (CLUTTER_STAGE (gobject));
    if (APPLET_IS_APPLICATION_BUTTON (focused))
    {
        applet_helper_scroll_to_focused_actor (priv->scroll_box);
    }
}

static void
scroll_bar_on_changed (GObject     *object,
                      GParamSpec  *pspec,
                      gpointer user_data)
{
    AppletFavoritesManager *self = APPLET_FAVORITES_MANAGER (user_data);
    AppletFavoritesManagerPrivate *priv = self->priv;
    GList *categories = priv->categories;
    while (categories)
    {
        AppletFavoritesCategory *category = categories->data;
        applet_category_base_close_all_popup_menu ((AppletCategoryBase *)category);
        categories = categories->next;
    }
}

AppletFavoritesManager *
applet_favorites_manager_new (void)
{
    AppletFavoritesManager *favorites_manager;
    favorites_manager = g_object_new (APPLET_TYPE_FAVORITES_MANAGER, NULL);
    return favorites_manager;
}

StBoxLayout *
applet_favorites_manager_get_box (AppletFavoritesManager *self)
{
    g_return_val_if_fail (APPLET_IS_FAVORITES_MANAGER(self), NULL);
    AppletFavoritesManagerPrivate *priv = self->priv;
    return priv->box;
}

void
applet_favorites_manager_start_key_focus (AppletFavoritesManager *self)
{
    g_return_if_fail (APPLET_IS_FAVORITES_MANAGER(self));
    AppletFavoritesManagerPrivate *priv = self->priv;
    AppletFavoritesCategory *category = NULL;
    if (priv->categories)
    {
        category = (AppletFavoritesCategory *)clutter_actor_get_child_at_index ((ClutterActor *)priv->list_box, 0);
        if (category)
            applet_category_base_buttons_navigate_focus ((AppletCategoryBase *)category, 0);
    }
}
