/**
 * Copyright (c) 2016-2018 CPU and Fundamental Software Research Center, CAS
 *
 * This software is published by the license of CPU-OS License，you can use and
 * distibuted this software under this License. See CPU-OS Liense for more detail.
 *
 * You should have receive a copy of CPU-OS License. If not, please contact us
 * by email <support_os@cpu-os.ac.cn>
 *
 **/
//
// Created by zhangtao on 18-2-24.
// Email: zhangtao@cpu-os.ac.cn
//
#include <glib/gi18n.h>
#include <syslog.h>
#include "applet-favorites.h"
#include "applet-menu-private.h"

G_DEFINE_TYPE (AppletFavorites, applet_favorites, G_TYPE_OBJECT)

#define APPLET_FAVORITES_GET_PRIVATE(obj) (G_TYPE_INSTANCE_GET_PRIVATE ((obj), APPLET_TYPE_FAVORITES, AppletFavoritesPrivate))

struct _AppletFavoritesPrivate {
    AppletParser    *parser;
    GList           *groups;
    CdosAppSystem   *app_system;
    GHashTable      *group_apps;
};

enum {
    APP_ADDED,
    APP_REMOVED,
    APP_MOVED,
    GROUP_ADDED,
    GROUP_LABEL_CHANGED,
    GROUP_REMOVED,
    GROUP_MOVED,
    LAST_SIGNAL
};

static guint signals[LAST_SIGNAL] = {0};
static void
applet_favorites_dispose (GObject *object);
static double
applet_favorites_find_max_group (AppletFavorites *self);
static AppletFavorites *
applet_favorites_new (void);
static void
applet_favorites_load_data (AppletFavorites *self);
static void
applet_favorites_destroy_group_apps (GList *items);
static void
app_system_on_installed_changed (CdosAppSystem   *app_system, CdosApp *app, AppletFavorites *self);

static void
applet_favorites_replace_hash_table_value (GHashTable *hash_table,
                                          gpointer key,
                                          gpointer value)
{
    gpointer tmp = g_hash_table_lookup (hash_table, key);
    //if the first element changed in the GList, the hash tabel value also changed!
    if (tmp != value)
    {
        g_hash_table_steal (hash_table, key);
        g_hash_table_insert (hash_table, key, value);
    }
}

static void
groups_free (gpointer data)
{
    GroupData *group = data;
    g_free (group->group_name);
    g_free (group->group_label);
    g_free (group);
}

static void
applet_favorites_dispose (GObject *object)
{
    AppletFavoritesPrivate *priv;
    priv = APPLET_FAVORITES_GET_PRIVATE (object);
    if (priv->group_apps)
    {
        g_hash_table_destroy (priv->group_apps);
    }
    if (priv->groups)
    {
        g_list_free_full (priv->groups, groups_free);
    }
    if (priv->parser)
    {
        //g_object_unref (priv->parser);
    }
    G_OBJECT_CLASS (applet_favorites_parent_class)->dispose (object);
}

static void
applet_favorites_init (AppletFavorites *self)
{
    AppletFavoritesPrivate *priv;
    priv = APPLET_FAVORITES_GET_PRIVATE (self);
    self->priv = priv;

    priv->group_apps = g_hash_table_new_full (g_direct_hash,
                                              g_direct_equal,
                                              NULL,
                                              (GDestroyNotify) applet_favorites_destroy_group_apps);
    priv->app_system = cdos_app_system_get_default ();
    g_signal_connect (priv->app_system,
					 "removed-changed",
					 G_CALLBACK (app_system_on_installed_changed),
					 self);
    if (g_file_test (FAVORITES_FILE_URI, G_FILE_TEST_EXISTS) == FALSE)
    {   gchar *dirname = g_path_get_dirname (FAVORITES_FILE_URI);
        gchar *cmd = NULL;
        if (g_file_test (dirname, G_FILE_TEST_EXISTS) == FALSE)
        {
            g_mkdir_with_parents (dirname, 0766);
        }
        g_free (dirname);
        cmd = g_strdup_printf ("cp %s %s", FAVORITES_SOURCE, FAVORITES_FILE_URI);
        system(cmd);
        g_free (cmd);
    }
    priv->parser = applet_parser_new (FAVORITES_FILE_URI);
    applet_favorites_load_data (self);
}

static void
applet_favorites_class_init (AppletFavoritesClass *klass)
{
    GObjectClass *gobject_class = G_OBJECT_CLASS (klass);
    g_type_class_add_private (klass, sizeof (AppletFavoritesPrivate));
    gobject_class->dispose = applet_favorites_dispose;

    signals[APP_ADDED] =
        g_signal_new ("app-added",
                     G_TYPE_FROM_CLASS (klass),
                     G_SIGNAL_RUN_LAST,
                     (guint)G_STRUCT_OFFSET (AppletFavoritesClass, app_added),
                     NULL,
                     NULL,
                     g_cclosure_marshal_generic,
                     G_TYPE_NONE,
                     2,
                     G_TYPE_POINTER,
                     G_TYPE_STRING);

    signals[APP_REMOVED] =
        g_signal_new ("app-removed",
                     G_TYPE_FROM_CLASS (klass),
                     G_SIGNAL_RUN_LAST,
                     (guint)G_STRUCT_OFFSET (AppletFavoritesClass, app_removed),
                     NULL,
                     NULL,
                     g_cclosure_marshal_generic,
                     G_TYPE_NONE,
                     2,
                     G_TYPE_POINTER,
                     G_TYPE_STRING);

    signals[APP_MOVED] =
        g_signal_new ("app-moved",
                     G_TYPE_FROM_CLASS (klass),
                     G_SIGNAL_RUN_LAST,
                     (guint)G_STRUCT_OFFSET (AppletFavoritesClass, app_moved),
                     NULL,
                     NULL,
                     g_cclosure_marshal_generic,
                     G_TYPE_NONE,
                     3,
                     G_TYPE_POINTER,
                     G_TYPE_POINTER,
                     G_TYPE_STRING);

    signals[GROUP_ADDED] =
        g_signal_new ("group-added",
                     G_TYPE_FROM_CLASS (klass),
                     G_SIGNAL_RUN_LAST,
                     (guint)G_STRUCT_OFFSET (AppletFavoritesClass, group_added),
                     NULL,
                     NULL,
                     g_cclosure_marshal_generic,
                     G_TYPE_NONE,
                     1,
                     G_TYPE_POINTER);

    signals[GROUP_LABEL_CHANGED] =
        g_signal_new ("group-label-changed",
                     G_TYPE_FROM_CLASS (klass),
                     G_SIGNAL_RUN_LAST,
                     (guint)G_STRUCT_OFFSET (AppletFavoritesClass, group_label_changed),
                     NULL,
                     NULL,
                     g_cclosure_marshal_generic,
                     G_TYPE_NONE,
                     1,
                     G_TYPE_POINTER);

    signals[GROUP_REMOVED] =
        g_signal_new ("group-removed",
                     G_TYPE_FROM_CLASS (klass),
                     G_SIGNAL_RUN_LAST,
                     (guint)G_STRUCT_OFFSET (AppletFavoritesClass, group_removed),
                     NULL,
                     NULL,
                     g_cclosure_marshal_generic,
                     G_TYPE_NONE,
                     1,
                     G_TYPE_POINTER);

    signals[GROUP_MOVED] =
        g_signal_new ("group-moved",
                     G_TYPE_FROM_CLASS (klass),
                     G_SIGNAL_RUN_LAST,
                     (guint)G_STRUCT_OFFSET (AppletFavoritesClass, group_moved),
                     NULL,
                     NULL,
                     g_cclosure_marshal_generic,
                     G_TYPE_NONE,
                     0);
}

static double
applet_favorites_find_max_group (AppletFavorites *self)
{
    AppletFavoritesPrivate *priv;
    priv = self->priv;
    GList *list = priv->groups;
    GroupData *data = NULL;
    double max = -1, tmp = -1;
    gchar *endptr = NULL;
    while (list)
    {
        data = list->data;
        tmp = g_strtod (data->group_name, &endptr);
        if (tmp > max) {
            max = tmp;
        }
        list = list->next;
    }
    return max;
}

static AppletFavorites *
applet_favorites_new (void)
{
    GObject *object = g_object_new (APPLET_TYPE_FAVORITES, NULL);
    return APPLET_FAVORITES (object);
}

static void
applet_favorites_load_data (AppletFavorites *self)
{
    AppletFavoritesPrivate *priv = self->priv;
    xmlNodeSetPtr nodeset;
    xmlXPathObjectPtr result;
    xmlNodePtr label_node;
    GroupData *data;
    int i;
    if (priv->parser == NULL)
    {
        //g_error ("\n%s: the xml parser is NULL!", __func__);
        return;
    }
    result = applet_parser_get_nodes_by_path (priv->parser, "/data/group");
    if (result)
    {
        nodeset = result->nodesetval;
        for (i = 0; i < nodeset->nodeNr; i++)
        {
            xmlNodePtr group = nodeset->nodeTab[i];
            gchar *group_name, *group_label, *translatable;
            gchar *section;

            group_name = applet_parser_get_node_prop (priv->parser,
                                                     group,
                                                     "name");
            section = applet_parser_get_node_prop (priv->parser,
                                                  group,
                                                  "section");
            label_node = group->xmlChildrenNode;
            group_label = applet_parser_get_node_content (priv->parser, label_node);
            //load apps;
            //skip the label node;
            xmlNodePtr item_node = group->xmlChildrenNode->next;
            gchar *origin_group_label = NULL;
            GList *items = NULL;
            while (item_node != NULL)
            {
                gchar *app_id = applet_parser_get_node_content (priv->parser, item_node);
                CdosApp *app = cdos_app_system_lookup_app (priv->app_system, app_id);
                if (app)
                {
                    ItemData *item = g_new0 (ItemData, 1);
                    item->item_node = item_node;
                    item->app_id = (const gchar *)g_strdup (app_id);
                    items = g_list_append (items, item);
                }
                g_free (app_id);
                item_node = item_node->next;
            }
            if (g_strcmp0 (group_label, "") == 0)
            {
                g_free (group_label);
                group_label = g_strdup (FAVORITES_GROUP_DEFAULT_LABEL);
                origin_group_label = g_strdup (group_label);
                applet_parser_set_node_content (priv->parser, label_node, group_label);
            } else {
                translatable = applet_parser_get_node_prop (priv->parser,
                                                            label_node,
                                                            "translatable");
                if (translatable && g_strcmp0 (translatable, "yes") == 0)
                {
                    origin_group_label = g_strdup (group_label);
                    group_label = g_strdup (_(group_label));
                }
                if (translatable)
                    g_free (translatable);
            }
            data = g_new0 (GroupData, 1);
            data->group_node = group;
            data->group_name = g_strdup (group_name);
            data->group_label = g_strdup (group_label);
            data->origin_group_label = origin_group_label;
            data->section = (int)g_strtod (section, NULL);

            priv->groups = g_list_append (priv->groups, data);
            g_hash_table_insert (priv->group_apps, data, items);
            g_free (group_name);
            g_free (group_label);
            g_free (section);
        }
        xmlXPathFreeObject (result);
    }
}

static GroupData *
applet_favorites_get_group_data_by_name (AppletFavorites *self, const gchar *group_name)
{
    AppletFavoritesPrivate *priv = self->priv;
    GList *groups = priv->groups;
    while (groups)
    {
        GroupData *g = groups->data;
        if (g_strcmp0 (g->group_name, group_name) == 0)
        {
            return g;
        }
        groups = groups->next;
    }
    return NULL;
}

static void
applet_favorites_destroy_group_apps (GList *items)
{
    if (items)
        g_list_free_full (items, g_free);
}

static void
app_system_on_installed_changed (CdosAppSystem   *app_system, CdosApp *app, AppletFavorites *self)
{
    AppletFavoritesPrivate *priv = self->priv;
    GList *removed_app = NULL;
    GList *groups = priv->groups;
    while (groups != NULL)
    {
        GroupData *group = groups->data;
        GList *apps = g_hash_table_lookup (priv->group_apps, group);
        while (apps)
        {
            ItemData *item = apps->data;
            CdosApp  *tmp = cdos_app_system_lookup_app (priv->app_system, item->app_id);
            if (tmp == NULL)
            {
                removed_app = g_list_append (removed_app, item);
            }
            apps = apps->next;
        }
        groups = groups->next;
    }

    while (removed_app != NULL)
    {
        ItemData *item = removed_app->data;
        applet_favorites_remove_app (self, NULL, item->app_id);
        removed_app = removed_app->next;
    }
}

AppletFavorites *
applet_favorites_get_default (void)
{
    static AppletFavorites *favorites;
    if (favorites == NULL)
    {
        favorites = applet_favorites_new ();
    }
    return favorites;
}

GList *
applet_favorites_get_groups (AppletFavorites *self)
{
    g_return_val_if_fail (APPLET_IS_FAVORITES (self), NULL);
    AppletFavoritesPrivate *priv = self->priv;
    return priv->groups;
}

GList *
applet_favorites_get_group_apps (AppletFavorites *self,
                                 const gchar *group_name)
{
    g_return_val_if_fail (APPLET_IS_FAVORITES (self), NULL);
    g_return_val_if_fail (group_name != NULL, NULL);
    AppletFavoritesPrivate *priv = self->priv;
    GList *items = NULL, *apps = NULL;
    GroupData *group_data;

    group_data = applet_favorites_get_group_data_by_name (self, group_name);
    items = g_hash_table_lookup (priv->group_apps, group_data);
    while (items != NULL)
    {
        ItemData *item = items->data;
        CdosApp *app = cdos_app_system_lookup_app (priv->app_system, item->app_id);
        if (app)
        {
            apps = g_list_append (apps, app);
        }
        items = items->next;
    }
    return apps;
}

GroupData *
applet_favorites_add_group (AppletFavorites *self, const gchar *group_label)
{
    g_return_val_if_fail (APPLET_IS_FAVORITES (self), NULL);
    g_return_val_if_fail (NULL != group_label, NULL);
    AppletFavoritesPrivate *priv = self->priv;
    gchar *group_name = NULL;
    gchar *section = NULL;
    section = g_strdup_printf ("%d", FAVORITES_SECTION_PRIMARY);
    GroupData *tmp = NULL;
    double max_number = applet_favorites_find_max_group (self);
    group_name = g_strdup_printf ("%.0lf", max_number + 1);

    xmlNodePtr group = applet_parser_add_text_child (priv->parser,
                                                    NULL,
                                                    "group",
                                                    NULL);
    gboolean res = applet_parser_set_node_prop (priv->parser,
                                                group,
                                                "name",
                                                group_name);
    applet_parser_set_node_prop (priv->parser,
                                 group,
                                 "section",
                                 section);
    if (res)
    {
        applet_parser_add_text_child (priv->parser,
                                     group,
                                     "label",
                                     group_label);
        tmp = g_new0 (GroupData, 1);
        tmp->group_name = g_strdup (group_name);
        tmp->group_label = g_strdup (group_label);
        tmp->origin_group_label = g_strdup (group_label);
        tmp->group_node = group;
        tmp->section = FAVORITES_SECTION_PRIMARY;
        priv->groups = g_list_append (priv->groups, tmp);
        g_hash_table_insert (priv->group_apps, tmp, NULL);
        g_signal_emit (self, signals[GROUP_ADDED], 0, tmp);
    }
    g_free (group_name);
    g_free (section);
    return tmp;
}

//添加组到指定位置
GroupData *
applet_favorites_add_group_at_position (AppletFavorites *self,
                                        const gchar *group_label,
                                        gint position)
{
    g_return_val_if_fail (APPLET_IS_FAVORITES (self), NULL);
    g_return_val_if_fail (NULL != group_label, NULL);
    AppletFavoritesPrivate *priv = self->priv;
    gchar *group_name = NULL;
    gchar *section = NULL;
    section = g_strdup_printf ("%d", FAVORITES_SECTION_PRIMARY);
    GroupData *tmp = NULL;
    double max_number = applet_favorites_find_max_group (self);
    group_name = g_strdup_printf ("%.0lf", max_number + 1);
    xmlNodePtr group = applet_parser_add_text_child (priv->parser,
                                                    NULL,
                                                    "group",
                                                    NULL);
    gboolean res = applet_parser_set_node_prop (priv->parser,
                                                group,
                                                "name",
                                                group_name);
    applet_parser_set_node_prop (priv->parser,
                                 group,
                                 "section",
                                 section);
    if (res)
    {
        applet_parser_add_text_child (priv->parser,
                                     group,
                                     "label",
                                     group_label);
        tmp = g_new0 (GroupData, 1);
        tmp->group_name = g_strdup (group_name);
        tmp->group_label = g_strdup (group_label);
        tmp->origin_group_label = g_strdup (group_label);
        tmp->group_node = group;
        tmp->section = FAVORITES_SECTION_PRIMARY;
        priv->groups = g_list_insert (priv->groups, tmp, position);
        g_hash_table_insert (priv->group_apps, tmp, NULL);
        g_signal_emit (self, signals[GROUP_ADDED], 0, tmp);
    }
    g_free (group_name);
    g_free (section);
    return tmp;
}

gboolean
applet_favorites_remove_group (AppletFavorites *self,
                               const gchar *group_name)
{
    g_return_val_if_fail (APPLET_IS_FAVORITES (self), FALSE);
    g_return_val_if_fail ((NULL != group_name), FALSE);
    AppletFavoritesPrivate *priv = self->priv;
    gboolean res = FALSE;
    GroupData *group_data = applet_favorites_get_group_data_by_name (self, group_name);
    if (group_data)
    {
        xmlNodePtr group_node = group_data->group_node;
        res = applet_parser_remove_node (priv->parser, group_node);
        if (res)
        {
            g_hash_table_remove (priv->group_apps, group_data);
            priv->groups = g_list_remove (priv->groups, group_data);

            g_signal_emit (self, signals[GROUP_REMOVED], 0, group_data);
            groups_free (group_data);
        }
    }
    return res;
}

gboolean
applet_favorites_remove_group_by_label (AppletFavorites *self,
                                        const gchar *group_label)
{
    g_return_val_if_fail (APPLET_IS_FAVORITES (self), FALSE);
    AppletFavoritesPrivate *priv = self->priv;
    gboolean res = FALSE;
    GroupData *group_data = applet_favorites_get_group_by_label (self, group_label);
    if (group_data)
    {
        xmlNodePtr group_node = group_data->group_node;
        res = applet_parser_remove_node (priv->parser, group_node);
        if (res)
        {
            g_hash_table_remove (priv->group_apps, group_data);
            priv->groups = g_list_remove (priv->groups, group_data);

            g_signal_emit (self, signals[GROUP_REMOVED], 0, group_data);
            groups_free (group_data);
        }
    }
    return res;
}

gboolean
applet_favorites_set_group_label (AppletFavorites *self,
                                  const gchar *group_name,
                                  const gchar *group_label)
{
    g_return_val_if_fail (APPLET_IS_FAVORITES (self), FALSE);
    g_return_val_if_fail ((group_name != NULL), FALSE);
    g_return_val_if_fail ((group_label != NULL), FALSE);
    AppletFavoritesPrivate *priv = self->priv;
    gboolean res = FALSE;
    GroupData *group_data = NULL;
    group_data = applet_favorites_get_group_data_by_name (self, group_name);
    if (group_data && g_strcmp0 (group_data->group_label, group_label))
    {
        xmlNodePtr label_node = group_data->group_node->xmlChildrenNode;
        gchar *gl = g_markup_escape_text (group_label, -1);
        res = applet_parser_set_node_content (priv->parser, label_node, gl);
        g_free (gl);
    }
    if (res)
    {
        gchar *old_label = group_data->group_label;
        group_data->group_label = g_strdup (group_label);
        g_signal_emit (self,  signals[GROUP_LABEL_CHANGED], 0, group_data);
        //last step;
        g_free (old_label);
    }
    return res;
}

gchar *
applet_favorites_get_group_label (AppletFavorites *self,
                                  gchar *group_name)
{
    g_return_val_if_fail (APPLET_IS_FAVORITES (self), NULL);
    g_return_val_if_fail (NULL != group_name, NULL);
    GroupData *group;
    group = applet_favorites_get_group_data_by_name (self, group_name);
    if (group)
    {
        return g_strdup (group->group_label);
    }
    return NULL;
}

gboolean
applet_favorites_add_app (AppletFavorites *self,
                          const gchar *group_name,
                          const gchar *app_id,
                          gboolean is_head)
{
    g_return_val_if_fail (APPLET_IS_FAVORITES (self), FALSE);
    g_return_val_if_fail (app_id != NULL, FALSE);
    g_return_val_if_fail (group_name != NULL, FALSE);

    AppletFavoritesPrivate *priv = self->priv;
    xmlNodePtr group_node = NULL;
    GList *items;
    CdosApp *app = cdos_app_system_lookup_app (priv->app_system, app_id);
    if (app == NULL)
    {
        return FALSE;
    }

    GroupData *group_data = applet_favorites_get_group_data_by_name (self, group_name);
    if (group_data)
    {
        group_node = group_data->group_node;
        xmlNodePtr new_node = applet_parser_add_text_child (priv->parser,
                                                        group_node,
                                                        "item",
                                                        app_id);
        items = g_hash_table_lookup (priv->group_apps, group_data);
        if (is_head && items != NULL)
        {
            ItemData *header = items->data;
            new_node = applet_parser_move_to_before (priv->parser, new_node, header->item_node);
        }
        if (new_node)
        {
            ItemData *item = g_new0 (ItemData, 1);
            item->item_node = new_node;
            item->app_id = (const gchar *)g_strdup (app_id);
            GList *items = g_hash_table_lookup (priv->group_apps, group_data);
            if (is_head)
                items = g_list_prepend (items, item);
            else
                items = g_list_append (items, item);
            applet_favorites_replace_hash_table_value (priv->group_apps, group_data, items);
            g_signal_emit (self, signals[APP_ADDED], 0, group_data, item->app_id);
            return TRUE;
        }
    }
    return FALSE;
}

gboolean
applet_favorites_remove_app (AppletFavorites *self,
                             gchar *group_name,
                             const gchar *app_id)
{
    g_return_val_if_fail (APPLET_IS_FAVORITES (self), FALSE);
    g_return_val_if_fail (app_id != NULL, FALSE);

    AppletFavoritesPrivate *priv = self->priv;
    GroupData *group_data;
    gboolean res = FALSE;
    xmlNodePtr item_node;
    ItemData *item_data;
    GList *items;
    item_data = applet_favorites_get_item_by_app_id (self, app_id);
    item_node = item_data->item_node;
    if (group_name)
    {
        group_data = applet_favorites_get_group_data_by_name (self, group_name);
    } else {
        group_data = applet_favorites_get_group_by_app_id (self, app_id);
    }
    res = applet_parser_remove_node (priv->parser, item_node);
    if (res)
    {
        GList *tmp = g_hash_table_lookup (priv->group_apps, group_data);
        items = g_list_remove (tmp, item_data);
        applet_favorites_replace_hash_table_value (priv->group_apps, group_data, items);
        //更新缓存数据，如果全部移除，则移除组；
        if (items)
        {
            // don't use the g_hash_table_replace. it will call the destroy functions, which can't be destroyed.
            g_signal_emit (self, signals[APP_REMOVED], 0, group_data, app_id);
        } else {
            applet_favorites_remove_group (self, group_data->group_name);
        }
        g_free (item_data);
    }
    return res;
}

gboolean
applet_favorites_group_move (AppletFavorites *self,
                             const gchar *source_group_name,
                             const gchar *dest_group_name,
                             FavoritesSection section,
                             MoveMode mode)
{
    g_return_val_if_fail (APPLET_IS_FAVORITES (self), FALSE);
    g_return_val_if_fail (source_group_name != NULL, FALSE);
    if (g_strcmp0 (source_group_name, dest_group_name) == 0)
    {
        return FALSE;
    }

    AppletFavoritesPrivate *priv;
    priv = self->priv;
    GList *source_list, *dest_list;
    GroupData *source, *dest;
    xmlNodePtr new_node;
    gboolean res;
    gchar *ss = g_strdup_printf ("%d", section);
    source = applet_favorites_get_group_data_by_name (self, source_group_name);
    if (dest_group_name)
    {
        dest = applet_favorites_get_group_data_by_name (self, dest_group_name);
        dest_list = g_list_find (priv->groups, dest);
    } else {
        if (mode == MOVE_AFTER)
        {
            dest_list = g_list_last (priv->groups);
        } else {
            dest_list = g_list_first (priv->groups);
        }
        dest = dest_list->data;
    }
    source_list = g_list_find (priv->groups, source);
    res = applet_parser_set_node_prop (priv->parser, source->group_node, "section", ss);
    g_free (ss);
    if (mode == MOVE_AFTER)
    {
        new_node = applet_parser_move_to_after (priv->parser, source->group_node, dest->group_node);
        //see g_list_insert_before: if the last one , the sibling is null;
        dest_list = dest_list->next;
    } else {
        new_node = applet_parser_move_to_before (priv->parser, source->group_node, dest->group_node);
    }
    if (new_node)
    {
        res = TRUE;
        source->group_node = new_node;
        priv->groups = g_list_remove_link (priv->groups, source_list);
        priv->groups = g_list_insert_before (priv->groups, dest_list, source);
    }
    if (res)
    {
        source->section = section;
        g_signal_emit (self, signals[GROUP_MOVED], 0);
        return TRUE;
    }
    return FALSE;
}

gboolean
applet_favorites_app_move (AppletFavorites *self,
                           const gchar   *source_app_id,
                           const gchar   *dest_app_id,
                           MoveMode    mode)
{
    g_return_val_if_fail (APPLET_IS_FAVORITES (self), FALSE);
    g_return_val_if_fail (source_app_id != NULL, FALSE);
    g_return_val_if_fail (dest_app_id != NULL, FALSE);

    AppletFavoritesPrivate *priv = self->priv;
    ItemData *source_item, *dest_item;
    GroupData *source_data, *dest_data;
    GList     *dest_link, *source_items, *dest_items;
    xmlNodePtr new_node;

    source_item = applet_favorites_get_item_by_app_id (self, source_app_id);
    dest_item = applet_favorites_get_item_by_app_id (self, dest_app_id);
    if (dest_item == NULL || source_item == NULL || g_strcmp0 (source_app_id, dest_app_id) == 0)
    {
        return FALSE;
    }
    if (mode == MOVE_AFTER)
    {
        if (source_item->item_node->prev == dest_item->item_node)
        {
            return FALSE;
        }
        new_node = applet_parser_move_to_after (priv->parser, source_item->item_node, dest_item->item_node);
    } else {
        if (source_item->item_node->next == dest_item->item_node)
        {
            return FALSE;
        }
        new_node = applet_parser_move_to_before (priv->parser, source_item->item_node, dest_item->item_node);
    }

    if (new_node)
    {
        source_data = applet_favorites_get_group_by_app_id (self, source_app_id);
        dest_data = applet_favorites_get_group_by_app_id (self, dest_app_id);
        source_items = g_hash_table_lookup (priv->group_apps, source_data);
        //更新缓存数据：
        source_item->item_node = new_node;
        GList *new_source_items, *new_dest_items;
        //for source group;
        new_source_items = g_list_remove (source_items, source_item);
        applet_favorites_replace_hash_table_value (priv->group_apps, source_data, new_source_items);
        //for dest group: "dest_items"的数据必须放在source更新完成之后获取，防止source和dest是同一组产生的问题。
        dest_items = g_hash_table_lookup (priv->group_apps, dest_data);
        dest_link = g_list_find (dest_items, dest_item);
        if (mode == MOVE_AFTER)
        {
            dest_link = dest_link->next;
        }
        new_dest_items = g_list_insert_before (dest_items, dest_link, source_item);
        applet_favorites_replace_hash_table_value (priv->group_apps, dest_data, new_dest_items);
        g_signal_emit (self, signals[APP_MOVED], 0, source_data, dest_data, source_app_id);

        if (new_source_items == NULL)
        {
            applet_favorites_remove_group (self, source_data->group_name);
        }
        return TRUE;

    }
    return FALSE;
}

gboolean
applet_favorites_app_query_exists (AppletFavorites *self, CdosApp *app)
{
    g_return_val_if_fail (APPLET_IS_FAVORITES (self), FALSE);
    g_return_val_if_fail (CDOS_IS_APP (app), FALSE);
    const gchar *app_id = cdos_app_get_id (app);
    ItemData *item_data = applet_favorites_get_item_by_app_id (self, app_id);

    return item_data != NULL ? TRUE : FALSE;
}

GroupData *
applet_favorites_get_default_group (AppletFavorites *self)
{
    g_return_val_if_fail (APPLET_IS_FAVORITES (self), NULL);
    AppletFavoritesPrivate *priv = self->priv;
    GList *groups = priv->groups;
    GroupData *default_group = NULL;
    while (groups)
    {
        GroupData *group_data = groups->data;
        if (g_strcmp0 (group_data->group_label, FAVORITES_GROUP_DEFAULT_LABEL) == 0)
        {
            default_group = group_data;
            break ;
        }
        groups = groups->next;
    }
    if (default_group == NULL)
    {
        default_group = applet_favorites_add_group (self, FAVORITES_GROUP_DEFAULT_LABEL);
    }
    return default_group;
}

GroupData *
applet_favorites_get_group_by_app_id (AppletFavorites *self, const gchar *app_id)
{
    g_return_val_if_fail (APPLET_IS_FAVORITES (self), FALSE);
    g_return_val_if_fail (app_id != NULL, FALSE);

    AppletFavoritesPrivate *priv = self->priv;
    GList *groups = priv->groups;
    while (groups != NULL)
    {
        GroupData *data = groups->data;
        GList *items = g_hash_table_lookup (priv->group_apps, data);
        while (items != NULL)
        {
            ItemData *item = items->data;
            if (g_strcmp0 (app_id, item->app_id) == 0)
            {
                return data;
            }
            items = items->next;
        }
        groups = groups->next;
    }
    return NULL;

}

ItemData *
applet_favorites_get_item_by_app_id (AppletFavorites *self, const gchar *app_id)
{
    g_return_val_if_fail (APPLET_IS_FAVORITES (self), NULL);
    g_return_val_if_fail (app_id != NULL, NULL);

    AppletFavoritesPrivate *priv = self->priv;
    GList *groups = priv->groups;
    while (groups != NULL)
    {
        GroupData *group = groups->data;
        GList *items = g_hash_table_lookup (priv->group_apps, group);
        while (items != NULL)
        {
            ItemData *item = items->data;
            if (g_strcmp0 (app_id, item->app_id) == 0)
            {
                return item;
            }
            items = items->next;
        }
        groups = groups->next;
    }
    return NULL;
}

FavoritesSection
applet_favorites_get_group_section (AppletFavorites *self, const gchar *group_name)
{
    //g_return_val_if_fail (APPLET_IS_FAVORITES (self), NULL);
    //g_return_val_if_fail (group_name != NULL, NULL);
    GroupData *group;
    group = applet_favorites_get_group_data_by_name (self, group_name);
    if (group)
    {
        return group->section;
    }
    return FAVORITES_SECTION_NONE;
}

void
applet_favorites_set_group_section (AppletFavorites *self, const gchar *group_name, FavoritesSection section)
{
    g_return_if_fail (APPLET_IS_FAVORITES (self));
    g_return_if_fail (group_name != NULL);

    AppletFavoritesPrivate *priv = self->priv;
    GroupData *group;
    gboolean  res;
    group = applet_favorites_get_group_data_by_name (self, group_name);
    if (group)
    {
        gchar *s = g_strdup_printf ("%d", section);
        res = applet_parser_set_node_prop (priv->parser,
                                          group->group_node,
                                          "section",
                                          s);
        if (res)
        {
            group->section = section;
        }
        g_free (s);
    }
}

GroupData *
applet_favorites_get_group_by_label (AppletFavorites *self, const gchar *group_label)
{
    g_return_val_if_fail (APPLET_IS_FAVORITES (self), NULL);

    GList *groups = NULL;
    GroupData *group;

    //groups = self->priv->groups;
    groups = applet_favorites_get_groups (self);

    while (groups)
    {
        group = groups->data;
        if (g_strcmp0 (group_label, group->group_label) == 0 ||
            g_strcmp0 (group_label, group->origin_group_label) == 0)
        {
            return group;
        }
        groups = groups->next;
    }
    return NULL;
}

//通过label判断group是否存在
gboolean
applet_favorites_query_group_exist_by_label (AppletFavorites *self, const gchar *group_label)
{
    g_return_val_if_fail (APPLET_IS_FAVORITES (self), FALSE);

    GList *groups = NULL;
    GroupData *group;

    groups = applet_favorites_get_groups (self);

    while (groups)
    {
        group = groups->data;
        if (g_strcmp0 (group_label, group->group_label) == 0 ||
            g_strcmp0 (group_label, group->origin_group_label) == 0)
        {
            return TRUE;
        }
        groups = groups->next;
    }
    return FALSE;
}

//通过label获取组名
gchar *
applet_favorites_get_group_name_by_label (AppletFavorites *self, const gchar *group_label)
{
    g_return_val_if_fail (APPLET_IS_FAVORITES (self), NULL);
    g_return_val_if_fail (group_label != NULL, NULL);
    GList *groups = NULL;
    GroupData *group;
    groups = applet_favorites_get_groups (self);
    while (groups)
    {
        group = groups->data;
        if (g_strcmp0 (group_label, group->group_label) == 0 ||
            g_strcmp0 (group_label, group->origin_group_label))
        {
            return group->group_name;
        }
        groups = groups->next;
    }
    return NULL;
}

//设置group label是否可以被汉化，值为"yes"或"no"
gboolean
applet_favorites_set_group_label_translatable (AppletFavorites *self,
                                               const gchar *group_name,
                                               const gchar *translatable)
{
    g_return_val_if_fail (APPLET_IS_FAVORITES (self), FALSE);
    g_return_val_if_fail ((group_name != NULL), FALSE);
    AppletFavoritesPrivate *priv = self->priv;
    gboolean res = FALSE;
    GroupData *group_data = NULL;
    if (g_strcmp0 (translatable, "yes") != 0 &&
        g_strcmp0 (translatable, "no") != 0)
    {
        return FALSE;
    }
    group_data = applet_favorites_get_group_data_by_name (self, group_name);
    xmlNodePtr label_node = group_data->group_node->xmlChildrenNode;
    res = applet_parser_set_node_prop (priv->parser,
                                       label_node,
                                       "translatable",
                                       translatable);
    return res;
}
