/** 
 * Copyright (c) 2016-2018 CPU and Fundamental Software Research Center, CAS
 *  
 * This software is published by the license of CPU-OS License，you can use and 
 * distibuted this software under this License. See CPU-OS Liense for more detail.
 *
 * You should have receive a copy of CPU-OS License. If not, please contact us 
 * by email <support_os@cpu-os.ac.cn>  
 * 
 **/
#include "applet-parser.h"

G_DEFINE_TYPE (AppletParser, applet_parser, G_TYPE_OBJECT)

// clang-format off
#define APPLET_PARSER_GET_PRIVATE(obj) (G_TYPE_INSTANCE_GET_PRIVATE((obj), APPLET_TYPE_PARSER, AppletParserPrivate))
static const gchar *encoding = "UTF-8";
// clang-format on

struct _AppletParserPrivate
{
    const gchar     *file_uri;
    xmlDocPtr       doc;
    xmlNodePtr      root_node;
};

enum {
  PROP_0,
  PROP_FILE_URI
};

// 参数为文件名、编码类型和枚举变量(下面一个或多个的组合）
// XML_PARSE_RECOVER = 1<<0, recover on errors 尝试修复错误语法
// XML_PARSE_NOERROR = 1<<5, suppress error reports 不输出错误日志
// XML_PARSE_NOWARNING = 1<<6, suppress warning reports 不输出警告日志
// XML_PARSE_PEDANTIC = 1<<7, pedantic error reporting 输出详细的错误日志
// XML_PARSE_NOERROR和XML_PARSE_NOWARNING的功能是“当运行命令行程序时不输出错误/警告日志”。
const int options = XML_PARSE_PEDANTIC | XML_PARSE_RECOVER;

static void
applet_parser_set_file_uri (AppletParser *self, const gchar *uri);
static void
applet_parser_finalize (GObject *object);
static xmlXPathObjectPtr
applet_parser_get_node_set (AppletParser *self, const xmlChar *xpath);
static gboolean 
applet_parser_save_doc (AppletParser *self);


static void
applet_parser_set_file_uri (AppletParser *self, const gchar *uri)
{
    AppletParserPrivate *priv = self->priv;
    //this value can be set only once!
    if (priv->file_uri == NULL)
    {
        priv->file_uri = uri;
        if (g_file_test (priv->file_uri, G_FILE_TEST_EXISTS))
        {
            priv->doc = xmlReadFile (priv->file_uri, encoding, options);
            priv->root_node = xmlDocGetRootElement (priv->doc);
        } else {
            //if the xml does't exist, we will create a new one.
        }
    }
}

static void
applet_parser_class_init(AppletParserClass *klass)
{
    GObjectClass *gobject_class = G_OBJECT_CLASS (klass);
    g_type_class_add_private (klass, sizeof (AppletParserPrivate));
    gobject_class->finalize = applet_parser_finalize;
}

static void
applet_parser_init(AppletParser *parser)
{
    AppletParserPrivate *priv;
    parser->priv = priv = APPLET_PARSER_GET_PRIVATE (parser);
    xmlKeepBlanksDefault (0);
}

static void
applet_parser_finalize(GObject *object)
{
    AppletParserPrivate *priv = APPLET_PARSER_GET_PRIVATE (object);
    if (priv->root_node)
    {
        xmlFreeNode (priv->root_node);
    }
    if (priv->doc)
    {
        xmlFreeDoc (priv->doc);
    }
    xmlCleanupParser ();
    G_OBJECT_CLASS (applet_parser_parent_class)->finalize (object);
}

static gboolean 
applet_parser_save_doc (AppletParser *self)
{
    AppletParserPrivate *priv = self->priv;
    if (priv->doc == NULL)
    {
        return FALSE;
    }
    int res = xmlSaveFormatFile (priv->file_uri, priv->doc, 4);
    if (res == -1)
    {
        g_warning ("xml save format file failed!\n");
        return FALSE;
    }
    return TRUE;
}

static xmlXPathObjectPtr
applet_parser_get_node_set (AppletParser *self, const xmlChar *xpath)
{
    AppletParserPrivate *priv = self->priv;
    xmlXPathContextPtr context;
    xmlXPathObjectPtr result = NULL;  // 存储查询结果
    // 创建一个xpath上下文
    context = xmlXPathNewContext (priv->doc);
    if (context)
    {
        // 查询XPath表达式
        result = xmlXPathEvalExpression (xpath, context);
        xmlXPathFreeContext (context);  // 释放上下文
        if (result)
        {
            // 检查结果集是否为空
            if (xmlXPathNodeSetIsEmpty (result->nodesetval))
            {
                xmlXPathFreeObject (result);  // 为空就释放
                return NULL;
            }
        }
    }
    return result;
}
 
gchar *
applet_parser_get_node_prop (AppletParser *self,
                            xmlNodePtr cur_node,
                            const gchar *prop_name)
{
    g_return_val_if_fail (NULL != cur_node, NULL);
    g_return_val_if_fail (NULL != prop_name, NULL);
    
    xmlChar *value = xmlGetProp (cur_node, (const xmlChar *)prop_name);
    return  (gchar *)value;
}

gboolean
applet_parser_set_node_prop (AppletParser *self,
                            xmlNodePtr cur_node,
                            const gchar *prop_name,
                            const gchar *prop_value)
{
    g_return_val_if_fail (NULL != cur_node, FALSE);
    g_return_val_if_fail (NULL != prop_name, FALSE);
    g_return_val_if_fail (NULL != prop_value, FALSE);
    AppletParserPrivate *priv = self->priv;
    if (priv->doc == NULL)
    {
        return FALSE;
    }
    xmlSetProp (cur_node, BAD_CAST prop_name, BAD_CAST prop_value);
    gboolean res = applet_parser_save_doc (self);
    return res;
}

// 获取结点
// prop_name: 属性
// prop_value: 是唯一性属性值
xmlNodePtr
applet_parser_get_node_by_prop (AppletParser *self,
                                const gchar *xpath,
                                const gchar *prop_name,
                                const gchar *prop_value)
{
    g_return_val_if_fail (APPLET_IS_PARSER (self), NULL);
    g_return_val_if_fail (NULL != xpath, NULL);
    g_return_val_if_fail (NULL != prop_name, NULL);
    g_return_val_if_fail (NULL != prop_value, NULL);
    AppletParserPrivate *priv = self->priv;
    
    int index;  // 索引
    xmlXPathObjectPtr result;  // 查询结果集
    xmlNodeSetPtr nodeset;  // 结果集成员变量
    gchar *keyword;
    xmlNodePtr tmp = NULL;
    if (priv->doc == NULL)
    {
        return NULL;
    }
    result = applet_parser_get_node_set (self, (xmlChar *)xpath);
    if (NULL == result)
    {
        return tmp;
    }
    nodeset = result->nodesetval;
    for (index = 0; index < nodeset->nodeNr; index++)
    {
        keyword = applet_parser_get_node_prop (self, nodeset->nodeTab[index], prop_name);
        if (xmlStrcmp ((const xmlChar *)keyword, (const xmlChar *)prop_value) == 0)
        {
            tmp = nodeset->nodeTab[index];
            break;
        }
        xmlFree (keyword);
    }
    xmlXPathFreeObject (result);
    return tmp;
}

gboolean
applet_parser_remove_node (AppletParser *self, xmlNodePtr cur_node)
{   
    g_return_val_if_fail (APPLET_IS_PARSER (self), FALSE);
    g_return_val_if_fail (NULL != cur_node, FALSE);
    gboolean res = FALSE;
    xmlUnlinkNode (cur_node);
    res = applet_parser_save_doc (self);
    if (res)
    {
        xmlFreeNode (cur_node);
    }
    return res;
}

xmlNodePtr
applet_parser_move_to_after (AppletParser *self,
                            xmlNodePtr cur_node,
                            xmlNodePtr dest_node)
{
    g_return_val_if_fail (APPLET_IS_PARSER (self), NULL);
    g_return_val_if_fail (NULL != cur_node, NULL);
    g_return_val_if_fail (NULL != dest_node, NULL);
    xmlNodePtr new_node;
    if (cur_node == dest_node || cur_node->prev == dest_node)
    {
        return NULL;
    }
    xmlUnlinkNode (cur_node);
    new_node = xmlAddNextSibling (dest_node, cur_node);
    if (new_node)
    {
        applet_parser_save_doc (self);
    }
    return new_node;
}

xmlNodePtr
applet_parser_move_to_before (AppletParser *self,
                             xmlNodePtr cur_node,
                             xmlNodePtr dest_node)
{
    g_return_val_if_fail (APPLET_IS_PARSER (self), NULL);
    g_return_val_if_fail (NULL != cur_node, NULL);
    g_return_val_if_fail (NULL != dest_node, NULL);
    xmlNodePtr new_node;
    if (cur_node == dest_node || cur_node->next == dest_node)
    {
        return NULL;
    }
    xmlUnlinkNode (cur_node);
    new_node = xmlAddPrevSibling (dest_node, cur_node);
    if (new_node)
    {
        applet_parser_save_doc (self);
    }
    return new_node;
}

AppletParser *
applet_parser_new (const gchar *file_uri)
{
    if (g_file_test (file_uri, G_FILE_TEST_EXISTS) == FALSE)
    {
        return NULL;
    }
    AppletParser *parser;
    parser = g_object_new (APPLET_TYPE_PARSER, NULL);
    applet_parser_set_file_uri (parser, file_uri);
    return parser;
}

xmlNodePtr
applet_parser_add_text_child (AppletParser *self,
                             xmlNodePtr parent, 
                             const gchar *name, 
                             const gchar *content)
{
    g_return_val_if_fail (APPLET_IS_PARSER (self), NULL);
    AppletParserPrivate *priv = self->priv;
    if (name == NULL)
    {
        return NULL;
    }
    if (parent == NULL)
    {
        parent = priv->root_node;
    }
    xmlNodePtr new_node;
    new_node = xmlNewTextChild (parent, NULL, BAD_CAST name, BAD_CAST content); 
    applet_parser_save_doc (self);
    return new_node;
}

gchar *
applet_parser_get_node_content (AppletParser *self, xmlNodePtr cur_node)
{
    g_return_val_if_fail (APPLET_IS_PARSER (self), NULL);
    g_return_val_if_fail (cur_node != NULL, NULL);
    return (gchar *)xmlNodeGetContent (cur_node);
}

gboolean
applet_parser_set_node_content (AppletParser *self, 
                                xmlNodePtr cur_node, 
                                const gchar *content)
{   
    g_return_val_if_fail (APPLET_IS_PARSER (self), FALSE);
    g_return_val_if_fail (NULL != cur_node, FALSE);

    xmlNodeSetContent (cur_node, BAD_CAST content);
    gboolean res = applet_parser_save_doc (self);
    return res;
}

xmlXPathObjectPtr
applet_parser_get_nodes_by_path (AppletParser *self, const gchar *xpath)
{
    g_return_val_if_fail (APPLET_IS_PARSER (self), NULL);
    g_return_val_if_fail (xpath != NULL, NULL);
    xmlXPathObjectPtr result;
    result = applet_parser_get_node_set (self, (xmlChar *)xpath);
    return result;
}
