/*
 * gcc apt-polkit.c `pkg-config --cflags --libs glib-2.0 polkit-gobject-1 gtk+-3.0`
 */
#include "config.h"
#include <stdlib.h>
#include <errno.h>
#include <glib.h>
#include <glib/gi18n-lib.h>
#include <gio/gdesktopappinfo.h>
#include <gtk/gtk.h>

static volatile gint apt_state;
static const gchar *css =
    ".uninstall-window {\n \
  background-color: #fff; \n \
}\n \
.uninstall-window .titlebar { \n \
  min-height: 32px;\n \
}\n\
.uninstall-status-center {\n\
  font-size: 18px;\n\
}\n\
progress, trough {\n\
  min-height: 12px;\n\
}\n\
.uninstall-bottom-box {\n \
  background-color: #f4f4f4;\n \
  padding: 16px; \n \
  border-top: 1px solid #dbdbdb; \n \
}\n\
.uninstall-bottom-box button{\n\
  min-width:64px;\n\
}";

static GMainLoop *main_loop;

typedef struct _AptPolkit
{
    /*
    PolkitAuthority *authority;
    */
    gchar *desktop_file;
    gchar *app_name;
    gchar *display_name;
    GDBusConnection *connection;
    GDBusProxy *transaction;
    GPid child_pid;
    int std_out;
    int std_err;
    guint watch_id;

    GtkWidget *window;
    GtkWidget *header;
    GtkWidget *vbox;
    GtkWidget *bottom_hbox;
    GtkWidget *close_button;
    GtkWidget *confirm_button;
    GtkWidget *status_label;
    GtkWidget *progress_bar;
    gchar *icon;
    GIcon *gicon;
} AptPolkit;

static void
apt_free (AptPolkit *apt)
{
    g_clear_object (&apt->connection);
    g_clear_object (&apt->transaction);
    /* g_clear_object (&apt->authority); */
    g_free (apt->desktop_file);
    g_free (apt->app_name);
    g_free (apt->display_name);

    if (apt->watch_id)
        g_source_remove (apt->watch_id);

    if (apt->window)
    {
        g_signal_handlers_disconnect_by_data (apt->window, apt);
        gtk_widget_destroy (apt->window);
    }

    g_atomic_int_set (&apt_state, 0);

    g_main_loop_quit (main_loop);
}

static void
remove_finished (GObject *source_object, GAsyncResult *res, gpointer user_data)
{
    GError *error = NULL;
    AptPolkit *apt = user_data;

    /*
    g_dbus_connection_call_finish (apt->connection, res, &error);
    */
    g_dbus_proxy_call_finish (apt->transaction, res, &error);
    if (error != NULL)
    {
        g_warning ("%s", error->message);
        g_clear_error (&error);
    }
}

static void
parse_transaction_props (GVariant *parameters, AptPolkit *apt)
{
    GVariant *v;
    gchar *prop;
    const gchar *value;
    gint code;

    g_variant_get (parameters, "(sv)", &prop, &v);

    if (g_strcmp0 (prop, "Status") == 0)
    {
        value = g_variant_get_string (v, NULL);
        g_message ("Uninstalling %s changed: %s", prop, value);
    }
    else if (g_strcmp0 (prop, "Progress") == 0)
    {
        code = g_variant_get_int32 (v);
        gtk_progress_bar_set_fraction ((GtkProgressBar *)apt->progress_bar, code / 100.0);
    }
    else if (g_strcmp0 (prop, "StatusDetails") == 0)
    {
        value = g_variant_get_string (v, NULL);
        g_message ("Uninstalling %s changed: %s", prop, value);
    }
    else if (g_strcmp0 (prop, "ExitState") == 0)
    {
        gchar *status = NULL;
        value = g_variant_get_string (v, NULL);
        g_message ("Uninstalling %s changed: %s", prop, value);
        if (g_strcmp0 (value, "exit-failed") == 0)
        {
            status = g_strdup_printf (_ ("Uninstalled %s failed"), apt->display_name);
        }
        else if (g_strcmp0 (value, "exit-success") == 0)
        {
            status = g_strdup_printf (_ ("Uninstalled %s finished"), apt->display_name);
        }
        if (status != NULL)
        {
            gtk_widget_show (apt->close_button);
            gtk_widget_set_sensitive (apt->close_button, TRUE);
            gtk_widget_hide (apt->progress_bar);
            gtk_label_set_text ((GtkLabel *)apt->status_label, status);
            gtk_widget_set_halign (apt->status_label, GTK_ALIGN_CENTER);
            GtkStyleContext *style_context = gtk_widget_get_style_context (apt->status_label);
            gtk_style_context_add_class (style_context, "uninstall-status-center");
            g_free (status);
        }
    }
    g_free (prop);
    g_variant_unref (v);
}

static void
transaction_signal_cb (GDBusProxy *proxy,
                       const gchar *sender,
                       const gchar *signal,
                       GVariant *parameters,
                       AptPolkit *apt)
{
    if (g_strcmp0 (signal, "Finished") == 0)
    {
        gchar *exit_state;
        g_variant_get (parameters, "(s)", &exit_state);
        g_message ("Remove finished state: %s", exit_state);
        g_free (exit_state);
    }
    else if (g_strcmp0 (signal, "PropertyChanged") == 0)
    {
        parse_transaction_props (parameters, apt);
    }
}

static void
remove_packages_cb (GObject *source_object, GAsyncResult *res, gpointer user_data)
{
    GVariant *ret;
    GError *error = NULL;
    const gchar *transaction;
    AptPolkit *apt = user_data;

    ret = g_dbus_connection_call_finish (apt->connection, res, &error);
    if (error != NULL)
    {
        g_warning ("%s", error->message);
        g_clear_error (&error);
        apt_free (apt);
        return;
    }
    g_variant_get (ret, "(s)", &transaction);
    if (transaction)
    {
        apt->transaction = \
            g_dbus_proxy_new_sync (apt->connection, 
                                   0, 
                                   NULL, 
                                   "org.debian.apt", 
                                   transaction, 
                                   "org.debian.apt.transaction", 
                                   NULL, 
                                   &error);
        if (error != NULL)
        {
            g_warning ("%s", error->message);
            g_clear_error (&error);
            g_variant_unref (ret);
            apt_free (apt);
            return;
        }
        g_dbus_proxy_call (apt->transaction, 
                           "Run", 
                           NULL, 
                           G_DBUS_CALL_FLAGS_NONE, 
                           -1, 
                           NULL, 
                           remove_finished, 
                           apt);
        g_signal_connect (apt->transaction, 
                          "g-signal", 
                          G_CALLBACK (transaction_signal_cb), 
                          apt);
        g_free (transaction);
    }
    g_variant_unref (ret);
}

static void
remove_package (AptPolkit *apt)
{
    GVariantBuilder *builder;
    builder = g_variant_builder_new (G_VARIANT_TYPE ("as"));
    g_variant_builder_add (builder, "s", apt->app_name);
    g_message ("Uninstall package: %s", apt->app_name);
    g_dbus_connection_call (apt->connection,
                            "org.debian.apt",
                            "/org/debian/apt",
                            "org.debian.apt",
                            "RemovePackages",
                            g_variant_new ("(as)", builder),
                            G_VARIANT_TYPE ("(s)"),
                            G_DBUS_CALL_FLAGS_NONE,
                            -1,
                            NULL,
                            remove_packages_cb,
                            apt);
    g_variant_builder_unref (builder);
}

static void
read_fd (int fd, char *buf, size_t size)
{
    if (!fd)
        return;
    while (1)
    {
        int res = read (fd, buf, size);
        if (res > 0)
        {
            buf += res;
            size -= res;
        }
        else if (res == -1)
        {
            if (errno == EINTR || errno == EAGAIN)
                continue;
            else
                break;
        }
        else
            break;
    }
}

static void
watch_child (GPid child_pid, gint status, gpointer user_data)
{
    GError *error = NULL;
    gchar std_out[128] = {0};
    gchar std_err[128] = {0};

    AptPolkit *apt = user_data;

    if (status != 0)
    {
        if (apt->std_err)
        {
            read_fd (apt->std_err, std_err, sizeof (std_err));
            g_warning ("%d: %s", apt->std_err, std_err);
        }
        goto finish;
    }

    if (apt->std_out)
        read_fd (apt->std_out, std_out, sizeof (std_out));

    gchar *p = std_out;
    while (p && p[0])
    {
        if (p[0] == ':')
        {
            p[0] = '\0';
            break;
        }
        p++;
    }
    apt->app_name = g_strdup (std_out);
    apt->connection = g_bus_get_sync (G_BUS_TYPE_SYSTEM, NULL, &error);
    if (error)
    {
        g_warning ("%s", error->message);
        g_clear_error (&error);
        goto finish;
    }

    // check_authorization (apt);
    remove_package (apt);

finish:
    close (apt->std_out);
    close (apt->std_err);
    g_spawn_close_pid (apt->child_pid);
    apt->watch_id = 0;
    return;
}

static void
apply_css (GtkWidget *widget, GtkStyleProvider *provider)
{
    gtk_style_context_add_provider (gtk_widget_get_style_context (widget), provider, G_MAXUINT);
    if (GTK_IS_CONTAINER (widget))
        gtk_container_forall (GTK_CONTAINER (widget), (GtkCallback)apply_css, provider);
}

static void
window_on_destroyed (GtkWidget *widget, AptPolkit *apt)
{
    gtk_widget_destroyed (widget, &apt->window);
    apt_free (apt);
}

static void
close_button_on_clicked (GtkWidget *widget, AptPolkit *apt)
{
    apt_free (apt);
}

static void
confirm_button_on_clicked (GtkWidget *widget, AptPolkit *apt)
{
    if (apt->watch_id)
        return;

    GError *error = NULL;
    const gchar *args[] = {"/usr/bin/dpkg", "-S", apt->desktop_file, NULL};
    g_message ("Uninstall package from file: %s", apt->desktop_file);
    if (!g_spawn_async_with_pipes (NULL,
                                   (gchar **)args,
                                   NULL,
                                   G_SPAWN_DO_NOT_REAP_CHILD,
                                   NULL,
                                   NULL,
                                   &apt->child_pid,
                                   NULL,
                                   &apt->std_out,
                                   &apt->std_err,
                                   &error))
    {
        g_warning ("%s", error->message);
        g_clear_error (&error);
        apt_free (apt);
        return;
    }

    gtk_widget_hide (apt->confirm_button);
    gtk_widget_set_sensitive (apt->close_button, FALSE);
    gtk_button_set_label (apt->close_button, _("_Close"));

    apt->watch_id = g_child_watch_add (apt->child_pid, watch_child, apt);
}

static void
create_main_window (GAppInfo *app_info)
{
    const gchar *desktop_file = g_desktop_app_info_get_filename ((GDesktopAppInfo *)app_info);
    AptPolkit *apt;

    g_return_if_fail (desktop_file != NULL);

    apt = g_new0 (AptPolkit, 1);
    apt->desktop_file = g_strdup (desktop_file);
    apt->gicon = g_app_info_get_icon ((GAppInfo *)app_info);
    apt->display_name = g_strdup (g_app_info_get_display_name (app_info));

    GtkStyleContext *style_context;
    GtkStyleProvider *provider;

    gtk_init (NULL, NULL);

    gchar *title = NULL;
    apt->window = gtk_window_new (GTK_WINDOW_TOPLEVEL);
    gtk_window_set_position ((GtkWindow *)apt->window, GTK_WIN_POS_CENTER);
    gtk_window_set_default_size ((GtkWindow *)apt->window, 450, 240);
    gtk_window_set_skip_pager_hint ((GtkWindow *)apt->window, TRUE);
    //gtk_window_set_skip_taskbar_hint ((GtkWindow *)apt->window, TRUE);
    gtk_window_set_resizable ((GtkWindow *)apt->window, FALSE);
    style_context = gtk_widget_get_style_context (apt->window);
    gtk_style_context_add_class (style_context, "uninstall-window");

    g_signal_connect (apt->window, 
                      "destroy", 
                      G_CALLBACK (window_on_destroyed), 
                      apt);

    apt->header = gtk_header_bar_new ();
    title = g_strdup_printf (_("Uninstalling %s"), apt->display_name);
    gtk_header_bar_set_title ((GtkHeaderBar *)apt->header, title);
    gtk_header_bar_set_show_close_button ((GtkHeaderBar *)apt->header, FALSE);
    gtk_header_bar_set_decoration_layout ((GtkHeaderBar *)apt->header, NULL);
    gtk_window_set_titlebar ((GtkWindow *)apt->window, apt->header);
    apt->vbox = gtk_box_new (GTK_ORIENTATION_VERTICAL, 0);
    gtk_container_add ((GtkContainer *)apt->window, apt->vbox);
    g_free (title);

    GtkWidget *vbox = gtk_box_new (GTK_ORIENTATION_VERTICAL, 18);
    gtk_widget_set_margin_start (vbox, 32);
    gtk_widget_set_margin_end (vbox, 32);
    style_context = gtk_widget_get_style_context (vbox);
    gtk_style_context_add_class (style_context, "uninstall-centerbox");
    gtk_box_pack_start ((GtkBox *)apt->vbox, vbox, TRUE, FALSE, 0);

    apt->status_label = gtk_label_new (_("Are you sure uninstall this application?"));
    gtk_label_set_xalign ((GtkLabel *)apt->status_label, 0);
    gtk_widget_set_valign (apt->status_label, GTK_ALIGN_CENTER);

    apt->progress_bar = gtk_progress_bar_new ();
    gtk_widget_set_valign (apt->progress_bar, GTK_ALIGN_CENTER);
    gtk_box_pack_start ((GtkBox *)vbox, apt->status_label, FALSE, FALSE, 0);
    gtk_box_pack_start ((GtkBox *)vbox, apt->progress_bar, FALSE, FALSE, 0);

    apt->bottom_hbox = gtk_box_new (GTK_ORIENTATION_HORIZONTAL, 16);
    /*
    gtk_widget_set_margin_start (apt->bottom_hbox, 16);
    gtk_widget_set_margin_end (apt->bottom_hbox, 16);
    gtk_widget_set_margin_bottom (apt->bottom_hbox, 16);
    */
    style_context = gtk_widget_get_style_context (apt->bottom_hbox);
    gtk_style_context_add_class (style_context, "uninstall-bottom-box");
    gtk_container_add ((GtkContainer *)apt->vbox, apt->bottom_hbox);

    GtkWidget *image = gtk_image_new_from_gicon (apt->gicon, GTK_ICON_SIZE_LARGE_TOOLBAR);
    gtk_box_pack_start ((GtkBox *)apt->bottom_hbox, image, FALSE, FALSE, 0);

    apt->confirm_button = gtk_button_new_with_label (_ ("_Uninstall"));
    gtk_button_set_use_underline (apt->confirm_button, TRUE);
    g_signal_connect (apt->confirm_button, 
                      "clicked", 
                      G_CALLBACK (confirm_button_on_clicked), 
                      apt);
    gtk_box_pack_end ((GtkBox *)apt->bottom_hbox, apt->confirm_button, FALSE, FALSE, 0);

    apt->close_button = gtk_button_new_with_label (_ ("_Cancel"));
    gtk_button_set_use_underline (apt->close_button, TRUE);
    g_signal_connect (apt->close_button, 
                      "clicked", 
                      G_CALLBACK (close_button_on_clicked), 
                      apt);
    gtk_box_pack_end ((GtkBox *)apt->bottom_hbox, apt->close_button, FALSE, FALSE, 0);

    style_context = gtk_widget_get_style_context (apt->confirm_button);
    gtk_style_context_add_class (style_context, GTK_STYLE_CLASS_SUGGESTED_ACTION);
    gtk_window_set_focus ((GtkWindow *)apt->window, apt->confirm_button);

    provider = (GtkStyleProvider *)gtk_css_provider_new ();
    gtk_css_provider_load_from_data (GTK_CSS_PROVIDER (provider), css, -1, NULL);
    apply_css (apt->window, provider);

    gtk_widget_show_all (apt->window);

    g_object_unref (app_info);
}

static gboolean
apt_polkit_remove_package (const char *desktop_file)
{
    g_return_val_if_fail (desktop_file != NULL, -1);
    GDesktopAppInfo *app_info = g_desktop_app_info_new (desktop_file);
    if (!app_info)
    {
        g_message ("The desktop file '%s' does not exist", desktop_file);
        return FALSE;
    }
    create_main_window ((GAppInfo *)app_info);
    return TRUE;
}

int
main (int argc, char **argv)
{
    if (argc < 2)
    {
        g_warning ("argc must > 1");
        return 0;
    }

    bindtextdomain (GETTEXT_PACKAGE, LOCALEDIR);
    textdomain (GETTEXT_PACKAGE);

    main_loop = g_main_loop_new (NULL, FALSE);

    apt_polkit_remove_package (argv[1]);

    g_main_loop_run (main_loop);

    return 0;
}
