#include <string.h>
#include <gio/gio.h>
#include "cdos-util.h"
#include "applet-nm.h"
#include "applet-nm-device.h"
#include "applet-nm-wired-section-title-menu-item.h"
#include "applet-nm-menu-item.h"
#include "applet-nm-connection.h"

#include "applet-nm-private.h"

typedef struct _NMDeviceTable
{
    gpointer delegate;
    gpointer device;
} NMDeviceTable;

gint applet_nm_device_connection_sort_function (gpointer one, gpointer two);
static void applet_nm_device_sub_state_changed (NMDevice *device, gpointer psp, AppletNMDevice *self);

static void
applet_nm_device_state_changed (NMDevice *device, gint newstate, gint oldstate, guint reason, gpointer user_data);

static void applet_nm_device_update_status_item (AppletNMDevice *self);

static CdosPopupMenuItem *applet_nm_device_create_connection_item (AppletNMDevice *self, NMConnection *ob);

static void applet_nm_device_create_active_connection_item (AppletNMDevice *self);

static void applet_nm_device_add_connection (AppletNMDevice *self, NMConnection *conn);

static void applet_nm_device_create_section (AppletNMDevice *self);
static void applet_nm_device_clear_section (AppletNMDevice *self);

static gboolean
applet_nm_device_should_show_connection_item (AppletNMDevice *self)
{
    return (nm_device_get_state (self->device) >= NM_DEVICE_STATE_DISCONNECTED);
}

void
applet_nm_device_destroy (AppletNMDevice *self)
{
    AppletNMDeviceClass *klass = self->klass;
    klass->destroy (self);
}

void
applet_nm_device_destroy_real (AppletNMDevice *self)
{
    if (self->device)
    {
        g_object_set_data ((GObject *)self->device, "delegate", NULL);
        self->device = NULL;
    }
    applet_nm_device_clear_section (self);
    if (self->status_item)
    {
        cdos_popup_base_menu_item_destroy (CDOS_POPUP_BASE_MENU_ITEM (self->status_item));
        self->status_item = NULL;
    }

    g_slist_free (self->connections);
    cdos_popup_menu_base_destroy (CDOS_POPUP_MENU_BASE (self->section));
    g_free (self);
}

void
applet_nm_device_deactivate (AppletNMDevice *self)
{
    AppletNMDeviceClass *klass = self->klass;
    klass->deactivate (self);
}

static void
applet_nm_device_deactivate_real (AppletNMDevice *self)
{
    NMDeviceState state = nm_device_get_state (self->device);
    if (state == NM_DEVICE_STATE_ACTIVATED)
        nm_device_disconnect (self->device, NULL, NULL);
}

void
applet_nm_device_activate (AppletNMDevice *self)
{
    AppletNMDeviceClass *klass = self->klass;
    klass->activate (self);
}

static void
activate_connection_cb (GObject *source,
                        GAsyncResult *res,
                        gpointer user_data)
{
    GError *error = NULL;
    AppletNMDevice *self = APPLET_NM_DEVICE (user_data);
    nm_client_activate_connection_finish (self->client, res, &error);
    if (error)
    {
        g_warning ("%s", error->message);
        g_error_free (error);
    }
}

static void
add_and_activate_connection_cb (GObject *source,
                                GAsyncResult *res,
                                gpointer user_data)
{
    GError *error = NULL;
    AppletNMDevice *self = APPLET_NM_DEVICE (user_data);
    nm_client_add_and_activate_connection_finish (self->client, res, &error);
    if (error)
    {
        g_warning ("%s", error->message);
        g_error_free (error);
    }
}

static void
applet_nm_device_activate_real (AppletNMDevice *self)
{
    if (self->active_connection)
        return;
    if (self->connections)
    {
        NMConnection *conn = self->connections->data;
        nm_client_activate_connection_async (self->client, conn,
                                             self->device, NULL, NULL,
                                             activate_connection_cb, self);
    }
    else if (self->auto_connection_name)
    {
        NMConnection *connection = nm_simple_connection_new ();
        if (NULL != connection)
        {
            nm_client_add_and_activate_connection_async (self->client,
                                                         connection,
                                                         self->device,
                                                         NULL, NULL,
                                                         add_and_activate_connection_cb, self);
        }
    }
}

gboolean
applet_nm_device_get_connected (AppletNMDevice *self)
{
    return (nm_device_get_state (self->device) == NM_DEVICE_STATE_ACTIVATED);
}

void
applet_nm_device_set_active_connection (AppletNMDevice *self, NMActiveConnection *activeConnection)
{
    if (activeConnection == self->active_connection)
        return;

    if (self->active_connection_item)
    {
        cdos_popup_base_menu_item_destroy (self->active_connection_item);
        self->active_connection_item = NULL;
    }
    self->active_connection = activeConnection;
    applet_nm_device_clear_section (self);
    applet_nm_device_create_section (self);
}

void
applet_nm_device_check_connection (AppletNMDevice *self, NMConnection *connection)
{
    gint exists = applet_nm_device_find_connection (self, nm_connection_get_uuid (connection));
    gboolean valid = nm_device_connection_valid (self->device, connection);

    if (exists != -1 && !valid)
    {
        g_assert_not_reached ();
        applet_nm_device_remove_connection (self, connection);
    }
    else if (exists == -1 && valid)
    {
        applet_nm_device_add_connection (self, connection);
    }
    else if (exists != -1 && valid)
    {
        NMSettingConnection *s_con = nm_connection_get_setting_connection (connection);
        g_message ("FIXME: Update for connection: %s", nm_setting_connection_get_id (s_con));
    }
}

static void
applet_nm_device_add_connection (AppletNMDevice *self, NMConnection *conn)
{
    self->klass->add_connection (self, conn);
}

void
applet_nm_device_remove_connection (AppletNMDevice *self, NMConnection *conn)
{
    self->klass->remove_connection (self, conn);
}

static void
applet_nm_device_add_connection_real (AppletNMDevice *self, NMConnection *conn)
{
    self->connections = g_slist_prepend (self->connections, conn);
    self->connections = g_slist_sort (self->connections, (GCompareFunc)applet_nm_device_connection_sort_function);
    applet_nm_device_clear_section (self);
    applet_nm_device_create_section (self);
}

static void
applet_nm_device_remove_connection_real (AppletNMDevice *self, NMConnection *conn)
{
    gint pos = applet_nm_device_find_connection (self, nm_connection_get_uuid (conn));
    NMConnection *bj;
    if (pos == -1)
        return;

    bj = g_slist_nth_data (self->connections, pos);
    AppletNMConnection *table = applet_nm_connection_get_data (bj);

    if (table->item)
    {
        cdos_popup_base_menu_item_destroy (CDOS_POPUP_BASE_MENU_ITEM (table->item));
    }

    self->connections = g_slist_remove (self->connections, conn);

    int len = g_slist_length (self->connections);

    if (len <= 1 || len == NUM_VISIBLE_NETWORKS)
    {
        applet_nm_device_clear_section (self);
        applet_nm_device_create_section (self);
    }
}

gint
applet_nm_device_connection_sort_function (gpointer one, gpointer two)
{
    NMSettingConnection *s_con0 = nm_connection_get_setting_connection (one);
    NMSettingConnection *s_con1 = nm_connection_get_setting_connection (two);
    guint64 time0 = nm_setting_connection_get_timestamp (s_con0);
    guint64 time1 = nm_setting_connection_get_timestamp (s_con1);

    if (time0 == time1)
        return g_utf8_collate (nm_setting_connection_get_id (s_con0), nm_setting_connection_get_id (s_con1));

    return time0 - time1;
}

void
applet_nm_device_set_enabled (AppletNMDevice *device, gboolean enabled)
{
    // do nothing by default, we want to keep the connection list visible
    // in the majority of cases (wired, wwan, vpn)
    return;
}

gchar *
applet_nm_device_get_status_label (AppletNMDevice *self)
{
    g_return_val_if_fail (NM_IS_DEVICE (self->device), NULL);

    switch (nm_device_get_state (self->device))
    {
        case NM_DEVICE_STATE_DISCONNECTED:
            return _("off");
        case NM_DEVICE_STATE_ACTIVATED:
            return _("connected");
        case NM_DEVICE_STATE_UNMANAGED:
            return _("unmanaged");
        case NM_DEVICE_STATE_DEACTIVATING:
            return _("disconnecting");
        case NM_DEVICE_STATE_PREPARE:
        case NM_DEVICE_STATE_CONFIG:
        case NM_DEVICE_STATE_IP_CONFIG:
        case NM_DEVICE_STATE_IP_CHECK:
        case NM_DEVICE_STATE_SECONDARIES:
            return _("connecting...");

        case NM_DEVICE_STATE_NEED_AUTH:
            return _("authentication required");

        case NM_DEVICE_STATE_UNAVAILABLE:
            if(!self->firmware_missing_id)
                self->firmware_missing_id = g_signal_connect(self->device,"notify::firware-missing",
                                                            G_CALLBACK(applet_nm_device_sub_state_changed), self);
            if(nm_device_get_firmware_missing(self->device))
                return _("firware missing");
            if(nm_device_get_capabilities (self->device) & NM_DEVICE_CAP_CARRIER_DETECT)
            {
                if(!self->carried_changed_id)
                    self->carried_changed_id = g_signal_connect(self->device,"notify::carrier",G_CALLBACK(applet_nm_device_sub_state_changed), self);
                if (!nm_device_ethernet_get_carrier ((NMDeviceEthernet *)self->device))
                    return _("cable unplugged");
            }
            return _("unavailable");
        case NM_DEVICE_STATE_FAILED:
            return _("connection failed");
        default:
            g_warning ("Device state invalid , is %d",nm_device_get_state (self->device));
            return _("invalid");
    }
}

NMConnection *
applet_nm_device_create_auto_connection (AppletNMDevice *self)
{
    AppletNMDeviceClass *klass = self->klass;
    return klass->create_auto_connection (self);
}

gint
applet_nm_device_find_connection (AppletNMDevice *self, const char *uuid)
{
    gint i = 0;
    GSList *iter = NULL;
    for (iter = self->connections; iter; iter = iter->next)
    {
        NMConnection *conn = iter->data;
        const char *conn_uuid = nm_connection_get_uuid (conn);
        if (conn_uuid && strcmp (conn_uuid, uuid) == 0)
            return i;
        i++;
    }
    return -1;
}

void
applet_nm_device_clear_section_real (AppletNMDevice *self)
{
    GList *items = cdos_popup_menu_base_get_menu_items (CDOS_POPUP_MENU_BASE (self->section));
    GList *item;
    for (item = items; item; item = item->next)
        cdos_popup_base_menu_item_destroy (item->data);
    g_list_free (items);

    self->auto_connection_name = NULL;
    self->active_connection_item = NULL;
    self->overflow_item = NULL;
    GSList *connections = self->connections;
    GSList *iter = NULL;
    for (iter = connections; iter; iter = iter->next)
    {
        NMConnection *connection = iter->data;
        AppletNMConnection *conn = applet_nm_connection_get_data (connection);
        conn->item = NULL;
    }
}

/*
static gboolean
applet_nm_device_should_show_connection_list (AppletNMDevice *self)
{
    return (nm_device_get_state (self->device) >= NM_DEVICE_STATE_DISCONNECTED);
}
*/

static void
auto_connection_item (CdosPopupMenuItem *_autoConnectionItem, gpointer user_data)
{
    AppletNMDevice *self = APPLET_NM_DEVICE (user_data);

    NMConnection *connection = nm_simple_connection_new ();
    if (NULL != connection)
    {
        nm_client_add_and_activate_connection_async (self->client,
                                                     connection,
                                                     self->device,
                                                     NULL, NULL,
                                                     add_and_activate_connection_cb, self);
    }
}

static void
applet_nm_device_create_section (AppletNMDevice *self)
{
    self->active_connection = nm_device_get_active_connection (self->device);
    self->klass->create_section (self);
}

static void
applet_nm_device_clear_section (AppletNMDevice *self)
{
    self->klass->clear_section (self);
}

void
applet_nm_device_create_section_real (AppletNMDevice *self)
{
    if (!(applet_nm_device_should_show_connection_item (self)))
        return;

    if (self->active_connection)
    {
        applet_nm_device_create_active_connection_item(self);
        cdos_popup_menu_base_add_menu_item( CDOS_POPUP_MENU_BASE(self->section),
                                            CDOS_POPUP_BASE_MENU_ITEM(self->active_connection_item), -1);
    }

    if (self->connections != NULL)
    {
        guint activeOffset = self->active_connection_item ? 1 : 0;
        int j;
        GSList *iter = self->connections;
        NMRemoteConnection *r_con = NULL;
        if (self->active_connection)
        {
            r_con = nm_active_connection_get_connection (self->active_connection);
        }
        for (j = 0; iter; iter = iter->next, j++)
        {
            NMConnection *connection = iter->data;
            if (connection == (NMConnection *)r_con)
                continue;
            AppletNMConnection *ob = applet_nm_connection_get_data (connection);

            ob->item = applet_nm_device_create_connection_item (self, connection);

            if ((j + activeOffset) >= NUM_VISIBLE_NETWORKS)
            {
                if (!self->overflow_item)
                {
                    self->overflow_item = cdos_popup_submenu_menu_item_new (_("More..."), FALSE);
                    cdos_popup_menu_base_add_menu_item ((CdosPopupMenuBase *)self->section,
                                                        (CdosPopupBaseMenuItem *)self->overflow_item, -1);
                }
                CdosPopupMenuBase *menu =
                    (CdosPopupMenuBase *)cdos_popup_submenu_menu_item_get_menu (self->overflow_item);
                clutter_actor_set_x_expand ((ClutterActor *)menu, TRUE);
                cdos_popup_menu_base_add_menu_item (menu, (CdosPopupBaseMenuItem *)ob->item, -1);
            }
            else
                cdos_popup_menu_base_add_menu_item (
                    (CdosPopupMenuBase *)self->section, (CdosPopupBaseMenuItem *)ob->item, -1);
        }
    }
    else if (self->auto_connection_name)
    {
        self->auto_connection_item = cdos_popup_menu_item_new_with_params (self->auto_connection_name, NULL);
        // StWidget *actor = (StWidget *)cdos_popup_base_menu_item_get_actor ((CdosPopupBaseMenuItem
        // *)self->auto_connection_item);
        // st_widget_add_style_class_name (actor, "nm-menu-item");
        g_signal_connect (self->auto_connection_item, "activate", G_CALLBACK (auto_connection_item), self);

        cdos_popup_menu_base_add_menu_item (
            (CdosPopupMenuBase *)self->section, (CdosPopupBaseMenuItem *)self->auto_connection_item, -1);
    }
}

typedef struct _AppletNMDeviceData
{
    AppletNMDevice *device;
    NMConnection *connection;
} AppletNMDeviceData;

static void
applet_nm_device_create_connection_item_signal (void *item,
                                                ClutterEvent *e,
                                                gboolean keepmenu,
                                                AppletNMDeviceData *user_data)
{
    AppletNMDevice *self = user_data->device;
    
    nm_client_activate_connection_async (self->client, user_data->connection,
                                         self->device, NULL, NULL,
                                         activate_connection_cb, self);
}

static CdosPopupMenuItem *
applet_nm_device_create_connection_item (AppletNMDevice *self, NMConnection *conn)
{
    NMSettingConnection *s_con = nm_connection_get_setting_connection (conn);
    CdosPopupMenuItem *item = cdos_popup_menu_item_new_with_params(nm_setting_connection_get_id (s_con), NULL);
    cdos_popup_base_menu_item_set_ornament ((CdosPopupBaseMenuItem *)item, CDOS_ORNAMENT_SPACE);

    AppletNMDeviceData *data = g_new0 (AppletNMDeviceData, 1);
    data->device = self;
    data->connection = conn;
    g_signal_connect (item, "activate", G_CALLBACK (applet_nm_device_create_connection_item_signal), data);

    StWidget *actor = (StWidget *)item;
    st_widget_add_style_class_name (actor, "nm-menu-item");

    return item;
}

static void
active_connection_item_signal (void *item, ClutterEvent *e, gboolean keepmenu, AppletNMDevice *user_data)
{
    /*
    AppletNMDevice *self = user_data;
    nm_client_deactivate_connection (self->client,
                                     self->active_connection,
                                     NULL,
                                     NULL);
    */
}

static void
applet_nm_device_create_active_connection_item (AppletNMDevice *self)
{
    const gchar *title;
    NMRemoteConnection *active = nm_active_connection_get_connection (self->active_connection);
    if (active)
    {
        AppletNMConnection *table = applet_nm_connection_get_data ((NMConnection *)active);
        title = table->name;
        if (!title)
            title = nm_connection_get_id ((NMConnection *)active);
    }
    else
    {
        title = _("Connected (private)");
    }

    self->active_connection_item = (CdosPopupBaseMenuItem *)cdos_popup_menu_item_new_with_params (title, NULL);
    g_signal_connect (self->active_connection_item, "activate", G_CALLBACK (active_connection_item_signal), self);

    cdos_popup_base_menu_item_set_ornament (self->active_connection_item, CDOS_ORNAMENT_DOT);
    st_widget_add_style_class_name ((StWidget *)self->active_connection_item, "nm-menu-item");
}

static void
applet_nm_device_state_changed (NMDevice *device,
                                gint newstate,
                                gint oldstate,
                                guint reason, 
                                gpointer user_data)
{
    g_return_if_fail (NM_IS_DEVICE (device));
    AppletNMDevice *self = APPLET_NM_DEVICE (user_data);

    g_return_if_fail (self->device == device);
    if (newstate == oldstate)
    {
        return;
    }
    /*
    if(oldstate == NM_DEVICE_STATE_ACTIVATED){
        g_signal_emit_by_name(self,"network-lost");
    }
    */
    if (newstate == NM_DEVICE_STATE_FAILED)
        applet_nm_activation_failed (self, reason);

    if (reason == NM_DEVICE_STATE_REASON_REMOVED)
        return;

    applet_nm_device_update_status_item (self);

    applet_nm_device_clear_section (self);
    applet_nm_device_create_section (self);
    // g_signal_emit_by_name(self,"state-changed");
    applet_nm_state_changed (self->data, self);
}

static void
applet_nm_device_active_connection_changed (NMDevice *device, GParamSpec *pspec, AppletNMDevice *self)
{
    // g_signal_handler_disconnect (self->active_connection, self->
    NMActiveConnection *connection = nm_device_get_active_connection (self->device);
    applet_nm_device_set_active_connection (self, connection);
}

static void
applet_nm_device_update_status_item (AppletNMDevice *self)
{
    if (self->carried_changed_id)
    {
        g_object_disconnect (self->device, "notify::carrier", G_CALLBACK (applet_nm_device_sub_state_changed), NULL);
        self->carried_changed_id = 0;
    }

    /*
    if (priv->_firmwareChangedId)
    {
        g_object_disconnect(self->device, priv->_firmwareChangedId);
        priv->_firmwareChangedId = 0;
    }
    */
    cdos_popup_switch_menu_item_set_toggle_status (self->status_item, applet_nm_device_get_connected (self));
}

static void
applet_nm_device_sub_state_changed (NMDevice *device, gpointer pspec, AppletNMDevice *self)
{
    // cdos_popup_switch_menu_item_set_status(self->status_item,applet_nm_device_get_status_label(self));
    g_debug ("%s", applet_nm_device_get_status_label (self));
}

static gchar *
applet_nm_device_get_description (AppletNMDevice *self)
{
    /*
    const gchar *dev_product = nm_device_get_product(self->device);
    const gchar *dev_vendor  = nm_device_get_vendor(self->device);

    if (!dev_product || !dev_vendor)
        return NULL;

    char *product = cdos_util_fixup_pci_description (dev_product);
    char *vendor = cdos_util_fixup_pci_description (dev_vendor);

    gchar *out = g_new0 (char, strlen (product) + strlen (vendor) + 2);
    if (strstr (product, vendor) == NULL)
        sprintf (out, "%s ", vendor);
    sprintf (out + strlen (out), "%s", product);
    g_free (product);
    g_free (vendor);

    return out;
    */
    return g_strdup_printf ("%s (%s)", nm_device_get_description (self->device),
                            nm_device_get_iface (self->device));
}

static void applet_nm_device_status_item (CdosPopupSwitchMenuItem *statusItem, gboolean state, gpointer user_data) //...
{
    g_return_if_fail (user_data != NULL);
    AppletNMDevice *self = APPLET_NM_DEVICE (user_data);

    if (state)
        applet_nm_device_activate (self);
    else
        applet_nm_device_deactivate (self);
}

AppletNMDeviceClass *
applet_nm_device_class_new (void)
{
    AppletNMDeviceClass *klass = g_new0 (AppletNMDeviceClass, 1);
    klass->create_auto_connection = NULL; // applet_nm_device_create_auto_connection;
    klass->create_section = applet_nm_device_create_section_real;
    klass->clear_section = applet_nm_device_clear_section_real;
    klass->activate = applet_nm_device_activate_real;
    klass->deactivate = applet_nm_device_deactivate_real;
    klass->add_connection = applet_nm_device_add_connection_real;
    klass->remove_connection = applet_nm_device_remove_connection_real;
    klass->destroy = applet_nm_device_destroy_real;
    return klass;
}

gboolean
applet_nm_device_init (AppletNMDevice *self, 
                       AppletNM *applet,
                       NMDevice *device,
                       const GPtrArray *connections)
{
    self->device = device; 

    if (NM_IS_DEVICE (device))
    {
        g_object_set_data ((GObject *)device, "delegate", self);
        self->state_changed_id = g_signal_connect (self->device, 
                                                   "state-changed",
                                                   G_CALLBACK(applet_nm_device_state_changed),
                                                   self);
        g_signal_connect (self->device,
                          "notify::active-connection",
                          G_CALLBACK (applet_nm_device_active_connection_changed),
                          self);
    }
    else
    {
        self->state_changed_id = 0;
        return FALSE;
    }
    self->data = applet;
    self->client = applet->client;
    self->connections = NULL;
    for (int i = 0; i < connections->len; i++)
    {
        NMConnection *conn = connections->pdata[i];
        if(!nm_connection_get_uuid(conn))
            continue;

        if (!nm_device_connection_valid (self->device,conn))
            continue;
        self->connections = g_slist_prepend (self->connections, conn);
    }

    self->connections = g_slist_sort (self->connections, (GCompareFunc)applet_nm_device_connection_sort_function);

    gchar *description = applet_nm_device_get_description (self);
    self->status_item = cdos_popup_switch_menu_item_new_with_params (description, self->connected, NULL);
    g_free (description);

    st_widget_set_style_class_name (ST_WIDGET(self->status_item),"popup-subtitle-menu-item");
    st_widget_add_style_class_name (ST_WIDGET(self->status_item),"nm-menu-item");
    self->status_changed_id = g_signal_connect(self->status_item,
            "toggled",
            G_CALLBACK(applet_nm_device_status_item),
            self);

    applet_nm_device_update_status_item (self);

    self->section = cdos_popup_menu_section_new ();
    clutter_actor_set_x_expand ((ClutterActor *)self->section, TRUE);
    applet_nm_device_create_section (self);
    return 1;
}


CdosPopupSwitchMenuItem *
applet_nm_device_get_statusItem (AppletNMDevice *self)
{
    return self->status_item;
}

guint8
applet_nm_signal_to_icon (guint8 strength)
{
    if (strength <= 5)
        return 0;
    else
        return (strength - 5) / 25 + 1;
    /*
    if (strength > 80)
        return 100;
    else if (strength > 55)
        return 75;
    else if (strength > 30)
        return 50;
    else if (strength > 5)
        return 25;
    else
        return 0;
    */
}
