#include <string.h>
#include "applet-nm.h"
#include "cdos-popup.h"
#include "cdos-popup-separator-menu-item.h"
#include "cdos-util.h"

#include "applet-nm-connection.h"
#include "applet-nm-wired-section-title-menu-item.h"
#include "applet-nm-wireless-section-menu-item.h"
#include "applet-nm-menu-item.h"
#include "applet-nm-device-wired.h"
#include "applet-nm-device-wireless.h"

#include "applet-nm-private.h"

/**
 * SECTION:applet-nm
 * @title: AppletNM
 * @short_description: Network applet
 *
 * 管理网络连接，启用或禁用网络设备
 */

#define DEFAULT_PERIODIC_UPDATE_FREQUENCY_SECONDS 10
#define FAST_PERIODIC_UPDATE_FREQUENCY_SECONDS 2

const AppletEnumValue applet_nm_device_info[DeviceCount] = {
    {DeviceWired, "wired"},
    {DeviceWireless, "wireless"},
    {DeviceWwan, "wwan"},
    {DeviceVPN, "vpn"}
};

static void applet_nm_sync_section_title (AppletNM *self, AppletNMDeviceType type);

static void applet_nm_connection_removed (NMClient *client, NMConnection *connection, AppletNM *self);

static void applet_nm_update_connection (NMConnection *connection, AppletNM *self);

static void applet_nm_device_added (NMClient *client, NMDevice *device, AppletNM *self);

static void applet_nm_set_ap_icon (AppletNM *self, NMAccessPoint *ap);

static void applet_nm_update_icon (NMClient *NmClient, gpointer spec, AppletNM *self);

const AppletEnumValue ctypes[] =
{
    {DeviceWireless, NM_SETTING_WIRELESS_SETTING_NAME},
    {DeviceWired, NM_SETTING_WIRED_SETTING_NAME},
    {DeviceWired, NM_SETTING_PPPOE_SETTING_NAME},
    {DeviceWired, NM_SETTING_PPP_SETTING_NAME},
    {DeviceWwan,  NM_SETTING_BLUETOOTH_SETTING_NAME},
    {DeviceWwan,  NM_SETTING_CDMA_SETTING_NAME},
    {DeviceWwan,  NM_SETTING_GSM_SETTING_NAME},
    {DeviceVPN,   NM_SETTING_VPN_SETTING_NAME}
};

#ifndef GLIB_VERSION_2_54
gint
nm_g_ptr_array_find (const GPtrArray *array, gpointer element)
{
    guint i;
    for (i = 0; array && (i < array->len); i++)
        if (array->pdata[i] == element)
            return i;
    return -1;
}
#endif

static AppletNMDeviceType
get_ctype_category (const gchar *type)
{
    if (type == NULL)
        return DeviceInvalid;
    int i;
    for (i = 0; i < G_N_ELEMENTS (ctypes); i++)
    {
        if (strcmp (type, ctypes[i].value) == 0)
            return ctypes[i].key;
    }
    return DeviceInvalid;
}

G_DEFINE_TYPE (AppletNM, applet_nm, CDOS_TYPE_ICONAPPLET)

static void
menu_item_enabled_changed (AppletNMWirelessSectionMenuItem *item,
                           gboolean enabled,
                           AppletNM *self)
{
    GList *devices = self->devices[item->device_type].devices;
    GList *iter = NULL;
    for (iter = devices; iter; iter = iter->next)
    {
        applet_nm_device_set_enabled (iter->data, enabled);
    }
    applet_nm_sync_section_title (self, item->device_type);
}
static AppletNMWirelessSectionMenuItem *
make_toggle_item (AppletNM *self, AppletNMDeviceType type, const char *title)
{
    AppletNMWirelessSectionMenuItem *item = applet_nm_wireless_section_menu_item_new (self->client, type, title);
    g_signal_connect (item, "enabled-changed", G_CALLBACK (menu_item_enabled_changed), self);
    return item;
}

static void
applet_nm_dispose (GObject *object)
{
    AppletNM *self = APPLET_NM (object);
    if (self->devices)
    {
        for (gint i = 0; i < DeviceCount; i++)
        {
            GList *dev;
            GList *devices = self->devices[i].devices;
            for (dev = devices; dev; dev = dev->next)
            {
                applet_nm_device_destroy (dev->data);
            }
            g_list_free (devices);
        }
        g_free (self->devices);
    }
    g_clear_object (&self->client);

    G_OBJECT_CLASS (applet_nm_parent_class)->dispose (object);
}

static void
applet_nm_finalize (GObject *object)
{
    AppletNM *applet = APPLET_NM (object);
    g_free (applet->base_class);
    g_free (applet->wifi_class);
    g_free (applet->ethernet_class);
    g_free (applet->gdm_class);
    g_free (applet->cdma_class);
    g_free (applet->wimax_class);

    G_OBJECT_CLASS (applet_nm_parent_class)->finalize (object);
}

static void
applet_nm_class_init (AppletNMClass *klass)
{
    GObjectClass *gobject_class = G_OBJECT_CLASS (klass);
    gobject_class->dispose = applet_nm_dispose;
    gobject_class->finalize = applet_nm_finalize;
}

static void
applet_nm_add_action_callback (CdosPopupBaseMenuItem *item,
                               ClutterEvent *event,
                               gboolean keep,
                               AppletNM *self)
{
    GError *error = NULL;
    if (self->client)
        nm_client_networking_set_enabled (self->client, TRUE, &error);
    if (error)
    {
        g_warning ("%s", error->message);
        g_error_free (error);
    }
}

static void
cdos_add_action_util_callback (void)
{
    cdos_util_spawn_command_line ("nm-connection-editor");
}

static void
applet_nm_set_icon (AppletNM *self, const gchar *name)
{
    if (g_strcmp0 (self->current_icon_name, name))
    {
        cdos_iconapplet_set_icon_name (CDOS_ICONAPPLET (self), name);
        cdos_iconapplet_set_icon_size (CDOS_ICONAPPLET (self), 16);
        g_free (self->current_icon_name);
        self->current_icon_name = g_strdup (name);
    }
}

static void
update_for_device (AppletNM *self, AppletNMDevice *device, AppletNMDeviceType type)
{
    CdosPopupSwitchMenuItem *item = self->devices[type].item;
    switch (type)
    {
        case DeviceWired:
            applet_nm_wired_section_title_menu_item_update_for_device (item, self->devices[type].devices);
            break;
        case DeviceWireless:
            applet_nm_wireless_section_menu_item_update_for_device (item, device);
            break;
        default:
            break;
    }
}

static void
applet_nm_sync_section_title (AppletNM *self, AppletNMDeviceType type)
{
    GList *devices = self->devices[type].devices;
    CdosPopupSwitchMenuItem *item = self->devices[type].item;
    CdosPopupMenuSection *section = self->devices[type].section;
    if (devices == NULL)
    {
        if (section)
            clutter_actor_hide ((ClutterActor *)section);
    }
    else
    {
        clutter_actor_show ((ClutterActor *)section);
        if (devices->next == NULL)
        {
            AppletNMDevice *dev = devices->data;
            clutter_actor_hide (CLUTTER_ACTOR (dev->status_item));
            update_for_device (self, dev, type);
        }
        else
        {
            GList *iter = NULL;
            AppletNMDevice *dev_active = NULL;
            for (iter = devices; iter; iter = iter->next)
            {
                //CdosPopupSwitchMenuItem *status_item = ((AppletNMDevice *)iter->data)->status_item;
                //clutter_actor_show ((ClutterActor *)(CDOS_POPUP_BASE_MENU_ITEM(status_item)->actor));
                AppletNMDevice *dev_tmp = (AppletNMDevice *)iter->data;
                CdosPopupSwitchMenuItem *status_item = dev_tmp->status_item;
                if (dev_tmp->category == DeviceWired)
                {
                    clutter_actor_hide ((ClutterActor *)status_item);
                }
                else
                {
                    clutter_actor_show ((ClutterActor *)status_item);
                }
                if (applet_nm_device_get_connected (dev_tmp))
                    dev_active = dev_tmp;
            }
            update_for_device (self, dev_active, type);
        }
    }
}

static void
applet_nm_read_devices (AppletNM *self)
{
    const GPtrArray *devices = nm_client_get_devices (self->client);
    gint i;
    for (i = 0; devices && i < devices->len; i++)
    {
        NMDevice *a = devices->pdata[i];
        g_debug ("add device: description:%s\niface: %s, ip_iface: %s",
                 nm_device_get_description (a),
                 nm_device_get_iface (a),
                 nm_device_get_ip_iface (a));
        applet_nm_device_added (self->client, a, self);
    }
}

void
applet_nm_state_changed (AppletNM *applet, AppletNMDevice *wrapper)
{
    g_return_if_fail (APPLET_IS_NM (applet));
    applet_nm_sync_section_title (applet, wrapper->category);
}

static void
applet_nm_device_added (NMClient *client, NMDevice *device, AppletNM *self)
{
    AppletNMDevice *delegate = g_object_get_data ((GObject *)device, "delegate");
    if (delegate)
    {
        // already seen, not adding again
        return;
    }
    AppletNMDeviceType type;
    NMDeviceType t = nm_device_get_device_type (device);
    if (t == NM_DEVICE_TYPE_ETHERNET)
        type = DeviceWired;
    else if (t == NM_DEVICE_TYPE_WIFI)
        type = DeviceWireless;
    else
    {
        g_message ("%s: %p %d %s", __func__, device, t, nm_device_get_type_description (device));
        return;
    }

    AppletNMDevice *wrapper;
    if (type == DeviceWired)
        wrapper = (AppletNMDevice *)applet_nm_device_wired_new (self, device, self->connections);
    else if (type == DeviceWireless)
        wrapper = (AppletNMDevice *)applet_nm_device_wireless_new (self, device, self->connections);

    g_object_set_data ((GObject *)device, "delegate", wrapper);

    cdos_popup_menu_base_add_actor (CDOS_POPUP_MENU_BASE (self->devices[type].section),
                                    (ClutterActor *)wrapper->section);

    cdos_popup_menu_base_add_menu_item (
        CDOS_POPUP_MENU_BASE (self->devices[type].section), CDOS_POPUP_BASE_MENU_ITEM (wrapper->status_item), 1);
    self->devices[type].devices = g_list_append (self->devices[type].devices, wrapper);

    applet_nm_sync_section_title (self, type);
}

static void
applet_nm_device_removed (NMClient *client, NMDevice *device, AppletNM *self)
{
    // if (!(applet_nm_device_hash_get_delegate(device)))
    AppletNMDevice *delegate = g_object_get_data ((GObject *)device, "delegate");
    if (delegate == NULL)
    {
        g_warning ("Removing a network device that was not added");
        return;
    }
    applet_nm_device_destroy (delegate);
    GList *devices = self->devices[delegate->category].devices;
    devices = g_list_remove (devices, delegate);
    self->devices[delegate->category].devices = devices;
    applet_nm_sync_section_title (self, delegate->category);
}

/*
static void
applet_nm_active_connection_changed (NMActiveConnection *activeConnection, gpointer spec, AppletNM *self)
{
    if ((nm_active_connection_get_state(activeConnection) == NM_ACTIVE_CONNECTION_STATE_ACTIVATED)
            && (applet_nm_active_connection_get_primary_device (activeConnection))
            && (applet_nm_device_hash_get_notification (applet_nm_active_connection_get_primary_device (activeConnection))))
    {
        cdos_notification_destroy (applet_nm_device_hash_get_notification (applet_nm_active_connection_get_primary_device (activeConnection)), 1);//暂时为1...
        applet_nm_device_hash_set_notification (applet_nm_active_connection_get_primary_device (activeConnection), NULL);
    }
}
*/

static void
applet_nm_read_connections (AppletNM *self)
{
    if (self->connections)
    {
        g_ptr_array_free (self->connections, TRUE);
    }
    self->connections = g_ptr_array_new ();
    const GPtrArray *connections = nm_client_get_connections (self->client);
    for (int i = 0; i < connections->len; i++)
    {
        NMConnection *conn = NM_CONNECTION (connections->pdata[i]);

        AppletNMConnection *table = applet_nm_connection_get_data (conn);
        if (table->uuid)
        {
            //connection was already seen (for example because NetworkManager was restarted)
            continue;
        }
        table->connection = conn;
        table->updated_id = g_signal_connect (conn, "changed", G_CALLBACK (applet_nm_update_connection), self);

        applet_nm_update_connection (conn, self);

        g_ptr_array_add (self->connections, conn);
    }
}

static void
applet_nm_update_connection (NMConnection *connection, AppletNM *self)
{
    NMSettingConnection *s_con = nm_connection_get_setting_connection (connection);
    /*
    const char *type = nm_setting_connection_get_connection_type (s_con);
    NMSetting *setting = nm_connection_get_setting_by_name (connection, NM_SETTING_CONNECTION_SETTING_NAME);
    */

    AppletNMConnection *table = applet_nm_connection_get_data (connection);
    table->type = nm_connection_get_connection_type (connection);
    table->section = get_ctype_category (table->type);
    table->name = nm_setting_connection_get_id (s_con);
    table->uuid = nm_setting_connection_get_uuid (s_con);
    table->timestamp = nm_setting_connection_get_timestamp (s_con);

    g_debug ("update connection: type: %s, id:%s, uuid:%s",
             table->type, table->name, table->uuid);

    GList *devices = self->devices[DeviceWired].devices;
    GList *iter;
    for (iter = devices; iter; iter = iter->next)
        applet_nm_device_check_connection (iter->data, connection);

    devices = self->devices[DeviceWireless].devices;
    for (iter = devices; iter; iter = iter->next)
        applet_nm_device_check_connection (iter->data, connection);
}

static void
applet_nm_new_connection (NMClient *client, NMRemoteConnection *remote_connection, AppletNM *self)
{
    NMConnection *connection = NM_CONNECTION (remote_connection);
    AppletNMConnection *table = applet_nm_connection_get_data (connection);
    if (table->uuid)
    {
        // connection was already seen
        return;
    }
    table->updated_id = g_signal_connect (connection, "changed", G_CALLBACK (applet_nm_update_connection), self);

    applet_nm_update_connection (connection, self);
    g_ptr_array_insert (self->connections, -1, connection);
}

static void
applet_nm_connection_removed (NMClient *client, NMConnection *connection, AppletNM *self)
{
    AppletNMConnection *conn = applet_nm_connection_get_data (connection);

    guint pos;
#ifdef GLIB_VERSION_2_54
    if (g_ptr_array_find (self->connections, connection, &pos))
#else
    if (nm_g_ptr_array_find (self->connections, connection) >= 0)
#endif
        g_ptr_array_remove (self->connections, connection);
    AppletNMDeviceType section = conn->section;
    if ((section != DeviceVPN) && (section != DeviceInvalid))
    {
        GList *devices = self->devices[section].devices;
        GList *iter = NULL;
        for (iter = devices; iter; iter = iter->next)
            applet_nm_device_remove_connection (iter->data, connection);
    }

    conn->uuid = 0;
    // g_signal_handler_disconnect (connection, conn->removed_id);
    // g_signal_handler_disconnect (connection, conn->updated_id);
    g_free (conn);
}

static void
applet_nm_hide_devices (AppletNM *self)
{
    self->devices_hidden = TRUE;
    /* FIXME */
    // if (self->devices[DeviceWired] != NULL)
    {
        clutter_actor_hide (CLUTTER_ACTOR (self->devices[DeviceWired].section));
        clutter_actor_hide (CLUTTER_ACTOR (self->devices[DeviceWireless].section));
    }
}

static void
applet_nm_show_normal (AppletNM *self)
{
    if (!self->devices_hidden)//nothing to do
        return;

    self->devices_hidden = FALSE;
    clutter_actor_hide ((ClutterActor *)self->status_section);

    applet_nm_sync_section_title (self, DeviceWired);
    applet_nm_sync_section_title (self, DeviceWireless);
    applet_nm_sync_section_title (self, DeviceWwan);

    // if( !applet_nm_device_vpn_get_empty() )
    //{
    // TODO:
    //}
}

static void
applet_nm_sync_connection (NMClient *client,
                           GParamSpec *pspec,
                           AppletNM *self)
{
    NMActiveConnection *connection;
    if (self->main_connection)
    {
        g_signal_handlers_disconnect_by_data (self->main_connection, self);
        self->main_connection = NULL;
    }
    connection = nm_client_get_primary_connection (self->client);
    self->main_connection = connection;
    if (!self->main_connection)
    {
        for (AppletNMDeviceType type = DeviceWired; type < DeviceCount; type++)
        {
            GList *devices = self->devices[type].devices;
            for (GList *iter = devices; iter; iter = iter->next)
            {
                AppletNMDevice *device = iter->data;
                NMDevice *dev = device->device;
                connection = nm_device_get_active_connection (dev);
                if (connection)
                {
                    self->main_connection = connection;
                    applet_nm_update_icon (NULL, NULL, self);
                    return;
                }
            }
        }
    }
    applet_nm_update_icon (NULL, NULL, self);
}

static void
applet_nm_active_connection_state_changed (NMActiveConnection *connection,
                                           GParamSpec *pspec,
                                           AppletNM *self)
{
    NMActiveConnectionState state = nm_active_connection_get_state (connection);
    if (state != NM_ACTIVE_CONNECTION_STATE_ACTIVATING)
    {
        // g_signal_handlers_disconnect_by_func (connection,
        //                                      applet_nm_active_connection_state_changed,
        //                                      self);
        applet_nm_sync_connection (NULL, NULL, self);
    }
}

/*
 * 如果Networkmanager没用配置自动检查connectivity，
 * 那么在网络由 正在连接-->已连接 转变后，
 * NMClient 不会收到 connectivity 变化的信号。
 * 可以通过 activating-connection 和 active-connections 的变化检测网络状态变化
 */
static void
applet_nm_sync_activating_connection (NMClient *client,
                                      GParamSpec *pspec,
                                      AppletNM *self)
{
    NMActiveConnection *connection;
    connection = nm_client_get_activating_connection (self->client);
    if (!connection)
    {
        applet_nm_sync_connection (NULL, NULL, self);
        return;
    }
    if (self->main_connection)
        g_signal_handlers_disconnect_by_data (self->main_connection, self);

    self->main_connection = connection;
    g_signal_connect (connection, "notify::state", G_CALLBACK (applet_nm_active_connection_state_changed), self);
    applet_nm_update_icon (NULL, NULL, self);
}

static gboolean
applet_nm_has_active_device (AppletNM *self)
{
    for (AppletNMDeviceType type = DeviceWired; type < DeviceCount; type++)
    {
        GList *devices = self->devices[type].devices;
        for (GList *iter = devices; iter; iter = iter->next)
        {
            AppletNMDevice *device = iter->data;
            NMDevice *dev = device->device;
            NMDeviceState state = nm_device_get_state (dev);
            if (state > NM_DEVICE_STATE_DISCONNECTED &&
                state < NM_DEVICE_STATE_DEACTIVATING)
            {
                return TRUE;
            }
        }
    }
    return FALSE;
}

static void
applet_nm_sync_active_connections (NMClient *client,
                                   GParamSpec *pspec,
                                   AppletNM *self)
{
    /*
     * FIXME
     * 这里好像是libnm-glib的问题，在所有网络设备都断开连接时，
     * 依然可以获取到active-connections，
     * 所以在这里对所有设备状态进行判断，
     * 注意不能使用 nm_client_get_state() 函数判断网络状态。
     */
    gboolean has_active_device = applet_nm_has_active_device (self);
    if (!has_active_device)
    {
        self->main_connection = NULL;
        applet_nm_update_icon (NULL, NULL, self);
        return;
    }
    const GPtrArray *connections = nm_client_get_active_connections (self->client);
    if (!connections || (connections->len == 0))
    {
        self->main_connection = NULL;
        applet_nm_update_icon (NULL, NULL, self);
        return;
    }
    /*
     * 在主网卡已连接时，NMClient不会收到activating-connection信号，
     * 也不能获取到 activating-connection，只能收到active-connetions信号。
     */
    gint i;
    for (i = 0; i < connections->len; i++)
    {
        NMActiveConnection *active = connections->pdata[i];
        if (nm_active_connection_get_state (active) == NM_ACTIVE_CONNECTION_STATE_ACTIVATING)
        {
            if (self->main_connection)
                g_signal_handlers_disconnect_by_data (self->main_connection, self);
            self->main_connection = active;
            g_signal_connect (active, "notify::state", G_CALLBACK (applet_nm_active_connection_state_changed), self);
            break;
        }
    }
    applet_nm_update_icon (NULL, NULL, self);
}

static void
applet_nm_sync_connectivity (NMClient *client,
                             GParamSpec *pspec,
                             AppletNM *self)
{
    applet_nm_update_icon (NULL, NULL, self);
}

static void
applet_nm_sync_nm_state (NMClient *NmClient, gpointer spec, AppletNM *self)
{
    if (!nm_client_get_nm_running (self->client))
    {
        g_debug ("NetworkManager is not running,hiding...");
        clutter_actor_hide ((ClutterActor *) self);
        return;
    }
    else
    {
        clutter_actor_show ((ClutterActor *) self);
    }

    if (!nm_client_networking_get_enabled (self->client))
    {
        applet_nm_set_icon (self, "found-network-offline");
        applet_nm_hide_devices (self);
        cdos_popup_menu_item_set_label (self->status_item, _("Networking is disabled"));
        clutter_actor_show ((ClutterActor *)self->status_section);
        return;
    }
    applet_nm_show_normal (self);
    applet_nm_sync_connection (NULL, NULL, self);
}

static void
applet_nm_ap_strength_changed (NMAccessPoint *ap, GParamSpec *pspec, AppletNM *self)
{
    applet_nm_set_ap_icon (self, ap);
}

static void
applet_nm_set_ap_icon (AppletNM *self, NMAccessPoint *ap)
{
    if (self->active_access_point && NM_IS_ACCESS_POINT (self->active_access_point))
    {
        g_signal_handlers_disconnect_by_func (self->active_access_point, applet_nm_ap_strength_changed, self);
    }
    if (!self->main_connection)
    {
        self->active_access_point = NULL;
        return;
    }
    self->active_access_point = ap;
    guint8 strength = nm_access_point_get_strength (ap);
    if (self->connectivity != 0x2)
        applet_nm_set_icon (self, "found-network-wireless-no-route");
    else
    {
        const gchar *icon = applet_nm_get_icon (ap, applet_nm_signal_to_icon (strength));
        applet_nm_set_icon (self, icon);
    }
    g_signal_connect (ap, "notify::strength", G_CALLBACK (applet_nm_ap_strength_changed), self);
    char *utf8_ssid = applet_nm_get_utf8_ssid (nm_access_point_get_ssid (ap));
    gchar *tooltip;
    if (self->connectivity == 0x1)
    {
        tooltip = g_strdup_printf ("%s: %s (%d%%)",
                                    _("Wireless connection, maybe need authenticate"),
                                    utf8_ssid,
                                    strength);
    }
    else if (self->connectivity == 0)
    {
        tooltip = g_strdup_printf ("%s: %s (%d%%)",
                                   _("Wireless connection, limited access"),
                                   utf8_ssid,
                                   strength);

    }
    else /* (self->connectivity == 0x2) */
    {
        tooltip = g_strdup_printf ("%s: %s (%d%%)",
                                   _("Wireless connection"),
                                   utf8_ssid,
                                   strength);
    }
    cdos_applet_set_applet_tooltip (CDOS_APPLET (self), tooltip);
    g_free (tooltip);
    g_free (utf8_ssid);
}

static void
access_point_changed (NMDeviceWifi *device, gpointer spec, AppletNM *self)
{
    NMDeviceWifi *wdev = NM_DEVICE_WIFI (device);
    NMAccessPoint *ap;
    ap = nm_device_wifi_get_active_access_point (wdev);
    NM80211Mode mode = nm_device_wifi_get_mode (wdev);
    if (!ap)
    {
        /* if (mode != NM_802_11_MODE_ADHOC)
        {
            g_warning (_("An active wireless connection, in infrastructure mode, involves no access point."));
        } */
        applet_nm_set_icon (self, "found-network-wireless-no-route");
        cdos_applet_set_applet_tooltip (CDOS_APPLET (self), _("Connected to the wireless network"));
    } else {
        applet_nm_set_ap_icon (self, ap);
    }
}

static void
applet_nm_update_icon (NMClient *client, gpointer spec, AppletNM *self)
{
    if (!nm_client_networking_get_enabled (self->client))
    {
        applet_nm_set_icon (self, "found-network-offline");
        return;
    }

    NMActiveConnection *mc;
    gboolean has_active_device = applet_nm_has_active_device (self);
    if (has_active_device)
        mc = self->main_connection;
    else
    {
        self->main_connection = NULL;
        mc = NULL;
    }
    if (!mc)
    {
        applet_nm_set_icon (self, "found-network-offline");
        cdos_applet_set_applet_tooltip (CDOS_APPLET (self), _("No connection"));
        return;
    }
    if (!NM_IS_ACTIVE_CONNECTION (mc))
        return;

    NMActiveConnectionState state = nm_active_connection_get_state (mc);
    if (state > NM_ACTIVE_CONNECTION_STATE_ACTIVATED)
        return;

    const GPtrArray *devices = nm_active_connection_get_devices (mc);
    if (!devices || !devices->len)
    {
        NMState state = nm_client_get_state (self->client);
        if (state == NM_STATE_CONNECTED_GLOBAL)
            applet_nm_set_icon (self, "found-network-wired");
        else
            applet_nm_set_icon (self, "found-network-wired-no-route");
        return;
    }
    NMDevice *device = devices->pdata[0];
    NMDeviceType type = nm_device_get_device_type (device);

    if (device && self->wireless_active_ap_changed > 0)
    {
        g_signal_handler_disconnect (device, self->wireless_active_ap_changed);
        self->wireless_active_ap_changed = 0;
    }

    if (state == NM_ACTIVE_CONNECTION_STATE_ACTIVATING)
    {
        switch (type)
        {
            case NM_DEVICE_TYPE_ETHERNET:
                applet_nm_set_icon (self, "found-network-wired-acquiring");
                cdos_applet_set_applet_tooltip (CDOS_APPLET (self), _("Connecting to the wired network..."));
                break;
            case NM_DEVICE_TYPE_MODEM:
                applet_nm_set_icon (self, "found-network-celluar-acquiring");
                cdos_applet_set_applet_tooltip (CDOS_APPLET (self), _("Connecting to the celluar network..."));
                break;
            case NM_DEVICE_TYPE_WIFI:
                applet_nm_set_icon (self, "found-network-wireless-acquiring");
                cdos_applet_set_applet_tooltip (CDOS_APPLET (self), _("Connecting to the wireless network..."));
                break;
            /*
            case DeviceVPN:
                applet_nm_set_icon (self, "network-vpn-acquiring");
                cdos_applet_set_applet_tooltip (CDOS_APPLET (self), _("Connecting to the VPN network..."));
                break;
            */
            default:
                applet_nm_set_icon (self, "found-network-wired-acquiring");
                cdos_applet_set_applet_tooltip (CDOS_APPLET (self), _("Connecting to the wired network..."));
                break;
        }
    }
    else if (state == NM_ACTIVE_CONNECTION_STATE_ACTIVATED)
    {
        NMConnectivityState connectivity;
        // connectivity = nm_client_get_connectivity (self->client);
        connectivity = nm_client_check_connectivity (self->client, NULL, NULL);
        gboolean is_portal = FALSE;
        gboolean is_global = FALSE;
        /*
         * portal: 因特网连接被网关劫持，用户需要在浏览器中向网关进行身份验证。
         * limited: 主机已经连接到网络，但是没用完全的因特网访问权限，而且也
         *          没用检测到中间网关
         * global: 已连接到因特网
         */
        if (connectivity == NM_CONNECTIVITY_PORTAL)
            is_portal = TRUE;
        else if (connectivity == NM_CONNECTIVITY_LIMITED)
            is_global = FALSE;
        else if (connectivity == NM_CONNECTIVITY_FULL ||
                 connectivity == NM_CONNECTIVITY_UNKNOWN)
            is_global = TRUE;
        self->connectivity = (is_global << 1) | is_portal;
        if (type != NM_DEVICE_TYPE_WIFI)
        {
            if (self->active_access_point && NM_IS_ACCESS_POINT (self->active_access_point))
            {
                g_signal_handlers_disconnect_by_func (self->active_access_point, applet_nm_ap_strength_changed, self);
                self->active_access_point = NULL;
            }
        }
        switch (type)
        {
            case NM_DEVICE_TYPE_ETHERNET:
                if (is_global)
                    applet_nm_set_icon (self, "found-network-wired");
                else if (is_portal || !is_global)
                {
                    applet_nm_set_icon (self, "found-network-wired-no-route");
                }
                /*
                else
                    applet_nm_set_icon (self, "network-error");
                */
                cdos_applet_set_applet_tooltip (CDOS_APPLET (self), _("Connected to the wired network"));
                break;
            case NM_DEVICE_TYPE_MODEM:
                applet_nm_set_icon (self, "found-network-celluar-connected");
                cdos_applet_set_applet_tooltip (CDOS_APPLET (self), _("Connected to the celluar network"));
                break;
            case NM_DEVICE_TYPE_WIFI:
            {
                NMRemoteConnection *r_con = nm_active_connection_get_connection (mc);
                if (r_con)
                {
                    NMSettingIPConfig *ip4_config = nm_connection_get_setting_ip4_config ((NMConnection *)r_con);
                    const char *method = nm_setting_ip_config_get_method (ip4_config);
                    if (g_strcmp0 (method, NM_SETTING_IP4_CONFIG_METHOD_SHARED) == 0)
                    {
                        applet_nm_set_icon (self, "found-network-wireless-hotspot");
                        cdos_applet_set_applet_tooltip (CDOS_APPLET (self), _("Hotspot Active"));
                        return;
                    }
                }
                if (device)
                {
                    self->wireless_active_ap_changed = \
                    g_signal_connect (device, "notify::active-access-point", G_CALLBACK (access_point_changed), self);
                    access_point_changed (device, NULL, self);

                    /* NMDeviceWifi *wdev = NM_DEVICE_WIFI (device);
                    NMAccessPoint *ap = nm_device_wifi_get_active_access_point (wdev);
                    NM80211Mode mode = nm_device_wifi_get_mode (wdev);
                    if (!ap)
                    {
                        if (mode != NM_802_11_MODE_ADHOC)
                        {
                            g_warning (_("An active wireless connection, in infrastructure mode, involves no access point?"));
                            //break;
                        }
                        if (is_global)
                            applet_nm_set_icon (self, "found-network-wireless-connected");
                        else
                            applet_nm_set_icon (self, "found-network-wireless-no-route");
                        cdos_applet_set_applet_tooltip (CDOS_APPLET (self), _("Connected to the wireless network"));
                    } else {
                        applet_nm_set_ap_icon (self, ap);
                    } */
                }
                break;
            }
            /*
            case DeviceVPN:
                applet_nm_set_icon (self, "network-vpn");
                cdos_applet_set_applet_tooltip (CDOS_APPLET (self), _("Connected to the VPN network"));
                break;
            */
            default:
                applet_nm_set_icon (self, "found-network-wired");
                cdos_applet_set_applet_tooltip (CDOS_APPLET (self), _("Connected to the wired network"));
                break;
        }
    }
    else
    {
        applet_nm_set_icon (self, "found-network-offline");
        cdos_applet_set_applet_tooltip (CDOS_APPLET (self), _("No connection"));
    }
}

static void
applet_nm_on_connections_read (AppletNM *self)
{
    applet_nm_read_connections (self);
    applet_nm_read_devices (self);
    applet_nm_sync_nm_state (NULL, NULL, self);

    if (!self->inited)
    {
        self->inited = TRUE;
        g_signal_connect (self->client,
                         "notify::nm-running",
                         G_CALLBACK (applet_nm_sync_nm_state),
                         self);
        g_signal_connect (self->client,
                         "notify::networking-enabled",
                         G_CALLBACK (applet_nm_sync_nm_state),
                         self);
        g_signal_connect (self->client,
                         "notify::state",
                         G_CALLBACK (applet_nm_sync_nm_state),
                         self);
        g_signal_connect (self->client,
                          "notify::active-connections",
                          /* G_CALLBACK (applet_nm_update_vpn_active_connections), */
                          G_CALLBACK (applet_nm_sync_active_connections),
                          self);
        g_signal_connect (self->client,
                          "notify::connectivity",
                          G_CALLBACK (applet_nm_sync_connectivity),
                          self);
        g_signal_connect (self->client,
                          "notify::primary-connection",
                          G_CALLBACK (applet_nm_sync_connection),
                          self);
        g_signal_connect (self->client,
                          "notify::activating-connection",
                          G_CALLBACK (applet_nm_sync_activating_connection),
                          self);
        g_signal_connect (self->client,
                         "device-added",
                         G_CALLBACK (applet_nm_device_added),
                         self);
        g_signal_connect (self->client,
                         "device-removed",
                         G_CALLBACK (applet_nm_device_removed),
                         self);
        g_signal_connect (self->client,
                         "connection-added",
                         G_CALLBACK (applet_nm_new_connection),
                         self);
        g_signal_connect (self->client,
                         "connection-removed",
                         G_CALLBACK (applet_nm_connection_removed),
                         self);
    }
}

static void
applet_nm_init (AppletNM *self)
{
    self->devices = g_new0 (AppletNMDeviceObj, DeviceCount);
    self->base_class = applet_nm_device_class_new ();
    self->ethernet_class = applet_nm_device_wired_class_new (self);
    self->wifi_class = applet_nm_device_wireless_class_new (self);
    self->client = NULL;
    self->current_icon_name = NULL;
    self->devices_hidden = TRUE;
    cdos_applet_set_popup_menu_active (CDOS_APPLET (self), TRUE);
    cdos_applet_set_context_menu_active (CDOS_APPLET (self), TRUE);
}

static void
applet_nm_try_init (AppletNM *self)
{
    CdosPopupMenuItem *network_settings;
    CdosPopupMenu *context_menu;
    if (!self->client)
        return;

    /* Wired Device */
    self->devices[DeviceWired].section = cdos_popup_menu_section_new ();
    clutter_actor_set_x_expand ((ClutterActor *)self->devices[DeviceWired].section, TRUE);
    self->devices[DeviceWired].devices = NULL;
    self->devices[DeviceWired].item =
        (CdosPopupSwitchMenuItem *)applet_nm_wired_section_title_menu_item_new (_("Wired"));
    cdos_popup_menu_base_add_menu_item (CDOS_POPUP_MENU_BASE (self->devices[DeviceWired].section),
                                        CDOS_POPUP_BASE_MENU_ITEM (self->devices[DeviceWired].item),
                                        -1);

    // clutter_actor_hide (CLUTTER_ACTOR (cdos_popup_menu_base_get_box (CDOS_POPUP_MENU_BASE
    // (self->devices[DeviceWired].section))));
    cdos_popup_menu_base_add_actor (CDOS_POPUP_MENU_BASE (self->menu),
                                    (ClutterActor *)self->devices[DeviceWired].section);

    cdos_popup_menu_base_add_menu_item (
        CDOS_POPUP_MENU_BASE (self->menu), CDOS_POPUP_BASE_MENU_ITEM (cdos_popup_separator_menu_item_new ()), -1);

    /* Wireless Device */
    self->devices[DeviceWireless].section = cdos_popup_menu_section_new ();
    clutter_actor_set_x_expand ((ClutterActor *)self->devices[DeviceWireless].section, TRUE);
    self->devices[DeviceWireless].item =
        (CdosPopupSwitchMenuItem *)make_toggle_item (self, DeviceWireless, _("Wireless"));
    cdos_popup_menu_base_add_menu_item ((CdosPopupMenuBase *)self->devices[DeviceWireless].section,
                                        (CdosPopupBaseMenuItem *)self->devices[DeviceWireless].item,
                                        -1);
    cdos_popup_menu_base_add_actor ((CdosPopupMenuBase *)self->menu,
                                    (ClutterActor *)self->devices[DeviceWireless].section);

    /* cdos_popup_menu_base_add_menu_item (
        CDOS_POPUP_MENU_BASE (self->menu), CDOS_POPUP_BASE_MENU_ITEM (cdos_popup_separator_menu_item_new ()), -1); */

    context_menu = cdos_applet_get_context_menu (CDOS_APPLET (self));
    // open the network settings in the panel
    network_settings = cdos_popup_menu_base_add_commandline_action (CDOS_POPUP_MENU_BASE (context_menu),
                                                                    _("Network Settings"),
                                                                    "found-control-center network");
    cdos_popup_menu_item_set_label_align (network_settings, ST_ALIGN_MIDDLE, 0);
    StLabel *network_settings_label = cdos_popup_menu_item_get_label (network_settings);
    st_widget_add_style_class_name (ST_WIDGET (network_settings_label), "popup-menu-item-label-link");

    /* CdosPopupMenuItem *network_connections = cdos_popup_menu_base_add_action (
        CDOS_POPUP_MENU_BASE (context_menu), _("Network Connections"), cdos_add_action_util_callback, NULL);
    cdos_popup_menu_item_set_label_align (network_connections, ST_ALIGN_MIDDLE, 0);
    StLabel *network_connections_label = cdos_popup_menu_item_get_label (network_connections);
    st_widget_add_style_class_name (ST_WIDGET (network_connections_label), "popup-menu-item-label-link"); */
    
    applet_nm_on_connections_read (self);
}

static void
new_client_finish (GObject *source,
                   GAsyncResult *res,
                   AppletNM *self)
{
    GError *error = NULL;
    self->client = nm_client_new_finish (res, &error);
    if (error)
    {
        g_warning ("%s: %s", __func__, error->message);
        return;
    }
    applet_nm_try_init (self);
}

static void
applet_nm_init_after (AppletNM *self)
{
    self->menu_manager = cdos_applet_get_menu_manager((CdosApplet *)self);
    self->menu = cdos_applet_get_popup_menu((CdosApplet *)self);

    st_widget_add_style_class_name(ST_WIDGET(self->menu->actor),"nm-popup-menu");
    cdos_popup_menu_manager_add_menu (self->menu_manager, CDOS_POPUP_MENU_BASE (self->menu), -1);

    applet_nm_set_icon (self, "found-network-offline");
    self->status_section = cdos_popup_menu_section_new ();
    PopupBaseMenuItemInit *params = g_new0 (PopupBaseMenuItemInit, 1);
    params->style_class = "popup-inactive-menu-item";
    params->reactive = 0;
    self->status_item = cdos_popup_menu_item_new ();
    cdos_popup_menu_item_set_label (self->status_item, _("Status"));

    cdos_popup_base_menu_item_set_params ((CdosPopupBaseMenuItem *)self->status_item, params);
    g_free (params);
    cdos_popup_menu_base_add_menu_item (
        CDOS_POPUP_MENU_BASE (self->status_section), (CdosPopupBaseMenuItem *)self->status_item, -1);

    cdos_popup_menu_base_add_action (CDOS_POPUP_MENU_BASE (self->status_section),
                                     _("Enable networking"),
                                     (GCallback)applet_nm_add_action_callback,
                                     self);
    clutter_actor_hide ((ClutterActor *)self->status_section);

    cdos_popup_menu_base_add_actor ((CdosPopupMenuBase *)self->menu,
                                    (ClutterActor *)self->status_section);

    cdos_popup_menu_base_add_menu_item (CDOS_POPUP_MENU_BASE (self->menu),
                                        CDOS_POPUP_BASE_MENU_ITEM (cdos_popup_separator_menu_item_new()),
                                        -1);

    nm_client_new_async (NULL, (GAsyncReadyCallback)new_client_finish, self);
}

static AppletNM *
applet_nm_new (MetaData *metadata, guint orientation, guint panel_height)
{
    AppletNM *self;
    self = g_object_new (APPLET_TYPE_NM, "orientation", orientation, NULL);
    applet_nm_init_after (self);

    return self;
}

/**
 * applet_network_main:
 *
 * @orientation: 文本的方向
 * @panel_height：panel的方向
 * @mata: 新的Meta结构体
 *
 * 创建一个新的applet
 * Returns:新创建的 #CdosApplet
 *
 * Since:1.0.0
 */

CdosApplet *
applet_network_main (MetaData *meta, guint orientation, guint panel_height)
{
    AppletNM *appletnm;
    appletnm = applet_nm_new (meta, orientation, panel_height);
    // cdos_iconapplet_reset_for_systray (CDOS_ICONAPPLET (appletnm));
    return CDOS_APPLET (appletnm);
}

static void
applet_nm_notify_for_device (AppletNMDevice *device,
                             const gchar *iconName,
                             const gchar *title,
                             const gchar *text,
                             gboolean urgency)
{
    if (device->notification)
    {
        cdos_notification_destroy (device->notification, CDOS_NOTIFICATION_EXPIRED);
    }
    /* TODO: send notification */
}

void
applet_nm_activation_failed (AppletNMDevice *device, guint reason) //...
{
    applet_nm_notify_for_device (device, "network-error", _("Connection failed"),
                                 _("Activation of network connection failed"),2);

}
