/**
 * SECTION: AppletNotification
 * @short_description: 通知应用
 * @include: applet-notification.h
 *
 * AppletNotificationClass 类用来接收，展示，处理 #CdosMessageTray 发送来地消息
 */
#include <sys/time.h>
#include <math.h>

#include "cdos-start.h"
#include "applet-notifications.h"
#include "ui/cdos-popup-base-menu-item.h"

#define APPLET_NOTIFICATIONS_MAX_COUNT 48

/* 注册函数 */
G_DEFINE_TYPE (AppletNotifications, applet_notifications, CDOS_TYPE_TEXT_ICON_APPLET)

/* 继承关系
 * AppletNotifications->CdosTextIconApplet->IconApplet->Applet
 */

typedef struct _DataItemClicked DataItemClicked;
struct _DataItemClicked
{
    AppletNotifications *self;
    gboolean destroyed;
};


static void applet_notifications_update_list (AppletNotifications *self);
static void applet_notifications_item_destroy_cb (CdosNotification *notification, AppletNotifications *self);
static gboolean applet_notifications_on_applet_clicked_cb (ClutterActor *actor, ClutterEvent *event, gpointer user_data);

static void applet_notifications_item_clicked_cb (GObject *obj, gpointer user_data);

static void applet_notifications_display (AppletNotifications *self);

static void applet_notifications_update_timestamp (AppletNotifications *self);

static void applet_notifications_notification_added_cb (GObject *obj, CdosNotification *notification, gpointer user_data);

static void applet_notifications_clear_all_cb (GObject *obj, gpointer event, gboolean keepMenu, gpointer user_data);

static void callback_scroll_start (GObject *obj, gpointer user_data);

static void callback_scroll_stop (GObject *obj, gpointer user_data);

static gchar *
applet_notifications_stringify (gint count)
{
    gchar *str = NULL;

    if (count == 0)
    {
        str = g_strdup (_("No notifications"));
    }
    else if (count == 1)
    {
        str = g_strdup_printf ("%s ( %d )", _(" notification"), count);
    }
    else if (count > 1)
    {
        str = g_strdup_printf ("%s ( %d )", _(" notifications"), count);
    }
    return str;
}

static void
applet_notifications_notification_added_cb (GObject *obj, CdosNotification *notification, gpointer user_data)
{
    g_return_if_fail (notification != NULL);
    AppletNotifications *self = APPLET_NOTIFICATIONS (user_data);

    if (self->ignore_transient_notifications && notification->is_transient)
    {
        return;
    }

    g_object_ref (notification); //引用计数加1，防止被释放

    ClutterActor *parent = clutter_actor_get_parent (CLUTTER_ACTOR (notification));
    if (parent != NULL)
    {
        clutter_actor_remove_child (parent, CLUTTER_ACTOR (notification));
    }
    gint existing_index = g_list_index (self->notifications, notification);
    if (existing_index != -1)
    {
        g_warn_if_reached ();
        notification->in_notification_center = TRUE;
        clutter_actor_add_child (CLUTTER_ACTOR (self->p_notification_bin), CLUTTER_ACTOR (notification));
        g_object_set (notification, 
                  "x-expand", TRUE, 
                  "y-expand", TRUE, 
                  "x-align", CLUTTER_ACTOR_ALIGN_FILL, 
                  "y-expand", CLUTTER_ACTOR_ALIGN_START,
                  NULL);
        cdos_notification_collapse (notification);
        clutter_actor_show (CLUTTER_ACTOR (notification->time_label));

        applet_notifications_update_list (self);
        return;
    }

    notification->in_notification_center = TRUE;

    if (self->notification_count == APPLET_NOTIFICATIONS_MAX_COUNT)
    {
        GList *firstNode = g_list_first (self->notifications);
        cdos_notification_destroy (CDOS_NOTIFICATION (firstNode->data), 0);
        self->notification_count = self->notification_count - 1;
    }

    self->notifications = g_list_append (self->notifications, notification);
    self->notification_count += 1;
    cdos_notification_collapse (notification);
    clutter_actor_add_child (CLUTTER_ACTOR (self->p_notification_bin), CLUTTER_ACTOR (notification));
    g_object_unref (notification); //减去1

    st_widget_add_style_class_name (ST_WIDGET (notification), "applet-notification-padding");

    DataItemClicked *data_clicked, *data_destroy;
    data_clicked = g_new0 (DataItemClicked, 1);
    data_destroy = g_new0 (DataItemClicked, 1);
    data_clicked->self = self;
    data_clicked->destroyed = FALSE;
    data_destroy->self = self;
    data_destroy->destroyed = TRUE;

    g_signal_connect (notification, "clicked", G_CALLBACK (applet_notifications_item_clicked_cb), data_clicked);
    g_signal_connect (notification, "destroy", G_CALLBACK (applet_notifications_item_destroy_cb), self);

    clutter_actor_show (CLUTTER_ACTOR (notification->time_label));
    /*
    gfloat height;
    guint count = self->notification_count;
    gfloat width, height;

    clutter_actor_get_size (CLUTTER_ACTOR (self->p_main_container), &width, &height);

    if ( height > 300 || count >= 5) {
        clutter_actor_set_size (CLUTTER_ACTOR (self->p_main_container), -1, 300);
    } else {
        clutter_actor_set_size (CLUTTER_ACTOR (self->p_main_container), -1, -1);
    }

    if (count == 0) {
        clutter_actor_set_size (CLUTTER_ACTOR(self->p_main_container), 0, 0);
    }
    */
    applet_notifications_update_list (self);
}

static void
applet_notifications_clear_all_cb (GObject *obj, gpointer event, gboolean keepMenu, gpointer user_data)
{
    AppletNotifications *self = APPLET_NOTIFICATIONS (user_data);

    guint count;
    count = g_list_length (self->notifications);
    CdosNotification *notification_tmp = NULL;
    GList *notifications = self->notifications;
    if (count > 0)
    {
        while (notifications != NULL)
        {
            if (notifications->data)
            {
                notification_tmp = CDOS_NOTIFICATION (notifications->data);
                notifications = notifications->next;
                cdos_notification_destroy (notification_tmp, CDOS_NOTIFICATION_DISMISSED);
            }
        }
        g_list_free (self->notifications);
    }
    self->notifications = NULL;
    self->notification_count = 0;
    applet_notifications_update_list (self);
}

static void
applet_notifications_item_destroy_cb (CdosNotification *notification, AppletNotifications *self)
{
    self->notifications = g_list_remove (self->notifications, notification);

    applet_notifications_update_list (self);
}


static void
applet_notifications_item_clicked_cb (GObject *obj, gpointer user_data)
{
    CdosNotification *notification = CDOS_NOTIFICATION (obj);
    DataItemClicked *data = (DataItemClicked *)user_data;
    AppletNotifications *self = APPLET_NOTIFICATIONS (data->self);

    gint i = g_list_index (self->notifications, notification);

    if (i != -1)
    {
        /*NULL if it is not found*/
        self->notifications = g_list_remove (self->notifications, notification);
        if (!data->destroyed)
        {
            cdos_notification_destroy (notification, CDOS_NOTIFICATION_DISMISSED);
        }
    }

    applet_notifications_update_list (self);
}

static void
callback_scroll_start (GObject *obj, gpointer user_data)
{
    AppletNotifications *self = APPLET_NOTIFICATIONS (user_data);
    CDOS_POPUP_MENU_BASE (self->menu)->pass_events = TRUE;
}

static void
callback_scroll_stop (GObject *obj, gpointer user_data)
{
    AppletNotifications *self = APPLET_NOTIFICATIONS (user_data);
    CDOS_POPUP_MENU_BASE (self->menu)->pass_events = FALSE;
}

static void
cdos_notification_dispose (GObject *gobject)
{
    AppletNotifications *self = APPLET_NOTIFICATIONS (gobject);
    if (self->blink_timeout_id)
    {
        g_source_remove (self->blink_timeout_id);
        self->blink_timeout_id = 0;
    }

    CdosMessageTray *message_tray = cdos_message_tray_get_default ();
    g_signal_handlers_disconnect_by_data (message_tray, self);

    for (GList *iter = self->notifications; iter; iter = iter->next)
    {
        g_signal_handlers_disconnect_by_data (iter->data, self);
    }
    G_OBJECT_CLASS (applet_notifications_parent_class)->dispose (gobject);
}

static void
cdos_notification_finalize (GObject *gobject)
{
    AppletNotifications *self = APPLET_NOTIFICATIONS (gobject);

    g_list_free (self->notifications);
    g_free (self->menu_text);

    G_OBJECT_CLASS (applet_notifications_parent_class)->finalize (gobject);
}

/* _class_init */
static void
applet_notifications_class_init (AppletNotificationsClass *klass)
{
    GObjectClass *gobject_class = G_OBJECT_CLASS (klass);
    gobject_class->dispose = cdos_notification_dispose;
    gobject_class->finalize = cdos_notification_finalize;

    // CdosAppletClass *parent_class = CDOS_APPLET_CLASS (klass);
    // parent_class->on_applet_clicked = applet_notifications_on_applet_clicked_cb;
}

/* _init */
static void
applet_notifications_init (AppletNotifications *self)
{
    self->menu_manager = cdos_applet_get_menu_manager ((CdosApplet *)self);
    self->notification_count = 0;
    self->notifications = NULL;

    cdos_applet_set_popup_menu_active (CDOS_APPLET (self), TRUE);

    CdosMessageTray *message_tray = cdos_message_tray_get_default ();

    g_signal_connect (message_tray, "notify-applet-update", G_CALLBACK (applet_notifications_notification_added_cb), self);
}

/**
 * applet_notifications_main:
 * @metadata: applet的元数据
 * @orientation: 方向
 * @panel_height: 面板高度
 * @instanceId： 字符串类型的id
 *
 * 新建一个 #CdosApplet 类的实例。
 *
 * Returns: 返回一个 #CdosApplet 的实例。
 */
CdosApplet *
applet_notifications_main (MetaData *metadata, StSide orientation, gfloat panel_height, gchar *instanceId)
{
    AppletNotifications *self = g_object_new (APPLET_TYPE_NOTIFICATIONS, "orientation", orientation, NULL);

    self->menu = cdos_applet_get_popup_menu ((CdosApplet *)self);

    applet_notifications_display (self);

    g_signal_connect (self, "button-press-event", 
                      G_CALLBACK (applet_notifications_on_applet_clicked_cb), self);

    return CDOS_APPLET (self);
}

static void
applet_notifications_display (AppletNotifications *self)
{
    PopupBaseMenuItemInit params_for_new_item;
    cdos_popup_base_menu_item_get_init_params (&params_for_new_item);

    cdos_iconapplet_set_icon_name (CDOS_ICONAPPLET (self), "found-notifications");
    cdos_iconapplet_set_icon_size (CDOS_ICONAPPLET (self), 16);
    cdos_applet_set_applet_tooltip (CDOS_APPLET (self), _("Notifications"));

    self->p_main_container = st_box_layout_new ();
    clutter_actor_set_name (CLUTTER_ACTOR (self->p_main_container), "traycontainer");
    st_box_layout_set_vertical (ST_BOX_LAYOUT (self->p_main_container), TRUE);

    self->p_notification_bin = st_box_layout_new ();
    st_box_layout_set_vertical (ST_BOX_LAYOUT (self->p_notification_bin), TRUE);
    st_widget_set_style_class_name (self->p_notification_bin, "applet-notification");

    self->button_label_box = ST_BOX_LAYOUT (st_box_layout_new ());

    self->menu_text = applet_notifications_stringify (self->notification_count);
    // self->menu_label = cdos_popup_menu_item_new(self->menu_text, &params_for_new_item);
    self->menu_label = cdos_popup_menu_item_new ();
    cdos_popup_menu_item_set_label (self->menu_label, self->menu_text);
    cdos_popup_menu_item_set_label_align (self->menu_label, ST_ALIGN_MIDDLE, 0);

    g_object_set (CLUTTER_ACTOR (self->menu_label),
                  "reactive",
                  FALSE,
                  "can-focus",
                  FALSE,
                  NULL);
    /*第一个参数类型,CdosGenericContainer *actor, 设置属性是在CdosPopupBaseMenuItem内完成的 */
    st_widget_add_style_class_name (ST_WIDGET (cdos_popup_menu_item_get_label (self->menu_label)),
                                    "applet-notification-title");
    self->clear_action = cdos_popup_menu_item_new_with_params (_("Clear notifications"), &params_for_new_item);
    g_signal_connect (self->clear_action,
                      "activate",
                      G_CALLBACK (applet_notifications_clear_all_cb),
                      self);
    clutter_actor_hide (CLUTTER_ACTOR (self->clear_action));
    st_widget_add_style_class_name (ST_WIDGET (self->clear_action), 
                                    "applet-notification-clear");

    cdos_popup_menu_base_add_menu_item (CDOS_POPUP_MENU_BASE (self->menu), 
                                        CDOS_POPUP_BASE_MENU_ITEM (self->menu_label), 
                                        -1);
    g_object_set (self->menu_label, 
                  "x-expand", FALSE, 
                  "x-align", CLUTTER_ACTOR_ALIGN_START, 
                  NULL);

    cdos_popup_menu_base_add_actor (CDOS_POPUP_MENU_BASE (self->menu), 
                                    CLUTTER_ACTOR (self->p_main_container));
    cdos_popup_menu_base_add_menu_item (CDOS_POPUP_MENU_BASE (self->menu), 
                                        CDOS_POPUP_BASE_MENU_ITEM (self->clear_action), 
                                        -1);
    g_object_set (self->clear_action, 
                  "x-expand", FALSE, 
                  "x-align", CLUTTER_ACTOR_ALIGN_END, 
                  NULL);

    self->scroll_view = st_scroll_view_new ();
    g_object_set (self->scroll_view, 
                  "x-fill", TRUE, 
                  "y-fill", TRUE, 
                  "y-align", ST_ALIGN_START, 
                  NULL);
    g_object_set (self->scroll_view, 
                  "style-class", "vfade", 
                  NULL);

    clutter_actor_add_child (CLUTTER_ACTOR (self->p_main_container), CLUTTER_ACTOR (self->scroll_view));
    st_scroll_view_add_child (ST_SCROLL_VIEW (self->scroll_view), CLUTTER_ACTOR (self->p_notification_bin));
    st_scroll_view_set_policy (ST_SCROLL_VIEW (self->scroll_view), GTK_POLICY_NEVER, GTK_POLICY_AUTOMATIC);

    ClutterActor *vscroll = st_scroll_view_get_vscroll_bar (ST_SCROLL_VIEW (self->scroll_view));
    g_signal_connect (vscroll, "scroll-start", G_CALLBACK (callback_scroll_start), self);
    g_signal_connect (vscroll, "scroll-stop", G_CALLBACK (callback_scroll_stop), self);

    applet_notifications_update_list (self);
}

/**
 * applet_notifications_on_orientation_changed:
 * @self: A #AppletNotifications
 * @orientation: 方向
 *
 * 处理通知展示的方向。
 *
 */
/*
static void
applet_notifications_on_orientation_changed (AppletNotifications *self, StSide orientation)
{
    self->p_orientation = orientation;

    if (self->menu)
    {
        cdos_popup_base_menu_item_destroy (CDOS_POPUP_BASE_MENU_ITEM (self->menu));
    }

    self->menu = cdos_applet_get_popup_menu (CDOS_APPLET (self));
    cdos_popup_menu_manager_add_menu (self->menu_manager, CDOS_POPUP_MENU_BASE (self->menu), -1);
    applet_notifications_display (self);
}
*/

/**
 * applet_notifications_update_list:
 * @self: A #AppletNotifications
 *
 * 更新applet里notifications的排序，包括特定情况下的图标，紧急程度高的上浮等等。
 *
 */
static void
applet_notifications_update_list (AppletNotifications *self)
{
    self->notification_count = g_list_length (self->notifications);

    gint max_urgency = -1;
    if (self->notification_count > 0)
    {
        clutter_actor_show (CLUTTER_ACTOR (self->clear_action));
        clutter_actor_show ((ClutterActor *)self->scroll_view);
        clutter_actor_show (CLUTTER_ACTOR (self->p_main_container));
        cdos_iconapplet_set_icon_name (CDOS_ICONAPPLET (self), "found-notifications");
        cdos_text_icon_applet_set_applet_label (CDOS_TEXT_ICON_APPLET (self),
                                                g_strdup_printf ("%d", self->notification_count));
        gint cur_urgency;
        GList *iter = NULL;
        for (iter = self->notifications; iter; iter = iter->next)
        {
            CdosNotification *notification_tmp = iter->data;
            cur_urgency = notification_tmp->urgency;

            if (cur_urgency > max_urgency)
            {
                max_urgency = cur_urgency;
            }
        }

        //测试critical_message时图标状态：max_urgency = 3;
        switch (max_urgency)
        {
            case CDOS_NOTIFICATION_LOW:
                break;

            case CDOS_NOTIFICATION_NORMAL:
            case CDOS_NOTIFICATION_CRITICAL:
                break;


            default:
                g_warning ("Have not correct max_urgency %s:%d %d", __func__, __LINE__, max_urgency);
                break;
        }
    }
    else
    {
        cdos_text_icon_applet_set_applet_label (CDOS_TEXT_ICON_APPLET (self), "");
        cdos_iconapplet_set_icon_name (CDOS_ICONAPPLET (self), "found-notifications-empty");
        clutter_actor_hide (CLUTTER_ACTOR (self->clear_action));
        clutter_actor_hide (CLUTTER_ACTOR (self->p_main_container));
    }
    gchar *strtmp = applet_notifications_stringify (self->notification_count);
    cdos_popup_menu_item_set_label (self->menu_label, strtmp);
    g_free (strtmp);

    gfloat width = 0, height = 0;
    clutter_actor_get_preferred_size (CLUTTER_ACTOR (self->p_notification_bin), NULL, NULL, &width, &height);
    guint count = self->notification_count;

    if (height > 300 || count >= 5)
    {
        clutter_actor_set_size (CLUTTER_ACTOR (self->p_main_container), -1, 300);
    }
    else
    {
        clutter_actor_set_size (CLUTTER_ACTOR (self->p_main_container), -1, -1);
        clutter_actor_queue_relayout (CLUTTER_ACTOR (self->p_notification_bin));
    }

    if (count == 0)
    {
        clutter_actor_set_size (CLUTTER_ACTOR (self->p_main_container), 0, 0);
    }
}

/**
 * applet_notifications_on_applet_clicked_cb:
 * @actor: A #ClutterActor
 * @event: 事件类型
 * @user_data: 回调函数传入参数
 *
 * 这个函数是对父类on_applet_clicked方法地重写，监听applet的“button-press-event” 信号回调；弹出，收起applet
 */
static gboolean
applet_notifications_on_applet_clicked_cb (ClutterActor *actor, ClutterEvent *event, gpointer user_data)
{
    if (clutter_event_get_button (event) != 1)
    {
        return FALSE;
    }

    AppletNotifications *self = APPLET_NOTIFICATIONS (user_data);
    gfloat width, height;
    guint count;

    applet_notifications_update_timestamp (self);

    return FALSE;

    cdos_popup_menu_base_toggle (CDOS_POPUP_MENU_BASE (self->menu));

    clutter_actor_get_size (CLUTTER_ACTOR (self->p_main_container), &width, &height);
    count = self->notification_count;
    if (height > 300 || count >= 5)
    {
        clutter_actor_set_size (CLUTTER_ACTOR (self->p_main_container), -1, 300);
    }
    else
    {
        clutter_actor_set_size (CLUTTER_ACTOR (self->p_main_container), -1, -1);
    }

    if (count == 0)
    {
        clutter_actor_set_size (CLUTTER_ACTOR (self->p_main_container), 0, 0);
        // set menu_label in middle
    }
}

static void
applet_notifications_update_timestamp (AppletNotifications *self)
{
    GList *iter = NULL;
    for (iter = self->notifications; iter; iter = iter->next)
    {
        CdosNotification *notification = iter->data;
        ClutterActor *clutter_text;
        clutter_text = st_label_get_clutter_text (notification->time_label);
        gchar *date = cdos_notification_timeify (notification, NULL);
        clutter_text_set_markup ((ClutterText *)clutter_text, date);
        g_free (date);
    }
}
