/**
 * Copyright (c) 2016-2018 CPU and Fundamental Software Research Center, CAS
 *
 * This software is published by the license of CPU-OS License，you can use and
 * distibuted this software under this License. See CPU-OS Liense for more detail.
 *
 * You should have receive a copy of CPU-OS License. If not, please contact us
 * by email <support_os@cpu-os.ac.cn>
 *
 **/
#include <linux/rfkill.h>

#include <unistd.h>
#include <sys/time.h>
#include <fcntl.h>
#include <errno.h>
#include <libupower-glib/upower.h>

#include "cdos-util.h"
#include "applet-notifications2.h"
#include "applet-airplane-mode-button.h"

#define BUF_SIZE 128

/*飞行模式按钮*/

typedef struct
{
    GSource source;
    AppletAirplaneModeButton *self;
}RfkillSource;

struct _AppletAirplaneModeButtonPrivate
{
    gboolean is_open;
    guint watch_id;
    int rfkill_fd;
    GSource *source;
    GHashTable *devices;
};

G_DEFINE_TYPE_WITH_PRIVATE (AppletAirplaneModeButton, applet_airplane_mode_button, APPLET_TYPE_SETTINGS_BUTTON);

static void
rfkill_read_events (AppletAirplaneModeButton *self, int fd)
{
    AppletAirplaneModeButtonPrivate *priv = self->priv;
    GHashTableIter iter;
    gpointer key, value;
    gboolean blocked = TRUE;

    while (1)
    {
        struct rfkill_event event;
        ssize_t len;
        len = read(fd, &event, sizeof(event));
        if (len < 0)
        {
            break;
        }
        if (len != RFKILL_EVENT_SIZE_V1)
        {
            g_warning ("%s: event size error", __func__);
            continue;
        }
        g_debug ("%s: idx %d, op %d, soft %d, hard %d", __func__, event.idx, event.op, event.soft, event.hard);
        if (event.type != RFKILL_TYPE_ALL &&
            event.type != RFKILL_TYPE_WLAN &&
            event.type != RFKILL_TYPE_BLUETOOTH &&
            event.type != RFKILL_TYPE_WIMAX &&
            event.type != RFKILL_TYPE_WWAN)
        {
            continue;
        }
        /* soft/hard: 0启用，1禁用设备 */
        if (event.op == RFKILL_OP_ADD ||
            event.op == RFKILL_OP_CHANGE)
        {
            g_hash_table_insert (priv->devices,
                                 GINT_TO_POINTER (event.idx),
                                 GINT_TO_POINTER (event.soft | event.hard));
        }
        else if (event.op == RFKILL_OP_DEL)
        {
            g_hash_table_remove (priv->devices,
                                 GINT_TO_POINTER (event.idx));
        }
        else if (event.op == RFKILL_OP_CHANGE_ALL)
        {
            if (event.soft == 0 && event.hard == 0)
                blocked = FALSE;
        }
    }

    g_hash_table_iter_init (&iter, priv->devices);
    while (g_hash_table_iter_next (&iter, &key, &value))
    {
        guint state = GPOINTER_TO_UINT (value);
        if (state == FALSE)
        {
            blocked = FALSE;
            break;
        }
    }
    if (priv->is_open != blocked)
    {
        priv->is_open = blocked;
        applet_settings_button_set_open (APPLET_SETTINGS_BUTTON (self), priv->is_open);
    }
}

static gboolean
rfkill_event_source_prepare (GSource *source, int *timeout)
{
    *timeout = -1;
    return FALSE;
}

static gboolean
rfkill_event_source_dispatch (GSource *source, GSourceFunc calback, gpointer data)
{
    RfkillSource *src = (RfkillSource *)source;
    AppletAirplaneModeButton *self = src->self;
    rfkill_read_events (self, self->priv->rfkill_fd);
    return TRUE;
}

static GSourceFuncs rfkill_source_funcs =
{
    rfkill_event_source_prepare,
    NULL,
    rfkill_event_source_dispatch,
    NULL
};

static GSource *
rfkill_event_source_new (int fd, AppletAirplaneModeButton *self)
{
    RfkillSource *source;
    source = (RfkillSource *)g_source_new (&rfkill_source_funcs, sizeof (RfkillSource));
    source->self = self;
    g_source_add_unix_fd (&source->source, fd, G_IO_IN | G_IO_HUP | G_IO_ERR);

    return &source->source;
}

static void
applet_airplane_mode_button_activate (AppletSettingsButton *self)
{
    AppletAirplaneModeButtonPrivate *priv;

    AppletAirplaneModeButton *airplane_button = APPLET_AIRPLANE_MODE_BUTTON (self);
    priv = airplane_button->priv;

    if (g_hash_table_size (priv->devices) == 0)
    {
        return;
    }

    gboolean block = !priv->is_open;

    GHashTableIter iter;
    gpointer key, value;
    struct rfkill_event event;

    g_hash_table_iter_init (&iter, priv->devices);
    while (g_hash_table_iter_next (&iter, &key, &value))
    {
        guint idx = GPOINTER_TO_UINT (key);
        event.idx = idx;
        event.op = RFKILL_OP_CHANGE;
        event.type = 0;
        event.soft = block ? 1 : 0;
        event.hard = 0;

        if (write (priv->rfkill_fd, &event, sizeof(struct rfkill_event)) < 0)
        {
            continue;
        }
    }
}

static void
on_open_changed_callback (GObject *gobject,
                          GParamSpec *pspec,
                          gpointer user_data)
{
    applet_settings_button_open_state_show (APPLET_SETTINGS_BUTTON (gobject));
}

static void
applet_airplane_mode_button_finalize (GObject *gobject)
{
    AppletAirplaneModeButton *self = APPLET_AIRPLANE_MODE_BUTTON (gobject);
    AppletAirplaneModeButtonPrivate *priv = self->priv;

    if (priv->rfkill_fd > 0)
        close(priv->rfkill_fd);
    g_hash_table_destroy (priv->devices);
    g_source_unref (priv->source);

    G_OBJECT_CLASS (applet_airplane_mode_button_parent_class)->finalize (gobject);
}

static void
applet_airplane_mode_button_init (AppletAirplaneModeButton *self)
{
    AppletAirplaneModeButtonPrivate *priv;
    priv = self->priv = applet_airplane_mode_button_get_instance_private (self);
    priv->is_open = FALSE;

    priv->devices = g_hash_table_new (NULL, NULL);

    priv->rfkill_fd = open("/dev/rfkill", O_RDWR);
    if (priv->rfkill_fd < 0)
    {
        g_warning ("Can't open /dev/rfkill: %s", strerror(errno));
        return;
    }
    if (fcntl(priv->rfkill_fd, F_SETFL, O_NONBLOCK) < 0)
    {
        close(priv->rfkill_fd);
        priv->rfkill_fd = -1;
        g_warning ("Can't set rfkill fd to NONBLOCK");
        return;
    }
    rfkill_read_events (self, priv->rfkill_fd);

    priv->source = rfkill_event_source_new (priv->rfkill_fd, self);
    g_source_attach (priv->source, NULL);
}

static void
applet_airplane_mode_button_class_init (AppletAirplaneModeButtonClass *klass)
{
    GObjectClass *gobject_class = G_OBJECT_CLASS(klass);
    gobject_class->finalize = applet_airplane_mode_button_finalize;

    AppletSettingsButtonClass *parent_class = APPLET_SETTINGS_BUTTON_CLASS (klass);
    parent_class->activate = applet_airplane_mode_button_activate;
}

AppletAirplaneModeButton *
applet_airplane_mode_button_new (AppletNotifications2 *notification,
                                 gboolean keep_notification,
                                 gboolean show_open_state)
{
    AppletAirplaneModeButton *airplane_button;

    airplane_button = g_object_new (APPLET_TYPE_AIRPLANE_MODE_BUTTON,
                                    "icon-name", "notifications-airplane-mode",
                                    "icon-size", 24,
                                    "text", _("Airplane Mode"),
                                    NULL);

    applet_settings_button_set_notification (APPLET_SETTINGS_BUTTON (airplane_button), notification);
    applet_settings_button_keep_notification (APPLET_SETTINGS_BUTTON (airplane_button), keep_notification);
    applet_settings_button_show_open_state (APPLET_SETTINGS_BUTTON (airplane_button), show_open_state);

    g_signal_connect (airplane_button,
                      "notify::open",
                      G_CALLBACK (on_open_changed_callback),
                      NULL);

    return airplane_button;
}
