/**
 * Copyright (c) 2016-2018 CPU and Fundamental Software Research Center, CAS
 *
 * This software is published by the license of CPU-OS License，you can use and
 * distibuted this software under this License. See CPU-OS Liense for more detail.
 *
 * You should have receive a copy of CPU-OS License. If not, please contact us
 * by email <support_os@cpu-os.ac.cn>
 *
 **/
#include <gio/gio.h>
#include "applet-luminance-button.h"
#include "applet-airplane-mode-button.h"
#include "config.h"

/*亮度调节按钮*/

#define DEFAULT_STEP 20
#define MAX_VALUE 100
#define MIN_VALUE 0
#define OFFSET 1
#define INTERVAL 5000

struct _AppletLuminanceButtonPrivate
{
    GDBusConnection *connection;
    GDBusProxy *proxy;
    gint32 current_brightness;
    guint step;
    guint timer_id;
    gulong brightness_changed_id;
    gboolean enabled;
};

G_DEFINE_TYPE_WITH_PRIVATE (AppletLuminanceButton, applet_luminance_button, APPLET_TYPE_SETTINGS_BUTTON);

static gboolean
applet_luminance_button_set_value (AppletLuminanceButton *self, gint32 brightness)
{
    GVariant *variant;
    GError *error = NULL;

    if (!self->priv->enabled)
        return TRUE;

    variant = g_variant_new_parsed ("('org.freedesktop.SettingsDaemon.Power.Screen',"
                                    "'Brightness', %v)",
                                    g_variant_new_int32 (brightness));
    g_dbus_proxy_call_sync (self->priv->proxy,
                            "org.freedesktop.DBus.Properties.Set",
                            variant,
                            G_DBUS_CALL_FLAGS_NONE,
                            -1,
                            NULL,
                            &error);
    if (error)
    {
        g_warning ("%s", error->message);
        g_error_free (error);
        return FALSE;
    }
    return TRUE;
}

static gboolean
applet_luminance_button_step_up (AppletLuminanceButton *luminance_button)
{
    AppletLuminanceButtonPrivate *priv;
    gboolean res;
    gint32 new_percentage;

    priv = luminance_button->priv;

    /*亮度值修正，往上加会少加1亮度值*/
    new_percentage = priv->current_brightness + priv->step + OFFSET;
    if (new_percentage >= MAX_VALUE)
    {
        new_percentage = MAX_VALUE;
    }

    res = applet_luminance_button_set_value (luminance_button, new_percentage);

    return res;
}

static gboolean
applet_luminance_button_step_down (AppletLuminanceButton *luminance_button)
{
    AppletLuminanceButtonPrivate *priv;
    gboolean res;
    guint new_percentage;

    priv = luminance_button->priv;

    /*亮度值修正，往下减会多减1亮度值*/
    new_percentage = priv->current_brightness - priv->step + OFFSET;
    if (new_percentage <= MIN_VALUE)
    {
        new_percentage = MIN_VALUE;
    }
    res = applet_luminance_button_set_value (luminance_button, new_percentage);

    return res;
}

static gboolean
applet_luminance_button_on_scroll_event (ClutterActor *actor,
                                         ClutterScrollEvent *event)
{
    AppletLuminanceButton *self;
    const gchar *method;
    GError *error = NULL;
    GVariant *res;

    self = (AppletLuminanceButton *)actor;
    ClutterScrollDirection dir = clutter_event_get_scroll_direction ((ClutterEvent *)event);
    if (dir == CLUTTER_SCROLL_UP)
        method = "StepUp";
    else
        method = "StepDown";

    res = g_dbus_proxy_call_sync (self->priv->proxy,
                                  method,
                                  NULL,
                                  G_DBUS_CALL_FLAGS_NONE,
                                  -1,
                                  NULL,
                                  &error);
    if (error)
    {
        g_warning ("%s", error->message);
        g_error_free (error);
    }
    else
        g_variant_unref (res);
    return TRUE;
}

static void
on_open_changed_callback (GObject *gobject,
                          GParamSpec *pspec,
                          gpointer user_data)
{
    applet_settings_button_open_state_show (APPLET_SETTINGS_BUTTON (gobject));
}

static gboolean
reset_text_to_default (gpointer user_data)
{
    AppletLuminanceButton *luminance_button;
    AppletLuminanceButtonPrivate *priv;

    luminance_button = (AppletLuminanceButton *)user_data;
    priv = luminance_button->priv;

    applet_settings_button_set_text (APPLET_SETTINGS_BUTTON (luminance_button),
                                     _("Luminance"));

    priv->timer_id = 0;

    return G_SOURCE_REMOVE;
}

static void
applet_luminance_button_add_timer (AppletLuminanceButton *luminance_button)
{
    AppletLuminanceButtonPrivate *priv;

    priv = luminance_button->priv;

    priv->timer_id = g_timeout_add (INTERVAL,
                                    (GSourceFunc) reset_text_to_default,
                                    luminance_button);
}

static void
applet_luminance_button_remove_timer (AppletLuminanceButton *luminance_button)
{
    AppletLuminanceButtonPrivate *priv;

    priv = luminance_button->priv;

    if (priv->timer_id)
    {
        g_source_remove (priv->timer_id);
        priv->timer_id = 0;
    }
}

static void
applet_luminance_button_activate (AppletSettingsButton *button)
{
    AppletLuminanceButton *luminance_button = APPLET_LUMINANCE_BUTTON (button);
    AppletLuminanceButtonPrivate *priv;

    priv = luminance_button->priv;
    applet_luminance_button_remove_timer (luminance_button);

    if (priv->current_brightness == MAX_VALUE)
        applet_luminance_button_step_down (luminance_button);
    else
        applet_luminance_button_step_up (luminance_button);
}

static void
applet_luminance_button_show_current_brightness (AppletLuminanceButton *self)
{
    gchar *percentage_str;
    applet_luminance_button_remove_timer (self);
    percentage_str = g_strdup_printf ("%u%%", self->priv->current_brightness);
    applet_settings_button_set_text (APPLET_SETTINGS_BUTTON (self),
                                            percentage_str);
    g_free (percentage_str);
    applet_luminance_button_add_timer (self);
}

static void
sync_screen_brightness (AppletLuminanceButton *self)
{
    GVariant *value;
    gint32 brightness;
    value = g_dbus_proxy_get_cached_property (self->priv->proxy, "Brightness");
    if (!value)
    {
        self->priv->current_brightness = -1;
        return;
    }
    brightness = g_variant_get_int32 (value);
    if (self->priv->current_brightness != brightness)
    {
        self->priv->current_brightness = brightness;
        applet_luminance_button_show_current_brightness (self);
    }
    g_variant_unref (value);
}

static void
on_brightness_changed (GDBusProxy *proxy,
                       GVariant *changed,
                       GVariant *invalidated,
                       AppletLuminanceButton *self)
{
    sync_screen_brightness (self);
}

static void
on_dbus_proxy_ready (GObject *source,
                     GAsyncResult *res,
                     AppletLuminanceButton *self)
{
    GError *error = NULL;
    AppletLuminanceButtonPrivate *priv = self->priv;
    priv->proxy = g_dbus_proxy_new_finish (res, &error);
    if (error)
    {
        g_warning ("%s", error->message);
        g_clear_error (&error);
        return;
    }

    sync_screen_brightness (self);
    if (priv->current_brightness >= MIN_VALUE && priv->current_brightness < MAX_VALUE)
    {
        priv->enabled = TRUE;
        applet_settings_button_set_open (APPLET_SETTINGS_BUTTON (self), TRUE);
    }
    else
    {
        priv->enabled = FALSE;
        applet_settings_button_set_open (APPLET_SETTINGS_BUTTON (self), FALSE);
    }
    /*
     *当调节亮度+/-的热键按下时，同步显示当前的亮度值。
     */
    priv->brightness_changed_id = g_signal_connect (priv->proxy, "g-properties-changed",
                                                    G_CALLBACK (on_brightness_changed),
                                                    self);
}

static void
applet_luminance_button_dispose (GObject *gobject)
{
    AppletLuminanceButton *self = (AppletLuminanceButton *)gobject;
    AppletLuminanceButtonPrivate *priv = self->priv;
    g_clear_signal_handler (&priv->brightness_changed_id, priv->proxy);
    if (priv->connection)
    {
        g_object_unref (priv->connection);
        priv->connection = NULL;
    }
    if (priv->proxy)
    {
        g_object_unref (priv->proxy);
        priv->proxy = NULL;
    }
    applet_luminance_button_remove_timer (self);
    G_OBJECT_CLASS (applet_luminance_button_parent_class)->dispose (gobject);
}

static void
applet_luminance_button_finalize (GObject *gobject)
{
    G_OBJECT_CLASS (applet_luminance_button_parent_class)->finalize (gobject);
}

static void
applet_luminance_button_init (AppletLuminanceButton *self)
{
    AppletLuminanceButtonPrivate *priv;
    GError *error = NULL;
    priv = self->priv = applet_luminance_button_get_instance_private (self);
    priv->step = DEFAULT_STEP;
    //priv->activate_timer_id = 0;
    priv->timer_id = 0;
    priv->enabled = FALSE;

    priv->connection = g_bus_get_sync (G_BUS_TYPE_SESSION, NULL, &error);
    g_dbus_proxy_new (priv->connection,
                      G_DBUS_PROXY_FLAGS_DO_NOT_AUTO_START,
                      NULL,
                      "org.freedesktop.SettingsDaemon",
                      "/org/freedesktop/SettingsDaemon/Power",
                      "org.freedesktop.SettingsDaemon.Power.Screen",
                      NULL,
                      (GAsyncReadyCallback)on_dbus_proxy_ready,
                      self);
}

static void
applet_luminance_button_class_init (AppletLuminanceButtonClass *klass)
{
    GObjectClass *gobject_class = G_OBJECT_CLASS(klass);

    gobject_class->dispose = applet_luminance_button_dispose;
    gobject_class->finalize = applet_luminance_button_finalize;

    ClutterActorClass *actor_class = CLUTTER_ACTOR_CLASS (klass);
    actor_class->scroll_event = applet_luminance_button_on_scroll_event;

    AppletSettingsButtonClass *parent_class = APPLET_SETTINGS_BUTTON_CLASS (klass);
    parent_class->activate = applet_luminance_button_activate;
}

AppletLuminanceButton *
applet_luminance_button_new (AppletNotifications2 *notification,
                             gboolean keep_notification,
                             gboolean show_open_state)
{
    AppletLuminanceButton *luminance_button;
    luminance_button = g_object_new (APPLET_TYPE_LUMINANCE_BUTTON,
                                    "icon-name", "notifications-luminance",
                                    "icon-size", 24,
                                    "text", _("Luminance"),
                                    NULL);
    applet_settings_button_set_notification (APPLET_SETTINGS_BUTTON (luminance_button), notification);
    applet_settings_button_keep_notification (APPLET_SETTINGS_BUTTON (luminance_button), keep_notification);
    applet_settings_button_show_open_state (APPLET_SETTINGS_BUTTON (luminance_button), show_open_state);

    g_signal_connect (luminance_button,
                      "notify::open",
                      G_CALLBACK (on_open_changed_callback),
                      NULL);

    return luminance_button;
}
