/** 
 * Copyright (c) 2016-2018 CPU and Fundamental Software Research Center, CAS
 *  
 * This software is published by the license of CPU-OS License，you can use and 
 * distibuted this software under this License. See CPU-OS Liense for more detail.
 *
 * You should have receive a copy of CPU-OS License. If not, please contact us 
 * by email <support_os@cpu-os.ac.cn>  
 * 
 **/
/**
 * SECTION: cdos-music-player
 * @include: cdos-music-player.h
 */

#define DBUS_NAME "org.freedesktop.DBus"
#define INTERFACE_NAME "org.freedesktop.DBus"
#define DBUS_PATH "/org/freedesktop/DBus"

#include <math.h>
#include <string.h>
#include <meta/util.h>

#include "cdos-mpris2-player.h"
#include "cdos-start.h"
#include "cdos-scale.h"
#include "applet-player.h"

struct _AppletPlayer
{
    StWidget parent;
    CdosMpris2Player *player;
    StLabel *title_label;
    StLabel *artists_label;
    StLabel *album_label;
    ClutterActor *cover;
    ClutterActor *seek_bar;
    StIcon *prev_icon;
    StIcon *play_pause_icon;
    StIcon *next_icon;
    ClutterActor *previous;
    ClutterActor *next;
    ClutterActor *play_pause;
    ClutterActor *buttons;
    StWidget *time_current;
    StWidget *time_length;
    guint64 position;
    guint length;
    gchar *trackid;

    StTextureCache *texture_cache;
    gfloat spacing;
    gfloat avail_width;

    gulong hover_changed_id;
};

G_DEFINE_TYPE (AppletPlayer, applet_player, ST_TYPE_WIDGET);

static gchar *
seconds_to_string (guint length)
{
    if (length == 0)
        return g_strdup ("0:00");

    gint hours, minutes, seconds;
    hours = length / 3600;
    length %= 3600;
    minutes = length / 60;
    seconds = length % 60;

    char min[3], sec[3];
    min[0] = minutes / 10 + '0';
    min[1] = minutes % 10 + '0';
    min[2] = '\0';
    sec[0] = seconds / 10 + '0';
    sec[1] = seconds % 10 + '0';
    sec[2] = '\0';

    gchar *time_text = NULL;
    if (hours > 0)
        time_text = g_strdup_printf ("%d:%s:%s", hours, min, sec);
    else
        time_text = g_strdup_printf ("%d:%s", minutes, sec);
    return time_text;
}

#define compute_child_size(SIZE) gfloat cover_min_##SIZE = 0, cover_nat_##SIZE = 0; \
    gfloat title_min_##SIZE, title_nat_##SIZE; \
    gfloat artists_min_##SIZE = 0, artists_nat_##SIZE = 0; \
    gfloat album_min_##SIZE, album_nat_##SIZE; \
    gfloat seek_min_##SIZE, seek_nat_##SIZE; \
    gfloat buttons_min_##SIZE, buttons_nat_##SIZE; \
    clutter_actor_get_preferred_##SIZE ((ClutterActor *)self->title_label, -1, &title_min_##SIZE, &title_nat_##SIZE); \
    clutter_actor_get_preferred_##SIZE ((ClutterActor *)self->artists_label, -1, &artists_min_##SIZE, &artists_nat_##SIZE); \
    clutter_actor_get_preferred_##SIZE ((ClutterActor *)self->album_label, -1, &album_min_##SIZE, &album_nat_##SIZE); \
    clutter_actor_get_preferred_##SIZE (self->cover, -1, &cover_min_##SIZE, &cover_nat_##SIZE); \
    clutter_actor_get_preferred_##SIZE (self->seek_bar, -1, &seek_min_##SIZE, &seek_nat_##SIZE); \
    clutter_actor_get_preferred_##SIZE (self->buttons, -1, &buttons_min_##SIZE, &buttons_nat_##SIZE);

static void
applet_player_get_preferred_width (ClutterActor *actor,
                                   gfloat        for_height,
                                   gfloat       *min_width_p,
                                   gfloat       *natural_width_p)
{
    StThemeNode *theme_node = st_widget_get_theme_node ((StWidget *)actor);
    st_theme_node_adjust_for_height (theme_node, &for_height);

    AppletPlayer *self = APPLET_PLAYER (actor);

    compute_child_size (width);

    gfloat left_min_width, right_min_width;
    gfloat left_nat_width, right_nat_width;

    right_min_width = MAX (MAX (title_min_width, artists_min_width), album_min_width);
    right_min_width = MAX (MAX (right_min_width, seek_min_width), buttons_min_width);
    left_min_width = cover_min_width;

    right_nat_width = MAX (MAX (title_nat_width, artists_nat_width), album_nat_width);
    right_nat_width = MAX (MAX (right_nat_width, seek_nat_width), buttons_nat_width);
    left_nat_width = cover_nat_width;

    *min_width_p = left_min_width + right_min_width;
    *natural_width_p = left_nat_width + right_nat_width + self->spacing;

    st_theme_node_adjust_preferred_width (theme_node, min_width_p, natural_width_p);
}

static void
applet_player_get_preferred_height (ClutterActor *actor,
                                    gfloat        for_width,
                                    gfloat       *min_height_p,
                                    gfloat       *natural_height_p)
{
    StThemeNode *theme_node = st_widget_get_theme_node ((StWidget *)actor);
    st_theme_node_adjust_for_width (theme_node, &for_width);

    AppletPlayer *self = APPLET_PLAYER (actor);

    compute_child_size (height);

    gfloat left_min_height, right_min_height;
    gfloat left_nat_height, right_nat_height;

    right_min_height = title_min_height + artists_min_height + album_min_height + seek_min_height + buttons_min_height;
    right_nat_height = title_nat_height + artists_nat_height + album_nat_height + seek_nat_height + buttons_nat_height;

    left_min_height = cover_min_height;
    left_nat_height = cover_nat_height;

    *min_height_p = MAX (left_min_height, right_min_height);
    *natural_height_p = MAX (left_nat_height, right_nat_height);

    st_theme_node_adjust_preferred_height (theme_node, min_height_p, natural_height_p);
}

static void
applet_player_allocate (ClutterActor *actor,
                        const ClutterActorBox *box,
                        ClutterAllocationFlags flags)
{
    AppletPlayer *self = APPLET_PLAYER (actor);
    StThemeNode *theme_node = st_widget_get_theme_node ((StWidget *)actor);

    clutter_actor_set_allocation (actor, box, flags);

    compute_child_size (width);
    compute_child_size (height);

    ClutterActorBox content_box, child_box;
    st_theme_node_get_content_box (theme_node, box, &content_box);

    int x, y;

    gfloat avail_width = content_box.x2 - content_box.x1;

    x = content_box.x1;
    y = content_box.y1;

    self->avail_width = avail_width;
    /* cover */
    child_box.x1 = x;
    child_box.x2 = child_box.x1 + cover_nat_width;
    child_box.y1 = y;
    child_box.y2 = child_box.y1 + cover_nat_height;
    clutter_actor_allocate (self->cover, &child_box, flags);

    x = child_box.x2 + self->spacing;
    avail_width -= x;
    /* title */
    gfloat w = MIN (avail_width, title_nat_width);
    child_box.x1 = x + (int)(avail_width - w) / 2;
    child_box.x2 = child_box.x1 + w;
    child_box.y1 = y;
    child_box.y2 = child_box.y1 + title_nat_height;
    clutter_actor_allocate ((ClutterActor *)self->title_label, &child_box, flags);

    y = child_box.y2;
    /* artist */
    w = artists_nat_width;
    if (w > avail_width)
        w = avail_width;
    child_box.x1 = x + (int)(avail_width - w) / 2;
    child_box.x2 = child_box.x1 + w;
    child_box.y1 = y;
    child_box.y2 = child_box.y1 + artists_nat_height;
    clutter_actor_allocate ((ClutterActor *)self->artists_label, &child_box, flags);

    y = child_box.y2;
    /* album */
    w = album_nat_width;
    if (w > avail_width)
        w = avail_width;
    child_box.x1 = x + (int)(avail_width - w) / 2;
    child_box.x2 = child_box.x1 + w;
    child_box.y1 = y;
    child_box.y2 = child_box.y1 + album_nat_height;
    clutter_actor_allocate ((ClutterActor *)self->album_label, &child_box, flags);

    y = child_box.y2;
    /* seek bar */
    w = avail_width;
    if (w < seek_nat_width)
        w = seek_nat_width;
    child_box.x1 = x + (avail_width - w) / 2.0;
    child_box.x2 = child_box.x1 + w;
    child_box.y1 = y;
    child_box.y2 = y + seek_nat_height;
    clutter_actor_allocate (self->seek_bar, &child_box, flags);
    y = child_box.y2;

    /* time */
    gfloat current_min_width, current_nat_width, current_min_height, current_nat_height;
    gfloat total_min_width, total_nat_width, total_min_height, total_nat_height;
    clutter_actor_get_preferred_width ((ClutterActor *)self->time_current, -1, &current_min_width, &current_nat_width),
    clutter_actor_get_preferred_height ((ClutterActor *)self->time_current, current_nat_width, &current_min_height, &current_nat_height);

    clutter_actor_get_preferred_width ((ClutterActor *)self->time_length, -1, &total_min_width, &total_nat_width);
    clutter_actor_get_preferred_height ((ClutterActor *)self->time_length, total_nat_width, &total_min_height, &total_nat_height);

    gfloat x_axis, y_axis;
    x_axis = child_box.x2;
    y_axis = child_box.y1 - total_nat_height;

    gint x0 = child_box.x1 + 0.5;
    child_box.x1 = x0;
    child_box.x2 = child_box.x1 + current_nat_width;
    child_box.y1 = (int)(y_axis + 0.5);
    child_box.y2 = child_box.y1 + current_nat_height;
    clutter_actor_allocate ((ClutterActor *)self->time_current, &child_box, flags);

    child_box.x1 = (int)(x_axis - total_nat_width);
    child_box.x2 = child_box.x1 + total_nat_width;
    child_box.y1 = (int)(y_axis + 0.5);
    child_box.y2 = y_axis + total_nat_height;
    clutter_actor_allocate ((ClutterActor *)self->time_length, &child_box, flags);

    /* buttons */
    w = buttons_nat_width;
    if (w > avail_width)
        w = avail_width;
    child_box.x1 = x + (avail_width - w) / 2.0;
    child_box.x2 = child_box.x1 + w;
    child_box.y1 = y;
    child_box.y2 = y + buttons_nat_height;
    clutter_actor_allocate (self->buttons, &child_box, flags);
}

static gboolean
applet_player_on_button_release (ClutterActor *actor, ClutterButtonEvent *event)
{
    return TRUE;
}

static void
applet_player_style_changed (StWidget *widget)
{
    AppletPlayer *self = APPLET_PLAYER (widget);

    StThemeNode *theme_node = st_widget_get_theme_node (widget);
    self->spacing = st_theme_node_get_length (theme_node, "spacing");

    ST_WIDGET_CLASS (applet_player_parent_class)->style_changed (widget);
}

static void
applet_player_on_connected (CdosMpris2Player *mpris2,
                            gboolean connected,
                            AppletPlayer *self)
{
    if (connected)
    {
        clutter_actor_show ((ClutterActor *)self);
        gboolean can_seek = cdos_mpris2_player_get_can_seek (mpris2);
        gboolean can_go_next = cdos_mpris2_player_get_can_go_next (mpris2);
        gboolean can_go_previous = cdos_mpris2_player_get_can_go_previous (mpris2);
        gboolean can_play = cdos_mpris2_player_get_can_play (mpris2);
        gboolean can_pause = cdos_mpris2_player_get_can_pause (mpris2);

        clutter_actor_set_reactive (self->seek_bar, can_seek);
        clutter_actor_set_reactive (self->previous, can_go_previous);
        clutter_actor_set_reactive (self->next, can_go_next);
        clutter_actor_set_reactive (self->play_pause, can_play || can_pause);
    }
    else
    {
        //clutter_actor_hide ((ClutterActor *)self);
        clutter_actor_destroy ((ClutterActor *)self);
    }
}

static void
applet_player_on_playback_status_changed (CdosMpris2Player *mpris2,
                                          CdosMpris2PlaybackStatus status,
                                          AppletPlayer *self)
{
    if (status == CDOS_MPRIS2_PLAYING)
    {
        st_icon_set_icon_name (self->play_pause_icon, "player-pause");
    }
    else
        st_icon_set_icon_name (self->play_pause_icon, "player-play");
}

static void
applet_player_on_playback_tick (CdosMpris2Player *mpris2,
                                gint64 position, // us
                                AppletPlayer *self)
{
    gint value = position / 1000000l;

    self->position = position;

    cdos_scale_set_value ((CdosScale *)self->seek_bar, value);

    gchar *time_text = seconds_to_string (value);
    st_label_set_text ((StLabel *)self->time_current, time_text);
    g_free (time_text);
}

static void
applet_player_on_metadata (CdosMpris2Player *mpris2,
                           CdosMpris2Metadata *metadata,
                           AppletPlayer *self)
{

    guint length = cdos_mpris2_metadata_get_length (metadata) / 1000000l;
    cdos_scale_set_range ((CdosScale *)self->seek_bar, 0, length, 1);
    self->position = 0;
    cdos_scale_set_value ((CdosScale *)self->seek_bar, 0);
    gchar *time_text = seconds_to_string (length);
    st_label_set_text ((StLabel *)self->time_length, time_text);
    g_free (time_text);

    const gchar *title = cdos_mpris2_metadata_get_title (metadata);
    gchar *artists = cdos_mpris2_metadata_get_artists (metadata);
    const gchar *album = cdos_mpris2_metadata_get_album (metadata);
    const gchar *arturl = cdos_mpris2_metadata_get_arturl (metadata);
    ClutterActor *cover = NULL;

    float paint_scale, resource_scale;
    clutter_actor_get_resource_scale (CLUTTER_ACTOR (self), &resource_scale);
    ClutterActor *stage = clutter_actor_get_stage (CLUTTER_ACTOR (self));
    StThemeContext *context = st_theme_context_get_for_stage (CLUTTER_STAGE (stage));
    g_object_get (context, "scale-factor", &paint_scale, NULL);

    if (!arturl)
    {
        GIcon *gicon = g_themed_icon_new ("album-cover");
        cover = st_texture_cache_load_gicon (self->texture_cache,
                                             NULL,
                                             gicon,
                                             120,
                                             paint_scale,
                                             resource_scale);
        g_object_unref (gicon);
    }
    else
    {
        GFile *icon_file = g_file_new_for_uri (arturl);
        cover = st_texture_cache_load_file_async (self->texture_cache, icon_file, 120, 120, paint_scale, resource_scale);
        g_object_unref (icon_file);
    }
    st_bin_set_child ((StBin *)self->cover, cover);

    if (title)
        st_label_set_text (self->title_label, title);
    else
        st_label_set_text (self->title_label, "");

    if (artists)
        st_label_set_text (self->artists_label, artists);
    else
        st_label_set_text (self->artists_label, "");

    if (album)
        st_label_set_text (self->album_label, album);
    else
        st_label_set_text (self->album_label, "");
    /*
    if (self->trackid)
        g_free (self->trackid);
    self->trackid = g_strdup (cdos_mpris2_metadata_get_trackid (metadata));
    */
}

static void
applet_player_on_control (CdosMpris2Player *mpris2,
                          guint control_flags,
                          AppletPlayer *self)
{
    gboolean can_seek = control_flags | CDOS_MPRIS2_CAN_SEEK;
    gboolean can_go_next = control_flags | CDOS_MPRIS2_CAN_GO_NEXT;
    gboolean can_go_previous = control_flags | CDOS_MPRIS2_CAN_GO_PREVIOUS;
    gboolean can_play = control_flags | CDOS_MPRIS2_CAN_PLAY;
    gboolean can_pause = control_flags | CDOS_MPRIS2_CAN_PAUSE;

    clutter_actor_set_reactive (self->seek_bar, can_seek);
    clutter_actor_set_reactive (self->previous, can_go_previous);
    clutter_actor_set_reactive (self->next, can_go_next);
    clutter_actor_set_reactive (self->play_pause, can_play || can_pause);
}

static void
applet_player_seeked (CdosScale *scale, gfloat value, AppletPlayer *self)
{
    gint64 position = cdos_mpris2_player_get_accurate_position (self->player);
    position = position / 1000000l;
    cdos_mpris2_player_seek (self->player, ((gint)value - position) * 1000000l);
}

static gboolean
applet_player_prev (ClutterActor *actor,
                    ClutterEvent *event,
                    AppletPlayer *self)
{
    cdos_mpris2_player_previous (self->player);
    return TRUE;
}

static gboolean
applet_player_play_pause (ClutterActor *actor,
                          ClutterEvent *event,
                          AppletPlayer *self)
{
    cdos_mpris2_player_play_pause (self->player);
    return TRUE;
}

static gboolean
applet_player_next (ClutterActor *actor,
                    ClutterEvent *event,
                    AppletPlayer *self)
{
    cdos_mpris2_player_next (self->player);
    return TRUE;
}

static gboolean
applet_player_raise (ClutterActor *actor,
                     ClutterEvent *event,
                     AppletPlayer *self)
{
    CdosMpris2Player *player = self->player;
    const gchar *desktop_entry = cdos_mpris2_player_get_player_desktop_entry (player);
    CdosApp *app = NULL;
    if (desktop_entry)
    {
        gchar *desktop_id = g_strconcat (desktop_entry, ".desktop", NULL);
        CdosAppSystem *app_system = cdos_app_system_get_default ();
        app = cdos_app_system_lookup_app (app_system, desktop_id);
        g_free (desktop_id);
    }
    if (app)
        cdos_app_activate (app);
    else
        cdos_mpris2_player_raise_player (self->player);
    return TRUE;
}

static void
applet_player_dispose (GObject *obj)
{
    AppletPlayer * self = APPLET_PLAYER (obj);
    if (self->hover_changed_id)
    {
        g_signal_handler_disconnect (obj, self->hover_changed_id);
        self->hover_changed_id = 0;
    }
    g_clear_pointer (&self->player, g_object_unref);

    G_OBJECT_CLASS (applet_player_parent_class)->dispose (obj);
}

static void
applet_player_finalize (GObject *obj)
{
    G_OBJECT_CLASS (applet_player_parent_class)->finalize (obj);
}

static void
applet_player_class_init (AppletPlayerClass *klass)
{
    GObjectClass *gobject_class = G_OBJECT_CLASS(klass);

    gobject_class->dispose = applet_player_dispose;
    gobject_class->finalize = applet_player_finalize;

    ClutterActorClass *actor_class = CLUTTER_ACTOR_CLASS (klass);
    actor_class->allocate = applet_player_allocate;
    actor_class->get_preferred_width = applet_player_get_preferred_width;
    actor_class->get_preferred_height = applet_player_get_preferred_height;
    actor_class->button_release_event = applet_player_on_button_release;

    StWidgetClass *widget_class = ST_WIDGET_CLASS (klass);
    widget_class->style_changed = applet_player_style_changed;
}

static void
applet_player_init (AppletPlayer *self)
{
    ClutterActor *actor = CLUTTER_ACTOR (self);

    clutter_actor_set_reactive (actor, TRUE);
    st_widget_set_track_hover ((StWidget *)self, TRUE);
    clutter_actor_set_x_expand (actor, TRUE);

    st_widget_set_style_class_name ((StWidget *)self, "music-player");
    clutter_actor_set_name (actor, "music-player");

    self->cover = (ClutterActor *)st_bin_new ();
    st_widget_add_style_class_name ((StWidget *)self->cover, "music-player-cover");
    clutter_actor_add_child (actor, self->cover);

    self->texture_cache = st_texture_cache_get_default ();
    ClutterActor *cover = st_icon_new ();
    st_icon_set_icon_name ((StIcon *)cover, "album-cover");
    st_icon_set_icon_size ((StIcon *)cover, 120);
    st_bin_set_child ((StBin *)self->cover, cover);

    self->title_label = (StLabel *)st_label_new ("Unknown");
    st_widget_add_style_class_name ((StWidget *)self->title_label, "music-player-title");
    clutter_actor_add_child (actor, (ClutterActor *)self->title_label);

    self->artists_label = (StLabel *)st_label_new ("Unknown");;
    st_widget_add_style_class_name ((StWidget *)self->artists_label, "music-player-artists");
    clutter_actor_add_child (actor, (ClutterActor *)self->artists_label);

    self->album_label = (StLabel *)st_label_new ("Unknow");;
    st_widget_add_style_class_name ((StWidget *)self->album_label, "music-player-album");
    clutter_actor_add_child (actor, (ClutterActor *)self->album_label);

    self->seek_bar = (ClutterActor *)cdos_scale_new ();
    st_widget_set_style_class_name ((StWidget *)self->seek_bar, "music-player-seekbar");
    clutter_actor_add_child (actor, self->seek_bar);

    self->time_current = st_label_new ("0:00");
    self->time_length = st_label_new ("0:00");
    st_widget_add_style_class_name (self->time_current, "music-player-time-current");
    st_widget_add_style_class_name (self->time_length, "music-player-time-length");
    clutter_actor_add_child (actor, (ClutterActor *)self->time_current);
    clutter_actor_add_child (actor, (ClutterActor *)self->time_length);

    self->buttons = (ClutterActor *)st_box_layout_new ();
    st_widget_add_style_class_name ((StWidget *)self->buttons, "music-player-buttons");
    clutter_actor_add_child (actor, self->buttons);

    ClutterActor *previous = g_object_new (ST_TYPE_ICON,
                                           "icon-size", 16,
                                           "icon-name", "player-prev",
                                           NULL);
    self->prev_icon = (StIcon *)previous;
    ClutterActor *play_pause = g_object_new (ST_TYPE_ICON,
                                             "icon-size", 16,
                                             "icon-name", "player-play",
                                             NULL);
    self->play_pause_icon = (StIcon *)play_pause;
    ClutterActor *next = g_object_new (ST_TYPE_ICON,
                                       "icon-size", 16,
                                       "icon-name", "player-next",
                                       NULL);
    self->next_icon = (StIcon *)next;

    self->previous = (ClutterActor *)st_bin_new ();
    clutter_actor_set_reactive (self->previous, TRUE);
    st_widget_set_track_hover ((StWidget *)self->previous, TRUE);
    st_bin_set_child ((StBin *)self->previous, previous);

    self->play_pause = (ClutterActor *)st_bin_new ();
    clutter_actor_set_reactive (self->play_pause, TRUE);
    st_widget_set_track_hover ((StWidget *)self->play_pause, TRUE);
    st_bin_set_child ((StBin *)self->play_pause, play_pause);

    self->next = (ClutterActor *)st_bin_new ();
    clutter_actor_set_reactive (self->next, TRUE);
    st_widget_set_track_hover ((StWidget *)self->next, TRUE);
    st_bin_set_child ((StBin *)self->next, next);

    clutter_actor_add_child (self->buttons, self->previous);
    clutter_actor_add_child (self->buttons, self->play_pause);
    clutter_actor_add_child (self->buttons, self->next);

    CdosMpris2Player *player = cdos_mpris2_player_new ();
    self->player = player;
    g_signal_connect (player, "connected", G_CALLBACK (applet_player_on_connected), self);
    g_signal_connect (player, "playback-status", G_CALLBACK (applet_player_on_playback_status_changed), self);
    g_signal_connect (player, "playback-tick", G_CALLBACK (applet_player_on_playback_tick), self);
    g_signal_connect (player, "metadata", G_CALLBACK (applet_player_on_metadata), self);
    g_signal_connect (player, "control-flags", G_CALLBACK (applet_player_on_control), self);

    g_signal_connect (self->seek_bar, "value-changed", G_CALLBACK (applet_player_seeked), self);
    g_signal_connect (self->previous, "button-release-event", G_CALLBACK (applet_player_prev), self);
    g_signal_connect (self->play_pause, "button-release-event", G_CALLBACK (applet_player_play_pause), self);
    g_signal_connect (self->next, "button-release-event", G_CALLBACK (applet_player_next), self);

    g_signal_connect (self, "button-release-event", G_CALLBACK (applet_player_raise), self);
}

/**
 * applet_player_new:
 * @source: A #CdosSource 
 * @title: 设置 #AppletPlayer 实例的标题 
 * @banner: 设置 #AppletPlayer 实例的内容
 * @params: 必要参数,参考 #AppletPlayerParams
 * 
 * 新建一个 #AppletPlayer 类的实例。
 *
 * Returns: 返回一个 #AppletPlayer 的实例。
 */
AppletPlayer *
applet_player_new (void)
{
    AppletPlayer *self = g_object_new (APPLET_TYPE_PLAYER, NULL);
    return self;
}

const gchar *
applet_player_get_name (AppletPlayer *player)
{
    return cdos_mpris2_player_get_player (player->player);
}


/*
 * AppletMpris2Listen
 */
typedef struct _AppletMpris2Listener
{
    GDBusConnection *connection;
    guint signal_handler_id;
    APlayerReadyCallback callback;
    gpointer user_data;
    GList *players;
}AppletMpris2Listener;

static AppletMpris2Listener *mpris2_listener = NULL;

static void
on_player_destroy (AppletPlayer *player, AppletMpris2Listener *self)
{
    self->players = g_list_remove (self->players, player);
}

static void
applet_mpris2_list_names (GObject *source_object,
                          GAsyncResult *res,
                          AppletMpris2Listener *self)
{
    GError *error = NULL;
    GVariant *v = g_dbus_connection_call_finish (self->connection, res, &error);
    if (v == NULL)
    {
        g_warning ("%s: %s", __func__, error->message);
        g_error_free (error);
    }

    GVariantIter *iter = NULL;
    const gchar *name = NULL;
    g_variant_get (v, "(as)", &iter);
    while (g_variant_iter_loop (iter, "&s", &name))
    {
        if (g_str_has_prefix (name, "org.mpris.MediaPlayer2."))
        {
            AppletPlayer *player = applet_player_new ();
            cdos_mpris2_player_set_player (player->player, self->connection, name);
            self->players = g_list_prepend (self->players, player);
            g_signal_connect (player, "destroy", G_CALLBACK (on_player_destroy), self);
            if (self->callback)
                self->callback (player, self->user_data);
        }
    }
}

static AppletPlayer *
find_player (AppletMpris2Listener *self, const char *name)
{
    for (GList *iter = self->players; iter; iter = iter->next)
    {
        AppletPlayer *player = iter->data;
        CdosMpris2Player *mpris2 = player->player;
        if (!g_strcmp0 (cdos_mpris2_player_get_player (mpris2), name))
            return player;
    }
    return NULL;
}

static void
applet_mpris2_name_owner_changed (GDBusConnection  *connection,
                                  const gchar      *sender,
                                  const gchar      *object_path,
                                  const gchar      *interface_name,
                                  const gchar      *signal_name,
                                  GVariant         *parameters,
                                  gpointer          user_data)
{
    const gchar *name;
    const gchar *old_owner;
    const gchar *new_owner;
    g_variant_get (parameters,
                   "(&s&s&s)",
                   &name,
                   &old_owner,
                   &new_owner);
    g_debug ("sender: %s, name:%s, old:%s, new:%s", sender, name, old_owner, new_owner);

    if (!g_str_has_prefix (name, "org.mpris.MediaPlayer2."))
        return;

    if (!new_owner)
        return;

    AppletPlayer *player = NULL; //applet_player_new ();
    AppletMpris2Listener *self = user_data;
    if (*new_owner && old_owner && *old_owner != '\0')
    {
        player = find_player (self, name);
        if (player)
            cdos_mpris2_player_set_player (player->player, NULL, name);
    }
    else if (!player && *new_owner != '\0')
    {
            player = applet_player_new ();
            cdos_mpris2_player_set_player (player->player, self->connection, name);
            self->players = g_list_prepend (self->players, player);
            g_signal_connect (player, "destroy", G_CALLBACK (on_player_destroy), self);
            if (self->callback)
                self->callback (player, self->user_data);
    }
}

static void
applet_mpris2_bus_get (GObject *source_object,
                       GAsyncResult *res,
                       AppletMpris2Listener *self)
{
    GError *error = NULL;
    GDBusConnection *connection = g_bus_get_finish (res, &error);
    if (connection == NULL)
    {
        g_warning ("%s", error->message);
        g_error_free (error);
        return;
    }
    self->connection = connection;
    g_dbus_connection_call (connection, DBUS_NAME, DBUS_PATH, INTERFACE_NAME,
                            "ListNames",
                            NULL,
                            G_VARIANT_TYPE ("(as)"),
                            G_DBUS_CALL_FLAGS_NONE,
                            -1,
                            NULL,
                            (GAsyncReadyCallback)applet_mpris2_list_names,
                            self);

    self->signal_handler_id = g_dbus_connection_signal_subscribe (self->connection,
                                                                  "org.freedesktop.DBus",
                                                                  INTERFACE_NAME,
                                                                  "NameOwnerChanged",
                                                                  DBUS_PATH,
                                                                  NULL,
                                                                  G_DBUS_SIGNAL_FLAGS_NONE,
                                                                  applet_mpris2_name_owner_changed,
                                                                  self,
                                                                  NULL);
}

void
applet_mpris2_listen (APlayerReadyCallback callback, gpointer user_data)
{
    if (!mpris2_listener)
    {
        mpris2_listener = g_new0 (AppletMpris2Listener, 1);
        mpris2_listener->callback = callback;
        mpris2_listener->user_data = user_data;
        g_bus_get (G_BUS_TYPE_SESSION, NULL, (GAsyncReadyCallback)applet_mpris2_bus_get, mpris2_listener);
        //g_dbus_connection_signal_unsubscribe (self->connection, self->signal_handler_id);
    }
}

void
applet_mpris2_unlisten (void)
{
    if (mpris2_listener)
    {
        AppletMpris2Listener *self = mpris2_listener;

        g_dbus_connection_signal_unsubscribe (self->connection, self->signal_handler_id);
        GList *iter = self->players;
        while (iter)
        {
            AppletPlayer *player = iter->data;
            iter = iter->next;
            clutter_actor_destroy ((ClutterActor *)player);
        }
        g_object_unref (self->connection);
        g_free (mpris2_listener);

        mpris2_listener = NULL;
    }
}
