/** 
 * Copyright (c) 2016-2018 CPU and Fundamental Software Research Center, CAS
 *  
 * This software is published by the license of CPU-OS License，you can use and 
 * distibuted this software under this License. See CPU-OS Liense for more detail.
 *
 * You should have receive a copy of CPU-OS License. If not, please contact us 
 * by email <support_os@cpu-os.ac.cn>  
 * 
 **/
#include <libupower-glib/upower.h>
#include "common.h"
#include "applet-settings-button.h"
#include "applet-notifications2.h"

#define DEFAULT_ICON_SIZE 24

enum
{
  PROP_ICON_NAME = 1,
  PROP_ICON_SIZE,
  PROP_TEXT,
  PROP_OPEN,
  NUM_PROPERTIES
};

enum
{
    ACTIVATE,
    LAST_SIGNAL
};

static GParamSpec *properties[NUM_PROPERTIES] = { NULL, };
static guint applet_settings_button_signals [LAST_SIGNAL] = {0};

struct _AppletSettingsButtonPrivate
{
    StIcon *icon;
    StLabel *label;
    
    /*属性*/
    gchar *icon_name;
    gint icon_size;
    gchar *text;
    gboolean open;    //打开/关闭状态
    
    AppletNotifications2 *notification;
    StBin *actor;
    gboolean keep_notification;   //是否显示通知中心
    gboolean show_open_state;   //是否显示打开状态
};

G_DEFINE_TYPE_WITH_PRIVATE (AppletSettingsButton, applet_settings_button, ST_TYPE_WIDGET);

void
applet_settings_button_set_icon_name (AppletSettingsButton *settings_button, 
                                      const gchar *icon_name)
{
    AppletSettingsButtonPrivate *priv;

    g_return_if_fail (APPLET_IS_SETTINGS_BUTTON (settings_button));

    priv = settings_button->priv;

    st_icon_set_icon_name (priv->icon, icon_name);
    g_object_notify_by_pspec (G_OBJECT (settings_button), properties[PROP_ICON_NAME]);
}

void
applet_settings_button_set_icon_size (AppletSettingsButton *settings_button,
                                      gint size)
{
    AppletSettingsButtonPrivate *priv;

    g_return_if_fail (APPLET_IS_SETTINGS_BUTTON (settings_button));

    priv = settings_button->priv;

    if (size < 0)
    {
        size = DEFAULT_ICON_SIZE;
    }
    st_icon_set_icon_size (priv->icon, size);

    g_object_notify_by_pspec (G_OBJECT (settings_button), properties[PROP_ICON_SIZE]);
}

void
applet_settings_button_set_text (AppletSettingsButton *settings_button, 
                                 const gchar *text)
{
    AppletSettingsButtonPrivate *priv;

    g_return_if_fail (APPLET_IS_SETTINGS_BUTTON (settings_button));
    g_return_if_fail (text != NULL);   //text不为NULL

    priv = settings_button->priv;

    st_label_set_text (priv->label, text);

    g_object_notify_by_pspec (G_OBJECT (settings_button), properties[PROP_TEXT]);
}

void
applet_settings_button_set_notification (AppletSettingsButton *settings_button,
                                         AppletNotifications2 *notification)
{
    AppletSettingsButtonPrivate *priv;

    g_return_if_fail (APPLET_IS_SETTINGS_BUTTON (settings_button));
    g_return_if_fail (APPLET_IS_NOTIFICATIONS2 (notification));

    priv = settings_button->priv;
    priv->notification = notification;
}

AppletNotifications2 *
applet_settings_button_get_notification (AppletSettingsButton *settings_button)
{
    g_return_val_if_fail (APPLET_IS_SETTINGS_BUTTON (settings_button), NULL);

    return settings_button->priv->notification;
}

void
applet_settings_button_keep_notification (AppletSettingsButton *settings_button,
                                          gboolean keep_notification)
{
    AppletSettingsButtonPrivate *priv;

    g_return_if_fail (APPLET_IS_SETTINGS_BUTTON (settings_button));

    priv = settings_button->priv;
    priv->keep_notification = keep_notification;
}

void
applet_settings_button_set_open (AppletSettingsButton *settings_button,
                                 gboolean open)
{
    AppletSettingsButtonPrivate *priv;

    g_return_if_fail (APPLET_IS_SETTINGS_BUTTON (settings_button));

    priv = settings_button->priv;
    if (priv->open == open)
        return;

    priv->open = open;

    g_object_notify_by_pspec (G_OBJECT (settings_button), properties[PROP_OPEN]);
}

gboolean
applet_settings_button_is_open (AppletSettingsButton *settings_button)
{
    g_return_val_if_fail (APPLET_IS_SETTINGS_BUTTON (settings_button), FALSE);

    return settings_button->priv->open;
}

void
applet_settings_button_show_open_state (AppletSettingsButton *settings_button,
                                        gboolean show_open_state)
{
    AppletSettingsButtonPrivate *priv;

    g_return_if_fail (APPLET_IS_SETTINGS_BUTTON (settings_button));

    priv = settings_button->priv;
    priv->show_open_state = show_open_state;
}

void
applet_settings_button_open_state_show (AppletSettingsButton *settings_button)
{
    AppletSettingsButtonPrivate *priv;

    g_return_if_fail (APPLET_IS_SETTINGS_BUTTON (settings_button));

    priv = settings_button->priv;

    if (priv->show_open_state)
    {
        if (priv->open)
        {
            st_widget_add_style_pseudo_class (ST_WIDGET (settings_button), "open");
        }
        else
        {
            st_widget_remove_style_pseudo_class (ST_WIDGET (settings_button), "open");
        }
    }
}

static void
applet_settings_button_activate (AppletSettingsButton *self)
{
    if (APPLET_SETTINGS_BUTTON_GET_CLASS (self))
        APPLET_SETTINGS_BUTTON_GET_CLASS (self)->activate (self);
}

static void
applet_settings_button_finalize (GObject *gobject)
{
    AppletSettingsButton *self = APPLET_SETTINGS_BUTTON (gobject);
    AppletSettingsButtonPrivate *priv = self->priv;
    g_free (priv->icon_name);
    g_free (priv->text);
    G_OBJECT_CLASS (applet_settings_button_parent_class)->finalize (gobject);
}

static void
applet_settings_button_set_property (GObject      *gobject,
                                     guint         prop_id,
                                     const GValue *value,
                                     GParamSpec   *pspec)
{
    AppletSettingsButton *settings_button = APPLET_SETTINGS_BUTTON (gobject);
    switch (prop_id)
    {
        case PROP_ICON_NAME:
            applet_settings_button_set_icon_name (settings_button, g_value_get_string (value));
            break;
        case PROP_ICON_SIZE:
            applet_settings_button_set_icon_size (settings_button, g_value_get_int (value));
            break;
        case PROP_TEXT:
            applet_settings_button_set_text (settings_button, g_value_get_string (value));
            break;
        case PROP_OPEN:
            applet_settings_button_set_open (settings_button, g_value_get_boolean (value));
            break;
        default:
            G_OBJECT_WARN_INVALID_PROPERTY_ID (gobject, prop_id, pspec);
            break;
    }
}

static void
applet_settings_button_get_property (GObject    *gobject,
                                     guint       prop_id,
                                     GValue     *value,
                                     GParamSpec *pspec)
{
    AppletSettingsButtonPrivate *priv;
    priv = APPLET_SETTINGS_BUTTON (gobject)->priv;
    switch (prop_id)
    {
        case PROP_ICON_NAME:
            g_value_set_string (value, priv->icon_name);
            break;
        case PROP_ICON_SIZE:
            g_value_set_int (value, priv->icon_size);
            break;
        case PROP_TEXT:
            g_value_set_string (value, priv->text);
            break;
        case PROP_OPEN:
            g_value_set_boolean (value, priv->open);
            break;
        default:
            G_OBJECT_WARN_INVALID_PROPERTY_ID (gobject, prop_id, pspec);
            break;
    }
}

static gboolean
applet_settings_button_release_event (ClutterActor *actor, ClutterButtonEvent *event)
{
    AppletSettingsButton *settings_button;
    AppletSettingsButtonPrivate *priv;
    
    settings_button = (AppletSettingsButton *)actor;
    priv = settings_button->priv;

    st_widget_remove_style_pseudo_class (ST_WIDGET (actor), "click");

    if (priv->notification)
    {
        if (!priv->keep_notification)
        {
            applet_notifications2_hide (priv->notification);
        }
    }

    if (priv->open)
        applet_settings_button_activate (settings_button);
    return TRUE;
}

static void
applet_settings_button_get_preferred_height (ClutterActor *actor,
                                             gfloat for_width,
                                             gfloat *min_height_p,
                                             gfloat *nat_height_p)
{
    StThemeNode *theme_node = st_widget_get_theme_node ((StWidget *)actor);
    st_theme_node_adjust_for_width (theme_node, &for_width);

    AppletSettingsButton *self = APPLET_SETTINGS_BUTTON (actor);

    gfloat icon_min_height, icon_nat_height;
    gfloat label_min_height, label_nat_height;

    clutter_actor_get_preferred_height ((ClutterActor *)self->priv->icon, for_width, &icon_min_height, &icon_nat_height);
    clutter_actor_get_preferred_height ((ClutterActor *)self->priv->label, for_width, &label_min_height, &label_nat_height);

    *min_height_p = icon_min_height + label_min_height;
    *nat_height_p = icon_nat_height + label_nat_height;

    st_theme_node_adjust_preferred_height (theme_node, min_height_p, nat_height_p);
}

static void
applet_settings_button_get_preferred_width (ClutterActor *actor,
                                            gfloat for_height,
                                            gfloat *min_width_p,
                                            gfloat *nat_width_p)
{
    StThemeNode *theme_node = st_widget_get_theme_node ((StWidget *)actor);
    st_theme_node_adjust_for_height (theme_node, &for_height);

    AppletSettingsButton *self = APPLET_SETTINGS_BUTTON (actor);

    gfloat icon_min_width, icon_nat_width;
    gfloat label_min_width, label_nat_width;

    clutter_actor_get_preferred_width ((ClutterActor *)self->priv->icon, for_height, &icon_min_width, &icon_nat_width);
    clutter_actor_get_preferred_width ((ClutterActor *)self->priv->label, for_height, &label_min_width, &label_nat_width);

    *min_width_p = MAX(icon_min_width, label_min_width);
    *nat_width_p = MAX(icon_nat_width, label_nat_width);

    st_theme_node_adjust_preferred_width (theme_node, min_width_p, nat_width_p);
}

static void
applet_settings_button_allocate (ClutterActor *actor,
                                 const ClutterActorBox *box,
                                 ClutterAllocationFlags flags)
{
    AppletSettingsButton *self = APPLET_SETTINGS_BUTTON (actor);
    StThemeNode *theme_node = st_widget_get_theme_node ((StWidget *)actor);

    clutter_actor_set_allocation (actor, box, flags);

    ClutterActorBox content_box, child_box;
    st_theme_node_get_content_box (theme_node, box, &content_box);

    gfloat icon_min_height, icon_nat_height;
    gfloat label_min_height, label_nat_height;

    clutter_actor_get_preferred_height ((ClutterActor *)self->priv->icon, -1, &icon_min_height, &icon_nat_height);
    clutter_actor_get_preferred_height ((ClutterActor *)self->priv->label, -1, &label_min_height, &label_nat_height);

    gfloat icon_min_width, icon_nat_width;
    gfloat label_min_width, label_nat_width;

    clutter_actor_get_preferred_width ((ClutterActor *)self->priv->icon, -1, &icon_min_width, &icon_nat_width);
    clutter_actor_get_preferred_width ((ClutterActor *)self->priv->label, -1, &label_min_width, &label_nat_width);

    child_box.x1 = (content_box.x2 + content_box.x1 - icon_nat_width) / 2;
    child_box.x2 = child_box.x1 + icon_nat_width;
    child_box.y1 = content_box.y1 + (content_box.y2 - content_box.y1 - icon_nat_height - label_nat_height) / 2;
    child_box.y2 = child_box.y1 + icon_nat_height;
    clutter_actor_allocate ((ClutterActor *)self->priv->icon, &child_box, flags);

    child_box.x1 = (int)(0.5 + (content_box.x2 + content_box.x1 - label_nat_width) / 2);
    child_box.x2 = child_box.x1 + label_nat_width;
    child_box.y2 = (int)(0.5 + (content_box.y2 + content_box.y1 + icon_nat_height + label_nat_height) / 2);
    child_box.y1 = child_box.y2 - label_nat_height;
    clutter_actor_allocate ((ClutterActor *)self->priv->label, &child_box, flags);
}
        
static void
applet_settings_button_init (AppletSettingsButton *self)
{
    AppletSettingsButtonPrivate *priv;
    ClutterActor *clutter_text;

    priv = self->priv = applet_settings_button_get_instance_private (self);

    priv->notification = NULL;
    priv->show_open_state = FALSE;
    //priv->open = FALSE;

    StWidget *actor = ST_WIDGET (self);
    st_widget_add_style_class_name (actor, "notification-settings-button");
    clutter_actor_set_reactive ((ClutterActor *)actor, TRUE);
    st_widget_set_track_hover (actor, TRUE);

    priv->icon = ST_ICON(st_icon_new ());
    /*设置默认的图标大小*/
    st_icon_set_icon_size (priv->icon, priv->icon_size);
    st_widget_add_style_class_name ((StWidget *)priv->icon, "notification-settings-button-icon");

	priv->label = g_object_new(ST_TYPE_LABEL,
	                           "style-class", "notification-settings-button-label",
	                           NULL);
    //test
    /*设置文本过长时，自动换行显示*/
    clutter_text = st_label_get_clutter_text (priv->label);
    clutter_text_set_ellipsize (CLUTTER_TEXT (clutter_text), PANGO_ELLIPSIZE_NONE);
    clutter_text_set_line_wrap (CLUTTER_TEXT (clutter_text), TRUE);

    _add_actor (self, priv->icon);
    _add_actor (self, priv->label);

}

static void
applet_settings_button_class_init (AppletSettingsButtonClass *klass)
{
    GObjectClass *gobject_class = G_OBJECT_CLASS (klass);

    gobject_class->finalize = applet_settings_button_finalize;
    gobject_class->set_property = applet_settings_button_set_property;
    gobject_class->get_property = applet_settings_button_get_property;

    ClutterActorClass *actor_class = CLUTTER_ACTOR_CLASS (klass);
    actor_class->get_preferred_width = applet_settings_button_get_preferred_width;
    actor_class->get_preferred_height = applet_settings_button_get_preferred_height;
    actor_class->allocate = applet_settings_button_allocate;
    actor_class->button_release_event = applet_settings_button_release_event;

    properties[PROP_ICON_NAME] = g_param_spec_string ("icon-name", 
                                 "Icon name", 
                                 "An icon name", 
                                 NULL, G_PARAM_READWRITE);
    g_object_class_install_property (gobject_class, PROP_ICON_NAME, properties[PROP_ICON_NAME]);

    properties[PROP_ICON_SIZE] = g_param_spec_int ("icon-size", 
                              "Icon size",
                              "The size if the icon, if positive.",
                              -1, G_MAXINT, -1, 
                              G_PARAM_READWRITE);
    g_object_class_install_property (gobject_class, PROP_ICON_SIZE, properties[PROP_ICON_SIZE]);

    properties[PROP_TEXT] = g_param_spec_string ("text", 
                                 "Text", 
                                 "Text of the label", 
                                 NULL, G_PARAM_READWRITE);
    g_object_class_install_property (gobject_class, PROP_TEXT, properties[PROP_TEXT]);

    properties[PROP_OPEN] = g_param_spec_boolean ("open",
                                  "Open",
                                  "Open state of the button",
                                  FALSE, G_PARAM_READWRITE);
    g_object_class_install_property (gobject_class, PROP_OPEN, properties[PROP_OPEN]);

    /*signals*/
    applet_settings_button_signals[ACTIVATE] = g_signal_new ("activate",
                                                                    G_TYPE_FROM_CLASS (klass),
                                                                    G_SIGNAL_RUN_LAST,
                                                                    G_STRUCT_OFFSET (AppletSettingsButtonClass, activate),
                                                                    NULL,
                                                                    NULL,
                                                                    g_cclosure_marshal_VOID__POINTER,
                                                                    G_TYPE_NONE,
                                                                    1,
                                                                    G_TYPE_POINTER);
}

AppletSettingsButton *
applet_settings_button_new (const gchar *icon_name, 
                                   gint icon_size, 
                                   const gchar *text)
{
    AppletSettingsButton *settings_button;

    settings_button = g_object_new (APPLET_TYPE_SETTINGS_BUTTON, 
                                    "icon-name", icon_name, 
                                    "icon-size", icon_size, 
                                    "text", text, 
                                    NULL);

    return settings_button;
}
