#include <gio/gio.h>
#include <math.h>
#include <glib.h>
#include <glib/gprintf.h>

#include "common.h"
#include "applet-power.h"
#include "cdos-systray.h"
#include "cdos-debug.h"
#include "cdos-popup-separator-menu-item.h"
#include "cdos-util.h"

#include "applet-power-private.h"

/**
 * SECTION: cdos-power
 * @Title: AppletPower
 * @Short_description: 电源管理器
 *
 * AppletPower 通过 settings daemon 的 power 模块来监听电源的状态变化
 * 比如拔出电源、电量变化等信号会触发 GUI 部分，从而显示变化。
 */

#define CDOS_NAME "org.freedesktop.UPower"
#define CDOS_PATH "/org/freedesktop/UPower/devices/DisplayDevice"
#define CDOS_PATH_UPOWER "/org/freedesktop/UPower"
#define CDOS_INTERFACE "org.freedesktop.UPower.Device"
#define POWER_SETTINGS_SCHEMAS "org.cdos.settings-daemon.plugins.power"

#define BATTERY_LOW_EFFECT_TIMES 3

/*
 * settings daemon 中接口编译老是不起作用，干脆在 power 中不使用 settings daemon 的 dbus
 * 直接使用 org.freedesktop.UPower 的接口
 * d-feet 可以查询的 object path 有好几个
 * {'DaemonVersion': '0.99.4',
 *  'LidIsClosed': False,
 *  'LidIsPresent': True,
 *  'OnBattery': True} 这几个是 org.freedesktop.UPower 的属性。
 *  其中 OnBattery 可以用于判断当前是否使用电池
 *
 * {'Capacity': 0.0,
 *  'Energy': 34.78,
 *  'EnergyEmpty': 0.0,
 *  'EnergyFull': 39.46,
 *  'EnergyFullDesign': 0.0,
 *  'EnergyRate': 14.699,
 *  'HasHistory': False,
 *  'HasStatistics': False,
 *  'IconName': 'battery-full',  图标名称
 *  'IsPresent': True,
 *  'IsRechargeable': False,
 *  'Luminosity': 0.0,
 *  'Model': '',
 *  'NativePath': '',
 *  'Online': False,
 *  'Percentage': 88.0,		剩余电量的百分比
 *  'PowerSupply': True,
 *  'Serial': '',
 *  'State': 2,			电池状态：0 未知 1 充电、2 放电、3 空liao、4 充满、 5 暂停充电、 6 暂停放电
 *  'Technology': 0,
 *  'Temperature': 0.0,
 *  'TimeToEmpty': 8518,	电池还可以坚持多久
 *  'TimeToFull': 0,		还需要多久电池充满
 *  'Type': 2,
 *  'UpdateTime': 1494825678,
 *  'Vendor': '',
 *  'Voltage': 0.0,
 *  'WarningLevel': 1}  这个是 org.freedesktop.UPower.Device 的属性
 *
 * */

/*
 *  UpDeviceState:
 *  UPower 中的结构，对比来看
 **/
typedef enum {
    UP_DEVICE_STATE_UNKNOWN,
    UP_DEVICE_STATE_CHARGING,
    UP_DEVICE_STATE_DISCHARGING,
    UP_DEVICE_STATE_EMPTY,
    UP_DEVICE_STATE_FULLY_CHARGED,
    UP_DEVICE_STATE_PENDING_CHARGE,
    UP_DEVICE_STATE_PENDING_DISCHARGE,
    UP_DEVICE_STATE_LAST
} UpDeviceState;

struct _AppletPowerPrivate
{
    GDBusConnection *connection;
    GDBusProxy *proxy;
    GDBusProxy *upower_proxy;
    guint refresh_id;
    guint refresh_times;
    //    GDBusProxy *daemon_proxy; 去掉这个成员，监听的是 settings daemon 的状态
    CdosPopupMenu *menu;
    CdosPopupMenuItem *action;
    // CdosPopupSeparatorMenuItem *separator;
    // MetaData *metadata;
    StWidget *if_charging;
    StIcon *battery_icon;
    StWidget *battery_percent;

    // GList *devices;
    GSettings *settings;
    gchar *primary_device_id;
    const gchar *icon_name;
    gboolean battery_low;
    gboolean battery_low_effect_flag;
    guint flag_num;
    guint battery_low_timeout_id;
    gint battery_low_threshold_percentage;
    gdouble primary_device_percentage;
    gint notify_low_num[3];
};

G_DEFINE_TYPE_WITH_PRIVATE (AppletPower, applet_power, CDOS_TYPE_ICONAPPLET);

static void applet_power_ui_frame (AppletPower *self);
static void
power_devices_changed (GDBusProxy *proxy, GVariant *changed_properties, GStrv invalidated_properties, gpointer data);
static gboolean on_action_clicked (ClutterActor *actor, ClutterEvent *event, gboolean keep, gpointer data);

static void applet_power_show_low_battery_tooltip (AppletPower *self, gchar *notify);
static void applet_power_low_battery (AppletPower *self);
static gboolean applet_power_low_battery_effect (AppletPower *self);
static void applet_power_settings_on_changed (GSettings *settings, gchar *key, AppletPower *self);

static void 
applet_power_ui_frame (AppletPower *self)
{
    AppletPowerPrivate *priv;
    priv = self->priv;
    priv->menu = cdos_applet_get_popup_menu (CDOS_APPLET (self));
    StBoxLayout *menu_box = cdos_popup_menu_base_get_box (CDOS_POPUP_MENU_BASE (priv->menu));
    st_widget_set_style_class_name (ST_WIDGET (menu_box), "battery-menu-box");
    st_box_layout_set_vertical (ST_BOX_LAYOUT (menu_box), TRUE);

    StWidget *title_box = st_box_layout_new ();
    st_widget_set_style_class_name (ST_WIDGET (title_box), "battery-title-box");
    st_box_layout_set_vertical (ST_BOX_LAYOUT (title_box), FALSE);
    _add_actor (menu_box, title_box);

    StWidget *percentage_box = st_box_layout_new ();
    st_widget_set_style_class_name (ST_WIDGET (percentage_box), "battery-percentage-box");
    st_box_layout_set_vertical (ST_BOX_LAYOUT (percentage_box), FALSE);
    _add_actor (menu_box, percentage_box);

    StWidget *title_label = st_label_new (_("Battery "));
    st_widget_set_style_class_name (ST_WIDGET (title_label), "battery-title-label");
    clutter_actor_set_x_expand (CLUTTER_ACTOR (title_label), TRUE);
    clutter_actor_set_x_align (CLUTTER_ACTOR (title_label), ST_ALIGN_START);
    _add_actor (title_box, title_label);

    priv->if_charging = st_label_new (_("Charging "));
    st_widget_set_style_class_name (ST_WIDGET (priv->if_charging), "battery-title-charging");
    clutter_actor_set_x_expand (CLUTTER_ACTOR (priv->if_charging), FALSE);
    clutter_actor_set_x_align (CLUTTER_ACTOR (priv->if_charging), ST_ALIGN_END);
    clutter_actor_hide (CLUTTER_ACTOR (priv->if_charging));
    _add_actor (title_box, priv->if_charging);

    priv->battery_icon = ST_ICON (st_icon_new ());
    st_icon_set_icon_size (ST_ICON (priv->battery_icon), 24);
    st_widget_set_style_class_name (ST_WIDGET (priv->battery_icon), "battery-icon");
    clutter_actor_set_x_expand (CLUTTER_ACTOR (priv->battery_icon), FALSE);
    clutter_actor_set_x_align (CLUTTER_ACTOR (priv->battery_icon), ST_ALIGN_START);
    _add_actor (percentage_box, priv->battery_icon);

    priv->battery_percent = st_label_new ("");
    st_widget_set_style_class_name (ST_WIDGET (priv->battery_percent), "battery-percentage");
    clutter_actor_set_x_expand (CLUTTER_ACTOR (priv->battery_percent), TRUE);
    clutter_actor_set_x_align (CLUTTER_ACTOR (priv->battery_percent), ST_ALIGN_START);
    _add_actor (percentage_box, priv->battery_percent);
}

static void
proxy_refresh_cb (GObject *source_object,
                  GAsyncResult *res,
                  gpointer user_data)
{
    GVariant *v;
    v = g_dbus_proxy_call_finish ((GDBusProxy *)source_object, res, NULL);
    g_debug ("power refresh done = %x", v);
    if (v)
    {
        g_variant_unref (v);
    }
}

static void
refresh_device (AppletPower *self, const gchar *path)
{
    AppletPowerPrivate *priv = self->priv;
    GDBusProxy *device_proxy;
    if (path == NULL)
        return ;
    device_proxy = g_dbus_proxy_new_sync (priv->connection, 
                                         G_DBUS_PROXY_FLAGS_NONE, 
                                         NULL, 
                                         CDOS_NAME, 
                                         path, 
                                         CDOS_INTERFACE, 
                                         NULL, 
                                         NULL);
    if (device_proxy)
    {
        g_dbus_proxy_call (device_proxy,
                        "Refresh",
                        NULL,
                        G_DBUS_CALL_FLAGS_NONE,
                        -1,
                        NULL,
                        (GAsyncReadyCallback)proxy_refresh_cb,
                        self);
        g_object_unref (device_proxy);
    }
}

static gboolean 
refresh_all_devices (AppletPower *self)
{
    /* Lenovo N7 laptop, the "EnergyRate" property is wrong (very low) 
    * after pull out of the power line, 
    * so I should refresh it with another interface, 
    * notice that the method "Refresh" of "org.freedesktop.UPower.Device" is invalid.
    */
    AppletPowerPrivate *priv = self->priv;
    GError *error = NULL;
    GVariant *v;
    if (priv->connection == NULL)
        return FALSE;
    if (priv->upower_proxy == NULL)
    {
        priv->upower_proxy = g_dbus_proxy_new_sync (priv->connection, 
                                                    G_DBUS_PROXY_FLAGS_NONE, 
                                                    NULL, 
                                                    CDOS_NAME, 
                                                    CDOS_PATH_UPOWER, 
                                                    CDOS_NAME, 
                                                    NULL, 
                                                    &error);
        if (error)
        {
            g_warning ("Find battery faild:%s", error->message);
            g_error_free (error);
            return FALSE;
        }
    }

    v = g_dbus_proxy_call_sync (priv->upower_proxy,
                                "EnumerateDevices",
                                NULL,
                                G_DBUS_CALL_FLAGS_NONE,
                                -1,
                                NULL,
                                NULL);
    if (v)
    {
        const gchar **paths = NULL; 
        guint len = 0;
        gint i;
        g_variant_get (v, "(^ao)", &paths);
        if (paths)
        {
            len = g_strv_length (paths);
            for (i = 0; i < len; i++)
            {
                g_debug ("upower device path[%d] = %s", i, paths[i]);
                refresh_device (self, paths[i]);
            }
            g_strfreev (paths);
            //拔出电源线之后的一小段时间内，“EnergyRate”很不稳定，因此，需要重复刷新几次。
            priv->refresh_times++;
        }
    }
    if (priv->refresh_times < 5)
        return TRUE;
    priv->refresh_id = 0;
    return FALSE;
}

static void
applet_power_dispose (GObject *object)
{
    AppletPower *self = APPLET_POWER (object);
    AppletPowerPrivate *priv = self->priv;
    if (priv->refresh_id > 0)
    {
        g_source_remove (priv->refresh_id);
        priv->refresh_id = 0;
    }
    g_clear_object (&priv->proxy);
    g_clear_object (&priv->upower_proxy);
    g_clear_object (&priv->settings);
    g_clear_object (&priv->connection);
    G_OBJECT_CLASS (applet_power_parent_class)->dispose (object);
}

static void
applet_power_finalize (GObject *object)
{
    AppletPower *self = APPLET_POWER (object);
    AppletPowerPrivate *priv = self->priv;

    if (priv->primary_device_id)
    {
        g_free (priv->primary_device_id);
    }
    G_OBJECT_CLASS (applet_power_parent_class)->finalize (object);
}

/*
static gboolean
applet_power_on_applet_clicked (gpointer data)
{
    AppletPower *self = APPLET_POWER (data);
    AppletPowerPrivate *priv = self->priv;

    cdos_popup_menu_base_toggle (CDOS_POPUP_MENU_BASE (priv->menu));
    return FALSE;
}
*/

static void
applet_power_class_init (AppletPowerClass *klass)
{
    GObjectClass *object_class = G_OBJECT_CLASS (klass);

    object_class->dispose = applet_power_dispose;
    object_class->finalize = applet_power_finalize;
    // parent_class->on_applet_clicked = applet_power_on_applet_clicked;
}

static void
applet_power_init (AppletPower *self)
{
    AppletPowerPrivate *priv;
    CdosPopupMenu *context_menu;
    priv = self->priv = applet_power_get_instance_private (self);

    priv->primary_device_id = NULL;
    priv->battery_low_timeout_id = 0;
    for (int i = 0; i< 3; i++)
    {

        priv->notify_low_num[i] = 0;
    }

    cdos_applet_set_popup_menu_active (CDOS_APPLET (self), TRUE);
    cdos_applet_set_context_menu_active (CDOS_APPLET (self), TRUE);
    context_menu = cdos_applet_get_context_menu (CDOS_APPLET (self));

    applet_power_ui_frame (self);

    priv->settings = g_settings_new (POWER_SETTINGS_SCHEMAS);
    g_signal_connect (priv->settings, "changed::percentage-low", G_CALLBACK (applet_power_settings_on_changed), self);
    gint percentage_low = g_settings_get_int (priv->settings, "percentage-low");
    priv->battery_low_threshold_percentage = percentage_low;

    GError *error = NULL;
    priv->connection = g_bus_get_sync (G_BUS_TYPE_SYSTEM, NULL, &error);
    if (error)
    {
        g_warning ("%s (%d):%s;", __func__, __LINE__, error->message);
        g_clear_error (&error);
    }
    priv->proxy = g_dbus_proxy_new_sync (priv->connection, 
                                         G_DBUS_PROXY_FLAGS_NONE, 
                                         NULL, 
                                         CDOS_NAME, 
                                         CDOS_PATH, 
                                         CDOS_INTERFACE, 
                                         NULL, 
                                         &error);
    if (error)
    {
        g_warning ("%s (%d):%s;", __func__, __LINE__, error->message);
        g_error_free (error);
    }
    else
    {
        g_signal_connect (priv->proxy, "g-properties-changed", G_CALLBACK (power_devices_changed), self);
    }

    /* priv->separator = cdos_popup_separator_menu_item_new ();
    cdos_popup_menu_base_add_menu_item (
        CDOS_POPUP_MENU_BASE (priv->menu), CDOS_POPUP_BASE_MENU_ITEM (priv->separator), -1); */

    priv->action = cdos_popup_menu_item_new ();
    cdos_popup_menu_item_set_label_align (priv->action, ST_ALIGN_START, 0);
    cdos_popup_menu_item_set_label (priv->action, _("Power Manager"));
    StLabel *label = cdos_popup_menu_item_get_label (priv->action);
    st_widget_add_style_class_name (ST_WIDGET (label), "popup-menu-item-label-link");

    g_signal_connect (priv->action, "activate", G_CALLBACK (on_action_clicked), self);
    cdos_popup_menu_base_add_menu_item (
        CDOS_POPUP_MENU_BASE (context_menu), CDOS_POPUP_BASE_MENU_ITEM (priv->action), -1);

    cdos_popup_menu_close (CDOS_POPUP_MENU_BASE (priv->menu), FALSE);
    clutter_actor_hide (CLUTTER_ACTOR (self));
    priv->battery_low = FALSE;
    if (priv->proxy)
    {
        //防止重启后，无法完成初始化；
        //先获取所有信息
        gchar **property_names = g_dbus_proxy_get_cached_property_names (priv->proxy);
        if (property_names)
        {
            power_devices_changed (priv->proxy, NULL, NULL, self);
            g_strfreev (property_names);
        }
    }
    refresh_all_devices (self);
}

static void
applet_power_settings_on_changed (GSettings *settings, gchar *key, AppletPower *self)
{
    AppletPowerPrivate *priv = self->priv;
    gint percentage_low = g_settings_get_int (settings, "percentage-low");
    priv->battery_low_threshold_percentage = percentage_low;

    if (priv->battery_low == FALSE && priv->primary_device_percentage <= priv->battery_low_threshold_percentage)
    {
        priv->battery_low = TRUE;
        applet_power_low_battery (self);
    }
    else if (priv->battery_low == TRUE && priv->primary_device_percentage > priv->battery_low_threshold_percentage)
    {
        priv->battery_low = FALSE;
    }
}

/*static void
applet_power_check_device_items(AppletPower *self){
    AppletPowerPrivate *priv = self->priv;
    GList *device_items = priv->device_items;
    GList *new_items = NULL;
    while (device_items != NULL){
        AppletPowerDeviceItem *device_item = device_items->data;
        gchar *item_device_id = applet_power_device_item_get_device_id(device_item);
        gboolean flag = FALSE;

        GList *devices = priv->devices;
        while (devices != NULL){
            gchar *device_id = devices->data;
            if(device_id && g_strcmp0(device_id,item_device_id) == FALSE){
                flag = TRUE;
                break ;
            }
            devices = devices->next;
        }
        if(flag == TRUE){
            new_items = g_list_append(new_items,device_item);
        }else{
            StWidget *box = applet_power_device_item_get_box(device_item);
            StBoxLayout *menu_box = cdos_popup_menu_base_get_box(CDOS_POPUP_MENU_BASE(priv->menu));
            clutter_container_remove_actor(CLUTTER_CONTAINER(menu_box),CLUTTER_ACTOR(box));
            g_object_unref(device_item);
        }
        device_items = device_items->next;
    }
    g_list_free(priv->device_items);
    priv->device_items = new_items;
}

static AppletPowerDeviceItem *
applet_power_get_device_item(AppletPower *self,gchar *device_id){
    AppletPowerPrivate *priv = self->priv;
    GList *device_items = priv->device_items;
    AppletPowerDeviceItem * device_item = NULL;
    while (device_items != NULL){
        device_item = device_items->data;
        gchar *item_device_id = applet_power_device_item_get_device_id(device_item);
        if(device_id && g_strcmp0(device_id,item_device_id) == FALSE){
            return device_item;
        }
        device_items = device_items->next;
    }
    return device_item;
}

static void
applet_power_update_device_items (AppletPower *self){
    AppletPowerPrivate *priv = self->priv;
    applet_power_check_device_items(self);
    GList *devices = priv->devices;
    while (devices != NULL){
        AppletPowerDevice *device = devices->data;
        if(device == NULL){
            continue;
        }
        if(device->device_type == BATTERY){
            AppletPowerDeviceItem *device_item = applet_power_get_device_item(self,device->device_id);
            if(device_item){
                applet_power_device_item_refresh(device_item);
                applet_power_device_item_set_icon_name(device_item,device->icon_name);
                applet_power_device_item_set_primary_device(device_item,device->is_primary);

            } else {
                device_item = applet_power_device_item_new(device);
                StWidget *box = applet_power_device_item_get_box(device_item);
                StBoxLayout *menu_box = cdos_popup_menu_base_get_box(CDOS_POPUP_MENU_BASE(priv->menu));
                st_box_layout_insert_actor(menu_box,CLUTTER_ACTOR(box),0);
                priv->device_items = g_list_append (priv->device_items, device_item);
            }
        }
        devices = devices->next;
    }
} */

static gboolean
applet_power_low_battery_effect (AppletPower *self)
{
    AppletPowerPrivate *priv = self->priv;
    ClutterActor *actor = cdos_iconapplet_get_applet_icon (CDOS_ICONAPPLET (self));
    if (priv->battery_low && priv->flag_num <= BATTERY_LOW_EFFECT_TIMES)
    {
        if (priv->battery_low_effect_flag == TRUE)
        {
            clutter_actor_hide (actor);
            priv->battery_low_effect_flag = FALSE;
        }
        else
        {
            priv->flag_num++;
            clutter_actor_show (actor);
            priv->battery_low_effect_flag = TRUE;
        }
        return TRUE;
    }
    else
    {
        clutter_actor_show (actor);
        priv->battery_low_effect_flag = FALSE;
    }
    priv->battery_low_timeout_id = 0;
    return FALSE;
}

static void
applet_power_low_battery (AppletPower *self)
{
    AppletPowerPrivate *priv = self->priv;
    priv->flag_num = 0;
    if (priv->battery_low_timeout_id == 0)
    {
        priv->battery_low_timeout_id = g_timeout_add (500, (GSourceFunc)applet_power_low_battery_effect, self);
    }
}

static void 
applet_power_show_low_battery_tooltip (AppletPower *self, gchar *notify)
{
    AppletPowerPrivate *priv = self->priv;
    gchar *notify_low = g_strconcat ("notify-send -i ", priv->icon_name, " ", _("Low Power Notification"), " ", notify, NULL);
    cdos_util_spawn_command_line (notify_low);
    g_free (notify_low);
}

/*static void
power_read_primary_device (AppletPower *self){
    AppletPowerPrivate *priv = self->priv;
    GVariant *retval = NULL;
    GError *error = NULL;
    //gchar **devices = NULL;
    retval = g_dbus_proxy_call_sync (priv->proxy, "GetPrimaryDevice", NULL, G_DBUS_CALL_FLAGS_NONE, -1,NULL, &error);
    if(priv->primary_device_id){
        g_free(priv->primary_device_id);
    }
    if (error){
        g_warning("\n%s (%d):%s;\n", __func__, __LINE__, error->message);
        priv->primary_device_id = NULL;
        priv->primary_device_percentage = 0;
        g_error_free(error);
        if(priv->battery_low){
            priv->battery_low = FALSE;
        }
        return;
    }
    gchar *device_id;
    guint device_type;
    gchar *icon;
    gdouble percent;
    guint state;
    guint64 time;
    // 对每条记录获取其中的字符串、整形等等
    g_variant_get (retval, "((susdut))", &device_id, &device_type, &icon,
                                    &percent, &state, &time);

    priv->primary_device_id = g_strdup(device_id);
    priv->primary_device_percentage = percent;
    if(priv->icon_name == NULL){
        priv->icon_name = icon;
    }
    if(device_type == BATTERY && state == DISCHARGING && priv->primary_device_percentage <= priv->battery_low_threshold_percentage){
        priv->battery_low = TRUE;
        applet_power_low_battery(self);
    }else{
        priv->battery_low = FALSE;
    }
    g_free(device_id);
    g_variant_unref(retval);
}

static void
power_read_other_devices (AppletPower *self){
    AppletPowerPrivate *priv = self->priv;
    GVariant *retval = NULL;
    GError *error = NULL;
    GVariantIter *iter = NULL;

    retval = g_dbus_proxy_call_sync (priv->proxy, "GetDevices", NULL, G_DBUS_CALL_FLAGS_NONE, -1,NULL, &error);
    if (error){
        g_warning("\n%s (%d):%s;\n", __func__, __LINE__, error->message);
        g_error_free(error);
        return;
    }

    GList *devices = priv->devices;
    while (devices != NULL){
        AppletPowerDevice *device = devices->data;
        if(device){
            g_free(device->icon_name);
            g_free(device->device_id);
            g_free(device);
        }
        devices= devices->next;
    }

    g_list_free(priv->devices);
    priv->devices = NULL;

    // 首先 GVariant 是个结构， 以 () 包围
    // 每个结构体都是一个数组类型的GVariant， 需要迭代器来遍历
    g_variant_get (retval, "(a(susdut))", &iter);
    gchar *device_id;
    guint device_type;
    gchar *icon;
    gdouble percent;
    guint state;
    guint64 time;
    // 对每条记录获取其中的字符串、整形等等
    // iter_loop 获取的 device_id, icon等在迭代结束后，会自动释放，不用管

    while ( g_variant_iter_loop (iter, "(susdut)", &device_id, &device_type, &icon, &percent, &state, &time))
    {
        AppletPowerDevice *device = g_malloc0 (sizeof (AppletPowerDevice));
        device->device_id = g_strdup(device_id);
        device->device_type = device_type;
        device->icon_name = g_strdup(icon);
        device->percent = percent;
        device->state = state;
        device->time = time;
        device->is_primary = FALSE;
        if(priv->primary_device_id && g_strcmp0(priv->primary_device_id,device_id) == FALSE){
            device->is_primary = TRUE;
        }
        if(priv->icon_name == NULL && device_type == AC_POWER){
            priv->icon_name = icon;
        }
        priv->devices = g_list_append(priv->devices,device);
    }
    g_variant_iter_free (iter);
    g_variant_unref(retval);
} */

static gboolean
on_action_clicked (ClutterActor *actor, ClutterEvent *event, gboolean keep, gpointer data)
{
    guint ret = cdos_util_spawn_command_line ("found-control-center power");

    if (ret < 0)
    {
        cdos_debug ();
        return FALSE;
    }
    return TRUE;
}

static void
power_devices_changed (GDBusProxy *proxy,
                       GVariant   *changed_properties,
                       GStrv       invalidated_properties,
                       gpointer    data)
{
    AppletPower *self = APPLET_POWER (data);
    AppletPowerPrivate *priv = self->priv;

    if (changed_properties)
    {
        GVariantIter iter;
        GVariant *value;
        gchar *key;
        g_variant_iter_init (&iter, changed_properties);
        while (g_variant_iter_loop (&iter, "{sv}", &key, &value))
        {
            g_debug ("Applet power:changed prop: %s", key);
            if (g_strcmp0 (key, "State") == 0)
            {
                if (priv->refresh_id > 0)
                {
                    g_source_remove (priv->refresh_id);
                    priv->refresh_id = 0;
                }
                priv->refresh_times = 0;
                priv->refresh_id = g_timeout_add_seconds (2, refresh_all_devices, self);
                break;
            }
        }
    }

    // 获取远程对象的 Icon 属性
    //GVariant *icon_name = g_dbus_proxy_get_cached_property (proxy, "IconName");
    // 电量百分比
    GVariant *percentage = g_dbus_proxy_get_cached_property (proxy, "Percentage");
    // 电源状态
    GVariant *power_state = g_dbus_proxy_get_cached_property (proxy, "State");
    // 以上都是 GVariant 类型的数据，需要转换类型使用
    //gchar *iconname;
    gdouble percent;
    guint state;
    //g_variant_get (icon_name, "s", &iconname);
    percent = g_variant_get_double (percentage);
    state = g_variant_get_uint32 (power_state);
    g_debug ("Power changed: state = %d", state);
    priv->primary_device_percentage = percent;
    // 以下，根据电源 state 的不同，需要获取不同的数据，state 描述详见文件头注释
    gint64 seconds = 0;
    GVariant *time_to_end;
    gchar *state_label = NULL;
    switch (state)
    {
        case 1:
            // 正在充电，获取 TimeToFull, INT64, 单位秒
            time_to_end = g_dbus_proxy_get_cached_property (proxy, "TimeToFull");
            state_label = _("Charging ");
            clutter_actor_show (CLUTTER_ACTOR (priv->if_charging));
            seconds = g_variant_get_int64 (time_to_end);
            for (int i = 0; i< 3; i++)
            {

                priv->notify_low_num[i] = 0;
            }
            break;
        case 2:
            // 正在耗电，获取 TimeToEmpty, INT64, 单位秒
            time_to_end = g_dbus_proxy_get_cached_property (proxy, "TimeToEmpty");
            state_label = _("Discharging ");
            seconds = g_variant_get_int64 (time_to_end);
            clutter_actor_hide (CLUTTER_ACTOR (priv->if_charging));
            if (priv->primary_device_percentage <= priv->battery_low_threshold_percentage)
            {
                priv->battery_low = TRUE;
                applet_power_low_battery (self);
            }
            else
            {
                priv->battery_low = FALSE;
            }

            break;
        case 3:
            clutter_actor_hide (CLUTTER_ACTOR (priv->if_charging));
            state_label = _("EmptyBat");
            for (int i = 0; i< 3; i++)
            {

                priv->notify_low_num[i] = 0;
            }
            seconds = 0;
            break;
        case 4:
            state_label = _("FullBat");
            for (int i = 0; i< 3; i++)
            {

                priv->notify_low_num[i] = 0;
            }
            seconds = 0;
            break;
        default:
            break;
    }
    guint minutes = round (seconds / 60);
    guint hour = round (minutes / 60);

    gchar *tooltip_text;
    if (hour > 0)
    {
        minutes = minutes % 60;
        tooltip_text = g_strdup_printf (_("%s %.0f%%\n Time left %d hour %d minutes"), state_label, percent, hour, minutes);
    }
    else
    {
        tooltip_text = g_strdup_printf (_("%s %.0f%%\n Time left %d minutes"), state_label, percent, minutes);
    }
    gchar *text = NULL, *percent_text = NULL;
    text = g_strdup_printf (" %.0f%%", percent);
    percent_text = g_strconcat (_("Electricity"), ":", text, NULL);
    st_label_set_text (ST_LABEL (priv->battery_percent), percent_text);
    g_free (percent_text);
    g_free (text);

    /*    if(iconname){
        if (priv->icon_name)
            g_free(priv->icon_name);
        priv->icon_name = g_strdup(iconname);
    }*/
    priv->icon_name = NULL;
    // upower 获取的 icon name 不太对，根据电量百分比主动取图
    if (state == 3)
    {
        priv->icon_name = "found-battery-empty";
    }
    else if (state == 4)
    {
        priv->icon_name = "found-battery-full";
    }

    if (percent >= 0 && percent <= 10.0)
    {
        if (state == 2)
            priv->icon_name = "found-battery-ten";
        else if (state == 1)
            priv->icon_name = "found-battery-ten-charging";
    }
    else if (percent > 10.0 && percent <= 20.0)
    {
        if (state == 2)
            priv->icon_name = "found-battery-twenty";
        else if (state == 1)
            priv->icon_name = "found-battery-twenty-charging";
    }
    else if (percent > 20.0 && percent <= 40.0)
    {
        if (state == 2)
            priv->icon_name = "found-battery-forty";
        else if (state == 1)
            priv->icon_name = "found-battery-forty-charging";
    }
    else if (percent > 40.0 && percent <= 60.0)
    {
        if (state == 2)
            priv->icon_name = "found-battery-sixty";
        else if (state == 1)
            priv->icon_name = "found-battery-sixty-charging";
    }
    else if (percent > 60.0 && percent <= 80.0)
    {
        if (state == 2)
            priv->icon_name = "found-battery-eighty";
        else if (state == 1)
            priv->icon_name = "found-battery-eighty-charging";
    }
    else if (percent > 80.0 && percent <= 90.0)
    {
        if (state == 2)
            priv->icon_name = "found-battery-ninety";
        else if (state == 1)
            priv->icon_name = "found-battery-ninety-charging";
    }
    else if (percent > 90.0 && percent <= 95.0)
    {
        if (state == 2)
            priv->icon_name = "found-battery-ninety-five";
        else if (state == 1)
            priv->icon_name = "found-battery-ninety-five-charging";
    }
    else if (percent > 95.0 && percent <= 100.0)
    {
        if (state == 2)
            priv->icon_name = "found-battery-full";
        else if (state == 1)
            priv->icon_name = "found-battery-full-charging";
    }

    cdos_applet_set_applet_tooltip (CDOS_APPLET (self), tooltip_text);
    if (state == 3 || state == 4)
    {
        cdos_applet_set_applet_tooltip (CDOS_APPLET (self), state_label);
    }

    /* 电池电量控制整合至found-settings-daemon中*/
    /* gchar *notify_body = NULL;
    if (percent <= 20 && percent > 10)
    {
        if (state == 2 && priv->notify_low_num[0] == 0)
        {
            notify_body = _("The battery is low, please connect it to the power supply");
            applet_power_show_low_battery_tooltip (self, notify_body);
            priv->notify_low_num[0]++;
        }
    }
    else if (percent <= 10 && percent > 6)
    {
        if (state == 2 && priv->notify_low_num[1] == 0)
        {
            notify_body = _("The battery is too low, please connect it to the power supply");
            applet_power_show_low_battery_tooltip (self, notify_body);
            priv->notify_low_num[1]++;
        }
    }
    else if (percent <= 6 && percent > 5)
    {
        if (state == 2 && priv->notify_low_num[2] == 0)
        {
            notify_body = _("The battery is too low, the computer will go to sleep");
            applet_power_show_low_battery_tooltip (self, notify_body);
            priv->notify_low_num[2]++;
        }
    }
    else if (percent <= 5.0)
    {
        if (state == 2 && priv->notify_low_num[3] == 0)
        {
            cdos_util_spawn_command_line("dbus-send --system --print-reply --dest=\"org.freedesktop.login1\" /org/freedesktop/login1 org.freedesktop.login1.Manager.Suspend boolean:'true'");
            priv->notify_low_num[3]++;
        }
        else if (state == 2 && priv->notify_low_num[3] == 1)
        {
            notify_body = _("The battery is seriously low, please save critical data");
            applet_power_show_low_battery_tooltip (self, notify_body);
            priv->notify_low_num[3]++;
        }
    } */

    //    power_read_primary_device (self);
    //    power_read_other_devices (self);
    //    applet_power_update_device_items(self);

    cdos_iconapplet_set_icon_name (CDOS_ICONAPPLET (self), "found-battery-missing");
    cdos_iconapplet_set_icon_size (CDOS_ICONAPPLET (self), 16);
    if (priv->icon_name && g_strcmp0 (priv->icon_name, "") != 0)
    {
        GError *error = NULL;
        GIcon *gicon = g_icon_new_for_string (priv->icon_name, &error);
        if (error)
        {
            g_warning ("%s (%d):%s", __func__, __LINE__, error->message);
            cdos_popup_menu_close (CDOS_POPUP_MENU_BASE (priv->menu), FALSE);
            clutter_actor_hide (CLUTTER_ACTOR (self));
            g_error_free (error);
        }
        else
        {
            cdos_iconapplet_set_applet_icon_gion (CDOS_ICONAPPLET (self), gicon);
            st_icon_set_icon_name (ST_ICON (priv->battery_icon), priv->icon_name);
            clutter_actor_show (CLUTTER_ACTOR (self));
        }
    }
    else
    {
        cdos_popup_menu_close (CDOS_POPUP_MENU_BASE (priv->menu), FALSE);
        clutter_actor_hide (CLUTTER_ACTOR (self));
    }
    //    g_free (iconname);
    g_free (tooltip_text);
}


static AppletPower *
applet_power_new (MetaData *meta, StSide orientation)
{
    AppletPower *self = NULL;
    self = g_object_new (APPLET_TYPE_POWER, "orientation", orientation, NULL);

    return self;
}

/**
 * applets_power_main:
 *
 * 新建对象，applet manager 新建 power 模块对象，将该 applet 的图标添加到任务栏右侧
 *
 * Return value: #AppletPower 对象
 */
AppletPower *
applets_power_main (MetaData *meta, guint orientation, guint panel_height, gchar *instance_id)
{
    static AppletPower *power;
    if (power == NULL)
    {
        power = applet_power_new (meta, orientation);
    }
    return power;
}
