/**
 * SECTION: AppletSound
 * @short_description: 声音应用
 * @include: applet-sound.h
 *
 * AppletSoundClass 类用来调节，开启或者禁用声音的输入输出
 */
#include "common.h"
#include <sys/fcntl.h>
#include <glib.h>
#include <glib/gstdio.h>
#include <pulse/volume.h>
#include "applet-sound.h"
#include <math.h>
#include "ui/cdos-util.h"
#include "ui/cdos-systray.h"
#include "ui/cdos-sound-manager.h"

G_DEFINE_TYPE (AppletSound, applet_sound, CDOS_TYPE_ICONAPPLET);

#define ANALOG_STEREO_NOT_MUTED (g_strconcat ("", g_get_home_dir (), "/.cdos/configs/analog_stereo_not_muted", NULL))
#define KEY_SOUNDS_SCHEMA "org.gnome.desktop.sound"
#define KEY_SOUND "allow-volume-above-100-percent"

/* 全局变量 */
// static PopupBaseMenuItemInit *params_for_new_item;

/* 自定义数据结构 */
typedef struct _paramsInChanged paramsInChanged;
typedef struct _paramsInActivate paramsInActivate;

typedef struct _paramsInChanged
{
    AppletSound *self;
    gboolean is_output;

} paramsInChanged;

typedef struct _paramsInActivate
{
    GvcMixerStream *sink;
    AppletSound *self;
} paramsInActivate;

/* 信号 */
enum
{
    PLAYERS_LOADED,
    LAST_SIGNAL
};

static guint applet_sound_signals[LAST_SIGNAL] = {0,};

static void applet_sound_create_output_ui_frame (AppletSound *self, StBoxLayout *menu_box);

static void applet_sound_create_input_ui_frame (AppletSound *self, StBoxLayout *menu_box);

static void applet_sound_show_fixed_elements (AppletSound *self);

static const gchar *applet_sound_volume_output_to_icon (AppletSound *self, gfloat volume);

static const gchar *applet_sound_volume_input_to_icon (AppletSound *self, gfloat volume);

/* CALL_BACK */
static void applet_sound_read_output_cb (GObject *obj, guint state, gpointer user_data);

static void applet_sound_read_input_cb (GObject *obj, guint state, gpointer user_data);

static void gvc_control_on_stream_changed (GvcMixerControl *control, guint id, gpointer user_data);

static void applet_sound_out_button_on_state (GObject *obj, GParamSpec *pspec, gpointer user_data);

static void applet_sound_in_button_on_state (GObject *obj, GParamSpec *pspec, gpointer user_data);

static void applet_sound_toggle_out_mute_cb (GObject *obj, gboolean state, gpointer user_data);

static void applet_sound_toggle_in_mute_cb (GObject *obj, gboolean state, gpointer user_data);

static void gvc_control_on_state_changed (GvcMixerControl *control, guint state, gpointer user_data);

static void applet_sound_slider_changed_cb (GObject *obj, gfloat value, gpointer user_data);

static void applet_sound_muted_changed_cb (GObject *obj, GParamSpec *param_spec, paramsInChanged *params);

static void applet_sound_volume_changed_cb (GObject *obj, GParamSpec *param_spec, paramsInChanged *params);

static gboolean applet_sound_on_click_sound_settings_cb (GObject *obj, ClutterEvent *event, gpointer user_data);

//输出声音(output)的ui框架
static void 
applet_sound_create_output_ui_frame (AppletSound *self, StBoxLayout *menu_box)
{
    self->output_box = st_box_layout_new ();
    st_box_layout_set_vertical (ST_BOX_LAYOUT (self->output_box), TRUE);
    _add_actor (menu_box, self->output_box);

    StWidget *p_output_title = st_label_new (_("Output Volume"));
    st_widget_set_style_class_name (ST_WIDGET (p_output_title), "sound-title");
    _add_actor (self->output_box, p_output_title);
    clutter_actor_set_x_expand (CLUTTER_ACTOR (p_output_title), TRUE);
    clutter_actor_set_x_align (CLUTTER_ACTOR (p_output_title), CLUTTER_ACTOR_ALIGN_FILL);

    StWidget *output_main_box = st_box_layout_new ();
    st_box_layout_set_vertical (ST_BOX_LAYOUT (output_main_box), TRUE);
    _add_actor (self->output_box, output_main_box);

    StWidget *output_item_box = st_box_layout_new ();
    st_box_layout_set_vertical (ST_BOX_LAYOUT (output_item_box), FALSE);
    _add_actor (output_main_box, output_item_box);

    self->mute_out_switch = st_button_new ();
    st_widget_set_style_class_name (ST_WIDGET (self->mute_out_switch), "sound-mute");
    _add_actor (output_item_box, self->mute_out_switch);
    st_button_set_toggle_mode (ST_BUTTON (self->mute_out_switch), TRUE);
    g_object_set (self->mute_out_switch,
                  "x-expand", TRUE,
                  "x-align", CLUTTER_ACTOR_ALIGN_START,
                  "y-align", CLUTTER_ACTOR_ALIGN_CENTER,
                  NULL);

    self->output_icon = ST_ICON (st_icon_new ());
    st_icon_set_icon_size (self->output_icon, 24);
    st_bin_set_child (ST_BIN (self->mute_out_switch), CLUTTER_ACTOR (self->output_icon));

    self->output_item_class = st_box_layout_new ();
    st_widget_set_style_class_name (ST_WIDGET (self->output_item_class), "sound-slider");
    _add_actor (output_item_box, self->output_item_class);

    self->output_text = st_label_new ("");
    ClutterText *output_text = (ClutterText *)st_label_get_clutter_text ((StLabel *)self->output_text);
    clutter_text_set_ellipsize (output_text, PANGO_ELLIPSIZE_NONE);
    st_widget_set_style_class_name (ST_WIDGET (self->output_text), "sound-text");
    _add_actor (output_item_box, self->output_text);
    
}

//输入声音(input)的ui框架
static void 
applet_sound_create_input_ui_frame (AppletSound *self, StBoxLayout *menu_box)
{
    self->input_box = st_box_layout_new ();
    st_box_layout_set_vertical (ST_BOX_LAYOUT(self->input_box), TRUE);
    _add_actor (menu_box, self->input_box);

    //“输入声音”
    StWidget *p_input_title = st_label_new (_("Input Volume"));
    st_widget_add_style_class_name (ST_WIDGET (p_input_title), "sound-title");
    _add_actor (self->input_box, p_input_title);
    clutter_actor_set_x_expand (CLUTTER_ACTOR (p_input_title), TRUE);
    clutter_actor_set_x_align (CLUTTER_ACTOR (p_input_title), CLUTTER_ACTOR_ALIGN_FILL);

    StWidget *input_main_box = st_box_layout_new ();
    st_box_layout_set_vertical (ST_BOX_LAYOUT (input_main_box), TRUE);
    _add_actor (self->input_box, input_main_box);

    StWidget *input_item_box = st_box_layout_new ();
    st_box_layout_set_vertical (ST_BOX_LAYOUT (input_item_box), FALSE);
    _add_actor (input_main_box,input_item_box);

    //静音按钮
    self->mute_in_switch = st_button_new ();
    st_widget_add_style_class_name (ST_WIDGET (self->mute_in_switch), "sound-mute");
    _add_actor (input_item_box, self->mute_in_switch);
    st_button_set_toggle_mode (ST_BUTTON (self->mute_in_switch), TRUE);
    g_object_set (self->mute_in_switch,
                  "x-expand", TRUE,
                  "x-align", CLUTTER_ACTOR_ALIGN_START,
                  "y-align", CLUTTER_ACTOR_ALIGN_CENTER,
                  NULL);
    //静音按钮内嵌图
    self->input_icon = ST_ICON (st_icon_new ());
    st_icon_set_icon_size (self->input_icon, 24);
    st_bin_set_child (ST_BIN (self->mute_in_switch), CLUTTER_ACTOR (self->input_icon));

    //滑杆
    self->input_item_class = st_box_layout_new ();
    st_widget_add_style_class_name (ST_WIDGET (self->input_item_class), "sound-slider");
    _add_actor (input_item_box, self->input_item_class);

    //音量百分比
    self->input_text = st_label_new ("");
    ClutterText *input_text = (ClutterText *)st_label_get_clutter_text ((StLabel *)self->input_text);
    clutter_text_set_ellipsize (input_text, PANGO_ELLIPSIZE_NONE);
    st_widget_add_style_class_name (ST_WIDGET (self->input_text), "sound-text");
    _add_actor (input_item_box, self->input_text);
}

static void
gvc_control_on_stream_changed (GvcMixerControl *control, guint id, gpointer user_data)
{
    AppletSound *self = (AppletSound *)user_data;
    g_debug ("Sound stream added or removed, id = %d", id);
    // only show input widgets if any application is recording audio
    self->p_show_input = FALSE;
    GSList *recording_apps = gvc_mixer_control_get_source_outputs (self->p_control);
    guint length, i;
    length = g_slist_length (recording_apps);
    if (self->p_input && length != 0)
    {
        for (i = 0; i < length; i++)
        {
            GvcMixerStream *output_stream = g_slist_nth_data (recording_apps, i);
            const gchar *id_str = gvc_mixer_stream_get_application_id (output_stream);
            // but skip gnome-volume-control and pavucontrol
            // (that appear as recording because they show the input level)
            gboolean tmp = (g_strcmp0 (id_str, "org.gnome.VolumeControl") != 0)
                           && (g_strcmp0 (id_str, "org.PulseAudio.pavucontrol") != 0);
            if (!id_str || tmp)
            {
                self->p_show_input = TRUE;
                break;
            }
        }
    }

    CdosPopupSeparatorMenuItem* item2 = cdos_popup_separator_menu_item_new();

    if (self->p_show_input)
    {
        clutter_actor_show (CLUTTER_ACTOR (item2));
        clutter_actor_show (CLUTTER_ACTOR (self->input_box));
    }
    else
    {
        clutter_actor_hide (CLUTTER_ACTOR (item2));
        clutter_actor_hide (CLUTTER_ACTOR (self->input_box));
    }
}

static void
applet_sound_toggle_out_mute_cb (GObject *obj, gboolean state, gpointer user_data)
{
    AppletSound *self = APPLET_SOUND (user_data);

    gboolean is_muted = gvc_mixer_stream_get_is_muted (self->p_output);

    if (is_muted)
    {
        gvc_mixer_stream_change_is_muted (self->p_output, FALSE);
        st_button_set_checked (ST_BUTTON (self->mute_out_switch), FALSE);
    }
    else
    {
        gvc_mixer_stream_change_is_muted (self->p_output, TRUE);
        st_button_set_checked (ST_BUTTON (self->mute_out_switch), TRUE);
    }
}

static void
applet_sound_toggle_in_mute_cb (GObject *obj, gboolean state, gpointer user_data)
{
    AppletSound *self = APPLET_SOUND (user_data);
    gboolean is_muted = gvc_mixer_stream_get_is_muted (self->p_input);

    if (is_muted)
    {
        gvc_mixer_stream_change_is_muted (self->p_input, FALSE);
        st_button_set_checked (ST_BUTTON (self->mute_in_switch), FALSE);
    }
    else
    {
        gvc_mixer_stream_change_is_muted (self->p_input, TRUE);
        st_button_set_checked (ST_BUTTON (self->mute_in_switch), TRUE);
    }
}

static gboolean
applet_sound_button_release_event (ClutterActor *actor, ClutterButtonEvent *button_event)
{
    AppletSound *self = APPLET_SOUND (actor);

    guint button = button_event->button;

    if (button == CLUTTER_BUTTON_MIDDLE)
    {
        gvc_mixer_stream_change_is_muted (self->p_output, !gvc_mixer_stream_get_is_muted (self->p_output));
    }

    return TRUE;
}

static gboolean
applet_sound_scroll_event (ClutterActor *actor, ClutterScrollEvent *scroll_event)
{
    AppletSound *self = APPLET_SOUND (actor);
    ClutterEvent *event = (ClutterEvent *)scroll_event;
    guint cal_volume_value;
    ClutterScrollDirection direction = clutter_event_get_scroll_direction (event);

    guint current_volume = gvc_mixer_stream_get_volume (self->p_output);

    if (direction == CLUTTER_SCROLL_DOWN)
    {
        gboolean prev_muted = gvc_mixer_stream_get_is_muted (self->p_output);
        if (self->p_volume_max * 0.01 >= current_volume)
            cal_volume_value = 0;
        else
            cal_volume_value = (guint) (current_volume - self->p_volume_max * 0.01);

        gvc_mixer_stream_set_volume (self->p_output, cal_volume_value);
        if (gvc_mixer_stream_get_volume (self->p_output) < 1)
        {
            gvc_mixer_stream_set_volume (self->p_output, 0);

            if (!prev_muted)
            {
                gvc_mixer_stream_change_is_muted (self->p_output, TRUE);
            }
        }

        gvc_mixer_stream_push_volume (self->p_output);
    }
    else if (direction == CLUTTER_SCROLL_UP)
    {
        cal_volume_value =
            MIN ((guint)self->p_volume_max, (guint) (current_volume + self->p_volume_max * 0.01));
        gvc_mixer_stream_set_volume (self->p_output, cal_volume_value);
        gvc_mixer_stream_change_is_muted (self->p_output, FALSE);
        gvc_mixer_stream_push_volume (self->p_output);
    }

    return TRUE;
}

//鼠标悬浮、点击时，静音图片状态
static void 
applet_sound_out_button_on_state (GObject *obj, GParamSpec *pspec, gpointer user_data)
{
    AppletSound *self = APPLET_SOUND (user_data);
    gboolean is_hover = FALSE, is_active = FALSE;
    const gchar *icon_str_sta = g_strdup (self->p_icon_name);
    is_hover = st_widget_has_style_pseudo_class (ST_WIDGET (self->mute_out_switch), "hover");
    is_active = st_widget_has_style_pseudo_class (ST_WIDGET (self->mute_out_switch), "active");
    if (is_hover)
    {
        g_free (icon_str_sta);
        icon_str_sta = g_strconcat (self->p_icon_name, "-hover", NULL);
    } 
    if (is_active)
    {
        g_free (icon_str_sta);
        icon_str_sta = g_strconcat (self->p_icon_name, "-active", NULL);
    }
    st_icon_set_icon_name (self->output_icon, icon_str_sta);
    g_free (icon_str_sta);
}

static void 
applet_sound_in_button_on_state (GObject *obj, GParamSpec *pspec, gpointer user_data)
{
    AppletSound *self = APPLET_SOUND (user_data);
    gboolean is_hover = FALSE, is_active = FALSE;
    gchar *icon_str = NULL;
    gboolean muted = gvc_mixer_stream_get_is_muted (self->p_input);
    is_hover = st_widget_has_style_pseudo_class (ST_WIDGET (self->mute_in_switch), "hover");
    is_active = st_widget_has_style_pseudo_class (ST_WIDGET (self->mute_in_switch), "active");
    if (muted)
    {
        icon_str = g_strdup ("found-audio-micro-volume-muted");
        if (is_hover)
        {
            g_free (icon_str);
            icon_str = g_strdup ("found-audio-micro-volume-muted-hover");
        }
        if (is_active)
        {
            g_free (icon_str);
            icon_str = g_strdup ("found-audio-micro-volume-muted-active");
        }
    }
    else
    {
        const gchar *icon_str_sta = applet_sound_volume_input_to_icon (self, gvc_mixer_stream_get_volume (self->p_input));
        icon_str = g_strdup (icon_str_sta);
        if (is_hover)
        {
            g_free (icon_str);
            icon_str = g_strconcat (icon_str_sta, "-hover", NULL);
        }
        if (is_active)
        {
            g_free (icon_str);
            icon_str = g_strconcat (icon_str_sta, "-active", NULL);
        }
    }
    st_icon_set_icon_name (self->input_icon, icon_str);
    g_free (icon_str);
}

static void
applet_sound_muted_changed_cb (GObject *obj, GParamSpec *param_spec, paramsInChanged *params)
{
    AppletSound *self = params->self;
    gboolean is_output = params->is_output;
    gboolean is_analog_stereo = FALSE;
    gboolean muted;
    
    if (is_output)
    {
        //_output,下面两句公有
        gfloat tmp = (gvc_mixer_stream_get_volume (self->p_output)) / (guint) (self->p_volume_max);
        muted = gvc_mixer_stream_get_is_muted (self->p_output);
        gchar *volume_tip = NULL, *volume_text = NULL;
        if (!g_strcmp0 (gvc_mixer_stream_get_description (self->p_output), _("Built-in Auido Analog Stereo"))) {
            is_analog_stereo = TRUE;
        }

        if (muted)
        {
            cdos_scale_set_value (self->output_slider, 0);
            applet_sound_set_icon_name (self, "found-audio-volume-muted");
            st_icon_set_icon_name (self->output_icon, "found-audio-volume-muted");
            volume_tip = g_strconcat (_("Muted"), NULL);
            volume_text = g_strconcat ("   0%", NULL);
            cdos_applet_set_applet_tooltip (CDOS_APPLET (self), volume_tip);
            st_label_set_text (ST_LABEL (self->output_text), volume_text);
            st_button_set_checked (ST_BUTTON (self->mute_out_switch), TRUE);
            g_free (volume_tip);
            g_free (volume_text);
            if (is_analog_stereo && g_file_test (ANALOG_STEREO_NOT_MUTED, G_FILE_TEST_EXISTS)) {
                g_remove (ANALOG_STEREO_NOT_MUTED);
            }
        }
        else
        {
            tmp = gvc_mixer_stream_get_volume (self->p_output) / self->p_volume_max;
            cdos_scale_set_value (self->output_slider, tmp);
            gchar *text;
            const char *icon_str;
            icon_str = applet_sound_volume_output_to_icon (self, gvc_mixer_stream_get_volume (self->p_output));
            applet_sound_set_icon_name (self, icon_str);
            st_icon_set_icon_name (self->output_icon, icon_str);

            gint percent_tooltip = (gint)round (gvc_mixer_stream_get_volume (self->p_output) * 100 / self->p_volume_max);
            text = g_strdup_printf (" %3d%%", percent_tooltip);
            volume_tip = g_strconcat (_("Volume"), ":", text, NULL);
            volume_text = g_strconcat (text, NULL);
            cdos_applet_set_applet_tooltip (CDOS_APPLET (self), volume_tip);
            st_label_set_text (ST_LABEL (self->output_text), volume_text);
            st_button_set_checked (ST_BUTTON (self->mute_out_switch), FALSE);

            g_free (text);
            g_free (volume_tip);
            g_free (volume_text);
            if (is_analog_stereo && !g_file_test (ANALOG_STEREO_NOT_MUTED, G_FILE_TEST_EXISTS))
            {
                int res = g_creat (ANALOG_STEREO_NOT_MUTED, 0644);
                if (res < 0)
                {
                    g_warning ("Fun:%s, Create analog stereo mute", __FUNCTION__);
                    return;
                }
            }
        }
    }
    else if (!is_output)
    {
        //_input,下面两句公有
        gfloat tmp = (gvc_mixer_stream_get_volume (self->p_input)) / (guint) (self->p_volume_max);
        muted = gvc_mixer_stream_get_is_muted (self->p_input);
        gchar *volume_text = NULL;

        if (muted)
        {
            cdos_scale_set_value (self->input_slider, 0);
            // applet_sound_set_icon_name (self, "found-audio-volume-muted");
            st_icon_set_icon_name (self->input_icon, "found-audio-micro-volume-muted");
            volume_text = g_strconcat ("   0%", NULL);
            // cdos_applet_set_applet_tooltip (CDOS_APPLET (self), volume_0);
            st_label_set_text (ST_LABEL (self->input_text), volume_text);
            st_button_set_checked (ST_BUTTON (self->mute_in_switch), TRUE);
            g_free (volume_text);
        }
        else
        {
            tmp = gvc_mixer_stream_get_volume (self->p_input) / self->p_volume_max;
            cdos_scale_set_value (self->input_slider, tmp);
            gchar *text;
            const char *icon_str;
            icon_str = applet_sound_volume_input_to_icon (self, gvc_mixer_stream_get_volume (self->p_input));
            // applet_sound_set_icon_name (self, (const gchar *)icon_str);
            st_icon_set_icon_name (self->input_icon, icon_str);
            
            gint percent_tooltip = (gint)round (gvc_mixer_stream_get_volume (self->p_input) * 100 / self->p_volume_max);
            text = g_strdup_printf (" %3d%%", percent_tooltip);
            volume_text = g_strconcat (text, NULL);
            // cdos_applet_set_applet_tooltip (CDOS_APPLET (self), volume_tip);
            st_label_set_text (ST_LABEL (self->input_text), volume_text);
            st_button_set_checked (ST_BUTTON (self->mute_in_switch), FALSE);

            g_free (text);
            g_free (volume_text);
        }
    }
}

static void
applet_sound_volume_changed_cb (GObject *obj, GParamSpec *param_spec, paramsInChanged *params)
{
    gboolean is_output = params->is_output;
    AppletSound *self = params->self;
    gfloat tmp;
    g_debug ("Sound volume changed, is output: %d, state: %d", is_output, gvc_mixer_control_get_state (self->p_control));
    if (is_output)
    {
        if (!self->p_output)
        {
            return;
        }
        if (!gvc_mixer_stream_get_is_muted (self->p_output))
        {
            tmp = (gfloat) (gvc_mixer_stream_get_volume (self->p_output) / self->p_volume_max);
            cdos_scale_set_value (self->output_slider, tmp);

            const gchar * str;
            str = applet_sound_volume_output_to_icon (self, gvc_mixer_stream_get_volume (self->p_output));
            st_icon_set_icon_name (self->output_icon, str);
            applet_sound_set_icon_name (self, str);

            gint percent_tooltip = (gint)round (gvc_mixer_stream_get_volume (self->p_output) * 100 / self->p_volume_max);
            gchar *text = NULL, *volume_tip = NULL, *volume_text = NULL;
            text = g_strdup_printf (" %3d%%", percent_tooltip); //输出%格式是%%
            volume_tip = g_strconcat (_("Volume"), ":", text, NULL);
            volume_text = g_strconcat (text, NULL);
            cdos_applet_set_applet_tooltip (CDOS_APPLET (self), volume_tip);
            st_label_set_text (ST_LABEL (self->output_text), volume_text);
            g_free (text);
            g_free (volume_tip);
            g_free (volume_text);
            
            if (gvc_mixer_control_get_state (self->p_control) != GVC_STATE_FAILED)
                cdos_sound_manager_play ("volume");
        }
    }
    else if (!is_output)
    {
        if (!self->p_input)
        {
            return;
        }

        if (!gvc_mixer_stream_get_is_muted (self->p_input))
        {
            tmp = (gfloat) (gvc_mixer_stream_get_volume (self->p_input) / self->p_volume_max);
            cdos_scale_set_value (self->input_slider, tmp);

            const char *str = applet_sound_volume_input_to_icon (self, gvc_mixer_stream_get_volume (self->p_input));
            // applet_sound_set_icon_name (self, str);
            st_icon_set_icon_name (self->input_icon, str);

            gint percent_tooltip = (gint)round (gvc_mixer_stream_get_volume (self->p_input) * 100 / self->p_volume_max);
            gchar *text = NULL, *volume_text = NULL;
            text = g_strdup_printf (" %3d%%", percent_tooltip); //输出%格式是%%
            volume_text = g_strconcat (text, NULL);
            // cdos_applet_set_applet_tooltip (CDOS_APPLET (self), volume_value);
            st_label_set_text (ST_LABEL (self->input_text), volume_text);

            g_free (text);
            g_free (volume_text);
        }
    }
    else
    {
        g_warning ("Property not correct!\n");
        return;
    }
}

static void
applet_sound_slider_changed_cb (GObject *obj, gfloat value, gpointer user_data)
{
    paramsInChanged *params = (paramsInChanged *)user_data;
    gboolean is_output = params->is_output;
    AppletSound *self = APPLET_SOUND (params->self);
    // CdosPopupBaseMenuItem *slider = CDOS_POPUP_BASE_MENU_ITEM (obj);
    /*
    if (g_strcmp0 (property, "_input") != 0 && g_strcmp0 (property, "_output") != 0)
    {
        g_warning ("Volume slider changed for %s, but %s does not exist", property, property);
        return;
    }
    */

    gfloat volume = value * self->p_volume_max;
    gboolean prev_muted;
    if (!is_output)
    {
        prev_muted = gvc_mixer_stream_get_is_muted (self->p_input);
        if (volume < 1)
        {
            gvc_mixer_stream_set_volume (self->p_input, 0);
            if (!prev_muted)
            {
                gvc_mixer_stream_change_is_muted (self->p_input, TRUE);
            }
        }
        else
        {
            gvc_mixer_stream_set_volume (self->p_input, (guint)volume);
            if (prev_muted)
            {
                gvc_mixer_stream_change_is_muted (self->p_input, FALSE);
                st_button_set_checked (ST_BUTTON (self->mute_in_switch), FALSE);
            }
        }

        gvc_mixer_stream_push_volume (self->p_input);
    }
    else
    {
        prev_muted = gvc_mixer_stream_get_is_muted (self->p_output);

        if (volume < 1)
        {
            gvc_mixer_stream_set_volume (self->p_output, 0);
            if (!prev_muted)
            {
                gvc_mixer_stream_change_is_muted (self->p_output, TRUE);
            }
        }
        else
        {
            gvc_mixer_stream_set_volume (self->p_output, (guint)volume);
            if (prev_muted)
            {
                gvc_mixer_stream_change_is_muted (self->p_output, FALSE);
                st_button_set_checked (ST_BUTTON (self->mute_out_switch), FALSE);
            }
        }

        gvc_mixer_stream_push_volume (self->p_output);
    }
}

static void
applet_sound_update (AppletSound *self)
{
    if (gvc_mixer_control_get_state (self->p_control) == GVC_STATE_READY)
    {
        applet_sound_read_output_cb (NULL, 0, self);
        applet_sound_read_input_cb (NULL, 0, self);
        clutter_actor_show (CLUTTER_ACTOR (self));
    }
    else
    {
        clutter_actor_hide (CLUTTER_ACTOR (self));
    }
}

static void
gvc_control_on_card_added (GvcMixerControl *control, guint index, gpointer user_data)
{
    g_message ("Sound control card added: index = %d", index);
    AppletSound *self = APPLET_SOUND (user_data);
    applet_sound_update (self);
}

static void
gvc_control_on_card_removed (GvcMixerControl *control, guint index, gpointer user_data)
{
    g_message ("Sound control card removed: index = %d", index);
    AppletSound *self = APPLET_SOUND (user_data);
    applet_sound_update (self);
}

static void
gvc_control_on_state_changed (GvcMixerControl *control, guint state, gpointer user_data)
{
    g_message ("Sound control state changed: state = %d", state);
    AppletSound *self = APPLET_SOUND (user_data);
    applet_sound_update (self);
}

static void
applet_sound_read_output_cb (GObject *obj, guint state, gpointer user_data)
{
    AppletSound *self = APPLET_SOUND (user_data);
    g_message ("Sound read output: state = %d", gvc_mixer_control_get_state (self->p_control));

    if (self->p_output_volume_id != 0)
    {
        g_signal_handler_disconnect (self->p_output, self->p_output_volume_id);
        g_signal_handler_disconnect (self->p_output, self->p_output_muted_id);
        self->p_output_volume_id = 0;
        self->p_output_muted_id = 0;
    }

    if (self->p_output_notify_id != 0)
    {
        g_signal_handler_disconnect (self->p_output, self->p_output_notify_id);
        self->p_output_notify_id = 0;
    }

    gboolean is_analog_stereo = FALSE;
    static guint once = 1;

    self->p_output = gvc_mixer_control_get_default_sink (self->p_control);
    if (self->p_output)
    {
        if (!g_strcmp0 (gvc_mixer_stream_get_description (self->p_output), _("Built-in Auido Analog Stereo"))) {
            is_analog_stereo = TRUE;
        }

        if (is_analog_stereo && g_file_test (ANALOG_STEREO_NOT_MUTED, G_FILE_TEST_EXISTS) && once) {
            // 模拟立体声，上一次状态是非静音
            if (gvc_mixer_stream_get_is_muted (self->p_output)) {
                gvc_mixer_stream_change_is_muted (self->p_output, FALSE);
                //g_timeout_add_seconds (5, timeout_fixed_cb, self);
            }
            // 只修正一次
            once = once - 1;
        }

        paramsInChanged *params_output = g_new0(paramsInChanged, 1);
        params_output->self = self;
        params_output->is_output = TRUE;
        self->p_output_muted_id = g_signal_connect (
            self->p_output, "notify::is-muted", G_CALLBACK (applet_sound_muted_changed_cb), params_output);
        self->p_output_volume_id = g_signal_connect (
            self->p_output, "notify::volume", G_CALLBACK (applet_sound_volume_changed_cb), params_output);

        applet_sound_muted_changed_cb (NULL, NULL, params_output);
        applet_sound_volume_changed_cb (NULL, NULL, params_output);
    }
    else
    {
        cdos_scale_set_value (self->output_slider, 0);
        applet_sound_set_icon_name (self, "found-audio-volume-muted");
    }
}

void
applet_sound_read_input_cb (GObject *obj, guint state, gpointer user_data)
{
    AppletSound *self = APPLET_SOUND (user_data);
    g_message ("Sound read input: state = %d", gvc_mixer_control_get_state (self->p_control));

    if (self->p_input_volume_id != 0)
    {
        g_signal_handler_disconnect (self->p_input, self->p_input_volume_id);
        g_signal_handler_disconnect (self->p_input, self->p_input_muted_id);
        self->p_input_volume_id = 0;
        self->p_input_muted_id = 0;
    }

    self->p_input = gvc_mixer_control_get_default_source (self->p_control);
    if (self->p_input)
    {
        paramsInChanged *params = g_new0 (paramsInChanged, 1);
        params->self = self;
        params->is_output = FALSE;

        self->p_input_muted_id =
            g_signal_connect (self->p_input, "notify::is-muted", G_CALLBACK (applet_sound_muted_changed_cb), params);
        self->p_input_volume_id =
            g_signal_connect (self->p_input, "notify::volume", G_CALLBACK (applet_sound_volume_changed_cb), params);

        applet_sound_muted_changed_cb (NULL, NULL, params);
        applet_sound_volume_changed_cb (NULL, NULL, params);
    }
    else
    {
        clutter_actor_hide (CLUTTER_ACTOR (self->input_box));
    }
}

static gboolean
applet_sound_on_click_sound_settings_cb (GObject *obj, ClutterEvent *event, gpointer user_data)
{
    // TODO:sound settings
    gint ret = cdos_util_spawn_command_line ("found-control-center sound");
    if (ret < 0)
    {
        return FALSE;
    }

    return TRUE;
}

static void
applet_sound_settings_changed (GSettings *settings, gchar *settings_key, gpointer user_data)
{
    AppletSound *self = APPLET_SOUND (user_data);
    if (g_strcmp0 (settings_key, KEY_SOUND) == 0)
    {
        gboolean state = g_settings_get_boolean (settings, KEY_SOUND);
        if (state)
            cdos_scale_set_range (self->output_slider, 0, 1.2, 0);
        else
            cdos_scale_set_range (self->output_slider, 0, 1, 0);
    }
}

static void
applet_sound_dispose (GObject *obj)
{
    AppletSound *self = APPLET_SOUND (obj);

    g_clear_object (&self->p_control);

    G_OBJECT_CLASS (applet_sound_parent_class)->dispose (obj);
}

static void
applet_sound_finalize (GObject *obj)
{
    AppletSound *self = APPLET_SOUND (obj);

    g_free (self->p_icon_name);
    g_free (self->p_icon_path);

    G_OBJECT_CLASS (applet_sound_parent_class)->finalize (obj);
}

static void
applet_sound_class_init (AppletSoundClass *klass)
{
    GObjectClass *gobject_class = G_OBJECT_CLASS (klass);
    gobject_class->dispose = applet_sound_dispose;
    gobject_class->finalize = applet_sound_finalize;

    ClutterActorClass *actor_class = CLUTTER_ACTOR_CLASS (klass);
    actor_class->scroll_event = applet_sound_scroll_event;
    actor_class->button_release_event = applet_sound_button_release_event;

    /*创建信号*/
    applet_sound_signals[PLAYERS_LOADED] = g_signal_new ("players-loaded",
                                                         G_TYPE_FROM_CLASS (klass),
                                                         G_SIGNAL_RUN_LAST,
                                                         G_STRUCT_OFFSET (AppletSoundClass, players_loaded),
                                                         NULL,
                                                         NULL,
                                                         g_cclosure_marshal_VOID__VOID,
                                                         G_TYPE_NONE,
                                                         0,
                                                         NULL);
}

/* _init */
static void
applet_sound_init (AppletSound *self)
{
    cdos_iconapplet_set_icon_name (CDOS_ICONAPPLET (self), "found-audio-volume-muted");
    cdos_iconapplet_set_icon_size (CDOS_ICONAPPLET (self), 16);

    cdos_applet_set_popup_menu_active (CDOS_APPLET (self), TRUE);
    cdos_applet_set_context_menu_active (CDOS_APPLET (self), TRUE);

    self->p_control = gvc_mixer_control_new ("Cdos Volume Control");
    g_signal_connect (self->p_control, "state-changed", G_CALLBACK (gvc_control_on_state_changed), self);
    g_signal_connect (self->p_control, "card-added", G_CALLBACK (gvc_control_on_card_added), self);
    g_signal_connect (self->p_control, "card-removed", G_CALLBACK (gvc_control_on_card_removed), self);
    g_signal_connect (self->p_control, "default-sink-changed", G_CALLBACK (applet_sound_read_output_cb), self);
    g_signal_connect (self->p_control, "default-source-changed", G_CALLBACK (applet_sound_read_input_cb), self);
    g_signal_connect (self->p_control, "stream-added", G_CALLBACK (gvc_control_on_stream_changed), self);
    g_signal_connect (self->p_control, "stream-removed", G_CALLBACK (gvc_control_on_stream_changed), self);

/* TODO:
     * previously was 1.5*this._control.get_vol_max_norm();,
     * but we'd need a little mark on the slider to make it obvious to the user we're going over 100%..
     */
#if 1
    self->p_volume_max = (gfloat) (1 * gvc_mixer_control_get_vol_max_norm (self->p_control));
#endif
#if 0
    self->p_volume_max = (gfloat)(1 * gvc_mixer_control_get_vol_max_amplified (self->p_control));
#endif
    self->p_output = NULL;
    self->p_output_volume_id = 0;
    self->p_output_muted_id = 0;
    self->p_output_notify_id = 0;

    self->p_input = NULL;
    self->p_input_volume_id = 0;
    self->p_input_muted_id = 0;

    self->p_icon_name = NULL;
    self->p_icon_path = NULL;
    self->p_volume_control_shown = FALSE;
}

/**
 * applet_sound_main:
 * @metadata: applet元数据
 * @orientation: 方向
 * @panel_height: 面板高度
 * @instanceId: 字符串类型的id
 *
 * 新建一个 #CdosApplet 类的实例。
 *
 * Returns: 返回一个 #CdosApplet 的实例。
 */
CdosApplet *
applet_sound_main (MetaData *metadata, StSide orientation, guint panel_height, gchar *instanceId)
{
    AppletSound *self = g_object_new (APPLET_TYPE_SOUND, NULL);

    cdos_applet_set_orientation (CDOS_APPLET (self), orientation);
    cdos_applet_set_panel_height (CDOS_APPLET (self), panel_height);

    self->menu = cdos_applet_get_popup_menu ((CdosApplet *)self);
    StBoxLayout *menu_box = cdos_popup_menu_base_get_box (CDOS_POPUP_MENU_BASE (self->menu));
    st_box_layout_set_vertical (ST_BOX_LAYOUT (menu_box), TRUE);
    applet_sound_create_output_ui_frame (self, menu_box);
    applet_sound_create_input_ui_frame (self, menu_box);

    applet_sound_show_fixed_elements (self);

    gboolean result = gvc_mixer_control_open (self->p_control);
    if (result)
    {
        g_warning ("gvc_mixer_control_open error!");
    }

    g_signal_connect (self->sound_settings,
                      "button-press-event", G_CALLBACK (applet_sound_on_click_sound_settings_cb), NULL);
    return CDOS_APPLET (self);
}

/**
 * applet_sound_on_applet_removed_from_panel:
 * @self: A #AppletSound
 *
 * 从面板移除applet
 *
 */
void
applet_sound_on_applet_removed_from_panel (AppletSound *self)
{
    if (self->p_icon_timeout_id)
    {
        g_source_remove (self->p_icon_timeout_id);
    }
}

/**
 * applet_sound_set_icon_name:
 * @self: A #AppletSound
 * @icon: 图标名称
 *
 * 设置图标名称
 *
 */
void
applet_sound_set_icon_name (AppletSound *self, const gchar *icon)
{
    if (g_strcmp0 (icon, self->p_icon_name) == 0)
        return;

    if (self->p_icon_name)
        g_free (self->p_icon_name);
    self->p_icon_name = g_strdup (icon);
    cdos_iconapplet_set_icon_name (CDOS_ICONAPPLET (self), icon);
}

/*
 * applet_sound_set_applet_icon:
 * @self: A #AppletSound
 * @player: player对象
 * @path: 路径
 *
 * 设置applet里player的名称，未完成
 *
 */
void
applet_sound_set_applet_icon (AppletSound *self, GObject *player, gchar *path)
{
    // FIXME
    // Player相关,暂时不写!
}

/*
 * applet_sound_set_applet_text:
 * @self: A #AppletSound
 * @player: player对象
 *
 * 设置applet里player的图标，未完成
 *
 */
void
applet_sound_set_applet_text (AppletSound *self, GObject *player)
{
    // FIXME
}

/*
 * applet_sound_set_applet_text_icon:
 * @self: A #AppletSound
 * @player: player对象
 * @path: 路径
 *
 * 设置applet里player的文本，未完成
 */
void
applet_sound_set_applet_text_icon (AppletSound *self, GObject *player, gchar *path)
{
    applet_sound_set_applet_icon (self, player, path);
    applet_sound_set_applet_text (self, player);
}

static void
applet_sound_show_fixed_elements (AppletSound *self)
{
    CdosPopupMenu *context_menu;
    if (self->p_volume_control_shown)
    {
        return;
    }

    self->p_volume_control_shown = TRUE;
    /* FIXME:
     * 是否需要?
     */
    st_button_set_checked (ST_BUTTON (self->mute_out_switch), FALSE);
    st_icon_set_icon_name (self->output_icon, "found-audio-volume-high");
    g_signal_connect (self->mute_out_switch, 
                      "notify::pseudo-class", 
                      G_CALLBACK (applet_sound_out_button_on_state), 
                      self);
    GSettings *settings = g_settings_new (KEY_SOUNDS_SCHEMA);
    g_signal_connect (settings, "changed::allow-volume-above-100-percent", G_CALLBACK (applet_sound_settings_changed), self);
    gboolean status = g_settings_get_boolean (settings, KEY_SOUND);
    gfloat max_slider = 1.0;
    if (status)
        max_slider = 1.2;
    self->output_slider = cdos_scale_new_with_range (CLUTTER_ORIENTATION_HORIZONTAL, 0, max_slider, 0);
    cdos_scale_set_direction (self->output_slider, CLUTTER_TEXT_DIRECTION_LTR);
    st_widget_add_style_class_name ((StWidget *)self->output_slider, "sound-scale");
    //track-hover指针是否悬浮在部件上，只在为TRUE时跟踪
    StWidget *output_item = st_box_layout_new ();
    g_object_set (output_item, "track-hover", FALSE, NULL);
    st_widget_add_style_class_name ((StWidget *)output_item, "sound-scale-item");
    ActorChild prop;
    prop.expand = TRUE;
    prop.align = -1;
    _add_actor (output_item, self->output_slider);
    clutter_actor_set_x_expand (CLUTTER_ACTOR (self->output_slider), prop.expand);
    clutter_actor_set_x_align (CLUTTER_ACTOR (self->output_slider), prop.align);
    StWidget *output_item_1 = output_item;

    paramsInChanged *params_out = g_new0 (paramsInChanged, 1);
    params_out->self = self;
    params_out->is_output = TRUE;

    g_signal_connect(self->output_slider, "value-changed",
                     G_CALLBACK (applet_sound_slider_changed_cb), params_out);

    _add_actor (self->output_item_class, output_item);
    g_signal_connect (self->mute_out_switch, "clicked", G_CALLBACK (applet_sound_toggle_out_mute_cb), self);

    st_widget_set_track_hover ((StWidget *)self->mute_in_switch, TRUE);
    st_icon_set_icon_name (ST_ICON (self->input_icon), "found-audio-micro-volume-high");
    g_signal_connect (self->mute_in_switch, 
                      "notify::pseudo-class", 
                      G_CALLBACK (applet_sound_in_button_on_state), 
                      self);

    self->input_slider = cdos_scale_new_with_range (CLUTTER_ORIENTATION_HORIZONTAL, 0, 1, 0);
    cdos_scale_set_direction (self->input_slider, CLUTTER_TEXT_DIRECTION_LTR);
    st_widget_add_style_class_name ((StWidget *)self->input_slider, "sound-scale");
    StWidget *input_item = st_box_layout_new ();
    g_object_set (input_item, "track-hover", FALSE, NULL);
    st_widget_add_style_class_name (input_item, "sound-scale-item");
    prop.expand = TRUE;
    prop.align = -1;
    _add_actor (input_item, self->input_slider);
    clutter_actor_set_x_expand (CLUTTER_ACTOR (self->input_slider), prop.expand);
    clutter_actor_set_x_align (CLUTTER_ACTOR (self->input_slider), prop.align);
    StWidget *input_item_1 = input_item;

    paramsInChanged *params_in = g_new0(paramsInChanged, 1);
    params_in->is_output = FALSE;
    params_in->self = self;

    g_signal_connect (self->input_slider,
                      "value-changed", G_CALLBACK (applet_sound_slider_changed_cb), params_in);
    _add_actor (self->input_item_class, input_item);
    g_signal_connect (self->mute_in_switch, "clicked", G_CALLBACK (applet_sound_toggle_in_mute_cb), self);

    /* FIXME:
     * 替换:cdos_popup_menu_base_add_settings_action (CDOS_POPUP_MENU_BASE (self->menu), "Sound Settings", "sound");
     * 暂时不写:self->p_select_device_item = cdos_popup_submenu_menu_item_new (_("Output device..."), TRUE);
     */

    /* CdosPopupSeparatorMenuItem *item2 = cdos_popup_separator_menu_item_new ();
    cdos_popup_menu_base_add_menu_item (CDOS_POPUP_MENU_BASE (self->menu), CDOS_POPUP_BASE_MENU_ITEM (item2), -1); */

    if (self->p_show_input)
    {
        clutter_actor_show (CLUTTER_ACTOR (self->input_box));
    }
    else
    {
        clutter_actor_hide (CLUTTER_ACTOR (self->input_box));
    }

    context_menu = cdos_applet_get_context_menu (CDOS_APPLET (self));
    self->sound_settings = cdos_popup_menu_item_new ();
    // self->sound_settings = cdos_popup_menu_item_new_with_params (_("Sound Settings"), params_for_new_item);
    cdos_popup_menu_item_set_label (self->sound_settings, _("Sound Settings"));
    cdos_popup_menu_item_set_label_align (self->sound_settings, ST_ALIGN_MIDDLE, 0);
    cdos_popup_menu_base_add_menu_item (
        CDOS_POPUP_MENU_BASE (context_menu), CDOS_POPUP_BASE_MENU_ITEM (self->sound_settings), -1);
    StLabel *label = cdos_popup_menu_item_get_label (self->sound_settings);
    st_widget_add_style_class_name (ST_WIDGET (label), "popup-menu-item-label-link");
    //CdosGenericContainer *sound_settings_actor = CLUTTER_ACTOR (self->sound_settings);
    // st_widget_add_style_class_name(ST_WIDGET(sound_settings_actor),"sound-menu-item-settings");

    paramsInChanged *params_input, *params_output;
    params_input = g_new0 (paramsInChanged, 1);
    params_input->self = self;
    params_input->is_output = FALSE;

    params_output = g_new0 (paramsInChanged, 1);
    params_output->self = self;
    params_output->is_output = TRUE;

    applet_sound_volume_changed_cb (NULL, NULL, params_output);
    applet_sound_volume_changed_cb (NULL, NULL, params_input);
    // g_free (params_for_new_item);
}

static const gchar *
applet_sound_volume_output_to_icon (AppletSound *self, gfloat volume)
{
    static const char *volumes[4] = {
        "found-audio-volume-muted",
        "found-audio-volume-low",
        "found-audio-volume-medium",
        "found-audio-volume-high"
    };

    if (volume <= 0) {
        return volumes[0];
    } else {
        guint n = 3 * volume / self->p_volume_max + 1;
        if (n >= 3) {
            return volumes[3];
        }
        return volumes[n];
    }
}

static const gchar *
applet_sound_volume_input_to_icon (AppletSound *self, gfloat volume)
{
    static const char *volumes[4] = {
        "found-audio-micro-volume-muted",
        "found-audio-micro-volume-low",
        "found-audio-micro-volume-medium",
        "found-audio-micro-volume-high"
    };

    if (volume <= 0) {
        return volumes[0];
    } else {
        guint n = 3 * volume / self->p_volume_max + 1;
        if (n >= 3) {
            return volumes[3];
        }
        return volumes[n];
    }
}
