#include <math.h>
#include <meta/meta-backend.h>
#include <meta/window.h>
#include <meta/workspace.h>
#include <meta/meta-monitor-manager.h>

#include "cdos-app-system.h"
#include "cdos-global.h"

#include "cdos-start.h"
#include "cdos-drag-drop.h"
#include "cdos-window-tracker.h"

#include "applet-app-list.h"
#include "applet-window-list.h"
#include "applet-pinned-favs.h"
#include "applet-app-group.h"
#include "applet-tree.h"
#include "applet-panel-launchers-manager.h"

typedef enum
{
    PAGE_PREV,
    PAGE_NEXT
} PaginationAction;

struct _AppletAppListPrivate
{
    ClutterActor *actor; /* StBoxLayout */
    MetaWorkspace *meta_workspace;

    ClutterActor *placeholder;
    guint page_count;

    ClutterActor *button_prev;
    ClutterActor *button_next;

    GTree *app_to_app_group;
    GTree *window_to_app;
    CdosWindowTracker *tracker;

    gboolean group_app_buttons;
    AppletPinnedFavs *pinned_favs;
    gboolean show_pinned_favs;

    gboolean need_refresh;

    StScrollView *scroll_view;

    GSettings *settings;
    GSettings *global_settings;
    MetaDisplay *display;
    MetaWorkspaceManager *workspace_manager;
    MetaMonitorManager *monitor_manager;
    CdosAppSystem *app_system;

    gfloat max_width;

    ClutterAction *drop_action;
    gint drag_position;
    gulong drag_motion_id;

    gulong workspace_changed_id;

    //watch launchers dir monitor;
    GFileMonitor *launchers_monitor;
    guint launchers_monitor_id;
};

G_DEFINE_TYPE_WITH_PRIVATE (AppletAppList, applet_app_list, ST_TYPE_BOX_LAYOUT);

static void applet_app_list_dispose (GObject *object);
static void applet_app_list_finalize (GObject *object);

static void applet_app_list_set_signals (AppletAppList *self);
static void applet_app_list_refresh_apps (AppletAppList *self);
static void applet_app_list_refresh_list (AppletAppList *self);
static void applet_app_list_window_added (AppletAppList *self,
                                          MetaWorkspace *meta_workspace,
                                          MetaWindow *meta_window,
                                          CdosApp *favapp,
                                          gboolean is_favapp);
static void applet_app_list_on_window_added (MetaWorkspace *meta_workspace, MetaWindow *meta_window, AppletAppList *self);
static void applet_app_list_remove_app (AppletAppList *self, CdosApp *app);

static void applet_app_list_load_favorites (AppletAppList *self);
static void
applet_app_list_window_removed (AppletAppList *self, MetaWorkspace *meta_workspace, MetaWindow *meta_window);
static void
applet_app_list_on_window_removed (MetaWorkspace *meta_workspace, MetaWindow *meta_window, AppletAppList *self);
static void panel_launchers_on_changed (GSettings *settings, gchar *key, AppletAppList *self);
static void refresh_apps_foreach_callback (gpointer meta_window, gpointer user_data);

static gboolean applet_app_list_remove_app_callback (AppletAppList *self);

static void applet_app_list_pagination_setup (AppletAppList *self);

static void applet_app_list_pagination_hide (AppletAppList *self);

static void applet_app_list_set_page_count (AppletAppList *self, guint new_count);
static void button_prev_on_page_up (StButton *button, gint clicked_button, AppletAppList *self);
static void button_next_on_page_down (StButton *button, gint clicked_button, AppletAppList *self);

static void applet_app_list_button_prev_set_state (AppletAppList *self, gboolean state);

static void applet_app_list_button_next_set_state (AppletAppList *self, gboolean state);

static gboolean applet_app_list_on_scroll (ClutterActor *actor, ClutterEvent *event, AppletAppList *self);

static void system_on_installed_changed (CdosAppSystem *app_system, CdosApp *app, AppletAppList *self);

static void on_monitors_changed (MetaMonitorManager *monitor_manager, AppletAppList *self);

static void applet_app_list_condense_app_groups (AppletAppList *self, gboolean condense);

static void applet_app_list_focus_app (AppletAppList *self);

static void applet_app_list_update_buttons_state (AppletAppList *self);

static char *current_desktop_id = NULL;

static void
should_show_placeholder (AppletAppList *self,
                         gfloat         event_x,
                         gfloat         event_y)
{
    ClutterDragAction *drag = cdos_get_current_drag_action ();
    if (!drag)
    {
        g_warn_if_reached ();
        drag_drop_successfull = FALSE;
        return;
    }
    AppletAppListPrivate *priv = self->priv;

    ClutterActor *drop_actor = (ClutterActor *)priv->actor;
    gfloat x, y;
    clutter_actor_transform_stage_point (drop_actor, event_x, event_y, &x, &y);

    ClutterActor *actor = drop_actor;
    gfloat pos_x, width;
    pos_x = clutter_actor_get_x (actor);
    width = clutter_actor_get_width (actor);
    if (x > pos_x && x < pos_x + width)
    {
        clutter_actor_transform_stage_point (actor, event_x, event_y, &x, &y);
        ClutterActor *placeholder = priv->placeholder;

        gint pos = 0;
        gint app_pos = 0;
        for (ClutterActor *child = clutter_actor_get_first_child (actor);
             child;
             child = clutter_actor_get_next_sibling (child))
        {
            if (!clutter_actor_is_visible (child))
            {
                if (child != placeholder)
                    pos++;
                continue;
            }
            if (child == placeholder)
                continue;
            gfloat gx, gw;
            gx = clutter_actor_get_x (child);
            gw = clutter_actor_get_width (child);
            if (x <= gx + gw / 2.0)
            {
                break;
            }
            pos++;
            if (APPLET_IS_APP_GROUP (child))
            {
                AppletAppGroup *app = (AppletAppGroup *)child;
                if (applet_app_group_get_is_favorite_app (app) ||
                    applet_app_group_get_was_favorite_app (app))
                {
                    ++app_pos;
                }
            }
        }
        clutter_actor_set_child_at_index (priv->actor, placeholder, pos);
        clutter_actor_show (placeholder);
        priv->drag_position = app_pos;
    }
}

typedef struct _LauncherData
{
    char *desktop_id;
    gint  new_pos;
}LauncherData;

static void
change_panel_launchers (GTask *task,
                        gpointer source_object,
                        LauncherData *data,
                        GCancellable *cancellable)
{
    char *desktop_id = data->desktop_id;
    gint new_pos = data->new_pos;

    GSettings *settings = g_settings_new ("org.cdos");
    char **launchers = g_settings_get_strv (settings, "panel-launchers");
    if (!launchers)
        return;

    gint len = g_strv_length (launchers);
    if (new_pos > len)
        new_pos = len - 1;
    gint i;
    for (i = 0; launchers[i] != NULL; i++)
    {
        if (g_strcmp0 (desktop_id, launchers[i]) == 0)
        {
            if (i == new_pos)
                goto out;

            //char *tmp0 = launchers[new_pos];
            char *tmp1 = launchers[i];
            if (i > new_pos)
            {
                for (int j = i; j > new_pos; j--)
                    launchers[j] = launchers[j - 1];
                launchers[new_pos] = tmp1;
            }
            else if (i < new_pos)
            {
                for (int j = i; j < new_pos; j++)
                    launchers[j] = launchers[j + 1];
                launchers[new_pos] = tmp1;
            }
            g_settings_set_strv (settings, "panel-launchers", (const char * const *)launchers);
            drag_drop_successfull = TRUE;
            goto out;
        }
    }
    guint max_apps = g_settings_get_int(settings, "panel-launchers-max-number");
    if (i >= max_apps)
        goto out;

    char **new_settings = g_malloc (sizeof (char *) * (len + 2));
    int j;
    for (j = 0; j < new_pos; j++)
        new_settings[j] = launchers[j];
    for (j = new_pos; j < len; j++)
        new_settings[j + 1] = launchers[j];
    new_settings[new_pos] = g_strdup (desktop_id);
    new_settings[j + 1] = NULL;
    g_settings_set_strv (settings, "panel-launchers", (const char * const *)new_settings);
    g_free (new_settings[new_pos]);
    g_free (new_settings);
    current_desktop_id = NULL;
    drag_drop_successfull = TRUE;

out:
    g_strfreev (launchers);
    g_free (desktop_id);
    g_free (data);
    g_object_unref (settings);
}

static void
on_drag_motion (ClutterDragAction *action,
                ClutterActor      *actor,
                gfloat             delta_x,
                gfloat             delta_y,
                AppletAppList     *self)
{
    gfloat motion_x, motion_y;
    clutter_drag_action_get_motion_coords (action, &motion_x, &motion_y);
    should_show_placeholder (self, motion_x, motion_y);
}

static void
on_drop_over_in (CdosDropAction    *action,
                 ClutterActor      *actor,
                 AppletAppList     *self)
{
    ClutterDragAction *drag = cdos_get_current_drag_action ();
    if (!drag)
    {
        g_warn_if_reached ();
        drag_drop_successfull = FALSE;
        return;
    }
    self->priv->drag_motion_id = 0;
    current_desktop_id = g_object_get_data ((GObject *)drag, "desktop-id");
    if (!current_desktop_id)
        return;

    self->priv->drag_motion_id = g_signal_connect (drag, "drag-motion",
                                             G_CALLBACK (on_drag_motion), self);

    gfloat motion_x, motion_y;
    clutter_drag_action_get_motion_coords (drag, &motion_x, &motion_y);
    should_show_placeholder (self, motion_x, motion_y);
}

static void
on_drop_over_out (CdosDropAction   *action,
                  ClutterActor     *actor,
                  AppletAppList    *self)
{
    ClutterActor *placeholder = self->priv->placeholder;
    clutter_actor_hide (placeholder);
    ClutterDragAction *drag = cdos_get_current_drag_action ();
    if (!drag)
    {
        g_warn_if_reached ();
        drag_drop_successfull = FALSE;
        return;
    }
    if (self->priv->drag_motion_id)
        g_signal_handler_disconnect (drag, self->priv->drag_motion_id);
    self->priv->drag_motion_id = 0;
    current_desktop_id = NULL;

    drag_drop_successfull = FALSE;
}

static void
on_drop_drop (CdosDropAction    *action,
              ClutterActor      *actor,
              gfloat             event_x,
              gfloat             event_y,
              AppletAppList  *self)
{
    AppletAppListPrivate *priv = self->priv;
    ClutterActor *placeholder = self->priv->placeholder;
    clutter_actor_hide (placeholder);

    ClutterDragAction *drag = cdos_get_current_drag_action ();
    if (!drag)
    {
        g_warn_if_reached ();
        drag_drop_successfull = FALSE;
        return;
    }

    if (self->priv->drag_motion_id)
        g_signal_handler_disconnect (drag, priv->drag_motion_id);
    priv->drag_motion_id = 0;

    const char *desktop_id;
    desktop_id = g_object_get_data ((GObject *)drag, "desktop-id");
    if (!desktop_id)
        return;

    LauncherData *data = g_new0 (LauncherData, 1);
    data->desktop_id = g_strdup (desktop_id);
    data->new_pos = priv->drag_position;
    /*
    g_timeout_add (50, (GSourceFunc)change_panel_launchers, data);
    */

    GTask *task = g_task_new (NULL, NULL, NULL, NULL);
    g_task_set_task_data (task, data, NULL);
    g_task_run_in_thread (task, (GTaskThreadFunc)change_panel_launchers);
    g_object_unref (task);
    drag_drop_successfull = TRUE;
}

static void
on_drop_cancel (CdosDropAction    *action,
                ClutterActor      *actor,
                gfloat             event_x,
                gfloat             event_y,
                AppletAppList  *self)
{
    drag_drop_successfull = FALSE;
}

static void
on_monitors_changed (MetaMonitorManager *monitor_manager, AppletAppList *self)
{
    applet_app_list_condense_app_groups (self, FALSE);
}

static void
applet_app_list_update_pinned_favs (AppletAppList *self)
{
    AppletAppListPrivate *priv = self->priv;
    priv->show_pinned_favs = FALSE;
    if (!priv->group_app_buttons)
    {
        applet_pinned_favs_reload (priv->pinned_favs);

        if (clutter_actor_is_visible ((ClutterActor *)priv->pinned_favs))
            priv->show_pinned_favs = TRUE;
    }
    else
    {
        applet_pinned_favs_unload (priv->pinned_favs);
        clutter_actor_hide ((ClutterActor *)priv->pinned_favs);
    }
}

static StAdjustment *
scroll_view_get_adjustment (AppletAppList *self)
{
    AppletAppListPrivate *priv = self->priv;
    ClutterActor *scroll_bar;
    StAdjustment *adjustment;

    scroll_bar = st_scroll_view_get_hscroll_bar (ST_SCROLL_VIEW (priv->scroll_view));
    adjustment = st_scroll_bar_get_adjustment ((StScrollBar *)scroll_bar);

    return adjustment;
}

static void
applet_app_list_scroll_view_update_page_size (AppletAppList *self)
{
    AppletAppListPrivate *priv = self->priv;
    StAdjustment *adjustment;

    gfloat sv_w;

    adjustment = scroll_view_get_adjustment (self);
    clutter_actor_get_transformed_size (CLUTTER_ACTOR (priv->scroll_view), &sv_w, NULL);

    g_object_set (adjustment, "page-size", sv_w, "page-increment", sv_w, NULL);
    applet_app_list_focus_app (self);
    applet_app_list_update_buttons_state (self);
}

static void
scroll_view_on_upper_changed (GObject *adjustment, GParamSpec *pspec, gpointer user_data)
{
    AppletAppList *self = APPLET_APP_LIST (user_data);
    // AppletAppListPrivate *priv = self->priv;

    gdouble page_size, upper;
    guint page_count;

    g_object_get (ST_ADJUSTMENT (adjustment), "page-size", &page_size, "upper", &upper, NULL);

    page_count = (uint)ceilf (upper / page_size);
    applet_app_list_set_page_count (self, page_count);
}

static void
applet_app_list_update_buttons_state (AppletAppList *self)
{
    // AppletAppListPrivate *priv = self->priv;
    StAdjustment *adjustment;
    gdouble lower, upper, value, page_size;

    adjustment = scroll_view_get_adjustment (self);
    g_object_get (ST_ADJUSTMENT (adjustment), "lower", &lower, "page-size", &page_size, "upper", &upper, NULL);

    value = st_adjustment_get_value (adjustment);
    gboolean prev_state = value == lower ? FALSE : TRUE;
    gboolean next_state = value + page_size == upper ? FALSE : TRUE;

    applet_app_list_button_prev_set_state (self, prev_state);
    applet_app_list_button_next_set_state (self, next_state);
}

static void
scroll_view_on_value_changed (GObject *adjustment, GParamSpec *pspec, gpointer user_data)
{
    AppletAppList *self = APPLET_APP_LIST (user_data);
    applet_app_list_update_buttons_state (self);
}

static void
applet_app_list_scroll_view_setup (AppletAppList *self)
{
    // AppletAppListPrivate *priv = self->priv;
    StAdjustment *adjustment;

    adjustment = scroll_view_get_adjustment (self);

    g_signal_connect (adjustment, "notify::upper", G_CALLBACK (scroll_view_on_upper_changed), self);
    g_signal_connect (adjustment, "notify::value", G_CALLBACK (scroll_view_on_value_changed), self);
}

static void
applet_app_list_scroll (AppletAppList *self, PaginationAction action, gboolean use_page_increment)
{
    // AppletAppListPrivate *priv = self->priv;
    StAdjustment *adjustment;
    gdouble step_increment, page_increment, value, new_value, delta, increment;

    adjustment = scroll_view_get_adjustment (self);
    g_object_get (
        adjustment, "step-increment", &step_increment, "page-increment", &page_increment, "value", &value, NULL);
    increment = use_page_increment == TRUE ? page_increment : step_increment;
    delta = action == PAGE_PREV ? -increment : increment;

    new_value = value + delta >= 0 ? (value + delta) : 0;
    st_adjustment_set_value (adjustment, new_value);
}

static void
app_list_on_allocation_changed (ClutterActor *actor, ClutterActorBox *box, ClutterAllocationFlags flags)
{
    return;
    AppletAppList *self = APPLET_APP_LIST (actor);
    // AppletAppListPrivate *priv = self->priv;

    applet_app_list_scroll_view_update_page_size (self);
}

static void
settings_on_changed (GSettings *settings, gchar *key, gpointer user_data)
{
    AppletAppList *self = APPLET_APP_LIST (user_data);
    AppletAppListPrivate *priv = self->priv;

    gboolean group_apps = g_settings_get_boolean (settings, CDOS_WINDOW_LIST_GROUP_APPS);
    priv->group_app_buttons = group_apps;
    applet_app_list_update_pinned_favs (self);
    applet_app_list_refresh_list (self);
}

static gboolean
actor_on_button_release (ClutterActor *actor, ClutterEvent *event, gpointer user_data)
{
    /*
    ClutterActor *source = clutter_event_get_source (event);
    if (source == actor)
    {
        CdosPanel *panel = cdos_monkey_plugin_get_panel (monkey_plugin_get ());
        cdos_panel_on_button_press (CLUTTER_ACTOR (panel), event, actor);
    }
    */
    return TRUE;
}

static void
applet_app_list_app_group_focused (AppletAppList *self, AppletAppGroup *app_group)
{
    AppletAppListPrivate *priv = self->priv;
    if (!app_group)
    {
        return;
    }

    if (priv->page_count > 1)
    {
        gfloat group_x1, group_x2;
        ClutterActorBox box;
        StAdjustment *adjustment;
        StWidget *group_actor = ST_WIDGET (app_group);
        clutter_actor_get_allocation_box (CLUTTER_ACTOR (group_actor), &box);

        group_x1 = box.x1;
        group_x2 = box.x2;

        gdouble upper, value, new_value = -1.0, page_size;
        adjustment = scroll_view_get_adjustment (self);
        g_object_get (ST_ADJUSTMENT (adjustment), "value", &value, "page-size", &page_size, "upper", &upper, NULL);

        if (group_x1 < value)
        {
            new_value = group_x1;
        }
        else if (group_x2 > value + page_size)
        {
            gdouble delta = (group_x2 - group_x1) - (value + page_size - group_x1);
            new_value = value + delta;
        }
        if (new_value >= 0)
        {
            st_adjustment_set_value (adjustment, (double)new_value);
        }
    }
}

static void
applet_app_list_focus_app (AppletAppList *self)
{
    AppletAppListPrivate *priv = self->priv;
    AppletAppGroup *app_group = NULL;

    CdosApp *app = cdos_window_tracker_get_focus_app (priv->tracker);
    if (app)
    {
        app_group = g_tree_lookup (priv->app_to_app_group, app);
        applet_app_list_app_group_focused (self, app_group);
    }
}

static void
tracker_on_notify_focus_app (GObject *tracker, GParamSpec *pspec, gpointer user_data)
{
    AppletAppList *self = APPLET_APP_LIST (user_data);
    applet_app_list_focus_app (self);
}

static void
applet_app_list_on_switch_workspace (CdosBaseWM *cdos_base_wm,
                                     gint previous_workspace_index,
                                     gint current_workspace_index,
                                     MetaMotionDirection direction,
                                     AppletAppList *self)
{
    AppletAppListPrivate *priv = self->priv;

    if (previous_workspace_index != current_workspace_index)
    {
        if (priv->meta_workspace)
            g_signal_handlers_disconnect_by_data (priv->meta_workspace, self);
        MetaWorkspace *meta_workspace = meta_workspace_manager_get_workspace_by_index (priv->workspace_manager, current_workspace_index);
        priv->meta_workspace = meta_workspace;

        applet_app_list_set_signals (self);
        applet_app_list_refresh_list (self);
    }
}

static void
applet_app_list_on_map (ClutterActor *actor)
{
    CLUTTER_ACTOR_CLASS (applet_app_list_parent_class)->map (actor);

    AppletAppList *self = APPLET_APP_LIST (actor);
    if (self->priv->need_refresh)
        applet_app_list_refresh_list (self);
}

static void
applet_app_list_allocate (ClutterActor *actor,
                          const ClutterActorBox *box,
                          ClutterAllocationFlags flags)
{
    gfloat avail_width;

    gfloat prev_min_width, prev_nat_width;
    gfloat next_min_width, next_nat_width;
    gfloat sv_min_width, sv_nat_width; /* StScrollView */
    gfloat favs_min_width = 0, favs_nat_width = 0;
    gfloat x;

    gboolean show_buttons = FALSE;

    AppletAppList *self = APPLET_APP_LIST (actor);
    AppletAppListPrivate *priv = self->priv;

    StThemeNode *theme_node = st_widget_get_theme_node ((StWidget *)actor);

    clutter_actor_set_allocation (actor, box, flags);

    ClutterActorBox content_box, child_box = {0};
    st_theme_node_get_content_box (theme_node, box, &content_box);

    avail_width = content_box.x2 - content_box.x1;

    if (priv->show_pinned_favs)
        clutter_actor_get_preferred_width ((ClutterActor *)priv->pinned_favs, -1, &favs_min_width, &favs_nat_width);
    clutter_actor_get_preferred_width ((ClutterActor *)priv->actor, -1, &sv_min_width, &sv_nat_width);

    show_buttons = (favs_nat_width + sv_min_width) > priv->max_width;

    if (!show_buttons)
    {
        clutter_actor_hide (priv->button_prev);
        clutter_actor_hide (priv->button_next);

        child_box.x2 = content_box.x1;
        if (priv->show_pinned_favs)
        {
            child_box.x1 = content_box.x1;
            child_box.x2 = child_box.x1 + favs_nat_width;
            child_box.y1 = content_box.y1;
            child_box.y2 = content_box.y2;
            clutter_actor_allocate ((ClutterActor *)priv->pinned_favs, &child_box, flags);
        }
        avail_width -= favs_nat_width;
        sv_nat_width = MIN(sv_nat_width, avail_width);
        child_box.x1 = child_box.x2;
        child_box.x2 = child_box.x1 + sv_nat_width;
        child_box.y1 = content_box.y1;
        child_box.y2 = content_box.y2;
        clutter_actor_allocate ((ClutterActor *)priv->scroll_view, &child_box, flags);
    }
    else
    {
        clutter_actor_show (priv->button_prev);
        clutter_actor_show (priv->button_next);
        clutter_actor_get_preferred_width (priv->button_prev, -1, &prev_min_width, &prev_nat_width);
        clutter_actor_get_preferred_width (priv->button_next, -1, &next_min_width, &next_nat_width);

        avail_width = avail_width - prev_nat_width - next_nat_width;
        x = content_box.x1;
        if (priv->show_pinned_favs)
        {
            favs_nat_width = MIN(favs_nat_width, avail_width);
            child_box.x1 = content_box.x1;
            child_box.x2 = child_box.x1 + favs_nat_width;
            x = child_box.x2;
            child_box.y1 = content_box.y1;
            child_box.y2 = content_box.y2;
            clutter_actor_allocate ((ClutterActor *)priv->pinned_favs, &child_box, flags);
        }
        child_box.x1 = x;
        child_box.x2 = child_box.x1 + prev_nat_width;
        child_box.y1 = content_box.y1;
        child_box.y2 = content_box.y2;
        clutter_actor_allocate (priv->button_prev, &child_box, flags);

        x = child_box.x2;

        child_box.x2 = content_box.x2;
        child_box.x1 = child_box.x2 - next_nat_width;
        child_box.y1 = content_box.y1;
        child_box.y2 = content_box.y2;
        clutter_actor_allocate (priv->button_next, &child_box, flags);

        avail_width -= favs_nat_width;
        sv_nat_width = MIN(sv_nat_width, avail_width);
        child_box.x1 = x;
        child_box.x2 = content_box.x2 - next_nat_width;
        child_box.y1 = content_box.y1;
        child_box.y2 = content_box.y2;
        clutter_actor_allocate ((ClutterActor *)priv->scroll_view, &child_box, flags);
    }
}

static void
applet_app_list_get_preferred_width (ClutterActor *actor,
                                     gfloat for_height,
                                     gfloat *min_width_p,
                                     gfloat *natural_width_p)
{
    AppletAppList *self = APPLET_APP_LIST (actor);
    AppletAppListPrivate *priv = self->priv;

    StThemeNode *theme_node = st_widget_get_theme_node ((StWidget *)actor);
    st_theme_node_adjust_for_height (theme_node, &for_height);

    gfloat sv_min_width, sv_nat_width; /* StScrollView */
    gfloat favs_min_width = 0, favs_nat_width = 0;

    if (priv->show_pinned_favs)
        clutter_actor_get_preferred_width ((ClutterActor *)priv->pinned_favs, for_height, &favs_min_width, &favs_nat_width);
    clutter_actor_get_preferred_width ((ClutterActor *)priv->actor, for_height, &sv_min_width, &sv_nat_width);

    if (min_width_p)
        *min_width_p = favs_min_width + 100; /* 暂时设为100 */
    if (natural_width_p)
        *natural_width_p = favs_nat_width + sv_nat_width;

    st_theme_node_adjust_preferred_width (theme_node, min_width_p, natural_width_p);
}

static void
applet_app_list_get_preferred_height (ClutterActor *actor,
                                      gfloat for_width,
                                      gfloat *min_height_p,
                                      gfloat *natural_height_p)
{
    AppletAppList *self = APPLET_APP_LIST (actor);
    AppletAppListPrivate *priv = self->priv;

    StThemeNode *theme_node = st_widget_get_theme_node ((StWidget *)actor);
    st_theme_node_adjust_for_width (theme_node, &for_width);

    gfloat sv_min_size, sv_nat_size; /* StScrollView */
    gfloat favs_min_size = 0, favs_nat_size = 0;

    if (priv->show_pinned_favs)
        clutter_actor_get_preferred_height ((ClutterActor *)priv->pinned_favs, -1, &favs_min_size, &favs_nat_size);
    clutter_actor_get_preferred_height ((ClutterActor *)priv->actor, -1, &sv_min_size, &sv_nat_size);

    if (min_height_p)
        *min_height_p = MAX(favs_min_size, sv_min_size);
    if (natural_height_p)
        *natural_height_p = MAX(favs_nat_size, sv_nat_size);

    st_theme_node_adjust_preferred_height (theme_node, min_height_p, natural_height_p);
}

static gboolean
launchers_monitor_update_delay (gpointer user_data)
{
    AppletAppList *self = APPLET_APP_LIST (user_data);
    AppletAppListPrivate *priv = self->priv;
    applet_app_list_update_pinned_favs (self);
    applet_app_list_refresh_list (self);
    priv->launchers_monitor_id = 0;
    return FALSE;
}

static void
launchers_dir_on_changed (GFileMonitor     *monitor,
                          GFile            *file,
                          GFile            *other_file,
                          GFileMonitorEvent event_type,
                          gpointer          user_data)
{
    AppletAppList *self = APPLET_APP_LIST (user_data);
    AppletAppListPrivate *priv = self->priv;
    if (event_type == G_FILE_MONITOR_EVENT_ATTRIBUTE_CHANGED)
    {
        if (priv->launchers_monitor_id > 0)
        {
            g_source_remove (priv->launchers_monitor_id);
        }
        priv->launchers_monitor_id = g_timeout_add (250, launchers_monitor_update_delay, self);
    }
}

static void
applet_app_list_init (AppletAppList *self)
{
    AppletAppListPrivate *priv;
    self->priv = applet_app_list_get_instance_private (self);
    priv = self->priv;
    gchar *launchers_dir;
    GFile *launchers_file;

    priv->settings = applet_window_list_get_settings ();
    priv->global_settings = cdos_get_cdos_settings ();

    launchers_dir = applet_panel_launchers_get_dirname ();
    launchers_file = g_file_new_for_path (launchers_dir);
    applet_panel_launchers_remove_invalid_files ();
    applet_panel_launchers_update_files ();
    priv->launchers_monitor = g_file_monitor (launchers_file, G_FILE_MONITOR_NONE, NULL, NULL);
    if (priv->launchers_monitor)
    {
        priv->launchers_monitor_id = 0;
        g_signal_connect (priv->launchers_monitor, 
                          "changed", 
                          G_CALLBACK (launchers_dir_on_changed), 
                          self);
    }
    g_object_unref (launchers_file);
    g_free (launchers_dir);

    priv->group_app_buttons = g_settings_get_boolean (priv->settings, CDOS_WINDOW_LIST_GROUP_APPS);
    g_signal_connect (priv->settings, "changed::group-apps", G_CALLBACK (settings_on_changed), self);

    priv->pinned_favs = applet_pinned_favs_new ();
    _add_actor (self, priv->pinned_favs);
    applet_app_list_update_pinned_favs (self);

    priv->scroll_view = g_object_new (ST_TYPE_SCROLL_VIEW,
                                      "style-class", "vfade",
                                      "x-expand", TRUE,
                                      "x-align", CLUTTER_ACTOR_ALIGN_START,
                                      NULL);
    st_scroll_view_set_policy (ST_SCROLL_VIEW (priv->scroll_view), GTK_POLICY_AUTOMATIC, GTK_POLICY_NEVER);
    ClutterActor *hscroll_bar = st_scroll_view_get_hscroll_bar (priv->scroll_view);
    clutter_actor_hide (hscroll_bar);
    st_scroll_view_set_auto_scrolling (ST_SCROLL_VIEW (priv->scroll_view), TRUE);
    clutter_actor_add_child (CLUTTER_ACTOR (self), CLUTTER_ACTOR (priv->scroll_view));
    applet_app_list_scroll_view_setup (self);

    priv->actor = (ClutterActor *)st_box_layout_new ();
    clutter_actor_set_reactive (priv->actor, TRUE);
    st_widget_add_style_class_name ((StWidget *)priv->actor, "window-list-box");
    st_scroll_view_add_child (priv->scroll_view, priv->actor);
    g_signal_connect (priv->actor, "button-release-event", G_CALLBACK (actor_on_button_release), self);

    ClutterAction *action = cdos_drop_action_new ((ClutterActor *)self);
    priv->drop_action = action;
    g_signal_connect (action, "over-in", G_CALLBACK (on_drop_over_in), self);
    g_signal_connect (action, "over-out", G_CALLBACK (on_drop_over_out), self);
    g_signal_connect (action, "drop", G_CALLBACK (on_drop_drop), self);
    g_signal_connect (action, "drop-cancel", G_CALLBACK (on_drop_cancel), self);

    priv->app_to_app_group = g_tree_new ((GCompareFunc)applet_tree_key_compare);
    priv->window_to_app = g_tree_new ((GCompareFunc)applet_tree_key_compare);
    priv->tracker = cdos_window_tracker_get_default ();

    priv->button_prev = (ClutterActor *)st_button_new ();
    st_widget_set_style_class_name ((StWidget *)priv->button_prev, "window-list-button-prev");
    clutter_actor_set_reactive (CLUTTER_ACTOR (priv->button_prev), TRUE);

    priv->button_next = (ClutterActor *)st_button_new ();
    st_widget_set_style_class_name ((StWidget *)priv->button_next, "window-list-button-next");
    clutter_actor_set_reactive (priv->button_next, TRUE);

    clutter_actor_insert_child_at_index (CLUTTER_ACTOR (self), priv->button_prev, 0);
    clutter_actor_add_child (CLUTTER_ACTOR (self), priv->button_next);

    applet_app_list_pagination_setup (self);

    priv->app_system = cdos_app_system_get_default ();
    g_signal_connect (priv->app_system, "installed-changed", G_CALLBACK (system_on_installed_changed), self);
    //为了兼顾开始菜单“最常使用”和“最近添加”功能，将原来“installed-changed”信号拆分成和“installed-changed”和“removed-changed”
    g_signal_connect (priv->app_system, "removed-changed", G_CALLBACK (system_on_installed_changed), self);
    priv->display = cdos_global_get_display (global);
    priv->workspace_manager = meta_display_get_workspace_manager (priv->display);
    priv->monitor_manager = meta_monitor_manager_get ();
    g_signal_connect (priv->monitor_manager, "monitors-changed", G_CALLBACK (on_monitors_changed), self);
    g_signal_connect_swapped (priv->scroll_view, "allocation-changed", G_CALLBACK (app_list_on_allocation_changed), self);
    // we hope when the app window is focused, make the app group visible;
    g_signal_connect (priv->tracker, "notify::focus-app", G_CALLBACK (tracker_on_notify_focus_app), self);

    CdosBaseWM *global_wm;
    global_wm = cdos_global_get_window_manager (global);
    priv->workspace_changed_id = g_signal_connect (global_wm, "switch-workspace",
                                    G_CALLBACK (applet_app_list_on_switch_workspace), self);
    gint active_workspace_index = meta_workspace_manager_get_active_workspace_index (priv->workspace_manager);
    applet_app_list_on_switch_workspace (global_wm, -1, active_workspace_index, META_MOTION_RIGHT, self);
}

static gboolean
disconnect_window_signals (MetaWindow *window, CdosApp *app, AppletAppList *self)
{
    g_signal_handlers_disconnect_by_data (window, self);
    return FALSE;
}

static void
applet_app_list_dispose (GObject *object)
{
    AppletAppList *self = APPLET_APP_LIST (object);
    AppletAppListPrivate *priv = self->priv;

    if (priv->meta_workspace)
    {
        g_signal_handlers_disconnect_by_data (priv->meta_workspace, self);
        priv->meta_workspace = NULL;
    }
    if (priv->settings)
    {
        g_signal_handlers_disconnect_by_data (priv->settings, self);
        priv->settings = NULL;
    }
    if (priv->global_settings)
    {
        g_signal_handlers_disconnect_by_data (priv->global_settings, self);
        priv->global_settings = NULL;
    }
    if (priv->display)
    {
        g_signal_handlers_disconnect_by_data (priv->display, self);
        priv->display = NULL;
    }
    if (priv->monitor_manager)
    {
        g_signal_handlers_disconnect_by_data (priv->monitor_manager, self);
        priv->monitor_manager = NULL;
    }
    if (priv->app_system)
    {
        g_signal_handlers_disconnect_by_data (priv->app_system, self);
        priv->app_system = NULL;
    }
    if (priv->tracker)
    {
        g_signal_handlers_disconnect_by_data (priv->tracker, self);
        priv->tracker = NULL;
    }
    if (priv->launchers_monitor_id > 0)
    {
        g_source_remove (priv->launchers_monitor_id);
        priv->launchers_monitor_id = 0;
    }
    if (priv->launchers_monitor)
    {
        g_object_unref (priv->launchers_monitor);
        priv->launchers_monitor = NULL;
    }

    g_clear_pointer (&priv->app_to_app_group, g_tree_destroy);

    if (priv->window_to_app)
    {
        g_tree_foreach (priv->window_to_app, (GTraverseFunc)disconnect_window_signals, self);

        g_tree_destroy (priv->window_to_app);
        priv->window_to_app = NULL;
    }

    G_OBJECT_CLASS (applet_app_list_parent_class)->dispose (object);
}

static void
applet_app_list_finalize (GObject *object)
{
    // AppletAppList *self = APPLET_APP_LIST (object);
    // AppletAppListPrivate *priv = self->priv;

    G_OBJECT_CLASS (applet_app_list_parent_class)->finalize (object);
}

static void
applet_app_list_class_init (AppletAppListClass *klass)
{
    GObjectClass *gobject_class = G_OBJECT_CLASS (klass);

    ClutterActorClass *actor_class = CLUTTER_ACTOR_CLASS (klass);
    actor_class->map = applet_app_list_on_map;
    actor_class->allocate = applet_app_list_allocate;
    actor_class->get_preferred_width = applet_app_list_get_preferred_width;
    actor_class->get_preferred_height = applet_app_list_get_preferred_height;

    gobject_class->dispose = applet_app_list_dispose;
    gobject_class->finalize = applet_app_list_finalize;
}

static void
applet_app_list_set_signals (AppletAppList *self)
{
    AppletAppListPrivate *priv = self->priv;

    g_signal_connect_after (priv->meta_workspace, "window-added", G_CALLBACK (applet_app_list_on_window_added), self);
    g_signal_connect_after (priv->meta_workspace, "window-removed", G_CALLBACK (applet_app_list_on_window_removed), self);

    if (priv->global_settings)
    {
        g_signal_connect (priv->global_settings, "changed::panel-launchers", G_CALLBACK (panel_launchers_on_changed), self);
    }
}

static void
applet_app_list_refresh_list (AppletAppList *self)
{
    AppletAppListPrivate *priv = self->priv;
    if (priv->app_to_app_group)
    {
        g_tree_destroy (priv->app_to_app_group);
    }
    if (priv->window_to_app)
    {
        g_tree_foreach (priv->window_to_app, (GTraverseFunc)disconnect_window_signals, self);
        g_tree_destroy (priv->window_to_app);
    }
    priv->app_to_app_group = g_tree_new ((GCompareFunc)applet_tree_key_compare);
    priv->window_to_app = g_tree_new ((GCompareFunc)applet_tree_key_compare);

    clutter_actor_destroy_all_children (self->priv->actor);

    //'placeholder' has been destroyed.
    priv->placeholder = (ClutterActor *)g_object_new (ST_TYPE_WIDGET, NULL);
    st_widget_set_style_class_name ((StWidget *)priv->placeholder, "panel-placeholder");
    _add_actor (priv->actor, priv->placeholder);
    clutter_actor_hide (priv->placeholder);

    applet_app_list_load_favorites (self);
    applet_app_list_refresh_apps (self);
}

static void
on_meta_window_prop_changed (MetaWindow *window,
                             GParamSpec *param,
                             AppletAppList *self)
{
    MetaWorkspace *workspace = meta_window_get_workspace (window);
    applet_app_list_window_removed (self, workspace, window);
    applet_app_list_window_added (self, workspace, window, NULL, FALSE);
}

static void
applet_app_list_window_added (AppletAppList *self,
                              MetaWorkspace *meta_workspace,
                              MetaWindow *meta_window,
                              CdosApp *favapp,
                              gboolean is_favapp)
{
    AppletAppListPrivate *priv = self->priv;
    CdosWindowTracker *tracker = priv->tracker;
    CdosApp *app;
    if (!priv->group_app_buttons && !meta_window && !is_favapp)
    {
        return;
    }
    if (meta_window)
    {
        MetaWindowType meta_window_type = meta_window_get_window_type (meta_window);
        if (meta_window_type != META_WINDOW_NORMAL)
        {
            return;
        }
    }
    if (favapp)
    {
        app = favapp;
    }
    else
    {
        app = cdos_window_tracker_get_window_app (tracker, meta_window);
    }

    if (!app)
        return;
    if (meta_window)
        g_tree_insert (priv->window_to_app, meta_window, app);

    AppletAppGroup *app_group = g_tree_lookup (priv->app_to_app_group, app);
    if (!app_group)
    {
        app_group = applet_app_group_new (app, is_favapp, meta_workspace);
        clutter_actor_add_child (priv->actor, (ClutterActor *)app_group);

        applet_app_group_update_meta_windows (app_group, meta_workspace);
        g_tree_insert (priv->app_to_app_group, app, app_group);

        if (meta_window)
        {
            g_signal_connect (meta_window, "notify::wm-class", G_CALLBACK (on_meta_window_prop_changed), self);
            g_signal_connect (meta_window, "notify::gtk-application-id", G_CALLBACK (on_meta_window_prop_changed), self);
        }
    }
    else
    {
        applet_app_group_update_meta_windows (app_group, meta_workspace);
    }
}

static void
applet_app_list_on_window_added (MetaWorkspace *meta_workspace, MetaWindow *meta_window, AppletAppList *self)
{
    MetaWindowType meta_window_type = meta_window_get_window_type (meta_window);
    if (meta_window_type == META_WINDOW_NORMAL)
    {
        applet_app_list_window_added (self, meta_workspace, meta_window, NULL, FALSE);
    }
}

static void
applet_app_list_remove_app (AppletAppList *self, CdosApp *app)
{
    AppletAppListPrivate *priv = self->priv;
    AppletAppGroup *app_group = g_tree_lookup (priv->app_to_app_group, app);

    if (app_group)
    {
        gboolean was_favorite_app = applet_app_group_get_was_favorite_app (app_group);
        gboolean is_favorite_app = applet_app_group_get_is_favorite_app (app_group);

        if (was_favorite_app || is_favorite_app)
        {
            applet_app_group_is_favorite (app_group, TRUE);
            return;
        }

        ClutterActor *app_group_actor = CLUTTER_ACTOR (app_group);
        clutter_actor_remove_child (priv->actor, app_group_actor);
        g_tree_remove (priv->app_to_app_group, app);

        g_timeout_add (15, (GSourceFunc)applet_app_list_remove_app_callback, self);
    }
}

static void
applet_app_list_load_favorites (AppletAppList *self)
{
    AppletAppListPrivate *priv = self->priv;
    if (!priv->group_app_buttons)
    {
        return;
    }
    gchar **launchers;
    launchers = g_settings_get_strv (priv->global_settings, "panel-launchers");
    CdosAppSystem *app_system = priv->app_system;
    guint i;

    if (launchers != NULL)
    {
        for (i = 0; launchers[i]; i++)
        {
            gchar *launcher = launchers[i];
            CdosApp *app;
            app = cdos_app_system_lookup_app (app_system, launcher);
            if (app)
            {
                applet_app_list_window_added (self, priv->meta_workspace, NULL, app, TRUE);
            }
        }
        g_strfreev (launchers);
    }
}

static void
applet_app_list_window_removed (AppletAppList *self, MetaWorkspace *meta_workspace, MetaWindow *meta_window)
{
    AppletAppListPrivate *priv = self->priv;
    CdosApp *app;
    app = g_tree_lookup (priv->window_to_app, meta_window);
    g_tree_remove (priv->window_to_app, meta_window);
    g_debug ("%s %p %p", __func__, app, meta_window);
    if (!app)
    {
        return;
    }

    g_signal_handlers_disconnect_by_data (meta_window, self);

    GSList *app_windows = cdos_app_get_windows (app);
    gboolean has_windows_on_workspace = FALSE;
    GSList *it = app_windows;
    while (it != NULL)
    {
        MetaWindow *meta_window_tmp = it->data;
        if (meta_window != meta_window_tmp)
        {
            MetaWindowType meta_window_type = meta_window_get_window_type (meta_window_tmp);
            MetaWorkspace *workspace = meta_window_get_workspace (meta_window_tmp);
            if (meta_window_type == META_WINDOW_NORMAL && workspace == meta_workspace)
            {
                has_windows_on_workspace = TRUE;
                break;
            }
        }
        it = it->next;
    }

    AppletAppGroup *app_group = g_tree_lookup (priv->app_to_app_group, app);
    if (app_group)
        applet_app_group_window_removed (app_group, meta_workspace, meta_window);
    if (!has_windows_on_workspace)
    {
        applet_app_list_remove_app (self, app);
    }
    else
    {
        applet_app_group_update_meta_windows (app_group, meta_workspace);
    }
}

static void
applet_app_list_on_window_removed (MetaWorkspace *meta_workspace, MetaWindow *meta_window, AppletAppList *self)
{
    MetaWindowType meta_window_type = meta_window_get_window_type (meta_window);
    if (meta_window_type == META_WINDOW_NORMAL)
    {
        applet_app_list_window_removed (self, meta_workspace, meta_window);
    }
}

static void
panel_launchers_on_changed (GSettings *settings, gchar *key, AppletAppList *self)
{
    AppletAppListPrivate *priv = self->priv;
    if (priv->group_app_buttons)
    {
        applet_panel_launchers_remove_invalid_files ();
        applet_panel_launchers_update_files ();
        applet_app_list_refresh_list (self);
    }
}

static void
refresh_apps_foreach_callback (gpointer meta_window, gpointer user_data)
{
    AppletAppList *self = APPLET_APP_LIST (user_data);
    AppletAppListPrivate *priv = self->priv;
    applet_app_list_window_added (self, priv->meta_workspace, meta_window, NULL, FALSE);
}

static gboolean
applet_app_list_remove_app_callback (AppletAppList *self)
{
    //applet_app_list_refresh_apps (self);
    return FALSE;
}

static void
applet_app_list_pagination_setup (AppletAppList *self)
{
    AppletAppListPrivate *priv = self->priv;
    priv->page_count = 1;

    applet_app_list_button_prev_set_state (self, FALSE);
    applet_app_list_button_next_set_state (self, FALSE);
    applet_app_list_pagination_hide (self);

    g_signal_connect (priv->button_prev, "clicked", G_CALLBACK (button_prev_on_page_up), self);
    g_signal_connect (priv->button_next, "clicked", G_CALLBACK (button_next_on_page_down), self);
    g_signal_connect (priv->scroll_view, "scroll-event", G_CALLBACK (applet_app_list_on_scroll), self);
}

static gboolean
condense_app_group (gpointer key, gpointer value, gpointer user_data)
{
    AppletAppGroup *app_group = APPLET_APP_GROUP (value);
    AppletButtonBox *button_box = applet_app_group_get_window_button_box (app_group);
    GList *buttons = applet_button_box_get_buttons (button_box);
    gboolean condense = applet_button_box_get_condense (button_box);

    if (g_list_length (buttons) > 1 && condense == FALSE)
    {
        g_debug ("%s", __func__);
        applet_app_group_set_buttons_condense (app_group, TRUE);
    }
    return FALSE;
}

static gboolean
scatter_app_group (gpointer key, gpointer value, gpointer user_data)
{
    AppletAppGroup *app_group = APPLET_APP_GROUP (value);
    AppletButtonBox *button_box = applet_app_group_get_window_button_box (app_group);
    GList *buttons = applet_button_box_get_buttons (button_box);
    gboolean condense = applet_button_box_get_condense (button_box);

    if (g_list_length (buttons) > 1 && condense == TRUE)
    {
        applet_app_group_set_buttons_condense (app_group, FALSE);
    }
    return FALSE;
}

static void
applet_app_list_condense_app_groups (AppletAppList *self, gboolean condense)
{
    AppletAppListPrivate *priv = self->priv;
    if (condense)
    {
        g_tree_foreach (priv->app_to_app_group, condense_app_group, self);
    }
    else
    {
        g_tree_foreach (priv->app_to_app_group, scatter_app_group, self);
    }
}

static void
applet_app_list_set_page_count (AppletAppList *self, guint new_count)
{
    AppletAppListPrivate *priv = self->priv;
    if (priv->page_count == new_count)
    {
        return;
    }

    guint old_count = priv->page_count;
    priv->page_count = new_count;
    gboolean button_state = FALSE;
    if (priv->page_count < 1)
    {
        applet_app_list_pagination_hide (self);
    }

    applet_app_list_button_prev_set_state (self, button_state);
    applet_app_list_button_next_set_state (self, button_state);

    if (!priv->group_app_buttons && priv->page_count > old_count)
    {
        applet_app_list_condense_app_groups (self, TRUE);
    }
    applet_app_list_update_buttons_state (self);
}

static void
button_prev_on_page_up (StButton *button, gint clicked_button, AppletAppList *self)
{
    if (clicked_button == 1)
    {
        applet_app_list_scroll (self, PAGE_PREV, TRUE);
    }
}

static void
button_next_on_page_down (StButton *button, gint clicked_button, AppletAppList *self)
{
    if (clicked_button == 1)
    {
        applet_app_list_scroll (self, PAGE_NEXT, TRUE);
    }
}

static void
applet_app_list_pagination_hide (AppletAppList *self)
{
    AppletAppListPrivate *priv = self->priv;
    clutter_actor_hide (priv->button_prev);
    clutter_actor_hide (priv->button_next);
}

static void
applet_app_list_button_prev_set_state (AppletAppList *self, gboolean state)
{
    AppletAppListPrivate *priv = self->priv;
    if (state)
    {
        //激活
        st_widget_remove_style_pseudo_class ((StWidget *)priv->button_prev, "disable");
    }
    else
    { //冻结
        st_widget_add_style_pseudo_class ((StWidget *)priv->button_prev, "disable");
    }
    clutter_actor_set_reactive (priv->button_prev, state);
    st_widget_set_track_hover (ST_WIDGET (priv->button_prev), state);
}

static void
applet_app_list_button_next_set_state (AppletAppList *self, gboolean state)
{
    AppletAppListPrivate *priv = self->priv;
    if (state == TRUE)
    { //激活
        st_widget_remove_style_pseudo_class ((StWidget *)priv->button_next, "disable");
    }
    else
    { //冻结
        st_widget_add_style_pseudo_class ((StWidget *)priv->button_next, "disable");
    }
    clutter_actor_set_reactive (priv->button_next, state);
    st_widget_set_track_hover (ST_WIDGET (priv->button_next), state);
}

static gboolean
applet_app_list_on_scroll (ClutterActor *actor, ClutterEvent *event, AppletAppList *self)
{
    AppletAppListPrivate *priv = self->priv;
    if (priv->page_count == 1)
    {
        return FALSE;
    }
    ClutterScrollDirection direction = clutter_event_get_scroll_direction (event);
    if (direction == CLUTTER_SCROLL_UP)
    {
        applet_app_list_scroll (self, PAGE_PREV, FALSE);
    }
    else if (direction == CLUTTER_SCROLL_DOWN)
    {
        applet_app_list_scroll (self, PAGE_NEXT, FALSE);
    }
    return FALSE;
}

static void
system_on_installed_changed (CdosAppSystem *app_system, CdosApp *app, AppletAppList *self)
{
    AppletAppListPrivate *priv = self->priv;
    GSettings *settings = priv->global_settings;

    gchar **launchers = g_settings_get_strv (settings, "panel-launchers");
    gchar **launchers_new = NULL;
    guint i, j = 0;
    guint l = g_strv_length (launchers);
    launchers_new = g_new0 (gchar *, l + 1);

    for (i = 0; i < l; i++)
    {
        gchar *launcher = launchers[i];
        if (launcher)
        {
            CdosApp *app;
            app = cdos_app_system_lookup_app (app_system, launcher);
            if (app)
            {
                launchers_new[j++] = g_strdup (launcher);
            }
        }
    }
    if (i != j)
    {
        g_settings_set_strv (settings, "panel-launchers", (const gchar *const *)launchers_new);
    }
    g_free (launchers_new);
    g_strfreev (launchers);
}

AppletAppList *
applet_app_list_new (MetaWorkspace *meta_workspace)
{
    AppletAppList *self;
    self = g_object_new (APPLET_TYPE_APP_LIST,
                         "x-align", CLUTTER_ACTOR_ALIGN_START,
                         "style-class", "window-list-app-list",
                         NULL);

    return self;
}

static void
applet_app_list_refresh_apps (AppletAppList *self)
{
    AppletAppListPrivate *priv = self->priv;

    if (!clutter_actor_is_mapped ((ClutterActor *)self))
    {
        //return;
    }

    GList *windows = NULL;
    windows = meta_workspace_list_windows (priv->meta_workspace);
    if (windows)
    {
        g_list_foreach (windows, refresh_apps_foreach_callback, self);
    }
    g_list_free (windows);
}

void
applet_app_list_set_max_width (AppletAppList *self, gfloat max_width)
{
    self->priv->max_width = max_width;
}

static gboolean
set_app_group_orientation (CdosApp *app, AppletAppGroup *app_group, gpointer data)
{
    StSide side = GPOINTER_TO_UINT (data);
    applet_app_group_set_orientation (app_group, side);
    return FALSE;
}

void
applet_app_list_set_orientation (AppletAppList *self, StSide orientation)
{
    g_tree_foreach (self->priv->app_to_app_group, (GTraverseFunc)set_app_group_orientation, GUINT_TO_POINTER(orientation));
}
