#include <string.h>
#include <stdio.h>
#include <gio/gdesktopappinfo.h>
#include <clutter/clutter.h>
#include "applet-panel-launchers-manager.h"
#include "cdos-start.h"
#include "st.h"

//path is: ~/.local/share/cdos/panel-launchers/
#define LAUNCHERS_FILE_DIR "/cdos/panel-launchers/"
#define DEFAULT_DIRECTORY_MODE (0755)

#define FILE_MANAGER_NAME "org.freedesktop.FileManager1"
#define FILE_MANAGER_PATH "/org/freedesktop/FileManager1"
#define FILE_MANAGER_INTERFACE "org.freedesktop.FileManager1"

typedef enum {
    LFO_COPY = 1,
    LFO_REMOVE
} LauncherFileOperation;

static gchar *
build_filename_path (const gchar *app_id)
{
    gchar *dirname, *file_path;
    dirname = applet_panel_launchers_get_dirname ();
    file_path = g_build_filename (dirname, app_id, NULL);
    g_free (dirname);
    return file_path;
}

static GDesktopAppInfo *
create_launcher_app_info (const gchar *app_id)
{
    GDesktopAppInfo *desktop_info;
    gchar *filename;

    filename = build_filename_path (app_id);
    if (g_file_test (filename, G_FILE_TEST_EXISTS))
    {
        desktop_info = g_desktop_app_info_new_from_filename (filename);
    }
    g_free (filename);
    return desktop_info;
}

static gboolean
launchers_file_operation (const gchar *app_id, LauncherFileOperation oper)
{
    gboolean res = FALSE;
    GDesktopAppInfo *app_info;
    const gchar *source_path;
    gchar *dest_path;
    GFile *source_file = NULL, *dest_file = NULL;
    GError *error = NULL;
    gboolean dest_exist;

    dest_path = build_filename_path (app_id);
    dest_file = g_file_new_for_path (dest_path);
    app_info = g_desktop_app_info_new (app_id);
    dest_exist = g_file_test (dest_path, G_FILE_TEST_EXISTS);

    if (oper == LFO_REMOVE)
    {
        if (dest_exist)
            res = g_file_trash (dest_file, NULL, &error);
    }
    else if (oper == LFO_COPY) 
    {
        if (app_info && !dest_exist)
        {
            source_path = g_desktop_app_info_get_filename (app_info);
            source_file = g_file_new_for_path (source_path);
            res = g_file_copy (source_file, dest_file, G_FILE_COPY_OVERWRITE, NULL, NULL, NULL, &error);
            if (res)
                g_chmod (dest_path, DEFAULT_DIRECTORY_MODE);
            g_object_unref (dest_file);
        }
    }

out:
    if (error)
    {
        g_warning ("launcher oper failed:%s", error->message);
        g_error_free (error);
    }

    if (app_info)
        g_object_unref (app_info);
    if (source_file)
        g_object_unref (source_file);
    g_object_unref (dest_file);
    g_free (dest_path);

    return res;
}

static gboolean 
desktop_file_is_panel_launcher (gchar **launchers, const gchar *file_name)
{
    gint i;
    for (i = 0; launchers[i] != NULL; i++)
    {
        if (g_strcmp0 (launchers[i], file_name) == 0)
        {
            return TRUE;
        }
    }
    return FALSE;
}

gchar *
applet_panel_launchers_get_dirname (void)
{
    gchar *dirname;
    dirname = g_build_filename (g_get_user_data_dir (), LAUNCHERS_FILE_DIR, NULL);
    if (g_file_test (dirname, G_FILE_TEST_EXISTS) == FALSE)
    {
        g_mkdir_with_parents (dirname, DEFAULT_DIRECTORY_MODE);
    }
    return dirname;
}

void
applet_panel_launchers_remove_invalid_files (void)
{
    GFile *dir = NULL;
    gchar *dirname;
    const gchar *child_name;
    GFileEnumerator *enumer = NULL;
    GSettings   *settings;
    gchar **launchers = NULL;

    dirname = applet_panel_launchers_get_dirname ();
    settings = cdos_get_cdos_settings ();
    if (!settings)
    {
        goto out;
    }
    launchers = g_settings_get_strv (settings, "panel-launchers");
    if (launchers == NULL)
    {
        goto out;
    }

    dir = g_file_new_for_path (dirname);
    enumer = g_file_enumerate_children (dir, "standard::*", G_FILE_QUERY_INFO_NOFOLLOW_SYMLINKS, NULL, NULL);
    if (enumer)
    {
        while (TRUE)
        {
            GFileInfo *child_info;
            if (!g_file_enumerator_iterate (enumer, &child_info, NULL, NULL, NULL))
                goto out;
            if (!child_info)
                break;
            child_name = g_file_info_get_name (child_info);
            if (desktop_file_is_panel_launcher (launchers, child_name) == FALSE)
            {
                launchers_file_operation (child_name, LFO_REMOVE);
            }
        }
    }

out:
    g_free (dirname);
    if (dir)
        g_object_unref (dir);
    if (enumer)
        g_object_unref (enumer);
    if (launchers)
        g_strfreev (launchers);
}    

void
applet_panel_launchers_update_files (void)
{
    gint i;
    gchar **launchers = NULL;
    GSettings *settings = NULL;

    settings = cdos_get_cdos_settings ();
    if (settings)
    {
        launchers = g_settings_get_strv (settings, "panel-launchers");
        if (launchers)
        {
            for (i = 0; launchers[i] != NULL; i++)
            {
                launchers_file_operation (launchers[i], LFO_COPY);
            }
            g_strfreev (launchers);
        }
    }
}

ClutterActor *
applet_panel_launchers_create_icon_texture (CdosApp *app, int size)
{
    g_return_val_if_fail (CDOS_IS_APP (app), NULL);
    GDesktopAppInfo *desktop_info;
    GIcon *icon = NULL;
    ClutterActor *ret = NULL;
    const gchar *app_id;

    app_id = cdos_app_get_id (app);
    
    desktop_info = create_launcher_app_info (app_id);

    ret = st_icon_new ();
    st_icon_set_icon_size (ST_ICON (ret), size);
    st_icon_set_fallback_icon_name (ST_ICON (ret), "application-x-executable");
    icon = g_app_info_get_icon ((GAppInfo *)desktop_info);
    if (icon != NULL)
        st_icon_set_gicon (ST_ICON (ret), icon);

    if (desktop_info)
        g_object_unref (desktop_info);

    return ret;
}

//return value should be free;
gchar *
applet_panel_launchers_get_display_name (CdosApp *app)
{
    g_return_val_if_fail (CDOS_IS_APP (app), NULL);
    GDesktopAppInfo *desktop_info;
    const gchar *display_name = NULL;
    const gchar *app_id;
    gchar *name = NULL;

    app_id = cdos_app_get_id (app);
    desktop_info = create_launcher_app_info (app_id);
    if (desktop_info)
    {
        display_name = g_app_info_get_display_name ((GAppInfo *)desktop_info);
        if (display_name)
            name = g_strdup (display_name);
        g_object_unref (desktop_info);
    }

    return name;
}

gchar *
applet_panel_launchers_get_uri (const gchar *app_id)
{
    gchar *path, *uri = NULL;
    path = build_filename_path (app_id);
    if (g_file_test (path, G_FILE_TEST_EXISTS))
    {
        uri = g_filename_to_uri (path, NULL, NULL);
    }
    g_free (path);
    return uri;
}

void
applet_panel_launchers_edit_properties (const gchar *app_id)
{
    GDBusProxy *nemo_proxy = NULL;
    gchar    *uri;
    GVariant *var, *param;
    GError   *error = NULL;
    gchar **uris = NULL;

    nemo_proxy = g_dbus_proxy_new_for_bus_sync (G_BUS_TYPE_SESSION, 
                                                G_DBUS_PROXY_FLAGS_NONE,
                                                NULL,
                                                FILE_MANAGER_NAME,
                                                FILE_MANAGER_PATH,
                                                FILE_MANAGER_INTERFACE,
                                                NULL, &error);
    if (!nemo_proxy)
    {
        g_message ("Connect nemo dbus failed: %s.", error->message);
        g_error_free (error);
        goto out;
    }

    uri = applet_panel_launchers_get_uri (app_id);
    if (!uri)
    {
        goto out;
    }

    uris = (gchar *)malloc (2 * sizeof (gchar *));
    uris[0] = uri;
    uris[1] = NULL;

    param = g_variant_new ("(^ass)", uris, "panel-launchers");
    var = g_dbus_proxy_call_sync (nemo_proxy, 
                                  "ShowItemProperties", 
                                  param, 
                                  G_DBUS_CALL_FLAGS_NONE, 
                                  -1, 
                                  NULL, 
                                  &error);
    if (error)
    {
        g_warning ("call nemo dbus method ShowItemProperties failed:%s.", error->message);
        g_error_free (error);
    }

out:
    if (uris)
        g_strfreev (uris);
    if (nemo_proxy)
        g_object_unref (nemo_proxy);
    if (var)
        g_variant_unref(var);
}
