/**
 * Copyright (c) 2016-2018 CPU and Fundamental Software Research Center, CAS
 * * Based on code from cinnamon-app-usage.c in Cinnamon.
 * Copyright (c) 1997-2003 by The XFree86 Project, Inc.
 * This software is published by the license of CPU-OS License, you can use and
 * distibuted this software under this License. See CPU-OS Liense for more detail.
 *
 * You should have receive a copy of CPU-OS License. If not, please contact us
 * by email <support_os@cpu-os.ac.cn>
 *
 **/
/* cdos-app-usage.c */

#include "cdos-app-usage.h"
#include <libxml/parser.h>
#include <libxml/xpath.h>
#include <sys/stat.h>


// 参数为文件名、编码类型和枚举变量(下面一个或多个的组合）
// XML_PARSE_RECOVER = 1<<0, recover on errors 尝试修复错误语法
// XML_PARSE_NOERROR = 1<<5, suppress error reports 不输出错误日志
// XML_PARSE_NOWARNING = 1<<6, suppress warning reports 不输出警告日志
// XML_PARSE_PEDANTIC = 1<<7, pedantic error reporting 输出详细的错误日志
// XML_PARSE_NOERROR和XML_PARSE_NOWARNING的功能是“当运行命令行程序时不输出错误/警告日志”。
const int usage_options = XML_PARSE_PEDANTIC | XML_PARSE_RECOVER;
const gchar *usage_encoding = "UTF-8";

const int FOCUS_TIME_MIN_SECONDS = 7; // Need 7 continuous seconds of focus
const int USAGE_CLEAN_DAYS = 7; // If after 7 days we haven't seen an app, purge it.
const gchar *PRIVACY_SCHEMA = "org.cdos.desktop.privacy";
const gchar *ENABLE_MONITORING_KEY = "remember-app-usage";

/*
 * 排序算法：
 * 每当有任一 APP 的 score 达到 SCORE_MAX 时，将所有 score 全部除以 2;
 * 每一个 SAVE_APPS_TIMEOUT 秒都提高一个单元的分数。当一个新的应用被密集使用的时候，
 * 这个机制允许列表跟新较快。
 * 为了保持列表的整洁，一个星期没都没有看到的应用程序，或者 score 小于 SCORE_MIN 都会被移除。
 */

// 多少秒保存一次结果
const int SAVE_APPS_TIMEOUT = 300;

// 试用上限50达到小时，认为被经常使用
#define SCORE_MAX (3600 * 50 / FOCUS_TIME_MIN_SECONDS)

// 保留数据的下限
#define SCORE_MIN (SCORE_MAX >> 3)

struct _CdosAppUsagePrivate
{
    GList *previously_running;
    CdosApp *watched_app;
    glong watch_start_time;
    xmlDocPtr doc;
    xmlNodePtr root_node;
};

G_DEFINE_TYPE_WITH_PRIVATE (CdosAppUsage, cdos_app_usage, G_TYPE_OBJECT);

enum
{
    USAGE_CHANGED,
    LAST_SIGNAL
};

static guint usage_signals[LAST_SIGNAL] = {
    0,
};

static void cdos_app_usage_finalize (GObject *obj);

static void remove_app_usage_for_xml (CdosAppSystem *appsys, const gchar *desktop_id, gpointer user_data);

static gchar *time_to_str (glong time_val);

static xmlDocPtr get_doc (CdosAppUsage *self);

static xmlNodePtr get_root (CdosAppUsage *self);

static xmlNodePtr get_node (CdosAppUsage *self, xmlChar *prop_name, xmlChar *prop_val);

static xmlNodePtr new_app (CdosAppUsage *self, gchar *desktop_id, gchar *score, gchar *last_seen);

static gboolean set_score (CdosAppUsage *self, gchar *desktop_id, gdouble score_val);

static gboolean set_last_seen (CdosAppUsage *self, gchar *desktop_id, gchar *last_seen);

static void ensure_list_saved (CdosAppUsage *self);

static gboolean interval_save_usage_for_xml (gpointer data);

static UsageData *get_usage_for_app (CdosAppUsage *self, CdosApp *app);

static void on_focus_app_changed (CdosWindowTracker *tracker, GParamSpec *spec, CdosAppUsage *self);

static void increment_useage_for_app (CdosAppUsage *self, CdosApp *watched_app);

static void increment_useage_for_app_at_time (CdosAppUsage *self, CdosApp *watched_app, glong time);

static void normalize_usage (CdosAppUsage *self);

static void update_enable_monitoring (CdosAppUsage *self);

static gboolean create_empty_file (const gchar *file_name);

static void cdos_app_usage_load_data (CdosAppUsage *self);


// ===== Private Method =====
static gint
compare_func (gconstpointer a, gconstpointer b)
{
    UsageData *usg_a = (UsageData *)a;
    UsageData *usg_b = (UsageData *)b;
    if (usg_a->score <= usg_b->score)
    {
        return 1;
    }
    return -1;
}

static GList *
sort_usage_struct_list (CdosAppUsage *self, GList *previously_running)
{
    previously_running = g_list_sort (previously_running, compare_func);
    return previously_running;
}

static guint64
get_time (void)
{
    return g_get_real_time () / 1000000;
}

static gchar *
time_to_str (glong time_val)
{
    return g_strdup_printf ("%ld", time_val);
}

static gboolean
create_empty_file (const gchar *file_name)
{
    g_debug ("Cannot find usage-app.xml, create one. %s\n", file_name);
    xmlDocPtr doc = xmlNewDoc (BAD_CAST "1.0");
    xmlNodePtr root_node = xmlNewNode (NULL, BAD_CAST "data");
    xmlDocSetRootElement (doc, root_node);

    gchar *file_path = g_strdup (file_name);
    gchar *last = g_strrstr (file_path, "/");
    if (last) *last = '\0';

    struct stat st = {0};
    if (stat (file_path, &st) == -1) {
        g_mkdir_with_parents (file_path, 0700);
	g_free (file_path);
    }
    int res = xmlSaveFormatFileEnc (file_name, doc, usage_encoding, 4);
    if (res == -1)
    {
        g_debug ("Create xml failed!");
        return FALSE;
    }

    return TRUE;
}

static xmlDocPtr
get_doc (CdosAppUsage *self)
{
    g_return_val_if_fail (CDOS_IS_APP_USAGE (self), NULL);
    // 设置输出内容不要保留空白
    xmlKeepBlanksDefault (0);
    CdosAppUsagePrivate *priv = self->priv;

    if (NULL == priv->doc)
    {
        if (g_file_test (USAGE_APP, G_FILE_TEST_EXISTS))
        {
            priv->doc = xmlReadFile (USAGE_APP, usage_encoding, usage_options);
        }
        else
        {
            g_print ("Haven't config file!");
            if (create_empty_file (USAGE_APP))
            {
                priv->doc = xmlReadFile (USAGE_APP, usage_encoding, usage_options);
            }
            else
            {
                return NULL;
            }
        }

        if (NULL == priv->doc)
        {
            g_print ("Document is not parsed successfully.");
	    // app_usage.xml file exist, but is empty.
	    create_empty_file (USAGE_APP);
        }
    }

    return priv->doc;
}

static xmlNodePtr
get_root (CdosAppUsage *self)
{
    // 设置输出内容不要保留空白
    xmlKeepBlanksDefault (0);
    CdosAppUsagePrivate *priv = self->priv;

    if (NULL == priv->doc)
    {
        g_debug ("Cannot get root node, xmlDoc is NULL.");
        return NULL;
    }

    if (NULL == priv->root_node)
    {
        priv->root_node = xmlDocGetRootElement (priv->doc);
        if (NULL == priv->root_node)
        {
            g_debug ("Cannot get usage-app.xml root node.");
            xmlFreeDoc (priv->doc);
            priv->doc = NULL;
            return NULL;
        }
    }

    return priv->root_node;
}

static xmlNodePtr
get_node (CdosAppUsage *self, xmlChar *prop_name, xmlChar *prop_val)
{
    CdosAppUsagePrivate *priv;
    priv = self->priv;
    xmlNodePtr root_node = priv->root_node;
    xmlNodePtr child = root_node->xmlChildrenNode, find_node = NULL;
    xmlChar *node_prop;
    while (child)
    {
        node_prop = xmlGetProp (child, BAD_CAST prop_name);
        if (!(xmlStrcmp (node_prop, prop_val)))
        {
            find_node = child;
            break;
        }
        child = child->next;
    }

    return find_node;
}

static xmlNodePtr
new_app (CdosAppUsage *self, gchar *desktop_id, gchar *score, gchar *last_seen)
{
    g_return_val_if_fail (CDOS_IS_APP_USAGE (self), NULL);
    g_return_val_if_fail ((NULL != desktop_id), NULL);
    CdosAppUsagePrivate *priv = self->priv;

    xmlNodePtr root = get_root (self);
    xmlNodePtr new_node = xmlNewChild (root, NULL, BAD_CAST "application", NULL);
    xmlNewProp (new_node, BAD_CAST "id", BAD_CAST desktop_id);
    xmlNewProp (new_node, BAD_CAST "score", BAD_CAST score);
    xmlNewProp (new_node, BAD_CAST "last-seen", BAD_CAST last_seen);
    int res = xmlSaveFormatFile (USAGE_APP, priv->doc, 8);
    if (res == -1)
    {
        g_warning ("%s: XML file save failed", __func__);
        return NULL;
    }

    return new_node;
}

static gboolean
remove_app (CdosAppUsage *self, const gchar *desktop_id)
{
    CdosAppUsagePrivate *priv = self->priv;
    xmlNodePtr cur_node = get_node (self, BAD_CAST "id", BAD_CAST desktop_id);
    if (NULL == cur_node)
    {
        return FALSE;
    }

    xmlDocPtr doc = priv->doc;
    xmlNodePtr tmp_node = cur_node->next;
    xmlUnlinkNode (cur_node);
    xmlFreeNode (cur_node);
    cur_node = tmp_node;

    int res = xmlSaveFormatFile (USAGE_APP, doc, 8);
    if (res == -1)
    {
        g_warning ("%s: XML file save failed", __func__);
        return FALSE;
    }
    return TRUE;
}

static gboolean G_GNUC_UNUSED
set_score (CdosAppUsage *self, gchar *desktop_id, gdouble score_val)
{
    g_return_val_if_fail (CDOS_IS_APP_USAGE (self), FALSE);
    g_return_val_if_fail ((NULL != desktop_id), FALSE);
    xmlDocPtr doc = get_doc (self);
    gchar *score = g_strdup_printf ("%lf", score_val);
    xmlNodePtr cur_node = get_node (self, BAD_CAST "id", BAD_CAST desktop_id);
    if (NULL == cur_node)
    {
        return FALSE;
    }

    xmlAttrPtr attr = xmlSetProp (cur_node, BAD_CAST "score", BAD_CAST score);
    if (NULL == attr)
    {
        return FALSE;
    }
    int res = xmlSaveFormatFile (USAGE_APP, doc, 8);
    if (res == -1)
    {
        g_warning ("%s: XML file save failed", __func__);
        return FALSE;
    }
    return TRUE;
}

static gboolean G_GNUC_UNUSED
set_last_seen (CdosAppUsage *self, gchar *desktop_id, gchar *last_seen)
{
    g_return_val_if_fail (CDOS_IS_APP_USAGE (self), FALSE);
    g_return_val_if_fail ((NULL != desktop_id), FALSE);

    xmlNodePtr cur_node = get_node (self, BAD_CAST "id", BAD_CAST desktop_id);
    xmlDocPtr doc = get_doc (self);
    if (NULL == cur_node)
    {
        return FALSE;
    }

    xmlAttrPtr attr = xmlSetProp (cur_node, BAD_CAST "last-seen", BAD_CAST last_seen);
    if (NULL == attr)
    {
        return FALSE;
    }

    int res = xmlSaveFormatFile (USAGE_APP, doc, 8);
    if (res == -1)
    {
        g_warning ("%s: XML file save failed", __func__);
        return FALSE;
    }
    return TRUE;
}

static gboolean
interval_save_usage_for_xml (gpointer data)
{
    CdosAppUsage *self = CDOS_APP_USAGE (data);
    self->save_id = 0;
    xmlDocPtr doc = get_doc (self);
    xmlNodePtr root_node = get_root (self);
    if ((NULL == doc) || (NULL == root_node))
    {
        g_warning ("\n root_node or doc is NULL\n");
        create_empty_file (USAGE_APP);
        return FALSE;
    }
    CdosAppUsagePrivate *priv = self->priv;
    GList *l = priv->previously_running;
    UsageData *usage = NULL;
    gchar *score_str = NULL;

    if (NULL != l)
    {
        xmlNodeSetContent (root_node, NULL);
        int res = xmlSaveFormatFile (USAGE_APP, doc, 4);
        if (res == -1)
        {
            g_warning ("%s: XML save failed", __func__);
            return FALSE;
        }

        while (l)
        {
            usage = l->data;
            score_str = g_strdup_printf ("%lf", usage->score);
            if (NULL != usage->desktop_id)
            {
                new_app (self, usage->desktop_id, score_str, usage->last_seen);
            }
            else
            {
                g_debug ("Cannot get desktop_id, will drop this item.");
            }
            l = l->next;
        }

        g_debug ("File saved on %s", g_date_time_format (g_date_time_new_now_local (), "%Y-%m-%d %H:%M:%S"));
        g_signal_emit_by_name (self, "usage-changed");
    }

    return FALSE;
}

// xml options end

static void
ensure_list_saved (CdosAppUsage *self)
{
    if (self->save_id != 0)
    {
        return;
    }

    g_debug ("Ensure list saved.");

    CdosAppUsagePrivate *priv = self->priv;
    // 过于频繁的排序操作，在同时打开几十个APP时，CDOS处理不过来会崩溃
    priv->previously_running = sort_usage_struct_list (self, priv->previously_running);
    self->save_id = g_timeout_add_seconds (SAVE_APPS_TIMEOUT, interval_save_usage_for_xml, self);
    g_source_set_name_by_id (self->save_id, "[cdos-desktop] interval_save_usage_for_xml");
}

static void
on_focus_app_changed (CdosWindowTracker *tracker, GParamSpec *spec, CdosAppUsage *self)
{
    CdosAppUsagePrivate *priv = self->priv;

    // if (!self->enable_monitoring)
    // {
    //     update_enable_monitoring (self);
    //     return;
    // }

    if (priv->watched_app)
    {
        increment_useage_for_app (self, priv->watched_app);
        g_object_unref (priv->watched_app);
    }

    priv->watched_app = cdos_window_tracker_get_focus_app (tracker);
    priv->watch_start_time = get_time ();
}

static void
on_app_state_changed (CdosAppSystem *app_system, CdosApp *app, gpointer user_data)
{
    g_return_if_fail (NULL != cdos_app_get_id (app));
    CdosAppUsage *self = CDOS_APP_USAGE (user_data);
    UsageData *usage = NULL;
    gboolean running = FALSE;

    // if (!self->enable_monitoring)
    // {
    //     return;
    // }

    if (cdos_app_is_window_backed (app))
    {
        return;
    }

    usage = get_usage_for_app (self, app);
    running = cdos_app_get_state (app) == CDOS_APP_STATE_RUNNING;
    glong time_val = get_time ();

    if (running && (usage != NULL))
    {
        usage->last_seen = time_to_str (time_val);
        g_debug ("Update for last_seen for running app!");
    }
}

// 每当由任一APP超过最大限定值，全员score除以2
static void
normalize_usage (CdosAppUsage *self)
{
    g_return_if_fail (CDOS_IS_APP_USAGE (self));
    CdosAppUsagePrivate *priv = self->priv;
    GList *l = priv->previously_running;
    UsageData *app_usage = NULL;
    while (l)
    {
        app_usage = l->data;
        app_usage->score = app_usage->score / 2;
        l = l->next;
    }
}

static void
increment_useage_for_app_at_time (CdosAppUsage *self, CdosApp *watched_app, glong time)
{
    CdosAppUsagePrivate *priv = self->priv;
    guint elapsed;
    guint usage_count;
    UsageData *usage;
    GDesktopAppInfo *info = NULL;
    gboolean no_display = FALSE;
    elapsed = (guint) (time - priv->watch_start_time);
    usage_count = elapsed / FOCUS_TIME_MIN_SECONDS;
    const gchar *app_id = cdos_app_get_id (watched_app);
    glong time_val = get_time ();
    priv->watched_app = watched_app;

    usage = get_usage_for_app (self, watched_app);
    // 不再限定关注超过7秒，记录APP
    // if (usage_count > 0)
    // {
    if (usage != NULL)
    {
        usage->score += usage_count;
        usage->last_seen = time_to_str (time_val);
    }
    else
    {
        usage = g_new0 (UsageData, 1);
        usage->desktop_id = g_strdup (app_id);
        usage->score += usage_count;
        usage->last_seen = time_to_str (time_val);
        usage->app = watched_app;
        info = cdos_app_get_app_info (usage->app);

        if (NULL != info)
        {
            // info为空，会造成段错误
            no_display = g_desktop_app_info_get_nodisplay (info);
        }

        if (usage->desktop_id && (!no_display))
        {
            priv->previously_running = g_list_append (priv->previously_running, usage);
            interval_save_usage_for_xml (self);
        }
        else
        {
            if (no_display)
            {
                g_debug ("Desktop file NoDisplay is true.");
            }
            else
            {
                g_debug ("Cannot get desktop_id, will not added.");
            }
            g_free (usage);
            return;
        }
    }

    if (usage->score > SCORE_MAX)
    {
        normalize_usage (self);
    }

    ensure_list_saved (self);
    g_debug ("id:%s score:%f list lenght:%d lastseen:%s",
             app_id,
             usage->score,
             g_list_length (priv->previously_running),
             usage->last_seen);
    // }
}

static void
increment_useage_for_app (CdosAppUsage *self, CdosApp *app)
{
    long cur_time = get_time ();
    increment_useage_for_app_at_time (self, app, cur_time);
}

static UsageData *
get_usage_for_app (CdosAppUsage *self, CdosApp *app)
{
    CdosAppUsagePrivate *priv = self->priv;
    GList *l = priv->previously_running;
    const gchar *app_id = cdos_app_get_id (app);

    UsageData *usage = NULL;
    while (l)
    {
        usage = l->data;
        if (!g_strcmp0 (app_id, usage->desktop_id))
        {
            usage->app = app;
            return usage;
        }
        l = l->next;
    }

    return NULL;
}

static CdosApp *
get_cdos_app_from_id (CdosAppUsage *self, gchar *desktop_id)
{
    g_return_val_if_fail (CDOS_IS_APP_USAGE (self), NULL);
    g_return_val_if_fail ((NULL != desktop_id), NULL);

    CdosAppSystem *appsys = cdos_app_system_get_default ();

    CdosApp *app = cdos_app_system_lookup_app (appsys, desktop_id);
    return app;
}

static void
update_enable_monitoring (CdosAppUsage *self)
{
    CdosAppUsagePrivate *priv = self->priv;
    // self->enable_monitoring =
    //     g_settings_get_boolean (self->privacy_settings, ENABLE_MONITORING_KEY);

    if (!self->enable_monitoring)
    {
        g_debug ("enable_monitoring is FALSE!");
        if (priv->watched_app)
        {
            g_object_unref (priv->watched_app);
        }

        priv->watched_app = NULL;
        if (self->save_id)
        {
            g_source_remove (self->save_id);
            self->save_id = 0;
        }
    }
}

static void
trans_struct_to_app_list (CdosAppUsage *self, GList *previously_running)
{
    UsageData *usg = NULL;
    self->app_list = NULL;
    gint i = 0;
    //“最常使用”默认显示12个
    while (previously_running)
    {
        usg = previously_running->data;
        if (!g_list_find (self->app_list, usg->app))
            self->app_list = g_list_append (self->app_list, usg->app);
        previously_running = previously_running->next;
        i++;
    }
}

void
cdos_app_usage_remove_app (CdosAppUsage *self, CdosApp *app)
{
    UsageData *usage;
    CdosAppSystem *appsys;
    usage = get_usage_for_app (self, app);
    appsys = cdos_app_system_get_default ();
    remove_app_usage_for_xml (appsys, usage->desktop_id, self);
}

static void
cdos_app_usage_load_data (CdosAppUsage *self)
{
    gchar *err = NULL;
    UsageData *usage = NULL;
    glong week_ago = get_time () - 7 * 24 * 60 * 60;
    gchar *desktop_id = NULL, *last_seen = NULL, *score_str = NULL;
    gdouble score = 0, last_seen_val = 0;
    GList *should_removed = NULL;
    CdosAppUsagePrivate *priv = self->priv;
    if (NULL == priv->previously_running)
    {
        xmlNodePtr root_node = get_root (self);
        if (NULL != root_node)
        {
            xmlNodePtr child = root_node->xmlChildrenNode;
            while (child)
            {
                desktop_id = (gchar *)xmlGetProp (child, BAD_CAST "id");
                last_seen = (gchar *)xmlGetProp (child, BAD_CAST "last-seen");
                score_str = (gchar *)xmlGetProp (child, BAD_CAST "score");
                score = g_strtod (score_str, &err);
                last_seen_val = g_strtod (last_seen, &err);

                usage = g_new0 (UsageData, 1);
                usage->desktop_id = g_strdup (desktop_id);
                usage->score = score;
                usage->app = get_cdos_app_from_id (self, usage->desktop_id);
                usage->last_seen = g_strdup (last_seen);
                if ((score < SCORE_MIN && last_seen_val < week_ago) || (usage->app == NULL))
                {
                    should_removed = g_list_append (should_removed, desktop_id);
                    g_free (usage);
                }
                else
                {
                    priv->previously_running = g_list_append (priv->previously_running, usage);
                }
                child = child->next;
            }

            // 从xml移除不合适的项目
            g_debug ("should removed length %d", g_list_length (should_removed));
            while (should_removed)
            {
                desktop_id = should_removed->data;
                remove_app (self, desktop_id);
                should_removed = should_removed->next;
            }
        }
    }
}

static void
remove_app_usage_for_xml (CdosAppSystem *appsys, const gchar *desktop_id, gpointer user_data)
{
    g_debug ("removed app callback.");
    CdosAppUsage *self = CDOS_APP_USAGE (user_data);
    if (NULL == desktop_id)
    {
        g_warning ("Will removed app cannot get desktop_id.");
        return;
    }
    gboolean res = remove_app (self, desktop_id);
    if (!res)
    {
        g_debug ("Remove_app failed in %s!", __func__);
        // 这里如果return可能会造成软件已经卸载，GList无法清空
        // return;
    }

    CdosAppUsagePrivate *priv = self->priv;
    GList *l = priv->previously_running;
    UsageData *usage = NULL;
    gboolean list_changed = FALSE;

    while (l)
    {
        usage = l->data;
        if (g_strcmp0 (usage->desktop_id, desktop_id) == 0)
        {
            priv->previously_running = g_list_remove (priv->previously_running, usage);
            list_changed = TRUE;
            break;
        }
        l = l->next;
    }

    if (list_changed)
    {
        interval_save_usage_for_xml (self);
    }
}


// ===== Public Method =====
static CdosAppUsage *instance = NULL;

CdosAppUsage *
cdos_app_usage_get_default (void)
{
    if (instance == NULL)
    {
        instance = g_object_new (CDOS_TYPE_APP_USAGE, NULL);
    }

    return instance;
}

GList *
cdos_app_usage_get_list (CdosAppUsage *self)
{
    g_return_val_if_fail (CDOS_IS_APP_USAGE (self), NULL);
    CdosAppUsagePrivate *priv = self->priv;

    if (self->app_list)
    {
        g_list_free (self->app_list);
        self->app_list = NULL;
    }
    trans_struct_to_app_list (self, priv->previously_running);
    return self->app_list;
}

static void
cdos_app_usage_dispose (GObject *gobject)
{
    CdosWindowTracker *tracker = cdos_window_tracker_get_default ();
    CdosAppSystem *appsys = cdos_app_system_get_default ();
    g_signal_handlers_disconnect_by_data (tracker, gobject);
    g_signal_handlers_disconnect_by_data (appsys, gobject);
    G_OBJECT_CLASS (cdos_app_usage_parent_class)->dispose (gobject);
}

static void
cdos_app_usage_class_init (CdosAppUsageClass *klass)
{
    GObjectClass *gobject_class = G_OBJECT_CLASS (klass);

    gobject_class->dispose = cdos_app_usage_dispose;
    gobject_class->finalize = cdos_app_usage_finalize;

    usage_signals[USAGE_CHANGED] = g_signal_new ("usage-changed",
                                                 G_TYPE_FROM_CLASS (klass),
                                                 G_SIGNAL_RUN_LAST,
                                                 G_STRUCT_OFFSET (CdosAppUsageClass, usage_changed),
                                                 NULL,
                                                 NULL,
                                                 NULL,
                                                 G_TYPE_NONE,
                                                 0,
                                                 NULL);
}

static void
cdos_app_usage_init (CdosAppUsage *self)
{
    CdosAppUsagePrivate *priv = cdos_app_usage_get_instance_private (self);
    CdosWindowTracker *tracker = cdos_window_tracker_get_default ();
    CdosAppSystem *appsys = cdos_app_system_get_default ();

    self->save_id = 0;
    self->enable_monitoring = TRUE;
    self->privacy_settings = NULL;
    self->app_list = NULL;

    self->priv = priv;
    priv->doc = get_doc (self);
    priv->root_node = get_root (self);
    priv->previously_running = NULL;

    g_signal_connect (tracker, "notify::focus-app", G_CALLBACK (on_focus_app_changed), self);
    g_signal_connect (appsys, "app-state-changed", G_CALLBACK (on_app_state_changed), self);
    // CdosAppSystem已将 installed-changed 拆分成 installed-changed 和 removed-changed
    g_signal_connect (appsys, "removed-changed", G_CALLBACK (remove_app_usage_for_xml), self);
    // self->privacy_settings = g_settings_new (PRIVACY_SCHEMA);
    // g_signal_connect (self->privacy_settings,
    //                   "changed::"ENABLE_MONITORING_KEY,
    //                   G_CALLBACK (on_enable_monitoring_key_changed),
    //                   self);

    update_enable_monitoring (self);
    cdos_app_usage_load_data (self);
}

static void
cdos_app_usage_finalize (GObject *obj)
{
    CdosAppUsage *self = CDOS_APP_USAGE (obj);

    interval_save_usage_for_xml (self);

    if (self->save_id > 0)
    {
        g_source_remove (self->save_id);
    }

    if (self->privacy_settings)
    {
        g_object_unref (self->privacy_settings);
    }

    if (NULL != self->app_list)
    {
        g_list_free (self->app_list);
    }

    CdosAppUsagePrivate *priv = self->priv;

    if (NULL != priv->previously_running)
    {
        g_list_free (priv->previously_running);
    }

    if (priv->root_node)
    {
        xmlFreeNode (priv->root_node);
    }
    if (priv->doc)
    {
        xmlFreeDoc (priv->doc);
    }

    xmlCleanupParser ();

    instance = NULL;

    G_OBJECT_CLASS (cdos_app_usage_parent_class)->finalize (obj);
}
