/* -*- mode: C; c-file-style: "gnu"; indent-tabs-mode: nil; -*- */

#include "config.h"

#include "cdos-base-util.h"
#include <glib/gi18n-lib.h>
#include <gtk/gtk.h>
#ifdef HAVE__NL_TIME_FIRST_WEEKDAY
#include <langinfo.h>
#endif

#include <libxml/parser.h>
#include <libxml/tree.h>
#include <libxml/xmlmemory.h>

/* Some code in this file adapted under the GPLv2+ from:
 *
 * GNOME panel utils: gnome-panel/gnome-panel/panel-util.c
 * (C) 1997, 1998, 1999, 2000 The Free Software Foundation
 * Copyright 2000 Helix Code, Inc.
 * Copyright 2000,2001 Eazel, Inc.
 * Copyright 2001 George Lebl
 * Copyright 2002 Sun Microsystems Inc.
 *
 * Authors: George Lebl
 *          Jacob Berkman
 *          Mark McLoughlin
 */

static GFile *
cdos_base_util_get_gfile_root (GFile *file)
{
    GFile *parent;
    GFile *parent_old;

    /* search for the root on the URI */
    parent_old = g_object_ref (file);
    parent = g_file_get_parent (file);
    while (parent != NULL)
    {
        g_object_unref (parent_old);
        parent_old = parent;
        parent = g_file_get_parent (parent);
    }

    return parent_old;
}

static char *
cdos_base_util_get_file_display_name_if_mount (GFile *file)
{
    GFile *compare;
    GVolumeMonitor *monitor;
    GList *mounts, *l;
    char *ret;

    ret = NULL;

    /* compare with all mounts */
    monitor = g_volume_monitor_get ();
    mounts = g_volume_monitor_get_mounts (monitor);
    for (l = mounts; l != NULL; l = l->next)
    {
        GMount *mount;
        mount = G_MOUNT (l->data);
        compare = g_mount_get_root (mount);
        if (!ret && g_file_equal (file, compare))
            ret = g_mount_get_name (mount);
        g_object_unref (mount);
    }
    g_list_free (mounts);
    g_object_unref (monitor);

    return ret;
}

static char *
cdos_base_util_get_file_display_for_common_files (GFile *file)
{
    GFile *compare;

    compare = g_file_new_for_path (g_get_home_dir ());
    if (g_file_equal (file, compare))
    {
        g_object_unref (compare);
        /* Translators: this is the same string as the one found in
       * nemo */
        return g_strdup (_("Home"));
    }

    compare = g_file_new_for_path ("/");
    if (g_file_equal (file, compare))
    {
        g_object_unref (compare);
        /* Translators: this is the same string as the one found in
       * nemo */
        return g_strdup (_("File System"));
    }
    g_object_unref (compare);

    return NULL;
}

static char *
cdos_base_util_get_file_description (GFile *file)
{
    GFileInfo *info;
    char *ret;

    ret = NULL;

    info = g_file_query_info (file, "standard::description", G_FILE_QUERY_INFO_NOFOLLOW_SYMLINKS, NULL, NULL);

    if (info)
    {
        ret = g_strdup (g_file_info_get_attribute_string (info, G_FILE_ATTRIBUTE_STANDARD_DESCRIPTION));
        g_object_unref (info);
    }

    return ret;
}

static char *
cdos_base_util_get_file_display_name (GFile *file, gboolean use_fallback)
{
    GFileInfo *info;
    char *ret;

    ret = NULL;

    info = g_file_query_info (file, "standard::display-name", G_FILE_QUERY_INFO_NOFOLLOW_SYMLINKS, NULL, NULL);

    if (info)
    {
        ret = g_strdup (g_file_info_get_display_name (info));
        g_object_unref (info);
    }

    if (!ret && use_fallback)
    {
        /* can happen with URI schemes non supported by gvfs */
        char *basename;

        basename = g_file_get_basename (file);
        ret = g_filename_display_name (basename);
        g_free (basename);
    }

    return ret;
}

static GIcon *
cdos_base_util_get_file_icon_if_mount (GFile *file)
{
    GFile *compare;
    GVolumeMonitor *monitor;
    GList *mounts, *l;
    GIcon *ret;

    ret = NULL;

    /* compare with all mounts */
    monitor = g_volume_monitor_get ();
    mounts = g_volume_monitor_get_mounts (monitor);
    for (l = mounts; l != NULL; l = l->next)
    {
        GMount *mount;
        mount = G_MOUNT (l->data);
        compare = g_mount_get_root (mount);
        if (!ret && g_file_equal (file, compare))
        {
            ret = g_mount_get_icon (mount);
        }
        g_object_unref (mount);
    }
    g_list_free (mounts);
    g_object_unref (monitor);

    return ret;
}

static const char *
cdos_base_util_get_icon_for_uri_known_folders (const char *uri)
{
    const char *icon;
    char *path;
    int len;

    icon = NULL;

    if (!g_str_has_prefix (uri, "file:"))
        return NULL;

    path = g_filename_from_uri (uri, NULL, NULL);

    len = strlen (path);
    if (path[len] == '/')
        path[len] = '\0';

    if (strcmp (path, "/") == 0)
        icon = "drive-harddisk";
    else if (strcmp (path, g_get_home_dir ()) == 0)
        icon = "user-home";
    else if (strcmp (path, g_get_user_special_dir (G_USER_DIRECTORY_DESKTOP)) == 0)
        icon = "user-desktop";

    g_free (path);

    return icon;
}

/* This is based on nautilus_compute_title_for_uri() and
 * nautilus_file_get_display_name_nocopy() */
char *
cdos_base_util_get_label_for_uri (const char *text_uri)
{
    GFile *file;
    char *label;
    GFile *root;
    char *root_display;

    /* Here's what we do:
   *  + x-nautilus-search: URI
   *  + check if the URI is a mount
   *  + if file: URI:
   *   - check for known file: URI
   *   - check for description of the GFile
   *   - use display name of the GFile
   *  + else:
   *   - check for description of the GFile
   *   - if the URI is a root: "root displayname"
   *   - else: "root displayname: displayname"
   */

    label = NULL;

    // FIXME: see nautilus_query_to_readable_string() to have a nice name
    if (g_str_has_prefix (text_uri, "x-nautilus-search:"))
        return g_strdup (_("Search"));

    file = g_file_new_for_uri (text_uri);

    label = cdos_base_util_get_file_display_name_if_mount (file);
    if (label)
    {
        g_object_unref (file);
        return label;
    }

    if (g_str_has_prefix (text_uri, "file:"))
    {
        label = cdos_base_util_get_file_display_for_common_files (file);
        if (!label)
            label = cdos_base_util_get_file_description (file);
        if (!label)
            label = cdos_base_util_get_file_display_name (file, TRUE);
        g_object_unref (file);

        return label;
    }

    label = cdos_base_util_get_file_description (file);
    if (label)
    {
        g_object_unref (file);
        return label;
    }

    root = cdos_base_util_get_gfile_root (file);
    root_display = cdos_base_util_get_file_description (root);
    if (!root_display)
        root_display = cdos_base_util_get_file_display_name (root, FALSE);
    if (!root_display)
        /* can happen with URI schemes non supported by gvfs */
        root_display = g_file_get_uri_scheme (root);

    if (g_file_equal (file, root))
        label = root_display;
    else
    {
        char *displayname;

        displayname = cdos_base_util_get_file_display_name (file, TRUE);
        /* Translators: the first string is the name of a gvfs
       * method, and the second string is a path. For
       * example, "Trash: some-directory". It means that the
       * directory called "some-directory" is in the trash.
       */
        label = g_strdup_printf (_("%1$s: %2$s"), root_display, displayname);
        g_free (root_display);
        g_free (displayname);
    }

    g_object_unref (root);
    g_object_unref (file);

    return label;
}

/**
 * cdos_base_util_get_icon_for_uri:
 * @text_uri: URI值
 *
 * 查找与提供URI相关的icon。处理多个特定的NOME-internal事例，
 * 如x-nautilus-search。
 *
 * Return Value: 新的GIcon对象
 */
GIcon *
cdos_base_util_get_icon_for_uri (const char *text_uri)
{
    const char *name;
    GFile *file;
    GFileInfo *info;
    GIcon *retval;

    /* Here's what we do:
   *  + check for known file: URI
   *  + x-nautilus-search: URI
   *  + override burn: URI icon
   *  + check if the URI is a mount
   *  + override trash: URI icon for subfolders
   *  + check for application/x-gnome-saved-search mime type and override
   *    icon of the GFile
   *  + use icon of the GFile
   */

    /* this only checks file: URI */
    name = cdos_base_util_get_icon_for_uri_known_folders (text_uri);
    if (name)
        return g_themed_icon_new (name);

    if (g_str_has_prefix (text_uri, "x-nautilus-search:"))
        return g_themed_icon_new ("folder-saved-search");

    /* gvfs doesn't give us a nice icon, so overriding */
    if (g_str_has_prefix (text_uri, "burn:"))
        return g_themed_icon_new ("nautilus-cd-burner");

    file = g_file_new_for_uri (text_uri);

    retval = cdos_base_util_get_file_icon_if_mount (file);
    if (retval)
        return retval;

    /* gvfs doesn't give us a nice icon for subfolders of the trash, so
   * overriding */
    if (g_str_has_prefix (text_uri, "trash:"))
    {
        GFile *root;

        root = cdos_base_util_get_gfile_root (file);
        g_object_unref (file);
        file = root;
    }

    info = g_file_query_info (file, "standard::icon", G_FILE_QUERY_INFO_NONE, NULL, NULL);
    g_object_unref (file);

    if (!info)
        return g_themed_icon_new ("gtk-file");

    retval = g_file_info_get_icon (info);
    if (retval)
        g_object_ref (retval);
    g_object_unref (info);

    if (retval)
        return retval;

    return g_themed_icon_new ("gtk-file");
}

static void
stop_pick (ClutterActor *actor, const ClutterColor *color)
{
    g_signal_stop_emission_by_name (actor, "pick");
}

/**
 * cdos_base_util_set_hidden_from_pick:
 * @actor: ClutterActor对象
 * @hidden: 是否应该隐藏选择的actor对象
 *
 * 如果hidden是TRUE，即使是CLUTTER_PICK_ALL模式，也需要隐藏选择的actor对
 * 象；如果hidden是FALSE，不隐藏actor对象
 */
void
cdos_base_util_set_hidden_from_pick (ClutterActor *actor, gboolean hidden)
{
    gpointer existing_handler_data;

    existing_handler_data = g_object_get_data (G_OBJECT (actor), "cdos-stop-pick");
    if (hidden)
    {
        if (existing_handler_data != NULL)
            return;
        g_signal_connect (actor, "pick", G_CALLBACK (stop_pick), NULL);
        g_object_set_data (G_OBJECT (actor), "cdos-stop-pick", GUINT_TO_POINTER (1));
    }
    else
    {
        if (existing_handler_data == NULL)
            return;
        g_signal_handlers_disconnect_by_func (actor, stop_pick, NULL);
        g_object_set_data (G_OBJECT (actor), "cdos-stop-pick", NULL);
    }
}

/**
 * cdos_base_util_get_transformed_allocation:
 * @actor: ClutterActor对象
 * @box: (out): location to store returned box in stage coordinates
 *
 * 这个函数与clutter_actor_get_transformed_position()和
 * clutter_actor_get_transformed_size()的组合是相似的，但和
 * clutter_actor_get_transformed_size()函数不相似，因为它总
 * 是返回当前值的变换，而clutter_actor_get_transformed_size()
 * 返回重新布局的错误值。
 *
 * 如果没有有效的变换，和GJS队列的限制，该函数比clutter_actor_get_abs_allocation_vertices()
 * 更合适。
 */
void
cdos_base_util_get_transformed_allocation (ClutterActor *actor, ClutterActorBox *box)
{
    /* Code adapted from clutter-actor.c:
   * Copyright 2006, 2007, 2008 OpenedHand Ltd
   */
    graphene_point3d_t v[4];
    gfloat x_min, x_max, y_min, y_max;
    gint i;

    g_return_if_fail (CLUTTER_IS_ACTOR (actor));

    clutter_actor_get_abs_allocation_vertices (actor, v);

    x_min = x_max = v[0].x;
    y_min = y_max = v[0].y;

    for (i = 1; i < G_N_ELEMENTS (v); ++i)
    {
        if (v[i].x < x_min)
            x_min = v[i].x;

        if (v[i].x > x_max)
            x_max = v[i].x;

        if (v[i].y < y_min)
            y_min = v[i].y;

        if (v[i].y > y_max)
            y_max = v[i].y;
    }

    box->x1 = x_min;
    box->y1 = y_min;
    box->x2 = x_max;
    box->y2 = y_max;
}

char *
cdos_base_util_normalize_and_casefold (const char *str)
{
    char *normalized, *result;

    if (str == NULL)
        return NULL;

    normalized = g_utf8_normalize (str, -1, G_NORMALIZE_ALL);
    result = g_utf8_casefold (normalized, -1);
    g_free (normalized);
    return result;
}

/**
 * cdos_base_util_get_week_start:
 *
 * 获取当前区域的第一个周的某一天，以从0到6的数字表示，代表着从周日到周六。
 *
 * Returns: 当前区域表示第一周七天中一天的数值。
 */
/* Copied from gtkcalendar.c */
int
cdos_base_util_get_week_start (void)
{
    int week_start;
#ifdef HAVE__NL_TIME_FIRST_WEEKDAY
    union
    {
        unsigned int word;
        char *string;
    } langinfo;
    int week_1stday = 0;
    int first_weekday = 1;
    guint week_origin;
#else
    char *gtk_week_start;
#endif

#ifdef HAVE__NL_TIME_FIRST_WEEKDAY
    langinfo.string = nl_langinfo (_NL_TIME_FIRST_WEEKDAY);
    first_weekday = langinfo.string[0];
    langinfo.string = nl_langinfo (_NL_TIME_WEEK_1STDAY);
    week_origin = langinfo.word;
    if (week_origin == 19971130) /* Sunday */
        week_1stday = 0;
    else if (week_origin == 19971201) /* Monday */
        week_1stday = 1;
    else
        g_warning ("Unknown value of _NL_TIME_WEEK_1STDAY.\n");

    week_start = (week_1stday + first_weekday - 1) % 7;
#else
/* Use a define to hide the string from xgettext */
#define GTK_WEEK_START "calendar:week_start:0"
    gtk_week_start = dgettext ("gtk30", GTK_WEEK_START);

    if (strncmp (gtk_week_start, "calendar:week_start:", 20) == 0)
        week_start = *(gtk_week_start + 20) - '0';
    else
        week_start = -1;

    if (week_start < 0 || week_start > 6)
    {
        g_warning (
            "Whoever translated calendar:week_start:0 for GTK+ "
            "did so wrongly.\n");
        week_start = 0;
    }
#endif

    return week_start;
}

/**
 * cdos_get_event_state:
 * @event: ClutterEvent对象
 *
 * 获取事件的当前状态。
 * Gets the current state of the event (the set of modifier keys that
 * are pressed down). Thhis is a wrapper around
 * clutter_event_get_state() that strips out any un-declared modifier
 * flags, to make gjs happy; see
 * https://bugzilla.gnome.org/show_bug.cgi?id=597292.
 *
 * Return value: 事件的状态
 */
ClutterModifierType
cdos_base_util_get_event_state (ClutterEvent *event)
{
    ClutterModifierType state = clutter_event_get_state (event);
    return state & CLUTTER_MODIFIER_MASK;
}

/**
 * cdos_write_string_to_stream:
 * @stream: GOutputStream对象
 * @str: 要赋值给stream的UTF-8格式的字符串
 * @error: 存储GError值的地址
 *
 * 将UTF-8格式的字符串写入GOutputStream对象。对没有二进制缓冲机制的GJS而言，这是一种解决方案。
 *
 * Return value: 如果写入成功，返回TRUE。
 */
gboolean
cdos_base_util_write_string_to_stream (GOutputStream *stream, const char *str, GError **error)
{
    return g_output_stream_write_all (stream, str, strlen (str), NULL, NULL, error);
}

/**
 * cdos_base_util_get_file_contents_utf8_sync:
 * @path: UTF-8编码格式的文件名路径
 * @error: GError对象
 *
 * 同步加载一个文件的内容作为一个NUL结尾的字符串，以UTF-8格式验证。嵌入NUL
 * 字符算作无效的内容。
 *
 * Returns: 文件内容
 */
char *
cdos_base_util_get_file_contents_utf8_sync (const char *path, GError **error)
{
    char *contents;
    gsize len;
    if (!g_file_get_contents (path, &contents, &len, error))
        return NULL;
    if (!g_utf8_validate (contents, len, NULL))
    {
        g_free (contents);
        g_set_error (error, G_IO_ERROR, G_IO_ERROR_FAILED, "File %s contains invalid UTF-8", path);
        return NULL;
    }
    return contents;
}

/**
 * cdos_breakpoint:
 *
 * 使用G_BREAKPOINT()终止当前进程。这用于调试，例如gdb。
 */
void
cdos_base_util_breakpoint (void)
{
    G_BREAKPOINT ();
}

/**
 * cdos_parse_search_provider:
 * @data: provider描述
 * @name: 存储显示名字的地址
 * @url: 存储url模板的地址
 * @langs: 所支持语言的列表
 * @icon_data_uri: 存储uri的地址
 * @error: 存储GError的地址
 *
 * Returns: 成功时返回TRUE
 */
gboolean
cdos_base_util_parse_search_provider (const char *data,
                                      char **name,
                                      char **url,
                                      GList **langs,
                                      char **icon_data_uri,
                                      GError **error)
{
    xmlDocPtr doc = xmlParseMemory (data, strlen (data));
    xmlNode *root;

    *name = NULL;
    *url = NULL;
    *icon_data_uri = NULL;
    *langs = NULL;

    if (!doc)
    {
        g_set_error (error, G_IO_ERROR, G_IO_ERROR_FAILED, "Malformed xml");
        return FALSE;
    }

    root = xmlDocGetRootElement (doc);
    if (root && root->name && xmlStrcmp (root->name, (const xmlChar *)"OpenSearchDescription") == 0)
    {
        xmlNode *child;
        for (child = root->children; child; child = child->next)
        {
            if (!child->name)
                continue;
            if (xmlStrcmp (child->name, (const xmlChar *)"Language") == 0)
            {
                xmlChar *val = xmlNodeListGetString (doc, child->xmlChildrenNode, 1);
                if (!val)
                    continue;
                *langs = g_list_append (*langs, g_strdup ((char *)val));
                xmlFree (val);
            }
            if (!*name && xmlStrcmp (child->name, (const xmlChar *)"ShortName") == 0)
            {
                xmlChar *val = xmlNodeListGetString (doc, child->xmlChildrenNode, 1);
                *name = g_strdup ((char *)val);
                xmlFree (val);
            }
            if (!*icon_data_uri && xmlStrcmp (child->name, (const xmlChar *)"Image") == 0)
            {
                xmlChar *val = xmlNodeListGetString (doc, child->xmlChildrenNode, 1);
                if (val)
                    *icon_data_uri = g_strdup ((char *)val);
                xmlFree (val);
            }
            if (!*url && xmlStrcmp (child->name, (const xmlChar *)"Url") == 0)
            {
                xmlChar *template;
                xmlChar *type;

                type = xmlGetProp (child, (const xmlChar *)"type");
                if (!type)
                    continue;

                if (xmlStrcmp (type, (const xmlChar *)"text/html") != 0)
                {
                    xmlFree (type);
                    continue;
                }
                xmlFree (type);

                template = xmlGetProp (child, (const xmlChar *)"template");
                if (!template)
                    continue;
                *url = g_strdup ((char *)template);
                xmlFree (template);
            }
        }
    }
    else
    {
        g_set_error (error, G_IO_ERROR, G_IO_ERROR_FAILED, "Invalid OpenSearch document");
        xmlFreeDoc (doc);
        return FALSE;
    }
    xmlFreeDoc (doc);
    if (*icon_data_uri && *name && *url)
        return TRUE;

    if (*icon_data_uri)
        g_free (*icon_data_uri);
    else
        g_set_error (error, G_IO_ERROR, G_IO_ERROR_FAILED, "search provider doesn't have icon");

    if (*name)
        g_free (*name);
    else if (error && !*error)
        g_set_error (error, G_IO_ERROR, G_IO_ERROR_FAILED, "search provider doesn't have ShortName");

    if (*url)
        g_free (*url);
    else if (error && !*error)
        g_set_error (error, G_IO_ERROR, G_IO_ERROR_FAILED, "search provider doesn't have template for url");

    if (*langs)
    {
        g_list_foreach (*langs, (GFunc)g_free, NULL);
        g_list_free (*langs);
    }

    *url = NULL;
    *name = NULL;
    *icon_data_uri = NULL;
    *langs = NULL;

    return FALSE;
}

/**
 * cdos_shader_effect_set_double_uniform:
 * @effect: ClutterShaderEffect对象
 * @name: uniform的名称
 * @value: 要设置的值
 *
 * 在ClutterShaderEffect对象上设置两个uniform
 *
 * 问题是JavaScript没有数据类型，gjs尝试自动猜测GValue的类型。如果数
 * 值是“1.0”或者其他，会猜测用用整型，在GLSL会导致错误。
 */
void
cdos_base_util_shader_effect_set_double_uniform (ClutterShaderEffect *effect, const gchar *name, gdouble value)
{
    GValue gvalue = G_VALUE_INIT;
    g_value_init (&gvalue, G_TYPE_DOUBLE);
    g_value_set_double (&gvalue, value);

    clutter_shader_effect_set_uniform_value (effect, name, &gvalue);
}

/**
 * cdos_base_util_str_tolower:
 * @str: 原始字符串
 * Returns 返回一个新字符串
 * 将原始字符串中的字符全部转换为小写字符,使用之后,通过@g_free予以释放;
 */
gchar *
cdos_base_util_str_tolower (gchar *str)
{
    if (str == NULL)
    {
        return NULL;
    }
    gint len = strlen (str);
    gchar *temp = g_new0 (char, len + 1);
    gint i;
    for (i = 0; i < len; i++)
    {
        gchar c = g_ascii_tolower (str[i]);
        temp[i] = c;
    }
    temp[len] = '\0';
    return temp;
}
/**
 * cdos_base_util_str_tolower:
 * @str: 原始字符串
 * Returns 返回一个新字符串
 * 将原始字符串中的字符全部转换为大写字符,使用之后,通过@g_free予以释放;
 */
gchar *
cdos_base_util_str_toupper (gchar *str)
{
    if (str == NULL)
    {
        return NULL;
    }
    gint len = strlen (str);
    gchar *temp = g_new0 (char, len + 1);
    gint i;
    for (i = 0; i < len; i++)
    {
        gchar c = g_ascii_toupper (str[i]);
        temp[i] = c;
    }
    temp[len] = '\0';
    return temp;
}
