/* -*- Mode: C; indent-tabs-mode: nil; c-basic-offset: 4; tab-width: 4 -*- */
/*
 * ccm-debug.c
 * Copyright (C) Nicolas Bruguier 2007-2011 <gandalfn@club-internet.fr>
 *
 * cairo-compmgr is free software: you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * cairo-compmgr is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/* addr2line.c -- convert addresses to line number and function name
 Copyright 1997, 1998, 1999, 2000, 2001, 2002 Free Software Foundation, Inc.
 Contributed by Ulrich Lauther <Ulrich.Lauther@mchp.siemens.de>

 This file was part of GNU Binutils.

 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2, or (at your option)
 any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  */

#include "config.h"

#include "cdos-debug.h"
#define HAVE_DEBUG 1

int debug_call = 0;
int debug_mode = 0;

#ifndef __SW64__

#ifndef __USE_GNU
#define __USE_GNU
#endif

#define _GNU_SOURCE
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <execinfo.h>

#include <bfd.h>
#include <dlfcn.h>
#include <link.h>

#include <time.h>
#include <unistd.h>

/* 150 isn't special; it's just an arbitrary non-ASCII char value.  */
#define OPTION_DEMANGLER (150)
static asymbol **syms; /* Symbol table.  */
#ifndef false
#define false 0
#endif
#ifndef true
#define true 1
#endif

static void slurp_symtab (bfd *abfd);
static void find_address_in_section (bfd *abfd, asection *section, void *data);

/* Read in the symbol table.  */

static void
slurp_symtab (bfd *abfd)
{
    long symcount;
    unsigned int size;

    if ((bfd_get_file_flags (abfd) & HAS_SYMS) == 0)
        return;

    symcount = bfd_read_minisymbols (abfd, false, (PTR)&syms, &size);
    if (symcount == 0)
        symcount = bfd_read_minisymbols (abfd, true /* dynamic */, (PTR)&syms, &size);

    if (symcount < 0)
    {
        printf ("%s\n", bfd_get_filename (abfd));
        exit (1);
    }
}

/* These global variables are used to pass information between
 translate_addresses and find_address_in_section.  */

static bfd_vma pc;
static const char *filename;
static const char *functionname;
static unsigned int line;
static int found;

/* Look for an address in a section.  This is called via
 bfd_map_over_sections.  */

static void
find_address_in_section (bfd *abfd, asection *section, void *data __attribute__((__unused__)))
{
    bfd_vma vma = 0;
    bfd_size_type size = 0;

    if (found)
        return;
    
    //这两个函数在debian11版本内不存在。
    /* if ((bfd_get_section_flags (abfd, section) & SEC_ALLOC) == 0)
        return;

    vma = bfd_get_section_vma (abfd, section);
    if (pc < vma)
        return; */

#ifdef HAVE_BFD_GET_SECTION_SIZE
    size = bfd_get_section_size (section);
#endif
#ifdef HAVE_BFD_SECTION_SIZE
    size = bfd_section_size (section);
#endif

    if (pc >= vma + size)
        return;

    found = bfd_find_nearest_line (abfd, section, syms, pc - vma, &filename, &functionname, &line);
}

static char **
translate_addresses_buf (bfd *abfd, bfd_vma *addr, int naddr)
{
    int naddr_orig = naddr;
    char b;
    int total = 0;
    enum
    {
        Count,
        Print
    } state;
    char *buf = &b;
    int len = 0;
    char **ret_buf = NULL;

    /* iterate over the formating twice.
     * the first time we count how much space we need
     * the second time we do the actual printing */
    for (state = Count; state <= Print; state++)
    {
        if (state == Print)
        {
            ret_buf = malloc (total + sizeof (char *) * naddr);
            buf = (char *)(ret_buf + naddr);
            len = total;
        }
        while (naddr)
        {
            if (state == Print)
                ret_buf[naddr - 1] = buf;
            pc = addr[naddr - 1];

            found = false;

            bfd_boolean has_reloc = (abfd->flags & HAS_RELOC);
            bfd_map_over_sections (abfd,
                                   find_address_in_section,
                                   /*(PTR) NULL*/ &has_reloc);

            if (!found)
            {
                total += snprintf (buf, len, "[0x%llx] \?\?() \?\?:0", (long long unsigned int)addr[naddr - 1]) + 1;
            }
            else
            {
                const char *name;

                name = functionname;
                if (name == NULL || *name == '\0')
                    name = "??";
                if (filename != NULL)
                {
                    char *h;

                    h = strrchr (filename, '/');
                    if (h != NULL)
                        filename = h + 1;
                }
                total += snprintf (buf, len, "%s:%u\t%s()", filename ? filename : "??", line, name) + 1;
            }
            if (state == Print)
            {
                /* set buf just past the end of string */
                buf = buf + total + 1;
            }
            naddr--;
        }
        naddr = naddr_orig;
    }
    return ret_buf;
}

/* Process a file.  */
static char **
process_file (const char *file_name, bfd_vma *addr, int naddr)
{
    bfd *abfd;
    char **matching;
    char **ret_buf;

    abfd = bfd_openr (file_name, NULL);

    if (abfd == NULL)
    {
        printf ("%s\n", file_name);
        exit (1);
    }

    if (bfd_check_format (abfd, bfd_archive))
    {
        printf ("%s: can not get addresses from archive", file_name);
        exit (1);
    }

    if (!bfd_check_format_matches (abfd, bfd_object, &matching))
    {
        // bfd_nonfatal(bfd_get_filename(abfd));
        if (bfd_get_error () == bfd_error_file_ambiguously_recognized)
        {
            // list_matching_formats(matching);
            free (matching);
        }
        exit (1);
    }

    slurp_symtab (abfd);

    ret_buf = translate_addresses_buf (abfd, addr, naddr);

    if (syms != NULL)
    {
        free (syms);
        syms = NULL;
    }

    bfd_close (abfd);
    return ret_buf;
}

#define MAX_DEPTH 16

struct file_match
{
    const char *file;
    void *address;
    void *base;
    void *hdr;
};

static int
find_matching_file (struct dl_phdr_info *info, size_t size, void *data)
{
    struct file_match *match = data;
    /* This code is modeled from Gfind_proc_info-lsb.c:callback() from libunwind */
    long n;
    const ElfW (Phdr) * phdr;
    ElfW (Addr) load_base = info->dlpi_addr;
    phdr = info->dlpi_phdr;
    for (n = info->dlpi_phnum; --n >= 0; phdr++)
    {
        if (phdr->p_type == PT_LOAD)
        {
            ElfW (Addr) vaddr = phdr->p_vaddr + load_base;
            if (match->address >= (void *)vaddr && match->address < (void *)(vaddr + phdr->p_memsz))
            {
                /* we found a match */
                match->file = info->dlpi_name;
                match->base = (void *)info->dlpi_addr;
                return 1;
            }
        }
    }
    return 0;
}

static char **
cdos_backtrace_symbols (void *const *buffer, int size)
{
    int stack_depth = size - 1;
    int x;
    int total = 0;

    char ***locations;
    char ** final;
    char *f_strings;

    locations = malloc (sizeof (char **) * (stack_depth + 1));

    bfd_init ();

    for (x = stack_depth; x >= 0; x--)
    {
        struct file_match match = {.address = buffer[x]};
        char **ret_buf;
        bfd_vma addr;
        dl_iterate_phdr (find_matching_file, &match);
        addr = (intptr_t)buffer[x] - (intptr_t)match.base;
        if (match.file && strlen (match.file))
            ret_buf = process_file (match.file, &addr, 1);
        else
            ret_buf = process_file ("/proc/self/exe", &addr, 1);
        locations[x] = ret_buf;
        total += strlen (ret_buf[0]) + 1;
    }
    /* allocate the array of char* we are going to return and extra space for
     * all of the strings */
    final = malloc (total + (stack_depth + 1) * sizeof (char *));
    /* get a pointer to the extra space */
    f_strings = (char *)(final + stack_depth + 1);

    /* fill in all of strings and pointers */
    for (x = stack_depth; x >= 0; x--)
    {
        strcpy (f_strings, locations[x][0]);
        free (locations[x]);
        final[x] = f_strings;
        f_strings += strlen (f_strings) + 1;
    }

    free (locations);

    return final;
}

void
cdos_backtrace (void)
{
    if (debug_mode == 0)
        return;

    void *array[30];
    int size;
    char **strings;
    int i;

    size = backtrace (array, 30);
    strings = cdos_backtrace_symbols (array, size);

    for (i = 0; i < size; i++)
        printf ("%s\n", strings[i]);
    if (strings)
        free (strings);
}

#else
void
cdos_backtrace (void)
{
}
#endif // __SW64__

/**
 * trace.log
 * https://balau82.wordpress.com/2010/10/06/trace-and-profile-function-calls-with-gcc/
 */
/*
#define HAVE_TRACE 0
#if HAVE_TRACE
#define TRACE_FILE "trace.log"

static FILE *fp_trace = NULL;

void __attribute__((constructor))
trace_begin ()
{
    if ((fp_trace = fopen (TRACE_FILE, "w")) == NULL)
    {
        printf ("open %s error.\n", TRACE_FILE);
        return;
    }
}

void __attribute__((destructor))
trace_end ()
{
    if (fp_trace)
        fclose (fp_trace);
}

#define NUM_ITEMS 20
struct syms_cache {
    void *ptr;
    char *name;
}cache[NUM_ITEMS] = {
    {NULL, NULL}
};

static int current_index = -1;

static int find_cache (const void *func)
{
    int i;
    for (i = current_index; i >= 0; i--)
    {
        if (cache[i].ptr == func)
            return i;
    }
    return i;
}

void __attribute__((no_instrument_function))
__cyg_profile_func_enter (void *func, void *caller)
{
    if (debug_call == 0)
        return;
    else if (debug_call == 1)
    {
        if (fp_trace != NULL)
            fprintf (fp_trace, "e %p %p %lu\n", func, caller, time (NULL));
    }
    else if (debug_call == 2)
    {
        void *arr[2];
        arr[0] = func;
        arr[1] = caller;
        char **strings;
        strings = cdos_backtrace_symbols (arr, 2);
        fprintf (fp_trace, "func: %s\n\tcaller: %s %lu\n", strings[0], strings[1], time (NULL));
    }
    else if (debug_call == 3)
    {
        int i = find_cache (func);
        if (i > 0)
        {
            fprintf (fp_trace, "> time: %.3f, func: %s\n",
                    (double)clock()/CLOCKS_PER_SEC*1000, cache[i].name);
        }
        else {
            void *arr[1];
            arr[0] = func;
            char **strings;
            strings = cdos_backtrace_symbols (arr, 1);
            fprintf (fp_trace, "> time: %.3f, func: %s\n",
                    (double)clock()/CLOCKS_PER_SEC*1000, strings[0]);
            if (current_index < NUM_ITEMS - 1)
            {
                i = current_index + 1;
                cache[i].ptr = func;
                cache[i].name = strings[0];
                current_index = i;
            }
        }
        fflush (fp_trace);
    }
}

void __attribute__((no_instrument_function))
__cyg_profile_func_exit (void *func, void *caller)
{
    if (debug_call == 0)
        return;
    if (debug_call == 1)
    {
      if (fp_trace != NULL)
        fprintf (fp_trace, "x %p %p %lu\n", func, caller, time (NULL));
    }
    else if (debug_call == 3)
    {
        clock_t end = clock();
        int i = find_cache (func);
        if (i > 0)
        {
            fprintf (fp_trace, "< time: %.3f, func: %s\n",
                    (double)end/CLOCKS_PER_SEC * 1000, cache[i].name);
            cache[i].ptr = NULL;
            cache[i].name = NULL;
            if (i == current_index)
                current_index -= 1;
        }
        else {
            void *arr[1];
            arr[0] = func;
            char **strings;
            strings = cdos_backtrace_symbols (arr, 1);
            fprintf (fp_trace, "< time: %.3f, func: %s\n",
                (double)end/CLOCKS_PER_SEC*1000, strings[0]);
        }
        fflush (fp_trace);
    }
}
#endif
*/
