/* -*- mode: C; c-file-style: "gnu"; indent-tabs-mode: nil; -*- */

#include "config.h"

#include "cdos-doc-system.h"

#include "cdos-global.h"


/**
 * SECTION:cdos-doc-system
 * 简述: 跟踪最近使用的文档
 *
 * 存储最近使用的文档信息，并添加到用于异步查询的API。
 */
enum
{
    CHANGED,
    DELETED,
    LAST_SIGNAL
};

static guint signals[LAST_SIGNAL] = {0};

struct _CdosDocSystemPrivate
{
    GtkRecentManager *manager;
    GHashTable *infos_by_uri;
    GSList *infos_by_timestamp;

    guint idle_recent_changed_id;

    GHashTable *deleted_infos;
    guint idle_emit_deleted_id;
};

G_DEFINE_TYPE_WITH_PRIVATE (CdosDocSystem, cdos_doc_system, G_TYPE_OBJECT);

/**
 * cdos_doc_system_get_all:
 * @system: CdosDocSystem对象
 *
 * 返回当前存储的最近文件集合。当变化时更新，最近文件从底层GtkRecentManager读取。
 * 该函数没有实现I/O。
 *
 * Returns: 存储最近的文件信息
 */
GSList *
cdos_doc_system_get_all (CdosDocSystem *self)
{
    return self->priv->infos_by_timestamp;
}

/**
 * cdos_doc_system_lookup_by_uri:
 * @system: CdosDocSystem对象
 * @uri: Url值
 *
 * Returns: 对应url的最近文件信息
 */
GtkRecentInfo *
cdos_doc_system_lookup_by_uri (CdosDocSystem *self, const char *uri)
{
    return g_hash_table_lookup (self->priv->infos_by_uri, uri);
}

static gboolean
cdos_doc_system_idle_emit_deleted (gpointer data)
{
    CdosDocSystem *self = CDOS_DOC_SYSTEM (data);
    GHashTableIter iter;
    gpointer key, value;

    self->priv->idle_emit_deleted_id = 0;

    g_hash_table_iter_init (&iter, self->priv->deleted_infos);

    while (g_hash_table_iter_next (&iter, &key, &value))
    {
        GtkRecentInfo *info = key;
        g_signal_emit (self, signals[DELETED], 0, info);
    }

    g_signal_emit (self, signals[CHANGED], 0);

    return FALSE;
}

typedef struct
{
    CdosDocSystem *self;
    GtkRecentInfo *info;
} CdosDocSystemRecentQueryData;

static void
on_recent_file_query_result (GObject *source, GAsyncResult *result, gpointer user_data)
{
    CdosDocSystemRecentQueryData *data = user_data;
    CdosDocSystem *self = data->self;
    GError *error = NULL;
    GFileInfo *fileinfo;

    fileinfo = g_file_query_info_finish (G_FILE (source), result, &error);
    if (fileinfo)
        g_object_unref (fileinfo);
    /* This is a strict error check; we don't want to cause recent files to
   * vanish for anything potentially transient.
   */
    if (error != NULL && error->domain == G_IO_ERROR && error->code == G_IO_ERROR_NOT_FOUND)
    {
        self->priv->infos_by_timestamp = g_slist_remove (self->priv->infos_by_timestamp, data->info);
        g_hash_table_remove (self->priv->infos_by_uri, gtk_recent_info_get_uri (data->info));

        g_hash_table_insert (self->priv->deleted_infos, gtk_recent_info_ref (data->info), NULL);

        if (self->priv->idle_emit_deleted_id == 0)
            self->priv->idle_emit_deleted_id = g_timeout_add (0, cdos_doc_system_idle_emit_deleted, self);
    }
    g_clear_error (&error);

    gtk_recent_info_unref (data->info);
    g_free (data);
}

/**
 * cdos_doc_system_queue_existence_check:
 * @system: CdosDocSystem对象
 * @n_items: 检查存在的列表项数目，从最后操作的那个文档开始
 *
 * 异步启动检查最近文件的数目，任何已被删除的文件将被CdosDocSystem对象发送删除信号。
 * 注意：该函数忽略非本地文件，这些文件一直存在的（直到这些文件被人为删除）。
 *
 * 该函数在CdosDocSystem对象发送改变信号后调用，用于显示文件的子集。
 */
void
cdos_doc_system_queue_existence_check (CdosDocSystem *self, guint n_items)
{
    GSList *iter;
    guint i;

    for (i = 0, iter = self->priv->infos_by_timestamp; i < n_items && iter; i++, iter = iter->next)
    {
        GtkRecentInfo *info = iter->data;
        const char *uri;
        GFile *file;
        CdosDocSystemRecentQueryData *data;

        if (!gtk_recent_info_is_local (info))
            continue;

        data = g_new0 (CdosDocSystemRecentQueryData, 1);
        data->self = self;
        data->info = gtk_recent_info_ref (info);

        uri = gtk_recent_info_get_uri (info);
        file = g_file_new_for_uri (uri);

        g_file_query_info_async (file,
                                 "standard::type",
                                 G_FILE_QUERY_INFO_NONE,
                                 G_PRIORITY_DEFAULT,
                                 NULL,
                                 on_recent_file_query_result,
                                 data);
        g_object_unref (file);
    }
}

static int
sort_infos_by_timestamp_descending (gconstpointer a, gconstpointer b)
{
    GtkRecentInfo *info_a = (GtkRecentInfo *)a;
    GtkRecentInfo *info_b = (GtkRecentInfo *)b;
    time_t modified_a, modified_b;

    modified_a = gtk_recent_info_get_modified (info_a);
    modified_b = gtk_recent_info_get_modified (info_b);

    return modified_b - modified_a;
}

static gboolean
idle_handle_recent_changed (gpointer data)
{
    CdosDocSystem *self = CDOS_DOC_SYSTEM (data);
    GList *items, *iter;

    self->priv->idle_recent_changed_id = 0;

    g_hash_table_remove_all (self->priv->deleted_infos);
    g_hash_table_remove_all (self->priv->infos_by_uri);
    g_slist_free (self->priv->infos_by_timestamp);
    self->priv->infos_by_timestamp = NULL;

    items = gtk_recent_manager_get_items (self->priv->manager);
    for (iter = items; iter; iter = iter->next)
    {
        GtkRecentInfo *info = iter->data;
        const char *uri = gtk_recent_info_get_uri (info);

        /* uri is owned by the info */
        g_hash_table_insert (self->priv->infos_by_uri, (char *)uri, info);

        self->priv->infos_by_timestamp = g_slist_prepend (self->priv->infos_by_timestamp, info);
    }
    g_list_free (items);

    self->priv->infos_by_timestamp = g_slist_sort (self->priv->infos_by_timestamp, sort_infos_by_timestamp_descending);

    g_signal_emit (self, signals[CHANGED], 0);

    return FALSE;
}

static void
cdos_doc_system_on_recent_changed (GtkRecentManager *manager, CdosDocSystem *self)
{
    if (self->priv->idle_recent_changed_id != 0)
        return;
    self->priv->idle_recent_changed_id = g_timeout_add (0, idle_handle_recent_changed, self);
}

/**
 * cdos_doc_system_open:
 * @system: CdosDocSystem对象
 * @info: GtkRecentInfo对象
 * @workspace: 打开工作区间，如果是-1，则为默认值
 *
 * 启动默认应用程序
 * Launch the default application associated with the mime type of
 * @info, using its uri.
 */
void
cdos_doc_system_open (CdosDocSystem *system, GtkRecentInfo *info, int workspace)
{
    GFile *file;
    GAppInfo *app_info;
    gboolean needs_uri;
    GAppLaunchContext *context;

    context = cdos_global_create_app_launch_context (cdos_global_get (), 0, -1);
    if (workspace != -1)
        gdk_app_launch_context_set_desktop ((GdkAppLaunchContext *)context, workspace);

    file = g_file_new_for_uri (gtk_recent_info_get_uri (info));
    needs_uri = g_file_get_path (file) == NULL;
    g_object_unref (file);

    app_info = g_app_info_get_default_for_type (gtk_recent_info_get_mime_type (info), needs_uri);
    if (app_info != NULL)
    {
        GList *uris;
        uris = g_list_prepend (NULL, (gpointer)gtk_recent_info_get_uri (info));
        g_app_info_launch_uris (app_info, uris, context, NULL);
        g_list_free (uris);
    }
    else
    {
        char *app_name;
        const char *app_exec;
        char *app_exec_quoted;
        guint count;
        time_t time;

        app_name = gtk_recent_info_last_application (info);
        if (gtk_recent_info_get_application_info (info, app_name, &app_exec, &count, &time))
        {
            GRegex *regex;

            /* TODO: Change this once better support for creating
             GAppInfo is added to GtkRecentInfo, as right now
             this relies on the fact that the file uri is
             already a part of appExec, so we don't supply any
             files to app_info.launch().

             The 'command line' passed to
             create_from_command_line is allowed to contain
             '%<something>' macros that are expanded to file
             name / icon name, etc, so we need to escape % as %%
           */

            regex = g_regex_new ("%", 0, 0, NULL);
            app_exec_quoted = g_regex_replace (regex, app_exec, -1, 0, "%%", 0, NULL);
            g_regex_unref (regex);

            app_info = g_app_info_create_from_commandline (app_exec_quoted, NULL, 0, NULL);
            g_free (app_exec_quoted);

            /* The point of passing an app launch context to
             launch() is mostly to get startup notification and
             associated benefits like the app appearing on the
             right desktop; but it doesn't really work for now
             because with the way we create the appInfo we
             aren't reading the application's desktop file, and
             thus don't find the StartupNotify=true in it. So,
             despite passing the app launch context, no startup
             notification occurs.
           */
            g_app_info_launch (app_info, NULL, context, NULL);
        }

        g_free (app_name);
    }

    g_object_unref (context);
}

static void
cdos_doc_system_class_init (CdosDocSystemClass *klass)
{
    signals[CHANGED] = g_signal_new ("changed",
                                     CDOS_TYPE_DOC_SYSTEM,
                                     G_SIGNAL_RUN_LAST,
                                     0,
                                     NULL,
                                     NULL,
                                     NULL,
                                     G_TYPE_NONE,
                                     0);

    signals[DELETED] = g_signal_new ("deleted",
                                     CDOS_TYPE_DOC_SYSTEM,
                                     G_SIGNAL_RUN_LAST,
                                     0,
                                     NULL,
                                     NULL,
                                     NULL,
                                     G_TYPE_NONE,
                                     1,
                                     GTK_TYPE_RECENT_INFO);
}

static void
cdos_doc_system_init (CdosDocSystem *self)
{
    CdosDocSystemPrivate *priv;

    self->priv = priv = cdos_doc_system_get_instance_private (self);
    self->priv->manager = gtk_recent_manager_get_default ();

    self->priv->deleted_infos = g_hash_table_new_full (NULL, NULL, (GDestroyNotify)gtk_recent_info_unref, NULL);
    self->priv->infos_by_uri =
        g_hash_table_new_full (g_str_hash, g_str_equal, NULL, (GDestroyNotify)gtk_recent_info_unref);

    g_signal_connect (self->priv->manager, "changed", G_CALLBACK (cdos_doc_system_on_recent_changed), self);
    cdos_doc_system_on_recent_changed (self->priv->manager, self);
}

/**
 * cdos_doc_system_get_default:
 *
 * Return Value: 全局CdosDocSystem单例模式
 */
CdosDocSystem *
cdos_doc_system_get_default (void)
{
    static CdosDocSystem *instance = NULL;

    if (instance == NULL)
        instance = g_object_new (CDOS_TYPE_DOC_SYSTEM, NULL);

    return instance;
}
