#include "config.h"
#include "common.h"
#include <meta/util.h>
#include "cdos-global.h"

#include <fcitx-utils/keysym.h>
#include <fcitx/ime.h>
#include <fcitx/frontend.h>

#include "cdos-input-method-fcitx.h"

#define FCITX_DBUS_SERVICE "org.fcitx.Fcitx"
#define FCITX_IM_DBUS_PATH "/inputmethod"
#define FCITX_IC_DBUS_PATH "/inputcontext_%d"

#define FCITX_IM_DBUS_INTERFACE "org.fcitx.Fcitx.InputMethod"
#define FCITX_IC_DBUS_INTERFACE "org.fcitx.Fcitx.InputContext"

enum
{
    FOCUS_IN,
    FOCUS_OUT,
    RESET,
    N_SIGNALS
};

static guint signals[N_SIGNALS] = { 0 };

typedef struct _CdosInputMethodFcitx
{
    ClutterInputMethod parent;
    ClutterInputContentHintFlags hint;
    ClutterInputContentPurpose purpose;
    ClutterInputFocus *focus;

    gchar *preedit;

    GDBusProxy *im_proxy;
    GDBusProxy *ic_proxy;
    gchar *icname;
    gchar *servicename;
    GCancellable *cancellable;
    guint bus_watch_id;
}CdosInputMethodFcitx;

G_DEFINE_TYPE (CdosInputMethodFcitx, cdos_input_method_fcitx, CLUTTER_TYPE_INPUT_METHOD)

static void
cdos_input_method_fcitx_dispose (GObject *gobject)
{
    CdosInputMethodFcitx *self = CDOS_INPUT_METHOD_FCITX (gobject);
    g_clear_object (&self->cancellable);
    g_clear_pointer (&self->icname, g_free);
    g_clear_pointer (&self->servicename, g_free);
    g_clear_object (&self->ic_proxy);
    g_clear_object (&self->im_proxy);
    if (self->bus_watch_id)
    {
        g_bus_unwatch_name (self->bus_watch_id);
        self->bus_watch_id = 0;
    }

    G_OBJECT_CLASS (cdos_input_method_fcitx_parent_class)->dispose (gobject);
}

static void
cdos_input_method_fcitx_call_method (CdosInputMethodFcitx *self, const char *method)
{
    if (self->ic_proxy)
    {
        g_dbus_proxy_call (self->ic_proxy,
                           method,
                           NULL,
                           G_DBUS_CALL_FLAGS_NONE,
                           -1,
                           NULL, NULL, NULL);
    }
}

static void
cdos_input_method_fcitx_focus_in (ClutterInputMethod *im, ClutterInputFocus *focus)
{
    CdosInputMethodFcitx *self = CDOS_INPUT_METHOD_FCITX (im);
    self->focus = focus;
    cdos_input_method_fcitx_call_method (self, "FocusIn");

    g_signal_emit (self, signals[FOCUS_IN], 0);
}

static void
cdos_input_method_fcitx_focus_out (ClutterInputMethod *im)
{
    CdosInputMethodFcitx *self = CDOS_INPUT_METHOD_FCITX (im);
    self->focus = NULL;

    cdos_input_method_fcitx_call_method (self, "FocusOut");
    if (self->preedit)
    {
        g_free (self->preedit);
        self->preedit = NULL;
        clutter_input_method_set_preedit_text (im, NULL, 0);
    }

    g_signal_emit (self, signals[FOCUS_OUT], 0);
}

static void
cdos_input_method_fcitx_reset (ClutterInputMethod *im)
{
    g_debug ("%s", __func__);
    CdosInputMethodFcitx *self = CDOS_INPUT_METHOD_FCITX (im);
    cdos_input_method_fcitx_call_method (self, "Reset");

    g_signal_emit (self, signals[RESET], 0);
}

static void
cdos_input_method_fcitx_set_cursor_location (ClutterInputMethod *im,
                                             const graphene_rect_t *rect)
{
    CdosInputMethodFcitx *self = CDOS_INPUT_METHOD_FCITX (im);
    int x, y, w, h;
    if (self->ic_proxy)
    {
        x = graphene_rect_get_x (rect);
        y = graphene_rect_get_y (rect);
        w = graphene_rect_get_width (rect);
        h = graphene_rect_get_height (rect);
        g_dbus_proxy_call (self->ic_proxy,
                           "SetCursorRect",
                           g_variant_new ("(iiii)", x, y, w, h),
                           G_DBUS_CALL_FLAGS_NONE, -1, NULL, NULL, NULL);
    }
    g_signal_emit_by_name (im, "cursor-location-changed", rect);
}

static void
cdos_input_method_fcitx_set_surrounding (ClutterInputMethod *im,
                                         const gchar        *text,
                                         guint               cursor,
                                         guint               anchor)
{
    g_debug ("%s", __func__);
    CdosInputMethodFcitx *self = CDOS_INPUT_METHOD_FCITX (im);
    if (self->ic_proxy)
    {
        if (text)
        {                                                       
            g_dbus_proxy_call (self->ic_proxy,
                               "SetSurroundingText",
                               g_variant_new ("(suu)", text, cursor, anchor),
                               G_DBUS_CALL_FLAGS_NONE, -1, NULL, NULL, NULL);
        }
        else
        {
            g_dbus_proxy_call (self->ic_proxy, "SetSurroundingTextPosition",
                               g_variant_new ("(uu)", cursor, anchor),        
                               G_DBUS_CALL_FLAGS_NONE, -1, NULL, NULL, NULL);
        }
    }
}
static void
cdos_input_method_fcitx_update_capacities (CdosInputMethodFcitx *self,
                                           guint32 caps)
{
    g_dbus_proxy_call (self->ic_proxy, "SetCapacity",
                       g_variant_new ("(u)", caps),
                       G_DBUS_CALL_FLAGS_NONE, -1, NULL, NULL, NULL);
}

static void
cdos_input_method_fcitx_update_content_hints (ClutterInputMethod *im,
                                              ClutterInputContentHintFlags hint)
{
    g_debug ("%s %u", __func__, hint);

    CdosInputMethodFcitx *self = CDOS_INPUT_METHOD_FCITX (im);
    self->hint = hint;
}

static void
cdos_input_method_fcitx_update_content_purpose (ClutterInputMethod         *im,
                                                ClutterInputContentPurpose  purpose)
{
    g_debug ("%s %u", __func__, purpose);
    guint32 caps;
    CdosInputMethodFcitx *self = CDOS_INPUT_METHOD_FCITX (im);
    self->purpose = purpose;
    if (purpose & CLUTTER_INPUT_CONTENT_PURPOSE_PASSWORD)
    {
        caps = CAPACITY_PASSWORD;
    }
    cdos_input_method_fcitx_update_capacities (self, caps);
}

typedef struct _KeyProcessData
{
    CdosInputMethodFcitx *self;
    ClutterEvent *event;
}KeyProcessData;

static void
fcitx_process_key_cb (GObject *source,
                      GAsyncResult *res,
                      gpointer user_data)
{
    gint ret = 0;
    GError *error = NULL;
    ClutterInputMethod *im;
    ClutterEvent *event;
    KeyProcessData *data = user_data;
    CdosInputMethodFcitx *self = data->self;

    g_clear_object (&self->cancellable);

    if (!self->ic_proxy)
        return;

    GVariant *result = g_dbus_proxy_call_finish (self->ic_proxy, res, &error);
    if (error)
    {
        g_warning ("%s: %s", __func__, error->message);
        g_error_free (error);
    }
    else
    {
        g_variant_get (result, "(i)", &ret);
        g_variant_unref (result);
    }
    im = CLUTTER_INPUT_METHOD (self);
    event = data->event;
    clutter_input_method_notify_key_event (im, event, ret);

    clutter_event_free (data->event);
    g_free (data);
}

static guint32
to_fcitx_modifiers (guint32 state)
{
    gint32 fcitx_mods = 0;

    if (state & CLUTTER_SHIFT_MASK)
        fcitx_mods |= FcitxKeyState_Shift;
    if (state & CLUTTER_LOCK_MASK)
        fcitx_mods |= FcitxKeyState_CapsLock;
    if (state & CLUTTER_CONTROL_MASK)
        fcitx_mods |= FcitxKeyState_Ctrl;
    if (state & CLUTTER_SUPER_MASK)
        fcitx_mods |= FcitxKeyState_Super;
    if (state & CLUTTER_META_MASK)
        fcitx_mods |= FcitxKeyState_Meta;

    return fcitx_mods;
}

static gboolean
cdos_input_method_fcitx_filter_key_event (ClutterInputMethod *im,
                                          const ClutterEvent *event)
{
    CdosInputMethodFcitx *self = CDOS_INPUT_METHOD_FCITX (im);

    if (self->purpose & CLUTTER_INPUT_CONTENT_PURPOSE_PASSWORD)
        return FALSE;

    if (!self->ic_proxy)
        return FALSE;
    else
    {
        ClutterEventType type;
        guint32 keyval, keycode, state, time;
        gint32 itype;
        KeyProcessData *data;

        type = clutter_event_type (event);
        if (type & CLUTTER_KEY_RELEASE)
            itype = FCITX_RELEASE_KEY;
        else
            itype = FCITX_PRESS_KEY;

        keyval = clutter_event_get_key_symbol (event);
        keycode = clutter_event_get_key_code (event); /* Convert XKB keycodes to evcodes? */

        if (event)
            time = clutter_event_get_time (event);
        else
            time = meta_display_get_current_time_roundtrip (cdos_global_get_display (cdos_global_get ()));

        state = clutter_event_get_state (event);
        state = to_fcitx_modifiers (state);

        if (!self->cancellable)
            self->cancellable = g_cancellable_new();
        data = g_new0 (KeyProcessData, 1);
        data->self = self;
        data->event = clutter_event_copy (event);
        g_dbus_proxy_call(self->ic_proxy,
                          "ProcessKeyEvent",
                          g_variant_new ("(uuuiu)", keyval, keycode, state, itype, time),
                          G_DBUS_CALL_FLAGS_NONE, -1, self->cancellable,
                          fcitx_process_key_cb, data);
    }

    return TRUE;
}

static void
on_fcitx_signal (GDBusProxy *proxy,
                 gchar *sender_name,
                 gchar *signal_name,
                 GVariant *parameters,
                 gpointer user_data)
{
    ClutterInputMethod *im;
    //CdosInputMethodFcitx *im_fcitx;

    if (g_strcmp0 (signal_name, "EnableIM") == 0)
    {
        g_debug ("%s: %s", __func__, signal_name);
    }
    else if (g_strcmp0 (signal_name, "CloseIM") == 0)
    {
        g_debug ("%s: %s", __func__, signal_name);
    }
    else if (g_strcmp0 (signal_name, "CommitString") == 0)
    {
        const gchar *data = NULL;
        g_variant_get (parameters, "(s)", &data);
        if (data)
        {
            im = CLUTTER_INPUT_METHOD (user_data);
            clutter_input_method_commit (im, data);
        }
    }
    else if (g_strcmp0 (signal_name, "ForwardKey") == 0)
    {
        g_debug ("%s: %s", __func__, signal_name);
        const ClutterEvent *event;
        guint32 keyval, state, time;
        guint32 keycode;
        gint32 type;
        gboolean is_press;
        g_variant_get (parameters, "(uui)", &keyval, &state, &type);
        is_press = type == FCITX_PRESS_KEY;
        //state &= ~FcitxKeyState
        event = clutter_get_current_event ();
        if (event)
            time = clutter_event_get_time (event);
        else
            time = meta_display_get_current_time_roundtrip (cdos_global_get_display (cdos_global_get ()));
        keycode = 0;
        im = CLUTTER_INPUT_METHOD (user_data);
        clutter_input_method_forward_key (im, keyval, keycode,
                                          state & CLUTTER_MODIFIER_MASK,
                                          time,
                                          is_press);
    }
    else if (g_strcmp0 (signal_name, "DeleteSurroundingText") == 0)
    {
        guint32 nchars;
        gint32 offset;
        g_variant_get (parameters, "(iu)", &offset, &nchars);
        g_debug ("%s: %s", __func__, signal_name);
        im = CLUTTER_INPUT_METHOD (user_data);
        clutter_input_method_delete_surrounding (im, offset, nchars);
    }
    else if (g_strcmp0 (signal_name, "UpdateClientSideUI") == 0)
    {
        g_debug ("%s: %s", __func__, signal_name);
        /*
        const gchar *auxup, *auxdown, *preedit, *candidate, *imname;
        int cursor;
        g_variant_get (parameters, "(sssssi)", &auxup, &auxdown, &preedit,
                      &candidate, &imname, &cursor);
        */
    }
    else if (g_strcmp0 (signal_name, "UpdateFormattedPreedit") == 0)
    {
        g_debug ("%s: %s (PACKAGE_NAME:PACKAGE_VERSION not implemented)", __func__, signal_name);
        int cursor_pos;
        GVariantIter *iter;
        g_variant_get(parameters, "(a(si)i)", &iter, &cursor_pos);

        gchar *string;
        int type;
        while (g_variant_iter_next (iter, "(si)", &string, &type, NULL))
        {
            g_debug ("string:%s, cursor_pos:%d, type:%d", string, cursor_pos, type);
            g_free (string);
        }
    }
}

static void
cdos_im_fcitx_create_ic_phase2_finished (GObject *source,
                                         GAsyncResult *res,
                                         gpointer user_data)
{
    GError *error = NULL;
    CdosInputMethodFcitx *self = CDOS_INPUT_METHOD_FCITX (user_data);

    g_clear_object (&self->cancellable);
    self->ic_proxy = g_dbus_proxy_new_finish (res, &error);
    if (error)
    {
        g_warning ("%s: %s", __func__, error->message);
        g_error_free (error);
        return;
    }
    g_signal_connect (self->ic_proxy, "g-signal",
                      G_CALLBACK (on_fcitx_signal), self);
}

static void
cdos_im_fcitx_create_ic_finished (GObject *source,
                                  GAsyncResult *res,
                                  gpointer user_data)
{
    GError *error = NULL;
    CdosInputMethodFcitx *self = CDOS_INPUT_METHOD_FCITX (user_data);

    g_clear_object (&self->cancellable);
    GVariant *result = g_dbus_proxy_call_finish (G_DBUS_PROXY (source), res, &error);
    if (error)
    {
        g_warning ("%s: %s", __func__, error->message);
        g_error_free (error);
        return;
    }
    gboolean enable;
    guint32 key1, state1, key2, state2;
    int id;
    g_variant_get (result, "(ibuuuu)", &id, &enable, &key1, &state1, &key2, &state2);
    g_variant_unref(result);

    self->icname = g_strdup_printf (FCITX_IC_DBUS_PATH, id);
    self->cancellable = g_cancellable_new();
    g_dbus_proxy_new_for_bus (G_BUS_TYPE_SESSION,
                              G_DBUS_PROXY_FLAGS_NONE,
                              NULL,
                              self->servicename,
                              self->icname, FCITX_IC_DBUS_INTERFACE,
                              self->cancellable,
                              cdos_im_fcitx_create_ic_phase2_finished,
                              self);
}

static void
cdos_im_fcitx_create_ic (CdosInputMethodFcitx *self)
{
    self->cancellable = g_cancellable_new();
    const char *appname = PACKAGE_NAME;
    pid_t pid = getpid();
    g_dbus_proxy_call (self->im_proxy,
                       "CreateICv3",
                       g_variant_new ("(si)", appname, pid),
                       G_DBUS_CALL_FLAGS_NONE, -1, /* timeout */
                       self->cancellable,
                       cdos_im_fcitx_create_ic_finished,
                       self);
}

static void
cdos_im_fcitx_create_proxy_finished (GObject *source,
                                     GAsyncResult *res,
                                     gpointer user_data)
{
    GError *error = NULL;
    CdosInputMethodFcitx *self = CDOS_INPUT_METHOD_FCITX (user_data);

    g_clear_object (&self->cancellable);
    self->im_proxy = g_dbus_proxy_new_for_bus_finish (res, &error);
    if (error)
    {
        g_warning ("%s: %s", __func__, error->message);
        g_error_free (error);
        return;
    }

    gchar *owner_name = g_dbus_proxy_get_name_owner (self->im_proxy);
    if (!owner_name)
    {
        g_debug ("%s %s %p", __func__, self->servicename, self->im_proxy);
        g_clear_object (&self->im_proxy);
        g_free(owner_name);
        return ;
    }
    else
        g_free(owner_name);

    cdos_im_fcitx_create_ic (self);
}

static void
cdos_im_fcitx_create_proxy (CdosInputMethodFcitx *self)
{
    self->cancellable = g_cancellable_new();
    g_dbus_proxy_new_for_bus (G_BUS_TYPE_SESSION,
                              G_DBUS_PROXY_FLAGS_DO_NOT_LOAD_PROPERTIES,
                              NULL,
                              self->servicename,
                              FCITX_IM_DBUS_PATH, FCITX_IM_DBUS_INTERFACE,
                              self->cancellable,
                              cdos_im_fcitx_create_proxy_finished,
                              self);
}

static void
cdos_input_method_fcitx_set_capacity (CdosInputMethodFcitx *self)
{
    g_debug ("%s", __func__);
    guint32 caps = CAPACITY_NONE;
    caps = CAPACITY_PREEDIT |
           CAPACITY_UPPERCASE |
           CAPACITY_LOWERCASE;
    cdos_input_method_fcitx_update_capacities (self, caps);
}

static void
on_fcitx_dbus_appeared_cb (GDBusConnection *connection, const gchar *name, const gchar *name_owner, gpointer user_data)
{
    g_message ("%s,%d,name:%s,name_owner:%s\n",__func__,__LINE__,name,name_owner);
    CdosInputMethodFcitx *self = CDOS_INPUT_METHOD_FCITX (user_data);

    if (self->ic_proxy)
        g_clear_object (&self->ic_proxy);
    if (self->im_proxy)
        g_clear_object (&self->im_proxy);

    cdos_im_fcitx_create_proxy (self);
}


static void
on_fcitx_dbus_vanished_cb (GDBusConnection *connection, const gchar *name, gpointer user_data)
{
    /* do nothing */
    g_message ("%s,%d,name:%s\n",__func__,__LINE__,name);
}

static void
cdos_input_method_fcitx_init (CdosInputMethodFcitx *self)
{
    int xdpy_num = 0;
    if (meta_is_wayland_compositor ())
        return;

    const char *display = g_getenv ("DISPLAY");
    if (display == NULL)
    {
        g_warning ("%s: Couldn't get DISPLAY environment", __func__);
        return;
    }
    xdpy_num = atoi (display + 1);
    self->servicename = g_strdup_printf ("%s-%d", FCITX_DBUS_SERVICE, xdpy_num);
    // cdos_im_fcitx_create_proxy (self);

    g_signal_connect (self, "notify::can-show-preedit",
                      G_CALLBACK (cdos_input_method_fcitx_set_capacity), self);

    self->bus_watch_id = g_bus_watch_name (  G_BUS_TYPE_SESSION,
                                    self->servicename,
                                    G_BUS_NAME_WATCHER_FLAGS_NONE,
                                    on_fcitx_dbus_appeared_cb,
                                    on_fcitx_dbus_vanished_cb,
                                    self,
                                    NULL);
}

static void
cdos_input_method_fcitx_class_init (CdosInputMethodFcitxClass *klass)
{
    GObjectClass *gobject_class = G_OBJECT_CLASS (klass);
    gobject_class->dispose = cdos_input_method_fcitx_dispose;

    ClutterInputMethodClass *parent_class = CLUTTER_INPUT_METHOD_CLASS (klass);
    parent_class->focus_in = cdos_input_method_fcitx_focus_in;
    parent_class->focus_out = cdos_input_method_fcitx_focus_out;
    parent_class->reset = cdos_input_method_fcitx_reset;
    parent_class->set_cursor_location = cdos_input_method_fcitx_set_cursor_location;
    parent_class->set_surrounding = cdos_input_method_fcitx_set_surrounding;
    parent_class->update_content_hints = cdos_input_method_fcitx_update_content_hints;
    parent_class->update_content_purpose = cdos_input_method_fcitx_update_content_purpose;
    parent_class->filter_key_event = cdos_input_method_fcitx_filter_key_event;

    signals[FOCUS_IN] =
        g_signal_new ("focus-in",
                      G_TYPE_FROM_CLASS (gobject_class),
                      G_SIGNAL_RUN_LAST,
                      0, NULL, NULL, NULL,
                      G_TYPE_NONE, 0);
    signals[FOCUS_OUT] =
        g_signal_new ("focus-out",
                      G_TYPE_FROM_CLASS (gobject_class),
                      G_SIGNAL_RUN_LAST,
                      0, NULL, NULL, NULL,
                      G_TYPE_NONE, 0);
    signals[RESET] =
        g_signal_new ("reset",
                      G_TYPE_FROM_CLASS (gobject_class),
                      G_SIGNAL_RUN_LAST,
                      0, NULL, NULL, NULL,
                      G_TYPE_NONE, 0);
}

ClutterInputMethod *
cdos_input_method_fcitx_new (void)
{
    return g_object_new (CDOS_TYPE_INPUT_METHOD_FCITX, NULL);
}
