/* -*- mode: C; c-file-style: "gnu"; indent-tabs-mode: nil; -*- */

/*
 * Copyright (c) 2008 Red Hat, Inc.
 * Copyright (c) 2008 Intel Corp.
 *
 * Based on plugin skeleton by:
 * Author: Tomas Frydrych <tf@linux.intel.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, see <http://www.gnu.org/licenses/>.
 */

#include "config.h"

#include <stdlib.h>
#include <string.h>

#include <clutter/clutter.h>
#include <clutter/x11/clutter-x11.h>
#include <meta/display.h>
#include <meta/meta-plugin.h>
#include <meta/meta-background.h>
#include <meta/meta-background-actor.h>
#include <meta/util.h>

#include "cdos-global-private.h"
#include "cdos-wm-private.h"

extern gint cdos_plugin_start (CdosGlobal *global);

static void cdos_plugin_outer_start (MetaPlugin *plugin);
static void cdos_plugin_minimize (MetaPlugin *plugin, MetaWindowActor *actor);
static void cdos_plugin_unminimize (MetaPlugin *plugin, MetaWindowActor *actor);
static void cdos_plugin_size_change (MetaPlugin *plugin,
                                     MetaWindowActor *actor,
                                     MetaSizeChange which_change,
                                     MetaRectangle *old_frame_rect,
                                     MetaRectangle *old_buffer_rect);
static void cdos_plugin_map (MetaPlugin *plugin, MetaWindowActor *actor);
static void cdos_plugin_destroy (MetaPlugin *plugin, MetaWindowActor *actor);

static void cdos_plugin_switch_workspace (MetaPlugin *plugin, gint from, gint to, MetaMotionDirection direction);

static void cdos_plugin_kill_window_effects (MetaPlugin *plugin, MetaWindowActor *actor);
static void cdos_plugin_kill_switch_workspace (MetaPlugin *plugin);

static void
cdos_plugin_show_tile_preview (MetaPlugin *plugin, MetaWindow *window, MetaRectangle *tile_rect, int tile_monitor);
static void cdos_plugin_hide_tile_preview (MetaPlugin *plugin);
static void
cdos_plugin_show_window_menu (MetaPlugin *plugin, MetaWindow *window, MetaWindowMenuType menu, int x, int y);
static void cdos_plugin_show_window_menu_for_rect (MetaPlugin *plugin,
                                                   MetaWindow *window,
                                                   MetaWindowMenuType menu,
                                                   MetaRectangle *rect);

static gboolean cdos_plugin_xevent_filter (MetaPlugin *plugin, XEvent *event);

static gboolean cdos_plugin_keybinding_filter (MetaPlugin *plugin, MetaKeyBinding *binding);

static void cdos_plugin_confirm_display_change (MetaPlugin *plugin);

static const MetaPluginInfo *cdos_plugin_plugin_info (MetaPlugin *plugin);


#define CDOS_TYPE_PLUGIN (cdos_plugin_get_type ())
#define CDOS_PLUGIN(obj) (G_TYPE_CHECK_INSTANCE_CAST ((obj), CDOS_TYPE_PLUGIN, CdosPlugin))
#define CDOS_PLUGIN_CLASS(klass) (G_TYPE_CHECK_CLASS_CAST ((klass), CDOS_TYPE_PLUGIN, CdosPluginClass))
#define CDOS_IS_PLUGIN(obj) (G_TYPE_CHECK_INSTANCE_TYPE ((obj), CDOS_PLUGIN_TYPE))
#define CDOS_IS_PLUGIN_CLASS(klass) (G_TYPE_CHECK_CLASS_TYPE ((klass), CDOS_TYPE_PLUGIN))
#define CDOS_PLUGIN_GET_CLASS(obj) (G_TYPE_INSTANCE_GET_CLASS ((obj), CDOS_TYPE_PLUGIN, CdosPluginClass))

typedef struct _CdosPlugin CdosPlugin;
typedef struct _CdosPluginClass CdosPluginClass;

struct _CdosPlugin
{
    MetaPlugin parent;

    int glx_error_base;
    int glx_event_base;
    guint have_swap_event : 1;
    CoglContext *cogl_context;

    CdosGlobal *global;
    ClutterActor *background_group;
};

struct _CdosPluginClass
{
    MetaPluginClass parent_class;
};

GType cdos_plugin_get_type (void);

G_DEFINE_TYPE (CdosPlugin, cdos_plugin, META_TYPE_PLUGIN)

static void
cdos_plugin_class_init (CdosPluginClass *klass)
{
    MetaPluginClass *plugin_class = META_PLUGIN_CLASS (klass);

    plugin_class->start = cdos_plugin_outer_start;
    plugin_class->map = cdos_plugin_map;
    plugin_class->minimize = cdos_plugin_minimize;
    plugin_class->unminimize = cdos_plugin_unminimize;
    plugin_class->size_change = cdos_plugin_size_change;
    plugin_class->destroy = cdos_plugin_destroy;

    plugin_class->switch_workspace = cdos_plugin_switch_workspace;

    plugin_class->kill_window_effects = cdos_plugin_kill_window_effects;
    plugin_class->kill_switch_workspace = cdos_plugin_kill_switch_workspace;

    plugin_class->show_tile_preview = cdos_plugin_show_tile_preview;
    plugin_class->hide_tile_preview = cdos_plugin_hide_tile_preview;
    plugin_class->show_window_menu = cdos_plugin_show_window_menu;
    plugin_class->show_window_menu_for_rect = cdos_plugin_show_window_menu_for_rect;

    plugin_class->xevent_filter = cdos_plugin_xevent_filter;
    plugin_class->keybinding_filter = cdos_plugin_keybinding_filter;

    plugin_class->confirm_display_change = cdos_plugin_confirm_display_change;

    plugin_class->plugin_info = cdos_plugin_plugin_info;
}

static void
cdos_plugin_init (CdosPlugin *cdos_base_plugin)
{
}

static gboolean
cdos_plugin_has_swap_event (CdosPlugin *cdos_base_plugin)
{
    CoglDisplay *cogl_display = cogl_context_get_display (cdos_base_plugin->cogl_context);
    CoglRenderer *renderer = cogl_display_get_renderer (cogl_display);
    const char *(*query_extensions_string) (Display * dpy, int screen);
    Bool (*query_extension) (Display * dpy, int *error, int *event);
    Display *xdisplay;
    const char *glx_extensions;

    /* We will only get swap events if Cogl is using GLX */
    if (cogl_renderer_get_winsys_id (renderer) != COGL_WINSYS_ID_GLX)
        return FALSE;

    xdisplay = clutter_x11_get_default_display ();

    query_extensions_string = (void *)cogl_get_proc_address ("glXQueryExtensionsString");
    query_extension = (void *)cogl_get_proc_address ("glXQueryExtension");

    query_extension (xdisplay, &cdos_base_plugin->glx_error_base, &cdos_base_plugin->glx_event_base);

    glx_extensions = query_extensions_string (xdisplay, XDefaultScreen (xdisplay));

    return strstr (glx_extensions, "GLX_INTEL_swap_event") != NULL;
}

static gboolean
start_ui_delay (CdosPlugin *cdos_base_plugin)
{
    cdos_plugin_start (cdos_base_plugin->global);
    return FALSE;
}

static void
cdos_plugin_outer_start (MetaPlugin *plugin)
{
    CdosPlugin *cdos_base_plugin = CDOS_PLUGIN (plugin);
    ClutterBackend *backend;

    backend = clutter_get_default_backend ();
    cdos_base_plugin->cogl_context = clutter_backend_get_cogl_context (backend);

    cdos_base_plugin->have_swap_event = cdos_plugin_has_swap_event (cdos_base_plugin);

    cdos_base_plugin->global = cdos_global_get ();
    _cdos_global_set_plugin (cdos_base_plugin->global, META_PLUGIN (cdos_base_plugin));

    g_idle_add ((GSourceFunc)start_ui_delay, cdos_base_plugin);
}

static CdosBaseWM *
get_cdos_base_wm (void)
{
    CdosBaseWM *wm;

    g_object_get (cdos_global_get (), "window-manager", &wm, NULL);
    /* drop extra ref added by g_object_get */
    g_object_unref (wm);

    return wm;
}

static void
cdos_plugin_minimize (MetaPlugin *plugin, MetaWindowActor *actor)
{
    _cdos_base_wm_minimize (get_cdos_base_wm (), actor);
}

static void
cdos_plugin_unminimize (MetaPlugin *plugin, MetaWindowActor *actor)
{
    _cdos_base_wm_unminimize (get_cdos_base_wm (), actor);
}

static void
cdos_plugin_size_change (MetaPlugin *plugin,
                         MetaWindowActor *actor,
                         MetaSizeChange which_change,
                         MetaRectangle *old_frame_rect,
                         MetaRectangle *old_buffer_rect)
{
    _cdos_base_wm_size_change (get_cdos_base_wm (), actor, which_change, old_frame_rect, old_buffer_rect);
}

static void
cdos_plugin_map (MetaPlugin *plugin, MetaWindowActor *actor)
{
    _cdos_base_wm_map (get_cdos_base_wm (), actor);
}

static void
cdos_plugin_destroy (MetaPlugin *plugin, MetaWindowActor *actor)
{
    _cdos_base_wm_destroy (get_cdos_base_wm (), actor);
}

static void
cdos_plugin_switch_workspace (MetaPlugin *plugin, gint from, gint to, MetaMotionDirection direction)
{
    _cdos_base_wm_switch_workspace (get_cdos_base_wm (), from, to, direction);
}

static void
cdos_plugin_kill_window_effects (MetaPlugin *plugin, MetaWindowActor *actor)
{
    _cdos_base_wm_kill_window_effects (get_cdos_base_wm (), actor);
}

static void
cdos_plugin_kill_switch_workspace (MetaPlugin *plugin)
{
    _cdos_base_wm_kill_switch_workspace (get_cdos_base_wm ());
}

static void
cdos_plugin_show_tile_preview (MetaPlugin *plugin, MetaWindow *window, MetaRectangle *tile_rect, int tile_monitor)
{
    _cdos_base_wm_show_tile_preview (get_cdos_base_wm (), window, tile_rect, tile_monitor);
}

static void
cdos_plugin_hide_tile_preview (MetaPlugin *plugin)
{
    _cdos_base_wm_hide_tile_preview (get_cdos_base_wm ());
}

static void
cdos_plugin_show_window_menu (MetaPlugin *plugin, MetaWindow *window, MetaWindowMenuType menu, int x, int y)
{
    _cdos_base_wm_show_window_menu (get_cdos_base_wm (), window, menu, x, y);
}

static void
cdos_plugin_show_window_menu_for_rect (MetaPlugin *plugin,
                                       MetaWindow *window,
                                       MetaWindowMenuType menu,
                                       MetaRectangle *rect)
{
    _cdos_base_wm_show_window_menu_for_rect (get_cdos_base_wm (), window, menu, rect);
}

static gboolean
cdos_plugin_xevent_filter (MetaPlugin *plugin, XEvent *xev)
{
#ifdef GLX_INTEL_swap_event
    CdosPlugin *cdos_base_plugin = CDOS_PLUGIN (plugin);

    if (cdos_base_plugin->have_swap_event && xev->type == (cdos_base_plugin->glx_event_base + GLX_BufferSwapComplete))
    {
        GLXBufferSwapComplete *swap_complete_event;
        swap_complete_event = (GLXBufferSwapComplete *)xev;

        /* Buggy early versions of the INTEL_swap_event implementation in Mesa
       * can send this with a ust of 0. Simplify life for consumers
       * by ignoring such events */
        if (swap_complete_event->ust != 0)
        {
        }
    }
#endif

    return FALSE;
}

static gboolean
cdos_plugin_keybinding_filter (MetaPlugin *plugin, MetaKeyBinding *binding)
{
    return _cdos_base_wm_filter_keybinding (get_cdos_base_wm (), binding);
}

static void
cdos_plugin_confirm_display_change (MetaPlugin *plugin)
{
    _cdos_base_wm_confirm_display_change (get_cdos_base_wm ());
}

static const MetaPluginInfo *
cdos_plugin_plugin_info (MetaPlugin *plugin)
{
    static const MetaPluginInfo info = {.name = "Cdos",
                                        .version = "0.1",
                                        .author = "Various",
                                        .license = "GPLv2+",
                                        .description = "Provides Cdos core functionality"};

    return &info;
}
