/**
 * Copyright (c) 2016-2018 CPU and Fundamental Software Research Center, CAS
 *
 * This software is published by the license of CPU-OS License, you can use and
 * distibuted this software under this License. See CPU-OS Liense for more detail.
 *
 * You should have receive a copy of CPU-OS License. If not, please contact us
 * by email <support_os@cpu-os.ac.cn>
 *
 **/
/* applet-recently-parser.c */

#include "cdos-app-system.h"
#include <libxml/parser.h>
#include <libxml/xpath.h>
#include <sys/stat.h>
#include "cdos-global.h"
#include "cdos-base-util.h"
#include "cdos-recently-parser.h"

static void cdos_recently_parser_finalize (GObject *obj);

static void cdos_recently_parser_dispose (GObject *obj);

static xmlNodePtr cdos_recently_parser_get_node (CdosRecentlyParser *self, xmlChar *prop_name, xmlChar *prop_val);


enum
{
    RECENT_CHANGED,
    LAST_SIGNAL
};

static guint signals[LAST_SIGNAL] = {0};

struct _CdosRecentlyParserPrivate
{
    xmlDocPtr doc;
    xmlNodePtr root_node;
    GList *list;
    // test
    guint timer_id;
    GSettings *settings;
};

#define RECENTLY_APP "/usr/share/found-desktop/data/newly-install.xml"

/* uncomment the following if you have defined any signals */
/* static guint signals[LAST_SIGNAL] = {0}; */

G_DEFINE_TYPE_WITH_PRIVATE (CdosRecentlyParser, cdos_recently_parser, G_TYPE_OBJECT);

const gchar *recently_encoding = "UTF-8";
// 参数为文件名、编码类型和枚举变量(下面一个或多个的组合）
// XML_PARSE_RECOVER = 1<<0, recover on errors 尝试修复错误语法
// XML_PARSE_NOERROR = 1<<5, suppress error reports 不输出错误日志
// XML_PARSE_NOWARNING = 1<<6, suppress warning reports 不输出警告日志
// XML_PARSE_PEDANTIC = 1<<7, pedantic error reporting 输出详细的错误日志
// XML_PARSE_NOERROR和XML_PARSE_NOWARNING的功能是“当运行命令行程序时不输出错误/警告日志”。
const int recently_options = XML_PARSE_PEDANTIC | XML_PARSE_RECOVER;

// test
static guint remove_interval = 0;

static xmlDocPtr
cdos_recently_parser_get_doc (CdosRecentlyParser *self)
{
    // 设置输出内容不要保留空白
    xmlKeepBlanksDefault (0);
    CdosRecentlyParserPrivate *priv =  cdos_recently_parser_get_instance_private (self);

    if (NULL == priv->doc)
    {
        if (g_file_test (RECENTLY_APP, G_FILE_TEST_EXISTS))
        {
            priv->doc = xmlReadFile (RECENTLY_APP, recently_encoding, recently_options);
        }
        else
        {
            g_debug ("Warning: cannot find user recently file.");
        }

        if (NULL == priv->doc)
        {
            g_debug ("Document is not parsed successfully.");
        }
    }

    return priv->doc;
}

static xmlNodePtr
cdos_recently_parser_get_root (CdosRecentlyParser *self)
{
    g_return_val_if_fail (CDOS_IS_RECENTLY_PARSER (self), NULL);
    // 设置输出内容不要保留空白
    xmlKeepBlanksDefault (0);
    CdosRecentlyParserPrivate *priv;
    priv = self->priv;

    if (NULL == priv->doc)
    {
        g_debug ("Cannot get root node: xmlDoc is NULL");
        return NULL;
    }

    if (NULL == priv->root_node)
    {
        priv->root_node = xmlDocGetRootElement (priv->doc);
        if (NULL == priv->root_node)
        {
            g_debug ("The document is empty.");
            xmlFreeDoc (priv->doc);
            return NULL;
        }
    }

    return priv->root_node;
}

static xmlNodePtr
cdos_recently_parser_get_node (CdosRecentlyParser *self, xmlChar *prop_name, xmlChar *prop_val)
{
    g_debug ("FIXME");
    return NULL;
    g_return_val_if_fail (CDOS_IS_RECENTLY_PARSER (self), NULL);
    CdosRecentlyParserPrivate *priv;
    priv = self->priv;
    xmlNodePtr root_node = priv->root_node;
    xmlNodePtr child = root_node->xmlChildrenNode, find_node = NULL;
    xmlChar *node_prop;
    while (child)
    {
        node_prop = xmlGetProp (child, BAD_CAST prop_name);
        if (!(xmlStrcmp (node_prop, prop_val)))
        {
            find_node = child;
            break;
        }
        child = child->next;
    }

    return find_node;
}

static gint
cdos_recently_parser_find_app_by_id (CdosRecentlyParser *self, const gchar *id)
{
    CdosRecentlyParserPrivate *priv = self->priv;
    GList *list = priv->list;
    CdosRecentlyAppData *data;
    guint index = 0;
    while (list)
    {
        data = list->data;
        if (g_strcmp0 (data->id, id) == 0)
        {
            return index;
        }
        ++index;
        list = list->next;
    }

    return -1;
}

// 删除结点之前，必须判断是否存在
static gboolean
cdos_recently_parser_remove_node (CdosRecentlyParser *self, xmlNodePtr cur_node)
{
    g_debug ("%s node will be delete.", cur_node->name);
    CdosRecentlyParserPrivate *priv = self->priv;
    xmlDocPtr doc = priv->doc;
    xmlNodePtr tmp_node = cur_node->next;
    xmlUnlinkNode (cur_node);
    xmlFreeNode (cur_node);
    cur_node = tmp_node;

    int res = xmlSaveFormatFile (RECENTLY_APP, doc, 4);
    if (res == -1)
    {
        g_warning ("xml save format file failed!");
        return FALSE;
    }
    return TRUE;
}

// ========== Public Method ==========
GList *
cdos_recently_parser_get_list (CdosRecentlyParser *self)
{
    g_return_val_if_fail (CDOS_IS_RECENTLY_PARSER (self), NULL);
    CdosRecentlyParserPrivate *priv;
    xmlNodePtr root_node = NULL, children = NULL;
    xmlChar *prop_id, *prop_last_install;
    CdosRecentlyAppData *data;
    priv = self->priv;
    if (NULL == priv->list)
    {
        root_node = cdos_recently_parser_get_root (self);
        if (NULL == root_node)
        {
            g_debug ("Get root node failed!");
            return NULL;
        }
        children = root_node->xmlChildrenNode;
        while (children)
        {
            prop_id = xmlGetProp (children, BAD_CAST "id");
            prop_last_install = xmlGetProp (children, BAD_CAST "last-install");
            data = g_new0 (CdosRecentlyAppData, 1);
            data->id = g_strdup ((gchar *)prop_id);
            data->last_install = g_strdup ((gchar *)prop_last_install);
            priv->list = g_list_prepend (priv->list, data);
            xmlFree (prop_id);
            xmlFree (prop_last_install);
            children = children->next;
        }
    }
    return priv->list;
}

GList *
cdos_recently_parser_get_app_list (CdosRecentlyParser *self)
{
    g_return_val_if_fail (CDOS_IS_RECENTLY_PARSER (self), NULL);
    CdosRecentlyParserPrivate *priv;
    priv = self->priv;
    GList *app_list = NULL, *l;
    CdosRecentlyAppData *data;
    CdosApp *app;
    CdosAppSystem *app_system;
    app_system = cdos_app_system_get_default ();
    if (priv->list)
    {
        l = priv->list;
        while (l != NULL)
        {
            data = l->data;
            app = cdos_app_system_lookup_app (app_system, data->id);
            if(app != NULL)
            {
                app_list = g_list_append (app_list, app);
            }
            l = l->next;
        }
    }
    return app_list;
}

static void
cdos_recently_parser_add_app (CdosRecentlyParser *self, const gchar *desktop_id, glong cur_time)
{
    CdosRecentlyParserPrivate *priv;
    CdosRecentlyAppData *data;
    priv = self->priv;
    xmlNodePtr root = NULL;
    xmlNodePtr has_node = cdos_recently_parser_get_node (self, BAD_CAST "id", BAD_CAST desktop_id);
    gchar *time_str = g_strdup_printf ("%ld", cur_time);
    if (NULL != has_node)
    {
        // update xml
        xmlAttrPtr attr = xmlSetProp (has_node, BAD_CAST "last-install", BAD_CAST time_str);
        if (NULL != attr)
        {
            int res = xmlSaveFormatFile (RECENTLY_APP, priv->doc, 4);
            if (res == -1)
            {
                g_debug ("xml save format file failed!");
                return;
            }
        }

        // update glist
        gint index = cdos_recently_parser_find_app_by_id (self, desktop_id);
        if (index < 0)
        {
            g_debug ("xml node found, but cannot found in list");
            return;
        }

        GList *list = g_list_nth (priv->list, index);
        if (list)
        {
            data = list->data;
            data->last_install = g_strdup (time_str);
        }
    }
    else
    {
        // update xml
        root = cdos_recently_parser_get_root (self);
        if (NULL == root)
        {
            return;
        }
        xmlNodePtr new_node = xmlNewChild (root, NULL, BAD_CAST "application", NULL);
        xmlNewProp (new_node, BAD_CAST "id", BAD_CAST desktop_id);
        xmlNewProp (new_node, BAD_CAST "last-install", BAD_CAST time_str);
        int res = xmlSaveFormatFile (RECENTLY_APP, priv->doc, 4);
        if (res == -1)
        {
            g_debug ("xml save format file failed!");
            return;
        }

        // update glist
        data = g_new0 (CdosRecentlyAppData, 1);
        data->id = g_strdup (desktop_id);
        data->last_install = g_strdup (time_str);
        priv->list = g_list_prepend (priv->list, data);
    }
}

void
cdos_recently_parser_remove_app (CdosRecentlyParser *self, const gchar *desktop_id)
{
    g_return_if_fail (CDOS_IS_RECENTLY_PARSER (self));
    g_return_if_fail (NULL != desktop_id);
    CdosRecentlyParserPrivate *priv;

    gboolean res = FALSE;
    xmlChar *node_prop = NULL;
    priv = self->priv;

    xmlNodePtr root = cdos_recently_parser_get_root (self);
    xmlNodePtr child = root->xmlChildrenNode;

    while (child)
    {
        node_prop = xmlGetProp (child, BAD_CAST "id");
        if (!xmlStrcmp (node_prop, BAD_CAST desktop_id))
        {
            // update xml
            res = cdos_recently_parser_remove_node (self, child);

            // update glist

            if (res)
            {
                gint index = cdos_recently_parser_find_app_by_id (self, desktop_id);
                if (index < 0)
                {
                    g_debug ("cannot find, but xml delete ok!\n");
                    return;
                }
                GList *tmp = g_list_nth (priv->list, index);
                if (NULL == tmp)
                {
                    return;
                }
                priv->list = g_list_remove (priv->list, tmp->data);
            }
            break;
        }
        child = child->next;
    }
}


// ========== GObject Method ==========
static void
cdos_recently_parser_class_init (CdosRecentlyParserClass *klass)
{
    GObjectClass *gobject_class;
    gobject_class = (GObjectClass *)klass;

    gobject_class->finalize = cdos_recently_parser_finalize;

    gobject_class->dispose = cdos_recently_parser_dispose;
    signals[RECENT_CHANGED] = g_signal_new ("recent-changed",
                                            G_TYPE_FROM_CLASS (klass),
                                            G_SIGNAL_RUN_LAST,
                                            G_STRUCT_OFFSET (CdosRecentlyParserClass, recent_changed),
                                            NULL,
                                            NULL,
                                            NULL,
                                            G_TYPE_NONE,
                                            0);
}

static long
get_time (void)
{
    guint64 t = g_get_real_time ();
    return t / 1000000;
}

static void
test_installed_changed (CdosAppSystem *app_sys, CdosApp *app, gpointer user_data)
{
    CdosRecentlyParser *self = CDOS_RECENTLY_PARSER (user_data);
    if (NULL == app || !CDOS_IS_APP (app))
    {
        return;
    }
    const gchar *app_name = cdos_app_get_name (app);
    if (!g_desktop_app_info_get_nodisplay (cdos_app_get_app_info (app)) && !g_str_has_prefix (app_name, "Uninstall"))
    {
        cdos_recently_parser_add_app (self, cdos_app_get_id (app), get_time ());
        g_signal_emit (self, signals[RECENT_CHANGED], 0);
    }
}

static void
test_removed_changed (CdosAppSystem *app_sys, const gchar *desktop_id, gpointer user_data)
{
    gint index;
    CdosRecentlyParser *self = CDOS_RECENTLY_PARSER (user_data);

    index = cdos_recently_parser_find_app_by_id (self, desktop_id);
    g_debug ("%s %s", __func__, desktop_id);
    if (index != -1)
    {
        cdos_recently_parser_remove_app (self, desktop_id);
        g_signal_emit (self, signals[RECENT_CHANGED], 0);
    }
}

// test
//如果新安装的应用超过remove_interval指定的时间，将从“最近添加”中移除
static gboolean
remove_app_timeout (gpointer user_data)
{
    g_debug ("remove_interval= %u", remove_interval); //////调试
    CdosRecentlyParser *self;
    GList *apps_list = NULL;
    CdosRecentlyAppData *data;
    gchar *install_time;
    long install_time_to_long;
    long current_time;
    gchar *ptr;
    gboolean should_emit = FALSE;

    self = (CdosRecentlyParser *)user_data;
    current_time = get_time ();
    apps_list = cdos_recently_parser_get_list (self);
    while (apps_list)
    {
        data = apps_list->data;

        install_time = data->last_install;
        install_time_to_long = strtol (install_time, &ptr, 10);

        if (current_time - install_time_to_long >= remove_interval)
        {
            should_emit = TRUE;
            cdos_recently_parser_remove_app (self, data->id);
        }

        apps_list = apps_list->next;
    }

    if (should_emit)
    {
        g_signal_emit (self, signals[RECENT_CHANGED], 0);
    }

    return G_SOURCE_CONTINUE;
}

static void
end_timer (CdosRecentlyParser *self)
{
    CdosRecentlyParserPrivate *priv;

    priv = self->priv;

    if (priv->timer_id > 0)
    {
        g_source_remove (priv->timer_id);
        priv->timer_id = 0;
    }
}

static void
start_timer (CdosRecentlyParser *self, GSourceFunc function, gpointer data)
{
    CdosRecentlyParserPrivate *priv;

    priv = self->priv;
    end_timer (self);

    priv->timer_id = g_timeout_add_seconds (10, function, data);
}

static guint
calculate_seconds (CdosRecentlyParser *self)
{
    CdosRecentlyParserPrivate *priv;
    gint days;

    priv = self->priv;
    days = g_settings_get_int (priv->settings, "newly-installed-remove-days");

    // TODO 改成需要的天数
    return 60 * 60 * 24 * days;
}

static void
on_settings_changed (GSettings *settings, gchar *key, gpointer user_data)
{
    // gint days;
    CdosRecentlyParser *self;

    self = (CdosRecentlyParser *)user_data;
    // days = g_settings_get_int (settings, "newly-installed-remove-days");
    // remove_interval = 60 * days;
    remove_interval = calculate_seconds (self);
    // GSettings中从“最近添加”移除时的天数设置发生改变，重启定时器
    start_timer (self, remove_app_timeout, self);
}

static void
cdos_recently_parser_init (CdosRecentlyParser *self)
{
    CdosRecentlyParserPrivate *priv;
    CdosGlobal *global;
    // test
    // gint days;
    priv = cdos_recently_parser_get_instance_private (self);
    self->priv = priv;
    priv->doc = cdos_recently_parser_get_doc (self);
    priv->root_node = cdos_recently_parser_get_root (self);
    priv->list = cdos_recently_parser_get_list (self);
    // test
    global = cdos_global_get ();
    priv->settings = cdos_global_get_settings (global);
    g_signal_connect (priv->settings, "changed::newly-installed-remove-days", G_CALLBACK (on_settings_changed), self);
    // days = g_settings_get_int (priv->settings, "newly-installed-remove-days");
    // remove_interval = 60 * days;
    remove_interval = calculate_seconds (self);
    start_timer (self, remove_app_timeout, self);
}

static void
cdos_recently_parser_dispose (GObject *obj)
{
    G_OBJECT_CLASS (cdos_recently_parser_parent_class)->dispose (obj);
    CdosRecentlyParser *self = CDOS_RECENTLY_PARSER (obj);
    CdosRecentlyParserPrivate *priv = self->priv;
    if (priv->list)
    {
        g_list_free (priv->list);
    }

    if (priv->doc)
    {
        xmlFree (priv->doc);
    }

    if (priv->root_node)
    {
        xmlFree (priv->root_node);
    }

    xmlCleanupParser ();

    // test
    end_timer (self);
}

static void
cdos_recently_parser_finalize (GObject *obj)
{
    G_OBJECT_CLASS (cdos_recently_parser_parent_class)->finalize (obj);
}

CdosRecentlyParser *
cdos_recently_parser_new (void)
{
    static CdosRecentlyParser *instance = NULL;
    CdosAppSystem *app_sys = NULL;

    if (NULL == instance)
    {
        instance = g_object_new (CDOS_TYPE_RECENTLY_PARSER, NULL);
        app_sys = cdos_app_system_get_default ();
        g_signal_connect (app_sys, "installed-changed", G_CALLBACK (test_installed_changed), instance);
        g_signal_connect (app_sys, "removed-changed", G_CALLBACK (test_removed_changed), instance);
    }

    return instance;
}
