/** 
 * Copyright (c) 2016-2018 CPU and Fundamental Software Research Center, CAS
 * * Based on code from cinnamon-screenshot.c in Cinnamon. 
 * Copyright (c) 1997-2003 by The XFree86 Project, Inc.
 * This software is published by the license of CPU-OS License, you can use and 
 * distibuted this software under this License. See CPU-OS Liense for more detail.
 *
 * You should have receive a copy of CPU-OS License. If not, please contact us 
 * by email <support_os@cpu-os.ac.cn>  
 * 
 **/
/* -*- mode: C; c-file-style: "gnu"; indent-tabs-mode: nil; -*- */

#include <meta/display.h>
#include <meta/util.h>
#include <meta/meta-shaped-texture.h>
#include <meta/meta-cursor-tracker.h>

#include "cdos-global.h"
#include "cdos-screenshot.h"
#include "cdos-sound-manager.h"
#include "cdos-start.h"

typedef struct _CdosScreenshotPrivate CdosScreenshotPrivate;

struct _CdosScreenshotClass
{
  GObjectClass parent_class;
};

struct _CdosScreenshot
{
  GObject parent_instance;

  CdosScreenshotPrivate *priv;
};

/* Used for async screenshot grabbing */
struct _CdosScreenshotPrivate
{
  CdosGlobal *global;

  char *filename;
  char *filename_used;

  cairo_surface_t *image;
  cairo_rectangle_int_t screenshot_area;

  gboolean include_cursor;
  gboolean include_frame;

  CdosScreenshotCallback callback;

  GdkPixbuf *thumb;
};

G_DEFINE_TYPE_WITH_PRIVATE (CdosScreenshot, cdos_screenshot, G_TYPE_OBJECT);

static void
cdos_screenshot_class_init (CdosScreenshotClass *screenshot_class)
{
  (void) screenshot_class;
}

static void
cdos_screenshot_init (CdosScreenshot *screenshot)
{
  screenshot->priv = cdos_screenshot_get_instance_private (screenshot);
  screenshot->priv->global = cdos_global_get ();
}

static void
on_screenshot_written (GObject *source,
                       GAsyncResult *result,
                       gpointer user_data)
{
  CdosScreenshot *screenshot = CDOS_SCREENSHOT (source);
  CdosScreenshotPrivate *priv = screenshot->priv;
  if (priv->callback)
    priv->callback (screenshot,
                    g_task_propagate_boolean (G_TASK (result), NULL),
                    &priv->screenshot_area);

  g_clear_pointer (&priv->image, cairo_surface_destroy);
  g_clear_pointer (&priv->filename, g_free);

  meta_enable_unredirect_for_display (cdos_global_get_display (priv->global));  

  ClutterContent *image;
  ClutterActor *actor;
  StBin *bin;
  guint8 * data;
  GdkPixbuf *thumb = priv->thumb;
  actor = clutter_actor_new ();
  image = clutter_image_new ();
  data = gdk_pixbuf_get_pixels (thumb);
  CoglPixelFormat format = gdk_pixbuf_get_has_alpha (thumb) ?
                             COGL_PIXEL_FORMAT_RGBA_8888 : COGL_PIXEL_FORMAT_RGB_888;
  gint dst_width = gdk_pixbuf_get_width (thumb);
  gint dst_height = gdk_pixbuf_get_height (thumb);
  gint rowstride = gdk_pixbuf_get_rowstride (thumb);
  clutter_image_set_data ((ClutterImage *)image,
                          data,
                          format, dst_width, dst_height, rowstride,
                          NULL);
  clutter_actor_set_size (actor, dst_width, dst_height);
  clutter_actor_set_y_align (actor, CLUTTER_ACTOR_ALIGN_CENTER);
  clutter_actor_set_content (actor, image);
  g_object_unref (thumb);

  bin = (StBin *)st_bin_new ();
  st_bin_set_child (bin, actor);
  st_bin_set_fill (bin, TRUE, TRUE);
  if (dst_width < dst_height)
      dst_width = dst_height;
  clutter_actor_set_size ((ClutterActor *)bin, dst_width, dst_width);

  gchar *uri;
  if (priv->filename_used[0] == '/')
      uri = g_strconcat ("file://", priv->filename_used, NULL);
  else
      uri = g_strdup (priv->filename_used);
  cdos_notify_info ((ClutterActor *)bin, _("Screenshot saved"), uri);
  g_free (uri);
}

/* called in an I/O thread */
static GOutputStream *
get_stream_for_unique_path (const gchar *path,
                            const gchar *filename,
                            gchar **filename_used)
{
  GOutputStream *stream;
  GFile *file;
  gchar *real_path, *real_filename, *name, *ptr;
  gint idx;

  ptr = g_strrstr (filename, ".png");

  if (ptr != NULL)
    real_filename = g_strndup (filename, ptr - filename);
  else
    real_filename = g_strdup (filename);

  idx = 0;
  real_path = NULL;

  do
    {
      if (idx == 0)
        name = g_strdup_printf ("%s.png", real_filename);
      else
        name = g_strdup_printf ("%s-%d.png", real_filename, idx);

      real_path = g_build_filename (path, name, NULL);
      g_free (name);

      file = g_file_new_for_path (real_path);
      stream = G_OUTPUT_STREAM (g_file_create (file, G_FILE_CREATE_NONE, NULL, NULL));
      g_object_unref (file);

      if (stream != NULL)
        *filename_used = real_path;
      else
        g_free (real_path);

      idx++;
    }
  while (stream == NULL);

  g_free (real_filename);

  return stream;
}

/* called in an I/O thread */
static GOutputStream *
get_stream_for_filename (const gchar *filename,
                         gchar **filename_used)
{
  const gchar *path;

  path = g_get_user_special_dir (G_USER_DIRECTORY_PICTURES);
  if (!g_file_test (path, G_FILE_TEST_EXISTS))
    {
      path = g_get_home_dir ();
      if (!g_file_test (path, G_FILE_TEST_EXISTS))
        return NULL;
    }

  return get_stream_for_unique_path (path, filename, filename_used);
}

static GOutputStream *
prepare_write_stream (const gchar *filename,
                      gchar **filename_used)
{
  GOutputStream *stream;
  GFile *file;
  gchar *filename1;
  gchar *time;

  GDateTime *date = g_date_time_new_now_local ();
  time = g_date_time_format (date, "%F_%H-%M-%S");

  if (!filename)
    filename = _("screenshot");
  filename1 = g_strconcat (time, filename, NULL);

  if (g_path_is_absolute (filename1))
    {
      file = g_file_new_for_path (filename1);
      *filename_used = g_strdup (filename1);
      stream = G_OUTPUT_STREAM (g_file_replace (file, NULL, FALSE, G_FILE_CREATE_NONE, NULL, NULL));
      g_object_unref (file);
    }
  else
    {
      stream = get_stream_for_filename (filename1, filename_used);
    }
  g_date_time_unref (date); 
  g_free (time);
  g_free (filename1);

  return stream;
}

static void
write_screenshot_thread (GTask          *result,
                         gpointer        object,
                         gpointer        task_data,
                         GCancellable   *cancellable)
{
  cdos_sound_manager_play ("screen-capture");

  cairo_status_t status;
  GOutputStream *stream;
  CdosScreenshot *screenshot = CDOS_SCREENSHOT (object);
  CdosScreenshotPrivate *priv;

  g_assert (screenshot != NULL);

  priv = screenshot->priv;

  /*
   * s3g 显卡驱动有bug，底层不支持 BGRA 格式。
   * 需要将获取到的像素数据 BR 部分交换。
   */
  /*
   * 纠正：这是cogl旧版本的bug。已在cogl中修复
   */
#if G_BYTE_ORDER == G_LITTLE_ENDIAN
#if 0
  uint8_t *data = cairo_image_surface_get_data (priv->image);
  int stride = cairo_image_surface_get_stride (priv->image);
  int width = cairo_image_surface_get_width (priv->image);
  if (stride != width * 4)
      goto _error;

  int height = cairo_image_surface_get_height (priv->image);
  for (int i = 0; i < height; i++)
  {
      for (int j = 0; j < width; j++)
      {
          uint8_t *d = data + i * stride + j * 4;
          uint8_t *r = d;
          uint8_t *b = d + 2;
          uint8_t t = *r;
          *r = *b;
          *b = t;
      }
  }
  cairo_surface_mark_dirty (priv->image);
#endif
#endif

  stream = prepare_write_stream (priv->filename,
                                 &priv->filename_used);
  if (stream == NULL)
    status = CAIRO_STATUS_FILE_NOT_FOUND;
  else
    {
      GdkPixbuf *pixbuf;
      int width, height;
      width = cairo_image_surface_get_width (priv->image);
      height = cairo_image_surface_get_height (priv->image);

      pixbuf = gdk_pixbuf_get_from_surface (priv->image,
                                            0, 0,
                                            width, height);

      if (gdk_pixbuf_save_to_stream (pixbuf, stream, "png", NULL, NULL,
                                    "tEXt::Software", "cdos-screenshot", NULL))
        {
          status = CAIRO_STATUS_SUCCESS;
          GdkPixbuf *thumb;
          gfloat scale_x, scale_y;
          gfloat size = 48.0;

          if (width > height)
            {
              scale_x = size / width;
              scale_y = scale_x;
            }
          else
            {
              scale_y = size / height;
              scale_x = scale_y;
            }

          thumb = gdk_pixbuf_scale_simple (pixbuf,
                                           (gint)(width * scale_x),
                                           (gint)(height * scale_y),
                                           GDK_INTERP_BILINEAR);
          screenshot->priv->thumb = thumb;
        }
      else
        status = CAIRO_STATUS_WRITE_ERROR;

      g_object_unref (pixbuf);
    }


  g_task_return_boolean (result, status == CAIRO_STATUS_SUCCESS);

  g_clear_object (&stream);

  /*
  status = cairo_surface_write_to_png (priv->image, priv->filename);
  g_task_return_boolean (result, status == CAIRO_STATUS_SUCCESS);
  */
}

static cairo_surface_t *
cdos_util_composite_capture_images (ClutterCapture  *captures,
                                    int              n_captures,
                                    int              x,
                                    int              y,
                                    int              width,
                                    int              height)
{
  int i;
  cairo_format_t format;
  cairo_surface_t *image;
  cairo_t *cr;

  format = cairo_image_surface_get_format (captures[0].image);
  image = cairo_image_surface_create (format, width, height);

  cr = cairo_create (image);

  for (i = 0; i < n_captures; i++)
    {
      ClutterCapture *capture = &captures[i];
      double capture_scale = 1.0;

      /*
       * Ignore capture regions with scale other than 1 for now; mutter can't
       * produce them yet, so there is no way to test them.
       */
      cairo_surface_get_device_scale (capture->image, &capture_scale, NULL);
      if ((int) capture_scale != 1)
        continue;

      cairo_save (cr);

      cairo_translate (cr,
                       capture->rect.x - x,
                       capture->rect.y - y);
      cairo_set_source_surface (cr, capture->image, 0, 0);
      cairo_paint (cr);

      cairo_restore (cr);
    }
  cairo_destroy (cr);

  return image;
}

static void
do_grab_screenshot (CdosScreenshot   *screenshot,
                    ClutterStage     *stage,
                    int               x,
                    int               y,
                    int               width,
                    int               height)
{
  CdosScreenshotPrivate *priv = screenshot->priv;
  ClutterCapture *captures;
  int n_captures;
  int i;

  clutter_stage_capture (stage, FALSE,
                         &(cairo_rectangle_int_t) {
                           .x = x,
                           .y = y,
                           .width = width,
                           .height = height
                         },
                         &captures,
                         &n_captures);
  if (n_captures == 0)
    return;
  else if (n_captures == 1)
    priv->image = cairo_surface_reference (captures[0].image);
  else
    priv->image = cdos_util_composite_capture_images (captures,
                                                      n_captures,
                                                      x, y,
                                                      width, height);

  for (i = 0; i < n_captures; i++)
    cairo_surface_destroy (captures[i].image);

  g_free (captures);
}

static void
_draw_cursor_image (MetaCursorTracker     *tracker,
                    cairo_surface_t       *surface,
                    cairo_rectangle_int_t  area)
{
  CoglTexture *texture;
  int width, height;
  int stride;
  guint8 *data;
  cairo_surface_t *cursor_surface;
  cairo_region_t *screenshot_region;
  cairo_t *cr;
  int x, y;
  int xhot, yhot;

  texture = meta_cursor_tracker_get_sprite (tracker);
  if (!texture)
    return;

  screenshot_region = cairo_region_create_rectangle (&area);
  meta_cursor_tracker_get_pointer (tracker, &x, &y, NULL);

  if (!cairo_region_contains_point (screenshot_region, x, y))
    {
      cairo_region_destroy (screenshot_region);
      return;
    }

  meta_cursor_tracker_get_hot (tracker, &xhot, &yhot);
  width = cogl_texture_get_width (texture);
  height = cogl_texture_get_height (texture);
  stride = 4 * width;
  data = g_new (guint8, stride * height);
  cogl_texture_get_data (texture, CLUTTER_CAIRO_FORMAT_ARGB32, stride, data);

  /* FIXME: cairo-gl? */
  cursor_surface = cairo_image_surface_create_for_data (data,
                                                        CAIRO_FORMAT_ARGB32,
                                                        width, height,
                                                        stride);
  cr = cairo_create (surface);
  cairo_set_source_surface (cr,
                            cursor_surface,
                            x - xhot - area.x,
                            y - yhot - area.y);
  cairo_paint (cr);

  cairo_destroy (cr);
  cairo_surface_destroy (cursor_surface);
  cairo_region_destroy (screenshot_region);
  g_free (data);
}

static void
grab_screenshot (ClutterActor *stage,
                 CdosScreenshot *screenshot)
{
  MetaDisplay *display;
  MetaCursorTracker *tracker;
  int width, height;
  GTask *result;
  CdosScreenshotPrivate *priv = screenshot->priv;

  display = cdos_global_get_display (priv->global);
  meta_display_get_size (display, &width, &height);

  do_grab_screenshot (screenshot, CLUTTER_STAGE (stage), 0, 0, width, height);

  if (meta_display_get_n_monitors (display) > 1)
    {
      cairo_region_t *screen_region = cairo_region_create ();
      cairo_region_t *stage_region;
      MetaRectangle monitor_rect;
      cairo_rectangle_int_t stage_rect;
      int i;
      cairo_t *cr;

      for (i = meta_display_get_n_monitors (display) - 1; i >= 0; i--)
        {
          meta_display_get_monitor_geometry (display, i, &monitor_rect);
          cairo_region_union_rectangle (screen_region, (const cairo_rectangle_int_t *) &monitor_rect);
        }

      stage_rect.x = 0;
      stage_rect.y = 0;
      stage_rect.width = width;
      stage_rect.height = height;

      stage_region = cairo_region_create_rectangle ((const cairo_rectangle_int_t *) &stage_rect);
      cairo_region_xor (stage_region, screen_region);
      cairo_region_destroy (screen_region);

      cr = cairo_create (priv->image);

      for (i = 0; i < cairo_region_num_rectangles (stage_region); i++)
        {
          cairo_rectangle_int_t rect;
          cairo_region_get_rectangle (stage_region, i, &rect);
          cairo_rectangle (cr, (double) rect.x, (double) rect.y, (double) rect.width, (double) rect.height);
          cairo_fill (cr);
        }

      cairo_destroy (cr);
      cairo_region_destroy (stage_region);
    }

  priv->screenshot_area.x = 0;
  priv->screenshot_area.y = 0;
  priv->screenshot_area.width = width;
  priv->screenshot_area.height = height;

  if (priv->include_cursor)
    {
      tracker = meta_cursor_tracker_get_for_display (display);
      _draw_cursor_image (tracker, priv->image, priv->screenshot_area);
    }

  g_signal_handlers_disconnect_by_func (stage, (void *)grab_screenshot, (gpointer)screenshot);

  result = g_task_new (screenshot, NULL, on_screenshot_written, NULL);
  g_task_run_in_thread (result, write_screenshot_thread);
  g_object_unref (result);
}

static void
grab_area_screenshot (ClutterActor *stage,
                      CdosScreenshot *screenshot)
{
  GTask *result;
  CdosScreenshotPrivate *priv = screenshot->priv;

  do_grab_screenshot (screenshot,
                      CLUTTER_STAGE (stage),
                      priv->screenshot_area.x,
                      priv->screenshot_area.y,
                      priv->screenshot_area.width,
                      priv->screenshot_area.height);

  g_signal_handlers_disconnect_by_func (stage, (void *)grab_area_screenshot, (gpointer)screenshot);
  result = g_task_new (screenshot, NULL, on_screenshot_written, NULL);
  g_task_run_in_thread (result, write_screenshot_thread);
  g_object_unref (result);
}

static void
grab_window_screenshot (ClutterActor *stage,
                        CdosScreenshot *screenshot)
{
  CdosScreenshotPrivate *priv = screenshot->priv;
  GTask *result;
  MetaDisplay *display = cdos_global_get_display (priv->global);
  MetaCursorTracker *tracker;
  MetaWindow *window = meta_display_get_focus_window (display);
  ClutterActor *window_actor;
  gfloat actor_x, actor_y;
  MetaShapedTexture *stex;
  MetaRectangle rect;
  cairo_rectangle_int_t clip;

  window_actor = CLUTTER_ACTOR (meta_window_get_compositor_private (window));
  clutter_actor_get_position (window_actor, &actor_x, &actor_y);

  meta_window_get_frame_rect (window, &rect);

  if (!priv->include_frame)
    meta_window_frame_rect_to_client_rect (window, &rect, &rect);

  priv->screenshot_area.x = rect.x;
  priv->screenshot_area.y = rect.y;
  clip.x = rect.x - (gint) actor_x;
  clip.y = rect.y - (gint) actor_y;

  clip.width = priv->screenshot_area.width = rect.width;
  clip.height = priv->screenshot_area.height = rect.height;

  stex = META_SHAPED_TEXTURE (meta_window_actor_get_texture (META_WINDOW_ACTOR (window_actor)));
  priv->image = meta_shaped_texture_get_image (stex, &clip);

  //settings = g_settings_new (A11Y_APPS_SCHEMA);
  if (priv->include_cursor) // && !g_settings_get_boolean (settings, MAGNIFIER_ACTIVE_KEY))
    {
      tracker = meta_cursor_tracker_get_for_display (display);
      _draw_cursor_image (tracker, priv->image, priv->screenshot_area);
    }
  //g_object_unref (settings);

  g_signal_handlers_disconnect_by_func (stage, (void *)grab_window_screenshot, (gpointer)screenshot);
  result = g_task_new (screenshot, NULL, on_screenshot_written, NULL);
  g_task_run_in_thread (result, write_screenshot_thread);
  g_object_unref (result);
}

/**
 * cdos_screenshot_screenshot:
 * @screenshot: CdosScreenshot对象
 * @include_cursor: 是否包含光标
 * @filename: 截图的文件名
 * @callback: 异步捕获时返回成功或失败时调用该函数
 *
 * 全屏截图，保存为png图像
 *
 */
void
cdos_screenshot_screenshot (CdosScreenshot *screenshot,
                             gboolean include_cursor,
                             const char *filename,
                             CdosScreenshotCallback callback)
{
  ClutterActor *stage;
  CdosScreenshotPrivate *priv = screenshot->priv;

  priv->filename = g_strdup (filename);
  priv->callback = callback;
  priv->include_cursor = include_cursor;

  stage = CLUTTER_ACTOR (cdos_global_get_stage (priv->global));

  meta_disable_unredirect_for_display (cdos_global_get_display (priv->global));

  g_signal_connect_after (stage, "paint", G_CALLBACK (grab_screenshot), (gpointer)screenshot);

  clutter_actor_queue_redraw (stage);
}

/**
 * cdos_screenshot_screenshot_area:
 * @screenshot: CdosScreenshot对象
 * @x: 区域X坐标
 * @y: 区域Y坐标
 * @width: 区域宽度
 * @height: 区域高度
 * @filename: 截图的文件名
 * @callback: 异步捕获时返回成功或失败时调用的函数
 *
 * Takes a screenshot of the passed in area and saves it
 * in @filename as png image.
 *
 */
void
cdos_screenshot_screenshot_area (CdosScreenshot *screenshot,
                                  gboolean include_cursor,
                                  int x,
                                  int y,
                                  int width,
                                  int height,
                                  const char *filename,
                                  CdosScreenshotCallback callback)
{
  ClutterActor *stage;
  CdosScreenshotPrivate *priv = screenshot->priv;

  priv->filename = g_strdup (filename);
  priv->screenshot_area.x = x;
  priv->screenshot_area.y = y;
  priv->screenshot_area.width = width;
  priv->screenshot_area.height = height;
  priv->callback = callback;
  priv->include_cursor = include_cursor;

  stage = CLUTTER_ACTOR (cdos_global_get_stage (priv->global));

  g_signal_connect_after (stage, "paint", G_CALLBACK (grab_area_screenshot), (gpointer)screenshot);

  clutter_actor_queue_redraw (stage);
}

/**
 * cdos_screenshot_screenshot_window:
 * @screenshot: CdosScreenshot对象
 * @include_frame: 是否包含帧
 * @include_cursor: 是否包含光标
 * @filename: 截图的文件名
 * @callback: 异步捕获时返回成功或失败时调用的函数
 *
 * 聚焦的窗口截图，保存为png图像
 *
 */
void
cdos_screenshot_screenshot_window (CdosScreenshot *screenshot,
                                    gboolean include_frame,
                                    gboolean include_cursor,
                                    const char *filename,
                                    CdosScreenshotCallback callback)
{
  CdosScreenshotPrivate *priv = screenshot->priv;

  MetaDisplay *display = cdos_global_get_display (cdos_global_get ());
  MetaWindow *window = meta_display_get_focus_window (display);

  if (filename != NULL || !window)
    {
	  if(callback)
		callback (screenshot, FALSE, NULL);
	  return;
    }

  ClutterActor	*stage;
  stage = CLUTTER_ACTOR (cdos_global_get_stage (priv->global));
  priv->filename = g_strdup (filename);
  priv->callback = callback;
  priv->include_frame = include_frame;
  priv->include_cursor = include_cursor;
  
  meta_disable_unredirect_for_display (cdos_global_get_display (cdos_global_get ()));

  g_signal_connect_after (stage, "paint", G_CALLBACK(grab_window_screenshot), (gpointer)screenshot);

  clutter_actor_queue_redraw (stage);

}

CdosScreenshot *
cdos_screenshot_new (void)
{
  return g_object_new (CDOS_TYPE_SCREENSHOT, NULL);
}
