#include <stdlib.h>
#include <grp.h>
#include <locale.h>
#include <pwd.h>
#include <security/pam_appl.h>

#undef G_LOG_DOMAIN
#define G_LOG_DOMAIN "cdos-user-verify"
#include <glib.h>
#include <gio/gio.h>
#include <glib/gi18n.h>
#include <glib-object.h>


/* 
 * RegisterInfo:注册服务名称和用户名
 * BeginVerification：开始认证
 */
static const gchar *introspect_xml =
    "<node name=\"/org/cdos/ScreenSaver/UserVerifier\">"
    "   <interface name=\"org.cdos.ScreenSaver.UserVerifier\">"
    "    <method name=\"RegistVerifyInfo\">"
    "      <arg name=\"service_name\" direction=\"in\" type=\"s\"/>"
    "      <arg name=\"username\" direction=\"in\" type=\"s\"/>"
    "      <arg name=\"result\" direction=\"out\" type=\"b\"/>"
    "    </method>"
    "    <method name=\"BeginVerification\">"
    "      <arg name=\"result\" direction=\"out\" type=\"b\"/>"
    "    </method>"
    "    <method name=\"AnswerQuery\">"
    "      <arg name=\"answer\" direction=\"in\" type=\"s\"/>"
    "      <arg name=\"result\" direction=\"out\" type=\"b\"/>"
    "    </method>"
    "    <method name=\"GetAccountLocked\">"
    "      <arg name=\"locked\" direction=\"out\" type=\"b\"/>"
    "    </method>"
    "    <method name=\"Cancel\">"
    "      <arg name=\"force\" direction=\"in\" type=\"b\"/>"
    "      <arg name=\"result\" direction=\"out\" type=\"b\"/>"
    "    </method>"
    "    <signal name=\"Error\">"
    "      <arg name=\"error\" type=\"s\"/>"
    "    </signal>"
    "    <signal name=\"Message\">"
    "      <arg name=\"message\" type=\"s\"/>"
    "    </signal>"
    "    <signal name=\"InfoQuery\">"
    "      <arg name=\"query\" type=\"s\"/>"
    "    </signal>"
    "    <signal name=\"SecretInfoQuery\">"
    "      <arg name=\"query\" type=\"s\"/>"
    "    </signal>"
    "    <signal name=\"VerificationFailed\">"
    "      <arg name=\"error\" type=\"s\"/>"
    "    </signal>"
    "    <signal name=\"VerificationComplete\">"
    "    </signal>"
    "    <signal name=\"LockStateChanged\">"
    "      <arg name=\"locked\" type=\"b\"/>"
    "    </signal>"
    "  </interface>"
    "</node>";

// clang-format off
typedef enum {
    IDLE,
    BUSY
} VerifyStatus;
// clang-format on

typedef struct _CdosUserVerifier
{
    GMainLoop *loop;
    GDBusConnection *connection;
    GDBusNodeInfo *dbus_node_info;
    guint bus_id;

    GThread *verify_thread;
    GMainLoop *verify_loop;
    GCond cond;
    GMutex mutex;

    pam_handle_t *pam_handle;
    char *username;
    gboolean require_password;
    char *service;
    char *answer;

    guint timeout_id;
    guint locked_id;
    guint idle_id;
    guint wait_timeout_id;
    VerifyStatus status;
    gboolean wait_timeout;
    gboolean canceled;
    gboolean lock_state;
} CdosUserVerifier;

static void handle_method_call (GDBusConnection *connection G_GNUC_UNUSED,
                                const gchar *sender,
                                const gchar *object_path G_GNUC_UNUSED,
                                const gchar *interface,
                                const gchar *method,
                                GVariant *parameters,
                                GDBusMethodInvocation *invocation,
                                CdosUserVerifier *self);

static const GDBusInterfaceVTable interface_vtable = {
    (GDBusInterfaceMethodCallFunc)handle_method_call,
    NULL,
    NULL,
};

static void 
verify_pam_stop (CdosUserVerifier *self, int status)
{
    int retval;
    if (!self->pam_handle)
        return;
    if ((retval = pam_end (self->pam_handle, status)) != PAM_SUCCESS)
    {
        g_warning ("Pam end error: %s", pam_strerror (self->pam_handle, retval));
    }
    self->pam_handle = NULL; 
}

static void
cdos_user_verifier_free (CdosUserVerifier *self)
{
    g_mutex_unlock (&self->mutex);

    g_cond_clear (&self->cond);
    g_mutex_clear (&self->mutex);

    verify_pam_stop (self, 0);

    if (self->timeout_id)
    {
        g_source_remove (self->timeout_id);
        self->timeout_id = 0;
    }
    if (self->locked_id)
    {
        g_source_remove (self->locked_id);
        self->locked_id = 0;
    }
    if (self->idle_id)
    {
        g_source_remove (self->idle_id);
        self->idle_id = 0;
    }
    if (self->wait_timeout_id)
    {
        g_source_remove (self->wait_timeout_id);
        self->wait_timeout_id = 0;
    }

    g_free (self->username);
    g_free (self->service);
    g_free (self->answer);

    if (self->connection)
    {
        g_object_unref (self->connection);
        self->connection = NULL;
    }
    if (self->dbus_node_info)
    {
        g_dbus_node_info_unref (self->dbus_node_info);
        self->dbus_node_info = NULL;
    }

    if (self->verify_loop)
    {
        g_main_loop_quit (self->verify_loop);
        self->verify_loop = NULL;
    }
    if (self->verify_thread)
    {
        g_thread_unref (self->verify_thread);
        self->verify_thread = NULL;
    }
    g_free (self);
}

static char *
convert_to_utf8 (const char *msg)
{
    char *utf8;
    utf8 = g_locale_to_utf8 (msg, -1, NULL, NULL, NULL);
    if (utf8 == NULL)
    {
        char *p, *q;
        utf8 = g_strdup (msg);
        p = utf8;
        while (*p != '\0' && !g_utf8_validate ((const char *)p, -1, (const char **)&q))
        {
            *q = '?';
            p = q + 1;
        }
    }
    return utf8;
}

static gboolean
emit_signal (CdosUserVerifier *self, const gchar *signal, GVariant *parameters)
{
    GError *error = NULL;
    GDBusConnection *conn = g_bus_get_sync (G_BUS_TYPE_SESSION, NULL, &error);
    if (error)
    {
        g_warning ("get bus error: %s", error->message);
        g_error_free (error);
        return FALSE;
    }
    if (!g_dbus_connection_emit_signal (conn,
                                        NULL,
                                        "/org/cdos/ScreenSaver/UserVerifier",
                                        "org.cdos.ScreenSaver.UserVerifier",
                                        signal,
                                        parameters,
                                        &error))
    {
        g_warning ("emit signal[%s] error: %s", signal, error->message);
        g_error_free (error);
        return FALSE;
    }
    return TRUE;
}

static gboolean
on_wait_timeout (CdosUserVerifier *self)
{
    self->wait_timeout = TRUE;
    self->wait_timeout_id =0;
    return FALSE;
}

static gboolean
cdos_user_verifier_ask_question (CdosUserVerifier *self, const char *msg, gchar **answer)
{
    GVariant *v;

    v = g_variant_new ("(s)", msg);
    if (!emit_signal (self, "InfoQuery", v))
    {
        return FALSE;
    }
    self->wait_timeout = FALSE;
    if (self->wait_timeout_id)
        g_source_remove (self->wait_timeout_id);

    self->wait_timeout_id = g_timeout_add_seconds (30, (GSourceFunc)on_wait_timeout, self);
    g_mutex_lock (&self->mutex);
    while (!self->answer && !self->wait_timeout && !self->canceled)
    {
        g_cond_wait (&self->cond, &self->mutex);
    }
    g_mutex_unlock (&self->mutex);

    if (answer)
        *answer = g_strdup (self->answer);

    g_clear_pointer (&self->answer, g_free);
    return TRUE;
}

static gboolean
cdos_user_verifier_ask_secret (CdosUserVerifier *self, const char *msg, gchar **answer)
{
    g_debug ("%s %lu", __func__, pthread_self ());
    GVariant *v;

    v = g_variant_new ("(s)", msg);
    if (!emit_signal (self, "SecretInfoQuery", v))
    {
        // g_variant_unref (v);
        return FALSE;
    }

    self->wait_timeout = FALSE;
    if (self->wait_timeout_id)
        g_source_remove (self->wait_timeout_id);
    self->wait_timeout_id = g_timeout_add_seconds (30, (GSourceFunc)on_wait_timeout, self);

    g_mutex_lock (&self->mutex);
    while (!self->answer && !self->wait_timeout && !self->canceled)
    {
        //g_main_context_iteration (g_main_loop_get_context (self->verify_loop), TRUE);
        g_cond_wait (&self->cond, &self->mutex);
    }
    g_mutex_unlock (&self->mutex);

    if (answer)
        *answer = g_strdup (self->answer);
    g_clear_pointer (&self->answer, g_free);
    return TRUE;
}

static gboolean
cdos_user_verifier_send_message (CdosUserVerifier *self, const char *msg)
{
    GVariant *v;

    v = g_variant_new ("(s)", msg);
    if (!emit_signal (self, "Message", v))
    {
        // g_variant_unref (v);
        return FALSE;
    }
    return TRUE;
}

static gboolean
cdos_user_verifier_send_error (CdosUserVerifier *self, const char *msg)
{
    GVariant *v;
    v = g_variant_new ("(s)", msg);
    g_warning ("Pam send error: %s", msg);
    if (!emit_signal (self, "Error", v))
    {
        // g_variant_unref (v);
        return FALSE;
    }
    return TRUE;
}

static gboolean
process_pam_message (CdosUserVerifier *self, const struct pam_message *message, gchar **response)
{
    gboolean res;
    char *msg;
    char *answer;

    res = FALSE;
    msg = convert_to_utf8 (message->msg);
    answer = NULL;

    g_debug ("%s: msg_style:%d, msg:%s", __func__, message->msg_style, msg);

    switch (message->msg_style)
    {
        case PAM_PROMPT_ECHO_ON:
            res = cdos_user_verifier_ask_question (self, msg, &answer);
            break;
        case PAM_PROMPT_ECHO_OFF:
            res = cdos_user_verifier_ask_secret (self, msg, &answer);
            break;
        case PAM_TEXT_INFO:
            res = cdos_user_verifier_send_message (self, msg);
            break;
        case PAM_ERROR_MSG:
            res = cdos_user_verifier_send_error (self, msg);
            break;
        default:
            break;
    }
    if (response)
    {
        if (answer)
            *response = strndup (answer, PAM_MAX_RESP_SIZE - 1);
        else
            *response = NULL;
    }
    g_free (answer);

    return res;
}

static int
pam_messages_handler (int n_messages,
                      const struct pam_message **messages,
                      struct pam_response **responses,
                      gpointer user_data)
{
    struct pam_response *replies;
    int ret;
    int i;
    CdosUserVerifier *self = user_data;
    ret = PAM_CONV_ERR;

    if (n_messages < 0)
        return ret;
    if (n_messages == 0)
    {
        if (responses)
            *responses = NULL;
        return PAM_SUCCESS;
    }

    replies = (struct pam_response *)calloc (n_messages, sizeof (struct pam_response));

    for (i = 0; i < n_messages; i++)
    {
        gchar *response = NULL;
        if (!process_pam_message (self, messages[i], &response))
        {
            ret = PAM_CONV_ERR;
            goto out;
        }
        replies[i].resp = response;
        replies[i].resp_retcode = PAM_SUCCESS;
    }
    ret = PAM_SUCCESS;
    if (responses)
        *responses = replies;
out:
    if (ret != PAM_SUCCESS)
    {
        for (i = 0; i < n_messages; i++)
        {
            if (replies[i].resp)
                free (replies[i].resp);
        }
        free (replies);
        if (responses)
            *responses = NULL;
        g_warning ("%s: %s", __func__, pam_strerror (self->pam_handle, ret));
    }

    return ret;
}

static int
init_pam (CdosUserVerifier *self)
{
    struct pam_conv pam_conversation;
    int error_code;

    pam_conversation.conv = pam_messages_handler;
    pam_conversation.appdata_ptr = self;

    error_code = pam_start (self->service, self->username, &pam_conversation, &self->pam_handle);
    if (error_code != PAM_SUCCESS)
    {
        g_warning ("pam_start(): %s", pam_strerror (self->pam_handle, error_code));
        return error_code;
    }
    error_code = pam_set_item (self->pam_handle, PAM_TTY, g_getenv ("DISPLAY"));
    if (error_code != PAM_SUCCESS)
        g_warning ("pam_set_item(): %s, DISPLAY: %s", pam_strerror (self->pam_handle, error_code), g_getenv ("DISPLAY"));

    return PAM_SUCCESS;
}

static gboolean
quit_loop (CdosUserVerifier *self)
{
    g_debug ("%s: quit loop now, status = %d", __func__, self->status);
    if (self->status == IDLE)
    {
        g_bus_unown_name (self->bus_id);
        g_main_loop_quit (self->loop);
        cdos_user_verifier_free (self);

        return FALSE;
    }
    return TRUE;
}

static const gchar *
get_pam_error (pam_handle_t *pamh, int retval)
{
    const gchar *error = pam_strerror (pamh, retval);
    return error;
}

static gboolean 
user_in_nopasswdlogin_group (const char *name) {
    struct group *data;
    int i;
    gboolean usr_in_nopwdlogin_group = FALSE;

    while (0 != (data = getgrent())) { 
        if (0 == strcmp(data->gr_name, "nopasswdlogin")) { 
            i = 0;
            while (data->gr_mem[i]) {
                if (0 == strcmp(name, data->gr_mem[i++])) {
                    usr_in_nopwdlogin_group = TRUE;
                    break;
                }
            }            
            break;
        }
    }
    endgrent ();

    return usr_in_nopwdlogin_group;
}

static void
idle_quit (CdosUserVerifier *self)
{
    self->status = IDLE;
    if (self->timeout_id)
        g_source_remove (self->timeout_id);
    g_debug ("%s:will quit in 30s.", __func__);
    self->timeout_id = g_timeout_add_seconds (30, (GSourceFunc)quit_loop, self);
}

static gboolean
check_acct_mgmt (CdosUserVerifier *self)
{
    int retval;
    gboolean locked = FALSE;

    if (self->pam_handle == NULL)
    {
        retval = init_pam (self);
        if (retval != PAM_SUCCESS)
            return FALSE;
    }

    retval = pam_acct_mgmt (self->pam_handle, 0);
    g_debug ("pam_acct_mgmt() retval = %d, lock_state %d", retval, self->lock_state);

    if (retval == PAM_MAXTRIES)
    {
        locked = TRUE;
    }
    if (locked != self->lock_state)
    {
        self->lock_state = locked;
        emit_signal (self, "LockStateChanged", g_variant_new ("(b)", self->lock_state));
    }
    if (!self->lock_state) 
    {
        verify_pam_stop (self, retval);
    }
    return TRUE;
}

static gboolean 
idle_acct_mgmt (CdosUserVerifier *self)
{    
    check_acct_mgmt (self);
    self->status = BUSY;
    g_debug ("%s::lock_state %d", __func__, self->lock_state);
    if (self->lock_state == FALSE)
    {
        self->locked_id = 0;
        idle_quit (self);
    }
    return self->lock_state;
}

static gboolean
idle_authenticate (CdosUserVerifier *self)
{
    int retval, retval2;

    retval = pam_authenticate (self->pam_handle, 0);
    g_debug ("pam_authenticate() retval = %d", retval);

    if (retval != PAM_SUCCESS)
    {
        const gchar *error;
        if (self->canceled)
            error = "Canceled";
        else if (self->wait_timeout)
            error = "Timeout";
        else
            error = get_pam_error (self->pam_handle, retval);
        g_warning ("Found screensaver pam auth faild(%d): %s", retval, error);
        emit_signal (self, "VerificationFailed", g_variant_new ("(s)", error));
        //帐号认证次数达到最大，则被锁定
        if (retval == PAM_MAXTRIES && !self->lock_state)
        {
            self->lock_state = TRUE;
            emit_signal (self, "LockStateChanged", g_variant_new ("(b)", self->lock_state));
            if (self->locked_id == 0)
                self->locked_id = g_timeout_add_seconds (1, (GSourceFunc)idle_acct_mgmt, self);
            goto out;
        }
    }
    else
    {
        emit_signal (self, "VerificationComplete", NULL);
    }

    verify_pam_stop (self, retval);
    idle_quit (self);

out:
    self->idle_id = 0;
    return FALSE;
}

static gboolean 
check_regist_info (CdosUserVerifier *self)
{
    if (self->service && self->username)
        return TRUE;
    g_warning ("The pam service name (%s) or user name (%s) missing.", self->service, self->username);
    return FALSE;
}

static gboolean 
update_pam_info (CdosUserVerifier *self, gchar *service, gchar *username)
{    
    gboolean changed =  FALSE;
    if (service && g_strcmp0 (self->service, service))
    {
        g_free (self->service);
        self->service = g_strdup (service);
        changed = TRUE;
    }
    if (username && g_strcmp0 (self->username, username))
    {
        g_free (self->username);
        self->username = g_strdup (username);
        changed = TRUE;
    }
    if (changed)
    {
        g_clear_pointer (&self->answer, g_free);
        verify_pam_stop (self, 0);
    }
    return changed;
}

static void
handle_method_call (GDBusConnection *connection G_GNUC_UNUSED,
                    const gchar *sender,
                    const gchar *object_path G_GNUC_UNUSED,
                    const gchar *interface,
                    const gchar *method,
                    GVariant *parameters,
                    GDBusMethodInvocation *invocation,
                    CdosUserVerifier *self)
{
    g_debug ("\n%s: %s %s", __func__, interface, method);
    self->status = BUSY;
    if (g_strcmp0 (method, "RegistVerifyInfo") == 0)
    {
        gchar *service;
        gchar *username;
        g_variant_get (parameters, "(ss)", &service, &username);
        g_message ("Regist user verify info, service: %s, username: %s", service, username);
        update_pam_info (self, service, username);
        g_free (service);
        g_free (username);
        g_dbus_method_invocation_return_value (invocation, g_variant_new ("(b)", TRUE));
        return;
    }
    else if (g_strcmp0 (method, "BeginVerification") == 0)
    {
        if (check_regist_info (self) == FALSE)
        {
            g_dbus_method_invocation_return_value (invocation, g_variant_new ("(b)", FALSE));
            return;
        }

        g_clear_pointer (&self->answer, g_free);

        if (user_in_nopasswdlogin_group (self->username))
        {
            emit_signal (self, "VerificationComplete", NULL);
        }
        else if (self->pam_handle == NULL)
        {
            gint retval = init_pam (self);
            g_debug ("%s:int pam, retval = %d", __func__, retval);
            if (retval != PAM_SUCCESS)
            {
                emit_signal (self, "VerificationFailed", g_variant_new ("(s)", get_pam_error (self->pam_handle, retval)));
            }
            else
            {
                if (self->idle_id)
                    g_source_remove (self->idle_id);
                self->idle_id = g_idle_add ((GSourceFunc)idle_authenticate, self);
            }
        }
        g_dbus_method_invocation_return_value (invocation, g_variant_new ("(b)", TRUE));
        return ;
    }
    else if (g_strcmp0 (method, "AnswerQuery") == 0)
    {
        if (check_regist_info (self) == FALSE)
        {
            g_dbus_method_invocation_return_value (invocation, g_variant_new ("(b)", FALSE));
            return;
        }
        gchar *answer;
        g_variant_get (parameters, "(s)", &answer);
        g_debug ("%s:::get answer:%s", __func__, answer);
        g_clear_pointer (&self->answer, g_free);
        if (answer)
            self->answer = g_strdup (answer);

        g_cond_signal (&self->cond);
        g_free (answer);
        g_dbus_method_invocation_return_value (invocation, g_variant_new ("(b)", TRUE));
        return;
    }
    else if (g_strcmp0 (method, "GetAccountLocked") == 0)
    {
        if (check_regist_info (self) == FALSE)
        {
            g_dbus_method_invocation_return_value (invocation, g_variant_new ("(b)", FALSE));
            return;
        }
        check_acct_mgmt (self);
        g_dbus_method_invocation_return_value (invocation, g_variant_new ("(b)", self->lock_state));
        return;
    }
    else if (g_strcmp0 (method, "Cancel") == 0)
    {
        gboolean force = FALSE;
        g_variant_get (parameters, "(b)", &force);
        g_dbus_method_invocation_return_value (invocation, g_variant_new ("(b)", !self->lock_state));
        //如果是锁定状态暂不退出
        if (self->lock_state == FALSE || force)
        {
            self->status = IDLE;
            quit_loop (self);
        }
    }
}

static void
on_userverifier_dbus_acquried (GDBusConnection *conn, const gchar *name, CdosUserVerifier *self)
{
    GError *error = NULL;
    if (!g_dbus_connection_register_object (conn,
                                            "/org/cdos/ScreenSaver/UserVerifier",
                                            self->dbus_node_info->interfaces[0],
                                            &interface_vtable,
                                            self,
                                            NULL,
                                            &error))
    {
        g_warning ("%s: %s", __func__, error->message);
        g_error_free (error);
        return;
    }
}

static gpointer
verify_thread (CdosUserVerifier *self)
{
    GMainContext *context = g_main_context_new ();
    self->verify_loop = g_main_loop_new (context, FALSE);
    g_main_context_push_thread_default (context);

    GError *error = NULL;

    GDBusConnection *conn;

    conn = g_bus_get_sync (G_BUS_TYPE_SESSION, NULL, &error);
    if (conn == NULL)
    {
        g_warning ("%s", error->message);
        g_error_free (error);
        return NULL;
    }

    self->lock_state = FALSE;
    self->connection = conn;
    self->dbus_node_info = g_dbus_node_info_new_for_xml (introspect_xml, &error);
    if (NULL == self->dbus_node_info)
    {
        g_warning ("%s", error->message);
        g_error_free (error);
        g_object_unref (G_OBJECT (conn));
        return NULL;
    }

    self->bus_id = g_bus_own_name (G_BUS_TYPE_SESSION,
                                   "org.cdos.ScreenSaver.UserVerifier",
                                   G_BUS_NAME_OWNER_FLAGS_REPLACE | G_BUS_NAME_OWNER_FLAGS_ALLOW_REPLACEMENT,
                                   (GBusAcquiredCallback)on_userverifier_dbus_acquried,
                                   NULL,
                                   NULL,
                                   self,
                                   NULL);
    if (!self->bus_id)
    {
        g_warning ("Failed to call g_bus_own_name()");
        g_dbus_node_info_unref (self->dbus_node_info);
        g_object_unref (G_OBJECT (conn));
        return NULL;
    }
    g_main_loop_run (self->verify_loop);
    return NULL;
}

static gboolean
cdos_user_verifier_init (CdosUserVerifier *self)
{
    g_debug ("%s %lu", __func__, pthread_self ());

    self->verify_thread = g_thread_new ("verify_thread",
                                        (GThreadFunc)verify_thread,
                                        self);

    g_cond_init (&self->cond);
    g_mutex_init (&self->mutex);

    return TRUE;
}

int
main (int argc, char **argv)
{
    GMainLoop *loop = NULL;
    CdosUserVerifier *user_verifier;

    setlocale (LC_ALL, "");
    loop = g_main_loop_new (NULL, FALSE);
    user_verifier = g_new0 (CdosUserVerifier, 1);
    user_verifier->loop = loop;
    gboolean ret = cdos_user_verifier_init (user_verifier);
    if (ret)
        g_main_loop_run (loop);
    return 0;
}
