/* -*- mode: C; c-file-style: "gnu"; indent-tabs-mode: nil; -*- */

#include "config.h"

#include <string.h>
#include <stdlib.h>

#include <X11/Xlib.h>
#include <X11/Xatom.h>
#include <gdk/gdk.h>
#include <gdk/gdkx.h>
#include <meta/display.h>
#include <meta/group.h>
#include <meta/util.h>
#include <meta/window.h>
#include <meta/meta-workspace-manager.h>

#include "cdos-window-tracker-private.h"
#include "cdos-app-private.h"
#include "cdos-global.h"
#include "cdos-marshal.h"
#include "st.h"

/* This file includes modified code from
 * desktop-data-engine/engine-dbus/hippo-application-monitor.c
 * in the functions collecting application usage data.
 * Written by Owen Taylor, originally licensed under LGPL 2.1.
 * Copyright Red Hat, Inc. 2006-2008
 */

/**
 * SECTION:cdos-window-tracker
 * 简述：和应用程序相关的窗口
 *
 * 保持从窗口到应用程序的影射。依据WM_CLASS X11属性的一些启发实现该功能；在未
 * 来，通过启动通知实现。
 */

struct _CdosWindowTracker
{
    GObject parent;

    CdosApp *focus_app;

    /* <MetaWindow * window, CdosApp *app> */
    GHashTable *window_to_app;

    /* <int, CdosApp *app> */
    GHashTable *launched_pid_to_app;
};

G_DEFINE_TYPE (CdosWindowTracker, cdos_window_tracker, G_TYPE_OBJECT);

enum
{
    PROP_0,
    PROP_FOCUS_APP
};

enum
{
    STARTUP_SEQUENCE_CHANGED,
    TRACKED_WINDOWS_CHANGED,

    LAST_SIGNAL
};

static guint signals[LAST_SIGNAL] = {0};

static void cdos_window_tracker_finalize (GObject *object);
static void set_focus_app (CdosWindowTracker *tracker, CdosApp *new_focus_app);
static void on_focus_window_changed (MetaDisplay *display, GParamSpec *spec, CdosWindowTracker *tracker);

static void track_window (CdosWindowTracker *tracker, MetaWindow *window);
static void disassociate_window (CdosWindowTracker *tracker, MetaWindow *window);
static CdosApp *cdos_startup_sequence_get_app (MetaStartupSequence *sequence);

static void
cdos_window_tracker_get_property (GObject *gobject, guint prop_id, GValue *value, GParamSpec *pspec)
{
    CdosWindowTracker *tracker = CDOS_WINDOW_TRACKER (gobject);

    switch (prop_id)
    {
        case PROP_FOCUS_APP:
            g_value_set_object (value, tracker->focus_app);
            break;
        default:
            G_OBJECT_WARN_INVALID_PROPERTY_ID (gobject, prop_id, pspec);
            break;
    }
}

static void
cdos_window_tracker_class_init (CdosWindowTrackerClass *klass)
{
    GObjectClass *gobject_class = G_OBJECT_CLASS (klass);

    gobject_class->get_property = cdos_window_tracker_get_property;
    gobject_class->finalize = cdos_window_tracker_finalize;

    g_object_class_install_property (gobject_class,
                                     PROP_FOCUS_APP,
                                     g_param_spec_object ("focus-app",
                                                          "Focus App",
                                                          "Focused application",
                                                          CDOS_TYPE_APP,
                                                          G_PARAM_READABLE));

    signals[STARTUP_SEQUENCE_CHANGED] = g_signal_new ("startup-sequence-changed",
                                                      CDOS_TYPE_WINDOW_TRACKER,
                                                      G_SIGNAL_RUN_LAST,
                                                      0,
                                                      NULL,
                                                      NULL,
                                                      NULL,
                                                      G_TYPE_NONE,
                                                      1,
                                                      META_TYPE_STARTUP_SEQUENCE);
    signals[TRACKED_WINDOWS_CHANGED] = g_signal_new ("tracked-windows-changed",
                                                     CDOS_TYPE_WINDOW_TRACKER,
                                                     G_SIGNAL_RUN_LAST,
                                                     0,
                                                     NULL,
                                                     NULL,
                                                     NULL,
                                                     G_TYPE_NONE,
                                                     0);
}

/**
 * cdos_window_tracker_is_window_interesting:
 *
 * The CdosWindowTracker associates certain kinds of windows with
 * applications; however, others we don't want to
 * appear in places where we want to give a list of windows
 * for an application, such as the alt-tab dialog.
 *
 * 想要显示的窗口例子是根桌面窗口。我们跳过所有的覆盖重定向类型，也排除其他的窗口
 * 类型，尽管这些的多数都是覆盖重定向的。
 *
 * Returns: 如果窗口是interesting，则返回TRUE。
 */
gboolean
cdos_window_tracker_is_window_interesting (MetaWindow *window)
{
    if (meta_window_is_override_redirect (window) || meta_window_is_skip_taskbar (window))
        return FALSE;

    switch (meta_window_get_window_type (window))
    {
        /* Definitely ignore these. */
        case META_WINDOW_DESKTOP:
        case META_WINDOW_DOCK:
        case META_WINDOW_SPLASHSCREEN:
        /* Should have already been handled by override_redirect above,
       * but explicitly list here so we get the "unhandled enum"
       * warning if in the future anything is added.*/
        case META_WINDOW_DROPDOWN_MENU:
        case META_WINDOW_POPUP_MENU:
        case META_WINDOW_TOOLTIP:
        case META_WINDOW_NOTIFICATION:
        case META_WINDOW_COMBO:
        case META_WINDOW_DND:
        case META_WINDOW_OVERRIDE_OTHER:
            return FALSE;
        case META_WINDOW_NORMAL:
        case META_WINDOW_DIALOG:
        case META_WINDOW_MODAL_DIALOG:
        case META_WINDOW_MENU:
        case META_WINDOW_TOOLBAR:
        case META_WINDOW_UTILITY:
            break;
    }

    return TRUE;
}

static CdosApp *
get_app_from_id (MetaWindow *window,
                 const char *id)
{
    CdosApp *app;
    CdosAppSystem *appsys;
    char *desktop_file;

    appsys = cdos_app_system_get_default ();

    desktop_file = g_strconcat (id, ".desktop", NULL);
    app = cdos_app_system_lookup_app (appsys, desktop_file);
    if (app)
        g_object_ref (app);

    g_free (desktop_file);
    return app;
}

static CdosApp *
get_app_from_gapplication_id (MetaWindow *window)
{
    const char *id;

    id = meta_window_get_gtk_application_id (window);
    if (!id)
        return NULL;

    return get_app_from_id (window, id);
}

/**
 * get_app_from_window_wmclass:
 *
 * 在给定的窗口中查找，并尝试检测基于WM_CLASS的应用。如果没有检测到，则返回NULL。
 *
 * Return value: 新的CdosApp对象，或者为NULL。
 */
static CdosApp *
get_app_from_window_wmclass (MetaWindow *window)
{
    CdosApp *app;
    CdosAppSystem *appsys;
    const char *wm_class;
    const char *wm_instance;

    appsys = cdos_app_system_get_default ();

    /* Notes on the heuristics used here:
     much of the complexity here comes from the desire to support
     Chrome apps.

     From https://bugzilla.gnome.org/show_bug.cgi?id=673657#c13

     Currently chrome sets WM_CLASS as follows (the first string is the 'instance',
     the second one is the 'class':

     For the normal browser:
     WM_CLASS(STRING) = "chromium", "Chromium"

     For a bookmarked page (through 'Tools -> Create application shortcuts')
     WM_CLASS(STRING) = "wiki.gnome.org__GnomeShell_ApplicationBased", "Chromium"

     For an application from the chrome store (with a .desktop file created through
     right click, "Create shortcuts" from Chrome's apps overview)
     WM_CLASS(STRING) = "crx_blpcfgokakmgnkcojhhkbfbldkacnbeo", "Chromium"

     The .desktop file has a matching StartupWMClass, but the name differs, e.g. for
     the store app (youtube) there is

     .local/share/applications/chrome-blpcfgokakmgnkcojhhkbfbldkacnbeo-Default.desktop

     with

     StartupWMClass=crx_blpcfgokakmgnkcojhhkbfbldkacnbeo

     Note that chromium (but not google-chrome!) includes a StartupWMClass=chromium
     in their .desktop file, so we must match the instance first.

     Also note that in the good case (regular gtk+ app without hacks), instance and
     class are the same except for case and there is no StartupWMClass at all.
  */

    /* first try a match from WM_CLASS (instance part) to StartupWMClass */
    wm_instance = meta_window_get_wm_class_instance (window);
    app = cdos_app_system_lookup_startup_wmclass (appsys, wm_instance);
    if (app != NULL)
        return g_object_ref (app);
    /* then try a match from WM_CLASS to StartupWMClass */
    wm_class = meta_window_get_wm_class (window);
    app = cdos_app_system_lookup_startup_wmclass (appsys, wm_class);
    if (app != NULL)
        return g_object_ref (app);
    /* then try a match from WM_CLASS (instance part) to .desktop */
    app = cdos_app_system_lookup_desktop_wmclass (appsys, wm_instance);
    if (app != NULL)
        return g_object_ref (app);

    /* finally, try a match from WM_CLASS to .desktop */
    app = cdos_app_system_lookup_desktop_wmclass (appsys, wm_class);
    if (app != NULL)
        return g_object_ref (app);

    return NULL;
}

/*
 * get_app_from_sandboxed_app_id:
 * @window: a #MetaWindow
 *
 * Looks only at the given window, and attempts to determine
 * an application based on its Flatpak or Snap ID.  If one can't be determined,
 * return %NULL.
 *
 * Return value: (transfer full): A newly-referenced #ShellApp, or %NULL
 */
static CdosApp *
get_app_from_sandboxed_app_id (MetaWindow  *window)
{
  const char *id;

  id = meta_window_get_sandboxed_app_id (window);
  if (!id)
    return NULL;

  return get_app_from_id (window, id);
}

/**
 * get_app_from_window_group:
 * @monitor: CdosWindowTracker对象
 * @window: MetaWindow对象
 *
 * 检查MetaWindow对象组内的其它窗口，看该应用是否是这些窗口中的一个。
 *
 * Return value: 新建的CdosApp对象，或者为NULL。
 */
static CdosApp *
get_app_from_window_group (CdosWindowTracker *tracker, MetaWindow *window)
{
    CdosApp *result;
    GSList *group_windows;
    MetaGroup *group;
    GSList *iter;

    group = meta_window_get_group (window);
    if (group == NULL)
        return NULL;

    group_windows = meta_group_list_windows (group);

    result = NULL;
    /* Try finding a window in the group of type NORMAL; if we
   * succeed, use that as our source. */
    for (iter = group_windows; iter; iter = iter->next)
    {
        MetaWindow *group_window = iter->data;

        if (meta_window_get_window_type (group_window) != META_WINDOW_NORMAL)
            continue;

        result = g_hash_table_lookup (tracker->window_to_app, group_window);
        if (result)
            break;
    }

    g_slist_free (group_windows);

    if (result)
        g_object_ref (result);

    return result;
}

/**
 * get_app_from_window_pid:
 * @tracker: CdosWindowTracker对象
 * @window: MetaWindow对象
 *
 * 检查和加载应用相关的窗口进程ID
 *
 * Return value: 新的CdosApp对象，或者为NULL。
 */
static CdosApp *
get_app_from_window_pid (CdosWindowTracker *tracker, MetaWindow *window)
{
    CdosApp *result;
    int pid;

    if (meta_window_is_remote (window))
        return NULL;

    pid = meta_window_get_pid (window);

    if (pid == -1)
        return NULL;

    result = g_hash_table_lookup (tracker->launched_pid_to_app, GINT_TO_POINTER (pid));
    if (result != NULL)
        g_object_ref (result);

    return result;
}

/**
 * get_app_for_window:
 *
 * 查找和该窗口相关的应用程序，使用所有可用的信息(如窗口的MetaGroup),
 * 和我们知道的其他窗口信息。
 *
 * Returns: CdosApp对象，如果没有发现，则返回NULL。
 */
static CdosApp *
get_app_for_window (CdosWindowTracker *tracker, MetaWindow *window)
{
    CdosApp *result = NULL;
    const char *startup_id;
    MetaWindow *transient_for;

    transient_for = meta_window_get_transient_for (window);
    if (transient_for)
        return get_app_for_window (tracker, transient_for);

    /* First, we check whether we already know about this window,
     * if so, just return that.
     */
    if (meta_window_get_window_type (window) == META_WINDOW_NORMAL ||
        meta_window_is_remote (window) ||
        meta_window_get_window_type (window) == META_WINDOW_DIALOG)
    {
        result = g_hash_table_lookup (tracker->window_to_app, window);
        if (result != NULL)
        {
            g_object_ref (result);
            return result;
        }
    }

    if (meta_window_is_remote (window))
        return _cdos_app_new_for_window (window);

    result = get_app_from_gapplication_id (window);
    if (result)
        return result;

    /* Check if the app's WM_CLASS specifies an app; this is
     * canonical if it does.
     */
    result = get_app_from_window_wmclass (window);
    if (result != NULL)
        return result;

    /* Check if the window was opened from within a sandbox; if this
     * is the case, a corresponding .desktop file is guaranteed to match;
     */
    result = get_app_from_sandboxed_app_id (window);
    if (result != NULL)
        return result;

    result = get_app_from_window_pid (tracker, window);
    if (result != NULL)
        return result;

    /* Now we check whether we have a match through startup-notification */
    startup_id = meta_window_get_startup_id (window);
    if (startup_id)
    {
        GSList *iter, *sequences;

        sequences = cdos_window_tracker_get_startup_sequences (tracker);
        for (iter = sequences; iter; iter = iter->next)
        {
            MetaStartupSequence *sequence = iter->data;
            const char *id = meta_startup_sequence_get_id (sequence);
            if (strcmp (id, startup_id) != 0)
                continue;

            result = cdos_startup_sequence_get_app (sequence);
            if (result)
            {
                result = g_object_ref (result);
                break;
            }
        }
    }

    /* If we didn't get a startup-notification match, see if we matched
   * any other windows in the group.
   */
    if (result == NULL)
        result = get_app_from_window_group (tracker, window);
    /* Our last resort - we create a fake app from the window */
    if (result == NULL)
        result = _cdos_app_new_for_window (window);

    return result;
}

const char *
_cdos_window_tracker_get_app_context (CdosWindowTracker *tracker, CdosApp *app)
{
    return "";
}

static void
update_focus_app (CdosWindowTracker *self)
{
    MetaWindow *new_focus_win;
    CdosApp *new_focus_app;

    new_focus_win = meta_display_get_focus_window (cdos_global_get_display (cdos_global_get ()));
    new_focus_app = new_focus_win ? cdos_window_tracker_get_window_app (self, new_focus_win) : NULL;

    set_focus_app (self, new_focus_app);
}

static void
tracked_window_changed (CdosWindowTracker *self,
                        MetaWindow         *window)
{
  /* It's simplest to just treat this as a remove + add. */
  disassociate_window (self, window);
  track_window (self, window);
  /* also just recalculate the focused app, in case it was the focused
     window that changed */
  update_focus_app (self);
}

static void
on_wm_class_changed (MetaWindow  *window,
                     GParamSpec  *pspec,
                     gpointer     user_data)
{
  CdosWindowTracker *self = CDOS_WINDOW_TRACKER (user_data);
  tracked_window_changed (self, window);
}

static void
on_gtk_application_id_changed (MetaWindow  *window,
                               GParamSpec  *pspec,
                               gpointer     user_data)
{
  CdosWindowTracker *self = CDOS_WINDOW_TRACKER (user_data);
  tracked_window_changed (self, window);
}

static void
track_window (CdosWindowTracker *self, MetaWindow *window)
{
    CdosApp *app;
    if (!cdos_window_tracker_is_window_interesting (window))
        return;

    app = get_app_for_window (self, window);

    if (!app)
        return;

    /* At this point we've stored the association from window -> application */
    g_hash_table_insert (self->window_to_app, window, app);
    g_signal_connect (window, "notify::wm-class", G_CALLBACK (on_wm_class_changed), self);
    g_signal_connect (window, "notify::gtk-application-id", G_CALLBACK (on_gtk_application_id_changed), self);

    _cdos_app_add_window (app, window);

    g_signal_emit (self, signals[TRACKED_WINDOWS_CHANGED], 0);
}

static void
cdos_window_tracker_on_window_added (MetaWorkspace *workspace, MetaWindow *window, gpointer user_data)
{
    CdosWindowTracker *self = CDOS_WINDOW_TRACKER (user_data);
    MetaWindowType window_type = meta_window_get_window_type (window);

    if (window_type == META_WINDOW_NORMAL ||
        window_type == META_WINDOW_DIALOG ||
        window_type == META_WINDOW_UTILITY ||
        window_type == META_WINDOW_MODAL_DIALOG)
        track_window (self, window);
}

static void
disassociate_window (CdosWindowTracker *self, MetaWindow *window)
{
    CdosApp *app;

    app = g_hash_table_lookup (self->window_to_app, window);
    if (!app)
        return;

    g_object_ref (app);

    g_hash_table_remove (self->window_to_app, window);

    _cdos_app_remove_window (app, window);
    g_signal_handlers_disconnect_by_func (window, G_CALLBACK (on_wm_class_changed), self);
    g_signal_handlers_disconnect_by_func (window, G_CALLBACK (on_gtk_application_id_changed), self);

    g_signal_emit (self, signals[TRACKED_WINDOWS_CHANGED], 0);

    g_object_unref (app);
}

static void
cdos_window_tracker_on_window_removed (MetaWorkspace *workspace, MetaWindow *window, gpointer user_data)
{
    disassociate_window (CDOS_WINDOW_TRACKER (user_data), window);
}

static void
load_initial_windows (CdosWindowTracker *tracker)
{
    MetaWorkspaceManager *wsm;
    GList *workspaces, *iter;
    MetaDisplay *display = cdos_global_get_display (cdos_global_get ());
    wsm = meta_display_get_workspace_manager (display);
    workspaces = meta_workspace_manager_get_workspaces (wsm);
    for (iter = workspaces; iter; iter = iter->next)
    {
        MetaWorkspace *workspace = iter->data;
        GList *windows = meta_workspace_list_windows (workspace);
        GList *window_iter;

        for (window_iter = windows; window_iter; window_iter = window_iter->next)
        {
            MetaWindow *window = window_iter->data;
            track_window (tracker, window);
        }

        g_list_free (windows);
    }
}

static void
cdos_window_tracker_on_n_workspaces_changed (MetaWorkspaceManager *wsm, GParamSpec *pspec, gpointer user_data)
{
    CdosWindowTracker *self = CDOS_WINDOW_TRACKER (user_data);
    GList *workspaces, *iter;
    workspaces = meta_workspace_manager_get_workspaces (wsm);

    for (iter = workspaces; iter; iter = iter->next)
    {
        MetaWorkspace *workspace = iter->data;

        /* This pair of disconnect/connect is idempotent if we were
       * already connected, while ensuring we get connected for
       * new workspaces.
       */
        g_signal_handlers_disconnect_by_func (workspace, cdos_window_tracker_on_window_added, self);
        g_signal_handlers_disconnect_by_func (workspace, cdos_window_tracker_on_window_removed, self);

        g_signal_connect (workspace, "window-added", G_CALLBACK (cdos_window_tracker_on_window_added), self);
        g_signal_connect (workspace, "window-removed", G_CALLBACK (cdos_window_tracker_on_window_removed), self);
    }
}

static void
init_window_tracking (CdosWindowTracker *self)
{
    MetaDisplay *display;
    display = cdos_global_get_display (cdos_global_get ());
    MetaWorkspaceManager *wsm;
    wsm = meta_display_get_workspace_manager (display);

    // 在 mutter 中 没有找到 display 的 n-workspaces 属性，但 WorkspaceManager 上有.
    // 这里必须要绑定工作区变化的信号，否则，新工作区上面的创建的窗口都会是 NULL !!!!!
    g_signal_connect (wsm, "notify::n-workspaces", G_CALLBACK (cdos_window_tracker_on_n_workspaces_changed), self);
    g_signal_connect (display, "notify::focus-window", G_CALLBACK (on_focus_window_changed), self);

    cdos_window_tracker_on_n_workspaces_changed (wsm, NULL, self);
}

static void
on_startup_sequence_changed (MetaDisplay *display, MetaStartupSequence *sequence, CdosWindowTracker *self)
{
    CdosApp *app;

    app = cdos_startup_sequence_get_app (sequence);
    if (app)
        _cdos_app_handle_startup_sequence (app, sequence);

    g_signal_emit (self, signals[STARTUP_SEQUENCE_CHANGED], 0, sequence);
}

static void
cdos_window_tracker_init (CdosWindowTracker *self)
{
    MetaDisplay *display;
    display = cdos_global_get_display (cdos_global_get ());
    MetaStartupNotification *sn = meta_display_get_startup_notification (display);

    self->window_to_app = g_hash_table_new_full (g_direct_hash, g_direct_equal, NULL, (GDestroyNotify)g_object_unref);

    self->launched_pid_to_app = g_hash_table_new_full (NULL, NULL, NULL, (GDestroyNotify)g_object_unref);


    g_signal_connect (sn, "changed", G_CALLBACK (on_startup_sequence_changed), self);

    load_initial_windows (self);
    init_window_tracking (self);
}

static void
cdos_window_tracker_finalize (GObject *object)
{
    CdosWindowTracker *self = CDOS_WINDOW_TRACKER (object);

    g_hash_table_destroy (self->window_to_app);
    g_hash_table_destroy (self->launched_pid_to_app);

    G_OBJECT_CLASS (cdos_window_tracker_parent_class)->finalize (object);
}

/**
 * cdos_window_tracker_get_window_app:
 * @tracker: 监控对象
 * @metawin: MetaWindow对象
 *
 * Returns: 和窗口相关的应用程序
 */
CdosApp *
cdos_window_tracker_get_window_app (CdosWindowTracker *tracker, MetaWindow *metawin)
{
    MetaWindow *transient_for;
    CdosApp *app;

    transient_for = meta_window_get_transient_for (metawin);
    if (transient_for != NULL)
        metawin = transient_for;

    app = g_hash_table_lookup (tracker->window_to_app, metawin);
    if (app)
        g_object_ref (app);

    return app;
}


/**
 * cdos_window_tracker_get_app_from_pid:
 * @self; CdosAppSystem对象
 * @pid: Unix进程标识
 *
 * 查找和该进程相关的应用
 *
 * Returns: CdosApp对象，如果没有，则返回NULL。
 */
CdosApp *
cdos_window_tracker_get_app_from_pid (CdosWindowTracker *self, int pid)
{
    GSList *running = cdos_app_system_get_running (cdos_app_system_get_default ());
    GSList *iter;
    CdosApp *result = NULL;

    for (iter = running; iter; iter = iter->next)
    {
        CdosApp *app = iter->data;
        GSList *pids = cdos_app_get_pids (app);
        GSList *pids_iter;

        for (pids_iter = pids; pids_iter; pids_iter = pids_iter->next)
        {
            int app_pid = GPOINTER_TO_INT (pids_iter->data);
            if (app_pid == pid)
            {
                result = app;
                break;
            }
        }
        g_slist_free (pids);

        if (result != NULL)
            break;
    }

    g_slist_free (running);

    return result;
}

static void
on_child_exited (GPid pid, gint status, gpointer unused_data)
{
    CdosWindowTracker *tracker;

    tracker = cdos_window_tracker_get_default ();

    g_hash_table_remove (tracker->launched_pid_to_app, GINT_TO_POINTER ((gint)pid));
}

void
_cdos_window_tracker_add_child_process_app (CdosWindowTracker *tracker, GPid pid, CdosApp *app)
{
    gpointer pid_ptr = GINT_TO_POINTER ((int)pid);

    if (g_hash_table_lookup (tracker->launched_pid_to_app, pid_ptr))
        return;

    g_hash_table_insert (tracker->launched_pid_to_app, pid_ptr, g_object_ref (app));
    g_child_watch_add (pid, on_child_exited, NULL);
    /* TODO: rescan unassociated windows
   * Unlikely in practice that the launched app gets ahead of us
   * enough to map an X window before we get scheduled after the fork(),
   * but adding this note for future reference.
   */
}

static void
set_focus_app (CdosWindowTracker *tracker, CdosApp *new_focus_app)
{
    if (new_focus_app == tracker->focus_app)
        return;

    if (tracker->focus_app != NULL)
        g_object_unref (tracker->focus_app);

    tracker->focus_app = new_focus_app;

    if (tracker->focus_app != NULL)
        g_object_ref (tracker->focus_app);

    g_object_notify (G_OBJECT (tracker), "focus-app");
}

static void
on_focus_window_changed (MetaDisplay *display, GParamSpec *spec, CdosWindowTracker *tracker)
{
    update_focus_app (tracker);
}

CdosApp *
cdos_window_tracker_get_focus_app (CdosWindowTracker *tracker)
{
    return tracker->focus_app;
}

/**
 * cdos_window_tracker_get_startup_sequences:
 * @tracker:CdosWindowTracke对象
 *
 *
 * Returns: 当前激活的启动序列
 */
GSList *
cdos_window_tracker_get_startup_sequences (CdosWindowTracker *self)
{
    CdosGlobal *global = cdos_global_get ();
    MetaDisplay *display = cdos_global_get_display (global);
    MetaStartupNotification *sn = meta_display_get_startup_notification (display);
    return meta_startup_notification_get_sequences (sn);
}

/**
 * cdos_startup_sequence_get_app:
 * @sequence: CdosStartupSequence对象
 *
 * Returns: 启动的应用程序，如果未知，则为NULL。
 */
CdosApp *
cdos_startup_sequence_get_app (MetaStartupSequence *sequence)
{
    const char *appid;
    CdosAppSystem *appsys;
    CdosApp *app;

    appid = meta_startup_sequence_get_application_id (sequence);
    if (!appid)
        return NULL;

    appsys = cdos_app_system_get_default ();
    app = cdos_app_system_lookup_app_for_path (appsys, appid);
    return app;
}

/**
 * cdos_window_tracker_get_default:
 *
 * Return Value: 全局CdosWindowTracker实例
 */
CdosWindowTracker *
cdos_window_tracker_get_default (void)
{
    static CdosWindowTracker *instance;

    if (instance == NULL)
    {
        instance = g_object_new (CDOS_TYPE_WINDOW_TRACKER, NULL);
        g_object_add_weak_pointer (G_OBJECT (instance), (gpointer *)&instance);
    }

    return instance;
}
