/* -*- mode: C; c-file-style: "gnu"; indent-tabs-mode: nil; -*- */
/*
 * st-tooltip.c: Plain tooltip actor
 *
 * Copyright 2008, 2009 Intel Corporation
 * Copyright 2009 Red Hat, Inc.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms and conditions of the GNU Lesser General Public License,
 * version 2.1, as published by the Free Software Foundation.
 *
 * This program is distributed in the hope it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for
 * more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 */

/**
 * SECTION:st-tooltip
 * @short_description: tooltip窗口部件
 *
 * StTooltip实现一个tooltip。这通常不应有应用程序创建，而是由tooltip的功能实现，例
 * 如st_button_set_tooltip()。
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <math.h>
#include <stdlib.h>
#include <string.h>

#include <glib.h>

#include <clutter/clutter.h>

#include "st-tooltip.h"

#include "st-label.h"
#include "st-private.h"

enum
{
  PROP_0,

  PROP_LABEL,
  PROP_SOURCE,

  N_PROPS
};

typedef struct _StTooltipPrivate StTooltipPrivate;

struct _StTooltipPrivate
{
  StLabel         *label;
  ClutterActor    *source;
};

G_DEFINE_TYPE_WITH_PRIVATE (StTooltip, st_tooltip, ST_TYPE_WIDGET);

static void st_tooltip_show (ClutterActor *self);

static void st_tooltip_constrain (StTooltip             *tooltip,
                                  const graphene_rect_t *geometry,
                                  graphene_rect_t       *adjusted_geometry);

static void
st_tooltip_set_property (GObject      *gobject,
                         guint         prop_id,
                         const GValue *value,
                         GParamSpec   *pspec)
{
  StTooltip *tooltip = ST_TOOLTIP (gobject);

  switch (prop_id)
    {
    case PROP_LABEL:
      st_tooltip_set_label (tooltip, g_value_get_string (value));
      break;

    case PROP_SOURCE:
      st_tooltip_set_source (tooltip, g_value_get_object (value));
      break;

    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (gobject, prop_id, pspec);
      break;
    }
}

static void
st_tooltip_get_property (GObject    *gobject,
                         guint       prop_id,
                         GValue     *value,
                         GParamSpec *pspec)
{
  StTooltipPrivate *priv = st_tooltip_get_instance_private (ST_TOOLTIP (gobject));

  switch (prop_id)
    {
    case PROP_LABEL:
      g_value_set_string (value, st_label_get_text (priv->label));
      break;

    case PROP_SOURCE:
      g_value_set_object (value, priv->source);
      break;

    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (gobject, prop_id, pspec);
      break;
    }
}

static void
st_tooltip_get_preferred_width (ClutterActor *self,
                                gfloat        for_height,
                                gfloat       *min_width_p,
                                gfloat       *natural_width_p)
{
  StTooltipPrivate *priv = st_tooltip_get_instance_private (ST_TOOLTIP (self));
  StThemeNode *theme_node = st_widget_get_theme_node (ST_WIDGET (self));
  gfloat label_height;

  st_theme_node_adjust_for_height (theme_node, &for_height);

  if (for_height > -1)
    {
      label_height = for_height;
    }
  else
    {
      label_height = -1;
    }

  clutter_actor_get_preferred_width (CLUTTER_ACTOR (priv->label),
                                     label_height,
                                     min_width_p,
                                     natural_width_p);

  st_theme_node_adjust_preferred_width (theme_node, min_width_p, natural_width_p);
}

static void
st_tooltip_get_preferred_height (ClutterActor *self,
                                 gfloat        for_width,
                                 gfloat       *min_height_p,
                                 gfloat       *natural_height_p)
{
  StTooltipPrivate *priv = st_tooltip_get_instance_private (ST_TOOLTIP (self));
  StThemeNode *theme_node = st_widget_get_theme_node (ST_WIDGET (self));
  gfloat min_label_h, natural_label_h;

  st_theme_node_adjust_for_width (theme_node, &for_width);

  clutter_actor_get_preferred_height (CLUTTER_ACTOR (priv->label),
                                      for_width,
                                      &min_label_h,
                                      &natural_label_h);
  *min_height_p = min_label_h;
  *natural_height_p = natural_label_h;

  st_theme_node_adjust_preferred_height (theme_node, min_height_p, natural_height_p);
}

static void
st_tooltip_allocate (ClutterActor          *self,
                     const ClutterActorBox *box,
                     ClutterAllocationFlags flags)
{
  StTooltipPrivate *priv = st_tooltip_get_instance_private (ST_TOOLTIP (self));
  StThemeNode *theme_node = st_widget_get_theme_node (ST_WIDGET (self));
  ClutterActorBox content_box;

  clutter_actor_set_allocation (self, box, flags);

  st_theme_node_get_content_box (theme_node, box, &content_box);

  clutter_actor_allocate (CLUTTER_ACTOR (priv->label), &content_box, flags);
}

static void
st_tooltip_paint (ClutterActor *self, ClutterPaintContext *paint_context)
{
  StTooltipPrivate *priv = st_tooltip_get_instance_private (ST_TOOLTIP (self));

  st_widget_paint_background (ST_WIDGET (self), paint_context);

  clutter_actor_paint (CLUTTER_ACTOR (priv->label), paint_context);
}

static void
st_tooltip_dispose (GObject *self)
{
  StTooltipPrivate *priv = st_tooltip_get_instance_private (ST_TOOLTIP (self));

  if (priv->label)
    {
      clutter_actor_destroy (CLUTTER_ACTOR (priv->label));
      priv->label = NULL;
    }

  G_OBJECT_CLASS (st_tooltip_parent_class)->dispose (self);
}

static void
st_tooltip_class_init (StTooltipClass *klass)
{
  GObjectClass *gobject_class = G_OBJECT_CLASS (klass);
  ClutterActorClass *actor_class = CLUTTER_ACTOR_CLASS (klass);
  GParamSpec *pspec;

  gobject_class->set_property = st_tooltip_set_property;
  gobject_class->get_property = st_tooltip_get_property;
  gobject_class->dispose = st_tooltip_dispose;

  actor_class->get_preferred_width = st_tooltip_get_preferred_width;
  actor_class->get_preferred_height = st_tooltip_get_preferred_height;
  actor_class->allocate = st_tooltip_allocate;
  actor_class->paint = st_tooltip_paint;
  actor_class->show = st_tooltip_show;

  pspec = g_param_spec_string ("label",
                               "Label",
                               "Label of the tooltip",
                               NULL, G_PARAM_READWRITE);
  g_object_class_install_property (gobject_class, PROP_LABEL, pspec);

  pspec = g_param_spec_object ("source",
                               "Source",
                               "Actor on the stage the tooltip applies to",
                               CLUTTER_TYPE_ACTOR,
                               G_PARAM_READWRITE);
  g_object_class_install_property (gobject_class, PROP_SOURCE, pspec);
}

static void
st_tooltip_init (StTooltip *tooltip)
{
  StTooltipPrivate *priv = st_tooltip_get_instance_private (tooltip);
  priv->label = ST_LABEL (st_label_new (NULL));

  clutter_actor_add_child (CLUTTER_ACTOR (tooltip),
                           CLUTTER_ACTOR (priv->label));

  g_object_set (tooltip, "show-on-set-parent", FALSE, NULL);

  clutter_actor_set_reactive (CLUTTER_ACTOR (tooltip), FALSE);
}

static void
st_tooltip_update_position (StTooltip *tooltip)
{
  StTooltipPrivate *priv = st_tooltip_get_instance_private (tooltip);
  graphene_rect_t geometry;
  graphene_rect_t adjusted_geometry;
  gfloat tooltip_w, tooltip_h, tooltip_x, tooltip_y;

  /* if no area set, just position ourselves top left */
  if (!priv->source)
    {
      clutter_actor_set_position ((ClutterActor*) tooltip, 0, 0);
      return;
    }

  /* we need to have a style in case there are padding/border values to take into
   * account when calculating width/height */
  st_widget_ensure_style ((StWidget *) tooltip);

  /* find out the tooltip's size */
  clutter_actor_get_size ((ClutterActor*) tooltip, &tooltip_w, &tooltip_h);

  /* attempt to place the tooltip */
  gfloat source_x, source_y, source_w, source_h;
  clutter_actor_get_transformed_position (priv->source, &source_x, &source_y);
  clutter_actor_get_transformed_size (priv->source, &source_w, &source_h);
  tooltip_x = (int)(source_x + (source_w / 2) - (tooltip_w / 2));
  tooltip_y = (int)(source_y + source_h);

  geometry.origin.x = tooltip_x;
  geometry.origin.y = tooltip_y;
  geometry.size.width = ceil (tooltip_w);
  geometry.size.height = ceil (tooltip_h);

  st_tooltip_constrain (tooltip, &geometry, &adjusted_geometry);

  tooltip_x = adjusted_geometry.origin.x;
  tooltip_y = adjusted_geometry.origin.y;

  /* FIXME */
  clutter_actor_set_anchor_point ((ClutterActor*) tooltip, -tooltip_x, -tooltip_y);
}

/**
 * st_tooltip_get_label:
 * @tooltip: StTooltip对象
 *
 * 获取tooltip的文本显示
 *
 * Returns: tooltip的文本。该文本不能由应用程序释放。
 */
const gchar *
st_tooltip_get_label (StTooltip *tooltip)
{
  g_return_val_if_fail (ST_IS_TOOLTIP (tooltip), NULL);

  StTooltipPrivate *priv = st_tooltip_get_instance_private (tooltip);
  return st_label_get_text (priv->label);
}

/**
 * st_tooltip_set_label:
 * @tooltip: StTooltip对象
 * @text: 设置为label的文本
 *
 * 设置tooltip的文本显示
 */
void
st_tooltip_set_label (StTooltip   *tooltip,
                      const gchar *text)
{
  StTooltipPrivate *priv;

  g_return_if_fail (ST_IS_TOOLTIP (tooltip));

  priv = st_tooltip_get_instance_private (tooltip);

  st_label_set_text (priv->label, text);

  g_object_notify (G_OBJECT (tooltip), "label");
}

static void
st_tooltip_show (ClutterActor *self)
{
  StTooltip *tooltip = ST_TOOLTIP (self);

  st_tooltip_update_position (tooltip);

  /* finally show the tooltip... */
  CLUTTER_ACTOR_CLASS (st_tooltip_parent_class)->show (self);
}

/**
 * st_tooltip_set_source:
 * @tooltip: StTooltip对象
 * @area: graphene_rect_t对象
 *
 * 设置tooltip应用的stage区域。
 */
void
st_tooltip_set_source (StTooltip    *tooltip,
                       ClutterActor *source)
{
  g_return_if_fail (ST_IS_TOOLTIP (tooltip));

  StTooltipPrivate *priv = st_tooltip_get_instance_private (tooltip);
  priv->source = source;
}

/**
 * st_tooltip_get_source:
 * @tooltip: StTooltip对象
 *
 * 回收tooltip当前应用的stage区域。
 *
 * Returns: ClutterActor对象，由tooltip获取，不能由应用程序释放。
 */
ClutterActor *
st_tooltip_get_source (StTooltip *tooltip)
{
  g_return_val_if_fail (ST_IS_TOOLTIP (tooltip), NULL);

  StTooltipPrivate *priv = st_tooltip_get_instance_private (tooltip);
  return priv->source;
}

typedef struct {
  StTooltipConstrainFunc func;
  gpointer data;
  GDestroyNotify notify;
} ConstrainFuncClosure;

static void
constrain_func_closure_free (gpointer data)
{
  ConstrainFuncClosure *closure = data;
  if (closure->notify)
    closure->notify (closure->data);
  g_slice_free (ConstrainFuncClosure, data);
}

static GQuark
st_tooltip_constrain_func_quark (void)
{
  static GQuark value = 0;
  if (G_UNLIKELY (value == 0))
    value = g_quark_from_static_string ("st-tooltip-constrain-func");
  return value;
}

/**
 * st_tooltip_set_constrain_func:
 * @stage: ClutterStage对象
 * @func: 用于约束tooltip位置的函数
 * @data: 传递到func的用户数据
 * @notify: 当data不再需要，需要掉用的函数
 *
 * 设置回调函数，该函数用于限制tooltips的位置。如果stage扩展到多个monitors，并
 * 且tooltips不通过monitors而被定位时调用该毁掉函数。
 */
void
st_tooltip_set_constrain_func (ClutterStage           *stage,
                               StTooltipConstrainFunc  func,
                               gpointer                data,
                               GDestroyNotify          notify)
{
  ConstrainFuncClosure *closure;

  g_return_if_fail (CLUTTER_IS_STAGE (stage));

  if (func)
    {
      closure = g_slice_new (ConstrainFuncClosure);
      closure->func = func;
      closure->data = data;
      closure->notify = notify;
    }
  else
    closure = NULL;

  g_object_set_qdata_full (G_OBJECT (stage), st_tooltip_constrain_func_quark (),
                           closure, constrain_func_closure_free);
}

static void
st_tooltip_constrain (StTooltip             *tooltip,
                      const graphene_rect_t *geometry,
                      graphene_rect_t       *adjusted_geometry)
{
  ConstrainFuncClosure *closure;

  ClutterActor *stage = clutter_actor_get_stage (CLUTTER_ACTOR (tooltip));

  *adjusted_geometry = *geometry;

  if (stage == NULL)
    return;

  closure = g_object_get_qdata (G_OBJECT (stage), st_tooltip_constrain_func_quark ());
  if (closure)
    {
      closure->func (tooltip, geometry, adjusted_geometry, closure->data);
    }
  else
    {
      ClutterActor *parent;
      gfloat parent_w, parent_h;

      parent = clutter_actor_get_parent ((ClutterActor *) tooltip);
      clutter_actor_get_size (parent, &parent_w, &parent_h);

      /* make sure the tooltip is not off parent horizontally */
      if (adjusted_geometry->origin.x < 0)
        adjusted_geometry->origin.x = 0;
      else if (adjusted_geometry->origin.x + adjusted_geometry->size.width > parent_w)
        adjusted_geometry->origin.x = (int)(parent_w) - adjusted_geometry->size.width;

      /* make sure the tooltip is not off parent vertically */
      if (adjusted_geometry->origin.y + adjusted_geometry->size.height > parent_h)
        adjusted_geometry->origin.y = parent_h - adjusted_geometry->size.height;
    }
}
