#include "cdos-app-favorites.h"
#include "cdos-app.h"
#include "cdos-app-system.h"
#include "cdos-global.h"
/**
 * SECTION: cdos-app-favorites
 * @short_description: 用来处理用户喜爱的应用程序
 * @include: cdos-app-favorites.h
 *
 * CdosAppFavorites 该类与用户喜爱的应用程序有关。主要是通过读取和设置gsettings org.cdos
 * favorite-apps，该类中使用GHashTable保存喜爱的应用程序列表。
 */

#define FAVORITE_APPS_KEY "favorite-apps"

static gchar **change_list_array (GList *list);

static gboolean cdos_app_favorites_remove_favorite_real (CdosAppFavorites *app, gchar *appId);

static gboolean cdos_app_favorites_add_favorite_real (CdosAppFavorites *app, gchar *appId, gint pos);

static void cdos_app_favorites_reload (CdosAppFavorites *app);

static void cdos_app_favorites_on_fav_changed (GSettings *settings, gchar *key, gpointer user_data);

static CdosAppFavorites *cdos_app_favorites_new (void);
static void cdos_app_favorites_dispose (GObject *gobject);

enum
{
    CHANGED,
    LAST_SIGNAL
};

static guint cdos_app_favorites_signals[LAST_SIGNAL] = {
    0,
};

struct _CdosAppFavoritesPrivate
{
    GHashTable *favorites;
    GSettings *settings;
};

G_DEFINE_TYPE_WITH_PRIVATE (CdosAppFavorites, cdos_app_favorites, G_TYPE_OBJECT);

static void
cdos_app_favorites_class_init (CdosAppFavoritesClass *klass)
{
    GObjectClass *gobject_class = G_OBJECT_CLASS (klass);

    gobject_class->dispose = cdos_app_favorites_dispose;

    gobject_class->dispose = cdos_app_favorites_dispose;
    cdos_app_favorites_signals[CHANGED] = g_signal_new ("changed",
                                                        G_TYPE_FROM_CLASS (klass),
                                                        G_SIGNAL_RUN_LAST,
                                                        G_STRUCT_OFFSET (CdosAppFavoritesClass, changed),
                                                        NULL,
                                                        NULL,
                                                        g_cclosure_marshal_VOID__VOID,
                                                        G_TYPE_NONE,
                                                        0);
}

static void
cdos_app_favorites_init (CdosAppFavorites *app)
{
    app->priv = cdos_app_favorites_get_instance_private (app);
    app->priv->favorites = g_hash_table_new_full (g_str_hash, g_direct_equal, NULL, NULL);
    app->priv->settings = cdos_global_get_settings (cdos_global_get ());
}

static void
cdos_app_favorites_dispose (GObject *gobject)
{
    CdosAppFavorites *self = CDOS_APP_FAVORITES (gobject);
    CdosAppFavoritesPrivate *priv = self->priv;
    if (priv->favorites)
    {
        g_hash_table_destroy (priv->favorites);
        priv->favorites = NULL;
    }
    G_OBJECT_CLASS (cdos_app_favorites_parent_class)->dispose (gobject);
}

static CdosAppFavorites *
cdos_app_favorites_new (void)
{
    CdosAppFavorites *app;
    app = CDOS_APP_FAVORITES (g_object_new (CDOS_TYPE_APP_FAVORITES, NULL));
    gchar *sig = g_strconcat ("changed::", FAVORITE_APPS_KEY, NULL);
    g_signal_connect (app->priv->settings, sig, G_CALLBACK (cdos_app_favorites_on_fav_changed), app);

    // added
    g_signal_connect (
        app->priv->settings, "changed::favorite-apps-max-number", G_CALLBACK (cdos_app_favorites_on_fav_changed), app);

    cdos_app_favorites_reload (app);
    g_free (sig);

    return app;
}

static void
cdos_app_favorites_on_fav_changed (GSettings *settings, gchar *key, gpointer user_data)
{
    cdos_app_favorites_reload (CDOS_APP_FAVORITES (user_data));
    g_signal_emit_by_name (CDOS_APP_FAVORITES (user_data), "changed");
}

static void
cdos_app_favorites_reload (CdosAppFavorites *app)
{
    gint i;
    CdosAppFavoritesPrivate *priv = app->priv;
    gchar **ids = g_settings_get_strv (priv->settings, FAVORITE_APPS_KEY);
    gint ids_length = g_strv_length (ids); // added
    // g_print ("ids_length=%d\n", ids_length);   //////调试
    // test
    /*for (i = 0; i < ids_length; i++)
        {
            g_print ("ids[%d]=%s\n", i, ids[i]);
        }*/

    CdosAppSystem *appSys = cdos_app_system_get_default ();
    CdosApp *ap = NULL;
    // added
    gint start_num;
    gchar *favorite;
    GPtrArray *apps = g_ptr_array_new ();
    gint max_number = g_settings_get_int (priv->settings, "favorite-apps-max-number");

    gchar **ids_tmp = ids;
    /*
        while(*ids_tmp)
        {
            ap = cdos_app_system_lookup_app(appSys, *ids_tmp);
            if (ap != NULL)
                    g_ptr_array_add(apps, ap);
            ids_tmp++;
        }
        */

    if (ids_length < max_number)
    {
        start_num = 0;
    }
    else
    {
        start_num = ids_length - max_number;
    }

    for (i = start_num; i < ids_length; i++)
    {
        ////g_print ("ids_tmp[%d]=%s\n", i, ids_tmp[i]);   //////调试
        favorite = g_strdup (ids_tmp[i]);
        if (favorite)
        {
            ap = cdos_app_system_lookup_app (appSys, favorite);
        }
        if (ap != NULL)
            g_ptr_array_add (apps, ap);

        g_free (favorite);
        favorite = NULL;
    }

    g_hash_table_remove_all (priv->favorites);

    for (i = 0; i < (apps->len); i++)
    {
        ap = g_ptr_array_index (apps, i);
        const gchar *id = cdos_app_get_id (ap);
        // g_print ("id=%s\n", id);   //////调试
        g_hash_table_insert (priv->favorites, (gchar *)id, ap);
    }

    g_ptr_array_free (apps, TRUE);
    g_strfreev (ids);
}

/**
 * cdos_app_favorites_get_favorites_map
 * @app: 一个#CdosAppFavorites对象
 *
 * 获取存有喜爱应用信息的哈希表
 *
 * Returns: #GHashTable对象
 */
GHashTable *
cdos_app_favorites_get_favorites_map (CdosAppFavorites *app)
{
    g_return_val_if_fail (CDOS_IS_APP_FAVORITES (app), NULL);
    return app->priv->favorites;
}

/**
 * cdos_app_favorites_get_favorites
 * @app: 一个#CdosAppFavorites对象
 *
 * 获取喜爱应用程序的列表，从GHashTable中取得values
 *
 * Returns: 返回喜爱应用程序的列表，一个#GList对象
 */
GList *
cdos_app_favorites_get_favorites (CdosAppFavorites *app)
{
    g_return_val_if_fail (CDOS_IS_APP_FAVORITES (app), NULL);
    return g_hash_table_get_values (app->priv->favorites);
}

/**
 * cdos_app_favorites_is_favorite
 * @app: 一个#CdosAppFavorites对象
 * @appId: 应用程序ID
 *
 * 判断一个应用是否是喜欢的应用程序
 *
 * Returns: TRUE，是喜欢的应用程序;FALSE，不是喜欢的应用程序
 */
gboolean
cdos_app_favorites_is_favorite (CdosAppFavorites *app, const gchar *appId)
{
    g_return_val_if_fail (CDOS_IS_APP_FAVORITES (app), FALSE);
    if (NULL == appId)
        return FALSE;

    return g_hash_table_contains (app->priv->favorites, appId);
}

static gboolean
cdos_app_favorites_add_favorite_real (CdosAppFavorites *app, gchar *appId, gint pos)
{
    if (NULL == appId)
        return FALSE;

    CdosAppSystem *appSystem = cdos_app_system_get_default ();
    CdosApp *ap = cdos_app_system_lookup_app (appSystem, appId);

    if (!ap)
        return FALSE;

    GList *ids = g_hash_table_get_keys (app->priv->favorites);

    if (pos == -1)
        ids = g_list_append (ids, appId);

    else
        ids = g_list_insert (ids, appId, pos);


    gchar **result = change_list_array (ids);
    g_hash_table_insert (app->priv->favorites, appId, ap);
    g_settings_set_strv (app->priv->settings, FAVORITE_APPS_KEY, (const gchar **)result);
    g_list_free (ids);
    g_strfreev (result);
    return TRUE;
}

/**
 * cdos_app_favorites_add_favorite
 * @app: 一个#CdosAppFavorites对象
 * @appId: 应用程序ID
 * @pos: 添加应用程序的位置
 *
 * 添加应用程序到pos指定的列表位置，并设置gsettings org.cdos
 * favorite-apps的值;如果pos=-1，将会以追加的方式把应用程序添加到列表
 *
 */
void
cdos_app_favorites_add_favorite (CdosAppFavorites *app, gchar *appId, gint pos)
{
    g_return_if_fail (CDOS_IS_APP_FAVORITES (app));
    cdos_app_favorites_add_favorite_real (app, appId, pos);
}

/**
 * cdos_app_favorites_add_favorite_to_pos
 * @app: 一个#CdosAppFavorites对象
 * @appId: 应用程序ID
 * @pos: 添加应用程序的位置
 *
 * 添加应用程序到pos指定的列表位置，可以参考cdos_app_favorites_add_favorite
 *
 */
void
cdos_app_favorites_add_favorite_to_pos (CdosAppFavorites *app, gchar *appId, gint pos)
{
    g_return_if_fail (CDOS_IS_APP_FAVORITES (app));
    cdos_app_favorites_add_favorite_real (app, appId, pos);
}

/**
 * cdos_app_favorites_move_favorite_to_pos
 * @app: 一个#CdosAppFavorites对象
 * @appId: 应用程序ID
 * @pos: 添加应用程序的位置
 *
 * 移动喜爱的应用程序到pos指定的列表位置;如果pos=-1，将会以追加的方式把应用程序添加到列表
 *
 */
void
cdos_app_favorites_move_favorite_to_pos (CdosAppFavorites *app, gchar *appId, gint pos)
{
    g_return_if_fail (CDOS_IS_APP_FAVORITES (app));
    cdos_app_favorites_remove_favorite (app, appId);
    cdos_app_favorites_add_favorite (app, appId, pos);
}

static gboolean
cdos_app_favorites_remove_favorite_real (CdosAppFavorites *app, gchar *appId)
{
    if (NULL == appId)
        return FALSE;

    if (!g_hash_table_contains (app->priv->favorites, appId))
        return FALSE;

    gboolean r = g_hash_table_remove (app->priv->favorites, appId);
    if (r == TRUE)
    {
        GList *res = g_hash_table_get_keys (app->priv->favorites);
        gchar **result = change_list_array (res);
        g_settings_set_strv (app->priv->settings, FAVORITE_APPS_KEY, (const gchar **)result);
        g_strfreev (result);
        g_list_free (res);
    }
    return TRUE;
}

/**
 * cdos_app_favorites_remove_favorite
 * @app: 一个#CdosAppFavorites对象
 * @appId: 应用程序ID
 *
 * 移除appId对应的应用程序，并重新设置gsettings org.cdos favorite-apps
 *
 */
void
cdos_app_favorites_remove_favorite (CdosAppFavorites *app, gchar *appId)
{
    g_return_if_fail (CDOS_IS_APP_FAVORITES (app));
    cdos_app_favorites_remove_favorite_real (app, appId);
}


static gchar **
change_list_array (GList *list)
{
    GList *tmp = list;
    guint len = g_list_length (tmp);
    gchar **res = g_malloc (sizeof (gchar *) * (len + 1));

    gchar **result = res;

    while (tmp)
    {
        const gchar *data = tmp->data;
        *result = g_strdup (data);
        result++;
        tmp = tmp->next;
    }
    *result = NULL;

    return res;
}

/**
 * cdos_app_favorites_get_default
 *
 * 获取#CdosAppFavorites对象
 *
 * Returns: 一个#CdosAppFavorites对象
 */
CdosAppFavorites *
cdos_app_favorites_get_default (void)
{
    static CdosAppFavorites *instance = NULL;

    if (NULL == instance)
    {
        instance = cdos_app_favorites_new ();
        g_object_add_weak_pointer (G_OBJECT (instance), (gpointer *)&instance);
    }

    return instance;
}
