#include "cdos-applet-manager.h"
#include "cdos-panel.h"
#include "cdos-extension.h"
#include "cdos-ui-plugin.h"
#include "cdos-utils-private.h"
#include "applets/group/applets-group.h"

#include "cdos-global.h"
#include "st.h"

#include <stdio.h>
#include <stdlib.h>

guint menu_flag = 0;
guint calendar_flag = 0;
guint network_flag = 0;
guint removable_flag = 0;
guint sound_flag = 0;
guint systray_flag = 0;

struct _AppletDefinition
{
    ClutterActor *panel;
    StSide orientation;
    ClutterActor *location;
    gboolean center;
    gint order;
    gchar *uuid;
};

typedef struct
{
    /* the raw from gsettings */
    gchar **raw;
    /* uuid_map: the each element */
    GHashTable *uuid_map;

} EnabledAppletDefinitions;

/* Maps uuid -> applet objects */
static GTree *applets_tree = NULL;

static gchar **grouped_applets = NULL;
static gboolean group_loaded = FALSE;
static CdosApplet *applets_group = NULL;
static GHashTable *add_to_group = NULL;

static gboolean applets_loaded = FALSE;

static EnabledAppletDefinitions *enabled_applet_definitions;
static CdosIconDispatcher *status_icon_dispatcher;

static void
cdos_applet_manager_add_applet_to_panels (CdosExtension *extension,
                                          AppletDefinition *definition);

static AppletDefinition *
cdos_applet_manager_get_applet_definition (const gchar *definition)
{
    CdosPanel *panel;
    StSide orientation;
    gint order;
    gboolean center;
    ClutterActor *location;

    /**
     * format used in gsettings is 'panel:location:order:uuid' where:
     * - location is either 'left', 'center' or 'right'
     * - order is an integer representing the order of the applet within the panel/location (i.e. 1st, 2nd etc...)
     * - uuid is a unique id assigned to the applet instance when added.
     */
    AppletDefinition *ret = g_new0 (AppletDefinition, 1);
    gchar **elements = g_strsplit (definition, ":", -1);

    if (g_strv_length (elements) > 2)
    {
        panel = cdos_ui_plugin_get_panel ();

        if (cdos_panel_get_bottom_position (panel))
        {
            orientation = ST_SIDE_BOTTOM;
        }
        else
        {
            orientation = ST_SIDE_TOP;
        }

        order = atoi (elements[1]);

        location = cdos_panel_get_leftbox (panel);

        if (g_strcmp0 ("center", elements[0]) == 0)
        {
            center = TRUE;
            location = cdos_panel_get_centerbox (panel);
        }
        else if (g_strcmp0 (elements[0], "right") == 0)
        {
            location = cdos_panel_get_rightbox (panel);
        }

        ret->panel = (ClutterActor *)panel;
        ret->orientation = orientation;
        ret->location = location;
        ret->center = center;
        ret->order = order;
        ret->uuid = g_strdup (elements[2]);
        g_strfreev(elements);
        return ret;
    }

    g_warning ("Bad applet definition: %s", definition);
    return NULL;
}

    static EnabledAppletDefinitions *
cdos_applet_manager_get_enabled_applet_definitions (void)
{
    int i;
    EnabledAppletDefinitions *result = g_new0 (EnabledAppletDefinitions, 1);

    GSettings *settings = g_settings_new ("org.cdos");
    result->raw = g_settings_get_strv (settings, "enabled-applets");
    result->uuid_map = g_hash_table_new (g_str_hash, g_str_equal);

    /* Parse all definitions */
    for (i = 0; result->raw[i] != NULL; i ++)
    {
        AppletDefinition *applet_definition  = cdos_applet_manager_get_applet_definition (result->raw[i]);
        if (applet_definition)
        {
            g_hash_table_insert (result->uuid_map, applet_definition->uuid, applet_definition);
        }
    }
    return result;
}

static void
cdos_applet_manager_remove_applet_from_panels (AppletDefinition *applet_definition)
{
    CdosApplet *applet;
    applet = g_tree_lookup (applets_tree, applet_definition->uuid);

    if (applet)
    {
        //cdos_applet_on_applet_removed_from_panel (applet);
        g_debug ("%s: %s", __func__, applet_definition->uuid);

        CdosExtension *extension = cdos_applet_get_extension (applet);
        GHashTable *definitions = cdos_extension_get_loaded_definitions (extension);
        g_hash_table_remove (definitions, applet_definition->uuid);

        g_tree_remove (applets_tree, applet_definition->uuid);

        ClutterActor *panel_location = cdos_applet_get_panel_location (applet);
        if (panel_location != NULL && g_strcmp0 (applet_definition->uuid, "show-desktop@cdos.org") != 0)
        {
            ClutterActor *actor = CLUTTER_ACTOR (applet);
            g_object_ref (actor);
            g_debug ("%s %d", __func__, ((GObject *)applet)->ref_count);
            clutter_actor_remove_child (panel_location, actor);
        }
    }
}

static ClutterActor *
cdos_applet_manager_get_panel_zone (CdosPanel *panel, gint key)
{
    if (panel)
    {
        switch (key)
        {
            case 0:
                return (ClutterActor *)cdos_panel_get_leftbox (panel);
            case 1:
                return (ClutterActor *)cdos_panel_get_centerbox (panel);
            case 2:
                return (ClutterActor *)cdos_panel_get_rightbox (panel);
            default:
                break;
        }
    }
    return NULL;
}

/**
 * cdos_applet_manager_save_applets_positions:
 *
 * Save applets positions
 *
 */
void
cdos_applet_manager_save_applets_positions (void)
{
    guint j;
    CdosPanel *panel;
    const gchar *zones_strings[] = {"left", "center", "right"};
    GSList *all_applets = NULL;
    gchar **strv_applets = {NULL, };
    GSettings *settings = g_settings_new ("org.cdos");

    panel = cdos_ui_plugin_get_panel ();
    if (!panel)
        return;
    for (j = 0; zones_strings[j] != NULL; j++)
    {
        ClutterActor *zone = cdos_applet_manager_get_panel_zone (panel, j);
        if (zone)
        {
            ClutterActor *child = clutter_actor_get_first_child ((ClutterActor *)zone);
            while (child)
            {
                CdosApplet *app = CDOS_APPLET (child);
                if (app)
                {
                    all_applets = g_slist_append (all_applets, app);
                }
                child = clutter_actor_get_next_sibling (child);
            }
        }
    }

    for (j = 0; zones_strings[j] != NULL; j++)
    {
        ClutterActor *zone = cdos_applet_manager_get_panel_zone (panel, j);
        for (GSList *iter = all_applets; iter != NULL; iter = iter->next)
        {
            CdosApplet *applet = iter->data;
            ClutterActor *applet_zone;
            ClutterActor *new_panel_location = cdos_applet_get_new_panel_location (applet);
            if (new_panel_location != NULL)
            {
                applet_zone = new_panel_location;
            }
            else
            {
                applet_zone = cdos_applet_get_panel_location (applet);
            }
            gint applet_order;
            gint new_order = cdos_applet_get_new_order (applet);
            if (new_order != -1)
            {
                applet_order = new_order;
            }
            else
            {
                applet_order = cdos_applet_get_order (applet);
            }
            if (applet_zone == zone)
            {
                /* "[left|center|right]:order:uuid */
                gchar *str = g_strconcat (zones_strings[j], ":",
                                          g_strdup_printf ("%d", applet_order), ":",
                                          cdos_applet_get_uuid (applet), NULL);
                gchar **tmp = strv_applets;
                strv_applets = cdos_strv_elem_add ((const gchar **)tmp, str);
                g_strfreev (tmp);
            }
        }
    }

    for (GSList *iter = all_applets; iter != NULL; iter = iter->next)
    {
        CdosApplet *applet = iter->data;
        cdos_applet_set_new_panel_location (applet, NULL);
        cdos_applet_set_new_order (applet, -1); // -1 instead of null that from js
    }

    g_settings_set_strv (settings, "enabled-applets", (const gchar * const *)strv_applets);
    g_slist_free(all_applets);
    //g_strfreev(zones_strings);
    g_strfreev(strv_applets);
}


/**
 * cdos_applet_manager_remove_applet_from_panel:
 * @obj: CdosApplet Object
 * @event: gchar * event
 * @keep_menu: gchar * keep_menu
 * @uuid: gchar * uuid
 *
 * Remove applet from panel
 *
 */
void
cdos_applet_manager_remove_applet_from_panel (CdosApplet *obj, gchar *event,
                                              gchar *keep_menu, const gchar *uuid)
{
    int i ;
    gchar **enabled_applets = enabled_applet_definitions->raw;
    for (i = 0; enabled_applets[i] != NULL; i ++)
    {
        AppletDefinition *applet_definition = cdos_applet_manager_get_applet_definition (enabled_applets[i]);
        if (applet_definition)
        {
            gboolean flag = FALSE;
            if (uuid &&
                !g_strcmp0 (applet_definition->uuid, uuid))
            {
                gchar **new_enabled_applets = g_strdupv (enabled_applets);
                GSettings *settings = g_settings_new ("org.cdos");
                new_enabled_applets = cdos_strv_elem_del_by_nth (new_enabled_applets, i);
                g_settings_set_strv (settings, "enabled-applets", (const gchar * const *)new_enabled_applets);
                g_strfreev(new_enabled_applets);
                flag = TRUE;
            }
            g_free(applet_definition->uuid);
            g_free(applet_definition);
            if(flag)
                break;
        }
    }
}

static CdosApplet *
cdos_applet_manager_create_applet (CdosExtension *extension, AppletDefinition *applet_definition)
{
    gchar *uuid = applet_definition->uuid;
    StSide orientation = applet_definition->orientation;
    gfloat panel_height = -1;

    CdosApplet *applet;
    applet = g_tree_lookup (applets_tree, uuid);
    if (applet != NULL)
    {
        cdos_applet_set_orientation (applet, orientation);

        return applet;
    }

    applet = cdos_extension_module_main (extension, orientation, panel_height, uuid);

    if (NULL == applet)
    {
        return NULL;
    }

    g_tree_insert (applets_tree, g_strdup (uuid), applet);
    cdos_applet_set_uuid (applet, uuid);

    return applet;
}

static gboolean
cdos_applet_manager_in_grouped_applets (gchar *uuid)
{
    g_return_val_if_fail (uuid != NULL, FALSE);
    for (guint i = 0; i < g_strv_length (grouped_applets); i++)
    {
        if (g_strcmp0 (grouped_applets[i], uuid) == 0)
        {
            return TRUE;
        }
    }
    return FALSE;
}

static void
cdos_applet_manager_load_group_applets (CdosExtension *extension, AppletDefinition *definition, CdosApplet *applet)
{
    cdos_applet_set_order (applet, definition->order); // 此order并不是针对grouped-applets的，并不是必要
    cdos_applet_set_extension (applet, extension);
    ClutterActor *org_location = cdos_applet_get_panel_location (applet);
    if (org_location != NULL)
    {
        g_object_ref (applet);
        clutter_actor_remove_child (CLUTTER_ACTOR (org_location), (ClutterActor *)applet);
        cdos_applet_set_panel_location (applet, NULL);
    }
    // CDOS_APPLET_CLASS(G_OBJECT_GET_CLASS(applets_group))->add_other_applet (applets_group, cdos_applet_get_actor (applet));
    applet_group_add_applet ((AppletGroup *)applets_group, CDOS_APPLET(applet));
    // cdos_applet_set_panel_location (applet, definition->location);
    GHashTable *loaded_definitions = cdos_extension_get_loaded_definitions (extension);
    if (!loaded_definitions) {
        loaded_definitions = g_hash_table_new (g_str_hash, g_str_equal);
        cdos_extension_set_loaded_definitions (extension, loaded_definitions);
    }
    g_hash_table_insert (loaded_definitions, definition->uuid, definition);
    CDOS_APPLET_CLASS(G_OBJECT_GET_CLASS(applet))->on_applet_added_to_panel(applet, applets_loaded);
}

static void
cdos_applet_manager_load_addtable_applets (gpointer key, gpointer value, gpointer user_data)
{
    CdosExtension *extension = key;
    AppletDefinition *definition = value;
    CdosApplet *applet = cdos_applet_manager_create_applet (extension, definition);
    cdos_applet_manager_load_group_applets (extension, definition, applet);
    g_hash_table_remove (add_to_group, extension);
}

static void
cdos_applet_manager_add_applet_to_panels (CdosExtension *extension, AppletDefinition *definition)
{
    /* create the applet */
    gchar *uuid = cdos_extension_get_uuid (extension);
    if (g_strcmp0 (uuid, "systray@cdos.org") == 0)
    {
        systray_flag ++;
        g_debug ("(%s %d) systray times %d\n",  __func__, __LINE__, systray_flag);
        if (systray_flag > 1)
            return;
    }
    CdosApplet *applet = cdos_applet_manager_create_applet (extension, definition);
    if (applet == NULL)
        return ;

    /* for grouped-applets */
    if (cdos_applet_manager_in_grouped_applets (definition->uuid))
    {
        if (group_loaded)
        {
            if (applets_group == NULL)
                group_loaded = FALSE;
            else
            {
                cdos_applet_manager_load_group_applets (extension, definition, applet);
                return ;
            }
        }
        if (!group_loaded)
        {
            if (add_to_group == NULL)
            {
                add_to_group = g_hash_table_new (g_direct_hash, g_direct_equal);
            }
            g_hash_table_insert (add_to_group, extension, definition);
            return ;
        }
    } // code for grouped-applets 与下文代码有重复，后续可进行优化整合，目前保证算法逻辑正常，以供测试


    cdos_applet_set_order (applet, definition->order);
    cdos_applet_set_extension (applet, extension);

    /* Remove it from its previous panel location (if it had one) */
    ClutterActor *panel_location = cdos_applet_get_panel_location (applet);
    if (panel_location != NULL)
    {
        g_object_ref (applet);
        clutter_actor_remove_child ((ClutterActor *)panel_location, (ClutterActor *)applet);
        cdos_applet_set_panel_location (applet, NULL);
    }
    /* Add it to its new panel location */
    ClutterActor *location = definition->location;

    ClutterActor *child = clutter_actor_get_first_child (location);
    ClutterActor *prev = NULL;
    while (child)
    {
        CdosApplet *app = CDOS_APPLET (child);
        if (app)
        {
            if (definition->order < cdos_applet_get_order (app))
                break;
        }
        prev = child;
        child = clutter_actor_get_next_sibling (child);
    }

    ClutterActor *applet_actor = (ClutterActor *)applet;
    if (prev)
        clutter_actor_insert_child_above ((ClutterActor *)location, applet_actor, prev);
    else
        clutter_actor_insert_child_below ((ClutterActor *)location, applet_actor, NULL);

    cdos_applet_set_panel_location (applet, location);

    GHashTable *loaded_definitions = cdos_extension_get_loaded_definitions (extension);
    if (!loaded_definitions)
    {
        loaded_definitions = g_hash_table_new (g_str_hash, g_str_equal);
        cdos_extension_set_loaded_definitions (extension, loaded_definitions);
    }
    g_hash_table_insert (loaded_definitions, definition->uuid, definition);

    CDOS_APPLET_CLASS(G_OBJECT_GET_CLASS(applet))->on_applet_added_to_panel(applet, applets_loaded);

    if (g_strcmp0 (definition->uuid, "grouped-applets@cdos.org") == 0)
    {
        applets_group = (CdosApplet *)applet;
        group_loaded = TRUE;
        if (add_to_group)
            g_hash_table_foreach (add_to_group, cdos_applet_manager_load_addtable_applets, NULL);
    }
}

static gboolean
cdos_applet_manager_applet_definitions_equal (AppletDefinition *a, AppletDefinition *b)
{
  if ((a->panel == b->panel) &&
    (a->orientation == b->orientation) &&
    (a->location == b->location) &&
    (a->order == b->order))
  {
    return TRUE;
  }
  return FALSE;
}

/**
 * cdos_applet_manager_finish_extension_load:
 * @extension: CdosExtension Object
 *
 * Returns: gboolean
 */
gboolean
cdos_applet_manager_finish_extension_load (CdosExtension *extension)
{
    AppletDefinition *definition;
    display_current_time (cdos_extension_get_uuid (extension));
    /* add all applet instances for this extension */
    definition = g_hash_table_lookup (enabled_applet_definitions->uuid_map, cdos_extension_get_uuid (extension));
    cdos_applet_manager_add_applet_to_panels (extension, definition);

    display_current_time (cdos_extension_get_uuid (extension));
    return TRUE;
}

static void
cdos_applet_manager_remove_all_cb (const gchar *key, gpointer value, gpointer user_data)
{
    EnabledAppletDefinitions *new_enabled_applet_definitions = user_data;
    if (!g_hash_table_lookup (new_enabled_applet_definitions->uuid_map, key))
    {
        g_debug ("%s %s", __func__, key);
        cdos_applet_manager_remove_applet_from_panels (value);
    }
}

static void
cdos_applet_manager_add_or_move_cb (gpointer key, gpointer value, gpointer user_data)
{
    AppletDefinition *new_def = value;
    AppletDefinition *old_def = g_hash_table_lookup (enabled_applet_definitions->uuid_map, key);
    if (!old_def || !cdos_applet_manager_applet_definitions_equal (new_def, old_def))
    {
        CdosExtension *extension = cdos_extension_get_object_by_uuid (new_def->uuid);
        if (extension)
        {
            cdos_applet_manager_add_applet_to_panels (extension, new_def);
        }
    }
}

static void
cdos_applet_manager_make_sure_cb (gpointer key, gpointer value, gpointer user_data)
{
    cdos_extension_load_extension (key);
}

static void
cdos_applet_manager_on_enabled_applets_changed (void)
{
  EnabledAppletDefinitions *new_enabled_applet_definitions =
        cdos_applet_manager_get_enabled_applet_definitions ();

  /* Remove all applet instances that do not exist in the definition anymore */
  g_hash_table_foreach (enabled_applet_definitions->uuid_map,
                        (GHFunc)cdos_applet_manager_remove_all_cb,
                        new_enabled_applet_definitions);

  /* Add or move applet instances of already loaded applet extensions */
  g_hash_table_foreach (new_enabled_applet_definitions->uuid_map,
                        cdos_applet_manager_add_or_move_cb,
                        NULL);

  enabled_applet_definitions = new_enabled_applet_definitions;

  /* Make sure all applet extensions are loaded.
   * Once loaded, the applets will add themselves via finishExtenisonLoad.
   * here via this (cdos_applet_manager_finish_extension_load).
   */
  g_hash_table_foreach (enabled_applet_definitions->uuid_map,
                        cdos_applet_manager_make_sure_cb,
                        NULL);

  //CdosIconDispatcher * status_icon_dispatcher = cdos_ui_plugin_get_status_icon_dispatcher (cdos_ui_plugin_get ());
  //cdos_icon_dispatcher_redisplay (status_icon_dispatcher);
}

static gchar **
cdos_applet_manager_get_grouped_applets (void)
{
  GSettings *settings = g_settings_new ("org.cdos");
  return g_settings_get_strv (settings, "grouped-applets");
}

static void
cdos_applet_manager_on_grouped_applets_changed (GSettings *settings, gchar *key,gpointer user_data)
{
  grouped_applets = cdos_applet_manager_get_grouped_applets ();

  gchar **enabled_applets_tmp = g_settings_get_strv (settings, "enabled-applets");
  /* 暂且通过cdos_applet_manager_on_enabled_applets_changed 来做grouped-applets的显示更新（实际上是更新panel） */
//  cdos_applet_manager_on_enabled_applets_changed ();
  applet_group_remove_all_applets ((AppletGroup *)applets_group);

  int i = 0;
  for (i = 0; enabled_applets_tmp[i] != NULL; i ++)
  {
    AppletDefinition *applet_definition  = cdos_applet_manager_get_applet_definition (enabled_applets_tmp[i]);
    if (applet_definition)
    {
        CdosExtension *extension = cdos_extension_get_object_by_uuid (applet_definition->uuid);
        if (extension)
        {
            cdos_applet_manager_add_applet_to_panels (extension, applet_definition);
        }

        g_free (applet_definition->uuid);
        g_free (applet_definition);

    }
  }
  g_strfreev (enabled_applets_tmp);
}

/**
 * cdos_applet_manager_init:
 *
 * init for applet manager
 *
 */
void
cdos_applet_manager_init (void)
{
  GSettings *settings = g_settings_new ("org.cdos");

  grouped_applets = cdos_applet_manager_get_grouped_applets ();
  status_icon_dispatcher = cdos_ui_plugin_get_status_icon_dispatcher ();

  applets_loaded = FALSE;

  /* Load all applet extensions, the applets themselves will be added in finishExtension load */
  enabled_applet_definitions = cdos_applet_manager_get_enabled_applet_definitions ();

  applets_loaded = TRUE;

  g_signal_connect (settings, "changed::enabled-applets", G_CALLBACK (cdos_applet_manager_on_enabled_applets_changed), NULL);
  g_signal_connect (settings, "changed::grouped-applets", G_CALLBACK (cdos_applet_manager_on_grouped_applets_changed), NULL);

  if (!applets_tree)
  {
      applets_tree = g_tree_new_full ((GCompareDataFunc)g_strcmp0, NULL, g_free, NULL);
  }
  cdos_applet_manager_on_enabled_applets_changed ();
}

static gboolean
set_applet_orientation (const char *uuid, CdosApplet *applet, gpointer data)
{
    cdos_applet_set_orientation (applet, GPOINTER_TO_UINT (data));
    return FALSE;
}

void
cdos_applet_manager_set_applets_orientation (StSide orientation)
{
    if (!applets_tree)
        return;
    g_tree_foreach (applets_tree, (GTraverseFunc)set_applet_orientation, GUINT_TO_POINTER (orientation));
}
