/**
 * SECTION: cdos-applet
 * @short_description: 小应用开发的抽象基类
 * @title: CdosApplet
 * @include: cdos-applet.h
 *
 * #CdosApplet为小应用开发提供抽象基类,该基类对存储了小应用的基本信息以及提供了对停泊元素(如#CdosPanel)方面的相关操作。
 *
 * #CdosApplet只能被继承后使用，不推荐实例化后使用。
 *
 *
 */
#include <math.h>

#include <gio/gio.h>
//#include <girepository.h>
#include "config.h"

#include "cdos-applet.h"
#include "cdos-start.h"
#include "cdos-layout-manager.h"
#include "cdos-popup.h"
#include "cdos-panel-item-tooltip.h"
#include "cdos-applet-manager.h"


struct _CdosAppletPrivate
{
    ClutterActor   *panel_location;
    ClutterActor   *newPanelLocation;
    guint           order;
    GSettings      *settings;
    CdosPanelItemTooltip *applet_tooltip;
    gchar          *uuid;
    AppletDefinition *definition;
    gfloat          panel_height;

    CdosExtension  *extension;
    guint           new_order;
    StSide          orientation;
    gboolean        with_menu;

    CdosPopupMenu         *popup_menu;
    CdosPopupMenu         *context_menu;
    CdosPopupMenuManager  *menu_manager;
    gboolean popup_menu_active;
    gboolean context_menu_active;
    gulong popup_menu_state_id;
    gulong context_menu_state_id;
};

G_DEFINE_TYPE_WITH_PRIVATE (CdosApplet, cdos_applet, ST_TYPE_WIDGET);

enum
{
    PROP_0,
    PROP_ORIENTATION,
    PROP_PANEL_HEIGHT,
    PROP_SCALEMODE,
    PROP_INSTANCE_ID,
    PROP_WITH_MENU
};

static void cdos_applet_set_popup_menu (CdosApplet *self, guint orientation);

static void menu_on_open_state_changed (CdosPopupMenuBase *popup_menu, gboolean open,CdosApplet *self);

static void
cdos_applet_dispose (GObject *gobject)
{
    g_debug ("%s", __func__);;
    CdosAppletPrivate *priv = CDOS_APPLET (gobject)->priv;

    g_clear_object (&priv->panel_location);
    g_clear_object (&priv->newPanelLocation);
    g_clear_pointer ((ClutterActor **)&priv->popup_menu, clutter_actor_destroy);
    g_clear_pointer ((ClutterActor **)&priv->context_menu, clutter_actor_destroy);
    g_clear_object (&priv->menu_manager);
    G_OBJECT_CLASS (cdos_applet_parent_class)->dispose (gobject);
}

static void
cdos_applet_set_property (GObject *gobject,
                          guint prop_id,
                          const GValue *value,
                          GParamSpec   *pspec)
{
  CdosApplet *obj = CDOS_APPLET (gobject);
  switch (prop_id)
  {
    case PROP_ORIENTATION:
      cdos_applet_set_orientation (obj, g_value_get_int(value));
      break;
    case PROP_PANEL_HEIGHT:
      cdos_applet_set_panel_height (obj, g_value_get_float (value));
      break;
    case PROP_WITH_MENU:
      g_debug ("set popup menu active");
      cdos_applet_set_popup_menu_active (obj ,g_value_get_boolean (value));
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (gobject, prop_id, pspec);
      break;
  }
}

static void
cdos_applet_get_property(GObject    *gobject,
                        guint       prop_id,
                        GValue     *value,
                        GParamSpec *pspec)
{

    CdosAppletPrivate *priv = NULL;

    priv = CDOS_APPLET (gobject)->priv;

    switch (prop_id)
    {
        case PROP_ORIENTATION:
            g_value_set_int (value, priv->order);
            break;
        case PROP_PANEL_HEIGHT:
            g_value_set_float (value, priv->panel_height);
            break;
        case PROP_WITH_MENU:
            g_value_set_boolean (value, priv->with_menu);
            break;
        default:
            G_OBJECT_WARN_INVALID_PROPERTY_ID (gobject, prop_id, pspec);
            break;
    }
}

static void
menu_on_open_state_changed (CdosPopupMenuBase *popup_menu, gboolean open, CdosApplet *self)
{
    CdosAppletPrivate *priv = self->priv;
    if (priv->popup_menu_active || priv->context_menu_active)
    {
        if (open)
        {
            st_widget_add_style_pseudo_class (ST_WIDGET (self), "active");
            cdos_tooltip_set_prevent_show (CDOS_TOOLTIP(priv->applet_tooltip), TRUE);
        } else {
            st_widget_remove_style_pseudo_class (ST_WIDGET (self), "active");
            cdos_tooltip_set_prevent_show (CDOS_TOOLTIP(priv->applet_tooltip), FALSE);
        }
    }
}

static void
cdos_applet_set_popup_menu (CdosApplet *self, guint orientation)
{
    CdosAppletPrivate *priv = self->priv;
    CdosPopupMenu *popup_menu;
    gchar style[20]= {0,};
    ClutterActor *ui_group = NULL;
    if (priv->popup_menu)
        return;
    ui_group = cdos_ui_plugin_get_ui_group ();
    popup_menu = cdos_popup_menu_new_with_params ((ClutterActor *)self, 0.5, orientation);
    if ( ui_group && popup_menu)
        clutter_actor_add_child (ui_group, CLUTTER_ACTOR(popup_menu->actor));
    clutter_actor_hide((ClutterActor *)popup_menu);

    const MetaRectangle *monitor = NULL;
    CdosLayoutManager *layout = cdos_layout_manager_get_default ();
    CdosPanel *panel = cdos_ui_plugin_get_panel ();
    monitor = cdos_layout_manager_get_primary_monitor (layout);
    gfloat panelheight = clutter_actor_get_height ((ClutterActor *)panel);

    StThemeNode *theme_node = st_widget_get_theme_node (ST_WIDGET (popup_menu));
    gdouble gap_length = st_theme_node_get_length ( theme_node, "-boxpointer-gap");

    int max_height = 1000;
    if (monitor)
        max_height = round(monitor->height - panelheight - gap_length);

    memset (style, 0, 20);
    g_snprintf (style, 20, "max-height: %dpx", max_height);
    st_widget_set_style ((StWidget *)popup_menu, style);
    priv->popup_menu = popup_menu;

    CdosBoxPointer *bp = cdos_popup_menu_get_box_pointer (popup_menu);
    st_widget_add_style_class_name ((StWidget *)bp, "applet-popup-menu-boxpointer");

    cdos_popup_menu_manager_add_menu (priv->menu_manager, CDOS_POPUP_MENU_BASE (priv->popup_menu), -1);
    priv->popup_menu_state_id = g_signal_connect (popup_menu,
                                                  "open-state-changed",
                                                  G_CALLBACK (menu_on_open_state_changed),
                                                  self);
}

static void
cdos_applet_set_context_menu (CdosApplet *self, guint orientation)
{
    CdosAppletPrivate *priv = self->priv;
    ClutterActor *ui_group = NULL;
    if (priv->context_menu)
        return ;

    ui_group = cdos_ui_plugin_get_ui_group ();
    priv->context_menu = cdos_popup_menu_new_with_params ((ClutterActor *)self, 0.5, orientation);
    clutter_actor_add_child (ui_group, CLUTTER_ACTOR(priv->context_menu->actor));
    clutter_actor_hide((ClutterActor *)priv->context_menu);

    CdosBoxPointer *bp = cdos_popup_menu_get_box_pointer (priv->context_menu);
    st_widget_add_style_class_name ((StWidget *)bp, "applet-context-menu-boxpointer");

    cdos_popup_menu_manager_add_menu (priv->menu_manager, 
                                      CDOS_POPUP_MENU_BASE (priv->context_menu), 
                                      -1);
    priv->context_menu_state_id = g_signal_connect (priv->context_menu,
                                                    "open-state-changed",
                                                    G_CALLBACK (menu_on_open_state_changed),
                                                    self);
}

static gboolean
cdos_applet_on_applet_clicked (CdosApplet *self)
{
    return FALSE;
}

/**
 * cdos_applet_set_applet_tooltip:
 * @self: 当前小应用对象实例
 * @text: 需要显示的tooltip内容
 *
 * 设置当前小应用需要显示的tooltip内容。
 */
void
cdos_applet_set_applet_tooltip (CdosApplet *self, const gchar *text)
{
    CdosAppletPrivate *priv = cdos_applet_get_instance_private (self);

    // FIXME StLabel *label = cdos_tooltip_get_label (priv->applet_tooltip);
    st_label_set_text ((StLabel *)priv->applet_tooltip, text);
}

/**
 * This function is called by appletManager when the applet is added to the panel.
 * This is meant to be overridden in individual applets.
 */
static gboolean
cdos_applet_on_applet_added_to_panel (CdosApplet *applet, gboolean user_enabled)
{
    return FALSE;
}

static gboolean
cdos_applet_on_applet_removed_from_panel (CdosApplet *applet)
{
  // CdosAppletPrivate *priv = CDOS_APPLET_GET_PRIVATE (applet);

  return FALSE;
}

static gboolean
cdos_applet_on_panel_height_changed (void)
{
    return FALSE;
}

/**
 * cdos_applet_set_orientation:
 * @applet: 当前小应用对象实例
 * @orientation: 对其方式，#StSide枚举成员
 *
 * 设置当前小应用及其tooltip的上下对其方式；
 */
void
cdos_applet_set_orientation (CdosApplet *self, StSide orientation)
{
    g_return_if_fail (CDOS_IS_APPLET (self));

    CdosAppletPrivate *priv = cdos_applet_get_instance_private (self);
    priv->orientation = orientation;
    if(priv->popup_menu)
        cdos_popup_menu_set_arrow_side (priv->popup_menu, orientation);

    if(priv->context_menu)
        cdos_popup_menu_set_arrow_side (priv->context_menu, orientation);

    CdosPanelItemTooltip *tooltip = priv->applet_tooltip;
    tooltip->orientation = orientation;

    if (CDOS_APPLET_CLASS (G_OBJECT_GET_CLASS (self))->on_orientation_changed)
        CDOS_APPLET_CLASS (G_OBJECT_GET_CLASS (self))->on_orientation_changed (self, orientation);
}

/**
 * cdos_applet_get_panel_height:
 * @applet: 当前小应用实例对象
 *
 * 获取面板高度
 *
 * Returns:返回高度尺寸
 */
gfloat
cdos_applet_get_panel_height (CdosApplet *self)
{
    CdosAppletPrivate *priv = cdos_applet_get_instance_private (self);
    return priv->panel_height;
}

/**
 * cdos_applet_set_panel_height:
 * @applet: 当前小应用实例对象
 * @panel_height: 面板高度
 *
 * 设置面板高度值，注意：此函数并不是要修改面板的实际高度，而是根据面板的实际高度，将其值保存在当前结构体内。
 */
void
cdos_applet_set_panel_height (CdosApplet *self, gfloat panel_height)
{
    CdosAppletPrivate *priv = cdos_applet_get_instance_private (self);
    priv->panel_height = panel_height;
}
/**
 * cdos_applet_get_actor：
 * @applet：当前小应用实例对象
 *
 * 获取当前小应用实例的主容器元素；
 *
 * Returns:返回一个#StBoxLayout的实例对象。
 *
 */
StWidget *
cdos_applet_get_actor (CdosApplet *applet)
{
  return ST_WIDGET (applet);
}

/**
 * cdos_applet_get_panel_location:
 * @applet: 当前小应用实例对象
 *
 * 获取当前小应用所在的面板区域的元素对象；
 *
 * Returns:返回一个#StBoxLayout实例对象。
 */
ClutterActor *
cdos_applet_get_panel_location (CdosApplet *applet)
{
  CdosAppletPrivate *priv = applet->priv;
  return priv->panel_location;
}
/**
 * cdos_applet_set_panel_location:
 * @self: 当前小应用实例对象
 * @panellocation: 面板区域元素
 *
 * 设置当前小应用所在的面板区域元素，该函数不推荐在内部使用。
 */
void
cdos_applet_set_panel_location (CdosApplet *self, ClutterActor *panellocation)
{
    CdosAppletPrivate *priv = cdos_applet_get_instance_private (self);
    if (NULL != panellocation)
    {
        priv->panel_location = g_object_ref (panellocation);
    }
    else
    {
        g_object_unref (priv->panel_location);
        priv->panel_location = NULL;
    }
}
/**
 * cdos_applet_get_new_panel_location:
 * @applet: 当前小应用实例对象
 *
 * 获取当前小应用的所在面板中的新区域元素，该成员主要用于小应用的移动。
 *
 * Returns: 返回一个#StBoxLayout的实例对象。
 */
ClutterActor *
cdos_applet_get_new_panel_location (CdosApplet *applet)
{
  g_return_val_if_fail (CDOS_IS_APPLET (applet), NULL);
  return applet->priv->newPanelLocation;
}
/**
 * cdos_applet_set_new_panel_location:
 * @applet: 当前小应用实例对象
 * @new_panel_location: 面板区域元素
 *
 * 给当前小应用设置一个新的面板区域，当移动当前小应用的时候。
 */
void
cdos_applet_set_new_panel_location (CdosApplet *self,
                                    ClutterActor *new_panel_location)
{
    g_return_if_fail (CDOS_IS_APPLET (self));
    CdosAppletPrivate *priv = cdos_applet_get_instance_private (self);
    priv->newPanelLocation = g_object_ref (new_panel_location);
}
/**
 * cdos_applet_get_order:
 * @applet: 当前小应用实例对象
 *
 * 获取当前小应用在面板中的位置序列号。
 *
 * Returns: 返回当前的位置序列号。
 */
gint
cdos_applet_get_order (CdosApplet *self)
{
  g_return_val_if_fail (CDOS_IS_APPLET (self), -1);
  return self->priv->order;
}
/**
 * cdos_applet_set_order:
 * @self: 当前小应用实例对象
 * @order: 新的序列号；
 *
 * 设置当前小应用所在面板中新的位置序列号，该函数没有更改对应的GSettings中对应的序列号，使用时，需要在外部对GSettings进行更新。
 */
void
cdos_applet_set_order (CdosApplet *self, guint order)
{
    g_return_if_fail (CDOS_IS_APPLET (self));

    CdosAppletPrivate *priv = cdos_applet_get_instance_private (self);
    priv->order = order;
}
/**
 * cdos_applet_set_extension:
 * @applet: 当前小应用实例对象
 * @extension: 扩展实例对象
 *
 * 为当前小应用绑定扩展对象。
 */
void cdos_applet_set_extension (CdosApplet *applet,
            CdosExtension *extension)
{
  g_return_if_fail (CDOS_IS_APPLET (applet));
  CdosAppletPrivate *priv = applet->priv;
  priv->extension = extension;
}
/**
 * cdos_applet_get_extension
 * @applet: 当前小应用实例对象
 *
 * 获取当前小应用绑定的扩展对象。
 *
 * Returns: 返回一个#CdosExtension对象。
 */
CdosExtension *
cdos_applet_get_extension (CdosApplet *applet)
{
  g_return_val_if_fail (CDOS_IS_APPLET (applet), NULL);
  return applet->priv->extension;
}

static gboolean
cdos_applet_on_button_press (ClutterActor *actor, ClutterButtonEvent *event)
{
    CdosApplet *self = CDOS_APPLET (actor);
    CdosAppletPrivate *priv = self->priv;
    if (clutter_event_get_button ((ClutterEvent *)event) == 1)
    {
        if (priv->context_menu_active)
            cdos_popup_menu_close (CDOS_POPUP_MENU_BASE(priv->context_menu), FALSE);
        if (priv->popup_menu_active)
        {
            if (CDOS_APPLET_CLASS (G_OBJECT_GET_CLASS (self))->on_applet_clicked (self))
                return TRUE;

            cdos_popup_menu_base_toggle (CDOS_POPUP_MENU_BASE (priv->popup_menu));
            return TRUE;
        }
    }
    else if (clutter_event_get_button ((ClutterEvent *)event) == 3)
    {
        if (priv->popup_menu_active)
            cdos_popup_menu_close (CDOS_POPUP_MENU_BASE(priv->popup_menu), FALSE);
        if (priv->context_menu_active)
            cdos_popup_menu_base_toggle (CDOS_POPUP_MENU_BASE (priv->context_menu));
        return TRUE;
    }
    return TRUE;
}

static gboolean
cdos_applet_on_button_release (ClutterActor *actor, ClutterButtonEvent *event)
{
    return TRUE;
}

static void
cdos_applet_allocate (ClutterActor *actor,
                      const ClutterActorBox *box,
                      ClutterAllocationFlags flags)
{
    StThemeNode *theme_node = st_widget_get_theme_node ((StWidget *)actor);
    ClutterActorBox content_box;

    clutter_actor_set_allocation (actor, box, flags);

    st_theme_node_get_content_box (theme_node, box, &content_box);

    gfloat width = content_box.x2 - content_box.x1;
    gfloat height = content_box.y2 - content_box.y1;

    ClutterActor *child = NULL;
    ClutterActorIter iter;
    clutter_actor_iter_init (&iter, actor);
    while (clutter_actor_iter_next (&iter, &child))
    {
        ClutterActorBox child_box;
        gfloat min_w, nat_w;
        gfloat min_h, nat_h;
        clutter_actor_get_preferred_size (child, &min_w, &min_h, &nat_w, &nat_h);
        child_box.x1 = content_box.x1 + (width - min_w) / 2;
        child_box.x2 = child_box.x1 + min_w;
        child_box.y1 = content_box.y1 + (height - min_h) / 2;
        child_box.y2 = child_box.y1 + min_h;
        clutter_actor_allocate (child, &child_box, flags);
    }
}

static void
cdos_applet_class_init (CdosAppletClass *klass)
{
    GObjectClass *gobject_class = G_OBJECT_CLASS (klass);

    gobject_class->dispose = cdos_applet_dispose;
    gobject_class->set_property = cdos_applet_set_property;
    gobject_class->get_property = cdos_applet_get_property;

    ClutterActorClass *actor_class = CLUTTER_ACTOR_CLASS (klass);
    actor_class->allocate = cdos_applet_allocate;
    actor_class->button_press_event = cdos_applet_on_button_press;
    actor_class->button_release_event = cdos_applet_on_button_release;

    klass->on_applet_clicked = cdos_applet_on_applet_clicked;
    klass->on_applet_added_to_panel = cdos_applet_on_applet_added_to_panel;
    klass->on_applet_removed_from_panel = cdos_applet_on_applet_removed_from_panel;
    klass->on_panel_height_changed = cdos_applet_on_panel_height_changed;

    GParamSpec *pspec;
    /**
     * CdosApplet:orientation:
     *
     * 小应用对齐方式
     */
    pspec = g_param_spec_int ("orientation",
                              "orientation",
                              "orientaion of this applet",
                              0,
                              3,
                              1,
                              G_PARAM_READWRITE);
    g_object_class_install_property (gobject_class,
                                     PROP_ORIENTATION,
                                     pspec);
    /**
     * CdosApplet:scalemode:
     *
     * 是否是变换模式
     */
    pspec = g_param_spec_boolean ("scalemode",
                                  "scale actor",
                                  "actor can scale or not",
                                  FALSE,
                                  G_PARAM_READWRITE);
    g_object_class_install_property (gobject_class,
                                     PROP_SCALEMODE,
                                     pspec);
    /**
     * CdosApplet:panel-height:
     *
     * 当前面板高度
     */
    pspec = g_param_spec_float ("panel-height",
                                "panel height",
                                "the height of panel",
                                -1,
                                200,
                                40,
                                G_PARAM_READWRITE);
    g_object_class_install_property (gobject_class,
                                     PROP_PANEL_HEIGHT,
                                     pspec);
    /**
     * CdosApplet:instance-id:
     *
     * 小应用实例id
     */
    pspec = g_param_spec_string ("instance-id",
                                 "instance id",
                                 "the instance id of the applet",
                                 NULL,
                                 G_PARAM_READWRITE);
    g_object_class_install_property (gobject_class,
                                     PROP_INSTANCE_ID,
                                     pspec);

    pspec = g_param_spec_boolean ("with-menu",
                                  "with popup menu",
                                  "The applet popup menu",
                                  FALSE,
                                  G_PARAM_READWRITE);
    g_object_class_install_property (gobject_class,
                                     PROP_WITH_MENU,
                                     pspec);
}

static void
cdos_applet_init (CdosApplet *self)
{
    CdosAppletPrivate *priv = NULL;

    priv = self->priv = cdos_applet_get_instance_private (self);

    priv->order = 0;
    priv->new_order = -1;
    StWidget *actor = (StWidget *)self;

    priv->panel_location = NULL;
    priv->newPanelLocation = NULL;
    CdosPanel *panel = cdos_ui_plugin_get_panel ();
    priv->panel_height = cdos_panel_get_height (panel);
    priv->uuid = NULL;
    priv->applet_tooltip = cdos_panel_item_tooltip_new ((ClutterActor *)actor,
                                                        "", ST_SIDE_BOTTOM);
    priv->menu_manager = cdos_popup_menu_manager_new ((ClutterActor *)self);
    st_widget_set_style_class_name (actor, "applet-box");
    clutter_actor_set_reactive ((ClutterActor*) actor, TRUE);
    st_widget_set_track_hover (actor, TRUE);
}

/**
 * cdos_applet_get_uuid:
 * @applet: 当前小应用实例对象
 *
 * 获取当前小应用的uuid。
 *
 * Returns: 当前小应用的uuid。
 */
gchar *
cdos_applet_get_uuid (CdosApplet *applet)
{
  g_return_val_if_fail (CDOS_IS_APPLET (applet), NULL);
  return applet->priv->uuid;
}
/**
 * cdos_applet_set_uuid:
 * @applet: 当前小应用实例对象
 * @uuid: 设置当前小应用的uuid
 *
 * 设置当前小应用的uuid,该方法在构造过程中使用，没有更新GSettings内相应的值，不推荐在外部使用。
 */
void cdos_applet_set_uuid (CdosApplet *applet, gchar *uuid)
{
  g_return_if_fail (CDOS_IS_APPLET (applet));
  applet->priv->uuid = uuid;
}

/**
 * cdos_applet_get_new_order:
 * @applet: 当前小应用实例对象
 *
 * 获取当前小应用新的序列号，通常在当前小应用移动之后使用。
 *
 * Returns: 返回新的序列号。
 */
gint
cdos_applet_get_new_order (CdosApplet *applet)
{
  g_return_val_if_fail (CDOS_IS_APPLET(applet), -1);
  return applet->priv->new_order;
}

/**
 * cdos_applet_set_new_order:
 * @applet: 当前小应用实例对象
 *
 * 设置当前小应用新的序列号，通常在当前小应用移动过程中使用。
 */
void cdos_applet_set_new_order (CdosApplet *applet,
                                gint new_order)
{
  g_return_if_fail (CDOS_IS_APPLET (applet));
  applet->priv->new_order = new_order;
}

void
cdos_applet_set_popup_menu_active (CdosApplet *self, gboolean active)
{
    g_return_if_fail (CDOS_IS_APPLET (self));
    CdosAppletPrivate *priv = self->priv;
    if (priv->popup_menu_active == active)
        return;

    if(active)
    {
        if(priv->popup_menu == NULL){
            cdos_applet_set_popup_menu(self,ST_SIDE_BOTTOM);
        }
        if(priv->popup_menu_state_id == 0){
            priv->popup_menu_state_id = \
            g_signal_connect(CDOS_POPUP_MENU_BASE(priv->popup_menu),"open-state-changed",G_CALLBACK(menu_on_open_state_changed),self);
        }
    }else{
        cdos_popup_menu_close (CDOS_POPUP_MENU_BASE(priv->popup_menu), FALSE);
        if (priv->popup_menu_state_id > 0)
        {
            g_signal_handler_disconnect (CDOS_POPUP_MENU_BASE(priv->popup_menu), priv->popup_menu_state_id);
            priv->popup_menu_state_id = 0L;
        }
    }
    priv->popup_menu_active = active;
}

CdosPopupMenu *
cdos_applet_get_popup_menu (CdosApplet *self)
{
    g_return_val_if_fail (CDOS_IS_APPLET (self),NULL);
    CdosAppletPrivate *priv = self->priv;

    if(priv->popup_menu == NULL)
        cdos_applet_set_popup_menu (self, priv->orientation);

    return priv->popup_menu;
}

gboolean
cdos_applet_get_popup_menu_active (CdosApplet *self)
{
    g_return_val_if_fail (CDOS_IS_APPLET (self),FALSE);
    CdosAppletPrivate *priv = self->priv;
    return priv->popup_menu_active;
}

CdosPopupMenuManager *
cdos_applet_get_menu_manager (CdosApplet *self)
{
    g_return_val_if_fail (CDOS_IS_APPLET (self),NULL);
    CdosAppletPrivate *priv = self->priv;
    if (!priv->menu_manager)
        priv->menu_manager = cdos_popup_menu_manager_new ((ClutterActor *)self);
    return priv->menu_manager;
}

void 
cdos_applet_set_context_menu_active (CdosApplet *self, gboolean active)
{
    g_return_if_fail (CDOS_IS_APPLET (self));
    CdosAppletPrivate *priv = self->priv;
    if (priv->context_menu_active == active)
        return;
    if(active)
    {
        if(priv->context_menu == NULL)
        {
            cdos_applet_set_context_menu (self, ST_SIDE_BOTTOM);
        }
        if(priv->context_menu_state_id == 0)
        {
            priv->context_menu_state_id = \
                g_signal_connect (CDOS_POPUP_MENU_BASE (priv->context_menu),
                                  "open-state-changed",
                                  G_CALLBACK (menu_on_open_state_changed),
                                  self);
        }
    } else {
        if (priv->context_menu)
        {
            cdos_popup_menu_close (CDOS_POPUP_MENU_BASE(priv->context_menu), FALSE);
            if (priv->context_menu_state_id > 0)
            {
                g_signal_handler_disconnect (CDOS_POPUP_MENU_BASE(priv->context_menu), 
                                            priv->context_menu_state_id);
                priv->context_menu_state_id = 0L;
            }
        }
    }
    priv->context_menu_active = active;
}

gboolean 
cdos_applet_get_context_menu_active (CdosApplet *self)
{
    g_return_val_if_fail (CDOS_IS_APPLET (self), FALSE);
    CdosAppletPrivate *priv = self->priv;
    return priv->context_menu_active;
}

CdosPopupMenu *
cdos_applet_get_context_menu (CdosApplet *self)
{
    g_return_val_if_fail (CDOS_IS_APPLET (self), NULL);
    CdosAppletPrivate *priv = self->priv;
    if (priv->context_menu == NULL)
        cdos_applet_set_context_menu (self, priv->orientation);
    return priv->context_menu;
}