#include <meta/meta-background-actor.h>
#include <meta/meta-background-group.h>
#include <sys/time.h>
#include <math.h>

#include "common.h"
#include "cdos-appswitcher-3d.h"
#include "cdos-ui-plugin.h"

#define INITIAL_DELAY_TIMEOUT 150
#define CHECK_DESTROYED_TIMEOUT 100
#define TRANSITION_TYPE CLUTTER_EASE_OUT_QUAD
#define ICON_SIZE 64
#define ICON_SIZE_BIG 128
#define ICON_TITLE_SPACING 10
#define PREVIEW_SCALE 0.5

#define TITLE_POSITION 7.0 / 8.0f
#define ANIMATION_TIME 250 // miscellaneousseconds
#define SWITCH_TIME_DELAY 100 // ms
#define DIM_FACTOR 0.4
#define BACKGROUND_SCHEMA "org.cdos.desktop.background"

static void
remove_preview_data (gpointer key, PreviewData *data, gpointer user_data)
{
    if (data != NULL)
        g_free (data);
}

/* 虚函数 */
static gboolean appswitcher3d_show (AppSwitcher *self);
static void appswitcher3d_hide (AppSwitcher *self);

static void appswitcher3d_on_destroy (AppSwitcher *self);

static void appswitcher3d_create_list (AppSwitcher *self);
static void appswitcher3d_update_list (AppSwitcher *self, gint index);

static void appswitcher3d_select_next (AppSwitcher *self);
static void appswitcher3d_select_previous (AppSwitcher *self);

static void appswitcher3d_on_workspace_selected (AppSwitcher *self, MetaWorkspace *workspace);
static gboolean appswitcher3d_check_switch_time (AppSwitcher *self);

static void appswitcher3d_set_current_window (AppSwitcher *self, MetaWindow *window);

static void appswitcher3d_destroy_actors (AppSwitcher *self);
/* ************** Virtual Function End ************** */

static void appswitcher3d_enable_monitor_fix (AppSwitcher3D *switcher);
static void appswitcher3d_disable_monitor_fix (AppSwitcher3D *switcher);
static void clone_clicked (StButton *button, gint clicked_button, AppSwitcher3D *switcher);

void
cdos_appswitcher3d_init (AppSwitcher3D *switcher, MetaKeyBinding *binding)
{
    g_return_if_fail (switcher != NULL);
    // AppSwitcher3D *switcher = g_new0 (AppSwitcher3D, 1);
    AppSwitcher *parent = (AppSwitcher *)switcher;

    appswitcher_init (parent, binding);

    switcher->preview_table = g_hash_table_new (0, 0);

    parent->actor = (ClutterActor *)g_object_new (ST_TYPE_WIDGET, NULL);
    clutter_actor_set_reactive (parent->actor, 1);
    parent->actor->flags |= CLUTTER_ACTOR_VISIBLE;
    clutter_actor_hide (parent->actor);

    switcher->background_group = meta_background_group_new ();
    clutter_actor_add_child (parent->actor, switcher->background_group);
    switcher->background = cdos_background_manager_new (switcher->background_group, 1, 0, TRUE,BACKGROUND_SCHEMA);
    // st_container_lower ((ClutterContainer *)parent->actor, switcher->background, NULL);
    switcher->preview_actor = (ClutterActor *)g_object_new (ST_TYPE_WIDGET, NULL);
    clutter_actor_set_reactive (switcher->preview_actor, 1);

    clutter_actor_add_child (parent->actor, switcher->preview_actor);

    ClutterActor *ui_group = cdos_ui_plugin_get_ui_group ();
    clutter_actor_add_child (ui_group, parent->actor);

    parent->show = appswitcher3d_show;
    parent->hide = appswitcher3d_hide;
    parent->check_switch_time = appswitcher3d_check_switch_time;
    parent->on_workspace_selected = appswitcher3d_on_workspace_selected;
    parent->create_list = appswitcher3d_create_list;
    parent->set_current_window = appswitcher3d_set_current_window;
    parent->destroy_actors = appswitcher3d_destroy_actors;
    parent->on_destroy = appswitcher3d_on_destroy;

    parent->select_next = appswitcher3d_select_next;
    parent->select_previous = appswitcher3d_select_previous;
    parent->update_list = appswitcher3d_update_list;
}

static gboolean
appswitcher3d_show (AppSwitcher *self)
{
    AppSwitcher3D *switcher = (AppSwitcher3D *)self;

    appswitcher3d_enable_monitor_fix (switcher);

    const MetaRectangle *monitor = self->active_monitor;
    clutter_actor_set_position (switcher->preview_actor, monitor->x, monitor->y);
    clutter_actor_set_size (switcher->preview_actor, monitor->width, monitor->height);
    clutter_actor_set_clip (switcher->preview_actor, 0, 0, monitor->width, monitor->height);

    // create previews
    appswitcher3d_create_list (self);

    // hide windows and show Coverflow actors
    ClutterActor *window_group = cdos_global_get_window_group (global);
    clutter_actor_hide (window_group);

    clutter_actor_show (self->actor);

    MetaDisplay *display = cdos_global_get_display (global);
    meta_disable_unredirect_for_display (display);

    appswitcher_next (self);

    return 0;
}

static void
appswitcher3d_hide_previews (AppSwitcher3D *switcher, guint8 end_opacity)
{
    AppSwitcher *parent = (AppSwitcher *)switcher;
    const MetaRectangle *monitor = parent->active_monitor;

    // preview windows

    GList *p = NULL, *w = NULL; // previews, windows
    for (p = parent->previews, w = parent->windows; p; p = p->next, w = w->next)
    {
        MetaWindow *meta_win = w->data;
        ClutterActor *actor =  (ClutterActor *)meta_window_get_compositor_private (meta_win);
        /* TODO */
        if (w != parent->current_index)
        {
            clutter_actor_set_child_below_sibling (switcher->preview_actor, p->data, NULL);
        }

        gboolean minimized;
        g_object_get (meta_win, "minimized", &minimized, NULL);
        ClutterTransition *transition;
        transition = cdos_transition_simple (p->data, TRANSITION_TYPE, ANIMATION_TIME,
                "opacity", !minimized ? end_opacity : 0,
                "x", minimized ? 0.f -monitor->x : clutter_actor_get_x (actor) - monitor->x,
                "y", minimized ? 0.f -monitor->y : clutter_actor_get_y (actor) - monitor->y,
                "width", minimized ? 0.f: clutter_actor_get_width (actor),
                "height", minimized ? 0.f: clutter_actor_get_height (actor),
                NULL
                );
        g_signal_connect_swapped (transition, "completed", G_CALLBACK (clutter_actor_destroy), p->data);
    }
    g_list_free (parent->previews);
    parent->previews = NULL;
}

static void
appswitcher3d_hide (AppSwitcher *self)
{
    /* 不使用 initail_timeout_id, 不需要该函数 */
    AppSwitcher3D *switcher = (AppSwitcher3D *)self;
    appswitcher3d_hide_previews (switcher, 255);
    appswitcher3d_disable_monitor_fix (switcher);

    MetaDisplay *display = cdos_global_get_display (global);
    meta_enable_unredirect_for_display (display);
}

static gboolean
appswitcher3d_check_switch_time (AppSwitcher *self)
{
    AppSwitcher3D *switcher = (AppSwitcher3D *)self;
    long t;
    struct timeval tm;
    gettimeofday (&tm, 0);
    t = tm.tv_sec * 1000 + tm.tv_usec / 1000;
    if (t - switcher->last_time < SWITCH_TIME_DELAY)
        return 0;
    switcher->last_time = t;
    return 1;
}

static void
appswitcher3d_on_workspace_selected (AppSwitcher *self,
                                     MetaWorkspace *workspace)
{
    AppSwitcher3D *switcher = (AppSwitcher3D *)self;
    appswitcher3d_hide_previews (switcher, 0);

    g_list_free (self->windows);
    self->windows = cdos_get_windows_for_binding (self->binding);
    if (!self->windows)
    {
        appswitcher_destroy (NULL, NULL, self);
        return;
    }

    MetaDisplay *display = cdos_global_get_display (global);
    MetaWindow *w = meta_display_get_focus_window (display);
    self->current_index = g_list_find (self->windows, w);
    if (!self->current_index)
    {
        self->current_index = self->windows;
    }

    appswitcher3d_create_list (self);
    appswitcher_next (self);
}

static void
appswitcher3d_on_destroy (AppSwitcher *self)
{
    AppSwitcher3D *switcher = (AppSwitcher3D *)self;
    appswitcher3d_disable_monitor_fix (switcher);
    if (switcher->application_icon_box)
    {
        clutter_actor_destroy (switcher->application_icon_box);
        switcher->application_icon_box = NULL;
    }

    if (switcher->window_title)
    {
        clutter_actor_destroy (switcher->window_title);
        switcher->window_title = NULL;
    }
    switcher->icon = NULL;

    g_hash_table_foreach (switcher->preview_table, (GHFunc)remove_preview_data, NULL);
    g_hash_table_destroy (switcher->preview_table);
    switcher->preview_table = NULL;

    if (self->previews)
    {
        g_list_foreach (self->previews, (GFunc)clutter_actor_destroy, NULL);
        g_list_free (self->previews);

        self->previews = NULL;
    }
    g_list_free (self->windows);
    self->windows = NULL;
}

static void
appswitcher3d_create_list (AppSwitcher *self)
{
    AppSwitcher3D *switcher = (AppSwitcher3D *)self;

    g_hash_table_foreach (switcher->preview_table, (GHFunc)remove_preview_data, NULL);
    g_hash_table_remove_all (switcher->preview_table);

    if (self->previews)
    {
        g_list_foreach (self->previews, (GFunc)clutter_actor_destroy, NULL);
        g_list_free (self->previews);
        self->previews = NULL;
    }

    const MetaRectangle *monitor = self->active_monitor;

    MetaDisplay *display = cdos_global_get_display (global);
    MetaWorkspaceManager *wsm = meta_display_get_workspace_manager (display);
    MetaWorkspace *current_workspace = meta_workspace_manager_get_active_workspace (wsm);

    GList *iter = NULL;
    for (iter = self->windows; iter; iter = iter->next)
    {
        MetaWindow *win = iter->data;
        // printf("%s: 0x%lx\n", __func__, meta_window_get_xwindow(win));
        MetaWindowActor *compositor = META_WINDOW_ACTOR (meta_window_get_compositor_private (win));

        if (compositor)
        {
            ClutterActor *texture = (ClutterActor *)compositor;
            gfloat width, height;
            clutter_actor_get_size (texture, &width, &height);

            gdouble scale = 1.0;
            gfloat preview_width = monitor->width * PREVIEW_SCALE;
            gfloat preview_height = monitor->height * PREVIEW_SCALE;
            if (width > preview_width || height > preview_height)
                scale = min (preview_width / width, preview_height / height);

            ClutterActor *preview = (ClutterActor *)st_button_new ();
            {
                gint minimized;
                g_object_get (win, "minimized", &minimized, NULL);
                MetaWorkspace *wks = meta_window_get_workspace (win);
                guint8 opacity = 0;
                if ((!minimized && wks == current_workspace) || meta_window_is_on_all_workspaces (win))
                    opacity = 255;
                clutter_actor_set_opacity (preview, opacity);
                clutter_actor_set_reactive (preview, 1);
                gfloat x = 0, y = 0;
                clutter_actor_set_pivot_point (preview, 0.5, 0.5);
                if (!minimized)
                {
                    /*
                    ClutterPoint pos = {.x = 0, .y = 0};
                    ClutterSize size = {.width = 0, .height = 0};
                    clutter_actor_get_position (CLUTTER_ACTOR (compositor), &pos.x, &pos.y);
                    clutter_actor_get_size ((ClutterActor *)compositor, &size.width, &size.height);

                    x = pos.x + size.width / 2 - monitor->x;
                    y = pos.y + size.height / 2 - monitor->y;
                    */
                    clutter_actor_get_position (CLUTTER_ACTOR (compositor), &x, &y);
                }
                clutter_actor_set_position (preview, x, y);
            }
            PreviewData *data = g_new0 (PreviewData, 1);
            data->target_width = round (width * scale);
            data->target_height = round (height * scale);
            data->target_width_side = data->target_width * 2 / 3;
            data->target_height_side = data->target_height;
            data->meta_window = win;
            g_hash_table_insert (switcher->preview_table, preview, data);

            ClutterActor *clone = clutter_clone_new (texture);
            st_bin_set_child ((StBin *)preview, clone);
            g_signal_connect (preview, "clicked", G_CALLBACK (clone_clicked), switcher);
            self->previews = g_list_append (self->previews, preview);
            clutter_actor_add_child (switcher->preview_actor, preview);
        }
    }
    // adapt_clones (); /* 空函数 */
}

static void
clone_clicked (StButton *button, gint clicked_button, AppSwitcher3D *switcher)
{
    AppSwitcher *parent = (AppSwitcher *)switcher;
    gint current_index = g_list_index (parent->previews, button);
    parent->current_index = g_list_nth (parent->windows, current_index);
    appswitcher_activate_selected (NULL, NULL, parent);
}

static void
appswitcher3d_update_list (AppSwitcher *self, gint index)
{
    g_warning ("virtual function.");
}

static void
appswitcher3d_select_next (AppSwitcher *self)
{
    g_warning ("virtual function.");
}

static void
appswitcher3d_select_previous (AppSwitcher *self)
{
    g_warning ("virtual function.");
}

static void
appswitcher3d_set_current_window (AppSwitcher *self, MetaWindow *window)
{
    const MetaRectangle *monitor = self->active_monitor;
    AppSwitcher3D *switcher = (AppSwitcher3D *)self;

    // 删除当前 window title label
    if (switcher->window_title != NULL)
    {
        clutter_actor_remove_child (self->actor, switcher->window_title);
    }

    MetaWindow *current_window = NULL;
    current_window = g_list_find (self->windows, self->current_index->data)->data;

    const char *title = meta_window_get_title (META_WINDOW (current_window));
    StWidget *window_title = st_label_new (title);
    switcher->window_title = (ClutterActor *)window_title;

    st_widget_set_style_class_name (window_title, "app-switcher-title");
    clutter_actor_set_opacity ((ClutterActor *)window_title, 0);

    // ellipsize if title is too long
    char style[128] = {0};
    snprintf (style, sizeof(style) - 1,
            "max-width:%dpx;font-size:14px;font-wight:bold;padding:14px;",
            (int)(monitor->width - 200));

    st_widget_set_style (window_title, style);
    ClutterActor *clutter_text = st_label_get_clutter_text ((StLabel *)window_title);
    clutter_text_set_ellipsize ((ClutterText *)clutter_text, PANGO_ELLIPSIZE_END);

    clutter_actor_add_child (self->actor, switcher->window_title);
    {
        clutter_actor_set_opacity (switcher->window_title, 0);
        clutter_actor_save_easing_state (switcher->window_title);
        clutter_actor_set_easing_mode (switcher->window_title, TRANSITION_TYPE);
        clutter_actor_set_easing_duration (switcher->window_title, ANIMATION_TIME);
        clutter_actor_set_opacity (switcher->window_title, 255);
        clutter_actor_restore_easing_state (switcher->window_title);
    }

    guint ui_scale = cdos_global_get_ui_scale (global);
    float cx = monitor->x + (monitor->width + (ICON_SIZE * ui_scale) +
                (ICON_TITLE_SPACING * ui_scale)) / 2;
    float cy = round(monitor->height * TITLE_POSITION);
    graphene_size_t title_size;
    clutter_actor_get_size (switcher->window_title, &title_size.width, &title_size.height);
    clutter_actor_set_position (switcher->window_title,
            cx - round(title_size.width / 2), cy - round(title_size.height / 2));

    // window icon
    if (switcher->application_icon_box != NULL)
    {
        clutter_actor_remove_child (self->actor, switcher->application_icon_box);
    }

    CdosApp *app = cdos_window_tracker_get_window_app (self->tracker, current_window);
    if (app != NULL)
        switcher->icon = cdos_app_create_icon_texture (app, ICON_SIZE);
    if (switcher->icon == NULL)
    {
        StIcon *icon = (StIcon *)st_icon_new ();
        switcher->icon = (ClutterActor *)icon;
        st_icon_set_icon_name (icon, "applications-other");
        st_icon_set_icon_size (icon, ICON_SIZE);
    }

    ClutterActor *application_icon_box = (ClutterActor *)st_bin_new ();
    switcher->application_icon_box = application_icon_box;
    st_widget_set_style_class_name ((StWidget *)application_icon_box, "app-switcher-iconbox");
    clutter_actor_set_opacity (application_icon_box, 0);
    {
        graphene_point3d_t point;
        clutter_actor_get_position (switcher->window_title, &point.x, &point.y);
        clutter_actor_set_position (application_icon_box,
                                    round (point.x - ICON_SIZE * ui_scale - ICON_TITLE_SPACING * ui_scale),
                                    round (cy - ICON_SIZE * ui_scale / 2.0));
    }

    st_bin_set_child ((StBin *)application_icon_box, switcher->icon);
    clutter_actor_add_child (self->actor, application_icon_box);

    {
        clutter_actor_set_opacity (switcher->application_icon_box, 0);
        clutter_actor_save_easing_state (switcher->application_icon_box);
        clutter_actor_set_easing_mode (switcher->application_icon_box, TRANSITION_TYPE);
        clutter_actor_set_easing_duration (switcher->application_icon_box, ANIMATION_TIME);
        clutter_actor_set_opacity (switcher->application_icon_box, 255);
        clutter_actor_restore_easing_state (switcher->application_icon_box);
    }
}

static void
appswitcher3d_destroy_actors (AppSwitcher *self)
{
    AppSwitcher3D *switcher = (AppSwitcher3D *)self;
    cdos_background_manager_destroy (switcher->background);
    clutter_actor_destroy (switcher->background_group);

    clutter_actor_destroy (self->actor);

    clutter_actor_show (cdos_global_get_window_group (global));
}

static void
appswitcher3d_enable_monitor_fix (AppSwitcher3D *switcher)
{
    return;

    switcher->monitor_fix = 1;
    ClutterStage *stage = cdos_global_get_stage (global);
    graphene_size_t size;
    clutter_actor_get_size (CLUTTER_ACTOR (stage), &size.width, &size.height);
    switcher->old_width = size.width;
    switcher->old_height = size.height;
    const MetaRectangle *monitor = ((AppSwitcher *)switcher)->active_monitor;
    gfloat width = 2 * (monitor->x + monitor->width / 2);
    gfloat height = 2 * (monitor->y + monitor->height / 2);

    clutter_actor_set_size ((ClutterActor *)stage, width, height);
}

static void
appswitcher3d_disable_monitor_fix (AppSwitcher3D *switcher)
{
    if (switcher->monitor_fix)
    {
        ClutterStage *stage = cdos_global_get_stage (global);
        clutter_actor_set_size ((ClutterActor *)stage, switcher->old_width, switcher->old_height);
        switcher->monitor_fix = 0;
    }
}
