#include "common.h"
#include "cdos-layout-manager.h"
#include "cdos-appswitcher.h"
#include "cdos-appswitcher-private.h"

#define INITIAL_DELAY_TIMEOUT 150
#define CHECK_DESTROYED_TIMEOUT 150
#define DISABLE_HOVER_TIMEOUT 150

// clang-format off
typedef enum {
    ICON_22 = 22,
    ICON_32 = 32,
    ICON_48 = 48,
    ICON_64 = 64,
    ICON_96 = 96
} IconSize;
// clang-format on

static void update_active_monitor (AppSwitcher *switcher);

static void disable_hover (AppSwitcher *switcher);
static gboolean key_press_event (ClutterActor *actor, ClutterEvent *event, AppSwitcher *switcher);
static gboolean key_release_event (ClutterActor *actor, ClutterEvent *event, AppSwitcher *switcher);
static gboolean scroll_event (ClutterActor *actor, ClutterEvent *event, AppSwitcher *switcher);

static int
sort_window_by_user_time (MetaWindow *a, MetaWindow *b)
{
    MetaWindow *win1 = META_WINDOW (a);
    MetaWindow *win2 = META_WINDOW (b);
    guint32 t1 = meta_window_get_user_time (win1);
    guint32 t2 = meta_window_get_user_time (win2);

    guint m1 = 0, m2 = 0;
    GValue m = {0,};
    g_value_init (&m, G_TYPE_BOOLEAN);
    g_object_get_property ((GObject *)win1, "minimized", &m);
    m1 = g_value_get_boolean (&m);

    g_value_reset (&m);
    g_object_get_property ((GObject *)win2, "minimized", &m);
    m2 = g_value_get_boolean (&m);

    // g_object_get (win2, "minimized", &m2, NULL);
    if (m1 == m2)
        return (t2 == t1) ? 0 : ((t2 > t1) ? 1 : -1);
    else
        return m1 ? 1 : -1;
}

static gboolean
match_skip_taskbar (MetaWindow *win)
{
    return !meta_window_is_skip_taskbar (win);
}

static gboolean
match_wm_class (MetaWindow *win, const char *klass)
{
    return (strcmp (meta_window_get_wm_class (win), klass) == 0) && (!meta_window_is_skip_taskbar (win));
}

static gboolean
match_workspace (MetaWindow *win, const void *workspace)
{
    return (meta_window_get_workspace (win) == workspace) && (!meta_window_is_skip_taskbar (win));
}

static guint
primary_modifier (guint mask)
{
    if (mask == 0)
        return 0;
    guint primary = 1;
    while (mask > 1)
    {
        mask >>= 1;
        primary <<= 1;
    }
    return primary;
}

/**
 * g_list_filter:
 * @list: GList
 * @filter_func: 过滤函数
 *
 * 对 %list 中的每个元素a调用 filter_func(a),对 %list 中的元素进行筛选。
 *
 * Returns: GList.
 */
GList *
g_list_filter (const GList *list, FilterFunc filter_func)
{
    const GList *iter = NULL;
    GList *ret = NULL;
    for (iter = list; iter; iter = iter->next)
    {
        if (!filter_func (iter->data))
        {
            // list = g_list_remove (list, iter->data);
            ret = g_list_prepend (ret, iter->data);
        }
    }
    ret = g_list_reverse (ret);
    return ret;
}


/**
 * g_list_filter_p:
 * @list: GList
 * @filter_func: 过滤函数
 * @data: user data
 *
 * 对 %list 中的每个元素a调用 filter_func(a, data),对 %list 中的元素进行筛选。
 *
 * Returns: GList.
 */
GList *
g_list_filter_p (const GList *list, FilterFunc_p filter_func, const void *data)
{
    const GList *iter = NULL;
    GList *ret = NULL;
    for (iter = list; iter; iter = iter->next)
    {
        if (filter_func (iter->data, data))
        {
            // list = g_list_remove (list, iter->data);
            ret = g_list_prepend (ret, iter->data);
        }
    }
    ret = g_list_reverse (ret);
    return ret;
}

/**
 * cdos_get_windows_for_binding:
 * @binding: MetaKeyBinding
 *
 * 根据 %binding 的值返回一个 元素类型为 MetaWindow 的GList。
 *
 * Returns: GList(元素类型为 MetaWindow)
 */
GList *
cdos_get_windows_for_binding (MetaKeyBinding *binding)
{
    GList *all_windows = NULL;
    GList *windows = NULL;
    GList *window_actors = cdos_global_get_window_actors (global);
    MetaWindow *window = NULL;
    GList *iter = NULL;
    for (iter = window_actors; iter; iter = iter->next)
    {
        window = meta_window_actor_get_meta_window (iter->data);
        if (cdos_is_interesting (window))
            all_windows = g_list_prepend (all_windows, window);
    }
    const char *name = meta_key_binding_get_name (binding);
    if (strcmp (name, "switch-panels") == 0)
    {
        windows = g_list_filter (all_windows, (FilterFunc)match_skip_taskbar);
    }
    else if (strcmp (name, "switch-group") == 0)
    {
        const char *klass = NULL;
        MetaWindow *focused = NULL;
        MetaDisplay *display = cdos_global_get_display (cdos_global_get ());
        focused = meta_display_get_focus_window (display);
        if (focused == NULL)
            focused = windows->data;
        klass = meta_window_get_wm_class (focused);

        windows = g_list_filter_p (all_windows, (FilterFunc_p)match_wm_class, klass);
    }
    else
    {
        MetaDisplay *display = cdos_global_get_display (global);
        MetaWorkspaceManager *wsm = meta_display_get_workspace_manager (display);
        MetaWorkspace *wks = meta_workspace_manager_get_active_workspace (wsm);
        windows = g_list_filter_p (all_windows, (FilterFunc_p)match_workspace, wks);
    }
    g_list_free (all_windows);
    windows = g_list_sort (windows, (GCompareFunc)sort_window_by_user_time);
    return windows;
}

void
appswitcher_init (AppSwitcher *switcher, MetaKeyBinding *binding)
{
    // AppSwitcher *switcher = g_new0 (AppSwitcher, 1);
    g_return_if_fail (switcher != NULL);

    switcher->binding = binding;
    switcher->windows = cdos_get_windows_for_binding (binding);

    MetaWindow *focused = NULL;
    MetaDisplay *display = cdos_global_get_display (global);
    focused = meta_display_get_focus_window (display);
    switcher->current_index = g_list_find (switcher->windows, focused);
    if (switcher->current_index == 0)
        switcher->current_index = switcher->windows;
    switcher->modifier_mask = primary_modifier (meta_key_binding_get_mask (binding));

    switcher->tracker = cdos_window_tracker_get_default ();
    switcher->window_manager = cdos_global_get_window_manager (global);

    switcher->dcid = g_signal_connect (switcher->window_manager, "destroy", G_CALLBACK (window_destroyed), switcher);
    /* 在 key_press_event() 中已经激活 window, 不需要连接 map 信号 */
    // switcher->mcid = g_signal_connect (switcher->window_manager, "map",
    //        G_CALLBACK (appswitcher_activate_selected), switcher);

    switcher->enforce_primary_monitor =
        g_settings_get_boolean (cdos_global_get_settings (global), "alttab-switcher-enforce-primary-monitor");
    update_active_monitor (switcher);
}

/* 虚函数 */
static gboolean appswitcher_show (AppSwitcher *switcher);
// static void appswitcher_hide (AppSwitcher *switcher);

static void appswitcher_on_destroy (AppSwitcher *switcher);

// static void appswitcher_create_list (AppSwitcher *switcher);
static void appswitcher_update_list (AppSwitcher *switcher, gint index);

static void appswitcher_select_next (AppSwitcher *switcher);
static void appswitcher_select_previous (AppSwitcher *switcher);

static void appswitcher_on_workspace_selected (AppSwitcher *switcher, MetaWorkspace *workspace);

static gboolean appswitcher_check_switch_time (AppSwitcher *switcher);

static void appswitcher_set_current_window (AppSwitcher *switcher, MetaWindow *window);
/* ************** Virtual Function end ************** */

// static void appswitcher_select (AppSwitcher *switcher, void *index);

static gboolean appswitcher_switch_workspace (AppSwitcher *switcher, int direction);

static void appswitcher_show_desktop (AppSwitcher *switcher);

typedef struct _CheckDestroyedData
{
    AppSwitcher *switcher;
    MetaWindow *window;
} CheckDestroyData;

static gboolean check_destroyed (CheckDestroyData *data);

static void remove_destroyed_window (AppSwitcher *switcher, MetaWindow *window);

gboolean
appswitcher_setup_modal (AppSwitcher *switcher)
{
    switcher->have_modal = cdos_push_modal (switcher->actor, 0, 0, 0);
    if (switcher->have_modal == 0)
        switcher->have_modal = cdos_push_modal (switcher->actor, 0, META_MODAL_POINTER_ALREADY_GRABBED, 0);
    if (switcher->have_modal == 0)
    {
        appswitcher_activate_selected (NULL, NULL, switcher);
    }
    else
    {
        // Initially disable hover so we ignore the enter-event if
        // the switcher appears underneath the current pointer location
        // disable_hover (switcher); // 不懂
        g_signal_connect (switcher->actor, "key-press-event", G_CALLBACK (key_press_event), switcher);
        g_signal_connect (switcher->actor, "key-release-event", G_CALLBACK (key_release_event), switcher);
        g_signal_connect (switcher->actor, "scroll-event", G_CALLBACK (scroll_event), switcher);
        g_signal_connect (switcher->actor, "button-press-event", G_CALLBACK (appswitcher_destroy), switcher);

        // There's a race condition; if the user released Alt before
        // we got the grab, then we won't be notified. So we check
        // now. (Have to do this after updating selection.)
        int x, y;
        ClutterModifierType mod;
        cdos_global_get_pointer (global, &x, &y, &mod);
        if (!(mod & switcher->modifier_mask))
        {
            appswitcher_activate_selected (NULL, NULL, switcher);
            return false;
        }

        // We delay showing the popup so that fast Alt+Tab users aren't
        // disturbed by the popup briefly flashing.
        // switcher->initial_delay_timeout_id = g_timeout_add (INITIAL_DELAY_TIMEOUT, (GSourceFunc)appswitcher_show,
        // switcher);

        appswitcher_show (switcher);
    }
    return switcher->have_modal;
}

static void
pop_modal (AppSwitcher *switcher)
{
    if (switcher->have_modal)
    {
        cdos_pop_modal (switcher->actor, 0);
        switcher->have_modal = false;
    }
}

static gboolean
appswitcher_show (AppSwitcher *switcher)
{
    if (switcher->show)
        return switcher->show (switcher);
    else
        return 0;
}

/*
static void
appswitcher_hide (AppSwitcher *switcher)
{
    if(switcher->hide)
        switcher->hide (switcher);
}
*/

static void
appswitcher_on_destroy (AppSwitcher *switcher)
{
    if (switcher->on_destroy)
        switcher->on_destroy (switcher);
}

/*
static void
appswitcher_create_list (AppSwitcher *switcher)
{
    if (switcher->create_list)
        switcher->create_list (switcher);
}
*/

static void
appswitcher_update_list (AppSwitcher *switcher, gint index)
{
    if (switcher->update_list)
        switcher->update_list (switcher, index);
}

static void
appswitcher_select_next (AppSwitcher *switcher)
{
    if (switcher->select_next)
        switcher->select_next (switcher);
}

static void
appswitcher_select_previous (AppSwitcher *switcher)
{
    if (switcher->select_previous)
        switcher->select_previous (switcher);
}

static void
appswitcher_on_workspace_selected (AppSwitcher *switcher, MetaWorkspace *workspace)
{
    if (switcher->on_workspace_selected)
        switcher->on_workspace_selected (switcher, workspace);
}

static gboolean
appswitcher_check_switch_time (AppSwitcher *switcher)
{
    if (switcher->check_switch_time)
        return switcher->check_switch_time (switcher);
    return true;
}

static void
appswitcher_set_current_window (AppSwitcher *switcher, MetaWindow *window)
{
    if (switcher->set_current_window)
        switcher->set_current_window (switcher, window);
}

void
appswitcher_next (AppSwitcher *switcher)
{
    if (switcher->windows == NULL || switcher->windows->next == NULL)
    {
        switcher->current_index = switcher->windows;
        appswitcher_update_list (switcher, 0);
    }
    else
    {
        clutter_actor_set_reactive (switcher->actor, 0);
        appswitcher_select_next (switcher);
        clutter_actor_set_reactive (switcher->actor, 1);
    }
    appswitcher_set_current_window (switcher, NULL);
}

void
appswitcher_previous (AppSwitcher *switcher)
{
    if (switcher->windows == NULL || switcher->windows->next == NULL)
    {
        switcher->current_index = switcher->windows;
        appswitcher_update_list (switcher, 0);
    }
    else
    {
        clutter_actor_set_reactive (switcher->actor, 0);
        appswitcher_select_previous (switcher);
        clutter_actor_set_reactive (switcher->actor, 1);
    }
    appswitcher_set_current_window (switcher, NULL);
}

/*
static void
appswitcher_select (AppSwitcher *switcher, void *index)
{
    switcher->current_index = index;
    appswitcher_set_current_window (switcher, NULL);
}
*/

static void
update_active_monitor (AppSwitcher *switcher)
{
    // g_free (switcher->active_monitor);
    switcher->active_monitor = NULL;
    CdosLayoutManager *layout = cdos_layout_manager_get_default ();
    if (!switcher->enforce_primary_monitor)
    {
        int x, y;
        ClutterModifierType mods;
        cdos_global_get_pointer (global, &x, &y, &mods);
        switcher->active_monitor = cdos_layout_manager_find_monitor_for_rect (layout, x, y, 0, 0);
    }
    if (switcher->active_monitor == NULL)
    {
        // switcher->active_monitor = g_new0 (MetaRectangle, 1);
        switcher->active_monitor = cdos_layout_manager_get_primary_monitor (layout);
        /*
        MetaRectangle rect;
        meta_screen_get_monitor_geometry (cdos_global_get_screen (global), 0, &rect);
        switcher->active_monitor = g_new0 (MetaRectangle, 1);
        *switcher->active_monitor = rect;
        */
    }
}

static gboolean
key_press_event (ClutterActor *actor, ClutterEvent *event, AppSwitcher *switcher)
{
    ClutterModifierType event_state = clutter_event_get_state (event) & CLUTTER_MODIFIER_MASK;
    disable_hover (switcher);

    gint sym = clutter_event_get_key_symbol (event);
    // Switch workspace
    if (event_state & CLUTTER_CONTROL_MASK)
    {
        switch (sym)
        {
            case CLUTTER_KEY_Right:
                if (appswitcher_switch_workspace (switcher, 1))
                    return true;
                break;
            case CLUTTER_KEY_Left:
                if (appswitcher_switch_workspace (switcher, -1))
                    return true;
                break;
            default:
                break;
        }
    }

    // Extra keys
    switch (sym)
    {
        case CLUTTER_KEY_Escape:
            appswitcher_destroy (NULL, event, switcher);
            return true;
        case CLUTTER_KEY_Return:
            // Enter -> Select active window
            appswitcher_activate_selected (NULL, NULL, switcher);
            return true;
        case CLUTTER_KEY_d:
        case CLUTTER_KEY_D:
            appswitcher_show_desktop (switcher);
            break;
        case CLUTTER_KEY_q:
        case CLUTTER_KEY_Q:
            // Q -> Close window
            {
                MetaWindow *window = switcher->current_index->data;
                meta_window_delete (META_WINDOW (window), cdos_global_get_current_time (global));
                CheckDestroyData *data = g_new0 (CheckDestroyData, 1);
                data->window = window;
                data->switcher = switcher;
                switcher->check_destroyed_timeout_id =
                    g_timeout_add (CHECK_DESTROYED_TIMEOUT, (GSourceFunc)check_destroyed, data);
                return true;
            }
            break;
        case CLUTTER_KEY_Right:
        case CLUTTER_KEY_Down:
            if (appswitcher_check_switch_time (switcher))
                appswitcher_previous (switcher);
            return true;
        default:
            break;
    }
    // Default alt-tab
    MetaDisplay *dpy = cdos_global_get_display (cdos_global_get ());
    MetaKeyBindingAction action =
        meta_display_get_keybinding_action (dpy, clutter_event_get_key_code (event), event_state);
    switch (action)
    {
        case META_KEYBINDING_ACTION_SWITCH_GROUP:
        case META_KEYBINDING_ACTION_SWITCH_WINDOWS:
        case META_KEYBINDING_ACTION_SWITCH_PANELS:
        case META_KEYBINDING_ACTION_SWITCH_APPLICATIONS:
            if (appswitcher_check_switch_time (switcher))
            {
                if (event_state & CLUTTER_SHIFT_MASK)
                    appswitcher_previous (switcher);
                else
                {
                    appswitcher_next (switcher);
                }
            }
            return true;
        case META_KEYBINDING_ACTION_SWITCH_GROUP_BACKWARD:
        case META_KEYBINDING_ACTION_SWITCH_WINDOWS_BACKWARD:
        case META_KEYBINDING_ACTION_SWITCH_PANELS_BACKWARD:
            if (appswitcher_check_switch_time (switcher))
                appswitcher_previous (switcher);
            return true;
        default:
            break;
    }
    return true;
}

static gboolean
key_release_event (ClutterActor *actor, ClutterEvent *event, AppSwitcher *switcher)
{
    int x, y;
    ClutterModifierType mods;
    cdos_global_get_pointer (global, &x, &y, &mods);
    int state = mods & switcher->modifier_mask;

    if (state == 0)
    {
        if (switcher->initial_delay_timeout_id != 0)
        {
            // GList *w = g_list_find (switcher->windows, switcher->current_index->data);
            // switcher->current_index =
            //        w->next ? w->next : switcher->windows;
        }
        /* 防止 map 最小化窗口时崩溃 ?*/
        /*
        MetaWindow *w = META_WINDOW (switcher->current_index->data);
        gboolean m = 0;
        g_object_get (w, "minimized", &m, NULL);
        if (meta_window_is_mapped (w) == 0 || m == 1)
        {
            g_signal_handler_disconnect (switcher->window_manager, switcher->mcid);
            switcher->mcid = 0;
        }
        */
        appswitcher_activate_selected (NULL, NULL, switcher);
    }
    return true;
}

static gboolean
scroll_event (ClutterActor *actor, ClutterEvent *event, AppSwitcher *switcher)
{
    if (appswitcher_check_switch_time (switcher))
    {
        clutter_actor_set_reactive (actor, 0);
        if (clutter_event_get_scroll_direction (event) == CLUTTER_SCROLL_UP)
            appswitcher_previous (switcher);
        else
            appswitcher_next (switcher);
        clutter_actor_set_reactive (actor, 1);
    }
    return true;
}

static void
disable_hover (AppSwitcher *switcher)
{
    switcher->mouse_active = 0;

    if (switcher->motion_timeout_id)
        g_source_remove (switcher->motion_timeout_id);

    // switcher->motion_timeout_id = g_timeout_add (DISABLE_HOVER_TIMEOUT, (GSourceFunc)mouse_timeout, switcher);
}

static gboolean
appswitcher_switch_workspace (AppSwitcher *switcher, int direction)
{
    MetaDisplay *display = cdos_global_get_display (global);
    MetaWorkspaceManager *wsm = meta_display_get_workspace_manager (display);

    int n_wks = meta_workspace_manager_get_n_workspaces (wsm);
    if (n_wks < 2)
        return false;

    int current = meta_workspace_manager_get_active_workspace_index (wsm);
    int next_index = (n_wks + current + direction) % n_wks;
    MetaWorkspace *workspace = meta_workspace_manager_get_workspace_by_index (wsm, next_index);
    meta_workspace_activate (workspace, cdos_global_get_current_time (global));
    if (current == meta_workspace_manager_get_active_workspace_index (wsm))
        return false;

    // TODO
    // cdos_base_wm_show_workspace_OSD ();
    appswitcher_on_workspace_selected (switcher, workspace);
    return true;
}

void
window_destroyed (CdosBaseWM *wm, MetaWindowActor *actor, AppSwitcher *switcher)
{
    remove_destroyed_window (switcher, meta_window_actor_get_meta_window (actor));
}

static gboolean
check_destroyed (CheckDestroyData *data)
{
    AppSwitcher *switcher = data->switcher;
    MetaWindow *window = data->window;
    switcher->check_destroyed_timeout_id = 0L;
    remove_destroyed_window (switcher, window);
    g_free (data);
    return false;
}

static void
remove_destroyed_window (AppSwitcher *switcher, MetaWindow *window)
{
    GList *iter = NULL, *preview = NULL;
    for (iter = switcher->windows, preview = switcher->previews; iter; iter = iter->next)
    {
        if (window == iter->data)
        {
            if (switcher->windows->next == NULL)
                appswitcher_destroy (NULL, NULL, switcher);
            else
            {
                switcher->windows = g_list_remove_link (switcher->windows, iter);
                clutter_actor_destroy (CLUTTER_ACTOR (preview->data));
                switcher->previews = g_list_remove_link (switcher->previews, preview);
                clutter_actor_destroy (preview->data);
                g_slice_free (GList, preview);

                if (iter == switcher->current_index)
                    switcher->current_index =
                        (iter->prev == NULL) ? (iter->next ? iter->next : switcher->windows) : iter->prev;
                g_slice_free (GList, iter);

                appswitcher_update_list (switcher, 0);
                appswitcher_set_current_window (switcher, switcher->current_index->data);
            }
            return;
        }
        preview = preview->next;
    }
}

void
appswitcher_activate_selected (CdosBaseWM *wm, MetaWindowActor *actor, AppSwitcher *switcher)
{
    cdos_activate_window (switcher->current_index->data, cdos_global_get_current_time (global), -1);
    appswitcher_destroy (NULL, NULL, switcher);
}

static void
appswitcher_show_desktop (AppSwitcher *switcher)
{
    GList *iter = NULL;

    for (iter = switcher->windows; iter; iter = iter->next)
    {
        guint min = 0;
        g_object_get ((GObject *)iter->data, "minimized", &min, NULL);
        if (!min)
            meta_window_minimize (iter->data);
    }
    appswitcher_destroy (NULL, NULL, switcher);
}

static void
appswitcher_destroy_actors (AppSwitcher *switcher)
{
    if (switcher->destroy_actors)
        switcher->destroy_actors (switcher);
}

gboolean
appswitcher_destroy (ClutterActor *actor, ClutterEvent *event, AppSwitcher *switcher)
{
    pop_modal (switcher);
    /* FIXME initial_delay_timeout_id 不懂 */
    // if (switcher->initial_delay_timeout_id != 0)

    appswitcher_on_destroy (switcher);

    appswitcher_destroy_actors (switcher);
    // else
    //    appswitcher_hide (switcher);

    if (switcher->initial_delay_timeout_id > 0)
    {
        g_source_remove (switcher->initial_delay_timeout_id);
        switcher->initial_delay_timeout_id = 0L;
    }

    if (switcher->check_destroyed_timeout_id != 0)
    {
        g_source_remove (switcher->check_destroyed_timeout_id);
        switcher->check_destroyed_timeout_id = 0L;
    }
    if (switcher->dcid)
        g_signal_handler_disconnect (switcher->window_manager, switcher->dcid);
    if (switcher->mcid)
        g_signal_handler_disconnect (switcher->window_manager, switcher->mcid);

    g_free (switcher);
    return 0;
}
