#include "cdos-background.h"
#include "cdos-global.h"
#include "cdos-layout-manager.h"
#include <gdk/gdk.h>
#include <json-glib/json-glib.h>
#include <meta/display.h>
#include <meta/meta-monitor-manager.h>

/**
 * SECTION:cdos-background
 * @title: CdosBackground
 * @short_description: 加载桌面的背景图片
 *
 * CdosBackground的功能为当启动桌面时,通过CdosBackground类来加载桌面背景图片
 */

#define BACKGROUND_SCHEMA "org.cdos.desktop.background"

#define DEFAULT_WATERMARK_DIR "/usr/share/backgrounds/"

#define NFS_LICENSE_NAME        "com.nfs.license"
#define NFS_LICENSE_PATH        "/com/nfs/license/Info"
#define NFS_LICENSE_INTERFACE   "com.nfs.license.Info"

struct _CdosBackgroundManager
{
    MetaDisplay *display;
    MetaMonitorManager *monitor_manager;
    GList *backgrounds;
    gulong monitor_changed_id;
    guint rebuild_backgrounds_id;
    ClutterActor *container;
    gboolean vignette;
    gboolean control_position;
    gchar *settings_schema;
    gboolean in_destroy;
    GDBusProxy *nemo_proxy;
    //是否显示注册状态？
    gboolean ignore_licenses;
};

static const gchar *current_background = NULL;

static gboolean cdos_background_update_animation (CdosBackground *self);
static void cdos_background_manager_on_monitors_changed (MetaMonitorManager *monitor_manager, CdosBackgroundManager *self);
static void update_watermark (CdosBackground *self, gchar *watermark_name, gint monitor_index);
static void cdos_background_check_registe_status_delay (CdosBackground *self);
static int get_state_from_data (gchar *data);

static FdBgSettings *default_settings = NULL;

static FdBgSettings *
get_default_settings (void)
{
    if (!default_settings)
        default_settings = fd_bg_settings_new (BACKGROUND_SCHEMA, "/etc/desktop/background");
    return default_settings;
}

static gboolean 
is_primary_monitor (CdosBackground *self)
{
    CdosGlobal *global = cdos_global_get ();
    MetaDisplay *display = cdos_global_get_display (global);
    gint primary_monitor = meta_display_get_primary_monitor (display);
    return primary_monitor == self->monitor_index ? TRUE : FALSE;
}

static const gchar *
get_register_watermark_name (gboolean is_registered)
{
    const gchar *lang = NULL;
    if (is_registered)
        return "bg-watermark.png";

    lang = g_getenv ("LANG");
    if (lang && g_str_has_prefix (lang, "en_"))
    {
        return "bg-watermark-unregistered-en.png";
    }
    return "bg-watermark-unregistered.png";
}

static const gchar *
get_bg_unregister_name (void)
{
    const gchar *lang = NULL;
    lang = g_getenv ("LANG");
    if (lang && g_str_has_prefix (lang, "en_"))
        return "bg-unregistered-en.png";
    return "bg-unregistered.png";
}

static void
free_resources (CdosBackground *self)
{
    if (self->meta_background)
        g_debug ("%s %p %d", __func__, self, ((GObject *)self->meta_background)->ref_count);
    if (self->image1)
        g_debug ("%s image1: %d", __func__, ((GObject *)self->image1)->ref_count);
    // g_clear_pointer (&self->file1, g_object_unref);
    g_clear_pointer (&self->file2, g_object_unref);
    if (self->image1)
    {
        g_signal_handlers_disconnect_by_data (self->image1, self);
        g_clear_pointer (&self->image1, g_object_unref);
    }
    if (self->image2)
    {
        g_signal_handlers_disconnect_by_data (self->image2, self);
        g_clear_pointer (&self->image2, g_object_unref);
    }

    if (self->animation_timeout_id)
    {
        g_source_remove (self->animation_timeout_id);
        self->animation_timeout_id = 0;
    }
    if (self->timeline)
    {
        if (clutter_timeline_is_playing (self->timeline))
            clutter_timeline_stop (self->timeline);
        g_object_unref (self->timeline);
        self->timeline = NULL;
    }
    if (self->dir)
    {
        g_dir_close (self->dir);
        self->dir = NULL;
    }
    /*
     * 在meta_background_actor_set_meta_background()中释放
     */
    // g_clear_object (&self->meta_background);
}

void
cdos_background_destroy (CdosBackground *self)
{
    g_debug ("%s %p", __func__, self);
    if (self->settings_changed_id)
    {
        g_signal_handler_disconnect (self->settings, self->settings_changed_id);
        self->settings_changed_id = 0;
    }
    if (self->check_registe_status_id > 0)
    {
        g_source_remove (self->check_registe_status_id);
        self->check_registe_status_id = 0;
    }
    if (self->watermark)
    {
        clutter_actor_destroy (self->watermark);
        //clutter_actor_remove_child (self->container, self->watermark);
        self->watermark = NULL;
    }
    if (self->register_proxy)
    {
        g_object_unref (self->register_proxy);
        self->register_proxy = NULL;
    }

    if (self->cancellable)
    {
        g_cancellable_cancel (self->cancellable);
        g_object_unref (self->cancellable);
        self->cancellable = NULL;
    }

    free_resources (self);
    g_clear_pointer (&self->file1, g_object_unref);
    g_clear_object (&self->meta_background);

    if (self->settings && !self->is_default)
    {
        g_object_unref (self->settings);
        self->settings = NULL;
    }
    g_free (self->register_name);
    clutter_actor_destroy (self->actor);
    self->actor = NULL;
    g_free (self);
}

static void
cdos_background_set_color_bg (CdosBackground *self)
{
    gchar *primary_color = fd_bg_settings_get_primary_color (self->settings);
    GDesktopBackgroundShading shading_type = fd_bg_settings_get_color_shading_type (self->settings);
    g_debug ("Background primary color:%s", primary_color);
    if (!primary_color)
        return;

    ClutterColor bg_color;
    clutter_color_from_string (&bg_color, primary_color);

    if (shading_type == G_DESKTOP_BACKGROUND_SHADING_SOLID)
        meta_background_set_color (self->meta_background, &bg_color);
    else
    {
        gchar *secondary_color = fd_bg_settings_get_secondary_color (self->settings);
        ClutterColor secondary_bg_color;
        if (secondary_color)
            clutter_color_from_string (&secondary_bg_color, secondary_color);
        g_debug ("Background secondary color:%s", secondary_color);
        meta_background_set_gradient (self->meta_background, shading_type, &bg_color, &secondary_bg_color);
        g_free (secondary_color);
    }
    g_free (primary_color);
}

static void
cdos_background_new_frame (ClutterTimeline *timeline,
                           gint msecs,
                           CdosBackground *self)
{
    gdouble progress = clutter_timeline_get_progress (self->timeline);

    meta_background_set_blend (self->meta_background,
                               self->file1,
                               self->file2,
                               self->file2 ? progress : 1 - progress, self->style);
}

static void
cdos_background_stop (ClutterTimeline *timeline, CdosBackground *self)
{
}

static void
cdos_background_load_new_bg_image (MetaBackgroundImage *image, CdosBackground *self)
{
    if (!self->file2)
        meta_background_set_blend (self->meta_background, self->file1, NULL, 0, self->style);
    else
        meta_background_set_blend (self->meta_background, self->file1, self->file2, 1.0, self->style);
    g_signal_handlers_disconnect_by_data (image, self);
}

/*
 * 每隔 self->duration 时间加载一张新的图片，并显示动画
 */
static void
cdos_background_image_loaded (MetaBackgroundImage *image, CdosBackground *self)
{
    if (!self->file2)
    {
        meta_background_set_blend (self->meta_background, self->file1, NULL, 0, self->style);
    }
    else
    {
        meta_background_set_blend (self->meta_background, self->file1, self->file2, 1.0, self->style);
        clutter_timeline_rewind (self->timeline);
        clutter_timeline_start (self->timeline);
    }
    if (!self->animation_timeout_id && self->use_animation)
    {
        self->animation_timeout_id =
            g_timeout_add_seconds (self->duration, (GSourceFunc)cdos_background_update_animation, self);
    }
    g_debug ("loaded %p ref_count: %d", image, ((GObject *)image)->ref_count);
    g_signal_handlers_disconnect_by_data (image, self);
    //g_object_unref (image);
}

static gboolean
cdos_background_update_animation (CdosBackground *self)
{
    if (self->animation_timeout_id && !clutter_actor_is_mapped (self->container))
        return TRUE;

    self->use_animation = TRUE;

    const char *filename;
    if (self->is_default && current_background && !self->timeline)
        filename = current_background;
    else
        filename = g_dir_read_name (self->dir);
    char *file_path = NULL;
    while (filename)
    {
        file_path = g_strconcat (self->picture_uri, "/", filename, NULL);
        if (g_file_test (file_path, G_FILE_TEST_IS_REGULAR))
        {
            if (self->is_default)
                current_background = filename;
            break;
        }
        else
        {
            filename = g_dir_read_name (self->dir);
            g_free (file_path);
            file_path = NULL;
        }
    }
    if (!file_path)
    {
        if (!self->file1 && !self->file2)
        {
            self->use_animation = FALSE;
            self->animation_timeout_id = 0;
            return FALSE;
        }
        else
        {
            g_dir_rewind (self->dir);
            return TRUE;
        }
    }
    self->animation_timeout_id = 0;
    /*
    if (g_file_test (file_path, G_FILE_TEST_IS_SYMLINK))
    {
        gchar *real_path = realpath (file_path, NULL);
        g_free (file_path);
        file_path = real_path;
    }
    */

    g_debug ("path: %s %d %s", __func__, __LINE__, file_path);

    GFile *file = g_file_new_for_commandline_arg (file_path);
    g_free (file_path);

    if (!self->timeline)
    {
        self->file1 = file;
        cdos_background_set_color_bg (self);

        self->timeline = clutter_timeline_new (5000); // duration = 5 seconds
        clutter_timeline_set_progress_mode (self->timeline, CLUTTER_EASE_IN_QUAD);
        g_signal_connect (self->timeline, "new-frame", G_CALLBACK (cdos_background_new_frame), self);
        g_signal_connect (self->timeline, "stopped", G_CALLBACK (cdos_background_stop), self);

        self->cache = meta_background_image_cache_get_default ();
        MetaBackgroundImage *image = meta_background_image_cache_load (self->cache, file);
        self->image1 = image;
        if (meta_background_image_is_loaded (image))
            cdos_background_image_loaded (image, self);
        else
            g_signal_connect (image, "loaded", G_CALLBACK (cdos_background_image_loaded), self);
        /*
        meta_background_set_blend (self->meta_background,
                                   self->file1,
                                   NULL,
                                   0,
                                   self->style);
        self->animation_timeout_id = g_timeout_add_seconds (30,
                                     (GSourceFunc)cdos_background_update_animation,
                                     self);
        */
        return FALSE;
    }
    if (g_file_equal (file, self->file1) && !self->file2)
    {
        g_object_unref (file);
        self->use_animation = FALSE;
        return FALSE;
    }
    g_signal_handlers_disconnect_by_data (self->image1, self);

    if (self->image2)
    {
        g_object_unref (self->image1);
        self->image1 = self->image2;
    }

    if (self->file2)
    {
        g_object_unref (self->file1);
        self->file1 = self->file2;
    }
    self->file2 = file;

    MetaBackgroundImage *image = meta_background_image_cache_load (self->cache, file);
    self->image2 = image;
    if (meta_background_image_is_loaded (image))
    {
        cdos_background_image_loaded (image, self);
    }
    else
        g_signal_connect (image, "loaded", G_CALLBACK (cdos_background_image_loaded), self);

    return FALSE;
}

static void
on_settings_changed (FdBgSettings *settings, CdosBackground *self)
{
    free_resources (self);

    GDesktopBackgroundStyle style = fd_bg_settings_get_image_style (self->settings);
    self->style = style;
    gint delay = fd_bg_settings_get_delay (self->settings);
    self->duration = delay > 0 ? delay : -1;
    gboolean use_color = FALSE;
    //是否使用未注册的壁纸。
    gboolean use_unregistered_bg = FALSE;
    //是否是live环境。
    gboolean is_live = cdos_boot_is_live ();
    gchar *uri = NULL;

    //register_status为3时（试用）期间，也不显示此未注册壁纸。
    if (self->is_default && 
        self->check_flags && 
        is_live == FALSE && 
        self->register_status >= 0 && 
        self->register_status != 1 && 
        self->register_status != 2 && 
        self->register_status != 3) 
    {
        const gchar *bg_name = get_bg_unregister_name ();
        gchar *bg_filename = g_build_filename (DEFAULT_WATERMARK_DIR, bg_name, NULL);
        if (g_file_test (bg_filename, G_FILE_TEST_EXISTS))
        {
            uri = g_filename_to_uri (bg_filename, NULL, NULL);
            g_message ("Use unregistered background: %s", uri);
            use_unregistered_bg = TRUE;
        }
        g_free (bg_filename);
    }
    self->check_flags = FALSE;

    if (style != G_DESKTOP_BACKGROUND_STYLE_NONE || use_unregistered_bg)
    {
        if (uri == NULL) 
            uri = fd_bg_settings_get_image_file (self->settings);
        g_message ("Background uri: %s", uri);

        if (self->picture_uri)
            g_free (self->picture_uri);
        self->picture_uri = uri;

        gchar *ptr = uri;
        gchar *unescaped = NULL;
        if (g_str_has_prefix (uri, "file:///"))
        {
            unescaped = g_uri_unescape_string (uri, NULL);
            ptr = unescaped + 7;
        }

        if (g_file_test (ptr, G_FILE_TEST_IS_REGULAR))
        {
            if (!self->file1)
            {
                self->file1 = g_file_new_for_commandline_arg (uri);
                meta_background_set_file (self->meta_background, self->file1, style);
            }
            else
            {
                self->file2 = g_file_new_for_commandline_arg (uri);
                self->cache = meta_background_image_cache_get_default ();
                self->image2 = meta_background_image_cache_load (self->cache, self->file2);     
                if (meta_background_image_is_loaded (self->image2))
                    cdos_background_load_new_bg_image (self->image2, self);
                else
                    g_signal_connect (self->image2, "loaded", G_CALLBACK (cdos_background_load_new_bg_image), self);            
            }
        }
        else if (g_file_test (ptr, G_FILE_TEST_IS_DIR))
        {
            GDir *dir = g_dir_open (uri, 0, NULL);
            if (dir)
            {
                self->dir = dir;
                const char *basename = g_dir_read_name (dir);
                //g_dir_read_name()接口读取dir下某一条目，在cdos_background_update_animation中遍历dir时，该条目被跳过导致若目录下只有1张图片时无法显示
                g_dir_rewind (dir);
                if (!basename)
                {
                    use_color = TRUE;
                    g_dir_close (self->dir);
                    self->dir = NULL;
                }
                else
                {
                    if (self->is_default && current_background)
                    {
                        while (g_strcmp0 (basename, current_background))
                        {
                            basename = g_dir_read_name (dir);
                            if (!basename)
                            {
                                g_dir_rewind (dir);
                                break;
                            }
                        }
                    }
                    cdos_background_update_animation (self);
                }
            }
            else
                use_color = TRUE;
        }
        else
            use_color = TRUE;
        g_free (unescaped);
    }
    else if (style == G_DESKTOP_BACKGROUND_STYLE_NONE || use_color)
        cdos_background_set_color_bg (self);
}

static void
update_watermark (CdosBackground *self, gchar *watermark_name, gint monitor_index)
{
    ClutterContent *image = NULL;
    GError *error = NULL;
    GdkPixbuf *pixbuf = NULL;
    gchar *file_path;
    gfloat wx, wy, dr, db;

    CdosGlobal *global = cdos_global_get ();
    MetaDisplay *display = cdos_global_get_display (global);
    guint ui_scale = cdos_global_get_ui_scale (global);

    if (!is_primary_monitor (self) || self->register_status < 0)
        return ;

    // MetaScreen *screen = cdos_global_get_screen (global);
    MetaRectangle geometry;
    meta_display_get_monitor_geometry (display, monitor_index, &geometry);

    file_path = g_strconcat (DEFAULT_WATERMARK_DIR, watermark_name, NULL);
    if (g_file_test (file_path, G_FILE_TEST_EXISTS) == FALSE)
    {
        g_warning ("The watermark file (%s) is not exists.", file_path);
        goto out;
    }

    pixbuf = gdk_pixbuf_new_from_file (file_path, &error);
    if (error)
    {
        g_warning ("Load watermark file failed: %s", error->message);
        goto out;
    } 

    image = clutter_image_new ();
    gint pixbuf_width = gdk_pixbuf_get_width (pixbuf);
    gint pixbuf_height = gdk_pixbuf_get_height (pixbuf);

    clutter_image_set_data (CLUTTER_IMAGE (image),
                            gdk_pixbuf_get_pixels (pixbuf),
                            gdk_pixbuf_get_has_alpha (pixbuf) 
                            ? COGL_PIXEL_FORMAT_RGBA_8888 
                            : COGL_PIXEL_FORMAT_RGB_888,
                            pixbuf_width,
                            pixbuf_height,
                            gdk_pixbuf_get_rowstride (pixbuf),
                            &error);
    if (error)
    {
        g_warning ("Set watermark file failed: %s", error->message);
        goto out;
    }

    if (self->watermark == NULL){
        self->watermark = clutter_actor_new ();
        clutter_actor_add_child (CLUTTER_ACTOR(self->container), CLUTTER_ACTOR(self->watermark));
    }

    clutter_actor_set_content (self->watermark, image);
    clutter_actor_set_width (self->watermark, pixbuf_width);
    clutter_actor_set_height (self->watermark, pixbuf_height);
    clutter_actor_set_scale (self->watermark, (float)ui_scale, (float)ui_scale);

    //设置右侧和底部的距离
    dr = (77 + pixbuf_width) * ui_scale;
    db = (97 + pixbuf_height) * ui_scale;

    wx = (float)(geometry.x + geometry.width - dr);
    wy = (float)(geometry.y + geometry.height - db);

    clutter_actor_set_position (self->watermark, wx, wy);

out:
    if (error)
        g_error_free (error);
    if (pixbuf)
        g_object_unref (pixbuf);
    if (image)
        g_object_unref (image);
    g_free (file_path);
}


static int
get_state_from_data (gchar *data)
{
    JsonParser *parser = json_parser_new ();
    json_parser_load_from_data (parser, data, -1, NULL);

    JsonReader *reader = json_reader_new (json_parser_get_root (parser)); 
    json_reader_read_member (reader, "state");
    gint state = json_reader_get_int_value (reader);
    json_reader_end_member (reader);


    g_object_unref (reader);
    g_object_unref (parser);

    return state;
}

static gchar *
cdos_get_register_name (void)
{
    gchar *reg_name = NULL;
    if (0 != access ("/etc/licenses.d/os", F_OK))
    {
        reg_name = g_strdup ("os");
    } else {
        FILE *stream;
        char strline[512];
        stream = popen ("echo", "r");
        fgets (strline, sizeof (strline), stream);
        stream = popen ("cat /etc/licenses.d/os | head -1", "r");
        fgets (strline, sizeof (strline), stream);
        pclose (stream);

        char *tmp = NULL;
        if ((tmp = strstr (strline, "\n")))
        {
        *tmp = '\0';
        }

        if (g_strcmp0 (strline, "") == 0)
            reg_name = g_strdup ("os");
        else
            reg_name = g_strdup (strline);
    }
    g_message ("Desktop register name: %s", reg_name);
    return reg_name;
}

static void
on_presence_status_changed (GDBusProxy *proxy,
                            gchar *sender,
                            gchar *signal,
                            GVariant *parameters,
                            CdosBackground *self)
{
    if (g_strcmp0 (signal, "licenseSuccess") == 0)
    {
        gchar *register_name;
        g_variant_get (parameters, "(s)", &register_name);
        if (g_strcmp0 (register_name, self->register_name) == 0)
        {
            cdos_background_check_registe_status_delay (self);
        }
        g_free (register_name);
    }
}

static void
update_watermark_from_status (CdosBackground *self)
{
    const gchar *watermark_name = NULL;
    if (self->register_status == 1 || self->register_status == 2)
    {
        watermark_name = get_register_watermark_name (TRUE);
    } else {
        watermark_name = get_register_watermark_name (FALSE);
    }   
    update_watermark (self, watermark_name, self->monitor_index);
}

static void
get_license_status_cb (GObject *source,
            GAsyncResult *res,
            CdosBackground *self)
{
    GError *error = NULL;
    GVariant *ret;
    gchar *result;
    if (!G_IS_DBUS_PROXY (self->register_proxy))
        return;

    ret = g_dbus_proxy_call_finish (self->register_proxy, res, &error);
    if (error != NULL)
    {
        g_warning ("Get license status failed: %s", error->message);
        g_clear_error (&error);
        goto out;
    }
    g_variant_get (ret, "(s)", &result);
    g_variant_unref (ret);
    if (!result)
    {
        g_warning ("Did't get any data from the registered server.");
        goto out;
    }
    self->register_status = get_state_from_data (result);
    g_free (result);

    //regidtered_status 0: 试用期(未激活)；1 已激活； 2激活(已过期)； 3试用期；4试用期(已过期)共5个状态
    g_message ("Register status changed: %d.", self->register_status);

    //刷新壁纸
    self->check_flags = TRUE;
    on_settings_changed (NULL, self);

out:
    update_watermark_from_status (self);
}

static gboolean
do_check_registe_status (CdosBackground *self)
{
    GError *error = NULL;
    GVariant *ret;
    const gchar *watermark_name = NULL;
    self->check_registe_status_id = 0;
    if (!self->register_proxy)
    {
        //dbus不存在时，可能无注册功能，此时显示已注册的水印。
        self->register_status = 1;
        update_watermark_from_status (self);
        return FALSE;
    }

    // 调用方法,同步改成异步
    g_cancellable_reset (self->cancellable);
    g_dbus_proxy_call(self->register_proxy,
                    "GetLicenseStatus",
                    g_variant_new ("(s)", self->register_name),
                    G_DBUS_CALL_FLAGS_NONE,
                    -1,
                    self->cancellable,
                    get_license_status_cb,
                    self);
    return FALSE;
}

static void
cdos_background_check_registe_status_delay (CdosBackground *self)
{
    if (self->check_registe_status_id > 0)
    {
        g_source_remove (self->check_registe_status_id);
    }
    self->check_registe_status_id = \
        g_timeout_add (1000, 
                       (GSourceFunc)do_check_registe_status, 
                       self);
}

static void
cdos_background_update_watermark (GObject *source,
                GAsyncResult *res,
                CdosBackground *self)
{
    GError *error = NULL;

    self->register_proxy = g_dbus_proxy_new_finish (res, &error);
    if (error)
    {
        g_warning ("Connect dbus failed:%s", error->message);
        g_clear_error (&error);
    } else {
        if (!self->register_name)
        {
            self->register_name = cdos_get_register_name ();
        }
        g_signal_connect (self->register_proxy, 
                            "g-signal", 
                            G_CALLBACK (on_presence_status_changed), 
                            self);
        cdos_background_check_registe_status_delay (self);
    }
}

/**
 * cdos_background_new:
 *
 * 创建新的背景
 *
 * Returns:新创建的 #CdosBackground
 *
 * Since:1.0.0
 */
CdosBackground *
cdos_background_new (ClutterActor *container,
                     guint monitor_index,
                     gboolean vignette,
                     gboolean control_position,
                     gboolean ignore_licenses,
                     const gchar *settings_schema)
{
    CdosBackground *self;
    self = g_new0 (CdosBackground, 1);
    if (!self)
        return NULL;

    self->container = container;

    g_debug ("Background settings schema: %s", settings_schema);

    if (settings_schema && g_strcmp0 (settings_schema, BACKGROUND_SCHEMA))
    {
        self->is_default = FALSE;
        self->settings = fd_bg_settings_new (settings_schema, NULL);
    } else {
        self->is_default = TRUE;
        self->settings = get_default_settings ();
    }
    self->vignette = vignette;
    self->monitor_index = monitor_index;
    self->control_position = control_position;
    self->register_status = -1;
    self->check_flags = FALSE;

    CdosGlobal *global = cdos_global_get ();
    MetaDisplay *display = cdos_global_get_display (global);
    self->display = display;
    MetaRectangle geometry;
    meta_display_get_monitor_geometry (display, monitor_index, &geometry);
    gint primary_monitor = meta_display_get_primary_monitor (display);
    // self->monitor_changed_id = g_signal_connect (screen,"monitors-changed",
    // G_CALLBACK(cdos_background_screen_on_monitors_changed), self);

    if (!self->actor)
    {
        ClutterActor *background_actor = meta_background_actor_new (display, self->monitor_index);
        if (self->vignette)
            meta_background_actor_set_vignette ((MetaBackgroundActor *)background_actor, 1, 1.0, 0.0);
        self->actor = background_actor;
        clutter_actor_add_child (self->container, self->actor);
    }
    self->meta_background = meta_background_new (display);
    meta_background_actor_set_background ((MetaBackgroundActor *)self->actor, self->meta_background);
    self->settings_changed_id = g_signal_connect (self->settings, 
                                                  "changed", 
                                                  G_CALLBACK (on_settings_changed), 
                                                  self);
    on_settings_changed (NULL, self);

    clutter_actor_set_size (self->actor, geometry.width, geometry.height);
    clutter_actor_set_position (self->actor, geometry.x, geometry.y);

    //if (primary_monitor == monitor_index)
    {
        self->cancellable = g_cancellable_new ();
        do_check_registe_status (self);
        // 同步调用改成异步调用
        if (ignore_licenses == FALSE)
            g_dbus_proxy_new_for_bus (G_BUS_TYPE_SYSTEM,
                        G_DBUS_PROXY_FLAGS_NONE,
                        NULL,
                        NFS_LICENSE_NAME,
                        NFS_LICENSE_PATH,
                        NFS_LICENSE_INTERFACE,
                        self->cancellable,
                        (GAsyncReadyCallback)cdos_background_update_watermark,
                        self);
    }

    return self;
}

/*
 * CdosBackgroundManager:
 */

static void
on_actor_destroy (ClutterActor *actor, CdosBackgroundManager *self)
{
    if (self->in_destroy)
        return;
    for (GList *iter = self->backgrounds; iter; iter = iter->next)
    {
        CdosBackground *background = iter->data;
        if (background->actor == actor)
        {
            self->backgrounds = g_list_remove (self->backgrounds, background);
            /*
            if (self->backgrounds == NULL)
                cdos_background_manager_destroy (self);
            */
            return;
        }
    }
}


void
cdos_background_manager_hide_register_watermark (CdosBackgroundManager *self)
{
    g_return_if_fail (self != NULL);

    for (GList *iter = self->backgrounds; iter; iter = iter->next)
    {
        CdosBackground *background = iter->data;
        if (background != NULL && background->watermark != NULL)
        {
            clutter_actor_hide (CLUTTER_ACTOR (background->watermark));
        }
    }
}

static gboolean
rebuild_backgrounds_delay (CdosBackgroundManager *self)
{
    CdosBackground *background;
    self->rebuild_backgrounds_id = 0;
    g_list_free_full (self->backgrounds, cdos_background_destroy);
    self->backgrounds = NULL;

    int n = meta_display_get_n_monitors (self->display);
    
    for (int i = 0; i < n; i++)
    {
        background = cdos_background_new (self->container, i, 
                                          self->vignette, 
                                          self->control_position, 
                                          self->ignore_licenses,
                                          self->settings_schema);
        self->backgrounds = g_list_prepend (self->backgrounds, background);
        //g_signal_connect (background->actor, "destroy", G_CALLBACK (on_actor_destroy), self);
    }
    return FALSE;
}

static void
cdos_background_manager_on_monitors_changed (MetaMonitorManager *monitor_manager, CdosBackgroundManager *self)
{
    if (self->rebuild_backgrounds_id > 0)
    {
        g_source_remove (self->rebuild_backgrounds_id);
        self->rebuild_backgrounds_id = 0;
    }
    self->rebuild_backgrounds_id = \
            g_timeout_add (250, 
                           (GSourceFunc)rebuild_backgrounds_delay, 
                           self);
}

CdosBackgroundManager *
cdos_background_manager_new (ClutterActor *container,
                             gboolean vignette,
                             gboolean control_position,
                             gboolean ignore_licenses,
                             const gchar *settings_schema)
{
    CdosBackgroundManager *self = g_new0 (CdosBackgroundManager, 1);
    self->container = container;
    self->vignette = vignette;
    self->control_position = control_position;
    self->settings_schema = g_strdup (settings_schema);
    self->ignore_licenses = ignore_licenses;

    CdosGlobal *global = cdos_global_get ();
    MetaDisplay *display = cdos_global_get_display (global);
    self->display = display;
    self->monitor_manager = meta_monitor_manager_get ();
    self->monitor_changed_id = \
        g_signal_connect_after (self->monitor_manager, 
                                "monitors-changed", 
                                G_CALLBACK (cdos_background_manager_on_monitors_changed), 
                                self);
    rebuild_backgrounds_delay (self);
    return self;
}

void
cdos_background_manager_destroy (CdosBackgroundManager *self)
{
    self->in_destroy = 1;
    g_signal_handler_disconnect (self->monitor_manager, self->monitor_changed_id);
    if (self->rebuild_backgrounds_id > 0)
    {
        g_source_remove (self->rebuild_backgrounds_id);
        self->rebuild_backgrounds_id = 0;
    }

    g_list_free_full (self->backgrounds, cdos_background_destroy);
    self->backgrounds = NULL;

    if (self->nemo_proxy)
    {
        g_object_unref (self->nemo_proxy);
        self->nemo_proxy = NULL;
    }
    g_free (self->settings_schema);
    g_free (self);
}
