/**
 * SECTION: cdos-box-pointer
 * @short_description: 为弹出框提供了带有明显指向标识（如指向关联元素的小三角）的外加容器。
 * @title: CdosBoxPointer
 * @include: cdos-box-pointer.h
 *
 * #CdosBoxPointer为弹出框(#CdosPopupMenu)提供带有小三角标识的外围容器元素，通常作为PopupMenu的包裹元素。
 *
 * CdosBoxPointer的结构成员包括：actor,container,bin,border。
 *
 * CdosBoxPointer结构关系:
 *
 * |__ actor:StBin:主容器元素，不显示
 *
 * |____ container:CdosGernerContainer:子容器元素，不显示
 *
 * |________ bin:StBin:包裹容器元素
 *
 * |________ border:StDrawingArea:标识元素
 *
 */
#include <math.h>
#include <st.h>
#include "cdos-transition.h"
#include "cdos-box-pointer.h"
#include "cdos-base-util.h"
#include "cdos-layout-manager.h"

struct _CdosBoxPointerPrivate
{
    gint                  arrow_side;
    gboolean              is_popup;
    gdouble               arrow_origin;
    gfloat                x_offset;
    gfloat                y_offset;
    gfloat                x_position;
    gfloat                y_position;
    gfloat                arrow_alignment;
    gfloat                source_alignment;
    ClutterActor         *source_actor;
    guint                 opacity;
    ClutterEffect        *blur_effect;
    graphene_point_t     *pos;
};

typedef enum _CdosBoxPointerAnimationType
{
    CDOS_BOX_POINTER_ANIMATION_NONE  = 0,
    CDOS_BOX_POINTER_ANIMATION_SLIDE = 1 << 0,
    CDOS_BOX_POINTER_ANIMATION_FADE  = 1 << 1,
    CDOS_BOX_POINTER_ANIMATION_ALL   = ~0
}CdosBoxPointerAnimationType;

enum
{
    PROP_0,
    PROP_IS_POPUP,
    PROP_ARROW_SIDE
};

G_DEFINE_TYPE_WITH_PRIVATE (CdosBoxPointer, cdos_box_pointer, ST_TYPE_WIDGET);

static void
cdos_box_pointer_adjust_allocation_for_arrow (CdosBoxPointer *self,
                                              gboolean        is_width,
                                              gfloat         *min_size,
                                              gfloat         *natural_size)
{
    CdosBoxPointerPrivate *priv;
    gdouble border_width, rise;
    priv = self->priv;
    StThemeNode *theme_node = st_widget_get_theme_node (ST_WIDGET (self));
    border_width = st_theme_node_get_length (theme_node, "-arrow-border-width");
    *min_size += border_width * 2;
    *natural_size += border_width * 2;
    if ((!is_width
         && (priv->arrow_side == ST_SIDE_TOP || priv->arrow_side == ST_SIDE_BOTTOM))
        || (is_width
         && (priv->arrow_side == ST_SIDE_LEFT || priv->arrow_side == ST_SIDE_RIGHT)))
    {
        rise = st_theme_node_get_length (theme_node, "-arrow-rise");
        *min_size += rise;
        *natural_size += rise;
    }
}

static void
cdos_box_pointer_reposition (CdosBoxPointer *self, ClutterActorBox *box)
{
    CdosBoxPointerPrivate *priv;
    priv = self->priv;

    gfloat res_x = 0;
    gfloat res_y = 0;
    if (priv->pos)
    {
        priv->x_position = priv->pos->x;
        priv->y_position = priv->pos->y;
        goto finish;
    }
    StThemeNode  *source_node;
    gfloat source_center_x;
    gfloat source_center_y;
    ClutterActorBox allocation_box;
    ClutterActorBox source_content_box;
    ClutterActorBox source_allocation;
    gfloat min_width;
    gfloat min_height;
    gfloat nat_width;
    gfloat nat_height;

    StThemeNode *theme_node;
    gdouble    border_width;
    gdouble    arrow_base;
    gdouble    border_raidus;
    gdouble    margin;
    gdouble    half_margin;
    gdouble    gap;

    gfloat alignment = priv->arrow_alignment;

    source_node = st_widget_get_theme_node (ST_WIDGET (priv->source_actor));

    clutter_actor_get_allocation_box (priv->source_actor, &allocation_box);
    st_theme_node_get_content_box (source_node, &allocation_box, &source_content_box);

    cdos_base_util_get_transformed_allocation (priv->source_actor, &source_allocation);

    source_center_x = source_allocation.x1 + source_content_box.x1
            + (source_content_box.x2 - source_content_box.x1) * priv->source_alignment;
    source_center_y = source_allocation.y1 + source_content_box.y1
            + (source_content_box.y2 - source_content_box.y1) * priv->source_alignment;

    clutter_actor_get_preferred_size (CLUTTER_ACTOR (self), &min_width, &min_height, &nat_width, &nat_height);

    CdosLayoutManager *layout = cdos_layout_manager_get_default ();
    const MetaRectangle *monitor = cdos_layout_manager_find_monitor_for_actor (layout, priv->source_actor);

    theme_node = st_widget_get_theme_node (ST_WIDGET (self));

    border_width = st_theme_node_get_length (theme_node, "-arrow-border-width");
    arrow_base = st_theme_node_get_length (theme_node, "-arrow-base");
    border_raidus = st_theme_node_get_length (theme_node, "-arrow-border-radius");

    margin = 4 * border_raidus + border_width + arrow_base;
    half_margin = margin / 2.0;
    gap = st_theme_node_get_length (theme_node, "-boxpointer-gap");

    switch (priv->arrow_side)
    {
        case ST_SIDE_TOP:
            res_y = source_allocation.y2 + gap;
            break;
        case ST_SIDE_BOTTOM:
            res_y = source_allocation.y1 - nat_height - gap;
            break;
        case ST_SIDE_LEFT:
            res_x = source_allocation.x2 + gap;
            break;
        case ST_SIDE_RIGHT:
            res_x = source_allocation.x1 - nat_width - gap;
            break;
    }
    // Now align and position the pointing axis, making sure
    // it fits on screen
    switch (priv->arrow_side)
    {
        case ST_SIDE_TOP:
        case ST_SIDE_BOTTOM:
            res_x = source_center_x - (half_margin + (nat_width - margin) * alignment);
            //res_x = MAX(res_x, monitor.x + 10);
            /*
             *调整CdosBoxPointer在x轴方向的偏移量为2 pix。如果偏移量<2，开始菜单在显示时会挤开屏幕上已经全屏显示的应用程序
             */
            if (monitor != NULL)
            {
                res_x = MAX(res_x, monitor->x + 2);
                res_x = MIN(res_x, monitor->x + monitor->width - (10 + nat_width));
            }
            cdos_box_pointer_set_arrow_origin (self, source_center_x - res_x);
            break;
        case ST_SIDE_LEFT:
        case ST_SIDE_RIGHT:
            res_y = source_center_y - (half_margin + (nat_height - margin) * alignment);
            if (monitor != NULL)
            {
                res_y = MAX(res_y, monitor->y + 10);
                res_y = MIN(res_y, monitor->y + monitor->height - (10 + nat_height));
            }
            cdos_box_pointer_set_arrow_origin(self, source_center_y - res_y);
            break;
    }
finish:
    if (priv->pos)
    {
        res_x = priv->pos->x;
        res_y = priv->pos->y;
    }
    ClutterActor *parent = clutter_actor_get_parent (CLUTTER_ACTOR (self));
    gboolean success = FALSE;

    gfloat x, y;
    while (!success)
    {
        success = clutter_actor_transform_stage_point (parent, res_x, res_y, &x, &y);
        parent = clutter_actor_get_parent (parent);
    }
    priv->x_position = floor(x);
    priv->y_position = floor(y);

    clutter_actor_box_set_origin (box, floor(x), floor(y));
}

static void
cdos_box_pointer_get_preferred_width (ClutterActor *actor,
                                      gfloat for_height,
                                      gfloat *min_width,
                                      gfloat *nat_width)
{
    CdosBoxPointer *self = (CdosBoxPointer *)actor;
    if (!self->priv->is_popup)
    {
        CLUTTER_ACTOR_CLASS (cdos_box_pointer_parent_class)->get_preferred_width (actor, for_height, min_width, nat_width);
        return;
    }

    StThemeNode *theme_node = st_widget_get_theme_node (ST_WIDGET (self));
    st_theme_node_adjust_for_height (theme_node, &for_height);

    clutter_actor_get_preferred_width (CLUTTER_ACTOR (self->bin), for_height, min_width, nat_width);
    cdos_box_pointer_adjust_allocation_for_arrow (self, TRUE, min_width, nat_width);
    st_theme_node_adjust_preferred_width (theme_node, min_width, nat_width);
}

static void
cdos_box_pointer_get_preferred_height (ClutterActor *actor,
                                       gfloat for_width,
                                       gfloat *min_height,
                                       gfloat *nat_height)
{
    CdosBoxPointer *self = (CdosBoxPointer *)actor;
    if (!self->priv->is_popup)
    {
        CLUTTER_ACTOR_CLASS (cdos_box_pointer_parent_class)->get_preferred_height (actor, for_width, min_height, nat_height);
        return;
    }

    StThemeNode *theme_node = st_widget_get_theme_node (ST_WIDGET (self));
    gdouble border_width = st_theme_node_get_length (theme_node, "-arrow-border-width");
    st_theme_node_adjust_for_width (theme_node, &for_width);

    if (st_theme_node_get_height (theme_node) == -1)
    {
        clutter_actor_get_preferred_height (CLUTTER_ACTOR(self->bin),
                                            for_width - 2 * border_width,
                                            min_height,
                                            nat_height);
    }
    cdos_box_pointer_adjust_allocation_for_arrow (self, FALSE, min_height, nat_height);
    st_theme_node_adjust_preferred_height (theme_node, min_height, nat_height);
}

static void
cdos_box_pointer_allocate (ClutterActor *actor,
                           const ClutterActorBox *box,
                           ClutterAllocationFlags flags)
{
    CdosBoxPointer *self = (CdosBoxPointer *)actor;
    CdosBoxPointerPrivate *priv = self->priv;

    ClutterActorBox child_box, content_box;

    clutter_actor_set_allocation (actor, box, flags);

    StThemeNode *theme_node = st_widget_get_theme_node (ST_WIDGET (self));

    st_theme_node_get_content_box (theme_node, box, &content_box);

    if (!priv->is_popup)
    {
        clutter_actor_set_allocation (actor, box, flags);
        clutter_actor_allocate (clutter_actor_get_first_child (actor), &content_box, flags);
        return;
    }

    gdouble border_width = st_theme_node_get_length (theme_node, "-arrow-border-width");
    gdouble rise = st_theme_node_get_length (theme_node, "-arrow-rise");

    /*
    gfloat avail_width, avail_height;
    avail_width = content_box.x2 - content_box.x1;
    avail_height = content_box.y2 - content_box.y1;
    */

    clutter_actor_allocate (CLUTTER_ACTOR (self->border), &content_box, flags);
    child_box.x1 = content_box.x1 + border_width;
    child_box.y1 = content_box.y1 + border_width;
    child_box.x2 = content_box.x2 - border_width;
    child_box.y2 = content_box.y2 - border_width;

    switch (priv->arrow_side)
    {
        case ST_SIDE_TOP:
            child_box.y1 += rise;
            break;
        case ST_SIDE_BOTTOM:
            child_box.y2 -= rise;
            break;
        case ST_SIDE_LEFT:
            child_box.x1 += rise;
            break;
        case ST_SIDE_RIGHT:
            child_box.x2 -= rise;
            break;
    }

    clutter_actor_allocate (CLUTTER_ACTOR (self->bin), &child_box, flags);

    if (priv->source_actor && clutter_actor_is_mapped (priv->source_actor))
    {
        ClutterActorBox box1 = *box;
        cdos_box_pointer_reposition (self, &box1);
        clutter_actor_set_allocation (actor, &box1, flags);
    }
}

static void
cdos_box_pointer_set_is_popup (CdosBoxPointer *self, gboolean is_popup)
{
    self->priv->is_popup = is_popup;
}

static void
draw_border (StDrawingArea *area, CdosBoxPointer *self)
{
    CdosBoxPointerPrivate *priv = self->priv;
    StThemeNode *theme_node = st_widget_get_theme_node (ST_WIDGET (self));

    gdouble border_width, border_base, border_rise;
    st_theme_node_lookup_length (theme_node, "-arrow-border-width", FALSE, &border_width);
    st_theme_node_lookup_length (theme_node, "-arrow-base", FALSE, &border_base);
    st_theme_node_lookup_length (theme_node, "-arrow-rise", FALSE, &border_rise);

    if (!border_base || !border_rise)
        return;

    gdouble border_raidus = st_theme_node_get_length (theme_node, "-arrow-border-radius");

    gdouble half_border = border_width / 2;
    gdouble half_base = floor (border_base / 2);

    ClutterColor border_color, background_color;
    GdkRGBA gdk_border_color, gdk_background_color;
    st_theme_node_lookup_color (theme_node, "-arrow-border-color", FALSE, &border_color);
    st_theme_node_lookup_color (theme_node, "-arrow-background-color", FALSE, &background_color);

    gdk_border_color.red   = border_color.red / 255.0;
    gdk_border_color.green = border_color.green / 255.0;
    gdk_border_color.blue  = border_color.blue / 255.0;
    gdk_border_color.alpha = border_color.alpha / 255.0;

    gdk_background_color.red   = background_color.red / 255.0;
    gdk_background_color.green = background_color.green / 255.0;
    gdk_background_color.blue  = background_color.blue / 255.0;
    gdk_background_color.alpha = background_color.blue / 255.0;

    guint width, height;
    st_drawing_area_get_surface_size (area, &width, &height);
    guint box_width, box_height;

    box_width = width;
    box_height = height;
    if (priv->arrow_side == ST_SIDE_TOP || priv->arrow_side == ST_SIDE_BOTTOM)
        box_height -= border_rise;
    else
        box_width -= border_rise;

    cairo_t *cr = st_drawing_area_get_context (area);

    gdk_cairo_set_source_rgba (cr, &gdk_border_color);

    if (priv->arrow_side == ST_SIDE_TOP)
        cairo_translate (cr, 0, border_rise);
    else
        cairo_translate (cr, border_rise, 0);

    gdouble x1 = half_border, y1 = half_border;
    gdouble x2 = box_width - half_border;
    gdouble y2 = box_height - half_border;
    cairo_move_to (cr, x1 + border_raidus, y1);

    if (priv->arrow_side == ST_SIDE_TOP)
    {
        if (priv->arrow_origin < (x1 + (border_raidus + half_base)))
        {
            cairo_line_to (cr, priv->arrow_origin, y1 - border_rise);
            cairo_line_to (cr, MAX(x1 + border_raidus, priv->arrow_origin) + half_base, y1);
        }
        else if (priv->arrow_origin > (x2 - (border_raidus + half_base)))
        {
            cairo_line_to (cr, MIN(x2 - border_raidus, priv->arrow_origin) - half_base, y1);
            cairo_line_to (cr, priv->arrow_origin, y1 - border_rise);
        }
        else
        {
            cairo_line_to (cr, priv->arrow_origin - half_base, y1);
            cairo_line_to (cr, priv->arrow_origin, y1 - border_rise);
            cairo_line_to (cr, priv->arrow_origin + half_base, y1);
        }
    }
    cairo_line_to (cr, x2 - border_raidus, y1);
    cairo_arc (cr, x2 - border_raidus, y1 + border_raidus, border_raidus, 3 * G_PI / 2, G_PI / 2);

    if (priv->arrow_side == ST_SIDE_RIGHT)
    {
        if (priv->arrow_origin < (y1 + (border_raidus + half_base)))
        {
            cairo_line_to (cr, x2 + border_rise, priv->arrow_origin);
            cairo_line_to (cr, x2, MAX(y1 + border_raidus, priv->arrow_origin) + half_base);
        }
        else if (priv->arrow_origin > (y2 - (border_raidus + half_base)))
        {
            cairo_line_to (cr, x2, MIN(y2 - border_raidus, priv->arrow_origin) - half_base);
            cairo_line_to (cr, x2 + border_rise, priv->arrow_origin);
        }
        else
        {
            cairo_line_to (cr, x2, priv->arrow_origin - half_base);
            cairo_line_to (cr, x2 + border_rise, priv->arrow_origin);
            cairo_line_to (cr, x2, priv->arrow_origin + half_base);
        }
    }
    cairo_line_to (cr, x2, y2 - border_raidus);
    cairo_arc (cr, x2 - border_raidus, y2 - border_raidus, border_raidus, 0, G_PI / 2);

    if (priv->arrow_side == ST_SIDE_BOTTOM)
    {
        if (priv->arrow_origin < (x1 + (border_raidus + half_base)))
        {
            cairo_line_to (cr, MAX(x1 + border_raidus, priv->arrow_origin) + half_base, y2);
            cairo_line_to (cr, priv->arrow_origin, y2 + border_rise);
        }
        else if (priv->arrow_origin > (x2 - (border_raidus + half_base)))
        {
            cairo_line_to (cr, priv->arrow_origin, y2 + border_rise);
            cairo_line_to (cr, MIN(x2 - border_raidus, priv->arrow_origin) - half_base, y2);
        }
        else
        {
            cairo_line_to (cr, priv->arrow_origin + half_base, y2);
            cairo_line_to (cr, priv->arrow_origin, y2 + border_rise);
            cairo_line_to (cr, priv->arrow_origin - half_base, y2);
        }
    }
    cairo_line_to (cr, x1 + border_raidus, y2);
    cairo_arc (cr, x1 + border_raidus, y2 - border_raidus, border_raidus, G_PI / 2, G_PI / 2);

    if (priv->arrow_side == ST_SIDE_LEFT)
    {
        if (priv->arrow_origin < (y1 + (border_raidus + half_base)))
        {
            cairo_line_to (cr, x1, MAX(y1 + border_raidus, priv->arrow_origin) + half_base);
            cairo_line_to (cr, x1 - border_rise, priv->arrow_origin);
        }
        else if (priv->arrow_origin > (y2 - (border_raidus + half_base)))
        {
            cairo_line_to (cr, x1 - border_rise, priv->arrow_origin);
            cairo_line_to (cr, x1, MIN(y2 - border_raidus, priv->arrow_origin) - half_base);
        }
        else
        {
            cairo_line_to (cr, x1, priv->arrow_origin + half_base);
            cairo_line_to (cr, x1 - border_rise, priv->arrow_origin);
            cairo_line_to (cr, x1, priv->arrow_origin - half_base);
        }
    }
    cairo_line_to (cr, x1, y1 + border_raidus);
    cairo_arc (cr, x1 + border_raidus, y1 + border_raidus, border_raidus, G_PI, 3 * G_PI / 2);

    gdk_cairo_set_source_rgba (cr, &gdk_background_color);
    cairo_fill_preserve (cr);
    gdk_cairo_set_source_rgba (cr, &gdk_border_color);
    cairo_set_line_width (cr, border_width);
    cairo_stroke (cr);
}

static void
cdos_box_pointer_set_property (GObject *gobject,
                               guint    prop_id,
                               const GValue *value,
                               GParamSpec   *pspec)
{
    CdosBoxPointer *self = CDOS_BOX_POINTER (gobject);
    switch (prop_id)
    {
        case PROP_IS_POPUP:
            cdos_box_pointer_set_is_popup (self, g_value_get_boolean (value));
            break;
        case PROP_ARROW_SIDE:
            cdos_box_pointer_set_arrow_side (self, g_value_get_int (value));
            break;
        default:
            G_OBJECT_WARN_INVALID_PROPERTY_ID (gobject, prop_id, pspec);
            break;
    }
}

static void
cdos_box_pointer_get_property (GObject    *gobject,
                               guint       prop_id,
                               GValue     *value,
                               GParamSpec *pspec)
{
    CdosBoxPointer *self = CDOS_BOX_POINTER (gobject);
    CdosBoxPointerPrivate *priv = self->priv;

    switch (prop_id)
    {
        case PROP_IS_POPUP:
            g_value_set_boolean (value, priv->is_popup);
            break;
        case PROP_ARROW_SIDE:
            g_value_set_int (value, priv->arrow_side);
            break;
        default:
            G_OBJECT_WARN_INVALID_PROPERTY_ID (gobject, prop_id, pspec);
            break;
    }
}

static void
cdos_box_pointer_constructed (GObject *gobject)
{
    CdosBoxPointer *self = CDOS_BOX_POINTER (gobject);
    CdosBoxPointerPrivate *priv = self->priv;

    ClutterActor *actor = CLUTTER_ACTOR (self);
    //clutter_actor_set_offscreen_redirect (actor, CLUTTER_OFFSCREEN_REDIRECT_AUTOMATIC_FOR_OPACITY);

    if (!priv->is_popup)
        goto finish;

    self->bin = g_object_new (ST_TYPE_BIN, "x-fill", TRUE, "y-fill", TRUE, "x-align", ST_ALIGN_START, "style-class", "box-pointer",NULL);
    st_bin_set_fill (self->bin, TRUE, TRUE);

    self->border = g_object_new (ST_TYPE_DRAWING_AREA, NULL);
    clutter_actor_set_reactive (CLUTTER_ACTOR (self->bin), TRUE);

    g_signal_connect (self->border, "repaint", G_CALLBACK(draw_border), self);

    clutter_actor_add_child (actor, CLUTTER_ACTOR (self->border));
    clutter_actor_add_child (actor, CLUTTER_ACTOR (self->bin));
    clutter_actor_set_child_above_sibling (actor, CLUTTER_ACTOR (self->bin), CLUTTER_ACTOR (self->border));

finish:
    G_OBJECT_CLASS (cdos_box_pointer_parent_class)->constructed (gobject);
}

static void
cdos_box_pointer_finalize (GObject *gobject)
{
    CdosBoxPointer *self = CDOS_BOX_POINTER (gobject);
    CdosBoxPointerPrivate *priv = self->priv;

    g_free (priv->pos);

    G_OBJECT_CLASS(cdos_box_pointer_parent_class)->finalize (gobject);
}

static void
cdos_box_pointer_class_init (CdosBoxPointerClass *klass)
{
    GObjectClass *gobject_class = G_OBJECT_CLASS (klass);
    ClutterActorClass *actor_class = CLUTTER_ACTOR_CLASS (klass);

    gobject_class->finalize = cdos_box_pointer_finalize;
    gobject_class->get_property = cdos_box_pointer_get_property;
    gobject_class->set_property = cdos_box_pointer_set_property;
    gobject_class->constructed = cdos_box_pointer_constructed;

    actor_class->get_preferred_width = cdos_box_pointer_get_preferred_width;
    actor_class->get_preferred_height = cdos_box_pointer_get_preferred_height;
    actor_class->allocate = cdos_box_pointer_allocate;

    GParamSpec *pspec;

    pspec = g_param_spec_boolean ("is-popup",
                                  "is-popup",
                                  "wether is popup menu",
                                  FALSE,
                                  G_PARAM_READWRITE | G_PARAM_CONSTRUCT_ONLY);

    g_object_class_install_property (gobject_class,
                                     PROP_IS_POPUP,
                                     pspec);

    pspec = g_param_spec_int ("arrow-side",
                              "Arrow side",
                              "Arrow side of popup menu",
                              ST_SIDE_TOP,
                              ST_SIDE_LEFT,
                              ST_SIDE_TOP,
                              G_PARAM_READWRITE);

    g_object_class_install_property (gobject_class,
                                     PROP_ARROW_SIDE,
                                     pspec);
}

static void
cdos_box_pointer_init (CdosBoxPointer *self)
{
    CdosBoxPointerPrivate *priv;
    priv = self->priv = cdos_box_pointer_get_instance_private (self);

    self->bin = NULL;
    self->border = NULL;
    priv->source_actor = NULL;

    priv->source_alignment = 0.5;
    priv->arrow_side = 0;
    priv->arrow_origin = 0;
    priv->x_offset = 0;
    priv->y_offset = 0;
    priv->x_position = 0;
    priv->y_position = 0;
    priv->pos = NULL;

    if (cdos_enable_blur_effect ())
    {
        priv->blur_effect = st_blur_effect_new ();
        st_blur_effect_set_mode (ST_BLUR_EFFECT (priv->blur_effect), ST_BLUR_MODE_BACKGROUND);
        st_blur_effect_set_sigma (ST_BLUR_EFFECT (priv->blur_effect), 2.0);
        clutter_actor_add_effect (CLUTTER_ACTOR (self), priv->blur_effect);
    }
}

/**
 * cdos_box_pointer_new:
 * @arrow_side: 锚点所在的方位，取值为StSide类型成员
 *
 * 创建#CdosBoxPointer实例对象
 *
 * Returns:返回一个#CdosBoxPointer实例。
 */
CdosBoxPointer *
cdos_box_pointer_new (gint arrow_side)
{
    CdosBoxPointer *self;
    self = g_object_new (CDOS_TYPE_BOX_POINTER, "is-popup", TRUE, NULL);
    cdos_box_pointer_set_arrow_side (self, arrow_side);
    return self;
}

/**
 * cdos_box_pointer_show:
 * @self: 当前实例对象
 * @animate: 是否使用过渡动画，TRUE-使用，FALSE-不使用
 *
 * 显示当前实例主容器元素。
 */
void
cdos_box_pointer_show (CdosBoxPointer *self, gboolean animate)
{
    g_return_if_fail (CDOS_IS_BOX_POINTER (self));
    CdosBoxPointerPrivate *priv;
    priv = self->priv;
    ClutterActor *actor = CLUTTER_ACTOR (self);
    StThemeNode *theme_node = st_widget_get_theme_node (ST_WIDGET(self));
    gdouble rise = st_theme_node_get_length(theme_node, "-arrow-rise");

    if (animate)
    {
        clutter_actor_set_opacity (actor, 0);
        clutter_actor_show (actor);

        if (rise > 0.0)
        {
            switch (priv->arrow_side)
            {
                case ST_SIDE_TOP:
                    clutter_actor_set_translation (actor, 0, -rise, 0);
                    break;
                case ST_SIDE_BOTTOM:
                    clutter_actor_set_translation (actor, 0, rise, 0);
                    break;
                case ST_SIDE_LEFT:
                    clutter_actor_set_translation (actor, -rise, 0, 0);
                    break;
                case ST_SIDE_RIGHT:
                    clutter_actor_set_translation (actor, rise, 0, 0);
                    break;
            }
        }
        cdos_transition_simple (actor,
                                CLUTTER_LINEAR,
                                150,
                                "opacity", 255,
                                "translation-x", 0,
                                "translation-y", 0,
                                NULL);
    }
    else
    {
        clutter_actor_set_opacity (actor, 255);
        clutter_actor_show (actor);
    }
}

/**
 * cdos_box_pointer_hide:
 * @self: 当前实例对象
 * @animate: 是否使用过渡动画，TRUE-使用，FALSE-不使用
 *
 * 隐藏当前实例主容器元素。
 */
void
cdos_box_pointer_hide (CdosBoxPointer *self, gboolean animate)
{
    /* TODO: 动画 */
    g_return_if_fail (CDOS_IS_BOX_POINTER (self));
    clutter_actor_hide (CLUTTER_ACTOR (self));
}

/**
 * cdos_box_pointer_set_arrow_side:
 * @self: 当前实例对象；
 * @side: 方位值，取值为StSide枚举成员
 *
 * 设置当前实例锚点所在方位；
 */
void
cdos_box_pointer_set_arrow_side (CdosBoxPointer *self, StSide side)
{
    g_return_if_fail (CDOS_IS_BOX_POINTER (self));
    CdosBoxPointerPrivate *priv;
    priv = self->priv;
    g_return_if_fail (priv->is_popup == TRUE);

    priv->arrow_side = side;
    clutter_actor_queue_relayout (CLUTTER_ACTOR (self));
}

/**
 *  cdos_box_pointer_set_position:
 *  @self: 当前实例对象
 *  @source_actor: 给定元素
 *  @alignment: 对齐方式
 *
 *  根据给定元素和对齐方式，重置当前实例元素相对给定元素的位置。
 */
gboolean
cdos_box_pointer_set_position (CdosBoxPointer *self, ClutterActor *source_actor, gfloat alignment)
{
    g_return_val_if_fail (CDOS_IS_BOX_POINTER (self), FALSE);
    // We need to show it now to force an allocation,
    // so that we can query the correct size.
    CdosBoxPointerPrivate *priv;
    priv = self->priv;
    g_return_val_if_fail (priv->is_popup == TRUE, FALSE);

    clutter_actor_show (CLUTTER_ACTOR (self));
    priv->source_actor = source_actor;
    priv->arrow_alignment = alignment;
    clutter_actor_queue_relayout (CLUTTER_ACTOR (self));
    return TRUE;
}

/**
 *  cdos_box_pointer_set_position_x_y:
 *  @self: 当前实例对象
 *  @source_actor: 给定元素
 *  @alignment: 对齐方式
 *  @x: x方向的位置偏移量
 *  @y: y方向的位置偏移量
 *
 *  根据给定元素，对齐方式和相对偏移量，重置当前实例元素和给定元素之间的位置。
 */
void
cdos_box_pointer_set_position_x_y (CdosBoxPointer *self,
                                   ClutterActor *source_actor,
                                   gfloat alignment,
                                   gfloat x,
                                   gfloat y)
{
    g_return_if_fail (CDOS_IS_BOX_POINTER (self));
    // We need to show it now to force an allocation,
    // so that we can query the correct size.
    CdosBoxPointerPrivate *priv;
    priv = self->priv;
    g_return_if_fail (priv->is_popup == TRUE);

    clutter_actor_show (CLUTTER_ACTOR (self));
    priv->source_actor = source_actor;
    if (!priv->pos)
        priv->pos = g_new0 (graphene_point_t, 1);
    priv->pos->x = x;
    priv->pos->y = y;
    priv->source_actor = source_actor;
    priv->arrow_alignment = alignment;
    clutter_actor_queue_relayout (CLUTTER_ACTOR (self));
}

/**
 * cdos_box_pointer_set_source_alignment:
 * @self:当前实例对象
 * @alignmentz:对其方式
 *
 * 根据给定的对其方式，重置当前实例元素和关联元素之间的位置。
 */
void
cdos_box_pointer_set_source_alignment (CdosBoxPointer *self, gfloat alignment)
{
    g_return_if_fail (CDOS_IS_BOX_POINTER (self));
    CdosBoxPointerPrivate *priv;
    priv = self->priv;
    g_return_if_fail (priv->is_popup == TRUE);

    priv->source_alignment = alignment;
    if (!priv->source_actor)
        return;
    cdos_box_pointer_set_position (self, priv->source_actor, priv->arrow_alignment);
}

/**
 * cdos_box_pointer_set_arrow_origin:
 * @self: 当前实例对象
 * @origin: 原点位置
 *
 * 统一设置原点所在的偏移距离
 */
void
cdos_box_pointer_set_arrow_origin (CdosBoxPointer *self, gfloat origin)
{
    g_return_if_fail (CDOS_IS_BOX_POINTER (self));
    CdosBoxPointerPrivate *priv;
    priv = self->priv;
    g_return_if_fail (priv->is_popup == TRUE);

    if (priv->arrow_origin != origin)
    {
        priv->arrow_origin = origin;
        st_drawing_area_queue_repaint (ST_DRAWING_AREA (self->border));
    }
}
