#include "cdos-coverflow-switcher.h"

/**
 * SECTION: cdos-coverflow-switcher
 * @Title: CoverflowSwitcher
 * @Short_description: 3d 模式切换应用程序
 *
 * 在应用程序及工作区间进行切换，关闭窗口、显示桌面
 */

#define TRANSITION_TYPE CLUTTER_EASE_OUT_QUAD
#define SIDE_ANGLE 60
#define BLEND_OUT_ANGLE 30

struct _CoverflowSwitcher
{
    AppSwitcher3D parent;
    gfloat y_offset;
    gfloat x_offset_left;
    gfloat x_offset_right;
    gfloat x_offset_center;

    gboolean looping;
    gboolean require_update;
};

/* Virtual function */
static void coverflow_switcher_select_next (AppSwitcher *switcher);
static void coverflow_switcher_select_previous (AppSwitcher *switcher);
static void coverflow_switcher_update_list (AppSwitcher *switcher, int index);
/*    end  */

typedef struct _FlipData FlipData;
typedef void (*FlipCompleteFunc) (ClutterTimeline *tml, FlipData *flip);

struct _FlipData
{
    CoverflowSwitcher *self;
    ClutterActor *preview;
    gint index;
    gint direction; /* -1: left; 1: right */
    gboolean cf_is_last;

    gfloat x_offset;
    guint8 opacity;
    gint rotation_angle_y;
    guint animation_time;
    ClutterAnimationMode transition;

    FlipCompleteFunc completed;

    guint32 flags;
};

#define FLIP_INDEX 1
#define FLIP_DIRECTION (1 << 1)
#define FLIP_CF_IS_LAST (1 << 2)
#define FLIP_X_OFFSET (1 << 3)
#define FLIP_OPACITY (1 << 4)
#define FLIP_ROTATION_ANGLE_Y (1 << 5)
#define FLIP_ANIMATION_TIME (1 << 6)
#define FLIP_TRANSITION (1 << 7)
#define FLIP_COMPLETED (1 << 8)

#define FLIP_SET_INDEX(f,v) f->index=(v);f->flags|=FLIP_INDEX
#define FLIP_SET_DIRECTION(f,v) f->direction=(v);f->flags|=FLIP_DIRECTION
#define FLIP_SET_CF_IS_LAST(f,v) f->cf_is_last=(v);f->flags|=FLIP_CF_IS_LAST
#define FLIP_SET_X_OFFSET(f,v) f->x_offset=(v);f->flags|=FLIP_X_OFFSET
#define FLIP_SET_OPACITY(f,v) f->opacity=(v);f->flags|=FLIP_OPACITY
#define FLIP_SET_ROTATION_ANGLE_Y(f,v) f->rotation_angle_y=(v);f->flags|=FLIP_ROTATION_ANGLE_Y
#define FLIP_SET_ANIMATION_TIME(f,v) f->animation_time=(v);f->flags|=FLIP_ANIMATION_TIME
#define FLIP_SET_TRANSITION(f,v) f->transition=(v);f->flags|=FLIP_TRANSITION
#define FLIP_SET_COMPLETED(f,v) f->completed=(v);f->flags|=FLIP_COMPLETED

static void coverflow_switcher_flip_stack (CoverflowSwitcher *self, int direction);
static void coverflow_switcher_on_flip_in (ClutterTimeline *tml, FlipData *flip);
static void coverflow_switcher_on_flip_complete (ClutterTimeline *tml, FlipData *flip);

/**
 * cdos_coverflow_switcher_new:
 * @binding: MetaKeyBinding
 *
 * 创建一个 CoverflowSwitcher.
 *
 * @Returns: CoverflowSwitcher.
 */
AppSwitcher *
cdos_coverflow_switcher_new (MetaKeyBinding *binding)
{
    CoverflowSwitcher *self = g_new0 (CoverflowSwitcher, 1);
    AppSwitcher3D *parent = (AppSwitcher3D *)self;
    AppSwitcher *switcher = (AppSwitcher *)self;

    cdos_appswitcher3d_init (parent, binding);
    switcher->select_next = coverflow_switcher_select_next;
    switcher->select_previous = coverflow_switcher_select_previous;
    switcher->update_list = coverflow_switcher_update_list;

    const MetaRectangle *monitor = switcher->active_monitor;
    if (monitor)
    {
        self->y_offset = monitor->height / 2.0;
        self->x_offset_left = monitor->width * 0.1;
        self->x_offset_right = monitor->width - self->x_offset_left;
        self->x_offset_center = monitor->width / 2.0;
    }

    /* setup_modal() 由 cdos_appswitcher3d_init() 移到这里，
     * offset 设置完成后再显示动画
     */
    appswitcher_setup_modal (switcher);
    return switcher;
}

static void
coverflow_switcher_select_next (AppSwitcher *switcher)
{
    CoverflowSwitcher *self = (CoverflowSwitcher *)switcher;

    if (switcher->current_index->next == NULL)
    {
        switcher->current_index = switcher->windows;
        coverflow_switcher_flip_stack (self, -1);
    }
    else
    {
        switcher->current_index = switcher->current_index->next;
        coverflow_switcher_update_list (switcher, 0);
    }
}

static void
coverflow_switcher_select_previous (AppSwitcher *switcher)
{
    CoverflowSwitcher *self = (CoverflowSwitcher *)switcher;

    if (switcher->current_index == switcher->windows)
    {
        switcher->current_index = g_list_last (switcher->windows);
        coverflow_switcher_flip_stack (self, 1);
    }
    else
    {
        switcher->current_index = switcher->current_index->prev;
        coverflow_switcher_update_list (switcher, 0);
    }
}

static void animate_preview_to_side (FlipData *flip);
static void animate_preview_to_mid (FlipData *flip);

static void
coverflow_switcher_flip_stack (CoverflowSwitcher *self, int direction)
{
    // cdos_debug ();
    AppSwitcher *switcher = (AppSwitcher *)self;

    self->looping = 1;

    gfloat x_offset = 0;
    gint angle = 0;

    if (direction == -1)
    {
        x_offset = -self->x_offset_left;
        angle = BLEND_OUT_ANGLE;
    }
    else if (switcher->active_monitor)
    {
        x_offset = switcher->active_monitor->width + self->x_offset_left;
        angle = -BLEND_OUT_ANGLE;
    }
#define ANIMATION_TIME 250 // miscellaneousseconds
    guint animation_time = ANIMATION_TIME * 2 / 3;

    GList *iter = NULL;
    gint i = 0;
    for (iter = switcher->previews; iter; iter = iter->next)
    {
        ClutterActor *preview = iter->data;
        FlipData *flip = g_new0 (FlipData, 1);

        flip->self = self;
        flip->preview = preview;
        /*
        flip->opacity = 0;
        flip->rotation_angle_y = angle;
        flip->animation_time = animation_time;
        flip->transition = TRANSITION_TYPE;
        flip->completed = (FlipCompleteFunc)coverflow_switcher_on_flip_in;
        flip->cf_is_last = (iter->next == NULL);
        */
        FLIP_SET_DIRECTION (flip, direction);
        FLIP_SET_INDEX (flip, i);
        FLIP_SET_X_OFFSET (flip, x_offset);
        FLIP_SET_ROTATION_ANGLE_Y (flip, angle);
        FLIP_SET_OPACITY (flip, 0);
        FLIP_SET_ANIMATION_TIME (flip, animation_time);
        FLIP_SET_TRANSITION (flip, TRANSITION_TYPE);
        FLIP_SET_CF_IS_LAST (flip, (iter->next == NULL));
        FLIP_SET_COMPLETED (flip, (FlipCompleteFunc)coverflow_switcher_on_flip_in);

        animate_preview_to_side (flip);

        i++;
    }
}

static void
coverflow_switcher_on_flip_in (ClutterTimeline *tml, FlipData *flip)
{
    // cdos_debug ();
    clutter_actor_remove_all_transitions (flip->preview);

    CoverflowSwitcher *self = flip->self;
    AppSwitcher3D *parent = (AppSwitcher3D *)self;
    AppSwitcher *switcher = (AppSwitcher *)self;

    gfloat x_offset_start, x_offset_end;
    gdouble angle_start, angle_end;

    if (flip->direction == -1)
    {
        /* left -> right
         *
         *  / -后-> \
         *  \ <-前- /
         * */
        if (switcher->active_monitor)
            x_offset_start = switcher->active_monitor->width + self->x_offset_left;
        else
            x_offset_start = self->x_offset_left;
        x_offset_end = self->x_offset_right;
        angle_start = -BLEND_OUT_ANGLE;
        angle_end = -SIDE_ANGLE;
    }
    else
    {
        x_offset_start = -self->x_offset_right;
        x_offset_end = self->x_offset_left;
        angle_start = BLEND_OUT_ANGLE;
        angle_end = SIDE_ANGLE;
    }
    guint animation_time = ANIMATION_TIME * 2 / 3;

    clutter_actor_set_rotation_angle (flip->preview, CLUTTER_Y_AXIS, angle_start);

    gint current_index = g_list_position (switcher->windows, switcher->current_index);
    clutter_actor_set_x (flip->preview, x_offset_start + 50 * (flip->index - current_index));
    // flip->completed = coverflow_switcher_on_flip_complete;

    gint opposite_direction = -flip->direction;

    flip->flags = 0;

    if (flip->index == current_index)
    {
        clutter_actor_set_child_above_sibling (parent->preview_actor, flip->preview, NULL);
        if (flip->cf_is_last != 0)
        {
            FLIP_SET_COMPLETED (flip, coverflow_switcher_on_flip_complete);
        }
        /*
        flip->direction = opposite_direction;
        flip->animation_time = animation_time;
        */
        FLIP_SET_DIRECTION (flip, opposite_direction);
        FLIP_SET_ANIMATION_TIME (flip, animation_time);
        animate_preview_to_mid (flip);
    }
    else
    {
        if (flip->direction == -1)
            clutter_actor_set_child_above_sibling (parent->preview_actor, flip->preview, NULL);
        else
            clutter_actor_set_child_below_sibling (parent->preview_actor, flip->preview, NULL);
        /*
        flip->opacity = 255;
        flip->rotation_angle_y = angle_end;
        flip->animation_time = animation_time;
        flip->transition = TRANSITION_TYPE;
        if (flip->cf_is_last)
            flip->completed = coverflow_switcher_on_flip_complete;

        flip->direction = opposite_direction;
        flip->x_offset = x_offset_end;
        */
        FLIP_SET_OPACITY (flip, 255);
        FLIP_SET_ROTATION_ANGLE_Y (flip, angle_end);
        FLIP_SET_ANIMATION_TIME (flip, animation_time);
        FLIP_SET_TRANSITION (flip, TRANSITION_TYPE);

        if (flip->cf_is_last)
        {
            FLIP_SET_COMPLETED (flip, coverflow_switcher_on_flip_complete);
        }
        FLIP_SET_DIRECTION (flip, opposite_direction);
        FLIP_SET_X_OFFSET (flip, x_offset_end);

        animate_preview_to_side (flip);
    }
}

static void
coverflow_switcher_on_flip_complete (ClutterTimeline *tml, FlipData *flip)
{
    CoverflowSwitcher *self = flip->self;
    AppSwitcher *switcher = (AppSwitcher *)self;

    self->looping = 0;
    // if (flip->completed != NULL && flip->flags & FLIP_COMPLETED)
    {
        g_free (flip);
    }
    if (self->require_update == 1)
    {
        self->require_update = 0;
        coverflow_switcher_update_list (switcher, 0);
    }
}

static void
remove_transition (ClutterTimeline *tml, FlipData *flip)
{
    g_free (flip);
}

static void
coverflow_switcher_update_list (AppSwitcher *switcher, int index)
{
    CoverflowSwitcher *self = (CoverflowSwitcher *)switcher;
    AppSwitcher3D *parent = (AppSwitcher3D *)self;

    if (self->looping == 1)
    {
        self->require_update = 1;
        return;
    }

    // MetaRectangle *monitor = switcher->active_monitor;
    guint animation_time = ANIMATION_TIME;

    GList *iter = NULL;
    gint i = 0;
    gint current_index = g_list_position (switcher->windows, switcher->current_index);
    for (iter = switcher->previews; iter; iter = iter->next)
    {
        ClutterActor *preview = iter->data;

        FlipData *flip = g_new0 (FlipData, 1);
        flip->self = self;
        flip->preview = preview;

        if (i == current_index)
        {
            FLIP_SET_ANIMATION_TIME (flip, animation_time);
            animate_preview_to_mid (flip);
        }
        else if (i < current_index)
        {
            clutter_actor_set_child_above_sibling (parent->preview_actor, preview, NULL);

            FLIP_SET_INDEX (flip, i);
            FLIP_SET_DIRECTION (flip, -1);
            FLIP_SET_X_OFFSET (flip, self->x_offset_left);
            FLIP_SET_OPACITY (flip, 255);
            FLIP_SET_ROTATION_ANGLE_Y (flip, SIDE_ANGLE);
            FLIP_SET_ANIMATION_TIME (flip, animation_time);
            FLIP_SET_TRANSITION (flip, TRANSITION_TYPE);

            animate_preview_to_side (flip);
        }
        else if (i > current_index)
        {
            clutter_actor_set_child_below_sibling (parent->preview_actor, preview, NULL);
            FLIP_SET_INDEX (flip, i);
            FLIP_SET_DIRECTION (flip, 1);
            FLIP_SET_X_OFFSET (flip, self->x_offset_right);
            FLIP_SET_OPACITY (flip, 255);
            FLIP_SET_ROTATION_ANGLE_Y (flip, -SIDE_ANGLE);
            FLIP_SET_ANIMATION_TIME (flip, animation_time);
            FLIP_SET_TRANSITION (flip, TRANSITION_TYPE);

            animate_preview_to_side (flip);
        }
        i++;
    }
}

static void
animate_preview_to_side (FlipData *flip)
{
    CoverflowSwitcher *self = flip->self;
    AppSwitcher *switcher = (AppSwitcher *)self;
    AppSwitcher3D *parent = (AppSwitcher3D *)self;
    ClutterActor *preview = flip->preview;

    if (flip->direction == -1)
        clutter_actor_set_pivot_point (preview, 0.0, 0.5);
    else if (flip->direction == 1)
        clutter_actor_set_pivot_point (preview, 1.0, 0.5);

    {
        gint current_index = g_list_position (switcher->windows, switcher->current_index);

        gfloat x = flip->x_offset + 50 * (flip->index - current_index);
        gfloat y = self->y_offset;
        gfloat width, height;
        PreviewData *data = g_hash_table_lookup (parent->preview_table, preview);
        width = MAX (data->target_width_side * (10 - abs (flip->index - current_index)) / 10, 0);
        height = MAX (data->target_height_side * (10 - abs (flip->index - current_index)) / 10, 0);

        if (flip->direction == 1)
            x = x - width;
        y -= height / 2.0;

        guint time = flip->animation_time;
        gdouble rotation_angle_y = flip->rotation_angle_y;
        guint8 opacity = flip->opacity;
        ClutterAnimationMode mode = flip->transition;
        ClutterTransition *transition = NULL;
        transition = cdos_transition_simple (preview,
                                             mode,
                                             time,
                                             "x", x,
                                             "y", y,
                                             "opacity", opacity,
                                             "width", width,
                                             "height", height,
                                             "rotation-angle-y", rotation_angle_y,
                                             NULL);
        if (flip->completed != NULL && (flip->flags & FLIP_COMPLETED))
            g_signal_connect (transition, "completed", G_CALLBACK (flip->completed), flip);
        else
        {
            g_signal_connect (transition, "completed", G_CALLBACK (remove_transition), flip);
        }
    }
}

static void
animate_preview_to_mid (FlipData *flip)
{
    CoverflowSwitcher *self = flip->self;
    // AppSwitcher *switcher = (AppSwitcher *)self;
    AppSwitcher3D *parent = (AppSwitcher3D *)self;
    ClutterActor *preview = flip->preview;

    // gfloat pwidth = clutter_actor_get_width (preview);
    // gfloat rotation_vertex_x = (flip->direction == 1) ? width / 2 : -width / 2;

    clutter_actor_set_pivot_point (preview, 0.5, 0.5);
    clutter_actor_set_child_above_sibling (parent->preview_actor, preview, NULL);
    {
        guint8 opacity = 255;
        gfloat x = self->x_offset_center;
        gfloat y = self->y_offset;
        guint time = flip->animation_time;
        ClutterAnimationMode mode = TRANSITION_TYPE;
        gfloat width, height;
        PreviewData *data = g_hash_table_lookup (parent->preview_table, preview);
        g_return_if_fail (data != NULL);
        width = data->target_width;
        height = data->target_height;

        x -= width / 2.0;
        y -= height / 2.0;
        gdouble rotation_angle_y = 0.0;

        ClutterTransition *transition = NULL;
        transition = cdos_transition_simple (preview,
                                             mode,
                                             time,
                                             "x", x,
                                             "y", y,
                                             "opacity", opacity,
                                             "width", width,
                                             "height", height,
                                             "rotation-angle-y", rotation_angle_y,
                                             NULL);
        if (flip->completed != NULL && (flip->flags & FLIP_COMPLETED))
            g_signal_connect (transition, "completed", G_CALLBACK (flip->completed), flip);
        else
        {
            g_signal_connect (transition, "completed", G_CALLBACK (remove_transition), flip);
        }
    }
}
