#include "cdos-doc-manager.h"
#include "cdos-doc-system.h"

/**
 * SECTION: cdos-doc-manager
 * @Title: CdosDocManager
 * @short_description: DocManager围绕着DocSystem，主要用于公开DocInfo对象
 * @include: cdos-doc-manager.h
 *
 * CdosDocManager 用来管理文档信息对象
 */

struct _CdosDocManagerPrivate
{
    CdosDocSystem *doc_system;
    GList *infos_by_timestamp;
    GHashTable *infos_by_uri;
};

enum
{
    CHANGED,

    LAST_SIGNAL
};

static guint cdos_doc_manager_signals[LAST_SIGNAL] = {
    0,
};

G_DEFINE_TYPE_WITH_PRIVATE (CdosDocManager, cdos_doc_manager, G_TYPE_OBJECT);

static void cdos_doc_manager_finalize (GObject *gobject);
static void cdos_doc_manager_dispose (GObject *gobject);

static void cdos_doc_manager_reload (CdosDocSystem *doc, CdosDocManager *doc_manager);

static void
cdos_doc_manager_init (CdosDocManager *manager)
{
    CdosDocManagerPrivate *priv;
    priv = manager->priv = cdos_doc_manager_get_instance_private (manager);

    priv->doc_system = cdos_doc_system_get_default ();

    priv->infos_by_timestamp = NULL;
    priv->infos_by_uri = g_hash_table_new_full (g_str_hash, NULL, g_free, g_object_unref);

    g_signal_connect (manager->priv->doc_system, "changed", G_CALLBACK (cdos_doc_manager_reload), manager);
    cdos_doc_manager_reload (NULL, manager);
}

static void
cdos_doc_manager_class_init (CdosDocManagerClass *klass)
{
    GObjectClass *gobject_class = G_OBJECT_CLASS (klass);

    gobject_class->finalize = cdos_doc_manager_finalize;
    gobject_class->dispose = cdos_doc_manager_dispose;

    cdos_doc_manager_signals[CHANGED] = g_signal_new ("changed",
                                                      G_TYPE_FROM_CLASS (klass),
                                                      G_SIGNAL_RUN_LAST,
                                                      G_STRUCT_OFFSET (CdosDocManagerClass, changed),
                                                      NULL,
                                                      NULL,
                                                      g_cclosure_marshal_VOID__VOID,
                                                      G_TYPE_NONE,
                                                      0);
}

static void
cdos_doc_manager_dispose (GObject *gobject)
{
    CdosDocManagerPrivate *priv = CDOS_DOC_MANAGER (gobject)->priv;
    g_clear_object (&priv->doc_system);
    if (priv->infos_by_timestamp)
    {
        g_list_free_full (priv->infos_by_timestamp, (GDestroyNotify)g_object_unref);
        priv->infos_by_timestamp = NULL;
    }

    if (priv->infos_by_uri)
    {
        g_hash_table_destroy (priv->infos_by_uri);
        priv->infos_by_uri = NULL;
    }

    G_OBJECT_CLASS (cdos_doc_manager_parent_class)->dispose (gobject);
}

static void
cdos_doc_manager_finalize (GObject *gobject)
{
    // CdosDocManagerPrivate *priv = CDOS_DOC_MANAGER(gobject)->priv;

    G_OBJECT_CLASS (cdos_doc_manager_parent_class)->finalize (gobject);
}

static void
cdos_doc_manager_reload (CdosDocSystem *doc, CdosDocManager *manager)
{
    GSList *l;
    GtkRecentInfo *recent_info;
    CdosDocInfo *doc_info;

    CdosDocManagerPrivate *priv = manager->priv;

    GSList *docs = cdos_doc_system_get_all (priv->doc_system);

    if (priv->infos_by_timestamp)
    {
        g_list_free (priv->infos_by_timestamp);
        priv->infos_by_timestamp = NULL;
    }

    if (priv->infos_by_uri)
    {
        g_hash_table_remove_all (priv->infos_by_uri);
    }

    for (l = docs; l != NULL; l = l->next)
    {
        recent_info = l->data;
        doc_info = cdos_doc_info_new (recent_info);
        priv->infos_by_timestamp = g_list_append (priv->infos_by_timestamp, doc_info);
        gchar *url = g_strdup (cdos_doc_info_get_uri (doc_info));
        g_hash_table_insert (priv->infos_by_uri, url, doc_info);
    }

    g_signal_emit_by_name (manager, "changed");
}

/**
 * cdos_doc_manager_get_time_stamp_ordered_infos
 * @manager: #CdosDocManager对象
 *
 * 获取以时间戳为顺序的文档信息
 *
 * Returns: GPtrArray信息数组
 */

GList *
cdos_doc_manager_get_time_stamp_ordered_infos (CdosDocManager *manager)
{
    g_return_val_if_fail (CDOS_IS_DOC_MANAGER (manager), NULL);
    return manager->priv->infos_by_timestamp;
}

/**
 * cdos_doc_manager_get_infos_by_uri
 * @manager: #CdosDocManager对象
 *
 * 获取#CdosDocManager的私有属性infos_by_uri
 *
 * Returns: 以哈希表的形式返回infos_by_uri
 */

GHashTable *
cdos_doc_manager_get_infos_by_uri (CdosDocManager *manager)
{
    g_return_val_if_fail (CDOS_IS_DOC_MANAGER (manager), NULL);
    return manager->priv->infos_by_uri;
}

/**
 * cdos_doc_manager_lookup_by_uri
 * @manager: #CdosDocManager对象
 * @uri: 需要查找的uri
 *
 * 根据uri查找对应的#CdosDocInfo
 *
 * Returns: 查找到的文档信息对象
 */

CdosDocInfo *
cdos_doc_manager_lookup_by_uri (CdosDocManager *manager, const gchar *uri)
{
    g_return_val_if_fail (CDOS_IS_DOC_MANAGER (manager), NULL);
    g_return_val_if_fail (uri != NULL, NULL);

    return g_hash_table_lookup (manager->priv->infos_by_uri, uri);
}

/**
 * cdos_doc_manager_get_default
 *
 * 获取#CdosDocManager对象
 *
 * Returns: #CdosDocManager对象
 */

CdosDocManager *
cdos_doc_manager_get_default (void)
{
    static CdosDocManager *manager = NULL;

    if (manager == NULL)
    {
        manager = CDOS_DOC_MANAGER (g_object_new (CDOS_TYPE_DOC_MANAGER, NULL));
        g_object_add_weak_pointer (G_OBJECT (manager), (gpointer *)&manager);
    }

    return manager;
}
