#include "config.h"
#include <gio/gio.h>
#include <time.h>

#include "cdos-applet.h"
#include "cdos-extension.h"
#include "applets/applets.h"

#include "cdos-applet-manager.h"
#include "cdos-showdesktop.h"

#define SYSTEM_PATH_LEN  50

G_DEFINE_TYPE (CdosExtension, cdos_extension, G_TYPE_OBJECT);

enum
{
    STATE_INITIALIZING = 0,
    STATE_LOADED,
    STATE_ERROR,
    STATE_OUT_OF_DATE,
    STATES_NUM
};

enum
{
    STATUS_MENU_LOADED,
    LAST_SIGNAL
};

static guint cdos_extension_signals[LAST_SIGNAL] = {0};
gboolean init_flag = FALSE;

/* Maps uuid -> Extension object */
static GHashTable *objects = NULL;

typedef struct _StrExtension
{
    gchar *key;
    CdosExtension *value;
} StrExtension;

static void cdos_extension_set_module (CdosExtension *extension, gint module);

/**
 * cdos_extension_get_object_by_uuid:
 * @uuid: uuid of extension or applet
 *
 * get object by uuid like that objects[uuid] from js
 *
 * Returns: a instance of #CdosExtension
 */
CdosExtension *
cdos_extension_get_object_by_uuid (const gchar *uuid)
{
    if (objects == NULL)
    {
        objects = g_hash_table_new (g_str_hash, g_str_equal);
        return NULL;
    }
    if (uuid != NULL)
    {
        return g_hash_table_lookup (objects, uuid);
    }
    return NULL;
}

static void
cdos_extension_finalize (GObject *object)
{
  CdosExtension *extension = CDOS_EXTENSION(object);
  g_free (extension->uuid);
  g_free (extension->style_sheet);

  G_OBJECT_CLASS (cdos_extension_parent_class)->finalize (object);
}
static void
cdos_extension_dispose (GObject *object)
{
    CdosExtension *extension = CDOS_EXTENSION (object);

    if (extension->loaded_definitions != NULL)
    {
        g_hash_table_remove_all (extension->loaded_definitions);
        g_hash_table_unref (extension->loaded_definitions);
        extension->loaded_definitions = NULL;
    }

    G_OBJECT_CLASS (cdos_extension_parent_class)->dispose (object);
}
static void
cdos_extension_init (CdosExtension *extension)
{
    extension->start_time = time (NULL);
}

static void
cdos_extension_class_init (CdosExtensionClass *klass)
{
    GObjectClass *object_class = G_OBJECT_CLASS (klass);

    object_class->finalize = cdos_extension_finalize;
    object_class->dispose = cdos_extension_dispose;

    cdos_extension_signals[STATUS_MENU_LOADED] = g_signal_new ("status-menu-loaded",
                                                               G_TYPE_FROM_CLASS (klass),
                                                               G_SIGNAL_RUN_LAST,
                                                               G_STRUCT_OFFSET (CdosExtensionClass, status_menu_loaded),
                                                               NULL,
                                                               NULL,
                                                               g_cclosure_marshal_VOID__OBJECT,
                                                               G_TYPE_NONE,
                                                               1,
                                                               G_TYPE_OBJECT);
}

static void
cdos_extension_set_uuid (CdosExtension *extension, const gchar *uuid)
{
    if (uuid)
    {
        extension->uuid = g_strdup (uuid);
    }
}

// 这里根据applet的uuid去设置module, 决定创建具体的applet类型（cdos_extension_module_main创建具体的applet）
// 后面可能将直接取消module, 用uuid代替其起到的作用
/* Cinnamon applets have a uuid property; this key applet which should be loaded.*/
static void
cdos_extension_set_module_by_uuid (CdosExtension *extension, const gchar *uuid)
{
    if (extension && uuid)
    {
        /* default settings key list */
        if (g_strcmp0 (uuid, "win10-menu@cdos.org") == 0 || g_strcmp0 (uuid, "menu@cdos.org") == 0)
        {
            cdos_extension_set_module (extension, CDOS_EXTENSION_MODULE_MENU);
        }
        else if (g_strcmp0 (uuid, "scale@cdos.org") == 0 || g_strcmp0 (uuid, "cdos-scale@cdos.org") == 0)
        {
            cdos_extension_set_module (extension, CDOS_EXTENSION_MODULE_SCALE);
        }
        else if (g_strcmp0 (uuid, "window-list@cdos.org") == 0 || g_strcmp0 (uuid, "cdos-window-list@cdos.org") == 0)
        {
            cdos_extension_set_module (extension, CDOS_EXTENSION_MODULE_WINDOW_LIST);
        }
        else if (g_strcmp0 (uuid, "notifications@cdos.org") == 0
                 || g_strcmp0 (uuid, "cdos-notifications@cdos.org") == 0)
        {
            cdos_extension_set_module (extension, CDOS_EXTENSION_MODULE_NOTIFICATIONS);
        }
        else if (g_strcmp0 (uuid, "notifications2@cdos.org") == 0
                 || g_strcmp0 (uuid, "cdos-notifications2@cdos.org") == 0)
        {
            cdos_extension_set_module (extension, CDOS_EXTENSION_MODULE_NOTIFICATIONS2);
        }
        else if (g_strcmp0 (uuid, "workspace@cdos.org") == 0)
        {
            cdos_extension_set_module (extension, CDOS_EXTENSION_MODULE_WORKSPACE);
        }
        else if (g_strcmp0 (uuid, "removable-drives@cdos.org") == 0)
        {
            cdos_extension_set_module (extension, CDOS_EXTENSION_MODULE_REMOVABLE_DRIVES);
        }
        else if (g_strcmp0 (uuid, "keyboard@cdos.org") == 0)
        {
            cdos_extension_set_module (extension, CDOS_EXTENSION_MODULE_KEYBOARD);
        }
        else if (g_strcmp0 (uuid, "bluetooth@cdos.org") == 0)
        {
            cdos_extension_set_module (extension, CDOS_EXTENSION_MODULE_BLUETOOTH);
        }
        else if (g_strcmp0 (uuid, "network@cdos.org") == 0 || g_strcmp0 (uuid, "cdos-network@cdos.org") == 0)
        {
            cdos_extension_set_module (extension, CDOS_EXTENSION_MODULE_NETWORK);
        }
        else if (g_strcmp0 (uuid, "sound@cdos.org") == 0 || g_strcmp0 (uuid, "cdos-sound@cdos.org") == 0)
        {
            cdos_extension_set_module (extension, CDOS_EXTENSION_MODULE_SOUND);
        }
        else if (g_strcmp0 (uuid, "power@cdos.org") == 0)
        {
            cdos_extension_set_module (extension, CDOS_EXTENSION_MODULE_POWER);
        }
        else if (g_strcmp0 (uuid, "systray@cdos.org") == 0)
        {
            cdos_extension_set_module (extension, CDOS_EXTENSION_MODULE_SYSTRAY);
        }
        else if (g_strcmp0 (uuid, "calendar@cdos.org") == 0 || g_strcmp0 (uuid, "cdos-calendar@cdos.org") == 0)
        {
            cdos_extension_set_module (extension, CDOS_EXTENSION_MODULE_CALENDAR);
        }
        else if (g_strcmp0 (uuid, "show-desktop@cdos.org") == 0 || g_strcmp0 (uuid, "cdos-show-desktop@cdos.org") == 0)
        {
            cdos_extension_set_module (extension, CDOS_EXTENSION_MODULE_SHOW_DESKTOP);
        }
        else if (g_strcmp0 (uuid, "grouped-applets@cdos.org") == 0)
        {
            cdos_extension_set_module (extension, CDOS_EXTENSION_MODULE_GROUPED_APPLETS);
        }
        else if (g_strcmp0 (uuid, "settings@cdos.org") == 0)
        {
            cdos_extension_set_module (extension, CDOS_EXTENSION_MODULE_SETTINGS);
        }
        else if (g_strcmp0 (uuid, "feedback@cdos.org") == 0)
        {
            cdos_extension_set_module (extension, CDOS_EXTENSION_MODULE_FEEDBACK);
        }
    }
}

static void
cdos_extension_init_private (void)
{
  if (objects == NULL)
    objects = g_hash_table_new (g_str_hash, g_str_equal);

  init_flag = TRUE;
}

/**
 * cdos_extension_new:
 * @dir: GFile * dir
 *
 * 新建一个 #CdosExtension 类的实例。
 *
 * Returns: 返回一个 #CdosExtension 的实例。
 */
CdosExtension *
cdos_extension_new (const gchar *uuid)
{
  if (uuid)
  {
    CdosExtension *extension;

    if (!init_flag)
    {
      cdos_extension_init_private ();
    }

    extension = g_object_new (CDOS_TYPE_EXTENSION, NULL);

    cdos_extension_set_uuid (extension, uuid);

    /* set extension module by uuid*/
    cdos_extension_set_module_by_uuid (extension, uuid);

    if (objects == NULL)
      objects = g_hash_table_new (g_str_hash, g_str_equal);

    extension->loaded_menu = FALSE;
    g_hash_table_insert (objects, extension->uuid, extension);

    return extension;
  }
  return NULL;
}

/**
 * cdos_extension_load_extension:
 * @uuid: gchar * uuid
 *
 * load extension by uuid
 *
 * Returns: 返回一个 #CdosExtension 的实例。
 */
CdosExtension *
cdos_extension_load_extension (const gchar *uuid)
{
    g_return_val_if_fail (uuid, NULL);
    CdosExtension *extension = cdos_extension_get_object_by_uuid (uuid);
    if (!extension)
    {
        extension = cdos_extension_new (uuid);

        if (!cdos_applet_manager_finish_extension_load (extension))
        {
            return NULL;
        }
    }

    return extension;
}

/**
 * cdos_extension_unload_extension:
 * @uuid: gchar * uuid
 *
 * unload extension
 *
 */
void
cdos_extension_unload_extension (const gchar *uuid)
{
}

/**
 * cdos_extension_module_main:
 * @extension: A #CdosExtension
 * @orientation: StSide orientation
 * @panel_height: gfloat panel_height
 * @applet_id: gchar * applet_id
 *
 * create a #CdosApplet instance
 *
 * Returns: a #CdosApplet instance.
 */
CdosApplet *
cdos_extension_module_main (CdosExtension *extension,
                            StSide orientation,
                            gfloat panel_height,
                            gchar *applet_id)
{
    g_return_val_if_fail (CDOS_IS_EXTENSION (extension), NULL);

    gint module = extension->module;
    CdosApplet *applet = NULL;

#define TIME_START          \
    do                      \
    {                       \
        clock_t start, end; \
        start = clock ()

#define TIME_END(applet) end=clock();\
    g_debug (#applet" used time: %gms", (double)(end-start) / CLOCKS_PER_SEC * 1000);\
}while(0)

    switch (module)
    {
        case CDOS_EXTENSION_MODULE_MENU:
            TIME_START;
            applet = (CdosApplet *)applets_menu_main (NULL, orientation, panel_height);
            TIME_END(menu);
            break;
        case CDOS_EXTENSION_MODULE_SHOW_DESKTOP:
#ifdef HAVE_META_WORKSPACE_MANAGER_SHOW_DESKTOP
            TIME_START;
            applet = (CdosApplet *)cdos_showdesktop_main (NULL, orientation, panel_height);
            TIME_END(show-desktop);
#else
            g_debug ("HAVE_META_WORKSPACE_MANAGER_SHOW_DESKTOP=false");
#endif
            break;
        case CDOS_EXTENSION_MODULE_PANEL_LAUNCHERS:
            break;
        case CDOS_EXTENSION_MODULE_WINDOW_LIST:
            TIME_START;
            applet = (CdosApplet *)applets_window_list_main (NULL, orientation, panel_height, applet_id);
            TIME_END(window-list);
            break;
        case CDOS_EXTENSION_MODULE_NOTIFICATIONS:
            TIME_START;
            applet = (CdosApplet *)applet_notifications_main (NULL, orientation, panel_height, applet_id);
            TIME_END(notifications);
            break;
        case CDOS_EXTENSION_MODULE_NOTIFICATIONS2:
            TIME_START;
            applet = (CdosApplet *)applet_notifications2_main (NULL, orientation, panel_height, applet_id);
            TIME_END(notifications2);
            break;
        case CDOS_EXTENSION_MODULE_REMOVABLE_DRIVES:
            TIME_START;
            applet = (CdosApplet *)applet_removable_drives_main (NULL, orientation, panel_height);
            TIME_END(removable-drivers);
            break;
        case CDOS_EXTENSION_MODULE_KEYBOARD:
            TIME_START;
            //applet = (CdosApplet *) applet_keyboard_main(NULL,orientation,panel_height,applet_id);
            TIME_END(keyboard);
            break;
        case CDOS_EXTENSION_MODULE_BLUETOOTH:
            break;
        case CDOS_EXTENSION_MODULE_NETWORK:
            TIME_START;
            applet = (CdosApplet *)applet_network_main (NULL, orientation, (guint)panel_height);
            TIME_END(network);
            break;
        case CDOS_EXTENSION_MODULE_SOUND:
            applet = (CdosApplet *)applet_sound_main (NULL, orientation, panel_height, applet_id);
            break;
        case CDOS_EXTENSION_MODULE_POWER:
            TIME_START;
            applet = (CdosApplet *)applets_power_main (NULL, orientation, panel_height, applet_id);
            TIME_END(power);
            break;
        case CDOS_EXTENSION_MODULE_SYSTRAY:
            TIME_START;
            applet = (CdosApplet *)applets_systray_main (NULL, orientation, panel_height);
            TIME_END(systray);
            break;
        case CDOS_EXTENSION_MODULE_CALENDAR:
            TIME_START;
            applet = (CdosApplet *)applets_calendar_main (NULL, orientation, panel_height);
            TIME_END(calendar);
            break;
        case CDOS_EXTENSION_MODULE_SCALE:
            TIME_START;
            applet = (CdosApplet *)applet_scale_new (NULL, orientation, panel_height, applet_id);
            TIME_END(scale);
            break;
        case CDOS_EXTENSION_MODULE_GROUPED_APPLETS:
            TIME_START;
            applet = (CdosApplet *)applet_group_main (NULL, orientation, panel_height);
            TIME_END(group);
            break;
        case CDOS_EXTENSION_MODULE_FEEDBACK:
            TIME_START;
            applet = (CdosApplet *)applet_feedback_main (NULL, orientation, panel_height);
            TIME_END(group);
            break;
        default:
            break;
    }
    return applet;
}

/**
 * cdos_extension_get_dir:
 * @extension: A #CdosExtension
 *
 * get dir of extension
 *
 * Returns: dir
 */
GFile *
cdos_extension_get_dir (CdosExtension *extension)
{
    g_return_val_if_fail (CDOS_IS_EXTENSION (extension), NULL);
    return extension->dir;
}

/**
 * cdos_extension_get_uuid:
 * @extension: A #CdosExtension
 *
 * get uuid of extension
 *
 * Returns: uuid
 */
gchar *
cdos_extension_get_uuid (CdosExtension *extension)
{
    g_return_val_if_fail (CDOS_IS_EXTENSION (extension), NULL);
    return extension->uuid;
}

/**
 * cdos_extension_get_loaded_definitions:
 * @extension: A #CdosExtension
 *
 * get loaded definitions of extension
 *
 * Returns: loaded_definitions
 */
GHashTable *
cdos_extension_get_loaded_definitions (CdosExtension *extension)
{
    g_return_val_if_fail (CDOS_IS_EXTENSION (extension), NULL);
    return extension->loaded_definitions;
}

/**
 * cdos_extension_set_loaded_definitions:
 * @extension: A #CdosExtension
 *
 * set loaded definitions of extension
 *
 */
void
cdos_extension_set_loaded_definitions (CdosExtension *extension, GHashTable *loaded_definitions)
{
    g_return_if_fail (CDOS_IS_EXTENSION (extension));
    if (extension->loaded_definitions != NULL)
    {
        g_hash_table_remove_all (extension->loaded_definitions);
        g_hash_table_unref (extension->loaded_definitions);
    }
    extension->loaded_definitions = loaded_definitions;
}

/**
 * cdos_extension_set_module:
 * @extension: A #CdosExtension
 * @module: module of extension
 *
 * set module of extension
 *
 */
static void
cdos_extension_set_module (CdosExtension *extension, gint module)
{
    extension->module = module;
}

/**
 * cdos_extension_get_max_instances:
 * @uuid: uuid of applet
 *
 * get max instances
 *
 * Returns: number of max instances。
 */
gint
cdos_extension_get_max_instances (const gchar *uuid)
{
    return 1;
}
