/*
 * Copyright 2021 <cdos>
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms and conditions of the GNU Lesser General Public License,
 * version 2.1, as published by the Free Software Foundation.
 *
 * This program is distributed in the hope it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for
 * more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
/*
 * SECTION:  cdos-indicator-manager.c
 * @Title: cdos-indicator-manager
 * @Short_Description:  libappindicator and StatusNotifier support.
 */

#include "common.h"
#include "cdos-sn-watcher.h"
#include "cdos-indicator-manager.h"
#include "cdos-sn-item.h"

struct _CdosIndicatorManager
{
    GObject parent;

    guint bus_id;
    CdosStatusNotifierWatcher *sn_skeleton;
};

G_DEFINE_TYPE(CdosIndicatorManager, cdos_indicator_manager, G_TYPE_OBJECT)

static gboolean
cdos_indicator_manager_register_status_notifier_item (CdosStatusNotifierWatcher *skeleton,
                                                      GDBusMethodInvocation     *invocation,
                                                      const gchar               *service,
                                                      CdosIndicatorManager      *self)
{
    g_debug ("%s", __func__);
    const gchar *bus_name, *object_path;
    CdosSnItem *item G_GNUC_UNUSED;
    const gchar *sender = g_dbus_method_invocation_get_sender (invocation);

    if (service[0] == '/')
    {
        /* /org/ayatana/NotificationItem */
        bus_name = sender;
        object_path = service;
    }
    else
    {
        /* org.kde.StatusNotifierItem */
        bus_name = service;
        object_path = "/StatusNotifierItem";
    }
    if (!g_dbus_is_name (bus_name))
    {
        g_dbus_method_invocation_return_error_literal (invocation,
                                                       G_IO_ERROR,
                                                       G_IO_ERROR_INVALID_ARGUMENT,
                                                       "Invalid bus name");
        return FALSE;
    }
    g_debug ("%s %s", bus_name, object_path);
    item = cdos_sn_item_new (bus_name, object_path);

    cdos_status_notifier_watcher_complete_register_status_notifier_item (skeleton, invocation);
    return TRUE;
}

static gboolean
cdos_indicator_manager_register_status_notifier_host (CdosStatusNotifierWatcher *skeleton,
                                                      GDBusMethodInvocation     *invocation,
                                                      const char                *service,
                                                      CdosIndicatorManager *self)
{
    g_debug ("%s %s", __func__, service);
    return TRUE;
}

static void
on_bus_acquired (GDBusConnection *connection,
                 const gchar     *name,
                 CdosIndicatorManager *self)
{
    g_debug ("%s: %s", __func__, name);
    GError *error = NULL;
    self->sn_skeleton = cdos_status_notifier_watcher_skeleton_new ();
    g_signal_connect (self->sn_skeleton, "handle-register-status-notifier-item",
                      G_CALLBACK (cdos_indicator_manager_register_status_notifier_item),
                      self);
    g_signal_connect (self->sn_skeleton, "handle-register-status-notifier-host",
                      G_CALLBACK (cdos_indicator_manager_register_status_notifier_host),
                      self);
    cdos_status_notifier_watcher_set_protocol_version (self->sn_skeleton, 0);
    cdos_status_notifier_watcher_set_is_status_notifier_host_registered (self->sn_skeleton, TRUE);

    g_dbus_interface_skeleton_export (G_DBUS_INTERFACE_SKELETON (self->sn_skeleton),
                                      connection,
                                      "/StatusNotifierWatcher",
                                      &error);
    if (error)
    {
        g_warning ("%s", error->message);
        g_error_free (error);
        g_clear_object (&self->sn_skeleton);
        return;
    }
}

static void
on_name_acquired (GDBusConnection *connection,
                  const char      *name,
                  gpointer         user_data)
{
    g_debug ("%s: %s", __func__, name);
}

static void
on_name_lost (GDBusConnection *connection,
              const gchar     *name,
              gpointer         user_data)
{
    g_debug ("%s: %s", __func__, name);
}

static void
cdos_indicator_manager_init (CdosIndicatorManager *self)
{
    self->bus_id = g_bus_own_name (G_BUS_TYPE_SESSION,
                                   "org.kde.StatusNotifierWatcher",
                                   G_BUS_NAME_OWNER_FLAGS_REPLACE,
                                   (GBusAcquiredCallback)on_bus_acquired,
                                   on_name_acquired,
                                   on_name_lost,
                                   self,
                                   NULL);
}

static void
cdos_indicator_manager_dispose (GObject *gobject)
{
    CdosIndicatorManager *self = CDOS_INDICATOR_MANAGER (gobject);
    g_clear_object (&self->sn_skeleton);

    G_OBJECT_CLASS (cdos_indicator_manager_parent_class)->dispose (gobject);
}

static void
cdos_indicator_manager_class_init (CdosIndicatorManagerClass *klass)
{
    GObjectClass *gobject_class = G_OBJECT_CLASS (klass);
    gobject_class->dispose = cdos_indicator_manager_dispose;
}

CdosIndicatorManager *
cdos_indicator_manager_get_default (void)
{
    static CdosIndicatorManager *indicator_manager = NULL;
    if (!indicator_manager)
    {
        indicator_manager = g_object_new (CDOS_TYPE_INDICATOR_MANAGER, NULL);
        g_object_add_weak_pointer ((GObject *)indicator_manager, (gpointer *)&indicator_manager);
    }
    return indicator_manager;
}
