#include <meta/keybindings.h>
#include <meta/display.h>
#include <meta/util.h>

#include "cdos-keybindings.h"
#include "cdos-global.h"
#include "cdos-util.h"

#include "cdos-key-grabber.h"

static CdosActionMode action_mode = CDOS_ACTION_MODE_NORMAL;

/**
 * 快捷键管理参考gnome-shell。不再在 cdos-desktop 中提供调用系
 * 统功能的快捷键。cdos-desktop 只对这些快捷键提供dbus接口。
 *
 * 窗口管理相关快捷方式在 mutter 中设置
 *
 * 桌面相关在 cdos-desktop 中设置
 *
 * 其余快捷键（系统功能、调用其他应用等）在 cdos-settings-daemon
 * 中注册，通过dbus接口调用cdos-desktop，最终调用到
 * meta_display_grab_accelerator.
 */

/**
 * SECTION: cdos-keybindings
 * @title: CdosKeybindings
 * @short_description: 用来管理用户自定义的快捷键
 * @include: cdos-keybindings.h
 *
 * CdosKeybindings 用来添加或者移除用户自定义的快捷键。
 */


struct _CdosKeybindings
{
    GObject parent_instance;
    MetaDisplay *display;
    GHashTable *bindings; /* key name -> action mode */
    GHashTable *grabbers; /* sender -> watch id */
    GHashTable *action_to_sender; /* action -> sender */

    GDBusConnection *connection;
    CdosKeyGrabber *kg_skeleton;
    guint           kg_bus_id;
};

typedef struct
{
    GDBusProxy *proxy;
    guint value;
} ProxyData;

G_DEFINE_TYPE (CdosKeybindings, cdos_keybindings, G_TYPE_OBJECT);

static void
cdos_keybindings_dispose (GObject *object)
{
    CdosKeybindings *self = CDOS_KEYBINDINGS (object);

    if (self->bindings)
    {
        g_hash_table_destroy (self->bindings);
        self->bindings = NULL;
    }
    if (self->grabbers)
    {
        g_hash_table_destroy (self->grabbers);
        self->grabbers = NULL;
    }

    G_OBJECT_CLASS (cdos_keybindings_parent_class)->dispose (object);
}

static void
cdos_keybindings_finalize (GObject *object)
{
    G_OBJECT_CLASS (cdos_keybindings_parent_class)->finalize (object);
}

static void
allow_keybinding (CdosKeybindings *self,
                  char *key,
                  guint mode)
{
    g_hash_table_replace (self->bindings, key, GUINT_TO_POINTER (mode));
}

static guint
grab_accel (CdosKeybindings *self,
            const gchar *accel,
            guint mode,
            guint flags)
{
    guint action;
    g_autofree char *key = NULL;
    action = meta_display_grab_accelerator (self->display, accel, flags);
    if (action == META_KEYBINDING_ACTION_NONE)
        return action;

    key = meta_external_binding_name_for_action (action);
    allow_keybinding (self, g_steal_pointer (&key), mode);

    return action;
}

static void
on_sender_vanished (GDBusConnection *conn,
                    const char      *name,
                    gpointer         user_data)
{
    GHashTableIter iter;
    gpointer key, value;
    guint action;
    CdosKeybindings *self = user_data;

    g_hash_table_remove (self->grabbers, name);

    g_hash_table_iter_init (&iter, self->action_to_sender);
    while (g_hash_table_iter_next (&iter, &key, &value))
    {
        if (g_strcmp0 (value, name) == 0)
        {
            action = GPOINTER_TO_UINT (key);
            meta_display_ungrab_accelerator (self->display, action);
            g_hash_table_iter_remove (&iter);
        }
    }
}

static void
watch_sender (CdosKeybindings *self,
              const gchar     *sender)
{

    guint watch_id = 0;
    if (g_hash_table_contains (self->grabbers, sender))
        return;
    watch_id = g_bus_watch_name_on_connection (self->connection,
                                               sender,
                                               G_BUS_NAME_WATCHER_FLAGS_NONE,
                                               NULL,
                                               on_sender_vanished,
                                               self,
                                               NULL);
    g_hash_table_insert (self->grabbers, g_strdup (sender), GUINT_TO_POINTER (watch_id));
}

static gboolean
cdos_keybindings_handle_grab_accelerator (CdosKeyGrabber  *grabber,
                                          GDBusMethodInvocation *invocation,
                                          const gchar     *accelerator,
                                          guint            mode,
                                          guint            flags,
                                          CdosKeybindings *self)
{
    const gchar *sender;
    char *ref_sender;
    guint action = 0;
    sender = g_dbus_method_invocation_get_sender (invocation);
    ref_sender = g_ref_string_new (sender);
    action = grab_accel (self, accelerator, mode, flags);

    if (action)
        g_hash_table_insert (self->action_to_sender, GUINT_TO_POINTER (action), ref_sender);

    cdos_key_grabber_complete_grab_accelerator (grabber,
                                                invocation,
                                                action);
    return TRUE;
}

static void
cdos_keybindings_handle_grab_accelerators (CdosKeyGrabber *grabber,
                                           GDBusMethodInvocation *invocation,
                                           GVariant *parameters,
                                           CdosKeybindings *self)
{
    GVariantIter iter;
    gchar *accel = NULL;
    guint mode = 0, flags = 0;
    GVariantBuilder *builder;
    const gchar *sender;
    char *ref_string;
    builder = g_variant_builder_new (G_VARIANT_TYPE ("au"));
    g_variant_iter_init (&iter, parameters);

    sender = g_dbus_method_invocation_get_sender (invocation);
    ref_string = g_ref_string_new (sender);
    //gsize n_accels = g_variant_n_children (parameters);
    //for (gsize i = 0; i < n_accels; i++)
    while (g_variant_iter_next (&iter, "(suu)", &accel, &mode, &flags))
    {
        guint action = grab_accel (self, accel, mode, flags);
        g_variant_builder_add (builder, "u", action);
        if (action)
            g_hash_table_insert (self->action_to_sender, GUINT_TO_POINTER (action), g_ref_string_acquire (ref_string));
    }

    g_ref_string_release (ref_string);
    watch_sender (self, sender);

    cdos_key_grabber_complete_grab_accelerators (grabber,
                                                invocation,
                                                g_variant_new ("au", builder));
    g_variant_builder_unref (builder);
}

static void
cdos_keybindings_handle_ungrab_accelerator (CdosKeyGrabber *grabber,
                                            GDBusMethodInvocation *invocation,
                                            guint action,
                                            CdosKeybindings *self)
{
    g_debug ("%s %u", __func__, action);
    const gchar *sender;
    gchar *old_sender;
    gboolean ret = FALSE;
    sender = g_dbus_method_invocation_get_sender (invocation);
    old_sender = g_hash_table_lookup (self->action_to_sender, GUINT_TO_POINTER (action));
    if (g_strcmp0 (sender, old_sender) == 0)
    {
        ret = meta_display_ungrab_accelerator (self->display, action);
        if (ret)
            g_hash_table_remove (self->action_to_sender, GUINT_TO_POINTER (action));
    }
    cdos_key_grabber_complete_ungrab_accelerator (grabber,
                                                  invocation,
                                                  ret);
}

static void
cdos_keybindings_handle_ungrab_accelerators (CdosKeyGrabber *grabber,
                                             GDBusMethodInvocation *invocation,
                                             GVariant *parameters,
                                             CdosKeybindings *self)
{
    g_debug ("%s %s", __func__, g_variant_get_type_string (parameters));
    guint action = 0;
    gboolean ret = FALSE;
    const gchar *sender;
    gchar *old_sender;
    GVariantIter iter;

    sender = g_dbus_method_invocation_get_sender (invocation);

    g_variant_iter_init (&iter, parameters);
    while (g_variant_iter_next (&iter, "u", &action))
    {
        gboolean r = FALSE;
        old_sender = g_hash_table_lookup (self->action_to_sender, GUINT_TO_POINTER (action));
        if (g_strcmp0 (sender, old_sender) == 0)
        {
            r = meta_display_ungrab_accelerator (self->display, action);
            if (r)
                g_hash_table_remove (self->action_to_sender, GUINT_TO_POINTER (action));
        }
        ret &= r;
    }
    cdos_key_grabber_complete_ungrab_accelerators (grabber,
                                                   invocation,
                                                   ret);
}

static void
on_accelerator_activated (MetaDisplay *display,
                          guint action,
                          ClutterInputDevice *device,
                          guint32 timestamp,
                          CdosKeybindings *self)
{
    g_debug ("%s %u", __func__, action);
    if (!g_hash_table_contains (self->action_to_sender, GUINT_TO_POINTER (action)))
        return;

    cdos_key_grabber_emit_accelerator_activated (self->kg_skeleton,
                                                 action,
                                                 timestamp,
                                                 clutter_input_device_get_device_node (device));
}

static void
on_bus_acquired (GDBusConnection *connection,
                 const gchar     *name,
                 CdosKeybindings *self)
{
    GError *error = NULL;
    self->kg_skeleton = cdos_key_grabber_skeleton_new ();
    if (error)
    {
        g_warning ("%s", error->message);
        g_error_free (error);
        g_clear_object (&self->kg_skeleton);
        return;
    }
    self->connection = connection;

    g_signal_connect (self->kg_skeleton, "handle-grab-accelerator",
                      G_CALLBACK (cdos_keybindings_handle_grab_accelerator),
                      self);
    g_signal_connect (self->kg_skeleton, "handle-grab-accelerators",
                      G_CALLBACK (cdos_keybindings_handle_grab_accelerators),
                      self);
    g_signal_connect (self->kg_skeleton, "handle-ungrab-accelerator",
                      G_CALLBACK (cdos_keybindings_handle_ungrab_accelerator),
                      self);
    g_signal_connect (self->kg_skeleton, "handle-ungrab-accelerators",
                      G_CALLBACK (cdos_keybindings_handle_ungrab_accelerators),
                      self);

    g_dbus_interface_skeleton_export (G_DBUS_INTERFACE_SKELETON (self->kg_skeleton),
                                      connection,
                                      "/org/freedesktop/KeyGrabber",
                                      &error);
    g_signal_connect (self->display, "accelerator-activated",
                      G_CALLBACK (on_accelerator_activated), self);
}

static void
on_name_acquired (GDBusConnection *connection,
                  const char      *name,
                  gpointer         user_data)
{
    g_debug ("%s: %s", __func__, name);
}

static void
on_name_lost (GDBusConnection *connection,
              const gchar     *name,
              gpointer         user_data)
{
    g_debug ("%s: %s", __func__, name);
}

static void
cdos_keybindings_init (CdosKeybindings *self)
{
    CdosGlobal *global = cdos_global_get ();
    self->display = cdos_global_get_display (global);

    self->kg_bus_id = g_bus_own_name (G_BUS_TYPE_SESSION,
                                   "org.freedesktop.KeyGrabber",
                                   G_BUS_NAME_OWNER_FLAGS_REPLACE,
                                   (GBusAcquiredCallback)on_bus_acquired,
                                   on_name_acquired,
                                   on_name_lost,
                                   self,
                                   NULL);

    self->bindings = g_hash_table_new_full (g_str_hash, g_str_equal, g_free, NULL);
    self->grabbers = g_hash_table_new_full (g_str_hash, g_str_equal, g_free, NULL);
    self->action_to_sender = g_hash_table_new_full (NULL, NULL, NULL, (GDestroyNotify)g_ref_string_release);
}

static void
cdos_keybindings_class_init (CdosKeybindingsClass *klass)
{
    GObjectClass *object_class;

    object_class = G_OBJECT_CLASS (klass);

    object_class->finalize = cdos_keybindings_finalize;
    object_class->dispose = cdos_keybindings_dispose;
}

/**
*cdos_keybindings_new
*
*创建#CdosKeybindings对象
*
*Returns: 一个#CdosKeybindings对象
*/

CdosKeybindings *
cdos_keybindings_new (void)
{
    static CdosKeybindings *keybindings = NULL;
    if (!keybindings)
    {
        keybindings = g_object_new (CDOS_TYPE_KEYBINDINGS, NULL);
        g_object_add_weak_pointer ((GObject *)keybindings, (gpointer *)&keybindings);
    }

    return keybindings;
}

gboolean
cdos_keybindings_allow_keybinding (CdosKeybindings *self,
                                   const gchar *key)
{
    guint mode;
    mode = GPOINTER_TO_UINT (g_hash_table_lookup (self->bindings, key));

    return !(mode & cdos_get_action_mode ());
}

CdosActionMode
cdos_get_action_mode (void)
{
    return action_mode;
}

void
cdos_set_action_mode (CdosActionMode mode)
{
    action_mode = mode;
}

