/*
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General
 * Public License along with this library; if not, see <http://www.gnu.org/licenses/>.
 */
#include "config.h"

#include <meta/meta-backend.h>
#include "cdos-global.h"
#include "cdos-layout-manager.h"
#include "cdos-input-method.h"
#include "cdos-keyboard-manager.h"
#include "cdos-keyboard.h"
#include "cdos-keyboard-private.h"

typedef struct _CdosKeyboardManager
{
    GObject parent;
    ClutterSeat *seat;
    ClutterVirtualInputDevice *virtual_device;
    ClutterInputDevice *last_device;
    ClutterInputMethod *im;
    GSettings *a11y_applications_settings;

    CdosKeyboard *keyboard;
    gboolean enabled;
    gboolean im_focus_in;
    gulong settings_change_id;
    gulong a11y_applications_settings_changed_id;
    gulong touch_mode_changed_id;
    gulong last_device_changed_id;
    gulong im_focus_in_id;
    gulong im_focus_out_id;
    gulong im_cursor_location_changed_id;
} CdosKeyboardManager;

G_DEFINE_TYPE (CdosKeyboardManager, cdos_keyboard_manager, G_TYPE_OBJECT)

static void
cdos_keyboard_manager_dispose (GObject *object)
{
    MetaBackend *backend;
    CdosKeyboardManager *self = CDOS_KEYBOARD_MANAGER (object);
    g_clear_object (&self->a11y_applications_settings);
    if (self->touch_mode_changed_id)
    {
        g_signal_handler_disconnect (self->seat, self->touch_mode_changed_id);
        self->touch_mode_changed_id = 0;
    }
    if (self->keyboard)
    {
        cdos_layout_manager_hide_keyboard (cdos_layout_manager_get_default (), FALSE);
        clutter_actor_destroy (CLUTTER_ACTOR (self->keyboard));
        self->keyboard = NULL;
    }

    g_clear_signal_handler (&self->im_focus_in_id, self->im);
    g_clear_signal_handler (&self->im_focus_out_id, self->im);
    g_clear_signal_handler (&self->im_cursor_location_changed_id, self->im);

    backend = meta_get_backend ();
    g_clear_signal_handler (&self->last_device_changed_id, backend);

    G_OBJECT_CLASS (cdos_keyboard_manager_parent_class)->dispose (object);
}

static void
on_im_focus_in (ClutterInputMethod *im,
                CdosKeyboardManager *self)
{
    /*
    CdosLayoutManager *layout_manager = cdos_layout_manager_get_default ();
    cdos_layout_manager_show_keyboard (layout_manager);
    */
    self->im_focus_in = TRUE;
    cdos_keyboard_show (self->keyboard);
}

static void
on_im_focus_out (ClutterInputMethod *im,
                 CdosKeyboardManager *self)
{
    /*
    CdosLayoutManager *layout_manager = cdos_layout_manager_get_default ();
    cdos_layout_manager_hide_keyboard (layout_manager, TRUE);
    */
    self->im_focus_in = FALSE;
    cdos_keyboard_hide (self->keyboard);
}

static void
on_cursor_location_changed (ClutterInputMethod *im,
                            const graphene_rect_t *rect,
                            CdosKeyboardManager *self)
{

}

static gboolean
is_touch_screen (CdosKeyboardManager *self)
{
    ClutterInputDeviceType device_type;
    if (!self->last_device)
        return FALSE;
    device_type = clutter_input_device_get_device_type (self->last_device);
    return device_type == CLUTTER_TOUCHSCREEN_DEVICE;
}

static void
sync_keyboard_state (CdosKeyboardManager *self)
{
    gboolean enabled = g_settings_get_boolean (self->a11y_applications_settings,
                                               "screen-keyboard-enabled");
    gboolean is_touch = clutter_seat_get_touch_mode (self->seat) &&
                        is_touch_screen (self);

    CdosLayoutManager *layout_manager = cdos_layout_manager_get_default ();

    self->enabled = enabled;

    if (!self->enabled)
    {
        g_clear_signal_handler (&self->im_focus_in_id, self->im);
        g_clear_signal_handler (&self->im_focus_out_id, self->im);
        g_clear_signal_handler (&self->im_cursor_location_changed_id, self->im);

        if (!self->keyboard)
            return;
    }

    if (self->enabled)
    {
        if (!self->keyboard)
        {
            self->keyboard = cdos_keyboard_new ();
            /*
            cdos_keyboard_show (self->keyboard);
            */
        }
        if (!self->im_focus_in_id)
        {
            self->im_focus_in_id =
                g_signal_connect (self->im,
                                  "focus-in",
                                  G_CALLBACK (on_im_focus_in),
                                  self);
            self->im_focus_out_id =
                g_signal_connect (self->im,
                                  "focus-out",
                                  G_CALLBACK (on_im_focus_out),
                                  self);
            self->im_cursor_location_changed_id =
                g_signal_connect (self->im,
                                  "cursor-location-changed",
                                  G_CALLBACK (on_cursor_location_changed),
                                  self);
        }
    }
    else if (self->keyboard)
    {
        cdos_layout_manager_hide_keyboard (layout_manager, FALSE);
        clutter_actor_destroy (CLUTTER_ACTOR (self->keyboard));
        self->keyboard = NULL;
    }
    g_debug ("%s %d", __func__, self->enabled);
}

static void
on_last_device_changed (MetaBackend *backend,
                        ClutterInputDevice *device,
                        CdosKeyboardManager *self)
{
    if (clutter_input_device_get_device_type (device) == CLUTTER_KEYBOARD_DEVICE)
    {
        return;
    }
    self->last_device = device;
    sync_keyboard_state (self);
}

static void
cdos_keyboard_manager_init (CdosKeyboardManager *self)
{
    MetaBackend *backend;

    self->seat = clutter_backend_get_default_seat (clutter_get_default_backend ());

    self->a11y_applications_settings = g_settings_new (A11Y_APPLICATIONS_SCHEMA);
    self->a11y_applications_settings_changed_id =
            g_signal_connect_swapped (self->a11y_applications_settings,
                                      "changed::screen-keyboard-enabled",
                                      G_CALLBACK (sync_keyboard_state),
                                      self);

    self->touch_mode_changed_id =
        g_signal_connect_swapped (self->seat,
                                  "notify::touch-mode",
                                  G_CALLBACK (sync_keyboard_state),
                                  self);
    backend = meta_get_backend ();
    self->last_device_changed_id = g_signal_connect (backend,
                                                     "last-device-changed",
                                                     G_CALLBACK (on_last_device_changed),
                                                     self);

    self->im = cdos_input_method_get ();
    self->virtual_device = clutter_seat_create_virtual_device (self->seat,
                                                               CLUTTER_KEYBOARD_DEVICE);

    sync_keyboard_state (self);
}

static void
cdos_keyboard_manager_class_init (CdosKeyboardManagerClass *klass)
{
    GObjectClass *gobject_class = G_OBJECT_CLASS (klass);
    gobject_class->dispose = cdos_keyboard_manager_dispose;
}

CdosKeyboardManager *
cdos_keyboard_manager_new (void)
{
    static CdosKeyboardManager *keyboard_manager = NULL;
    if (!keyboard_manager)
    {
        keyboard_manager = g_object_new (CDOS_TYPE_KEYBOARD_MANAGER,
                                         NULL);
        g_object_add_weak_pointer ((GObject *)keyboard_manager,
                                   (void **)&keyboard_manager);
    }
    return keyboard_manager;
}

CdosKeyboardManager *
cdos_keyboard_manager_get (void)
{
    return cdos_keyboard_manager_new ();
}

gboolean
cdos_keyboard_manager_should_take_event (CdosKeyboardManager *self,
                                         ClutterEvent *event)
{
    ClutterActor *source;
    g_return_val_if_fail (CDOS_IS_KEYBOARD_MANAGER (self), FALSE);

    if (!self->keyboard)
        return FALSE;

    source = clutter_event_get_source (event);
    if (clutter_actor_contains ((ClutterActor *)self->keyboard, source))
        return TRUE;

    return FALSE;
}

gboolean
cdos_keyboard_manager_get_im_focused (CdosKeyboardManager *self)
{
    g_return_val_if_fail (CDOS_IS_KEYBOARD_MANAGER (self), FALSE);
    return self->im_focus_in;
}

void
cdos_keyboard_manager_keyval_press (CdosKeyboardManager *self, guint keyval)
{
    g_return_if_fail (CDOS_IS_KEYBOARD_MANAGER (self));

    clutter_virtual_input_device_notify_keyval (self->virtual_device,
                                                clutter_get_current_event_time (),
                                                keyval,
                                                CLUTTER_KEY_STATE_PRESSED);
}

void
cdos_keyboard_manager_keyval_release (CdosKeyboardManager *self, guint keyval)
{
    g_return_if_fail (CDOS_IS_KEYBOARD_MANAGER (self));

    clutter_virtual_input_device_notify_keyval (self->virtual_device,
                                                clutter_get_current_event_time (),
                                                keyval,
                                                CLUTTER_KEY_STATE_RELEASED);
}
