/*
 * Created by zhangtao on 16-7-5.
 */

/**
 * SECTION: cdos-keyboard
 * @short_description: 提供虚拟键盘输入相关功能。
 * @title: CdosKeyboard
 * @include: cdos-keyboard.h
 *
 * #CdosKeyboard 主要提供了桌面虚拟键盘输入相关的功能:控制虚拟键盘布局，加载虚拟按键，控制虚拟键盘的是显示和隐藏，响应鼠标和键盘的聚焦事件等。
 *
 * #CdosKeyboard 是基于caribou库，在此基础上进行了一定的扩展和修改。
 */


#include <math.h>
#include "cdos-global.h"
#include "cdos-keyboard.h"
#include "cdos-keyboard-private.h"
#include "cdos-key.h"
#include "cdos-caribou-dbus.h"
#include "cdos-message-tray.h"
#include "cdos-layout-manager.h"
#include "cdos-keyboard-manager.h"
#include "cdos-start.h"

typedef struct _CdosKeyboardSignalIds CdosKeyboardSignalIds;
struct _CdosKeyboardSignalIds
{
    gulong keyboard_settings_changed_id;
    gulong a11y_applications_settings_changed_id;
    gulong keyboard_notify_active_group_id;
    gulong global_stage_notify_key_focus_id;
    gulong group_notify_active_level_id;
    gulong tray_button_press_id;
    gulong overview_showing_id;
    gulong overview_hiding_id;
    gulong key_released_id;
    gulong keyboard_group_added_id;
    gulong keyboard_group_removed_id;
};

typedef struct _CdosKeyboard
{
    StBoxLayout parent;

    ClutterSeat *seat;
    guint32 timestamp;
    guint num_of_horiz_keys;
    guint num_of_vert_keys;
    GSettings *keyboard_settings;
    GSettings *a11y_applications_settings;
    CdosKeyboardSignalIds *signal_ids;
    CaribouKeyboardModel *keyboard;
    StBoxLayout *keyboard_box;
    ClutterAction *drag_action;

    GTree *groups;
    StWidget *current_page;
    gboolean showing;
    const gchar *icon_name;
    const gchar *name;
    /*dbus*/
    GDBusInterfaceSkeleton *interface;

    gulong touch_mode_changed_id;
    gulong monitors_changed_id;
    guint caribou_id;

    MetaWindow *focus_window;
}CdosKeyboard;

G_DEFINE_TYPE (CdosKeyboard, cdos_keyboard, ST_TYPE_BOX_LAYOUT)

static void cdos_keyboard_dispose (GObject *object);
static void cdos_keyboard_finalize (GObject *object);

static void cdos_keyboard_settings_changed (CdosKeyboard *self, gchar *key);

static void cdos_keyboard_register_dbus (CdosKeyboard *self);

static void cdos_keyboard_destroy (CdosKeyboard *self);

static void cdos_keyboard_setup (CdosKeyboard *self);

static void cdos_keyboard_add_keys (CdosKeyboard *self);

static void cdos_keyboard_add_rows (CdosKeyboard *self, CaribouRowModel *row, StWidget *layout);

static void cdos_keyboard_load_rows (CdosKeyboard *self, CaribouLevelModel *level, StWidget *layout);

static void cdos_keyboard_redraw (CdosKeyboard *self);

static void cdos_keyboard_set_active_layer (CdosKeyboard *self);

static void cdos_keyboard_on_notify_active_group (GObject *keyboard, GParamSpec *pspec, CdosKeyboard *self);

static void cdos_keyboard_group_on_notify_active_level (GObject *group, GParamSpec *pspec, CdosKeyboard *self);

static CaribouKeyboardModel* cdos_keyboard_create_caribou_model (CdosKeyboard *self);

static GTree* cdos_keyboard_create_layers_for_group (CdosKeyboard *self, gchar *gname);

static void cdos_keyboard_group_added (CaribouKeyboardModel *keyboard, gchar *gname, CdosKeyboard *self);

static void cdos_keyboard_group_removed (CaribouKeyboardModel *keyboard, gchar *gname, CdosKeyboard *self);

static void cdos_keyboard_on_monitors_changed (CdosLayoutManager *layout_manager, CdosKeyboard *self);

//D-Bus methods;
static gboolean dbus_handle_show (Keyboard *object,
                                  GDBusMethodInvocation *invocation,
                                  guint32 timestamp,
                                  CdosKeyboard *self);
static gboolean dbus_handle_hide (Keyboard *object,
                                  GDBusMethodInvocation *invocation,
                                  guint32 timestamp,
                                  CdosKeyboard *self);
static gboolean dbus_handle_set_cursor_location (Keyboard *object,
                                                 GDBusMethodInvocation *invocation,
                                                 gint x,
                                                 gint y,
                                                 gint w,
                                                 gint h,
                                                 CdosKeyboard *self);
static gboolean dbus_handle_set_entry_location (Keyboard *object,
                                                GDBusMethodInvocation *invocation,
                                                gint x,
                                                gint y,
                                                gint w,
                                                gint h,
                                                CdosKeyboard *self);
static void
cdos_keyboard_dispose (GObject *object)
{
    CdosKeyboard *self = CDOS_KEYBOARD (object);
    CdosLayoutManager *layout_manager = cdos_layout_manager_get_default ();

    g_clear_object (&self->keyboard_settings);

    if (self->monitors_changed_id)
    {
        g_signal_handler_disconnect (layout_manager, self->monitors_changed_id);
        self->monitors_changed_id = 0;
    }

    if (self->groups)
    {
        g_tree_destroy (self->groups);
        self->groups = NULL;
    }

    if (self->interface)
    {
        g_object_unref (self->interface);
        self->interface = NULL;
    }

    /*
    if (self->drag_action)
    {
        clutter_actor_remove_action (CLUTTER_ACTOR (self->keyboard_box), self->drag_action);
        self->drag_action = NULL;
    }
    */

    if (self->caribou_id)
    {
        g_bus_unown_name (self->caribou_id);
        self->caribou_id = 0;
    }

    cdos_keyboard_destroy (self);

    G_OBJECT_CLASS (cdos_keyboard_parent_class)->dispose (object);
}

static void
cdos_keyboard_finalize (GObject *object)
{
    CdosKeyboard *self = CDOS_KEYBOARD (object);

    g_free (self->signal_ids);

    G_OBJECT_CLASS (cdos_keyboard_parent_class)->finalize (object);
}

static void
cdos_keyboard_init (CdosKeyboard *self)
{
    CdosLayoutManager *layout_manager = cdos_layout_manager_get_default ();

    self->signal_ids = g_new0 (CdosKeyboardSignalIds, 1);
    self->timestamp = 0;
    self->keyboard_settings = g_settings_new (KEYBOARD_SCHEMA);
    self->icon_name = "virtual-keyboard";
    self->name = _("Cdos virtual keyboard");

    self->keyboard_box = cdos_layout_manager_get_keyboardbox (layout_manager);

    /*
    self->drag_action = clutter_drag_action_new ();
    clutter_actor_add_action (CLUTTER_ACTOR (self->keyboard_box), self->drag_action);
    */

    self->monitors_changed_id =
            g_signal_connect (layout_manager,
                              "monitors-changed",
                              G_CALLBACK (cdos_keyboard_on_monitors_changed),
                              self);
}

static void
cdos_keyboard_class_init (CdosKeyboardClass *klass)
{
    GObjectClass *gobject_class = G_OBJECT_CLASS (klass);

    gobject_class->dispose = cdos_keyboard_dispose;
    gobject_class->finalize = cdos_keyboard_finalize;
}

static void
cdos_keyboard_settings_changed (CdosKeyboard *self, gchar *key)
{
    g_debug ("%s", __func__);
    //MetaDisplay *display;
    const gchar *keyboard_type = NULL;
    gchar *keyboard_type_s = g_settings_get_string (self->keyboard_settings, KEYBOARD_TYPE);

    if (self->keyboard)
    {
        keyboard_type = caribou_keyboard_model_get_keyboard_type (self->keyboard);
        cdos_keyboard_destroy (self);
    }
    g_debug ("%s %s", keyboard_type, keyboard_type_s);

    if (self->keyboard != NULL &&
        g_strcmp0 (keyboard_type, keyboard_type_s) != 0)
    {
        return ;
    }

    /*
    display = cdos_global_get_display (global);
    self->timestamp = meta_display_get_current_time_roundtrip (display) + 1;
    */
    cdos_keyboard_setup (self);
}

static void
on_name_lost_handler (GDBusConnection *connection,
                      const gchar *name,
                      gpointer user_data)
{
    /* Name was already taken, or the bus went away */
    g_warning ("Name taken or bus went away - shutting down");
}

static void
on_name_acquired_handler (GDBusConnection *connection,
                          const gchar *name,
                          gpointer data)
{
    g_debug ("%s: %s", __func__, name);

    CdosKeyboard *self = CDOS_KEYBOARD (data);
    GDBusInterfaceSkeleton *interface = keyboard_skeleton_new ();
    self->interface = interface;
    g_dbus_interface_skeleton_export (G_DBUS_INTERFACE_SKELETON (interface),
                                      connection,
                                      KEYBOARD_DBUS_PATH,
                                      NULL);
    g_signal_connect (interface,
                      "handle-show",
                      G_CALLBACK (dbus_handle_show),
                      self);
    g_signal_connect (interface,
                      "handle-hide",
                      G_CALLBACK (dbus_handle_hide),
                      self);
    g_signal_connect (interface,
                      "handle-set-cursor-location",
                      G_CALLBACK (dbus_handle_set_cursor_location),
                      self);
    g_signal_connect (interface,
                      "handle-set-entry-location",
                      G_CALLBACK (dbus_handle_set_entry_location),
                      self);
    keyboard_set_name ((Keyboard *)interface, self->name);
    keyboard_set_icon ((Keyboard *)interface, self->icon_name);
    //keyboard_set_enabled ((Keyboard *)interface, self->keyboard_enabled);
}

static void
cdos_keyboard_register_dbus (CdosKeyboard *self)
{
    g_debug ("register: %s", KEYBOARD_DBUS_INTERFACE);
    self->caribou_id = g_bus_own_name (G_BUS_TYPE_SESSION,
                                       KEYBOARD_DBUS_INTERFACE,
                                       G_BUS_NAME_OWNER_FLAGS_NONE,
                                       NULL,
                                       (GBusNameAcquiredCallback) on_name_acquired_handler,
                                       (GBusNameLostCallback) on_name_lost_handler,
                                       self,
                                       NULL);
}

static void
cdos_keyboard_destroy (CdosKeyboard *self)
{
    ClutterStage *stage = cdos_global_get_stage (global);
    if (self->signal_ids->global_stage_notify_key_focus_id)
    {
        g_signal_handler_disconnect (stage, self->signal_ids->global_stage_notify_key_focus_id);
        self->signal_ids->global_stage_notify_key_focus_id = 0;
    }

    if (self->keyboard)
    {
        g_object_unref (self->keyboard);
        self->keyboard = NULL;
    }
}

static CaribouKeyboardModel*
cdos_keyboard_create_caribou_model (CdosKeyboard *self)
{
    gchar *keyboard_type;
    CaribouKeyboardModel * model = NULL;
    keyboard_type = g_settings_get_string (self->keyboard_settings,KEYBOARD_TYPE);
    if (keyboard_type)
    {
        model = (CaribouKeyboardModel*) g_object_new (CARIBOU_TYPE_KEYBOARD_MODEL,
                                                      "keyboard-type", g_strdup(keyboard_type),
                                                      NULL);
        g_free(keyboard_type);
    }
    return model;
}

static void
cdos_keyboard_group_added (CaribouKeyboardModel *keyboard, gchar *gname, CdosKeyboard *self)
{
    GTree *layers = cdos_keyboard_create_layers_for_group (self, gname);
    g_tree_insert (self->groups, g_strdup (gname), layers);

}

static void
cdos_keyboard_group_removed (CaribouKeyboardModel *keyboard, gchar *gname, CdosKeyboard *self)
{
    g_tree_remove (self->groups, gname);
}

static void
cdos_keyboard_setup (CdosKeyboard *self)
{
    g_debug ("%s", __func__);
    clutter_actor_set_name (CLUTTER_ACTOR (self), "keyboard");
    clutter_actor_set_reactive (CLUTTER_ACTOR (self), TRUE);
    st_box_layout_set_vertical (ST_BOX_LAYOUT (self), TRUE);

    clutter_actor_add_child (CLUTTER_ACTOR (self->keyboard_box), CLUTTER_ACTOR (self));
    self->keyboard = cdos_keyboard_create_caribou_model (self);

    self->groups = g_tree_new_full ((GCompareDataFunc)g_strcmp0, NULL,
                                    g_free, (GDestroyNotify)g_tree_destroy);
    self->current_page = NULL;

    self->num_of_horiz_keys = 0;
    self->num_of_vert_keys = 0;
    cdos_keyboard_add_keys (self);

    self->signal_ids->keyboard_notify_active_group_id =
        g_signal_connect (self->keyboard,
                          "notify::active-group",
                          G_CALLBACK (cdos_keyboard_on_notify_active_group),
                          self);

    //ClutterStage *global_stage = cdos_global_get_stage (global);
    //关闭聚焦后弹出键盘；
    /*self->signal_ids->global_stage_notify_key_focus_id =
        g_signal_connect (G_OBJECT (global_stage),
                          "notify::key-focus",
                          G_CALLBACK (cdos_keyboard_global_stage_on_notify_key_focus),
                          self);*/
    self->signal_ids->keyboard_group_added_id =
        g_signal_connect (self->keyboard,
                          "group-added",
                          G_CALLBACK(cdos_keyboard_group_added),
                          self);
    self->signal_ids->keyboard_group_removed_id =
        g_signal_connect (self->keyboard,
                          "group-removed",
                          G_CALLBACK(cdos_keyboard_group_removed),
                          self);
}

static GTree *
cdos_keyboard_create_layers_for_group (CdosKeyboard *self, gchar *gname)
{
    g_debug ("%s %s", __func__, gname);
    CaribouGroupModel *group = caribou_keyboard_model_get_group (self->keyboard, gname);
    g_signal_connect (group,
                      "notify::active-level",
                      G_CALLBACK (cdos_keyboard_group_on_notify_active_level),
                      self);
    gint levels_length, j;
    GTree *layers = g_tree_new_full ((GCompareDataFunc)g_strcmp0, NULL,
                                     g_free, NULL);
    gchar **levels = caribou_group_model_get_levels (group, &levels_length);
    for (j = 0; j < levels_length; j++)
    {
        gchar *lname = levels[j];
        g_print ("\t%s", lname);
        CaribouLevelModel *level = caribou_group_model_get_level (group, lname);
        StWidget *layout = st_box_layout_new ();
        st_widget_set_style_class_name (layout, "keyboard-layout");
        st_box_layout_set_vertical (ST_BOX_LAYOUT (layout), TRUE);
        cdos_keyboard_load_rows (self, level, layout);
        g_tree_insert (layers, g_strdup (lname), layout);
        clutter_actor_add_child (CLUTTER_ACTOR (self), CLUTTER_ACTOR (layout));
        clutter_actor_hide (CLUTTER_ACTOR (layout));
    }
    g_print("\n");
    g_strfreev (levels);
    return layers;
}

static void
cdos_keyboard_add_keys (CdosKeyboard *self)
{
    g_debug ("%s", __func__);
    gint groups_length;
    gchar **groups = caribou_keyboard_model_get_groups (self->keyboard, &groups_length);
    guint i;
    for (i = 0; i < groups_length; i++)
    {
        gchar *gname = groups[i];
        GTree *layers = cdos_keyboard_create_layers_for_group (self, gname);
        g_tree_insert (self->groups, g_strdup (gname), layers);
    }
    g_strfreev (groups);
    cdos_keyboard_set_active_layer (self);
}

static void
cdos_keyboard_key_on_prefs_key_released (CaribouIKeyboardObject* sender,
                                         CaribouKeyModel* key,
                                         CdosKeyboard *self)
{
    cdos_keyboard_hide (self);
}

static void
cdos_keyboard_add_rows (CdosKeyboard *self, CaribouRowModel *row, StWidget *layout)
{
    StWidget *keyboard_row = st_box_layout_new ();
    st_widget_set_style_class_name (keyboard_row, "keyboard-rows");
    gint keys_length, i;
    CaribouColumnModel **column_keys = caribou_row_model_get_columns (row, &keys_length);
    for (i = 0; i < keys_length; i++)
    {
        CaribouColumnModel *column_key = column_keys[i];
        gint children_length, j;

        CaribouKeyModel **children =
        (CaribouKeyModel **)caribou_ikeyboard_object_get_children ((CaribouIKeyboardObject *)column_key, &children_length);

        StWidget *right_box = st_box_layout_new ();
        StWidget *center_box = st_box_layout_new ();
        StWidget *left_box = st_box_layout_new ();
        clutter_actor_set_x_expand ((ClutterActor *)left_box, TRUE);
        clutter_actor_set_x_expand ((ClutterActor *)center_box, TRUE);
        clutter_actor_set_x_expand ((ClutterActor *)right_box, TRUE);
        st_widget_set_style_class_name (right_box, "keyboard-row");
        st_widget_set_style_pseudo_class (right_box, "right");
        st_widget_set_style_class_name (center_box, "keyboard-row");
        st_widget_set_style_pseudo_class (center_box, "center");
        st_widget_set_style_class_name (left_box, "keyboard-row");
        st_widget_set_style_pseudo_class (left_box, "left");

        for (j = 0; j < children_length; j++)
        {
            if (self->num_of_horiz_keys == 0)
            {
                self->num_of_horiz_keys = children_length;
            }

            CaribouKeyModel *key = (CaribouKeyModel *)children[j];
            CdosKey *button = cdos_key_new (key);
            const gchar *key_align = caribou_key_model_get_align (key);
            const gchar *key_name = caribou_key_model_get_name (key);
            StWidget *button_actor = ST_WIDGET (button);
            if (g_strcmp0 (key_align, "right") == 0)
            {
                clutter_actor_add_child (CLUTTER_ACTOR (right_box), CLUTTER_ACTOR (button_actor));
            }
            else if (g_strcmp0 (key_align, "left") == 0)
            {
                clutter_actor_add_child (CLUTTER_ACTOR (left_box), CLUTTER_ACTOR (button_actor));
            }
            else
            {
                clutter_actor_add_child (CLUTTER_ACTOR (center_box), CLUTTER_ACTOR (button_actor));
            }

            if (g_strcmp0 (key_name, "Caribou_Prefs") == 0)
            {
                g_signal_connect (key,
                                  "key-released",
                                  G_CALLBACK (cdos_keyboard_key_on_prefs_key_released),
                                  self);
            }
        }
        clutter_actor_add_child (CLUTTER_ACTOR (keyboard_row), CLUTTER_ACTOR (left_box));
        clutter_actor_add_child (CLUTTER_ACTOR (keyboard_row), CLUTTER_ACTOR (center_box));
        clutter_actor_add_child (CLUTTER_ACTOR (keyboard_row), CLUTTER_ACTOR (right_box));
    }
    clutter_actor_add_child (CLUTTER_ACTOR (layout), CLUTTER_ACTOR (keyboard_row));
}

static void
cdos_keyboard_load_rows (CdosKeyboard *self, CaribouLevelModel *level, StWidget *layout)
{
    gint rows_length, i;
    CaribouRowModel **rows = caribou_level_model_get_rows (level, &rows_length);
    for (i = 0; i < rows_length; i++)
    {
        CaribouRowModel *row = rows[i];
        if (self->num_of_vert_keys == 0)
        {
            self->num_of_vert_keys = rows_length;
        }
        cdos_keyboard_add_rows (self, row, layout);
    }
}

static void
cdos_keyboard_redraw (CdosKeyboard *self)
{
    const MetaRectangle *monitor;
    gint focus_index;
    gfloat max_height, max_width;
    CdosLayoutManager *layout_manager = cdos_layout_manager_get_default ();
    focus_index = cdos_layout_manager_get_focus_index (layout_manager);
    cdos_layout_manager_set_keyboard_index (layout_manager, focus_index);
    monitor = cdos_layout_manager_get_monitor_geometry (layout_manager, focus_index);
    if (!monitor)
        return;

    max_height = monitor->height / 3;
    max_width = monitor->width * 0.667;
    clutter_actor_set_width (CLUTTER_ACTOR (self), max_width);

    StWidget *layout = self->current_page;
    StThemeNode *layout_node;

    if (layout)
    {
        layout_node = st_widget_get_theme_node (layout);
    }
    gdouble vertical_spacing = st_theme_node_get_length (layout_node, "spacing");
    gdouble padding = st_theme_node_get_length (layout_node, "padding");

    ClutterActor *first_child = clutter_actor_get_first_child (CLUTTER_ACTOR (layout));
    ClutterActor *box = NULL;
    if (first_child)
        box = clutter_actor_get_first_child (first_child);
    if (!box)
        return;

    StThemeNode *box_node = st_widget_get_theme_node (ST_WIDGET (box));
    gdouble horizontal_spacing = st_theme_node_get_length (box_node, "spacing");

    gdouble all_horizontal_spacing = (self->num_of_horiz_keys - 1) * horizontal_spacing;
    gdouble key_width = floor ((max_width - all_horizontal_spacing - 2 * padding) / self->num_of_horiz_keys);

    gdouble all_vertical_spacing = (self->num_of_vert_keys - 1) * vertical_spacing;
    gdouble key_height = floor ((max_height - all_vertical_spacing - 2 * padding) / self->num_of_vert_keys);

    gdouble key_size = MIN (key_width, key_height);
    gfloat actor_height = key_size * self->num_of_vert_keys + all_vertical_spacing + 2 * padding;
    clutter_actor_set_height (CLUTTER_ACTOR (self), actor_height);

    ClutterActorIter iter;
    ClutterActor *keyboard_row;
    clutter_actor_iter_init (&iter, CLUTTER_ACTOR (self->current_page));
    while (clutter_actor_iter_next (&iter, &keyboard_row))
    {
        ClutterActorIter it;
        clutter_actor_iter_init (&it, keyboard_row);
        ClutterActor *box;
        while (clutter_actor_iter_next (&it, &box))
        {
            ClutterActorIter key_iter;
            clutter_actor_iter_init (&key_iter, box);
            ClutterActor *child;
            while (clutter_actor_iter_next (&key_iter, &child))
            {
                CdosKey *key = CDOS_KEY (child);
                gfloat key_width = cdos_key_get_key_width (key);
                clutter_actor_set_width (child, key_width * (float)key_size);
                clutter_actor_set_height (child, (float)key_size);
                if (cdos_key_get_actor_extended_keyboard (key))
                {
                    GList *extended_keys = clutter_actor_get_children (child);
                    for (GList *l = extended_keys; l; l = l->next)
                    {
                        ClutterActor *extended_key = l->data;
                        clutter_actor_set_width (extended_key, (float)key_size);
                        clutter_actor_set_height (extended_key, (float)key_size);
                    }
                    g_list_free (extended_keys);
                }
            }
        }
    }
}

static void
cdos_keyboard_set_active_layer (CdosKeyboard *self)
{
    const gchar *active_group_name = caribou_keyboard_model_get_active_group (self->keyboard);
    CaribouGroupModel *active_group = caribou_keyboard_model_get_group (self->keyboard, active_group_name);
    const gchar *active_level = caribou_group_model_get_active_level (active_group);

    GTree *layers = g_tree_lookup (self->groups, active_group_name);
    if (self->current_page != NULL)
    {
        clutter_actor_hide (CLUTTER_ACTOR (self->current_page));
    }
    self->current_page = g_tree_lookup (layers, active_level);
    clutter_actor_show (CLUTTER_ACTOR (self->current_page));
}

static void
cdos_keyboard_move_temporarily (CdosKeyboard *self)
{
    gint x, y;
    gfloat height;
    MetaRectangle rect;
    CdosGlobal *global = cdos_global_get ();
    MetaDisplay *meta_display = cdos_global_get_display (global);
    MetaWindow *current_window = meta_display_get_focus_window (meta_display);

    meta_window_get_frame_rect (current_window, &rect);
    x = rect.x;
    height = clutter_actor_get_height (CLUTTER_ACTOR (self));
    y = 3 * height / 2;
    meta_window_move_frame (current_window, TRUE, x, y);
}

static void
cdos_keyboard_set_cursor_location (CdosKeyboard *self,
                                   gint x,
                                   gint y,
                                   gint w,
                                   gint h)
{
    g_debug ("TODO: %s", __func__);
    return;
    gfloat actor_height = clutter_actor_get_height (CLUTTER_ACTOR (self));
    if (y >= 2 * actor_height)
    {
        cdos_keyboard_move_temporarily (self);
    }
}

static void
cdos_keyboard_on_notify_active_group (GObject *keyboard,
                                      GParamSpec *pspec,
                                      CdosKeyboard *self)
{
    cdos_keyboard_set_active_layer (self);

    if (self->showing)
    {
        cdos_keyboard_redraw (self);
    }
}

static void
cdos_keyboard_group_on_notify_active_level (GObject *group,
                                            GParamSpec *pspec,
                                            CdosKeyboard *self)
{
    cdos_keyboard_set_active_layer (self);

    if (self->showing)
    {
        cdos_keyboard_redraw (self);
    }
}

static void G_GNUC_UNUSED
cdos_keyboard_set_focus_window (CdosKeyboard *self)
{
    MetaWindow *focus_window = NULL;
    MetaDisplay *display;
    CdosKeyboardManager *keyboard_manager;
    keyboard_manager = cdos_keyboard_manager_get ();
    display = cdos_global_get_display (cdos_global_get ());
    if (cdos_keyboard_manager_get_im_focused (keyboard_manager))
        focus_window = meta_display_get_focus_window (display);

    self->focus_window = focus_window;
}

/*
static gboolean
cdos_keyboard_tray_button_on_press (ClutterActor *actor,
                                    ClutterEvent *event,
                                    CdosKeyboard *self)
{
    CdosMessageTray *main_message_tray = cdos_message_tray_get_default ();
    if (main_message_tray)
    {
        cdos_message_tray_toggle (main_message_tray);
    }
    return FALSE;
}

static void
cdos_keyboard_on_overview_showing (CdosOverview *overview, StWidget *tray_button)
{
    st_widget_add_style_pseudo_class (tray_button, "grayed");
}

static void
cdos_keyboard_on_overview_hiding (CdosOverview *overview, StWidget *tray_button)
{
    clutter_actor_set_reactive (CLUTTER_ACTOR (tray_button), TRUE);
    st_widget_remove_style_pseudo_class (tray_button, "grayed");
}
*/

/**
 * cdos_keyboard_new:
 *
 * 创建#CdosKeyboard实例化对象。
 *
 * Returns :(transfer full): 实例化对象
 */
CdosKeyboard *
cdos_keyboard_new (void)
{
    static CdosKeyboard *keyboard = NULL;
    if (keyboard == NULL)
    {
        keyboard = g_object_new (CDOS_TYPE_KEYBOARD, NULL);
        g_object_add_weak_pointer (G_OBJECT (keyboard), (gpointer *)&keyboard);
        cdos_keyboard_register_dbus (keyboard);
        cdos_keyboard_settings_changed (keyboard, NULL);
        cdos_keyboard_start (keyboard);
    }
    return keyboard;
}

/**
 * cdos_keyboard_start:
 * @self: #CdosKeyboard 当前实例对象
 *
 * 当前实例对象构造完成之后，在特定场景下，做了部分的补充，该方法通常在实例化之后被调用。
 */
void
cdos_keyboard_start (CdosKeyboard *self)
{
    g_return_if_fail (CDOS_IS_KEYBOARD (self));
    cdos_keyboard_redraw (self);
}

/**
 * cdos_keyboard_show:
 * @self: #CdosKeyboard. 当前实例对象
 *
 * 显示当前桌面的虚拟键盘。
 */
void
cdos_keyboard_show (CdosKeyboard *self)
{
    CdosLayoutManager *layout_manager;
    g_return_if_fail (CDOS_IS_KEYBOARD (self));

    if (self->showing)
    {
        return ;
    }
    self->showing = TRUE;
    /*
    cdos_keyboard_redraw (self);
    */
    layout_manager = cdos_layout_manager_get_default ();
    cdos_layout_manager_show_keyboard (layout_manager);
    if (self->interface)
        keyboard_set_enabled ((Keyboard *)self->interface, TRUE);
}

/**
 * cdos_keyboard_hide:
 * @self:当前实例对象
 *
 * 隐藏当前桌面的虚拟键盘。
 */
void
cdos_keyboard_hide (CdosKeyboard *self)
{
    CdosLayoutManager *layout_manager;

    g_return_if_fail (CDOS_IS_KEYBOARD (self));
    if (!self->showing)
    {
        return ;
    }
    self->showing = FALSE;
    layout_manager = cdos_layout_manager_get_default ();
    cdos_layout_manager_hide_keyboard (layout_manager, TRUE);

    if (self->interface)
        keyboard_set_enabled ((Keyboard *)self->interface, FALSE);
}

/**
 * dbus_handle_show:
 * @self:实例对象
 * @timestamp:时间戳
 *
 * 通过DBUS，显示当前桌面的虚拟键盘。
 *
 * DBUS专用接口函数。
 *
 */
static gboolean
dbus_handle_show (Keyboard *object,
                    GDBusMethodInvocation *invocation,
                    guint32 timestamp,
                    CdosKeyboard *self)
{
    g_debug ("%s", __func__);
    self->timestamp = timestamp;
    cdos_keyboard_show (self);
    keyboard_complete_show (object, invocation, TRUE);
    return TRUE;
}

/**
 * dbus_handle_hide:
 * @self:实例对象
 * @timestamp:时间戳
 *
 * 通过DBUS，隐藏当前桌面的虚拟键盘。
 *
 * DBUS专用接口函数。
 *
 */
static gboolean
dbus_handle_hide (Keyboard *object,
                GDBusMethodInvocation *invocation,
                guint32 timestamp,
                CdosKeyboard *self)
{
    g_debug ("%s", __func__);
    self->timestamp = timestamp;
    /*
     * cdos-desktop 输入框激活时，不处理
     */
    CdosKeyboardManager *keyboard_manager;
    keyboard_manager = cdos_keyboard_manager_get ();
    if (!cdos_keyboard_manager_get_im_focused (keyboard_manager))
        cdos_keyboard_hide (self);
    keyboard_complete_hide (object, invocation, TRUE);

    return TRUE;
}

static gboolean
dbus_handle_set_cursor_location (Keyboard *object,
                                 GDBusMethodInvocation *invocation,
                                 gint x,
                                 gint y,
                                 gint w,
                                 gint h,
                                 CdosKeyboard *self)
{
    g_debug ("%s", __func__);
    cdos_keyboard_set_cursor_location (self, x, y, w, h);
    keyboard_complete_set_cursor_location (object, invocation, TRUE);
    return TRUE;
}

static gboolean
dbus_handle_set_entry_location (Keyboard *object,
                                GDBusMethodInvocation *invocation,
                                gint x,
                                gint y,
                                gint w,
                                gint h,
                                CdosKeyboard *self)
{
    g_debug ("%s", __func__);
    cdos_keyboard_set_cursor_location (self, x, y, w, h);
    return TRUE;
}

static void
cdos_keyboard_on_monitors_changed (CdosLayoutManager *layout_manager,
                                   CdosKeyboard *self)
{
    if (self->showing)
    {
        cdos_keyboard_redraw (self);
    }
}


const gchar*
cdos_keyboard_get_icon_name (CdosKeyboard *self)
{
    g_return_val_if_fail (CDOS_IS_KEYBOARD (self), NULL);
    return self->icon_name;
}

const gchar*
cdos_keyboard_get_name (CdosKeyboard *self)
{
    g_return_val_if_fail (CDOS_IS_KEYBOARD (self), NULL);
    return self->name;
}

gboolean
cdos_keyboard_is_showing (CdosKeyboard *self)
{
    g_return_val_if_fail (CDOS_IS_KEYBOARD (self), FALSE);
    return self->showing;
}
