#include "config.h"
#include <meta/display.h>
#include <meta/meta-monitor-manager.h>

#include "cdos-start.h"
#include "cdos-layout-manager.h"
#include "cdos-screen-saver.h"

/**
 * SECTION:cdos-layout-manager
 * @Title: CdosLayoutManager
 * @Short_description: 桌面布局管理器
 *
 * CdosLayoutManager 负责桌面上所有组件的显示与输入区域更新，比如面板、弹出菜单等，
 * 桌面上有控件显示，如果这个控件需要与用户交互，必须要获取这个控件的起始位置和宽高，
 * 将它添加到 CdosLayoutManager 的 chrome 中，并实时更新输入区域。
 */

#define DESKTOPS_ON_PRIMARY "true::false"
#define DESKTOPS_ON_ALL "true::true"
#define DESKTOPS_ON_NON_PRIMARY "false::true"
#define DESKTOPS_ON_NONE "false::false"
#define DESKTOPS_DEFAULT DESKTOPS_ON_PRIMARY

#define META_LAYER_FULLSCREEN 5

typedef struct _ActorStruct
{
    gulong visible_id;
    gulong allocation_id;
    ClutterActor *actor;
    gint flags;
    /*
    gint track_fullscreen;
    gint affect_struts;
    gint affect_input_region;
    */
} ActorStruct;

struct _CdosLayoutManagerPrivate
{
    guint primaryindex;
    guint bottomindex;
    guint update_region_idle;
    StWidget *panelbox;
    ClutterActor *keyboardbox;
    MetaRectangle *primarymonitor;
    GList *tracked_actors;
    GPtrArray *monitors;
    gint keyboard_index;
    gulong keyboard_height_notify_id;
};

G_DEFINE_TYPE_WITH_PRIVATE (CdosLayoutManager, cdos_layout_manager, G_TYPE_OBJECT);

enum
{
    MONITORS_CHANGED,
    LAST_SIGNAL
};

static guint cdos_layout_manager_signals[LAST_SIGNAL] = {0};

static void cdos_layout_manager_queue_update_regions (ClutterActor *actor, gpointer unused, gpointer data);

static void
cdos_layout_manager_dispose (GObject *gobject)
{
    CdosLayoutManager *self = CDOS_LAYOUT_MANAGER (gobject);
    CdosLayoutManagerPrivate *priv = self->priv;

    cdos_layout_manager_remove_chrome (self, (ClutterActor *)priv->panelbox);

    g_clear_pointer ((ClutterActor **)&priv->panelbox, clutter_actor_destroy);
    g_clear_pointer ((ClutterActor **)&priv->keyboardbox, clutter_actor_destroy);

    g_list_free_full (g_steal_pointer (&priv->tracked_actors), g_free);

    if (priv->monitors)
    {
        g_ptr_array_free (priv->monitors, 1);
        priv->monitors = NULL;
    }

    G_OBJECT_CLASS (cdos_layout_manager_parent_class)->dispose (gobject);
}

static void
cdos_layout_manager_finalize (GObject *gobject)
{
    CdosLayoutManager *self = CDOS_LAYOUT_MANAGER (gobject);
    CdosLayoutManagerPrivate *priv = self->priv;

    g_free (priv->primarymonitor);
    /* 调用父类的 finalize() 方法 */
    G_OBJECT_CLASS (cdos_layout_manager_parent_class)->finalize (gobject);
}

static void
cdos_layout_manager_class_init (CdosLayoutManagerClass *klass)
{
    GObjectClass *gobject_class = G_OBJECT_CLASS (klass); // 自己的父类型是什么？

    gobject_class->finalize = cdos_layout_manager_finalize;
    gobject_class->dispose = cdos_layout_manager_dispose;

    /**
     * CdosLayoutManager::monitors-changed:
     * @actor: 接受这个信号的对象
     *
     * CdosLayoutManager 监听桌面 MetaDisplay 的 monitors-changed 信号并转发
     *
     * Since: 1.0.0
     */
    cdos_layout_manager_signals[MONITORS_CHANGED] = g_signal_new ("monitors-changed",
                                                                  CDOS_TYPE_LAYOUT_MANAGER,
                                                                  G_SIGNAL_RUN_LAST,
                                                                  0,
                                                                  NULL,
                                                                  NULL,
                                                                  NULL,
                                                                  G_TYPE_NONE,
                                                                  0);
}

static gboolean
is_above_or_below_primary (CdosLayoutManager *self, MetaRectangle *monitor)
{
    CdosLayoutManagerPrivate *priv = self->priv;
    MetaRectangle *primary = priv->monitors->pdata[priv->primaryindex];
    int m_left = monitor->x;
    int m_right = monitor->x + monitor->width;
    int p_left = primary->x;
    int p_right = primary->x + primary->width;

    if ((m_left >= p_left && m_left < p_right) || (m_right > p_left && m_right <= p_right)
        || (p_left >= m_left && p_left < m_right)
        || (p_right > m_left && p_right <= m_right))
        return TRUE;
    return FALSE;
}

static gint
find_monitor_for_rect (CdosLayoutManager *self, guint x, guint y, guint w, guint h)
{
    guint centerX = x + w / 2;
    guint centerY = y + h / 2;

    GPtrArray *monitors = self->priv->monitors;

    gint i = 0;
    while (monitors != NULL && i < monitors->len)
    {
        MetaRectangle *tmprect = monitors->pdata[i];

        if (centerX >= tmprect->x && centerX < tmprect->x + tmprect->width &&
            centerY >= tmprect->y && centerY < tmprect->y + tmprect->height)
            return i;

        i++;
    }
    i = 0;
    while (monitors != NULL && i < monitors->len)
    {
        MetaRectangle *tmprect = monitors->pdata[i];

        if (x + w > tmprect->x && x < tmprect->x + tmprect->width &&
            y + h > tmprect->y && y < tmprect->y + tmprect->height)
            return i;
        i++;
    }

    return -1;
}

static gint
find_monitor_for_actor (CdosLayoutManager *self, ClutterActor *actor)
{
    gfloat x, y, w, h;
    clutter_actor_get_transformed_size (actor, &w, &h);
    clutter_actor_get_transformed_position (actor, &x, &y);
    return find_monitor_for_rect (self, x, y, w, h);
}

static void
cdos_layout_manager_update_monitors (CdosLayoutManager *self)
{
    MetaDisplay *display;
    MetaRectangle *rect;
    CdosGlobal *global;
    int n_monitors = 0;
    int i;
    CdosLayoutManagerPrivate *priv;
    MetaMonitorSwitchConfigType switch_config;
    MetaMonitorManager *monitor_manager;

    priv = self->priv;

    global = cdos_global_get ();
    display = cdos_global_get_display (global);
    n_monitors = meta_display_get_n_monitors (display);
    monitor_manager = meta_monitor_manager_get ();

    if (priv->monitors)
        g_ptr_array_free (priv->monitors, 1);

    if (n_monitors == 0)
    {
        priv->monitors = NULL;
        priv->primarymonitor = NULL;
        return;
    }
    priv->monitors = g_ptr_array_new_full (n_monitors, g_free);
    priv->monitors->len = n_monitors;
    for (i = 0; i < n_monitors; i++)
    {
        rect = g_malloc0 (sizeof (MetaRectangle));
        meta_display_get_monitor_geometry (display, i, rect);
        // g_ptr_array_add(priv->monitors, rect);
        priv->monitors->pdata[i] = rect;
    }

    if (n_monitors == 1)
    {
        priv->primaryindex = 0;
        priv->bottomindex = 0;
    }
    else
    {
        priv->primaryindex = meta_display_get_primary_monitor (display);
        priv->bottomindex = priv->primaryindex;
        for (i = 0; i < n_monitors; i++)
        {
            MetaRectangle *monitor = priv->monitors->pdata[i];
            if (is_above_or_below_primary (self, monitor))
            {
                if (monitor->y > ((MetaRectangle *)priv->monitors->pdata[priv->bottomindex])->y)
                    priv->bottomindex = i;
            }
        }
        switch_config = meta_monitor_manager_get_switch_config (monitor_manager);
        if (switch_config == META_MONITOR_SWITCH_CONFIG_UNKNOWN)
        {
            meta_monitor_manager_switch_config (monitor_manager, META_MONITOR_SWITCH_CONFIG_ALL_LINEAR);
        }
    }
    priv->primarymonitor = priv->monitors->pdata[priv->primaryindex];
}

/**
 * cdos_layout_manager_update_boxes:
 * @self: #CdosLayoutManager 对象
 *
 * 这个函数用于实时更新显示，例如分辨率发生变化的时候，更改 panel 的显示位置等，
 * 更新桌面上所有控件的接受输入的状态
 */
void
cdos_layout_manager_update_boxes (CdosLayoutManager *self)
{
    g_return_if_fail (CDOS_IS_LAYOUT_MANAGER (self));
    CdosLayoutManagerPrivate *priv = self->priv;
    g_debug ("%s", __func__);

    if (!priv->monitors)
        return;

    gfloat height, p_height;
    ClutterActor *panel = cdos_panel;

    height = clutter_actor_get_height (panel);
    MetaRectangle *bottom = priv->monitors->pdata[priv->bottomindex];
    ClutterActor *panelbox = (ClutterActor *)priv->panelbox;
    if (g_strcmp0 (desktop_layout, LAYOUT_TRADITIONAL) == 0)
    {
        /* a panel at the bottom */
        p_height = height;
        clutter_actor_set_position (panelbox,
                                    bottom->x,
                                    bottom->y + bottom->height - p_height);
        cdos_panel_set_monitor ((CdosPanel *)panel, priv->bottomindex);
    }
    else if (g_strcmp0 (desktop_layout, LAYOUT_FLIPPED) == 0)
    {
        /* panel on top */
        clutter_actor_set_position (panelbox,
                                    (gfloat)priv->primarymonitor->x,
                                    (gfloat)priv->primarymonitor->y);
        cdos_panel_set_monitor ((CdosPanel *)panel, priv->primaryindex);
    }
    else if (g_strcmp0 (desktop_layout, LAYOUT_CLASSIC) == 0)
    {
        p_height = height;
        clutter_actor_set_position (panelbox,
                                    (gfloat)priv->primarymonitor->x,
                                    (gfloat)priv->primarymonitor->y + (gfloat)priv->primarymonitor->height - p_height);
        cdos_panel_set_monitor ((CdosPanel *)panel, priv->primaryindex);
    }

    cdos_layout_manager_set_keyboard_index (self, priv->primaryindex);

    cdos_layout_manager_queue_update_regions (NULL, NULL, self);
}

static void
cdos_layout_manager_update_visibility (CdosLayoutManager *self)
{
    CdosLayoutManagerPrivate *priv = self->priv;
    GList *iter = NULL;
    MetaDisplay *display;
    display = cdos_global_get_display (cdos_global_get ());

    CdosScreenSaver *screen_saver = cdos_screen_saver_get ();
    if (screen_saver && clutter_actor_is_visible ((ClutterActor *)screen_saver))
        return;

    for (iter = priv->tracked_actors; iter; iter = iter->next)
    {
        ActorStruct *actor_data = iter->data;

        if (!actor_data->flags & TRACK_FULLSCREEN)
            continue;
        gboolean visible = TRUE;
        /* TODO screen_saver, overview show*/

        CdosGlobal *global = cdos_global_get ();
        ClutterActor *window_group = cdos_global_get_window_group (global);
        if (clutter_actor_is_visible (window_group))
        {
            gint m = find_monitor_for_actor (self, actor_data->actor);
            if (m >= 0 && meta_display_get_monitor_in_fullscreen (display, m))
                visible = FALSE;
        }
        if (visible)
            clutter_actor_show (actor_data->actor);
        else
            clutter_actor_hide (actor_data->actor);
    }
}

void
cdos_layout_manager_on_window_restacked (MetaDisplay *display, CdosLayoutManager *self)
{
    cdos_global_sync_pointer (cdos_global_get ());
    cdos_layout_manager_update_visibility (self);
    cdos_layout_manager_queue_update_regions (NULL, NULL, self);
}

static void
cdos_layout_manager_relayout (CdosLayoutManager *self)
{
    cdos_layout_manager_update_visibility (self);
    cdos_layout_manager_queue_update_regions (NULL, NULL, self);
}
/*
 * 处理 screen 的 monitors-changed 信号
 */
static void
on_monitors_changed (MetaWorkspaceManager *workspace_manager, CdosLayoutManager *self)
{
    cdos_layout_manager_update_monitors (self);
    cdos_layout_manager_update_boxes (self);
    cdos_layout_manager_update_visibility (self);
    cdos_start_center_pointer_on_screen (self);

    g_signal_emit (self, cdos_layout_manager_signals[MONITORS_CHANGED], 0);
}

static void
cdos_layout_manager_init (CdosLayoutManager *self)
{
    CdosLayoutManagerPrivate *priv;
    CdosGlobal *global = cdos_global_get ();
    MetaDisplay *display = cdos_global_get_display (global);
    MetaMonitorManager *monitor_manager = meta_monitor_manager_get ();

    priv = self->priv = cdos_layout_manager_get_instance_private (self);
    StWidget *panelbox = g_object_new (ST_TYPE_BOX_LAYOUT,
                                       "style-class", "panel-box",
                                       NULL);
    clutter_actor_set_name ((ClutterActor *)panelbox, "panel");

    if (cdos_enable_blur_effect ())
    {
        ClutterEffect *blur_effect = st_blur_effect_new ();
        st_blur_effect_set_mode (ST_BLUR_EFFECT (blur_effect), ST_BLUR_MODE_BACKGROUND);
        st_blur_effect_set_sigma (ST_BLUR_EFFECT (blur_effect), 2.0);
        clutter_actor_add_effect (CLUTTER_ACTOR (panelbox), blur_effect);
    }

    ClutterActor *keyboardbox = (ClutterActor *)st_box_layout_new ();
    //clutter_actor_set_name ((ClutterActor *)keyboardbox, "keyboard");
    clutter_actor_set_reactive (keyboardbox, TRUE);
    clutter_actor_hide (keyboardbox);
    st_widget_set_track_hover ((StWidget *)keyboardbox, TRUE);
    clutter_actor_add_child (cdos_ui_plugin_get_ui_group (), keyboardbox);

    priv->panelbox = panelbox;

    priv->keyboardbox = keyboardbox;

    priv->primaryindex = 0;
    priv->update_region_idle = 0;
    priv->tracked_actors = NULL;
    priv->keyboard_index = -1;
    priv->keyboard_height_notify_id = 0;

    g_signal_connect (display, "restacked", G_CALLBACK (cdos_layout_manager_on_window_restacked), self);
    g_signal_connect (monitor_manager, "monitors-changed", G_CALLBACK (on_monitors_changed), self);
    cdos_layout_manager_update_monitors (self);
    cdos_layout_manager_add_chrome (self, (ClutterActor *)panelbox, 0x7);
    cdos_layout_manager_relayout (self);
}

/**
 * cdos_layout_manager_get_panelbox:
 * @self: #CdosLayoutManager 对象
 *
 * 获取 CdosLayoutManager 对象的 panelbox
 *
 * Returns: (transfer full): 返回 #StWidget 对象.
 */
StWidget *
cdos_layout_manager_get_panelbox (CdosLayoutManager *self)
{
    // 同上
    g_return_val_if_fail (CDOS_IS_LAYOUT_MANAGER (self), NULL);

    return self->priv->panelbox;
}

/**
 * cdos_layout_manager_get_keyboardbox:
 * @self: #CdosLayoutManager 对象
 *
 * 获取 CdosLayoutManager的 keyboardBox 属性
 *
 * Returns: (transfer full): 返回 #StBoxLayout.
 */
StBoxLayout *
cdos_layout_manager_get_keyboardbox (CdosLayoutManager *self)
{
    // 同上
    g_return_val_if_fail (CDOS_IS_LAYOUT_MANAGER (self), NULL);

    return (StBoxLayout *)self->priv->keyboardbox;
}

/**
 * cdos_layout_manager_get_primary_monitor:
 * @self: #CdosLayoutManager 对象
 *
 * 获取 CdosLayoutManager 的屏幕属性
 *
 * Returns: (transfer full): 返回 #MetaRectangle 结构数据
 */
const MetaRectangle *
cdos_layout_manager_get_primary_monitor (CdosLayoutManager *self)
{
    // 同上
    CdosLayoutManagerPrivate *priv;
    g_return_val_if_fail (CDOS_IS_LAYOUT_MANAGER (self), NULL);

    priv = self->priv;
    return priv->primarymonitor;
}

const MetaRectangle *
cdos_layout_manager_find_monitor_for_rect (CdosLayoutManager *self, int x, int y, int width, int height)
{
    CdosLayoutManagerPrivate *priv;
    g_return_val_if_fail (CDOS_IS_LAYOUT_MANAGER (self), NULL);
    priv = self->priv;
    GPtrArray *monitors = priv->monitors;
    for (int i = 0; i < monitors->len; i++)
    {
        MetaRectangle *rect = monitors->pdata[i];
        if (rect->x <= x && rect->y <= y && rect->x + rect->width > x && rect->y + rect->height > y)
            return rect;
    }
    return NULL;
}

const MetaRectangle *
cdos_layout_manager_find_monitor_for_actor (CdosLayoutManager *self, ClutterActor *actor)
{
    CdosLayoutManagerPrivate *priv;
    g_return_val_if_fail (CDOS_IS_LAYOUT_MANAGER (self), NULL);
    priv = self->priv;
    GPtrArray *monitors = priv->monitors;
    gint i = find_monitor_for_actor (self, actor);
    if (i >= 0)
        return monitors->pdata[i];
    return NULL;
}

const MetaRectangle *
cdos_layout_manager_get_monitor_geometry (CdosLayoutManager *self, gint monitor_index)
{
    CdosLayoutManagerPrivate *priv;
    g_return_val_if_fail (CDOS_IS_LAYOUT_MANAGER (self), NULL);

    priv = self->priv;
    GPtrArray *monitors = priv->monitors;
    if (monitor_index >= 0 && monitors && monitor_index < monitors->len)
        return monitors->pdata[monitor_index];
    return NULL;
}

/**
 * cdos_layout_manager_get_monitors:
 * @self: a #CdosLayoutManager
 *
 * Get all monitors
 *
 * Return value: (transfer container) (element-type MetaRectangle): the monitor's list.
 */
GPtrArray *
cdos_layout_manager_get_monitors (CdosLayoutManager *self)
{
    // 同上
    g_return_val_if_fail (CDOS_IS_LAYOUT_MANAGER (self), NULL);

    CdosLayoutManagerPrivate *priv = self->priv;
    return priv->monitors;
}

/**
 * cdos_layout_manager_get_index:
 * @self: a #CdosLayoutManager
 *
 * Get index property
 *
 * Returns: index
 */
gint
cdos_layout_manager_get_index (void)
{
    CdosGlobal *global = cdos_global_get ();
    MetaDisplay *display = cdos_global_get_display (global);
    gint index = meta_display_get_current_monitor (display);
    return index;
}

static void
cdos_layout_manager_queue_update_regions (ClutterActor *actor, gpointer unused, gpointer data)
{
    CdosLayoutManager *self = CDOS_LAYOUT_MANAGER (data);
    CdosLayoutManagerPrivate *priv = self->priv;

    if (priv->update_region_idle == 0)
    {
        priv->update_region_idle = meta_later_add (META_LATER_BEFORE_REDRAW,
                                                   (GSourceFunc)cdos_layout_manager_update_regions,
                                                   self, NULL);
        /*
        priv->update_region_idle = g_idle_add ((GSourceFunc)cdos_layout_manager_update_regions, self);
        */
    }
}

/**
 * cdos_layout_manager_add_chrome:
 * @self:  #CdosLayoutManager 对象
 * @actor: #ClutterActor 对象
 *
 * 将 #ClutterActor 对象添加到显示容器 uiGroup 之中，并修改它的的 input region，
 * 只有给这个对象设置了合适的输入区域后，这个 actor 才会对鼠标等输入事件做响应
 *
 * Returns: none
 */
void
cdos_layout_manager_add_chrome (CdosLayoutManager *self, ClutterActor *actor, gint flags)
{
    g_return_if_fail (CDOS_IS_LAYOUT_MANAGER (self));
    CdosLayoutManagerPrivate *priv = self->priv;
    ActorStruct *actor_data;
    gulong handleone, handletwo;
    ClutterActor *ui_group;
    ClutterActor *top_window_group;

    // FIXME ADD_TO_UI_GROUP
    GList *iter;
    for (iter = priv->tracked_actors; iter; iter = iter->next)
    {
        ActorStruct *strut = iter->data;
        if (strut->actor == actor)
        {
            g_debug ("Trying to re-track existing ClutterActor %p", actor);
            return;
        }
    }
    ClutterActor *parent = clutter_actor_get_parent (actor);
    ui_group = cdos_ui_plugin_get_ui_group ();
    if (!parent)
        clutter_actor_add_child (ui_group, actor);

    top_window_group = cdos_global_get_top_window_group (cdos_global_get ());
    if (clutter_actor_contains (ui_group, top_window_group))
    {
        if (!(flags & ABOVE_TOP_WINDOW_GROUP))
            clutter_actor_set_child_above_sibling (ui_group, top_window_group, actor);
        else
            clutter_actor_set_child_above_sibling (ui_group, actor, top_window_group);
    }

    actor_data = g_malloc0 (sizeof (ActorStruct));
    handleone = g_signal_connect (actor,
                                  "notify::visible",
                                  G_CALLBACK (cdos_layout_manager_queue_update_regions),
                                  self);
    handletwo = g_signal_connect (actor,
                                  "notify::allocation",
                                  G_CALLBACK (cdos_layout_manager_queue_update_regions),
                                  self);
    actor_data->visible_id = handleone;
    actor_data->allocation_id = handletwo;
    actor_data->actor = actor;
    actor_data->flags = flags;
    priv->tracked_actors = g_list_append (priv->tracked_actors, actor_data);
    cdos_layout_manager_queue_update_regions (actor, NULL, self);
}


/**
 * cdos_layout_manager_remove_chrome:
 * @self: #CdosLayoutManager 对象
 * @actor: #ClutterActor 对象
 *
 * 当这个 actor 不需要再现实，将它从 chrome 中移除，不再接受输入事件
 *
 * Returns: none
 */
void
cdos_layout_manager_remove_chrome (CdosLayoutManager *self, ClutterActor *actor)
{
    g_return_if_fail (CDOS_IS_LAYOUT_MANAGER (self));

    GList *actors = NULL;
    CdosLayoutManagerPrivate *priv;

    priv = self->priv;
    actors = priv->tracked_actors;
    GList *recorder = priv->tracked_actors;

    while (actors != NULL)
    {
        ActorStruct *dataitem = actors->data;
        if (dataitem->actor == actor)
        {
            recorder = g_list_remove (recorder, dataitem);
            g_signal_handler_disconnect (dataitem->actor, dataitem->visible_id);
            g_signal_handler_disconnect (dataitem->actor, dataitem->allocation_id);
            g_free (dataitem);
            break;
        }
        actors = actors->next;
    }

    priv->tracked_actors = recorder;
    cdos_layout_manager_queue_update_regions (actor, NULL, self);
}

/**
 * cdos_layout_manager_update_regions:
 * @self: a #CdosLayoutManager
 *
 * 设置输入区域，这里先获取所有添加到 #CdosLayoutManager 中的 actors，再一一设置
 *
 * Returns: success
 */
gboolean
cdos_layout_manager_update_regions (CdosLayoutManager *self)
{
    g_return_val_if_fail (CDOS_IS_LAYOUT_MANAGER (self), FALSE);
    CdosLayoutManagerPrivate *priv;
    priv = self->priv;

    const MetaRectangle *rect;
    GList *actors = NULL;
    GSList *rects = NULL; // FOR INPUT REGION SET
    GSList *struts = NULL;
    MetaRectangle *newRect;

    if (!priv->monitors)
        return FALSE;

    if (priv->update_region_idle != 0ul)
    {
        /*
        g_source_remove (priv->update_region_idle);
        */
        meta_later_remove (priv->update_region_idle);
        priv->update_region_idle = 0ul;
    }

    // Limit struts to the size of the screen
    gint screen_width, screen_height;
    MetaDisplay *display = cdos_global_get_display (cdos_global_get ());
    MetaWorkspaceManager *wsm = meta_display_get_workspace_manager (display);
    meta_display_get_size (display, &screen_width, &screen_height);
    actors = priv->tracked_actors;
    while (actors != NULL)
    {
        GList *next = actors->next;
        ActorStruct *dataitem;
        dataitem = actors->data;
        actors = next;
        gint affect_struts = dataitem->flags & AFFECT_STRUTS;
        gint affect_input_region = dataitem->flags & AFFECT_INPUT_REGION;
        if (!affect_input_region && !affect_struts)
            continue;

        if (!clutter_actor_is_visible (dataitem->actor))
            continue;

        gfloat box_x, box_y, width, height;
        gint x, y, w, h;
        gint x1, x2, y1, y2;

        clutter_actor_get_transformed_position (dataitem->actor, &box_x, &box_y);
        clutter_actor_get_transformed_size (dataitem->actor, &width, &height);
        x = (int)(box_x + 0.5);
        y = (int)(box_y + 0.5);
        w = (int)(width + 0.5);
        h = (int)(height + 0.5);

        newRect = g_malloc0 (sizeof (MetaRectangle));
        newRect->x = x;
        newRect->y = y;
        newRect->width = w;
        newRect->height = h;

        rects = g_slist_append (rects, newRect);

        if (affect_struts)
        {
            x1 = MAX (x, 0);
            x2 = MIN (x + w, screen_width);
            y1 = MAX (y, 0);
            y2 = MIN (y + h, screen_height);

            rect = cdos_layout_manager_find_monitor_for_rect (self, x, y, w, h);

            MetaSide side;
            if (rect && x1 <= rect->x && x2 >= rect->x + rect->width)
            {
                if (y1 <= rect->y)
                    side = META_SIDE_TOP;
                else if (y2 >= rect->y + rect->height)
                    side = META_SIDE_BOTTOM;
                else
                    continue;
            }
            else if (rect && y1 <= rect->y && y2 >= rect->y + rect->height)
            {
                if (x1 < 0)
                    side = META_SIDE_LEFT;
                else if (x2 >= screen_width)
                    side = META_SIDE_RIGHT;
                else
                    continue;
            }
            else if (x1 <= 0)
                side = META_SIDE_LEFT;
            else if (y1 <= 0)
                side = META_SIDE_TOP;
            else if (x2 >= screen_width)
                side = META_SIDE_RIGHT;
            else if (y2 >= screen_height)
                side = META_SIDE_BOTTOM;
            else
                continue;

            switch (side)
            {
                case META_SIDE_TOP:
                    y1 = 0;
                    break;
                case META_SIDE_BOTTOM:
                    y2 = screen_height;
                    break;
                case META_SIDE_LEFT:
                    x1 = 0;
                    break;
                case META_SIDE_RIGHT:
                    x2 = screen_width;
                    break;
            }

            MetaRectangle strutRect;
            strutRect.x = x1;
            strutRect.y = y1;
            strutRect.width = x2 - x1;
            strutRect.height = y2 - y1;

            MetaStrut *strut = g_malloc0 (sizeof (MetaStrut));
            strut->rect = strutRect;
            strut->side = side;
            struts = g_slist_append (struts, strut);
        }
    }

    cdos_global_set_stage_input_region (global, rects);

    gint n_spaces = meta_workspace_manager_get_n_workspaces (wsm);
    gint i = 0;

    while (i < n_spaces)
    {
        MetaWorkspace *space;

        space = meta_workspace_manager_get_workspace_by_index (wsm, i);
        meta_workspace_set_builtin_struts (space, struts);
        ++i;
    }
    if (struts)
    {
        g_slist_free_full (struts, (GDestroyNotify)g_free);
    }
    if (rects)
    {
        g_slist_free_full (rects, (GDestroyNotify)g_free);
    }
    return FALSE;
}

CdosLayoutManager *
cdos_layout_manager_get_default (void)
{
    static CdosLayoutManager *layout_manager = NULL;
    if (layout_manager == NULL)
    {
        layout_manager = g_object_new (CDOS_TYPE_LAYOUT_MANAGER, NULL);
        g_object_add_weak_pointer (G_OBJECT (layout_manager), (gpointer *)&layout_manager);
    }
    return layout_manager;
}

void
cdos_layout_manager_set_keyboard_index (CdosLayoutManager *self, gint index)
{
    g_return_if_fail (CDOS_IS_LAYOUT_MANAGER (self));
    CdosLayoutManagerPrivate * priv = self->priv;
    if (index < 0)
        return;
    if (priv->keyboard_index == index)
        return;
    if (!priv->monitors || priv->monitors->len < 1)
        return;

    priv->keyboard_index = index;
    const MetaRectangle *monitor = priv->monitors->pdata[index];
    if (monitor)
    {
        gfloat x, y, w, h;

        w = monitor->width * 0.667;
        h = -1;
        x = (monitor->x + monitor->width - w ) / 2;
        y = monitor->y + monitor->height;
        g_debug ("%s::index = %d, x = %d, y = %d, w = %d, h = %d", __func__, index, (int)monitor->x, (int)monitor->y, (int)monitor->width, (int)monitor->height);
        clutter_actor_set_position (CLUTTER_ACTOR(priv->keyboardbox), x, y);
        clutter_actor_set_size (CLUTTER_ACTOR(priv->keyboardbox), w, h);
    }
}

gint
cdos_layout_manager_get_keyboard_index (CdosLayoutManager *self)
{
    g_return_val_if_fail (CDOS_IS_LAYOUT_MANAGER (self), -1);
    CdosLayoutManagerPrivate * priv = self->priv;
    return priv->keyboard_index;
}

static void
on_keyboard_height_changed (ClutterActor *keyboard,
                            GParamSpec *spec,
                            CdosLayoutManager *self)
{
    CdosLayoutManagerPrivate * priv = self->priv;
    gfloat kb_height;
    kb_height = clutter_actor_get_height (priv->keyboardbox);
    clutter_actor_set_translation (priv->keyboardbox,
                                   0, kb_height, 0);
}

static void
on_keyboard_shown (ClutterTimeline *tml,
                   CdosLayoutManager *self)
{
    CdosLayoutManagerPrivate * priv = self->priv;
    if (priv->keyboard_height_notify_id == 0)
    {
        priv->keyboard_height_notify_id =
            g_signal_connect (priv->keyboardbox,
                              "notify::height",
                              G_CALLBACK (on_keyboard_height_changed),
                              self);
    }
    cdos_layout_manager_queue_update_regions (NULL, NULL, self);
}

void
cdos_layout_manager_show_keyboard (CdosLayoutManager *self)
{
    g_return_if_fail (CDOS_IS_LAYOUT_MANAGER (self));
    CdosLayoutManagerPrivate * priv = self->priv;
    gfloat kb_height;
    ClutterTransition *trans;
    ClutterActor *ui_group;

    clutter_actor_remove_all_transitions (priv->keyboardbox);
    if (clutter_actor_is_visible (priv->keyboardbox))
    {
        return;
    }
    cdos_layout_manager_add_chrome (self, priv->keyboardbox,
                                    AFFECT_INPUT_REGION | ABOVE_TOP_WINDOW_GROUP);
    clutter_actor_show (CLUTTER_ACTOR (priv->keyboardbox));
    ui_group = cdos_ui_plugin_get_ui_group ();
    clutter_actor_set_child_above_sibling (ui_group, priv->keyboardbox, NULL);
    kb_height = clutter_actor_get_height (priv->keyboardbox);
    trans = cdos_transition_simple (priv->keyboardbox,
                                    CLUTTER_EASE_OUT_QUAD,
                                    300,
                                    "opacity", 255,
                                    "translation-y", -kb_height,
                                    NULL);
    g_signal_connect (trans, "completed",
                      G_CALLBACK (on_keyboard_shown), self);
}

static void
on_keyboard_hidden (CdosLayoutManager *self)
{
    CdosLayoutManagerPrivate * priv = self->priv;
    clutter_actor_hide (priv->keyboardbox);
}

void
cdos_layout_manager_hide_keyboard (CdosLayoutManager *self, gboolean animate)
{
    g_return_if_fail (CDOS_IS_LAYOUT_MANAGER (self));
    ClutterTransition *trans;
    CdosLayoutManagerPrivate * priv = self->priv;
    cdos_layout_manager_remove_chrome (self, priv->keyboardbox);
    g_clear_signal_handler (&priv->keyboard_height_notify_id, priv->keyboardbox);
    if (!animate)
    {
        clutter_actor_hide (priv->keyboardbox);
        return;
    }
    if (!clutter_actor_is_visible (priv->keyboardbox))
    {
        return;
    }
    trans = cdos_transition_simple (priv->keyboardbox,
                                    CLUTTER_EASE_OUT_QUAD,
                                    300,
                                    "opacity", 0,
                                    "translation-y", 0,
                                    NULL);
    g_signal_connect_swapped (trans, "completed",
                              G_CALLBACK (on_keyboard_hidden), self);
    g_signal_connect_swapped (trans, "stopped",
                              G_CALLBACK (on_keyboard_hidden), self);
}

gint
cdos_layout_manager_get_focus_index (CdosLayoutManager *self)
{
    CdosGlobal *global;
    ClutterActor *focused_actor;

    g_return_val_if_fail (CDOS_IS_LAYOUT_MANAGER (self), -1);

    gint i = self->priv->primaryindex;
    global = cdos_global_get ();
    focused_actor = clutter_stage_get_key_focus (cdos_global_get_stage (global));

    if (focused_actor == NULL)
    {
        MetaWindow *focused_window;
        MetaDisplay *display = cdos_global_get_display (global);
        focused_window = meta_display_get_focus_window (display);
        if (focused_window)
            i = meta_window_get_monitor (focused_window);
    }
    return i;
}

const MetaRectangle *
cdos_layout_manager_get_focus_monitor (CdosLayoutManager *self)
{
    gint focus_index = -1;

    g_return_val_if_fail (CDOS_IS_LAYOUT_MANAGER (self), NULL);
    focus_index = cdos_layout_manager_get_focus_index (self);
    if (focus_index > -1)
    {
        return self->priv->monitors->pdata[focus_index];
    }
    return NULL;
}
