#include <meta/util.h>

#include "cdos-lightbox.h"
#include "cdos-start.h"

G_DEFINE_TYPE (CdosLightbox, cdos_lightbox, ST_TYPE_WIDGET);

/**
 * SECTION:cdos-lightbox
 * @title: CdosLightbox
 * @short_description: 设置lightbox
 *
 * 对lightbox显示,隐藏和创建,设置lightbox的变暗时间
 */

static void
cdos_lightbox_dispose (GObject *gobject)
{
    CdosLightbox *box = CDOS_LIGHTBOX (gobject);

    if(box->allocation_changed_handler > 0)
    {
        g_signal_handler_disconnect (box->container, box->allocation_changed_handler);
        box->allocation_changed_handler = 0;
    }
    g_object_unref(box->container);
    box->highlighted = NULL;
    if (box->children)
    {
        g_list_free (box->children);
        box->children = NULL;
    }
    G_OBJECT_CLASS (cdos_lightbox_parent_class)->dispose (gobject);
}

static gboolean
change_width_height (CdosLightbox *data)
{
    CdosLightbox *box = data;
    ClutterActor *actor = box->actor;
    clutter_actor_set_width (actor, box->width);
    clutter_actor_set_height (actor, box->height);
    return 0;
}
static void
allocation_changed (ClutterActor *container,
                    ClutterActorBox *clutter_box,
                    guint    flags,
                    CdosLightbox *lightbox)
{
    CdosLightbox *box = lightbox;
    meta_later_add (META_LATER_BEFORE_REDRAW, (GSourceFunc)change_width_height, box, NULL);

    ClutterActor *c = container;
    float width = clutter_actor_get_width (c);
    float height = clutter_actor_get_height (c);
    box->width = width;
    box->height = height;
}

static void
transition_complete_show (ClutterTimeline *tml, CdosLightbox *box)
{
    clutter_actor_remove_transition (box->actor, "animate-box-opacity");
    clutter_actor_show (box->actor);
}

/**
 * cdos_lightbox_show:
 * @box: 作为其他actor的一个容器
 *
 * 显示lightbox
 *
 * Returns: void
 */
void
cdos_lightbox_show (CdosLightbox *box)
{
    if (box->fade_time > 0)
    {
        clutter_actor_set_opacity (box->actor, 0x0);
        clutter_actor_show (box->actor);
        /*
        ClutterAnimation *anim = clutter_actor_animate (box->actor,
                                        CLUTTER_EASE_OUT_QUAD,
                                        box->fade_time,
                                        "opacity", 0xff,
                                        NULL);
        //ClutterTimeline *tml = clutter_animation_get_timeline (anim);
        g_signal_connect (anim, "completed",
                          G_CALLBACK (clutter_actor_show),
                          box->actor);
                          */
        ClutterTransition *transition = clutter_property_transition_new ("opacity");
        clutter_actor_set_easing_mode (box->actor, CLUTTER_EASE_OUT_QUAD);
        clutter_timeline_set_duration (CLUTTER_TIMELINE (transition), box->fade_time);
        clutter_timeline_set_repeat_count (CLUTTER_TIMELINE (transition), 0);
        clutter_transition_set_from (transition, G_TYPE_UCHAR, 0x0);
        clutter_transition_set_to (transition, G_TYPE_UCHAR, 255);
        clutter_actor_add_transition (box->actor, "animate-box-opacity", transition);
        g_signal_connect (transition, "completed", G_CALLBACK (transition_complete_show), box);
    }
    else{
        clutter_actor_set_opacity (box->actor, 0xff);
        clutter_actor_show (box->actor);
    }
}

static void
transition_complete_hide (ClutterTimeline *tml, CdosLightbox *box)
{
    clutter_actor_remove_transition (box->actor, "animate-box-opacity");
    clutter_actor_hide (box->actor);
}


/**
 * cdos_lightbox_hide:
 * @box: 作为其他actor的一个容器
 *
 * 隐藏lightbox
 *
 * Returns: void
 */
void
cdos_lightbox_hide (CdosLightbox *box)
{
    if(box->fade_time > 0)
    {
        /*
        ClutterAnimation *anim = clutter_actor_animate (box->actor,
                                        CLUTTER_EASE_OUT_QUAD,
                                        box->fade_time,
                                        "opacity", 0x0,
                                        NULL);
        //ClutterTimeline *tml = clutter_animation_get_timeline (anim);
        g_signal_connect (anim, "completed",
                          G_CALLBACK (clutter_actor_hide),
                          box->actor);
        */
        ClutterTransition *transition = clutter_property_transition_new ("opacity");
        clutter_actor_set_easing_mode (box->actor, CLUTTER_EASE_OUT_QUAD);
        clutter_timeline_set_duration (CLUTTER_TIMELINE (transition), box->fade_time);
        clutter_timeline_set_repeat_count (CLUTTER_TIMELINE (transition), 0);
        clutter_transition_set_from (transition, G_TYPE_UCHAR, 0xff);
        clutter_transition_set_to (transition, G_TYPE_UCHAR, 0x0);
        clutter_actor_add_transition (box->actor, "animate-box-opacity", transition);
        g_signal_connect (transition, "completed", G_CALLBACK (transition_complete_hide), box);

    }else{
        clutter_actor_hide (box->actor);
    }
}


/**
 * cdos_lightbox_highlight:
 * @box: 作为其他actor的一个容器
 * @window: 一个clutter的actor
 *
 * 设置lightbox的highlight
 *
 * Returns: void
 */
void
cdos_lightbox_highlight (CdosLightbox *box, ClutterActor *window)
{
    if(box->highlighted == window)
        return;
    GList *iter = NULL;
    for (iter = g_list_last (box->children); iter; iter = iter->prev)
    {
        if(iter->data == window)
        {
            clutter_actor_set_child_above_sibling (box->container, window, NULL);
            return;
        }
        else if(iter->data == box->highlighted)
            clutter_actor_set_child_above_sibling (box->container, iter->data, NULL);
    }
    box->highlighted = window;
}

static void
create (CdosLightbox *lightbox)
{
    ClutterActor *actor = (ClutterActor *)lightbox;
    StWidget *widget = ST_WIDGET (actor);

    st_widget_set_style_class_name (widget, "lightbox");

    clutter_actor_add_child (lightbox->container, (ClutterActor *)widget);
    clutter_actor_set_position (actor, 0.0, 0.0);
    clutter_actor_set_child_above_sibling (lightbox->container, actor, NULL);
    clutter_actor_hide (actor);

    lightbox->allocation_changed_handler =
        g_signal_connect (lightbox->container, "allocation-changed",
                G_CALLBACK(allocation_changed), lightbox);

    lightbox->highlighted = NULL;
}

static void
cdos_lightbox_init (CdosLightbox *self)
{
    create (self);
}

static void
cdos_lightbox_class_init (CdosLightboxClass *klass)
{
    GObjectClass *gobject_class = G_OBJECT_CLASS (klass);
    gobject_class->dispose = cdos_lightbox_dispose;
}

/**
 * cdos_lightbox_set_fade_time:
 * @box: 作为其他actor的一个容器
 * @msec: 设置的时间
 *
 * 设置lightbox的变暗时间
 *
 * Returns: void
 */
void
cdos_lightbox_set_fade_time(CdosLightbox *box, guint msec)
{
    if(box == NULL)
        return;
    box->fade_time = msec;
}


/**
 * cdos_lightbox_new:
 * @container: 作为其他actor的一个容器
 * @width: lightbox的宽度
 * @height: lightbox的高度
 * @inhibit_events: 是否允许容器的事件发生
 *
 * 新建lightbox
 *
 * Returns: lightbox
 */
CdosLightbox *
cdos_lightbox_new (ClutterActor    *container,
                   gfloat           width,
                   gfloat           height,
                   gboolean         inhibit_events)
{
    container = g_object_ref(container);
    CdosLightbox *lightbox = g_object_new (CDOS_TYPE_LIGHTBOX, "reactive", inhibit_events, NULL);
    lightbox->width          = (width > 0.f) ? width : clutter_actor_get_width ((ClutterActor *)container);
    lightbox->height         = (height > 0.f) ? height : clutter_actor_get_height ((ClutterActor *)container);

    lightbox->container      = container;
    lightbox->children       = clutter_actor_get_children (container);

    //lightbox->fade_time      = fade_time;
    return lightbox;
}
