/**
 * SECTION: cdos-login-manager
 * @short_description: logind
 *
 */
#include <gio/gio.h>
#include "cdos-login-manager.h"

enum
{
    PREPARE_FOR_SLEEP,
    LAST_SIGNAL
};
static guint logind_signals[LAST_SIGNAL] = {0};

struct _CdosLoginManager
{
    GObject parent_instance;

    GDBusProxy *proxy;
};

G_DEFINE_TYPE (CdosLoginManager, cdos_login_manager, G_TYPE_OBJECT);

static void
logind_signal_cb (GDBusProxy *proxy,
                  gchar *sender,
                  gchar *signal,
                  GVariant *parameters,
                  CdosLoginManager *self)
{
    if (g_strcmp0 (signal, "PrepareForSleep") == 0)
    {
        gboolean is_about_to_suspend;
        g_variant_get (parameters, "(b)", &is_about_to_suspend);
        g_signal_emit (self, logind_signals[PREPARE_FOR_SLEEP], 0, is_about_to_suspend);
    }
}

static void
cdos_login_manager_dispose (GObject *gobject)
{
    CdosLoginManager *self = CDOS_LOGIN_MANAGER (gobject);

    g_clear_object (&self->proxy);

    G_OBJECT_CLASS (cdos_login_manager_parent_class)->dispose (gobject);
}

static void
on_proxy_get (GObject *source,
              GAsyncResult *res,
              CdosLoginManager *self)
{
    GError *error = NULL;
    self->proxy = g_dbus_proxy_new_for_bus_finish (res, &error);
    if (error)
    {
        g_warning ("%s: %s", __func__, error->message);
        return;
    }
    g_signal_connect (self->proxy, "g-signal", G_CALLBACK (logind_signal_cb), self);
}

static void
cdos_login_manager_init (CdosLoginManager *self)
{
    g_dbus_proxy_new_for_bus (G_BUS_TYPE_SYSTEM,
                              G_DBUS_PROXY_FLAGS_DO_NOT_AUTO_START,
                              NULL,
                              "org.freedesktop.login1",
                              "/org/freedesktop/login1",
                              "org.freedesktop.login1.Manager",
                              NULL,
                              (GAsyncReadyCallback)on_proxy_get,
                              self);

}

static void
cdos_login_manager_class_init (CdosLoginManagerClass *klass)
{
    GObjectClass *gobject_class = G_OBJECT_CLASS (klass);

    gobject_class->dispose = cdos_login_manager_dispose;

    logind_signals[PREPARE_FOR_SLEEP] = g_signal_new ("prepare-for-sleep",
                                                      CDOS_TYPE_LOGIN_MANAGER,
                                                      G_SIGNAL_RUN_LAST,
                                                      0,
                                                      NULL,
                                                      NULL,
                                                      NULL,
                                                      G_TYPE_NONE,
                                                      1,
                                                      G_TYPE_BOOLEAN);
}

/**
 * cdos_login_manager_new:
 *
 * 新建一个 #CdosLoginManager 类的实例。
 *
 * Returns: 返回一个 #CdosLoginManager 的实例。
 */
CdosLoginManager *
cdos_login_manager_new (void)
{
    static CdosLoginManager *self = NULL;
    if (!self)
    {
        self = g_object_new (CDOS_TYPE_LOGIN_MANAGER, NULL);
        g_object_add_weak_pointer ((GObject *)self, (gpointer *)&self);
    }
    return self;
}

static void
call_method_done (GObject *proxy,
                  GAsyncResult *res,
                  gpointer user_data)
{
    GError *error = NULL;
    g_dbus_proxy_call_finish ((GDBusProxy *)proxy, res, &error);
    if (error)
    {
        g_warning ("%s", error->message);
        g_error_free (error);
    }
}

static void
call_method (CdosLoginManager *self, const gchar *method, gboolean interactive)
{
    g_return_if_fail (CDOS_IS_LOGIN_MANAGER (self));
    g_return_if_fail (G_IS_DBUS_PROXY (self->proxy));

    g_dbus_proxy_call (self->proxy,
                       method,
                       g_variant_new ("(b)",
                                      interactive),
                       G_DBUS_CALL_FLAGS_NONE,
                       -1,
                       NULL,
                       (GAsyncReadyCallback) call_method_done,
                       self);
}

void
cdos_login_manager_halt (CdosLoginManager *self, gboolean interactive)
{
    call_method (self, "Halt", interactive);
}

void
cdos_login_manager_hibernate (CdosLoginManager *self, gboolean interactive)
{
    call_method (self, "Hibernate", interactive);
}

void
cdos_login_manager_hybrid_sleep (CdosLoginManager *self, gboolean interactive)
{
    call_method (self, "HybridSleep", interactive);
}

void
cdos_login_manager_power_off (CdosLoginManager *self, gboolean interactive)
{
    call_method (self, "PowerOff", interactive);
}

void
cdos_login_manager_reboot (CdosLoginManager *self, gboolean interactive)
{
    call_method (self, "Reboot", interactive);
}

void
cdos_login_manager_suspend (CdosLoginManager *self, gboolean interactive)
{
    call_method (self, "Suspend", interactive);
}


struct CallMethodData
{
    CdosLogin1Callback callback;
    gpointer user_data;
};

static CdosLogin1Result
get_sd_result_from_string (const char *str)
{
    if (!g_strcmp0 (str, "na"))
    {
        return CDOS_LOGIN1_RESULT_NA;
    }
    else if (!g_strcmp0 (str, "yes"))
    {
        return CDOS_LOGIN1_RESULT_YES;
    }
    else if (!g_strcmp0 (str, "no"))
    {
        return CDOS_LOGIN1_RESULT_NO;
    }
    else if (!g_strcmp0 (str, "challenge"))
    {
        return CDOS_LOGIN1_RESULT_CHALLENGE;
    }
    else
        return CDOS_LOGIN1_RESULT_YES;
}

static void
call_method_with_callback_done (GObject *proxy,
                                GAsyncResult *res,
                                gpointer user_data)
{
    GError *error = NULL;
    GVariant *v = NULL;
    char *str = NULL;
    v = g_dbus_proxy_call_finish ((GDBusProxy *)proxy, res, &error);
    if (error)
    {
        g_warning ("%s", error->message);
        g_error_free (error);
        return;
    }
    struct CallMethodData *data = user_data;
    if (data)
    {
        g_variant_get (v, "(s)", &str);
        data->callback (data->user_data, get_sd_result_from_string (str));
        g_free (data);
    }
    g_variant_unref (v);
}

static void
call_method_with_callback (CdosLoginManager *self,
                           const gchar *method,
                           CdosLogin1Callback callback,
                           gpointer user_data)
{
    struct CallMethodData *data = NULL;
    if (callback)
    {
        data = g_new0 (struct CallMethodData, 1);
        data->callback = callback;
        data->user_data = user_data;
    }

    g_dbus_proxy_call (self->proxy,
                       method,
                       NULL,
                       G_DBUS_CALL_FLAGS_NONE,
                       -1,
                       NULL,
                       (GAsyncReadyCallback) call_method_with_callback_done,
                       data);
}

void
cdos_login_manager_can_halt (CdosLoginManager *self,
                             CdosLogin1Callback callback,
                             gpointer user_data)
{
    call_method_with_callback (self, "CanHalt", callback, user_data);
}

void
cdos_login_manager_can_hibernate (CdosLoginManager *self,
                                  CdosLogin1Callback callback,
                                  gpointer user_data)
{
    call_method_with_callback (self, "CanHibernate", callback, user_data);
}

void
cdos_login_manager_can_hybrid_sleep (CdosLoginManager *self,
                                     CdosLogin1Callback callback,
                                     gpointer user_data)
{
    call_method_with_callback (self, "CanHybridSleep", callback, user_data);
}

void
cdos_login_manager_can_power_off (CdosLoginManager *self,
                                  CdosLogin1Callback callback,
                                  gpointer user_data)
{
    call_method_with_callback (self, "CanPowerOff", callback, user_data);
}

void
cdos_login_manager_can_reboot (CdosLoginManager *self,
                               CdosLogin1Callback callback,
                               gpointer user_data)
{
    call_method_with_callback (self, "CanReboot", callback, user_data);
}

void
cdos_login_manager_can_suspend (CdosLoginManager *self,
                                CdosLogin1Callback callback,
                                gpointer user_data)
{
    call_method_with_callback (self, "CanSuspend", callback, user_data);
}
