#include "st.h"
#include "cdos-start.h"
#include "cdos-looking-glass.h"
#include "cdos-types.h"

/**
 * SECTION: cdos-looking-glass
 * @Title: CdosLookingGlass
 * @short_description: 用于UI调试
 *
 * 查看一个 #ClutterActor 的所有可读属性。
 */

struct _CdosLookingGlass
{
    gboolean open;
    gfloat   offset;
    GList   *raw_results;
    gint     max_items;

    ClutterActor *icon; // for cdos_pop/push_modal
    ClutterActor *actor;
    ClutterActor *orig;
    gulong   inspect_id;
    ClutterEffect *effect;

    gboolean pass_event;
};

struct _Inspector
{
    ClutterActor *self;
    GList        *parents;
    GList        *children;
};

static gboolean on_button_press_event (ClutterActor *actor,
                                       ClutterEvent *event,
                                       CdosLookingGlass *lg);

static gboolean on_motion_event (ClutterActor *actor,
                                 ClutterEvent *event,
                                 CdosLookingGlass *lg);

static gboolean on_scroll_event (ClutterActor *actor,
                                 ClutterEvent *event,
                                 CdosLookingGlass *lg);

static void
end_inspect (CdosLookingGlass *lg)
{
    ClutterActor *ui_group = cdos_ui_plugin_get_ui_group ();
    clutter_actor_remove_child (ui_group, lg->icon);
    clutter_actor_destroy (lg->icon);

    if (lg->inspect_id > 0)
        g_signal_handler_disconnect (cdos_global_get_stage (global), lg->inspect_id);

    g_free (lg);
}

static gboolean
on_captured_event (ClutterActor *actor,
                   ClutterEvent *event,
                   CdosLookingGlass *lg)
{
    ClutterEventType type = clutter_event_type (event);
    //cdos_debug ("pass_event: %d\n", lg->pass_event);
    if (type == CLUTTER_KEY_PRESS)
    {
        guint sym = clutter_event_get_key_symbol (event);
        if ((sym == CLUTTER_KEY_Control_L) || (sym == CLUTTER_KEY_Control_R) ||
            (sym == CLUTTER_KEY_Pause))
        {
            lg->pass_event = !lg->pass_event;
            return TRUE;
        }
        if (sym == CLUTTER_KEY_Escape)
        {
            cdos_pop_modal (lg->icon, 0);
            g_signal_handler_disconnect (cdos_global_get_stage (global), lg->inspect_id);
            lg->inspect_id = 0L;
            if (lg->effect != NULL)
            {
                clutter_actor_remove_effect (lg->actor, lg->effect);
                while (CLUTTER_IS_EFFECT (lg->effect))
                    g_object_unref (lg->effect);
                lg->actor = NULL;
                lg->orig = NULL;
                lg->effect = NULL;
            }
            end_inspect (lg);
            return TRUE;
        }
    }

    if (lg->pass_event)
        return FALSE;

    switch (type)
    {
        case CLUTTER_BUTTON_PRESS:
            return on_button_press_event (actor, event, lg);
        case CLUTTER_MOTION:
            return on_motion_event (actor, event, lg);
        case CLUTTER_SCROLL:
            return on_scroll_event (actor, event, lg);
        default:
            return TRUE;
    }

    return TRUE;
}

static gboolean
on_motion_event (ClutterActor *actor,
                 ClutterEvent *event,
                 CdosLookingGlass *lg)
{
    graphene_point_t coord;
    clutter_event_get_position (event, &coord);

    ClutterStage *stage = cdos_global_get_stage (global);
    ClutterActor *target = clutter_stage_get_actor_at_pos (stage, CLUTTER_PICK_ALL,
                                coord.x, coord.y);

    if (target == lg->actor)
        return TRUE;
    if (lg->orig == NULL) lg->orig = target;
    ClutterActor *tmp = lg->actor;
    lg->actor = target;

    if (tmp && lg->effect != 0)
    {
        clutter_actor_remove_effect (tmp, lg->effect);
    }
    if (target != lg->orig)
    {
        if (tmp != lg->orig)
            clutter_actor_remove_effect (lg->orig, lg->effect);
        lg->orig = target;
    }
    while (CLUTTER_IS_EFFECT (lg->effect))
        g_object_unref (lg->effect);

    lg->effect = clutter_blur_effect_new ();
    clutter_actor_add_effect (lg->actor, lg->effect);

    return TRUE;
}

static ClutterEffect *
create_shader_effect (CdosLookingGlass *lg)
{
    if (CLUTTER_IS_SHADER_EFFECT (lg->effect))
        return lg->effect;

    ClutterEffect *effect = clutter_shader_effect_new (CLUTTER_FRAGMENT_SHADER);

    const char *shader =
        "uniform sampler2D tex;\n"
        "uniform float height;\n"
        "uniform float width;\n"
        "void main()\n"
        "{\n"
        "   vec4 pos = cogl_tex_coord_in[0];\n"
        "   vec4 color = cogl_color_in * texture2D(tex, pos.xy);\n"
        "   if(pos.x < width || pos.y < height ||"
        "       abs(pos.x - 1) < width || abs(pos.y - 1.0) < height)\n"
        "       cogl_color_out = vec4(color.rgb, 0.8) * vec4(0.5, 1.0, 0.3, 1.0);\n"
        "   else\n"
        "       cogl_color_out = vec4(color.rgb * color.rgb, color.a * 0.7);\n"
        "}";
    clutter_shader_effect_set_shader_source ((ClutterShaderEffect *)effect, shader);
    g_object_ref (effect);
    return effect;
}

static gboolean
on_scroll_event (ClutterActor *actor,
                 ClutterEvent *event,
                 CdosLookingGlass *lg)
{
    ClutterScrollDirection dir = clutter_event_get_scroll_direction (event);

    if (dir == CLUTTER_SCROLL_UP)
    {
        ClutterActor *parent = clutter_actor_get_parent (lg->actor);
        if(parent != NULL)
        {
            clutter_actor_remove_effect (lg->actor, lg->effect);
            lg->actor = parent;
            //while(CLUTTER_IS_EFFECT (lg->effect))
            //    g_object_unref (lg->effect);

            lg->effect = create_shader_effect (lg);
            ClutterShaderEffect *effect = CLUTTER_SHADER_EFFECT (lg->effect);
            clutter_shader_effect_set_uniform (effect,
                    "tex", G_TYPE_INT, 1, 0);
            clutter_shader_effect_set_uniform (effect, "width",
                G_TYPE_FLOAT, 1,
                2.0/clutter_actor_get_width (parent));
            clutter_shader_effect_set_uniform (effect, "height",
                G_TYPE_FLOAT, 1,
                2.0/clutter_actor_get_height (parent));
            clutter_actor_add_effect(lg->actor, lg->effect);
        }
    }
    else if(dir == CLUTTER_SCROLL_DOWN)
    {
        if(lg->actor == lg->orig || lg->orig == NULL)
            return TRUE;
        clutter_actor_remove_effect (lg->actor, lg->effect);
        lg->actor = lg->orig;

        lg->effect = create_shader_effect (lg);
        ClutterShaderEffect *effect = CLUTTER_SHADER_EFFECT (lg->effect);
        clutter_shader_effect_set_uniform (effect,
                "tex", G_TYPE_INT, 1, 0);
        clutter_shader_effect_set_uniform (effect, "width",
                G_TYPE_FLOAT, 1,
                2.0/clutter_actor_get_width (lg->actor));
        clutter_shader_effect_set_uniform (effect, "height",
                G_TYPE_FLOAT, 1,
                2.0/clutter_actor_get_height (lg->actor));
        clutter_actor_add_effect(lg->actor, lg->effect);
    }
    return TRUE;
}


/**
 * cdos_get_actor_properties:
 * @actor: #ClutterActor
 *
 * 获取 #ClutterActor 的所有属性
 *
 */
void
cdos_get_actor_properties (void *actor)
{
    g_return_if_fail (CLUTTER_IS_ACTOR (actor));
    GParamSpec **spec = NULL;
    guint n_props = 0;
    spec = g_object_class_list_properties (G_OBJECT_GET_CLASS (actor), &n_props);
    guint i;
    for (i = 0; i < n_props; i++)
    {
        GParamSpec *sp = spec[i];

        if (!(sp->flags & G_PARAM_READABLE))
            continue;

        GValue v = {0,};
        g_value_init (&v, sp->value_type);
        g_object_get_property ((GObject *)actor, sp->name, &v);
        switch (sp->value_type)
        {
#define PRINT(type,TYPE,format) do {\
                type m; \
                m=g_value_get_##TYPE(&v);\
                printf("\t%s: " format "\n",sp->name, m);\
            }while(0)

            case G_TYPE_INT:
                PRINT(int, int, "%d");
                break;
            case G_TYPE_INT64:
                PRINT(gint64, int64, "%ld");
                break;
            case G_TYPE_BOOLEAN:
                PRINT(gboolean, boolean, "%d");
                break;
            case G_TYPE_ENUM:
                PRINT(gint, enum, "%d");
                break;
            case G_TYPE_FLAGS:
                PRINT(guint, flags, "%u");
                break;
            case G_TYPE_LONG:
                PRINT(glong, long, "%ld");
                break;
            case G_TYPE_UCHAR:
                PRINT(guchar, uchar, "%c");
                break;
            case G_TYPE_UINT:
                PRINT(guint, uint, "%u");
                break;
            case G_TYPE_UINT64:
                PRINT(guint64, uint64, "%lu");
                break;
            case G_TYPE_ULONG:
                PRINT(gulong, ulong, "%lu");
                break;
            case G_TYPE_CHAR:
                PRINT(gint8, schar, "%c");
                break;
            case G_TYPE_OBJECT:
                {
                    gpointer *obj = g_value_get_object(&v);
                    printf("\t%s: %s %p\n", sp->name, g_type_name(G_OBJECT_TYPE(obj)), obj);
                }
                break;
            case G_TYPE_POINTER:
                PRINT(gpointer, pointer, "%p");
                break;
            case G_TYPE_STRING:
                PRINT(const gchar *, string, "%s");
                break;
            case G_TYPE_FLOAT:
                PRINT(gfloat, float, "%f");
                break;
            case G_TYPE_DOUBLE:
                PRINT(gdouble, double, "%f");
                break;
            default:
            {
                gpointer p = NULL;
                if(G_VALUE_HOLDS_POINTER (&v))
                    p = g_value_get_pointer (&v);
                else if (G_VALUE_HOLDS_BOXED (&v))
                    p = g_value_get_boxed (&v);

                if(sp->value_type == GRAPHENE_TYPE_POINT)
                {
                    graphene_point_t *point = p;
                    printf("\t%s: %f %f\n", sp->name, point->x, point->y);
                }
                else if(sp->value_type == CLUTTER_TYPE_ACTOR_BOX)
                {
                    ClutterActorBox *box = p;
                    printf("\t%s:\n\t    x1=%-4.4f, y1=%-4.4f\n\t    x2=%-4.4f, y2=%-4.4f\n", sp->name,
                            box->x1, box->y1, box->x2, box->y2);
                }
                else if(sp->value_type == GRAPHENE_TYPE_SIZE)
                {
                    graphene_size_t *sz = p;
                    printf("\t%s: w=%f, h=%f\n", sp->name, sz->width, sz->height);
                }
                else if(sp->value_type == GRAPHENE_TYPE_RECT)
                {
                    graphene_rect_t *rect = p;
                    printf("\t%s:\n\t    x=%-4.4f, y=%-4.4f\n\t    w=%-4.4f, h=%-4.4f\n",
                            sp->name, rect->origin.x, rect->origin.y,
                            rect->size.width, rect->size.height);
                }
                else if(sp->value_type == GRAPHENE_TYPE_POINT3D)
                {
                    graphene_point3d_t *vertex = p;
                    printf("\t%s: x=%f, y=%f, z=%f\n", sp->name,
                            vertex->x, vertex->y, vertex->z);
                }
                else if(sp->value_type == CLUTTER_TYPE_MATRIX)
                {
                    CoglMatrix *matrix = p;
                    printf("\t%s:\n", sp->name);
                    printf("\t    %-4.2f %-4.2f %-4.2f %-4.2f\n",
                            matrix->xx, matrix->xy, matrix->xz, matrix->xw);
                    printf("\t    %-4.2f %-4.2f %-4.2f %-4.2f\n",
                            matrix->yx, matrix->yy, matrix->yz, matrix->yw);
                    printf("\t    %-4.2f %-4.2f %-4.2f %-4.2f\n",
                            matrix->zx, matrix->zy, matrix->zz, matrix->zw);
                    printf("\t    %-4.2f %-4.2f %-4.2f %-4.2f\n",
                            matrix->wx, matrix->wy, matrix->wz, matrix->ww);
                }
                else if(G_VALUE_HOLDS_ENUM (&v))
                {
                    gint e = g_value_get_enum (&v);
                    printf("\t%s: %d\n", sp->name, e);
                }
                else if(G_VALUE_HOLDS_FLAGS (&v))
                {
                    guint f = g_value_get_flags(&v);
                    printf("\t%s: %u\n", sp->name, f);
                }
            }
            break;
        }
    }

    g_free (spec);
#undef PRINT
}

static gboolean
on_button_press_event (ClutterActor *actor,
                       ClutterEvent *event,
                       CdosLookingGlass *lg)
{
    graphene_point_t coord;
    clutter_event_get_position (event, &coord);
    //coord.x = event->button.x;
    //coord.y = event->button.y;

    ClutterStage *stage = cdos_global_get_stage (global);
    ClutterActor *target = lg->actor;
    if (lg->actor == NULL)
        target = clutter_stage_get_actor_at_pos (stage, CLUTTER_PICK_ALL,
                                coord.x, coord.y);

    g_message ("[inspect x: %f, y: %f]", coord.x, coord.y);
    g_return_val_if_fail (CLUTTER_IS_ACTOR (target), FALSE);

    printf ("%s: %p\n", g_type_name (G_OBJECT_TYPE (target)), target);

    /* 获取属性 */
    cdos_get_actor_properties (target);

    g_signal_handler_disconnect (cdos_global_get_stage (global), lg->inspect_id);
    lg->inspect_id = 0L;

    ClutterActor *parent = clutter_actor_get_parent (target);
    ClutterActor *child = target;
    while (parent)
    {
        printf("\nparent of %s: %p is %s: %p\n",
                g_type_name(G_OBJECT_TYPE(child)), child,
                g_type_name(G_OBJECT_TYPE(parent)), parent);
        child = parent;
        parent = clutter_actor_get_parent (child);
    }
    if (CLUTTER_IS_EFFECT (lg->effect))
    {
        clutter_actor_remove_effect (lg->actor, lg->effect);
        while (CLUTTER_IS_EFFECT (lg->effect))
            g_object_unref (lg->effect);
    }
    lg->actor = NULL;
    lg->effect = NULL;
    lg->orig = NULL;

    cdos_pop_modal (lg->icon, 0);
    end_inspect (lg);

    return TRUE;
}

static gboolean
start_inspect (ClutterActor *icon,
               ClutterEvent *event,
               CdosLookingGlass *lg)
{
    cdos_push_modal (icon, 0, 0, CDOS_ACTION_MODE_NONE);
    ClutterStage *stage = cdos_global_get_stage (global);
    lg->inspect_id = g_signal_connect (stage, "captured-event",
            G_CALLBACK (on_captured_event), lg);
    return TRUE;
}

/**
 * cdos_looking_glass_init:
 *
 * 新建一个 CdosLookingGlass
 *
 */
void
cdos_looking_glass_init (void)
{
    CdosLookingGlass *lg = g_new0 (CdosLookingGlass, 1);
    ClutterActor *inspect_icon = st_icon_new ();
    clutter_actor_set_reactive (inspect_icon, 1);

    st_icon_set_icon_name ((StIcon *)inspect_icon, "gtk-color-picker");
    st_icon_set_icon_size ((StIcon *)inspect_icon, 24);

    ClutterActor *ui_group = cdos_ui_plugin_get_ui_group ();
    clutter_actor_add_child (ui_group, inspect_icon);

    lg->icon = inspect_icon;
    lg->actor = inspect_icon;
    start_inspect (lg->icon, NULL, lg);
}
