/** 
 * Copyright (c) 2016-2018 CPU and Fundamental Software Research Center, CAS
 *  
 * This software is published by the license of CPU-OS License，you can use and 
 * distibuted this software under this License. See CPU-OS Liense for more detail.
 *
 * You should have receive a copy of CPU-OS License. If not, please contact us 
 * by email <support_os@cpu-os.ac.cn>  
 * 
 **/
#include <clutter/clutter.h>
#include <gio/gio.h>
#include <meta/meta-shaped-texture.h>
#include <meta/meta-cursor-tracker.h>

#include "cdos-texture-content.h"
#include "cdos-magnifier.h"
#include "cdos-zoom-region.h"

#define     SHOW_KEY        "screen-magnifier-enabled"
#define     MOUSE_POLL_FREQUENCY        15

struct _CdosMagnifierPrivate
{
    ClutterActor *mouse_sprite;
    ClutterActor *cursor_root;
    CdosGlobal *global;

    gboolean sys_mouse_over_any;
};

G_DEFINE_TYPE_WITH_PRIVATE (CdosMagnifier, cdos_magnifier, G_TYPE_OBJECT);

static gboolean
cdos_magnifier_on_settings_changed (GSettings *settings, gchar *key, gpointer user_data);

static void
cdos_magnifier_dispose (GObject *magnifier)
{
    CdosMagnifier *self = (CdosMagnifier *)magnifier;
    CdosMagnifierPrivate *priv;
    priv = self->priv;

    g_clear_object (&self->cursor_tracker);
    g_clear_pointer (&priv->cursor_root, clutter_actor_destroy);
    g_clear_pointer (&priv->mouse_sprite, clutter_actor_destroy);

    g_list_foreach (self->zoom_regions, (GFunc)g_object_unref, NULL);
    self->zoom_regions = NULL;
    G_OBJECT_CLASS (cdos_magnifier_parent_class)->dispose (magnifier);
}

static void
cdos_magnifier_finalize (GObject *magnifier)
{
    G_OBJECT_CLASS (cdos_magnifier_parent_class)->finalize (magnifier);
}

static void
cdos_magnifier_class_init (CdosMagnifierClass *klass)
{
    GObjectClass *gobject_class = G_OBJECT_CLASS (klass);

    gobject_class->finalize = cdos_magnifier_finalize;
    gobject_class->dispose = cdos_magnifier_dispose;
}

static void
cdos_util_cursor_tracker_to_clutter (MetaCursorTracker *tracker,
                                     ClutterActor      *actor)
{
  CoglTexture *sprite;

  sprite = meta_cursor_tracker_get_sprite (tracker);
  if (sprite)
    {
      clutter_actor_show (actor);
    }
  else
    {
      clutter_actor_hide (actor);
    }
}

static gboolean
cdos_magnifier_update_mouse_sprite (MetaCursorTracker *tracker, gpointer user_data)
{
    CdosMagnifier *self = (CdosMagnifier *)user_data;
    CdosMagnifierPrivate *priv;
    priv = self->priv;

    cdos_util_cursor_tracker_to_clutter (tracker, priv->mouse_sprite);
    gint xhot, yhot;

    meta_cursor_tracker_get_hot (tracker, &xhot, &yhot);

    return FALSE;
}

static gboolean
cdos_magnifier_settings_init (CdosMagnifier *self, CdosZoomRegion *zoom_region)
{
    if (zoom_region != NULL)
    {
//        gfloat mag_factor = g_settings_get_double (self->zoom_regions, "mag-factor");
//        g_debug ("magnifier: %f\n", mag_factor);
//        cdos_zoom_region_set_mag_factor (zoom_region, mag_factor);
        cdos_zoom_region_set_clamp_scrolling_at_edges (zoom_region, FALSE);

        cdos_zoom_region_set_magnifier (zoom_region, self);// FIXME
        //cdos_magnifier_update_zoom_region (magnifier, zoom_region);

        cdos_zoom_region_set_mouse_tracking_mode (zoom_region, CDOS_MOUSE_TRACKING_MODE_PUSH);

    //    g_signal_connect (self->zoom_settings, "changed::mag-factor", 
    //                                G_CALLBACK (cdos_magnifier_on_settings_changed), self);

    }
    return FALSE;
}

static void
cdos_magnifier_show_system_cursor (CdosMagnifier *self)
{
    meta_cursor_tracker_set_pointer_visible (self->cursor_tracker, TRUE); 
    //cdos_xfixes_cursor_show (self->xfixes_cursor);
}

void cdos_magnifier_hide_system_cursor (CdosMagnifier *self)
{
    meta_cursor_tracker_set_pointer_visible (self->cursor_tracker, FALSE); 
    //cdos_xfixes_cursor_hide (self->xfixes_cursor);
}

static void
active_each_zoom_region (gpointer data, gpointer user_data)
{
    CdosZoomRegion *zoom_region = (CdosZoomRegion *)data;
    CdosMagnifier *self = (CdosMagnifier *)user_data;
    CdosMagnifierPrivate *priv;
    priv = self->priv;

    if (self->enabled)
    {
        cdos_zoom_region_set_mouse_source_actor (zoom_region, priv->cursor_root);
    }
    cdos_zoom_region_set_active (zoom_region, self->enabled);
    
    if (!self->enabled){
        cdos_zoom_region_remove_mouse_source_actor (zoom_region);
    }
}

static void
each_zoom_region_scroll_mouse (gpointer data, gpointer user_data)
{
    CdosZoomRegion *zoom_region = (CdosZoomRegion *)data;
    CdosMagnifier *self = (CdosMagnifier *)user_data;
    CdosMagnifierPrivate *priv;
    priv = self->priv;

    if (cdos_zoom_region_scroll_to_mouse_position (zoom_region))
    {
        priv->sys_mouse_over_any = TRUE;        
    }
    //gint x_pos, y_pos;
    //cdos_zoom_region_screen_to_viewport (zoom_region, self->x_mouse, self->y_mouse, &x_pos, &y_pos);
    //clutter_actor_set_position (priv->cursor_root, x_pos, y_pos);
    //clutter_actor_show (priv->cursor_root);
    //cdos_magnifier_show_system_cursor (self);
}



static gboolean
cdos_magnifier_scroll_to_mouse_pos (CdosMagnifier *self)
{
    gint x_mouse, y_mouse;
    ClutterModifierType mask;
    CdosMagnifierPrivate *priv;
    priv = self->priv;

    cdos_global_get_pointer (priv->global, &x_mouse, &y_mouse, &mask);
    
    if (x_mouse != self->x_mouse || y_mouse != self->y_mouse)
    {
        self->x_mouse = x_mouse;
        self->y_mouse = y_mouse;

        priv->sys_mouse_over_any = FALSE;
        g_list_foreach (self->zoom_regions, (GFunc)each_zoom_region_scroll_mouse, self);
    }

    // 这里不可以直接 show cursor，没有进行坐标转换，位置不对 
    //cdos_magnifier_show_system_cursor (self);
    // cdos_zoom_region_update_mouse_position ();
    //cdos_zoom_region_screen_to_viewport ()
    if (priv->sys_mouse_over_any)
    {
        // 在放大器情况下，不显示原本的系统鼠标，显示的是 mag-mouse
        //g_debug ("hide cursor\n");
        cdos_magnifier_hide_system_cursor (self);
    }else {
        cdos_magnifier_show_system_cursor (self);
    }
    return TRUE;
}

static void
cdos_magnifier_start_tracking_mouse (CdosMagnifier *self)
{
    if (self->mouse_tracking_id == 0)
    {
        self->mouse_tracking_id = g_timeout_add (MOUSE_POLL_FREQUENCY, 
                                                 (GSourceFunc)cdos_magnifier_scroll_to_mouse_pos, self); 
    }
}

static void
cdos_magnifier_stop_tracking_mouse (CdosMagnifier *self)
{
    if (self->mouse_tracking_id != 0)
    {
        g_source_remove (self->mouse_tracking_id);
    }
    self->mouse_tracking_id = 0;
}

static void
cdos_magnifier_set_active (CdosMagnifier *self, gboolean enabled)
{
    CdosMagnifierPrivate *priv;
    priv = self->priv;
   

    // cursor-root，每次激活的时候传给 zoom-region，这样每次激活都可以看到鼠标
    // 问题是退出 zoom 后，右下角有个小鼠标
    // 原因 maybe zoom-region 的 mouse-source-actor 没有被销毁(（大误）)
    // 实际是因为 connect cursor-tracker 的回调没有被停止，disconnect 之后就好了
    if (enabled)
    {
        priv->mouse_sprite = clutter_actor_new ();
        cdos_util_cursor_tracker_to_clutter (self->cursor_tracker, priv->mouse_sprite);

        priv->cursor_root = clutter_actor_new ();
        clutter_actor_add_child (priv->cursor_root, priv->mouse_sprite);
        self->cursor_tracker_id = g_signal_connect (self->cursor_tracker, "cursor-changed", G_CALLBACK (cdos_magnifier_update_mouse_sprite), self);
    } else {
        if (priv->mouse_sprite)
        {
            clutter_actor_destroy (priv->mouse_sprite);
            priv->mouse_sprite = NULL;
        }
        if (priv->cursor_root)
        {
            clutter_actor_destroy (priv->cursor_root);
            priv->cursor_root = NULL;
        }
        if (self->cursor_tracker_id)
        {
            g_signal_handler_disconnect (self->cursor_tracker, self->cursor_tracker_id);
            self->cursor_tracker_id = 0;
        }
    }
    g_return_if_fail (self->zoom_regions != NULL);
    g_list_foreach (self->zoom_regions, active_each_zoom_region, self);

    if (enabled)
    {
        //start tracking mouse
        cdos_magnifier_start_tracking_mouse (self);
    }else
    {
        //stop tracking mouse
        cdos_magnifier_stop_tracking_mouse (self);
        
    }

    if (!enabled)
    {
       //cdos_xfixes_cursor_show (self->xfixes_cursor);
        meta_cursor_tracker_set_pointer_visible (self->cursor_tracker, TRUE); 
    }

}

static gboolean
cdos_magnifier_on_settings_changed (GSettings *settings, gchar *key, gpointer user_data)
{
    CdosMagnifier *self = (CdosMagnifier *)user_data;

    if (g_strcmp0 (key, SHOW_KEY) == 0)
    {
        self->enabled = g_settings_get_boolean (self->settings, SHOW_KEY);

        cdos_magnifier_set_active (self, self->enabled);
        /*if (!self->enabled)
        {
            ClutterActor *ui_group = monkey_plugin_get_ui_group (monkey_plugin_get ());
            clutter_actor_remove_child (ui_group, priv->cursor_root);
        }*/
    }
    
    if (g_strcmp0 (key, "mag-factor") == 0)
    {
        if (self->zoom_regions != NULL)
        {
            gfloat mag_factor = g_settings_get_double (self->zoom_settings, "mag-factor");
            GList *zoom_region = g_list_first (self->zoom_regions);
            cdos_zoom_region_set_mag_factor (zoom_region->data, mag_factor);
            cdos_magnifier_set_active (self, self->enabled && mag_factor > 1.0); 
            //cdos_magnifier_set_active (self, TRUE); 
        }
    }
    return FALSE;
}

static void
cdos_magnifier_init (CdosMagnifier *self)
{
    CdosMagnifierPrivate *priv;
    priv = self->priv = cdos_magnifier_get_instance_private (self);


    self->zoom_regions = NULL;
    priv->global = cdos_global_get ();
    MetaDisplay *display = cdos_global_get_display (cdos_global_get ());
    MetaCursorTracker *tracker = meta_cursor_tracker_get_for_display (display);
    //CdosXFixesCursor *xfixes_cursor = cdos_xfixes_cursor_get_for_stage (stage);
    
    /*priv->mouse_sprite = clutter_texture_new ();
    cdos_util_cursor_tracker_to_clutter (tracker, priv->mouse_sprite);
    //cdos_xfixes_cursor_update_texture_image (xfixes_cursor, CLUTTER_TEXTURE (priv->mouse_sprite));

    priv->cursor_root = clutter_actor_new ();
    clutter_actor_add_child (priv->cursor_root, priv->mouse_sprite);*/

    //ClutterActor *ui_group = monkey_plugin_get_ui_group (monkey_plugin_get ());
    //clutter_actor_add_child (ui_group, priv->cursor_root);

    gint x_mouse, y_mouse;
    ClutterModifierType mask;
    cdos_global_get_pointer (priv->global, &x_mouse, &y_mouse, &mask);

    CdosZoomRegion *a_zoom_region = cdos_zoom_region_new (NULL, NULL);
    //cdos_zoom_region_set_mouse_source_actor (a_zoom_region, priv->cursor_root);

    self->zoom_regions = g_list_append (self->zoom_regions, a_zoom_region);

    self->cursor_tracker = tracker;
    self->settings = g_settings_new ("org.cdos.desktop.a11y.applications");
    self->zoom_settings = g_settings_new ("org.cdos.desktop.a11y.magnifier");

    g_signal_connect (self->settings, "changed::screen-magnifier-enabled", 
                                    G_CALLBACK (cdos_magnifier_on_settings_changed), self);


    self->enabled = g_settings_get_boolean (self->settings, SHOW_KEY);
    self->x_mouse = 0;
    self->y_mouse = 0;

    cdos_magnifier_settings_init (self, a_zoom_region);
    cdos_magnifier_set_active (self, self->enabled);
}

CdosMagnifier *
cdos_magnifier_new (void)
{
    CdosMagnifier *magnifier = g_object_new (CDOS_TYPE_MAGNIFIER, NULL);
    return magnifier;
}
