/**
 * Copyright (c) 2016-2018 CPU and Fundamental Software Research Center, CAS
 *
 * This software is published by the license of CPU-OS License，you can use and
 * distibuted this software under this License. See CPU-OS Liense for more detail.
 *
 * You should have receive a copy of CPU-OS License. If not, please contact us
 * by email <support_os@cpu-os.ac.cn>
 *
 **/
/**
 * SECTION: cdos-message-tray
 * @short_description: 消息展示与处理.
 * @include: cdos-message-tray.h
 *
 * CdosMessageTrayClass 类用来展示消息,转发或者删除消息.
 */

#include "cdos-message-tray.h"

#include "cdos-global.h"
#include "cdos-notification.h"
#include "cdos-source.h"
#include "cdos-layout-manager.h"
#include "cdos-start.h"
#include "cdos-notification-daemon-source.h"
#include "cdos-sound-manager.h"
#include "cdos-screen-saver.h"

#define NOTIFICATION_TIMEOUT 3000
#define NOTIFICATION_CRITICAL_TIMEOUT 6000
#define NOTIFICATION_ANIMATION_TIME 300

static gboolean want_send_to_applet = TRUE;

G_DEFINE_TYPE (CdosMessageTray, cdos_message_tray, ST_TYPE_WIDGET);

enum
{
    NOTIFIY_APPLET_UPDATE,
    SOURCE_ADDED,
    NOTIFIY_APPLET_LAST_SIGNAL
};

static guint cdos_message_tray_signals [NOTIFIY_APPLET_LAST_SIGNAL] = {0};

/**
 * MessageTrayCtate:
 * @HIDDEN: 表示已经隐藏
 * @SHOWING: 动画过程中,正在显示
 * @SHOWN: 已经显示
 * @HIDING: 动画过程中,正在隐藏
 *
 * CdosMessageTray 的状态枚举变量
 */
typedef enum _MessageTrayCtate
{
    HIDDEN = 0,
    SHOWING,
    SHOWN,
    HIDING
} MessageTrayCtate;

typedef struct _CategorySound
{
    const char *category;
    const char *sound;
} CategorySound;

static CategorySound sounds [] =
{
    {"device", NULL},
    {"device.added", "device-added"},
    {"device.error", NULL},
    {"device.removed", "device-removed"},
    {"email", NULL},
    {"email.arrived", "message-new-email"},
    {"email.bounced", NULL},
    {"im", "message"},
    {"im.error", NULL},
    {"im.received", "message-new-instant"},
    {"network", NULL},
    {"network.connect", "network-connectivity-established"},
    {"network.disconnect", "network-connectivity-lost"},
    {"network.error", NULL},
    {"presence", NULL},
    {"presence.offline", "service-login"},
    {"presence.online", "service-logout"},
    {"transfer", NULL},
    {"transfer.complete", "complete"},
    {"transfer.error", NULL}
};

static gint cdos_message_tray_get_source_index (CdosMessageTray *self, CdosSource *source);
static void cdos_message_tray_show_notification (CdosMessageTray *self);
static void cdos_message_tray_update_showing_notification (CdosMessageTray *self);
static void cdos_message_tray_update_notification_timeout (CdosMessageTray *self, guint timeout);
static void cdos_message_tray_hide_notification (CdosMessageTray *self);

/* Private */
static gint cdos_notification_daemon_source_level (CdosNotification *notification)
{
// check notification level
    CdosSource *source = notification->source;
    if (CDOS_IS_NOTIFICATION_DAEMON_SOURCE (source))
        return cdos_notification_daemon_source_get_level ((CdosNotificationDaemonSource *)source);
    return -1;
}

static void
cdos_message_tray_update_state (CdosMessageTray *self)
{
    gboolean pending = FALSE;
    gboolean expired = FALSE;
    gboolean show = FALSE;

    CdosNotification *notification = NULL;
    if (self->notification_queue)
        notification = g_queue_peek_tail (self->notification_queue);

    pending = notification != NULL;

    expired = (self->notification_timeout_id == 0 &&
               !(self->notification && self->notification->urgency == CDOS_NOTIFICATION_CRITICAL)
              ) || self->notification_removed;

    // 消息不为空，且不是勿扰模式
    show = pending && self->notifications_enabled;
    g_debug ("%s::notification state = %d, show = %d, pending= %d, enabled = %d, dnd = %d.", __func__, self->notification_state, show, pending, self->notifications_enabled, self->dnd);
    if (self->notification_state == HIDDEN)
    {
        if (show)
        {
            // 如果通知面板弹出的情况下，不显示横幅
            if (self->dnd)
            {
                self->notification = g_queue_pop_tail (self->notification_queue);
                clutter_actor_add_child (CLUTTER_ACTOR (self), CLUTTER_ACTOR (self->notification));
                cdos_message_tray_update_notification_timeout (self, 0);
                self->notification_state = SHOWN;
                self->notification_timeout_id = 0;
                cdos_message_tray_update_state (self);
            } else {
                cdos_message_tray_show_notification (self);
            }
        }
        else if (pending || !self->notifications_enabled)
        {
            self->notification = g_queue_pop_tail (self->notification_queue);
            clutter_actor_add_child (CLUTTER_ACTOR (self), CLUTTER_ACTOR (self->notification));
            g_signal_handlers_disconnect_by_data (notification, self);
            g_signal_emit_by_name (self, "notify-applet-update", self->notification);
        }
    }
    else if (self->notification_state == SHOWN)
    {
        if (expired)
        {
            cdos_message_tray_hide_notification (self);
        }
    }
}

static gint
cdos_message_tray_get_source_index (CdosMessageTray *self, CdosSource *source)
{
    gint index = g_list_index (self->sources, source);
    return index;
}

static void
cdos_message_tray_show_notification_completed_cb (ClutterTimeline *tml,
                                                  CdosMessageTray *self)
{
    if (self->notification_state != SHOWING)
        return;

    if (self->notification == NULL)
    {
        // 这里为空的原因是通知在 showing 阶段就被销毁掉了
        // 比如不停的连接、断开有线网络
        cdos_message_tray_update_notification_timeout (self, 0);
        self->notification_state = SHOWN;
        cdos_message_tray_update_state (self);
        return;
    }
    self->notification_state = SHOWN;
    self->notification_timeout_id = 0;

    if (self->notification_removed)
    {
        cdos_message_tray_update_notification_timeout (self, 0);
        cdos_message_tray_update_state (self);
        return;
    }

    if (self->notification->urgency != CDOS_NOTIFICATION_CRITICAL)
        cdos_message_tray_update_notification_timeout (self, NOTIFICATION_TIMEOUT);
    else
        cdos_message_tray_update_notification_timeout (self, NOTIFICATION_CRITICAL_TIMEOUT);
}

static void
cdos_message_tray_update_showing_notification (CdosMessageTray *self)
{
    ClutterActor *actor = CLUTTER_ACTOR (self);
    clutter_actor_remove_all_transitions (actor);

    self->notification_state = SHOWING;
    CdosLayoutManager *layout_manager = cdos_layout_manager_get_default ();
    const MetaRectangle *monitor = cdos_layout_manager_get_primary_monitor (layout_manager);
    if (!monitor)
    {
        cdos_message_tray_show_notification_completed_cb (NULL, self);
        return;
    }

    //锁屏时不展示通知
    if (cdos_screen_saver_get_active ())
    {
        g_message ("Don't show notification, the screensaver is active.");
        cdos_message_tray_show_notification_completed_cb (NULL, self);
        return;
    }

    gfloat x = monitor->x + monitor->width - clutter_actor_get_width (actor) -12;
    self->x = x;

    ClutterTransition *trans;
    trans = cdos_transition_simple (actor,
                                    CLUTTER_EASE_OUT_QUAD,
                                    NOTIFICATION_ANIMATION_TIME,
                                    "opacity", 255,
                                    "x", x,
                                    NULL);
    g_signal_connect (trans, "completed", G_CALLBACK (cdos_message_tray_show_notification_completed_cb), self);
}

static const char *
get_sound_by_category (const char *category)
{
    for (int i = 0; i < G_N_ELEMENTS (sounds); i++)
    {
        CategorySound *sound = &sounds[i];
        if (g_strcmp0 (sound->category, category) == 0)
            return sound->sound;
    }
    return NULL;
}

static void
cdos_message_tray_show_notification (CdosMessageTray *self)
{
    ClutterActor *actor = CLUTTER_ACTOR (self);
    const MetaRectangle *monitor;
    CdosPanel *panel = cdos_ui_plugin_get_panel ();

    self->notification_timeout_id = 0;

    self->notification = g_queue_pop_tail (self->notification_queue);

    CdosNotificationDaemonSourceLevel level = CDOS_NOTIFICATION_DAEMON_SOURCE_DND;
    level = cdos_notification_daemon_source_level (self->notification);
    switch (level)
    {
        case CDOS_NOTIFICATION_DAEMON_SOURCE_CLEAN:
            cdos_message_tray_update_notification_timeout (self, 0);
            self->notification_state = SHOWN;
            self->notification_timeout_id = 0;
            cdos_message_tray_update_state (self);
            return;
        case CDOS_NOTIFICATION_DAEMON_SOURCE_DND:
            //FIXME 不显示横幅
            cdos_message_tray_update_notification_timeout (self, 0);
            self->notification_state = SHOWN;
            self->notification_timeout_id = 0;
            cdos_message_tray_update_state (self);
            // 直接 return 有个问题， dnd 模式应该有声音的
            break;
        default:
            break;
    }
    if (level != CDOS_NOTIFICATION_DAEMON_SOURCE_SILENT)
    {
        const gchar *sound = self->notification->sound_name;
        if (sound && *sound != '\0')
        {
            g_debug ("sound name: %s\n", sound);
            //cdos_global_play_theme_sound (global, 0, sound, NULL, NULL);
            //cdos_sound_manager_play (sound);
            cdos_global_play_theme_sound (global, 0, sound, NULL, NULL);
            goto play_sound_end;
        }
        sound = self->notification->sound_file;
        if (sound && *sound != '\0')
        {
            cdos_global_play_sound_file (global, 0, sound, NULL, NULL);
            goto play_sound_end;
        }
        sound = get_sound_by_category (self->notification->category);
        if (sound && *sound != '\0')
        {
            cdos_global_play_theme_sound (global, 0, sound, NULL, NULL);
            goto play_sound_end;
        }
        sound = "logout";
        cdos_sound_manager_play (sound);
    }

play_sound_end:
    if (level != CDOS_NOTIFICATION_DAEMON_SOURCE_DND)
    {
        clutter_actor_add_child (actor, CLUTTER_ACTOR (self->notification));

        clutter_actor_set_opacity (actor, 0);

        CdosLayoutManager *layout_manager = cdos_layout_manager_get_default ();
        monitor = cdos_layout_manager_get_primary_monitor (layout_manager);
        if (monitor)
        {
            gfloat y = monitor->y + monitor->height
                       - clutter_actor_get_height (actor)
                       - clutter_actor_get_height (CLUTTER_ACTOR(panel))
                       - 12;
            clutter_actor_set_y (actor, y);
            clutter_actor_set_x (actor, monitor->x + monitor->width);
            clutter_actor_show (actor);
        }

        cdos_message_tray_update_showing_notification (self);
    }
}

static gint
notification_compare_func (CdosNotification *notification1,
                           CdosNotification *notification2)
{
    return (notification2->urgency - notification1->urgency);
}

static void
cdos_message_tray_on_notification_destroy_cb (CdosNotification *notification,
                                              CdosMessageTray *self)
{
    g_queue_remove (self->notification_queue, notification);

    if (self->notification == notification)
    {
        cdos_message_tray_update_notification_timeout (self, 0);
        self->notification_removed = TRUE;
        self->notification = NULL;
        cdos_message_tray_update_state (self);

        return;
    }

}

static void
cdos_message_tray_on_notify_cb (CdosSource *source,
                                CdosNotification *notification,
                                CdosMessageTray *self)
{
    // If a notification that is being shown is updated, we update
    // how it is shown and extend the time until it auto-hides.
    // If a new notification is updated while it is being hidden,
    // we stop hiding it and show it again.

    if (self->notification == notification)
    {
        cdos_message_tray_update_showing_notification (self);
    }
    else if (g_queue_index (self->notification_queue, notification) == -1)
    {
        g_signal_connect (notification, "destroy", G_CALLBACK (cdos_message_tray_on_notification_destroy_cb), self);
        g_queue_push_head (self->notification_queue, notification);
        g_queue_sort (self->notification_queue, (GCompareDataFunc)notification_compare_func, NULL);
    }
    cdos_message_tray_update_state (self);
}

static void
cdos_message_tray_on_source_destroy_cb (CdosSource *source, gpointer user_data)
{
    CdosMessageTray *self = CDOS_MESSAGE_TRAY (user_data);

    gint index = cdos_message_tray_get_source_index (self, source);
    if (index == -1) {
        return ;
    }

    self->sources = g_list_remove (self->sources, source);

    gboolean needUpdate = FALSE;

    if (self->notification != NULL && self->notification->source == source)
    {
        cdos_message_tray_update_notification_timeout (self, 0);
        //self->notification_removed = TRUE;
        needUpdate = TRUE;
    }

    if (needUpdate)
        cdos_message_tray_update_state (self);
}

static gboolean
cdos_message_tray_notification_timeout_cb (CdosMessageTray *self)
{
    self->notification_timeout_id = 0;
    cdos_message_tray_update_state (self);
    return FALSE;
}

static void
cdos_message_tray_update_notification_timeout (CdosMessageTray *self, guint timeout)
{
    if (self->notification_timeout_id)
    {
        g_source_remove (self->notification_timeout_id);
        self->notification_timeout_id = 0;
    }

    if (timeout > 0)
        self->notification_timeout_id =  g_timeout_add (timeout, (GSourceFunc)cdos_message_tray_notification_timeout_cb, self);
}

static void
cdos_message_tray_hide_notification_completed_cb (ClutterTimeline *tml, CdosMessageTray *self)
{
    if (self->notification_state == HIDDEN)
        return;
    self->notification_state = HIDDEN;

    clutter_actor_hide (CLUTTER_ACTOR (self));

    if(self->notification_clicked_id > 0)
    {
        g_signal_handler_disconnect(self->notification, self->notification_clicked_id);
        self->notification_clicked_id = 0;
    }
    CdosNotification *notification = self->notification;
    g_debug ("%s::notification hide completed, removed = %d.", __func__, self->notification_removed);
    if (notification && !self->notification_removed)
    {
        g_signal_handlers_disconnect_by_data (notification, self);
        g_signal_emit_by_name (self, "notify-applet-update", notification);
    }
    else if (notification != NULL)
    {
        cdos_notification_destroy (notification, CDOS_NOTIFICATION_EXPIRED);
    }

    self->notification = NULL;
    self->notification_removed = FALSE;
    cdos_message_tray_update_state (self);
}

static void
cdos_message_tray_hide_notification_stopped (ClutterTimeline *tml, gboolean is_finished, CdosMessageTray *self)
{
    if (!is_finished)
        cdos_message_tray_hide_notification_completed_cb (tml, self);
}

static void
cdos_message_tray_hide_notification (CdosMessageTray *self)
{
    clutter_actor_remove_all_transitions ((ClutterActor *)self);

    CdosLayoutManager *layout_manager = cdos_layout_manager_get_default ();
    const MetaRectangle *monitor = cdos_layout_manager_get_primary_monitor (layout_manager);
    self->notification_state = HIDING;
    if (!monitor)
    {
        cdos_message_tray_hide_notification_completed_cb (NULL, self);
        return;
    }

    if (self->notification)
    {
        if (clutter_actor_get_parent (CLUTTER_ACTOR (self->notification)) == NULL)
            clutter_actor_add_child (CLUTTER_ACTOR (self), CLUTTER_ACTOR (self->notification));
    }
    ClutterTransition *trans;
    trans = cdos_transition_simple (CLUTTER_ACTOR (self),
                                    CLUTTER_EASE_OUT_QUAD,
                                    NOTIFICATION_ANIMATION_TIME,
                                    "opacity", 0,
                                    "x", (gfloat) (monitor->x + monitor->width),
                                    NULL);
    g_signal_connect (trans, "completed", G_CALLBACK (cdos_message_tray_hide_notification_completed_cb), self);
    g_signal_connect (trans, "stopped", G_CALLBACK (cdos_message_tray_hide_notification_stopped), self);
}

static void
cdos_message_tray_on_notification_enabled_updated_cb (GSettings *settings,
                                                      GParamSpec *pspec,
                                                      CdosMessageTray *self)
{
    self->notifications_enabled = g_settings_get_boolean (settings, "display-notifications");
}

/* public */
/**
  * cdos_message_tray_contains:
  * @self: A #CdosMessageTray
  * @source: #CdosSource 类型的参数
  *
  * 判断source已经存在于 (#GList)self->sources 里。
  *
  * Returns (boolean): 若存在，返回%TRUE；否则，返回%FALSE。
  */
gboolean
cdos_message_tray_contains (CdosMessageTray *self, CdosSource *source)
{
    g_return_val_if_fail (CDOS_IS_MESSAGE_TRAY(self),FALSE);
    gboolean res = cdos_message_tray_get_source_index (self, source) >= 0;

    return res;
}

/**
  * cdos_message_tray_add:
  * @self: A #CdosMessageTray
  * @source: #CdosSource 参数,用来绑定信号.
  *
  * 判断source是否在 #CdosMessageTray 成员变量sources里：如果存在就退出，不存在就将对source绑定信号。
  */
void
cdos_message_tray_add (CdosMessageTray *self, CdosSource *source)
{
    g_return_if_fail (CDOS_IS_MESSAGE_TRAY (self));
    g_return_if_fail (CDOS_IS_SOURCE (source));
    if (cdos_message_tray_contains (self, source)) {
        g_warning ("Trying to re-add source %s", source->name);
        return ;
    }
    self->sources = g_list_prepend (self->sources, source);
    g_signal_connect (source, "notif", G_CALLBACK (cdos_message_tray_on_notify_cb), self);
    g_signal_connect (source, "destroy", G_CALLBACK (cdos_message_tray_on_source_destroy_cb), self);
    g_signal_emit (self, cdos_message_tray_signals[SOURCE_ADDED], 0, source);
}

void cdos_message_tray_set_dnd (CdosMessageTray *self, gboolean dnd)
{
    // 配合 applet notification，当面板弹出时，新通知不弹出横幅
    self->dnd = dnd;
    if (dnd)
    {
        self->notification_timeout_id = 0;
        cdos_message_tray_update_state (self);
        //clutter_actor_hide (CLUTTER_ACTOR (self));
    } else {
        //clutter_actor_show (CLUTTER_ACTOR (self));
        cdos_message_tray_update_notification_timeout (self, NOTIFICATION_TIMEOUT);
    }
}

/**
  * cdos_message_tray_toggle:
  * @self: A #CdosMessageTray
  *
  * 切换显示 #CdosMessageTray 实例
  *
  */
void cdos_message_tray_toggle (CdosMessageTray *self)
{
    g_return_if_fail(CDOS_IS_MESSAGE_TRAY(self));
    cdos_message_tray_update_state (self);
}

/**
  * cdos_message_tray_hide:
  * @self: A #CdosMessageTray
  *
  * 隐藏 #CdosMessageTray 的实例
  *
  */
void cdos_message_tray_hide (CdosMessageTray *self)
{
    g_return_if_fail(CDOS_IS_MESSAGE_TRAY(self));
    cdos_message_tray_update_state (self);
}

/**
 * cdos_message_tray_want_send_to_applet:
 * @willSend: 是否需要通知发送到applet
 *
 * 是否发送到applet
 *
 */
void cdos_message_tray_want_send_to_applet (gboolean willSend)
{
    want_send_to_applet = willSend;
}

static void
cdos_message_tray_dispose (GObject *object)
{
    CdosMessageTray * self = CDOS_MESSAGE_TRAY(object);
    CdosLayoutManager *layout_manager = cdos_layout_manager_get_default ();
    cdos_layout_manager_remove_chrome (layout_manager, CLUTTER_ACTOR(self));
    if(self->notification_clicked_id > 0)
    {
        g_signal_handler_disconnect(self->notification,self->notification_clicked_id);
        self->notification_clicked_id = 0;
    }
    g_clear_pointer (&self->notification_queue, g_queue_free);
    g_clear_pointer (&self->sources, g_list_free);

    G_OBJECT_CLASS(cdos_message_tray_parent_class)->dispose(object);
}

static void
cdos_message_tray_get_preferred_width (ClutterActor *actor,
                                       gfloat        for_height,
                                       gfloat       *min_width_p,
                                       gfloat       *natural_width_p)
{
    StThemeNode *theme_node = st_widget_get_theme_node ((StWidget *)actor);
    st_theme_node_adjust_for_height (theme_node, &for_height);
    st_theme_node_adjust_preferred_width (theme_node, min_width_p, natural_width_p);
}

static void
cdos_message_tray_get_preferred_height (ClutterActor *actor,
                                        gfloat        for_width,
                                        gfloat       *min_height_p,
                                        gfloat       *natural_height_p)
{
    StThemeNode *theme_node = st_widget_get_theme_node ((StWidget *)actor);
    st_theme_node_adjust_for_width (theme_node, &for_width);

    gfloat min_h = 0, nat_h = 0;
    ClutterActor *child = clutter_actor_get_first_child (actor);
    if (child)
        clutter_actor_get_preferred_height (child, for_width, &min_h, &nat_h);
    if (min_height_p)
        *min_height_p = min_h;
    if (natural_height_p)
        *natural_height_p = nat_h;

    st_theme_node_adjust_preferred_height (theme_node, min_height_p, natural_height_p);
}

static void
cdos_message_tray_allocate (ClutterActor *actor,
                            const ClutterActorBox *box,
                            ClutterAllocationFlags flags)
{
    StThemeNode *theme_node = st_widget_get_theme_node ((StWidget *)actor);
    ClutterActorBox content_box;
    clutter_actor_set_allocation (actor, box, flags);

    st_theme_node_get_content_box (theme_node, box, &content_box);
    ClutterActor *child = clutter_actor_get_first_child (actor);
    if (child)
        clutter_actor_allocate (child, &content_box, flags);
}

static void
on_allocation_changed (ClutterActor *actor,
                       ClutterActorBox *box,
                       ClutterAllocationFlags flags,
                       CdosMessageTray *self)
{
    gfloat x;
    gfloat width;
    x = box->x1;
    width = box->x2 - box->x1;
    clutter_actor_set_clip (actor,
                            0, 0,
                            width - (x - self->x),
                            box->y2 - box->y1);
}

static void
cdos_message_tray_class_init (CdosMessageTrayClass *klass)
{
    GObjectClass *gobject_class = G_OBJECT_CLASS(klass);
    //gobject_class->finalize = cdos_message_tray_finalize;
    gobject_class->dispose = cdos_message_tray_dispose;

    ClutterActorClass *actor_class = CLUTTER_ACTOR_CLASS (klass);
    actor_class->get_preferred_width = cdos_message_tray_get_preferred_width;
    actor_class->get_preferred_height = cdos_message_tray_get_preferred_height;
    actor_class->allocate = cdos_message_tray_allocate;
    /**
     * CdosMessageTray::notify-applet-update:
     * @object (GObject): 绑定信号的对象
     * @notification (CdosNotification): 发送的notification
     * @user_data (gpointer): 真对回调函数所传入的参数
     *
     * 当满足notification的条件可以被下发到 #AppletNotifications 时，
     * 可以捕获该信号
     */
    cdos_message_tray_signals [NOTIFIY_APPLET_UPDATE] =
            g_signal_new ("notify-applet-update",
                          G_OBJECT_CLASS_TYPE (klass),
                          G_SIGNAL_RUN_LAST,
                          0,
                          NULL, NULL,
                          g_cclosure_marshal_VOID__OBJECT,
                          G_TYPE_NONE, 1,
                          G_TYPE_OBJECT);
    cdos_message_tray_signals [SOURCE_ADDED] =
            g_signal_new ("source-added",
                          G_OBJECT_CLASS_TYPE (klass),
                          G_SIGNAL_RUN_LAST,
                          0,
                          NULL, NULL,
                          g_cclosure_marshal_VOID__OBJECT,
                          G_TYPE_NONE, 1,
                          G_TYPE_OBJECT);
}

static void
cdos_message_tray_init (CdosMessageTray *self)
{
    CdosGlobal *global = cdos_global_get ();

    GSettings *global_settings = cdos_global_get_settings (global);

    CdosLayoutManager *layout_manager = cdos_layout_manager_get_default ();

    clutter_actor_hide ((ClutterActor *)self);
    st_widget_add_style_class_name ((StWidget *)self, "message-tray");

    self->notification_queue = g_queue_new ();
    self->notification = NULL;
    self->notification_clicked_id = 0;

    self->notification_state = HIDDEN;
    self->notification_timeout_id = 0;
    self->notification_removed = FALSE;

    self->sources = NULL;

    cdos_layout_manager_add_chrome (layout_manager, CLUTTER_ACTOR(self), 4);

    //g_signal_connect (layout_manager, "monitors-changed", G_CALLBACK (cdos_message_tray_set_size_position_cb), self);

    g_signal_connect (global_settings, 
                      "changed::display-notifications", 
                      G_CALLBACK (cdos_message_tray_on_notification_enabled_updated_cb), 
                      self );

    cdos_message_tray_on_notification_enabled_updated_cb (global_settings, NULL, self);

    g_signal_connect (self, "allocation-changed", G_CALLBACK (on_allocation_changed), self);
}

/**
 * cdos_message_tray_get_default:
 *
 * Returns (CdosMessageTray): 返回一个 #CdosMessageTray 类的实例.
 */
CdosMessageTray *
cdos_message_tray_get_default (void)
{
    static CdosMessageTray *self = NULL;
    if (!self)
    {
        self = g_object_new (CDOS_TYPE_MESSAGE_TRAY, NULL);
        g_object_add_weak_pointer (G_OBJECT (self), (gpointer *)&self);
    }
    return self;
}
