/*
 * Author: leit
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, see <http://www.gnu.org/licenses/>.
 */
#include "config.h"

#include <meta/display.h>
#include <meta/boxes.h>

#include "cdos-monitor-constraint.h"
#include "cdos-layout-manager.h"

typedef struct _CdosMonitorConstraint
{
    ClutterConstraint parent;
    ClutterActor *actor;

    MetaDisplay *display;
    CdosLayoutManager *layout_manager;
    gint monitor_index;
    gboolean on_monitor_center;
    gulong monitors_changed_id;
}CdosMonitorConstraint;

G_DEFINE_TYPE (CdosMonitorConstraint, cdos_monitor_constraint, CLUTTER_TYPE_CONSTRAINT)

enum
{
    PROP_0,
    PROP_INDEX,
    PROP_ON_CENTER,
    PROP_LAST
};

static GParamSpec *obj_props[PROP_LAST];

static void
cdos_monitor_constraint_dispose (GObject *object)
{
    CdosMonitorConstraint *self = CDOS_MONITOR_CONSTRAINT (object);
    g_clear_signal_handler (&self->monitors_changed_id, self->layout_manager);
    self->actor = NULL;

    G_OBJECT_CLASS (cdos_monitor_constraint_parent_class)->dispose (object);
}

static void
cdos_monitor_constraint_get_property(GObject         *object,
                                     guint            prop_id,
                                     GValue          *value,
                                     GParamSpec      *pspec)  
{
    CdosMonitorConstraint *self = CDOS_MONITOR_CONSTRAINT (object);
    switch (prop_id)
     {
         case PROP_INDEX:
             g_value_set_int (value, self->monitor_index);
             break;
         case PROP_ON_CENTER:
             g_value_set_boolean (value, self->on_monitor_center);
             break;
        default:
            G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
            break;
    }                     
}

static void
cdos_monitor_constraint_set_property(GObject         *object,
                                     guint            prop_id,
                                     const GValue    *value,
                                     GParamSpec      *pspec)  
{
    CdosMonitorConstraint *self = CDOS_MONITOR_CONSTRAINT (object);
    switch (prop_id)
    {
         case PROP_INDEX:
             self->monitor_index = g_value_get_int (value);
             if (self->actor)
                 clutter_actor_queue_relayout (self->actor);
             break;
         case PROP_ON_CENTER:
             self->on_monitor_center = g_value_get_boolean (value);
             if (self->actor)
                 clutter_actor_queue_relayout (self->actor);
             break;
        default:
            G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
            break;
    }                     
}

static void
on_monitors_changed (CdosLayoutManager *layout_manager, CdosMonitorConstraint *self)
{
    clutter_actor_queue_relayout (self->actor);
}

static void
cdos_monitor_constraint_set_actor (ClutterActorMeta *meta,
                                   ClutterActor *actor)
{
    CdosMonitorConstraint *self = CDOS_MONITOR_CONSTRAINT (meta);
    if (actor)
    {
        self->actor = actor;
        self->monitors_changed_id = g_signal_connect (self->layout_manager, "monitors-changed",
                                                      G_CALLBACK (on_monitors_changed), self);
    }
    else
    {
        self->actor = NULL;
        g_clear_signal_handler (&self->monitors_changed_id, self->layout_manager);
    }
    CLUTTER_ACTOR_META_CLASS (cdos_monitor_constraint_parent_class)->set_actor (meta, actor);
}

static void
cdos_monitor_constraint_init (CdosMonitorConstraint *self)
{
    CdosGlobal *global;
    global = cdos_global_get ();
    self->display = cdos_global_get_display (global);
    self->layout_manager = cdos_layout_manager_get_default ();
}

static void
cdos_monitor_constraint_update_allocation (ClutterConstraint *constraint,
                                           ClutterActor *actor,
                                           ClutterActorBox *box)
{
    MetaRectangle rect;
    int n_monitors;
    gfloat x, y, w, h;
    CdosMonitorConstraint *self = CDOS_MONITOR_CONSTRAINT (constraint);

    n_monitors = meta_display_get_n_monitors (self->display);

    if (n_monitors == 0)
        return;

    if (self->monitor_index < 0 || self->monitor_index >= n_monitors)
        self->monitor_index = meta_display_get_primary_monitor (self->display);
     meta_display_get_monitor_geometry (self->display, self->monitor_index, &rect);

     if (self->on_monitor_center)
     {
         //clutter_actor_get_preferred_size (actor, NULL, NULL, &w, &h);
         w = box->x2 - box->x1;
         h = box->y2 - box->y1;
         x = rect.x + (rect.width - w) / 2.0;
         y = rect.y + (rect.height - h) / 2.0;
         clutter_actor_box_set_origin (box, x, y);
     }
     else
         clutter_actor_box_init_rect (box, rect.x, rect.y, rect.width, rect.height);
}

static void
cdos_monitor_constraint_class_init (CdosMonitorConstraintClass *klass)
{
    GObjectClass *gobject_class = G_OBJECT_CLASS (klass);
    ClutterActorMetaClass *meta_class = CLUTTER_ACTOR_META_CLASS (klass);
    ClutterConstraintClass *parent_class = CLUTTER_CONSTRAINT_CLASS (klass);

    gobject_class->get_property = cdos_monitor_constraint_get_property;
    gobject_class->set_property = cdos_monitor_constraint_set_property;
    gobject_class->dispose = cdos_monitor_constraint_dispose;

    obj_props[PROP_INDEX] =
    g_param_spec_int ("index",
                      "Monitor index",
                      "Monitor index. -1 is primary",
                      -1, G_MAXINT,
                      -1,
                      G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS);

    obj_props[PROP_ON_CENTER] =
    g_param_spec_boolean ("on-center",
                          "On monitor center",
                          "Whether the actor is set to appear on monitor center",
                          FALSE,
                          G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS);
    g_object_class_install_properties (gobject_class, PROP_LAST, obj_props);

    meta_class->set_actor = cdos_monitor_constraint_set_actor;
    parent_class->update_allocation = cdos_monitor_constraint_update_allocation;
}

ClutterConstraint *
cdos_monitor_constraint_new (gint monitor_index,
                             gboolean on_monitor_center)
{
    ClutterConstraint *constraint = g_object_new (CDOS_TYPE_MONITOR_CONSTRAINT,
                                                  "index", monitor_index,
                                                  "on-center", on_monitor_center,
                                                  NULL);
    return constraint;
}

void
cdos_monitor_constraint_set_index (CdosMonitorConstraint *self,
                                   gint index)
{
    g_return_if_fail (CDOS_IS_MONITOR_CONSTRAINT (self));
    if (self->monitor_index == index)
        return;
    self->monitor_index = index;
    if (self->actor)
        clutter_actor_queue_relayout (self->actor);
}

void
cdos_monitor_constraint_set_on_center (CdosMonitorConstraint *self,
                                       gboolean on_center)
{
    g_return_if_fail (CDOS_IS_MONITOR_CONSTRAINT (self));
    if (self->on_monitor_center == on_center)
        return;
    self->on_monitor_center = on_center;
    if (self->actor)
        clutter_actor_queue_relayout (self->actor);
}
