/*
 * Copyright 2021 <cdos>
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms and conditions of the GNU Lesser General Public License,
 * version 2.1, as published by the Free Software Foundation.
 *
 * This program is distributed in the hope it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for
 * more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 */

/*
 * SECTION:  cdos-monitor-switcher.c
 * @Title: cdos-monitor-switcher
 * @Short_Description:  libappindicator and StatusNotifier support.
 */

#include "config.h"
#include <meta/common.h>
#include <meta/meta-monitor-manager.h>
#include <meta/boxes.h>

#include "common.h"
#include "cdos-layout-manager.h"
#include "cdos-ui-plugin.h"
#include "cdos-monitor-switcher.h"
#include "cdos-start.h"

struct _CdosMonitorSwitcher
{
    StWidget parent;

    MetaDisplay *display;
    ClutterActor *stage;
    StWidget *box;
    ClutterActor *buttons[META_MONITOR_SWITCH_CONFIG_UNKNOWN];
    guint focus;
    ClutterInputDevice *mouse;
    gulong event_capture_id;
};

G_DEFINE_TYPE(CdosMonitorSwitcher, cdos_monitor_switcher, ST_TYPE_WIDGET)

static MetaMonitorSwitchConfigType config_types[META_MONITOR_SWITCH_CONFIG_UNKNOWN + 1] =
{
     META_MONITOR_SWITCH_CONFIG_ALL_MIRROR,
     META_MONITOR_SWITCH_CONFIG_ALL_LINEAR,
     META_MONITOR_SWITCH_CONFIG_EXTERNAL,
     META_MONITOR_SWITCH_CONFIG_BUILTIN,
     META_MONITOR_SWITCH_CONFIG_UNKNOWN
};

static const gchar *config_desc[META_MONITOR_SWITCH_CONFIG_UNKNOWN + 1] =
{
    "Mirror",
    "Join Displays",
    "External Display",
    "Built-in Display",
    "Unknown"
};

static const gchar *config_icons[META_MONITOR_SWITCH_CONFIG_UNKNOWN] =
{
    "view-mirror-symbolic",
    "video-joined-displays-symbolic",
    "video-single-display-symbolic",
    "computer-symbolic",
    NULL
};

static void
close_monitor_switcher (CdosMonitorSwitcher *self)
{
    CdosLayoutManager *layout_manager;
    ClutterActor *actor = CLUTTER_ACTOR (self);
    if (self->event_capture_id)
    {
        g_signal_handler_disconnect (self->stage, self->event_capture_id);
        self->event_capture_id = 0;
    }
    st_focus_manager_remove_group (st_focus_manager_get_for_stage ((ClutterStage *)self->stage), ST_WIDGET (actor));
    layout_manager = cdos_layout_manager_get_default ();
    cdos_layout_manager_remove_chrome (layout_manager, actor);
    clutter_actor_hide (actor);
    cdos_pop_modal (actor, 0);
}

static void
cdos_monitor_switcher_apply_config (CdosMonitorSwitcher *self)
{
    MetaMonitorManager *monitor_manager = meta_monitor_manager_get ();
    if (meta_monitor_manager_can_switch_config (monitor_manager) && self->focus != -1)
    {
        g_message ("Desktop monitor switcher, Apply config: %s", config_desc[config_types[self->focus]]);
        meta_monitor_manager_switch_config (monitor_manager,
                                            config_types[self->focus]);
    }
    close_monitor_switcher (self);
}

static void
button_on_button_focus (ClutterActor *button,
                        CdosMonitorSwitcher *self)
{
    gpointer index;
    index = g_object_get_data ((GObject *)button, "index");
    self->focus = GPOINTER_TO_UINT (index);
}

static gboolean
button_on_button_release (ClutterActor *button,
                          ClutterEvent *event,
                          CdosMonitorSwitcher *self)
{
    gpointer index;
    index = g_object_get_data ((GObject *)button, "index");
    self->focus = GPOINTER_TO_UINT (index);
    cdos_monitor_switcher_apply_config (self);
    return TRUE;
}

static void
add_button (CdosMonitorSwitcher *self, guint i)
{
    StWidget *desc;
    StIcon *icon;
    const gchar *display_name;
    /*
    MetaMonitorSwitchConfigType config;
    config = config_types[i];
    */

    StBin *bin = g_object_new (ST_TYPE_BIN,
                               "style-class", "monitor-switcher-button",
                               "track-hover", TRUE,
                               "can-focus", TRUE,
                               "reactive", TRUE,
                               "x-expand", TRUE,
                               "y-expand", TRUE,
                               NULL);
    g_object_set_data ((GObject *)bin, "index", GUINT_TO_POINTER (i));
    g_signal_connect (bin, "key-focus-in", G_CALLBACK (button_on_button_focus), self);
    g_signal_connect (bin, "button-release-event", G_CALLBACK (button_on_button_release), self);

    StWidget *vbox = g_object_new (ST_TYPE_BOX_LAYOUT,
                                   "x-expand", TRUE,
                                   NULL);
    st_box_layout_set_vertical ((StBoxLayout *)vbox, TRUE);

    MetaMonitorManager *monitor_manager = meta_monitor_manager_get ();
    //台式机等设备
    if (i > 1 && meta_monitor_manager_has_laptop_panel (monitor_manager) == FALSE)
    {
        /** 
         * read display name from monitor manager.
         * 0-primary monitor or as primary monitor,
         * 1-is extenal monitor.
         */
        gint monitor_index = i == 2 ? 1 : 0;
        display_name = meta_monitor_manager_get_monitor_display_name (monitor_manager, monitor_index);
    } else {
        display_name = _(config_desc[i]);
    }

    desc = st_label_new (display_name);

    _add_actor (vbox, desc);
    icon = g_object_new (ST_TYPE_ICON,
                         "icon-name", config_icons[i],
                         "icon-size", 128,
                         NULL);
    _add_actor (vbox, icon);

    st_bin_set_child (bin, CLUTTER_ACTOR (vbox));
    self->buttons[i] = (ClutterActor *)bin;

    _add_actor (self->box, bin);
}

static void
add_buttons (CdosMonitorSwitcher *self)
{
    add_button (self, META_MONITOR_SWITCH_CONFIG_ALL_MIRROR);
    add_button (self, META_MONITOR_SWITCH_CONFIG_ALL_LINEAR);
    add_button (self, META_MONITOR_SWITCH_CONFIG_EXTERNAL);
    add_button (self, META_MONITOR_SWITCH_CONFIG_BUILTIN);
}

static void
cdos_monitor_switcher_dispose (GObject *gobject)
{
    CdosMonitorSwitcher *self = CDOS_MONITOR_SWITCHER (gobject);
    close_monitor_switcher (self);
    if (self->mouse)
    {
        clutter_input_device_ungrab (self->mouse);
        self->mouse = NULL;
    }
    G_OBJECT_CLASS (cdos_monitor_switcher_parent_class)->dispose (gobject);
}

static void
cdos_monitor_switcher_get_preferred_width (ClutterActor *actor,
                                           gfloat        for_height,
                                           gfloat       *min_width_p,
                                           gfloat       *natural_width_p)
{
    CdosMonitorSwitcher *self = (CdosMonitorSwitcher *)actor;
    StThemeNode *theme_node = st_widget_get_theme_node ((StWidget *)actor);
    st_theme_node_adjust_for_height (theme_node, &for_height);
    gfloat min_w = 0, nat_w = 0;
    ClutterActor *child = CLUTTER_ACTOR (self->box);
    clutter_actor_get_preferred_width (child, for_height, &min_w, &nat_w);
    if (min_width_p)
        *min_width_p = min_w;
    if (natural_width_p)
        *natural_width_p = nat_w;

    st_theme_node_adjust_preferred_width (theme_node, min_width_p, natural_width_p);
}

static void
cdos_monitor_switcher_get_preferred_height (ClutterActor *actor,
                                            gfloat        for_width,
                                            gfloat       *min_height_p,
                                            gfloat       *natural_height_p)
{
    CdosMonitorSwitcher *self = (CdosMonitorSwitcher *)actor;
    StThemeNode *theme_node = st_widget_get_theme_node ((StWidget *)actor);
    st_theme_node_adjust_for_width (theme_node, &for_width);

    gfloat min_h = 0, nat_h = 0;
    ClutterActor *child = CLUTTER_ACTOR (self->box);
    clutter_actor_get_preferred_height (child, for_width, &min_h, &nat_h);
    if (min_height_p)
        *min_height_p = min_h;
    if (natural_height_p)
        *natural_height_p = nat_h;

    st_theme_node_adjust_preferred_height (theme_node, min_height_p, natural_height_p);
}

static void
cdos_monitor_switcher_allocate (ClutterActor *actor,
                                const ClutterActorBox *box,
                                ClutterAllocationFlags flags)
{
    StThemeNode *theme_node = st_widget_get_theme_node ((StWidget *)actor);
    ClutterActorBox content_box;
    clutter_actor_set_allocation (actor, box, flags);

    st_theme_node_get_content_box (theme_node, box, &content_box);
    ClutterActor *child = clutter_actor_get_first_child (actor);
    if (child)
        clutter_actor_allocate (child, &content_box, flags);
}

static gboolean 
response_key_release_event (ClutterActor *actor,
                            ClutterEvent *event)
{
    CdosMonitorSwitcher *self = (CdosMonitorSwitcher *)actor;
    guint key = clutter_event_get_key_symbol (event);
    gboolean is_enter = (key == CLUTTER_KEY_ISO_Enter ||
                         key == CLUTTER_KEY_Return ||
                         key == CLUTTER_KEY_KP_Enter);
    gboolean is_escape = key == CLUTTER_KEY_Escape;
    if (is_escape)
    {
        close_monitor_switcher ((CdosMonitorSwitcher *)actor);
    } 
    if (is_enter)
    {
        cdos_monitor_switcher_apply_config (self);
        return TRUE;
    }
    return FALSE;
}

static gboolean
cdos_monitor_switcher_on_key_release (ClutterActor *actor,
                                      ClutterKeyEvent *event)
{
    ClutterInputDevice *device;
    GtkDirectionType dir;
    ClutterModifierType mod;
    gboolean is_nav_key = TRUE;

    device = clutter_event_get_device ((ClutterEvent *)event);
    if (!device)
        return FALSE;
    mod = clutter_event_get_state ((ClutterEvent *)event);
    mod &= CLUTTER_MODIFIER_MASK;
    mod &= ~CLUTTER_LOCK_MASK;
    if (mod)
        return FALSE;

    guint key = clutter_event_get_key_symbol ((ClutterEvent *)event);
    gboolean is_enter = (key == CLUTTER_KEY_ISO_Enter ||
                         key == CLUTTER_KEY_Return ||
                         key == CLUTTER_KEY_KP_Enter);
    gboolean is_escape = key == CLUTTER_KEY_Escape;

    if (is_escape)
    {
        close_monitor_switcher ((CdosMonitorSwitcher *)actor);
    }
    else if (is_enter)
    {
        CdosMonitorSwitcher *self = (CdosMonitorSwitcher *)actor;
        cdos_monitor_switcher_apply_config (self);
        return TRUE;
    }
    else
    {
        switch (key)
        {
            case CLUTTER_KEY_Left:
                dir = GTK_DIR_LEFT;
                break;
            case CLUTTER_KEY_Right:
            case CLUTTER_KEY_Tab:
                dir = GTK_DIR_RIGHT;
                break;
            default:
                is_nav_key = FALSE;
                break;
        }
        if (is_nav_key)
        {
            ClutterActor *focus;
            ClutterStage *stage = cdos_global_get_stage (global);
            focus = clutter_stage_get_key_focus (stage);
            if (focus == actor)
            {
                CdosMonitorSwitcher *self = (CdosMonitorSwitcher *)actor;
                if (self->focus != -1)
                    st_widget_navigate_focus ((StWidget *)actor, self->buttons[self->focus], dir, TRUE);
                else
                    st_widget_navigate_focus ((StWidget *)actor, NULL, dir, FALSE);
                return TRUE;
            }
        }
    }
    return FALSE;
}

static gboolean
on_button_event (ClutterActor *actor,
                 ClutterEvent *event,
                 gboolean is_release)
{
    ClutterInputDevice *device;
    gfloat ex, ey, x, y, w, h;

    device = clutter_event_get_device (event);
    if (!device)
        return FALSE;

    clutter_event_get_coords (event, &ex, &ey);
    clutter_actor_get_transformed_position (actor, &x, &y);
    clutter_actor_get_transformed_size (actor, &w, &h);
    if (ex < x || ex > x + w ||
        ey < y || ey > y + h)
    {
        if (is_release)
            close_monitor_switcher ((CdosMonitorSwitcher *)actor);
        //return TRUE;
    }
    return FALSE;
}

static gboolean
cdos_monitor_switcher_on_button_release (ClutterActor *actor,
                                         ClutterButtonEvent *button_event)
{
    return on_button_event (actor, (ClutterEvent *)button_event, TRUE);
}

static gboolean
cdos_monitor_switcher_on_button_press (ClutterActor *actor,
                                       ClutterButtonEvent *button_event)
{
    return on_button_event (actor, (ClutterEvent *)button_event, FALSE);
}

static gboolean
cdos_monitor_switcher_on_motion (ClutterActor *actor,
                                 ClutterMotionEvent *motion_event)
{
    /*
    CdosMonitorSwitcher *self;
    self = (CdosMonitorSwitcher *)actor;
    */
    return on_button_event (actor, (ClutterEvent *)motion_event, FALSE);
}

static gboolean
cdos_monitor_switcher_on_event_capture (ClutterActor *stage,
                                        ClutterEvent *event,
                                        ClutterActor *actor)
{
    ClutterEventType event_type;
    event_type = clutter_event_type (event);

    if (event_type == CLUTTER_MOTION)
        return TRUE;
    if (event_type == CLUTTER_BUTTON_PRESS)
        return on_button_event (actor, event, FALSE);
    if (event_type == CLUTTER_BUTTON_RELEASE)
        return on_button_event (actor, event, TRUE);
    if (event_type == CLUTTER_KEY_RELEASE)
        return response_key_release_event (actor, event);

    return FALSE;
}

static void
cdos_monitor_switcher_init (CdosMonitorSwitcher *self)
{
    ClutterActor *actor;
    CdosLayoutManager *layout_manager;
    ClutterLayoutManager *layout;

    CdosGlobal *global = cdos_global_get ();
    self->display = cdos_global_get_display (global);
    /*
    ClutterSeat *seat = clutter_backend_get_default_seat (clutter_get_default_backend ());
    ClutterInputDevice *mouse = clutter_seat_get_pointer (seat);
    self->mouse = mouse;
    */

    actor = CLUTTER_ACTOR (self);
    st_widget_add_style_class_name (ST_WIDGET (self), "monitor-switcher");

    self->box = g_object_new (ST_TYPE_BOX_LAYOUT,
                              "style-class", "monitor-switcher-box",
                              NULL);
    layout = clutter_actor_get_layout_manager ((ClutterActor *)self->box);
    clutter_box_layout_set_homogeneous (CLUTTER_BOX_LAYOUT (layout), TRUE);
    clutter_actor_add_child (actor, CLUTTER_ACTOR (self->box));
    add_buttons (self);

    layout_manager = cdos_layout_manager_get_default ();
    cdos_layout_manager_add_chrome (layout_manager, actor, AFFECT_INPUT_REGION);
}

static void
cdos_monitor_switcher_class_init (CdosMonitorSwitcherClass *klass)
{
    GObjectClass *gobject_class = G_OBJECT_CLASS (klass);
    gobject_class->dispose = cdos_monitor_switcher_dispose;

    ClutterActorClass *actor_class = CLUTTER_ACTOR_CLASS (klass);
    actor_class->get_preferred_width = cdos_monitor_switcher_get_preferred_width;
    actor_class->get_preferred_height = cdos_monitor_switcher_get_preferred_height;
    actor_class->allocate = cdos_monitor_switcher_allocate;
    actor_class->key_release_event = cdos_monitor_switcher_on_key_release;
    actor_class->button_release_event = cdos_monitor_switcher_on_button_release;
    actor_class->button_press_event = cdos_monitor_switcher_on_button_press;
    actor_class->motion_event = cdos_monitor_switcher_on_motion;
}

static void
cdos_monitor_switcher_show (CdosMonitorSwitcher *self)
{
    gfloat width, height;
    gfloat x, y;
    int current_monitor;
    MetaRectangle rect;
    ClutterActor *ui_group, *actor;
    ClutterActor *stage;
    MetaMonitorSwitchConfigType config;

    actor = CLUTTER_ACTOR (self);
    MetaMonitorManager *monitor_manager = meta_monitor_manager_get ();

    if (meta_monitor_manager_can_switch_config (monitor_manager) == FALSE)
    {
        g_message ("Desktop monitor switcher: Can't switch config.");
        close_monitor_switcher (self);
        return;
    }

    ui_group = cdos_ui_plugin_get_ui_group ();
    clutter_actor_set_child_above_sibling (ui_group, actor, NULL);

    /* if (meta_monitor_manager_has_laptop_panel (monitor_manager)) 
    {
        clutter_actor_show (self->buttons[META_MONITOR_SWITCH_CONFIG_EXTERNAL]);
        clutter_actor_show (self->buttons[META_MONITOR_SWITCH_CONFIG_BUILTIN]);
    } else {
        clutter_actor_hide (self->buttons[META_MONITOR_SWITCH_CONFIG_EXTERNAL]);
        clutter_actor_hide (self->buttons[META_MONITOR_SWITCH_CONFIG_BUILTIN]);
    } */

    current_monitor = meta_display_get_current_monitor (self->display);
    if (current_monitor >= 0)
        meta_display_get_monitor_geometry (self->display, current_monitor, &rect);
    clutter_actor_get_size (actor, &width, &height);
    x = rect.x + (rect.width - width) / 2.0;
    y = rect.y + (rect.height - height) / 2.0;

    clutter_actor_set_position (actor, x, y);
    clutter_actor_show (actor);
    cdos_push_modal (actor, 0, 0, CDOS_ACTION_MODE_SYSTEM_MODAL);

    /*
    if (self->mouse)
        clutter_input_device_grab (self->mouse, actor);
    */

    stage = clutter_actor_get_stage (actor);
    st_focus_manager_add_group (st_focus_manager_get_for_stage ((ClutterStage *)stage), ST_WIDGET (actor));
    self->event_capture_id = g_signal_connect (stage,
                                               "captured-event",
                                               G_CALLBACK(cdos_monitor_switcher_on_event_capture),
                                               self);
    self->stage = stage;

    config = meta_monitor_manager_get_switch_config (monitor_manager);
    g_message ("Desktop monitor switcher, current config: %d(%s)", config, _(config_desc[config]));
    self->focus = -1;
    for (guint i = 0; i < G_N_ELEMENTS (config_types) - 1; i++)
    {
        if (config == config_types[i])
        {
            clutter_actor_grab_key_focus (CLUTTER_ACTOR (self->buttons[i]));
            self->focus = i;
            break;
        }
    }
}

CdosMonitorSwitcher *
cdos_monitor_switcher_get_default (void)
{
    static CdosMonitorSwitcher *monitor_switcher = NULL;
    if (!monitor_switcher)
    {
        monitor_switcher = g_object_new (CDOS_TYPE_MONITOR_SWITCHER, NULL);
        //g_object_add_weak_pointer ((GObject *)monitor_switcher, (gpointer *)&monitor_switcher);
    }
    cdos_monitor_switcher_show (monitor_switcher);
    return monitor_switcher;
}
