/**
 * Copyright (c) 2016-2018 CPU and Fundamental Software Research Center, CAS
 *
 * This software is published by the license of CPU-OS License，you can use and
 * distibuted this software under this License. See CPU-OS Liense for more detail.
 *
 * You should have receive a copy of CPU-OS License. If not, please contact us
 * by email <support_os@cpu-os.ac.cn>
 *
 **/
#include <gio/gio.h>
#include "cdos-mpris2-player.h"


#define MPRIS2_PLAYER_PREFIX "org.mpris.MediaPlayer2."

struct _CdosMpris2Player
{
    GObject parent_instance;

    GDBusConnection *connection;
    GDBusProxy      *props_proxy;
    GDBusProxy      *player_proxy;
    gchar           *dbus_name;
    guint            signal_handler_id;
    guint            watch_id;
    gboolean         connected;
    gulong           playback_timer_id;

    gchar           *player;
    gboolean         strict_mode;

    /* MediaPlayer2 Properties */
    gboolean         can_quit;
    gboolean         can_raise;
    gboolean         has_tracklist;
    gchar           *identity;
    gchar          **supported_uri_schemes;
    gchar          **supported_mime_types;
    gchar           *desktop_entry; /* Optional */

    /* Interface MediaPlayer2.Player Properties */
    CdosMpris2PlaybackStatus   playback_status;
    gdouble          rate;
    gdouble          volume;
    gint64           position;
    gdouble          minimum_rate;
    gdouble          maximum_rate;
    gboolean         can_go_next;
    gboolean         can_go_previous;
    gboolean         can_play;
    gboolean         can_pause;
    gboolean         can_seek;
    gboolean         can_control;

    CdosMpris2Metadata *metadata;
};

enum
{
    CONNECTED,
    CONTROL_FLAGS,
    PLAYBACK_STATUS,
    PLAYBACK_TICK,
    METADATA,
    VOLUME,
    LAST_SIGNAL
};
static guint cdos_mpris2_player_signals[LAST_SIGNAL] = {0};

G_DEFINE_TYPE (CdosMpris2Player, cdos_mpris2_player, G_TYPE_OBJECT);

static void cdos_mpris2_player_parse_playback_status (CdosMpris2Player *mpris2, const gchar *playback_status);

struct _CdosMpris2Metadata
{
    gchar  *trackid;
    gchar  *url;
    gchar  *title;
    gchar  *artists;
    gchar  *album;
    gint64  length;
    guint   track_no;
    gchar  *arturl;
};

static void
cdos_mpris2_metadata_set_trackid (CdosMpris2Metadata *metadata, const gchar *trackid)
{
    if (metadata->trackid)
        g_free (metadata->trackid);
    metadata->trackid = g_strdup (trackid);
}

const gchar *
cdos_mpris2_metadata_get_trackid (CdosMpris2Metadata *metadata)
{
    return metadata->trackid;
}

static void
cdos_mpris2_metadata_set_url (CdosMpris2Metadata *metadata, const gchar *url)
{
    if (metadata->url)
        g_free (metadata->url);
    metadata->url = g_strdup (url);
}

const gchar *
cdos_mpris2_metadata_get_url (CdosMpris2Metadata *metadata)
{
    return metadata->url;
}

static void
cdos_mpris2_metadata_set_title (CdosMpris2Metadata *metadata, const gchar *title)
{
    if (metadata->title)
        g_free (metadata->title);
    metadata->title = g_strdup (title);
}

const gchar *
cdos_mpris2_metadata_get_title (CdosMpris2Metadata *metadata)
{
    return metadata->title;
}

static void
cdos_mpris2_metadata_set_artists (CdosMpris2Metadata *metadata, gchar *artists)
{
    if (metadata->artists)
        g_free (metadata->artists);
    metadata->artists = artists;
}

gchar *
cdos_mpris2_metadata_get_artists (CdosMpris2Metadata *metadata)
{
    return metadata->artists;
}

static void
cdos_mpris2_metadata_set_album (CdosMpris2Metadata *metadata, const gchar *album)
{
    if (metadata->album)
        g_free (metadata->album);
    metadata->album = g_strdup (album);
}

const gchar *
cdos_mpris2_metadata_get_album (CdosMpris2Metadata *metadata)
{
    return metadata->album;
}

static void
cdos_mpris2_metadata_set_length (CdosMpris2Metadata *metadata, gint64 length)
{
    metadata->length = length;
}

gint64
cdos_mpris2_metadata_get_length (CdosMpris2Metadata *metadata)
{
    return metadata->length;
}

static void G_GNUC_UNUSED
cdos_mpris2_metadata_set_track_no (CdosMpris2Metadata *metadata, guint track_no)
{
    metadata->track_no = track_no;
}

guint
cdos_mpris2_metadata_get_track_no (CdosMpris2Metadata *metadata)
{
    return metadata->track_no;
}

static void
cdos_mpris2_metadata_set_arturl (CdosMpris2Metadata *metadata, const gchar *arturl)
{
    if (metadata->arturl)
        g_free (metadata->arturl);
    metadata->arturl = g_strdup (arturl);
}

const gchar *
cdos_mpris2_metadata_get_arturl (CdosMpris2Metadata *metadata)
{
    return metadata->arturl;
}

static CdosMpris2Metadata *
cdos_mpris2_metadata_new_from_variant (GVariant *dict)
{
    GVariantIter iter;
    GVariant *value;
    gchar *key;

    gint64 length = 0;

    CdosMpris2Metadata *metadata;

    //metadata = g_slice_new0 (CdosMpris2Metadata);
    metadata = g_new0 (CdosMpris2Metadata, 1);

    g_variant_iter_init (&iter, dict);
    while (g_variant_iter_loop (&iter, "{sv}", &key, &value))
    {
        if (0 == g_ascii_strcasecmp (key, "mpris:trackid"))
            cdos_mpris2_metadata_set_trackid (metadata, g_variant_get_string (value, NULL));
        else if (0 == g_ascii_strcasecmp (key, "mpris:artUrl"))
            cdos_mpris2_metadata_set_arturl (metadata, g_variant_get_string (value, NULL));
        else if (0 == g_ascii_strcasecmp (key, "mpris:length"))
            length = g_variant_get_int64 (value);
        else if (0 == g_ascii_strcasecmp (key, "xesam:album"))
            cdos_mpris2_metadata_set_album (metadata, g_variant_get_string (value, NULL));
        else if (0 == g_ascii_strcasecmp (key, "xesam:albumArtist"));
            // List of Strings.
        else if (0 == g_ascii_strcasecmp (key, "xesam:artist"))
        {
            const gchar **artists = g_variant_get_strv (value, NULL);
            gchar *tmp = g_strjoinv (" ", (gchar **)artists);
            cdos_mpris2_metadata_set_artists (metadata, tmp);
            g_free (artists);
        }
        else if (0 == g_ascii_strcasecmp (key, "xesam:url"))
            cdos_mpris2_metadata_set_url (metadata, g_variant_get_string (value, NULL));
        else if (0 == g_ascii_strcasecmp (key, "xesam:title"))
            cdos_mpris2_metadata_set_title (metadata, g_variant_get_string (value, NULL));
        else if (0 == g_ascii_strcasecmp (key, "xesam:genre"));
            /* (List of Strings.) Not use genre */
        else if (0 == g_ascii_strcasecmp (key, "xesam:comment"));
            /* (List of Strings) Not use comment */
        else if (0 == g_ascii_strcasecmp (key, "xesam:audioBitrate"));
            /* (uint32) Not use audioBitrate */
        else if (0 == g_ascii_strcasecmp (key, "xesam:trackNumber"));
            //cdos_mpris2_metadata_set_track_no (metadata, g_variant_get_int32 (value));
        else if (0 == g_ascii_strcasecmp (key, "xesam:useCount"));
            /* (Integer) Not use useCount */
        else if (0 == g_ascii_strcasecmp (key, "xesam:userRating"));
            /* (Float) Not use userRating */
        else if (0 == g_ascii_strcasecmp (key, "xesam:contentCreated"));
            /* has type 's' */
        else if (0 == g_ascii_strcasecmp (key, "audio-bitrate"));
            /* has type 'i' */
        else if (0 == g_ascii_strcasecmp (key, "audio-channels"));
            /* has type 'i' */
        else if (0 == g_ascii_strcasecmp (key, "audio-samplerate"));
            /* has type 'i' */
        else if (0 == g_ascii_strcasecmp (key, "xesam:contentCreated"));
            /* has type 's' */
        else if (0 == g_ascii_strcasecmp (key, "audio-bitrate"));
            /* has type 'i' */
        else if (0 == g_ascii_strcasecmp (key, "audio-channels"));
            /* has type 'i' */
        else if (0 == g_ascii_strcasecmp (key, "audio-samplerate"));
            /* has type 'i'*/
        else
            g_debug ("GVariant '%s' has type '%s'", key,
                     g_variant_get_type_string (value));
    }

    cdos_mpris2_metadata_set_length (metadata, length);

    return metadata;
}

static void
cdos_mpris2_metadata_free (CdosMpris2Metadata *metadata)
{
    if(metadata == NULL)
        return;

    if(metadata->trackid)
        g_free (metadata->trackid);
    if(metadata->url)
        g_free (metadata->url);
    if(metadata->title)
        g_free (metadata->title);
    if(metadata->artists)
        g_free (metadata->artists);
    if(metadata->album)
        g_free (metadata->album);
    if(metadata->arturl)
        g_free (metadata->arturl);

    //g_slice_free (CdosMpris2Metadata, metadata);
    g_free (metadata);
}

static void
cdos_mpris2_player_reset (CdosMpris2Player *mpris2)
{
    mpris2->connection            = NULL;
    mpris2->props_proxy           = NULL;
    mpris2->player_proxy          = NULL;
    mpris2->dbus_name             = NULL;
    mpris2->watch_id              = 0;

    mpris2->player                = NULL;

    mpris2->can_quit              = FALSE;
    mpris2->can_raise             = FALSE;
    mpris2->has_tracklist         = FALSE;
    mpris2->identity              = NULL;
    mpris2->desktop_entry         = NULL;
    mpris2->supported_uri_schemes = NULL;
    mpris2->supported_mime_types  = NULL;

    mpris2->playback_status       = CDOS_MPRIS2_STOPPED;
    mpris2->rate                  = 1.0;
    mpris2->metadata              = NULL;
    mpris2->volume                = -1;
    mpris2->position              = 0;
    mpris2->minimum_rate          = 1.0;
    mpris2->maximum_rate          = 1.0;
    mpris2->can_go_next           = FALSE;
    mpris2->can_go_previous       = FALSE;
    mpris2->can_play              = FALSE;
    mpris2->can_pause             = FALSE;
    mpris2->can_seek              = FALSE;
    mpris2->can_control           = FALSE;

    mpris2->connected             = FALSE;
    mpris2->strict_mode           = FALSE;
}

static void
cdos_mpris2_player_call_player_method (CdosMpris2Player *mpris2, const char *method)
{
    GDBusMessage *message;
    GError       *error = NULL;

    message = g_dbus_message_new_method_call (mpris2->dbus_name,
                                              "/org/mpris/MediaPlayer2",
                                              "org.mpris.MediaPlayer2.Player",
                                              method);

    g_dbus_connection_send_message (mpris2->connection,
                                    message,
                                    G_DBUS_SEND_MESSAGE_FLAGS_NONE,
                                    NULL,
                                    &error);
    if (error != NULL)
    {
        g_warning ("unable to send message: %s", error->message);
        g_clear_error (&error);
        error = NULL;
    }
    /*
    g_dbus_connection_flush_sync (mpris2->connection, NULL, &error);
    if (error != NULL)
    {
        g_warning ("unable to flush message queue: %s", error->message);
        g_clear_error (&error);
    }
    */
    g_dbus_connection_flush (mpris2->connection, NULL, NULL, NULL);

    g_object_unref (message);
}

/* Send mesages to use methods of org.mpris.MediaPlayer2 interfase. */

static void
cdos_mpris2_player_call_method (CdosMpris2Player *mpris2, const char *method)
{
    GDBusMessage *message;
    GError       *error = NULL;

    message = g_dbus_message_new_method_call (mpris2->dbus_name,
                                              "/org/mpris/MediaPlayer2",
                                              "org.mpris.MediaPlayer2",
                                              method);

    g_dbus_connection_send_message (mpris2->connection,
                                    message,
                                    G_DBUS_SEND_MESSAGE_FLAGS_NONE,
                                    NULL,
                                    &error);
    if (error != NULL)
    {
        g_warning ("unable to send message: %s", error->message);
        g_clear_error (&error);
        error = NULL;
    }

    g_dbus_connection_flush_sync (mpris2->connection, NULL, &error);
    if (error != NULL)
    {
        g_warning ("unable to flush message queue: %s", error->message);
        g_clear_error (&error);
    }

    g_object_unref (message);
}

static GVariant *
cdos_mpris2_player_get_properties (CdosMpris2Player *self)
{
    GVariantIter iter;
    GVariant *result, *child = NULL;

    result = g_dbus_connection_call_sync (self->connection,
                                          self->dbus_name,
                                          "/org/mpris/MediaPlayer2",
                                          "org.freedesktop.DBus.Properties",
                                          "GetAll",
                                          g_variant_new ("(s)", "org.mpris.MediaPlayer2"),
                                          G_VARIANT_TYPE ("(a{sv})"),
                                          G_DBUS_CALL_FLAGS_NONE,
                                          1000,
                                          NULL,
                                          NULL);

    if(result)
    {
        g_variant_iter_init (&iter, result);
        child = g_variant_iter_next_value (&iter);
    }

    return child;
}

static void
cdos_mpris2_player_parse_properties (CdosMpris2Player *self, GVariant *properties)
{
    GVariantIter iter;
    GVariant *value;
    const gchar *key;

    g_variant_iter_init (&iter, properties);

    while (g_variant_iter_loop (&iter, "{sv}", &key, &value))
    {
        if (0 == g_ascii_strcasecmp (key, "CanQuit"))
            self->can_quit = g_variant_get_boolean (value);
        else if (0 == g_ascii_strcasecmp (key, "CanRaise"))
            self->can_raise = g_variant_get_boolean (value);
        else if (0 == g_ascii_strcasecmp (key, "HasTrackList"))
            self->has_tracklist = g_variant_get_boolean (value);
        else if (0 == g_ascii_strcasecmp (key, "Identity"))
        {
            if (self->identity)
                g_free (self->identity);
            self->identity = g_variant_dup_string (value, NULL);
        }
        else if (0 == g_ascii_strcasecmp (key, "DesktopEntry"))
        {
            if (self->desktop_entry)
                g_free (self->desktop_entry);
            self->desktop_entry = g_variant_dup_string (value, NULL);
        }
        else if (0 == g_ascii_strcasecmp (key, "SupportedUriSchemes"))
        {
            if (self->supported_uri_schemes)
                g_strfreev (self->supported_uri_schemes);
            self->supported_uri_schemes = g_variant_dup_strv (value, NULL);
        }
        else if (0 == g_ascii_strcasecmp (key, "SupportedMimeTypes"))
        {
            if (self->supported_mime_types)
                g_strfreev (self->supported_mime_types);
            self->supported_mime_types = g_variant_dup_strv (value, NULL);
        }
    }
}

static GVariant *
cdos_mpris2_player_get_player_properties (CdosMpris2Player *self)
{
    GVariantIter iter;
    GVariant *result, *child = NULL;
    GError *error = NULL;

    result = g_dbus_connection_call_sync (self->connection,
                                          self->dbus_name,
                                          "/org/mpris/MediaPlayer2",
                                          "org.freedesktop.DBus.Properties",
                                          "GetAll",
                                          g_variant_new ("(s)", "org.mpris.MediaPlayer2.Player"),
                                          G_VARIANT_TYPE ("(a{sv})"),
                                          G_DBUS_CALL_FLAGS_NONE,
                                          1000,
                                          NULL,
                                          &error);

    if(result)
    {
        g_variant_iter_init (&iter, result);
        child = g_variant_iter_next_value (&iter);
    }
    else
    {
        g_critical ("%s: %s", __func__, error->message);
        g_clear_error (&error);
    }

    return child;
}

static void
cdos_mpris2_player_parse_player_properties (CdosMpris2Player *mpris2, GVariant *properties)
{
    GVariantIter iter;
    GVariant *value;
    const gchar *key;
    const gchar *playback_status = NULL;
    const gchar *loop_status G_GNUC_UNUSED = NULL;
    CdosMpris2Metadata *metadata = NULL;
    gdouble volume = -1;
    gboolean loop_status_changed G_GNUC_UNUSED = FALSE;
    guint control_flags = 0;

    g_variant_iter_init (&iter, properties);

    while (g_variant_iter_loop (&iter, "{sv}", &key, &value))
    {
        if (0 == g_strcmp0 (key, "PlaybackStatus"))
        {
            playback_status = g_variant_get_string(value, NULL);
        }
        else if (0 == g_strcmp0 (key, "Rate"))
        {
            mpris2->rate = g_variant_get_double(value);
        }
        else if (0 == g_strcmp0 (key, "Metadata"))
        {
            metadata = cdos_mpris2_metadata_new_from_variant (value);
        }
        else if (0 == g_strcmp0 (key, "Volume"))
        {
            volume = g_variant_get_double(value);
        }
        else if (0 == g_strcmp0 (key, "MinimumRate"))
        {
            mpris2->minimum_rate = g_variant_get_double(value);
        }
        else if (0 == g_strcmp0 (key, "MaximumRate"))
        {
            mpris2->maximum_rate = g_variant_get_double(value);
        }
        else if (0 == g_strcmp0 (key, "CanGoNext"))
        {
            mpris2->can_go_next = g_variant_get_boolean(value);
            control_flags |= CDOS_MPRIS2_CAN_GO_NEXT;
        }
        else if (0 == g_strcmp0 (key, "CanGoPrevious"))
        {
            mpris2->can_go_previous = g_variant_get_boolean(value);
            control_flags |= CDOS_MPRIS2_CAN_GO_PREVIOUS;
        }
        else if (0 == g_strcmp0 (key, "CanPlay"))
        {
            mpris2->can_play = g_variant_get_boolean(value);
            control_flags |= CDOS_MPRIS2_CAN_PLAY;
        }
        else if (0 == g_strcmp0 (key, "CanPause"))
        {
            mpris2->can_pause = g_variant_get_boolean(value);
            control_flags |= CDOS_MPRIS2_CAN_PAUSE;
        }
        else if (0 == g_strcmp0 (key, "CanSeek"))
        {
            mpris2->can_seek = g_variant_get_boolean(value);
            control_flags |= CDOS_MPRIS2_CAN_SEEK;
        }
        else if (0 == g_strcmp0 (key, "CanControl"))
        {
            mpris2->can_control = g_variant_get_boolean(value);
            control_flags |= CDOS_MPRIS2_CAN_CONTROL;
        }
        /* Optionals */
        else if (0 == g_strcmp0 (key, "LoopStatus"))
        {
            loop_status_changed = TRUE;
            loop_status = g_variant_get_string (value, NULL);
        }
        else
            continue;
    }
    if (control_flags)
        g_signal_emit (mpris2, cdos_mpris2_player_signals[CONTROL_FLAGS], 0, control_flags);

    if (metadata != NULL)
    {
        if (mpris2->metadata != NULL)
            cdos_mpris2_metadata_free (mpris2->metadata);
        mpris2->metadata = metadata;

        g_signal_emit (mpris2, cdos_mpris2_player_signals[METADATA], 0, metadata);
    }

    if (playback_status != NULL)
    {
        cdos_mpris2_player_parse_playback_status (mpris2, playback_status);
        g_signal_emit (mpris2, cdos_mpris2_player_signals[PLAYBACK_STATUS], 0, mpris2->playback_status);
    }

    if (volume != -1)
    {
        if (mpris2->volume != volume)
        {
            mpris2->volume = volume;
            g_signal_emit (mpris2, cdos_mpris2_player_signals[VOLUME], 0, volume);
        }
    }
}

static GVariant *
cdos_mpris2_player_get_player_property (CdosMpris2Player *self, const gchar *prop)
{
    GVariant *v, *iter;
    GError *error = NULL;

    v = g_dbus_connection_call_sync (self->connection,
                                     self->dbus_name,
                                     "/org/mpris/MediaPlayer2",
                                     "org.freedesktop.DBus.Properties",
                                     "Get",
                                     g_variant_new ("(ss)",
                                                    "org.mpris.MediaPlayer2.Player",
                                                    prop),
                                     G_VARIANT_TYPE ("(v)"),
                                     G_DBUS_CALL_FLAGS_NONE,
                                     1000, // timeout: 1000ms
                                     NULL,
                                     &error);
    if (error)
    {
        g_critical ("Could not get properties on org.mpris.MediaPlayer2, %s",
                    error ? error->message : "no error given");
        g_clear_error (&error);
        return NULL;
    }

    g_variant_get (v, "(v)", &iter);

    return iter;
}

/* Change any player property using org.freedesktop.DBus.Properties interfase. */

static void G_GNUC_UNUSED
cdos_mpris2_player_set_player_property (CdosMpris2Player *self, const gchar *prop, GVariant *vprop)
{
    GVariant *reply;
    GError   *error = NULL;

    reply = g_dbus_connection_call_sync (self->connection,
                                         self->dbus_name,
                                         "/org/mpris/MediaPlayer2",
                                         "org.freedesktop.DBus.Properties",
                                         "Set",
                                         g_variant_new ("(ssv)",
                                                        "org.mpris.MediaPlayer2.Player",
                                                        prop,
                                                        vprop),
                                         NULL,
                                         G_DBUS_CALL_FLAGS_NONE,
                                         1000,
                                         NULL,
                                         &error);
    if (reply == NULL)
    {
        g_warning ("Unable to set session: %s", error->message);
        g_error_free (error);
        return;
    }
    g_variant_unref(reply);
}

static gboolean
playback_tick_emit_cb (gpointer user_data)
{
    CdosMpris2Player *mpris2 = user_data;

    mpris2->position += (mpris2->rate) * 1000000;

    g_signal_emit (mpris2, cdos_mpris2_player_signals[PLAYBACK_TICK], 0, mpris2->position);

    return TRUE;
}

static void
cdos_mpris2_player_parse_playback_status (CdosMpris2Player *mpris2, const gchar *playback_status)
{
    GVariant *value;

    if (0 == g_strcmp0 (playback_status, "Playing"))
        mpris2->playback_status = CDOS_MPRIS2_PLAYING;
    else if (0 == g_strcmp0 (playback_status, "Paused"))
        mpris2->playback_status = CDOS_MPRIS2_PAUSED;
    else
        mpris2->playback_status = CDOS_MPRIS2_STOPPED;

    if (mpris2->playback_status == CDOS_MPRIS2_PLAYING)
    {
        value = cdos_mpris2_player_get_player_property (mpris2, "Position");
        mpris2->position = g_variant_get_int64 (value);

        if (mpris2->playback_timer_id == 0)
            mpris2->playback_timer_id = g_timeout_add_seconds (1, playback_tick_emit_cb, mpris2);
    }
    else
    {
        if (mpris2->playback_timer_id > 0)
        {
            g_source_remove (mpris2->playback_timer_id);
            mpris2->playback_timer_id = 0;
        }
    }
}

static void
cdos_mpris2_player_connected_dbus (GDBusConnection *connection,
                                   const gchar *name,
                                   const gchar *owner,
                                   gpointer     user_data)
{
    GVariant *reply;
    CdosMpris2Player *self = user_data;
    self->connected = TRUE;

    reply = cdos_mpris2_player_get_properties (self);
    cdos_mpris2_player_parse_properties (self, reply);
    g_variant_unref (reply);

    reply = cdos_mpris2_player_get_player_properties (self);
    cdos_mpris2_player_parse_player_properties (self, reply);
    g_variant_unref (reply);

    g_signal_emit (self, cdos_mpris2_player_signals[CONNECTED], 0, self->connected);
}

static void
cdos_mpris2_player_lose_dbus (GDBusConnection *connection,
                              const gchar *name,
                              gpointer user_data)
{
    CdosMpris2Player *mpris2 = user_data;

    /* Interface MediaPlayer2 */

    mpris2->can_quit        = FALSE;
    mpris2->can_raise       = FALSE;
    mpris2->has_tracklist   = FALSE;
    if (mpris2->identity)
    {
        g_free (mpris2->identity);
        mpris2->identity = NULL;
    }
    if (mpris2->supported_uri_schemes)
    {
        g_strfreev(mpris2->supported_uri_schemes);
        mpris2->supported_uri_schemes = NULL;
    }
    if (mpris2->supported_mime_types)
    {
        g_strfreev (mpris2->supported_mime_types);
        mpris2->supported_mime_types = NULL;
    }

    /* Optionals Interface MediaPlayer2 */
    if (mpris2->desktop_entry)
    {
        g_free (mpris2->desktop_entry);
        mpris2->desktop_entry = NULL;
    }

    /* Interface MediaPlayer2.Player */
    mpris2->playback_status = CDOS_MPRIS2_STOPPED;
    mpris2->rate            = 1.0;
    if (mpris2->metadata != NULL)
    {
        cdos_mpris2_metadata_free (mpris2->metadata);
        mpris2->metadata = NULL;
    }
    mpris2->volume          = -1;
    mpris2->position        = 0;
    mpris2->minimum_rate    = 1.0;
    mpris2->maximum_rate    = 1.0;
    mpris2->can_go_next     = FALSE;
    mpris2->can_go_previous = FALSE;
    mpris2->can_play        = FALSE;
    mpris2->can_pause       = FALSE;
    mpris2->can_seek        = FALSE;
    mpris2->can_control     = FALSE;

    /* Optionals Interface MediaPlayer2.Player */
    //mpris2->has_loop_status = FALSE;

    mpris2->connected = FALSE;
    g_signal_emit (mpris2, cdos_mpris2_player_signals[CONNECTED], 0, mpris2->connected);
}

static void
cdos_mpris2_player_on_dbus_props_signal (GDBusProxy *proxy,
                                         gchar      *sender,
                                         gchar      *signal_name,
                                         GVariant   *parameters,
                                         gpointer    user_data)
{
    GVariantIter iter;
    GVariant *child;

    CdosMpris2Player *self = user_data;

    if (g_strcmp0 (signal_name, "PropertiesChanged"))
        return;

    g_variant_iter_init (&iter, parameters);

    child = g_variant_iter_next_value (&iter); /* Interface name. */
    g_variant_unref (child);

    child = g_variant_iter_next_value (&iter); /* Property name. */
    cdos_mpris2_player_parse_player_properties (self, child);
    g_variant_unref (child);
}

static void
cdos_mpris2_player_on_dbus_player_signal (GDBusProxy   *proxy,
                                          gchar        *sender,
                                          gchar        *signal_name,
                                          GVariant     *parameters,
                                          gpointer      user_data)
{
    GVariantIter iter;
    GVariant *child;

    CdosMpris2Player *self = user_data;

    if (g_strcmp0 (signal_name, "Seeked"))
        return;

    g_variant_iter_init (&iter, parameters);

    child = g_variant_iter_next_value (&iter);

    self->position = g_variant_get_int64 (child);
    g_signal_emit (self, cdos_mpris2_player_signals[PLAYBACK_TICK], 0, self->position);
    g_variant_unref (child);
}

static void
cdos_mpris2_player_connect_dbus (CdosMpris2Player *mpris2)
{
    GDBusProxy *proxy;
    GError     *error = NULL;
    guint       watch_id;

    watch_id = g_bus_watch_name_on_connection (mpris2->connection,
                                               mpris2->dbus_name,
                                               G_BUS_NAME_OWNER_FLAGS_REPLACE,
                                               cdos_mpris2_player_connected_dbus,
                                               cdos_mpris2_player_lose_dbus,
                                               mpris2,
                                               NULL);

    /* interface=org.freedesktop.DBus.Properties */
    proxy = g_dbus_proxy_new_sync (mpris2->connection,
                                   G_DBUS_PROXY_FLAGS_NONE,
                                   NULL,
                                   mpris2->dbus_name,
                                   "/org/mpris/MediaPlayer2",
                                   "org.freedesktop.DBus.Properties",
                                   NULL, /* GCancellable */
                                   &error);

    if (proxy == NULL)
    {
        g_warning ("Error creating proxy: %s", error->message);
        g_error_free (error);
        error = NULL;
    }
    else
    {
        g_signal_connect (proxy, "g-signal",
                          G_CALLBACK (cdos_mpris2_player_on_dbus_props_signal), mpris2);
        mpris2->props_proxy = proxy;
    }

    /* interface=org.mpris.MediaPlayer2.Player */
    proxy = g_dbus_proxy_new_sync (mpris2->connection,
                                   G_DBUS_PROXY_FLAGS_NONE,
                                   NULL,
                                   mpris2->dbus_name,
                                   "/org/mpris/MediaPlayer2",
                                   "org.mpris.MediaPlayer2.Player",
                                   NULL, /* GCancellable */
                                   &error);

    if (proxy == NULL)
    {
        g_warning ("Error creating proxy: %s", error->message);
        g_error_free (error);
        error = NULL;
    }
    else
    {
        g_signal_connect (proxy, "g-signal",
                          G_CALLBACK (cdos_mpris2_player_on_dbus_player_signal), mpris2);
        mpris2->player_proxy = proxy;
    }

    mpris2->watch_id = watch_id;
}

void
cdos_mpris2_player_set_player (CdosMpris2Player *self, GDBusConnection *connection, const char *name)
{
    self->dbus_name = g_strdup (name);
    if (connection)
        self->connection = g_object_ref (connection);

    cdos_mpris2_player_connect_dbus (self);
}

const gchar *
cdos_mpris2_player_get_player (CdosMpris2Player *self)
{
    return self->dbus_name;
}

void
cdos_mpris2_player_raise_player (CdosMpris2Player *mpris2)
{
    if (mpris2->can_raise)
        cdos_mpris2_player_call_method (mpris2, "Raise");
}

void
cdos_mpris2_player_quit_player (CdosMpris2Player *mpris2)
{
    cdos_mpris2_player_call_method (mpris2, "Quite");
}

const gchar
*cdos_mpris2_player_get_player_desktop_entry  (CdosMpris2Player *mpris2)
{
    return mpris2->desktop_entry;
}

static void
cdos_mpris2_player_dispose (GObject *gobject)
{
    CdosMpris2Player *self = CDOS_MPRIS2_PLAYER (gobject);
    g_clear_pointer (&self->connection, g_object_unref);
    if (self->playback_timer_id > 0)
    {
        g_source_remove (self->playback_timer_id);
        self->playback_timer_id = 0;
    }
    if (self->watch_id)
    {
        g_bus_unwatch_name (self->watch_id);
        self->watch_id = 0;
    }
    g_clear_pointer (&self->props_proxy, g_object_unref);
    g_clear_pointer (&self->player_proxy, g_object_unref);

    G_OBJECT_CLASS (cdos_mpris2_player_parent_class)->dispose (gobject);
}

static void
cdos_mpris2_player_finalize (GObject *gobject)
{
    CdosMpris2Player *mpris2 = CDOS_MPRIS2_PLAYER (gobject);

    if (mpris2->dbus_name)
        g_free (mpris2->dbus_name);

    if (mpris2->identity)
        g_free (mpris2->identity);

    if (mpris2->desktop_entry)
        g_free (mpris2->desktop_entry);

    if (mpris2->supported_uri_schemes)
        g_strfreev (mpris2->supported_uri_schemes);
    if (mpris2->supported_mime_types)
        g_strfreev (mpris2->supported_mime_types);

    if (mpris2->metadata != NULL)
        cdos_mpris2_metadata_free (mpris2->metadata);

    G_OBJECT_CLASS (cdos_mpris2_player_parent_class)->finalize (gobject);
}

static void
cdos_mpris2_player_init (CdosMpris2Player *self)
{
    cdos_mpris2_player_reset (self);
}

static void
cdos_mpris2_player_class_init (CdosMpris2PlayerClass *klass)
{
    GObjectClass  *gobject_class;

    gobject_class = G_OBJECT_CLASS (klass);
    gobject_class->dispose = cdos_mpris2_player_dispose;
    gobject_class->finalize = cdos_mpris2_player_finalize;

    cdos_mpris2_player_signals[CONNECTED] =
        g_signal_new ("connected",
                      CDOS_TYPE_MPRIS2_PLAYER,
                      G_SIGNAL_RUN_LAST,
                      0,
                      NULL, NULL,
                      g_cclosure_marshal_VOID__BOOLEAN,
                      G_TYPE_NONE, 1, G_TYPE_BOOLEAN);

    cdos_mpris2_player_signals[CONTROL_FLAGS] =
        g_signal_new ("control-flags",
                      CDOS_TYPE_MPRIS2_PLAYER,
                      G_SIGNAL_RUN_LAST,
                      0,
                      NULL, NULL,
                      g_cclosure_marshal_VOID__UINT,
                      G_TYPE_NONE, 1, G_TYPE_UINT);

    cdos_mpris2_player_signals[PLAYBACK_STATUS] =
        g_signal_new ("playback-status",
                      G_TYPE_FROM_CLASS (gobject_class),
                      G_SIGNAL_RUN_LAST,
                      0,
                      NULL, NULL,
                      g_cclosure_marshal_VOID__ENUM,
                      G_TYPE_NONE, 1, G_TYPE_INT);

    cdos_mpris2_player_signals[PLAYBACK_TICK] =
        g_signal_new ("playback-tick",
                      G_TYPE_FROM_CLASS (gobject_class),
                      G_SIGNAL_RUN_LAST,
                      0,
                      NULL, NULL,
                      g_cclosure_marshal_VOID__LONG,
                      G_TYPE_NONE, 1, G_TYPE_INT64);

    cdos_mpris2_player_signals[METADATA] =
        g_signal_new ("metadata",
                      G_TYPE_FROM_CLASS (gobject_class),
                      G_SIGNAL_RUN_LAST,
                      0,
                      NULL, NULL,
                      g_cclosure_marshal_VOID__POINTER,
                      G_TYPE_NONE, 1, G_TYPE_POINTER);

    cdos_mpris2_player_signals[VOLUME] =
        g_signal_new ("volume",
                      G_TYPE_FROM_CLASS (gobject_class),
                      G_SIGNAL_RUN_LAST,
                      0,
                      NULL, NULL,
                      g_cclosure_marshal_VOID__DOUBLE,
                      G_TYPE_NONE, 1, G_TYPE_DOUBLE);
}

CdosMpris2Player *
cdos_mpris2_player_new (void)
{
    CdosMpris2Player *mpris2 = g_object_new (CDOS_TYPE_MPRIS2_PLAYER, NULL);

    return mpris2;
}

void
cdos_mpris2_player_previous (CdosMpris2Player *mpris2)
{
    if (!mpris2->connected)
        return;

    if (!mpris2->can_control)
        return;

    if (!mpris2->can_go_previous)
        return;

    cdos_mpris2_player_call_player_method (mpris2, "Previous");
}

void
cdos_mpris2_player_next (CdosMpris2Player *mpris2)
{
    if (!mpris2->connected)
        return;

    if (!mpris2->can_control)
        return;

    if (!mpris2->can_go_next)
        return;

    cdos_mpris2_player_call_player_method (mpris2, "Next");
}

void
cdos_mpris2_player_pause (CdosMpris2Player *mpris2)
{
    if (!mpris2->connected)
        return;

    if (!mpris2->can_control)
        return;

    if (!mpris2->can_pause)
        return;

    cdos_mpris2_player_call_player_method (mpris2, "Pause");
}

void
cdos_mpris2_player_play_pause (CdosMpris2Player *mpris2)
{
    if (!mpris2->connected)
        return;

    if (!mpris2->can_control)
        return;

    if (!mpris2->can_pause)
        return;

    cdos_mpris2_player_call_player_method (mpris2, "PlayPause");
}

void
cdos_mpris2_player_stop (CdosMpris2Player *mpris2)
{
    if (!mpris2->connected)
        return;

    if (!mpris2->can_control)
        return;

    if (mpris2->strict_mode && !mpris2->can_control)
        return;

    cdos_mpris2_player_call_player_method (mpris2, "Stop");
}

void
cdos_mpris2_player_play (CdosMpris2Player *mpris2)
{
    if (!mpris2->connected)
        return;

    if (!mpris2->can_control)
        return;

    if (mpris2->strict_mode && !mpris2->can_play)
        return;

    cdos_mpris2_player_call_player_method (mpris2, "Play");
}

void
cdos_mpris2_player_seek (CdosMpris2Player *mpris2, gint64 offset)
{
    if (mpris2->position + offset > mpris2->metadata->length)
        return;

    GDBusMessage *message;
    GError       *error = NULL;
    message = g_dbus_message_new_method_call (mpris2->dbus_name,
                                              "/org/mpris/MediaPlayer2",
                                              "org.mpris.MediaPlayer2.Player",
                                              "Seek");
    g_dbus_message_set_body (message, g_variant_new ("(x)", offset));

    g_dbus_connection_send_message (mpris2->connection,
                                    message,
                                    G_DBUS_SEND_MESSAGE_FLAGS_NONE,
                                    NULL,
                                    &error);
    g_dbus_connection_flush_sync (mpris2->connection, NULL, &error);

    /*
    GVariant *v;
    v = g_dbus_connection_call_sync (mpris2->connection,
                                     mpris2->dbus_name,
                                     "/org/mpris/MediaPlayer2",
                                     "org.mpris.MediaPlayer2.Player",
                                     "Seek",
                                     g_variant_new ("(x)", offset),
                                     NULL, //G_VARIANT_TYPE ("(v)"),
                                     G_DBUS_CALL_FLAGS_NONE,
                                     1000, // timeout: 1000ms
                                     NULL,
                                     &error);
    */
    if (error != NULL)
    {
        g_warning ("unable to send message: %s", error->message);
        g_clear_error (&error);
        error = NULL;
    }
    g_object_unref (message);
}

gint64
cdos_mpris2_player_get_position (CdosMpris2Player *mpris2)
{
    return mpris2->position;
}

gint64
cdos_mpris2_player_get_accurate_position (CdosMpris2Player *mpris2)
{
    GVariant *v = cdos_mpris2_player_get_player_property (mpris2, "Position");
    if (v)
        return g_variant_get_int64 (v);
    return 0;
}

gdouble
cdos_mpris2_player_get_volume (CdosMpris2Player *mpris2)
{
    return mpris2->volume;
}

gboolean
cdos_mpris2_player_get_can_go_next (CdosMpris2Player *mpris2)
{
    return mpris2->can_go_next;
}

gboolean
cdos_mpris2_player_get_can_go_previous (CdosMpris2Player *mpris2)
{
    return mpris2->can_go_previous;
}

gboolean
cdos_mpris2_player_get_can_play (CdosMpris2Player *mpris2)
{
    return mpris2->can_play;
}

gboolean
cdos_mpris2_player_get_can_pause (CdosMpris2Player *mpris2)
{
    return mpris2->can_pause;
}

gboolean
cdos_mpris2_player_get_can_seek (CdosMpris2Player *mpris2)
{
    return mpris2->can_seek;
}

gboolean
cdos_mpris2_player_get_can_control (CdosMpris2Player *mpris2)
{
    return mpris2->can_control;
}

/*
int main ()
{
    GMainLoop *loop = NULL;
    loop = g_main_loop_new (NULL, FALSE);
    CdosMpris2Player *player = cdos_mpris2_player_new ();
    g_main_loop_run (loop);
    return 0;
}
*/
