/**
 * Copyright (c) 2016-2018 CPU and Fundamental Software Research Center, CAS
 *
 * This software is published by the license of CPU-OS License，you can use and
 * distibuted this software under this License. See CPU-OS Liense for more detail.
 *
 * You should have receive a copy of CPU-OS License. If not, please contact us
 * by email <support_os@cpu-os.ac.cn>
 *
 **/
/**
 * SECTION: cdos-notification-daemon-source
 * @short_description: 通知守护进程数据结构
 * @include: cdos-notification-daemon-source.h
 *
 * CdosNotificationDaemonSourceClass 类用来处理通知
 */
#include <gio/gio.h>
#include "cdos-notification-daemon-source.h"

enum {
    PROP_0,
    PROP_PID,
    PROP_SENDER,
    PROP_LEVEL,
    NUM_PROPERTIES
};

struct _CdosNotificationDaemonSourcePrivate
{
    gchar *sender;
    guint pid;
    guint level;
    ClutterActor *icon;
    guint watch_id;
};

G_DEFINE_TYPE_WITH_PRIVATE (CdosNotificationDaemonSource, cdos_notification_daemon_source, CDOS_TYPE_SOURCE);

static void cdos_notification_daemon_source_on_name_vanished (GDBusConnection *connection, const gchar *name,
                                                              CdosNotificationDaemonSource *self);

static void
cdos_notification_daemon_source_dispose (GObject *gobject)
{
    CdosNotificationDaemonSource *self = CDOS_NOTIFICATION_DAEMON_SOURCE (gobject);
    if (self->priv->watch_id)
    {
        g_bus_unwatch_name (self->priv->watch_id);
        self->priv->watch_id = 0;
    }
    G_OBJECT_CLASS (cdos_notification_daemon_source_parent_class)->dispose (gobject);
}

static void
cdos_notification_daemon_source_finalize (GObject *gobject)
{
    G_OBJECT_CLASS (cdos_notification_daemon_source_parent_class)->finalize (gobject);
}

static void
cdos_notification_daemon_source_set_property (GObject *gobject,
                                              guint prop_id,
                                              const GValue *value,
                                              GParamSpec *pspec)
{
    CdosNotificationDaemonSource *self = CDOS_NOTIFICATION_DAEMON_SOURCE (gobject);

    switch (prop_id)
    {
        case PROP_PID:
            cdos_notification_daemon_source_set_pid (self, g_value_get_uint (value));
            break;
        case PROP_SENDER:
            cdos_notification_daemon_source_set_sender (self, (gchar *)g_value_get_string (value));
            break;
        case PROP_LEVEL:
            cdos_notification_daemon_source_set_level (self, g_value_get_uint (value));
            break;
        default:
            G_OBJECT_WARN_INVALID_PROPERTY_ID (gobject, prop_id, pspec);
            break;
    }
}

static void
cdos_notification_daemon_source_get_property (GObject *gobject,
                                              guint prop_id,
                                              GValue *value,
                                              GParamSpec *pspec)
{
    CdosNotificationDaemonSource *
    new_obj = CDOS_NOTIFICATION_DAEMON_SOURCE (gobject);

    switch (prop_id)
    {
        case PROP_PID:
            g_value_set_int (value, new_obj->priv->pid);
            break;

        case PROP_SENDER:
            g_value_set_string (value, new_obj->priv->sender);
            break;

        case PROP_LEVEL:
            g_value_set_int (value, new_obj->priv->level);
            break;

        default:
            G_OBJECT_WARN_INVALID_PROPERTY_ID (gobject, prop_id, pspec);
            break;
    }
}

/**
 * cdos_notification_daemon_source_destory:
 * @self: A #CdosNotificationDaemonSource
 *
 * 调用父类方法删除self
 *
 */
static void
cdos_notification_daemon_source_destory (CdosNotificationDaemonSource *self)
{
    cdos_source_destroy (CDOS_SOURCE(self), CDOS_NOTIFICATION_UNDEFINED_RESERVED);
}

/**
 * cdos_notification_daemon_source_set_sender:
 * @self: A #CdosNotificationDaemonSource
 * @sender: 发送者
 *
 * 设置self的"sender"属性
 *
 */
void
cdos_notification_daemon_source_set_sender (CdosNotificationDaemonSource *self, const gchar *sender)
{
    if (self->priv->sender)
    {
        g_free (self->priv->sender);
        self->priv->sender = NULL;
    }
    self->priv->sender = g_strdup (sender);
}

/**
 * cdos_notification_daemon_source_set_pid:
 * @self: A #CdosNotificationDaemonSource
 * @pid: pid
 *
 * 设置self的"pid"属性
 *
 */
void
cdos_notification_daemon_source_set_pid (CdosNotificationDaemonSource *self, guint pid)
{
    self->priv->pid = pid;
}

void
cdos_notification_daemon_source_set_level (CdosNotificationDaemonSource *self, guint level)
{
    self->priv->level = level;
}

/**
 * cdos_notification_daemon_source_get_sender:
 * @self: A #CdosNotificationDaemonSource
 * @sender: 发送者
 *
 * 读取self的"sender"属性
 *
 * Returns: sender.
 */
gchar *
cdos_notification_daemon_source_get_sender (CdosNotificationDaemonSource *self)
{
    return self->priv->sender;
}

/**
 * cdos_notification_daemon_source_set_pid:
 * @self: A #CdosNotificationDaemonSource
 * @pid: pid
 *
 * 获取self的"pid"属性
 *
 * Returns: pid.
 */
guint
cdos_notification_daemon_source_get_pid (CdosNotificationDaemonSource *self)
{
    return self->priv->pid;
}

guint
cdos_notification_daemon_source_get_level (CdosNotificationDaemonSource *self)
{
    return self->priv->level;
}

void
cdos_notification_daemon_source_on_name_vanished (GDBusConnection *connection, const gchar *name,
                                                  CdosNotificationDaemonSource *self)
{
    /*
     * Destroy the notification source when its sender is removed from DBus.
     * Only do so if this.app is set to avoid removing "notify-send" sources, senders
     * of which аre removed from DBus immediately.
     * Sender being removed from DBus would normally result in a tray icon being removed,
     * so allow the code path that handles the tray icon being removed to handle that case.
     */
    if (self->app)
    {
        cdos_notification_daemon_source_destory (self);
    }
}

static void
cdos_notification_daemon_source_set_app (CdosNotificationDaemonSource *self)
{
    if (self->app)
    {
        return;
    }

    guint pid = self->priv->pid;
    CdosWindowTracker *tracker;
    tracker = cdos_window_tracker_get_default ();
    self->app = cdos_window_tracker_get_app_from_pid (tracker, pid);
}

static void
cdos_notification_daemon_source_open_app (CdosNotificationDaemonSource *self)
{
    if (self->app == NULL)
        return;
    GSList *windows;
    windows = cdos_app_get_windows (self->app);
    if (windows)
    {
        MetaWindow *window = windows->data;
        meta_window_activate (window, 0);
    }
}

/**
 * cdos_notification_daemon_source_open:
 * @self: A #CdosNotificationDaemonSource
 * @notification: 打开notification
 *
 * 打开APP
 *
 */
static void
cdos_notification_daemon_source_open (CdosSource *source)
{
    //cdos_source_destroy_non_resident_notifications (source);
    cdos_notification_daemon_source_open_app ((CdosNotificationDaemonSource *)source);
}

/**
 * cdos_notification_daemon_source_process_notification:
 * @self: A #CdosNotificationDaemonSource
 * @notification: 接收的通知
 * @icon: 图标
 *
 * 处理通知的进程程序。
 */
void
cdos_notification_daemon_source_process_notification (CdosNotificationDaemonSource *self,
                                                      CdosNotification *notification)
{
    if (!self->app)
    {
        cdos_notification_daemon_source_set_app (self);
    }
    CdosWindowTracker *tracker = cdos_window_tracker_get_default ();
    CdosApp *focus_app;
    g_object_get (tracker, "focus-app", &focus_app, NULL);

    if (notification->resident && (notification->in_notification_center))
        /*原为:notification.resident && this.app && tracker.focus_app == this.app*/
        cdos_source_push_notification (CDOS_SOURCE (self), notification);
    else
        cdos_source_notify (CDOS_SOURCE (self), notification);
}

static ClutterActor *
cdos_notification_daemon_source_create_icon (CdosSource *source)
{
    CdosNotificationDaemonSource *self = CDOS_NOTIFICATION_DAEMON_SOURCE (source);
    if (self->app)
        return cdos_app_create_icon_texture (self->app, SOURCE_ICON_SIZE);
    else
    {
        return CDOS_SOURCE_CLASS (cdos_notification_daemon_source_parent_class)->create_icon (source);
    }
}

static void
cdos_notification_daemon_source_class_init (CdosNotificationDaemonSourceClass *klass)
{
    GObjectClass *gobject_class = G_OBJECT_CLASS(klass);
    gobject_class->dispose = cdos_notification_daemon_source_dispose;
    gobject_class->finalize = cdos_notification_daemon_source_finalize;

    gobject_class->get_property = cdos_notification_daemon_source_get_property;
    gobject_class->set_property = cdos_notification_daemon_source_set_property;

    CdosSourceClass *cdos_source_class = CDOS_SOURCE_CLASS (klass);
    cdos_source_class->open = cdos_notification_daemon_source_open;
    cdos_source_class->create_icon = cdos_notification_daemon_source_create_icon;

    /**
     * CdosNotificationDaemonSource:pid:
     * @gobject_class: 属性绑定的对象
     *
     * #CdosNotificationDaemonSource 的pid属性
     */
    g_object_class_install_property (gobject_class,
                                     PROP_PID,
                                     g_param_spec_uint ("pid", /*identifier*/
                                                        "PID", /*nickname*/
                                                        "The PID of the Notification Daemon Source",/*discreption*/
                                                        0,/*min*/
                                                        G_MAXUINT,/*max*/
                                                        0,/*default*/
                                                        G_PARAM_READWRITE/*flags*/
                                     ));

    /**
     * CdosNotificationDaemonSource:sender:
     * @gobject_class: 属性绑定的对象
     *
     * #CdosNotificationDaemonSource 的sender属性
     */
    g_object_class_install_property (gobject_class,
                                     PROP_SENDER,
                                     g_param_spec_string ("sender",
                                                          "Sender",
                                                          "The sender of Notification Daemon Source",
                                                          NULL,
                                                          G_PARAM_READWRITE
                                     ));
}

static void
cdos_notification_daemon_source_init (CdosNotificationDaemonSource *self)
{
    self->priv = cdos_notification_daemon_source_get_instance_private (self);
}

/**
 * cdos_notification_daemon_source_new:
 * @name: 标题
 * @pid: pid
 * @sender: 发送者
 *
 * 新建一个 #CdosNotificationDaemonSource 类的实例。
 *
 * Returns: 返回一个 #CdosNotificationDaemonSource 的实例。
 */
CdosNotificationDaemonSource *
cdos_notification_daemon_source_new (const gchar *name, guint pid, const gchar *sender)
{
    CdosNotificationDaemonSource *self = g_object_new (CDOS_TYPE_NOTIFICATION_DAEMON_SOURCE, NULL);
    cdos_source_set_name (CDOS_SOURCE (self), name);
    cdos_source_set_app_name (CDOS_SOURCE (self), name);
    cdos_notification_daemon_source_set_pid (self, pid);
    cdos_notification_daemon_source_set_sender (self, sender);

    if (sender)
    {
        /*
         * TODO:
         * dbus-glib implementation of watch_name() doesn’t return an id to be used for
         * unwatch_name() or implement unwatch_name(), however when we move to using GDBus implementation,
         * we should save the id here and call unwatch_name() with it in destroy().
         * Moving to GDBus is the work in progress: https://bugzilla.gnome.org/show_bug.cgi?id=648651
         * and https://bugzilla.gnome.org/show_bug.cgi?id=622921 .
         */

        self->priv->watch_id = g_bus_watch_name (G_BUS_TYPE_SESSION,
                                                 sender,
                                                 G_BUS_NAME_WATCHER_FLAGS_NONE,
                                                 NULL,
                                                 (GBusNameVanishedCallback) cdos_notification_daemon_source_on_name_vanished,
                                                 self,
                                                 NULL);

        cdos_notification_daemon_source_set_app (self);
        if (self->app)
            cdos_source_set_name ((CdosSource *)self, cdos_app_get_name (self->app));
        ClutterActor *icon = cdos_notification_daemon_source_create_icon ((CdosSource *)self);
        CDOS_SOURCE (self)->icon = icon;
        clutter_actor_insert_child_at_index ((ClutterActor *)self, icon, 0);
    }

    return self;
}
