/**
 * Copyright (c) 2016-2018 CPU and Fundamental Software Research Center, CAS
 *
 * This software is published by the license of CPU-OS License，you can use and
 * distibuted this software under this License. See CPU-OS Liense for more detail.
 *
 * You should have receive a copy of CPU-OS License. If not, please contact us
 * by email <support_os@cpu-os.ac.cn>
 *
 **/
/**
 * SECTION: cdos-notification-daemon
 * @short_description: 通知守护进程
 * @include: cdos-notification-daemon.h
 *
 * CdosNotificationDaemonClass 类用来处理消息转发
 */
#include <stdio.h>
#include <string.h>
#include <glib.h>
#include <regex.h>

#include "cdos-ui-plugin.h"
#include "cdos-notification-daemon-source.h"
#include "cdos-notification-daemon.h"
#include "cdos-utils-private.h"
#include "cdos-sound-manager.h"
#include "cdos-message-tray.h"

#define IMAGE_SIZE 125

G_DEFINE_TYPE (CdosNotificationDaemon, cdos_notification_daemon, G_TYPE_OBJECT)

/*G_CALLBACK user_data*/
/*use in notification.connect-destroy*/
typedef struct _notificationParams notificationParams;
/*use in Notify, 二级回调*/
typedef struct _sourceParams sourceParams;
typedef struct _callMethodParams callMethodParams;
typedef struct _ndataStruct ndataStruct;
typedef struct _ActionInvokedData ActionInvokedData;

static const gchar introspection_xml[] =
"<node name=\"/org/freedesktop/Notifications\">"
 "<interface name=\"org.freedesktop.Notifications\">"
   "<method name=\"Notify\">"
     "<arg type=\"s\" name=\"app_name\" direction=\"in\" />"
     "<arg type=\"u\" name=\"id\" direction=\"in\" />"
     "<arg type=\"s\" name=\"icon\" direction=\"in\" />"
     "<arg type=\"s\" name=\"summary\" direction=\"in\" />"
     "<arg type=\"s\" name=\"body\" direction=\"in\" />"
     "<arg type=\"as\" name=\"actions\" direction=\"in\" />"
     "<arg type=\"a{sv}\" name=\"hints\" direction=\"in\" />"
     "<arg type=\"i\" name=\"timeout\" direction=\"in\" />"
     "<arg type=\"u\" name=\"return_id\" direction=\"out\" />"
   "</method>"
   "<method name=\"CloseNotification\">"
     "<arg  type=\"u\" direction=\"in\" />"
   "</method>"
   "<method name=\"GetCapabilities\">"
     "<arg  type=\"as\" direction=\"out\" />"
   "</method>"
   "<method name=\"GetServerInformation\">"
     "<arg  type=\"s\" name=\"return_name\"         direction=\"out\" />"
     "<arg  type=\"s\" name=\"return_vendor\"       direction=\"out\" />"
     "<arg  type=\"s\" name=\"return_version\"      direction=\"out\" />"
     "<arg  type=\"s\" name=\"return_spec_version\" direction=\"out\" />"
   "</method>"
   "<signal name=\"NotificationClosed\">"
     "<arg  type=\"uu\" />"
   "</signal>"
   "<signal name=\"ActionInvoked\">"
     "<arg  type=\"us\" />"
   "</signal>"
 "</interface>"
"</node>";

struct _notificationParams
{
    guint ndata_id;
    CdosNotificationDaemon *self;
};

struct _sourceParams
{
    gchar *sender;
    CdosNotificationDaemon *self;
};

struct _callMethodParams
{
    guint id;
    gchar *sender;
    CdosNotificationDaemon *self;
    GObject *ex;//没用到
};

struct _ActionInvokedData
{
    CdosNotificationDaemon *daemon;
    guint id;
    gchar *action;
};

/*ndata结构*/
struct _ndataStruct
{
    gchar **actions;
    gchar *app_name;
    gchar *body;
    GHashTable *hints;
    guint id;
    gchar *icon;
    CdosNotification *notification;
    gchar *summary;
    guint timeout;
};

static CdosNotificationDaemonSource *cdos_notification_daemon_lookup_source (CdosNotificationDaemon *self,
                                                                             const gchar *title,
                                                                             pid_t pid);

static CdosSource *cdos_notification_daemon_get_source (CdosNotificationDaemon *self, const gchar *title,
                                                        pid_t pid, ndataStruct *ndata, const gchar *sender);

static void cdos_notification_daemon_notify_for_source (CdosNotificationDaemon *self,
                                                        CdosNotificationDaemonSource *source, ndataStruct *ndata);

static void cdos_notification_daemon_emit_notification_closed (CdosNotificationDaemon *self, guint id, CdosNotificationClosedReason reason);

static void cdos_notification_daemon_emit_action_invoked (CdosNotificationDaemon *self, guint id, const gchar *action);

static void generate_new_proxy (CdosNotificationDaemon *self);

static guint get_hints_urgency (GHashTable *hints);

static gboolean get_hints_transient (GHashTable *hints);

static gboolean get_hints_resident (GHashTable *hints);

static void proxy_sender_cb (GObject *source_obj , GAsyncResult *result, gpointer user_data);
static void on_notification_dbus_acquried(GObject *obj, const gchar *name, gpointer user_data);
static void handle_method_call (GDBusConnection *connection G_GNUC_UNUSED,
                                const gchar *sender G_GNUC_UNUSED,
                                const gchar *object_path G_GNUC_UNUSED,
                                const gchar *interface_name,
                                const gchar *method_name,
                                GVariant *parameters,
                                GDBusMethodInvocation *invocation,
                                CdosNotificationDaemon *self);

static const GDBusInterfaceVTable interface_vtable = {
    (GDBusInterfaceMethodCallFunc)handle_method_call,
    NULL,
    NULL,
};

static guint next_notification_id = 1;

static void generate_new_proxy (CdosNotificationDaemon *self)
{
    GError *proxy_error = NULL;
    self->bus_proxy = NULL;
    /*GDBusInterfaceInfo *interfaceInfo = org_freedesktop_dbus_interface_info();*/

    /*新建代理*/
    const gchar *bus_name = "org.freedesktop.DBus";
    const gchar *bus_path = "/org/freedesktop/DBus";
    const gchar *interface_name = "org.freedesktop.DBus";

    /*原型:g_dbus_proxy_new_sync(connection, G_DBUS_PROXY_FLAGS_NONE, interfaceInfo, bus_name,bus_path, cancellable, &error);*/
    /*该函数是生成函数, ,类型转换:OrgFreedesktopDBus->GDBusProxy*/

    self->bus_proxy = g_dbus_proxy_new_for_bus_sync (G_BUS_TYPE_SESSION,
                                                     G_DBUS_PROXY_FLAGS_DO_NOT_LOAD_PROPERTIES,
                                                     NULL,
                                                     bus_name,
                                                     bus_path,
                                                     interface_name,
                                                     NULL, &proxy_error);
    if (self->bus_proxy == NULL)
    {
        g_warning ("Couldn't get org.freedesktop.DBus proxy: %s", proxy_error->message);
        if (proxy_error)
            g_error_free (proxy_error);
        return;
    }

}

/*
 image-path, image-data: > V1.2, latest version
 image_path, image_data: v1.1, old version
 icon_data: early version, only use in image-path not available
 hints_urgency: 0, 1, 2
 hints_transient,
 hints_resident,
 hints_action-icons
*/


/**
 * image-data: (iiibiiay)
 * icon-data : (iiibiiay)
 */
static ClutterImage *
cdos_image_from_data (GVariant *image_data)
{
    ClutterImage *image = NULL;
    gint32 width, height, rowstride, bits_per_sample, channels;
    gboolean has_alpha;
    GVariant *pixel_data;
    gsize correct_len;
    const guchar *data;
    if (!g_variant_is_of_type (image_data, G_VARIANT_TYPE ("(iiibiiay)")))
    {
        g_warning ("Image data is not the correct type");
        return NULL;
    }
    g_variant_get (image_data,
                   "(iiibii@ay)",
                   &width,
                   &height,
                   &rowstride,
                   &has_alpha,
                   &bits_per_sample,
                   &channels,
                   &pixel_data);

    correct_len = (height - 1) * rowstride + width
                  * ((channels * bits_per_sample + 7) / 8);
    if (correct_len != g_variant_get_size (pixel_data))
    {
        g_warning ("Pixel data length (%lu) did not match expected value (%u)",
                   g_variant_get_size (pixel_data), (guint)correct_len);
        return NULL;
    }

    /*
    data = (guchar *) g_memdup (g_variant_get_data (pixel_data),
                                g_variant_get_size (pixel_data));
    */
    data = g_variant_get_data (pixel_data);

    image = (ClutterImage *)clutter_image_new ();
    GError *error = NULL;
    clutter_image_set_data (image,
                            data,
                            has_alpha ? COGL_PIXEL_FORMAT_RGBA_8888 : COGL_PIXEL_FORMAT_RGB_888,
                            width, height,
                            rowstride,
                            &error);
    if (error)
    {
        g_warning ("%s", error->message);
        g_object_unref (image);
        return NULL;
    }
    return image;
}

/*解析hints:所有版本的image path, 解析hints:image-path, Latest Version*/
static const
gchar *get_hints_image_path (GHashTable *hints)
{
    const gchar *image_path = NULL;
    image_path = g_hash_table_lookup (hints, "image-path");
    if (!image_path)
        image_path = g_hash_table_lookup (hints, "image_path");

    return image_path;
}

static guint
get_hints_urgency (GHashTable *hints)
{
    if (!g_hash_table_contains (hints, "urgency"))
        return CDOS_NOTIFICATION_NORMAL;

    gulong urgency = (gulong)(gpointer)g_hash_table_lookup (hints, "urgency");
    return urgency;
}

static gboolean
get_hints_transient (GHashTable *hints)
{
    gboolean transient = (gulong)(gpointer)g_hash_table_lookup (hints, "transient");
    return transient;
}

static gboolean
get_hints_action_icons (GHashTable *hints)
{
    gboolean hints_action_icons = (gboolean)(gulong)g_hash_table_lookup (hints, "action-icons");

    return hints_action_icons;
}

static gboolean
get_hints_resident (GHashTable *hints)
{
    gboolean resident = (gulong)g_hash_table_lookup (hints, "resident");

    return resident;
}

static const gchar *
get_hints_sound_file (GHashTable *hints)
{
    const gchar *sound_file = g_hash_table_lookup (hints, "sound-file");
    return sound_file;
}

static const gchar *
get_hints_sound_name (GHashTable *hints)
{
    const gchar *sound_name = g_hash_table_lookup (hints, "sound-name");
    return sound_name;
}

static const gchar *
get_hints_category (GHashTable *hints)
{
    const gchar *category = g_hash_table_lookup (hints, "category");
    return category;
}

static void
splice_source_cb (CdosSource *source, CdosNotificationDaemon *self)
{
    self->sources = g_list_remove (self->sources, source);
}

static void
source_destroy_cb (CdosSource *source, gpointer user_data)
{
    sourceParams *params = (sourceParams *)user_data;

    CdosNotificationDaemon *self = params->self;

    gchar *sender = params->sender;

    if (!g_hash_table_remove (self->sender_to_pid, sender))
        g_warning ("The key of sender(%s) is not found!", sender);
    g_free(params->sender);
    g_free(params);
}

static void
notification_destroy_cb (CdosNotification *notification, gpointer user_data)
{
    notificationParams *params = (notificationParams *)user_data;
    CdosNotificationDaemon *self = params->self;
    guint ndata_id = params->ndata_id;

    g_hash_table_remove (self->notifications, GUINT_TO_POINTER (ndata_id));

    cdos_notification_daemon_emit_notification_closed (self, ndata_id, notification->closed_reason);
    g_free(params);
}

static void
proxy_sender_cb (GObject *source_obj, GAsyncResult *result, gpointer user_data)
{
    callMethodParams *params = (callMethodParams *)user_data;
    CdosNotificationDaemon *self = params->self;
    guint id = params->id;
    gchar *sender = params->sender;
    GError *call_error = NULL;
    g_free(user_data);
    GVariant *res;
    guint pid;
    GDBusProxy *proxy = G_DBUS_PROXY (source_obj);

    res = g_dbus_proxy_call_finish (proxy, result, &call_error);
    g_variant_get (res, "(u)", &pid);

    if (res == NULL)
    {
        g_warning ("couldn't logout using proxy_sender_cb: %s",
                   call_error->message);
    }
    else
    {
        g_variant_unref (res);
    }
    if (call_error)
        g_error_free (call_error);
    ndataStruct *ndata = g_hash_table_lookup (self->notifications, GUINT_TO_POINTER(id));

    if (!ndata)
    {
        return;
    }

    gchar *app_name = ndata->app_name;
    CdosSource *source = cdos_notification_daemon_get_source (self, app_name,
                                                              pid, ndata,
                                                              sender);
    /* We only store sender-pid entries for persistent sources.*/
    /* Removing the entries once the source is destroyed*/
    /* would result in the entries associated with transient*/
    /* sources removed once the notification is shown anyway.*/
    /* However, keeping these pairs would mean that we would*/
    /* possibly remove an entry associated with a persistent*/
    /* source when a transient source for the same sender is*/
    /* distroyed.*/

    if (!source->is_transient)
    {
        if (!g_hash_table_insert (self->sender_to_pid, sender, GSIZE_TO_POINTER(pid)))
            g_warning ("key (%s) already exist", sender);
        sourceParams *data = g_new0(sourceParams, 1);
        data->self = self;
        data->sender = g_strdup (sender);
        g_signal_connect (source, "destroy", G_CALLBACK (source_destroy_cb), data);
    }

    cdos_notification_daemon_notify_for_source (self,
            CDOS_NOTIFICATION_DAEMON_SOURCE (source), ndata);

}

static void
cdos_notification_daemon_dispose (GObject *gobject)
{
    CdosNotificationDaemon *self = CDOS_NOTIFICATION_DAEMON(gobject);

    if(self->sources)
    {
        GList *sources = self->sources;
        while (sources != NULL)
        {
            if(sources->data)
            {
                g_object_unref(sources->data);
                sources->data = NULL;
            }
            sources = sources->next;
        }
        g_list_free (self->sources);
        self->sources = NULL;
    }
    if (self->notifications)
    {
        GHashTable *notifications = self->notifications;
        GHashTableIter iter;
        gpointer key, value;
        g_hash_table_iter_init (&iter, notifications);
        while(g_hash_table_iter_next (&iter, &key, &value))
        {
                ndataStruct *ndata = value;
                g_hash_table_unref(ndata->hints);
                g_strfreev (ndata->actions);
        }
        g_hash_table_destroy (self->notifications);
        self->notifications = NULL;
    }

    g_clear_object (&self->bus_proxy);
    if (self->notification_node_info)
    {
        g_dbus_node_info_unref (self->notification_node_info);
        self->notification_node_info = NULL;
    }

    G_OBJECT_CLASS (cdos_notification_daemon_parent_class)->dispose (gobject);
}

static void
cdos_notification_daemon_finalize (GObject *gobject)
{
    //CdosNotificationDaemon *self = CDOS_NOTIFICATION_DAEMON(gobject);

    G_OBJECT_CLASS(cdos_notification_daemon_parent_class)->finalize (gobject);
}

static void
cdos_notification_daemon_class_init (CdosNotificationDaemonClass *klass)
{
    GObjectClass *gobject_class = G_OBJECT_CLASS(klass);
    gobject_class->finalize = cdos_notification_daemon_finalize;
    gobject_class->dispose = cdos_notification_daemon_dispose;
}

static void
cdos_notification_daemon_init (CdosNotificationDaemon *self)
{
    GDBusConnection *conn;

    GError *conn_error = NULL;
    conn = g_bus_get_sync (G_BUS_TYPE_SESSION, NULL, &conn_error);
    if (conn == NULL)
    {
        /*handle connection error*/
        g_warning ("Failed to open connection to bus in notification Daemon:%s", conn_error->message);
        g_error_free(conn_error);
        return;
    }

    self->notification_node_info = g_dbus_node_info_new_for_xml (introspection_xml, &conn_error);
    if (NULL == self->notification_node_info)
    {
        g_error ("Failed to call g_dbus_node_info_new_for_xml: %s!", conn_error->message);
        g_error_free(conn_error);
        g_object_unref (G_OBJECT (conn));
        return;
    }

    if (0 == g_bus_own_name(G_BUS_TYPE_SESSION,
                            "org.freedesktop.Notifications",
                            G_BUS_NAME_OWNER_FLAGS_REPLACE | G_BUS_NAME_OWNER_FLAGS_ALLOW_REPLACEMENT,
                            (GBusAcquiredCallback )on_notification_dbus_acquried,
                            NULL,
                            NULL,
                            self,
                            NULL)) {
        g_warning ("Failed to call g_bus_own_name_on_connection!\n");
        g_dbus_node_info_unref (self->notification_node_info);
        g_object_unref (G_OBJECT (conn));
        return;
    }

    self->sources = NULL;
    self->sender_to_pid = g_hash_table_new (g_str_hash, g_str_equal); //FIXME
    self->notifications = g_hash_table_new (NULL, NULL);

    self->settings = cdos_global_get_settings (cdos_global_get());
}

static ClutterActor *
cdos_notification_daemon_icon_for_notification_data (const gchar *icon, GHashTable *hints, int size)
{
    /*
     * If an icon is not specified, we use 'image-data' or 'image-path' hint for an icon
     * and don't show a large image. There are currently many applications that use
     * notify_notification_set_icon_from_pixbuf() from libnotify, which in turn sets
     * the 'image-data' hint. These applications don't typically pass in 'app_icon'
     * argument to Notify() and actually expect the pixbuf to be shown as an icon.
     * So the logic here does the right thing for this case. If both an icon and either
     * one of 'image-data' or 'image-path' are specified, we show both an icon and
     * a large image.
     */
    gchar *uri;
    const gchar *path_begin_f = "file://";

    ClutterActor *actor;

    /**
     * 优先级：
     * 1. image-data
     * 2. image-path
     * 3. app_icon
     * 4. icon_data
     * 如果同时指定了几个上面的属性，只显示一个
     */
    GVariant *image_data = NULL;
    image_data = g_hash_table_lookup (hints, "image-data");
    if (!image_data)
        image_data = g_hash_table_lookup (hints, "image_data");
    if (image_data)
    {
        ClutterImage *image = cdos_image_from_data (image_data);
        g_variant_unref (image_data);
        if (image)
        {
            actor = clutter_actor_new ();
            clutter_actor_set_content (actor, (ClutterContent *)image);
            clutter_actor_set_size (actor, size, size);
            return actor;
        }
    }

    StTextureCache *texture_cache;
    texture_cache = st_texture_cache_get_default ();

    const gchar *image_path;
    StThemeContext *context;
    GFile *file;
    float paint_scale, resource_scale;
    ClutterActor *ui_group = cdos_ui_plugin_get_ui_group ();
    clutter_actor_get_resource_scale (ui_group, &resource_scale);
    ClutterActor *stage = clutter_actor_get_stage (ui_group);
    context = st_theme_context_get_for_stage (CLUTTER_STAGE (stage));
    g_object_get (context, "scale-factor", &paint_scale, NULL);

    image_path = get_hints_image_path (hints);

    if (image_path)
    {
        uri = g_filename_to_uri (image_path, NULL, NULL);
        file = g_file_new_for_uri (uri);
        actor = st_texture_cache_load_file_async (texture_cache,
                                                  file,
                                                  size, size,
                                                  paint_scale, resource_scale);
        g_object_unref (file);
        return actor;
    }
    else if (icon && (*icon != '\0'))
    {
        if (strncmp (icon, path_begin_f, 7) == 0)
        {
            file = g_file_new_for_uri (icon);
            actor = st_texture_cache_load_file_async (texture_cache,
                                                      file,
                                                      size, size,
                                                      paint_scale, resource_scale);
            g_object_unref (file);
        }
        else if (icon[0] == '/')
        {
            file = g_file_new_for_path (icon);
            actor = st_texture_cache_load_file_async (texture_cache,
                                                      file,
                                                      size, size,
                                                      paint_scale, resource_scale);
            g_object_unref (file);
        }
        else
        {
            actor = st_icon_new ();
            st_icon_set_icon_name (ST_ICON (actor), icon);
            st_icon_set_icon_size (ST_ICON (actor), (gint) size);
        }
        return actor;
    }
    else
    {
        /* 处理 icon_data */
        GVariant *icon_data = NULL;
        icon_data = g_hash_table_lookup (hints, "icon_data");
        if (icon_data)
        {
            ClutterImage *image = cdos_image_from_data (icon_data);
            g_variant_unref (icon_data);
            if (image)
            {
                actor = clutter_actor_new ();
                clutter_actor_set_content (actor, (ClutterContent *)image);
                clutter_actor_set_size (actor, size, size);
                return actor;
            }
        }
    }
#if 1
    {
        const gchar *stock_icon;
        gint hints_urgency = get_hints_urgency (hints);

        switch (hints_urgency)
        {
            case CDOS_NOTIFICATION_LOW :
            case CDOS_NOTIFICATION_NORMAL:
                stock_icon = "gtk-dialog-info";
                break;
            case CDOS_NOTIFICATION_CRITICAL :
                stock_icon = "gtk-dialog-error";
                break;
            default:
                stock_icon = "gtk-dialog-info";
                break;
        }

        actor = st_icon_new ();

        st_icon_set_icon_name (ST_ICON (actor), stock_icon);
        st_icon_set_icon_size (ST_ICON (actor), (gint) size);
        return actor;
    }
#endif
    return NULL;
}

static CdosNotificationDaemonSource *
cdos_notification_daemon_lookup_source (CdosNotificationDaemon *self,
                                        const gchar *title,
                                        pid_t pid)
{
    // g_debug ("title is: %s, pid: %d\n", title, pid);
    // 将所有的 notify-send 的通知统一作为同一个 source 来处理
    if (!pid )
        return NULL;

    GList *sources = self->sources;
    while(sources != NULL)
    {
        CdosNotificationDaemonSource *source = sources->data;
        guint source_pid = cdos_notification_daemon_source_get_pid (source);
        if (source_pid == pid)// && (g_strcmp0 (source->initialTitle, title) == 0))
            return source;

        if (g_strcmp0 (title, "notify-send") == 0 &&
            g_strcmp0 (cdos_source_get_name ((CdosSource *)source), "notify-send") == 0)
        {
            return source;
        }
        sources = sources->next;
    }
    return NULL;
}

/*
 *Returns the source associated with ndata.notification if it is set.
 *Otherwise, returns the source associated with the title and pid if
 *such source is stored in self._sources and the notification is not
 *transient. If the existing or requested source is associated with
 *a tray icon and passed in pid matches a pid of an existing source,
 *the title match is ignored to enable representing a tray icon and
 *notifications from the same application with a single source.
 *
 *If no existing source is found, a new source is created as long as
 *pid is provided.
 *
 *Either a pid or ndata.notification is needed to retrieve or
 *create a source.
 */

static CdosSource *
cdos_notification_daemon_get_source (CdosNotificationDaemon *self,
                                     const gchar *title,
                                     pid_t pid,
                                     ndataStruct *ndata,
                                     const gchar *sender)
{
    if (!pid && !(ndata && ndata->notification))
    {
        return NULL;
    }

    CdosNotification *notification = ndata->notification;

    /*
     * We use notification's source for the notifications we still have
     * around that are getting replaced because we don't keep sources
     * for transient notifications in self._sources, but we still want
     * the notification associated with them to get replaced correctly.
     */
    if (ndata && notification)
    {
        CdosSource *source = notification->source;
        return source;
    }

    gboolean is_transient = (ndata && get_hints_transient (ndata->hints) == TRUE);
    /*
     * We don't want to override a persistent notification
     * with a transient one from the same sender, so we
     * always create a new source object for new transient notifications
     * and never add it to self._sources .
     */

    CdosNotificationDaemonSource *source = NULL;
    source = cdos_notification_daemon_lookup_source (self, title, pid);
    if (source)
    {
        /*调用CdosNotificationDaemonSource的父类CdosSource中的方法*/
        cdos_source_set_name (CDOS_SOURCE (source), title);
        cdos_source_set_app_name (CDOS_SOURCE (source), title);

        return (CdosSource *)source;
    }

    source = cdos_notification_daemon_source_new (title, pid, sender);
    /*调用CdosNotificationDaemonSource的父类CdosSource中的方法*/
    cdos_source_set_transient (CDOS_SOURCE(source), is_transient);

    self->sources = g_list_append (self->sources, source);
    g_signal_connect(source, "destroy", G_CALLBACK (splice_source_cb), self);

    CdosMessageTray *message_tray = cdos_message_tray_get_default ();
    cdos_message_tray_add (message_tray, CDOS_SOURCE(source));

    return CDOS_SOURCE (source);
}

/**
 * cdos_notification_daemon_notify:
 * @self: A #CdosNotificationDaemon
 * @app_name: 应用名
 * @replace_id: 待替换元素的位置
 * @icon: 图标
 * @summary: 消息内容
 * @body: 消息主体
 * @actions: notification自定义的按钮动作
 * @hints: notification DBus相关数据
 * @timeout: 过期时间
 * @out_id: 返回结果
 * @sender: 发送者(C语言中不好获取,所以在此当参数传过来使用).
 *
 * notification的守护进程,负责转发通知.
 *
 * Returns: %TRUE 表示工作正常.
 */
static gboolean
cdos_notification_daemon_notify (CdosNotificationDaemon *self,
                                 gchar                  *app_name,
                                 guint                   replace_id,
                                 gchar                  *icon,
                                 gchar                  *summary,
                                 gchar                  *body,
                                 gchar                 **actions,
                                 GHashTable             *hints,
                                 guint                   timeout,
                                 guint                  *out_id,
                                 const gchar            *sender)
{
    guint id;

    ndataStruct *ndata = NULL;
    ndata = g_hash_table_lookup (self->notifications, GUINT_TO_POINTER(replace_id));


    if (replace_id != 0 && ndata)
    {
        id = replace_id;
    }
    else
    {
        ndata = g_new0 (ndataStruct, 1);
        replace_id = 0;
        id = next_notification_id;
        next_notification_id++;
        ndata->id = id;
        g_hash_table_insert (self->notifications, GUINT_TO_POINTER(id), ndata);
    }
    ndata->app_name = app_name;
    ndata->icon = icon;
    ndata->summary = summary;
    ndata->body = body;
    if (ndata->actions)
        g_strfreev (ndata->actions);
    ndata->actions = actions;
    if (ndata->hints)
        g_hash_table_destroy (ndata->hints);
    ndata->hints = hints;
    ndata->timeout = timeout;

    guint pid;
    pid = GPOINTER_TO_UINT (g_hash_table_lookup (self->sender_to_pid, sender));
    CdosSource *source = cdos_notification_daemon_get_source (self, app_name, pid, ndata, sender);
    if (source)
    {
        g_debug ("%s replaces_id: %d, source: %p, pid: %d", __func__, id, source, pid);

        cdos_notification_daemon_notify_for_source (self, (CdosNotificationDaemonSource *)source, ndata);
        *out_id = id;
        return TRUE;
    }
    if (replace_id)
    {
        /*
         * There's already a pending call to GetConnectionUnixProcessID,
         * which will see the new notification data when it finishes,
         * so we don't have to do anything.
         */
        *out_id = id;
        return TRUE;
    }
    /*
     * replace_id = 0:
     */
    callMethodParams *call_methodparams = g_new0(callMethodParams, 1);
    call_methodparams->self = self;
    call_methodparams->id = id;
    call_methodparams->sender = g_strdup (sender);
    g_dbus_proxy_call (self->bus_proxy,
                       "GetConnectionUnixProcessID",
                       g_variant_new ("(s)", sender),
                       G_DBUS_CALL_FLAGS_NONE,
                       G_MAXINT,
                       NULL,
                       (GAsyncReadyCallback) proxy_sender_cb,
                       call_methodparams);
    *out_id = id;

    return TRUE;
}

static void
notification_action_invoked (CdosNotification *notificaton,
                             guint replaces_id,
                             const gchar *action,
                             CdosNotificationDaemon *self)
{
    cdos_notification_daemon_emit_action_invoked (self, replaces_id, action);
}

static void
cdos_notification_daemon_notify_for_source (CdosNotificationDaemon *self,
                                            CdosNotificationDaemonSource *source,
                                            ndataStruct *ndata)
{
    ClutterActor *iconActor = cdos_notification_daemon_icon_for_notification_data (ndata->icon,
                                                                                   ndata->hints,
                                                                                   24);
    CdosNotification *notification = ndata->notification;
    if (notification == NULL)
    {
        CdosNotificationParams new_params;
        memset (&new_params, 0, sizeof(CdosNotificationParams));
        new_params.icon = iconActor;
        new_params.replaces_id = ndata->id;
        new_params.category = get_hints_category (ndata->hints);
        new_params.sound_file = get_hints_sound_file (ndata->hints);
        new_params.sound_name = get_hints_sound_name (ndata->hints);

        notification = cdos_notification_new (CDOS_SOURCE(source), ndata->summary,
                                              ndata->body, &new_params);

        ndata->notification = notification;

        notificationParams *destroyParams = g_new0(notificationParams, 1);
        destroyParams->ndata_id = ndata->id;
        destroyParams->self = self;
        g_signal_connect (notification, "destroy", G_CALLBACK (notification_destroy_cb), destroyParams);
    }
    else
    {
        CdosNotificationParams update_params;
        memset (&update_params, 0, sizeof(CdosNotificationParams));
        update_params.icon = iconActor;
        update_params.clear = TRUE;
        update_params.category = get_hints_category (ndata->hints);
        update_params.sound_file = get_hints_sound_file (ndata->hints);
        update_params.sound_name = get_hints_sound_name (ndata->hints);

        cdos_notification_update (notification, ndata->summary, ndata->body, &update_params);
    }

    if (ndata->actions)
    {
        gboolean actions_icon = get_hints_action_icons (ndata->hints);
        if (actions_icon)
            cdos_notification_set_use_action_icons (notification, TRUE);

        char **actions = ndata->actions;
        cdos_notification_set_actions (notification, (const char **)actions);
        g_signal_connect (notification, "action-invoked", G_CALLBACK (notification_action_invoked), self);
    }
    gint hints_urgency = get_hints_urgency (ndata->hints);
    cdos_notification_set_urgency (notification, hints_urgency);

    gboolean hints_resident = get_hints_resident (ndata->hints);
    cdos_notification_set_resident (notification, (hints_resident == TRUE));

    gboolean hints_transient = get_hints_transient (ndata->hints);
    cdos_notification_set_transient (notification, (hints_transient == TRUE));

    cdos_notification_daemon_source_process_notification (source, notification);
}

/**
 * cdos_notification_daemon_close_notification:
 * @self: A #CdosNotificationDaemon
 * @reason: 关闭原因
 *
 * 关闭通知，并说明原因。
 *
 * Returns: 正常调用，返回 %TURE
 */
static gboolean
cdos_notification_daemon_close_notification (CdosNotificationDaemon *self, guint id)
{
    ndataStruct *ndata = g_hash_table_lookup (self->notifications, GUINT_TO_POINTER(id));
    if (ndata)
    {
        if (ndata->notification)
        {
            cdos_notification_destroy (ndata->notification, CDOS_NOTIFICATION_SOURCE_CLOSED);
        }
        g_strfreev (ndata->actions);
        /* TODO 释放 hints 中的 string 内存 */
        g_hash_table_unref(ndata->hints);
        g_hash_table_remove (self->notifications, GUINT_TO_POINTER(id));
        g_free (ndata);
    }
    return TRUE;
}

/**
 * cdos_notification_daemon_get_capabilities:
 * @self: A #CdosNotificationDaemon
 * @vBuilder: 能被dbus读取的数据结构
 *
 * 被dbus使用到，获取某些属性。
 */
static void
cdos_notification_daemon_get_capabilities (CdosNotificationDaemon* self, GVariantBuilder* vBuilder)
{
    g_variant_builder_add (vBuilder, "s", "actions");
    g_variant_builder_add (vBuilder, "s", "action-icons");
    g_variant_builder_add (vBuilder, "s", "body");
    g_variant_builder_add (vBuilder, "s", "body-markup");
    g_variant_builder_add (vBuilder, "s", "body-hyperlinks");
    g_variant_builder_add (vBuilder, "s", "body-images");
    g_variant_builder_add (vBuilder, "s", "icon-multi");
    g_variant_builder_add (vBuilder, "s", "persistence");
    g_variant_builder_add (vBuilder, "s", "sound");
}

/**
 * cdos_notification_daemon_get_server_information:
 * @self: A #CdosNotificationDaemon
 * @out_name: 桌面名称
 * @out_vendor: GNOME
 * @out_version: 1.0
 * @out_sepc_ver: 0.1
 *
 * 获取系统服务信息
 */
static void
cdos_notification_daemon_get_server_information (CdosNotificationDaemon *self,
                                                 const gchar **out_name,
                                                 const gchar **out_vendor,
                                                 const gchar **out_version,
                                                 const gchar **out_spec_ver)
{
    if (out_name)
        *out_name = "cdos";
    if (out_vendor)
        *out_vendor = "GNOME";
    if (out_version)
        *out_version = "1.0";
    if (out_spec_ver)
        *out_spec_ver = "0.1";
}

static void
cdos_notification_daemon_emit_notification_closed (CdosNotificationDaemon *self,
                                                   guint id,
                                                   CdosNotificationClosedReason reason)
{
    GError *error = NULL;
    GDBusConnection *connection = g_bus_get_sync (G_BUS_TYPE_SESSION, NULL, &error);
    if (connection == NULL)
    {
        g_warning ("Failed to open connection to bus in notification Daemon:%s", error->message);
        g_object_unref (connection);
        if (error)
            g_error_free(error);
        return;
    }

    g_dbus_connection_emit_signal (connection, NULL, "/org/freedesktop/Notifications",
                                   "org.freedesktop.Notifications", "NotificationClosed",
                                   g_variant_new ("(uu)", id, reason), NULL);
    g_clear_error (&error);
}

static void
cdos_notification_daemon_emit_action_invoked (CdosNotificationDaemon *self, guint id, const gchar *action)
{
    GError *error = NULL;

    GDBusConnection *connection = g_bus_get_sync (G_BUS_TYPE_SESSION, NULL, &error);
    if (connection == NULL)
    {
        g_warning ("Failed to open connection to bus in notification Daemon:%s", error->message);
        g_object_unref (connection);
        if (error)
            g_error_free(error);
        return;
    }

    if (!g_dbus_connection_emit_signal (connection,
                                        NULL,
                                        "/org/freedesktop/Notifications",
                                        "org.freedesktop.Notifications",
                                        "ActionInvoked",
                                        g_variant_new ("(us)", id, action),
                                        &error))
    {
        /*return TURE unless error, handle error*/
        g_warning ("g_dbus_connection_emit_signal wrong");
    }
    g_clear_error(&error);
}

static void
on_notification_dbus_acquried(GObject *obj, const gchar *name, gpointer user_data)
{

    GDBusConnection *connection = G_DBUS_CONNECTION (obj);
    GError *conn_error = NULL;
    //GDBusInterfaceSkeleton *interface = G_DBUS_INTERFACE_SKELETON (org_freedesktop_notifications_skeleton_new());
    CdosNotificationDaemon* self = CDOS_NOTIFICATION_DAEMON(user_data);

    if (!g_dbus_connection_register_object(connection,
                                           "/org/freedesktop/Notifications",
                                           self->notification_node_info->interfaces[0],
                                           &interface_vtable,
                                           self,
                                           NULL,
                                           &conn_error))
    {
        /*handle export error */
        g_warning ("g_dbus_connection_register_object error: %s", conn_error->message);
    }

    //gboolean res = g_dbus_interface_skeleton_export (interface, connection, "org/freedesktop/Notifications", &conn_error);
    generate_new_proxy (self);
    g_clear_error(&conn_error);

}

static void
handle_method_call(GDBusConnection *connection G_GNUC_UNUSED,
                   const gchar *sender,
                   const gchar *object_path G_GNUC_UNUSED,
                   const gchar *interface_name,
                   const gchar *method_name,
                   GVariant *parameters,
                   GDBusMethodInvocation *invocation,
                   CdosNotificationDaemon *self)
{
    if (0 == g_strcmp0 (method_name, "GetServerInformation"))
    {
       const gchar* out_name, *out_vendor, *out_version, *out_spec_ver;
       cdos_notification_daemon_get_server_information (self,
                                                        &out_name,
                                                        &out_vendor,
                                                        &out_version,
                                                        &out_spec_ver);
       g_dbus_method_invocation_return_value(invocation, g_variant_new("(ssss)", out_name, out_vendor, out_version, out_spec_ver));
    }
    else if (0 == g_strcmp0 (method_name, "GetCapabilities"))
    {
        GVariantBuilder * caps_builder = g_variant_builder_new (G_VARIANT_TYPE ("as"));
        cdos_notification_daemon_get_capabilities (self, caps_builder);
        g_dbus_method_invocation_return_value (invocation, g_variant_new ("(as)", caps_builder));
        g_variant_builder_unref (caps_builder);
    }
    else if (0 == g_strcmp0 (method_name, "CloseNotification"))
    {
        guint id = 0;
        g_variant_get(parameters, "(u)", &id);
        g_debug ("CloseNotification: %d", id);
        cdos_notification_daemon_close_notification (self, id);
        g_dbus_method_invocation_return_value (invocation, NULL);
    }
    else if (0 == g_strcmp0 (method_name, "Notify"))
    {
        gchar *app_name;
        guint replace_id;
        gchar *icon = NULL;
        gchar *summary;
        gchar *body;
        GVariantIter *actions_iter; // 为"as"类型的GVariant
        GVariantIter *hints; // 为"a{sv}"类型的GVariant
        guint timeout;
        guint out_id = 0;
        g_variant_get (parameters, "(&susssasa{sv}i)", &app_name, &replace_id, &icon, &summary, &body, &actions_iter, &hints, &timeout);
        g_debug ("%s: app_name(%s), replace_id(%u), icon(%s), summary(%s), body(%s), timeout(%u)!",
                __func__, app_name, replace_id, icon, summary, body, timeout);

        GHashTable *hints_hashtbl = g_hash_table_new_full (g_str_hash, g_str_equal, NULL, NULL);
        gchar **actions = NULL;
        gchar *key = NULL;
        GVariant *value;
        gsize n_actions = g_variant_iter_n_children (actions_iter);
        if (n_actions > 0)
        {
            actions = g_new (gchar *, n_actions + 1);
            // g_variant_get_strv
            int i = 0;
            while (g_variant_iter_loop (actions_iter, "s", &key))
            {
                actions[i] = g_strdup (key);
                ++i;
                g_debug ("action: %s", key);
            }
            actions[i] = NULL;
        }

        //while (g_variant_iter_loop(hints, "{sv}", &key, &value)) {
        while (g_variant_iter_next (hints, "{sv}", &key, &value))
        {
            const gchar *value_type = g_variant_get_type_string (value);
            if (value_type[0] == '(')
            {

                if ((g_strcmp0 (key, "image-data") == 0) ||
                    (g_strcmp0 (key, "image_data") == 0) ||
                    (g_strcmp0 (key, "icon-data") == 0) ||
                    (g_strcmp0 (key, "icon_data") == 0))
                {
                    g_debug ("%s(%s) %p", key, value_type, value);
                    g_hash_table_insert (hints_hashtbl, key, value);
                }
            }
            else
            {
                gulong val = 0;
                switch (*value_type)
                {
#define VAL_TYPE_CASE(format,type, t) case format:{\
    type v;\
    v = g_variant_get_##t(value);\
    val = (gulong)v;\
    g_debug ("key(%s), value type(%c), value: 0x%lx", key, *value_type, val); \
    g_hash_table_insert (hints_hashtbl, key, (gpointer)val); \
    break;}
                    VAL_TYPE_CASE('y',guchar, byte)
                    VAL_TYPE_CASE('b',gboolean, boolean)
                    VAL_TYPE_CASE('n',gint16, int16)
                    VAL_TYPE_CASE('q',guint16, uint16)
                    VAL_TYPE_CASE('i',gint32, int32)
                    VAL_TYPE_CASE('u',guint32, uint32)
                    VAL_TYPE_CASE('x',gint64, int64)
                    VAL_TYPE_CASE('t',guint64, uint64)
                    VAL_TYPE_CASE('d',gdouble, double)
#undef VAL_TYPE_CASE
#define VAL_TYPE_CASE(format) case format:{\
    gchar *v;\
    g_variant_get(value, value_type, &v);\
    g_debug ("key(%s), value type(%c), value: %s", key, *value_type, v); \
    g_hash_table_insert (hints_hashtbl, key, v); \
    break;}
                    VAL_TYPE_CASE('s')
                    VAL_TYPE_CASE('o')
                    VAL_TYPE_CASE('g')
#undef VAL_TYPE_CASE
                    default:
                        g_warning ("hint: %s(type: %s) NOT handled", key, value_type);
                        break;
                }
                g_variant_unref (value);
            }
        }

        if (cdos_notification_daemon_notify (self,
                                             app_name, replace_id, icon,
                                             summary, body, actions,
                                             hints_hashtbl, timeout,
                                             &out_id, sender))
        {
            g_debug ("%s replace_id=%d", __func__, out_id);
            g_dbus_method_invocation_return_value(invocation, g_variant_new("(u)", out_id));
        }
    }
}
