/**
 * Copyright (c) 2016-2018 CPU and Fundamental Software Research Center, CAS
 *
 * This software is published by the license of CPU-OS License，you can use and
 * distibuted this software under this License. See CPU-OS Liense for more detail.
 *
 * You should have receive a copy of CPU-OS License. If not, please contact us
 * by email <support_os@cpu-os.ac.cn>
 *
 **/
/**
 * SECTION: cdos-notification
 * @short_description: 通知结构体
 * @include: cdos-notification.h
 *
 * CdosNotificationClass 类用来显示通知
 */

#include <math.h>
#include <string.h>
#include <meta/util.h>

#include "cdos-util.h"
#include "cdos-notification.h"
#include "cdos-global-private.h"
#include "cdos-base-util.h"
#include "cdos-source.h"
#include "cdos-start.h"

#define NOTIFY_HEIGHT 11
static const gchar *DATE_FORMAT = "%F %H:%M"; // _("%l:%M %p")

struct _CdosNotificationPrivate
{
    ClutterActor *close_button;
    ClutterActor *expand_button;

    gfloat spacing;
    gfloat avail_width;

    gboolean need_expand;
    gboolean expanded;
    gboolean use_action_icons;
    gboolean has_default_action;
    gboolean compact;

    gulong hover_changed_id;
};

G_DEFINE_TYPE_WITH_PRIVATE (CdosNotification, cdos_notification, ST_TYPE_WIDGET);

enum
{
    EXPANDED,
    COLLAPSED,
    CLICKED,
    ACTION_INVOKED,
    LAST_SIGNAL
};

static guint cdos_notification_signals [LAST_SIGNAL] = { 0 };

static gboolean
close_notification (ClutterActor *actor,
                    ClutterEvent *event,
                    CdosNotification *self)
{
    cdos_notification_destroy (self, CDOS_NOTIFICATION_UNDEFINED_RESERVED);
    return TRUE;
}

static gboolean
expand_notification (ClutterActor *button,
                     ClutterEvent *event,
                     CdosNotification *self)
{
    CdosNotificationPrivate *priv = self->priv;

    StLabel *actor = self->body_label;
    gfloat label_height;
    ClutterActor *clutter_text = st_label_get_clutter_text (actor);

    clutter_actor_save_easing_state ((ClutterActor *)self->body_label);
    clutter_actor_set_easing_duration ((ClutterActor *)self->body_label, 200);
    clutter_actor_set_easing_mode ((ClutterActor *)self->body_label, CLUTTER_EASE_IN_OUT_QUAD);

    if (priv->expanded)
    {
        clutter_text_set_ellipsize ((ClutterText *)st_label_get_clutter_text (actor),
                                    PANGO_ELLIPSIZE_END);
        clutter_actor_get_preferred_height (clutter_text, -1, NULL, &label_height);
        clutter_actor_set_size ((ClutterActor *)self->body_label, priv->avail_width, label_height);
        st_widget_set_style_class_name ((StWidget *)priv->expand_button, "expand-notification");
        clutter_actor_restore_easing_state ((ClutterActor *)self->body_label);
        priv->expanded = FALSE;
        return TRUE;
    }

    clutter_text_set_line_wrap ((ClutterText *)clutter_text, TRUE);
    clutter_text_set_line_wrap_mode ((ClutterText *)clutter_text, PANGO_WRAP_WORD_CHAR);
    clutter_text_set_ellipsize ((ClutterText *)clutter_text, PANGO_ELLIPSIZE_NONE);
    /*
    gfloat natural_text_width ;
    clutter_actor_get_preferred_height (clutter_text, -1, NULL, &label_height);
    clutter_actor_get_preferred_width (clutter_text, -1, NULL, &natural_text_width);
    guint row = (guint) natural_text_width / priv->avail_width + 1;
    clutter_actor_set_size ((ClutterActor *)self->body_label, priv->avail_width, row * label_height);
    */
    /* 多个url折行显示时，用上面的方法显示不全 */
    clutter_actor_get_preferred_height (clutter_text, priv->avail_width, NULL, &label_height);
    clutter_actor_set_size ((ClutterActor *)self->body_label, priv->avail_width, label_height + NOTIFY_HEIGHT);
    clutter_actor_restore_easing_state ((ClutterActor *)self->body_label);

    st_widget_set_style_class_name ((StWidget *)priv->expand_button, "collapse-notification");
    priv->expanded = TRUE;
    return TRUE;
}

#define compute_child_size(SIZE) gfloat image_min_##SIZE = 0, image_nat_##SIZE = 0; \
    gfloat title_min_##SIZE, title_nat_##SIZE; \
    gfloat body_min_##SIZE = 0, body_nat_##SIZE = 0; \
    gfloat time_min_##SIZE = 0, time_nat_##SIZE = 0; \
    gfloat close_min_##SIZE, close_nat_##SIZE; \
    gfloat expand_min_##SIZE = 0, expand_nat_##SIZE = 0; \
    gfloat action_min_##SIZE = 0, action_nat_##SIZE = 0; \
    if (self->image) \
      clutter_actor_get_preferred_##SIZE (self->image, -1, &image_min_##SIZE, &image_nat_##SIZE); \
    clutter_actor_get_preferred_##SIZE ((ClutterActor *)self->title_label, -1, &title_min_##SIZE, &title_nat_##SIZE); \
    if (self->body_label) \
      clutter_actor_get_preferred_##SIZE ((ClutterActor *)self->body_label, -1, &body_min_##SIZE, &body_nat_##SIZE); \
    if (clutter_actor_is_visible ((ClutterActor *)self->time_label)) \
      clutter_actor_get_preferred_##SIZE ((ClutterActor *)self->time_label, -1, &time_min_##SIZE, &time_nat_##SIZE); \
    clutter_actor_get_preferred_##SIZE (priv->close_button, -1, &close_min_##SIZE, &close_nat_##SIZE); \
    clutter_actor_get_preferred_##SIZE (priv->expand_button, -1, &expand_min_##SIZE, &expand_nat_##SIZE); \
    if (self->action_area) \
      clutter_actor_get_preferred_##SIZE (self->action_area, -1, &action_min_##SIZE, &action_nat_##SIZE);

static void
cdos_notification_get_preferred_width (ClutterActor *actor,
                                       gfloat        for_height,
                                       gfloat       *min_width_p,
                                       gfloat       *natural_width_p)
{
    StThemeNode *theme_node = st_widget_get_theme_node ((StWidget *)actor);
    st_theme_node_adjust_for_height (theme_node, &for_height);

    CdosNotification *self = CDOS_NOTIFICATION (actor);
    CdosNotificationPrivate *priv = self->priv;

    compute_child_size (width);

    gfloat top_min_width, center_min_width, right_min_width;
    gfloat top_nat_width, center_nat_width, right_nat_width;

    if (priv->compact)
        center_min_width = MAX (title_min_width + time_min_width, body_min_width);
    else
        center_min_width = MAX (MAX (title_min_width, body_min_width), time_min_width);
    right_min_width = MAX (close_min_width, expand_min_width);
    top_min_width = image_min_width + center_min_width + right_min_width;

    if (priv->compact)
        center_nat_width = MAX (title_nat_width + time_nat_width, body_nat_width);
    else
        center_nat_width = MAX (MAX (title_nat_width, body_nat_width), time_nat_width);
    right_nat_width = MAX (close_nat_width, expand_nat_width);
    top_nat_width = image_nat_width + center_nat_width + right_nat_width + priv->spacing;

    *min_width_p = MAX (top_min_width, action_min_width);
    *natural_width_p = MAX (top_nat_width, action_nat_width);

    st_theme_node_adjust_preferred_width (theme_node, min_width_p, natural_width_p);
}

static void
cdos_notification_get_preferred_height (ClutterActor *actor,
                                        gfloat        for_width,
                                        gfloat       *min_height_p,
                                        gfloat       *natural_height_p)
{
    StThemeNode *theme_node = st_widget_get_theme_node ((StWidget *)actor);
    st_theme_node_adjust_for_width (theme_node, &for_width);

    CdosNotification *self = CDOS_NOTIFICATION (actor);
    CdosNotificationPrivate *priv = self->priv;

    compute_child_size (height);

    gfloat top_min_height, center_min_height, right_min_height;
    gfloat top_nat_height, center_nat_height, right_nat_height;

    if (priv->compact)
        center_min_height = title_min_height + body_min_height;
    else
        center_min_height = title_min_height + body_min_height + time_min_height;
    right_min_height = close_min_height + expand_min_height;
    top_min_height = MAX (MAX (image_min_height, center_min_height), right_min_height);

    if (priv->compact)
        center_nat_height = title_nat_height + body_nat_height;
    else
        center_nat_height = title_nat_height + body_nat_height + time_nat_height;
    right_nat_height = close_nat_height + expand_nat_height;
    top_nat_height = MAX (MAX (image_nat_height, center_nat_height), right_nat_height);

    *min_height_p = top_min_height + action_min_height;
    *natural_height_p = top_nat_height + action_nat_height;
    if (action_nat_height > 0)
    {
        *natural_height_p += priv->spacing * 0.5;
    }
    if (self->in_notification_center)
        *min_height_p = *natural_height_p;

    st_theme_node_adjust_preferred_height (theme_node, min_height_p, natural_height_p);
}

static void
cdos_notification_allocate (ClutterActor *actor,
                            const ClutterActorBox *box,
                            ClutterAllocationFlags flags)
{
    CdosNotification *self = CDOS_NOTIFICATION (actor);
    CdosNotificationPrivate *priv = self->priv;
    StThemeNode *theme_node = st_widget_get_theme_node ((StWidget *)actor);

    clutter_actor_set_allocation (actor, box, flags);

    compute_child_size (width);
    compute_child_size (height);

    ClutterActorBox content_box, child_box;
    st_theme_node_get_content_box (theme_node, box, &content_box);

    gint x, y;

    /* left, center, right area */
    gint left_y = 0, center_y = 0, right_y = 0;
    gfloat avail_width;

    x = content_box.x1;
    y = content_box.y1;

    if (self->image)
    {
        child_box.x1 = x;
        child_box.x2 = child_box.x1 + image_nat_width;
        child_box.y1 = y;
        child_box.y2 = child_box.y1 + image_nat_height;
        clutter_actor_allocate (self->image, &child_box, flags);
        x = child_box.x2 + priv->spacing;
        left_y = child_box.y2 + priv->spacing * 0.5;
    }

    child_box.x1 = content_box.x2 - close_nat_width;
    child_box.x2 = content_box.x2;
    child_box.y1 = content_box.y1;
    child_box.y2 = child_box.y1 + close_nat_height;
    clutter_actor_allocate (priv->close_button, &child_box, flags);
    right_y = child_box.y2;

    avail_width = content_box.x2 - x - close_nat_width;
    if (avail_width > priv->spacing)
        avail_width -= priv->spacing;

    priv->avail_width = avail_width;

    {
        child_box.x1 = x;
        gfloat tnw = priv->compact ? time_nat_width : 0;
        gfloat w = MIN (avail_width - tnw, title_nat_width);
        child_box.x2 = child_box.x1 +w;
        child_box.y1 = content_box.y1;
        child_box.y2 = child_box.y1 + title_nat_height;
        clutter_actor_allocate ((ClutterActor *)self->title_label, &child_box, flags);
        y = child_box.y2;
    }
    if (priv->compact && clutter_actor_is_visible (CLUTTER_ACTOR (self->time_label)))
    {
        child_box.x1 = child_box.x2;
        child_box.x2 = child_box.x1 + time_nat_width;
        child_box.y1 = (int)(0.5 + content_box.y1 + (title_nat_height - time_nat_height) / 2);
        child_box.y2 = child_box.y1 + time_nat_height;
        clutter_actor_allocate ((ClutterActor *)self->time_label, &child_box, flags);
    }

    if (self->body_label)
    {
        child_box.x1 = x;
        gfloat w = body_nat_width;
        if (w > avail_width)
        {
            priv->need_expand = TRUE;
            w = avail_width;
        }
        child_box.x2 = child_box.x1 + w;
        child_box.y1 = y;
        child_box.y2 = child_box.y1 + body_nat_height;
        clutter_actor_allocate ((ClutterActor *)self->body_label, &child_box, flags);
        center_y = child_box.y2;
    }
    if (priv->need_expand)
    {
        child_box.x1 = (int)(content_box.x2 - expand_nat_width + 0.5);
        child_box.x2 = content_box.x2;
        child_box.y1 = right_y;
        child_box.y2 = child_box.y1 + expand_nat_height;
        clutter_actor_allocate (priv->expand_button, &child_box, flags);
    }
    if (!priv->compact)
    {
        child_box.x1 = x;
        child_box.x2 = child_box.x1 + time_nat_width;
        child_box.y1 = center_y;
        child_box.y2 = child_box.y1 + time_nat_height;
        clutter_actor_allocate ((ClutterActor *)self->time_label, &child_box, flags);
        center_y = child_box.y2 + priv->spacing;
    }
    if (self->action_area)
    {
        if (action_nat_width > avail_width)
            action_nat_width = avail_width;
        child_box.x1 = (int)(content_box.x1 + (content_box.x2 - content_box.x1 - action_nat_width) / 2 + 0.5);
        child_box.x2 = content_box.x2;
        y = MAX (left_y, center_y);
        child_box.y1 = y;
        child_box.y2 = child_box.y1 + action_nat_height;
        clutter_actor_allocate ((ClutterActor *)self->action_area, &child_box, flags);
    }
}

static gboolean
cdos_notification_on_button_release (ClutterActor *actor, ClutterButtonEvent *event)
{
    CdosNotification *self = CDOS_NOTIFICATION (actor);

    g_signal_emit (self, cdos_notification_signals[CLICKED], 0);

    if (self->priv->has_default_action)
        g_signal_emit (self, cdos_notification_signals[ACTION_INVOKED], 0, self->replaces_id, "default");

    // We hide all types of notifications once the user clicks on them because the common
    // outcome of clicking should be the relevant window being brought forward and the user's
    // attention switching to the window.
    if (!self->resident)
        cdos_notification_destroy(self, CDOS_NOTIFICATION_UNDEFINED_RESERVED);

    return TRUE;
}

static void
cdos_notification_style_changed (StWidget *widget)
{
    CdosNotification *self = CDOS_NOTIFICATION (widget);
    CdosNotificationPrivate *priv = self->priv;

    StThemeNode *theme_node = st_widget_get_theme_node (widget);
    priv->spacing = st_theme_node_get_length (theme_node, "spacing");

    ST_WIDGET_CLASS (cdos_notification_parent_class)->style_changed (widget);
}

static void
cdos_notification_on_hover (StWidget *widget,
                            GParamSpec *pspec,
                            CdosNotification *self)
{
    CdosNotificationPrivate *priv = self->priv;

    gboolean hover = st_widget_get_hover (widget);
    if (hover)
    {
        clutter_actor_show (priv->close_button);
        if (priv->need_expand)
            clutter_actor_show (priv->expand_button);
    }
    else
    {
        clutter_actor_hide (self->priv->close_button);
        if (priv->need_expand)
            clutter_actor_hide (priv->expand_button);
    }
}

static void
cdos_notification_dispose (GObject *obj)
{
    CdosNotification * self = CDOS_NOTIFICATION (obj);
    if (self->priv->hover_changed_id)
    {
        g_signal_handler_disconnect (obj, self->priv->hover_changed_id);
        self->priv->hover_changed_id = 0;
    }

    G_OBJECT_CLASS (cdos_notification_parent_class)->dispose (obj);
}

static void
cdos_notification_finalize (GObject *obj)
{
    CdosNotification *self = CDOS_NOTIFICATION (obj);

    g_clear_pointer (&self->timestamp, g_date_time_unref);
    g_free (self->category);
    g_free (self->sound_file);
    g_free (self->sound_name);
    g_free (self->title);
    g_free (self->body);

    G_OBJECT_CLASS (cdos_notification_parent_class)->finalize (obj);
}

static void
cdos_notification_class_init (CdosNotificationClass *klass)
{
    GObjectClass *gobject_class = G_OBJECT_CLASS(klass);

    gobject_class->dispose = cdos_notification_dispose;
    gobject_class->finalize = cdos_notification_finalize;

    ClutterActorClass *actor_class = CLUTTER_ACTOR_CLASS (klass);
    actor_class->allocate = cdos_notification_allocate;
    actor_class->get_preferred_width = cdos_notification_get_preferred_width;
    actor_class->get_preferred_height = cdos_notification_get_preferred_height;
    actor_class->button_release_event = cdos_notification_on_button_release;

    StWidgetClass *widget_class = ST_WIDGET_CLASS (klass);
    widget_class->style_changed = cdos_notification_style_changed;

    /**
     * CdosNotification::expanded:
     * @object: 绑定信号的对象
     * @user_data: 真对回调函数所传入的参数
     *
     * notification能根据内容来自适应大小,当默认大小不满足时,就会触发该信号.
     */
    cdos_notification_signals [EXPANDED] =
        g_signal_new ("expanded",
                      G_TYPE_FROM_CLASS (klass),
                      G_SIGNAL_RUN_LAST,
                      0,
                      NULL, NULL,
                      g_cclosure_marshal_VOID__VOID,
                      G_TYPE_NONE, 0);

    /**
     * CdosNotification::collapsed:
     * @object: 绑定信号的对象
     * @user_data: 真对回调函数所传入的参数
     *
     * 信号发送条件:
     * 若notification的destroyed成员变量为 %TRUE, 则返回;
     * 否则,不调整大小,不对文本进行自动换行,然后发送该信号.
     */
    cdos_notification_signals [COLLAPSED] =
            g_signal_new ("collapsed",
                    G_TYPE_FROM_CLASS (klass),
                    G_SIGNAL_RUN_LAST,
                    0,
                    NULL, NULL,
                    g_cclosure_marshal_VOID__VOID,
                    G_TYPE_NONE, 0);

    /**
     * CdosNotification::clicked:
     * @object: 绑定信号的对象
     * @user_data: 真对回调函数所传入的参数
     *
     * 当触发notification的"button-release-event"信号时,会在回调函数时发送.
     */
    cdos_notification_signals [CLICKED] =
        g_signal_new ("clicked",
                      G_TYPE_FROM_CLASS (klass),
                      G_SIGNAL_RUN_LAST,
                      0,
                      NULL, NULL,
                      g_cclosure_marshal_VOID__VOID,
                      G_TYPE_NONE, 0);
    /**
     * CdosNotification::action-invoked:
     * @object: 绑定信号的对象
     * @user_data: 真对回调函数所传入的参数
     *
     * 当触发notification的"clicked"信号时,会在回调函数时发送.
     */
    cdos_notification_signals [ACTION_INVOKED] =
            g_signal_new ("action-invoked",
                    G_TYPE_FROM_CLASS (klass),
                    G_SIGNAL_RUN_LAST,
                    0,
                    NULL, NULL,
                    g_cclosure_marshal_VOID__UINT_POINTER,
                    G_TYPE_NONE, 2,
                    G_TYPE_UINT, G_TYPE_STRING);
}

static void
cdos_notification_init (CdosNotification *self)
{
    CdosNotificationPrivate *priv;
    self->priv = priv = cdos_notification_get_instance_private (self);
    ClutterActor *actor = CLUTTER_ACTOR (self);

    self->urgency = CDOS_NOTIFICATION_NORMAL;
    self->resident = FALSE;
    self->is_transient = FALSE;
    priv->expanded = FALSE;
    priv->compact = TRUE;

    self->timestamp = g_date_time_new_now_local ();
    self->in_notification_center = FALSE;
    self->unshown = TRUE;

    clutter_actor_set_reactive (actor, TRUE);
    clutter_actor_set_x_expand (actor, TRUE);
    st_widget_set_track_hover ((StWidget *)self, TRUE);

    st_widget_set_style_class_name ((StWidget *)self, "notification");
    clutter_actor_set_name (actor, "notification");

    self->title_label = (StLabel *)st_label_new (NULL);
    st_widget_add_style_class_name ((StWidget *)self->title_label, "notification-title");
    clutter_actor_add_child (actor, (ClutterActor *)self->title_label);

    self->time_label = (StLabel *)st_label_new (NULL);
    st_widget_add_style_class_name ((StWidget *)self->time_label, "notification-time");
    clutter_actor_hide ((ClutterActor *)self->time_label);
    clutter_actor_add_child (actor, (ClutterActor *)self->time_label);

    self->body_label = cdos_url_highlighter_new (" ", TRUE);
    st_widget_add_style_class_name ((StWidget *)self->body_label, "notification-body");
    clutter_actor_add_child (actor, (ClutterActor *)self->body_label);

    priv->close_button = g_object_new (ST_TYPE_WIDGET, "style-class", "close-notification", NULL);
    clutter_actor_set_reactive (priv->close_button, TRUE);
    clutter_actor_hide (priv->close_button);
    g_signal_connect (priv->close_button, "button-release-event", G_CALLBACK (close_notification), self);
    clutter_actor_add_child (actor, priv->close_button);

    priv->expand_button = g_object_new (ST_TYPE_WIDGET, "style-class", "expand-notification", NULL);
    clutter_actor_set_reactive (priv->expand_button, TRUE);
    clutter_actor_hide (priv->expand_button);
    g_signal_connect (priv->expand_button, "button-release-event", G_CALLBACK (expand_notification), self);
    clutter_actor_add_child (actor, priv->expand_button);

    priv->hover_changed_id = g_signal_connect (self, "notify::hover", G_CALLBACK (cdos_notification_on_hover), self);
}

/**
 * cdos_notification_new:
 * @source: A #CdosSource
 * @title: 设置 #CdosNotification 实例的标题
 * @banner: 设置 #CdosNotification 实例的内容
 * @params: 必要参数,参考 #CdosNotificationParams
 *
 * 新建一个 #CdosNotification 类的实例。
 *
 * Returns: 返回一个 #CdosNotification 的实例。
 */
CdosNotification *
cdos_notification_new (CdosSource *source, const gchar *title, const gchar *body, CdosNotificationParams *params)
{
    CdosNotification *self = g_object_new (CDOS_TYPE_NOTIFICATION, NULL);
    self->source = source;
    self->replaces_id = params->replaces_id;
    cdos_notification_update (self, title, body, params);

    return self;
}

/**
 * cdos_notification_update:
 * @title: the new title
 * @banner: the new banner
 * @params: as in the CdosNotification constructor
 *
 * Updates the notification by regenerating its icon and updating
 * the title/banner. If @params.clear is %TRUE, it will also
 * remove any additional actors/action buttons previously added.
 */

void
cdos_notification_update (CdosNotification *self, const gchar *title, const gchar *body, CdosNotificationParams *params)
{
    if (NULL != self->timestamp)
        g_date_time_unref(self->timestamp);

    self->category = g_strdup (params->category);
    self->sound_file = g_strdup (params->sound_file);
    self->sound_name = g_strdup (params->sound_name);

    self->timestamp = g_date_time_new_now_local ();
    if (self->image && (params->icon || params->clear))
    {
        clutter_actor_destroy (self->image);
        self->image = NULL;
    }

    if (!self->image && params->icon)
    {
        self->image = params->icon;
        //if (!self->image)
        //    self->image = cdos_source_create_icon (self->source);
        clutter_actor_insert_child_at_index ((ClutterActor *)self, self->image, 0);
    }
    if (g_strcmp0 (self->title, title) != 0)
    {
        g_free (self->title);
        self->title = g_strdup (title);
        for (char *p = self->title; *p != '\0'; p++)
            if (*p == '\n')
                *p = ' ';

        clutter_text_set_markup ((ClutterText *)st_label_get_clutter_text (self->title_label), self->title);
        clutter_text_set_ellipsize ((ClutterText *)st_label_get_clutter_text (self->title_label),
                PANGO_ELLIPSIZE_END);
        clutter_actor_queue_relayout (CLUTTER_ACTOR (self->title_label));
    }

    gchar *time = g_date_time_format (self->timestamp, DATE_FORMAT);
    clutter_text_set_markup (CLUTTER_TEXT (st_label_get_clutter_text(self->time_label)), time);
    g_free (time);
    clutter_actor_hide (CLUTTER_ACTOR (self->time_label));
    /*
    if (pango_find_base_dir (new_title, -1) == PANGO_DIRECTION_RTL)
        priv->titleDirection = CT_TEXT_DIRECTION_RTL;
    else
        priv->titleDirection = CT_TEXT_DIRECTION_LTR;

    //st_widget_set_direction (self->table, self->titleDirection);
    */
    if (body && *body != '\0')
    {
        if (!self->body_label)
        {
            self->body_label = cdos_url_highlighter_new (NULL, TRUE);
            st_widget_add_style_class_name ((StWidget *)self->body_label, "notification-body");
            //clutter_actor_set_x_expand ((ClutterActor *)self->body_label, TRUE);
            clutter_actor_set_x_align ((ClutterActor *)self->body_label, CLUTTER_ACTOR_ALIGN_START);

            clutter_actor_insert_child_above ((ClutterActor *)self, (ClutterActor *)self->body_label, (ClutterActor *)self->title_label);
        }
        gchar *new_body = NULL;
        GRegex *regex = g_regex_new ("\n", 0, 0, NULL);
        new_body = g_regex_replace (regex, body, -1, 0, "  ", 0, NULL);
        g_regex_unref (regex);
        self->body = g_strdup (new_body);
        cdos_url_highlighter_set_markup ((CdosUrlHighlighter *)self->body_label, new_body);
        g_free (new_body);
        clutter_text_set_ellipsize ((ClutterText *)st_label_get_clutter_text (self->body_label),
                PANGO_ELLIPSIZE_END);
        clutter_actor_queue_relayout (CLUTTER_ACTOR (self->body_label));
    }
}

/**
 * cdos_notification_set_icon_visible:
 * @self: A #CdosNotification
 * @visible: 是否可见
 *
 * 设置notification的icon是否可见
 *
 */
void cdos_notification_set_icon_visible (CdosNotification *self, gboolean visible)
{
    if (visible)
        clutter_actor_show (self->image);
    else
        clutter_actor_hide (self->image);
}

static void
cdos_notification_invoke_action (ClutterActor *actor, guint button, CdosNotification *self)
{
    const gchar *action = clutter_actor_get_name (actor);
    g_signal_emit (self, cdos_notification_signals[ACTION_INVOKED], 0, self->replaces_id, action);

    cdos_notification_destroy(self, CDOS_NOTIFICATION_UNDEFINED_RESERVED);
}

static void
cdos_notification_add_button (CdosNotification *self,
                              const gchar *id,
                              const gchar *label)
{
    StWidget *button;
    button = st_button_new ();
    st_widget_set_can_focus (button, TRUE);

    if (gtk_icon_theme_has_icon (gtk_icon_theme_get_default (), id) && self->priv->use_action_icons)
    {
        StIcon *icon = ST_ICON (st_icon_new ());
        st_icon_set_icon_size (icon, 16);
        st_icon_set_icon_name (icon, id);
        st_widget_add_style_class_name (button, "notification-icon-button");
        st_bin_set_child (ST_BIN(button), CLUTTER_ACTOR(icon));
    }
    else
    {
        st_widget_add_style_class_name (button, "notification-button");
        st_button_set_label ((StButton *)button, label);
    }

    clutter_actor_set_name ((ClutterActor *)button, id);
    g_signal_connect (button, "clicked", G_CALLBACK (cdos_notification_invoke_action), self);
    clutter_actor_add_child (self->action_area, (ClutterActor *)button);
}

void
cdos_notification_set_actions (CdosNotification *self, const gchar **actions)
{
    g_return_if_fail (CDOS_IS_NOTIFICATION (self));
    if (!self->action_area)
    {
        self->action_area = (ClutterActor *)st_box_layout_new ();
        clutter_actor_add_child ((ClutterActor *)self, self->action_area);
        st_widget_add_style_class_name ((StWidget *)self->action_area, "notification-actions");
    }
    else
        clutter_actor_destroy_all_children (self->action_area);

    int i = 0;
    while (actions[i] != NULL)
    {
        const gchar *actions_i = actions[i++];
        const gchar *actions_i_add_1 = actions[i++];
        if (g_strcmp0 (actions_i, "default") == 0)
        {
            self->priv->has_default_action = TRUE;
        }
        else if (actions_i_add_1)
        {
            cdos_notification_add_button (self, actions_i, _(actions_i_add_1));
        }
        else
            break;
    }
}
/**
 * cdos_notification_set_urgency:
 * @self: A #CdosNotification
 * @urgency: 通知的紧急程度
 *
 * 设置通知的紧急程度
 *
 */
void
cdos_notification_set_urgency (CdosNotification *self, guint urgency)
{
    self->urgency = urgency;
}

/**
 * cdos_notification_set_resident:
 * @self: A #CdosNotification
 * @resident: 是否是常驻通知
 *
 * 设置通知是否是常驻
 *
 */
void
cdos_notification_set_resident (CdosNotification *self, gboolean resident)
{
    self->resident = resident;
}

/**
 * cdos_notification_set_transient:
 * @self: A #CdosNotification
 * @isTransient: 是否是短暂的
 *
 * 设置通知是否是短暂的
 *
 */
void
cdos_notification_set_transient (CdosNotification *self, gboolean transient)
{
    self->is_transient = transient;
}

gboolean
cdos_notification_get_transient (CdosNotification *self)
{
    return self->is_transient;
}

/**
 * cdos_notification_expand:
 * @self: A #CdosNotification
 * @animate: 默认大小能完全显示时,是否扩展通知
 *
 * 设置通知是否是自动调节大小
 *
 */
void
cdos_notification_expand (CdosNotification *self , gboolean animate)
{
    if (!self->priv->expanded)
    {
        expand_notification (NULL, NULL, self);
        g_signal_emit (self, cdos_notification_signals[EXPANDED], 0);
    }
    return;
}

/**
 * cdos_notification_collapse:
 * @self: A #CdosNotification
 *
 * 收缩通知:如果可删除,什么都不做;否则,不自动换行.
 *
 */
void
cdos_notification_collapse (CdosNotification *self)
{
    if (self->priv->expanded)
    {
        expand_notification (NULL, NULL, self);
        g_signal_emit (self, cdos_notification_signals[COLLAPSED], 0);
    }
}

/**
 * cdos_notification_destroy:
 * @self: A #CdosNotification
 * @reason: 删除的原因
 *
 * 删除通知,并给出原因.
 *
 */
void
cdos_notification_destroy (CdosNotification *self, CdosNotificationClosedReason reason)
{
    g_return_if_fail (CDOS_IS_NOTIFICATION(self));
    self->closed_reason = reason;
    ClutterActor *actor = CLUTTER_ACTOR (self);
    ClutterActor *parent = clutter_actor_get_parent (actor);
    if (!parent)
    {
        clutter_actor_hide (actor);
        clutter_actor_add_child (cdos_ui_plugin_get_ui_group (), actor);
    }
    clutter_actor_destroy (actor);
}

void
cdos_notification_set_use_action_icons (CdosNotification *self, gboolean value)
{
    self->priv->use_action_icons = value;
}

gchar *
cdos_notification_timeify (CdosNotification *self, const char *format)
{
    g_return_val_if_fail (CDOS_IS_NOTIFICATION (self), NULL);

    const gchar *default_format = "%p %I:%M";
    const gchar *default_format1 = _("%b%e %H:%M");

    GDateTime *orig_time = self->timestamp;
    GDateTime *now_timestamp = g_date_time_new_now_local ();
    GTimeSpan diff = g_date_time_difference (now_timestamp, orig_time) / 1000000;
    gchar *str = NULL, *str_diff = NULL;
    gchar *orig;

    if (diff <= 15)
    {
        str_diff = g_strdup (_("Just now"));
    }
    else if (diff > 15 && diff <= 59)
    {
        /*
        gchar s_diff[3];
        s_diff[0] = '0' + diff / 10;
        s_diff[1] = '0' + diff % 10;
        s_diff[2] = '\0';
        */
        str_diff = g_strdup_printf (_("%ld seconds ago"), diff);
    }
    else if (diff > 59 && diff <= 119)
    {
        str_diff = g_strdup (_("1 minute ago"));
    }
    else if (diff > 119 && diff <= 599)
    {
        /*
        gchar s_diff[2];
        s_diff[0] = '0' + diff / 60;
        s_diff[1] = '\0';
        str_diff = g_strdup_printf (_("%s minutes ago"), s_diff);
        */
        gint df = diff / 60;
        str_diff = g_strdup_printf (_("%d minutes ago"), df);
    }
    else if (diff > 599 && diff <= 3540)
    {
        gint d = diff / 60;
        /*
        gchar s_diff[3];
        s_diff[0] = '0' + d / 10;
        s_diff[1] = '0' + d % 10;
        s_diff[2] = '\0';
        str_diff = g_strdup_printf (_("%s minutes ago"), s_diff);
        */
        str_diff = g_strdup_printf (_("%d minutes ago"), d);
    }

    if (diff <= 3540)
    {
        str = g_strconcat (" \u2022 ", str_diff, NULL);
        g_free (str_diff);
    }
    else
    {
        if (!format)
        {
            if (g_date_time_get_day_of_week (orig_time) !=
                    g_date_time_get_day_of_week (now_timestamp))
                format = default_format1;
            else
                format = default_format;
        }
        orig = g_date_time_format (orig_time, format);
        str = g_strconcat (" \u2022 ", orig, NULL);
        g_free (orig);
    }

    g_date_time_unref (now_timestamp);

    return str;
}
