#include "glib-2.62.h"

#include "cdos-overview-workspace-thumbnail.h"
#include "cdos-overview-window-clone.h"
#include "cdos-overview-thumbnails-box.h"

#include "cdos-layout-manager.h"
#include "cdos-window-manager.h"
#include "cdos-base-util.h"

#include <math.h>

/**
 * SECTION:cdos-overview-workspace-thumbnail
 * @Title: CdosOverviewWorkspaceThumbnail
 * @Short_description: 单个工作区的缩略图
 *
 * 桌面默认有四个工作区，每个工作区缩略图除了工作区本身，还有个 #GtkEntry 显示这个缩略图的名称
 * 在当前工作区中打开、关闭窗口等事件，都需要这个对象来监听信号并更新窗口列表内容
 * 由于获取的窗口列表是所有已经打开的窗口，还需要过滤这些窗口是不是普通要显示的窗口，是不是当前工作区的窗口等。
 */
#define INACTIVE_OPACITY 128
#define ACTIVE_OPACITY 48

#define ICON_OFFSET -5

#define REARRANGE_TIME_ON 100
#define REARRANGE_TIME_OFF 300
#define ICON_SIZE 128

gfloat thumbnail_box_scale = 0.125;

//外部定义: main中尚未实现的函数

static gboolean isMyWindow (CdosOverviewWorkspaceThumbnail *self, MetaWindowActor *win);
static gboolean isOverviewWindow (CdosOverviewWorkspaceThumbnail *self, MetaWindow *win);

G_DEFINE_TYPE (CdosOverviewWorkspaceThumbnail, cdos_overview_workspace_thumbnail, CLUTTER_TYPE_ACTOR);

static void do_add_window (CdosOverviewWorkspaceThumbnail *self, MetaWindow *metaWin);
static void doRemoveWindow (CdosOverviewWorkspaceThumbnail *self, MetaWindow *metaWin);
static void restack (CdosOverviewWorkspaceThumbnail *self, gboolean force);

static void
onWorkspaceChanged (GObject *gobject, ClutterEvent *event, gpointer user_data)
{
    CdosOverviewWindowClone *clone = CDOS_OVERVIEW_WINDOW_CLONE (gobject);
    CdosOverviewWorkspaceThumbnail *self = CDOS_OVERVIEW_WORKSPACE_THUMBNAIL (user_data);
    MetaWindow *metaWindow = clone->metaWindow;

    doRemoveWindow (self, metaWindow);
}

/**
 * add_window_clone:
 * @self: #CdosOverviewWorkspaceThumbnail 对象
 * @win: 普通窗口对象
 *
 * 添加窗口克隆链表：在 #CdosOverviewWindowClone 对象的 cdos_overview_window_clone_refresh_clone
 * 函数中，大概介绍了这个过程。
 * 传递进来的普通窗口，根据需要，可以为它的临时窗口也一起创建克隆体，保存在一个链表中，一一遍历保存到
 * #CdosOverviewWindowClone 的 clone 成员中，
 * 它的这个 clone 不直接对外使用，而是继续添加到 actor 中，也就是我们这个函数中的 clone.
 * 每个窗口对应一个 #CdosOverviewWindowClone 对象，工作区中所有窗口的克隆体就保存在 self->windows 链表中
 * 需要注意的是：每新加一个窗口，就将它放置在最顶层。
 */
static void
add_window_clone (CdosOverviewWorkspaceThumbnail *self, MetaWindowActor *win)
{
    CdosOverviewWindowClone *clone = cdos_overview_window_clone_new (win);

    g_signal_connect (clone, "workspace-changed", G_CALLBACK (onWorkspaceChanged), self);

    clutter_actor_add_child (self->contents, (ClutterActor *)clone);
    /*
    length = g_list_length (self->windows);

        if (length == 0)
        cdos_overview_window_clone_set_stack_above (clone, self->background);
        else
    {
        CdosOverviewWindowClone *tmpClone = g_list_nth_data (self->windows, length - 1);
        cdos_overview_window_clone_set_stack_above (clone, (ClutterActor *)tmpClone);
    }
    */
    clutter_actor_set_child_above_sibling (self->contents, (ClutterActor *)clone, NULL);
    self->windows = g_list_append (self->windows, clone);
}

static void
onScrollEvent (GObject *gobject, ClutterEvent *event, gpointer user_data)
{
    // CdosOverviewWorkspaceThumbnail *self = CDOS_OVERVIEW_WORKSPACE_THUMBNAIL (user_data);
    ClutterScrollDirection direction = clutter_event_get_scroll_direction (event);
    switch (direction)
    {
        case CLUTTER_SCROLL_UP:
            cdos_wm_action_move_workspace_left ();
            break;
        case CLUTTER_SCROLL_DOWN:
            cdos_wm_action_move_workspace_right ();
            break;
        default:
            break;
    }
}

static GList *
lookupIndex (CdosOverviewWorkspaceThumbnail *self, MetaWindow *metaWindow)
{
    GList *idx = NULL;
    for (idx = self->windows; idx; idx = idx->next)
    {
        CdosOverviewWindowClone *clone = idx->data;

        if (clone->metaWindow == metaWindow)
            return idx;
    }
    return NULL;
}

static void
add_window (CdosOverviewWorkspaceThumbnail *self, MetaWindow *window)
{
    MetaWindowActor *win = (MetaWindowActor *)meta_window_get_compositor_private (window);
    if (!isMyWindow (self, win) || !isOverviewWindow (self, window))
        return;

    add_window_clone (self, win);
}

typedef struct
{
    CdosOverviewWorkspaceThumbnail *self;
    MetaWindow *window;
    guint tid;
} ParamForCallback;

static gint
find_id (CdosOverviewWorkspaceThumbnail *self, MetaWindow *window)
{
    gint i;
    for (i = 0; i < self->add_window_ids->len; i++)
        if (window == ((ParamForCallback *)self->add_window_ids->pdata[i])->window)
            return i;
    return -1;
}

static gboolean
onIdle_cb (ParamForCallback *id)
{
    CdosOverviewWorkspaceThumbnail *self = id->self;

    {
        MetaWindow *metaWin = id->window;

        /*will be null if we're closing down*/
        MetaWindowActor *actor = META_WINDOW_ACTOR (meta_window_get_compositor_private (metaWin));
        if (actor)
        {
            add_window (self, metaWin);
            g_ptr_array_remove (self->add_window_ids, id);
            return FALSE;
        }
    }
    return TRUE;
}

static void
do_add_window (CdosOverviewWorkspaceThumbnail *self, MetaWindow *window)
{
    if (lookupIndex (self, window) != NULL)
        return;
    if (meta_window_get_compositor_private (window))
    {
        add_window (self, window);
        return;
    }

    gint i = find_id (self, window);
    if (i >= 0)
        return;

    ParamForCallback *param;

    param = g_new0 (ParamForCallback, 1);
    if (!param)
        return;
    g_ptr_array_add (self->add_window_ids, param);
    param->self = self;
    param->window = window;

    // Newly-created windows are added to a workspace before
    // the compositor finds out about them...
    param->tid = g_idle_add ((GSourceFunc)onIdle_cb, param);
}

static gboolean
isMyWindow (CdosOverviewWorkspaceThumbnail *self, MetaWindowActor *win)
{
    MetaWindow *window = meta_window_actor_get_meta_window (win);
    return meta_window_located_on_workspace (window, self->metaWorkspace)
           && (meta_window_get_monitor (window) == self->monitor_index);
}

static gboolean
isOverviewWindow (CdosOverviewWorkspaceThumbnail *self, MetaWindow *win)
{
    MetaWindowType type;

    if (meta_window_is_override_redirect (win))
        return FALSE;

    type = meta_window_get_window_type (win);
    return type != META_WINDOW_DESKTOP && type != META_WINDOW_DOCK;
}

static void
doRemoveWindow (CdosOverviewWorkspaceThumbnail *self, MetaWindow *metaWin)
{
    MetaWindowActor *win = META_WINDOW_ACTOR (meta_window_get_compositor_private (metaWin));

    //找出actor在windows中的位置
    GList *index = lookupIndex (self, metaWin);

    if (index == NULL)
        return;

    // Check if window still should be here
    if (win && isMyWindow (self, win) && isOverviewWindow (self, metaWin))
        return;

    //找出index
    ClutterActor *clone = index->data;

    self->windows = g_list_remove (self->windows, clone);
}

static gint
minimizedDiff (CdosOverviewWindowClone *a, CdosOverviewWindowClone *b)
{
    gboolean minimize_a = FALSE;
    gboolean minimize_b = FALSE;

    g_object_get (a->metaWindow, "minimized", &minimize_a, NULL);
    g_object_get (b->metaWindow, "minimized", &minimize_b, NULL);

    return minimize_b - minimize_a;
}


static gint
transientRelation (CdosOverviewWindowClone *a, CdosOverviewWindowClone *b)
{
    gint overviewDifference = 1; // noOverviewDiff (a,b);

    if (overviewDifference)
    {
        gint transientA = (meta_window_get_transient_for (a->metaWindow) == b->metaWindow) ? 1 : 0;
        gint transientB = !transientA && (meta_window_get_transient_for (b->metaWindow) == a->metaWindow) ? 1 : 0;
        if (transientA - transientB)
            return transientA - transientB;
        else
            return overviewDifference;
    }
    return 0;
}

static gint
window_sort_func (CdosOverviewWindowClone *a, CdosOverviewWindowClone *b, GList *windows)
{
    gint ret;
    if ((ret = transientRelation (a, b)))
        return ret;
    else if ((ret = minimizedDiff (a, b)))
        return ret;
    else
    {
        guint32 index_a = -1;
        guint32 index_b = -1;

        GList *iter;
        gint i = 0;
        for (iter = windows; iter; iter = iter->next)
        {
            if (a->metaWindow == iter->data)
                index_a = i;
            if (b->metaWindow == iter->data)
                index_b = i;
            i++;
            if (index_a >= 0 && index_b >= 0)
                break;
        }
        gint diff_a_b = index_a - index_b;
        return diff_a_b;
    }
}

static void
syncStacking (CdosOverviewWorkspaceThumbnail *self)
{
    //将window做一个排序
    GList *windows = cdos_global_get_window_actors (global);
    self->windows = g_list_sort_with_data (self->windows, (GCompareDataFunc)window_sort_func, windows);
    self->windows = g_list_reverse (self->windows);
    GList *iter;
    for (iter = self->windows; iter; iter = iter->next)
    {
        CdosOverviewWindowClone *clone = iter->data;

        /*
        MetaWindow *metaWindow = clone->metaWindow;

        g_object_get (metaWindow, "minimized", &minimize, NULL);
        if (minimize){
            continue;
        }
        */
        if (iter == self->windows)
        {
            cdos_overview_window_clone_set_stack_above (clone, self->background->actor);
        }
        else
        {
            CdosOverviewWindowClone *previousClone = iter->prev->data;
            cdos_overview_window_clone_set_stack_above (clone, (ClutterActor *)previousClone);
        }
    }
}

static void
restack (CdosOverviewWorkspaceThumbnail *self, gboolean force)
{
    syncStacking (self);
}

static void
disconnectOtherSignals (CdosOverviewWorkspaceThumbnail *self)
{
    if (self->overview_hiding_id)
    {
        CdosOverview *overview = cdos_overview_get ();
        g_signal_handler_disconnect (overview, self->overview_hiding_id);
        self->overview_hiding_id = 0L;
    }
    MetaDisplay *display = cdos_global_get_display (cdos_global_get ());
    g_clear_signal_handler (&self->restackedNotifyId, display);
    g_clear_signal_handler (&self->windowAddedId, self->metaWorkspace);
    g_clear_signal_handler (&self->windowRemovedId, self->metaWorkspace);
    g_clear_signal_handler (&self->windowEnteredMonitorId, display);
    g_clear_signal_handler (&self->windowLeftMonitorId, display);
}

static void
windowAdded (GObject *gobject, MetaWindow *metaWin, CdosOverviewWorkspaceThumbnail *self)
{
    do_add_window (self, metaWin);
    restack (self, TRUE);
}

static void
windowRemoved (GObject *gobject, MetaWindow *metaWin, CdosOverviewWorkspaceThumbnail *self)
{
    doRemoveWindow (self, metaWin);
}

static void
windowEnteredMonitor (GObject *gobject, gint monitorIndex, MetaWindow *metaWin, CdosOverviewWorkspaceThumbnail *self)
{
    // important if workspaces-only-on-primary is in effect
    // do_add_window (self, metaWin);
}

static void
windowLeftMonitor (GObject *gobject, gint monitorIndex, MetaWindow *metaWin, CdosOverviewWorkspaceThumbnail *self)
{
    // important if workspaces-only-on-primary is in effect
    doRemoveWindow (self, metaWin);
}

static void
on_restack (GObject *gobject, CdosOverviewWorkspaceThumbnail *self)
{
    restack (self, TRUE);
}

static void
on_overview_hidding (CdosOverview *overview, CdosOverviewWorkspaceThumbnail *self)
{
    disconnectOtherSignals (self);
    if (self->add_window_ids)
    {
        guint i;
        for (i = 0; i < self->add_window_ids->len; i++)
        {
            ParamForCallback *id = self->add_window_ids->pdata[i];
            g_source_remove (id->tid);
        }
        g_ptr_array_free (self->add_window_ids, 1);
        self->add_window_ids = NULL;
    }
}
static void
cdos_overview_workspace_thumbnail_dispose (GObject *object)
{
    CdosOverviewWorkspaceThumbnail *self = CDOS_OVERVIEW_WORKSPACE_THUMBNAIL (object);
    if (self->background)
    {
        cdos_background_destroy (self->background);
        self->background = NULL;
    }
    G_OBJECT_CLASS (cdos_overview_workspace_thumbnail_parent_class)->dispose (object);
}

static void
cdos_overview_workspace_thumbnail_finalize (GObject *object)
{
    CdosOverviewWorkspaceThumbnail *self = CDOS_OVERVIEW_WORKSPACE_THUMBNAIL (object);
    on_overview_hidding (NULL, self);

    if (self->windows)
    {
        g_list_free (self->windows);
        self->windows = NULL;
    }

    G_OBJECT_CLASS (cdos_overview_workspace_thumbnail_parent_class)->finalize (object);
}

static gboolean
thumbnail_on_button_press (ClutterActor *actor, ClutterButtonEvent *e)
{
    return TRUE;
}

static gboolean
thumbnail_on_button_release (ClutterActor *actor, ClutterButtonEvent *e)
{
    CdosOverviewWorkspaceThumbnail *self = CDOS_OVERVIEW_WORKSPACE_THUMBNAIL (actor);

    MetaDisplay *display = cdos_global_get_display (global);
    MetaWorkspaceManager *wsm = meta_display_get_workspace_manager (display);
    MetaWorkspace *wks = meta_workspace_manager_get_active_workspace (wsm);
    if (self->metaWorkspace == wks)
    {
        CdosOverview *overview = cdos_overview_get ();
        clutter_actor_hide ((ClutterActor *)overview);
    }
    else
        meta_workspace_activate (self->metaWorkspace, cdos_global_get_current_time (global));
    return TRUE;
}

static void
cdos_overview_workspace_thumbnail_class_init (CdosOverviewWorkspaceThumbnailClass *klass)
{
    GObjectClass *gobject_class = G_OBJECT_CLASS (klass);

    gobject_class->finalize = cdos_overview_workspace_thumbnail_finalize;
    gobject_class->dispose = cdos_overview_workspace_thumbnail_dispose;

    ClutterActorClass *actor_class = CLUTTER_ACTOR_CLASS (klass);
    actor_class->button_press_event = thumbnail_on_button_press;
    actor_class->button_release_event = thumbnail_on_button_release;
}

static void
cdos_overview_workspace_thumbnail_init (CdosOverviewWorkspaceThumbnail *self)
{
    self->add_window_ids = g_ptr_array_new_with_free_func ((GDestroyNotify)g_free);
    self->monitor_index = cdos_layout_manager_get_index ();
}

/**
 * cdos_overview_workspace_thumbnail_show:
 * @self: #CdosOverviewWorkspaceThumbnail 对象
 *
 * 显示这个工作区中所有窗口的缩略图
 *
 */
static void
cdos_overview_workspace_thumbnail_show (CdosOverviewWorkspaceThumbnail *self, gboolean force, gboolean override)
{
    return;

    CdosLayoutManager *layout_manager = cdos_layout_manager_get_default ();
    GPtrArray *monitors = cdos_layout_manager_get_monitors (layout_manager);

    gint monitorIndex; //遍历monitors
    for (monitorIndex = 0; monitors && monitorIndex < monitors->len; monitorIndex++)
    {
        GList *iter;
        for (iter = self->windows; iter; iter = iter->next)
        {
            CdosOverviewWindowClone *window = iter->data;
            MetaWindow *metaWindow = window->metaWindow;

            if (monitorIndex == meta_window_get_monitor (metaWindow))
            {
                clutter_actor_show ((ClutterActor *)window);

                gboolean a = !meta_window_showing_on_its_workspace (metaWindow);
                clutter_actor_set_opacity ((ClutterActor *)window, 255 >> a);
            }
        }
    }
}

static void
on_workspace_close_button_clicked (StButton *button,
                                int click_button,
                                CdosOverviewWorkspaceThumbnail *self)
{
    MetaDisplay *display = cdos_global_get_display (cdos_global_get ());
    MetaWorkspaceManager *wsm = meta_display_get_workspace_manager (display);
    
    meta_workspace_manager_remove_workspace (wsm, self->metaWorkspace, cdos_global_get_current_time (global));    // 触发信号，去跑 onWorkspaceRemoved
}

static void
cdos_overview_workspace_thumbnail_post_init (CdosOverviewWorkspaceThumbnail *self, MetaWorkspace *metaWorkspace)
{
    ClutterActor *close_icon;
    MetaDisplay *display = cdos_global_get_display (global);
    MetaWorkspaceManager *wsm = meta_display_get_workspace_manager (display);

    self->metaWorkspace = metaWorkspace;

    clutter_actor_set_reactive ((ClutterActor *)self, TRUE);
    clutter_actor_set_clip_to_allocation ((ClutterActor *)self, TRUE);
    // st_widget_set_style_class_name ((StWidget *)self, "workspace-thumbnail");

    gint w, h;
    cdos_global_get_screen_size (global, &w, &h);
    clutter_actor_set_size ((ClutterActor *)self, (gfloat)w, (gfloat)h);

    self->contents = st_box_layout_new ();
    clutter_actor_add_child ((ClutterActor *)self, self->contents);

    self->lastButtonPressTimeStamp = 0;
    self->lastButtonPressActor = NULL;

    g_signal_connect (self, "scroll-event", G_CALLBACK (onScrollEvent), self);

    self->background = cdos_background_new (self->contents, 0, 0, 0, TRUE, NULL);

    self->shader = st_bin_new ();
    st_widget_set_style (self->shader, "background-color: black;");

    clutter_actor_add_child ((ClutterActor *)self, (ClutterActor *)self->shader);

    clutter_actor_set_size ((ClutterActor *)self->shader, (gfloat)w, (gfloat)h);

    clutter_actor_set_opacity ((ClutterActor *)self->shader, INACTIVE_OPACITY);

    if (metaWorkspace == meta_workspace_manager_get_active_workspace (wsm))
        clutter_actor_set_opacity ((ClutterActor *)self->shader, ACTIVE_OPACITY);

    self->workspace_title_box = g_object_new (ST_TYPE_BOX_LAYOUT, 
                                                        NULL);

    gchar * workspace_name = g_strdup (_("Workspace"));
    self->workspace_title = g_object_new (ST_TYPE_LABEL,
                                                "text", workspace_name,
                                                "visible", TRUE,
                                                "x-expand", TRUE,
                                                "y-expand", TRUE, 
                                                "x-align", CLUTTER_ACTOR_ALIGN_CENTER, 
                                                "y-align", CLUTTER_ACTOR_ALIGN_CENTER, 
                                                NULL);
    st_widget_add_style_class_name (self->workspace_title, "window-caption");

    self->workspace_close_button = g_object_new (ST_TYPE_BUTTON,
                                                "style-class", "workspace-close-button",
                                                "visible", TRUE,
                                                "x-expand", TRUE,
                                                "y-expand", TRUE, 
                                                "x-align", CLUTTER_ACTOR_ALIGN_END, 
                                                "y-align", CLUTTER_ACTOR_ALIGN_END, 
                                                NULL);
    close_icon = g_object_new (ST_TYPE_ICON,
                               "icon-name", "close-window",
                               NULL);
    st_bin_set_child ((StBin *)self->workspace_close_button, close_icon);
    g_signal_connect (self->workspace_close_button, "clicked", G_CALLBACK (on_workspace_close_button_clicked), self);

    _add_actor (self->workspace_title_box, self->workspace_title);


    GList *windows = cdos_global_get_window_actors (cdos_global_get ()); // filter(self->isMyWindow, self);
    GList *window_filters = NULL;

    // 先过滤非本工作区的
    while (windows != NULL)
    {
        GList *next = windows->next;
        MetaWindowActor *win = windows->data;
        MetaWindow *window = meta_window_actor_get_meta_window (win);
        if (meta_window_located_on_workspace (window, metaWorkspace))
            window_filters = g_list_prepend (window_filters, win);
        windows = next;
    }
    // 再过滤非普通窗口的
    GList *yes_windows = NULL;
    while (window_filters != NULL)
    {
        GList *next = window_filters->next;
        MetaWindowActor *win = window_filters->data;
        if (isOverviewWindow (self, meta_window_actor_get_meta_window (win)))
        {
            yes_windows = g_list_prepend (yes_windows, win);
        }
        window_filters = next;
    }
    // yes_windows = g_list_reverse (yes_windows);
    // Create clones for windows that should be visible in the Overview
    self->count = 0;
    self->windows = NULL;

    // 在这里添加对最小化窗口的过滤试试，初始化的时候…… 貌似不太对的样子
    GList *iter;
    gboolean minimize = FALSE;
    for (iter = yes_windows; iter; iter = iter->next)
    {
        MetaWindowActor *index_window = iter->data;

        MetaWindow *metaWindow = meta_window_actor_get_meta_window (index_window);
        g_object_get (metaWindow, "minimized", &minimize, NULL);

        if (!minimize)
            add_window_clone (self, index_window);
    }

    g_list_free (window_filters);
    g_list_free (yes_windows);

    self->windowAddedId = g_signal_connect (self->metaWorkspace, "window-added", G_CALLBACK (windowAdded), self);
    self->windowRemovedId = g_signal_connect (self->metaWorkspace, "window-removed", G_CALLBACK (windowRemoved), self);

    self->windowEnteredMonitorId =
        g_signal_connect (display, "window-entered-monitor", G_CALLBACK (windowEnteredMonitor), self);
    self->windowLeftMonitorId =
        g_signal_connect (display, "window-left-monitor", G_CALLBACK (windowLeftMonitor), self);

    self->restackedNotifyId = g_signal_connect (display, "restacked", G_CALLBACK (on_restack), self);

    CdosOverview *overview = cdos_overview_get ();
    self->overview_hiding_id = g_signal_connect (overview, "hiding", G_CALLBACK (on_overview_hidding), self);
    self->isActive = FALSE;

    restack (self, TRUE);

    self->slidePosition = 0; // Fully slid in
    cdos_overview_workspace_thumbnail_show (self, 0, 0);
}

/**
 * cdos_overview_workspace_thumbnail_new:
 * @metaWorkspace: #MetaWorkspace 工作区对象
 * @box: #CdosOverviewThumbnailsBox 对象
 *
 * 新建 #CdosOverviewWorkspaceThumbnail 对象，代表一个工作区的缩略图
 */
CdosOverviewWorkspaceThumbnail *
cdos_overview_workspace_thumbnail_new (MetaWorkspace *metaWorkspace)
{
    CdosOverviewWorkspaceThumbnail *instance;
    instance = g_object_new (CDOS_TYPE_OVERVIEW_WORKSPACE_THUMBNAIL, NULL);
    cdos_overview_workspace_thumbnail_post_init (instance, metaWorkspace);

    return instance;
}

/**
 * cdos_overview_workspace_thumbnail_activate_workspace:
 * @self: #CdosOverviewWorkspaceThumbnail 对象
 *
 * 激活这个缩略图对应的工作区，进入这个工作区。
 */
void
cdos_overview_workspace_thumbnail_activate_workspace (CdosOverviewWorkspaceThumbnail *self)
{
    g_return_if_fail (CDOS_IS_OVERVIEW_WORKSPACE_THUMBNAIL (self));
    MetaDisplay *display = cdos_global_get_display (global);
    MetaWorkspaceManager *wsm = meta_display_get_workspace_manager (display);
    CdosOverview *overview = cdos_overview_get ();

    if (self->metaWorkspace != meta_workspace_manager_get_active_workspace (wsm))
    {
        guint timestamp = cdos_global_get_current_time (cdos_global_get ());

        meta_workspace_activate (self->metaWorkspace, timestamp);
    }
    clutter_actor_hide ((ClutterActor *)overview);
}

/**
 * cdos_overview_workspace_thumbnail_shade:
 * @self: #CdosOverviewWorkspaceThumbnail 对象
 * @force: 是否强制设置
 *
 * 上下左右键切换工作区时，离开的工作区阴影效果，不再高亮
 */
void
cdos_overview_workspace_thumbnail_shade (CdosOverviewWorkspaceThumbnail *self, gboolean force)
{
    g_return_if_fail (CDOS_IS_OVERVIEW_WORKSPACE_THUMBNAIL (self));
    if (!self->isSelected || force)
    {
        cdos_transition_simple ((ClutterActor *)self->shader,
                                CLUTTER_EASE_IN_QUAD,
                                200,
                                "opacity", INACTIVE_OPACITY,
                                NULL);
    }
}

/**
 * cdos_overview_workspace_thumbnail_highlight:
 * @self: #CdosOverviewWorkspaceThumbnail 对象
 *
 * 当进入某个工作区缩略图时，高亮显示
 */
void
cdos_overview_workspace_thumbnail_highlight (CdosOverviewWorkspaceThumbnail *self)
{
    g_return_if_fail (CDOS_IS_OVERVIEW_WORKSPACE_THUMBNAIL (self));
    cdos_transition_simple ((ClutterActor *)self->shader,
                            CLUTTER_EASE_IN_QUAD,
                            200,
                            "opacity", ACTIVE_OPACITY,
                            NULL);
}
