//
// Created by zhangtao on 16-3-17.
// source:overview.js
/**
 * SECTION: cdos-overview
 * @short_description:用于工作起切换时，生成各工作区的概观图示。
 * @title: CdosOverview
 * @include: cdos-overview.h
 *
 * #CdosOverview在激活工作区切换时，生成各工作区相应可视化缩略图，辅助标识当前各个工作区的桌面视图状态。
 *
 * CdosOverview结构关系:
 *
 * |________ cover_pane:ClutterActor
 *
 * |____ background.actor:ClutterActor
 *
 * |____ background_shade:StWidget
 *
 * |____ workspaces_view_actor:ClutterActor
 */
//添加依赖库;

#define CLUTTER_DISABLE_DEPRECATION_WARNINGS

#include <math.h>
#include <meta/boxes.h>
#include <meta/meta-background-group.h>

#include "cdos-overview.h"
#include "cdos-overview-thumbnails-box.h"
#include "cdos-workspaces-view.h"

#include "cdos-layout-manager.h"
#include "cdos-global.h"
#include "cdos-start.h"
#include "cdos-background.h"
#include "cdos-monitor-switcher.h"

#define ANIMATION_TIME 200
#define BACKGROUND_SCHEMA "org.cdos.desktop.background"

static const gchar cdos_overview_xml[] =
"<node>"
"  <interface name='org.freedesktop.Desktop'>"
"    <method name='OverviewToggle'>"
"    </method>"
"    <method name='RunMonitorSwitchDialog'>"
"    </method>"
"  </interface>"
"</node>";
#define CDOS_DBUS_PATH "/org/freedesktop/Desktop"

enum
{
    OVERVIEW_DESTROY,
    OVERVIEW_HIDING,
    OVERVIEW_HIDDEN,
    OVERVIEW_SHOWING,
    OVERVIEW_SHOWN,
    LAST_SIGNAL
};

static CdosOverview *overview = NULL;

struct _CdosOverviewPrivate
{
    CdosWorkspacesView *workspaces_view;
    ClutterActor *thumbnails_box;

    gdouble spacing;
    gboolean visible;
    gboolean shown;
    gboolean show_temporarily;
    gboolean modal;
    gboolean hide_in_progress;

    gfloat drag_start_x;
    gfloat drag_start_y;

    gfloat drag_x;
    gfloat drag_y;

    gdouble drag_start_value;
    gint last_motion_time;

    ClutterActor *cover_pane;
    ClutterActor *background_group;
    CdosBackgroundManager *background;

    guint button_press_id;
    guint button_release_id;
    guint32 window_switch_timestamp;
    guint window_switch_timeout_id;
    guint last_active_workspace_index;
    MetaWindow *last_hovered_window;

    GDBusNodeInfo   *introspection_data;
    GDBusConnection *connection;
    GCancellable    *bus_cancellable;

    gint result; // 为了传递动画complete 回调参数
};

G_DEFINE_TYPE_WITH_PRIVATE (CdosOverview, cdos_overview, ST_TYPE_WIDGET);

static void cdos_overview_dispose (GObject *object);
static void cdos_overview_finalize (GObject *object);

static gint cdos_overview_signals[LAST_SIGNAL] = {0};

static void cdos_overview_animate_visible (CdosOverview *self);

static gboolean cdos_overview_sync_input_mode (CdosOverview *self);

static void cdos_overview_animate_not_visible (CdosOverview *self);

static void cdos_overview_show_done (ClutterTimeline *tml, CdosOverview *self);

static void cdos_overview_hide_done (ClutterTimeline *tml, CdosOverview *self);

static gboolean cdos_overview_cover_pane_event_callback (ClutterActor *actor, ClutterEvent *event, CdosOverview *self);

static void
handle_method_call (GDBusConnection       *connection,
                    const gchar           *sender,
                    const gchar           *object_path,
                    const gchar           *interface_name,
                    const gchar           *method_name,
                    GVariant              *parameters,
                    GDBusMethodInvocation *invocation,
                    gpointer               user_data)
{
    CdosOverview *self = CDOS_OVERVIEW (user_data);
    CdosOverviewPrivate *priv = self->priv;

    if (g_strcmp0 (method_name, "OverviewToggle") == 0) 
    {
        cdos_overview_toggle (self);
        g_dbus_method_invocation_return_value (invocation, NULL);
    }
    if (g_strcmp0 (method_name, "RunMonitorSwitchDialog") == 0) 
    {
        cdos_monitor_switcher_get_default ();
        g_dbus_method_invocation_return_value (invocation, NULL);
    }
}

static const GDBusInterfaceVTable interface_vtable =
{
        handle_method_call,
        NULL, /* Get Property */
        NULL, /* Set Property */
};

static void
on_bus_gotten (GObject          *source_object,
               GAsyncResult     *res,
               gpointer         *user_data)
{
    CdosOverview *self = CDOS_OVERVIEW (user_data);
    CdosOverviewPrivate *priv = self->priv;
    GDBusConnection *connection;
    GError *error = NULL;

    if (priv->bus_cancellable == NULL ||
        g_cancellable_is_cancelled (priv->bus_cancellable)) {
            g_warning ("Operation has been cancelled, so not retrieving session bus");
            return;
    }

    connection = g_bus_get_finish (res, &error);
    if (connection == NULL) {
            g_warning ("Could not get session bus: %s", error->message);
            g_error_free (error);
            return;
    }
    priv->connection = connection;
    g_dbus_connection_register_object (connection,
                                      CDOS_DBUS_PATH,
                                      priv->introspection_data->interfaces[0],
                                      &interface_vtable,
                                      self,
                                      NULL,
                                      NULL);
}

static void
cdos_overview_init (CdosOverview *self)
{
    CdosOverviewPrivate *priv;
    priv = self->priv = cdos_overview_get_instance_private (self);

    priv->spacing = 0;
    priv->button_press_id = 0;
    priv->button_release_id = 0;

    priv->shown = FALSE;
    priv->modal = FALSE;
    self->animation_in_progress = FALSE;
    priv->hide_in_progress = FALSE;
    priv->visible = FALSE;

    priv->window_switch_timeout_id = 0;
    priv->window_switch_timestamp = 0;
    priv->last_hovered_window = NULL;

    priv->introspection_data = g_dbus_node_info_new_for_xml (cdos_overview_xml, NULL);
    priv->bus_cancellable = g_cancellable_new ();
    g_bus_get (G_BUS_TYPE_SESSION,
               priv->bus_cancellable,
               (GAsyncReadyCallback) on_bus_gotten,
               self);
}

static void
cdos_overview_monitors_changed (CdosLayoutManager *layout_manager,
                                CdosOverview *self)
{
    clutter_actor_hide ((ClutterActor *)self);

    const MetaRectangle *primary = cdos_layout_manager_get_primary_monitor (layout_manager);
    if (primary && self->priv->cover_pane)
    {
        clutter_actor_set_position (self->priv->cover_pane, primary->x, primary->y);
        clutter_actor_set_size (self->priv->cover_pane, primary->width, primary->height);
    }
}

static gboolean
cdos_overview_on_button_press (ClutterActor *actor, ClutterButtonEvent *event)
{
    return TRUE;
}

static gboolean
cdos_overview_on_button_release (ClutterActor *actor, ClutterButtonEvent *event)
{
    CdosOverview *self = CDOS_OVERVIEW (actor);
    if (self->animation_in_progress)
        return FALSE;
    clutter_actor_hide (actor);
    return TRUE;
}

static gboolean
cdos_overview_on_key_press (ClutterActor *actor, ClutterKeyEvent *event)
{
    CdosOverview *self = CDOS_OVERVIEW (actor);
    CdosOverviewPrivate *priv;
    priv = self->priv;

    if (self->animation_in_progress)
        return FALSE;

    if (!cdos_workspaces_view_key_press_event (priv->workspaces_view, event))
        return cdos_overview_thumbnails_box_key_press_event ((CdosOverviewThumbnailsBox *)priv->thumbnails_box, (ClutterEvent *)event);

    return TRUE;
}

static gboolean
cdos_overview_on_key_release (ClutterActor *actor, ClutterKeyEvent *event)
{
    guint symbol = clutter_event_get_key_symbol ((ClutterEvent *)event);
    CdosOverview *self = CDOS_OVERVIEW (actor);
    if (self->animation_in_progress)
        return FALSE;
    if (symbol == CLUTTER_KEY_Escape)
    {
        clutter_actor_hide (actor);
        return TRUE;
    }
    return FALSE;
}

static gboolean
cdos_overview_sync_input_mode (CdosOverview *self)
{
    CdosOverviewPrivate *priv;
    priv = self->priv;

    if (self->animation_in_progress)
    {
        return TRUE;
    }

    if (priv->shown)
    {
        if (!priv->modal)
        {
            if (cdos_push_modal ((ClutterActor *)self, 0, 0, CDOS_ACTION_MODE_OVERVIEW))
            {
                priv->modal = TRUE;
            }
            else
            {
                clutter_actor_hide ((ClutterActor *)self);
                return FALSE;
            }
        }
    }
    else
    {
        if (priv->modal)
        {
            cdos_pop_modal ((ClutterActor *)self, 0);
            priv->modal = FALSE;
        }
    }
    return TRUE;
}

/*
static gboolean
emit_showing_signal (CdosOverview *self)
{
    g_signal_emit (self, cdos_overview_signals[OVERVIEW_SHOWING], 0, NULL);
    return FALSE;
}
*/

static void
cdos_overview_animate_visible (CdosOverview *self)
{
    gfloat height = 0;
    CdosOverviewPrivate *priv;
    priv = self->priv;

    if (priv->visible || self->animation_in_progress)
        return;
    MetaDisplay *display = cdos_global_get_display (global);
    gint screen_width, screen_height;
    meta_display_get_size (display, &screen_width, &screen_height);
    clutter_actor_set_size ((ClutterActor *)self, screen_width, screen_height);

    priv->background_group = meta_background_group_new ();
    clutter_actor_add_child ((ClutterActor *)self, priv->background_group);
    priv->background = cdos_background_manager_new (priv->background_group, 1, 0, TRUE, BACKGROUND_SCHEMA);
    clutter_actor_set_reactive (priv->background_group, TRUE);
    clutter_actor_show (priv->background_group);

    priv->visible = TRUE;
    self->animation_in_progress = TRUE;

    priv->cover_pane = clutter_actor_new ();
    clutter_actor_set_reactive (priv->cover_pane, TRUE);

    clutter_actor_add_child ((ClutterActor *)self, priv->cover_pane);
    clutter_actor_set_position (priv->cover_pane, 0.f, 0.f);
    clutter_actor_set_size (priv->cover_pane, (gfloat)screen_width, (gfloat)screen_height);
    g_signal_connect (priv->cover_pane, "event", G_CALLBACK (cdos_overview_cover_pane_event_callback), self);

    meta_disable_unredirect_for_display (display);
    ClutterActor *window_group;
    window_group = cdos_global_get_window_group (global);
    clutter_actor_hide (window_group);

    CdosOverviewThumbnailsBox *thumbnails_box = cdos_overview_thumbnails_box_new ();
    clutter_actor_add_child ((ClutterActor *)self, (ClutterActor *)thumbnails_box);
    priv->thumbnails_box = (ClutterActor *)thumbnails_box;
    clutter_actor_set_opacity ((ClutterActor *)thumbnails_box, 0);
    cdos_transition_simple ((ClutterActor *)thumbnails_box,
                            CLUTTER_EASE_OUT_QUAD,
                            ANIMATION_TIME,
                            "opacity", 255,
                            NULL);

    CdosWorkspacesView *workspaces_view = cdos_workspaces_view_new ();
    _add_actor (self, workspaces_view);
    priv->workspaces_view = workspaces_view;
    clutter_actor_hide ((ClutterActor *)workspaces_view);
    clutter_actor_get_preferred_height (priv->thumbnails_box,
                                        screen_width,
                                        NULL, &height);
    CdosLayoutManager *layout_manager = cdos_layout_manager_get_default ();
    const MetaRectangle *monitor = cdos_layout_manager_get_monitor_geometry (layout_manager, meta_display_get_primary_monitor (display));
    cdos_workspaces_view_set_avail_size (workspaces_view, monitor->height - height);

    CLUTTER_ACTOR_CLASS (cdos_overview_parent_class)->show ((ClutterActor *)self);

    ClutterActor *ui_group = cdos_ui_plugin_get_ui_group ();
    clutter_actor_set_child_above_sibling (ui_group, (ClutterActor *)self, NULL);

    ClutterActor *bg = clutter_actor_get_first_child (priv->background_group);
    while (bg)
    {
        cdos_transition_simple (bg,
                                CLUTTER_EASE_OUT_QUAD,
                                ANIMATION_TIME,
                                "brightness", 0.55, NULL);
        bg = clutter_actor_get_next_sibling (bg);
    }

    clutter_actor_set_child_above_sibling ((ClutterActor *)self, priv->cover_pane, NULL);
    clutter_actor_show (priv->cover_pane);
    display_current_time ("Showing overview");
    ClutterTransition *trans = cdos_transition_simple ((ClutterActor *)self,
                                                        CLUTTER_EASE_OUT_QUAD,
                                                        ANIMATION_TIME,
                                                        "opacity", 255,
                                                        NULL);
    g_signal_connect (trans, "completed", G_CALLBACK(cdos_overview_show_done), self);
    //g_idle_add ((GSourceFunc)emit_showing_signal, self);
    g_signal_emit (self, cdos_overview_signals[OVERVIEW_SHOWING], 0);
}

static void
cdos_overview_animate_not_visible (CdosOverview *self)
{
    CdosOverviewPrivate *priv;
    priv = self->priv;

    if (!priv->visible || self->animation_in_progress)
    {
        return;
    }

    self->animation_in_progress = TRUE;
    priv->hide_in_progress = TRUE;

    // cdos_start_enable_panels();

    clutter_actor_hide ((ClutterActor *)priv->workspaces_view);
    clutter_actor_hide (priv->thumbnails_box);

    ClutterTransition *trans = cdos_transition_simple ((ClutterActor *)self,
                                                        CLUTTER_EASE_OUT,
                                                        ANIMATION_TIME,
                                                        "opacity", 255,
                                                        NULL);
    g_signal_connect (trans, "completed", G_CALLBACK (cdos_overview_hide_done), self);

    clutter_actor_set_child_above_sibling ((ClutterActor *)self, priv->cover_pane, NULL);
    clutter_actor_show (priv->cover_pane);
    g_signal_emit (self, cdos_overview_signals[OVERVIEW_HIDING], 0, NULL);
}

static void
cdos_overview_show_done (ClutterTimeline *tml, CdosOverview *self)
{
    display_current_time (__func__);
    CdosOverviewPrivate *priv;
    priv = self->priv;
    self->animation_in_progress = FALSE;
    clutter_actor_hide (priv->cover_pane);

    g_signal_emit (self, cdos_overview_signals[OVERVIEW_SHOWN], 0, NULL);

    if (!priv->shown)
    {
        cdos_overview_animate_not_visible (self);
    }
    cdos_overview_sync_input_mode (self);
    cdos_global_sync_pointer (global);

    ClutterStage *stage = cdos_global_get_stage (global);
    StFocusManager *focus_manager = st_focus_manager_get_for_stage (stage);
    st_focus_manager_add_group (focus_manager, (StWidget *)self);
}

static void
cdos_overview_hide_done (ClutterTimeline *tml, CdosOverview *self)
{
    CdosOverviewPrivate *priv;
    priv = self->priv;

    ClutterActor *ui_group = cdos_ui_plugin_get_ui_group ();
    clutter_actor_set_child_below_sibling (ui_group, (ClutterActor *)self, NULL);

    if (priv->cover_pane)
    {
        clutter_actor_destroy (priv->cover_pane);
        priv->cover_pane = NULL;
    }

    meta_enable_unredirect_for_display (cdos_global_get_display (global));
    ClutterActor *window_group = cdos_global_get_window_group (global);
    clutter_actor_show (window_group);

    cdos_background_manager_destroy (priv->background);
    // clutter_actor_remove_child ((ClutterActor *)self, priv->background_group);
    clutter_actor_destroy (priv->background_group);
    priv->background_group = NULL;

    if (priv->workspaces_view)
    {
        clutter_actor_destroy ((ClutterActor *)priv->workspaces_view);
        priv->workspaces_view = NULL;
    }
    if (priv->thumbnails_box)
    {
        clutter_actor_destroy ((ClutterActor *)priv->thumbnails_box);
        priv->thumbnails_box = NULL;
    }

    priv->visible = FALSE;
    self->animation_in_progress = FALSE;
    priv->hide_in_progress = FALSE;
    priv->shown = FALSE;

    CLUTTER_ACTOR_CLASS (cdos_overview_parent_class)->hide ((ClutterActor *)self);

    g_signal_emit (self, cdos_overview_signals[OVERVIEW_HIDDEN], 0, NULL);

    if (priv->shown || self->animation_in_progress)
    {
        cdos_overview_animate_visible (self);
    }
    cdos_overview_sync_input_mode (self);

    CdosLayoutManager *layout_manager = cdos_layout_manager_get_default ();
    cdos_layout_manager_update_regions (layout_manager);
    g_debug ("%s", __func__);
}

static gboolean
cdos_overview_cover_pane_event_callback (ClutterActor *actor, ClutterEvent *event, CdosOverview *self)
{
    return TRUE;
}


/**
 * cdos_overview_new:
 *
 * 创建一个新的实例对象；
 *
 * Returns:返回一个#CdosOverview的实例对象
 */
CdosOverview *
cdos_overview_new (void)
{
    if (overview)
        return overview;

    CdosOverview *self;
    self = g_object_new (CDOS_TYPE_OVERVIEW,
                         "name", "overview",
                         "style-class", "overview-group",
                         "reactive", TRUE,
                         "visible", FALSE,
                         NULL);

    CdosLayoutManager *layout = cdos_layout_manager_get_default ();
    const MetaRectangle *primary = cdos_layout_manager_get_primary_monitor (layout);
    if (primary)
    {
        clutter_actor_set_position ((ClutterActor *)self, primary->x, primary->y);
        clutter_actor_set_size ((ClutterActor *)self, primary->width, primary->height);
    }

    ClutterActor *ui_group = cdos_ui_plugin_get_ui_group ();
    _add_actor (ui_group, self);

    g_signal_connect (layout, "monitors-changed", G_CALLBACK (cdos_overview_monitors_changed), self);

    overview = self;
    return self;
}

CdosOverview *
cdos_overview_get (void)
{
    return cdos_overview_new ();
}
/**
 * cdos_overview_toggle:
 * @self: 当前实例对象
 *
 * 来回切换当前实例元素的显示和隐藏。
 */
void
cdos_overview_toggle (CdosOverview *self)
{
    g_return_if_fail (CDOS_IS_OVERVIEW (self));
    CdosOverviewPrivate *priv;
    priv = self->priv;
    if (self->animation_in_progress)
        return ;

    if (priv->shown)
    {
        clutter_actor_hide ((ClutterActor *)self);
    }
    else
    {
        clutter_actor_show ((ClutterActor *)self);
    }
}

static void
cdos_overview_show (ClutterActor *actor)
{
    CdosOverview *self = (CdosOverview *)actor;
    CdosOverviewPrivate *priv;
    priv = self->priv;
    if (priv->shown)
    {
        return;
    }
    priv->shown = TRUE;

    cdos_overview_sync_input_mode (self);

    cdos_overview_animate_visible (self);
}

static void
cdos_overview_hide (ClutterActor *actor)
{
    CdosOverview *self = (gpointer)actor;

    CdosOverviewPrivate *priv;
    priv = self->priv;

    if (!priv->shown)
    {
        return;
    }
    ClutterStage *stage = cdos_global_get_stage (global);
    StFocusManager *focus_manager = st_focus_manager_get_for_stage (stage);
    st_focus_manager_remove_group (focus_manager, (StWidget *)self);

    priv->shown = FALSE;
    cdos_overview_animate_not_visible (self);
    cdos_overview_sync_input_mode (self);
}

static void
cdos_overview_dispose (GObject *object)
{
    CdosOverview *self = CDOS_OVERVIEW(object);
    CdosOverviewPrivate *priv = self->priv;
    if (priv->bus_cancellable != NULL)
    {
        g_cancellable_cancel (priv->bus_cancellable);
        g_object_unref (priv->bus_cancellable);
        priv->bus_cancellable = NULL;
    }
    if (priv->introspection_data) 
    {
        g_dbus_node_info_unref (priv->introspection_data);
        priv->introspection_data = NULL;
    }
    if (priv->connection != NULL) 
    {
        g_object_unref (priv->connection);
        priv->connection = NULL;
    }
    G_OBJECT_CLASS (cdos_overview_parent_class)->dispose (object);
}

static void
cdos_overview_finalize (GObject *object)
{
    // CdosOverview *self = CDOS_OVERVIEW(object);
    overview = NULL;
    G_OBJECT_CLASS (cdos_overview_parent_class)->finalize (object);
}

static void
cdos_overview_allocate (ClutterActor *actor,
                        const ClutterActorBox *box,
                        ClutterAllocationFlags flags)
{
    StThemeNode *theme_node = st_widget_get_theme_node ((StWidget *)actor);
    ClutterActorBox content_box;
    st_theme_node_get_content_box (theme_node, box, &content_box);
    clutter_actor_set_allocation (actor, &content_box, flags);

    CdosOverview *self = (CdosOverview *)actor;
    CdosOverviewPrivate *priv = self->priv;
    if (priv->background_group)
        clutter_actor_allocate_preferred_size (priv->background_group, flags);

    if (priv->thumbnails_box)
    {
        CdosLayoutManager *layout = cdos_layout_manager_get_default ();
        const MetaRectangle *rect = cdos_layout_manager_get_primary_monitor (layout);
        ClutterActorBox box1;
        gfloat height = 0;

        clutter_actor_get_preferred_height (priv->thumbnails_box,
                                            content_box.x2 - content_box.x1,
                                            NULL, &height);
        box1.x1 = rect->x;
        box1.x2 = rect->x + rect->width;
        box1.y1 = rect->y + rect->height - height;
        box1.y2 = rect->y + rect->height;
        clutter_actor_allocate (priv->thumbnails_box, &box1, flags);

        clutter_actor_allocate (priv->cover_pane, &content_box, flags);
        clutter_actor_allocate ((ClutterActor *)priv->workspaces_view, &content_box, flags);
    }
}

static void
cdos_overview_class_init (CdosOverviewClass *klass)
{
    GObjectClass *gobject_class = G_OBJECT_CLASS (klass);
    gobject_class->dispose = cdos_overview_dispose;
    gobject_class->finalize = cdos_overview_finalize;

    ClutterActorClass *actor_class = CLUTTER_ACTOR_CLASS (klass);
    actor_class->allocate = cdos_overview_allocate;
    /*
    actor_class->get_prefered_width = cdos_overview_get_preferred_width;
    actor_class->get_prefered_height = cdos_overview_get_preferred_width;
    */
    actor_class->show = cdos_overview_show;
    actor_class->hide = cdos_overview_hide;

    actor_class->button_press_event = cdos_overview_on_button_press;
    actor_class->button_release_event = cdos_overview_on_button_release;
    actor_class->key_press_event = cdos_overview_on_key_press;
    actor_class->key_release_event = cdos_overview_on_key_release;

    /**
     * CdosOverview::hidden
     * @self: 信号发射实例
     *
     * 实例元素动画隐藏已经完成。
     */
    cdos_overview_signals[OVERVIEW_HIDDEN] = g_signal_new ("hidden",
                                                           G_TYPE_FROM_CLASS (klass),
                                                           G_SIGNAL_RUN_LAST,
                                                           0,
                                                           NULL,
                                                           NULL,
                                                           g_cclosure_marshal_VOID__VOID,
                                                           G_TYPE_NONE,
                                                           0);
    /**
     * CdosOverview::hiding
     * @self: 信号发射实例
     *
     * 实例元素动画隐藏正在进行中。
     */
    cdos_overview_signals[OVERVIEW_HIDING] = g_signal_new ("hiding",
                                                           G_TYPE_FROM_CLASS (klass),
                                                           G_SIGNAL_RUN_LAST,
                                                           0,
                                                           NULL,
                                                           NULL,
                                                           g_cclosure_marshal_VOID__VOID,
                                                           G_TYPE_NONE,
                                                           0);
    /**
     * CdosOverview::showing
     * @self: 信号发射实例
     *
     * 实例元素动画显示正在进行中。
     */
    cdos_overview_signals[OVERVIEW_SHOWING] = g_signal_new ("showing",
                                                            G_TYPE_FROM_CLASS (klass),
                                                            G_SIGNAL_RUN_LAST,
                                                            0,
                                                            NULL,
                                                            NULL,
                                                            g_cclosure_marshal_VOID__VOID,
                                                            G_TYPE_NONE,
                                                            0);
    /**
     * CdosOverview::shown
     * @self: 信号发射实例
     *
     * 实例元素动画显示已经完成。
     */
    cdos_overview_signals[OVERVIEW_SHOWN] = g_signal_new ("shown",
                                                          G_TYPE_FROM_CLASS (klass),
                                                          G_SIGNAL_RUN_LAST,
                                                          0,
                                                          NULL,
                                                          NULL,
                                                          g_cclosure_marshal_VOID__VOID,
                                                          G_TYPE_NONE,
                                                          0);
}
