#include "config.h"
#include <meta/display.h>
#include "cdos-panel.h"
#include "cdos-window-manager.h"

#include "cdos-popup-menu-item.h"
#include "cdos-popup-switch-menu-item.h"
#include "cdos-popup-submenu-menu-item.h"
#include "cdos-popup-separator-menu-item.h"
#include "cdos-popup-menu.h"
#include "cdos-popup-menu-manager.h"
#include "cdos-layout-manager.h"
#include "cdos-drag-drop.h"

#define DEFAULT_PANEL_HEIGHT 40.0

#ifndef HAVE_META_WINDOW_TILE
# ifdef HAVE_META_WINDOW_TILE_SYMBOL
# pragma message "Couldn't find meta_window_tile() function."
# define HAVE_META_WINDOW_TILE
typedef enum
{
  META_TILE_NONE,
  META_TILE_LEFT,
  META_TILE_RIGHT,
  META_TILE_MAXIMIZED
} MetaTileMode;

extern void meta_window_tile (MetaWindow *window, MetaTileMode mode);
# endif
#endif

#ifndef HAVE_META_WINDOW_CAN_TILE_SIDE_BY_SIDE
# ifdef HAVE_META_WINDOW_CAN_TILE_SIDE_BY_SIDE_SYMBOL
# pragma message "Couldn't find meta_window_can_tile_side_by_side() function."
# define HAVE_META_WINDOW_CAN_TILE_SIDE_BY_SIDE
  extern gboolean meta_window_can_tile_side_by_side (MetaWindow *window);
# endif
#endif

/**
 * SECTION:cdos-panel
 * @title: CdosPanel
 * @short_description: 任务栏
 *
 * 任务栏分为三部分：左侧，中间，右侧。左侧放menu、scale,
 * 中间为window-list，
 * 右侧放grouped-applets、可移动磁盘、网络、声音、systray、showdesktop等
 */
#define AUTOHIDE_ANIMATION_TIME  200
// 添加右键菜单
//

typedef struct _SettingsLauncher
{
    CdosPopupBaseMenuItem *laucher;
    StLabel *label;
    ClutterActor *icon;
    CdosPanel *panel;
    GCallback callback;
    gpointer user_data;
} SettingsLauncher;


struct _CdosPanel
{
    StWidget     parent;
    StWidget    *left_box;
    StWidget    *center_box;
    StWidget    *right_box;
    gboolean     bottom_position :1;
    gboolean     hideable;
    gint         hide_delay;
    gint         show_delay;
    guint        show_delay_id;
    guint        hide_timer;
    guint        show_timer;
    guint        loop_timer;
    gboolean     hidden;
    gboolean     disabled;
    gboolean     is_primary :1;
    CdosPopupMenu *context_menu;
    CdosPopupMenuManager *menu_manager;
    GList *settings_launchers;
    GSettings *settings;
    gint         monitor;
};

G_DEFINE_TYPE (CdosPanel, cdos_panel, ST_TYPE_WIDGET);

enum {
  PROP_NUM = 0,
  PROP_POSITION,
  PROP_HIDEABLE,
  PROP_PRIMARY,
  NUM_PROPERTIES
};

static GParamSpec *pspecs[NUM_PROPERTIES] = { NULL, };

static void cdos_panel_set_primary (CdosPanel *self, gboolean is_primary);
static void cdos_panel_clear_timers (CdosPanel *self);
static gboolean cdos_panel_show_panel (CdosPanel *self);

/**
 * cdos_panel_get_leftbox:
 * @panel: #CdosPanel类的一个对象
 *
 * 获取左侧的box
 *
 * Returns:左侧的box
 */
ClutterActor *
cdos_panel_get_leftbox (CdosPanel *panel)
{
    g_return_val_if_fail (CDOS_IS_PANEL (panel), NULL);
    return (ClutterActor *)panel->left_box;
}


/**
 * cdos_panel_get_centerbox:
 * @panel: #CdosPanel类的一个对象
 *
 * 获取中间的box
 *
 * Returns:中间的box
 */
ClutterActor *
cdos_panel_get_centerbox (CdosPanel *panel)
{
    g_return_val_if_fail (CDOS_IS_PANEL (panel), NULL);
    return (ClutterActor *)panel->center_box;
}

/**
 * cdos_panel_get_rightbox:
 * @panel: #CdosPanel类的一个对象
 *
 * 获取右侧的box
 *
 * Returns:右侧的box
 */
ClutterActor *
cdos_panel_get_rightbox (CdosPanel *panel)
{
    g_return_val_if_fail (CDOS_IS_PANEL (panel), NULL);
    return (ClutterActor *)panel->right_box;
}

static void
on_settings_changed (GSettings *settings, gchar *key, CdosPanel *self)
{
    self->hideable = g_settings_get_boolean (settings, "panel-autohide");
    self->hide_delay = g_settings_get_int (settings, "panel-hide-delay");
    self->show_delay = g_settings_get_int (settings, "panel-show-delay");

    if (self->hideable)
        cdos_panel_disable (self, -1);
    else
        cdos_panel_enable (self, 250);
}

static void
cdos_panel_dispose (GObject *gobject)
{
    CdosPanel *self = CDOS_PANEL (gobject);
    cdos_panel_clear_timers (self);
    if(self->settings_launchers)
    {
        GList *launchers = self->settings_launchers;
        while (launchers != NULL)
        {
            SettingsLauncher *launcher = launchers->data;
            clutter_actor_destroy ((ClutterActor *)launcher->laucher);
            launchers = launchers->next;
        }
        g_list_free (self->settings_launchers);
        self->settings_launchers = NULL;
    }
    if (self->settings)
    {
        g_signal_handlers_disconnect_by_data (self->settings, gobject);
        self->settings = NULL;
    }

    g_clear_pointer ((ClutterActor **)&self->context_menu, clutter_actor_destroy);
    g_clear_object (&self->menu_manager);

    G_OBJECT_CLASS (cdos_panel_parent_class)->dispose (gobject);
}

static void
cdos_panel_finalize (GObject *gobject)
{
  G_OBJECT_CLASS (cdos_panel_parent_class)->finalize (gobject);
}

static void
cdos_panel_set_property (GObject      *gobject,
                         guint         prop_id,
                         const GValue *value,
                         GParamSpec   *pspec)
{
  CdosPanel *obj = CDOS_PANEL (gobject);

  switch (prop_id)
    {
    case PROP_POSITION:
      cdos_panel_set_bottom_position (obj, g_value_get_boolean (value));
      break;
    case PROP_HIDEABLE:
      cdos_panel_set_hideable (obj, g_value_get_boolean (value));
      break;
    case PROP_PRIMARY:
      cdos_panel_set_primary (obj, g_value_get_boolean (value));
      break;

    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (gobject, prop_id, pspec);
      break;
    }
}

static void
cdos_panel_get_property (GObject    *gobject,
                         guint       prop_id,
                         GValue     *value,
                         GParamSpec *pspec)
{
  CdosPanel *self = CDOS_PANEL (gobject);

  switch (prop_id)
    {
    case PROP_POSITION:
      g_value_set_boolean(value, self->bottom_position);
      break;
    case PROP_HIDEABLE:
      g_value_set_boolean(value, self->hideable);
      break;
    case PROP_PRIMARY:
      g_value_set_boolean (value, self->is_primary);
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (gobject, prop_id, pspec);
      break;
    }
}

static void
cdos_panel_get_preferred_height (ClutterActor *actor,
                                 gfloat for_width,
                                 gfloat *min_height_p,
                                 gfloat *natural_height_p)
{
    CdosPanel *self = CDOS_PANEL (actor);
    StThemeNode *theme_node = st_widget_get_theme_node ((StWidget *)self);
    st_theme_node_adjust_for_width (theme_node, &for_width);

    gfloat left_height, center_height, right_height = 0;
    clutter_actor_get_preferred_height ((ClutterActor *)self->left_box, -1.0,
                                       NULL,
                                       &left_height);
    clutter_actor_get_preferred_height ((ClutterActor *)self->center_box, -1.0,
                                       NULL,
                                       &center_height);
    clutter_actor_get_preferred_height ((ClutterActor *)self->right_box, -1.0,
                                       NULL,
                                       &right_height);
    *natural_height_p = *min_height_p = MAX(left_height, MAX(center_height, right_height));
    st_theme_node_adjust_preferred_height (theme_node, min_height_p, natural_height_p);
}

static void
cdos_panel_get_preferred_width (ClutterActor *actor,
                                gfloat for_height,
                                gfloat *min_width_p,
                                gfloat *natural_width_p)
{
    CdosPanel *self = CDOS_PANEL (actor);
    StThemeNode *node = st_widget_get_theme_node ((StWidget *)self);
    st_theme_node_adjust_for_height (node, &for_height);

    *min_width_p = 1;

    const MetaRectangle *monitor;
    monitor = cdos_layout_manager_get_monitor_geometry (cdos_layout_manager_get_default (),
                                                        self->monitor);
    if (monitor)
        *natural_width_p = *min_width_p = monitor->width;
    else
    {
        gfloat left_width, center_width, right_width = 0;
        clutter_actor_get_preferred_width ((ClutterActor *)self->left_box, -1.0,
                                            NULL,
                                            &left_width);
        clutter_actor_get_preferred_width ((ClutterActor *)self->center_box, -1.0,
                                           NULL,
                                           &center_width);
        clutter_actor_get_preferred_width ((ClutterActor *)self->right_box, -1.0,
                                           NULL,
                                           &right_width);
        *natural_width_p = *min_width_p = left_width + center_width + right_width;
    }
    st_theme_node_adjust_preferred_width (node, min_width_p, natural_width_p);
}

static void
cdos_panel_allocate (ClutterActor           *actor,
                     const ClutterActorBox  *box,
                     ClutterAllocationFlags  flags)
{
    gfloat width;
    gfloat x = 0;
    StThemeNode *theme_node;
    ClutterActorBox content_box, child_box;

    CdosPanel *self = CDOS_PANEL (actor);

    theme_node = st_widget_get_theme_node (ST_WIDGET (self));

    clutter_actor_set_allocation (actor, box, flags);

    st_theme_node_get_content_box (theme_node, box, &content_box);

    width = content_box.x2 - content_box.x1;

    gfloat left_width, center_width, right_width = 0;
    clutter_actor_get_preferred_width ((ClutterActor *)self->left_box, -1.0,
                                       NULL,
                                       &left_width);
    clutter_actor_get_preferred_width ((ClutterActor *)self->center_box, -1.0,
                                       NULL,
                                       &center_width);
    clutter_actor_get_preferred_width ((ClutterActor *)self->right_box, -1.0,
                                       NULL,
                                       &right_width);

    gfloat spaceNeeded = 0.f;
    spaceNeeded = left_width + right_width;
    if (spaceNeeded < width)
        center_width = width - spaceNeeded;

    child_box.x1 = content_box.x1;
    child_box.x2 = child_box.x1 + left_width;
    child_box.y1 = content_box.y1;
    child_box.y2 = content_box.y2;
    clutter_actor_allocate ((ClutterActor *)self->left_box, &child_box, flags);

    x = child_box.x2;

    child_box.x1 = x;
    child_box.x2 = width -right_width;
    child_box.y1 = content_box.y1;
    child_box.y2 = content_box.y2;
    clutter_actor_allocate ((ClutterActor *)self->center_box, &child_box, flags);

    x = child_box.x2;
    child_box.x1 = x;
    child_box.x2 = content_box.x2;
    child_box.y1 = content_box.y1;
    child_box.y2 = content_box.y2;
    clutter_actor_allocate ((ClutterActor *)self->right_box, &child_box, flags);
}

static void
cdos_panel_paint (ClutterActor *actor, ClutterPaintContext *ctx)
{
    st_widget_paint_background (ST_WIDGET (actor), ctx);
    CdosPanel *self = CDOS_PANEL(actor);
    clutter_actor_paint ((ClutterActor *)self->left_box, ctx);
    clutter_actor_paint ((ClutterActor *)self->center_box, ctx);
    clutter_actor_paint ((ClutterActor *)self->right_box, ctx);
}

static gboolean
loop_check_status (CdosPanel *self)
{
    if (menu_stack_depth > 0)
        return TRUE;
    cdos_panel_disable (self, -1);
    return FALSE;
}

static gboolean
check_menu_open (CdosPanel *self)
{
    if (menu_stack_depth > 0)
    {
        if (!self->loop_timer)
        {
            guint timeout = self->hide_delay > 0 ? self->hide_delay : 3000;
            self->loop_timer = g_timeout_add (timeout, (GSourceFunc)loop_check_status, self);
        }
        return FALSE;
    }

    cdos_panel_disable (self, -1);
    return FALSE;
}

static gboolean
cdos_panel_on_leave (ClutterActor *actor,
                     ClutterCrossingEvent *event)
{
    CdosPanel *self = CDOS_PANEL (actor);

    if (!self->hideable)
        return FALSE;

    return check_menu_open (self);
}

static gboolean
cdos_panel_on_enter (ClutterActor *actor,
                     ClutterCrossingEvent *event)
{
    CdosPanel *self = CDOS_PANEL (actor);
    CdosGlobal *global = cdos_global_get ();
    MetaDisplay *display = cdos_global_get_display (global);
    gint primary = meta_display_get_primary_monitor (display);

    if (primary < 0)
        return FALSE;
    if (meta_display_get_monitor_in_fullscreen (display, primary))
        return FALSE;

    cdos_panel_clear_timers (self);
    self->show_delay_id = g_timeout_add (self->show_delay, 
                                        (GSourceFunc)cdos_panel_show_panel, 
                                        self);  
    return FALSE;
}

/**
 * cdos_panel_on_button_press:
 * @actor: #ClutterActor
 * @event: #ClutterEvent
 * @data: user data
 *
 * 点击出现右键菜单
 *
 * Returns: TRUE,出现右键菜单
 */
gboolean
cdos_panel_on_button_press (ClutterActor *actor,
                            ClutterEvent *event,
                            gpointer data)
{
    CdosPanel *self = (CdosPanel *)actor;
    guint event_type;

    event_type = clutter_event_get_button (event);
    if (event_type == 1)
    {
        if (self->context_menu && CDOS_POPUP_MENU_BASE (self->context_menu)->is_open)
        {
            cdos_popup_menu_close ((CdosPopupMenuBase *)self->context_menu, TRUE);
        }
    }
    else if (event_type == 3)
    {
        gfloat x, y;
        clutter_event_get_coords (event, &x, &y);
        ClutterStage *stage;
        stage = cdos_global_get_stage (cdos_global_get ());
        ClutterActor *target = clutter_stage_get_actor_at_pos (stage, CLUTTER_PICK_ALL, x, y);

        if (clutter_actor_contains ((ClutterActor *)self->center_box, target))
            cdos_popup_menu_base_toggle (CDOS_POPUP_MENU_BASE (self->context_menu));
        if (!CDOS_POPUP_MENU_BASE (self->context_menu)->is_open)
        {
            return FALSE;
        }

        CdosBoxPointer *box_pointer = cdos_popup_menu_get_box_pointer (self->context_menu);
        gfloat width = clutter_actor_get_width (actor);
        gfloat x1;
        clutter_actor_transform_stage_point (actor, x, 0, &x1, NULL);
        gfloat alignment = x1 / width;
        cdos_box_pointer_set_source_alignment (box_pointer, alignment);
    }
    return TRUE;
}

static void
panel_layout_window (guint mode)
{
    MetaMaximizeFlags max1, max2;

    GList *window_filter = NULL;
    MetaDisplay *display = cdos_global_get_display (cdos_global_get ());
    MetaWorkspaceManager *wsm = meta_display_get_workspace_manager (display);
    MetaWorkspace *workspace = meta_workspace_manager_get_active_workspace (wsm);
    GList *window_actors = cdos_global_get_window_actors (cdos_global_get ());
    GList *wa = window_actors;
    while (wa != NULL)
    {
        MetaWindow *mwin = meta_window_actor_get_meta_window (wa->data);
        if (meta_window_get_workspace (mwin) == workspace)
        {
            if (cdos_window_tracker_is_window_interesting (mwin))
            {
                window_filter = g_list_prepend (window_filter, mwin);
            }
        }
        wa = wa->next;
    }
    g_list_free (window_actors);

    MetaWindow *layout_windows[2];

    gint window_num = 0;
    GList *wf = window_filter;
    while (wf != NULL && window_num < G_N_ELEMENTS (layout_windows))
    {
        MetaWindow *meta_win = wf->data;
        gboolean minimized = FALSE;
        g_object_get (meta_win, "minimized", &minimized, NULL);

        if (!minimized)
        {
            layout_windows[window_num++] = meta_win;
        }
        wf = wf->next;
    }
    if (window_num == 0)
    {
        return;
    }
    else if (window_num == 1)
    {
        MetaWindow *to_be_shown = layout_windows[0];
        max1 = meta_window_get_maximized (to_be_shown);
        if (max1 & META_MAXIMIZE_BOTH)
            meta_window_unmaximize (to_be_shown, META_MAXIMIZE_BOTH);
        else
            meta_window_maximize (to_be_shown, META_MAXIMIZE_BOTH);
    }
    else
    {
        if (mode == 1)
        {
#ifdef HAVE_META_WINDOW_TILE
            MetaWindow *to_be_shown1 = layout_windows[0];
            MetaWindow *to_be_shown2 = layout_windows[1];
#ifdef HAVE_META_WINDOW_CAN_TILE_SIDE_BY_SIDE
            if (meta_window_can_tile_side_by_side (to_be_shown1))
#endif
            {
                meta_window_tile (to_be_shown1, META_TILE_LEFT);
            }
            max2 = meta_window_get_maximized (to_be_shown2);
#ifdef HAVE_META_WINDOW_CAN_TILE_SIDE_BY_SIDE
            if (!max2 && meta_window_can_tile_side_by_side (to_be_shown2))
#endif
            {
                meta_window_tile (to_be_shown2, META_TILE_RIGHT);
            }
#endif
        }
        else if (mode == 2)
        {
            MetaWindow *to_be_shown1 = layout_windows[0];
            MetaWindow *to_be_shown2 = layout_windows[1];
            max1 = meta_window_get_maximized (to_be_shown1);
            max2 = meta_window_get_maximized (to_be_shown2);

            if (max1 & META_MAXIMIZE_BOTH)
                meta_window_unmaximize (to_be_shown1, META_MAXIMIZE_BOTH);

            if (max2 & META_MAXIMIZE_BOTH)
                meta_window_unmaximize (to_be_shown2, META_MAXIMIZE_BOTH);

            meta_window_raise (to_be_shown2);
            meta_window_raise (to_be_shown1);
        }else if (mode == 3)
        {

        }
    }
}

#ifdef HAVE_META_WORKSPACE_MANAGER_SHOW_DESKTOP
static void
show_desktop (CdosPanel *self)
{
    CdosGlobal *global = cdos_global_get ();
    MetaWorkspaceManager *wsm = cdos_global_get_workspace_manager (global);

    MetaWorkspace *workspace = meta_workspace_manager_get_active_workspace (wsm);
    gboolean is_showing_desktop = meta_workspace_get_showing_desktop (workspace);
    if (!is_showing_desktop)
    {
        meta_workspace_manager_show_desktop (wsm, cdos_global_get_current_time (global));
    }
    else
    {
        meta_workspace_manager_unshow_desktop (wsm);
    }
}
#endif

static void
open_system_monitor (void)
{
    cdos_util_spawn_command_line ("gnome-system-monitor");
}

static void
stack_windows (void)
{
    panel_layout_window (2);
}

#ifdef HAVE_META_WINDOW_TILE
static void
abreast_windows (void)
{
    panel_layout_window (1);
}
#endif

static void
panel_settings_launcher_activate (CdosPopupBaseMenuItem *actor,
                                  ClutterEvent *event,
                                  gboolean keep_menu,
                                  SettingsLauncher *launcher)
{
    if (launcher->callback)
        ((void (*)(void *))(launcher->callback)) (launcher->user_data);

    cdos_popup_menu_base_toggle (CDOS_POPUP_MENU_BASE (launcher->panel->context_menu));
}

static SettingsLauncher *
panel_new_settings_launcher (CdosPanel *self,
                             const gchar *label_text,
                             const gchar *icon_name,
                             GCallback callback,
                             gpointer user_data)
{
    ClutterActor *icon = NULL;
    SettingsLauncher *launcher = g_new0 (SettingsLauncher, 1);
    CdosPopupBaseMenuItem *settings_launcher = cdos_popup_base_menu_item_new ();
    launcher->laucher = settings_launcher;
    ClutterActor *launcher_actor = CLUTTER_ACTOR (settings_launcher);
    st_widget_add_style_class_name (ST_WIDGET (launcher_actor), "popup-contextmenu-item");

    StWidget *label = st_label_new (label_text);
    launcher->label = (StLabel *)label;
    if (icon_name)
    {
        ClutterActor *icon = g_object_new (ST_TYPE_ICON,
                                       "icon-name", icon_name,
                                       "icon-size", 16,
                                       NULL);
        launcher->icon = icon;
        cdos_popup_base_menu_item_add_actor (settings_launcher, icon, NULL);
    } else {
        st_widget_set_style_class_name(label,"popup-menu-item-label");
    }

    ActorChild child;
    child.expand = TRUE;
    child.align = -1;
    cdos_popup_base_menu_item_add_actor (settings_launcher, (ClutterActor *)label, &child);

    if (callback)
    {
        launcher->panel = self;
        launcher->callback = callback;
        launcher->user_data = user_data;
        g_signal_connect (settings_launcher, "activate", G_CALLBACK (panel_settings_launcher_activate), launcher);
    }
    return launcher;
}

static gboolean
panel_position_get_mapping (GValue   *value,
                            GVariant *variant,
                            gpointer  user_data)
{
    const char *str = NULL;
    gboolean checked = FALSE;

    /* get a pointer out of the variant, instead of copying the string out */
    g_variant_get (variant, "&s", &str);

    if (str ==NULL || *str == '\0')
        return FALSE;

    if (g_strcmp0 (str, "flipped") == 0)
        checked = TRUE;
    g_value_set_boolean (value, checked);

    return TRUE;
}

static GVariant *
panel_position_set_mapping (const GValue   *value,
                            const GVariantType *variant_type,
                            gpointer  user_data)
{
    GVariant *v;
    const char *str[] = {"classic", "flipped"};
    gboolean checked = FALSE;

    checked = g_value_get_boolean (value);
    v = g_variant_new_string (str[checked]);

    return v;
}

static CdosPopupMenu *
panel_get_context_menu (CdosPanel *self, guint orientation)
{
    ClutterActor *source_actor = CLUTTER_ACTOR (self);
    CdosPopupSwitchMenuItem *auto_hide_item;
    CdosPopupSwitchMenuItem *panel_position_item;

    GList *launchers = self->settings_launchers;
    CdosPopupMenu *menu = cdos_popup_menu_new_with_params (source_actor, 0.0, ST_SIDE_BOTTOM);
    self->context_menu = menu;

    ClutterActor *menu_actor = (ClutterActor *)menu->actor;
    ClutterActor *ui_group = cdos_ui_plugin_get_ui_group ();
    clutter_actor_add_child (ui_group, menu_actor);
    st_widget_add_style_class_name (ST_WIDGET (menu_actor),"popup-contextmenu");

    // 显示桌面
#ifdef HAVE_META_WORKSPACE_MANAGER_SHOW_DESKTOP
    SettingsLauncher *show_desktop_item = panel_new_settings_launcher (self,
                                                                       _("Show desktop"),
                                                                       NULL,
                                                                       (GCallback)show_desktop, self);
    cdos_popup_menu_base_add_menu_item (CDOS_POPUP_MENU_BASE (menu), show_desktop_item->laucher, -1);
    launchers = g_list_append (launchers, show_desktop_item);

    CdosPopupSeparatorMenuItem  *s1 = cdos_popup_separator_menu_item_new ();
    cdos_popup_menu_base_add_menu_item (CDOS_POPUP_MENU_BASE (menu), CDOS_POPUP_BASE_MENU_ITEM (s1), -1);
#endif

    // 系统
    SettingsLauncher *menu_settings_item = panel_new_settings_launcher (self,
                                                                        _("Open the system monitor"),
                                                                        NULL,
                                                                        open_system_monitor, NULL);
    cdos_popup_menu_base_add_menu_item (CDOS_POPUP_MENU_BASE (menu), menu_settings_item->laucher, -1);
    launchers = g_list_append (launchers, menu_settings_item);

    SettingsLauncher *stack_windows_item = panel_new_settings_launcher (self,
                                                                        _("Stack windows"),
                                                                        NULL,
                                                                        stack_windows, NULL);
    cdos_popup_menu_base_add_menu_item (CDOS_POPUP_MENU_BASE (menu), stack_windows_item->laucher, -1);
    launchers = g_list_append (launchers,stack_windows_item);

#ifdef HAVE_META_WINDOW_TILE
    SettingsLauncher *abreast_windows_item = panel_new_settings_launcher (self,
                                                                          _("Abreast windows"),
                                                                          NULL,
                                                                          abreast_windows, NULL);
    cdos_popup_menu_base_add_menu_item (CDOS_POPUP_MENU_BASE (menu), abreast_windows_item->laucher, -1);
    launchers = g_list_append (launchers, abreast_windows_item);
#endif

    CdosPopupSeparatorMenuItem  *s4 = cdos_popup_separator_menu_item_new();
    cdos_popup_menu_base_add_menu_item (CDOS_POPUP_MENU_BASE (menu), CDOS_POPUP_BASE_MENU_ITEM(s4), -1);

/*    SettingsLauncher *overlapping_windows_item = panel_new_settings_launcher (_("Overlapping windows"),
                                                            "overlapping-windows", "show-desktop", menu);
    cdos_popup_menu_base_add_menu_item (CDOS_POPUP_MENU_BASE (menu), overlapping_windows_item->laucher, -1);
     launchers = g_list_append(launchers,overlapping_windows_item);

    ClutterActor *show_desktop_actor = (ClutterActor *)cdos_popup_base_menu_item_get_actor ((CdosPopupBaseMenuItem *)show_desktop_item->laucher);
   // g_signal_connect (show_desktop_actor, "button-release-event", G_CALLBACK (panel_settings_launcher_activate), "show-desktop");
    */
    self->settings_launchers = launchers;

    panel_position_item = cdos_popup_switch_menu_item_new_with_params (_("Placed at the top of the screen"), FALSE, NULL);
    cdos_popup_base_menu_item_set_keep_menu (CDOS_POPUP_BASE_MENU_ITEM (panel_position_item), FALSE);
    cdos_panel_add_menu_item (self, (CdosPopupBaseMenuItem *)panel_position_item);
    g_settings_bind_with_mapping (self->settings, "desktop-layout",
                                  panel_position_item, "state",
                                  G_SETTINGS_BIND_DEFAULT,
                                  panel_position_get_mapping,
                                  panel_position_set_mapping,
                                  self,
                                  NULL);

    auto_hide_item = cdos_popup_switch_menu_item_new_with_params (_("Auto hide panel"), FALSE, NULL);
    cdos_popup_base_menu_item_set_keep_menu (CDOS_POPUP_BASE_MENU_ITEM (auto_hide_item), FALSE);
    cdos_panel_add_menu_item (self, (CdosPopupBaseMenuItem *)auto_hide_item);
    g_settings_bind (self->settings, "panel-autohide", auto_hide_item, "state", G_SETTINGS_BIND_DEFAULT);

    return menu;
}

/* class init: 类的继承关系，将子类自己的方法绑定到父类,
 * 定义属性和信号
 */
static void
cdos_panel_class_init (CdosPanelClass *klass)
{
    GObjectClass *gobject_class = G_OBJECT_CLASS (klass);
    ClutterActorClass *actor_class = CLUTTER_ACTOR_CLASS (klass);

    gobject_class->finalize = cdos_panel_finalize;
    gobject_class->dispose = cdos_panel_dispose;
    gobject_class->set_property = cdos_panel_set_property;
    gobject_class->get_property = cdos_panel_get_property;

    actor_class->get_preferred_height = cdos_panel_get_preferred_height;
    actor_class->get_preferred_width = cdos_panel_get_preferred_width;
    actor_class->allocate = cdos_panel_allocate;
    actor_class->paint = cdos_panel_paint;
    actor_class->leave_event = cdos_panel_on_leave;
    actor_class->enter_event = cdos_panel_on_enter;

    pspecs[PROP_POSITION] = g_param_spec_boolean ("bottom-position",
                                                  "Bottom  position",
                                                  "The panel is bottom or top",
                                                  TRUE,
                                                  G_PARAM_READWRITE);
    pspecs[PROP_HIDEABLE] = g_param_spec_boolean ("hideable",
                                                  "hide panel",
                                                  "The panel can hide or not",
                                                  FALSE,
                                                  G_PARAM_READWRITE);

    pspecs[PROP_PRIMARY] = g_param_spec_boolean ("is-primary",
                                                 "is primary",
                                                 "the panel is primary or not",
                                                 TRUE,
                                                 G_PARAM_READWRITE);


    g_object_class_install_properties (gobject_class, NUM_PROPERTIES, pspecs);
}

static void
cdos_panel_init (CdosPanel *self)
{
    ClutterActor *actor = CLUTTER_ACTOR (self);
    clutter_actor_set_offscreen_redirect (actor, CLUTTER_OFFSCREEN_REDIRECT_ALWAYS);

    clutter_actor_set_x_expand (actor, TRUE);
    clutter_actor_set_x_align (actor, CLUTTER_ACTOR_ALIGN_FILL);

    ClutterActor *left_box = g_object_new (ST_TYPE_WIDGET, NULL);
    ClutterActor *center_box = g_object_new (ST_TYPE_WIDGET, NULL);
    ClutterActor *right_box = g_object_new (ST_TYPE_WIDGET, NULL);

    ClutterLayoutManager *box_layout;
    box_layout = clutter_box_layout_new ();
    clutter_actor_set_layout_manager (left_box, box_layout);
    box_layout = clutter_box_layout_new ();
    clutter_actor_set_layout_manager (center_box, box_layout);
    box_layout = clutter_box_layout_new ();
    clutter_actor_set_layout_manager (right_box, box_layout);

    clutter_actor_set_name (left_box, "panel-left");
    clutter_actor_set_name (center_box, "panel-center");
    clutter_actor_set_name (right_box, "panel-right");

    self->left_box = (StWidget *)left_box;
    self->center_box = (StWidget *)center_box;
    self->right_box = (StWidget *)right_box;

    self->bottom_position = TRUE;
    clutter_actor_add_child (actor, left_box);
    clutter_actor_add_child (actor, center_box);
    clutter_actor_add_child (actor, right_box);

    self->settings = cdos_get_cdos_settings ();
    g_signal_connect (self->settings, "changed", G_CALLBACK(on_settings_changed), self);
    on_settings_changed (self->settings, NULL, self);
}

static void
cdos_panel_clear_timers (CdosPanel *self)
{
    if(self->show_timer)
    {
        g_source_remove (self->show_timer);
        self->show_timer = 0;
    }

    if(self->hide_timer)
    {
        g_source_remove (self->hide_timer);
        self->hide_timer = 0;
    }

    if(self->loop_timer)
    {
        g_source_remove (self->loop_timer);
        self->loop_timer = 0;
    }
    if (self->show_delay_id)
    {
        g_source_remove (self->show_delay_id);
        self->show_delay_id = 0;
    }
}

static gboolean
cdos_panel_hide_panel (CdosPanel *self)
{
    self->hide_timer = 0;

    ClutterSeat *seat = clutter_backend_get_default_seat (clutter_get_default_backend ());
    ClutterInputDevice *mouse = clutter_seat_get_pointer (seat);
    if (mouse)
    {
        graphene_point_t point;
        clutter_input_device_get_coords (mouse, NULL, &point);
        gfloat x, y, w, h;
        clutter_actor_get_transformed_size ((ClutterActor *)self, &w, &h);
        clutter_actor_get_transformed_position ((ClutterActor *)self, &x, &y);
        if (point.x <= x || point.x >= x + w ||
            point.y <= y || point.y >= y + h)
        {
        }
        else
        {
            cdos_panel_clear_timers (self);
            return FALSE;
        }
    }

    CdosGlobal *global = cdos_global_get ();
    MetaDisplay *display = cdos_global_get_display (global);
    CdosLayoutManager *layout = cdos_layout_manager_get_default ();

    cdos_layout_manager_on_window_restacked (display, layout);

    const MetaRectangle *monitor = cdos_layout_manager_get_primary_monitor (layout);
    if (!monitor)
        return FALSE;

    guint animation_time = AUTOHIDE_ANIMATION_TIME;
    gfloat height = clutter_actor_get_height ((ClutterActor *)self);
    float y = self->bottom_position ? (monitor->y + monitor->height - 1) : (monitor->y - height + 1);

    ClutterActor *actor_parent = clutter_actor_get_parent ((ClutterActor *)self);
    cdos_transition_simple (actor_parent,
                            CLUTTER_EASE_OUT_QUAD,
                            animation_time,
                            "y", y,
                            // "opacity",TRANS_PROP_UCHAR,255,
                            NULL);
    self->hidden = TRUE;
    return FALSE;
}

static gboolean
cdos_panel_show_panel (CdosPanel *self)
{
    self->show_delay_id = 0;
    if(!self->hidden)
        return FALSE;

    CdosGlobal *global = cdos_global_get ();
    MetaDisplay *display = cdos_global_get_display (global);
    CdosLayoutManager *layout = cdos_layout_manager_get_default ();
    cdos_layout_manager_on_window_restacked (display, layout);

    const MetaRectangle *monitor = cdos_layout_manager_get_primary_monitor (layout);
    if (!monitor)
        return FALSE;

    gfloat animation_time = AUTOHIDE_ANIMATION_TIME;
    gfloat height = clutter_actor_get_height ((ClutterActor *)self);
    float y = self->bottom_position ? (monitor->y + monitor->height - height) : (monitor->y);

    ClutterActor *actor_parent = clutter_actor_get_parent ((ClutterActor *)self);
    cdos_transition_simple (actor_parent,
                            CLUTTER_EASE_IN,
                            animation_time,
                            "y", y,
                            NULL);

    self->hidden = FALSE;

    if (self->hideable)
    {
        // 这里需要判断是否有 menu 处于 open 状态，如果 menu 处于open状态，则不能直接 disable panel，等 menu 关闭了之后才能 disable panel
        g_timeout_add (CHECK_MENU_OPEN_DELAY, (GSourceFunc)check_menu_open, self);
    }

    return FALSE;
}

static void
cdos_panel_leave_panel (CdosPanel *self, gint delay)
{
    cdos_panel_clear_timers (self);

    delay = delay >= 0 ? delay : self->hide_delay;
    delay = MAX(delay, 500);

    self->hide_timer = g_timeout_add (delay, (GSourceFunc)cdos_panel_hide_panel, self);
}

/**
 * cdos_panel_enable:
 * @self: #CdosPanel 类的对象
 *
 * 显示面板
 */
void
cdos_panel_enable (CdosPanel *self, gint delay)
{
    g_return_if_fail (CDOS_IS_PANEL (self));
    cdos_panel_clear_timers (self);
    delay = delay >= 0 ? delay : 250;
    self->show_delay_id = g_timeout_add (delay, 
                                        (GSourceFunc)cdos_panel_show_panel, 
                                        self);
}

/**
 * cdos_panel_disable:
 * @self: #CdosPanel 类的对象
 * @delay: #gint
 *
 * 隐藏面板
 */
void
cdos_panel_disable (CdosPanel *self, gint delay)
{
    g_return_if_fail (CDOS_IS_PANEL (self));
    cdos_panel_leave_panel (self, delay);
}

gboolean 
cdos_panel_get_hidden (CdosPanel  *self)
{
    return self->hidden;
}

/**
 * cdos_panel_set_bottom_position:
 * @self: #CdosPanel的一个类
 * @bottom: 是否把panel放到底部
 *
 * 设置panel的位置，是否把面板放到底部
 */
void
cdos_panel_set_bottom_position (CdosPanel *self,
                                gboolean bottom)
{
    g_return_if_fail (CDOS_IS_PANEL (self));

    self->bottom_position = bottom;
    if (self->context_menu)
    {
        cdos_popup_menu_set_arrow_side (self->context_menu, bottom ? ST_SIDE_BOTTOM : ST_SIDE_TOP);
    }
}

/**
 * cdos_panel_set_hideable:
 * @self: #CdosPanel类的一个对象
 * @hideable: 是否隐藏panel的boolean值
 *
 * 设置隐藏panel的boolean值
 *
 */
void
cdos_panel_set_hideable (CdosPanel  *self,
                         gboolean    hideable)
{
    g_return_if_fail (CDOS_IS_PANEL (self));
    if (self->hideable == hideable)
        return;

    g_settings_set_boolean (self->settings, "panel-autohide", hideable);
}

gboolean
cdos_panel_get_hideable (CdosPanel *self)
{
    g_return_val_if_fail (CDOS_IS_PANEL (self), FALSE);
    return self->hideable;
}

static void
cdos_panel_set_primary (CdosPanel   *self,
                        gboolean     is_primary)
{
    self->is_primary = is_primary;
}

/**
 * cdos_panel_get_bottom_position:
 * @self: #CdosPanel的一个类
 *
 * 获取是否把面板放到底部的boolean值
 *
 * Returns: 获取是否把panel放到底部的boolean值
 */
gboolean
cdos_panel_get_bottom_position (CdosPanel *self)
{
    g_return_val_if_fail (CDOS_IS_PANEL (self), FALSE);

    return self->bottom_position;
}

static void
on_drop_over_in (ClutterDropAction *action,
                 ClutterActor      *actor,
                 void  *self)
{
    g_debug ("over-in");
    clutter_actor_save_easing_state (actor);
    clutter_actor_set_easing_mode (actor, CLUTTER_EASE_OUT_CIRC);
    clutter_actor_set_opacity (actor, 196);
    clutter_actor_restore_easing_state (actor);
}

static void
on_drop_over_out (ClutterDropAction *action,
                  ClutterActor      *actor,
                  void  *self)
{
    g_debug ("over-out");
    clutter_actor_save_easing_state (actor);
    clutter_actor_set_easing_mode (actor, CLUTTER_EASE_OUT_CIRC);
    clutter_actor_set_opacity (actor, 255);
    clutter_actor_restore_easing_state (actor);
    drag_drop_successfull = FALSE;
}

static void
on_drop_drop (ClutterDropAction *action,
              ClutterActor      *actor,
              gfloat             event_x,
              gfloat             event_y,
              void              *self)
{
    g_debug ("panel drop");
    GSettings *settings = cdos_get_cdos_settings ();
    char **launchers = g_settings_get_strv (settings, "panel-launchers");
    if (!launchers)
        return;

    ClutterDragAction *drag = cdos_get_current_drag_action ();
    if (!drag)
    {
        g_warn_if_reached ();
        drag_drop_successfull = FALSE;
        return;
    }
    char *desktop_id = g_object_get_data ((GObject *)drag, "desktop-id");
    if (!desktop_id)
        return;

    gint i;
    for (i = 0; launchers[i] != NULL; i++)
        if (g_strcmp0 (desktop_id, launchers[i]) == 0)
            goto out;
    guint max_apps = g_settings_get_int(settings, "panel-launchers-max-number");
    if (i >= max_apps)
        goto out;

    gint len = i;
    char **new_settings = g_malloc (sizeof (char *) * (len + 2));
    for (i = 0; i < len; i++)
        new_settings[i] = launchers[i];
    new_settings[len] = desktop_id;
    new_settings[len + 1] = NULL;
    g_settings_set_strv (settings, "panel-launchers", (const char * const *)new_settings);
    g_free (new_settings);
    drag_drop_successfull = TRUE;

out:
    g_strfreev (launchers);
    return;
}

static void
on_drop_cancel (ClutterDropAction *action,
                ClutterActor      *actor,
                gfloat             event_x,
                gfloat             event_y,
                void  *self)
{
    g_debug ("drop-candel");
    drag_drop_successfull = FALSE;
}
/**
 * cdos_panel_new:
 * @bottom_position: 是否把panel放到底部的boolean值
 * @is_primary: TRUE
 *
 * 新建一个面板
 *
 * Returns: panel
 */
CdosPanel *
cdos_panel_new (gboolean bottom_position, gboolean is_primary)
{
    CdosPanel *panel;

    panel = g_object_new (CDOS_TYPE_PANEL,
                          "bottom-position", bottom_position,
                          "is-primary", is_primary,
                          NULL);
    panel->context_menu = panel_get_context_menu (panel, bottom_position ? ST_SIDE_BOTTOM : ST_SIDE_TOP);
    panel->menu_manager = cdos_popup_menu_manager_new (CLUTTER_ACTOR(panel));
    cdos_popup_menu_manager_add_menu (panel->menu_manager, CDOS_POPUP_MENU_BASE (panel->context_menu), -1);

    //  g_signal_connect (CLUTTER_ACTOR(panel), "button-press-event", G_CALLBACK (on_button_press), panel);

    ClutterAction *action = cdos_drop_action_new ((ClutterActor *)panel->center_box);
    g_signal_connect (action, "over-in", G_CALLBACK (on_drop_over_in), NULL);
    g_signal_connect (action, "over-out", G_CALLBACK (on_drop_over_out), NULL);
    g_signal_connect (action, "drop", G_CALLBACK (on_drop_drop), NULL);
    g_signal_connect (action, "drop-cancel", G_CALLBACK (on_drop_cancel), NULL);
    return panel;
}

gfloat
cdos_panel_get_height (CdosPanel *cdos_panel)
{
    g_return_val_if_fail (CDOS_IS_PANEL (cdos_panel), 0);
    return clutter_actor_get_height (CLUTTER_ACTOR (cdos_panel));
}

void
cdos_panel_add_launcher (CdosPanel *self, const gchar *label, const gchar *icon_name, GCallback callback, gint position, gpointer user_data)
{
    SettingsLauncher *item = panel_new_settings_launcher (self,
                                                          label,
                                                          icon_name,
                                                          callback, user_data);
    cdos_popup_menu_base_add_menu_item (CDOS_POPUP_MENU_BASE (self->context_menu), item->laucher, position);
}

void
cdos_panel_add_separator_item (CdosPanel *self)
{
    CdosPopupSeparatorMenuItem  *s1 = cdos_popup_separator_menu_item_new ();
    cdos_popup_menu_base_add_menu_item (CDOS_POPUP_MENU_BASE (self->context_menu), CDOS_POPUP_BASE_MENU_ITEM (s1), -1);
}

void
cdos_panel_add_menu_item (CdosPanel *self, CdosPopupBaseMenuItem *item)
{
    cdos_popup_menu_base_add_menu_item (CDOS_POPUP_MENU_BASE (self->context_menu), item, -1);
    st_widget_add_style_class_name (ST_WIDGET (item), "popup-contextmenu-item");
}

void
cdos_panel_set_monitor (CdosPanel *self, gint monitor)
{
    self->monitor = monitor;
    clutter_actor_queue_relayout (CLUTTER_ACTOR (self));
}
