#include <math.h>
#include <glib.h>
#include <clutter/clutter.h>
#include "cdos-popup-base-menu-item.h"
#include "cdos-start.h"
#include "cdos-marshal.h"

/**
 * SECTION: cdos-popup-base-menu-item.c
 * @Title: CdosPopupBaseMenuItem
 * @short_description: CdosPopupBaseMenuItem的作用是为menu提供一个弹出菜单里的选项，
 * 这个控件必须放在Menu里。这个控件为我们提供了一个便于加入menu的item。
 *
 * CdosPopupBaseMenuItem 提供一个可以和menu配套使用的控件。其其实就是由box构成。
 */

typedef struct  _CdosPopupBaseMenuItemPrivate
{
    ClutterActor   *ornament;
    CdosOrnament    ornament_type;
    gfloat          spacing;
    gboolean        active;
    gboolean        activatable;
    gboolean        sensitive;
    gulong          notify_hover_id;
    gulong          key_focus_in_id;
    gulong          key_focus_out_id;
    gboolean        keep_menu;
}CdosPopupBaseMenuItemPrivate;

G_DEFINE_TYPE_WITH_PRIVATE (CdosPopupBaseMenuItem, cdos_popup_base_menu_item, ST_TYPE_WIDGET);

enum
{
    PROP_0,

    PROP_ACTIVE,
    PROP_ACTIVETABLE,
    PROP_SENSITIVE
};

enum
{
    ACTIVATE,
    ACTIVE_CHANGED,
    SENSITIVE_CHANGED,
    LAST_SIGNAL
};
static guint cdos_popup_base_menu_item_signals[LAST_SIGNAL] =  { 0, };

typedef struct _RequestedSize
{
    ClutterActor *actor;
    gboolean expand;
    gfloat min;
    gfloat nat;
}RequestedSize;

static void
cdos_popup_base_menu_item_on_hover_changed (ClutterActor          *actor,
                                            GParamSpec            *spec,
                                            CdosPopupBaseMenuItem *item);

static void
cdos_popup_base_menu_item_dispose (GObject *gobject)
{
    G_OBJECT_CLASS (cdos_popup_base_menu_item_parent_class)->dispose (gobject);
}

static void
cdos_popup_base_menu_item_set_property (GObject     *gobject,
                                        guint     prop_id,
                                        const GValue *value,
                                        GParamSpec *pspec)
{
    CdosPopupBaseMenuItem *item = CDOS_POPUP_BASE_MENU_ITEM (gobject);

    switch(prop_id)
    {
        case PROP_ACTIVE:
            cdos_popup_base_menu_item_set_active (item, g_value_get_boolean(value));
            break;
        case PROP_ACTIVETABLE:
            cdos_popup_base_menu_item_set_activatable (item, g_value_get_boolean(value));
            break;
        case PROP_SENSITIVE:
            cdos_popup_base_menu_item_set_sensitive (item, g_value_get_boolean (value));
            break;
        default:
        G_OBJECT_WARN_INVALID_PROPERTY_ID (gobject, prop_id, pspec);
    }
}

static void
cdos_popup_base_menu_item_get_property (GObject    *gobject,
                                        guint     prop_id,
                                        GValue      *value,
                                        GParamSpec   *pspec)
{
    CdosPopupBaseMenuItem *item = CDOS_POPUP_BASE_MENU_ITEM (gobject);
    CdosPopupBaseMenuItemPrivate *priv = cdos_popup_base_menu_item_get_instance_private (item);

    switch(prop_id)
    {
        case PROP_ACTIVE:
            g_value_set_boolean (value, priv->active);
            break;
        case PROP_ACTIVETABLE:
            g_value_set_boolean (value, priv->activatable);
            break;
        case PROP_SENSITIVE:
            g_value_set_boolean (value, priv->sensitive);
            break;
        default:
            G_OBJECT_WARN_INVALID_PROPERTY_ID (gobject, prop_id, pspec);
    }
}

static void
cdos_popup_base_menu_item_get_preferred_width (ClutterActor *actor,
                                               gfloat        for_height,
                                               gfloat       *min_width_p,
                                               gfloat       *natural_width_p)
{
    gfloat  minimum, natural;
    gint n_children = 0;
    ClutterActorIter iter;
    ClutterActor *child;

    CdosPopupBaseMenuItem *item = CDOS_POPUP_BASE_MENU_ITEM (actor);
    CdosPopupBaseMenuItemPrivate *priv = cdos_popup_base_menu_item_get_instance_private (item);
    gfloat spacing = priv->spacing;

    StThemeNode *theme_node = st_widget_get_theme_node (ST_WIDGET (actor));
    st_theme_node_adjust_for_height (theme_node, &for_height);

    minimum = natural = 0;

    clutter_actor_iter_init (&iter, actor);
    while (clutter_actor_iter_next (&iter, &child))
    {
        gfloat child_min = 0, child_nat = 0;
        if (!clutter_actor_is_visible (child))
            continue;
        n_children++;
        clutter_actor_get_preferred_width (child, for_height, &child_min, &child_nat);
        minimum += child_min;
        natural += child_nat;
    }
    if (n_children > 1)
    {
        minimum += spacing * (n_children - 1);
        natural += spacing * (n_children - 1);
    }

    st_theme_node_adjust_preferred_width (theme_node, &minimum, &natural);

    *min_width_p = minimum;
    *natural_width_p = natural;
}

static void
cdos_popup_base_menu_item_get_preferred_height (ClutterActor *actor,
                                                gfloat        for_width,
                                                gfloat       *min_height_p,
                                                gfloat       *natural_height_p)
{
    gfloat minimum, natural;
    ClutterActorIter iter;
    ClutterActor *child;

    minimum = natural = 0;

    StThemeNode *theme_node = st_widget_get_theme_node (ST_WIDGET (actor));
    st_theme_node_adjust_for_width (theme_node, &for_width);

    clutter_actor_iter_init (&iter, actor);
    while (clutter_actor_iter_next (&iter, &child))
    {
        gfloat child_min = 0, child_nat = 0;
        if (!clutter_actor_is_visible (child))
            continue;
        clutter_actor_get_preferred_height (child, for_width, &child_min, &child_nat);
        minimum = MAX (minimum, child_min);
        natural = MAX (natural, child_nat);
    }

    st_theme_node_adjust_preferred_height (theme_node, &minimum, &natural);

    *min_height_p = minimum;
    *natural_height_p = natural;
}

/*
 * Pulled from ClutterBoxLayout
 */
static gint
compare_gap (gconstpointer p1,
             gconstpointer p2,
             gpointer      data)
{
    RequestedSize *sizes = data;
    const guint *c1 = p1;
    const guint *c2 = p2;
    const gint d1 = MAX (sizes[*c1].nat - sizes[*c1].min, 0);
    const gint d2 = MAX (sizes[*c2].nat - sizes[*c2].min, 0);

    gint delta = (d2 - d1);

    if (0 == delta)
        delta = (*c2 - *c1);
    return delta;
}

static void
cdos_popup_base_menu_item_allocate (ClutterActor           *actor,
                                    const ClutterActorBox  *box,
                                    ClutterAllocationFlags  flags)
{
    StThemeNode *theme_node;
    ClutterActorBox content_box;
    clutter_actor_set_allocation (actor, box, flags);

    theme_node = st_widget_get_theme_node (ST_WIDGET (actor));
    st_theme_node_get_content_box (theme_node, box, &content_box);

    ClutterActor *child;
    ClutterActorIter iter;

    ClutterActorBox child_allocation;

    gboolean is_rtl = FALSE;
    gfloat width, height;
    gint size, extra;
    gint n_vis_children = 0;
    gint n_expand_children = 0;
    gint n_extra_widgets = 0;

    gint x = 0;

    gfloat child_size;
    RequestedSize *sizes;

    clutter_actor_iter_init (&iter, actor);
    while (clutter_actor_iter_next (&iter, &child))
    {
        if (!clutter_actor_is_visible (child))
            continue;
        n_vis_children++;

        if (clutter_actor_needs_expand (child, CLUTTER_ORIENTATION_HORIZONTAL))
            n_expand_children++;
    }
    if (n_vis_children <= 0)
        return;

    CdosPopupBaseMenuItem *item = CDOS_POPUP_BASE_MENU_ITEM (actor);
    CdosPopupBaseMenuItemPrivate *priv = cdos_popup_base_menu_item_get_instance_private (item);
    gfloat spacing = priv->spacing;

    width = content_box.x2 - content_box.x1;
    height = content_box.y2 - content_box.y1;
    size = width - (n_vis_children - 1) * spacing;

    sizes = g_newa (RequestedSize, n_vis_children);
    gint i = 0;

    clutter_actor_iter_init (&iter, actor);
    while (clutter_actor_iter_next (&iter, &child))
    {
        if (!clutter_actor_is_visible (child))
            continue;
        clutter_actor_get_preferred_width (child,
                                           height,
                                           &sizes[i].min,
                                           &sizes[i].nat);
        size -= sizes[i].min;
        sizes[i].actor = child;
        sizes[i].expand = clutter_actor_needs_expand (child, CLUTTER_ORIENTATION_HORIZONTAL);
        i++;
    }
    guint *spreading = g_newa (guint, n_vis_children);
    for (i = 0; i < n_vis_children; i++)
        spreading[i] = i;

    g_qsort_with_data (spreading,
                       n_vis_children, sizeof (guint),
                       compare_gap, sizes);

    for (i = n_vis_children - 1; i >= 0 && size > 0; i--)
    {
        gint glue = (size + i) / (i + 1);
        gint gap = sizes[spreading[i]].nat - sizes[spreading[i]].min;
        extra = MIN (glue, gap);
        sizes[spreading[i]].min += extra;
        size -= extra;
    }
    if (n_expand_children > 0)
    {
        extra = size / n_expand_children;
        n_extra_widgets = size % n_expand_children;
    }
    else
        extra = 0;

    ClutterTextDirection text_dir;
    text_dir = clutter_actor_get_text_direction (actor);
    is_rtl = text_dir == CLUTTER_TEXT_DIRECTION_RTL;

    x = content_box.x1;

    for (i = 0; i < n_vis_children; i++)
    {
        child = sizes[i].actor;
        child_size = sizes[i].min;
        if (sizes[i].expand)
        {
            child_size += extra;
            if (n_extra_widgets > 0)
            {
                child_size++;
                n_extra_widgets--;
            }
        }
        child_allocation.x1 = x;
        child_allocation.x2 = child_allocation.x1 + child_size;

        x += child_size + spacing;

        gfloat min_h, nat_h;
        clutter_actor_get_preferred_height (child, child_size,
                                            &min_h, &nat_h);
        child_allocation.y1 = (int)(content_box.y1 + (height - nat_h) / 2);
        child_allocation.y2 = child_allocation.y1 + nat_h;

        if (is_rtl)
        {
            gfloat w = child_allocation.x2 - child_allocation.x1;
            child_allocation.x2 = content_box.x1 + (content_box.x2 - child_allocation.x1);
            child_allocation.x1 = child_allocation.x2 - w;
        }
        clutter_actor_allocate (child, &child_allocation, flags);
        /*
        if (sizes[i].expand)
        {
            clutter_actor_allocate (child, &child_allocation, flags);
        }
        else
        {
            clutter_actor_allocate (child, &child_allocation, flags);
        }
        */
    }
}

static gboolean
cdos_popup_base_menu_item_on_key_press (ClutterActor *actor,
                                        ClutterKeyEvent *event)
{
    guint symbol = clutter_event_get_key_symbol ((ClutterEvent *)event);
    if (symbol == CLUTTER_KEY_space || symbol == CLUTTER_KEY_Return)
    {
        CdosPopupBaseMenuItem *item = CDOS_POPUP_BASE_MENU_ITEM (actor);
        CdosPopupBaseMenuItemPrivate *priv = cdos_popup_base_menu_item_get_instance_private (item);
        cdos_popup_base_menu_item_activate(item, (ClutterEvent *)event, priv->keep_menu);
        return TRUE;
    }
    return CLUTTER_ACTOR_CLASS (cdos_popup_base_menu_item_parent_class)->key_press_event (actor, event);
}

static gboolean
cdos_popup_base_menu_item_on_button_press (ClutterActor *actor,
                                           ClutterButtonEvent *event)
{
    st_widget_set_style_pseudo_class (ST_WIDGET (actor), "clicked");
    return TRUE;
}

static gboolean
cdos_popup_base_menu_item_on_button_release (ClutterActor *actor,
                                             ClutterButtonEvent *event)
{
    CdosPopupBaseMenuItem *item = CDOS_POPUP_BASE_MENU_ITEM (actor);
    CdosPopupBaseMenuItemPrivate *priv = cdos_popup_base_menu_item_get_instance_private (item);
    st_widget_remove_style_pseudo_class (ST_WIDGET (actor), "clicked");
    cdos_popup_base_menu_item_activate(item, (ClutterEvent *)event, priv->keep_menu);
    return TRUE;
}

static gboolean
cdos_popup_base_menu_item_on_leave (ClutterActor *actor,
                                    ClutterCrossingEvent *event)
{
    /*
    ClutterActor *related = event->related;
    g_debug ("%s %s", __func__, G_OBJECT_TYPE_NAME (related));
    */
    st_widget_remove_style_pseudo_class (ST_WIDGET (actor), "clicked");

    if (CLUTTER_ACTOR_CLASS (cdos_popup_base_menu_item_parent_class)->leave_event)
        return CLUTTER_ACTOR_CLASS (cdos_popup_base_menu_item_parent_class)->leave_event (actor, event);
    else
        return FALSE;
}

static void
cdos_popup_base_menu_item_key_focus_in (ClutterActor *actor)
{
    cdos_popup_base_menu_item_set_active (CDOS_POPUP_BASE_MENU_ITEM (actor), TRUE);
}

static void
cdos_popup_base_menu_item_key_focus_out (ClutterActor *actor)
{
    cdos_popup_base_menu_item_set_active (CDOS_POPUP_BASE_MENU_ITEM (actor), FALSE);
}

static void
cdos_popup_base_menu_item_on_hover_changed (ClutterActor *actor,
                                            GParamSpec *spec,
                                            CdosPopupBaseMenuItem *item)
{
    gboolean hover = st_widget_get_hover (ST_WIDGET(actor));
    cdos_popup_base_menu_item_set_active (item, hover);
}

static void
cdos_popup_base_menu_item_style_changed (StWidget *widget)
{
    CdosPopupBaseMenuItem *item = CDOS_POPUP_BASE_MENU_ITEM (widget);
    CdosPopupBaseMenuItemPrivate *priv = cdos_popup_base_menu_item_get_instance_private (item);

    StThemeNode *theme_node = st_widget_get_theme_node (widget);
    priv->spacing = st_theme_node_get_length (theme_node, "spacing");

    ST_WIDGET_CLASS (cdos_popup_base_menu_item_parent_class)->style_changed (widget);
}

static void
cdos_popup_base_menu_item_init (CdosPopupBaseMenuItem  *item)
{
    CdosPopupBaseMenuItemPrivate *priv;
    priv = cdos_popup_base_menu_item_get_instance_private (item);

    clutter_actor_set_reactive ((ClutterActor *)item, TRUE);

    st_widget_set_style_class_name ((StWidget *)item, "popup-menu-item");
    st_widget_set_track_hover ((StWidget *)item, TRUE);
    st_widget_set_can_focus ((StWidget *)item, TRUE);

    priv->ornament = NULL;
    priv->ornament_type = CDOS_ORNAMENT_NONE;
    priv->active = FALSE;
    priv->sensitive = TRUE;
    priv->activatable = TRUE;
    priv->keep_menu = FALSE;

    priv->notify_hover_id = g_signal_connect (item,
                                              "notify::hover",
                                              G_CALLBACK (cdos_popup_base_menu_item_on_hover_changed),
                                              item);
}

static void
cdos_popup_base_menu_item_class_init (CdosPopupBaseMenuItemClass *klass)
{
    GObjectClass *gobject_class = G_OBJECT_CLASS(klass);
    GParamSpec *pspec;

    gobject_class->set_property = cdos_popup_base_menu_item_set_property;
    gobject_class->get_property = cdos_popup_base_menu_item_get_property;
    gobject_class->dispose = cdos_popup_base_menu_item_dispose;

    ClutterActorClass *actor_class = CLUTTER_ACTOR_CLASS (klass);
    actor_class->get_preferred_width = cdos_popup_base_menu_item_get_preferred_width;
    actor_class->get_preferred_height = cdos_popup_base_menu_item_get_preferred_height;
    actor_class->allocate = cdos_popup_base_menu_item_allocate;

    actor_class->key_press_event = cdos_popup_base_menu_item_on_key_press;
    actor_class->button_press_event = cdos_popup_base_menu_item_on_button_press;
    actor_class->button_release_event = cdos_popup_base_menu_item_on_button_release;
    actor_class->leave_event = cdos_popup_base_menu_item_on_leave;
    actor_class->key_focus_in = cdos_popup_base_menu_item_key_focus_in;
    actor_class->key_focus_out = cdos_popup_base_menu_item_key_focus_out;

    StWidgetClass *widget_class = ST_WIDGET_CLASS (klass);
    widget_class->style_changed = cdos_popup_base_menu_item_style_changed;

    pspec = g_param_spec_boolean("active",
                "Active",
                "The active of item",
                TRUE,
                G_PARAM_READWRITE);
    g_object_class_install_property(gobject_class, PROP_ACTIVE, pspec);

    pspec = g_param_spec_boolean("activatable",
                "Activatable",
                "The activatable of item",
                TRUE,
                G_PARAM_READWRITE);
    g_object_class_install_property(gobject_class, PROP_ACTIVETABLE, pspec);

    pspec = g_param_spec_boolean("sensitive",
                "sensitive",
                "The sensitive of item",
                TRUE,
                G_PARAM_READWRITE);
    g_object_class_install_property(gobject_class, PROP_SENSITIVE, pspec);

//signals
    cdos_popup_base_menu_item_signals[ACTIVATE] = g_signal_new (
            "activate",
            G_TYPE_FROM_CLASS(klass),
            G_SIGNAL_RUN_LAST,
            G_STRUCT_OFFSET(CdosPopupBaseMenuItemClass, activate),
            NULL,
            NULL,
            _cdos_marshal_VOID__POINTER_BOOLEAN,
            G_TYPE_NONE,
            2,
            G_TYPE_POINTER,
            G_TYPE_BOOLEAN);

    cdos_popup_base_menu_item_signals[ACTIVE_CHANGED] = g_signal_new (
            "active-changed",
            G_TYPE_FROM_CLASS(klass),
            G_SIGNAL_RUN_LAST,
            G_STRUCT_OFFSET(CdosPopupBaseMenuItemClass, active_changed),
            NULL,
            NULL,
            NULL,
            G_TYPE_NONE,
            1,
            G_TYPE_BOOLEAN);

    cdos_popup_base_menu_item_signals[SENSITIVE_CHANGED] = g_signal_new (
            "sensitive-changed",
            G_TYPE_FROM_CLASS(klass),
            G_SIGNAL_RUN_LAST,
            G_STRUCT_OFFSET(CdosPopupBaseMenuItemClass, sensitive_changed),
            NULL,
            NULL,
            NULL,
            G_TYPE_NONE,
            1,
            G_TYPE_BOOLEAN);
}

/**
 * cdos_popup_base_menu_item_new：
 *
 * 新建一个 #CdosPopupBaseMenuItem 类的实例。
 *
 * Returns value: 返回一个 #CdosPopupBaseMenuItem 的实例。
 */
CdosPopupBaseMenuItem *
cdos_popup_base_menu_item_new (void)
{
    return g_object_new (CDOS_TYPE_POPUP_BASE_MENU_ITEM, NULL);
}

/**
 * cdos_popup_base_menu_item_new_with_params:
 * @params: 初始化类需要的基本信息
 *
 * 新建一个 #CdosPopupBaseMenuItem 类的实例。
 *
 * Return value: 返回一个 #CdosPopupBaseMenuItem 的实例。
 */
CdosPopupBaseMenuItem *
cdos_popup_base_menu_item_new_with_params (PopupBaseMenuItemInit *paramso)
{
    CdosPopupBaseMenuItem *item;
    item = g_object_new (CDOS_TYPE_POPUP_BASE_MENU_ITEM, NULL);
    PopupBaseMenuItemInit *params = g_slice_copy (sizeof (PopupBaseMenuItemInit), paramso);
    cdos_popup_base_menu_item_set_params(item, params);
    g_slice_free (PopupBaseMenuItemInit, params);
    return item;

}

/**
 * cdos_popup_base_menu_item_set_params:
 * @item: 需要设置类的属性
 * @params: 设置类的基本信息
 *
 */
void
cdos_popup_base_menu_item_set_params (CdosPopupBaseMenuItem *item,
                                      PopupBaseMenuItemInit *params)
{
    CdosPopupBaseMenuItemPrivate *priv;
    g_return_if_fail (CDOS_IS_POPUP_BASE_MENU_ITEM (item));

    if (NULL == params)
        return;

    priv = cdos_popup_base_menu_item_get_instance_private (item);

    if ((params->reactive == TRUE) &&
        (params->activate == TRUE) &&
        (params->hover == TRUE)  &&
        (params->sensitive == TRUE) &&
        (params->style_class == NULL) &&
        (params->focusOnHover == TRUE))
    {
        return;
    }

    priv->activatable = params->reactive && params->activate;

    cdos_popup_base_menu_item_set_sensitive (item, (priv->activatable && params->sensitive));

    if (params->style_class)
    {
        st_widget_remove_style_class_name (ST_WIDGET (item), "popup-menu-item");
        st_widget_set_style_class_name (ST_WIDGET (item), params->style_class);
    }

    if (!(params->reactive && params->hover))
    {
        if (priv->notify_hover_id)
        {
            g_signal_handler_disconnect (item, priv->notify_hover_id);
            priv->notify_hover_id = 0L;
        }
    }
}

/**
 * cdos_popup_base_menu_item_activate:
 * @item: A #CdosPopupBaseMenuItem
 * @event: 事件:
 * @keepMenu: ？？
 *
 * 使item处于activate状态
 *
 */
void
cdos_popup_base_menu_item_activate (CdosPopupBaseMenuItem *item,
                                    ClutterEvent          *event,
                                    gboolean               keep_menu)
{
    g_signal_emit (item, cdos_popup_base_menu_item_signals[ACTIVATE], 0, event, keep_menu);
}

/**
 * cdos_popup_base_menu_item_set_active:
 * @self: A #CdosPopupBaseMenuItem
 * @active: 设置activate状态
 *
 * 设置self的activate状态
 *
 */
void
cdos_popup_base_menu_item_set_active (CdosPopupBaseMenuItem *item,
                                      gboolean               active)
{
    CdosPopupBaseMenuItemPrivate *priv;
    g_return_if_fail (CDOS_IS_POPUP_BASE_MENU_ITEM (item));
    priv = cdos_popup_base_menu_item_get_instance_private (item);
    if(active != priv->active)
    {
        priv->active = active;
        if (active)
        {
            st_widget_add_style_pseudo_class(ST_WIDGET (item),  "active");
            clutter_actor_grab_key_focus (CLUTTER_ACTOR (item));
        }
        else
            st_widget_set_style_pseudo_class (ST_WIDGET (item), NULL);
        g_signal_emit (item, cdos_popup_base_menu_item_signals[ACTIVE_CHANGED], 0, active);
    }
}

/**
 * cdos_popup_base_menu_item_set_sensitive:
 * @item: A #CdosPopupBaseMenuItem
 * @sensitive: 是否是可以点击状态
 *
 * 设置self的是否是可以点击
 */
void
cdos_popup_base_menu_item_set_sensitive (CdosPopupBaseMenuItem *item,
                                         gboolean               sensitive)
{
    CdosPopupBaseMenuItemPrivate *priv;
    g_return_if_fail (CDOS_IS_POPUP_BASE_MENU_ITEM (item));

    priv = cdos_popup_base_menu_item_get_instance_private (item);

    if (priv->sensitive == sensitive)
        return;

    priv->sensitive = sensitive;
    clutter_actor_set_reactive ((ClutterActor *)item, sensitive);
    st_widget_set_can_focus ((StWidget *)item, sensitive);
    if (sensitive)
        st_widget_remove_style_pseudo_class ((StWidget *)item, "insensitive");
    else
        st_widget_add_style_pseudo_class ((StWidget *)item, "insensitive");

    g_signal_emit (item, cdos_popup_base_menu_item_signals[SENSITIVE_CHANGED], 0, sensitive);
}

/**
 * cdos_popup_base_menu_item_destroy:
 * @item: A #CdosPopupBaseMenuItem
 *
 * 销毁item
 */
void
cdos_popup_base_menu_item_destroy (CdosPopupBaseMenuItem *item)
{
    g_return_if_fail (CDOS_IS_POPUP_BASE_MENU_ITEM (item));
    ClutterActor *actor = CLUTTER_ACTOR (item);
    ClutterActor *parent = clutter_actor_get_parent (actor);
    if (!parent)
    {
        clutter_actor_hide (actor);
        clutter_actor_add_child (cdos_ui_plugin_get_ui_group (), actor);
    }
    clutter_actor_destroy (actor);
}

/**
 * cdos_popup_base_menu_item_add_actor:
 * @self: A #CdosPopupBaseMenuItem
 * @child: 需要添加的actor
 * @params: 添加的布局信息
 *
 * 往CdosPopupBaseMenuItem添加一个＃ClutterActor
 *
 */
void
cdos_popup_base_menu_item_add_actor (CdosPopupBaseMenuItem *item,
                                     ClutterActor          *child,
                                     ActorChild            *params)
{
    g_return_if_fail (CDOS_IS_POPUP_BASE_MENU_ITEM (item));
    g_return_if_fail (CLUTTER_IS_ACTOR (child));
    gboolean expand = FALSE;
    ClutterActorAlign align = CLUTTER_ACTOR_ALIGN_START;
    if (params)
    {
        expand = params->expand;
        align = params->align;
    }
    ClutterActor *actor = CLUTTER_ACTOR (item);
    clutter_actor_add_child (actor, child);

    /*
    if (align == ST_ALIGN_MIDDLE || expand)
        clutter_actor_set_x_expand (child, TRUE);
    else
        clutter_actor_set_x_expand (child, FALSE);
    */
    clutter_actor_set_x_expand (child, expand);
    clutter_actor_set_x_align (child, align);
}

/**
 * cdos_popup_base_menu_item_set_ornament:
 * @item: A #CdosPopupBaseMenuItem
 * @show: 是否显示
 *
 * 是否显示item中的dot
 *
 */
void
cdos_popup_base_menu_item_set_ornament (CdosPopupBaseMenuItem *item,
                                        CdosOrnament           ornament)
{
    g_return_if_fail (CDOS_IS_POPUP_BASE_MENU_ITEM (item));

    CdosPopupBaseMenuItemPrivate *priv;
    priv = cdos_popup_base_menu_item_get_instance_private (item);

    if (ornament >= CDOS_ORNAMENT_DOT &&
        ornament <= CDOS_ORNAMENT_SPACE)
    {
        if (priv->ornament_type == ornament)
            return;

        priv->ornament_type = ornament;
        //const char *ornaments[CDOS_ORNAMENT_NONE] = {"\u2022", "\u2713", ""};
        const char *ornaments[CDOS_ORNAMENT_NONE] = {"\u2022",
                                                     "\u2713",
                                                     "\u2715", /* \u2717  */
                                                     "\u25A2",
                                                     "\u2602",
                                                     "\u2601",
                                                     "\u2600",
                                                     ""};
        if (!priv->ornament)
        {
            priv->ornament = g_object_new (ST_TYPE_LABEL,
                                           "style-class", "popup-menu-item-ornament",
                                           "text", ornaments[ornament],
                                           NULL);
            clutter_actor_insert_child_below ((ClutterActor *)item, priv->ornament, NULL);
            clutter_actor_set_x_align (priv->ornament, CLUTTER_ACTOR_ALIGN_START);
            st_widget_set_style ((StWidget *)priv->ornament, "text-align: center");
        }
        else
            st_label_set_text ((StLabel *)priv->ornament, ornaments[ornament]);
    }
    else
    {
        if (!priv->ornament)
            return;
        clutter_actor_destroy (priv->ornament);
        priv->ornament = NULL;
        priv->ornament_type = CDOS_ORNAMENT_NONE;
    }
}

/**
 * cdos_popup_base_menu_item_get_active:
 * @item: A #CdosPopupBaseMenuItem
 *
 * 获取item的activate属性
 *
 * Return value： A #gboolean
 */
gboolean
cdos_popup_base_menu_item_get_active (CdosPopupBaseMenuItem *item)
{
    CdosPopupBaseMenuItemPrivate *priv;
    g_return_val_if_fail (CDOS_IS_POPUP_BASE_MENU_ITEM (item), FALSE);
    priv = cdos_popup_base_menu_item_get_instance_private (item);

    return priv->active;
}

/**
 * cdos_popup_base_menu_item_set_activatable:
 * @item: A #CdosPopupBaseMenuItem
 * @activatable: 是否activatable
 *
 * 设置的activatable
 *
 */
void
cdos_popup_base_menu_item_set_activatable (CdosPopupBaseMenuItem *item,
                                           gboolean      activatable)
{
    CdosPopupBaseMenuItemPrivate *priv;
    g_return_if_fail (CDOS_IS_POPUP_BASE_MENU_ITEM (item));

    priv = cdos_popup_base_menu_item_get_instance_private (item);
    priv->activatable = activatable;
    cdos_popup_base_menu_item_set_sensitive (item, priv->activatable && priv->sensitive);
}

/**
 * cdos_popup_base_menu_item_get_init_params:
 * @params: A #PopupBaseMenuItemInit
 *
 * 当我们创建一个PopupBaseMenuItem的控件时需要的默认参数，可以更改
 *
 * Return value： A#PopupBaseMenuItemInit
 */
PopupBaseMenuItemInit *
cdos_popup_base_menu_item_get_init_params (PopupBaseMenuItemInit *params)
{
    if (!params)
        return NULL;

    params->reactive = TRUE;
    params->activate = TRUE;
    params->hover = TRUE;
    params->sensitive = TRUE;
    params->style_class = NULL;

    return params;
}

/**
 * cdos_popup_base_menu_item_get_actor_child:
 * @child: A #ActorChild
 *
 * 当我们添加一个控件需要的必备参数
 *
 * Return value: a #ActorChild
 */
ActorChild *
cdos_popup_base_menu_item_get_actor_child (ActorChild *child)
{
    if (NULL == child)
        return NULL;

    child->expand = FALSE;
    child->align = ST_ALIGN_START;

    return child;
}

/**
 * cdos_popup_base_menu_item_set_keep_menu:
 * @item: A #CdosPopupBaseMenuItem
 * @keep_menu: #gboolean
 *
 * 设置item的keep_menu属性
 *
 */
void
cdos_popup_base_menu_item_set_keep_menu (CdosPopupBaseMenuItem *item,
                                         gboolean               keep_menu)
{
    g_return_if_fail (CDOS_IS_POPUP_BASE_MENU_ITEM (item));
    CdosPopupBaseMenuItemPrivate *priv;
    priv = cdos_popup_base_menu_item_get_instance_private (item);
    priv->keep_menu = keep_menu;
}

/**
 * cdos_popup_base_menu_item_get_keep_menu:
 * @item: A #CdosPopupBaseMenuItem
 *
 * 获取self的keep_menu成员
 *
 * Returns: ＃gboolean
 */
gboolean
cdos_popup_base_menu_item_get_keep_menu (CdosPopupBaseMenuItem *item)
{
    g_return_val_if_fail (CDOS_IS_POPUP_BASE_MENU_ITEM (item), FALSE);
    CdosPopupBaseMenuItemPrivate *priv;
    priv = cdos_popup_base_menu_item_get_instance_private (item);
    return priv->keep_menu;
}
