#include "cdos-popup-menu-base.h"
#include "cdos-popup-sub-menu.h"
#include "cdos-popup-menu.h"
#include "cdos-popup-menu-section.h"
#include "cdos-popup-submenu-menu-item.h"
#include <string.h>
#include "cdos-debug.h"
#include "cdos-start.h"
#include "cdos-util.h"
#include "cdos-marshal.h"

/**
 * SECTION: cdos-popup-menu-base.c
 * @Title: CdosPopupMenuBase
 * @short_description: 菜单控件的一个抽象基类，我们不可以直接创建一个这样的对象。
 *　
 *
 * CdosPopupImageMenuItem 实现菜单控件的基本功能
 */


struct _CdosPopupMenuBasePrivate
{
    ClutterActor *source_actor;
    StBoxLayout *box;
    CdosPopupBaseMenuItem *active_menu_item;
    GList *child_menus;
    CdosPopupMenuBase *parent_menu;
    gboolean in_destruction;
};

G_DEFINE_TYPE_WITH_PRIVATE (CdosPopupMenuBase, cdos_popup_menu_base, CDOS_TYPE_BOX_POINTER);

enum
{
    PROP_0,

    PROP_SOURCE_ACTOR,
    PROP_BOX
};

enum
{
    ACTIVATE,
    ACTIVE_CHANGED,
    CHILD_MENU_ADDED,
    CHILD_MENU_REMOVED,
    MENU_ITEM_ADD,
    OPEN_STATE_CHANGED,

    LAST_SIGNAL
};

static guint cdos_popup_menu_base_signals[LAST_SIGNAL] = {0,};

static void
cdos_popup_menu_base_set_box (CdosPopupMenuBase *self, StBoxLayout *box);

static void
cdos_popup_menu_base_dispose (GObject *gobject)
{
    CdosPopupMenuBase *self = CDOS_POPUP_MENU_BASE (gobject);
    CdosPopupMenuBasePrivate * priv = self->priv;

    if(priv->child_menus != NULL)
    {
        GList *children_menus = g_list_copy (priv->child_menus);
        for (GList *iter = children_menus; iter; iter = iter->next)
        {
            if (iter->data)
            {
                cdos_popup_menu_base_remove_child_menu (self, iter->data);
            }
        }
        g_list_free (children_menus);
        priv->child_menus = NULL;
    }
    if (priv->parent_menu)
    {
        cdos_popup_menu_base_remove_child_menu (priv->parent_menu, self);
        priv->parent_menu = NULL;
    }
    G_OBJECT_CLASS (cdos_popup_menu_base_parent_class)->dispose (gobject);
}

static void
cdos_popup_menu_base_finalize (GObject *gobject)
{
    G_OBJECT_CLASS (cdos_popup_menu_base_parent_class)->finalize (gobject);
}

static void
cdos_popup_menu_base_set_property (GObject      *gobject,
                                   guint         prop_id,
                                   const GValue *value,
                                   GParamSpec   *pspec)
{
    CdosPopupMenuBase *obj = CDOS_POPUP_MENU_BASE (gobject);

    switch (prop_id)
    {
        case PROP_SOURCE_ACTOR:
            cdos_popup_menu_base_set_source_actor (obj, g_value_get_object (value));
            break;
        case PROP_BOX:
            cdos_popup_menu_base_set_box (obj, g_value_get_object (value));
            break;
        default:
            G_OBJECT_WARN_INVALID_PROPERTY_ID(gobject, prop_id, pspec);
            break;
    }
}

static void
cdos_popup_menu_base_get_property (GObject    *gobject,
                                   guint       prop_id,
                                   GValue     *value,
                                   GParamSpec *pspec)
{
    CdosPopupMenuBasePrivate *priv = CDOS_POPUP_MENU_BASE (gobject)->priv;

    switch (prop_id)
    {
        case PROP_SOURCE_ACTOR:
            g_value_set_object (value, priv->source_actor);
            break;
        case PROP_BOX:
            g_value_set_object (value, priv->box);
            break;
        default:
            G_OBJECT_WARN_INVALID_PROPERTY_ID(gobject, prop_id, pspec);
            break;
    }
}

static ClutterActor *
cdos_popup_menu_base_get_actor (CdosPopupMenuBase *self)
{
    return CLUTTER_ACTOR (self);
}

static void
cdos_popup_menu_base_open (CdosPopupMenuBase *self, gboolean animate)
{
    clutter_actor_show (CLUTTER_ACTOR (self));
}

static void
cdos_popup_menu_base_close (CdosPopupMenuBase *self, gboolean animate)
{
    clutter_actor_hide (CLUTTER_ACTOR (self));
}

static void
cdos_popup_menu_base_class_init (CdosPopupMenuBaseClass * klass)
{
    GObjectClass *gobject_class = G_OBJECT_CLASS (klass);
    GParamSpec *pspec;

    gobject_class->finalize = cdos_popup_menu_base_finalize;
    gobject_class->dispose = cdos_popup_menu_base_dispose;
    gobject_class->set_property = cdos_popup_menu_base_set_property;
    gobject_class->get_property = cdos_popup_menu_base_get_property;

    klass->get_actor = cdos_popup_menu_base_get_actor;
    klass->open = cdos_popup_menu_base_open;
    klass->close = cdos_popup_menu_base_close;

    pspec = g_param_spec_object ("source-actor",
                                 "SourceActor",
                                 "SourceActor of the object",
                                 ST_TYPE_WIDGET,
                                 G_PARAM_READWRITE);
    g_object_class_install_property (gobject_class, PROP_SOURCE_ACTOR, pspec);

    cdos_popup_menu_base_signals[ACTIVATE] = g_signal_new ("activate",
                                                            CDOS_TYPE_POPUP_MENU_BASE,
                                                            G_SIGNAL_RUN_LAST,
                                                            0,
                                                            NULL,
                                                            NULL,
                                                            _cdos_marshal_VOID__POINTER_BOOLEAN,
                                                            G_TYPE_NONE,
                                                            2,
                                                            G_TYPE_POINTER,
                                                            G_TYPE_BOOLEAN);

    cdos_popup_menu_base_signals[ACTIVE_CHANGED] = g_signal_new ("active-changed",
                                                                  G_TYPE_FROM_CLASS(klass),
                                                                  G_SIGNAL_RUN_LAST,
                                                                  G_STRUCT_OFFSET(CdosPopupMenuBaseClass, active_changed),
                                                                  NULL,
                                                                  NULL,
                                                                  g_cclosure_marshal_VOID__OBJECT,
                                                                  G_TYPE_NONE,
                                                                  1,
                                                                  G_TYPE_OBJECT);

    cdos_popup_menu_base_signals[CHILD_MENU_ADDED] = g_signal_new ("child-menu-added",
                                                                   G_TYPE_FROM_CLASS(klass),
                                                                   G_SIGNAL_RUN_LAST,
                                                                   G_STRUCT_OFFSET(CdosPopupMenuBaseClass, child_menu_added),
                                                                   NULL,
                                                                   NULL,
                                                                   g_cclosure_marshal_VOID__OBJECT,
                                                                   G_TYPE_NONE,
                                                                   1,
                                                                   G_TYPE_OBJECT);

    cdos_popup_menu_base_signals[CHILD_MENU_REMOVED] = g_signal_new ("child-menu-removed",
                                                                     G_TYPE_FROM_CLASS(klass),
                                                                     G_SIGNAL_RUN_LAST,
                                                                     G_STRUCT_OFFSET(CdosPopupMenuBaseClass, child_menu_removed),
                                                                     NULL,
                                                                     NULL,
                                                                     g_cclosure_marshal_VOID__OBJECT,
                                                                     G_TYPE_NONE,
                                                                     1,
                                                                     G_TYPE_OBJECT);

    cdos_popup_menu_base_signals[MENU_ITEM_ADD] = g_signal_new ("menu-item-add",
                                                                G_TYPE_FROM_CLASS(klass),
                                                                G_SIGNAL_RUN_LAST,
                                                                G_STRUCT_OFFSET(CdosPopupMenuBaseClass, menu_item_add),
                                                                NULL,
                                                                NULL,
                                                                g_cclosure_marshal_VOID__VOID,
                                                                G_TYPE_NONE,
                                                                0);

    cdos_popup_menu_base_signals[OPEN_STATE_CHANGED] = g_signal_new ("open-state-changed",
                                                                     G_TYPE_FROM_CLASS(klass),
                                                                     G_SIGNAL_RUN_LAST,
                                                                     G_STRUCT_OFFSET(CdosPopupMenuBaseClass, open_state_changed),
                                                                     NULL,
                                                                     NULL,
                                                                     g_cclosure_marshal_VOID__BOOLEAN,
                                                                     G_TYPE_NONE,
                                                                     1,
                                                                     G_TYPE_BOOLEAN);
}

static void
cdos_popup_menu_base_init (CdosPopupMenuBase * self)
{
    CdosPopupMenuBasePrivate * priv;
    StBoxLayout *box;

    priv = self->priv = cdos_popup_menu_base_get_instance_private (self);

    box = ST_BOX_LAYOUT (st_box_layout_new ());
    priv->box = box;
    self->box = box;
    clutter_actor_set_x_expand (CLUTTER_ACTOR (box), TRUE);
    //clutter_actor_set_y_expand (CLUTTER_ACTOR (box), TRUE);
    st_box_layout_set_vertical (priv->box, TRUE);
    st_widget_set_style (ST_WIDGET (priv->box), "overflow:hidden;");

    clutter_actor_add_child (CLUTTER_ACTOR (self), CLUTTER_ACTOR (priv->box));

    priv->active_menu_item = NULL;
    self->length = 0;
    self->is_open = FALSE;
    self->block_source_events = FALSE;
    self->pass_events = FALSE;
    priv->child_menus = NULL;
}

/**
 * cdos_popup_menu_base_set_source_actor:
 * @self: A #CdosPopupMenuBase
 * @actor: 设置的source_actor
 *
 * 设置self的source_actor
 *
 */
void
cdos_popup_menu_base_set_source_actor (CdosPopupMenuBase *self, ClutterActor *actor)
{
    CdosPopupMenuBasePrivate *priv;
    g_return_if_fail (CDOS_IS_POPUP_MENU_BASE (self));
    g_return_if_fail (CLUTTER_IS_ACTOR (actor));

    priv = self->priv;

    if (actor)
    {
        priv->source_actor = actor;
        if (ST_IS_WIDGET (actor))
            st_widget_add_style_class_name ((StWidget *)actor, "source-actor");
    }
    else
        priv->source_actor = NULL;
}

/**
 * cdos_popup_menu_base_get_source_actor:
 * @self: A #CdosPopupMenuBase
 *
 * 获取self的source_actor
 *
 * Returns: ＃StWidget
 */
ClutterActor *
cdos_popup_menu_base_get_source_actor (CdosPopupMenuBase *self)
{
  g_return_val_if_fail(CDOS_IS_POPUP_MENU_BASE(self), NULL);
  return self->priv->source_actor;
}

/**
 * cdos_popup_menu_base_set_box:
 * @self: A #CdosPopupMenuBase
 * @box: 需要设置的box
 *
 * 设置self的box
 *
 */
static void
cdos_popup_menu_base_set_box (CdosPopupMenuBase *self, StBoxLayout *box)
{
  CdosPopupMenuBasePrivate *priv;
  g_return_if_fail(CDOS_IS_POPUP_MENU_BASE(self));
  priv = self->priv;

  if(priv->box){
    g_object_unref(priv->box);
  }
  if (box)
  {
    priv->box = box;
    self->box = box;
  }
  else
  {
    priv->box = NULL;
    self->box = NULL;
  }
}

/**
 * cdos_popup_menu_base_get_box:
 * @self: A #CdosPopupMenuBase
 *
 * 获取self的box成员
 *
 * Returns: #StBoxLayout
 */
StBoxLayout *
cdos_popup_menu_base_get_box(CdosPopupMenuBase *self)
{
  g_return_val_if_fail(CDOS_IS_POPUP_MENU_BASE(self), NULL);
  return self->priv->box;
}

/**
 * cdos_popup_menu_base_new:
 *
 * 新建一个 #CdosPopupMenuBase 类的实例。
 *
 * Returns: 返回一个 #CdosPopupMenuBase 的实例。
 */
CdosPopupMenuBase *
cdos_popup_menu_base_new (void)
{
  CdosPopupMenuBase *self;
  self = g_object_new (CDOS_TYPE_POPUP_MENU_BASE, NULL);
  return self;
}

static void
sub_menu_activate_callback (CdosPopupMenuBase *menu,
                            ClutterEvent      *event,
                            gboolean           keep_menu,
                            CdosPopupMenuBase *self)
{
  CdosPopupMenuBaseClass *obj_class = CDOS_POPUP_MENU_BASE_GET_CLASS(self);
  g_signal_emit (self, cdos_popup_menu_base_signals[ACTIVATE], 0, event, keep_menu);

  if (!keep_menu)
    obj_class->close(self, TRUE);
}

static void
sub_menu_active_change_callback (CdosPopupSubMenu  *menu,
                                   CdosPopupBaseMenuItem *menu_item,
                                   CdosPopupMenuBase *self)
{
  CdosPopupMenuBasePrivate *priv = self->priv;

  if (priv->active_menu_item != menu_item)
    {
      if (priv->active_menu_item)
              cdos_popup_base_menu_item_set_active (priv->active_menu_item, FALSE);
      priv->active_menu_item = menu_item;
      g_signal_emit_by_name (self, "active-changed", menu_item);
    }
}

static void
cdos_popup_menu_base_connect_sub_menu_signals (CdosPopupMenuBase     *self,
                                               CdosPopupBaseMenuItem *object,
                                               CdosPopupSubMenu *menu)
{
  gulong sub_menu_activate_id, sub_menu_activate_change_id;
  sub_menu_activate_id = g_signal_connect(
      menu, "activate", G_CALLBACK(sub_menu_activate_callback), self);

  sub_menu_activate_change_id = g_signal_connect(
                        menu, "active-changed",
                        G_CALLBACK(sub_menu_active_change_callback),
                        self);

  g_object_set_data(G_OBJECT(object),"sub_menu_activate_id", GSIZE_TO_POINTER (sub_menu_activate_id));
  g_object_set_data(G_OBJECT(object), "sub_menu_activate_change_id", GSIZE_TO_POINTER (sub_menu_activate_change_id));
}

static void
menu_item_active_changed_callback (CdosPopupBaseMenuItem          *menu_item,
                                   gboolean           active,
                                   CdosPopupMenuBase *self)
{
  CdosPopupMenuBasePrivate *priv = self->priv;

  if (active && priv->active_menu_item != menu_item)
    {
      if (priv->active_menu_item)
              cdos_popup_base_menu_item_set_active (priv->active_menu_item, FALSE);
      priv->active_menu_item = menu_item;
      g_signal_emit_by_name (self, "active-changed", menu_item);
    }
  else if (!active && priv->active_menu_item == menu_item)
    {
      priv->active_menu_item = NULL;
      g_signal_emit_by_name (self, "active-changed", NULL);
    }
}

static void
menu_item_sensitive_change_callback (CdosPopupBaseMenuItem *menu_item,
                                     gboolean sensitive,
                                     CdosPopupMenuBase *self)
{
    CdosPopupMenuBasePrivate *priv = self->priv;
    CdosGlobal *global = cdos_global_get ();
    ClutterStage *stage = cdos_global_get_stage (global);
    if (!sensitive && priv->active_menu_item == menu_item)
    {
        if (!st_widget_navigate_focus (ST_WIDGET (self), CLUTTER_ACTOR (menu_item),
                                       GTK_DIR_TAB_FORWARD,
                                       TRUE))
        {
            clutter_actor_grab_key_focus (CLUTTER_ACTOR (self));
        }
    }
    else if (sensitive && priv->active_menu_item == NULL)
    {
        if (clutter_stage_get_key_focus (stage) == (ClutterActor *)self)
            clutter_actor_grab_key_focus (CLUTTER_ACTOR (menu_item));
    }
}

static void
menu_item_destroy_callback (CdosPopupBaseMenuItem *menu_item, CdosPopupMenuBase *self)
{
    CdosPopupMenuBasePrivate *priv = self->priv;
  //原JS判断 menu_item->menu
  /*
  if (CDOS_IS_POPUP_SUBMENU_MENU_ITEM(menu_item)) {
      gulong sub_menu_activate_id,
             sub_menu_activate_change_id,
             closing_id;
      sub_menu_activate_id = GPOINTER_TO_SIZE (g_object_get_data(G_OBJECT(menu_item), "sub_menu_activate_id"));
      sub_menu_activate_change_id = GPOINTER_TO_SIZE (g_object_get_data(G_OBJECT(menu_item), "sub_menu_activate_change_id"));
      closing_id = GPOINTER_TO_SIZE (g_object_get_data(G_OBJECT(menu_item), "closing_id"));
      CdosPopupSubMenu *submenu;
      submenu = cdos_popup_submenu_menu_item_get_menu ((CdosPopupSubmenuMenuItem *)menu_item);
      if(sub_menu_activate_id){
          g_signal_handler_disconnect (submenu, sub_menu_activate_id);
      }
      if(sub_menu_activate_change_id){
          g_signal_handler_disconnect (submenu, sub_menu_activate_change_id);
      }
      if(closing_id){
          g_signal_handler_disconnect (self, closing_id);
      }
  }
  */
  //gulong item_activate = GPOINTER_TO_SIZE (g_object_get_data ((GObject *)menu_item, "activate_id"));
  //g_signal_handler_disconnect (menu_item, item_activate);
    if (menu_item == priv->active_menu_item)
        priv->active_menu_item = NULL;
}

static void
menu_item_activate_callback (CdosPopupBaseMenuItem *menu_item,
                             ClutterEvent *event,
                             gboolean keep_menu,
                             CdosPopupMenuBase *self)
{
    if (!CDOS_IS_POPUP_MENU_BASE (self))
    {
        g_warn_if_reached ();
        return;
    }

    CdosPopupMenuBaseClass *obj_class = CDOS_POPUP_MENU_BASE_GET_CLASS(self);
    g_signal_emit (self, cdos_popup_menu_base_signals[ACTIVATE], 0, event, keep_menu);

    if(!keep_menu)
    {
        obj_class->close(self,TRUE);
    }
}

static void
cdos_popup_menu_base_connect_item_signals (CdosPopupMenuBase     *self,
                                           CdosPopupBaseMenuItem *menu_item)
{
  g_signal_connect(menu_item,
                "active-changed",
                G_CALLBACK(menu_item_active_changed_callback),
                self);

  g_signal_connect(menu_item,
                  "sensitive-changed",
                  G_CALLBACK(menu_item_sensitive_change_callback),
                  self);

  g_signal_connect(menu_item,
                   "destroy",
                   G_CALLBACK(menu_item_destroy_callback),
                   self);

  //gulong item_activate;
  g_signal_connect(menu_item,"activate",G_CALLBACK(menu_item_activate_callback),self);
  //g_object_set_data ((GObject *)menu_item, "activate_id", GSIZE_TO_POINTER (item_activate));
}

static void
cdos_popup_menu_base_update_separator_visibility (CdosPopupMenuBase *self,
                                                  CdosPopupBaseMenuItem *menu_item)
{
    ClutterActor *actor = (ClutterActor *)menu_item;
    ClutterActor *before, *after;

    before = clutter_actor_get_previous_sibling (actor);

    while (before && !clutter_actor_is_visible (before))
    {
        before = clutter_actor_get_previous_sibling (before);
    }
    if (before == NULL ||
        CDOS_IS_POPUP_SEPARATOR_MENU_ITEM (before))
    {
        clutter_actor_hide (actor);
        return;
    }

    after = clutter_actor_get_next_sibling (actor);
    while (after && !clutter_actor_is_visible(after))
        after = clutter_actor_get_next_sibling (after);

    if (after == NULL ||
        CDOS_IS_POPUP_SEPARATOR_MENU_ITEM (after))
    {
        clutter_actor_hide (actor);
        return;
    }
    clutter_actor_show (actor);
}

/**
 * cdos_popup_menu_base_get_menu_items:
 * @self: A #CdosPopupMenuBase
 *
 * 获取self的所有子控件
 *
 * Returns: ＃GList
 */
GList *
cdos_popup_menu_base_get_menu_items (CdosPopupMenuBase *self)
{
    g_return_val_if_fail (CDOS_IS_POPUP_MENU_BASE (self), NULL);
    CdosPopupMenuBasePrivate *priv = self->priv;

    ClutterActor *child = clutter_actor_get_first_child (CLUTTER_ACTOR (priv->box));

    GList *result = NULL;
    while (child)
    {
        if (CDOS_IS_POPUP_BASE_MENU_ITEM (child) || CDOS_IS_POPUP_MENU_SECTION (child))
            result = g_list_prepend (result, child);
        child = clutter_actor_get_next_sibling (child);
    }
    result = g_list_reverse (result);
    return result;
}

/**
 * cdos_popup_menu_base_add_action:
 * @self: A #CdosPopupMenuBase
 * @title: 添加的menuitem显示的text
　* @callback: 点击时回调的函数
　* @data: 传递给回调函数的数据
 *
 * 为CdosPopupMenuBase添加一个可以点击的item，我们可以指定显示item的信息，还有
 *　就是点击之后响应的动作
 *
 * Returns: $CdosPopupMenuItem
 */
CdosPopupMenuItem *
cdos_popup_menu_base_add_action (CdosPopupMenuBase *self,
                                 const gchar *title,
                                 GCallback callback,
                                 gpointer user_data)
{
    g_return_val_if_fail (CDOS_IS_POPUP_MENU_BASE (self), NULL);

    CdosPopupMenuItem * menu_item;
    menu_item = cdos_popup_menu_item_new ();
    cdos_popup_menu_item_set_label (menu_item, title);
    g_signal_connect ((CdosPopupBaseMenuItem *)menu_item, "activate", G_CALLBACK(callback),user_data);
    cdos_popup_menu_base_add_menu_item (self, (CdosPopupBaseMenuItem *)menu_item, -1);
    return menu_item;
}

static void
add_commandline_action_callback (CdosPopupBaseMenuItem *item,
                                 ClutterEvent *e,
                                 gboolean k,
                                 const gchar *cmd)
{
  gint *argcp = 0;
  gchar ***argvp = NULL;
  GError **error = NULL;
  g_shell_parse_argv (cmd, argcp, argvp, error);
  cdos_util_spawn_command_line (cmd);
}

/**
 * cdos_popup_menu_base_add_settings_action:
 * @self: A #CdosPopupMenuBase
 * @title: 添加的menuitem显示的text
　* @cmd:　你期望执行的命令
 *
 * 当我们希望为CdosPopupMenuBase中添加一个item,并且这个item点击的目的
 *　是执行一个命令时，这个函数就能满足你的需求
 *
 * Returns: $CdosPopupMenuItem
 */
CdosPopupMenuItem *
cdos_popup_menu_base_add_commandline_action (CdosPopupMenuBase *self,
                                             const gchar *title,
                                             const gchar *cmd)
{
    CdosPopupMenuItem *item;
    gchar *command_line;
    g_return_val_if_fail(CDOS_IS_POPUP_MENU_BASE(self), NULL);

    command_line = g_strdup (cmd);
    item =  cdos_popup_menu_base_add_action (self, title, G_CALLBACK(add_commandline_action_callback), command_line);
    g_object_set_data_full ((GObject *)item, "command-line", command_line, g_free);
    return item;
}

/**
 * cdos_popup_menu_base_is_child_menu:
 * @self: A #CdosPopupMenuBase
 * @menu: 需要判断是否是子控件的CdosPopupMenuBase
 *
 * 判断一个CdosPopupMenuBase是否是另一个CdosPopupMenuBase的子控件
 *
 * Returns: gboolean 　True 是子控件，False 则不是
 */
gboolean
cdos_popup_menu_base_is_child_menu (CdosPopupMenuBase *self, CdosPopupMenuBase *menu)
{
    g_return_val_if_fail(CDOS_IS_POPUP_MENU_BASE(self), FALSE);
    CdosPopupMenuBasePrivate *priv = self->priv;
    return (g_list_index (priv->child_menus, menu) != -1);
}

/**
 * cdos_popup_menu_base_remove_child_menu:
 * @self: A #CdosPopupMenuBase
 * @menu: 需要添加到父控件CdosPopupMenuBase中的menu
 *
 * 将一个CdosPopupMenuBase添加到另一个CdosPopupMenuBase中
 *
 */
void
cdos_popup_menu_base_add_child_menu (CdosPopupMenuBase *self, CdosPopupMenuBase *menu)
{
    g_return_if_fail (CDOS_IS_POPUP_MENU_BASE (self));
    g_return_if_fail (CDOS_IS_POPUP_MENU_BASE (menu));

    g_return_if_fail (!menu->priv->parent_menu);

    CdosPopupMenuBasePrivate *priv = self->priv;
    if (cdos_popup_menu_base_is_child_menu (self, menu))
        return;
    priv->child_menus = g_list_append (priv->child_menus, g_object_ref (menu));
    menu->priv->parent_menu = self;

    g_signal_emit (self, cdos_popup_menu_base_signals[CHILD_MENU_ADDED], 0, menu);
}

/**
 * cdos_popup_menu_base_remove_child_menu:
 * @self: A #CdosPopupMenuBase
 * @menu: 需要从 父控件CdosPopupMenuBase中移除的menu
 *
 * 将一个CdosPopupMenuBase控件从其父控件CdosPopupMenuBase中移除
 *
 */
void
cdos_popup_menu_base_remove_child_menu (CdosPopupMenuBase *self, CdosPopupMenuBase *menu)
{
    g_return_if_fail (CDOS_IS_POPUP_MENU_BASE (self));
    g_return_if_fail (CDOS_IS_POPUP_MENU_BASE (menu));
    g_return_if_fail (menu->priv->parent_menu == self);
    menu->priv->parent_menu = NULL;

    CdosPopupMenuBasePrivate *priv = self->priv;
    priv->child_menus = g_list_remove (priv->child_menus, menu);
    g_signal_emit (self, cdos_popup_menu_base_signals[CHILD_MENU_REMOVED], 0, menu);
    g_object_unref(menu);
}

/*
 * g_signal_connect callbacks .
 *
 */
static void
menu_section_destroy_callback (CdosPopupBaseMenuItem *menu_item, CdosPopupMenuBase *self)
{
    gulong *sub_menu_activate_id, *sub_menu_activate_change_id;
    sub_menu_activate_id = g_object_get_data(G_OBJECT(menu_item), "sub_menu_activate_id");
    sub_menu_activate_change_id = g_object_get_data(G_OBJECT(menu_item), "sub_menu_activate_change_id");
    if(sub_menu_activate_id)
        g_signal_handler_disconnect (menu_item, *sub_menu_activate_id);
    if(sub_menu_activate_change_id)
        g_signal_handler_disconnect (menu_item, *sub_menu_activate_change_id);

    self->length--;
}

static void
sub_menu_menu_item_open_state_changed_callback (CdosPopupMenuBase *self,
                                                gboolean open)
{
    CdosPopupMenuBaseClass *obj_class = CDOS_POPUP_MENU_BASE_GET_CLASS(self);
    if (open)
        obj_class->close(self, FALSE);
}

static void
separator_menu_item_open_state_changed_callback (CdosPopupMenuBase     *self,
                                                 gboolean               open,
                                                 CdosPopupBaseMenuItem *menu_item)
{
    if (open)
        cdos_popup_menu_base_update_separator_visibility (self, menu_item);
}

static void
on_separator_destroy (CdosPopupBaseMenuItem *item,
                      CdosPopupMenuBase  *self)
{
    g_signal_handlers_disconnect_by_data (self, item);
}

/**
 * cdos_popup_menu_base_add_menu_item:
 * @self: A #CdosPopupMenuBase
 * @menu_item: 要添加到menu中的item #CdosPopupBaseMenuItem
 * @position: 添加的位置，-1为默认不指定
 *
 * 将CdosPopupBaseMenuItem添加到CdosPopupMenuBase中，并且可以指定添加的位置
 *
 */
void
cdos_popup_menu_base_add_menu_item (CdosPopupMenuBase     *self,
                                    CdosPopupBaseMenuItem *menu_item,
                                    gint position)
{
    g_return_if_fail (CDOS_IS_POPUP_MENU_BASE (self));
    g_return_if_fail (CDOS_IS_POPUP_BASE_MENU_ITEM (menu_item));

    CdosPopupMenuBasePrivate *priv = self->priv;
    gulong closing_id;
    ClutterActor *actor;

    actor = (ClutterActor *)menu_item;
    if (position == -1)
    {
        clutter_actor_add_child ((ClutterActor *)priv->box, actor);

        //clutter_actor_show(actor);
    }
    else
    {
        //cdos_popup_menu_base_get_menu_items 返回glist
        // 其中包括两种类型，CDOS_POPUP_MENU_SECTION 和 CDOS_IS_POPUP_BASE_MENU_ITEM
        //items = cdos_popup_menu_base_get_menu_items(self);
        if(CDOS_IS_POPUP_BASE_MENU_ITEM(menu_item))
        {
            clutter_actor_insert_child_at_index ((ClutterActor *)priv->box, actor, position);
        }
    }
    g_signal_emit (self, cdos_popup_menu_base_signals[MENU_ITEM_ADD], 0);

    if (CDOS_IS_POPUP_MENU_SECTION (menu_item))
    {
        // FIXME
        //cdos_popup_menu_base_connect_sub_menu_signals (self, menu_item, (CdosPopupBaseMenuItem *)menu_item);
        g_signal_connect (menu_item, "destroy",
                          G_CALLBACK(menu_section_destroy_callback), self);

    }
    else if (CDOS_IS_POPUP_SUBMENU_MENU_ITEM (menu_item))
    {
        //g_object_get(menu_item, "menu", &menu_item_menu, NULL);
        CdosPopupSubMenu *menu = cdos_popup_submenu_menu_item_get_menu ((CdosPopupSubmenuMenuItem *)menu_item);
        ClutterActor *menu_actor = CLUTTER_ACTOR (menu);
        if (position == -1)
            clutter_actor_add_child ((ClutterActor *)priv->box, menu_actor);
        else
            clutter_actor_insert_child_below ((ClutterActor *)priv->box, menu_actor, actor);
        cdos_popup_menu_base_connect_sub_menu_signals (self, menu_item, menu);
        cdos_popup_menu_base_connect_item_signals (self, menu_item);

        closing_id = g_signal_connect (self, "open-state-changed", G_CALLBACK(sub_menu_menu_item_open_state_changed_callback), NULL);

        g_object_set_data (G_OBJECT(menu_item), "closing-id", GSIZE_TO_POINTER (closing_id));
    }
    else if (CDOS_IS_POPUP_SEPARATOR_MENU_ITEM (menu_item))
    {
        cdos_popup_menu_base_connect_item_signals (self, menu_item);
        // updateSeparatorVisibility needs to get called any time the
        // separator's adjacent siblings change visibility or position.
        // open-state-changed isn't exactly that, but doing it in more
        // precise ways would require a lot more bookkeeping.
        g_signal_connect (self, "open-state-changed",
                          G_CALLBACK(separator_menu_item_open_state_changed_callback), menu_item);
        g_signal_connect (menu_item, "destroy", G_CALLBACK (on_separator_destroy), self);
    }
    else if (CDOS_IS_POPUP_BASE_MENU_ITEM (menu_item))
    {
        cdos_popup_menu_base_connect_item_signals (self, menu_item);
    }
    self->length++;
}

/**
 * cdos_popup_menu_base_add_actor:
 * @self: A #CdosPopupMenuBase
 * @actor: 需要添加的ClutterActor
 *
 * 将一个ClutterActor添加到CdosPopupMenuBase对象的box中
 *
 */
void
cdos_popup_menu_base_add_actor (CdosPopupMenuBase *self, ClutterActor *actor)
{
    g_return_if_fail (CDOS_IS_POPUP_MENU_BASE (self));
    g_return_if_fail (CLUTTER_IS_ACTOR(actor));
    clutter_actor_add_child ((ClutterActor *)self->priv->box, actor);
}

/**
 * cdos_popup_menu_base_remove_all:
 * @self: A #CdosPopupMenuBase
 *
 * 移除CdosPopupMenuBase中所有的控件
 *
 */
void
cdos_popup_menu_base_remove_all (CdosPopupMenuBase *self)
{
    g_return_if_fail (NULL != self && CDOS_IS_POPUP_MENU_BASE (self));
    CdosPopupMenuBasePrivate *priv = self->priv;
    clutter_actor_destroy_all_children (CLUTTER_ACTOR (priv->box));
}

/**
 * cdos_popup_menu_base_toggle:
 * @self: A #CdosPopupMenuBase
 *
 * 改变一个CdosPopupMenuBase的打开和关闭状态
 * 如果CdosPopupMenuBase是打开的，函数则会将CdosPopupMenuBase关闭
 *　反之亦然
 */
void
cdos_popup_menu_base_toggle (CdosPopupMenuBase *self)
{
    CdosPopupMenuBaseClass *obj_class = CDOS_POPUP_MENU_BASE_GET_CLASS (self);
    if (self->is_open)
        obj_class->close (self, TRUE);
    else
        obj_class->open (self, TRUE);
}

/**
 * cdos_popup_menu_base_toggle:
 * @self: A #CdosPopupMenuBase
 * @animate: ??
 * @onComplete:  ??
 *
 * 改变一个CdosPopupMenuBase的打开和关闭状态
 * 如果CdosPopupMenuBase是打开的，函数则会将CdosPopupMenuBase关闭
 *　反之亦然
 */
void
cdos_popup_menu_base_toggle_with_options (CdosPopupMenuBase *self,
                                          gboolean animate,
                                          void *on_complete)
{
    CdosPopupMenuBaseClass *obj_class = CDOS_POPUP_MENU_BASE_GET_CLASS(self);
    if (self->is_open)
        obj_class->close (self, animate);
    else
        obj_class->open (self, animate);
}

/**
 * cdos_popup_menu_base_destroy:
 * @self: A #CdosPopupMenuBase
 *
 * 销毁CdosPopupMenuBase
 *
 */
void
cdos_popup_menu_base_destroy (CdosPopupMenuBase *self)
{
    clutter_actor_destroy ((ClutterActor *)self);
}

void
cdos_popup_menu_base_remove_menu_item (CdosPopupMenuBase *self, CdosPopupBaseMenuItem *item)
{
    g_return_if_fail (CDOS_IS_POPUP_MENU_BASE (self));
    g_return_if_fail (CDOS_IS_POPUP_BASE_MENU_ITEM (item));

    CdosPopupMenuBasePrivate *priv = self->priv;

    g_object_set_data (G_OBJECT(item), "closing-id", NULL);
    ClutterActor *actor = (ClutterActor *)item;
    clutter_actor_remove_child ((ClutterActor *)priv->box, actor);

    self->length--;
}
