#include "cdos-popup-menu.h"
#include "cdos-popup-base-menu-item.h"
#include "cdos-layout-manager.h"
#include "cdos-start.h"

#include <sys/time.h>


/**
 * SECTION: cdos-popup-menu.c
 * @Title: CdosPopupMenu
 * @short_description: 继承自CdosPopupMenuBase, 可以创建一个menu实例。
 *　 你可以往menu里添加menuitem.
 *
 * CdosPopupMenu 实现一个菜单功能
 */
struct _CdosPopupMenuPrivate
{
  gfloat               arrow_alignment;
  gint                 arrow_side;
  CdosBoxPointer      *box_pointer;
  gboolean             is_popup;
};

G_DEFINE_TYPE_WITH_PRIVATE (CdosPopupMenu, cdos_popup_menu, CDOS_TYPE_POPUP_MENU_BASE);

enum
{
  PROP_0,

  PROP_ARROW_ALIGNMENT,
};

static void
cdos_popup_menu_finalize (GObject *gobject)
{
    G_OBJECT_CLASS (cdos_popup_menu_parent_class )->finalize (gobject);
}

static void
cdos_popup_menu_set_property (GObject      *gobject,
                              guint         prop_id,
                              const GValue *value,
                              GParamSpec   *pspec)
{
    CdosPopupMenu * obj = CDOS_POPUP_MENU (gobject);
    switch (prop_id)
    {
    case PROP_ARROW_ALIGNMENT:
          cdos_popup_menu_set_arrow_alignment (obj, g_value_get_float (value));
          break;
    default:
        G_OBJECT_WARN_INVALID_PROPERTY_ID (gobject, prop_id, pspec);
        break;
    }
}

static void
cdos_popup_menu_get_property (GObject    *gobject,
                              guint       prop_id,
                              GValue     *value,
                              GParamSpec *pspec )
{
    CdosPopupMenuPrivate * priv = CDOS_POPUP_MENU (gobject)->priv;
    switch (prop_id)
    {
    case PROP_ARROW_ALIGNMENT:
          g_value_set_float (value, priv->arrow_alignment);
         break;
    default:
        G_OBJECT_WARN_INVALID_PROPERTY_ID (gobject, prop_id, pspec);
        break;
    }
}

static void
cdos_popup_menu_constructed (GObject *gobject)
{
    G_OBJECT_CLASS (cdos_popup_menu_parent_class)->constructed (gobject);

    ClutterActor *parent;
    StBoxLayout *box;
    CdosPopupMenu *self = CDOS_POPUP_MENU (gobject);
    CdosPopupMenuPrivate *priv = self->priv;

    /* FIXME: 绘制border会导致菜单出现随机颜色的背景,所以暂时隐藏边框 */
    if (priv->is_popup)
        clutter_actor_hide (CLUTTER_ACTOR (priv->box_pointer->border));

    box = cdos_popup_menu_base_get_box (CDOS_POPUP_MENU_BASE (self));
    parent = clutter_actor_get_parent ((ClutterActor *)box);
    g_object_ref (box);
    if (parent)
        clutter_actor_remove_child (parent, (ClutterActor *)box);

    st_bin_set_child (priv->box_pointer->bin, (ClutterActor *)box);
    g_object_unref (box);
}

static gboolean
cdos_popup_menu_key_press_event (ClutterActor *actor,
                                 ClutterKeyEvent *kevent)
{
    GtkDirectionType dir;
    gboolean is_nav_key = TRUE;

    ClutterEvent *event = (ClutterEvent *)kevent;
    guint key = clutter_event_get_key_symbol (event);
    ClutterModifierType mod = clutter_event_get_state (event);

    mod &= CLUTTER_MODIFIER_MASK;
    mod &= ~CLUTTER_LOCK_MASK;

    if (mod)
        return FALSE;

    switch (key)
    {
        case CLUTTER_KEY_Up:
            dir = GTK_DIR_UP;
            break;
        case CLUTTER_KEY_Down:
            dir = GTK_DIR_DOWN;
            break;
        case CLUTTER_KEY_Left:
            dir = GTK_DIR_LEFT;
            break;
        case CLUTTER_KEY_Right:
            dir = GTK_DIR_RIGHT;
            break;
        default:
            is_nav_key = FALSE;
            break;
    }

    if (is_nav_key)
    {
        ClutterActor *focus;
        ClutterStage *stage = cdos_global_get_stage (global);
        focus = clutter_stage_get_key_focus (stage);
        if (focus == actor)
        {
            st_widget_navigate_focus ((StWidget *)actor, NULL, dir, FALSE);
            return TRUE;
        }
    }
    if (key == CLUTTER_KEY_Escape)
    {
        cdos_popup_menu_close ((CdosPopupMenuBase *)actor, TRUE);
        return TRUE;
    }

    return FALSE;
}

static ClutterActor *
cdos_popup_menu_get_actor (CdosPopupMenuBase *self)
{
    return CLUTTER_ACTOR (self);
}

static void
cdos_popup_menu_class_init (CdosPopupMenuClass *klass)
{
    CdosPopupMenuBaseClass *parent_class = CDOS_POPUP_MENU_BASE_CLASS (klass);
    GObjectClass *gobject_class = G_OBJECT_CLASS (klass);
    ClutterActorClass *actor_class = CLUTTER_ACTOR_CLASS (klass);
    GParamSpec *pspec;

    parent_class->open = cdos_popup_menu_open;
    parent_class->close = cdos_popup_menu_close;
    parent_class->get_actor = cdos_popup_menu_get_actor;
    gobject_class->finalize = cdos_popup_menu_finalize;
    gobject_class->set_property = cdos_popup_menu_set_property;
    gobject_class->get_property = cdos_popup_menu_get_property;
    gobject_class->constructed = cdos_popup_menu_constructed;

    actor_class->key_press_event = cdos_popup_menu_key_press_event;

    pspec = g_param_spec_float ("arrow-alignment",
                                "ArrowAlignemnt",
                                "ArrowAlignemnt of the object",
                                0.0, 1.0, 0.5, G_PARAM_READWRITE);
    g_object_class_install_property (gobject_class, PROP_ARROW_ALIGNMENT, pspec);
}

static void
cdos_popup_menu_init (CdosPopupMenu *self)
{
    CdosPopupMenuPrivate *priv = self->priv = cdos_popup_menu_get_instance_private (self);

    CdosGlobal *global = cdos_global_get ();
    StFocusManager *focus_manager = cdos_global_get_focus_manager (global);

    priv = self->priv;
    priv->box_pointer = CDOS_BOX_POINTER (self);

    self->actor = (ClutterActor *)self;

    st_widget_set_style_class_name (ST_WIDGET (self->actor), "popup-menu-boxpointer");

    clutter_actor_hide (CLUTTER_ACTOR (self));

    st_widget_add_style_class_name (ST_WIDGET (self->actor), "popup-menu");
    st_focus_manager_add_group (focus_manager, ST_WIDGET (self));
    clutter_actor_set_reactive (CLUTTER_ACTOR (self->actor), TRUE);

    g_object_set ((GObject *)self, "is-popup", TRUE, NULL);
    priv->is_popup = TRUE;
}

/**
 * cdos_popup_menu_get_arrow_alignment:
 * @self: A #CdosPopupMenu
 *
 * 获取self的arrow_alignment属性
 *
 * Returns: ＃gint
 */
gfloat
cdos_popup_menu_get_arrow_alignment (CdosPopupMenu *self)
{
    g_return_val_if_fail (CDOS_IS_POPUP_MENU (self), 0.5);
    return self->priv->arrow_alignment;
}

/**
 * cdos_popup_menu_set_arrow_alignment:
 * @self: A #CdosPopupMenu
 * @alignment: alignment
 *
 * 设置self的arrow_alignment
 *
 */
void
cdos_popup_menu_set_arrow_alignment (CdosPopupMenu *self, gfloat alignment)
{
  CdosPopupMenuPrivate *priv;
  g_return_if_fail(CDOS_IS_POPUP_MENU(self));
  priv = self->priv;
  priv->arrow_alignment = alignment;
}

/**
 * cdos_popup_menu_get_arrow_side:
 * @self: A #CdosPopupMenu
 *
 * 获取self的arrow_side成员
 *
 * Returns: ＃gint
 */
gint
cdos_popup_menu_get_arrow_side (CdosPopupMenu *self)
{
    g_return_val_if_fail (CDOS_IS_POPUP_MENU (self), 0);
    return self->priv->arrow_side;
}

/**
 * cdos_popup_menu_set_arrow_side:
 * @self: A #CdosPopupMenu
 * @side: 方向
 *
 * 设置self的arrow_side的值
 *
 */
void
cdos_popup_menu_set_arrow_side (CdosPopupMenu *self, gint side)
{
    CdosPopupMenuPrivate *priv;
    g_return_if_fail (CDOS_IS_POPUP_MENU (self));
    priv = self->priv;
    priv->arrow_side = side;

    cdos_box_pointer_set_arrow_side (priv->box_pointer, side);
}

/**
 * cdos_popup_menu_new:
 *
 * 新建一个 #CdosPopupMenu 类的实例。
 *
 * Returns: 返回一个 #CdosPopupMenu 的实例。
 */
CdosPopupMenu *
cdos_popup_menu_new (ClutterActor *source)
{
    CdosPopupMenu *self = NULL;
    self = g_object_new (CDOS_TYPE_POPUP_MENU,
                         "source-actor", source,
                         "is-popup", TRUE,
                         NULL);
    return self;
}

/**
 * cdos_popup_menu_new:
 * @source_actor:  #ClutterActor
 * @arrow_alignment: #gfloat
 * @arrow_side:      #gint
 *
 * 新建一个 #CdosPopupMenu 类的实例。
 *
 * Returns: 返回一个 #CdosPopupMenu 的实例。
 */
CdosPopupMenu *
cdos_popup_menu_new_with_params (ClutterActor *source_actor,
                                 gfloat        arrow_alignment,
                                 gint          arrow_side)
{
    CdosPopupMenu *self;
    self = g_object_new (CDOS_TYPE_POPUP_MENU,
                         "source-actor", source_actor,
                         "is-popup", TRUE,
                         "arrow-alignment", arrow_alignment,
                         "arrow-side", arrow_side,
                         NULL);
    return self;
}

/**
 * cdos_popup_menu_set_source_alignment:
 * @self: A #CdosPopupMenu
 * @alignment:　对其方式
 *
 * 设置menu的对其方式
 *
 */
void
cdos_popup_menu_set_source_alignment (CdosPopupMenu *self, gfloat alignment)
{
    g_return_if_fail(CDOS_IS_POPUP_MENU(self));
    CdosPopupMenuPrivate *priv = self->priv;
    cdos_box_pointer_set_source_alignment(priv->box_pointer, alignment);
}


/**
 * cdos_popup_menu_open:
 * @obj: A #CdosPopupMenuBase
 * @animate: 是否显示boxpointer中的actor
 *
 * 将menu显示出来
 *
 */
void
cdos_popup_menu_open (CdosPopupMenuBase *obj, gboolean animate)
{
    g_return_if_fail (CDOS_IS_POPUP_MENU_BASE (obj));
    CdosPopupMenuPrivate *priv = CDOS_POPUP_MENU(obj)->priv;
    if (!obj->is_open)
    {
        CdosLayoutManager *layout_manager = cdos_layout_manager_get_default ();
        cdos_layout_manager_add_chrome (layout_manager, CLUTTER_ACTOR (CDOS_POPUP_MENU(obj)->actor), 0x4);
        cdos_popup_menu_set_max_height (CDOS_POPUP_MENU (obj));
        obj->is_open = TRUE;
        menu_stack_depth += 1;
        gboolean ret = cdos_box_pointer_set_position (priv->box_pointer,
                (ClutterActor *)cdos_popup_menu_base_get_source_actor(obj), priv->arrow_alignment);

        if (ret)
        {
            cdos_box_pointer_show (priv->box_pointer, animate);
            g_signal_emit_by_name (obj, "open-state-changed", TRUE);
        }
    }
}

/**
 * cdos_popup_menu_open_to_x_y:
 * @obj: A #CdosPopupMenuBase
 * @animate: 是否显示boxpointer中的actor
　* @x: 新位置的x坐标
 * @y: 新位置的y坐标
 *
 * 将一个弹出的menu显示在你指定的位置
 *
 */
void
cdos_popup_menu_open_to_x_y (CdosPopupMenuBase *obj,
                             gboolean animate,
                             gfloat x,
                             gfloat y)
{
    g_return_if_fail (CDOS_IS_POPUP_MENU (obj));
    CdosPopupMenuPrivate *priv = CDOS_POPUP_MENU (obj)->priv;
    if(!obj->is_open)
    {
        cdos_popup_menu_set_max_height (CDOS_POPUP_MENU (obj));
        cdos_layout_manager_add_chrome (cdos_layout_manager_get_default (),
	                                    CLUTTER_ACTOR (obj), 4);
        obj->is_open = TRUE;
        menu_stack_depth += 1;

        cdos_box_pointer_set_position_x_y (priv->box_pointer, (ClutterActor *)cdos_popup_menu_base_get_source_actor(obj), priv->arrow_alignment, x, y);
        cdos_box_pointer_show(priv->box_pointer, animate);
        g_signal_emit_by_name (obj, "open-state-changed", TRUE);
    }
}

/**
 * cdos_popup_menu_close:
 * @obj: A #CdosPopupMenuBase
 * @animate: ？？
 *
 * 将popupmenu隐藏起来
 *
 */
void
cdos_popup_menu_close (CdosPopupMenuBase *obj, gboolean animate)
{
    g_return_if_fail(CDOS_IS_POPUP_MENU_BASE(obj));
    CdosPopupMenuPrivate *priv = CDOS_POPUP_MENU(obj)->priv;
    if(obj->is_open)
    {
      obj->is_open = FALSE;
      menu_stack_depth -= 1;
      if(obj->active_menu_item)
        cdos_popup_base_menu_item_set_active(obj->active_menu_item, FALSE);

      cdos_box_pointer_hide (priv->box_pointer, FALSE);

      CdosLayoutManager *layout_manager = cdos_layout_manager_get_default ();
      cdos_layout_manager_remove_chrome (layout_manager,  CLUTTER_ACTOR(CDOS_POPUP_MENU(obj)->actor));
      g_signal_emit_by_name(obj, "open-state-changed", FALSE);
    }
}

/**
 * cdos_popup_menu_set_max_height:
 * @self: A #CdosPopupMenu
 *
 * 设置self的最大的高度
 *
 */
void
cdos_popup_menu_set_max_height (CdosPopupMenu *self)
{
    ClutterActor *source_actor;
    source_actor = cdos_popup_menu_base_get_source_actor ((CdosPopupMenuBase *)self);

    CdosLayoutManager *layout = cdos_layout_manager_get_default ();
    const MetaRectangle *monitor;
    monitor = cdos_layout_manager_find_monitor_for_actor (layout, source_actor);
    gint max_height = -1, max_width = -1;
    if (monitor)
    {
        max_height = monitor->height - 100;
        max_width = monitor->width;
    }

    StThemeNode *theme_node = st_widget_get_theme_node ((StWidget *)self->actor);
    int top    = st_theme_node_get_border_width (theme_node, ST_SIDE_TOP);
    int bottom = st_theme_node_get_border_width (theme_node, ST_SIDE_BOTTOM);
    int left   = st_theme_node_get_border_width (theme_node, ST_SIDE_LEFT);
    int right  = st_theme_node_get_border_width (theme_node, ST_SIDE_RIGHT);

    max_height -= top + bottom;
    max_width -= left + right;

    ClutterActor *stage = clutter_actor_get_stage ((ClutterActor *)self);
    StThemeContext *context = st_theme_context_get_for_stage ((ClutterStage *)stage);
    gint scale_factor = 1;
    g_object_get (context, "scale-factor", &scale_factor, NULL);

    gchar style[255] = {0};
    snprintf(style, sizeof(style) - 1, "max-height:%dpx;max-width:%dpx;",
             max_height / scale_factor,
             max_width / scale_factor);
    st_widget_set_style ((StWidget *)self->actor, style);
}

/**
 * cdos_popup_menu_get_box_pointer:
 * @menu: A #CdosPopupMenu
 *
 * 获取self的box_pointer成员
 *
 * Returns: #CdosBoxPointer
 */
CdosBoxPointer *
cdos_popup_menu_get_box_pointer (CdosPopupMenu *menu)
{
    CdosPopupMenuPrivate *priv = CDOS_POPUP_MENU(menu)->priv;
    // panel 右键菜单需要定位弹出菜单的位置
    return priv->box_pointer;
}

