#include "cdos-popup-sub-menu.h"
#include "cdos-popup-menu.h"

/**
 * SECTION: cdos-popup-sub-menu.c
 * @Title: CdosPopupSubMenu
 * @short_description: 提供一个子菜单控件。
 *　
 *
 * CdosPopupSubMenu 提供一个子菜单的控件
 */


struct _CdosPopupSubMenuPrivate
{
    ClutterActor *source_actor;
    ClutterActor *arrow;
    StScrollView *actor;
    guint rotation_center_z_gravity;
    guint rotation_angle_z;
    gulong close_id;
    gulong open_id;
};

G_DEFINE_TYPE_WITH_PRIVATE (CdosPopupSubMenu, cdos_popup_sub_menu, CDOS_TYPE_POPUP_MENU_BASE);

static void
cdos_popup_sub_menu_dispose (GObject *gobject)
{
    CdosPopupSubMenu *self = CDOS_POPUP_SUB_MENU (gobject);
    CdosPopupSubMenuPrivate * priv = self->priv;

    g_clear_pointer (&priv->arrow, clutter_actor_destroy);

    G_OBJECT_CLASS (cdos_popup_sub_menu_parent_class)->dispose (gobject);
}

static ClutterActor *
cdos_popup_sub_menu_get_actor (CdosPopupMenuBase *obj)
{
    CdosPopupSubMenu *self = CDOS_POPUP_SUB_MENU (obj);
    return CLUTTER_ACTOR (self);
}

static void
cdos_popup_sub_menu_class_init (CdosPopupSubMenuClass *klass)
{
    CdosPopupMenuBaseClass *parent_class = CDOS_POPUP_MENU_BASE_CLASS (klass);
    GObjectClass * gobject_class = G_OBJECT_CLASS (klass);
    parent_class->open = cdos_popup_sub_menu_open;
    parent_class->close = cdos_popup_sub_menu_close;
    parent_class->get_actor=cdos_popup_sub_menu_get_actor;
    //gobject_class->finalize = cdos_popup_sub_menu_finalize;
    gobject_class->dispose = cdos_popup_sub_menu_dispose;
}

static void
cdos_popup_sub_menu_init (CdosPopupSubMenu *self)
{
    self->priv = cdos_popup_sub_menu_get_instance_private (self);
}

static void
_vscroll_start_callback (CdosPopupSubMenu *self)
{
    CdosPopupMenuBase *top_menu = cdos_popup_sub_menu_get_top_menu (self);
    if (top_menu)
        CDOS_POPUP_MENU_BASE (top_menu)->pass_events = TRUE;
}

static void
_vscroll_stop_callback (CdosPopupSubMenu *self)
{
    CdosPopupMenuBase *top_menu = cdos_popup_sub_menu_get_top_menu (self);
    if (top_menu)
        CDOS_POPUP_MENU_BASE (top_menu)->pass_events = FALSE;
}

static gboolean
cdos_popup_sub_menu_on_key_press_event (ClutterActor *actor, ClutterEvent *event, CdosPopupSubMenu *self)
{

    if(CDOS_POPUP_MENU_BASE (self)->is_open && clutter_event_get_key_symbol(event) == CLUTTER_KEY_Left)
    {
        cdos_popup_sub_menu_close (CDOS_POPUP_MENU_BASE (self), TRUE);
        CdosPopupBaseMenuItem *item = (CdosPopupBaseMenuItem *)self->priv->source_actor;
        cdos_popup_base_menu_item_set_active (item, TRUE);
        return TRUE;
    }
    return FALSE;
}

/**
 * cdos_popup_sub_menu_new:
 * @source_actor: source_actor
 * @source_arrow: source_arrow
 *
 * 新建一个 #CdosPopupSubMenu 类的实例。
 *
 * Returns: 返回一个 #CdosPopupSubMenu 的实例。
 */
CdosPopupSubMenu *
cdos_popup_sub_menu_new (ClutterActor *source_actor, ClutterActor *source_arrow)
{
    CdosPopupSubMenu *self;
    CdosPopupSubMenuPrivate *priv;
    ClutterActor *box, *parent;

    self = g_object_new (CDOS_TYPE_POPUP_SUB_MENU, NULL);
    priv = self->priv;

    //g_object_set(CDOS_POPUP_MENU_BASE(self), "source_actor", source_actor, NULL);
    priv->source_actor = source_actor;
    priv->arrow = NULL;
    if(source_arrow)
    {
        priv->arrow = g_object_ref(source_arrow);
        clutter_actor_set_pivot_point (priv->arrow, 0.5f, 0.5f);
    }
    //with long content, we make it scrollable - the scrollbar will only take
    //effect if a CSS max-height is set on the top menu.
    priv->actor = g_object_new (ST_TYPE_SCROLL_VIEW,
                                "style-class", "popup-sub-menu",
                                "hscrollbar-policy", GTK_POLICY_NEVER,
                                "vscrollbar-policy", GTK_POLICY_NEVER, NULL);
    // StScrollbar plays dirty tricks with events, calling
    // clutter_set_motion_events_enabled (FALSE) during the scroll; this
    // confuses our event tracking, so we just turn it off during the
    // scroll.
    ClutterActor *vscroll = st_scroll_view_get_vscroll_bar (priv->actor);
    g_signal_connect (vscroll, "scroll-start", G_CALLBACK(_vscroll_start_callback), self);
    g_signal_connect (vscroll, "scroll-stop", G_CALLBACK(_vscroll_stop_callback), self);

    box = (ClutterActor *)cdos_popup_menu_base_get_box (CDOS_POPUP_MENU_BASE (self));
    parent = clutter_actor_get_parent (box);
    g_object_ref (box);
    if (parent)
        clutter_actor_remove_child (parent, box);

    st_scroll_view_add_child (priv->actor, box);
    g_object_unref (box);

    clutter_actor_add_child (parent, CLUTTER_ACTOR (priv->actor));
    clutter_actor_set_clip_to_allocation (CLUTTER_ACTOR (priv->actor), TRUE);
    g_signal_connect (priv->actor, "key-press-event", G_CALLBACK(cdos_popup_sub_menu_on_key_press_event), self);
    clutter_actor_hide (CLUTTER_ACTOR (self));

    return self;
}

/**
 * cdos_popup_sub_menu_get_top_menu:
 * @self: A #CdosPopupSubMenu
 *
 * 获取CdosPopupMenuBase的顶端menu
 *
 * Returns: #CdosPopupMenuBase
 */
CdosPopupMenuBase *
cdos_popup_sub_menu_get_top_menu (CdosPopupSubMenu *self)
{
    g_return_val_if_fail (CDOS_IS_POPUP_SUB_MENU (self),NULL);
    ClutterActor *actor = clutter_actor_get_parent (CLUTTER_ACTOR (self));
    while (actor != NULL )
    {
        if (CDOS_IS_POPUP_MENU (actor))
          return CDOS_POPUP_MENU_BASE (actor);
        actor = clutter_actor_get_parent (actor);
    }
    return NULL;
}

/**
 * cdos_popup_sub_menu_needs_scrollbar:
 * @self: A #CdosPopupSubMenu
 *
 * 获取CdosPopupMenuBase是否需要scrollbar
 *
 * Returns: #gboolean 是否需要scrollbar 1需要　0不需要
 */
static gboolean
cdos_popup_sub_menu_needs_scrollbar (CdosPopupSubMenu *self)
{
    CdosPopupMenuBase *top_menu = cdos_popup_sub_menu_get_top_menu (self);
    if (!top_menu)
        return FALSE;

    gfloat top_min_height, top_natural_height;
    ClutterActor *actor = CLUTTER_ACTOR (self);
    clutter_actor_get_preferred_height (actor, -1, &top_min_height, &top_natural_height);
    StThemeNode *top_theme_node = st_widget_get_theme_node (ST_WIDGET (actor));
    int top_max_height = st_theme_node_get_max_height (top_theme_node);
    return top_max_height >= 0 && top_natural_height >= top_max_height;
}

/**
 * cdos_popup_sub_menu_open:
 * @obj: A #CdosPopupMenuBase
 * @animate: 是否显示
 *
 *　打开menu
 *
 */
void
cdos_popup_sub_menu_open (CdosPopupMenuBase *obj, gboolean animate)
{
    CdosPopupSubMenu *self = CDOS_POPUP_SUB_MENU (obj);
    g_return_if_fail (CDOS_IS_POPUP_SUB_MENU (self));

    if(obj->is_open)
      return;
    obj->is_open = TRUE;

    CdosPopupSubMenuPrivate *priv;
    priv = self->priv;
    ClutterActor *actor = (ClutterActor *)self;

    gboolean needs_scrollbar = cdos_popup_sub_menu_needs_scrollbar (self);
    GtkPolicyType policy = needs_scrollbar ? GTK_POLICY_AUTOMATIC : GTK_POLICY_NEVER;
    st_scroll_view_set_policy (priv->actor, GTK_POLICY_NEVER, policy);

    if (animate && needs_scrollbar)
        animate = FALSE;

    gdouble rotation_angle = 90;
    if (clutter_actor_get_text_direction ((ClutterActor *)self) == CLUTTER_TEXT_DIRECTION_RTL)
        rotation_angle = 270.0;

    if (self->priv->arrow)
    {
        clutter_actor_save_easing_state (self->priv->arrow);
        clutter_actor_set_easing_mode (self->priv->arrow, CLUTTER_EASE_OUT_EXPO);
        clutter_actor_set_easing_duration (self->priv->arrow, 300);
        clutter_actor_set_rotation_angle (self->priv->arrow,
                                          CLUTTER_Z_AXIS,
                                          rotation_angle);
        clutter_actor_restore_easing_state (self->priv->arrow);
    }

    if (animate)
    {
        gfloat min_h, nat_h;
        clutter_actor_get_preferred_height (actor, -1, &min_h, &nat_h);
        clutter_actor_set_height (actor, 0);
        clutter_actor_save_easing_state (actor);
        clutter_actor_set_easing_mode (actor, CLUTTER_EASE_OUT_EXPO);
        clutter_actor_set_easing_duration (actor, 200);
        clutter_actor_show (actor);
        clutter_actor_set_height (actor, nat_h);
        clutter_actor_restore_easing_state (actor);
    }
    else
        clutter_actor_show (actor);

    g_signal_emit_by_name (obj, "open-state-changed", TRUE);
}

static void
on_close_completed (ClutterActor *actor,
                    CdosPopupSubMenu *self)
{
    clutter_actor_hide (actor);
    clutter_actor_set_height (actor, -1);
    g_signal_handler_disconnect (actor, self->priv->close_id);
}

/**
 * cdos_popup_sub_menu_close:
 * @obj: A #CdosPopupMenuBase
 * @animate: ??
 *
 *　关闭menu
 *
 */
void
cdos_popup_sub_menu_close (CdosPopupMenuBase *obj, gboolean animate)
{
    CdosPopupSubMenu *self = CDOS_POPUP_SUB_MENU (obj);
    g_return_if_fail (CDOS_IS_POPUP_SUB_MENU (self));

    if (!obj->is_open)
        return;

    obj->is_open = FALSE;

    if (obj->active_menu_item)
        cdos_popup_base_menu_item_set_active (obj->active_menu_item, FALSE);
    gboolean needs_scrollbar = cdos_popup_sub_menu_needs_scrollbar (self);
    //GtkPolicyType policy = needs_scrollbar ? GTK_POLICY_AUTOMATIC : GTK_POLICY_NEVER;

    if (animate && needs_scrollbar)
        animate = TRUE;

    if (self->priv->arrow)
    {
        clutter_actor_save_easing_state (self->priv->arrow);
        clutter_actor_set_easing_mode (self->priv->arrow, CLUTTER_LINEAR);
        clutter_actor_set_easing_duration (self->priv->arrow, 300);
        clutter_actor_set_rotation_angle (self->priv->arrow,
                                          CLUTTER_Z_AXIS,
                                          0.0);
        clutter_actor_restore_easing_state (self->priv->arrow);
    }

    ClutterActor *actor = (ClutterActor *)self;
    if (animate)
    {
        clutter_actor_save_easing_state (actor);
        clutter_actor_set_easing_mode (actor, CLUTTER_EASE_OUT_QUAD);
        clutter_actor_set_easing_duration (actor, 200);
        clutter_actor_set_height (actor, 0);
        clutter_actor_restore_easing_state (actor);
        self->priv->close_id = g_signal_connect (actor, "transitions-completed", G_CALLBACK (on_close_completed), self);
    }
    else
        clutter_actor_hide (actor);

    g_signal_emit_by_name (obj, "open-state-changed", FALSE);
}
