/**
 * Copyright (c) 2016-2018 CPU and Fundamental Software Research Center, CAS
 *
 * This software is published by the license of CPU-OS License，you can use and
 * distibuted this software under this License. See CPU-OS Liense for more detail.
 *
 * You should have receive a copy of CPU-OS License. If not, please contact us
 * by email <support_os@cpu-os.ac.cn>
 *
 **/

#include <time.h>
#include <langinfo.h>
#include <fcntl.h>

#include <meta/boxes.h>
#include <meta/meta-cursor-tracker.h>
#include <meta/meta-idle-monitor.h>
#include <meta/meta-x11-display.h>
#include <meta/display.h>

#include <act/act.h>
#include "cdos-app.h"
#include "cdos-global.h"
#include "cdos-background.h"
#include "cdos-layout-manager.h"
#include "cdos-session.h"
#include "cdos-start.h"
#include "cdos-screen-saver.h"

#include "cdos-power-dialog.h"
#include "cdos-login-manager.h"
#include "cdos-util.h"
#include <grp.h>

typedef enum
{
    CS_HIDE,
    CS_SHOWN,
} ContentsState;

struct _CdosPowerDialog
{
    StWidget parent_instance;

    gboolean is_active;
    gboolean is_modal;
    ContentsState contents_state;

    GDBusNodeInfo *dbus_node_info;
    GDBusConnection *connection;
    GDBusProxy *proxy;

    ActUser *user;

    gchar *username;
    CdosGlobal *global;
    ClutterActor *stage;

    ClutterActor *background;
    CdosBackgroundManager *background_manager;
    ClutterActor *mask_actor;

    ClutterActor *contents;

    StWidget *info_label;
    StWidget *button_cancel;
    StWidget *button_enforce;

    StWidget *logged_users_box;
    StWidget *logged_users_list;

    StWidget *inhibitors_box;
    StWidget *inhibitors_list;

    StWidget *button_box;

    StWidget *time_label;
    StWidget *date_label;
    StWidget *calendar;

    StWidget *message_box;
    StWidget *message_label;

    gulong time_update_id;

    PowerActionId   action_id;
    gboolean busy;
};

typedef struct 
{
    const gchar     *name;
    const gchar     *stock_id;
    const gchar     *label;
    const gchar     *accelerator;
    const gchar     *tooltip;
    PowerActionId   action_id;
    GCallback  callback;
} PowerActionEntry;

G_DEFINE_TYPE (CdosPowerDialog, cdos_power_dialog, ST_TYPE_WIDGET);

static CdosPowerDialog *global_dialog = NULL;

#define POWER_DIALOG_DBUS_NAME    "org.freedesktop.Desktop"
#define POWER_DIALOG_PATH         "/org/freedesktop/Desktop"
#define POWER_DIALOG_INTERFACE    "org.freedesktop.Desktop.PowerDialog"

static const char *dbus_xml = \
"<node name=\"/org/freedesktop/Desktop\">"
"    <interface name=\"org.freedesktop.Desktop.PowerDialog\">"
"        <method name=\"GetActionList\">"
"           <arg name=\"actions\" direction=\"out\" type=\"as\"/>"
"        </method>"
"        <method name=\"Run\">"
"           <arg name=\"timeout\" direction=\"in\" type=\"i\"/>"
"           <arg name=\"action\" direction=\"in\" type=\"s\"/>"
"           <arg name=\"result\" direction=\"out\" type=\"b\"/>"
"        </method>"
"        <method name=\"ShowInhibitorDialog\">"
"           <arg name=\"action\" direction=\"in\" type=\"s\"/>"
"           <arg name=\"result\" direction=\"out\" type=\"b\"/>"
"        </method>"
"        <method name=\"ShowMessage\">"
"           <arg name=\"message\" direction=\"in\" type=\"s\"/>"
"        </method>"
"        <method name=\"Cancel\">"
"           <arg name=\"result\" direction=\"out\" type=\"b\"/>"
"        </method>"
"        <method name=\"IsActive\">"
"           <arg name=\"result\" direction=\"out\" type=\"b\"/>"
"        </method>"
"        <signal name=\"Cancel\">"
"           <arg name=\"action_name\" type=\"s\"/>"
"        </signal>"
"        <signal name=\"Response\">"
"           <arg name=\"action_name\" type=\"s\"/>"
"           <arg name=\"enforce\" type=\"b\"/>"
"        </signal>"
"    </interface>"
"</node>";

static void handle_method_call (GDBusConnection *connection G_GNUC_UNUSED,
                                const gchar *sender,
                                const gchar *object_path G_GNUC_UNUSED,
                                const gchar *interface,
                                const gchar *method,
                                GVariant    *parameters,
                                GDBusMethodInvocation *invocation,
                                CdosPowerDialog *self);

static void show_calendar (CdosPowerDialog *self, gboolean animate);
static void hide_calendar (CdosPowerDialog *self);

static void on_cancel_clicked (StButton *btn,
                               gint      button,
                               CdosPowerDialog *self);
static gboolean
update_inhibitors_list (CdosPowerDialog *self);
static gboolean
update_logged_users_list (CdosPowerDialog *self);
static void
focus_default_action_button (CdosPowerDialog *self);
static ClutterActor *
get_action_button_by_id (CdosPowerDialog *self, PowerActionId action_id);
static PowerActionId
get_action_id_by_name (gchar *action_name);
static const gchar *
get_action_name_by_id (PowerActionId action_id);
static const gchar *
get_action_label_by_id (PowerActionId action_id);
static GSList *
act_get_logged_other_users (CdosPowerDialog *self);
static void
check_valid_actions (CdosPowerDialog *self);
static gboolean 
is_valid_action (PowerActionId action_id);
static void
power_dialog_reset_message (CdosPowerDialog *self);

static gchar **
get_action_name_list (void);

static const GDBusInterfaceVTable interface_vtable =
{
    (GDBusInterfaceMethodCallFunc)handle_method_call,
    NULL,
    NULL,
};

static gboolean
dbus_emit_signal (CdosPowerDialog *self, const gchar *signal, GVariant *parameters)
{
    GError *error = NULL;
    GDBusConnection *conn = g_bus_get_sync (G_BUS_TYPE_SESSION, NULL, &error);
    if (error)
    {
        g_warning ("Get bus connection failed: %s", error->message);
        g_error_free (error);
        return FALSE;
    }
    if (!g_dbus_connection_emit_signal (conn,
                                        NULL,
                                        POWER_DIALOG_PATH,
                                        POWER_DIALOG_INTERFACE,
                                        signal,
                                        parameters,
                                        &error))
    {
        g_warning ("Emit signal[%s] failed: %s", signal, error->message);
        g_error_free (error);
        return FALSE;
    }
    return TRUE;
}

static void
cancel_dialog (CdosPowerDialog *self, gboolean emit_signal)
{
    CdosLayoutManager *layout;
    const gchar *action_name;
    GVariant *v = NULL;

    if (!self->is_modal || !self->is_active)
        return;
    cdos_pop_modal ((ClutterActor *)self, 0);

    layout = cdos_layout_manager_get_default ();
    cdos_layout_manager_remove_chrome (layout, CLUTTER_ACTOR (self));
    clutter_actor_hide (CLUTTER_ACTOR (self));
    self->contents_state = CS_HIDE;
    self->is_modal = FALSE;
    self->is_active = FALSE;
    self->busy = FALSE;
    if (emit_signal)
    {
        action_name = get_action_name_by_id (self->action_id);
        if (action_name)
        {
            v = g_variant_new ("(s)", action_name);
            dbus_emit_signal (self, "Cancel", v);
        }
    }
}

static void
send_escape_event (CdosPowerDialog *self)
{
    XKeyEvent xkevent;
    Display *xdisplay;
    MetaX11Display *x11display;
    Window xwindow;
    Window xrootwindow;

    guint16 keycode = 9;

    MetaDisplay *display;

    display = cdos_global_get_display (global);
    x11display = meta_display_get_x11_display (display);
    xrootwindow = clutter_x11_get_root_window ();
    xdisplay = meta_x11_display_get_xdisplay (x11display);
    xwindow = xrootwindow;

    int revert;
    XGetInputFocus(xdisplay, &xwindow, &revert);
    g_debug ("%s 0x%lx", __func__, xwindow);
    {
      xkevent.display = xdisplay;
      xkevent.window = xwindow;
      xkevent.root = xrootwindow;
      xkevent.subwindow = None;
      xkevent.time = CurrentTime;
      xkevent.x = 1;
      xkevent.y = 1;
      xkevent.x_root = 1;
      xkevent.y_root = 1;
      xkevent.state = 0;
      xkevent.same_screen = True;
      xkevent.keycode = keycode;

      xkevent.type = KeyPress;
      XSendEvent (xdisplay, xwindow, False, 0, (XEvent *)&xkevent);
      XFlush (xdisplay);

      /* If the application takes a grab on KeyPress, we don't
       * want to send it a KeyRelease. There's no good way of
       * knowing whether a tray icon will take a grab, so just
       * assume it does, and don't send the KeyRelease. That might
       * make the tracking for key events messed up if it doesn't take
       * a grab, but the tray icon won't get key focus in normal cases,
       * so let's hope this isn't too damaging...
       */
      xkevent.type = KeyRelease;
      XSendEvent (xdisplay, xwindow, False, 0, (XEvent *)&xkevent);
      XFlush (xdisplay);
    }
}

/* 仅判断是否有多个用户登陆，inhibitors由session判断（session可能会生成新的inhibitor,如针对wps）。 */
static gboolean 
should_prevent_action (CdosPowerDialog *self)
{
    GSList *logged_users;
    //注销不会影响其它用户，但关机和重启会。
    if (self->action_id == PA_LOGOUT)
        return FALSE;

    logged_users = act_get_logged_other_users (self);
    if (logged_users == NULL)
        return FALSE;

    g_slist_free (logged_users);
    return TRUE;
}

static gboolean
reset_dialog_view (CdosPowerDialog *self)
{
    clutter_actor_hide (self->contents);
    show_calendar (self, TRUE);

    check_valid_actions (self);

    clutter_actor_show (self->button_box);
    self->contents_state = CS_HIDE;
    self->busy = FALSE;
    power_dialog_reset_message (self);

    if (!clutter_actor_has_key_focus (self))
        clutter_stage_set_key_focus ((ClutterStage *)self->stage, self->button_box);

    focus_default_action_button (self);
    return FALSE;
}

static void
update_enforce_button_label (CdosPowerDialog *self)
{
    const gchar *label = NULL;
    switch (self->action_id)
    {
        case PA_LOGOUT:
            label = _("Logout Force");
            break;
        case PA_REBOOT:
            label = _("Reboot Force");
            break;
        case PA_POWER_OFF:
            label = _("ShutDown Force");
            break;
        default:
            break;
    }
    if (label)
    {
        st_button_set_label (self->button_enforce, label);
    }
}

static gboolean
update_contents_view (CdosPowerDialog *self)
{
    if (self->action_id == PA_NONE)
        return FALSE;
    hide_calendar (self);
    clutter_actor_hide (self->button_box);

    self->busy = FALSE;

    if (update_inhibitors_list (self) == FALSE)
    {
        if (update_logged_users_list (self) == FALSE)
            return FALSE;
    }
    self->contents_state = CS_SHOWN;
    clutter_actor_show (self->contents);
    update_enforce_button_label (self);
    clutter_actor_grab_key_focus ((ClutterActor *)self->button_cancel);
    return TRUE;
}

static gboolean 
power_dialog_run (CdosPowerDialog *self)
{
    ClutterActor *actor = CLUTTER_ACTOR (self);
    CdosLayoutManager *layout;
    if (!self->is_active)
    {
        self->is_modal = cdos_push_modal (actor, 0, 0, CDOS_ACTION_MODE_SYSTEM_MODAL);
        if (!self->is_modal)
        {
            self->is_modal = cdos_push_modal (actor,
                                                0,
                                                META_MODAL_POINTER_ALREADY_GRABBED,
                                                CDOS_ACTION_MODE_SYSTEM_MODAL);
        }
        if (!self->is_modal)
        {
            if (clutter_check_windowing_backend (CLUTTER_WINDOWING_X11))
            {
                send_escape_event (self);
                self->is_modal = cdos_push_modal (actor, 0, 0, CDOS_ACTION_MODE_SYSTEM_MODAL);
            }
            if (!self->is_modal)
                return FALSE;
        }

        clutter_actor_show (actor);
        layout = cdos_layout_manager_get_default ();
        cdos_layout_manager_add_chrome (layout, actor, AFFECT_INPUT_REGION | ABOVE_TOP_WINDOW_GROUP);
        self->is_active = TRUE;
    }

    if (!self->busy)
        reset_dialog_view (self);
    return !self->busy;
}

static void
power_dialog_reset_message (CdosPowerDialog *self)
{
    st_label_set_text (self->message_label, "");
    clutter_actor_hide (self->message_label);
}

static void
power_dialog_show_message (CdosPowerDialog *self, gchar *message)
{
    st_label_set_text (self->message_label, message);
    clutter_actor_show (self->message_label);
}

static void
handle_method_call (GDBusConnection *connection G_GNUC_UNUSED,
                    const gchar *sender,
                    const gchar *object_path G_GNUC_UNUSED,
                    const gchar *interface,
                    const gchar *method,
                    GVariant    *parameters,
                    GDBusMethodInvocation *invocation,
                    CdosPowerDialog *self)
{
    g_debug ("%s: %s %s", __func__, interface, method);
    if (g_strcmp0 (method, "Run") == 0)
    {
        gchar *action;
        gint timeout;
        PowerActionId action_id;
        gboolean ret = FALSE;
        g_variant_get (parameters, "(is)", &timeout, &action);
        action_id = get_action_id_by_name (action);
        if (is_valid_action (action_id))
            self->action_id = action_id;
        else 
            self->action_id = PA_NONE;

        ret = power_dialog_run (self);
        g_message ("PowerDialog Run: action: %s, action_id: %d.", action, self->action_id);
        g_free (action);
        g_dbus_method_invocation_return_value (invocation, 
                                               g_variant_new ("(b)", ret));
        return;
    }
    else if (g_strcmp0 (method, "ShowInhibitorDialog") == 0)
    {
        gchar *action;
        PowerActionId action_id;
        gboolean ret = FALSE;
        g_variant_get (parameters, "(s)", &action);
        action_id = get_action_id_by_name (action);
        if (is_valid_action (action_id))
        {
            self->action_id = action_id;
            if (self->action_id > PA_NONE)
            {
                ret = power_dialog_run (self);
                update_contents_view (self);
            }
        }
        g_message ("PowerDialog ShowInhibitorDialog: action: %s, action_id: %d.", action, self->action_id);
        g_free (action);
        g_dbus_method_invocation_return_value (invocation, 
                                               g_variant_new ("(b)", ret));
        return;
    }
    else if (g_strcmp0 (method, "ShowMessage") == 0)
    {
        gchar *message;
        g_variant_get (parameters, "(s)", &message);
        power_dialog_show_message (self, message);
        g_free (message);
        g_dbus_method_invocation_return_value (invocation, NULL);
        return;
    }
    else if (g_strcmp0 (method, "IsActive") == 0)
    {
        g_dbus_method_invocation_return_value (invocation, 
                                               g_variant_new ("(b)", self->is_active));
        return;
    }
    else if (g_strcmp0 (method, "Cancel") == 0)
    {
        if (!self->busy)
            cancel_dialog (self, FALSE);
        g_dbus_method_invocation_return_value (invocation, 
                                               g_variant_new ("(b)", !self->busy));
        return;
    }
    else if (g_strcmp0 (method, "GetActionList") == 0)
    {
        gchar **actions = get_action_name_list ();
        g_dbus_method_invocation_return_value (invocation, 
                                               g_variant_new ("(^as)", actions));
        g_strfreev (actions);
    }
}

static void
on_power_dialog_dbus_acquried (GDBusConnection *conn,
                              const gchar *name,
                              CdosPowerDialog *self)
{
    GError *error = NULL;
    if (!g_dbus_connection_register_object (conn,
                                            POWER_DIALOG_PATH,
                                            self->dbus_node_info->interfaces[0],
                                            &interface_vtable,
                                            self,
                                            NULL,
                                            &error))
    {
        g_warning ("Register dbus object failed: %s", error->message);
        g_error_free (error);
    }
}

static void
on_calendar_hidden (ClutterTimeline *tml,
                    CdosPowerDialog *self)
{
    clutter_actor_hide (self->calendar);
    g_source_remove (self->time_update_id);
    self->time_update_id = 0;
    g_signal_handlers_disconnect_by_data (tml, self);
}

static void
hide_calendar (CdosPowerDialog *self)
{
    MetaRectangle *geometry;
    gfloat x, y;
    gfloat nat_w, nat_h;
    CdosLayoutManager *layout_manager = cdos_layout_manager_get_default ();
    geometry = cdos_layout_manager_get_primary_monitor (layout_manager);

    clutter_actor_get_preferred_size (self->calendar, NULL, NULL, &nat_w, &nat_h);

    clutter_actor_set_opacity (self->calendar, 255);
    x = geometry->x + (geometry->width - nat_w) / 2;
    y = geometry->y - nat_h - 100;
    ClutterTransition *transition = cdos_transition_simple (self->calendar,
                                                            CLUTTER_EASE_OUT,
                                                            250,
                                                            "opacity", 0,
                                                            "x", x,
                                                            "y", y,
                                                            NULL);
    if (transition)
        g_signal_connect (transition, 
                        "completed", 
                        G_CALLBACK (on_calendar_hidden), 
                        self);
}

static const gchar *
get_date_format (void)
{
    const gchar *format;
    const gchar *lc_time;
    lc_time = g_getenv ("LC_TIME"); 
    g_debug ("%s:: lc_time = %s", __func__, lc_time);
    if (g_strcmp0 (lc_time, "zh_CN.UTF-8") == 0)
        format = "%m月%d日 %A";
    else 
        format = "%m/%d  %A";
    return format; 
}

static gboolean
update_time (CdosPowerDialog *self)
{
    GDateTime *now = g_date_time_new_now_local ();
    gchar *time = g_date_time_format (now, cdos_get_time_format ());
    gchar *date = g_date_time_format (now, get_date_format ());
    //gchar *date = g_date_time_format (now, "%b%e  %A");
    st_label_set_text ((StLabel *)self->time_label, time);
    st_label_set_text ((StLabel *)self->date_label, date);

    g_free (time);
    g_free (date);
    g_date_time_unref (now);

    return TRUE;
}

static void
show_calendar (CdosPowerDialog *self, gboolean animate)
{
    if (self->time_update_id == 0)
    {
        self->time_update_id = g_timeout_add_seconds (1, (GSourceFunc) update_time, self);
        update_time (self);
    }

    MetaRectangle *geometry;
    gfloat orig_x, orig_y, dest_x, dest_y;
    gfloat nat_w, nat_h;
    CdosLayoutManager *layout_manager = cdos_layout_manager_get_default ();
    geometry = cdos_layout_manager_get_primary_monitor (layout_manager);
    clutter_actor_get_preferred_size (self->calendar, NULL, NULL, &nat_w, &nat_h);

    dest_x = geometry->x + (geometry->width - nat_w) / 2;
    dest_y = geometry->y + 80;
    if (!animate)
    {
        clutter_actor_set_position (self->calendar, dest_x, dest_y);
        clutter_actor_set_opacity (self->calendar, 255);
        clutter_actor_show (self->calendar);
        return;
    }

    orig_x = dest_x;
    orig_y = geometry->y - nat_h - 100;
    clutter_actor_set_position (self->calendar, orig_x, orig_y);
    clutter_actor_set_opacity (self->calendar, 0);
    clutter_actor_show (self->calendar);
    cdos_transition_simple (self->calendar,
                            CLUTTER_EASE_IN,
                            250,
                            "opacity", 255,
                            "x", dest_x,
                            "y", dest_y,
                            NULL);
}

static const gchar *
get_real_or_user_name (ActUser *user)
{
  const gchar *name;

  name = act_user_get_real_name (user);
  if (name == NULL)
    name = act_user_get_user_name (user);

  return name;
}

static GSList *
act_get_logged_other_users (CdosPowerDialog *self)
{
    ActUser *user;
    GSList *user_list, *l, *logged_users = NULL;
    const gchar *user_name;
    gboolean login_state;
    ActUserManager *act_manager;
    
    act_manager = act_user_manager_get_default ();
    user_list = act_user_manager_list_users (act_manager);

    for (l = user_list; l; l = l->next) 
    {
        user = l->data;
        user_name = act_user_get_user_name (user);

        if (g_strcmp0(self->username, user_name) == 0)
            continue; 
        if (act_user_is_logged_in (user) == FALSE)
            continue;

        logged_users = g_slist_append (logged_users, user);
    }
    return logged_users;
}

static void
users_list_add_user_item (CdosPowerDialog *self, ActUser *user)
{
    ClutterActor *user_icon;
    ClutterActor *user_name;
    ClutterActor *user_item;
    const char *icon_file;

    user_item = g_object_new (ST_TYPE_BOX_LAYOUT,
                              "style-class", "pd-user-item",
                              "x-expand", TRUE,
                              "vertical", TRUE,
                              NULL);
    //用户头像
    user_icon = g_object_new (ST_TYPE_BIN, 
                              "style-class", "pd-user-icon", 
                              NULL);

    clutter_actor_set_x_expand (user_icon, TRUE);
    clutter_actor_set_x_align (user_icon, CLUTTER_ACTOR_ALIGN_CENTER);
    clutter_actor_set_y_align (user_icon, CLUTTER_ACTOR_ALIGN_START);
    clutter_actor_add_child (user_item, user_icon);

    icon_file = act_user_get_icon_file (user);
    if (icon_file && g_file_test (icon_file, G_FILE_TEST_EXISTS))
    {
        st_bin_set_child ((StBin *)user_icon, NULL);
        gchar *style = NULL;
        style = g_strconcat ("background-image:url(\"", icon_file, "\");", NULL);
        st_widget_set_style ((StWidget *)user_icon, style);
        g_free (style);
    }
    else
    {
        st_widget_set_style ((StWidget *)user_icon, NULL);
        ClutterActor *icon = g_object_new (ST_TYPE_ICON,
                                        "icon-size", 48,
                                        "icon-name", "avatar-default-symbolic",
                                        NULL);
        st_bin_set_child ((StBin *)user_icon, icon);
    }

    user_name = (ClutterActor *)st_label_new (get_real_or_user_name (user));
    st_widget_set_style_class_name ((StWidget *)user_name, "pd-user-label");
    clutter_actor_set_x_expand (user_name, TRUE);
    clutter_actor_set_x_align (user_name, CLUTTER_ACTOR_ALIGN_CENTER);
    clutter_actor_set_y_align (user_name, CLUTTER_ACTOR_ALIGN_START);
    clutter_actor_add_child (user_item, user_name);

    clutter_actor_add_child (self->logged_users_list, user_item);
}

static gboolean
update_logged_users_list (CdosPowerDialog *self)
{
    GSList *logged_users, *l;
    guint n = 0;
    ActUser *user;

    logged_users = act_get_logged_other_users (self);
    clutter_actor_remove_all_children (CLUTTER_ACTOR(self->logged_users_list));
    st_label_set_text ((StLabel *)self->info_label, "");
    if (!logged_users)
        return FALSE;

    n = g_slist_length (logged_users);
    if (n > 0)
    {
        for (l = logged_users; l; l = l->next)
        {
            user = l->data;
            users_list_add_user_item (self, user);
        }
    }
    st_label_set_text ((StLabel *)self->info_label, 
                        _("The computer has other logged users.\nMaybe your action will make the data loss.Are you sure?"));
    g_slist_free (logged_users);
    return TRUE;
}

static StWidget *
create_inhibitor_item (CdosPowerDialog *self, gchar *app_id, gchar *reason)
{
    StWidget *item_box;
    StWidget *app_icon_box;
    StWidget *label_box;
    StWidget *app_name_label;
    StWidget *reason_label;
    ClutterActor *app_icon;
    ClutterActor *contents_text;
    CdosApp *app;
    gchar *app_name;

    app = cdos_app_system_lookup_app (cdos_app_system_get_default (), app_id);
    g_message ("Inhibitor::app_id: %s reason:%s, app:%x\n", app_id, reason, app);
    if (app)
    {
        app_name = cdos_app_get_name (app);
        app_icon = cdos_app_create_icon_texture (app, 48);
    } else {
        app_name = g_strdup (app_id);
        app_icon = st_icon_new ();
        st_icon_set_icon_size (ST_ICON (app_icon), 48);
        st_icon_set_fallback_icon_name (ST_ICON (app_icon), "application-x-executable");
    }

    g_debug ("%s:::app_name: %s.\n", __func__, app_name);

    item_box = g_object_new (ST_TYPE_BOX_LAYOUT,
                             "style-class", "pd-inhibitor-item",
                             "x-expand", TRUE,
                             "x-align", CLUTTER_ACTOR_ALIGN_FILL,
                             "y-align", CLUTTER_ACTOR_ALIGN_START,
                             NULL);

    app_icon_box = g_object_new (ST_TYPE_BIN, 
                                 "style-class", "pd-inhibitor-icon", 
                                 "x-align", CLUTTER_ACTOR_ALIGN_START,
                                 "y-align", CLUTTER_ACTOR_ALIGN_START,
                                 NULL);
    clutter_actor_add_child (item_box, app_icon_box);

    
    st_bin_set_child ((StBin *)app_icon_box, app_icon);

    label_box = g_object_new (ST_TYPE_BOX_LAYOUT,
                              "vertical", TRUE,
                              "x-align", CLUTTER_ACTOR_ALIGN_FILL,
                              "y-align", CLUTTER_ACTOR_ALIGN_START,
                              NULL);
    clutter_actor_add_child (item_box, label_box);

    app_name_label  = st_label_new (app_name);
    contents_text = st_label_get_clutter_text (app_name_label);
    st_widget_set_style_class_name ((StWidget *)app_name_label, "pd-app-label");
    clutter_text_set_line_wrap ((ClutterText *)contents_text, TRUE);
    clutter_text_set_line_wrap_mode ((ClutterText *)contents_text, PANGO_WRAP_WORD_CHAR);
    clutter_text_set_ellipsize ((ClutterText *)contents_text, PANGO_ELLIPSIZE_END);
    //clutter_text_set_line_alignment ((ClutterText *)contents_text, PANGO_ALIGN_LEFT);
    clutter_actor_set_x_expand (app_name_label, TRUE);
    clutter_actor_set_x_align (app_name_label, CLUTTER_ACTOR_ALIGN_START);
    clutter_actor_add_child (label_box, app_name_label);

    reason_label = st_label_new (reason);
    contents_text = st_label_get_clutter_text (app_name_label);
    st_widget_set_style_class_name ((StWidget *)reason_label, "pd-reason-label");
    clutter_text_set_line_wrap ((ClutterText *)contents_text, TRUE);
    clutter_text_set_line_wrap_mode ((ClutterText *)contents_text, PANGO_WRAP_WORD_CHAR);
    clutter_text_set_ellipsize ((ClutterText *)contents_text, PANGO_ELLIPSIZE_END);
    //clutter_text_set_line_alignment ((ClutterText *)contents_text, PANGO_ALIGN_LEFT);
    clutter_actor_set_x_expand (reason_label, TRUE);
    clutter_actor_set_x_align (reason_label, CLUTTER_ACTOR_ALIGN_START);
    clutter_actor_add_child (label_box, reason_label);

    return item_box;
}

static gchar *
inhibitor_get_app_id (CdosSessionInhibitor *inhibitor)
{
    gchar *desktop_filename = NULL, *app_id;
    GDesktopAppInfo *app_info;
    app_id = cdos_session_inhibitor_get_app_id (inhibitor);

    if (app_id == NULL || g_strcmp0 (app_id, "") == 0) 
    {
        desktop_filename = NULL;
    } else if (!g_str_has_suffix (app_id, ".desktop")) {
        desktop_filename = g_strdup_printf ("%s.desktop", app_id);
    } else {
        desktop_filename = g_strdup (app_id);
    }

    if (desktop_filename)
    {
        //非标准路径下的
        if (g_path_is_absolute (desktop_filename))
        {
            app_info = g_desktop_app_info_new_from_filename (desktop_filename);
            //desktop_filename无效或者不存在,则返回app_id
            if (!app_info)
            {
                g_free (desktop_filename);
                desktop_filename = g_strdup (app_id);
            } else {
                g_object_unref (app_info);
            }
        }
    }
    g_free (app_id);
    return desktop_filename;
}

static void
inhibitors_list_add_item (CdosPowerDialog *self, gchar *inhibitor_path)
{
    
    gchar   *reason, *app_id;
    CdosSessionInhibitor *session_inhibitor;
    StWidget *item;

    session_inhibitor = cdos_session_inhibitor_new (inhibitor_path);
    app_id = inhibitor_get_app_id (session_inhibitor);
    if (!app_id)
        goto out;

    reason = cdos_session_inhibitor_get_reason (session_inhibitor);

    item = create_inhibitor_item (self, app_id, reason);
    clutter_actor_add_child (self->inhibitors_list, item);

    g_free (reason);
    g_free (app_id);

out:
    g_object_unref (session_inhibitor);    
}

static gboolean
update_inhibitors_list (CdosPowerDialog *self)
{
    gchar **inhibitors;
    gchar *message;
    const gchar *action_label;

    clutter_actor_remove_all_children (CLUTTER_ACTOR(self->inhibitors_list));
    inhibitors =  cdos_session_get_inhibitors ();
    st_label_set_text ((StLabel *)self->info_label, "");

    if (inhibitors == NULL)
        return FALSE;

    guint i;
    gchar *inhibitor;

    for (i = 0; inhibitors[i] != NULL; i++)
    {
        inhibitor = inhibitors[i];
        g_debug("%s::inhibitor = %s \n", __func__, inhibitor);
        inhibitors_list_add_item (self, inhibitor);
    }

    action_label = get_action_label_by_id (self->action_id);
    message = g_strdup_printf (_("The above program prevents %s. Forced operation may cause data loss.\nYou can click “cancel”，and complete the operation you want to perform."), gettext (action_label));
    st_label_set_text ((StLabel *)self->info_label, message);
    g_strfreev (inhibitors);
    g_free (message);
    return TRUE;
}

static gboolean 
key_is_enter (guint key)
{
    gboolean is_enter = (key == CLUTTER_KEY_ISO_Enter ||
                         key == CLUTTER_KEY_Return ||
                         key == CLUTTER_KEY_KP_Enter);
    return is_enter;
}

static void
emit_power_signal (CdosPowerDialog *self, gboolean enforce)
{
    const gchar *action_name;
    GVariant *v;
    action_name = get_action_name_by_id (self->action_id);
    if (action_name == NULL)
        action_name = "None";
    v = g_variant_new ("(sb)", action_name, enforce);
    g_message ("Emit signal: action: %s enforce: %d", action_name, enforce);
    dbus_emit_signal (self, "Response", v);
    //self->busy = FALSE;
}

//强制退出session的操作
static gboolean
start_session_quit_enforce_delay (CdosPowerDialog *self)
{
    emit_power_signal (self, TRUE);
    return FALSE;
}

//常规退出session的操作
static gboolean
start_session_quit_delay (CdosPowerDialog *self)
{
    gboolean prevent = FALSE;
    prevent = should_prevent_action (self);
    g_debug ("%s::prevent = %d", __func__, prevent);
    if (prevent)
        update_contents_view (self);
    else 
        emit_power_signal (self, FALSE);
    return FALSE;
}

static gboolean 
should_apply_event (ClutterEvent *event)
{
    ClutterEventType event_type;
    event_type = clutter_event_type (event);
    if (event_type ==  CLUTTER_BUTTON_PRESS || event_type == CLUTTER_BUTTON_RELEASE)
    {
        guint32 button = clutter_event_get_button (event);
        return button == 1 ? TRUE : FALSE;
    }

    if (event_type ==  CLUTTER_KEY_PRESS || event_type == CLUTTER_KEY_RELEASE)
    {
        guint key = clutter_event_get_key_symbol (event);
        return key_is_enter (key);
    }
    return FALSE;
}

static gboolean
action_logout_callback (ClutterActor *btn,
                        ClutterEvent *event,
                        CdosPowerDialog *self)
{
    if (should_apply_event (event) && !self->busy)
    {
        self->action_id = PA_LOGOUT;
        self->busy = TRUE;
        g_idle_add (start_session_quit_delay, self);
    }
    return FALSE;
}

static gboolean
action_reboot_callback (ClutterActor *btn,
                        ClutterEvent *event,
                        CdosPowerDialog *self)
{
    if (should_apply_event (event) && !self->busy)
    {
        self->action_id = PA_REBOOT;
        self->busy = TRUE;
        g_idle_add (start_session_quit_delay, self);
    }
    return FALSE;
}

static gboolean
action_suspend_callback (ClutterActor *btn,
                         ClutterEvent *event,
                         CdosPowerDialog *self)
{
    if (should_apply_event (event) && !self->busy)
    {
        self->action_id = PA_SUSPEND;
        self->busy = TRUE;
        cdos_session_suspend ();
        cancel_dialog (self, FALSE);
    }
    return FALSE;
}

static gboolean
action_hibernate_callback (ClutterActor *btn,
                           ClutterEvent *event,
                           CdosPowerDialog *self)
{
    if (should_apply_event (event) && !self->busy)
    {
        self->action_id = PA_HIBERNATE;
        self->busy = TRUE;
        cdos_session_hibernate ();
        cancel_dialog (self, FALSE);
    }
    return FALSE;
}

static gboolean
action_power_off_callback (ClutterActor *btn,
                           ClutterEvent *event,
                           CdosPowerDialog *self)
{
    if (should_apply_event (event) && !self->busy)
    {
        self->action_id = PA_POWER_OFF;
        self->busy = TRUE;
        g_idle_add (start_session_quit_delay, self);
    }
    return FALSE;
}

static gboolean
action_screensaver_callback (ClutterActor *btn,
                             ClutterEvent *event,
                             CdosPowerDialog *self)
{
    if (should_apply_event (event) && !self->busy)
    {
        self->action_id = PA_SCREENSAVER_LOCK;
        cdos_screen_saver_lock ();
        cancel_dialog (self, FALSE);
    }
    return FALSE;
}

static gboolean
action_switch_user_callback (ClutterActor *btn,
                             ClutterEvent *event,
                             CdosPowerDialog *self)
{
    if (should_apply_event (event) && !self->busy)
    {
        self->action_id = PA_SWITCH_USER;
        self->busy = TRUE;
        cdos_util_spawn_command_line ("dm-tool switch-to-greeter"); 
        cancel_dialog (self, FALSE);
    }
    return FALSE;
}

static void
on_cancel_clicked (StButton *btn,
                   gint      button,
                   CdosPowerDialog *self)
{
    if (button == 1 && !self->busy)
    {
        cancel_dialog (self, TRUE);
    }
}

static void
on_enforce_clicked (StButton *btn,
                   gint      button,
                   CdosPowerDialog *self)
{
    if (button == 1 && !self->busy)
    {
        self->busy = TRUE;
        g_idle_add (start_session_quit_enforce_delay, self);
    }
}

static void
button_on_key_focus_in (ClutterActor *actor, CdosPowerDialog *self)
{
    PowerActionId action_id;
    const gchar *name;
    name = clutter_actor_get_name (actor);
    action_id = get_action_id_by_name (name);
    self->action_id = action_id;
}

static StWidget *
create_action_button (CdosPowerDialog *self, PowerActionEntry *action_entry)
{
    StWidget *button;
    StWidget *button_box;
    StWidget *icon;
    StLabel  *label;

    button = st_button_new ();
    st_widget_add_style_class_name ((StWidget *)button, "pd-action-button");
    st_widget_set_can_focus ((StWidget *)button, TRUE);
    //clutter_actor_set_reactive ((ClutterActor *)button, TRUE);
    clutter_actor_set_name (CLUTTER_ACTOR (button), action_entry->name);
    g_signal_connect (CLUTTER_ACTOR (button), 
                      "button-release-event", 
                      action_entry->callback, 
                      self);
    g_signal_connect (CLUTTER_ACTOR (button), 
                      "key-release-event", 
                      action_entry->callback, 
                      self);
    g_signal_connect (CLUTTER_ACTOR (button), 
                      "key-focus-in", 
                      button_on_key_focus_in, 
                      self);
    button_box = g_object_new (ST_TYPE_BOX_LAYOUT,
                               "x-expand", TRUE,
                               "y-expand", TRUE,
                               "x-align", CLUTTER_ACTOR_ALIGN_CENTER,
                               "y-align", CLUTTER_ACTOR_ALIGN_CENTER,
                               "vertical", TRUE,
                               NULL);
    st_bin_set_child ((StBin *)button, button_box);

    //button icon
    icon = g_object_new (ST_TYPE_ICON,
                        "icon-name", action_entry->stock_id,
                        "icon-size", 48,
                        "x-align", CLUTTER_ACTOR_ALIGN_CENTER,
                        "y-align", CLUTTER_ACTOR_ALIGN_CENTER,
                        NULL);
    clutter_actor_add_child (button_box, icon);

    //button label
    label = g_object_new (ST_TYPE_LABEL, 
                          "style-class", "pd-action-label",
                          NULL);

    st_label_set_text ((StLabel *)label, gettext (action_entry->label));
    clutter_actor_add_child (button_box, CLUTTER_ACTOR(label));
    return button;
}

static void
init_background (CdosPowerDialog *self)
{
    self->background = clutter_actor_new ();
    self->background_manager = \
        cdos_background_manager_new (self->background, 
                                     FALSE, 
                                     FALSE,
                                     TRUE,
                                     "org.cdos.desktop.background");
    cdos_background_manager_hide_register_watermark (self->background_manager);

    clutter_actor_add_child (CLUTTER_ACTOR (self), self->background);
}

static void
init_calendar_box (CdosPowerDialog *self)
{
    self->calendar = g_object_new (ST_TYPE_WIDGET, 
                                   NULL);
    st_widget_add_style_class_name (self->calendar,"cs-calendar");
    ClutterLayoutManager *layout = clutter_box_layout_new ();
    clutter_box_layout_set_orientation (CLUTTER_BOX_LAYOUT (layout), CLUTTER_ORIENTATION_VERTICAL);
    clutter_actor_set_layout_manager (self->calendar, layout);
    clutter_actor_add_child (CLUTTER_ACTOR (self), self->calendar);

    self->time_label = st_label_new (NULL);
    st_widget_add_style_class_name (self->time_label,"cs-time-label");
    self->date_label = st_label_new (NULL);
    st_widget_add_style_class_name (self->date_label,"cs-date-label");
    clutter_actor_add_child (self->calendar, self->time_label);
    clutter_actor_add_child (self->calendar, self->date_label);

    clutter_actor_set_child_above_sibling (CLUTTER_ACTOR (self), self->calendar, self->mask_actor);
}

static void
init_users_box (CdosPowerDialog *self)
{
    StWidget *users_list_box;
    StWidget *scroll_view;

    self->logged_users_box = st_box_layout_new ();
    st_widget_set_style_class_name ((StWidget *)self->logged_users_box, "pd-users-box");
    clutter_actor_set_x_align ((ClutterActor *)self->logged_users_box, CLUTTER_ACTOR_ALIGN_CENTER);
    clutter_actor_set_y_align ((ClutterActor *)self->logged_users_box, CLUTTER_ACTOR_ALIGN_START);
    clutter_actor_add_child (self->contents, self->logged_users_box);

    scroll_view =  g_object_new (ST_TYPE_SCROLL_VIEW,
                                "style-class" , "hfade menu-applications-scrollbox",
                                "x-fill", FALSE,
                                "y-fill", FALSE,
                                "x-expand", FALSE,
                                "x-align", CLUTTER_ACTOR_ALIGN_CENTER,
                                "y-align", CLUTTER_ACTOR_ALIGN_START,
                                NULL);
    st_scroll_view_set_policy (scroll_view, GTK_POLICY_AUTOMATIC, GTK_POLICY_NEVER);
    st_scroll_view_set_auto_scrolling (scroll_view, TRUE);
    ClutterActor *hscroll_bar = st_scroll_view_get_hscroll_bar (scroll_view);
    clutter_actor_hide (hscroll_bar);
    clutter_actor_add_child (self->logged_users_box, (ClutterActor *)scroll_view);

    users_list_box = ST_BOX_LAYOUT (st_box_layout_new ());
    st_box_layout_set_vertical (users_list_box, FALSE);
    clutter_actor_set_x_expand ((ClutterActor *)users_list_box, TRUE);
    clutter_actor_set_x_align ((ClutterActor *)users_list_box, CLUTTER_ACTOR_ALIGN_FILL);
    clutter_actor_set_y_expand ((ClutterActor *)users_list_box, TRUE);
    clutter_actor_set_y_align ((ClutterActor *)users_list_box, CLUTTER_ACTOR_ALIGN_START);
    st_scroll_view_add_child (scroll_view, (ClutterActor *)users_list_box);
    self->logged_users_list = users_list_box;
}

static void
focus_default_action_button (CdosPowerDialog *self)
{
    ClutterActor *button;
    if (self->action_id == PA_NONE)
        return;
    button = get_action_button_by_id (self, self->action_id);
    if (button)
        clutter_actor_grab_key_focus (button);
}

static const PowerActionEntry action_entries[] = 
{
    {
        "ScreensaverLock", "found-system-screenlock",
        N_("Lock Screen"), "<alt>L",
        NULL, PA_SCREENSAVER_LOCK,
        G_CALLBACK (action_screensaver_callback)
    },
    {
        "SwitchUser", "found-system-switch-user",
        N_("Switch User"), "<alt>U",
        NULL, PA_SWITCH_USER,
        G_CALLBACK (action_switch_user_callback)
    },
    {
        "Suspend", "found-system-suspend",
        N_("Suspend"), "<alt>S",
        NULL, PA_SUSPEND,
        G_CALLBACK (action_suspend_callback)
    },
    {
        "Hibernate", "found-system-hibernate",
        N_("Hibernate"), "<alt>H",
        NULL, PA_HIBERNATE,
        G_CALLBACK (action_hibernate_callback)
    },
    {
        "Logout", "found-system-logout",
        N_("Logout"), "<alt>O",
        NULL, PA_LOGOUT,
        G_CALLBACK (action_logout_callback)
    },
    {
        "Reboot", "found-system-reboot",
        N_("Reboot"), "<alt>R",
        NULL, PA_REBOOT,
        G_CALLBACK (action_reboot_callback)
    },
    {
        "PowerOff", "found-system-power",
        N_("Power Off"), "<alt>P",
        NULL, PA_POWER_OFF,
        G_CALLBACK (action_power_off_callback)
    }
};

static gboolean 
is_valid_action (PowerActionId action_id)
{
    gboolean is_valid = TRUE;
    if (action_id  == PA_SWITCH_USER)
        is_valid = can_switch_user ();
    if (action_id  == PA_SUSPEND)
        is_valid = check_acpi_supports ("S3");
    if (action_id  == PA_HIBERNATE)
        is_valid = check_acpi_supports ("S4");
    //暂时先不做判断
    /* if (action_id  == PA_POWER_OFF)
        is_valid = cdos_session_can_shutdown ();
    if (action_id  == PA_REBOOT)
        is_valid = cdos_session_can_reboot (); */

    return is_valid;
}

static void
check_valid_actions (CdosPowerDialog *self)
{
    ClutterActor *action_button;
    GList *children, *l;
    PowerActionId action_id;
    const gchar *action_name;

    children = clutter_actor_get_children (self->button_box);
    for (l = children; l != NULL; l = l->next)
    {
        action_button = l ->data;
        action_name = clutter_actor_get_name (action_button);
        action_id = get_action_id_by_name (action_name);
        if (is_valid_action (action_id) == FALSE)
            clutter_actor_hide (action_button);
        else 
            clutter_actor_show (action_button);
    }
    g_list_free (children);
}

static gchar **
get_action_name_list (void)
{
    gint n = G_N_ELEMENTS (action_entries);
    guint i;
    gchar **actions = g_malloc0_n (n + 1, sizeof(char *));
    for (i = 0; i < n; i++)
    {
        actions[i] = g_strdup (action_entries[i].name);
    }
    actions[i] = NULL;
    return actions;
}

static void
init_actions_box (CdosPowerDialog *self)
{
    StWidget *button, *last_button = NULL;
    GList *children ;
    self->button_box = g_object_new (ST_TYPE_BOX_LAYOUT,
                                     "style-class", "pd-button-box",
                                     "x-expand", TRUE,
                                     "vertical", FALSE,
                                     NULL);
    clutter_actor_add_child (CLUTTER_ACTOR (self), self->button_box);
    clutter_actor_show (self->button_box);
    st_widget_set_can_focus (self->button_box, TRUE);
    clutter_actor_set_reactive ((ClutterActor *)self->button_box, TRUE);

    /* Add action buttons */
    gint n = G_N_ELEMENTS (action_entries);
    guint i;
    for (i = 0; i < n; i++)
    {
        button = create_action_button (self, &action_entries[i]);
        clutter_actor_add_child (self->button_box, button);
    }
}

static void
init_inhibitors_box (CdosPowerDialog *self)
{
    StWidget *scroll_view;

    self->inhibitors_box = ST_BOX_LAYOUT (st_box_layout_new ());
    st_widget_set_style_class_name ((StWidget *)self->inhibitors_box, "pd-inhibitors-box");
    clutter_actor_set_x_align ((ClutterActor *)self->inhibitors_box, CLUTTER_ACTOR_ALIGN_CENTER);
    clutter_actor_set_y_align ((ClutterActor *)self->inhibitors_box, CLUTTER_ACTOR_ALIGN_START);
    clutter_actor_add_child ((ClutterActor *)self->contents, (ClutterActor *)self->inhibitors_box);

    scroll_view =  g_object_new (ST_TYPE_SCROLL_VIEW,
                                 "style-class", "vfade menu-applications-scrollbox",
                                 "x-fill", TRUE,
                                 "y-fill", FALSE,
                                 "x-expand", TRUE,
                                 "x-align", CLUTTER_ACTOR_ALIGN_FILL,
                                 "y-align", CLUTTER_ACTOR_ALIGN_START,
                                 NULL);
    st_scroll_view_set_policy (scroll_view, GTK_POLICY_NEVER, GTK_POLICY_AUTOMATIC);
    st_scroll_view_set_auto_scrolling (scroll_view, TRUE);
    clutter_actor_add_child ((ClutterActor *)self->inhibitors_box, (ClutterActor *)scroll_view);

    self->inhibitors_list = g_object_new (ST_TYPE_BOX_LAYOUT,
                                          "style-class", "pd-inhibitor-list",
                                          "x-expand", TRUE,
                                          "x-align", CLUTTER_ACTOR_ALIGN_FILL,
                                          "y-align", CLUTTER_ACTOR_ALIGN_START,
                                          "vertical", TRUE,
                                          NULL);
    st_scroll_view_add_child (scroll_view, (ClutterActor *)self->inhibitors_list);
}

static void
init_message_box (CdosPowerDialog *self)
{
    self->message_box = g_object_new (ST_TYPE_BOX_LAYOUT,
                                      "style-class", "pd-message-box",
                                      "x-expand", TRUE,
                                      "y-expand", TRUE,
                                      "vertical", FALSE,
                                      NULL);
    clutter_actor_add_child ((ClutterActor *)self, (ClutterActor *)self->message_box);

    self->message_label = st_label_new (NULL);
    clutter_actor_set_x_align ((ClutterActor *)self->message_label, CLUTTER_ACTOR_ALIGN_CENTER);
    clutter_actor_set_y_align ((ClutterActor *)self->message_label, CLUTTER_ACTOR_ALIGN_START);
    clutter_actor_set_y_expand ((ClutterActor *)self->message_label, TRUE);
    clutter_actor_set_x_expand ((ClutterActor *)self->message_label, TRUE);

    clutter_actor_add_child ((ClutterActor *)self->message_box, 
                             (ClutterActor *)self->message_label);
    power_dialog_reset_message (self);
}

static void
prevent_box_info_label_setup (StWidget *label)
{
    st_widget_set_style_class_name ((StWidget *)label, "pd-message-label");
    ClutterActor *contents_text = st_label_get_clutter_text (label);
    clutter_text_set_line_wrap ((ClutterText *)contents_text, TRUE);
    clutter_text_set_line_wrap_mode ((ClutterText *)contents_text, PANGO_WRAP_WORD_CHAR);
    clutter_text_set_ellipsize ((ClutterText *)contents_text, PANGO_ELLIPSIZE_NONE);
    //clutter_text_set_line_alignment ((ClutterText *)contents_text, PANGO_ALIGN_LEFT);
}

static void
init_prevent_box (CdosPowerDialog *self)
{
    init_users_box (self);
    init_inhibitors_box (self);

    //message label;
    self->info_label  = (ClutterActor *)st_label_new (NULL);
    prevent_box_info_label_setup (self->info_label);
    clutter_actor_add_child (self->contents, self->info_label);

    ClutterActor *button_box = g_object_new (ST_TYPE_BOX_LAYOUT,
                                             "x-expand", TRUE,
                                             "vertical", TRUE,
                                             NULL);
    clutter_actor_add_child (self->contents, button_box);

    //cancel action button 
    self->button_cancel =  (ClutterActor *)st_button_new_with_label (_("Cancel"));
    st_widget_add_style_class_name ((StWidget *)self->button_cancel, "pd-button-cancel");
    st_widget_set_can_focus ((StWidget *)self->button_cancel, TRUE);
    clutter_actor_set_x_align (self->button_cancel, CLUTTER_ACTOR_ALIGN_CENTER);
    clutter_actor_set_y_align (self->button_cancel, CLUTTER_ACTOR_ALIGN_START);
    g_signal_connect (self->button_cancel, 
                      "clicked", 
                      G_CALLBACK (on_cancel_clicked), 
                      self);
    clutter_actor_add_child (button_box, self->button_cancel);

    //enforce action button 
    self->button_enforce =  (ClutterActor *)st_button_new ();
    st_widget_add_style_class_name ((StWidget *)self->button_enforce, "pd-button-enforce");
    st_widget_set_can_focus ((StWidget *)self->button_enforce, TRUE);
    clutter_actor_set_x_align (self->button_enforce, CLUTTER_ACTOR_ALIGN_CENTER);
    clutter_actor_set_y_align (self->button_enforce, CLUTTER_ACTOR_ALIGN_START);
    g_signal_connect (self->button_enforce, 
                      "clicked", 
                      G_CALLBACK (on_enforce_clicked), 
                      self);
    clutter_actor_add_child (button_box, self->button_enforce);
}

static PowerActionId
get_action_id_by_name (gchar *action_name)
{
    if (!action_name)
        return PA_NONE;

    gint n = G_N_ELEMENTS (action_entries);
    guint i;
    for (i = 0; i < n; i++)
    {
        if (g_strcmp0 (action_name, action_entries[i].name) == 0)
        {
            return action_entries[i].action_id;
        }
    }
    return PA_NONE;
}

static const gchar *
get_action_name_by_id (PowerActionId action_id)
{
    if (action_id == PA_NONE)
        return NULL;

    gint n = G_N_ELEMENTS (action_entries);
    guint i;
    for (i = 0; i < n; i++)
    {
        if (action_id == action_entries[i].action_id)
        {
            return action_entries[i].name;
        }
    }
    return NULL;
}

static const gchar *
get_action_label_by_id (PowerActionId action_id)
{
    if (action_id == PA_NONE)
        return NULL;

    gint n = G_N_ELEMENTS (action_entries);
    guint i;
    for (i = 0; i < n; i++)
    {
        if (action_id == action_entries[i].action_id)
        {
            return action_entries[i].label;
        }
    }
    return NULL;
}

static ClutterActor *
get_action_button_by_id (CdosPowerDialog *self, PowerActionId action_id)
{
    ClutterActor *button = NULL, *action_button;
    GList *children, *l;
    const gchar *action_name = NULL;
    const gchar *name;

    action_name = get_action_name_by_id (action_id);
    if (!action_name)
        return NULL;

    children = clutter_actor_get_children (self->button_box);
    for (l = children; l != NULL; l = l->next)
    {
        action_button = l ->data;
        name = clutter_actor_get_name (action_button);
        g_debug ("%s::action name = %s", __func__, name);
        if (name && g_strcmp0 (name, action_name) == 0)
            button = action_button;
    }
    g_list_free (children);
    return button;
}

static void
init_ui (CdosPowerDialog *self)
{
    ClutterActor *actor;
    //ClutterEffect *bg_effect;

    actor = CLUTTER_ACTOR (self);
    ClutterColor color = {255, 0, 0, 0};
    clutter_actor_set_background_color (actor, &color);

    init_background (self);

    self->mask_actor = clutter_actor_new ();
    clutter_actor_add_child (actor, self->mask_actor);
    clutter_actor_set_child_above_sibling (CLUTTER_ACTOR (self), self->mask_actor, self->background);
    ClutterColor mask_color = {0, 0, 0, 200};
    clutter_actor_set_background_color (self->mask_actor, &mask_color);

    if (cdos_enable_blur_effect ())
    {
        //bg_effect = clutter_blur_effect_new ();
        //bg_effect = st_blur_effect_new ();
        //st_blur_effect_set_mode (bg_effect, ST_BLUR_MODE_BACKGROUND);
        //st_blur_effect_set_sigma (bg_effect, 4.0);
        //clutter_actor_add_effect (self->mask_actor, bg_effect);
    }

    self->contents = g_object_new (ST_TYPE_BOX_LAYOUT,
                                   "style-class", "pd-contents", 
                                   "x-expand", TRUE,
                                   "vertical", TRUE,
                                   "x-align", CLUTTER_ACTOR_ALIGN_CENTER,
                                   NULL);
    clutter_actor_set_reactive ((ClutterActor *)self->contents, TRUE);
    clutter_actor_add_child (actor, self->contents);
    clutter_actor_hide (self->contents);

    init_calendar_box (self);
    init_actions_box (self);
    init_prevent_box (self);
    init_message_box (self);

    clutter_actor_hide ((ClutterActor *)self);
    clutter_actor_set_reactive ((ClutterActor *)self, TRUE);
}

static void
cdos_power_dialog_navigate_focus (CdosPowerDialog *self, ClutterActor *box, guint key)
{
    gboolean is_nav_key = TRUE;
    GtkDirectionType dir;
    ClutterActor *focused, *next_sibling = NULL;
    GList *children, *l;
    if (self->busy)
        return;

    focused = clutter_stage_get_key_focus (CLUTTER_STAGE (self->stage));
    switch (key)
    {
        case CLUTTER_KEY_Up:
        case CLUTTER_KEY_Left:
            dir = GTK_DIR_LEFT;
            break;
        case CLUTTER_KEY_Right:
        case CLUTTER_KEY_Tab:
        case CLUTTER_KEY_Down:
            dir = GTK_DIR_RIGHT;
            break;
        default:
            is_nav_key = FALSE;
            break;
    }
    if (is_nav_key)
    {
        children = clutter_actor_get_children (box);
        if (clutter_actor_contains ((ClutterActor *)box, focused))
        {
            if (dir == GTK_DIR_LEFT)
                next_sibling = clutter_actor_get_previous_sibling (focused);
            else 
                next_sibling = clutter_actor_get_next_sibling (focused);
        }
        if (!next_sibling)
        {
            if (dir == GTK_DIR_LEFT)
                next_sibling = (ClutterActor *)g_list_last (children)->data;
            else 
                next_sibling = (ClutterActor *)children->data;
        }
        //跳过不显示的
        if (clutter_actor_is_visible (next_sibling) == FALSE)
        {
            if (dir == GTK_DIR_LEFT)
                l = g_list_last (children);
            else 
                l = children;
            gboolean skip = TRUE;
            while (l)
            {
                ClutterActor *child = l->data;
                if (dir == GTK_DIR_LEFT)
                    l = l->prev;
                else 
                    l = l->next;

                //下一个开始不直接跳过;
                if (child == next_sibling)
                {
                    skip = FALSE;
                    continue;
                }
                if (!skip && clutter_actor_is_visible (child))
                {
                    next_sibling = child;
                    break;
                }                
            }
        }
        if (next_sibling && next_sibling != focused)
            clutter_actor_grab_key_focus (next_sibling);
        g_list_free (children);
    }
}

static gboolean
cdos_power_dialog_key_press_event (ClutterActor *actor,
                                   ClutterKeyEvent *event)
{
    CdosPowerDialog *self = CDOS_POWER_DIALOG (actor);
    if (!self->is_active)
        return FALSE;
    guint key = clutter_event_get_key_symbol ((ClutterEvent *)event);
    if (key == CLUTTER_KEY_Escape && !self->busy)
    {
        cancel_dialog (self, TRUE);
        return TRUE;
    }

    //强制操作不做导航键，防止误操作。
    if (clutter_actor_is_visible (self->button_box))
        cdos_power_dialog_navigate_focus (self, self->button_box, key);
    return FALSE;
}

static gboolean
cdos_power_dialog_button_release_event (ClutterActor *actor,
                                        ClutterButtonEvent *event)
{
    CdosPowerDialog *self = CDOS_POWER_DIALOG (actor);
    ClutterActor *source = clutter_event_get_source (event);

    if (self->is_active && (source == actor) && !self->busy)
    {
        cancel_dialog (self, TRUE);
    }
    return FALSE;
}

static void
cdos_power_dialog_get_preferred_width (ClutterActor *actor,
                                       gfloat for_height,
                                       gfloat   *min_width,
                                       gfloat   *nat_width)
{
    CdosPowerDialog *self = CDOS_POWER_DIALOG (actor);
    gfloat width = clutter_actor_get_width (self->stage);

    *min_width = width;
    *nat_width = width;
}

static void
cdos_power_dialog_get_preferred_height (ClutterActor *actor,
                                        gfloat for_width,
                                        gfloat   *min_height,
                                        gfloat   *nat_height)
{
    CdosPowerDialog *self = CDOS_POWER_DIALOG (actor);
    gfloat height = clutter_actor_get_height (self->stage);

    *min_height = height;
    *nat_height = height;
}

static void
cdos_power_dialog_allocate (ClutterActor *actor,
                            const ClutterActorBox *box,
                            ClutterAllocationFlags flags)
{
    CdosPowerDialog *self = CDOS_POWER_DIALOG (actor);
    gfloat min_width, nat_width, min_height, nat_height;
    MetaRectangle *geometry;

    CLUTTER_ACTOR_CLASS (cdos_power_dialog_parent_class)->allocate (actor, box, flags);

    CdosLayoutManager *layout_manager = cdos_layout_manager_get_default ();
    geometry = cdos_layout_manager_get_primary_monitor (layout_manager);

    ClutterActorBox child_box;

    if (self->mask_actor)
    {
        ClutterActorBox stage_box;
        clutter_actor_get_allocation_box (self->stage, &stage_box);
        child_box.x1 = stage_box.x1;
        child_box.y1 = stage_box.y1;
        child_box.x2 = stage_box.x2;
        child_box.y2 = stage_box.y2;
        clutter_actor_allocate (self->mask_actor, &child_box, flags);
    }
    if (self->contents && self->contents_state != CS_HIDE)
    {
        clutter_actor_get_preferred_size (self->contents, &min_width, &min_height, &nat_width, &nat_height);
        child_box.x1 = (int)(geometry->x + (geometry->width - nat_width) * 0.5);
        child_box.y1 = (int)(geometry->y + (geometry->height - nat_height) * 0.5);
        child_box.x2 = child_box.x1 + nat_width;
        child_box.y2 = child_box.y1 + nat_height;
        clutter_actor_allocate (self->contents, &child_box, flags);
    }

    if (self->button_box && clutter_actor_is_visible ((ClutterActor *)self->button_box))
    {
        clutter_actor_get_preferred_size (self->button_box, &min_width, &min_height, &nat_width, &nat_height);
        child_box.x1 = (int)(geometry->x + (geometry->width - nat_width) * 0.5);
        child_box.y1 = (int)(geometry->y + (geometry->height - nat_height) * 0.5);
        child_box.x2 = child_box.x1 + nat_width + 0.5;
        child_box.y2 = child_box.y1 + nat_height + 0.5;
        clutter_actor_allocate (self->button_box, &child_box, flags);
    }

    if (self->message_box && clutter_actor_is_visible ((ClutterActor *)self->message_box))
    {
        clutter_actor_get_preferred_size (self->message_box, &min_width, &min_height, &nat_width, &nat_height);
        child_box.x1 = (int)(geometry->x + (geometry->width - nat_width) * 0.5);
        child_box.y1 = (int)child_box.y2 + 36;
        child_box.x2 = child_box.x1 + nat_width + 0.5;
        child_box.y2 = child_box.y1 + nat_height + 0.5;
        clutter_actor_allocate (self->message_box, &child_box, flags);
    }
}

static void
cdos_power_dialog_dispose (GObject *gobject)
{
    CdosPowerDialog *self = CDOS_POWER_DIALOG (gobject);
    ActUserManager *act_manager = act_user_manager_get_default ();
    g_signal_handlers_disconnect_by_data (act_manager, self);

    g_clear_object (&self->proxy);
    g_clear_object (&self->user);
    g_clear_pointer (&self->username, g_free);

    G_OBJECT_CLASS (cdos_power_dialog_parent_class)->dispose(gobject);
}

static void
cdos_power_dialog_finalize (GObject *gobject)
{
    CdosPowerDialog *self = CDOS_POWER_DIALOG (gobject);
    
    G_OBJECT_CLASS (cdos_power_dialog_parent_class)->finalize(gobject);
}

static void
cdos_power_dialog_class_init (CdosPowerDialogClass *klass)
{
    GObjectClass *gobject_class = G_OBJECT_CLASS(klass);
    ClutterActorClass *actor_class = CLUTTER_ACTOR_CLASS (klass);

    gobject_class->finalize = cdos_power_dialog_finalize;
    gobject_class->dispose = cdos_power_dialog_dispose;

    actor_class->get_preferred_width = cdos_power_dialog_get_preferred_width;
    actor_class->get_preferred_height = cdos_power_dialog_get_preferred_height;
    actor_class->allocate = cdos_power_dialog_allocate;
    actor_class->key_press_event = cdos_power_dialog_key_press_event;
    actor_class->button_release_event = cdos_power_dialog_button_release_event;
}

static void
cdos_power_dialog_init (CdosPowerDialog *self)
{
    GError *error = NULL;
    self->is_active = FALSE;
    self->proxy = NULL;
    self->username = g_strdup (g_get_user_name ());
    self->action_id = PA_NONE;
    self->busy = FALSE;

    self->dbus_node_info = g_dbus_node_info_new_for_xml (dbus_xml, &error);
    if (error)
    {
        g_warning ("Dbus node info new failed:%s", error->message);
        g_error_free (error);
        return;
    }
    g_bus_own_name (G_BUS_TYPE_SESSION,
                    POWER_DIALOG_DBUS_NAME,
                    G_BUS_NAME_OWNER_FLAGS_REPLACE | G_BUS_NAME_OWNER_FLAGS_ALLOW_REPLACEMENT,
                    (GBusAcquiredCallback )on_power_dialog_dbus_acquried,
                    NULL,
                    NULL,
                    self,
                    NULL);

    ActUserManager *act_manager = act_user_manager_get_default ();
    self->user = act_user_manager_get_user (act_manager, self->username);

    self->global = cdos_global_get ();
    self->stage = (ClutterActor *)cdos_global_get_stage (self->global);

    init_ui (self);
}

CdosPowerDialog *
cdos_power_dialog_new (void)
{
    if (!global_dialog)
    {
        global_dialog = g_object_new (CDOS_TYPE_POWER_DIALOG, NULL);
        g_object_add_weak_pointer ((GObject *)global_dialog,
                                   (gpointer *)&global_dialog);
    }
    return global_dialog;
}

CdosPowerDialog *
cdos_power_dialog_get (void)
{
    if (!global_dialog)
    {
        global_dialog = cdos_power_dialog_new ();
    }
    return global_dialog;
}

gboolean
cdos_power_dialog_run (PowerActionId action_id)
{   
    CdosPowerDialog *self;
    self = cdos_power_dialog_get ();
    self->action_id = action_id;
    return power_dialog_run (self);
}
