#include <gio/gio.h>
#include "cdos-run-dialog.h"
#include "cdos-modal-dialog.h"
#include "cdos-util.h"
#include "config.h"
#include "cdos-start.h"
#include "cdos-debug.h"
#include "cdos-looking-glass.h"

#include <meta/main.h>

#define LOCKDOWN_SCHEMA "org.cdos.desktop.lockdown"
#define TERMINAL_SCHEMA "org.cdos.desktop.default-applications.terminal"
#define DISABLE_COMMAND_LINE_KEY "disable-command-line"

extern int debug_call;
extern int debug_mode;

static void
reexec_self (void)
{
    cdos_global_reexec_self (global);
}

static void
debug_quit (void)
{
    meta_quit (META_EXIT_ERROR);
}

static void
set_debug_call (void)
{
    g_atomic_int_xor (&debug_call, 1);
}

static void
set_debug_mode (void)
{
    g_atomic_int_xor (&debug_mode, 1);
}

typedef void (*ICFunc) (void);

struct InternalCmd
{
    const char *cmd;
    const ICFunc func;
};

struct InternalCmd internal_cmds[] = {
    {"lg", cdos_looking_glass_init},
    {"r", reexec_self},
    {"restart", reexec_self},
    {"debugexit", debug_quit},
    {"rt", cdos_load_theme},
    {"dc", set_debug_call},
    {"dm", set_debug_mode}
};

struct _CdosRunDialogPrivate
{
    ModalDialog    *dialog;
    GSettings      *lock_down_settings;
    GSettings      *terminal_settings;
    gboolean        enable_internal_commands;
    ClutterActor   *entry_text;
    StLabel        *errorMessage;
    ClutterActor   *errorBox;
    StLabel        *completionBox;
    StWidget       *label;
    StWidget       *entry;
    StWidget       *errorIcon;
};

G_DEFINE_TYPE_WITH_PRIVATE (CdosRunDialog, cdos_run_dialog, G_TYPE_OBJECT);

static gint cdos_run_dialog_run (const gchar *command);

static gboolean
cdos_run_dialog_key_press_event (ClutterActor *actor,
                                 ClutterEvent *event,
                                 gpointer      user_data);

static void
cdos_run_dialog_development_tools (GSettings *settings,
                                   gchar     *key,
                                   CdosRunDialog *user_data);
static CdosRunDialog *
cdos_run_dialog_new(void)
{
    return g_object_new(CDOS_TYPE_RUN_DIALOG, NULL);
}

static void
cdos_run_dialog_dispose (GObject *object)
{
    CdosRunDialog *self = CDOS_RUN_DIALOG (object);
    CdosRunDialogPrivate *priv = self->priv;

    g_clear_object (&priv->lock_down_settings);
    g_clear_object (&priv->terminal_settings);

    cdos_modal_dialog_destroy (priv->dialog);
    G_OBJECT_CLASS (cdos_run_dialog_parent_class)->dispose (object);
}

static void
cdos_run_dialog_finalize (GObject *object)
{
    G_OBJECT_CLASS (cdos_run_dialog_parent_class)->finalize (object);
}

static void
cdos_run_dialog_init (CdosRunDialog *dialog)
{
    CdosRunDialogPrivate *priv;
    priv = dialog->priv = cdos_run_dialog_get_instance_private (dialog);
    priv->dialog = g_malloc0 (sizeof (ModalDialog));

    priv->dialog = cdos_modal_dialog_init (priv->dialog, TRUE, "run-dialog");
    priv->lock_down_settings = g_settings_new (LOCKDOWN_SCHEMA);
    priv->terminal_settings = g_settings_new (TERMINAL_SCHEMA);

    GSettings *gloabl_settings = cdos_global_get_settings (cdos_global_get());
    g_signal_connect (gloabl_settings, "changed::development-tools",
            G_CALLBACK (cdos_run_dialog_development_tools),
            dialog);

    priv->enable_internal_commands = g_settings_get_boolean(gloabl_settings, "development-tools");

    priv->errorBox = g_object_new (ST_TYPE_BOX_LAYOUT, "style-class", "run-dialog-error-box", NULL);

    priv->errorMessage = g_object_new(ST_TYPE_LABEL, "style-class", "run-dialog-error-label", NULL);

    st_label_set_text(priv->errorMessage, _("This command can not run!"));

    priv->errorIcon = g_object_new(ST_TYPE_ICON,
                    "icon-name", "dialog-error",
                    "style-class", "run-dialog-error-icon",
                    "icon-size", 24,
                    NULL);

    priv->label = g_object_new(ST_TYPE_LABEL, "style-class", "run-dialog-label",
                    "text", _("Please enter a command"),
                    NULL);

    clutter_actor_add_child (priv->errorBox, CLUTTER_ACTOR(priv->errorIcon));
    clutter_actor_add_child (priv->errorBox, CLUTTER_ACTOR(priv->errorMessage));
    clutter_actor_hide (CLUTTER_ACTOR(priv->errorBox));

    priv->entry = g_object_new (ST_TYPE_ENTRY, "style-class", "run-dialog-entry", NULL);

    priv->entry_text = st_entry_get_clutter_text(ST_ENTRY(priv->entry));
    modal_dialog_set_initial_key_focus(priv->dialog, (ClutterActor *)priv->entry);

    clutter_actor_add_child ((ClutterActor *)priv->dialog->content_layout, CLUTTER_ACTOR(priv->label));
    clutter_actor_add_child ((ClutterActor *)priv->dialog->content_layout, CLUTTER_ACTOR(priv->entry));
    clutter_actor_add_child ((ClutterActor *)priv->dialog->content_layout, CLUTTER_ACTOR(priv->errorBox));

    g_signal_connect(priv->entry_text, "key-press-event",
             G_CALLBACK(cdos_run_dialog_key_press_event),
             dialog);
}

static gboolean
cdos_run_dialog_key_press_event (ClutterActor *actor,
                                 ClutterEvent *event,
                                 gpointer      user_data)
{
    CdosRunDialogPrivate *priv = CDOS_RUN_DIALOG(user_data)->priv;

    clutter_actor_hide(CLUTTER_ACTOR(priv->errorBox));
    guint symbol = clutter_event_get_key_symbol(event);
    if (symbol == CLUTTER_KEY_Return || symbol == CLUTTER_KEY_KP_Enter)
    {
        modal_dialog_pop_modal(priv->dialog, 0);
        const gchar *command = clutter_text_get_text(CLUTTER_TEXT(actor));
        if (!cdos_run_dialog_run(command))
        {
            clutter_text_set_text(CLUTTER_TEXT(actor), NULL);
            if (!modal_dialog_push_modal(priv->dialog, 0))
            {
                cdos_modal_dialog_close (priv->dialog, 0);
                return TRUE;
            }
            clutter_actor_show(CLUTTER_ACTOR(priv->errorBox));
            return TRUE;
        }

        cdos_modal_dialog_close(priv->dialog, 0);
        return TRUE;
    }

    if (symbol == CLUTTER_KEY_Escape)
    {
        cdos_modal_dialog_close(priv->dialog, 0);
        return TRUE;
    }

    return FALSE;
}

static void
cdos_run_dialog_class_init(CdosRunDialogClass *klass)
{
    GObjectClass *gobject_class = G_OBJECT_CLASS(klass);

    gobject_class->dispose = cdos_run_dialog_dispose;
    gobject_class->finalize = cdos_run_dialog_finalize;
}

static void
cdos_run_dialog_development_tools (GSettings *settings,
                                   gchar     *key,
                                   CdosRunDialog *user_data)
{
    CdosRunDialogPrivate *priv = user_data->priv;
    GSettings *gloabl_settings = cdos_global_get_settings(cdos_global_get());
    priv->enable_internal_commands = g_settings_get_boolean(gloabl_settings, "development-tools");
}

void
cdos_run_dialog_open(CdosRunDialog *dialog)
{
    g_return_if_fail(CDOS_IS_RUN_DIALOG(dialog));
    CdosRunDialogPrivate *priv = dialog->priv;

    clutter_text_set_text(CLUTTER_TEXT(priv->entry_text), NULL);
    if (g_settings_get_boolean(priv->lock_down_settings, DISABLE_COMMAND_LINE_KEY))
        return;

    cdos_modal_dialog_open(priv->dialog, 0);
}


static gint
cdos_run_dialog_run(const gchar *command_line)
{

    GPid pid;
    GError *error = NULL;
    gint argcp;
    gchar **argvp;
    gboolean success;

    gint i;
    for (i = 0; i < G_N_ELEMENTS (internal_cmds); i++)
    {
        if (g_strcmp0 (command_line, internal_cmds[i].cmd) == 0)
        {
            ICFunc f = internal_cmds[i].func;
            f ();
            return TRUE;
        }
    }

    g_shell_parse_argv (command_line, &argcp, &argvp, &error);
    if (error)
    {
        return 0;
    }

    success = g_spawn_async (NULL, argvp, NULL,  G_SPAWN_SEARCH_PATH |
                             G_SPAWN_STDOUT_TO_DEV_NULL | G_SPAWN_STDERR_TO_DEV_NULL,
                             NULL, NULL, &pid, &error);
    if (!success)
        return 0;
    return pid;


}

CdosRunDialog *
cdos_run_dialog_get_default(void)
{
    static CdosRunDialog *dialog = NULL;

    if (dialog == NULL)
        dialog = cdos_run_dialog_new();

    return dialog;
}
