/**
 * Copyright (c) 2016-2018 CPU and Fundamental Software Research Center, CAS
 *
 * This software is published by the license of CPU-OS License，you can use and
 * distibuted this software under this License. See CPU-OS Liense for more detail.
 *
 * You should have receive a copy of CPU-OS License. If not, please contact us
 * by email <support_os@cpu-os.ac.cn>
 *
 **/
/*
 *      CdosScreenSaver                 CdosUserVerifier
 *      -----> BeginVerification      ----->.
 *                                          |
 *              <------------------发送AskSecretQuery信号，循环等待
 *        --->AnswerQuery返回密码   --->  验证
 **/

#include <time.h>
#include <langinfo.h>
#include <fcntl.h>

#include <meta/boxes.h>
#include <meta/meta-cursor-tracker.h>
#include <meta/meta-idle-monitor.h>
#include <meta/meta-x11-display.h>

#include <act/act.h>

#include "cdos-global.h"
#include "cdos-background.h"
#include "cdos-layout-manager.h"
#include "cdos-session.h"
#include "cdos-start.h"
#include "cdos-popup-menu.h"
#include "cdos-popup-menu-manager.h"

#include "cdos-screen-saver.h"
#include "cdos-message-dialog.h"
#include "cdos-message-tray.h"
#include "cdos-login-manager.h"
#include "cdos-util.h"
#include <grp.h>
#include <cheese/cheese.h>
#include <cheese/cheese-camera.h>


typedef enum
{
    VERIFY_IDLE,        //空闲状态
    VERIFY_BEGIN,       //开始认证
    ANSWER_QUERY,       //返回密码
    VERIFYING,          //正在认证
    ACCT_LOCKED,        //帐号锁定
    VERIFY_END,         //认证结束
} VerifyState;

typedef enum
{
    CS_HIDE,
    CS_HIDING,
    CS_SHOWING,
    CS_SHOWN,
}ContentsState;

typedef enum
{
    IS_SCREENSAVER,
    IS_SCREENPROTECT,
    IS_NOTHING,
} ScreenSaverState;

typedef enum
{
    AUTH_PASSWORD,      //密码
    AUTH_FINGERPRINT,   //指纹
    AUTH_FINGERVEIN,    //指静脉
    AUTH_IRIS,          //虹膜
    AUTH_VOICEPRINT,    //声纹
    AUTH_FACE,          //人脸识别
    AUTH_UKEY           //UKey认证
} AuthType;

enum
{
    ACTIVE_CHANGED,
    PROTECT_ACTIVE_CHANGED,
    LAST_SIGNAL
};

static guint signals[LAST_SIGNAL] = {0};

struct _CdosScreenSaver
{
    StWidget parent_instance;

    gboolean is_active;
    gboolean is_modal;
    VerifyState state;
    ContentsState contents_state;

    GPid child_pid;
    guint watch_id;
    GDBusNodeInfo *dbus_node_info;
    CdosSessionPresence *presence;
    GDBusConnection *connection;
    GDBusConnection *system_connection;
    GDBusProxy *verify_proxy;
    GDBusProxy *session_proxy;
    GDBusProxy *bio_proxy;
    GDBusProxy *login1_proxy;

    GSettings *settings;

    ActUser *user;

    gchar *service;
    gchar *username;
    gchar *cached_password;

    gchar *cached_clipboard;

    CdosGlobal *global;
    ClutterActor *stage;
    int monitor;

    CdosLayoutManager *layout;
    CdosMessageTray *message_tray;

    ClutterActor *background;
    CdosBackgroundManager *background_manager;
    ClutterActor *background_protect;
    CdosBackgroundManager *background_manager_protect;
    ClutterActor *mask_actor;

    ClutterActor *contents;
    ClutterActor *arrow;
    ClutterActor *user_icon;
    ClutterActor *username_label;
    ClutterActor *prompt_entry;
    ClutterActor *prompt_label;
    ClutterActor *text_info_label;
    ClutterActor *message_label;
    ClutterActor *switch_user_button;
    ClutterActor *auth_type_button;
    ClutterActor *power_button;
    ClutterActor *button_box;
    StWidget     *entry_primary_icon;

    CdosPopupMenu *menu;
    CdosMessageDialog *dialog;

    ClutterActor *time_label;
    ClutterActor *date_label;
    ClutterActor *lunar_label;
    ClutterActor *motto_label;
    ClutterActor *calendar;

    gulong monitors_changed_id;
    gulong time_update_id;
    gulong user_changed_id;
    gulong source_added_id;

    guint prompt_lock_id;

    MetaIdleMonitor *idle_monitor;
    guint idle_watch_id;
    guint active_watch_id;
    guint timeout_hide_id;

    gboolean skip_unlock;
    gboolean verify_success;
    gboolean session_active;

    //gint fprint_status;
    gboolean auto_unlock;
    ScreenSaverState screensaver_state;
    gboolean is_protect_active;
    gulong motion_event_id;

    //是否使用方德统一认证
    gboolean use_bio_auth;
    ClutterActor *bio_face;
    StWidget *bio_info_box;
    StWidget *bio_type_label;
    StWidget *bio_selection_button;
    StWidget *bio_menu;
    AuthType auth_type;
    //双因子混合认证
    gint mix_auth;

    //是否支持多个生物认证进行切换
    gboolean bio_auth_can_switch;

    //是否显示密码明文
    gboolean is_show_password;

    //人脸识别
    CheeseCamera *cheese_camera;
    ClutterActor *camera_drawing;

    //生物认证动态图
    GdkPixbufAnimation *pixbuf_ani;
    GdkPixbufAnimationIter *ani_iter;
    GTimeVal animation_time;
    guint ani_id;

    guint try_again_id;
    guint regist_info_id;
};

G_DEFINE_TYPE (CdosScreenSaver, cdos_screen_saver, ST_TYPE_WIDGET);

static CdosScreenSaver *global_screensaver = NULL;

/* cdos-session 需要用到GetActive, ActiveChanged */

//"            <arg type=\"u\" name=\"return_id\" direction=\"out\" />"

#define SCREESAVER_DBUS_NAME    "org.freedesktop.ScreenSaver"
#define SCREESAVER_PATH         "/org/freedesktop/ScreenSaver"
#define SCREESAVER_INTERFACE    "org.freedesktop.ScreenSaver"

#define SCREEN_LOCK_NAME    "com.nfs.securitycenter"
#define SCREEN_LOCK_PATH    "/SecuritySetting/accountsetting"
#define SCREEN_LOCK_INTERFACE   "com.nfs.securitycenter.accountsettinginterface"

#define USER_FACE_SIZE 144

static const char *screensaver_dbus_xml =
"<node name=\"/org/freedesktop/ScreenSaver\">"
"    <interface name=\"org.freedesktop.ScreenSaver\">"
"        <method name=\"Lock\">"
"        </method>"
"        <method name=\"GetActive\">"
"            <arg type=\"b\" name=\"active\" direction=\"out\" />"
"        </method>"
"        <method name=\"SetActive\">"
"            <arg type=\"b\" name=\"value\" direction=\"in\" />"
"        </method>"
"        <method name=\"GetActiveTime\">"
"            <arg type=\"u\" name=\"value\" direction=\"out\" />"
"        </method>"
"        <method name=\"ProtectScreen\">"
"        </method>"
"        <method name=\"GetProtectActive\">"
"            <arg type=\"b\" name=\"active\" direction=\"out\" />"
"        </method>"
"        <signal name=\"ProtectActiveChanged\">"
"            <arg type=\"b\" name=\"new_value\" />"
"        </signal>"
"        <signal name=\"ActiveChanged\">"
"            <arg type=\"b\" name=\"new_value\" />"
"        </signal>"
"        <signal name=\"WakeUpScreen\" />"
"    </interface>"
"</node>";

static void handle_method_call (GDBusConnection *connection G_GNUC_UNUSED,
                                const gchar *sender,
                                const gchar *object_path G_GNUC_UNUSED,
                                const gchar *interface,
                                const gchar *method,
                                GVariant    *parameters,
                                GDBusMethodInvocation *invocation,
                                CdosScreenSaver *self);

static void active_changed (CdosScreenSaver *self, gboolean active);
static void protect_active_changed (CdosScreenSaver *self, gboolean active);
static void emulate_key_release_event (ClutterActor *actor,
                           guint key,
                           gboolean is_print);
static void verification_complete (CdosScreenSaver *self);
static void verification_failed (CdosScreenSaver *self);
static void answer_query (CdosScreenSaver *self);

static void show_calendar (CdosScreenSaver *self, gboolean animate);
static void hide_calendar (CdosScreenSaver *self);

static void on_cancel_clicked (StButton *btn,
                               gint      button,
                               CdosScreenSaver *self);

static gboolean user_in_nopasswdlogin_group(const char *name);
static void
update_bio_auth_shown (CdosScreenSaver *self);
static gboolean
use_auth_password (CdosScreenSaver *self);
static gboolean
use_auth_bio (CdosScreenSaver *self);
static gboolean 
bio_device_is_on_work_state (CdosScreenSaver *self, gchar *device_id);
static gboolean
cs_unlock_screen (CdosScreenSaver *self);
static void 
regist_verify_info_again (CdosScreenSaver *self, guint interval);
static void
draw_user_icon (CdosScreenSaver *self);

static const GDBusInterfaceVTable interface_vtable =
{
    (GDBusInterfaceMethodCallFunc)handle_method_call,
    NULL,
    NULL,
};

static void
set_verify_state (CdosScreenSaver *self, VerifyState state, gboolean skip)
{
    if (self->state == ACCT_LOCKED && !skip)
        return;
    self->state = state;
}

static void
user_verify_timeout_clear (CdosScreenSaver *self)
{
    if (self->try_again_id)
    {
        g_source_remove (self->try_again_id);
        self->try_again_id = 0;
    }
    if (self->regist_info_id)
    {
        g_source_remove (self->regist_info_id);
        self->regist_info_id = 0;
    }
}

/* 关闭人脸识别摄像头 */
static void
bio_camera_stop (CdosScreenSaver *self)
{
    if (self->camera_drawing)
    {
        clutter_actor_destroy (self->camera_drawing);
        self->camera_drawing = NULL;
    }
    if (self->cheese_camera)
    {
        cheese_camera_stop (self->cheese_camera);
        g_clear_object (&self->cheese_camera);
    }
    draw_user_icon (self);
}

/* 激活人脸识别摄像头 */
static void
bio_camera_play (CdosScreenSaver *self)
{
    if (self->auth_type == AUTH_FACE)
    {
        if (self->cheese_camera == NULL)
        {
            self->camera_drawing = clutter_actor_new ();
            clutter_actor_set_size (self->camera_drawing, USER_FACE_SIZE, USER_FACE_SIZE);
            self->cheese_camera = cheese_camera_new (self->camera_drawing, NULL, 320, 240);
            cheese_camera_setup (self->cheese_camera, NULL, NULL);
        }

        CheeseCameraDevice *camera_device = cheese_camera_get_selected_device (self->cheese_camera);
        if (camera_device)
        {
            g_message ("Selected camera: %s", cheese_camera_device_get_name (camera_device));
        } else {
            g_message ("Selected camera: None");
        }
        st_bin_set_child ((StBin *)self->user_icon, self->camera_drawing);
        cheese_camera_play (self->cheese_camera);
    } else {
        bio_camera_stop (self);
    }
}

/* 设置设备是否被选中 */
static gboolean 
bio_device_set_selected (CdosScreenSaver *self, gchar *device_id, gboolean select)
{
    GVariant *v;
    gint s = 0;
    GError *error = NULL;

    if (!self->bio_proxy || device_id == NULL)
        return FALSE;
    v = g_dbus_proxy_call_sync (self->bio_proxy,
                                "setDeviceSelectedById", 
                                g_variant_new ("(ss)", device_id, select ? "1" : "0"),
                                G_DBUS_CALL_FLAGS_NONE,
                                -1,
                                NULL, &error);
    if (error)
    {
        g_warning ("Set bio device selected failed:%s", error->message);
        g_clear_error (&error);
        return FALSE;
    }

    g_variant_get (v, "(i)", &s);
    g_variant_unref (v);
    g_message ("Set bio device selected: %d", s);
    return s == 0 ? TRUE : FALSE;
}

/* 判断生物认证是否激活 */
static gboolean 
bio_get_enabled (CdosScreenSaver *self)
{
    GVariant *v;
    gint s = 0;
    GError *error = NULL;

    if (!self->bio_proxy)
        return FALSE;
    v = g_dbus_proxy_call_sync (self->bio_proxy,
                                "getBioEnabled", 
                                g_variant_new ("(s)", self->username),
                                G_DBUS_CALL_FLAGS_NONE,
                                -1,
                                NULL, &error);
    if (error)
    {
        g_warning ("Get bio enabled failed:%s", error->message);
        g_clear_error (&error);
        return FALSE;
    }

    g_variant_get (v, "(i)", &s);
    g_variant_unref (v);
    g_message ("Bio enabled state:%d", s);

    return s > 0 ? TRUE : FALSE;
}

/* 获取生物认证的标识图 */
static gchar *
bio_get_device_type_face (CdosScreenSaver *self, gint device_type, gint state)
{
    GVariant *v;
    gchar *face = NULL;
    GError *error = NULL;

    if (!self->bio_proxy)
        return face;
    v = g_dbus_proxy_call_sync (self->bio_proxy,
                                "getDeviceFace", 
                                g_variant_new ("(ii)", device_type, state),
                                G_DBUS_CALL_FLAGS_NONE,
                                -1,
                                NULL, &error);
    if (error)
    {
        g_warning ("Get bio device face failed:%s", error->message);
        g_clear_error (&error);
        return face;
    }

    g_variant_get (v, "(s)", &face);
    g_variant_unref (v);

    return face;
}

/* 获取生物认证名称 */
static gchar *
bio_get_device_type_name (CdosScreenSaver *self, gint device_type)
{
    GVariant *v;
    gchar *name = NULL;
    GError *error = NULL;

    if (!self->bio_proxy)
        return name;
    v = g_dbus_proxy_call_sync (self->bio_proxy,
                                "getDeviceTypeName", 
                                g_variant_new ("(i)", device_type),
                                G_DBUS_CALL_FLAGS_NONE,
                                -1,
                                NULL, &error);
    if (error)
    {
        g_warning ("Get bio device type name failed:%s", error->message);
        g_clear_error (&error);
        return name;
    }

    g_variant_get (v, "(s)", &name);
    g_variant_unref (v);

    return name;
}

/* 获取用户支持的设备和认证类型 */
static gchar **
bio_get_user_supported_devices (CdosScreenSaver *self, const gchar *user_name)
{
    GVariant *v;
    gchar **devices = NULL;
    GError *error = NULL;

    if (!self->bio_proxy)
        return devices;
    v = g_dbus_proxy_call_sync (self->bio_proxy,
                                "getUserSupportedDevices", 
                                g_variant_new ("(s)", user_name),
                                G_DBUS_CALL_FLAGS_NONE,
                                -1,
                                NULL, &error);
    if (error)
    {
        g_warning ("Get bio user supported devices failed:%s", error->message);
        g_clear_error (&error);
        return devices;
    }

    GVariantIter *iter;
    gsize n;
    gchar *id;
    gint i = 0;

    g_variant_get (v, "(as)", &iter); 
    n = g_variant_iter_n_children (iter);
    if (n > 0)
    {
        devices = g_malloc0_n (n + 1, sizeof(char *));
        while (g_variant_iter_loop (iter, "s", &id))
        {
            gboolean work_state = bio_device_is_on_work_state (self, id);
            if (work_state)
            {
                devices[i] = g_strdup (id);
                i++;
            }
        }
        devices[i] = NULL;
    }

    g_variant_iter_free (iter);
    g_variant_unref (v);

    return devices;
}

/* 关闭设备 */
static gboolean 
bio_device_stop (CdosScreenSaver *self)
{
    GVariant *v;
    gint s = 0;
    GError *error = NULL;

    bio_camera_stop (self);

    if (!self->bio_proxy)
        return FALSE;
    v = g_dbus_proxy_call_sync (self->bio_proxy,
                                "stopBioVerify", 
                                NULL,
                                G_DBUS_CALL_FLAGS_NONE,
                                -1,
                                NULL, &error);
    if (error)
    {
        g_warning ("Stop bio device failed:%s", error->message);
        g_clear_error (&error);
        return FALSE;
    }

    g_variant_get (v, "(i)", &s);
    g_variant_unref (v);
    g_message ("Stop bio verify device: %d", s);

    return s > 0 ? TRUE : FALSE;
}

/* 获取设备是否处于工作状态 */
static gboolean 
bio_device_is_on_work_state (CdosScreenSaver *self, gchar *device_id)
{
    GVariant *v;
    gint s = 0;
    GError *error = NULL;

    if (!self->bio_proxy)
        return FALSE;
    v = g_dbus_proxy_call_sync (self->bio_proxy,
                                "getDeviceStatues", 
                                g_variant_new ("(s)", device_id),
                                G_DBUS_CALL_FLAGS_NONE,
                                -1,
                                NULL, &error);
    if (error)
    {
        g_warning ("Get bio device status failed:%s", error->message);
        g_clear_error (&error);
        return FALSE;
    }

    g_variant_get (v, "(i)", &s);
    g_variant_unref (v);
    if (s == 0)
        g_message ("Device %s is not on working state.", device_id);

    return s > 0 ? TRUE : FALSE;
}

/* 获取设备认证类型 */
static int 
bio_get_device_type (CdosScreenSaver *self, gchar *device_id)
{
    GVariant *v;
    gint s = 0;
    GError *error = NULL;

    if (!self->bio_proxy || !device_id)
        return 0;

    v = g_dbus_proxy_call_sync (self->bio_proxy,
                                "getDeviceType", 
                                g_variant_new ("(s)",
                                       device_id),
                                G_DBUS_CALL_FLAGS_NONE,
                                -1,
                                NULL, &error);
    if (error)
    {
        g_warning ("Get bio device type failed:%s", error->message);
        g_clear_error (&error);
        return 0;
    }

    g_variant_get (v, "(i)", &s);
    g_variant_unref (v);

    return s;
}

/* 判断设备是否被选择 */
static gboolean 
bio_device_is_selected (CdosScreenSaver *self, gchar *device_id) 
{
    GVariant *v;
    gint s = 0;
    GError *error = NULL;

    if (!self->bio_proxy)
        return FALSE;

    v = g_dbus_proxy_call_sync (self->bio_proxy,
                                "getDeviceSelected", 
                                g_variant_new ("(s)",
                                       device_id),
                                G_DBUS_CALL_FLAGS_NONE,
                                -1,
                                NULL, &error);
    if (error)
    {
        g_warning ("Get bio selected device failed:%s", error->message);
        g_clear_error (&error);
        return FALSE;
    }

    g_variant_get (v, "(i)", &s);
    g_variant_unref (v);

    return s > 0 ? TRUE : FALSE;
}

/* 获取默认的生物特征认证方式，如果为NULL,则用密码 */
static gchar *
bio_get_primary_device (CdosScreenSaver *self)
{
    gchar **devices = NULL;
    guint i, len;
    //首选设备
    gchar *primary_device = NULL;

    devices = bio_get_user_supported_devices (self, self->username);
    if (!devices)
        goto out;

    len = g_strv_length (devices);
    for (i = 0; i < len; i++)
    {
        gchar *device_id = devices[i];
        //gboolean work_state = bio_device_is_on_work_state (self, device_id);
        gboolean selected = bio_device_is_selected (self, device_id);
        if (selected)
        {
            primary_device = g_strdup (device_id);
            break;
        }
    }
    g_strfreev (devices);

out:
    g_message ("Find primary bio device, id is %s", primary_device);
    return primary_device; 
}

/* 判断是否有有效的设备 */
static gboolean
bio_check_user_supported_devices (CdosScreenSaver *self)
{
    gchar **devices = NULL;

    devices = bio_get_user_supported_devices (self, self->username);
    if (!devices)
    {
        g_message ("Don't find any valid bio devices, please check it");
        return FALSE;
    }
    g_strfreev (devices);
    return TRUE;
}

/**
 *  获取是否使用双因子认证(strict模式) 
 *  返回值：0-没有启用，1-打开（生物认证+密码认证），2-打开（密码认证+生物认证）
 */
static gint
bio_get_mix_auth_type (CdosScreenSaver *self)
{
    GVariant *v;
    gint s = 0;
    GError *error = NULL;

    if (!self->bio_proxy)
        return s;

    v = g_dbus_proxy_call_sync (self->bio_proxy,
                                "supportTwoFactorAuth", 
                                NULL,
                                G_DBUS_CALL_FLAGS_NONE,
                                -1,
                                NULL, &error);
    if (error)
    {
        g_warning ("Get support mix auth failed:%s", error->message);
        g_clear_error (&error);
        return s;
    }

    g_variant_get (v, "(i)", &s);
    g_variant_unref (v);

    return s;
}

/**
 *  生物特征识别相关信号 
 *  参数1：tf_type
 *  int型，用于标识双因子认证类型。
 *  0 表示单因子认证。当前版本在单因子认证时不会发送信号，此类型不会发生；
 *  1 表示认证顺序为 生物认证——>密码认证 ，当 tf_type=1 时，表示此信号在生物认证结束时发出，即将切换成其他类型的认证（密码认证）；
 *  2 表示认证顺序为 密码认证——>生物认证 ，当 tf_type=2 时，表示此信号在生物认证开始时发出，即其他类型的认证（密码认证）刚结束并切换成生物认证。
 *  参数2：bioauth_type
 *  int型，当前生物认证类型，与之前的生物认证类型定义保持一致。
 *  
 */
static void
bio_proxy_on_signal_cb (CdosScreenSaver *self,
                        char *sender_name,
                        char *signal_name,
                        GVariant   *parameters)
{
    if (g_strcmp0 (signal_name, "bioauthStateChanged") == 0)
    {
        gint tf_type, auth_type;
        const gchar *notice = NULL;
        g_variant_get (parameters, "(ii)", &tf_type, &auth_type);
        g_message ("Bioauth state changed,mix auth:%d, tf_type:%d, auth_type:%d", 
                    self->mix_auth, tf_type, auth_type); 
        if (!self->mix_auth)
            return;
        if (tf_type == 1)
        {
            //切换到密码认证
            use_auth_password (self);
            notice = _("Notice: you have passed with bio auth, please continue to use password auth.");
        } 
        else if (tf_type == 2)
        {
            //切换到生物特征认证
            use_auth_bio (self);
            notice = _("Notice: you have passed with password auth, please continue to use bio auth.");
        }
        st_label_set_text ((StLabel *)self->text_info_label, notice);
    }
}

/* 判断是否使用生物特征认证 */
static gboolean
should_use_bio_auth (CdosScreenSaver *self)
{
    GError *error = NULL;
    gboolean ret = FALSE;

    if (self->bio_proxy == NULL)
    {
        self->bio_proxy = g_dbus_proxy_new_for_bus_sync (G_BUS_TYPE_SYSTEM,
                                                0,
                                                NULL,
                                                "com.nfs.bioauth",
                                                "/com/nfs/bioauth",
                                                "com.nfs.bioauthInterface",
                                                NULL,
                                                &error);
        if (error)
        {
            g_warning ("Create nfs bioauth proxy faild: %s", error->message);
            g_error_free (error);
            goto out;
        }
        g_signal_connect_swapped (self->bio_proxy, 
                                  "g-signal", 
                                  G_CALLBACK (bio_proxy_on_signal_cb), 
                                  self);
    }

    //判断总开关是否打开，此处需要提供信号监听，但是接口内并没有信号
    if (bio_get_enabled (self) == FALSE)
        goto out;

    //判断当前用户支持的认证方式
    ret = bio_check_user_supported_devices (self);

out:
    self->use_bio_auth = ret;

    //判断是否是双因子混合认证？
    self->mix_auth = ret == TRUE ? bio_get_mix_auth_type (self) : 0;

    g_message ("Found screensaver auth should use bio auth: %d, mix auth: %d", self->use_bio_auth, self->mix_auth);

    return ret;
}

static void
remove_idle_monitor (CdosScreenSaver *self)
{
    if (self->idle_watch_id)
    {
        meta_idle_monitor_remove_watch (self->idle_monitor, self->idle_watch_id);
        self->idle_watch_id = 0;
    }
    if (self->active_watch_id)
    {
        meta_idle_monitor_remove_watch (self->idle_monitor, self->active_watch_id);
        self->active_watch_id = 0;
    }
    if (self->timeout_hide_id)
    {
        g_source_remove (self->timeout_hide_id);
        self->timeout_hide_id = 0;
    }
}

static void
active_monitor_cb (MetaIdleMonitor *monitor,
                   guint            watch_id,
                   CdosScreenSaver *self)
{
    self->active_watch_id = 0;
    g_source_remove (self->timeout_hide_id);
    self->timeout_hide_id = 0;
}

static gboolean
hide_contents (CdosScreenSaver *self)
{
    self->timeout_hide_id = 0;
    if (self->menu)
        cdos_popup_menu_close ((CdosPopupMenuBase *)self->menu, FALSE);
    on_cancel_clicked (NULL, 0, self);
    return FALSE;
}

static void
idle_monitor_cb (MetaIdleMonitor *monitor,
                 guint            watch_id,
                 CdosScreenSaver *self)
{
    //st_label_set_text ((StLabel *)self->message_label, _("Timeout"));
    self->timeout_hide_id = g_timeout_add_seconds (5, (GSourceFunc)hide_contents, self);
    self->active_watch_id = meta_idle_monitor_add_user_active_watch (self->idle_monitor,
                                                                     (MetaIdleMonitorWatchFunc)active_monitor_cb,
                                                                     self,
                                                                     NULL);
}

static gboolean
do_try_verify_once (CdosScreenSaver *self)
{
    g_debug ("%s:: contents state: %d", __func__, self->contents_state);
    if (self->contents_state == CS_SHOWN)
        cs_unlock_screen (self);
    self->try_again_id = 0;
    return FALSE;
}

/* 重新发起验证 */
static void 
try_verify_agin (CdosScreenSaver *self, guint interval)
{
    user_verify_timeout_clear (self);
    self->try_again_id = g_timeout_add (interval, (GSourceFunc) do_try_verify_once, self);
}

//更新帐号锁定状态
static void
update_account_lock_state (CdosScreenSaver *self, gboolean locked)
{
    g_message ("LockStateChanged: Account Locked = %d.", locked);
    if (locked)
    {
        set_verify_state (self, ACCT_LOCKED, TRUE);
    } else {
        set_verify_state (self, VERIFY_END, TRUE);
    }

    //根据当前认证状态更新UI显示
    if (self->auth_type > AUTH_PASSWORD && self->use_bio_auth)
    {
        use_auth_bio (self);
    } else {
        use_auth_password (self);
    }
}

static void
get_account_locked_done (GDBusProxy *proxy,
                         GAsyncResult *res,
                         CdosScreenSaver *self)
{
    GVariant *v;
    GError *error = NULL;
    gboolean locked = FALSE;

    v = g_dbus_proxy_call_finish (proxy, res, &error);
    if (error)
    {
        g_warning ("Get account lock state failed:%s", error->message);
        g_clear_error (&error);
        return;
    }
    if (v)
    {
        g_variant_get (v, "(b)", &locked);
        g_variant_unref (v);
    }
    g_message ("Get username locked state: %d", locked);
    update_account_lock_state (self, locked);

    //注册认证信息，为开始认证做准备
    regist_verify_info_again (self, 250);
}

//获取当前帐号是否被锁定
static void
check_account_lock_state (CdosScreenSaver *self)
{
    if (self->verify_proxy)
    {
        g_dbus_proxy_call (self->verify_proxy,
                           "GetAccountLocked",
                           NULL,
                           G_DBUS_CALL_FLAGS_NONE,
                           -1,
                           NULL,
                           (GAsyncReadyCallback) get_account_locked_done,
                           self);
    }
}

//更新输入框
static void
update_prompt_entry (CdosScreenSaver *self, gboolean focus_entry)
{
    //如果是密文显示，输错误密码后，直接清掉;明文显示时，不清除，以便用户看到错误之处。
    if (self->auth_type == AUTH_PASSWORD || self->auth_type == AUTH_UKEY)
    {
        if (!self->is_show_password)
            st_entry_set_text ((StEntry *)self->prompt_entry, NULL);

        if (focus_entry)
            clutter_actor_grab_key_focus ((ClutterActor *)self->prompt_entry);
    }
}

static void
on_verify_signal (GDBusProxy      *proxy,
                  const gchar     *sender,
                  const gchar     *signal,
                  GVariant        *variant,
                  CdosScreenSaver *self)
{
    g_debug ("%s: sender(%s) signal(%s)", __func__, sender, signal);
    self->verify_success = FALSE;
    if (g_strcmp0 (signal, "Message") == 0)
    {
        char *message = NULL;
        g_variant_get (variant, "(s)", &message);

        //Ukey认证时，主动聚焦到输入框，先设置焦点，再显示内容。
        /* if (self->auth_type == AUTH_UKEY && self->cached_password == NULL)
        {
            clutter_actor_grab_key_focus ((ClutterActor *)self->prompt_entry);
        } else {
            clutter_actor_grab_key_focus ((ClutterActor *)self);
        } */
        clutter_actor_grab_key_focus ((ClutterActor *)self);

        update_prompt_entry (self, TRUE);

        g_debug ("pam auth message: %s", message);
        st_label_set_text ((StLabel *)self->message_label, message);
        //self->skip_unlock = TRUE;
        g_free (message);
    }
    else if (g_strcmp0 (signal, "Error") == 0)
    {
        char *message = NULL;
        g_variant_get (variant, "(s)", &message);
        g_warning ("Auth error:%s", message);
        st_label_set_text ((StLabel *)self->message_label, _("Unkown error, please try angin."));
        g_free (message);
        self->skip_unlock = TRUE;
        clutter_actor_grab_key_focus ((ClutterActor *)self);
        regist_verify_info_again (self, 300);
    }
    else if (g_strcmp0 (signal, "SecretInfoQuery") == 0)
    {
        self->skip_unlock = FALSE;
        //st_label_set_text ((StLabel *)self->text_info_label, "");
        set_verify_state (self, ANSWER_QUERY, FALSE);
        answer_query (self);
    }
    else if (g_strcmp0 (signal, "InfoQuery") == 0)
    {
        self->skip_unlock = FALSE;
        char *message = NULL;
        g_variant_get (variant, "(s)", &message);
        //st_label_set_text ((StLabel *)self->text_info_label, "");
        set_verify_state (self, ANSWER_QUERY, FALSE);
        g_message ("Pam asked: %s", message);
        g_free (message);

        //当生物识别认证第一次认证失败后，会询问：1-再次验证，2-取消;此处选择2
        if (self->auth_type > AUTH_PASSWORD && self->auth_type != AUTH_UKEY)
        {
            g_clear_pointer (&self->cached_password, g_free);
            self->cached_password = g_strdup ("2");
        }
        answer_query (self);
    }
    else if (g_strcmp0 (signal, "VerificationFailed") == 0)
    {
        char *message = NULL;
        g_variant_get (variant, "(s)", &message);
        g_warning ("Pam auth failed: %s", message);
        gboolean is_message = (g_strcmp0 (message, "Timeout") == 0 ||
                               g_strcmp0 (message, "Canceled") == 0);
        if (is_message)
        {
            if (g_strcmp0 (message, "Timeout"))
                st_label_set_text ((StLabel *)self->message_label, _(message));
        }
        verification_failed (self);
        g_free (message);
    }
    else if (g_strcmp0 (signal, "LockStateChanged") == 0)
    {
        gboolean locked;
        g_variant_get (variant, "(b)", &locked);
        update_account_lock_state (self, locked);
        //2秒后，再次发起认证
        try_verify_agin (self, 2000);
    }
    else if (g_strcmp0 (signal, "VerificationComplete") == 0)
    {
        if (self->skip_unlock)
        {
            self->verify_success = TRUE;
            return;
        }
        else
            verification_complete (self);
    }
}

static void
verification_failed (CdosScreenSaver *self)
{
    set_verify_state (self, VERIFY_END, FALSE);
    g_clear_pointer (&self->cached_password, g_free);

    //3秒后，再次发起认证
    if (self->auth_type > AUTH_PASSWORD && self->auth_type != AUTH_UKEY)
    {
        try_verify_agin (self, 3000);
    }
}

static void
verification_complete (CdosScreenSaver *self)
{
    CdosLayoutManager *layout;
    if (!self->is_modal)
        return;
    g_message ("Screensaver verify completed, unlock it.");
    cdos_pop_modal ((ClutterActor *)self, 0);
    self->is_modal = FALSE;
    self->skip_unlock = FALSE;
    self->verify_success = FALSE;
    self->auto_unlock = FALSE;

    layout = cdos_layout_manager_get_default ();
    cdos_layout_manager_remove_chrome (layout, CLUTTER_ACTOR (self));
    clutter_actor_hide (CLUTTER_ACTOR (self));
    clutter_actor_hide (self->contents);
    //st_label_set_text ((StLabel *)self->text_info_label, "");
    self->contents_state = CS_HIDE;
    st_entry_set_text ((StEntry *)self->prompt_entry, NULL);

    bio_camera_stop (self);

    /* ClutterColor color = {0, 0, 0, 90};
    clutter_actor_set_background_color (self->mask_actor, &color); */

    active_changed (self, FALSE);

    if (self->motion_event_id > 0)
    {
        g_signal_handler_disconnect (self, self->motion_event_id);
        self->motion_event_id = 0;
    }
    protect_active_changed (self, FALSE);
    self->screensaver_state = IS_NOTHING;

    //meta_cursor_tracker_set_pointer_visible (meta_cursor_tracker_get_for_display (NULL), TRUE);
    g_clear_pointer (&self->cached_password, g_free);
    if (self->cached_clipboard)
        st_clipboard_set_text (st_clipboard_get_default (), ST_CLIPBOARD_TYPE_CLIPBOARD, self->cached_clipboard);
    g_clear_pointer (&self->cached_clipboard, g_free);
    set_verify_state (self, VERIFY_END, FALSE);
    remove_idle_monitor (self);
}

/* force:是否强制退出。如果用户锁定则不进行强制退出，等待倒计时结束 */
static gboolean
user_verify_cancel (CdosScreenSaver *self, gboolean force)
{
    gboolean canceled = FALSE;
    GError *error = NULL;
    GVariant *v;
    if (self->verify_proxy)
    {
        v = g_dbus_proxy_call_sync (self->verify_proxy,
                                    "Cancel",
                                    g_variant_new ("(b)", force),
                                    G_DBUS_CALL_FLAGS_NONE,
                                    -1,
                                    NULL,
                                    &error);
        if (error)
        {
            g_warning ("User verify canceled failed: %s", error->message);
            g_clear_error (&error);
        }
        if (v)
        {
            g_variant_get (v, "(b)", &canceled);
            g_variant_unref (v);
        }

        //如果正在使用生物特征认证，则退出后停止设备，防止设备提示不正确
        if (self->use_bio_auth)
        {
            bio_device_stop (self);
        }
    }
    return canceled;
}

static void
begin_verification_done (GDBusProxy *proxy,
                         GAsyncResult *res,
                         CdosScreenSaver *self)
{
    GVariant *v;
    GError *error = NULL;

    v = g_dbus_proxy_call_finish (proxy, res, &error);
    if (error)
    {
        g_warning ("Begin verify failed: %s", error->message);
        st_label_set_text ((StLabel *)self->text_info_label, _("Unkown error, please try angin."));
        clutter_actor_show (self->text_info_label);
        g_clear_error (&error);
    }
    if (v)
    {
        gboolean result = FALSE;
        g_variant_get (v, "(b)", &result);
        if (result == FALSE)
        {
            g_message ("Verify failed, some info missming, you should regist agin.");
            regist_verify_info_again (self, 300);
        }
        g_variant_unref (v);
    }
}

static void
answer_query_done (GDBusProxy *proxy,
                   GAsyncResult *res,
                   CdosScreenSaver *self)
{
    GVariant *v;
    GError *error = NULL;

    v = g_dbus_proxy_call_finish (proxy, res, &error);
    if (error)
    {
        g_warning ("Answer query failed: %s", error->message);
        g_clear_error (&error);
        set_verify_state (self, VERIFY_END, FALSE);
        return;
    }

    set_verify_state (self, VERIFYING, FALSE);
    if (v)
        g_variant_unref (v);
}

static void
answer_query (CdosScreenSaver *self)
{
    if (self->cached_password)
    {
        g_dbus_proxy_call (self->verify_proxy,
                        "AnswerQuery",
                        g_variant_new ("(s)",
                                        self->cached_password),
                        G_DBUS_CALL_FLAGS_NONE,
                        -1,
                        NULL,
                        (GAsyncReadyCallback) answer_query_done,
                        self);
    }
}

/*
 * 缓存剪切板内容。
 * 将剪切板清空，防止剪切板被粘贴到密码框
 */
static void
cache_clipboard_text (StClipboard *clipboard,
                      const gchar *text,
                      CdosScreenSaver *self)
{
    st_clipboard_set_text (clipboard, ST_CLIPBOARD_TYPE_CLIPBOARD, "");

    if (text)
        self->cached_clipboard = g_strdup (text);
}

static void
on_source_added (CdosMessageTray *message_tray, CdosSource *source, CdosScreenSaver *self)
{
    if (self->is_active)
    {
        GError *error = NULL;
        g_dbus_connection_emit_signal (self->connection,
                                       SCREESAVER_DBUS_NAME,
                                       SCREESAVER_PATH,
                                       SCREESAVER_INTERFACE,
                                       "WakeUpScreen",
                                       NULL,
                                       &error);
        if (error)
        {
            g_warning ("Emit WakeUpScreen: %s", error->message);
            g_error_free (error);
        }
    }
}

static void
send_escape_event (CdosScreenSaver *self)
{
    XKeyEvent xkevent;
    Display *xdisplay;
    MetaX11Display *x11display;
    Window xwindow;
    Window xrootwindow;

    guint16 keycode = 9;

    MetaDisplay *display;

    display = cdos_global_get_display (global);
    x11display = meta_display_get_x11_display (display);
    xrootwindow = clutter_x11_get_root_window ();
    xdisplay = meta_x11_display_get_xdisplay (x11display);
    xwindow = xrootwindow;

    int revert;
    XGetInputFocus(xdisplay, &xwindow, &revert);
    g_debug ("%s 0x%lx", __func__, xwindow);
    {
      xkevent.display = xdisplay;
      xkevent.window = xwindow;
      xkevent.root = xrootwindow;
      xkevent.subwindow = None;
      xkevent.time = CurrentTime;
      xkevent.x = 1;
      xkevent.y = 1;
      xkevent.x_root = 1;
      xkevent.y_root = 1;
      xkevent.state = 0;
      xkevent.same_screen = True;
      xkevent.keycode = keycode;

      xkevent.type = KeyPress;
      XSendEvent (xdisplay, xwindow, False, 0, (XEvent *)&xkevent);
      XFlush (xdisplay);

      /* If the application takes a grab on KeyPress, we don't
       * want to send it a KeyRelease. There's no good way of
       * knowing whether a tray icon will take a grab, so just
       * assume it does, and don't send the KeyRelease. That might
       * make the tracking for key events messed up if it doesn't take
       * a grab, but the tray icon won't get key focus in normal cases,
       * so let's hope this isn't too damaging...
       */
      xkevent.type = KeyRelease;
      XSendEvent (xdisplay, xwindow, False, 0, (XEvent *)&xkevent);
      XFlush (xdisplay);
    }
}

static gboolean
cdos_screen_saver_motion_event (ClutterActor *actor,
                        ClutterEvent *event,
                        CdosScreenSaver *data)
{
    CdosScreenSaver *self = CDOS_SCREEN_SAVER (actor);
    ClutterEventType event_type = clutter_event_type (event);

    if (event_type == CLUTTER_MOTION && self->is_protect_active) 
    {
        protect_active_changed (self, FALSE);
        clutter_actor_hide (self->background_protect);
        clutter_actor_show (self->background);     
        emulate_key_release_event (actor, CLUTTER_KEY_KP_Enter, FALSE);
    }
    return FALSE;
}

static gboolean
cdos_screen_saver_idle_connect_signal (gpointer data)
{
    CdosScreenSaver *self = CDOS_SCREEN_SAVER (data);

    if (self->motion_event_id > 0)
    {
        g_signal_handler_disconnect (self, self->motion_event_id);
        self->motion_event_id = 0;
    }
    self->motion_event_id = g_signal_connect (self, "motion-event", G_CALLBACK (cdos_screen_saver_motion_event), self);

    return FALSE;
}

/* 如果锁屏->切换用户->登陆，则自动解除锁屏 */
static gboolean
should_lock_screen (CdosScreenSaver *self)
{
    gboolean should_lock = TRUE;
    gchar *filename = g_build_filename ("/tmp/lightdm/authtime/", self->username, NULL);
    gchar *authtime_str = NULL;
    gsize len;
    gint64 authtime, current_time;
    GError *error = NULL;
    GDateTime *dt;
    gint64 delta;

    if (g_file_test (filename, G_FILE_TEST_EXISTS) == FALSE)
    {
        g_warning ("Can't find authtime file:%s", filename);
        goto out;
    }

    if (g_file_get_contents (filename, &authtime_str, &len, &error) == FALSE)
    {
        g_warning ("Get authtime from file faild:%s", error->message);
        g_error_free (error);
        goto out;
    }

    sscanf (authtime_str, "%ld", &authtime);
    dt = g_date_time_new_now_local ();
    current_time = g_date_time_to_unix (dt);
    delta = current_time - authtime;
    g_message ("Get session authtime: %ld, current: %ld, delta: %ld", authtime, current_time, delta);

    //10秒内存在登陆认证则不自动锁屏，防止重复输入密码。
    if (ABS(delta) < 10)
        should_lock = FALSE;

    g_date_time_unref (dt);
    g_free (authtime_str);
out:
    g_free (filename);
    return should_lock;
}

static void
cs_lock_screen (CdosScreenSaver *self)
{   
    gboolean should_lock = should_lock_screen (self);
    g_debug ("%s::==>> session is active = %d, username = %s, should_lock = %d", __func__, self->session_active, self->username, should_lock);
    if (!self->session_active || !should_lock)
        return;

    if (self->is_active || self->is_protect_active)
    {
        // 如果此时“屏保"已激活，又要去激活“锁屏”时，依旧维持”屏保“界面，但 self->is_active 需要将其改成true，之后走解锁的认证逻辑。
        if (self->is_protect_active && self->screensaver_state == IS_SCREENSAVER)
            active_changed (self, TRUE);
        
        // 如果此时“锁屏”已激活，不能切换成“屏保”，之后要走解锁的认证逻辑的。但背景换成屏保的
        if (self->is_active && self->screensaver_state == IS_SCREENPROTECT)
        {
            self->screensaver_state = IS_SCREENSAVER;
            protect_active_changed (self, TRUE);
            clutter_actor_show (self->background_protect);
            clutter_actor_hide (self->background);
        }
        return;
    }
    g_message ("Screensaver will lock.");
    ClutterActor *actor = CLUTTER_ACTOR (self);
    CdosLayoutManager *layout;

    if (self->screensaver_state == IS_SCREENSAVER)
    {
        clutter_actor_show (self->background);
        clutter_actor_hide (self->background_protect);
    }
    else if (self->screensaver_state == IS_SCREENPROTECT)
    {
        clutter_actor_show (self->background_protect);
        clutter_actor_hide (self->background);
    }

    self->is_modal = cdos_push_modal (actor, 0, 0, CDOS_ACTION_MODE_LOCK_SCREEN);
    if (!self->is_modal)
    {
        self->is_modal = cdos_push_modal (actor,
                                          0,
                                          META_MODAL_POINTER_ALREADY_GRABBED,
                                          CDOS_ACTION_MODE_LOCK_SCREEN);
    }
    if (!self->is_modal)
    {
        if (clutter_check_windowing_backend (CLUTTER_WINDOWING_X11))
        {
            send_escape_event (self);
            sleep(1);
            self->is_modal = cdos_push_modal (actor, 0, 0, CDOS_ACTION_MODE_LOCK_SCREEN);
        }
        /*
        cdos_notify_error (_("Unable lock screen"),
                           _("Lock was blocked by another application."));
        */
        if (!self->is_modal)
            return;
    }
    //meta_cursor_tracker_set_pointer_visible (meta_cursor_tracker_get_for_display (NULL), FALSE);

    clutter_actor_show (actor);
    layout = cdos_layout_manager_get_default ();
    cdos_layout_manager_add_chrome (layout, actor, AFFECT_INPUT_REGION | ABOVE_TOP_WINDOW_GROUP);
    show_calendar (self, FALSE);

    if (!clutter_actor_has_key_focus (actor))
        clutter_stage_set_key_focus ((ClutterStage *)self->stage, actor);

    // 延时 0.5秒 绑定”鼠标移动“信号
    g_timeout_add (500, cdos_screen_saver_idle_connect_signal, self);
    if (self->screensaver_state == IS_SCREENSAVER)
        active_changed (self, TRUE);
    else if (self->screensaver_state == IS_SCREENPROTECT)
        protect_active_changed (self, TRUE);

    StClipboard *clipboard;
    clipboard = st_clipboard_get_default ();
    st_clipboard_get_text (clipboard, ST_CLIPBOARD_TYPE_CLIPBOARD, (StClipboardCallbackFunc) cache_clipboard_text, self);
}

static gboolean
cs_unlock_screen (CdosScreenSaver *self)
{
    GError *error = NULL;
    if (!self->is_active)
        return TRUE;

    if (self->prompt_lock_id)
        return FALSE;

    set_verify_state (self, VERIFY_BEGIN, FALSE);
    g_message ("Unlock screensaver begin, auth type = %d", self->auth_type);
    /* if (self->state != ACCT_LOCKED)
        st_label_set_text ((StLabel *)self->message_label, ""); */
    bio_camera_play (self);
    g_dbus_proxy_call (self->verify_proxy,
                       "BeginVerification", NULL,
                       G_DBUS_CALL_FLAGS_NONE,
                       -1,
                       NULL,
                       (GAsyncReadyCallback) begin_verification_done,
                       self);
    //if (self->cached_password)
    //return TRUE;
    return TRUE;
}

static void
handle_method_call (GDBusConnection *connection G_GNUC_UNUSED,
                    const gchar *sender,
                    const gchar *object_path G_GNUC_UNUSED,
                    const gchar *interface,
                    const gchar *method,
                    GVariant    *parameters,
                    GDBusMethodInvocation *invocation,
                    CdosScreenSaver *self)
{
    g_debug ("%s: %s %s", __func__, interface, method);

    if (g_strcmp0 (method, "Lock") == 0)
    {
        self->screensaver_state = IS_SCREENSAVER;
        cs_lock_screen (self);
        g_dbus_method_invocation_return_value (invocation, NULL);
    }
    else if (g_strcmp0 (method, "GetActive") == 0)
    {
        g_dbus_method_invocation_return_value (invocation, g_variant_new ("(b)", self->is_active));
    }
    else if (g_strcmp0 (method, "SetActive") == 0)
    {
        //This method do nothing.
        g_dbus_method_invocation_return_value (invocation, NULL);
        /* gboolean is_active;
        g_variant_get (parameters, "(b)", &is_active);
        active_changed (self, is_active); */
    }
    else if (g_strcmp0 (method, "ProtectScreen") == 0)
    {
        self->screensaver_state = IS_SCREENPROTECT;
        cs_lock_screen (self);
        g_dbus_method_invocation_return_value (invocation, NULL);
    }
    else if (g_strcmp0 (method, "GetProtectActive") == 0)
    {
        g_dbus_method_invocation_return_value (invocation, g_variant_new ("(b)", self->is_protect_active));
    }
}

static void
protect_active_changed (CdosScreenSaver *self, gboolean active)
{
    if (self->is_protect_active != active)
    {
        GError *error = NULL;
        self->is_protect_active = active;
        g_signal_emit (self, signals[PROTECT_ACTIVE_CHANGED], 0, active);
        g_dbus_connection_emit_signal (self->connection,
                                       NULL,
                                       SCREESAVER_PATH,
                                       SCREESAVER_INTERFACE,
                                       "ProtectActiveChanged",
                                       g_variant_new ("(b)", active),
                                       &error);
        if (error)
        {
            g_warning ("%s", error->message);
            g_error_free (error);
        }
    }
 }

static void
active_changed (CdosScreenSaver *self, gboolean active)
{
    if (self->is_active != active)
    {
        GError *error = NULL;
        self->is_active = active;
        g_signal_emit (self, signals[ACTIVE_CHANGED], 0, active);
        g_dbus_connection_emit_signal (self->connection,
                                       NULL,
                                       SCREESAVER_PATH,
                                       SCREESAVER_INTERFACE,
                                       "ActiveChanged",
                                       g_variant_new ("(b)", active),
                                       &error);
        if (error)
        {
            g_warning ("%s", error->message);
            g_error_free (error);
        }
    }
}

static void
on_screensaver_dbus_acquried (GDBusConnection *conn,
                              const gchar *name,
                              CdosScreenSaver *self)
{
    GError *error = NULL;
    if (!g_dbus_connection_register_object (conn,
                                            SCREESAVER_PATH,
                                            self->dbus_node_info->interfaces[0],
                                            &interface_vtable,
                                            self,
                                            NULL,
                                            &error))
    {
        g_warning ("%s: %s", __func__, error->message);
        g_error_free (error);
        return;
    }
}

static gboolean
is_cdrom (void)
{
    gboolean ret = FALSE;
    int fd = 0;
    gchar buffer[80]= {0};
    const gchar *file = "/proc/cmdline";
    fd = open (file, O_RDONLY);
    if (fd > 0)
    {
        int sz = read (fd, (void *)buffer, 80);
        if (sz > 0 && strstr (buffer, "file=/cdrom"))
        {
            ret = TRUE;
        }
    }
    close (fd);

    return ret;
}

static void
on_presence_status_changed (CdosSessionPresence *presence,
                            guint status,
                            CdosScreenSaver *self)
{
    g_debug ("%s:%s %d", __FILE__, __func__, status);

    if (status == CDOS_SESSION_IDLE)
    {
        if (is_cdrom ())
            return;

        self->screensaver_state = IS_SCREENSAVER;
        cs_lock_screen (self);
    }
    else
    {
        if (self->is_active && !clutter_actor_is_visible (self->contents))
            clutter_actor_grab_key_focus ((ClutterActor *)self);
    }
}

static void
on_contents_hidden (ClutterTimeline *tml,
                    CdosScreenSaver *self)
{
    remove_idle_monitor (self);
    clutter_actor_hide (self->contents);
    self->contents_state = CS_HIDE;
    g_signal_handlers_disconnect_by_data (tml, self);
    g_clear_pointer (&self->cached_password, g_free);
    st_entry_set_text ((StEntry *)self->prompt_entry, NULL);
    //st_label_set_text ((StLabel *)self->text_info_label, "");

    //meta_cursor_tracker_set_pointer_visible (meta_cursor_tracker_get_for_display (NULL), FALSE);
    clutter_actor_grab_key_focus ((ClutterActor *)self);

    if (self->dialog)
        clutter_actor_hide ((ClutterActor *)self->dialog);
    bio_camera_stop (self);
}

static void
on_calendar_hidden (ClutterTimeline *tml,
                    CdosScreenSaver *self)
{
    clutter_actor_hide (self->calendar);
    g_signal_handlers_disconnect_by_data (tml, self);
}

static void
hide_calendar (CdosScreenSaver *self)
{
    MetaRectangle geometry;
    gfloat x, y;
    gfloat nat_w, nat_h;
    meta_display_get_monitor_geometry (cdos_global_get_display (self->global),
                                      self->monitor,
                                      &geometry);
    clutter_actor_get_preferred_size (self->calendar, NULL, NULL, &nat_w, &nat_h);

    clutter_actor_set_opacity (self->calendar, 255);
    x = geometry.x + (geometry.width - nat_w) / 2;
    y = geometry.y - nat_h - 100;
    ClutterTransition *transition = cdos_transition_simple (self->calendar,
                                                            CLUTTER_EASE_OUT,
                                                            250,
                                                            "opacity", 0,
                                                            "x", x,
                                                            "y", y,
                                                            NULL);
    g_signal_connect (transition, "completed", G_CALLBACK (on_calendar_hidden), self);
}

static const gchar *
get_date_format (void)
{
    const gchar *format;
    const gchar *lc_time;
    lc_time = g_getenv ("LC_TIME"); 
    g_debug ("%s:: lc_time = %s", __func__, lc_time);
    if (g_strcmp0 (lc_time, "zh_CN.UTF-8") == 0)
        format = "%m月%d日 %A";
    else 
        format = "%m/%d  %A";
    return format; 
}

static gboolean
update_time (CdosScreenSaver *self)
{
    GDateTime *now = g_date_time_new_now_local ();
    gchar *time = g_date_time_format (now, cdos_get_time_format ());
    gchar *date = g_date_time_format (now, get_date_format ());
    //gchar *date = g_date_time_format (now, "%b%e  %A");
    st_label_set_text ((StLabel *)self->time_label, time);
    st_label_set_text ((StLabel *)self->date_label, date);

    g_free (time);
    g_free (date);
    g_date_time_unref (now);

    return TRUE;
}

static void
show_calendar (CdosScreenSaver *self, gboolean animate)
{
    GDateTime *now = g_date_time_new_now_local ();
    gchar *time = g_date_time_format (now, cdos_get_time_format ());
    //gchar *date = g_date_time_format (now, "%x  %A");
    gchar *date = g_date_time_format (now, get_date_format ());
    st_label_set_text ((StLabel *)self->time_label, time);
    st_label_set_text ((StLabel *)self->date_label, date);

    g_free (time);
    g_free (date);
    g_date_time_unref (now);

    self->time_update_id = g_timeout_add_seconds (1, (GSourceFunc) update_time, self);

    MetaRectangle geometry;
    gfloat orig_x, orig_y, dest_x, dest_y;
    gfloat nat_w, nat_h;
    meta_display_get_monitor_geometry (cdos_global_get_display (self->global),
                                       self->monitor,
                                       &geometry);
    clutter_actor_get_preferred_size (self->calendar, NULL, NULL, &nat_w, &nat_h);

    dest_x = geometry.x + (geometry.width - nat_w) / 2;
    dest_y = geometry.y + 230;
    if (!animate)
    {
        clutter_actor_set_position (self->calendar, dest_x, dest_y);
        clutter_actor_set_opacity (self->calendar, 255);
        clutter_actor_show (self->calendar);
        return;
    }

    orig_x = dest_x;
    orig_y = geometry.y - nat_h - 100;
    clutter_actor_set_position (self->calendar, orig_x, orig_y);
    clutter_actor_set_opacity (self->calendar, 0);
    clutter_actor_show (self->calendar);
    cdos_transition_simple (self->calendar,
                            CLUTTER_EASE_IN,
                            250,
                            "opacity", 255,
                            "x", dest_x,
                            "y", dest_y,
                            NULL);
}

static void
real_logout (void)
{
    cdos_session_logout (2);
}

static void
do_logout (CdosPopupBaseMenuItem *item,
        ClutterEvent *event,
        gboolean keep_menu,
        CdosScreenSaver *self)
{
    on_cancel_clicked (NULL, 0, self);
    if (cdos_session_get_inhibited_actions () ^ GSM_INHIBITOR_FLAG_LOGOUT)
    {
        real_logout ();
        return;
    }
    cdos_message_dialog_set_title (self->dialog, _("Log out"));
    cdos_message_dialog_set_message (self->dialog,
                                     _("The current user has unsaved data."
                                       "Forced logout may cause the user's data to be lost."
                                       "\nAre you sure to continue logout?"));
    cdos_message_dialog_set_callback (self->dialog, (GCallback)real_logout, NULL);
    clutter_actor_set_child_above_sibling ((ClutterActor *)self, (ClutterActor *)self->dialog, NULL);
    clutter_actor_show ((ClutterActor *)self->dialog);
}

static void
real_shutdown (void)
{
    cdos_login_manager_power_off (cdos_login_manager_new (), FALSE);
}

static void
do_shutdown (CdosPopupBaseMenuItem *item,
          ClutterEvent *event,
          gboolean keep_menu,
          CdosScreenSaver *self)
{
    on_cancel_clicked (NULL, 0, self);
    if (cdos_session_get_inhibited_actions () ^ GSM_INHIBITOR_FLAG_LOGOUT)
    {
        real_shutdown ();
        return;
    }
    cdos_message_dialog_set_title (self->dialog, _("Shutdown"));
    cdos_message_dialog_set_message (self->dialog,
                                     _("The current user has unsaved data."
                                       "Forced shutdown may cause the user's data to be lost."
                                       "\nAre you sure to continue shutdown?"));
    cdos_message_dialog_set_callback (self->dialog, (GCallback)real_shutdown, NULL);
    clutter_actor_set_child_above_sibling ((ClutterActor *)self, (ClutterActor *)self->dialog, NULL);
    clutter_actor_show ((ClutterActor *)self->dialog);
    clutter_actor_grab_key_focus ((ClutterActor *)self->dialog);
}

static void
real_reboot (void)
{
    cdos_login_manager_reboot (cdos_login_manager_new (), FALSE);
}

static void
do_reboot (CdosPopupBaseMenuItem *item,
        ClutterEvent *event,
        gboolean keep_menu,
        CdosScreenSaver *self)
{
    on_cancel_clicked (NULL, 0, self);
    if (cdos_session_get_inhibited_actions () ^ GSM_INHIBITOR_FLAG_LOGOUT)
    {
        real_reboot ();
        return;
    }
    cdos_message_dialog_set_title (self->dialog, _("Reboot"));
    cdos_message_dialog_set_message (self->dialog,
                                     _("The current user has unsaved data."
                                       "Forced reboot may cause the user's data to be lost."
                                       "\nAre you sure to continue reboot?"));
    cdos_message_dialog_set_callback (self->dialog, (GCallback)real_reboot, NULL);
    clutter_actor_set_child_above_sibling ((ClutterActor *)self, (ClutterActor *)self->dialog, NULL);
    clutter_actor_show ((ClutterActor *)self->dialog);
    clutter_actor_grab_key_focus ((ClutterActor *)self->dialog);
}

static void
real_hibernate (void)
{
    pid_t status1, status2;
    status1 = system("sed \"s/yes/no/g\" /etc/cdos-fastboot/cdos-fastboot.conf > /tmp/cdos-fastboot-tmp");
    status2 = system("cat /tmp/cdos-fastboot-tmp > /etc/cdos-fastboot/cdos-fastboot.conf");
    if (status1 == -1 || status2 == -1){
        printf("system error \n");
    }
    cdos_util_spawn_command_line("dbus-send --system --print-reply --dest=\"org.freedesktop.login1\" /org/freedesktop/login1 org.freedesktop.login1.Manager.Hibernate boolean:'true'");
}

static void
do_hibernate (CdosPopupBaseMenuItem *item,
        ClutterEvent *event,
        gboolean keep_menu,
        CdosScreenSaver *self)
{
    on_cancel_clicked (NULL, 0, self);
    if (cdos_session_get_inhibited_actions () ^ GSM_INHIBITOR_FLAG_LOGOUT)
    {
        real_hibernate ();
        return;
    }
    cdos_message_dialog_set_title (self->dialog, _("Hibernate"));
    cdos_message_dialog_set_message (self->dialog,
                                     _("The current user has unsaved data."
                                       "Forced hibernate may cause the user's data to be lost."
                                       "\nAre you sure to continue hibernate?"));
    cdos_message_dialog_set_callback (self->dialog, (GCallback)real_hibernate, NULL);
    clutter_actor_set_child_above_sibling ((ClutterActor *)self, (ClutterActor *)self->dialog, NULL);
    clutter_actor_show ((ClutterActor *)self->dialog);
    clutter_actor_grab_key_focus ((ClutterActor *)self->dialog);
}

static void
real_suspend (void)
{
    cdos_util_spawn_command_line("dbus-send --system --print-reply --dest=\"org.freedesktop.login1\" /org/freedesktop/login1 org.freedesktop.login1.Manager.Suspend boolean:'true'");
}

static void
do_suspend (CdosPopupBaseMenuItem *item,
        ClutterEvent *event,
        gboolean keep_menu,
        CdosScreenSaver *self)
{
    /* (N 3060 GPU) after suspend resume, the user face icon does't show.*/
    on_cancel_clicked (NULL, 0, self);
    if (cdos_session_get_inhibited_actions () ^ GSM_INHIBITOR_FLAG_LOGOUT)
    {
        real_suspend ();
        return;
    }
    cdos_message_dialog_set_title (self->dialog, _("Suspend"));
    cdos_message_dialog_set_message (self->dialog,
                                     _("The current user has unsaved data."
                                       "Forced suspend may cause the user's data to be lost."
                                       "\nAre you sure to continue suspend?"));
    cdos_message_dialog_set_callback (self->dialog, (GCallback)real_suspend, NULL);
    clutter_actor_set_child_above_sibling ((ClutterActor *)self, (ClutterActor *)self->dialog, NULL);
    clutter_actor_show ((ClutterActor *)self->dialog);
    clutter_actor_grab_key_focus ((ClutterActor *)self->dialog);
}

static void
on_power_clicked (StButton *btn,
                  gint      button,
                  CdosScreenSaver *self)
{
    if (!self->menu)
    {
        self->menu = cdos_popup_menu_new_with_params (self->power_button, 0.5, ST_SIDE_BOTTOM);
        clutter_actor_add_child ((ClutterActor *)self->stage, (ClutterActor *)self->menu);
        CdosPopupMenuBase *menu = (CdosPopupMenuBase *)self->menu;
        st_widget_add_style_class_name ((StWidget *)menu, "cs-menu");
        //cdos_popup_menu_set_arrow_alignment (self->menu, 1);
        CdosPopupMenuManager *manager = cdos_popup_menu_manager_new (self->power_button);
        cdos_popup_menu_manager_add_menu (manager, menu, -1);
        if (check_acpi_supports ("S3") == TRUE)
            cdos_popup_menu_base_add_action (menu, _("Suspend"), (GCallback)do_suspend, self);
        if (check_acpi_supports ("S4") == TRUE)
            cdos_popup_menu_base_add_action (menu, _("Hibernate"), (GCallback)do_hibernate, self);
        cdos_popup_menu_base_add_action (menu, _("Logout"), (GCallback)do_logout, self);
        cdos_popup_menu_base_add_action (menu, _("Reboot"), (GCallback)do_reboot, self);
        cdos_popup_menu_base_add_action (menu, _("Shutdown"), (GCallback)do_shutdown, self);

        self->dialog = cdos_message_dialog_new_with_buttons(_("Logout"),
                                                              NULL,
                                                              NULL,
                                                              NULL);
        _add_actor (self, self->dialog);
        clutter_actor_hide ((ClutterActor *)self->dialog);
    }
    cdos_popup_menu_base_toggle ((CdosPopupMenuBase *)self->menu);
}

static void
on_switch_user_clicked (StButton *btn,
                        gint      button,
                        CdosScreenSaver *self)
{
    if (button == 1)
    {
    	self->auto_unlock = TRUE;
        on_cancel_clicked (NULL, 0, self);
        cdos_util_spawn_command_line ("dm-tool switch-to-greeter"); 
    }
}

static void
on_cancel_clicked (StButton *btn,
                   gint      button,
                   CdosScreenSaver *self)
{
    user_verify_timeout_clear (self);
    //退出认证状态
    user_verify_cancel (self, FALSE);
    if (self->contents_state == CS_SHOWN)
    {
        self->contents_state = CS_HIDING;

        /* ClutterColor color = {0, 0, 0, 90};
        cdos_transition_simple (self->mask_actor,
                                CLUTTER_EASE_IN,
                                250,
                                "background-color", &color,
                                NULL); */
        clutter_actor_set_opacity (self->contents, 255);
        ClutterTransition *transition = cdos_transition_simple (self->contents,
                                                                CLUTTER_EASE_IN,
                                                                250,
                                                                "opacity", 0,
                                                                NULL);
        g_signal_connect (transition, "completed", G_CALLBACK (on_contents_hidden), self);

        /* date, time label */
        show_calendar (self, TRUE);
    }
}

static void
on_unlock_clicked (StButton *btn,
                   gint      button,
                   CdosScreenSaver *self)
{
    ActUserPasswordMode password_mode;
    g_debug ("%s:::skip_unlock = %d", __func__, self->skip_unlock);
    if (self->skip_unlock)
    {
        self->skip_unlock = FALSE;
        if (self->verify_success)
        {
            verification_complete (self);
            return;
        }
        return;
    }

    password_mode = act_user_get_password_mode (self->user);

    const gchar *password = st_entry_get_text ((StEntry *)self->prompt_entry);
    g_clear_pointer (&self->cached_password, g_free);
    if (password_mode != ACT_USER_PASSWORD_MODE_NONE &&
        (!password || password[0] == '\0'))
    {
        st_widget_add_style_pseudo_class ((StWidget *)self->prompt_entry, "warning");
        if (!is_cdrom ())
            return;
    }

    self->cached_password = g_strdup (password);
    g_debug ("%s:::password = %s, state = %d", __func__, password, self->state);
    if (self->state == VERIFY_END)
    {
        cs_unlock_screen (self);
    }
    else if (self->state != VERIFYING)
    {
        answer_query (self);
    }
}

static gboolean
on_prompt_key_released (ClutterActor *entry,
                        ClutterKeyEvent *event,
                        CdosScreenSaver *self)
{
    guint key = clutter_event_get_key_symbol ((ClutterEvent *)event);

    gboolean is_enter = (key == CLUTTER_KEY_ISO_Enter ||
                         key == CLUTTER_KEY_Return ||
                         key == CLUTTER_KEY_KP_Enter);

    if (is_enter)
    {
        on_unlock_clicked (NULL, 0, self);
        return TRUE;
    }
    return FALSE;
}

static void
on_prompt_text_changed (ClutterText *clutter_text,
                        CdosScreenSaver *self)
{
    StWidget *widget = ST_WIDGET (self->prompt_entry);
    const char *entry_text = clutter_text_get_text (clutter_text);
    const gchar *message = NULL;
    if (entry_text)
    {
        st_widget_remove_style_pseudo_class (widget, "warning");
    }
    if (self->idle_watch_id && self->state != ACCT_LOCKED)
    {
        message = entry_text && g_strcmp0 (entry_text, "") ? "" : _("Please input your password or PIN code");
        st_label_set_text ((StLabel *)self->message_label, message);
    }
}

static void
draw_user_icon (CdosScreenSaver *self)
{
    guint ui_scale = 1;
    ClutterActor *face_icon = NULL;

    CdosGlobal *global = cdos_global_get();
    ui_scale = cdos_global_get_ui_scale (global);

    face_icon = cdos_get_user_face_icon (self->user, USER_FACE_SIZE * (int)ui_scale);
    if (face_icon == NULL)
    {
        face_icon = CLUTTER_ACTOR (st_icon_new ());
        st_icon_set_icon_name ((StIcon *)face_icon, "avatar-default-symbolic");
        st_icon_set_icon_size ((StIcon *)face_icon, USER_FACE_SIZE);
    }
    st_bin_set_child ((StBin *)self->user_icon, face_icon);
    clutter_actor_queue_redraw (self->user_icon);
}

static void
on_user_changed (ActUser *user,
                 CdosScreenSaver *self)
{
    draw_user_icon (self);
    const gchar *user_name = cdos_user_get_display_name (self->user);
    st_label_set_text ((StLabel *)self->username_label, user_name);
}

/*
 * TODO
 * 鼠标按下并移动时渐变动画
 */
static gboolean
on_gesture_begin (ClutterGestureAction *gesture,
                  ClutterActor *actor,
                  CdosScreenSaver *self)
{
    if (self->contents_state == CS_HIDE)
        return TRUE;
    return FALSE;
}

static gboolean
on_gesture_progress (ClutterGestureAction *gesture,
                     ClutterActor *actor,
                     CdosScreenSaver *self)
{
    return FALSE;
}

static void
on_gesture_end (ClutterGestureAction *gesture,
                ClutterActor *actor,
                CdosScreenSaver *self)
{

}

static void
settings_on_changed (CdosScreenSaver *self,
                     gchar           *key)
{
    if (g_strcmp0 (key, "default-message") == 0)
    {
        char *msg = g_settings_get_string (self->settings, key);
        st_label_set_text ((StLabel *)self->motto_label, msg);
    }
    else if (g_strcmp0 (key, "bio-auth-switch-enabled") == 0)
    {
        if (g_settings_has_key (self->settings, key))
            self->bio_auth_can_switch = g_settings_get_boolean (self->settings, key);
        else 
            self->bio_auth_can_switch = FALSE;
    }
}

static void
entry_on_primary_icon_clicked (StEntry *entry, CdosScreenSaver *self)
{
    ClutterText *clutter_text;
    gunichar pw_char;
    clutter_text = st_entry_get_clutter_text (entry);
    pw_char = clutter_text_get_password_char (clutter_text);

    if (pw_char != 0)
    {
        clutter_text_set_password_char (clutter_text, 0);
        st_icon_set_icon_name (self->entry_primary_icon, "password-show");
        self->is_show_password = TRUE;
    } else {
        clutter_text_set_password_char (clutter_text, 0x25cf);
        st_icon_set_icon_name (self->entry_primary_icon, "password-hide"); 
        self->is_show_password = FALSE;
    }
}

static void 
entry_on_secondary_icon_clicked (StEntry *entry, CdosScreenSaver *self)
{
    on_unlock_clicked (NULL, 0, self);
}

static void
entry_on_key_focus_in (ClutterActor *actor, CdosScreenSaver *self)
{
    //st_entry_set_text ((StEntry *)self->prompt_entry, NULL);
    update_prompt_entry (self, FALSE);
}

static gboolean 
should_show_switch_user (void)
{
    ActUserManager *act_manager = act_user_manager_get_default ();
    GSList *user_list = act_user_manager_list_users (act_manager);
    guint len = g_slist_length (user_list);
    g_slist_free (user_list);
    return len > 1 ? TRUE : FALSE;
}

static void
users_on_changed (ActUserManager *gobject,
               ActUser        *user,
               gpointer        user_data)
{
    CdosScreenSaver *self = CDOS_SCREEN_SAVER (user_data);
    if (can_switch_user() && should_show_switch_user())
    {
        clutter_actor_show (self->switch_user_button);
    }
    else
    {
        clutter_actor_hide (self->switch_user_button);
    }
}

/* 更新生物认证标识动画帧 */
static void
update_face_image (CdosScreenSaver *self, GdkPixbuf *pixbuf)
{
    ClutterContent *image;
    gint pixbuf_width;
    gint pixbuf_height;
    GBytes *pixels;

    if (!pixbuf)
        return;
    
    pixbuf_width = gdk_pixbuf_get_width (pixbuf);
    pixbuf_height = gdk_pixbuf_get_height (pixbuf);

    image = clutter_image_new ();
    pixels = gdk_pixbuf_read_pixel_bytes (pixbuf);

    clutter_image_set_bytes (CLUTTER_IMAGE (image),
                            pixels,
                            gdk_pixbuf_get_has_alpha (pixbuf)
                            ? COGL_PIXEL_FORMAT_RGBA_8888
                            : COGL_PIXEL_FORMAT_RGB_888,
                            pixbuf_width,
                            pixbuf_height,
                            gdk_pixbuf_get_rowstride (pixbuf),
                            NULL);
    clutter_actor_set_content (CLUTTER_ACTOR(self->bio_face), image);
    g_object_unref (image);
    g_clear_pointer (&pixels, g_bytes_unref);
}

/* 播放动画图片，如gif图片 */
static gboolean
play_images_animation (CdosScreenSaver *self)
{
    GdkPixbuf *pixbuf;
    gint delay_time; 
    gboolean advance = TRUE;

    if (!self->ani_iter)
        self->ani_iter = gdk_pixbuf_animation_get_iter (self->pixbuf_ani, &self->animation_time);
    else 
        advance = gdk_pixbuf_animation_iter_advance (self->ani_iter, &self->animation_time);

    if (!advance)
        return TRUE;

    delay_time = gdk_pixbuf_animation_iter_get_delay_time (self->ani_iter);
    g_time_val_add (&self->animation_time, delay_time * 1000);

    pixbuf = gdk_pixbuf_animation_iter_get_pixbuf (self->ani_iter);
    if (pixbuf == NULL)
        return FALSE;

    update_face_image (self, pixbuf);
    return TRUE;
}

static void
bio_face_image_clear (CdosScreenSaver *self)
{
    if (self->ani_id > 0)
    {
        g_source_remove (self->ani_id);
        self->ani_id = 0;
    }
    if (self->pixbuf_ani)
    {
        gdk_pixbuf_animation_unref (self->pixbuf_ani);
        self->pixbuf_ani = NULL;
    }
    if (self->ani_iter)
    {
        g_object_unref (self->ani_iter);
        self->ani_iter = NULL;
    }
    self->animation_time.tv_sec = 0;
    self->animation_time.tv_usec = 0;
}

/* 加载生物认证标识图片 */
static void
bio_load_face_image (CdosScreenSaver *self, gchar *file_path)
{
    gint pixbuf_width;
    gint pixbuf_height;

    bio_face_image_clear (self);

    self->pixbuf_ani = gdk_pixbuf_animation_new_from_file (file_path, NULL);
    g_debug ("%s:: bio face image: %s", __func__, file_path);
    if (!self->pixbuf_ani)
        return;

    pixbuf_width = gdk_pixbuf_animation_get_width (self->pixbuf_ani);
    pixbuf_height = gdk_pixbuf_animation_get_height (self->pixbuf_ani);
    clutter_actor_set_width (self->bio_face, pixbuf_width);
    clutter_actor_set_height (self->bio_face, pixbuf_height);

    self->ani_id = g_timeout_add (100, (GSourceFunc)play_images_animation, self);
}

//更新生物认证标识
static void
bio_update_bio_face (CdosScreenSaver *self, gint state)
{
    gchar *face_logo = NULL;
    if (self->auth_type > AUTH_PASSWORD && self->auth_type != AUTH_UKEY)
    {
        face_logo = bio_get_device_type_face (self, self->auth_type, state);
        if (face_logo && g_file_test (face_logo, G_FILE_TEST_EXISTS))
        {
            g_message ("Bio device face image: %s", face_logo);
            bio_load_face_image (self, face_logo);
        } else {
            g_warning ("Can't find the bio face file: %s", face_logo);
        }
        g_free (face_logo);

        //人脸识别：打开摄像头
        bio_camera_play (self);
    }
}

/* 是否显示多个生物认证下拉菜单 */
static gboolean 
should_show_selection_button (CdosScreenSaver *self)
{
    gchar **supported;
    gboolean should = FALSE;
    //双英子认证不显示切换
    if (self->mix_auth || !self->bio_auth_can_switch)
        return should;
    supported = bio_get_user_supported_devices (self, self->username);
    if (supported && g_strv_length (supported) > 1)
        should = TRUE;
    g_strfreev (supported);
    return should;
}

static gboolean 
do_regist_verify_info_once (CdosScreenSaver *self)
{
    if (self->verify_proxy)
    {
        GVariant *v;
        GError *error = NULL;
        v = g_dbus_proxy_call_sync (self->verify_proxy,
                           "RegistVerifyInfo",
                           g_variant_new ("(ss)",
                                      self->service,
                                      self->username),
                           G_DBUS_CALL_FLAGS_NONE,
                           -1,
                           NULL, &error);
        if (error)
        {
            g_warning ("Regist verify info failed:%s", error->message);
            g_clear_error (&error);
        }
        if (v)
        {
            g_debug ("%s:: contents state: %d", __func__, self->contents_state);
            g_variant_unref (v);
            //显示状态下才开始解锁，并开始认证
            if (self->contents_state == CS_SHOWN)
                cs_unlock_screen (self);
        }
    }
    self->regist_info_id = 0;
    return FALSE;
}

/* 重新注册认证 */
static void 
regist_verify_info_again (CdosScreenSaver *self, guint interval)
{
    g_debug ("%s::state = %d", __func__, self->state);
    user_verify_timeout_clear (self);
    //退出当前的认证服务
    user_verify_cancel (self, FALSE);
    self->regist_info_id = \
            g_timeout_add (interval, 
                           (GSourceFunc) do_regist_verify_info_once, 
                           self);
}

/**
 * 修改pam认证服务的配置文件
 * cdos-screensaver-mix: 使用common-auth做为认证组件，可以支持密码，生物，双因子等认证方式
 * cdos-screensaver: 仅支持密码认证
 */
static void
update_pam_service (CdosScreenSaver *self, const gchar *service_name)
{
    if (!service_name || g_strcmp0 (self->service, service_name) == 0)
        return ;

    g_free (self->service);
    self->service = g_strdup (service_name);
}

/* 使用密码认证 */
static gboolean
use_auth_password (CdosScreenSaver *self)
{
    gboolean lock_state;
    //停止生物设备认证
    /* if (self->auth_type > AUTH_PASSWORD)
        bio_device_stop (self); */

    lock_state = self->state == ACCT_LOCKED ? TRUE : FALSE;

    self->auth_type = AUTH_PASSWORD;
    clutter_actor_hide (self->bio_face);
    bio_face_image_clear (self);
    g_clear_pointer (&self->cached_password, g_free);
    draw_user_icon (self);

    st_button_set_label (ST_BUTTON (self->auth_type_button), _("Auth with bio"));
    //双因子认证不显示
    if (self->use_bio_auth && self->mix_auth == 0)
        clutter_actor_show (self->auth_type_button);
    else
        clutter_actor_hide (self->auth_type_button);
    clutter_actor_hide (self->bio_info_box);

    st_entry_set_text ((StEntry *)self->prompt_entry, NULL);
    clutter_actor_set_reactive ((ClutterActor *)self->prompt_entry, !lock_state);
    if (lock_state)
    {
        clutter_actor_hide ((ClutterActor *)self->prompt_entry);
        clutter_actor_grab_key_focus ((ClutterActor *)self);
    } else {
        st_label_set_text ((StLabel *)self->message_label, _("Please input your password or PIN code"));
        clutter_actor_show ((ClutterActor *)self->prompt_entry);
        clutter_actor_grab_key_focus ((ClutterActor *)self->prompt_entry);
    }

    update_pam_service (self, "cdos-screensaver-mix");

    return TRUE;
}

/* 使用ukey认证 */
static gboolean
use_auth_ukey (CdosScreenSaver *self)
{
    gboolean lock_state;
    lock_state = self->state == ACCT_LOCKED ? TRUE : FALSE;

    clutter_actor_hide (self->bio_face);
    bio_face_image_clear (self);
    g_clear_pointer (&self->cached_password, g_free);

    st_entry_set_text ((StEntry *)self->prompt_entry, NULL);
    clutter_actor_set_reactive ((ClutterActor *)self->prompt_entry, !lock_state);
    if (lock_state)
    {
        clutter_actor_hide ((ClutterActor *)self->prompt_entry);
        clutter_actor_grab_key_focus ((ClutterActor *)self);
    } else {
        st_label_set_text ((StLabel *)self->message_label, _("Please input your password or PIN code"));
        clutter_actor_show ((ClutterActor *)self->prompt_entry);
        clutter_actor_grab_key_focus ((ClutterActor *)self->prompt_entry);
    }

    update_pam_service (self, "cdos-screensaver-mix");

    return TRUE;
}

/* 使用生物识别认证 */
static gboolean
use_auth_bio (CdosScreenSaver *self)
{
    gchar *primary_device = NULL;
    const gchar *auth_type_name;
    gboolean lock_state;
    lock_state = self->state == ACCT_LOCKED ? TRUE : FALSE;

    bio_camera_stop (self);

    if (!lock_state)
        st_label_set_text ((StLabel *)self->message_label, "");

    primary_device = bio_get_primary_device (self);
    if (primary_device == NULL)
        return FALSE;

    self->auth_type = bio_get_device_type (self, primary_device);
    auth_type_name = bio_get_device_type_name (self, self->auth_type);
    g_message ("Current bio auth type: %d (%s).", self->auth_type, auth_type_name);
    st_label_set_text (self->bio_type_label, auth_type_name);
    g_free (primary_device);
    g_free (auth_type_name);

    st_button_set_label (ST_BUTTON (self->auth_type_button), _("Auth with password"));
    //双因子认证不显示
    if (self->mix_auth == 0)
        clutter_actor_show (self->auth_type_button);
    else
        clutter_actor_hide (self->auth_type_button);

    clutter_actor_show (self->bio_info_box);
    //是否显示切换按钮和菜单
    if (should_show_selection_button (self))
        clutter_actor_show (self->bio_selection_button);
    else 
        clutter_actor_hide (self->bio_selection_button);

    //使用ukey密码认证
    if (self->auth_type == AUTH_UKEY)
    {
        return use_auth_ukey (self);
    }

    clutter_actor_hide (self->prompt_entry);
    clutter_actor_show (self->bio_face);

    bio_update_bio_face (self, lock_state ? 0 : 1);
    clutter_actor_grab_key_focus ((ClutterActor *)self);

    update_pam_service (self, "cdos-screensaver-mix");

    return TRUE;
}

/* 更新显示界面 */
static void
update_bio_auth_shown (CdosScreenSaver *self)
{
    should_use_bio_auth (self);

    if (self->use_bio_auth && use_auth_bio (self))
        goto out;
    use_auth_password (self);

out:
    regist_verify_info_again (self, 300);
}

//切换密码和生物认证模式
static void
on_auth_type_clicked (StButton *btn,
                      gint      button,
                      CdosScreenSaver *self)
{
    if (button == 1)
    {
        user_verify_timeout_clear (self);
        if (self->auth_type == AUTH_PASSWORD)
        {
            use_auth_bio (self);
        } else {
            use_auth_password (self);
            /**
             * 强制使用"cdos-screensaver"认证，
             * 默认情况下仍使用"cdos-screensaver-mix"做为密码认证配置
             * 从其它认证方式切换到密码认证，后端存在一些问题，导致切换受阻，因此密码认证使用单独的配置
             */
            update_pam_service (self, "cdos-screensaver");
        }
        regist_verify_info_again (self, 300);
    }
}

static void
bio_menu_item_activated (CdosPopupBaseMenuItem *item,
                        ClutterEvent *event,
                        gboolean keep_menu,
                        CdosScreenSaver *self)
{
    gchar *device_id;
    gint device_type;
    gboolean ok;

    device_id = g_object_get_data (item, "device-id");
    device_type = bio_get_device_type (self, device_id);
    if (device_type == self->auth_type)
        return ;

    ok = bio_device_set_selected (self, device_id, TRUE);
    if (ok)
    {
        use_auth_bio (self);
    }
}

//选择生物认证方式
static void
on_bio_selection_clicked (StButton *btn,
                          gint      button,
                          CdosScreenSaver *self)
{
    if (!self->bio_menu)
    {
        gchar **devices;
        gchar *device_id;
        gint device_type;
        const gchar *device_name;
        CdosPopupMenuItem *menu_item;
        devices = bio_get_user_supported_devices (self, self->username);
        if (!devices)
            return;
        self->bio_menu = cdos_popup_menu_new_with_params (self->bio_selection_button, 0.5, ST_SIDE_TOP);
        clutter_actor_add_child ((ClutterActor *)self->stage, (ClutterActor *)self->bio_menu);
        st_widget_add_style_class_name ((StWidget *)self->bio_menu, "cs-menu");
        //cdos_popup_menu_set_arrow_alignment (self->menu, 1);
        CdosPopupMenuManager *manager = cdos_popup_menu_manager_new (self->bio_selection_button);
        cdos_popup_menu_manager_add_menu (manager, self->bio_menu, -1);

        for (int i = 0; devices[i] != NULL; i++)
        {
            device_id = devices[i];
            device_type = bio_get_device_type (self, device_id);
            device_name = bio_get_device_type_name (self, device_type);
            menu_item = cdos_popup_menu_base_add_action ((CdosPopupMenuBase *)self->bio_menu, 
                                             device_name, 
                                             (GCallback)bio_menu_item_activated, 
                                             self);
            g_object_set_data_full (menu_item, 
                                    "device-id", g_strdup (device_id),
                                    g_free);
            g_free (device_name); 
        }
        g_strfreev (devices);
    }
    cdos_popup_menu_base_toggle ((CdosPopupMenuBase *)self->bio_menu);
}

static void
init_ui (CdosScreenSaver *self)
{
    ClutterText *clutter_text, *prompt_clutter_text;
    ClutterActor *actor;
    StWidget  *auth_box;
    actor = CLUTTER_ACTOR (self);
    ClutterColor color = {0, 0, 1, 1};
    clutter_actor_set_background_color (actor, &color);

    self->background = clutter_actor_new ();
    self->background_manager =
        cdos_background_manager_new (self->background, FALSE, FALSE, FALSE, "org.cdos.desktop.background.screensaver");

    cdos_background_manager_hide_register_watermark (self->background_manager);

    self->background_protect = clutter_actor_new ();
    self->background_manager_protect =
        cdos_background_manager_new (self->background_protect, FALSE, FALSE, FALSE, "org.cdos.desktop.background.slideshow");
    cdos_background_manager_hide_register_watermark (self->background_manager_protect);

    clutter_actor_add_child (actor, self->background);
    clutter_actor_add_child (actor, self->background_protect);

    self->layout = cdos_layout_manager_get_default ();
    self->monitor = cdos_layout_manager_get_index ();

    self->mask_actor = clutter_actor_new ();
    clutter_actor_add_child (actor, self->mask_actor);
    ClutterColor mask_color = {0, 0, 0, 90};
    clutter_actor_set_background_color (self->mask_actor, &mask_color);

    self->contents = g_object_new (ST_TYPE_BOX_LAYOUT, "style-class", "cs-contents", "vertical", TRUE, NULL);
    clutter_actor_add_child (actor, self->contents);
    auth_box = st_box_layout_new ();
    st_box_layout_set_vertical (auth_box, TRUE);
    st_widget_add_style_class_name ((StWidget *)auth_box, "cs-auth-box");

    self->user_icon = g_object_new (ST_TYPE_BIN,
                                    "style-class", "cs-user-icon",
                                    NULL);
    self->username_label  = (ClutterActor *)st_label_new (NULL);
    self->text_info_label = (ClutterActor *)st_label_new (NULL);
    self->prompt_entry    = (ClutterActor *)st_entry_new (NULL);
    self->message_label   = (ClutterActor *)st_label_new ("");
    //生物识别
    //self->bio_face = st_icon_new ();
    self->bio_face = clutter_actor_new ();
    self->bio_info_box = st_box_layout_new ();
    st_box_layout_set_vertical (self->bio_info_box, FALSE);
    self->bio_type_label = st_label_new (NULL);
    self->bio_selection_button = st_button_new ();
    ClutterActor *bio_type_icon = (ClutterActor *)g_object_new (ST_TYPE_ICON,
                                                             "icon-name", "found-go-down",
                                                             "icon-size", 16,
                                                             NULL);
    st_bin_set_child ((StBin *)self->bio_selection_button, bio_type_icon);
    clutter_actor_add_child (self->bio_info_box, self->bio_type_label);
    clutter_actor_add_child (self->bio_info_box, self->bio_selection_button);

    self->power_button   = (ClutterActor *)st_button_new ();
    ClutterActor *power_icon = (ClutterActor *)g_object_new (ST_TYPE_ICON,
                                                             "icon-name", "found-menu-power",
                                                             "icon-size", 16,
                                                             NULL);
    st_bin_set_child ((StBin *)self->power_button, power_icon);

    self->switch_user_button = (ClutterActor *)st_button_new ();
    ClutterActor *swtch_user_icon = (ClutterActor *)g_object_new (ST_TYPE_ICON,
                                                             "icon-name", "system-switch-user",
                                                             "icon-size", 16,
                                                             NULL);
    st_bin_set_child ((StBin *)self->switch_user_button, swtch_user_icon);

    self->user_changed_id = g_signal_connect (self->user, 
                                              "changed", 
                                              G_CALLBACK(on_user_changed), 
                                              self);
    on_user_changed (self->user, self);

    self->auth_type_button = (ClutterActor *)st_button_new_with_label (_("Auth with password"));

    st_widget_set_style_class_name ((StWidget *)self->username_label, "cs-user-label");
    clutter_actor_set_x_expand (self->user_icon, TRUE);
    clutter_actor_set_x_align (self->user_icon, CLUTTER_ACTOR_ALIGN_CENTER);
    clutter_actor_set_y_align (self->user_icon, CLUTTER_ACTOR_ALIGN_START);

    clutter_actor_set_x_expand (self->username_label, TRUE);
    clutter_actor_set_x_align (self->username_label, CLUTTER_ACTOR_ALIGN_FILL);
    clutter_actor_set_y_align (self->username_label, CLUTTER_ACTOR_ALIGN_START);
    clutter_text = (ClutterText *)st_label_get_clutter_text ((StLabel *)self->username_label);
    clutter_text_set_line_wrap (clutter_text, TRUE);
    clutter_text_set_line_alignment (clutter_text, PANGO_ALIGN_CENTER);
    clutter_actor_set_x_align (clutter_text, CLUTTER_ACTOR_ALIGN_CENTER);
    clutter_actor_set_x_expand (clutter_text, TRUE);

    ClutterActor *userbox = g_object_new (ST_TYPE_BOX_LAYOUT,
                                          "style-class", "cs-user-box",
                                          "x-expand", TRUE,
                                          "vertical", TRUE,
                                          NULL);
    clutter_actor_add_child (userbox, self->user_icon);
    clutter_actor_add_child (userbox, self->username_label);

    clutter_actor_set_x_align (self->message_label, CLUTTER_ACTOR_ALIGN_CENTER);
    st_widget_set_style_class_name ((StWidget *)self->message_label, "cs-message-label");
    clutter_text = (ClutterText *)st_label_get_clutter_text ((StLabel *)self->message_label);
    clutter_text_set_line_wrap (clutter_text, TRUE);
    clutter_text_set_ellipsize (clutter_text, PANGO_ELLIPSIZE_NONE);

    clutter_actor_set_x_expand (self->prompt_entry, TRUE);
    clutter_actor_set_x_align (self->prompt_entry, CLUTTER_ACTOR_ALIGN_FILL);
    st_widget_add_style_class_name ((StWidget *)self->prompt_entry, "cs-prompt-entry");
    prompt_clutter_text = (ClutterText *)st_entry_get_clutter_text ((StEntry *)self->prompt_entry);
    clutter_text_set_password_char (prompt_clutter_text, 0x25cf);
    st_entry_set_input_purpose ((StEntry *)self->prompt_entry, CLUTTER_INPUT_CONTENT_PURPOSE_PASSWORD);
    clutter_text_set_max_length (prompt_clutter_text, 24);
    st_entry_set_hint_text (self->prompt_entry, _("Please input your password or PIN code"));

    StIcon *primary_icon = ST_ICON (st_icon_new ());
    st_icon_set_icon_size (primary_icon, 16); 
    st_icon_set_icon_name (primary_icon, "password-hide"); 

    st_entry_set_primary_icon (self->prompt_entry, (ClutterActor *)primary_icon); 
    st_widget_set_style_class_name ((StWidget *)primary_icon, "cs-prompt-primary-icon");
    st_widget_set_track_hover ((StWidget *)primary_icon, TRUE);
    self->entry_primary_icon = primary_icon;

    StIcon *secondary_icon = ST_ICON (st_icon_new ());
    st_icon_set_icon_size (secondary_icon, 16); 
    st_icon_set_icon_name (secondary_icon, "go-enter"); 
    st_entry_set_secondary_icon (self->prompt_entry, (ClutterActor *)secondary_icon); 
    st_widget_set_style_class_name ((StWidget *)secondary_icon, "cs-prompt-secondary-icon");
    st_widget_set_track_hover ((StWidget *)secondary_icon, TRUE);

    clutter_actor_set_x_expand (self->text_info_label, TRUE);
    clutter_actor_set_x_align (self->text_info_label, CLUTTER_ACTOR_ALIGN_CENTER);
    st_widget_set_style_class_name ((StWidget *)self->text_info_label, "cs-text-info-label");
    clutter_text = (ClutterText *)st_label_get_clutter_text ((StLabel *)self->text_info_label);
    clutter_text_set_line_wrap (clutter_text, TRUE);
    clutter_text_set_ellipsize (clutter_text, PANGO_ELLIPSIZE_NONE);

    clutter_actor_set_x_align (self->power_button, CLUTTER_ACTOR_ALIGN_START);
    clutter_actor_set_x_align (self->switch_user_button, CLUTTER_ACTOR_ALIGN_START);
    clutter_actor_set_x_align (self->auth_type_button, CLUTTER_ACTOR_ALIGN_CENTER);
    clutter_actor_set_x_align (auth_box, CLUTTER_ACTOR_ALIGN_FILL);
    clutter_actor_set_y_align (auth_box, CLUTTER_ACTOR_ALIGN_FILL);

    /* ClutterActor *expander = clutter_actor_new ();
    clutter_actor_set_x_expand (expander, TRUE);
    clutter_actor_set_x_align (expander, CLUTTER_ACTOR_ALIGN_CENTER); */

    self->button_box = g_object_new (ST_TYPE_BOX_LAYOUT,
                                            "style-class", "cs-button-box",
                                            "x-expand", TRUE,
                                            "vertical", FALSE,
                                            NULL);

    clutter_actor_add_child (self->button_box, self->switch_user_button);
    clutter_actor_add_child (self->button_box, self->power_button);
    //clutter_actor_add_child (self->button_box, expander);

    st_widget_add_style_class_name ((StWidget *)self->power_button, "cs-power-button");
    st_widget_add_style_class_name ((StWidget *)self->switch_user_button, "cs-switch-user-button");
    st_widget_add_style_class_name ((StWidget *)self->auth_type_button, "cs-switch-auth-button");
    //st_widget_add_style_class_name ((StWidget *)self->bio_face, "cs-bio-face");
    st_widget_add_style_class_name ((StWidget *)self->bio_type_label, "cs-bio-type");
    st_widget_add_style_class_name ((StWidget *)self->bio_selection_button, "cs-bio-selection");

    clutter_actor_set_y_align (self->message_label, CLUTTER_ACTOR_ALIGN_START);
    clutter_actor_set_y_align (self->text_info_label, CLUTTER_ACTOR_ALIGN_START);
    clutter_actor_set_y_align (self->prompt_entry, CLUTTER_ACTOR_ALIGN_START);

    clutter_actor_add_child (self->contents, auth_box);
    clutter_actor_add_child (auth_box, userbox);
    clutter_actor_add_child (auth_box, self->message_label);
    clutter_actor_add_child (auth_box, self->prompt_entry);
    clutter_actor_add_child (auth_box, self->bio_face);
    clutter_actor_add_child (auth_box, self->bio_info_box);
    clutter_actor_add_child (auth_box, self->text_info_label);

    clutter_actor_add_child (self->contents, self->auth_type_button);

    clutter_actor_add_child (actor, self->button_box);
    clutter_actor_show (self->button_box);

    clutter_actor_hide (self->auth_type_button);

    update_bio_auth_shown (self);

    g_signal_connect (self->power_button, "clicked", G_CALLBACK (on_power_clicked), self);
    g_signal_connect (self->switch_user_button, "clicked", G_CALLBACK (on_switch_user_clicked), self);
    g_signal_connect (self->auth_type_button, "clicked", G_CALLBACK (on_auth_type_clicked), self);
    g_signal_connect (self->bio_selection_button, "clicked", G_CALLBACK (on_bio_selection_clicked), self);
    g_signal_connect (prompt_clutter_text, "key-release-event", G_CALLBACK (on_prompt_key_released), self);
    g_signal_connect (prompt_clutter_text, "text-changed", G_CALLBACK (on_prompt_text_changed), self);
    g_signal_connect (prompt_clutter_text, 
                      "key-focus-in", 
                      G_CALLBACK (entry_on_key_focus_in), 
                      self);
    g_signal_connect (self->prompt_entry, 
                     "primary-icon-clicked", 
                     G_CALLBACK (entry_on_primary_icon_clicked), 
                     self);
    g_signal_connect (self->prompt_entry, 
                     "secondary-icon-clicked", 
                     G_CALLBACK (entry_on_secondary_icon_clicked), 
                     self);

    ActUserManager *act_manager = act_user_manager_get_default ();
    g_signal_connect (act_manager, 
                      "user-added", 
                      G_CALLBACK (users_on_changed), 
                      self);
    g_signal_connect (act_manager, 
                      "user-removed",
                      G_CALLBACK (users_on_changed), 
                      self);
    users_on_changed (act_manager, NULL, self);

    ClutterAction *gesture = clutter_gesture_action_new ();
    clutter_actor_add_action (actor, gesture);
    g_signal_connect (gesture, "gesture-begin", G_CALLBACK (on_gesture_begin), self);
    g_signal_connect (gesture, "gesture-progress", G_CALLBACK (on_gesture_progress), self);
    g_signal_connect (gesture, "gesture-cancel", G_CALLBACK (on_gesture_end), self);
    g_signal_connect (gesture, "gesture-end", G_CALLBACK (on_gesture_end), self);

    clutter_actor_hide (self->contents);

    self->calendar = g_object_new (ST_TYPE_WIDGET, "style-class", "cs-calendar", NULL);
    ClutterLayoutManager *vlayout1 = clutter_box_layout_new ();
    clutter_box_layout_set_orientation (CLUTTER_BOX_LAYOUT (vlayout1), CLUTTER_ORIENTATION_VERTICAL);
    clutter_actor_set_layout_manager (self->calendar, vlayout1);

    self->time_label = g_object_new (ST_TYPE_LABEL, "style-class", "cs-time-label",
                                     NULL);
    self->date_label = g_object_new (ST_TYPE_LABEL, "style-class", "cs-date-label",
                                     NULL);
    self->lunar_label = g_object_new (ST_TYPE_LABEL, "style-class", "cs-lunar-label",
                                     NULL);
    self->motto_label = g_object_new (ST_TYPE_LABEL, "style-class", "cs-motto-label", NULL);

    clutter_actor_add_child (self->calendar, self->time_label);
    clutter_actor_add_child (self->calendar, self->date_label);
    clutter_actor_add_child (self->calendar, self->lunar_label);
    clutter_actor_add_child (self->calendar, self->motto_label);

    clutter_actor_hide (self->lunar_label);
    clutter_actor_hide (self->motto_label);

    g_signal_connect_swapped (self->settings, 
                              "changed", 
                              G_CALLBACK(settings_on_changed), 
                              self);
    settings_on_changed (self, "default-message");

    clutter_actor_add_child (actor, self->calendar);
    clutter_actor_set_child_above_sibling (actor, self->calendar, self->mask_actor);
}

static void
user_loaded (ActUser *user,
             GParamSpec *spec,
             CdosScreenSaver *self)
{
    if (act_user_is_loaded (user))
        init_ui (self);
}

/* 注册认证服务信息 */
static void
create_user_verify_proxy (CdosScreenSaver *self)
{
    GError *error = NULL;
    if (self->verify_proxy == NULL)
    {
        self->verify_proxy = g_dbus_proxy_new_for_bus_sync (G_BUS_TYPE_SESSION,
                                                0,
                                                NULL,
                                                "org.cdos.ScreenSaver.UserVerifier",
                                                "/org/cdos/ScreenSaver/UserVerifier",
                                                "org.cdos.ScreenSaver.UserVerifier",
                                                NULL,
                                                &error);
        if (error)
        {
            g_warning ("Create screensaver verify proxy faild: %s", error->message);
            g_error_free (error);
            return;
        }
        g_signal_connect (self->verify_proxy, 
                          "g-signal", 
                          G_CALLBACK(on_verify_signal), 
                          self);
    }
}

static void
on_contents_shown (ClutterTimeline *tml,
                   CdosScreenSaver *self)
{
    self->contents_state = CS_SHOWN;
    g_signal_handlers_disconnect_by_data (tml, self);

    g_source_remove (self->time_update_id);
    self->time_update_id = 0;

    if (!self->idle_monitor)
        self->idle_monitor = meta_idle_monitor_get_core ();
    if (self->idle_watch_id)
        meta_idle_monitor_remove_watch (self->idle_monitor, self->idle_watch_id);
    if (self->prompt_lock_id == 0)
    {
        self->idle_watch_id = meta_idle_monitor_add_idle_watch (self->idle_monitor,
                                                                30 * 1000,
                                                                (MetaIdleMonitorWatchFunc)idle_monitor_cb,
                                                                self,
                                                                NULL);
        //meta_cursor_tracker_set_pointer_visible (meta_cursor_tracker_get_for_display (NULL), TRUE);
        //判断当前生物认证是否可用和有效。
        update_bio_auth_shown (self);
        //如果是双因子认证，则给出提示。
        if (self->mix_auth)
        {
            st_label_set_text ((StLabel *)self->text_info_label, 
                               _("Notice: it will use two auth types together, you can't pass until both success."));
            clutter_actor_show (self->text_info_label);
        } else {
            st_label_set_text ((StLabel *)self->text_info_label, "");
        }
        //验证用户锁定状态，并根据锁定状态更新界面显示以及发起认证注册
        check_account_lock_state (self);
    }
}

static gboolean
cdos_screen_saver_key_press_event (ClutterActor *actor,
                                   ClutterKeyEvent *event)
{
    guint key = clutter_event_get_key_symbol ((ClutterEvent *)event);
    gboolean is_escape = (key == CLUTTER_KEY_Escape);
    g_debug ("Key press event, is escape: %d", is_escape);
    if (is_escape)
    {
        on_cancel_clicked (NULL, 0, (CdosScreenSaver *)actor);
    }
    return TRUE;
}

static void
emulate_key_release_event (ClutterActor *actor,
                           guint key,
                           gboolean is_print)
{
    CdosScreenSaver *self;
    self = CDOS_SCREEN_SAVER (actor);

    gboolean is_enter = (key == CLUTTER_KEY_ISO_Enter ||
                         key == CLUTTER_KEY_Return ||
                         key == CLUTTER_KEY_KP_Enter);

    if (is_enter || is_print)
    {
        if (self->screensaver_state == IS_SCREENSAVER)
        {
            if (self->contents_state == CS_HIDE)
            {
                if (user_in_nopasswdlogin_group (self->username))
                { 
                    //添加免密登录，跳过显示密码输入框
                    //cs_unlock_screen (self);
                    verification_complete (self);
                }
                else 
                {
                    self->contents_state = CS_SHOWING;

                    /* ClutterColor color = {0, 0, 0, 100};
                    cdos_transition_simple (self->mask_actor,
                                            CLUTTER_EASE_OUT,
                                            250,
                                            "background-color", &color,
                                            NULL); */
                    hide_calendar (self);

                    clutter_actor_set_opacity (self->contents, 0);
                    clutter_actor_show (self->contents);
                    ClutterTransition *transition = cdos_transition_simple (self->contents,
                                                                            CLUTTER_EASE_IN,
                                                                            250,
                                                                            "opacity", 255,
                                                                            NULL);
                    g_signal_connect (transition, "completed", G_CALLBACK (on_contents_shown), self);
                }
            }
            else if (is_enter)
            {
                on_unlock_clicked (NULL, 1, self);
            }
        }
        else if (self->screensaver_state == IS_SCREENPROTECT)
        {
            verification_complete (self); 
        }
    }
}

static gboolean
cdos_screen_saver_key_release_event (ClutterActor *actor,
                                     ClutterKeyEvent *event)
{
    ClutterInputDevice *device;
    ClutterModifierType modifier;

    device = clutter_event_get_device ((ClutterEvent *)event);
    if (!device)
        return FALSE;

    modifier = clutter_input_device_get_modifier_state (device);

    if (modifier &
        (CLUTTER_SHIFT_MASK |
         CLUTTER_LOCK_MASK |
         CLUTTER_CONTROL_MASK |
         CLUTTER_MOD1_MASK |
         CLUTTER_MOD2_MASK |
         CLUTTER_MOD3_MASK |
         CLUTTER_MOD4_MASK |
         CLUTTER_MOD5_MASK |
         CLUTTER_SUPER_MASK |
         CLUTTER_META_MASK))
    {
        return FALSE;
    }

    guint key = clutter_event_get_key_symbol ((ClutterEvent *)event);
    gunichar unichar = clutter_event_get_key_unicode ((ClutterEvent *)event);

    gboolean is_print =  g_unichar_isprint (unichar);
    emulate_key_release_event (actor, key, is_print);

    return TRUE;
}

static gboolean
cdos_screen_saver_button_release_event (ClutterActor *actor,
                                        ClutterButtonEvent *event)
{
    CdosScreenSaver *self = CDOS_SCREEN_SAVER (actor);
    if (self->contents_state != CS_HIDE)
        return TRUE;

    emulate_key_release_event (actor, CLUTTER_KEY_KP_Enter, FALSE);
    return TRUE;
}

static void
cdos_screen_saver_get_preferred_width (ClutterActor *actor,
                                       gfloat for_height,
                                       gfloat   *min_width,
                                       gfloat   *nat_width)
{
    CdosScreenSaver *self = CDOS_SCREEN_SAVER (actor);
    gfloat width = clutter_actor_get_width (self->stage);

    *min_width = width;
    *nat_width = width;
}

static void
cdos_screen_saver_get_preferred_height (ClutterActor *actor,
                                        gfloat for_width,
                                        gfloat   *min_height,
                                        gfloat   *nat_height)
{
    CdosScreenSaver *self = CDOS_SCREEN_SAVER (actor);
    gfloat height = clutter_actor_get_height (self->stage);

    *min_height = height;
    *nat_height = height;
}

static void
cdos_screen_saver_allocate (ClutterActor *actor,
                            const ClutterActorBox *box,
                            ClutterAllocationFlags flags)
{
    CdosScreenSaver *self = CDOS_SCREEN_SAVER (actor);
    gfloat min_width, nat_width, min_height, nat_height;
    MetaRectangle geometry;

    CLUTTER_ACTOR_CLASS (cdos_screen_saver_parent_class)->allocate (actor, box, flags);

    self->monitor = cdos_layout_manager_get_index ();
    meta_display_get_monitor_geometry (cdos_global_get_display (self->global),
                                       self->monitor,
                                       &geometry);
    ClutterActorBox child_box;

    if (self->mask_actor)
    {
        ClutterActorBox stage_box;
        clutter_actor_get_allocation_box (self->stage, &stage_box);
        child_box.x1 = stage_box.x1;
        child_box.y1 = stage_box.y1;
        child_box.x2 = stage_box.x2;
        child_box.y2 = stage_box.y2;
        clutter_actor_allocate (self->mask_actor, &child_box, flags);
    }
    if (self->contents && self->contents_state != CS_HIDE)
    {
        clutter_actor_get_preferred_size (self->contents, &min_width, &min_height, &nat_width, &nat_height);
        child_box.x1 = (int)(geometry.x + (geometry.width - nat_width) * 0.5);
        child_box.y1 = (int)(geometry.y + (geometry.height - nat_height) * 0.5);
        child_box.x2 = child_box.x1 + nat_width;
        child_box.y2 = child_box.y1 + nat_height;
        clutter_actor_allocate (self->contents, &child_box, flags);
    }
    if (self->dialog && clutter_actor_is_visible ((ClutterActor *)self->dialog))
    {
        clutter_actor_get_preferred_size ((ClutterActor *)self->dialog, &min_width, &min_height, &nat_width, &nat_height);
        child_box.x1 = (int)(geometry.x + (geometry.width - nat_width) * 0.5);
        child_box.y1 = (int)(geometry.y + (geometry.height - nat_height) * 0.5);
        child_box.x2 = child_box.x1 + nat_width + 0.5;
        child_box.y2 = child_box.y1 + nat_height + 0.5;
        clutter_actor_allocate ((ClutterActor *)self->dialog, &child_box, flags);
    }
    if (self->button_box && clutter_actor_is_visible ((ClutterActor *)self->button_box))
    {
        clutter_actor_get_preferred_size (self->button_box, &min_width, &min_height, &nat_width, &nat_height);
        child_box.x1 = (int)(geometry.x + geometry.width - nat_width - 100);
        child_box.y1 = (int)(geometry.y + geometry.height - nat_height - 100);
        child_box.x2 = child_box.x1 + nat_width;
        child_box.y2 = child_box.y1 + nat_height;
        clutter_actor_allocate (self->button_box, &child_box, flags);
    } 
}

static void
cdos_screen_saver_dispose (GObject *gobject)
{
    CdosScreenSaver *self = CDOS_SCREEN_SAVER (gobject);
    ActUserManager *act_manager = act_user_manager_get_default ();
    g_signal_handlers_disconnect_by_data (act_manager, self);
    if (self->source_added_id)
    {
        g_signal_handler_disconnect (self->message_tray, self->source_added_id);
        self->source_added_id = 0;
    }

    bio_face_image_clear (self);
    user_verify_timeout_clear (self);
    bio_camera_stop (self);

    g_clear_object (&self->verify_proxy);
    g_clear_object (&self->session_proxy);
    g_clear_object (&self->login1_proxy);
    g_clear_object (&self->user);
    g_clear_object (&self->cheese_camera);
    g_clear_pointer (&self->cached_password, g_free);
    g_clear_pointer (&self->cached_clipboard, g_free);
    g_clear_pointer (&self->username, g_free);
    g_clear_pointer (&self->service, g_free);

    remove_idle_monitor (self);

    G_OBJECT_CLASS (cdos_screen_saver_parent_class)->dispose(gobject);
}

static void
cdos_screen_saver_finalize (GObject *gobject)
{
    if (global_screensaver)
        global_screensaver = NULL;
    G_OBJECT_CLASS (cdos_screen_saver_parent_class)->finalize(gobject);
}

static void
cdos_screen_saver_class_init (CdosScreenSaverClass *klass)
{
    GObjectClass *gobject_class = G_OBJECT_CLASS(klass);
    ClutterActorClass *actor_class = CLUTTER_ACTOR_CLASS (klass);

    gobject_class->finalize = cdos_screen_saver_finalize;
    gobject_class->dispose = cdos_screen_saver_dispose;

    actor_class->get_preferred_width = cdos_screen_saver_get_preferred_width;
    actor_class->get_preferred_height = cdos_screen_saver_get_preferred_height;
    actor_class->allocate = cdos_screen_saver_allocate;
    actor_class->key_release_event = cdos_screen_saver_key_release_event;
    actor_class->key_press_event = cdos_screen_saver_key_press_event;
    actor_class->button_release_event = cdos_screen_saver_button_release_event;

    signals[ACTIVE_CHANGED] = g_signal_new ("active-changed",
                                            CDOS_TYPE_SCREEN_SAVER,
                                            G_SIGNAL_RUN_LAST,
                                            0,
                                            NULL,
                                            NULL,
                                            NULL,
                                            G_TYPE_NONE,
                                            1,
                                            G_TYPE_BOOLEAN);

    signals[PROTECT_ACTIVE_CHANGED] = g_signal_new ("protect-active-changed",
                                            CDOS_TYPE_SCREEN_SAVER,
                                            G_SIGNAL_RUN_LAST,
                                            0,
                                            NULL,
                                            NULL,
                                            NULL,
                                            G_TYPE_NONE,
                                            1,
                                            G_TYPE_BOOLEAN);
}

/*
 * CoglFrameInfo
 * CoglFrameEvent
 **/
/*
static void
on_stage_presented (CdosScreenSaver *self)
{
    struct timespec ts;
    clock_gettime (CLOCK_REALTIME, &ts);
    g_message ("%lu on_stage_presented", ts.tv_sec);
}
*/

static void
session_manager_on_properties_changed (GDBusProxy *proxy,
                                    GVariant   *changed_properties,
                                    GStrv       invalidated_properties,
                                    gpointer    user_data)
{
    CdosScreenSaver *self = CDOS_SCREEN_SAVER (user_data);
    GVariantIter iter;
    GVariant *value;
    gchar *key;

    g_variant_iter_init (&iter, changed_properties);
    while (g_variant_iter_loop (&iter, "{sv}", &key, &value))
    {
        if (g_strcmp0 (key, "SessionIsActive") == 0)
        {
            gboolean is_active = g_variant_get_boolean (value);
            self->session_active = is_active;
            g_message ("Session active changed, active:%d, screensaver active state:%d, user:%s", is_active, self->is_active, self->username);
            if (is_active && self->is_active && self->auto_unlock)
            {
                verification_complete (self);
            } else {
                on_cancel_clicked (NULL, 0, self);
            }
            return;
        }
    }
}

static void
login1_session_on_signal_callback (GDBusConnection *connection,
                                    const gchar *sender_name,
                                    const gchar *object_path,
                                    const gchar *interface_name,
                                    const gchar *signal_name,
                                    GVariant *parameters,
                                    gpointer user_data)
{
    CdosScreenSaver *self = CDOS_SCREEN_SAVER (user_data);
    g_message ("Received siganl from %s, signal name: %s", object_path, signal_name);
    if (g_strcmp0 (signal_name, "Lock") == 0)
    {
        cdos_screen_saver_lock ();
    } 
    else if (g_strcmp0 (signal_name, "Unlock") == 0)
    {
        cs_unlock_screen (self);
    }
}

static void
login1_get_session_done (GObject *source_object,
                         GAsyncResult *res,
                         gpointer user_data)
{
    CdosScreenSaver *self = CDOS_SCREEN_SAVER (user_data);
    GVariant *v;
    GError *error = NULL;
    gchar *session_path;
    GDBusConnection *bus;

    v = g_dbus_proxy_call_finish (self->login1_proxy, res, &error);
    if (error)
    {
        g_warning ("Get login1 session failed: %s", error->message);
        g_error_free (error);
        return;
    }
    g_variant_get (v, "(o)", &session_path);
    g_message ("Login1 session path:%s", session_path);

    bus = g_dbus_proxy_get_connection (self->login1_proxy);
    g_dbus_connection_signal_subscribe (bus,
                                    NULL,
                                    "org.freedesktop.login1.Session",
                                    NULL,
                                    session_path,
                                    NULL,
                                    G_DBUS_SIGNAL_FLAGS_NONE,
                                    (GDBusSignalCallback) login1_session_on_signal_callback,
                                    self,
                                    NULL);
    g_free (session_path);
    g_variant_unref (v);
}

static void
cdos_screen_saver_init (CdosScreenSaver *self)
{
    GError *error = NULL;
    self->is_active = FALSE;
    self->is_protect_active = FALSE;
    self->screensaver_state = IS_NOTHING;
    self->watch_id = 0;
    set_verify_state (self, VERIFY_IDLE, FALSE);
    self->username = g_strdup (g_get_user_name ());
    self->session_active = TRUE;

    self->settings = g_settings_new ("org.cdos.desktop.screensaver");
    settings_on_changed (self, "bio-auth-switch-enabled");

    /* if (g_file_test ("/etc/IDAuthStatus", G_FILE_TEST_EXISTS)) */
    GDBusConnection *conn;

    conn = g_bus_get_sync (G_BUS_TYPE_SESSION, NULL, &error);
    if (conn == NULL)
    {
        g_warning ("%s", error->message);
        g_error_free (error);
        return;
    }

    self->connection = conn;
    self->system_connection = g_bus_get_sync (G_BUS_TYPE_SYSTEM, NULL, &error);

    self->dbus_node_info = g_dbus_node_info_new_for_xml (screensaver_dbus_xml, &error);
    if (NULL == self->dbus_node_info)
    {
        g_warning ("%s", error->message);
        g_error_free (error);
        g_object_unref (conn);
        return;
    }

    if (0 == g_bus_own_name (G_BUS_TYPE_SESSION,
                             SCREESAVER_DBUS_NAME,
                             G_BUS_NAME_OWNER_FLAGS_REPLACE | G_BUS_NAME_OWNER_FLAGS_ALLOW_REPLACEMENT,
                             (GBusAcquiredCallback )on_screensaver_dbus_acquried,
                             NULL,
                             NULL,
                             self,
                             NULL))
    {
        g_warning ("Failed to call g_bus_own_name()");
        g_dbus_node_info_unref (self->dbus_node_info);
        g_object_unref (conn);
        return;
    }

    self->session_proxy = g_dbus_proxy_new_for_bus_sync (G_BUS_TYPE_SESSION, 
                                                         G_DBUS_PROXY_FLAGS_DO_NOT_AUTO_START, 
                                                         NULL, 
                                                         "org.gnome.SessionManager", 
                                                         "/org/gnome/SessionManager",
                                                         "org.gnome.SessionManager", 
                                                         NULL, 
                                                         NULL);
    if (self->session_proxy)
    {
        g_signal_connect (self->session_proxy, 
                          "g-properties-changed", 
                          G_CALLBACK (session_manager_on_properties_changed), 
                          self);
    }

    self->login1_proxy = g_dbus_proxy_new_for_bus_sync (G_BUS_TYPE_SYSTEM, 
                                                        G_DBUS_PROXY_FLAGS_DO_NOT_AUTO_START, 
                                                        NULL, 
                                                        "org.freedesktop.login1", 
                                                        "/org/freedesktop/login1",
                                                        "org.freedesktop.login1.Manager", 
                                                        NULL, 
                                                        NULL);
    const gchar *session_id;
    session_id = g_getenv ("XDG_SESSION_ID");
    if (session_id && self->login1_proxy)
    {
        g_dbus_proxy_call (self->login1_proxy, 
                           "GetSession", 
                           g_variant_new ("(s)", session_id),
                           G_DBUS_CALL_FLAGS_NONE,
                           -1,
                           NULL,
                           (GAsyncReadyCallback) login1_get_session_done,
                           self);
    }

    //人脸识别
    cheese_init (NULL, NULL);

    self->presence = cdos_session_presence_new ();
    g_signal_connect (self->presence, "status-changed", G_CALLBACK (on_presence_status_changed), self);

    ActUserManager *act_manager = act_user_manager_get_default ();
    self->user = act_user_manager_get_user (act_manager, self->username);

    self->global = cdos_global_get ();
    self->stage = (ClutterActor *)cdos_global_get_stage (self->global);

    //g_signal_connect (self->stage, "presented", G_CALLBACK(on_stage_presented), self);

    g_signal_connect (self->user, "notify::is-loaded", G_CALLBACK (user_loaded), self);

    clutter_actor_hide ((ClutterActor *)self);
    clutter_actor_set_reactive ((ClutterActor *)self, TRUE);
    create_user_verify_proxy (self);

    self->message_tray = cdos_message_tray_get_default ();
    self->source_added_id = g_signal_connect (self->message_tray,
                                              "source-added",
                                              G_CALLBACK (on_source_added),
                                              self);
}

CdosScreenSaver *
cdos_screen_saver_new (void)
{
    if (!global_screensaver)
    {
        global_screensaver = g_object_new (CDOS_TYPE_SCREEN_SAVER, NULL);
        g_object_add_weak_pointer ((GObject *)global_screensaver,
                                   (gpointer *)&global_screensaver);
    }
    return global_screensaver;
}

CdosScreenSaver *
cdos_screen_saver_get (void)
{
    return global_screensaver;
}

void
cdos_screen_saver_lock (void)
{
    CdosScreenSaver *self = cdos_screen_saver_get ();
    if (self)
    {
        self->screensaver_state = IS_SCREENSAVER;
        cs_lock_screen (self);
    }
}

static gboolean 
user_in_nopasswdlogin_group(const char *name) {
    struct group *data;
    int i;
    gboolean usr_in_nopwdlogin_group = FALSE;

    while (0 != (data = getgrent())) { 
        if (0 == strcmp(data->gr_name, "nopasswdlogin")) { 
            i = 0;
            while (data->gr_mem[i]) {
                if (0 == strcmp(name, data->gr_mem[i++])) {
                    usr_in_nopwdlogin_group = TRUE;
                    break;
                }
            }
            break;
        }
    }
    endgrent ();

    return usr_in_nopwdlogin_group;
}

gboolean
cdos_screen_saver_get_active (void)
{
    CdosScreenSaver *self = cdos_screen_saver_get ();
    gboolean is_active = FALSE;
    if (self)
    {
        is_active = self->is_active || self->is_protect_active;
    }
    return is_active;
}

