/*
*
* This program is free software; you can redistribute it and/or modify it
* under the terms and conditions of the GNU Lesser General Public License,
* version 2.1, as published by the Free Software Foundation.
*
* This program is distributed in the hope it will be useful, but WITHOUT ANY
* WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
* FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for
* more details.
*
* You should have received a copy of the GNU Lesser General Public License
* along with this program. If not, see <http://www.gnu.org/licenses/>.
*/

#include <glib.h>
#include <gio/gio.h>
#include <gio/gunixinputstream.h>

#include "cdos-search-plocate.h"

typedef struct _CdosSearchPlocate
{
    CdosSearchProvider parent;

    gchar *plocate_path;
    GInputStream *stream;
    GCancellable *cancellable;

    pid_t child_pid;
    gint std_out;
    guint watch_id;

    gint count;
    gchar *buffer;
}CdosSearchPlocate;

G_DEFINE_TYPE (CdosSearchPlocate, cdos_search_plocate, CDOS_TYPE_SEARCH_PROVIDER)

static void
clear_fds (CdosSearchPlocate *self)
{
    GError *error = NULL;
    if (self->watch_id)
    {
        g_source_remove (self->watch_id);
        self->watch_id = 0;
    }
    if (self->child_pid)
    {
        g_spawn_close_pid (self->child_pid);
        self->child_pid = 0;
    }
    if (self->stream)
    {
        g_input_stream_clear_pending (self->stream);
        g_input_stream_close (self->stream,
                              NULL,
                              &error);
        g_object_set_data ((GObject *)self->stream, "cancelled", GUINT_TO_POINTER (1));
        if (error)
        {
            g_warning ("(warn if readched)%s: %s", __func__, error->message);
            g_clear_error (&error);
        }
        else
        {
            g_clear_object (&self->stream);
            g_clear_pointer (&self->buffer, g_free);
        }
        g_clear_object (&self->cancellable);
        self->std_out = 0;
    }
}

static void
cdos_search_plocate_dispose (GObject *object)
{
    CdosSearchPlocate *self = CDOS_SEARCH_PLOCATE (object);

    g_clear_pointer (&self->plocate_path, g_free);
    if (self->cancellable)
    {
        /* 不会立即取消异步 io */
        g_cancellable_cancel (self->cancellable);
    }
    clear_fds (self);

    G_OBJECT_CLASS (cdos_search_plocate_parent_class)->dispose (object);
}

static void
read_stream_cb (GObject *source,
                GAsyncResult *res,
                gpointer user_data)
{
    GError *error = NULL;
    gsize size = 0;
    gchar *data;
    gchar **strv = NULL;
    GPtrArray *results = NULL;
    gboolean success = FALSE;
    guint cancelled;
    CdosSearchPlocate *self;;

    /* CdosSearchPlocate 已经 dispose，而数据可能还没读完 */
    cancelled = GPOINTER_TO_UINT (g_object_get_data (source, "cancelled"));
    if (cancelled)
    {
        g_debug ("%s: Operation has been cancelled", __func__);
        return;
    }

    self = CDOS_SEARCH_PLOCATE (user_data);

    if (g_cancellable_is_cancelled (self->cancellable))
    {
        g_clear_object (&self->cancellable);
        clear_fds (self);
        return;
    }

    success = g_input_stream_read_all_finish (self->stream,
                                              res,
                                              &size,
                                              &error);
    if (!success && size == 0)
    {
        if (error)
        {
            g_warning ("%s: %s", __func__, error->message);
            g_clear_error (&error);
        }
    }
    else if (size > 0)
    {
        data = self->buffer;
        if (data)
        {
            for (gsize i = size - 1; i > 0; i--)
            {
                if (data[i] == '\n')
                {
                    if (i != size -1)
                        data[i + 1] = '\0';
                    break;
                }
            }
            strv = g_strsplit (data, "\n", -1);
        }
    }
    if (strv)
    {
        results = g_ptr_array_new ();
        for (int i = 0; strv[i] && strv[i][0] != '\0'; i++)
        {
            char *path = strv[i];
            if (g_file_test (path, G_FILE_TEST_EXISTS))
            {
                GFile *file;
                CdosSearchResult *result = g_new0 (CdosSearchResult, sizeof (*result));
                file = g_file_new_for_path (path);
                result->name = g_path_get_basename (path);
                result->description = g_path_get_dirname (path);
                result->icon = g_file_icon_new (file);
                g_object_unref (file);
                result->path = path;
                g_ptr_array_add (results, result);
            }
            else
                g_free (strv[i]);
        }
        g_free (strv);
        g_signal_emit_by_name (self, "finished", results);
    }
    clear_fds (self);
}

static void
watch_child (GPid child_pid, gint status, gpointer user_data)
{
    CdosSearchPlocate *self = CDOS_SEARCH_PLOCATE (user_data);

    if (status != 0 || g_cancellable_is_cancelled (self->cancellable))
    {
        g_clear_object (&self->cancellable);
		g_warning ("%s: exec plocate failed or be cancelled", __func__);
        clear_fds (self);
        return;
    }
    self->watch_id = 0;

    gsize size = self->count * 512;
    self->buffer = g_new0 (gchar, size);
    self->stream = g_unix_input_stream_new (self->std_out, TRUE);
    g_object_set_data ((GObject *)self->stream, "cancelled", GUINT_TO_POINTER (0));
    g_input_stream_read_all_async (self->stream,
                                   self->buffer,
                                   size,
                                   G_PRIORITY_DEFAULT,
                                   self->cancellable,
                                   read_stream_cb,
                                   self);

}

static void
cdos_search_plocate_search (CdosSearchProvider *provider,
                            const gchar *str,
                            gint count)
{
    GError *error = NULL;
    char count_str[4] = {0};
    CdosSearchPlocate *self = CDOS_SEARCH_PLOCATE (provider);

    g_clear_object (&self->cancellable);
    self->cancellable = g_cancellable_new ();

	if (count < 0)
		count = 10;
    self->count = count;
    g_snprintf (count_str, sizeof (count_str) - 1, "%d", count);

    const gchar *args[] = {"/usr/bin/plocate", "-l", count_str, str, NULL};
    if (!g_spawn_async_with_pipes (NULL,
                                   (gchar **)args,
                                   NULL,
                                   G_SPAWN_DO_NOT_REAP_CHILD,
                                   NULL,
                                   NULL,
                                   &self->child_pid,
                                   NULL,
                                   &self->std_out,
                                   NULL,
                                   &error))
    {
        g_warning ("%s", error->message);
        g_clear_error (&error);
        return;
    }

    self->watch_id = g_child_watch_add (self->child_pid, watch_child, self);
}

static void
cdos_search_plocate_cancel (CdosSearchProvider *provider)
{
    CdosSearchPlocate *self = CDOS_SEARCH_PLOCATE (provider);
    g_cancellable_cancel (self->cancellable);
    clear_fds (self);
}

static void
cdos_search_plocate_init (CdosSearchPlocate *self)
{
    self->plocate_path = g_find_program_in_path ("plocate");
}

static void
cdos_search_plocate_class_init (CdosSearchPlocateClass *klass)
{
    GObjectClass *gobject_class = G_OBJECT_CLASS (klass);
    gobject_class->dispose = cdos_search_plocate_dispose;

    CdosSearchProviderClass *parent_class = CDOS_SEARCH_PROVIDER_CLASS (klass);
    parent_class->search = cdos_search_plocate_search;
    parent_class->cancel = cdos_search_plocate_cancel;
}

CdosSearchProvider *
cdos_search_plocate_new (void)
{
    return g_object_new (CDOS_TYPE_SEARCH_PLOCATE, NULL);
}
