/**
 * SECTION: cdos-session
 * @short_description: 登入登出会话进程
 * @include: cdos-session.h
 *
 * CdosSessionPresenceClass, CdosSessionInhibitor 类
 * 以及相关的dbus来处理登入登出会话进程.
 */
#include "cdos-global.h"
#include "cdos-session.h"
#include "cdos-util.h"

#define SESSION_BUS_NAME "org.gnome.SessionManager"
#define SESSION_INHIBITOR_INTERFACE "org.gnome.SessionManager.Inhibitor"

/* Presence */
enum
{
    STATUS_CHANGED,
    LAST_SIGNAL0
};
static guint presence_signals[LAST_SIGNAL0] = {0};

typedef struct _CdosSessionPresence
{
    GObject parent_instance;

    GDBusProxy *proxy;
    guint32 status;
}CdosSessionPresence;

G_DEFINE_TYPE (CdosSessionPresence, cdos_session_presence, G_TYPE_OBJECT);

static void
on_presence_status_changed (GDBusProxy *proxy,
                            gchar *sender,
                            gchar *signal,
                            GVariant *parameters,
                            CdosSessionPresence *self)
{
    if (g_strcmp0 (signal, "StatusChanged") == 0)
    {
        GVariantIter iter;
        GVariant *child;
        guint status;

        g_variant_iter_init (&iter, parameters);
        child = g_variant_iter_next_value (&iter);
        status = g_variant_get_uint32 (child);
        g_variant_unref (child);
        self->status = status;
        g_signal_emit (self, presence_signals[STATUS_CHANGED], 0, status);
    }
}

static void
cdos_session_presence_finalize (GObject *gobject)
{
    G_OBJECT_CLASS (cdos_session_presence_parent_class)->finalize (gobject);
}

static void
cdos_session_presence_dispose (GObject *gobject)
{
    CdosSessionPresence *self = CDOS_SESSION_PRESENCE (gobject);

    g_clear_object (&self->proxy);

    G_OBJECT_CLASS (cdos_session_presence_parent_class)->dispose (gobject);
}

static void
cdos_session_presence_class_init (CdosSessionPresenceClass *klass)
{
    GObjectClass *gobject_class = G_OBJECT_CLASS (klass);

    gobject_class->finalize = cdos_session_presence_finalize;
    gobject_class->dispose = cdos_session_presence_dispose;

    presence_signals[STATUS_CHANGED] = g_signal_new ("status-changed",
                                                     CDOS_TYPE_SESSION_PRESENCE,
                                                     G_SIGNAL_RUN_LAST,
                                                     0,
                                                     NULL,
                                                     NULL,
                                                     NULL,
                                                     G_TYPE_NONE,
                                                     1,
                                                     G_TYPE_UINT);
}


static void
cdos_session_presence_init (CdosSessionPresence *self)
{
    GError *error = NULL;

    self->proxy = NULL;

    //建立连接到 /org/gnome/SessionManager/Presence的代理
    self->proxy = g_dbus_proxy_new_for_bus_sync (G_BUS_TYPE_SESSION,
                                                 G_DBUS_PROXY_FLAGS_DO_NOT_LOAD_PROPERTIES,
                                                 NULL,
                                                 SESSION_BUS_NAME,
                                                 "/org/gnome/SessionManager/Presence",
                                                 "org.gnome.SessionManager.Presence",
                                                 NULL,
                                                 &error);

    if (self->proxy == NULL)
    {
        g_warning ("Couldn't get org.gnome.SessionManager.Presence proxy: %s", error->message);
        g_error_free (error);
        return;
    }
    g_signal_connect (self->proxy, "g-signal", G_CALLBACK (on_presence_status_changed), self);
}

/**
 * cdos_session_presence_new:
 *
 * 新建一个 #CdosSessionPresence 类的实例。
 *
 * Returns: 返回一个 #CdosSessionPresence 的实例。
 */
CdosSessionPresence *
cdos_session_presence_new (void)
{
    static CdosSessionPresence *self = NULL;
    if (!self)
    {
        self = g_object_new (CDOS_TYPE_SESSION_PRESENCE, NULL);
    }
    return self;
}

/**
 * cdos_session_presence_get_status:
 * @self: A #CdosSessionPresence
 * @callback: 回调函数
 *
 * 这里通过dbus同步Proxy直接获取类返回值,因此没有使用callback
 *
 * Returns (guint): status
 */
guint
cdos_session_presence_get_status (CdosSessionPresence *self, GFunc callback)
{
    g_return_val_if_fail (CDOS_IS_SESSION_PRESENCE (self), 0);
    // GError *error = NULL;
    guint status = 0;
    GVariant *result = NULL;

    /* not yet connected to cdos-session */
    if (self->proxy == NULL)
    {
        g_warning ("session is not available");
    }

    /* get the session status */
    result = g_dbus_proxy_get_cached_property (self->proxy, "status");

    if (result == NULL)
    {
        g_warning ("no readable status property on %s", g_dbus_proxy_get_interface_name (self->proxy));
    }
    else
    {
        g_variant_get (result, "u", &status);

        g_variant_unref (result);
    }

    return status;
}


/**
 * cdos_session_presence_set_status:
 * @self: A #CdosSessionPresence
 * @status: 状态
 *
 * 这里通过dbus同步Proxy设置获取类返回值,因此没有使用callback
 *
 * Returns: 设置成功返回 %TRUE, 否则返回 %FALSE
 */
gboolean
cdos_session_presence_set_status (CdosSessionPresence *self, guint status)
{
    g_return_val_if_fail (CDOS_IS_SESSION_PRESENCE (self), FALSE);
    GError *error = NULL;
    GVariant *retval;

    retval = g_dbus_proxy_call_sync (
        self->proxy, "SetStatus", g_variant_new ("(u)", status), G_DBUS_CALL_FLAGS_NONE, -1, NULL, &error);

    //发送一个同步dbus设置status
    if (retval == NULL)
    {
        g_warning ("%s:Couldn't call org.gnome.SessionManager: %s", __func__, error->message);
        g_error_free (error);
        return FALSE;
    }
    g_dbus_proxy_set_cached_property (self->proxy, "status", retval);
    g_variant_unref (retval);
    return TRUE;
}


/* Inhibitor */

typedef struct _CdosSessionInhibitor
{
    GObject parent_instance;
    GDBusProxy *proxy;
    gboolean is_loaded;

    char *app_id;
    char *client_id; /* Object Path */
    guint32 flags;
    char *reason;
    guint32 xid;
} CdosSessionInhibitor;

G_DEFINE_TYPE (CdosSessionInhibitor, cdos_session_inhibitor, G_TYPE_OBJECT);

/* signal */
enum
{
    IS_LOADED, // expanded
    LAST_SIGNAL
};
static guint cdos_session_inhibitor_signals[LAST_SIGNAL] = {0};

static void
cdos_session_inhibitor_class_init (CdosSessionInhibitorClass *klass)
{
    // GObjectClass *gobject_class = G_OBJECT_CLASS(klass);

    /**
    * CdosSessionInhibitor::is-loaded:
    * @object: 绑定信号的对象
    * @user_data: 真对回调函数所传入的参数
    *
    * 该信号对 #CdosSessionInhibitor 的载入进行监听
    */
    cdos_session_inhibitor_signals[IS_LOADED] = g_signal_new ("is-loaded",
                                                              G_TYPE_FROM_CLASS (klass),
                                                              G_SIGNAL_RUN_LAST,
                                                              0,
                                                              NULL,
                                                              NULL,
                                                              NULL,
                                                              G_TYPE_NONE,
                                                              0);
}

static void
cdos_session_inhibitor_init (CdosSessionInhibitor *self)
{

}

gchar *
cdos_session_inhibitor_get_app_id (CdosSessionInhibitor *self)
{
    GError *error = NULL;
    GVariant *v;
    gchar *app_id = NULL, *new_app_id = NULL;

    v = g_dbus_proxy_call_sync (self->proxy, 
                                "GetAppId", 
                                NULL, 
                                G_DBUS_CALL_FLAGS_NONE, 
                                -1, 
                                NULL, 
                                &error);
    if (error)
    {
        g_warning ("Inhibitor get app id failed:%s", error->message);
        g_error_free (error);
        return app_id;
    }
    if (v)
    {
        g_variant_get (v, "(s)", &app_id);
        g_variant_unref (v);
        if (app_id)
            new_app_id = g_strdup (app_id);
        g_free (app_id);
    }
    return new_app_id;
}

gchar *
cdos_session_inhibitor_get_reason (CdosSessionInhibitor *self)
{
    GError *error = NULL;
    GVariant *v;
    gchar *reason = NULL;

    v = g_dbus_proxy_call_sync (self->proxy, 
                                "GetReason", 
                                NULL, 
                                G_DBUS_CALL_FLAGS_NONE, 
                                -1, 
                                NULL, 
                                &error);
    if (error)
    {
        g_warning ("Inhibitor get reason failed:%s", error->message);
        g_error_free (error);
        return reason;
    }
    if (v)
    {
        g_variant_get (v, "(s)", &reason);
        g_variant_unref (v);
    }
    return reason;
}

static void
cdos_session_inhibitor_post_init (CdosSessionInhibitor *self, gchar *object_path)
{
    GError *error = NULL;
    self->proxy = g_dbus_proxy_new_for_bus_sync (G_BUS_TYPE_SESSION,
                                                G_DBUS_PROXY_FLAGS_DO_NOT_LOAD_PROPERTIES,
                                                NULL,
                                                SESSION_BUS_NAME,
                                                object_path,
                                                SESSION_INHIBITOR_INTERFACE,
                                                NULL,
                                                &error);
    if (error)
    {
        g_warning ("Create dbus proxy failed: %s", error->message);
        g_error_free (error);
        return;
    }
}

CdosSessionInhibitor *
cdos_session_inhibitor_new (gchar *object_path)
{
    CdosSessionInhibitor *instance = g_object_new (CDOS_TYPE_SESSION_INHIBITOR, NULL);
    cdos_session_inhibitor_post_init (instance, object_path);
    return instance;
}

/* Session
 * 与session manager建立dbus连线
 */

typedef struct _CdosSessionManager
{
    GObject parent;
    GDBusProxy *proxy;
    guint32 inhibited_actions;
} CdosSessionManager;

G_DEFINE_TYPE (CdosSessionManager, cdos_session_manager, G_TYPE_OBJECT)

static CdosSessionManager *cdos_session_manager = NULL;

static void
cdos_session_manager_init (CdosSessionManager *self)
{
    static gboolean is_connecting = FALSE;
    GError *error = NULL;
    if (is_connecting)
        return;

    is_connecting = TRUE;

    self->proxy = g_dbus_proxy_new_for_bus_sync (G_BUS_TYPE_SESSION,
                              G_DBUS_PROXY_FLAGS_DO_NOT_AUTO_START,
                              NULL,
                              SESSION_BUS_NAME,
                              "/org/gnome/SessionManager",
                              "org.gnome.SessionManager",
                              NULL,
                              &error);
    if (error)
    {
        g_warning ("Connect session manager failed: %s", error->message);
        g_error_free (error);
        return;
    }
    cdos_session_manager = self;
}

static void
cdos_session_manager_class_init (CdosSessionManagerClass *klass)
{

}

/**
 * cdos_session_manager_new:
 *
 * 获取一个 #CdosSessionManager 类的实例。如果没有就新建一个。
 *
 * Return value: (transfer none): #CdosSessionManager
 */
CdosSessionManager *
cdos_session_new (void)
{
    if (cdos_session_manager)
        return cdos_session_manager;
    cdos_session_manager = g_object_new (CDOS_TYPE_SESSION_MANAGER, NULL);
    return cdos_session_manager;
}

static void
session_logout_cb (GObject *source_object, GAsyncResult *res, gpointer user_data)
{
    GVariant *result;
    GError *error = NULL;

    result = g_dbus_proxy_call_finish (G_DBUS_PROXY (source_object), res, &error);
    if (result == NULL)
    {
        g_warning (_("Couldn't logout using cdos-session: %s"), error->message);
        g_error_free (error);
    }
    else
    {
        g_variant_unref (result);
    }
}

static void
session_shutdown_or_reboot_cb (GObject *source_object, GAsyncResult *res, gpointer user_data)
{
    GVariant *result;
    GError *error = NULL;

    result = g_dbus_proxy_call_finish (G_DBUS_PROXY (source_object), res, &error);
    if (result == NULL)
    {
        g_warning ("%s", error->message);
        g_error_free (error);
    }
    else
        g_variant_unref (result);
}

/**
 * cdos_session_logout:
 * @mode: 模式
 *
 * 通过DBus代理获取当前模式进行注销;
 *
 */
void
cdos_session_logout (guint mode)
{
    if (!cdos_session_manager)
        return;
    if (!cdos_session_manager->proxy)
        return;

    g_dbus_proxy_call (cdos_session_manager->proxy,
                       "Logout",
                       g_variant_new ("(u)", mode),
                       G_DBUS_CALL_FLAGS_NONE,
                       -1, NULL,
                       session_logout_cb,
                       NULL);
}

/**
 * cdos_session_shutdown:
 *
 * 通过session会话进行关机
 *
 */
void
cdos_session_shutdown (void)
{
    if (!cdos_session_manager)
        return;
    if (!cdos_session_manager->proxy)
        return;
    g_dbus_proxy_call (cdos_session_manager->proxy,
                       "Shutdown",
                       NULL,
                       G_DBUS_CALL_FLAGS_NONE,
                       -1,
                       NULL,
                       session_shutdown_or_reboot_cb,
                       NULL);
}

/**
 * cdos_session_can_shutdown:
 *
 * 通过DBus获取当前代理的 "Canshutdown", 查询是否可以关机
 *
 * Return value: %TRUE 可以关机; %FALSE 则不可以.
 */
gboolean
cdos_session_can_shutdown (void)
{
    if (!cdos_session_manager)
        return TRUE;
    if (!cdos_session_manager->proxy)
        return TRUE;

    GError *error = NULL;
    g_autoptr (GVariant) result = NULL;
    gboolean can_shutdown;

    result = g_dbus_proxy_call_sync (cdos_session_manager->proxy,
                                     "CanShutdown",
                                     NULL,
                                     G_DBUS_CALL_FLAGS_NONE,
                                     -1,
                                     NULL,
                                     &error);
    if (result == NULL)
    {
        g_warning ("%s", error->message);
        g_error_free (error);
        return TRUE;
    }
    g_variant_get (result, "(b)", &can_shutdown);
    return can_shutdown;
}

/**
 * cdos_session_reboot:
 *
 * 通过session会话进行关机
 */
void
cdos_session_reboot (void)
{
    if (!cdos_session_manager)
        return;
    if (!cdos_session_manager->proxy)
        return;
    g_dbus_proxy_call (cdos_session_manager->proxy,
                       "Reboot",
                       NULL,
                       G_DBUS_CALL_FLAGS_NONE,
                       -1,
                       NULL,
                       session_shutdown_or_reboot_cb,
                       NULL);
}

/**
 * cdos_session_get_inhibitors:
 *
 * 获取当前阻止注销、锁屏、关机等操作的应用
 *
 * Return value: (transfer full): 包含Inhibitor Dbus path 的数组
 */
char **
cdos_session_get_inhibitors (void)
{
    if (!cdos_session_manager)
        return NULL;
    if (!cdos_session_manager->proxy)
        return NULL;

    GError *error = NULL;
    g_autoptr (GVariant) v = NULL;
    gchar **results= NULL;

    v = g_dbus_proxy_call_sync (cdos_session_manager->proxy,
                                   "GetInhibitors",
                                   NULL,
                                   G_DBUS_CALL_FLAGS_NONE,
                                   -1,
                                   NULL,
                                   &error);
    if (v == NULL)
    {
        g_warning ("%s", error->message);
        g_error_free (error);
        return NULL;
    }
    if (v)
    {
        g_variant_get (v, "(^ao)", &results);
    }
    if (results && g_strv_length (results) == 0)
    {
        g_strfreev (results);
        results = NULL;
    }
    return results;
}

void
cdos_session_suspend (void)
{
    cdos_util_spawn_command_line("dbus-send --system --print-reply --dest=\"org.freedesktop.login1\" /org/freedesktop/login1 org.freedesktop.login1.Manager.Suspend boolean:'true'");
}

void 
cdos_session_hibernate (void)
{
    cdos_util_spawn_command_line("dbus-send --system --print-reply --dest=\"org.freedesktop.login1\" /org/freedesktop/login1 org.freedesktop.login1.Manager.Hibernate boolean:'true'");
}

guint32
cdos_session_get_inhibited_actions (void)
{
    GVariant *v;
    if (!cdos_session_manager)
        return 0;
    if (!cdos_session_manager->proxy)
        return 0;

    v = g_dbus_proxy_get_cached_property (cdos_session_manager->proxy,
                                          "InhibitedActions");
    if (!v)
        return 0;
    cdos_session_manager->inhibited_actions = g_variant_get_uint32 (v);
    g_variant_unref (v);
    return cdos_session_manager->inhibited_actions;
}
