/*
 * Copyright 2021 <cdos>
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms and conditions of the GNU Lesser General Public License,
 * version 2.1, as published by the Free Software Foundation.
 *
 * This program is distributed in the hope it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for
 * more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
/*
 * SECTION:  cdos-indicator-manager.c
 * @Title: cdos-indicator-manager
 * @Short_Description:  libappindicator and StatusNotifier support.
 */

#include "common.h"
#include "cdos-sn-item.h"
#include "cdos-indicator-manager.h"
#include "cdos-popup-menu.h"

typedef enum _CdosSnItemStatus
{
    CDOS_SN_ITEM_PASSIVE,
    CDOS_SN_ITEM_ACTIVE,
    CDOS_SN_ITEM_NEEDS_ATTENTION
}CdosSnItemStatus;

struct _CdosSnItem
{
    StWidget          parent;

    gchar            *bus_name;
    gchar            *object_path;
    gchar            *bus_id;
    GDBusProxy       *proxy;
    GCancellable     *cancellable;

    char             *category;
    gchar            *id;
    gchar            *title;
    CdosSnItemStatus  status;
    guint32           window_id;

    gchar            *icon_name;
    ClutterImage     *icon_image;
    gchar            *overlay_icon_name;
    ClutterImage     *overlay_icon_image;
    gchar            *attention_icon_name;
    ClutterImage     *attention_icon_image;
    gchar            *attention_movie_name;

    gchar            *tooltip_icon_name;
    ClutterImage     *tooltip_icon_image;
    gchar            *tooltip_title;
    gchar            *tooltip_description;

    gboolean          item_is_menu;
    gchar            *menu_object_path;
};

G_DEFINE_TYPE(CdosSnItem, cdos_sn_item, ST_TYPE_WIDGET)

enum
{
    PROP_0,
    PROP_BUS_NAME,
    PROP_OBJECT_PATH,
    N_PROPERTIES
};

static GParamSpec *sn_item_properties[N_PROPERTIES] = {NULL, };

static void
cdos_sn_item_on_name_owner_changed (GDBusProxy   *proxy,
                                    GVariant     *parameters,
                                    CdosSnItem   *self)
{
    gchar *name_owner = g_dbus_proxy_get_name_owner (proxy);
    g_debug ("%s: %s", __func__, name_owner);
    g_free (name_owner);
}

static void
cdos_sn_item_title_changed (CdosSnItem *self)
{
    gchar *title = NULL;
    GVariant *v = g_dbus_proxy_get_cached_property (self->proxy, "Title");
    g_clear_pointer (&self->title, g_free);
    if (v)
    {
        title = g_variant_dup_string (v, NULL);
        g_debug ("%s %s", __func__, title);
        self->title = title;
        g_variant_unref (v);
    }
}

static void
cdos_sn_item_set_image_data (ClutterImage *image, GVariant *image_data)
{
    GVariantIter iter;
    gint32 width, height;
    GVariant *pixel_data = NULL, *tmp_data = NULL;
    gsize size = 0;
    const guchar *data;
    if (!g_variant_is_of_type (image_data, G_VARIANT_TYPE ("a(iiay)")))
    {
        g_warning ("Image data is not the correct type");
        return;
    }
    g_variant_iter_init (&iter, image_data);
    while (g_variant_iter_loop (&iter, "(ii@ay)", &width, &height, &pixel_data))
    {
        if (width > 0 && height > 0 &&
            pixel_data &&
            (gsize)width * height > size)
        {
            /* Find the largest image */
            size = g_variant_get_size (pixel_data) / 4;
            if (size == (gsize)(width * height))
            {
                tmp_data = pixel_data;
                break;
            }
        }
    }

    /*
    data = (guchar *) g_memdup (g_variant_get_data (pixel_data),
                                g_variant_get_size (pixel_data));
    */
    data = g_variant_get_data (tmp_data);

    GError *error = NULL;
    clutter_image_set_data (image,
                            data,
                            COGL_PIXEL_FORMAT_ARGB_8888,
                            width, height,
                            4 * width,
                            &error);
    g_variant_unref (pixel_data);

    if (error)
    {
        g_warning ("%s", error->message);
        g_error_free (error);
    }
}

static void
cdos_sn_item_icon_changed (CdosSnItem *self)
{
    g_debug ("%s", __func__);

    gchar *icon_name = NULL;
    GVariant *v = NULL;
    v = g_dbus_proxy_get_cached_property (self->proxy, "IconName");
    if (v)
    {
        icon_name = g_strdup (g_variant_get_string (v, NULL));
        g_free (self->icon_name);
        self->icon_name = icon_name;
        g_variant_unref (v);
        g_debug ("IconName %s", icon_name);
    }

    if (!v)
    {
        v = g_dbus_proxy_get_cached_property (self->proxy, "IconPixmap");
        if (v)
        {
            g_free (self->icon_name);
            self->icon_name = NULL;
            if (!self->icon_image)
            {
                self->icon_image = (ClutterImage *)clutter_image_new ();
            }
            cdos_sn_item_set_image_data (self->icon_image, v);
            g_variant_unref (v);
        }
    }

    v = g_dbus_proxy_get_cached_property (self->proxy, "OverlayIconName");
    if (v)
    {
        gchar *overlay_icon_name = g_strdup (g_variant_get_string (v, NULL));
        g_free (self->overlay_icon_name);
        self->overlay_icon_name = overlay_icon_name;
        g_variant_unref (v);
        g_debug ("OverlayIconName %s", overlay_icon_name);
    }
    else
    {
        v = g_dbus_proxy_get_cached_property (self->proxy, "OverlayIconPixmap");
        if (v)
        {
            g_free (self->overlay_icon_name);
            self->overlay_icon_name = NULL;
            g_variant_unref (v);
        }
    }

    v = g_dbus_proxy_get_cached_property (self->proxy, "AttentionIconName");
    if (v)
    {
        gchar *attention_icon_name = g_strdup (g_variant_get_string (v, NULL));
        g_free (self->attention_icon_name);
        self->attention_icon_name = attention_icon_name;
        g_variant_unref (v);
        g_debug ("AttentionIconName %s", attention_icon_name);
    }
    else
    {
        v = g_dbus_proxy_get_cached_property (self->proxy, "AttentionIconPixmap");
        if (v)
        {
            g_free (self->attention_icon_name);
            self->attention_icon_name = NULL;
            g_variant_unref (v);
        }
    }
}

static void
cdos_sn_item_tooltip_changed (CdosSnItem *self)
{
    g_debug ("%s", __func__);
    GVariant *v = g_dbus_proxy_get_cached_property (self->proxy, "ToolTip");
    if (!v)
        return;
    const gchar *tooltip_type = g_variant_get_type_string (v);
    if (g_strcmp0 (tooltip_type, "(sa(iiay)ss)") == 0)
    {
        /* Get Title and Description text */
        gchar *title = NULL, *description = NULL;
        g_variant_get (v, "(sa(iiay)ss)", NULL, NULL, &title, &description);
        g_debug ("ToolTip title:%s, description: %s", title, description);
        g_free (self->tooltip_title);
        g_free (self->tooltip_description);
        self->tooltip_title = title;
        self->tooltip_description = description;
    }
    else if (g_strcmp0 (tooltip_type, "(s)") == 0)
    {
        const gchar *tooltip = g_variant_get_string (v, NULL);
        g_debug ("tooltip: %s", tooltip);
        g_free (self->tooltip_title);
        g_free (self->tooltip_description);
        self->tooltip_description = NULL;
        self->tooltip_title = g_strdup (tooltip);
    }
    else
    {
        g_free (self->tooltip_title);
        g_free (self->tooltip_description);
        self->tooltip_title = NULL;
        self->tooltip_description = NULL;
    }
}

static void
cdos_sn_item_status_changed (CdosSnItem *self)
{
    const char *status;
    GVariant *v = g_dbus_proxy_get_cached_property (self->proxy, "Status");
    if (v)
    {
        status = g_variant_get_string (v, NULL);
        g_debug ("%s %s", __func__, status);
        if (g_strcmp0 (status, "Active") == 0)
            self->status = CDOS_SN_ITEM_ACTIVE;
        else if (g_strcmp0 (status, "NeedsAttention") == 0)
            self->status = CDOS_SN_ITEM_NEEDS_ATTENTION;
        else
            self->status = CDOS_SN_ITEM_PASSIVE;
        g_variant_unref (v);
    }
}

static void
cdos_sn_item_get_menu (CdosSnItem *self)
{
    GVariant *v = g_dbus_proxy_get_cached_property (self->proxy, "ItemIsMenu");
    if (v)
    {
        self->item_is_menu = g_variant_get_boolean (v);
        g_variant_unref (v);
    }

    v = g_dbus_proxy_get_cached_property (self->proxy, "Menu");
    g_clear_pointer (&self->menu_object_path, g_free);
    if (v)
    {
        self->menu_object_path = g_strdup (g_variant_get_string (v, NULL));
        g_variant_unref (v);
    }
}

static void
cdos_sn_item_on_signal (GDBusProxy  *proxy,
                        char        *sender,
                        char        *signal,
                        GVariant    *parameters,
                        CdosSnItem  *self)
{
#define CMP(sig) g_strcmp0(signal, #sig)
    if (!CMP(NewTitle))
    {
        cdos_sn_item_title_changed (self);
    }
    else if (!CMP(NewIcon) ||
             !CMP(NewAttentionIcon) ||
             !CMP(NewOverlayIcon))
    {
        cdos_sn_item_icon_changed (self);
    }
    else if (!CMP(NewToolTip))
    {
        cdos_sn_item_tooltip_changed (self);
    }
    else if (!CMP(NewStatus))
    {
        cdos_sn_item_status_changed (self);
    }
#undef CMP
}

static void
cdos_sn_item_get_properties (CdosSnItem *self)
{
    cdos_sn_item_title_changed (self);
    cdos_sn_item_icon_changed (self);
    cdos_sn_item_tooltip_changed (self);
    cdos_sn_item_get_menu (self);
}

static void
on_item_proxy_ready (GObject *source,
                     GAsyncResult *res,
                     CdosSnItem *self)
{
    GError *error = NULL;
    self->proxy = g_dbus_proxy_new_for_bus_finish (res, &error);
    if (error)
    {
        g_warning ("%s", error->message);
        g_error_free (error);
        return;
    }

    g_signal_connect (self->proxy, "notify::g-name-owner", G_CALLBACK(cdos_sn_item_on_name_owner_changed), self);

    g_signal_connect (self->proxy, "g-signal",
                      G_CALLBACK (cdos_sn_item_on_signal), self);
    cdos_sn_item_get_properties (self);
}

static void
cdos_sn_item_get_property (GObject    *object,
                           guint       prop_id,
                           GValue     *value,
                           GParamSpec *pspec)
 {
    CdosSnItem *self = CDOS_SN_ITEM (object);

    switch (prop_id)
    {
        case PROP_BUS_NAME:
            g_value_set_string (value, self->bus_name);
            break;

        case PROP_OBJECT_PATH:
            g_value_set_string (value, self->object_path);
            break;

        default:
            G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
            break;
    }
}

static void
cdos_sn_item_set_property (GObject      *object,
                           guint         prop_id,
                           const GValue *value,
                           GParamSpec   *pspec)
{
    CdosSnItem *self = CDOS_SN_ITEM (object);

    switch (prop_id)
    {
        case PROP_BUS_NAME:
            g_free (self->bus_name);
            self->bus_name = g_value_dup_string (value);
            break;

        case PROP_OBJECT_PATH:
            g_free (self->object_path);
            self->object_path = g_value_dup_string (value);
            break;

        default:
            G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
            break;
    }
}

static void
cdos_sn_item_constructed (GObject *object)
{
    CdosSnItem *self = CDOS_SN_ITEM (object);
    GObjectClass *parent_class = G_OBJECT_CLASS (cdos_sn_item_parent_class);
    if (parent_class->constructed &&
        parent_class->constructed != cdos_sn_item_constructed)
    {
        parent_class->constructed (object);
    }
    g_debug ("%s", __func__);
    self->id = g_strconcat (self->bus_name, self->object_path, NULL);
    self->cancellable = g_cancellable_new ();
    g_dbus_proxy_new_for_bus (G_BUS_TYPE_SESSION,
                              G_DBUS_PROXY_FLAGS_GET_INVALIDATED_PROPERTIES,
                              NULL,
                              self->bus_name,
                              self->object_path,
                              "org.kde.StatusNotifierItem",
                              self->cancellable,
                              (GAsyncReadyCallback)on_item_proxy_ready,
                              self);
}

static void
cdos_sn_item_dispose (GObject *object)
{
    CdosSnItem *self = CDOS_SN_ITEM (object);

    g_clear_object (&self->icon_image);
    g_clear_object (&self->attention_icon_image);
    g_clear_object (&self->overlay_icon_image);
    g_clear_object (&self->tooltip_icon_image);

    g_clear_object (&self->cancellable);
    g_clear_object (&self->proxy);

    G_OBJECT_CLASS (cdos_sn_item_parent_class)->dispose (object);
}

static void
cdos_sn_item_finalize (GObject *object)
{
    CdosSnItem *self = CDOS_SN_ITEM (object);

    g_free (self->bus_name);
    g_free (self->object_path);
    g_free (self->bus_id);
    g_free (self->id);

    g_free (self->title);
    g_free (self->tooltip_icon_name);
    g_free (self->tooltip_title);
    g_free (self->tooltip_description);

    g_free (self->icon_name);
    g_free (self->attention_icon_name);
    g_free (self->attention_movie_name);
    g_free (self->overlay_icon_name);

    g_free (self->menu_object_path);

    G_OBJECT_CLASS (cdos_sn_item_parent_class)->finalize (object);
}

static void
cdos_sn_item_init (CdosSnItem *self)
{
}

static void
cdos_sn_item_class_init (CdosSnItemClass *klass)
{
    GObjectClass *object_class = G_OBJECT_CLASS (klass);
    object_class->get_property = cdos_sn_item_get_property;
    object_class->set_property = cdos_sn_item_set_property;
    object_class->constructed  = cdos_sn_item_constructed;
    object_class->dispose      = cdos_sn_item_dispose;
    object_class->finalize     = cdos_sn_item_finalize;
  
    sn_item_properties[PROP_BUS_NAME] = 
        g_param_spec_string ("bus-name", "Bus name", "Bus name",
                             NULL,
                             G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS | G_PARAM_CONSTRUCT);
  
    sn_item_properties[PROP_OBJECT_PATH] = 
        g_param_spec_string ("object-path", "Object path", "Object path",
                             NULL,
                             G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS | G_PARAM_CONSTRUCT);
  
    g_object_class_install_properties (object_class,
                                       N_PROPERTIES,
                                       sn_item_properties);
}

CdosSnItem *
cdos_sn_item_new (const gchar *bus_name, const gchar *object_path)
{
    CdosSnItem *self;
    self = g_object_new (CDOS_TYPE_SN_ITEM,
                         "bus-name", bus_name,
                         "object-path", object_path,
                         NULL);
    return self;
}

void
cdos_sn_item_activate (CdosSnItem *self,
                       int         x,
                       int         y)
{

}

void
cdos_sn_item_secondary_activate (CdosSnItem *self,
                                 int         x,
                                 int         y)
{

}

void
cdos_sn_item_scroll (CdosSnItem *self,
                     int         delta,
                     const char *direction)
{

}
