/**
 * SECTION: cdos-sound-manager
 * @short_description: 管理声音
 * @include: cdos-sound-manager.h
 *
 * 用来处理 #CdosSoundManagerClass 的相关方法.
 */
#include <gio/gio.h>
#include "cdos-sound-manager.h"
#include "string.h"
#include "cdos-global.h"

enum SoundKeys
{
    Switch,
    Close,
    Map,
    Minimize,
    Maximize,
    Unmaximize,
    Tile,
    Login,
    Plug,
    Unplug,
    SoundKeysLast,
};
enum DesktopKeys
{
    Volume,
    DesktopKeysLast
};

typedef struct _KVTable
{
    const char *name;
    gint type;
    gboolean enabled;
    char *file;
} KVTable;

static KVTable kvtable[] = {
    {"switch", Switch, 0, NULL},
    {"close",  Close,  0, NULL},
    {"map",    Map,    0, NULL},
    {"minimize", Minimize, 0, NULL},
    {"maximize", Maximize, 0, NULL},
    {"unmaximize", Unmaximize, 0, NULL},
    {"tile", Tile, 0, NULL},
    {"login",  Login,  0, NULL},
    {"plug", Plug,     0, NULL},
    {"unplug", Unplug, 0, NULL},
    {"volume", Volume, 0, NULL}
};

/*声明*/
static void cdos_sound_manager_cache_settings (GSettings *settings, gchar *key, CdosSoundManager *self);
static void cdos_sound_manager_cache_desktop_settings (GSettings *desptop_settings, gchar *key, CdosSoundManager *self);

static void
cdos_sound_manager_init (CdosSoundManager *self)
{
    self->settings = g_settings_new ("org.cdos.sounds");
    self->desktop_settings = g_settings_new ("org.cdos.desktop.sound");
    cdos_sound_manager_cache_settings (self->settings, NULL, self);
    cdos_sound_manager_cache_desktop_settings (self->settings, NULL, self);

    /*参数依次是:实例, 信号名, 回调函数, 回调函数参数*/
    g_signal_connect (self->settings, "changed", G_CALLBACK (cdos_sound_manager_cache_settings), self);
    g_signal_connect (self->desktop_settings, "changed", G_CALLBACK (cdos_sound_manager_cache_desktop_settings), self);
}

static void
cdos_sound_manager_cache_settings (GSettings *settings, gchar *key, CdosSoundManager *self)
{
    const gchar *str_enabled = "-enabled";
    const gchar *str_file = "-file";

    gint i;
    for (i = Switch; i < SoundKeysLast; i++)
    {
        KVTable *t = &kvtable[i];

        gchar *str_key_enabled = g_strconcat (t->name, str_enabled, NULL);
        t->enabled = g_settings_get_boolean (self->settings, str_key_enabled);

        gchar *str_key_file = g_strconcat (t->name, str_file, NULL);
        t->file = g_settings_get_string (self->settings, str_key_file);

        g_free (str_key_enabled);
        g_free (str_key_file);
    }
}

static void
cdos_sound_manager_cache_desktop_settings (GSettings *desktop_settings, gchar *key, CdosSoundManager *self)
{
    // gchar *k = self->desktop_keys[0];//volume
    const gchar *str_sound_enabled = "-sound-enabled";
    const gchar *str_sound_file = "-sound-file";

    gint i = 0;
    for (i = Volume; i < DesktopKeysLast; i++)
    {
        KVTable *t = &kvtable[i + 10]; // i=10
        gchar *str_key_sound_enabled = g_strconcat (t->name, str_sound_enabled, NULL);
        t->enabled = g_settings_get_boolean (self->desktop_settings, str_key_sound_enabled);
        gchar *str_key_sound_file = g_strconcat (t->name, str_sound_file, NULL);
        t->file = g_settings_get_string (self->desktop_settings, str_key_sound_file);
        g_free (str_key_sound_enabled);
        g_free (str_key_sound_file);
    }
}

/**
 * cdos_sound_manager_play:
 * @sound: 要播放的声音名称
 *
 * 播放声音
 *
 */
void
cdos_sound_manager_play (const gchar *sound)
{
    if (sound == NULL)
        return;

    char *file = NULL;
    gint i;
    for (i = Switch; i < SoundKeysLast + DesktopKeysLast; i++)
    {
        KVTable *t = &kvtable[i];
        if (strcmp (t->name, sound) == 0 && t->enabled)
        {
            file = t->file;
            g_debug ("Play sound: %s, sound file: %s", t->name, file);
            cdos_global_play_sound_file (cdos_global_get (), 0, file, NULL, NULL);
            return;
        }
    }
}

/**
 * cdos_sound_manager_new:
 *
 * 新建一个 #CdosSoundManager 类的实例。
 *
 * Returns (CdosSoundManager): 返回一个 #CdosSoundManager 的实例。
 */
CdosSoundManager *
cdos_sound_manager_new (void)
{
    static CdosSoundManager *sound_manager = NULL;
    if (sound_manager == NULL)
    {
        sound_manager = g_new0 (CdosSoundManager, 1);
        cdos_sound_manager_init (sound_manager);
    }
    return sound_manager;
}

void
cdos_sound_manager_destroy (CdosSoundManager *self)
{
    g_clear_object (&self->settings);
    g_clear_object (&self->desktop_settings);
    g_free (self);
}
