/**
 * Copyright (c) 2016-2018 CPU and Fundamental Software Research Center, CAS
 *
 * This software is published by the license of CPU-OS License，you can use and
 * distibuted this software under this License. See CPU-OS Liense for more detail.
 *
 * You should have receive a copy of CPU-OS License. If not, please contact us
 * by email <support_os@cpu-os.ac.cn>
 *
 **/
/**
 * SECTION: cdos-source
 * @short_description: 创建,处理 CdosSource
 * @include: cdos-source.h
 *
 * CdostypeClass 类用来创建,处理 #CdosSource
 */
#include "cdos-source.h"
#include "cdos-start.h"

G_DEFINE_TYPE (CdosSource, cdos_source, ST_TYPE_WIDGET);

/* signal */
enum {
    NOTIFICATION_ADDED,      //notification-added
    NOTIF,         //notify
    CLOSE_ALL,         //close all
    LAST_SIGNAL
};

static guint cdos_source_signals[LAST_SIGNAL] = {0};

enum {
    PROP_0,
    PROP_NAME,
    PROP_NUM
};

static void
notification_destroyed (CdosNotification *notification, gpointer user_data)
{
    CdosSource *self = CDOS_SOURCE (user_data);
    self->notifications = g_list_remove (self->notifications, notification);
    if (self->notifications == NULL)
    {
        cdos_source_destroy (self, 0);
        return;
    }
}

static gboolean
close_source (ClutterActor *close,
              ClutterEvent *event,
              CdosSource *self)
{
    cdos_source_clear_all (self);
    return FALSE;
}

static void
cdos_source_set_property (GObject      *gobject,
                          guint         prop_id,
                          const GValue *value,
                          GParamSpec   *pspec)
{
    CdosSource *self = CDOS_SOURCE (gobject);

    switch (prop_id)
    {
        case PROP_NAME:
            cdos_source_set_name (self, g_value_get_string (value));
            break;

        default:
            G_OBJECT_WARN_INVALID_PROPERTY_ID (gobject, prop_id, pspec);
            break;
    }
}

static void
cdos_source_get_property (GObject    *gobject,
                         guint       prop_id,
                         GValue     *value,
                         GParamSpec *pspec)
{
    CdosSource *self = CDOS_SOURCE (gobject);
    switch (prop_id)
    {
        case PROP_NAME:
            g_value_set_string (value, self->name);
            break;

        default:
            G_OBJECT_WARN_INVALID_PROPERTY_ID (gobject, prop_id, pspec);
            break;
    }
}

static void
cdos_source_dispose (GObject *source)
{
    CdosSource *self = CDOS_SOURCE (source);
    for (GList *l = self->notifications; l; l = l->next)
    {
        CdosNotification *notification = l->data;
        g_signal_handlers_disconnect_by_data (notification, self);
        clutter_actor_destroy ((ClutterActor *)notification);
    }

    G_OBJECT_CLASS(cdos_source_parent_class)->dispose (source);
}

static void
cdos_source_finalize (GObject *obj)
{
    CdosSource *self = CDOS_SOURCE (obj);

    g_list_free (self->notifications);

    G_OBJECT_CLASS(cdos_source_parent_class)->finalize(obj);
}

static void
cdos_source_allocate (ClutterActor *actor,
                      const ClutterActorBox *box,
                      ClutterAllocationFlags flags)
{
    CdosSource *self = CDOS_SOURCE (actor);
    StThemeNode *theme_node = st_widget_get_theme_node ((StWidget *)actor);
    ClutterActorBox content_box, child_box = {0,};
    gfloat avail_width;

    clutter_actor_set_allocation (actor, box, flags);

    st_theme_node_get_content_box (theme_node, box, &content_box);

    gfloat width = content_box.x2 - content_box.x1;
    gfloat height = content_box.y2 - content_box.y1;
    gfloat spacing = 8;

    gfloat icon_min_width, icon_nat_width, icon_min_height, icon_nat_height;
    gfloat name_min_width, name_nat_width, name_min_height, name_nat_height;
    gfloat close_min_width, close_nat_width, close_min_height, close_nat_height;

    clutter_actor_get_preferred_size (self->icon,
                                      &icon_min_width, &icon_min_height,
                                      &icon_nat_width, &icon_nat_height);
    clutter_actor_get_preferred_size ((ClutterActor *)self->name_label,
                                      &name_min_width, &name_min_height,
                                      &name_nat_width, &name_nat_height);
    clutter_actor_get_preferred_size ((ClutterActor *)self->close,
                                      &close_min_width, &close_min_height,
                                      &close_nat_width, &close_nat_height);
    /* close button */
    child_box.x2 = content_box.x2;
    child_box.x1 = child_box.x2 - close_nat_width;
    child_box.y1 = (height - close_nat_height) / 2 + content_box.y1;
    child_box.y2 = child_box.y1 + close_nat_height;
    clutter_actor_allocate ((ClutterActor *)self->close, &child_box, flags);

    /* name */
    avail_width = width - close_nat_width;

    if (avail_width > name_nat_width + icon_nat_width + 2 * spacing)
    {
        //gfloat center = width / 2.0;
        gfloat x = (width - name_nat_width - icon_nat_width) / 2.0;
        child_box.x1 = x;
        child_box.x2 = child_box.x1 + icon_nat_width;
        child_box.y1 = content_box.y1 + (height - icon_nat_height) / 2.0;
        child_box.y2 = child_box.y1 + icon_nat_height;
        clutter_actor_allocate (self->icon, &child_box, flags);

        x = child_box.x2 + spacing;
        child_box.x1 = (int)(x + 0.5);
        child_box.x2 = child_box.x1 + name_nat_width;
        child_box.y1 = (int)(content_box.y1 + (height - name_nat_height) / 2.0 + 0.5);
        child_box.y2 = child_box.y1 + name_nat_height;
        clutter_actor_allocate ((ClutterActor *)self->name_label, &child_box, flags);
    }
}

static void
cdos_source_on_hover (StWidget *widget,
                      GParamSpec *pspec,
                      CdosSource *self)
{
    gboolean hover = st_widget_get_hover (widget);

    if (hover)
        clutter_actor_show ((ClutterActor *)self->close);
    else
        clutter_actor_hide ((ClutterActor *)self->close);
}

static ClutterActor *
cdos_source_create_icon_real (CdosSource *self)
{
    self->icon =  st_icon_new ();
    st_icon_set_icon_size ((StIcon *)self->icon, SOURCE_ICON_SIZE);
    /*
    CdosNotification *notification = self->notifications->data;
    CdosNotificationUrgency urgency = notification->urgency;
    char *icon = NULL;
    switch (urgency) {
        case CDOS_NOTIFICATION_CRITICAL :
            icon = "gtk-dialog-error";
            break;
        default:
            icon = "gtk-dialog-info";
            break;
    }
    */
    st_icon_set_icon_name ((StIcon *)self->icon, "gtk-dialog-info");
    return self->icon;
}

static void
cdos_source_open_real (CdosSource *self)
{
    g_debug ("%s", __func__);
}

static void
cdos_source_class_init (CdosSourceClass *klass)
{
    GObjectClass *gobject_class = G_OBJECT_CLASS(klass);
    gobject_class->set_property = cdos_source_set_property;
    gobject_class->get_property = cdos_source_get_property;
    gobject_class->dispose = cdos_source_dispose;
    gobject_class->finalize = cdos_source_finalize;

    ClutterActorClass *actor_class = CLUTTER_ACTOR_CLASS (klass);
    actor_class->allocate = cdos_source_allocate;

    GParamSpec *pspec;
    pspec = g_param_spec_string ("app-name",
                                 "AppName",
                                 "App name of the source",
                                 NULL, G_PARAM_READWRITE);
    g_object_class_install_property (gobject_class, PROP_NAME, pspec);
    /* singal */
    /**
     * CdosSource::notification-added:
     * @object: 绑定信号的对象
     * @user_data: 真对回调函数所传入的参数
     *
     * 该信号对#CdosSource 的添加notification进行监听
     */
     cdos_source_signals[NOTIFICATION_ADDED] =
     g_signal_new ("notification-added",
                   G_OBJECT_CLASS_TYPE (klass),
                   G_SIGNAL_RUN_LAST,
                   G_STRUCT_OFFSET (CdosSourceClass, notification_added),
                   NULL, NULL,
                   g_cclosure_marshal_VOID__OBJECT,
                   G_TYPE_NONE, 1,
                   CDOS_TYPE_NOTIFICATION);

    /**
     * CdosSource::cdos-source-notify:
     * @object: 绑定信号的对象
     * @user_data: 真对回调函数所传入的参数
     *
     * 该信号对#CdosSource 的发送通知进行监听
     */
     cdos_source_signals[NOTIF] =
     g_signal_new ("notif",
                   G_OBJECT_CLASS_TYPE (klass),
                   G_SIGNAL_RUN_LAST,
                   0,
                   NULL, NULL,
                   g_cclosure_marshal_VOID__OBJECT,
                   G_TYPE_NONE, 1,
                   CDOS_TYPE_NOTIFICATION);

    /**
     * CdosSource::cdos-source-notify:
     * @object: 绑定信号的对象
     * @user_data: 真对回调函数所传入的参数
     *
     * 该信号对#CdosSource 的发送通知进行监听
     */
     cdos_source_signals[CLOSE_ALL] =
     g_signal_new ("close-all",
                   G_OBJECT_CLASS_TYPE (klass),
                   G_SIGNAL_RUN_LAST,
                   0,
                   NULL, NULL,
                   g_cclosure_marshal_VOID__VOID,
                   G_TYPE_NONE, 0);


    /* override */
     klass->open = cdos_source_open_real;
     klass->create_icon = cdos_source_create_icon_real;
}

static void
cdos_source_init (CdosSource *self)
{
    ClutterActor *actor = CLUTTER_ACTOR (self);
    clutter_actor_set_x_expand (actor, TRUE);

    self->name_label = ST_LABEL (st_label_new (NULL));
    clutter_actor_add_child (actor, CLUTTER_ACTOR (self->name_label));

    self->close = g_object_new (ST_TYPE_WIDGET, "style-class", "close-notification", NULL);
    clutter_actor_add_child (actor, CLUTTER_ACTOR (self->close));
    clutter_actor_hide ((ClutterActor *)self->close);
    clutter_actor_set_reactive ((ClutterActor *)self->close, TRUE);
    g_signal_connect (self->close, "button-release-event", G_CALLBACK (close_source), self);

    clutter_actor_set_reactive (actor, TRUE);
    st_widget_set_track_hover ((StWidget *)self, TRUE);

    g_signal_connect (self, "notify::hover", G_CALLBACK (cdos_source_on_hover), self);

    self->is_transient = FALSE;

    self->notifications = NULL;
}

/*公有函数定义*/

/**
 * cdos_source_set_transient:
 * @self: A #CdosSource
 * @isTransient: 设置self是否是暂时的
 *
 * 设置 #CdosSource 实例是否时暂时的.
 *
 */
void
cdos_source_set_transient (CdosSource *self, gboolean is_transient)
{
    self->is_transient = is_transient;
}

/**
 * cdos_source_create_notification:
 * @self: A #CdosSource
 *
 * Called to create a new icon actor (of size this.ICON_SIZE).
 * Must be overridden by the subclass if you do not pass icons
 * explicitly to the cdos_notification_new () constructor.
 */
void
cdos_source_create_notification (CdosSource *self)
{
    g_warning ("no implementation of create_notification() in this.");
}

static void
cdos_source_open_notification (CdosNotification *notification, CdosSource *self)
{
    cdos_source_open (self);
}
/**
 * cdos_source_push_notification:
 * @self: A #CdosSource
 * @notificatin: 待加入链表的节点
 *
 * 判断notification是否在self->notifications里,若不存在就加入链表;
 * 为notification绑定两个信号: clicked和destroy.
 *
 */
void
cdos_source_push_notification (CdosSource *self, CdosNotification *notification)
{
    if (g_list_index (self->notifications, notification) < 0)
    {
        self->notifications = g_list_prepend (self->notifications, notification);
        g_signal_emit (self, cdos_source_signals[NOTIFICATION_ADDED], 0, notification);
        g_signal_connect (notification, "clicked", G_CALLBACK (cdos_source_open_notification), self);
        g_signal_connect (notification, "destroy", G_CALLBACK (notification_destroyed), self);
    }
}

/**
 * cdos_source_notify:
 * @self: A #CdosSource
 * @notification: 待发送的data
 *
 * 发送notification
 *
 */
void
cdos_source_notify (CdosSource *self, CdosNotification *notification)
{
    cdos_source_push_notification (self, notification);
    g_signal_emit (self, cdos_source_signals[NOTIF], 0, notification);
}

/**
 * cdos_source_open:
 * @obj: A #GObject
 * @user_data: 传送的数据
 *
 * 默认什么都不做,但是子类可以重写
 *
 */
void
cdos_source_open (CdosSource *self)
{
    CDOS_SOURCE_GET_CLASS (self)->open (self);
}

/**
 * cdos_source_new:
 * @name: 设置实例的标题
 *
 * 新建一个 #CdosSource 类的实例。
 *
 * Returns: 返回一个 #CdosSource 的实例。
 */
CdosSource *
cdos_source_new (const gchar *name)
{
    CdosSource *self = g_object_new (CDOS_TYPE_SOURCE, "reactive", TRUE, "name", name, NULL);
    cdos_source_create_icon_real (self);
    clutter_actor_insert_child_at_index ((ClutterActor *)self, self->icon, 0);

    return self;
}

void
cdos_source_set_name (CdosSource *self, const char *name)
{
    if (!name)
        return;
    if (g_strcmp0 (self->name, name) == 0)
        return;

    g_clear_pointer (&self->name, g_free);
    self->name = g_strdup (name);
    st_label_set_text (self->name_label, name);
}

const gchar *
cdos_source_get_name (CdosSource *self)
{
    return self->name;
}

void
cdos_source_set_app_name (CdosSource *self, const char *name)
{
    if (!name)
        return;
    if (g_strcmp0 (self->app_name, name) == 0)
        return;

    g_clear_pointer (&self->app_name, g_free);
    self->app_name = g_strdup (name);
}

const gchar *
cdos_source_get_app_name (CdosSource *self)
{
    return self->app_name;
}


void
cdos_source_set_summary_icon (CdosSource *self, ClutterActor *icon)
{
}

ClutterActor *
cdos_source_create_icon (CdosSource *self)
{
    g_return_val_if_fail (CDOS_IS_SOURCE (self), NULL);
    return CDOS_SOURCE_GET_CLASS (self)->create_icon (self);
}

void
cdos_source_destroy (CdosSource *self, guint reason)
{
    g_return_if_fail (CDOS_IS_SOURCE (self));
    ClutterActor *actor = CLUTTER_ACTOR (self);
    ClutterActor *parent = clutter_actor_get_parent (actor);
    clutter_actor_hide (actor);
    if (!parent)
        clutter_actor_add_child (cdos_ui_plugin_get_ui_group (), actor);
    clutter_actor_destroy ((ClutterActor *)self);
}

void
cdos_source_clear_all (CdosSource *self)
{
    g_return_if_fail (CDOS_IS_SOURCE (self));
    GList *l = self->notifications;
    while (l)
    {
        CdosNotification *notification = l->data;
        l = l->next;
        //clutter_actor_destroy ((ClutterActor *)notification);
        cdos_notification_destroy (notification, CDOS_NOTIFICATION_UNDEFINED_RESERVED);
    }
    g_signal_emit (self, cdos_source_signals[CLOSE_ALL], 0);
}
