/**
 * SECTION: cdos-systray
 * @short_description: 管理 Systray
 * @include: cdos-systray.h
 *
 * CdosSystrayManagerClass 类用来处理Systray相关的方法
 */
#include "cdos-systray.h"
#include <string.h>

/*注册函数*/
G_DEFINE_TYPE (CdosSystrayManager, cdos_systray_manager, G_TYPE_OBJECT);

/*声明*/

typedef struct _RoleNode RoleNode;

struct _RoleNode
{
    const gchar *role;
    const gchar *id_str;
};

enum
{
    SYSTRAY_CHANGED,
    LAST_SIGNAL
};

static guint cdos_systray_manager_signals[LAST_SIGNAL] = {0,};

static void
cdos_systray_manager_finalize (GObject *gobject)
{
    CdosSystrayManager *self = CDOS_SYSTRAY_MANAGER (gobject);

    /*释放父类的GList*/
    GList *roles = self->roles;
    if (roles)
    {
        while (roles != NULL)
        {
            RoleNode *node_data = roles->data;
            if (node_data)
            {
                g_free (node_data);
                roles->data = NULL;
            }
            roles = roles->next;
        }
        g_list_free (self->roles);
    }

    /* 调用父类的 finalize() 方法 */
    G_OBJECT_CLASS (cdos_systray_manager_parent_class)->finalize (gobject);
};

/*_class_init & _init*/
static void
cdos_systray_manager_class_init (CdosSystrayManagerClass *klass)
{
    /*继承*/
    GObjectClass *gobject_class = G_OBJECT_CLASS (klass);
    gobject_class->finalize = cdos_systray_manager_finalize;

    /*创建信号*/
    cdos_systray_manager_signals[SYSTRAY_CHANGED] =
        g_signal_new ("changed",
                      G_TYPE_FROM_CLASS (klass),
                      G_SIGNAL_RUN_LAST,
                      G_STRUCT_OFFSET (CdosSystrayManagerClass, systray_changed),
                      NULL,
                      NULL,
                      g_cclosure_marshal_VOID__VOID,
                      G_TYPE_NONE,
                      0,
                      NULL);
}

static void
cdos_systray_manager_init (CdosSystrayManager *self)
{
    self->roles = NULL;
};

/**
 * cdos_systray_manager_register_role:
 * @self: A #CdosSystrayManager
 * @role: 成员
 * @id: 序号
 *
 * 为 #CdosSystrayManager 注册成员
 *
 */
void
cdos_systray_manager_register_role (CdosSystrayManager *self, const gchar *role, const gchar *id)
{
    g_return_if_fail (CDOS_IS_SYSTRAY_MANAGER (self));

    RoleNode *node_data = g_new0 (RoleNode, 1);
    node_data->role = role;
    node_data->id_str = id;

    self->roles = g_list_append (self->roles, node_data);

    g_signal_emit_by_name (self, "changed");
}

/**
 * cdos_systray_manager_unregister_role:
 * @self: A #CdosSystrayManager
 * @role: 成员
 * @id: 序号
 *
 * 为 #CdosSystrayManager 注销成员
 *
 */
void
cdos_systray_manager_unregister_role (CdosSystrayManager *self, const gchar *role, const gchar *id)
{
    g_return_if_fail (CDOS_IS_SYSTRAY_MANAGER (self));
    GList *roles = self->roles;
    gboolean flag = FALSE;
    if (roles)
    {
        while (roles != NULL)
        {
            RoleNode *tmp_node = roles->data;
            if (tmp_node)
            {
                if (g_utf8_collate (tmp_node->id_str, id) == 0 && g_utf8_collate (tmp_node->role, role) == 0)
                {
                    self->roles = g_list_remove (self->roles, tmp_node);
                    g_free (tmp_node);
                    flag = TRUE;
                    break;
                }
            }
            roles = roles->next;
        }
    }
    if (flag)
        g_signal_emit_by_name (self, "changed");
};

/**
 * cdos_systray_manager_unregister_id:
 * @self: A #CdosSystrayManager
 * @id: 序号
 *
 * 通过(const #gchar *) id 注销 #CdosSystrayManager 的成员
 *
 */
void
cdos_systray_manager_unregister_id (CdosSystrayManager *self, const gchar *id)
{
    g_return_if_fail (CDOS_IS_SYSTRAY_MANAGER (self));
    GList *roles = self->roles;
    gboolean flag = FALSE;
    if (roles)
    {
        while (roles != NULL)
        {
            RoleNode *tmp_node = roles->data;
            if (tmp_node)
            {
                if (g_utf8_collate (tmp_node->id_str, id) == 0)
                {
                    self->roles = g_list_remove (self->roles, tmp_node);
                    g_free (tmp_node);
                    flag = TRUE;
                    break;
                }
            }
            roles = roles->next;
        }
    }
    if (flag)
        g_signal_emit_by_name (self, "changed");
}

/**
 * cdos_systray_manager_get_roles:
 * @self: #CdosSystrayManager
 * @id: 序号,id参数没有用到.
 *
 * 遍历所有成员并放入一个新的链表之中.
 *
 * Returns (GList *): 由所有成员构建的新链表.
 */
GList *
cdos_systray_manager_get_roles (CdosSystrayManager *self, const gchar *id)
{
    GList *iter;
    GList *ret = NULL;
    for (iter = self->roles; iter; iter = iter->next)
    {
        RoleNode *node = iter->data;
        ret = g_list_append (ret, (gchar *)node->role);
    }
    return ret;
}

/**
 * cdos_systray_manager_new:
 *
 * 新建一个 #CdosSystrayManager 类的实例。
 *
 * Returns (CdosSystrayManager): 返回一个 #CdosSystrayManager 的实例。
 */
CdosSystrayManager *
cdos_systray_manager_new (void)
{
    CdosSystrayManager *self = g_object_new (CDOS_TYPE_SYSTRAY_MANAGER, NULL);
    return self;
}
