/**
 * SECTION: cdos-theme-manager
 * @short_description: 管理主题
 * @include: cdos-theme-manager.h
 *
 * CdosThemeManagerClass 类用来处理主题相关操作.
 */
#include <gio/gio.h>
#include <gtk/gtk.h>
#include <string.h>

#include "glib-2.62.h"
#include "cdos-start.h"
#include "cdos-theme-manager.h"

#define SETTINGS_SCHEMA  "org.cdos.theme"
#define SETTINGS_KEY  "name"
#define HOME_THEMES  "/.themes/"
#define CDOS_CSS_FILE  "style.css"
#define CDOS_DIR "/found/"

G_DEFINE_TYPE (CdosThemeManager, cdos_theme_manager, G_TYPE_OBJECT);

static gchar* cdos_theme_manager_create_path (const gchar *originPath, const gchar *theme_name);

/*寻找主题*/
static gchar* cdos_theme_manager_find_theme (const gchar *theme_name);

/*改变主题*/
static void cdos_theme_manager_changed_theme (GSettings *settings,
                                              gchar *key,
                                              CdosThemeManager *tm);

static void
cdos_theme_manager_finalize (GObject *gobject)
{
    CdosThemeManager *tm = CDOS_THEME_MANAGER (gobject);
    g_free (tm->theme_path);

    G_OBJECT_CLASS (cdos_theme_manager_parent_class)->finalize (gobject);
}

static void
cdos_theme_manager_dispose (GObject *gobject)
{
    CdosThemeManager *tm = CDOS_THEME_MANAGER (gobject);

    g_clear_signal_handler (&tm->changed_id, &tm->settings);

    g_clear_object (&tm->settings);

    G_OBJECT_CLASS (cdos_theme_manager_parent_class)->dispose (gobject);
}

static void
cdos_theme_manager_class_init (CdosThemeManagerClass *klass)
{
    GObjectClass *gobject_class = G_OBJECT_CLASS(klass);
    gobject_class->finalize = cdos_theme_manager_finalize;
    gobject_class->dispose = cdos_theme_manager_dispose;
}

static void
cdos_theme_manager_init (CdosThemeManager *tm)
{
    tm->settings = g_settings_new (SETTINGS_SCHEMA);
    tm->changed_id = g_signal_connect (tm->settings, "changed::name", G_CALLBACK(cdos_theme_manager_changed_theme),tm);
    tm->theme_path = NULL;
    cdos_theme_manager_changed_theme (tm->settings, NULL, tm);
    //cdos_load_theme ();
}

static gchar*
cdos_theme_manager_create_path (const gchar *originPath, const gchar *theme_name)
{
    gchar *data = NULL;
    gint len = strlen(originPath) + strlen(HOME_THEMES) +
               strlen(theme_name) + strlen (CDOS_DIR) + strlen(CDOS_CSS_FILE);
    data = g_malloc0 (len + 1);
    snprintf(data, len, "%s%s%s%s%s",
             originPath, HOME_THEMES, theme_name, CDOS_DIR, CDOS_CSS_FILE);

    return data;
}

static gchar*
cdos_theme_manager_find_theme (const gchar *theme_name)
{
    gchar *theme_path = NULL;
    const gchar *homeDir = g_get_home_dir ();

    gchar *user_path = cdos_theme_manager_create_path (homeDir, theme_name);

    if (g_file_test (user_path, G_FILE_TEST_EXISTS))
    {
        theme_path = g_strdup (user_path);
    }
    else
    {
        const gchar *const *sysdirs = g_get_system_data_dirs ();

        for (gint i = 0; sysdirs[i] != NULL; i++)
        {
            gchar *syspath = g_strjoin (NULL, sysdirs[i], "/themes/", theme_name,
                                        "/found/", NULL);

            if (g_file_test (syspath, G_FILE_TEST_EXISTS))
            {
                theme_path = syspath;
                break;
            }
            else
            {
                g_free (syspath);
            }
        }
    }
    g_free (user_path);
    return theme_path;
}

static void
cdos_theme_manager_changed_theme (GSettings *settings,
                                  gchar     *key,
                                  CdosThemeManager *tm)
{
    StTextureCache *cache = st_texture_cache_get_default ();
    GtkIconTheme *icon_theme = st_texture_cache_get_icon_theme (cache);

    if (tm->theme_path)
    {
        int len = strlen(tm->theme_path) - strlen(CDOS_CSS_FILE);
        gchar *tmp_path = g_strndup (tm->theme_path, len);

        gchar **search_path = NULL;
        gint i, n_elements;
        /*后面两个参数是接收第一个参数传入后返回的两个结果*/
        gtk_icon_theme_get_search_path (icon_theme, &search_path, &n_elements);
        gboolean find_icon_dir = FALSE;
        for (i = 0; i < n_elements; i++)
        {
            if (g_strcmp0 (search_path[i], tmp_path) == 0)
            {
                find_icon_dir = TRUE;
                break;
            }
        }
        g_free (tmp_path);
        gint j = 0, k = 0;
        gchar **tmp = g_new (gchar *, n_elements - find_icon_dir);

        while(j < n_elements)
        {
            if(i != j)
            {
                tmp[k] = g_strdup (search_path[j]);
                k++;
            }
            j++;
        }
        gtk_icon_theme_set_search_path (icon_theme, (const gchar **)tmp, n_elements-find_icon_dir);
        g_strfreev (search_path);
        //g_strfreev (tmp);
    }

    gchar *stylesheet = NULL;
    gchar *theme_name = g_settings_get_string (tm->settings, SETTINGS_KEY);

    if (theme_name)
    {
        g_free (tm->theme_path);
        tm->theme_path = NULL;
        tm->theme_path = cdos_theme_manager_find_theme (theme_name);
        g_free (theme_name);

        if(tm->theme_path != NULL)
        {
            stylesheet = tm->theme_path;
        }
    }

    if (stylesheet)
    {
        g_debug ("loading user theme: %s", stylesheet);
    }
    else
    {
        g_debug ("loading default theme");
        cdos_set_theme_style_sheet (NULL);
        cdos_load_theme ();
        return;
    }

    if (g_str_has_suffix (tm->theme_path, CDOS_CSS_FILE))
    {
        int len = strlen(tm->theme_path) - strlen(CDOS_CSS_FILE);
        gchar *tmp = g_strndup (tm->theme_path, len);
        gtk_icon_theme_append_search_path (icon_theme, tmp);
        g_free (tmp);
    }
    else if (g_file_test (tm->theme_path, G_FILE_TEST_IS_DIR))
    {
        gtk_icon_theme_prepend_search_path (icon_theme, tm->theme_path);
    }
    cdos_set_theme_style_sheet (stylesheet);
    cdos_load_theme ();
}
