/**
 * Copyright (c) 2016-2018 CPU and Fundamental Software Research Center, CAS
 *
 * This software is published by the license of CPU-OS License，you can use and
 * distibuted this software under this License. See CPU-OS Liense for more detail.
 *
 * You should have receive a copy of CPU-OS License. If not, please contact us
 * by email <support_os@cpu-os.ac.cn>
 *
 **/
/**
 * SECTION: cdos-url-highlighter
 * @short_description:  用于对通知的URL进行高亮显示
 * @include: ui/cdos-url-highlighter.h
 *
 * CdosUrlHighlighterClass 类用来处理通知高亮和换行
 */
#include <stdlib.h>
#include <string.h>
#include "cdos-url-highlighter.h"
#include "cdos-global.h"
#include "cdos-start.h"

G_DEFINE_TYPE (CdosUrlHighlighter, cdos_url_highlighter, ST_TYPE_LABEL);

static void cdos_url_highlighter_highlight_urls (CdosUrlHighlighter *self);
static gint cdos_url_highlighter_find_url_at_pos (CdosUrlHighlighter *self, ClutterEvent *event);

/**
 * cdos_url_highlighter_set_markup:
 * @self: A #CdosUrlHighlighter
 * @text: 待处理文本
 * @allowMarkup: 是否允许处理
 *
 * 判断是否要对传入文本进行处理
 *
 */
void
cdos_url_highlighter_set_markup (CdosUrlHighlighter *self, const gchar *text)
{
    g_return_if_fail (CDOS_IS_URL_HIGHLIGHTER(self));
    if (!text)
        return;
    gchar *new_text = NULL;
    if (text != NULL)
    {
        new_text = cdos_url_highlighter_fix_markup (text);
    }
    else
    {
        new_text = g_strdup("");
    }
    g_free (self->text);
    self->text = new_text;
    cdos_url_highlighter_highlight_urls (self);
}

static char *
cdos_url_sub_string(const char* src, int startpos, int substrlen)
{
    if(src == NULL || substrlen == 0)
        return NULL;

    int srclen = strlen(src);
    if (srclen <= startpos)
        return NULL;

    char* ret = NULL;
    int mallocsize = 0;
    if (0 < substrlen)
    {
        mallocsize = substrlen + 1;
    }
    else
    {
        mallocsize = (srclen - startpos) + 1;
    }

    ret = (char*)malloc(mallocsize);
    int i = 0;
    while ('\0' != *(src + startpos + i) && (i < mallocsize - 1))
    {
        *(ret + i) = *(src + startpos + i);
        ++i;
    }
    ret[i] = '\0';
    return ret;
}

 /**
  * cdos_url_highlighter_fix_markup:
  * @text: 待处理文本
  * @allowMarkup: 是否允许进行标记
  *
  * 利用正则表达式处理文本内部的html转移字符，并保留加粗，斜体，下划线样式。
  *
  * Returns (gchar): 处理之后的文本
  */

gchar *
cdos_url_highlighter_fix_markup (const gchar *text)
{
    g_return_val_if_fail(text != NULL,NULL);
    if (1/*allowMarkup */)
    {
        gchar *new_str = NULL, *text_str = NULL;
        GRegex *regex_1, *regex_2;

        regex_1 = g_regex_new ("(&quot;|&apos;|&lt;|&gt;)", 0, 0, NULL);
        new_str = g_regex_replace (regex_1, text, -1, 0, "&amp;", 0, NULL);
        g_regex_unref (regex_1);

        //regex_2 = g_regex_new ("<(?!\/?[biu]>)", 0, 0, NULL);
        regex_2 = g_regex_new ("<[/]?([^biu]|[a-zA-Z]{2,}[/]?)>", 0, 0, NULL);
        text_str = g_regex_replace (regex_2, new_str, -1, 0, "&lt;", 0, NULL);
        g_regex_unref (regex_2);

        pango_parse_markup (text_str , -1 , 0 , NULL , NULL , NULL , NULL);

        g_free (new_str);

        return text_str;
    }
    return g_markup_escape_text (text , -1);
}

static void
cdos_url_highlighter_highlight_urls (CdosUrlHighlighter *self)
{
    gchar *markup = g_strdup ("");
    gint pos = 0;
    GList *urls = cdos_util_find_urls (self->text);     //返回一个url的阵列，url类型未知
    self->urls = urls;
    gchar *temp = NULL;
    while (urls != NULL)
    {
        MatchedUrl *url = urls->data;
        gint start_pos = cdos_util_get_url_start_pos (url);
        char *str = cdos_url_sub_string (self->text, pos, start_pos - pos);
        if (!str)
            temp = g_strconcat (markup, "<span foreground=\"", self->link_color, "\"><u>", cdos_util_get_url (url), "</u></span>", NULL);
        else
            temp = g_strconcat (markup, str, "<span foreground=\"", self->link_color, "\"><u>", cdos_util_get_url (url), "</u></span>", NULL);
        //g_debug ("%s %s", str, cdos_util_get_url(url));
        pos = cdos_util_get_url_end_pos (url);
        g_free(markup);
        markup = temp;
        g_free(str);
        urls = urls->next;
    }

    char *substr = cdos_url_sub_string (self->text, pos, -1);
    temp = g_strconcat (markup, substr , NULL );

    g_free(markup);
    g_free(substr);

    markup = temp;

    ClutterActor *clutter_text = st_label_get_clutter_text((StLabel *)self);
    clutter_text_set_markup (CLUTTER_TEXT(clutter_text), markup);

    g_free(markup);
}

static gint
cdos_url_highlighter_find_url_at_pos (CdosUrlHighlighter *self, ClutterEvent *event)
{
    gfloat x, y;
    gfloat x_out, y_out;
    gint i, j = 0;
    gint find_pos = -1;
    ClutterText *actor_clutter_text;


    clutter_event_get_coords (event, &x, &y);
    clutter_actor_transform_stage_point (CLUTTER_ACTOR(self), x , y , &x_out, &y_out);

    actor_clutter_text = CLUTTER_TEXT (st_label_get_clutter_text ((StLabel *)self));

    const gchar *clutter_text = clutter_text_get_text (actor_clutter_text);
    glong clutter_text_len = strlen (clutter_text);

    for (i = 0 ; i < clutter_text_len; i++)
    {
        gfloat px, py, line_height;
        clutter_text_position_to_coords (actor_clutter_text, i, &px, &py, &line_height);
        if (py > y_out || py + line_height < y_out  || x_out < px)
        {
            continue ;
        }
        find_pos = i;
    }

    if (find_pos != -1)
    {
        GList *urls = self->urls;
        while(urls != NULL)
        {
            MatchedUrl *url = urls->data;
            if(url)
            {
                gint url_pos = cdos_util_get_url_start_pos (url);
                glong url_len = strlen (cdos_util_get_url (url));
                if ((find_pos >= url_pos) && (url_pos + url_len > find_pos))
                {
                    return j;
                }
            }
            j++;
            urls = urls->next;
        }
    }

    return -1;
}

/* callback func */
static void
on_style_changed (StWidget *actor)
{
    CdosUrlHighlighter *self = CDOS_URL_HIGHLIGHTER (actor);

    gboolean hasColor;
    ClutterColor color;
    hasColor = st_theme_node_lookup_color (st_widget_get_theme_node (actor),
                                           "link-color",
                                           FALSE,
                                           &color);
    if (hasColor)
    {
        gchar *color_string = clutter_color_to_string (&color);
        gchar *link_color = g_strndup (color_string, 9);

        if (g_strcmp0 (link_color, self->link_color) != 0)
        {
            g_free (self->link_color);
            self->link_color = link_color;
            cdos_url_highlighter_highlight_urls (self);
        }
        else
            g_free (link_color);
        g_free (color_string);
    }
    ST_WIDGET_CLASS (cdos_url_highlighter_parent_class)->style_changed (actor);
}

static gboolean
on_button_press_event (ClutterActor *actor, ClutterButtonEvent *button_event)
{
    CdosUrlHighlighter *self = CDOS_URL_HIGHLIGHTER (actor);
    ClutterEvent *event = (ClutterEvent *)button_event;
    //g_debug ("\n INFO: button-press-event! (%s) \n", __FUNCTION__);
    // Don't try to URL highlight when invisible.
    // The MessageTray doesn't actually hide us, so
    // we need to check for paint opacities as well.

    if (!clutter_actor_get_paint_visibility (actor)
        || clutter_actor_get_paint_opacity (actor) == 0)
        return FALSE;

    // Keep Notification from seeing this and taking
    // a pointer grab, which would block our button-release-event
    // handler, if an URL is clicked
    return (cdos_url_highlighter_find_url_at_pos (self, event) == -1);
}


static gboolean
on_button_release_event (ClutterActor *actor, ClutterButtonEvent *button_event)
{
    CdosUrlHighlighter *self = CDOS_URL_HIGHLIGHTER (actor);
    ClutterEvent *event = (ClutterEvent *)button_event;
    gint urlId = -1;
    if (!clutter_actor_get_paint_visibility (actor)
        || clutter_actor_get_paint_opacity (actor) == 0)
        return FALSE;

    urlId = cdos_url_highlighter_find_url_at_pos (self, event);

    if (urlId != -1)
    {
        GError *error = NULL;
        GAppLaunchContext *context;
        //gchar *url = self->urls[urlId].url;  //不确定
        MatchedUrl *matchedUrl = g_list_nth_data (self->urls , (guint) urlId);
        gchar *url = cdos_util_get_url (matchedUrl);

        GRegex *regex_1;
        gchar *new_str;

        regex_1 = g_regex_new ("(\\s)", 0, 0, NULL);
        new_str = g_regex_replace (regex_1, url, -1, 0, "", 0, NULL);
        g_regex_unref (regex_1);

        gchar *tmp;
        if (index (new_str, ':') == NULL &&
            index (new_str, '@') == NULL)
        {
            tmp = g_strconcat ("http://", new_str, NULL);
        }
        else
            tmp = g_strdup (new_str);

        url = tmp;

        context = cdos_global_create_app_launch_context (cdos_global_get (), 0, -1);

        g_app_info_launch_default_for_uri (url, context, &error);

        g_object_unref (context);

        if (error == NULL)
        {
            g_free (url);
            g_free (new_str);
            return TRUE;
        }
        else
        {
            g_debug ("%s %s", __func__, error->message);
            cdos_notify_info (NULL, _("Error opening file"), error->message);
            gint pid;
            gchar *cmd = g_strconcat ("gvfs-open ", url, NULL);
            const gchar  *new_cmd = (const gchar *)cmd;

            pid = cdos_util_spawn_command_line (new_cmd);
            if (pid <= 0)
            {
                g_debug ("%s:%d: %s", __func__, __LINE__, error->message);
            }
            g_free (url);
            g_error_free (error);
            g_free (cmd);
            return TRUE;
        }
    }
    return FALSE;
}


static gboolean
on_motion_event (ClutterActor *actor, ClutterMotionEvent *motion_event)
{
    CdosUrlHighlighter *self = CDOS_URL_HIGHLIGHTER (actor);
    ClutterEvent *event = (ClutterEvent *)motion_event;
    gint urlId;
    CdosGlobal *global = cdos_global_get();

    if (!clutter_actor_get_paint_visibility (actor)
        || clutter_actor_get_paint_opacity (actor) == 0)
        return FALSE;

    urlId = cdos_url_highlighter_find_url_at_pos (self, event);
    MetaDisplay *display = cdos_global_get_display (global);
    if (urlId != -1 && !self->cursor_changed)
    {
        meta_display_set_cursor (display, META_CURSOR_POINTING_HAND);
        self->cursor_changed = TRUE;
    }
    else if (urlId == -1)
    {
        meta_display_set_cursor (display, META_CURSOR_DEFAULT);
        self->cursor_changed = FALSE;
    }
    return FALSE;
}


static gboolean
on_leave_event (ClutterActor *actor, ClutterCrossingEvent* crossing_event)
{
    CdosUrlHighlighter *self = CDOS_URL_HIGHLIGHTER (actor);
    CdosGlobal *global = cdos_global_get ();

    if (!clutter_actor_get_paint_visibility (actor)
        || clutter_actor_get_paint_opacity (actor) == 0)
    {
        return FALSE;
    }

    if (self->cursor_changed)
    {
        self->cursor_changed = FALSE;
        meta_display_set_cursor (cdos_global_get_display (global), META_CURSOR_DEFAULT);
    }
    return FALSE;
}

static void
cdos_url_highlighter_finalize (GObject *object)
{
    CdosUrlHighlighter *self = CDOS_URL_HIGHLIGHTER (object);
    GList *urls = self->urls;
    while (urls != NULL)
    {
        MatchedUrl *url = urls->data;
        if (url)
        {
            gchar *u = cdos_util_get_url (url);
            g_free(u);
            g_free(url);
        }
        urls = urls->next;
    }
    g_list_free (self->urls);
    g_free (self->link_color);
    G_OBJECT_CLASS (cdos_url_highlighter_parent_class)->finalize (object);
}

static void
cdos_url_highlighter_class_init (CdosUrlHighlighterClass *klass)
{
    GObjectClass *gobject_class = G_OBJECT_CLASS(klass);
    gobject_class->finalize = cdos_url_highlighter_finalize;

    ClutterActorClass *actor_class = CLUTTER_ACTOR_CLASS (klass);
    actor_class->button_release_event = on_button_release_event;
    actor_class->button_press_event = on_button_press_event;
    actor_class->motion_event = on_motion_event;
    actor_class->leave_event = on_leave_event;

    StWidgetClass *widget_class = ST_WIDGET_CLASS (klass);
    widget_class->style_changed = on_style_changed;
}

static void
cdos_url_highlighter_init (CdosUrlHighlighter *self)
{
    st_widget_set_style_class_name (ST_WIDGET (self), "url-highlighter");
    clutter_actor_set_reactive (CLUTTER_ACTOR (self), TRUE);

    self->link_color = g_strdup ("#ccccff");
    self->text = NULL;
    self->urls = NULL;
    self->cursor_changed = FALSE;
}

 /**
  * cdos_url_highlighter_new:
  * @text: 待处理文本
  * @lineWrap: 是否允许断行
  * @allowMarkup: 是否允许标记
  *
  * 新建一个#CdosUrlHighlighter类的实例。
  *
  * Returns (CdosUrlHighlighter): 返回一个#CdosUrlHighlighter类的实例。
  */
StLabel *
cdos_url_highlighter_new (const gchar *text, gboolean line_wrap)
{
    CdosUrlHighlighter *self = g_object_new (CDOS_TYPE_URL_HIGHLIGHTER, NULL);

    if (line_wrap)
    {
        ClutterText *actor_clutter_text = CLUTTER_TEXT(st_label_get_clutter_text ((StLabel *)self));

        clutter_text_set_line_wrap (actor_clutter_text, TRUE);
        clutter_text_set_line_wrap_mode (actor_clutter_text, PANGO_WRAP_CHAR);
        clutter_text_set_ellipsize (actor_clutter_text, PANGO_ELLIPSIZE_END);
    }

    cdos_url_highlighter_set_markup (self, text);

    st_widget_set_style_class_name (ST_WIDGET (self), "url-highlighter");

    return (StLabel *)self;
}

