#include <math.h>

#include "common.h"
#include "cdos-window-tracker.h"
#include "cdos-global.h"
#include "cdos-layout-manager.h"
#include "cdos-base-util.h"
#include "cdos-transition.h"

#include "cdos-window-clone-private.h"
#include "cdos-window-clone.h"

#define ICON_SIZE 32
#define ICON_OVERLAP 5
#define IDLE_HIDE_OVERLAY_TIMEOUT 500
#define OVERLAY_FADE_TIME 300
#define WINDOW_CLONE_EASE_TIME 500
#define WINDOW_CLONE_SCALE 1.02

/**
 * SECTION:cdos-window-clone
 * @Title: CdosWindowClone
 * @Short_description: 窗口克隆体
 *
 * CdosWindowClone 就是当前工作区内，窗口的克隆体。我们通过组合键进入 overview 缩略图模式，
 * 这个对象就是每个窗口的克隆体，由于是克隆体，退出 overview 模式时，需要将这些克隆体销毁，
 * 否则桌面上会多出窗口的复制体，不可使用。
 */

G_DEFINE_TYPE (CdosWindowClone, cdos_window_clone, ST_TYPE_WIDGET);

enum
{
    WORKSPACE_CHANGED,
    SELECTED,
    SHOW_OVERLAY,
    ACTIVATED,
    CLOSED,
    DRAG_BEGIN,
    DRAG_CANCELLED,
    DRAG_END,
    LAST_SIGNAL
};

static guint cdos_window_clone_signals[LAST_SIGNAL] = {0};

static void
on_meta_window_size_changed (MetaWindow *window, CdosWindowClone *self)
{
    gfloat border_x, border_y;
    MetaRectangle frame_rect, buffer_rect;

    meta_window_get_frame_rect (self->meta_window, &frame_rect);
    meta_window_get_buffer_rect (self->meta_window, &buffer_rect);

    border_x = frame_rect.x - buffer_rect.x;
    border_y = frame_rect.y - buffer_rect.y;

    self->orig_x += border_x;
    self->orig_y += border_y;
    self->orig_width = frame_rect.width;
    self->orig_height = frame_rect.height;
}

static void
on_workspace_changed (MetaWindow *w, gpointer user_data)
{
    CdosWindowClone *self = CDOS_WINDOW_CLONE (user_data);
    g_signal_emit (self, cdos_window_clone_signals[WORKSPACE_CHANGED], 0);
}

static void
disconnect_window_signals (MetaWindowActor *window, CdosWindowClone *self)
{
    if (self->workspace_changed_id)
    {
        g_signal_handler_disconnect (self->meta_window, self->workspace_changed_id);
        self->workspace_changed_id = 0L;
    }

    if (self->size_changed_id)
    {
        g_signal_handler_disconnect (self->meta_window, self->size_changed_id);
        self->size_changed_id = 0L;
    }
    if (self->window_actor_destroyed_id)
    {
        g_signal_handler_disconnect (self->window_actor, self->window_actor_destroyed_id);
        self->window_actor_destroyed_id = 0L;
    }
}

static void
on_real_window_destroyed (ClutterActor *actor, CdosWindowClone *self)
{
    clutter_actor_destroy ((ClutterActor *)self);
}

static gboolean
on_button_release_event (ClutterActor *actor, ClutterButtonEvent *button_event)
{
    CdosWindowClone *self = CDOS_WINDOW_CLONE (actor);
    ClutterEvent *event = (ClutterEvent *)button_event;
    guint32 event_type = clutter_event_get_button (event);

    if (event_type != 2)
    {
        self->selected = TRUE;
        g_signal_emit (self, cdos_window_clone_signals[ACTIVATED], 0);
        return TRUE;
    }
    else if (event_type == 2)
    {
        g_signal_emit (self, cdos_window_clone_signals[CLOSED], 0);

        return TRUE;
    }
    return FALSE;
}

static gboolean
on_button_press_event (ClutterActor *actor, ClutterButtonEvent *event)
{
    g_signal_emit (actor, cdos_window_clone_signals[SELECTED], 0);
    return TRUE;
}

static void
cdos_window_clone_show_overlay (CdosWindowClone *self, gboolean animate)
{
    if (self->overlay_shown)
        return;
    self->overlay_shown = TRUE;

    cdos_window_clone_set_stack_above (self, NULL);

    ClutterActor *button = CLUTTER_ACTOR (self->close_button);
    clutter_actor_set_opacity (button, 0);
    clutter_actor_show (button);
    cdos_transition_simple (button,
                            CLUTTER_EASE_OUT_EXPO,
                            OVERLAY_FADE_TIME,
                            "opacity", 255,
                            NULL);
    cdos_transition_simple (self->clone,
                            CLUTTER_EASE_OUT_EXPO,
                            WINDOW_CLONE_EASE_TIME,
                            "scale-x", WINDOW_CLONE_SCALE,
                            "scale-y", WINDOW_CLONE_SCALE,
                            NULL);
    g_signal_emit (self, cdos_window_clone_signals[SHOW_OVERLAY], 0);
}

void
cdos_window_clone_hide_overlay (CdosWindowClone *self, gboolean animate)
{
    ClutterTransition *trans;

    g_return_if_fail (CDOS_IS_WINDOW_CLONE (self));

    if (!self->overlay_shown)
        return;
    self->overlay_shown = FALSE;

    ClutterActor *button = CLUTTER_ACTOR (self->close_button);
    trans = cdos_transition_simple (button,
                                    CLUTTER_EASE_OUT_EXPO,
                                    OVERLAY_FADE_TIME,
                                    "opacity", 0,
                                    NULL);
    g_signal_connect_swapped (trans, "completed",
                              G_CALLBACK (clutter_actor_hide), button);

    cdos_transition_simple (self->clone,
                            CLUTTER_EASE_OUT_EXPO,
                            WINDOW_CLONE_EASE_TIME,
                            "scale-x", 1.0,
                            "scale-y", 1.0,
                            NULL);
}

static gboolean
on_enter_event (ClutterActor *actor, ClutterCrossingEvent *event)
{
    CdosWindowClone *self = CDOS_WINDOW_CLONE (actor);
    if (self->idle_hide_overlay_id)
    {
        g_source_remove (self->idle_hide_overlay_id);
        self->idle_hide_overlay_id = 0;
    }
    cdos_window_clone_show_overlay (self, TRUE);
    return CLUTTER_ACTOR_CLASS (cdos_window_clone_parent_class)->enter_event (actor, event);
}

static gboolean
idle_hide_overlay (CdosWindowClone *self)
{
    if (clutter_actor_has_pointer ((ClutterActor *)self->close_button) ||
        clutter_actor_has_pointer ((ClutterActor *)self->icon_title_box))
        return G_SOURCE_CONTINUE;

    if (!clutter_actor_has_pointer ((ClutterActor *)self))
        cdos_window_clone_hide_overlay (self, TRUE);
    self->idle_hide_overlay_id = 0;
    return G_SOURCE_REMOVE;
}

static gboolean
on_leave_event (ClutterActor *actor, ClutterCrossingEvent *event)
{
    CdosWindowClone *self = CDOS_WINDOW_CLONE (actor);
    if (self->idle_hide_overlay_id)
        g_source_remove (self->idle_hide_overlay_id);
    self->idle_hide_overlay_id = g_timeout_add (IDLE_HIDE_OVERLAY_TIMEOUT,
                                                (GSourceFunc)idle_hide_overlay,
                                                self);
    return CLUTTER_ACTOR_CLASS (cdos_window_clone_parent_class)->leave_event (actor, event);
}

static void
cdos_window_clone_dispose (GObject *gobject)
{
    CdosWindowClone *self = CDOS_WINDOW_CLONE (gobject);
    disconnect_window_signals (NULL, self);
    if (self->idle_hide_overlay_id)
    {
        g_source_remove (self->idle_hide_overlay_id);
        self->idle_hide_overlay_id = 0;
    }
    G_OBJECT_CLASS (cdos_window_clone_parent_class)->dispose (gobject);
}

static void
cdos_window_clone_get_preferred_height (ClutterActor *actor,
                                        gfloat for_width,
                                        gfloat *min_height,
                                        gfloat *nat_height)
{
    CdosWindowClone *clone = CDOS_WINDOW_CLONE (actor);
    MetaWindow *window = clone->meta_window;
    MetaRectangle rect;
    meta_window_get_frame_rect (window, &rect);
    if (min_height)
        *min_height = rect.height;
    if (nat_height)
        *nat_height = rect.height;
}

static void
cdos_window_clone_get_preferred_width (ClutterActor *actor,
                                       gfloat for_height,
                                       gfloat *min_width,
                                       gfloat *nat_width)
{
    CdosWindowClone *clone = CDOS_WINDOW_CLONE (actor);
    MetaWindow *window = clone->meta_window;
    MetaRectangle rect;
    meta_window_get_frame_rect (window, &rect);
    if (min_width)
        *min_width = rect.width;
    if (nat_width)
        *nat_width = rect.width;
}

static void
cdos_window_clone_allocate (ClutterActor *actor,
                            const ClutterActorBox *box,
                            ClutterAllocationFlags flags)
{
    MetaRectangle frame_rect, buffer_rect;
    CdosWindowClone *self;
    MetaWindow *window;
    ClutterActorBox box1;
    gfloat scale_x, scale_y;
    gfloat left_over_size, right_over_size;
    gfloat top_over_size, bottom_over_size;
    gfloat border_x, border_y;

    clutter_actor_set_allocation (actor, box, flags);

    self = CDOS_WINDOW_CLONE (actor);
    window = self->meta_window;
    meta_window_get_frame_rect (window, &frame_rect);
    meta_window_get_buffer_rect (window, &buffer_rect);
    box1.x1 = border_x = buffer_rect.x - frame_rect.x;
    box1.y1 = border_y = buffer_rect.y - frame_rect.y;
    box1.x2 = box1.x1 + buffer_rect.width;
    box1.y2 = box1.y1 + buffer_rect.height;
    /*
    gfloat clone_width, clone_height;
    clutter_actor_get_preferred_size (self->clone, NULL, NULL, &clone_width, &clone_height);
    */

    cdos_window_clone_chrome_widths (self,
                                     &left_over_size,
                                     &right_over_size);
    cdos_window_clone_chrome_heights (self,
                                      &top_over_size,
                                      &bottom_over_size);

    scale_x = (box->x2 - box->x1 - left_over_size - right_over_size) / frame_rect.width;
    scale_y = (box->y2 - box->y1 - top_over_size - bottom_over_size) / frame_rect.height;
    box1.x1 *= scale_x;
    box1.x2 *= scale_x;
    box1.y1 *= scale_y;
    box1.y2 *= scale_y;

    clutter_actor_allocate (self->clone, &box1, flags);

    /*
    clutter_actor_allocate_available_size (self->icon,
                                           0, 0,
                                           box->x2 - box->x1,
                                           box->y2 - box->y1, flags);
    clutter_actor_allocate_available_size ((ClutterActor *)self->close_button,
                                           0, 0,
                                           box->x2 - box->x1,
                                           box->y2 - box->y1, flags);
    */
    clutter_actor_allocate_available_size ((ClutterActor *)self->icon_title_box,
                                           0, 0,
                                           box->x2 - box->x1,
                                           box->y2 - box->y1, flags);
    if (self->close_button_side == ST_SIDE_LEFT)
    {
        box1.x1 = box1.x1 - left_over_size;
        box1.x2 = box1.x1 + left_over_size * 2;
    }
    else
    {
        box1.x1 = box1.x2 - right_over_size + border_x * scale_x;
        box1.x2 = box1.x1 + right_over_size * 2;
    }
    box1.y1 = box1.y1 - top_over_size - border_y * scale_y;
    box1.y2 = box1.y1 + top_over_size * 2;
    clutter_actor_allocate ((ClutterActor *)self->close_button, &box1, flags);
}

static const gchar *
get_window_caption (CdosWindowClone *self, CdosApp *app)
{
    const gchar *title;
    title = meta_window_get_title (self->meta_window);
    if (title)
        return title;
    else
    {
        return cdos_app_get_name (app);
    }
}

static void
on_window_close_button_clicked (StButton *button,
                                int click_button,
                                CdosWindowClone *self)
{
    meta_window_delete (self->meta_window,
                        cdos_global_get_current_time (cdos_global_get ()));
}

static void
adjust_overlay_offsets (ClutterActor *actor,
                        GParamSpec *param,
                        CdosWindowClone *self)
{
    gdouble scale_x = 0, scale_y = 0;
    ClutterActorBox box;
    gfloat width = 0, height = 0;
    gint width_increase, height_increase;
    clutter_actor_get_scale (self->clone, &scale_x, &scale_y);
    clutter_actor_get_allocation_box (self->clone, &box);
    width = box.x2 - box.x1;
    height = box.y2 - box.y1;
    width_increase = width * (scale_x - 1.0) / 2.0;
    height_increase = height * (scale_y - 1.0) / 2.0;

    /*
    clutter_actor_set_translation (self->icon, 0, height_increase, 0);
    */
    clutter_actor_set_translation ((ClutterActor *)self->icon_title_box,
                                   0, ICON_SIZE + ICON_OVERLAP + height_increase, 0);
    if (self->close_button_side == ST_SIDE_LEFT)
        clutter_actor_set_translation ((ClutterActor *)self->close_button,
                                       -width_increase, -height_increase, 0);
    else
        clutter_actor_set_translation ((ClutterActor *)self->close_button,
                                       width_increase, -height_increase, 0);
}

static void
cdos_window_clone_class_init (CdosWindowCloneClass *klass)
{
    GObjectClass *gobject_class = G_OBJECT_CLASS (klass);
    gobject_class->dispose = cdos_window_clone_dispose;

    ClutterActorClass *actor_class = CLUTTER_ACTOR_CLASS (klass);
    actor_class->get_preferred_width = cdos_window_clone_get_preferred_width;
    actor_class->get_preferred_height = cdos_window_clone_get_preferred_height;
    actor_class->allocate = cdos_window_clone_allocate;
    actor_class->button_press_event = on_button_press_event;
    actor_class->button_release_event = on_button_release_event;
    actor_class->leave_event = on_leave_event;
    actor_class->enter_event = on_enter_event;

    cdos_window_clone_signals[WORKSPACE_CHANGED] = g_signal_new ("workspace-changed",
                                                                 G_TYPE_FROM_CLASS (klass),
                                                                 G_SIGNAL_RUN_LAST,
                                                                 0,
                                                                 NULL,
                                                                 NULL,
                                                                 NULL,
                                                                 G_TYPE_NONE,
                                                                 0);

    cdos_window_clone_signals[SELECTED] = g_signal_new ("selected",
                                                        G_TYPE_FROM_CLASS (klass),
                                                        G_SIGNAL_RUN_LAST,
                                                        0,
                                                        NULL,
                                                        NULL,
                                                        NULL,
                                                        G_TYPE_NONE,
                                                        0);

    cdos_window_clone_signals[SHOW_OVERLAY] =
        g_signal_new ("show-overlay",
                      G_TYPE_FROM_CLASS (klass),
                      G_SIGNAL_RUN_LAST,
                      0,
                      NULL,
                      NULL,
                      NULL,
                      G_TYPE_NONE,
                      0);

    cdos_window_clone_signals[ACTIVATED] =
        g_signal_new ("activated",
                      G_TYPE_FROM_CLASS (klass),
                      G_SIGNAL_RUN_LAST,
                      0,
                      NULL,
                      NULL,
                      NULL,
                      G_TYPE_NONE,
                      0);

    cdos_window_clone_signals[CLOSED] =
        g_signal_new ("closed",
                      G_TYPE_FROM_CLASS (klass),
                      G_SIGNAL_RUN_LAST,
                      0,
                      NULL,
                      NULL,
                      NULL,
                      G_TYPE_NONE,
                      0);
}

static void
cdos_window_clone_init (CdosWindowClone *self)
{
    self->stack_above = NULL;
    self->selected = FALSE;
}

static void
cdos_window_clone_post_init (CdosWindowClone *self, MetaWindowActor *window_actor)
{
    CdosWindowTracker *tracker;
    CdosApp *app;
    ClutterActor *icon;
    StWidget *title;
    ClutterConstraint *constraint;
    MetaButtonLayout button_layout;
    StSide close_button_side;
    StWidget *close_button;
    ClutterActor *close_icon;

    if (window_actor == NULL)
        return;

    self->window_actor = window_actor;
    self->meta_window = meta_window_actor_get_meta_window (self->window_actor);

    clutter_actor_get_position ((ClutterActor *)window_actor, &self->orig_x, &self->orig_y);

    //MetaShapedTexture *texture = meta_window_actor_get_texture (window_actor);
    self->clone = clutter_clone_new (CLUTTER_ACTOR (window_actor));
    cdos_base_util_set_hidden_from_pick (self->clone, TRUE);
    clutter_actor_add_child ((ClutterActor *)self, self->clone);
    clutter_actor_set_pivot_point (self->clone, 0.5, 0.5);

    self->icon_title_box = g_object_new (ST_TYPE_BOX_LAYOUT,
                                         NULL);
    clutter_actor_set_pivot_point ((ClutterActor *)self->icon_title_box, 0.5, 0);
    constraint = clutter_align_constraint_new (self->clone, CLUTTER_ALIGN_X_AXIS, 0.5);
    clutter_actor_add_constraint ((ClutterActor *)self->icon_title_box, constraint);
    constraint = clutter_align_constraint_new (self->clone, CLUTTER_ALIGN_Y_AXIS, 1);
    clutter_actor_add_constraint ((ClutterActor *)self->icon_title_box, constraint);
    clutter_actor_set_translation ((ClutterActor *)self->icon_title_box,
                                   0, ICON_SIZE + ICON_OVERLAP, 0);

    tracker = cdos_window_tracker_get_default ();
    app = cdos_window_tracker_get_window_app (tracker, self->meta_window);
    icon = cdos_app_create_icon_texture (app, ICON_SIZE);
    /*
    clutter_actor_set_pivot_point (icon, 0.5, 0.5);
    constraint = clutter_align_constraint_new (self->clone, CLUTTER_ALIGN_X_AXIS, 0.5);
    clutter_actor_add_constraint (icon, constraint);
    constraint = clutter_align_constraint_new (self->clone, CLUTTER_ALIGN_Y_AXIS, 1);
    clutter_actor_add_constraint (icon, constraint);
    */

    self->icon = icon;

    title = st_label_new (get_window_caption (self, app));
    self->title = title;
    st_widget_add_style_class_name (title, "window-caption");

    meta_prefs_get_button_layout (&button_layout);
    if (button_layout.right_buttons[META_BUTTON_FUNCTION_CLOSE])
        close_button_side = ST_SIDE_RIGHT;
    else
        close_button_side = ST_SIDE_LEFT;
    self->close_button_side = close_button_side;

    close_button = g_object_new (ST_TYPE_BUTTON,
                                 "style-class", "window-close",
                                 "visible", FALSE,
                                 NULL);
    self->close_button = close_button;
    clutter_actor_set_pivot_point ((ClutterActor *)close_button, 0.5, 0.5);
    close_icon = g_object_new (ST_TYPE_ICON,
                               "icon-name", "close-window",
                               NULL);
    st_bin_set_child ((StBin *)close_button, close_icon);
    /*
     * 低版本 align constraint 的 pivot point 不生效，
     * 手动在allocate 函数中申请空间
     */
    /*
    constraint = clutter_align_constraint_new (self->clone,
                                               CLUTTER_ALIGN_X_AXIS,
                                               close_button_side == ST_SIDE_LEFT ? 0 : 1);
    clutter_actor_add_constraint ((ClutterActor *)close_button, constraint);
    constraint = clutter_align_constraint_new (self->clone, CLUTTER_ALIGN_Y_AXIS, 0);
    clutter_actor_add_constraint ((ClutterActor *)close_button, constraint);
    */
    g_signal_connect (close_button, "clicked",
                      G_CALLBACK (on_window_close_button_clicked), self);

    _add_actor (self, self->icon_title_box);
    _add_actor (self->icon_title_box, icon);
    _add_actor (self->icon_title_box, title);
    _add_actor (self, close_button);

    self->size_changed_id = g_signal_connect (self->meta_window, "size-changed",
                                              G_CALLBACK (on_meta_window_size_changed), self);
    on_meta_window_size_changed (self->meta_window, self);

    self->workspace_changed_id = g_signal_connect (self->meta_window,
                                                   "workspace-changed",
                                                   G_CALLBACK (on_workspace_changed), self);

    self->window_actor_destroyed_id = g_signal_connect (self->window_actor, "destroy",
                                                        G_CALLBACK (on_real_window_destroyed), self);
    g_signal_connect (self->clone, "notify::scale-x",
                      G_CALLBACK (adjust_overlay_offsets), self);
}

/**
 * cdos_window_clone_new:
 * @real_window: #MetaWindowActor 对象，实际打开的窗口
 * @my_container: #CdosWorkspaceMonitor 对象
 *
 * 新建窗口克隆体
 *
 * Return value: #CdosWindowClone 对象
 */
CdosWindowClone *
cdos_window_clone_new (MetaWindowActor *window)
{
    CdosWindowClone *instance = g_object_new (CDOS_TYPE_WINDOW_CLONE, "reactive", TRUE, NULL);

    cdos_window_clone_post_init (instance, window);

    return instance;
}

/**
 * cdos_window_clone_set_stack_above:
 * @self: #CdosWindowClone 对象
 * @actor: 需要置于顶层的对象，可空
 *
 * 将 actor 置于最顶层。 #CdosWindowClone 对象的 actor 成员用于处理所有的交互事件，
 * 例如鼠标点击等，之所以用 actor 而不用 clone［存放克隆窗口的组］，是因为 actor
 * 才能对事件做出响应。
 */
void
cdos_window_clone_set_stack_above (CdosWindowClone *self, ClutterActor *actor)
{
    self->stack_above = actor;

    if (self->stack_above == NULL)
    {
        ClutterActor *parent = clutter_actor_get_parent ((ClutterActor *)self);
        clutter_actor_set_child_above_sibling (parent, (ClutterActor *)self, NULL);
    }
    else
    {
        ClutterActor *parent = clutter_actor_get_parent (self->stack_above);
        clutter_actor_set_child_above_sibling (parent, (ClutterActor *)self, self->stack_above);
    }
}

void
cdos_window_clone_chrome_heights (CdosWindowClone *self,
                                  gfloat *top_over_size,
                                  gfloat *bottom_over_size)
{
    if (top_over_size)
    {
        gfloat height = 0;
        clutter_actor_get_preferred_height ((ClutterActor *)self->close_button,
                                            -1,
                                            NULL,
                                            &height);
        *top_over_size = height / 2;
    }
    if (bottom_over_size)
    {
        gfloat height = 0;
        clutter_actor_get_preferred_height ((ClutterActor *)self->icon_title_box,
                                            -1,
                                            NULL,
                                            &height);
        *bottom_over_size = height / 2;
    }
}

void
cdos_window_clone_chrome_widths (CdosWindowClone *self,
                                 gfloat *left_over_size,
                                 gfloat *right_over_size)
{
    gfloat width = 0;
    clutter_actor_get_preferred_width ((ClutterActor *)self->close_button,
                                       -1,
                                       NULL,
                                       &width);
    if (self->close_button_side == ST_SIDE_LEFT)
    {
        if (left_over_size)
            *left_over_size = width / 2;
        if (right_over_size)
            *right_over_size = 0.0;
    }
    else
    {
        if (left_over_size)
            *left_over_size = 0.0;
        if (right_over_size)
            *right_over_size = width / 2;
    }
}
