#include <stdarg.h>
#include <math.h>

#include "cdos-debug.h"
#include <meta/display.h>
#include <meta/keybindings.h>
#include <meta/common.h>
#include <meta/boxes.h>
#include <meta/meta-monitor-manager.h>

#include "cdos-coverflow-switcher.h"
#include "cdos-layout-manager.h"
#include "cdos-monitor-switcher.h"
#include "cdos-overview.h"
#include "cdos-run-dialog.h"
#include "cdos-screenshot.h"
#include "cdos-screen-saver.h"
#include "cdos-sound-manager.h"
#include "cdos-tile-preview.h"
#include "cdos-transition.h"
#include "cdos-window-manager.h"
#include "cdos-keybindings.h"
#include "cdos-message-dialog.h"
#include "cdos-monitor-constraint.h"

/**
 * SECTION: cdos-window-manager
 * @Title: CdosWM
 * @short_description: 管理窗口及工作区切换映射等
 *
 * 创建一个 CdosWM, 管理窗口切换、映射、关闭和切换工作区等的动画及绑定快捷键。
 */

struct _CdosWM
{
    CdosBaseWM *cdos_wm;

    /* <private> */
    MetaWindow *moving_window;
    int maximizing;
    int minimizing;
    int unmaximizing;
    int tiling;
    int mapping;
    int destroying;
    gboolean should_animate;

    GSettings *settings;
    ClutterActor *tile_preview;
    CdosKeybindings *keybindings;

    CdosMessageDialog *confirm;
    guint confirm_timeout_id;
    guint remaining_time;
    guint monitor_watch_id;
};


#define WINDOW_ANIMATION_TIME 250
#define TILE_HUD_ANIMATION_TIME 150
#define DIM_TIME 500
#define DIM_DESATURATION 0.600
#define DIM_BRIGHTNESS -0.2
#define UNDIM_TIME 250
#define MAP_TIME 250
#define MAXIMIZE_TIME 250
#define MINIMIZE_TIME 150
#define DESTROY_TIME 150
#define SWITCH_TIME 300
#define MONITOR_CHANGE_TIMEOUT 10000

// add by liting
static ClutterActor *snap_osd = NULL; // StBoxLayout
static ClutterActor *workspace_osd = NULL; // StLabel
// end
static CdosBaseWM *base_wm = NULL;

static gboolean desktop_effects = FALSE;

static void
_should_animate (GSettings *settings, gchar *key)
{
    if (g_strcmp0 (key, "desktop-effects") == 0)
    {
        desktop_effects = g_settings_get_boolean (settings, "desktop-effects");
    }
}

static void
should_animate (void)
{
    GSettings *settings = cdos_global_get_settings (global);
    desktop_effects = g_settings_get_boolean (settings, "desktop-effects");
    g_signal_connect (settings, "changed", G_CALLBACK (_should_animate), NULL);
}

static gboolean
enable_desktop_effects (void)
{
    if (cdos_get_modal_count () || cdos_is_software_rendering ())
    {
        return FALSE;
    }
    return desktop_effects;
}

/* 动画 */
typedef struct _DefaultTransData
{
    ClutterActor *actor;
    const gchar *name;
} DefaultTransData;

static void
default_transition_complete_callback (ClutterTimeline *tml, DefaultTransData *data)
{
    if (data->name)
        clutter_actor_remove_transition (data->actor, data->name);
    g_free (data);
}

static void
default_transition_stop_callback (ClutterTimeline *tml, gboolean is_finished, DefaultTransData *data)
{
    if (!is_finished)
    {
        if (data->name)
            clutter_actor_remove_transition (data->actor, data->name);
        g_free (data);
    }
}

static void
cdos_set_actor_on_monitor_center (ClutterActor *actor)
{
    int current_monitor;
    ClutterActor *ui_group;
    MetaDisplay *display;
    ClutterConstraint *constraint;

    display = cdos_global_get_display (cdos_global_get ());
    ui_group = cdos_ui_plugin_get_ui_group ();
    clutter_actor_set_child_above_sibling (ui_group, actor, NULL);

    current_monitor = meta_display_get_current_monitor (display);
    constraint = cdos_monitor_constraint_new (current_monitor, TRUE);
    clutter_actor_add_constraint (actor, constraint);
}

/**
 * 不需要回调函数的动画
 */
ClutterTransition *
cdos_transition_simple (ClutterActor *actor,
                        ClutterAnimationMode mode,
                        guint duration,
                        const char *first_property_name,
                        ...)
{
    g_return_val_if_fail (CLUTTER_IS_ACTOR (actor), NULL);
    g_return_val_if_fail (first_property_name != NULL, NULL);
    va_list args;

    clutter_actor_save_easing_state (actor);
    clutter_actor_set_easing_mode (actor, mode);
    clutter_actor_set_easing_duration (actor, duration);

    va_start (args, first_property_name);
    g_object_set_valist ((GObject *)actor, first_property_name, args);
    va_end (args);
    ClutterTransition *transition = clutter_actor_get_transition (actor, first_property_name);
    clutter_actor_restore_easing_state (actor);
    return transition;
}

/**
 * cdos_transition_to:
 * @actor: A #ClutterActor
 * @mode: 动画模式, #ClutterAnimationMode
 * @duration: 一次动画持续时间
 * @repeat_count: 动画重复次数。-1 无限重复。0 不重复。
 * @callback: 动画结束回调函数, 可以为 %NULL。
 * @data: @callback 参数
 * @name: 动画名称
 * @first_property_name: @actor 的属性，不能为 %NULL
 *
 * 例如：
 *     cdos_transition_to (actor,
 *                         CLUTTER_EASE_OUT_QUAD,
 *                         WINDOW_ANIMATION_TIME,
 *                         0,
 *                         NULL, NULL,
 *                         "switch-workspace",
 *                         "x", TRANS_PROP_FLOAT, 0.f,
 *                         "y", TRANS_PROP_FLOAT, 0.f,
 *                         NULL);
 *
 * Returns: #ClutterTransition 对象
 */
ClutterTransition *
cdos_transition_to (ClutterActor *actor,
                    ClutterAnimationMode mode,
                    guint duration,
                    guint repeat_count,
                    TransCallback callback,
                    void *data,
                    const char *name,
                    const char *first_property_name,
                    ...)
{
    /* 可变参数格式:
     * prop_name, (x, y, signal::new-frame, ...)
     * prop_type, (TRANS_PROP_FLOAT, ...)
     * to value
     */
    g_return_val_if_fail (CLUTTER_IS_ACTOR (actor), NULL);
    g_return_val_if_fail (first_property_name != NULL, NULL);

    va_list args;

    clutter_actor_save_easing_state (actor);

    ClutterTransition *trans = clutter_transition_group_new ();
    const char *prop_name = first_property_name;
    va_start (args, first_property_name);
    while (1)
    {
        if (g_str_has_prefix (prop_name, "signal::"))
        {
            GCallback callback = va_arg (args, GCallback);
            gpointer user_data = va_arg (args, gpointer);
            g_signal_connect (trans, prop_name + 8, callback, user_data);
        }
        else
        {
            gint value_type = va_arg (args, gint);
            if (value_type < TRANS_PROP_CHAR || value_type > TRANS_PROP_DOUBLE)
            {
                break;
            }
            ClutterTransition *trans_p = NULL;
            trans_p = clutter_property_transition_new (prop_name);
            switch (value_type)
            {
                case TRANS_PROP_CHAR:
                {
                    gchar p = va_arg (args, gint);
                    clutter_transition_set_to (trans_p, G_TYPE_CHAR, p);
                    break;
                }
                case TRANS_PROP_UCHAR:
                {
                    guchar p = va_arg (args, guint);
                    clutter_transition_set_to (trans_p, G_TYPE_UCHAR, p);
                    break;
                }
                case TRANS_PROP_INT:
                {
                    gint p = va_arg (args, gint);
                    clutter_transition_set_to (trans_p, G_TYPE_INT, p);
                    break;
                }
                case TRANS_PROP_UINT:
                {
                    guint p = va_arg (args, guint);
                    clutter_transition_set_to (trans_p, G_TYPE_UINT, p);
                    break;
                }
                case TRANS_PROP_FLOAT:
                {
                    gfloat p = va_arg (args, double);
                    clutter_transition_set_to (trans_p, G_TYPE_FLOAT, p);
                    break;
                }
                case TRANS_PROP_DOUBLE:
                {
                    gdouble p = va_arg (args, double);
                    clutter_transition_set_to (trans_p, G_TYPE_DOUBLE, p);
                    break;
                }
            }
            clutter_transition_group_add_transition ((ClutterTransitionGroup *)trans, trans_p);
        }
        prop_name = va_arg (args, char *);
        if (prop_name == NULL)
            break;
    }

    va_end (args);

    clutter_actor_set_easing_mode (actor, mode);
    clutter_timeline_set_duration (CLUTTER_TIMELINE (trans), duration);

    clutter_timeline_set_repeat_count (CLUTTER_TIMELINE (trans), repeat_count);

    clutter_actor_add_transition (actor, name, trans);
    if (callback)
    {
        g_signal_connect (trans, "completed", G_CALLBACK (callback), data);
    }
    else
    {
        DefaultTransData *data = g_new0 (DefaultTransData, 1);
        data->actor = actor;
        data->name = name;
        g_signal_connect (trans, "completed", G_CALLBACK (default_transition_complete_callback), data);
        g_signal_connect (trans, "stopped", G_CALLBACK (default_transition_stop_callback), data);
    }
    clutter_actor_restore_easing_state (actor);
    return trans;
}

typedef struct _WindowData
{
    CdosBaseWM *wm;
    ClutterActor *actor;
    gdouble orig_scale_x;
    gdouble orig_scale_y;
    gfloat  orig_x;
    gfloat  orig_y;
    guint8  orig_opacity;
}WindowData;

typedef struct _EffectData
{
#define MINIMIZING 1
#define MAPPING 2
#define DESTROYING 4
    guint state;
    guint elapsed_time;
    gdouble scale_x;
    gdouble scale_y;
    gfloat x;
    gfloat y;
    ClutterTimeline *tml;
} EffectData;

static EffectData *
get_effect_data (MetaWindowActor *actor)
{
#define EFFECT_DATA "effect-data"
    EffectData *data = NULL;
    data = g_object_get_data ((GObject *)actor, EFFECT_DATA);

    if (!data)
    {
        data = g_new0 (EffectData, 1);
        g_object_set_data_full ((GObject *)actor, EFFECT_DATA, data, g_free);
    }
    return data;
}


/* Switch Workspace */

struct TransData
{
    MetaWindowActor *actor;
    gboolean visible;
    graphene_point_t orig;
};

/* 是否正在切换工作区 */

static GMutex mutex;
static gint animation_count = 0;

static void
kill_window_effects (CdosBaseWM *cdos_wm, MetaWindowActor *actor)
{
    EffectData *effect = get_effect_data (actor);
    if (effect->tml)
    {
        g_object_ref (effect->tml);
        clutter_timeline_stop (effect->tml);
        g_signal_emit_by_name (effect->tml, "completed", NULL);
        g_object_unref (effect->tml);
    }
}

#define TRANSITION_SWITCH_WKS "switch-workspace"

static void
kill_switch_workspace (CdosBaseWM *cdos_wm, CdosWM *wm)
{
}

static void
on_switch_wks_complete (ClutterTimeline *tml, struct TransData *data)
{
    clutter_actor_set_position ((ClutterActor *)data->actor, data->orig.x, data->orig.y);
    MetaWindow *window = meta_window_actor_get_meta_window (data->actor);
    MetaRectangle rect;
    meta_window_get_buffer_rect (window, &rect);
    clutter_actor_set_position ((ClutterActor *)data->actor, rect.x, rect.y);
    if (data->visible == 0)
        clutter_actor_hide ((ClutterActor *)data->actor);

    g_free (data);

    g_mutex_lock (&mutex);
    animation_count--;
    g_mutex_unlock (&mutex);

    if (animation_count == 0)
    {
        CdosBaseWM *wm = cdos_global_get_window_manager (global);
        cdos_base_wm_completed_switch_workspace (wm);
        g_mutex_clear (&mutex);
    }
}

static void
on_switch_wks_stop (ClutterTimeline *tml, gboolean is_finished, struct TransData *data)
{
    if (!is_finished)
    {
        on_switch_wks_complete (tml, data);
    }
}

static void
switch_workspace (gpointer cdos_wm, gint from, gint to, gint direction, CdosWM *wm)
{
    ClutterTransition *trans = NULL;
    /* FIXME
     * 切换工作区动画可能会导致窗口不能正常最大/最小化等bug
     * 如果 expo、overview 正在显示，不显示动画
     */
    // if (!enable_desktop_effects ())
    {
        cdos_base_wm_completed_switch_workspace (CDOS_BASE_WM (cdos_wm));
        return;
    }
    GList *windows = cdos_global_get_window_actors (global);
    gint width, height;
    cdos_global_get_screen_size (global, &width, &height);

    /* @direction 是相机移动方向。
     * 窗口 actor 应该反向移动
     */
    gint x_dest = 0, y_dest = 0;
    if (direction == META_MOTION_UP ||
        direction == META_MOTION_UP_LEFT ||
        direction == META_MOTION_UP_RIGHT)
        y_dest = height;
    else if (direction == META_MOTION_DOWN ||
             direction == META_MOTION_DOWN_LEFT ||
             direction == META_MOTION_DOWN_RIGHT)
        y_dest = -height;

    if (direction == META_MOTION_LEFT ||
        direction == META_MOTION_UP_LEFT ||
        direction == META_MOTION_DOWN_LEFT)
        x_dest = width;
    else if (direction == META_MOTION_RIGHT ||
             direction == META_MOTION_UP_RIGHT ||
             direction == META_MOTION_DOWN_RIGHT)
        x_dest = -width;

    GList *iter = NULL;
    MetaDisplay *display = cdos_global_get_display (global);
    MetaWorkspaceManager *wsm = meta_display_get_workspace_manager (display);

    g_mutex_init (&mutex);

    for (iter = windows; iter != NULL; iter = iter->next)
    {
        MetaWindowActor *actor = iter->data;
        MetaWindow *window = meta_window_actor_get_meta_window (actor);

        if (!meta_window_showing_on_its_workspace (window))
            continue;
        if (window == wm->moving_window ||
                ((meta_display_get_grab_op (display) == META_GRAB_OP_MOVING
                  || meta_display_get_grab_op (display) == META_GRAB_OP_KEYBOARD_MOVING)
                  && window == meta_display_get_focus_window (display)))
        {
            /* clutter_actor_show_all () */
            clutter_actor_show (CLUTTER_ACTOR (actor));
            wm->moving_window = NULL;
        }
        else if (meta_workspace_manager_get_active_workspace_index (wsm) == from)
        {
            animation_count++;
            struct TransData *data = g_new0 (struct TransData, 1);
            data->actor = actor;
            data->visible = 0;
            clutter_actor_get_position ((ClutterActor *)actor, &data->orig.x, &data->orig.y);
            trans = cdos_transition_simple ((ClutterActor *)actor,
                                            CLUTTER_EASE_OUT_QUAD,
                                            WINDOW_ANIMATION_TIME,
                                            "x", data->orig.x + x_dest,
                                            "y", data->orig.y + y_dest,
                                            NULL);
            g_signal_connect (trans, "complete", G_CALLBACK(on_switch_wks_complete), data);
            g_signal_connect (trans, "stopped", G_CALLBACK(on_switch_wks_stop), data);
        }
        else if (meta_workspace_manager_get_active_workspace_index (wsm) == to)
        {
            animation_count++;
            struct TransData *data = g_new0 (struct TransData, 1);
            data->actor = actor;
            data->visible = 1;
            clutter_actor_get_position ((ClutterActor *)actor, &data->orig.x, &data->orig.y);
            clutter_actor_set_position ((ClutterActor *)actor, data->orig.x - x_dest, data->orig.y - y_dest);
            trans = cdos_transition_simple ((ClutterActor *)actor,
                                            CLUTTER_EASE_OUT_QUAD,
                                            WINDOW_ANIMATION_TIME,
                                            "x", data->orig.x,
                                            "y", data->orig.y,
                                            NULL);
            g_signal_connect (trans, "complete", G_CALLBACK(on_switch_wks_complete), data);
            g_signal_connect (trans, "stopped", G_CALLBACK(on_switch_wks_stop), data);
            clutter_actor_show ((ClutterActor *)actor);
        }
    }
}

static void
show_tile_preview (CdosBaseWM *base_wm, MetaWindow *window, MetaRectangle *rect, gint monitor, CdosWM *self)
{
    if (!self->tile_preview)
        self->tile_preview = cdos_tile_preview_new ();
    cdos_tile_preview_show ((CdosTilePreview *)self->tile_preview, window, rect, monitor);
}

static void
hide_tile_preview (CdosBaseWM *base_wm, CdosWM *self)
{
    if (!self->tile_preview)
        return;
    cdos_tile_preview_hide ((CdosTilePreview *)self->tile_preview);
}

static void
show_window_menu (CdosBaseWM *base_wm,
                  MetaWindow *window,
                  MetaWindowMenuType menu_type,
                  MetaRectangle *rect,
                  CdosWM *self)
{
    g_debug ("%s: %d", __func__, __LINE__);
}
static void
on_minimize_effect_complete (ClutterTimeline *tml, WindowData *data)
{
    ClutterActor *actor = data->actor;

    clutter_actor_set_scale (actor, 1.0f, 1.0f);
    clutter_actor_set_pivot_point (actor, 0.0f, 0.0f);
    clutter_actor_set_opacity (actor, data->orig_opacity);

    cdos_base_wm_completed_minimize (data->wm, META_WINDOW_ACTOR (actor));

    /*
    gfloat x, y;
    clutter_actor_get_transformed_position (actor, &x, &y);
    cdos_debug ("%f %f", x, y);
    clutter_actor_get_position (actor, &x, &y);
    cdos_debug ("%f %f", x, y);
    */

    EffectData *effect = get_effect_data ((MetaWindowActor *)data->actor);
    effect->state = 0;
    effect->tml = NULL;

    g_free (data);
}

static void
on_minimize_effect_stop (ClutterTimeline *tml, gboolean is_finished, WindowData *data)
{
    if (!is_finished)
    {
        EffectData *effect = get_effect_data ((MetaWindowActor *)data->actor);
        effect->state = MINIMIZING;
        effect->elapsed_time = clutter_timeline_get_elapsed_time (tml);
        clutter_actor_get_scale (data->actor, &effect->scale_x, &effect->scale_y);
        clutter_actor_get_position (data->actor, &effect->x, &effect->y);
        on_minimize_effect_complete (tml, data);
    }
}

static void
minimize_window (CdosBaseWM *cdos_wm, gpointer actor)
{
    cdos_sound_manager_play ("minimize");
    MetaWindow *window = meta_window_actor_get_meta_window (actor);

    gboolean is_skip = FALSE;
    const char *wm_class = meta_window_get_wm_class (window);
    if (g_strcmp0 (wm_class, "Nw") == 0 || g_strcmp0 (wm_class, "Nfs-browser") == 0
        || g_strcmp0 (wm_class, "Cdos-browser2") == 0
        || g_strcmp0 (wm_class, "cdos-picturebrowser") == 0)
        is_skip = TRUE;

    if (!enable_desktop_effects () || is_skip)
    {
        cdos_base_wm_completed_minimize (cdos_wm, actor);
        return;
    }
    clutter_actor_show (actor);

    WindowData *data = g_new0 (WindowData, 1);
    data->actor = actor;
    data->wm = cdos_wm;
    clutter_actor_get_scale (actor, &data->orig_scale_x, &data->orig_scale_y);
    data->orig_opacity = clutter_actor_get_opacity (actor);

    g_object_set_data ((GObject *)actor, "prev-state", GINT_TO_POINTER (TRUE));

    MetaRectangle rect;
    ClutterTransition *trans = NULL;
    EffectData *effect = get_effect_data (actor);
    if (meta_window_get_icon_geometry (window, &rect))
    {
        clutter_actor_set_pivot_point (actor, 0.0f, 0.0f);

        if (effect->state == MAPPING)
        {
            clutter_actor_set_scale (actor, effect->scale_x, effect->scale_y);
            clutter_actor_set_position (actor, effect->x, effect->y);
            effect->state = MINIMIZING;
        }
        trans = cdos_transition_simple (actor,
                                        CLUTTER_EASE_OUT_QUAD,
                                        WINDOW_ANIMATION_TIME,
                                        "scale-x", 0.0,
                                        "scale-y", 0.0,
                                        "opacity", 0,
                                        "x",  rect.x + rect.width / 2.0,
                                        "y", (double)rect.y,
                                        NULL);
    }
    else
    {
        clutter_actor_set_pivot_point (actor, 0.5f, 0.5f);
        trans = cdos_transition_simple (actor,
                                        CLUTTER_EASE_OUT_QUAD,
                                        WINDOW_ANIMATION_TIME,
                                        "scale-x",  0.0,
                                        "scale-y",  0.0,
                                        "opacity", 0,
                                        NULL);
    }
    effect->tml = (ClutterTimeline *)trans;
    g_signal_connect (trans, "completed", G_CALLBACK (on_minimize_effect_complete), data);
    g_signal_connect (trans, "stopped", G_CALLBACK (on_minimize_effect_stop), data);
}

static void
on_size_change_effect_complete (ClutterTimeline *tml, ClutterActor *actor)
{
    clutter_actor_set_scale (actor, 1.0, 1.0);
    clutter_actor_remove_all_transitions (actor);
    cdos_base_wm_completed_size_change (base_wm, (MetaWindowActor *)actor);
}

static void
size_change_window (CdosBaseWM *cdos_wm,
                    ClutterActor *actor,
                    MetaSizeChange which_change,
                    MetaRectangle *old_frame_rect,
                    MetaRectangle *old_buffer_rect)
{
    MetaWindow *window = meta_window_actor_get_meta_window ((MetaWindowActor *)actor);

    MetaMaximizeFlags max_flags = meta_window_get_maximized (window);

    if (max_flags == META_MAXIMIZE_BOTH)
    {
        cdos_sound_manager_play ("maximize");
    }
    else
    {
        cdos_sound_manager_play ("unmaximize");
    }

    gboolean is_skip = FALSE;
    const char *wm_class = meta_window_get_wm_class (window);
    if (g_strcmp0 (wm_class, "Nw") == 0 || g_strcmp0 (wm_class, "Nfs-browser") == 0
        || g_strcmp0 (wm_class, "Cdos-browser2") == 0
        || g_strcmp0 (wm_class, "cdos-picturebrowser") == 0
        || g_strcmp0 (wm_class, "smplayer") == 0)
        is_skip = TRUE;

    MetaWindowType type = meta_window_get_window_type (window);

    if (type != META_WINDOW_NORMAL || !enable_desktop_effects () || is_skip
        || (which_change == META_SIZE_CHANGE_FULLSCREEN)
        || (which_change == META_SIZE_CHANGE_UNFULLSCREEN))
    {
        cdos_base_wm_completed_size_change (cdos_wm, (MetaWindowActor *)actor);
        return;
    }
    MetaRectangle rect;
    meta_window_get_frame_rect (window, &rect);
    ClutterTransition *transition = NULL;
    transition = cdos_transition_simple (actor,
                                         CLUTTER_EASE_OUT_QUAD,
                                         50,
                                         "width", (double)rect.width,
                                         "height", (double)rect.height,
                                         "x", (double)rect.x,
                                         "y", (double)rect.y,
                                         NULL);
    g_signal_connect (transition, "completed", G_CALLBACK (on_size_change_effect_complete), actor);
}

static void
unminimize_window (CdosBaseWM *cdos_wm, MetaWindowActor *actor)
{
    cdos_sound_manager_play ("unminimize");
    cdos_base_wm_completed_unminimize (cdos_wm, actor);
}

static void
on_map_effect_complete (ClutterTimeline *tml, ClutterActor *actor)
{
    MetaWindowActor *window_actor = META_WINDOW_ACTOR (actor);

    clutter_actor_remove_all_transitions (actor);
    clutter_actor_set_opacity (actor, 255);
    clutter_actor_set_scale (actor, 1, 1);
    clutter_actor_set_translation (actor, 0, 0, 0);
    clutter_actor_set_pivot_point (actor, 0, 0);

    if (cdos_enable_blur_effect ())
    {
        ClutterEffect *blur_effect = st_blur_effect_new ();
        st_blur_effect_set_mode (ST_BLUR_EFFECT (blur_effect), ST_BLUR_MODE_BACKGROUND);
        st_blur_effect_set_sigma (ST_BLUR_EFFECT (blur_effect), 2.0);
        clutter_actor_add_effect (CLUTTER_ACTOR (actor), blur_effect);
    }
    cdos_base_wm_completed_map (cdos_global_get_window_manager (global), window_actor);

    EffectData *effect = get_effect_data (window_actor);
    effect->state = 0;
    effect->tml = NULL;
}

static void
on_map_effect_stop (ClutterTimeline *tml, gboolean is_finished, ClutterActor *actor)
{
    if (!is_finished)
    {
        EffectData *effect = get_effect_data ((MetaWindowActor *)actor);
        effect->state = MAPPING;
        effect->elapsed_time = clutter_timeline_get_elapsed_time (tml);
        clutter_actor_get_scale (actor, &effect->scale_x, &effect->scale_y);
        clutter_actor_get_position (actor, &effect->x, &effect->y);
        on_map_effect_complete (tml, actor);
    }
}

static void
on_dialog_closed (GPid pid, gint status, gpointer user_data)
{
    CdosWM *self = (CdosWM *)user_data;
    MetaPlugin *plugin = cdos_base_wm_get_plugin (self->cdos_wm);
    gboolean ok;

    ok = g_spawn_check_exit_status (status, NULL);
    self->monitor_watch_id = 0;
    meta_plugin_complete_display_change (plugin, ok);
}

static guint
xunli_confirm_display_change (CdosBaseWM *cdos_wm, CdosWM *self)
{
    GPid pid;
    MetaPlugin *plugin;
    //关闭窗口管理器内的倒计时。
    GMainContext *main_context;
    MetaMonitorManager *monitor_manager;
    GSource *source;

    main_context = g_main_context_default ();
    monitor_manager = meta_monitor_manager_get ();
    source = g_main_context_find_source_by_user_data (main_context, monitor_manager);
    if (source)
    {
        const gchar *name = g_source_get_name (source);
        if (g_strcmp0 (name, "[mutter] save_config_timeout") == 0)
        {
            guint id = g_source_get_id (source);
            if (id > 0)
            {
                g_source_remove (id);
            }
        }
    }
    if (self->monitor_watch_id > 0)
        return 0;

    plugin = cdos_base_wm_get_plugin (cdos_wm);
    pid = meta_show_dialog ("--question",
                            _("Does the display look OK?"),
                            "-1",
                            NULL,
                            _("_Keep This Configuration"),
                            _("_Restore Previous Configuration"),
                            "preferences-desktop-display",
                            0,
                            NULL,
                            NULL);
    self->monitor_watch_id = g_child_watch_add (pid, on_dialog_closed, self);
    return self->monitor_watch_id;
}

/*
 * 在特殊模式下过滤掉其他所有快捷键，除了 overlay-key
 * overlay-key 的过滤在 menu applet 中实现。
 * 补充： overlay-key 就是 super-L 打开隐藏菜单的键
 * */
extern int debug_mode;

static gboolean
cdos_filter_keybinding (CdosBaseWM *cdos_wm, MetaKeyBinding *binding, CdosWM *self)
{
    g_debug ("key name is : %s, current action mode: %d", meta_key_binding_get_name (binding), cdos_get_action_mode ());
    /* test */
    /*
    const gchar *screenshot = meta_key_binding_get_name (binding);
    if (debug_mode && g_str_has_prefix (screenshot, "external-grab-"))
    {
        static CdosScreenshot *sh = NULL;
        if (!sh)
            sh = cdos_screenshot_new ();
        cdos_screenshot_screenshot (sh, 1, "screenshot", NULL);
        return FALSE;
    }
    if (cdos_get_action_mode () == CDOS_ACTION_MODE_NONE)
    {
        if (g_strcmp0 (meta_key_binding_get_name (binding), "overlay-key") != 0)
            return TRUE;
        else
            return FALSE;
    }
    */
    if (cdos_get_action_mode () == CDOS_ACTION_MODE_POPUP)
    {
        if (g_strcmp0 (meta_key_binding_get_name (binding), "overlay-key") == 0)
            return FALSE;
    }
    if (cdos_get_action_mode () == CDOS_ACTION_MODE_NORMAL && meta_key_binding_is_builtin (binding))
    {
        return FALSE;
    }
    return cdos_keybindings_allow_keybinding (self->keybindings, meta_key_binding_get_name (binding));
}

static void
map_window (CdosBaseWM *cdos_wm, gpointer actor)
{
    MetaWindowActor *window_actor = META_WINDOW_ACTOR (actor);
    MetaWindow *window = meta_window_actor_get_meta_window (window_actor);
    ClutterTransition *trans = NULL;
    guint duration = 0;

    MetaWindowType type = meta_window_get_window_type (window);
    ClutterActor *clutter_actor = CLUTTER_ACTOR (actor);

    gboolean is_skip = FALSE;
    const char *wm_class = meta_window_get_wm_class (window);
    if (g_strcmp0 (wm_class, "Nw") == 0 || g_strcmp0 (wm_class, "Nfs-browser") == 0
        || g_strcmp0 (wm_class, "Cdos-browser2") == 0
        || g_strcmp0 (wm_class, "cdos-picturebrowser") == 0)
        is_skip = TRUE;

    if (!enable_desktop_effects () || is_skip)
    {
        cdos_base_wm_completed_map (cdos_wm, window_actor);
        return;
    }

    EffectData *effect = get_effect_data (window_actor);
    if ((type == META_WINDOW_NORMAL) || (type == META_WINDOW_DIALOG))
    {
        graphene_point_t dst;
        // clutter_actor_get_transformed_position (actor, &dst.x , &dst.y);
        MetaRectangle input_rect;
        meta_window_get_buffer_rect (window, &input_rect);
        dst.x = input_rect.x;
        dst.y = input_rect.y;
        clutter_actor_set_scale (actor, 0.5, 0.5);
        clutter_actor_show (actor);

        MetaRectangle rect;
        /* previous window state is minimized ? */
        gboolean prev_state = FALSE;
        prev_state = GPOINTER_TO_INT (g_object_get_data ((GObject *)actor, "prev-state"));

        if (!meta_window_is_attached_dialog (window) && meta_window_get_icon_geometry (window, &rect) && prev_state)
        {
            // clutter_actor_set_position (actor, (gfloat)(rect.x + rect.width/2), (gfloat)rect.y);
            clutter_actor_set_pivot_point (actor, 0.0f, 0.0f);

            if (effect->state == MINIMIZING)
            {
                effect->state = MAPPING;
                clutter_actor_set_position (actor, effect->x, effect->y);
                clutter_actor_set_scale (actor, effect->scale_x, effect->scale_y);
                duration = effect->elapsed_time;
            }
            else
            {
                clutter_actor_set_position (actor, rect.x + rect.width / 2.0, rect.y);
                clutter_actor_set_scale (actor, 0, 0);
                duration = MAP_TIME;
            }
            trans = cdos_transition_simple (actor,
                                            CLUTTER_EASE_IN_SINE,
                                            duration,
                                            "x", dst.x,
                                            "y", dst.y,
                                            "scale-x", 1.0,
                                            "scale-y", 1.0,
                                            NULL);
        }
        else
        {
            cdos_sound_manager_play ("map");
            clutter_actor_set_pivot_point (actor, 0.5f, 0.5f);

            trans = cdos_transition_simple (actor,
                                            CLUTTER_EASE_OUT_QUAD,
                                            MAP_TIME,
                                            "scale-x", 1.0,
                                            "scale-y", 1.0,
                                            NULL);
        }
    }
    else if (/* other window type */ 0)
    {
    }

    if (trans != NULL)
    {
        /* ClutterTransition 继承自 ClutterTimeline, 设置timeline 的 completed 信号 */
        g_signal_connect (trans, "completed", G_CALLBACK (on_map_effect_complete), clutter_actor);
        g_signal_connect (trans, "stopped", G_CALLBACK (on_map_effect_stop), clutter_actor);
        effect->tml = (ClutterTimeline *)trans;
    }
    else
        cdos_base_wm_completed_map (cdos_wm, window_actor);
}

static void
on_destroy_effect_complete (/*ClutterAnimation *anim*/ ClutterTimeline *tml, ClutterActor *actor)
{
    clutter_actor_remove_transition (actor, "unmap-window-scale-trans");
    MetaWindowActor *window_actor = META_WINDOW_ACTOR (actor);
    cdos_base_wm_completed_destroy (cdos_global_get_window_manager (global), window_actor);
}

static void
destroy_window (CdosBaseWM *cdos_wm, ClutterActor *actor)
{
    cdos_sound_manager_play ("close");
    MetaWindowActor *window_actor = META_WINDOW_ACTOR (actor);
    MetaWindow *window = meta_window_actor_get_meta_window (window_actor);

    MetaWindowType type = meta_window_get_window_type (window);
    if (type != META_WINDOW_NORMAL || !enable_desktop_effects ())
    {
        cdos_base_wm_completed_destroy (cdos_wm, window_actor);
        return;
    }
    clutter_actor_show (actor);
    clutter_actor_set_pivot_point (actor, 0.5f, 0.5f);
    ClutterTransition *trans;
    trans = cdos_transition_simple (actor,
                                    CLUTTER_EASE_IN_OUT_SINE,
                                    MAP_TIME,
                                    "scale-x", 0.0f,
                                    "scale-y", 0.0f,
                                    "opacity", 0,
                                    NULL);
    g_signal_connect (trans, "completed", G_CALLBACK (on_destroy_effect_complete), actor);
}


static void
shift_window_to_workspace (CdosWM *wm, MetaWindow *window, MetaMotionDirection direction)
{
    if (meta_window_get_window_type (window) == META_WINDOW_DESKTOP)
        return;
    wm->moving_window = window;
    MetaDisplay *display = cdos_global_get_display (global);
    MetaWorkspaceManager *wsm = meta_display_get_workspace_manager (display);
    MetaWorkspace *active = meta_workspace_manager_get_active_workspace (wsm);
    MetaWorkspace *wks = meta_workspace_get_neighbor (active, direction);
    if (wks != active)
    {
        meta_window_change_workspace (window, wks);
        meta_workspace_activate_with_focus (wks, window, cdos_global_get_current_time (global));
        show_workspace_OSD (display);
    }
}
/*
 * typedef void (* MetaKeyHandleFunc) (MetaDisplay *display,
 *                                     MetaScreen  *screen,
 *                                     MetaWindow  *window,
 *                                     XEvent      *event,
 *                                     MetaKeyBinding *binding,
 *                                     gpointer    user_data);
 */
static void
move_window_to_workspace_left (MetaDisplay    *display,
                               MetaWindow     *window,
                               XEvent         *event,
                               MetaKeyBinding *binding,
                               gpointer        data)
{
    CdosWM *wm = data;
    shift_window_to_workspace (wm, window, META_MOTION_LEFT);
}

static void
move_window_to_workspace_right (MetaDisplay    *display,
                                MetaWindow     *window,
                                XEvent         *event,
                                MetaKeyBinding *binding,
                                gpointer        data)
{
    CdosWM *wm = data;
    shift_window_to_workspace (wm, window, META_MOTION_RIGHT);
}

/**
 * cdos_wm_action_move_workspace_left:
 *
 * 切换到左侧工作区
 *
 */
void
cdos_wm_action_move_workspace_left (void)
{
    MetaDisplay *display = cdos_global_get_display (global);
    MetaWorkspaceManager *wsm = meta_display_get_workspace_manager (display);
    MetaWorkspace *active = meta_workspace_manager_get_active_workspace (wsm);
    MetaWorkspace *neighbour = meta_workspace_get_neighbor (active, META_MOTION_LEFT);
    if (active != neighbour)
    {
        cdos_sound_manager_play ("switch");
        meta_workspace_activate (neighbour, cdos_global_get_current_time (global));
    }
}

/**
 * cdos_wm_action_move_workspace_right:
 *
 * 切换到右侧工作区
 *
 */
void
cdos_wm_action_move_workspace_right (void)
{
    MetaDisplay *display = cdos_global_get_display (global);
    MetaWorkspaceManager *wsm = meta_display_get_workspace_manager (display);
    MetaWorkspace *active = meta_workspace_manager_get_active_workspace (wsm);
    MetaWorkspace *neighbour = meta_workspace_get_neighbor (active, META_MOTION_RIGHT);
    if (active != neighbour)
    {
        cdos_sound_manager_play ("switch");
        meta_workspace_activate (neighbour, cdos_global_get_current_time (global));
    }
}

/**
 * cdos_wm_action_move_workspace_up
 *
 * 切换到上一个工作区
 */
void
cdos_wm_action_move_workspace_up (void)
{
    MetaDisplay *display = cdos_global_get_display (global);
    MetaWorkspaceManager *wsm = meta_display_get_workspace_manager (display);
    MetaWorkspace *active = meta_workspace_manager_get_active_workspace (wsm);
    MetaWorkspace *neighbour = meta_workspace_get_neighbor (active, META_MOTION_UP);

    meta_workspace_activate (neighbour, cdos_global_get_current_time (global));
}

/**
 * cdos_wm_action_move_workspace_down:
 *
 * 切换到下一个工作区
 */
void
cdos_wm_action_move_workspace_down (void)
{
    MetaDisplay *display = cdos_global_get_display (global);
    MetaWorkspaceManager *wsm = meta_display_get_workspace_manager (display);
    MetaWorkspace *active = meta_workspace_manager_get_active_workspace (wsm);
    MetaWorkspace *neighbour = meta_workspace_get_neighbor (active, META_MOTION_DOWN);

    meta_workspace_activate (neighbour, cdos_global_get_current_time (global));
}

/**
 * cdos_wm_action_flip_workspace_left:
 *
 * 切到左侧工作区，并重置光标位置
 */
void
cdos_wm_action_flip_workspace_left (void)
{
    MetaDisplay *display = cdos_global_get_display (global);
    MetaWorkspaceManager *wsm = meta_display_get_workspace_manager (display);
    MetaWorkspace *active = meta_workspace_manager_get_active_workspace (wsm);
    MetaWorkspace *neighbour = meta_workspace_get_neighbor (active, META_MOTION_LEFT);
    if (active == neighbour)
        return;
    meta_workspace_activate (neighbour, cdos_global_get_current_time (global));
}

/**
 * cdos_wm_action_flip_workspace_right:
 *
 * 切到右侧工作区，并重置光标位置
 */
void
cdos_wm_action_flip_workspace_right (void)
{
    MetaDisplay *display = cdos_global_get_display (global);
    MetaWorkspaceManager *wsm = meta_display_get_workspace_manager (display);
    MetaWorkspace *active = meta_workspace_manager_get_active_workspace (wsm);
    MetaWorkspace *neighbour = meta_workspace_get_neighbor (active, META_MOTION_RIGHT);
    if (active == neighbour)
        return;
    meta_workspace_activate (neighbour, cdos_global_get_current_time (global));
}

// add by liting
static void
hide_snap_OSD (void)
{
    if (snap_osd != NULL)
    {
        clutter_actor_hide (CLUTTER_ACTOR (snap_osd));
        CdosLayoutManager *layout = cdos_layout_manager_get_default ();
        cdos_layout_manager_remove_chrome (layout, snap_osd);
        clutter_actor_destroy (snap_osd);
        snap_osd = NULL;
    }
}

static void
hide_workspace_OSD (ClutterTimeline *tml, void *data)
{
    if (workspace_osd != NULL)
    {
        clutter_actor_hide (CLUTTER_ACTOR (workspace_osd));
        clutter_actor_destroy (workspace_osd);
        workspace_osd = NULL;
    }
}

void
show_workspace_OSD (MetaDisplay *display)
{
    MetaWorkspaceManager *wsm = meta_display_get_workspace_manager (display);

    hide_snap_OSD ();
    hide_workspace_OSD (NULL, NULL);
    GSettings *settings = cdos_get_cdos_settings ();
    gboolean visible = g_settings_get_boolean (settings, "workspace-osd-visible");
    CdosLayoutManager *layout = cdos_layout_manager_get_default ();
    if (visible)
    {
        guint current_workspace_index = meta_workspace_manager_get_active_workspace_index (wsm);

        const MetaRectangle *monitor = cdos_layout_manager_get_primary_monitor (layout);
        if (workspace_osd == NULL)
        {
            workspace_osd = g_object_new (ST_TYPE_LABEL, "style-class", "workspace-osd", NULL);
        }
        st_label_set_text ((StLabel *)workspace_osd, cdos_get_workspace_name (current_workspace_index));
        clutter_actor_set_opacity (workspace_osd, 0);
        clutter_actor_add_child (cdos_ui_plugin_get_ui_group (), workspace_osd);

        gfloat label_width = clutter_actor_get_width (workspace_osd);
        gfloat label_height = clutter_actor_get_height (workspace_osd);
        guint x = monitor->x + ((gint)(monitor->width - label_width) >> 1);
        guint y = monitor->y + ((gint)(monitor->height - label_height) >> 1);
        clutter_actor_set_position (workspace_osd, x, y);
        // 开始动画
        guint duration = g_settings_get_int (settings, "workspace-osd-duration");

        clutter_actor_save_easing_state (workspace_osd);
        ClutterTransition *transition = clutter_property_transition_new ("opacity");
        clutter_actor_set_easing_mode (workspace_osd, CLUTTER_EASE_IN_SINE);
        clutter_timeline_set_duration (CLUTTER_TIMELINE (transition), duration);
        clutter_timeline_set_repeat_count (CLUTTER_TIMELINE (transition), 0);
        clutter_transition_set_from (transition, G_TYPE_UCHAR, 0);
        clutter_transition_set_to (transition, G_TYPE_UCHAR, 255);
        clutter_actor_add_transition (workspace_osd, "show-workspace-osd", transition);
        g_signal_connect (transition, "completed", G_CALLBACK (hide_workspace_OSD), display);
        clutter_actor_restore_easing_state (workspace_osd);
    }
}


/* 切换工作区时，先执行 show_workspace_switcher(),
 * 切换完成后，执行 switch_workspace ();
 */
static void
show_workspace_switcher (MetaDisplay    *display,
                         MetaWindow     *window,
                         XEvent         *event,
                         MetaKeyBinding *binding,
                         gpointer       user_data)
{
    const char *name = meta_key_binding_get_name (binding);
    if (name == NULL)
        return;
    if (strcmp (name, "switch-to-workspace-up") == 0)
    {
        CdosOverview *overview = cdos_overview_get ();
        cdos_overview_toggle (overview);
        return;
    }
    if (strcmp (name, "switch-to-workspace-down") == 0)
    {
        CdosOverview *overview = cdos_overview_get ();
        cdos_overview_toggle (overview);
        return;
    }
    MetaWorkspaceManager *wsm = meta_display_get_workspace_manager (display);
    if (meta_workspace_manager_get_n_workspaces (wsm) == 1)
        return;

    gint current_wks_idx = meta_workspace_manager_get_active_workspace_index (wsm);
    if (strcmp (name, "switch-to-workspace-left") == 0)
    {
        cdos_wm_action_move_workspace_left ();
        if (current_wks_idx != meta_workspace_manager_get_active_workspace_index (wsm))
            show_workspace_OSD (display);
    }
    else if (strcmp (name, "switch-to-workspace-right") == 0)
    {
        cdos_wm_action_move_workspace_right ();
        if (current_wks_idx != meta_workspace_manager_get_active_workspace_index (wsm))
            show_workspace_OSD (display);
    }
}

static void
start_appswitcher (MetaDisplay      *display,
                    MetaWindow      *window,
                    ClutterKeyEvent *event,
                    MetaKeyBinding  *binding,
                    gpointer         user_data)
{
    if (cdos_get_windows_for_binding (binding) == NULL)
        return;
    cdos_coverflow_switcher_new (binding);
}

/*
static void
screensaver_lock (void)
{
    cdos_screen_saver_lock ();
}
*/

static void
show_run_dialog (void)
{
    cdos_run_dialog_open (cdos_run_dialog_get_default ());
}

static void
switch_monitor (void)
{
    g_debug ("%s TODO", __func__);
    cdos_monitor_switcher_get_default ();
}
/**
 * cdos_wm_new:
 * @cdos_wm: #CdosBaseWM
 *
 * 创建一个 #CdosWM, 管理窗口切换、映射、关闭和切换工作区等的动画及绑定快捷键。
 *
 * Returns: #CdosWM
 *
 */
CdosWM *
cdos_wm_new (CdosBaseWM *cdos_wm)
{
    base_wm = cdos_wm;
    CdosWM *wm = g_new0 (CdosWM, 1);
    wm->should_animate = 1;
    wm->cdos_wm = cdos_wm;
    wm->keybindings = cdos_keybindings_new ();

    should_animate ();

    g_signal_connect (cdos_wm, "kill-window-effects", G_CALLBACK (kill_window_effects), NULL);
    g_signal_connect (cdos_wm, "kill-switch-workspace", G_CALLBACK (kill_switch_workspace), wm);
    g_signal_connect (cdos_wm, "switch-workspace", G_CALLBACK (switch_workspace), wm);
    g_signal_connect (cdos_wm, "minimize", G_CALLBACK (minimize_window), NULL);
    g_signal_connect (cdos_wm, "unminimize", G_CALLBACK (unminimize_window), NULL);
    g_signal_connect (cdos_wm, "size-change", G_CALLBACK (size_change_window), NULL);
    g_signal_connect (cdos_wm, "map", G_CALLBACK (map_window), NULL);
    g_signal_connect (cdos_wm, "filter-keybinding", G_CALLBACK (cdos_filter_keybinding), wm);
    g_signal_connect (cdos_wm, "confirm-display-change", G_CALLBACK (xunli_confirm_display_change), wm);
    g_signal_connect (cdos_wm, "destroy", G_CALLBACK (destroy_window), NULL);
    g_signal_connect (cdos_wm, "show-tile-preview", G_CALLBACK (show_tile_preview), wm);
    g_signal_connect (cdos_wm, "hide-tile-preview", G_CALLBACK (hide_tile_preview), wm);
    g_signal_connect (cdos_wm, "show-window-menu", G_CALLBACK (show_window_menu), wm);

    meta_keybindings_set_custom_handler ("move-to-workspace-left", (MetaKeyHandlerFunc)move_window_to_workspace_left, wm, NULL);
    meta_keybindings_set_custom_handler ("move-to-workspace-right", (MetaKeyHandlerFunc)move_window_to_workspace_right, wm, NULL);
    meta_keybindings_set_custom_handler ("switch-to-workspace-left", (MetaKeyHandlerFunc)show_workspace_switcher, wm, NULL);
    meta_keybindings_set_custom_handler ("switch-to-workspace-right", (MetaKeyHandlerFunc)show_workspace_switcher, wm, NULL);
    meta_keybindings_set_custom_handler ("switch-to-workspace-up", (MetaKeyHandlerFunc)show_workspace_switcher, wm, NULL);
    meta_keybindings_set_custom_handler ("switch-to-workspace-down", (MetaKeyHandlerFunc)show_workspace_switcher, wm, NULL);

    meta_keybindings_set_custom_handler ("switch-applications", (MetaKeyHandlerFunc)start_appswitcher, NULL, NULL);
    meta_keybindings_set_custom_handler ("switch-group", (MetaKeyHandlerFunc)start_appswitcher, NULL, NULL);
    meta_keybindings_set_custom_handler ("switch-windows-backward", (MetaKeyHandlerFunc)start_appswitcher, NULL, NULL);
    meta_keybindings_set_custom_handler ("cycle-windows", (MetaKeyHandlerFunc)start_appswitcher, NULL, NULL);
    meta_keybindings_set_custom_handler ("switch-group-backward", (MetaKeyHandlerFunc)start_appswitcher, NULL, NULL);
    meta_keybindings_set_custom_handler ("switch-panels", (MetaKeyHandlerFunc)start_appswitcher, NULL, NULL);
    meta_keybindings_set_custom_handler ("switch-panels-backward", (MetaKeyHandlerFunc)start_appswitcher, NULL, NULL);
    meta_keybindings_set_custom_handler ("panel-run-dialog", (MetaKeyHandlerFunc)show_run_dialog, NULL, NULL);
    meta_keybindings_set_custom_handler ("switch-monitor", (MetaKeyHandlerFunc)switch_monitor, NULL, NULL);

    return wm;
}
