#include <math.h>
#include "cdos-workspace.h"
#include "cdos-window-tracker.h"
#include "cdos-window-clone.h"
#include "cdos-window-clone-private.h"
#include "cdos-marshal.h"
#include "cdos-window-manager.h"
#include "cdos-layout-manager.h"

/**
 * SECTION:cdos-workspace
 * @Title: CdosWorkspace
 * @Short_description: 工作区
 *
 * CdosWorkspace 就是单个工作区，不同于工作区缩略图，这个对象与 CdosWorkspaceMonitor 一起
 * 管理应用窗口缩略图。
 */

G_DEFINE_TYPE (CdosWorkspaceMonitor, cdos_workspace_monitor, CLUTTER_TYPE_ACTOR);

#define DEFAULT_SLOT_FRACTION 0.825
#define WINDOWOVERLAY_ICON_SIZE 32
#define ANIMATION_TIME 250

enum
{
    FOCUS_REFRESH_REQUIRED,
    SELECTION_CHANGED,
    LAST_SIGNAL
};

#define N_POSITIONS 6

static graphene_point3d_t positions[] = {
                                    {0.5, 0.525, 0.875},
                                    {0.25, 0.525, 0.48},
                                    {0.75, 0.525, 0.48},
                                    {0.25, 0.275, 0.48},
                                    {0.75, 0.275, 0.48},
                                    {0.5, 0.75, 0.48},
                                    /* 4 */
                                    {0.25, 0.275, 0.47},
                                    {0.75, 0.275, 0.47},
                                    {0.25, 0.75, 0.47},
                                    {0.75, 0.75, 0.47},
                                    /* 5 */
                                    {0.165, 0.25, 0.32},
                                    {0.495, 0.25, 0.32},
                                    {0.825, 0.25, 0.32},
                                    {0.25, 0.75, 0.32},
                                    {0.75, 0.75, 0.32},
                                    /* 6 */
                                    {0.165, 0.25, 0.32},
                                    {0.495, 0.25, 0.32},
                                    {0.825, 0.25, 0.32},
                                    {0.165, 0.75, 0.32},
                                    {0.495, 0.75, 0.32},
                                    {0.825, 0.75, 0.32}};

static guint cdos_workspace_signals[LAST_SIGNAL] = {0};
static void cdos_workspace_select_active_clone (CdosWorkspace *workspace,
                                                CdosWindowClone *clone,
                                                CdosWorkspaceMonitor *cwm);

static void cdos_workspace_monitor_select_clone (CdosWorkspaceMonitor *cwm, CdosWindowClone *clone);

static void cdos_workspace_monitor_select_index (CdosWorkspaceMonitor *cwm, guint index);

static CdosWindowClone *cdos_workspace_monitor_add_window_clone (CdosWorkspaceMonitor *obj, MetaWindowActor *win);

static int sort_by_stable_sequence (void *a, void *b, GList *windows);

static void cdos_workspace_monitor_position_windows (CdosWorkspaceMonitor *cwm, guint flag);

static void cdos_workspace_monitor_do_window_added (CdosWorkspaceMonitor *obj, MetaWindow *mwindow);

struct _CdosWorkspacePrivate
{
    GList *monitors; //成员为 CdosWorkspaceMonitor
    CdosWindowClone *activeClone;
    guint currentMonitorIndex;
    gulong restack_id;
};

G_DEFINE_TYPE_WITH_PRIVATE (CdosWorkspace, cdos_workspace, CLUTTER_TYPE_ACTOR);

static gboolean
is_my_window (CdosWorkspaceMonitor *cwm, MetaWindowActor *win)
{
    MetaWindow *mwin = meta_window_actor_get_meta_window (win);
    if (meta_window_get_transient_for (mwin))
        return FALSE;

    MetaDisplay *display = cdos_global_get_display (cdos_global_get ());
    int primary_monitor_index = meta_display_get_primary_monitor (display);

    return (meta_window_located_on_workspace (mwin, cwm->meta_workspace)
        && (!mwin || cwm->monitor_index == primary_monitor_index));
}

static gboolean
cdos_workspace_monitor_is_overview_win (MetaWindowActor *win)
{
    MetaWindow *mwindow = NULL;

    mwindow = meta_window_actor_get_meta_window (win);

    return cdos_is_interesting (mwindow);
}

static gint
cdos_workspace_monitor_lookup_index (CdosWorkspaceMonitor *obj, MetaWindow *mwin)
{
    guint index = 0;
    GList *windows = obj->windows;
    while (windows != NULL)
    {
        GList *next = windows->next;
        CdosWindowClone *clone = windows->data;

        if (clone->meta_window == mwin)
        {
            return index;
        }
        index++;
        windows = next;
    }
    return -1;
}

static gboolean
delayed_window_repositioning (CdosWorkspaceMonitor *self)
{
    cdos_workspace_monitor_position_windows (self, WINDOW_POSITION_ANIMATE);
    return FALSE;
}
static void
cdos_workspace_monitor_do_remove_window (CdosWorkspaceMonitor *obj, MetaWindow *mwin)
{
    MetaWindowActor *win = (MetaWindowActor *)meta_window_get_compositor_private (mwin);
    GList *iter;
    for (iter = obj->windows; iter; iter = iter->next)
    {
        CdosWindowClone *clone = iter->data;
        if (clone->meta_window == mwin)
            break;
    }
    if (iter == NULL)
    {
        /*
        for (iter = obj->windows; iter; iter = iter->next)
        {
            CdosWindowClone *clone = iter->data;
            MetaWindow *w = clone->meta_window;
            Window W = meta_window_get_xwindow (w);
        }
        */
        return;
    }

    if (win && is_my_window (obj, win))
    {
        return;
    }

    CdosWindowClone *clone = iter->data;
    obj->windows = g_list_remove (obj->windows, clone);
    if (win)
    {
        /*
        ClutterPoint xy;
        ClutterSize size;
        clutter_actor_get_transformed_position ((ClutterActor *)clone, &xy.x, &xy.y);
        clutter_actor_get_transformed_size ((ClutterActor *)clone, &size.width, &size.height);
        // FIXME
        */
    }
    gint len = g_list_length (obj->windows);
    if (obj->kb_window_index >= len)
        obj->kb_window_index = len - 1;

    if (clone == obj->my_workspace->priv->activeClone)
        cdos_workspace_monitor_select_index (obj, obj->kb_window_index);

    meta_later_add (META_LATER_BEFORE_REDRAW, (GSourceFunc)delayed_window_repositioning, obj, NULL);
}

static void
add_window_actor (CdosWorkspaceMonitor *cwm, MetaWindowActor *actor, MetaWindow *window)
{
    if (cdos_workspace_monitor_lookup_index (cwm, window) != -1)
        return;

    if (!is_my_window (cwm, actor) || !cdos_workspace_monitor_is_overview_win (actor))
    {
        return;
    }
    CdosWindowClone *clone = cdos_workspace_monitor_add_window_clone (cwm, actor);

    clutter_actor_set_position ((ClutterActor *)clone, clone->orig_x, clone->orig_y);
    ClutterActor *stage = NULL;
    stage = clutter_actor_get_stage ((ClutterActor *)cwm);
    if (stage)
    {
        cdos_workspace_monitor_position_windows (cwm, WINDOW_POSITION_ANIMATE);
        //g_signal_emit_by_name (cwm->my_workspace, "focus-refresh-required");
    }
}

typedef struct _WindowAddId
{
    CdosWorkspaceMonitor *self;
    MetaWindow *window;
    guint id;
} WindowAddId;

static WindowAddId *
find_id (GPtrArray *arr, MetaWindow *window)
{
    guint i = 0;
    for (; i < arr->len; i++)
    {
        if (window == ((WindowAddId *)arr->pdata[i])->window)
            return arr->pdata[i];
    }
    return NULL;
}

static gboolean
add_window_idle (WindowAddId *id)
{
    CdosWorkspaceMonitor *obj = id->self;
    MetaWindow *window = id->window;

    MetaWindowActor *actor = META_WINDOW_ACTOR (meta_window_get_compositor_private (window));
    MetaWorkspace *my_workspace = meta_window_get_workspace (window);

    if (actor && (my_workspace == obj->meta_workspace))
    {
        add_window_actor (obj, actor, window);
        g_source_remove (id->id);
        id->id = 0;
        g_ptr_array_remove (obj->add_window_ids, id);
        return FALSE;
    }

    return TRUE;
}

static void
cdos_workspace_monitor_do_window_added (CdosWorkspaceMonitor *obj, MetaWindow *window)
{
    if (obj->leaving_overview)
        return;

    MetaWindowActor *actor = META_WINDOW_ACTOR (meta_window_get_compositor_private (window));

    if (!actor)
    {
        WindowAddId *id;
        id = find_id (obj->add_window_ids, window);
        if (id)
            return;
        id = g_new0 (WindowAddId, 1);
        g_ptr_array_add (obj->add_window_ids, id);
        id->window = window;
        id->self = obj;
        id->id = g_idle_add ((GSourceFunc)add_window_idle, id);

        return;
    }
    add_window_actor (obj, actor, window);
}

static void
clone_on_workspace_changed (CdosWindowClone *clone, gpointer data)
{
    CdosWorkspaceMonitor *cwm = (CdosWorkspaceMonitor *)data;

    cdos_workspace_monitor_do_remove_window (cwm, clone->meta_window);
}

static void
cdos_workspace_monitor_on_clone_selected (CdosWindowClone *clone, CdosWorkspaceMonitor *cwm)
{
    cdos_workspace_monitor_select_clone (cwm, clone);
}

static void
cdos_workspace_monitor_on_clone_show_overlay (CdosWindowClone *clone,
                                              CdosWorkspaceMonitor *cwm)
{
    GList *iter;
    for (iter = cwm->windows; iter; iter = iter->next)
    {
        CdosWindowClone *tmp = iter->data;
        if (tmp != clone)
        {
            cdos_window_clone_hide_overlay (tmp, TRUE);
        }
    }
}

static void
cdos_workspace_monitor_on_clone_activated (CdosWindowClone *clone, CdosWorkspaceMonitor *cwm)
{
    guint32 time = cdos_global_get_current_time (cdos_global_get ());
    guint wsIndex = -1;
    if (cwm->meta_workspace)
    {
        wsIndex = meta_workspace_index (cwm->meta_workspace);
        cdos_activate_window (clone->meta_window, time, wsIndex);
    }
}

static void
cdos_workspace_monitor_on_clone_closed (CdosWindowClone *clone, gpointer data)
{
    guint32 time = cdos_global_get_current_time (cdos_global_get ());
    meta_window_delete (clone->meta_window, time);
}

static void G_GNUC_UNUSED
on_clone_size_changed (CdosWindowClone *clone, CdosWorkspaceMonitor *cwm)
{
    cdos_workspace_monitor_position_windows (cwm, 0);
}

static CdosWindowClone *
cdos_workspace_monitor_add_window_clone (CdosWorkspaceMonitor *cwm, MetaWindowActor *win)
{
    CdosWindowClone *clone = NULL;
    clone = cdos_window_clone_new (win);
    g_signal_connect (clone, "workspace-changed", G_CALLBACK (clone_on_workspace_changed), cwm);
    g_signal_connect (clone, "selected", G_CALLBACK (cdos_workspace_monitor_on_clone_selected), cwm);
    g_signal_connect (clone, "show-overlay", G_CALLBACK (cdos_workspace_monitor_on_clone_show_overlay), cwm);
    g_signal_connect (clone, "activated", G_CALLBACK (cdos_workspace_monitor_on_clone_activated), cwm);
    g_signal_connect (clone, "closed", G_CALLBACK (cdos_workspace_monitor_on_clone_closed), cwm);

    // g_signal_connect (clone, "size-changed", G_CALLBACK(on_clone_size_changed), cwm);
    _add_actor (cwm, clone);
    cwm->windows = g_list_append (cwm->windows, clone);
    return clone;
}

static void
on_window_added (MetaWorkspace *workspace, MetaWindow *window, CdosWorkspaceMonitor *self)
{
    cdos_workspace_monitor_do_window_added (self, window);
}

static void
on_window_removed (MetaWorkspace *workspace, MetaWindow *window, CdosWorkspaceMonitor *self)
{
    // 这个回调函数发生在： overview 中关闭窗口
    cdos_workspace_monitor_do_remove_window (self, window);
}

static guint
calc_new_index (guint itemCount, guint curRow, guint numCols, guint numRows, guint curCol, gint rowDelta)
{
    guint newIndex = (curRow + rowDelta) * numCols + curCol;
    if (rowDelta >= 0)
    {
        return newIndex < itemCount ? newIndex : (curCol < numCols - 1 ? curCol + 1 : 0);
    }
    else
    {
        guint numFullRows = floor (itemCount / numCols);
        guint numIOILR = itemCount % numCols;

        return newIndex >= 0
                   ? newIndex
                   : (curCol == 0 ? (numFullRows * numCols) - 1
                                  : (numIOILR && curCol > numIOILR ? (((numFullRows - 1) * numCols) + curCol - 1)
                                                                   : (((numRows - 1) * numCols) + curCol - 1)));
    }
}

/**
 * grid_navigator_next_index:
 * @itemCount: 总数
 * @numCols: 列数
 * @currentIndex: 索引值
 * @symbol: 上下左右键
 *
 * 在网格模式中计算键盘按上下左右键切换到窗口或者工作区缩略图的索引值
 *
 * Return value: 索引值
 */
guint
grid_navigator_next_index (guint itemCount, guint numCols, guint currentIndex, guint symbol)
{
    gint result = -1;
    if (itemCount > 3 && (symbol == CLUTTER_KEY_Down || symbol == CLUTTER_KEY_Up))
    {
        guint numRows = ceil (itemCount / numCols);
        guint curRow = currentIndex / numCols;
        guint curCol = currentIndex % numCols;
        if (symbol == CLUTTER_KEY_Down)
        {
            result = calc_new_index (itemCount, curRow, numCols, numRows, curCol, 1);
        }

        if (symbol == CLUTTER_KEY_Up)
        {
            result = calc_new_index (itemCount, curRow, numCols, numRows, curCol, -1);
        }
    }
    else if (symbol == CLUTTER_KEY_Left || symbol == CLUTTER_KEY_Up)
    {
        result = (currentIndex < 1 ? itemCount : currentIndex) - 1;
    }
    else if (symbol == CLUTTER_KEY_Right || symbol == CLUTTER_KEY_Down)
    {
        result = (currentIndex + 1) % itemCount;
    }
    else if (symbol == CLUTTER_KEY_Home)
    {
        result = 0;
    }
    else if (symbol == CLUTTER_KEY_End)
    {
        result = itemCount - 1;
    }
    return result;
}

static gboolean
cdos_workspace_monitor_select_another_window (CdosWorkspaceMonitor *cwm, guint key_val)
{
    g_debug ("TODO: %s", __func__);
    return FALSE;
    guint numWindows = g_list_length (cwm->windows);
    if (numWindows == 0)
    {
        return FALSE;
    }
    guint currentIndex = cwm->kb_window_index;
    guint numCols = ceil (sqrt (numWindows));
    guint nextIndex = grid_navigator_next_index (numWindows, numCols, currentIndex, key_val);
    if (nextIndex < 0)
    {
        return FALSE;
    }
    cdos_workspace_monitor_select_index (cwm, nextIndex);
    return TRUE;
}

static gboolean
initial_position_idle (CdosWorkspaceMonitor *cwm)
{
    if (cwm->kb_window_index == 0)
    {
        cdos_workspace_monitor_select_another_window (cwm, CLUTTER_KEY_Home);
    }
    return FALSE;
}

/**
 * cdos_workspace_monitor_on_initial_position_windows:
 * @cwm: #CdosWorkspaceMonitor 对象
 *
 * 初始化时调用，设置默认选中的
 *
 *
 */
static void
cdos_workspace_monitor_on_initial_position_windows (CdosWorkspaceMonitor *cwm)
{
    cdos_workspace_monitor_select_another_window (cwm, CLUTTER_KEY_Home);
    cwm->initial_position_idle_id = g_idle_add ((GSourceFunc)initial_position_idle, cwm);
}

static CdosWorkspaceMonitor *
cdos_workspace_monitor_new (MetaWorkspace *meta_workspace,
                            guint monitor_index,
                            CdosWorkspace *workspace,
                            gboolean hasFocus)
{
    CdosWorkspaceMonitor *cwm = g_object_new (CDOS_TYPE_WORKSPACE_MONITOR, NULL);

    cwm->meta_workspace = meta_workspace;
    cwm->my_workspace = workspace;
    cwm->add_window_ids = g_ptr_array_new_with_free_func ((GDestroyNotify)g_free);

    cwm->monitor_index = monitor_index;
    cwm->drop_rect = clutter_actor_new ();
    _add_actor (cwm, cwm->drop_rect);

    /*
    cwm->windowOverlaysGroup = clutter_actor_new ();
    clutter_actor_set_size (cwm->windowOverlaysGroup, 0, 0);
    _add_actor (cwm, cwm->windowOverlaysGroup);
    */

    cwm->windows = NULL;
    GList *window_actors = cdos_global_get_window_actors (global);
    GList *iter = window_actors;
    while (iter != NULL)
    {
        MetaWindowActor *win = iter->data;
        if (is_my_window (cwm, win))
        {
            // if (meta_window_located_on_workspace (meta_window_actor_get_meta_window(win), meta_workspace))
            if (cdos_workspace_monitor_is_overview_win (win))
            {
                cdos_workspace_monitor_add_window_clone (cwm, win);
            }
        }
        iter = iter->next;
    }
    g_list_free (window_actors);

    if (cwm->meta_workspace != NULL)
    {
        cwm->window_added_id = g_signal_connect (meta_workspace, "window-added", G_CALLBACK (on_window_added), cwm);
        cwm->window_removed_id = g_signal_connect (meta_workspace, "window-removed", G_CALLBACK (on_window_removed), cwm);
    }
    cwm->leaving_overview = FALSE;
    cwm->stacked_once = FALSE;
    cwm->initial_position_idle_id = 0;
    if (hasFocus)
    {
        cwm->on_initial_position_windows = TRUE; //对应的函数只执行一次
    }

    return cwm;
}

typedef struct _SlotGeometry
{
    gfloat x;
    gfloat y;
    gfloat width;
    gfloat height;
} SlotGeometry;

typedef struct _AnimateCompleteData
{
    CdosWorkspaceMonitor *cwm;
    CdosWindowClone *clone;
} AnimateCompleteData;

/*
 * 计算出每个 window clone 位置的相对系数，比如 0.75,0.25 之类的数值
 */
static graphene_point3d_t *
compute_all_window_slots (CdosWorkspaceMonitor *cwm, guint window_count)
{
    if (window_count == 0)
    {
        return NULL;
    }
    graphene_point3d_t *slots = g_new0 (graphene_point3d_t, window_count);

    guint gridW = ceil (sqrt (window_count));
    guint gridH = ceil ((float)window_count / gridW);

    gfloat fraction = DEFAULT_SLOT_FRACTION * ((gfloat) (1.0 / gridW));
    cwm->slot_width = floor (fraction * cwm->width);
    guint i;
    for (i = 0; i < window_count; i++)
    {
        if (window_count <= N_POSITIONS)
        {
            gint idx = window_count * (window_count - 1) / 2 + i;
            slots[i] = positions[idx];
        }
        else
        {
            slots[i].z = fraction;
            slots[i].x = 0.5 / gridW + (gfloat) (i % gridW) / gridW;
            slots[i].y = 0.5 / gridH + floor (i / gridW) / gridH;
        }
    }
    return slots;
}

static SlotGeometry *
cdos_workspace_monitor_get_slot_geometry (CdosWorkspaceMonitor *cwm, graphene_point3d_t *slot)
{
    gfloat width;
    gfloat height;
    gfloat x;
    gfloat y;

    width = (gfloat) ((cwm->width - cwm->margin * 2) * slot->z);
    height = (gfloat) ((cwm->height - cwm->margin * 2) * slot->z);

    x = (gfloat) (cwm->x + cwm->margin + slot->x * (cwm->width - cwm->margin * 2) - width / 2);
    y = (gfloat) (cwm->y + cwm->margin + slot->y * (cwm->height - cwm->margin * 2) - height / 2);

    SlotGeometry *sg = g_malloc0 (sizeof (SlotGeometry));
    sg->x = x;
    sg->y = y;
    sg->width = width;
    sg->height = height;
    return sg;
}

static void
cdos_workspace_monitor_compute_window_layout (CdosWorkspaceMonitor *cwm,
                                              CdosWindowClone *clone,
                                              graphene_point3d_t *slot,
                                              graphene_point3d_t *res)
{
    MetaRectangle rect;
    SlotGeometry *sg;
    gfloat button_outer_height = 0.0, caption_icon_height = 0.0;
    gfloat button_outer_width = 0.0;
    gfloat x, y, scale;

    sg = cdos_workspace_monitor_get_slot_geometry (cwm, slot);
    meta_window_get_frame_rect (clone->meta_window, &rect);
    cdos_window_clone_chrome_heights (clone, &button_outer_height, &caption_icon_height);
    cdos_window_clone_chrome_widths (clone, &button_outer_width, &button_outer_width);

    scale = MIN ((gfloat) ((sg->width - button_outer_width) / rect.width),
                 (gfloat) ((sg->height - button_outer_height - caption_icon_height) / rect.height));
    scale = MIN (scale, 1.0);
    x = floor (sg->x + (sg->width - scale * rect.width) / 2.0);
    y = floor (sg->y + ((sg->height - (scale * rect.height)) / 2.0) - caption_icon_height);

    res->x = x;
    res->y = y;
    res->z = scale;

    g_free (sg);
}

static void
cdos_workspace_monitor_position_windows (CdosWorkspaceMonitor *cwm, guint flag)
{
    GList *clones = cwm->windows;
    guint initial_positioning = flag & WINDOW_POSITION_INITIAL;
    guint animate = flag & WINDOW_POSITION_ANIMATE;

    graphene_point3d_t *slots = compute_all_window_slots (cwm, g_list_length (clones));

    MetaDisplay *display = cdos_global_get_display (global);
    MetaWorkspaceManager *wsm = meta_display_get_workspace_manager (display);
    MetaWorkspace *current_workspace = meta_workspace_manager_get_active_workspace (wsm);
    gboolean is_on_current_workspace = (cwm->meta_workspace == NULL) || (cwm->meta_workspace == current_workspace);

    GList *tmp = g_list_copy (clones);
    tmp = g_list_sort_with_data (tmp, (GCompareDataFunc)sort_by_stable_sequence, clones);
    GList *iter = NULL;
    guint i = 0;
    for (iter = tmp; iter; iter = iter->next, i++)
    {
        gfloat width, height;
        graphene_point3d_t *slot = &slots[i];
        CdosWindowClone *clone = iter->data;

        MetaWindow *mwin = clone->meta_window;

        graphene_point3d_t res;
        cdos_workspace_monitor_compute_window_layout (cwm, clone, slot, &res);
        clutter_actor_set_pivot_point ((ClutterActor *)clone, 0.0, 0.0);
        width = clone->orig_width * res.z;
        height = clone->orig_height * res.z;

        if (animate && is_on_current_workspace)
        {
            guint8 target_opacity = clutter_actor_get_opacity ((ClutterActor *)clone);
            gboolean animate_a = FALSE;
            if (!meta_window_showing_on_its_workspace (mwin))
            {
                if (initial_positioning)
                {
                    clutter_actor_set_opacity ((ClutterActor *)clone, 0);
                    clutter_actor_set_position ((ClutterActor *)clone, res.x, res.y);
                    clutter_actor_set_size ((ClutterActor *)clone, 0, 0);
                }
                target_opacity = 255;
                animate_a = TRUE;
            }
            if (initial_positioning)
                clutter_actor_set_position ((ClutterActor *)clone, clone->orig_x, clone->orig_y);
            cdos_transition_simple ((ClutterActor *)clone,
                                    CLUTTER_EASE_OUT_QUAD,
                                    ANIMATION_TIME,
                                    "x", res.x,
                                    "y", res.y,
                                    "width", width,
                                    "height", height,
                                    NULL);
            if (animate_a)
            {
                clutter_actor_set_opacity ((ClutterActor *)clone, target_opacity);
            }
        }
        else
        {
            clutter_actor_set_position ((ClutterActor *)clone, res.x, res.y);
            clutter_actor_set_size ((ClutterActor *)clone, width, height);
        }
    }
    if (cwm->on_initial_position_windows)
    {
        cdos_workspace_monitor_on_initial_position_windows (cwm);
        g_source_remove (cwm->initial_position_idle_id);
        cwm->initial_position_idle_id = 0;
        cwm->on_initial_position_windows = FALSE;
    }
    g_free (slots);
    g_list_free (tmp);
    display_current_time ("position windows done");
}

static int
sort_by_stable_sequence (void *a, void *b, GList *windows)
{
    CdosWindowClone *clone_a = (CdosWindowClone *)a;
    CdosWindowClone *clone_b = (CdosWindowClone *)b;

    gboolean minimize_a = FALSE;
    g_object_get (clone_a->meta_window, "minimized", &minimize_a, NULL);
    gboolean minimize_b = FALSE;
    g_object_get (clone_b->meta_window, "minimized", &minimize_b, NULL);

    gint minimizedDiff = minimize_a - minimize_b;
    if (minimizedDiff)
        return minimizedDiff;

    gint index_a = -1;
    gint index_b = -1;

    GList *iter;
    gint i = 0;
    for (iter = windows; iter; iter = iter->next)
    {
        if (clone_a->window_actor == iter->data)
            index_a = i;
        else if (clone_b->window_actor == iter->data)
            index_b = i;
        i++;
        if (index_a >= 0 && index_b >= 0)
            break;
    }

    // js: return minimizedDiff || (index_a - index_b );
    /*
     * C 里 (-1||*) 返回值为1, js里返回 -1.
     */
    return index_a - index_b;
}

static void
cdos_workspace_monitor_sync_stacking (CdosWorkspaceMonitor *cwm, GList *windows)
{
    GList *clones = cwm->windows;
    cwm->stacked_once = TRUE;

    GList *tmp;
    tmp = g_list_copy (clones);
    tmp = g_list_sort_with_data (tmp, (GCompareDataFunc)sort_by_stable_sequence, windows);

    GList *reverse = g_list_reverse (tmp);
    ClutterActor *below = cwm->drop_rect;
    GList *iter;
    for (iter = reverse; iter; iter = iter->next)
    {
        CdosWindowClone *clone = iter->data;
        cdos_window_clone_set_stack_above (clone, below);
        below = (ClutterActor *)clone;
    }
    g_list_free (reverse);
}

static void
cdos_workspace_monitor_set_geometry (CdosWorkspaceMonitor *cwm, MetaRectangle *rect, guint margin)
{
    cwm->x = rect->x;
    cwm->y = rect->y;
    cwm->width = rect->width;
    cwm->height = rect->height;
    cwm->margin = margin;
}

static void
cdos_workspace_monitor_zoom_to_overview (CdosWorkspaceMonitor *cwm, guint avail_height)
{
    cwm->leaving_overview = FALSE;
    cwm->height = avail_height;

    CdosOverview *overview = cdos_overview_get ();
    gboolean animationInProgress = overview->animation_in_progress;
    if (animationInProgress)
    {
        cdos_workspace_monitor_position_windows (cwm, WINDOW_POSITION_ANIMATE | WINDOW_POSITION_INITIAL);
    }
    else
    {
        cdos_workspace_monitor_position_windows (cwm, WINDOW_POSITION_INITIAL);
    }
}

static void
on_overview_hidden (CdosOverview *self, gpointer data)
{
    CdosWorkspaceMonitor *cwm = (CdosWorkspaceMonitor *)data;

    cwm->leaving_overview = FALSE;
}

static void
cdos_workspace_monitor_zoom_from_overview (CdosWorkspaceMonitor *cwm)
{
    MetaDisplay *display = cdos_global_get_display (global);
    MetaWorkspaceManager *wsm = meta_display_get_workspace_manager (display);
    MetaWorkspace *current_workspace = meta_workspace_manager_get_active_workspace (wsm);

    cwm->leaving_overview = TRUE;

    CdosOverview *overview = cdos_overview_get ();
    cwm->overview_hidden_id = g_signal_connect (overview, "hidden", G_CALLBACK (on_overview_hidden), cwm);

    if (cwm->meta_workspace != NULL && cwm->meta_workspace != current_workspace)
        return;

    ClutterAnimationMode trans_in = CLUTTER_EASE_IN_QUAD;
    ClutterAnimationMode trans_out = CLUTTER_EASE_OUT_QUAD;
    guint8 opacity = 255;
    GList *clones = cwm->windows;
    GList *tmp = g_list_copy (clones);
    tmp = g_list_sort_with_data (tmp, (GCompareDataFunc)sort_by_stable_sequence, clones);
    GList *iter = NULL;
    for (iter = tmp; iter; iter = iter->next)
    {
        CdosWindowClone *clone = iter->data;
        MetaWindow *mwin = clone->meta_window;
        ClutterActor *actor = (ClutterActor *)clone;
        if (meta_window_showing_on_its_workspace (mwin))
        {
            clutter_actor_save_easing_state (actor);
            clutter_actor_set_easing_mode (actor, trans_in);
            clutter_actor_set_easing_duration (actor, ANIMATION_TIME);
            clutter_actor_set_position (actor, clone->orig_x, clone->orig_y);
            clutter_actor_set_size (actor, clone->orig_width, clone->orig_height);
            clutter_actor_set_opacity (actor, opacity);
            clutter_actor_restore_easing_state (actor);
        }
        else
        {
            clutter_actor_save_easing_state (actor);
            clutter_actor_set_pivot_point (actor, 0.5, 0.5);
            clutter_actor_set_easing_mode (actor, trans_out);
            clutter_actor_set_easing_duration (actor, ANIMATION_TIME);
            clutter_actor_set_size (actor, 0, 0);
            clutter_actor_set_opacity (actor, 0);
            clutter_actor_restore_easing_state (actor);
        }
    }
    g_list_free (tmp);
}

static void
cdos_workspace_monitor_select_index (CdosWorkspaceMonitor *cwm, guint index)
{
    cwm->kb_window_index = index;

    CdosWindowClone *activeClone = NULL;
    if (index > -1 && index < g_list_length (cwm->windows))
    {
        activeClone = g_list_nth_data (cwm->windows, cwm->kb_window_index);
    }
    cdos_workspace_select_active_clone (cwm->my_workspace, activeClone, cwm);
}

static void
cdos_workspace_monitor_select_clone (CdosWorkspaceMonitor *cwm, CdosWindowClone *clone)
{
    // guint index = g_list_index (temptest, clone);
    guint index = 0;

    index = g_list_index (cwm->windows, clone);
    cdos_workspace_monitor_select_index (cwm, index);
}

static gboolean
cdos_workspace_monitor_activate_selected_window (CdosWorkspaceMonitor *cwm)
{
    if (cwm->kb_window_index > -1 && cwm->kb_window_index < g_list_length (cwm->windows))
    {
        CdosWindowClone *clone = g_list_nth_data (cwm->windows, cwm->kb_window_index);
        cdos_workspace_monitor_on_clone_activated (clone, cwm);
        return TRUE;
    }
    return FALSE;
}

static void
cdos_workspace_dispose (GObject *gobject)
{
    CdosWorkspace *self = CDOS_WORKSPACE (gobject);
    CdosWorkspacePrivate *priv = cdos_workspace_get_instance_private (self);

    if (priv->restack_id)
    {
        MetaDisplay *display = cdos_global_get_display (global);
        g_signal_handler_disconnect (display, priv->restack_id);
        priv->restack_id = 0L;
    }

    if (priv->monitors)
    {
        g_list_free (priv->monitors);
        priv->monitors = NULL;
    }

    G_OBJECT_CLASS (cdos_workspace_parent_class)->dispose (gobject);
}

static void
cdos_workspace_class_init (CdosWorkspaceClass *klass)
{
    GObjectClass *gobject_class = G_OBJECT_CLASS (klass);

    gobject_class->dispose = cdos_workspace_dispose;

    cdos_workspace_signals[FOCUS_REFRESH_REQUIRED] = g_signal_new ("focus-refresh-required",
                                                                   G_TYPE_FROM_CLASS (klass),
                                                                   G_SIGNAL_RUN_LAST,
                                                                   0,
                                                                   NULL,
                                                                   NULL,
                                                                   _cdos_marshal_VOID__INT_INT,
                                                                   G_TYPE_NONE,
                                                                   2,
                                                                   G_TYPE_INT,
                                                                   G_TYPE_INT);

    cdos_workspace_signals[SELECTION_CHANGED] = g_signal_new ("selection-changed",
                                                              G_TYPE_FROM_CLASS (klass),
                                                              G_SIGNAL_RUN_LAST,
                                                              0,
                                                              NULL,
                                                              NULL,
                                                              g_cclosure_marshal_VOID__VOID,
                                                              G_TYPE_NONE,
                                                              0);
}

static void
cdos_workspace_init (CdosWorkspace *self)
{
    CdosWorkspacePrivate *priv;
    priv = self->priv = cdos_workspace_get_instance_private (self);

    priv->monitors = NULL;
    priv->activeClone = NULL;
}

/**
 * cdos_workspace_zoom_to_overview:
 * @self: #CdosWorkspace 对象
 *
 * 进入窗口缩略图模式
 *
 */
void
cdos_workspace_zoom_to_overview (CdosWorkspace *self, guint avail_height)
{
    CdosWorkspacePrivate *priv = self->priv;
    GList *monitors = priv->monitors;
    GList *iter;
    for (iter = monitors; iter; iter = iter->next)
    {
        CdosWorkspaceMonitor *cwm = iter->data;
        cdos_workspace_monitor_zoom_to_overview (cwm, avail_height);
    }
}

/**
 * cdos_workspace_zoom_from_overview:
 * @self: #CdosWorkspace 对象
 *
 * 从窗口缩略图模式退出，回到正常桌面
 *
 */
void
cdos_workspace_zoom_from_overview (CdosWorkspace *self)
{
    CdosWorkspacePrivate *priv = self->priv;
    GList *monitors = priv->monitors;
    while (monitors != NULL)
    {
        GList *next = monitors->next;
        CdosWorkspaceMonitor *cwm = monitors->data;
        cdos_workspace_monitor_zoom_from_overview (cwm);
        monitors = next;
    }
}

/**
 * cdos_workspace_position_windows:
 * @self: #CdosWorkspace 对象
 * @flag: 是否第一次设置
 *
 * 获取当前工作区的所有窗口的克隆列表，根据屏幕宽高、窗口个数计算出每个克隆窗口的起始坐标和缩放比例，
 * 并设置显示的位置
 */
void
cdos_workspace_position_windows (CdosWorkspace *self, guint flag)
{
    CdosWorkspacePrivate *priv = self->priv;
    GList *monitors = priv->monitors;

    while (monitors != NULL)
    {
        GList *next = monitors->next;
        CdosWorkspaceMonitor *cwm = monitors->data;
        if (cwm->my_workspace == self)
        {
            cdos_workspace_monitor_position_windows (cwm, flag);
            break;
        }
        monitors = next;
    }
}

static void
cdos_workspace_sync_stacking (MetaDisplay *display, CdosWorkspace *self)
{
    CdosWorkspacePrivate *priv = self->priv;

    GList *monitors = priv->monitors;
    GList *windows = cdos_global_get_window_actors (global);
    while (monitors != NULL)
    {
        GList *next = monitors->next;
        CdosWorkspaceMonitor *cwm = monitors->data;
        cdos_workspace_monitor_sync_stacking (cwm, windows);
        monitors = next;
    }
}

/**
 * cdos_workspace_on_key_press:
 * @self: #CdosWorkspace 对象
 * @actor: 可空
 * @event: 键盘事件
 *
 * 对键盘事件做出响应，空格、回车键激活选中的窗口，并退出缩略图模式回到正常桌面
 */
gboolean
cdos_workspace_on_key_press (CdosWorkspace *self, ClutterEvent *event)
{
    ClutterModifierType modifiers = clutter_event_get_state (event);
    guint symbol = clutter_event_get_key_symbol (event);
    CdosWorkspacePrivate *priv = self->priv;

    ClutterModifierType ctrlAltMask = CLUTTER_CONTROL_MASK | CLUTTER_MOD1_MASK;

    CdosWorkspaceMonitor *currentMonitor = g_list_nth_data (priv->monitors, priv->currentMonitorIndex);
    if ((symbol == CLUTTER_KEY_Return) ||
        (symbol == CLUTTER_KEY_space) ||
        (symbol == CLUTTER_KEY_KP_Enter))
    {
        if (cdos_workspace_monitor_activate_selected_window (currentMonitor))
        {
            return TRUE;
        }
        return FALSE;
    }

    if (modifiers & ctrlAltMask)
    {
        return FALSE;
    }
    return cdos_workspace_monitor_select_another_window (currentMonitor, symbol);
}

static void
cdos_workspace_select_active_clone (CdosWorkspace *workspace, CdosWindowClone *clone, CdosWorkspaceMonitor *cwm)
{
    CdosWorkspacePrivate *priv = workspace->priv;

    if (clone)
    {
        priv->currentMonitorIndex = cwm->monitor_index;
    }
    priv->activeClone = clone;
    g_signal_emit_by_name (workspace, "selection-changed");
}

/**
 * cdos_workspace_new:
 * @meta_workspace: #MetaWorkspace 对象
 * @view: #CdosWorkspacesView 对象
 *
 * 新建 #CdosWorkspace 对象
 */
CdosWorkspace *
cdos_workspace_new (MetaWorkspace *meta_workspace, CdosWorkspacesView *view)
{
    CdosWorkspace *self = g_object_new (CDOS_TYPE_WORKSPACE, NULL);
    CdosWorkspacePrivate *priv = self->priv;
    self->meta_workspace = meta_workspace;

    CdosLayoutManager *layout = cdos_layout_manager_get_default ();
    priv->currentMonitorIndex = cdos_layout_manager_get_index ();

    GPtrArray *monitors = cdos_layout_manager_get_monitors (layout);

    guint index = 0;
    while (monitors != NULL && index < monitors->len)
    {
        gboolean hasFocus = FALSE;
        if (index == priv->currentMonitorIndex)
        {
            hasFocus = TRUE;
        }

        CdosWorkspaceMonitor *cwm = NULL;
        cwm = cdos_workspace_monitor_new (meta_workspace, index, self, hasFocus);
        MetaRectangle *rect = monitors->pdata[index];

        /* ClutterGeometry */
        cdos_workspace_monitor_set_geometry (cwm, rect, 0);
        priv->monitors = g_list_append (priv->monitors, cwm);
        _add_actor (self, cwm);
        index++;
    }
    MetaDisplay *display = cdos_global_get_display (global);
    priv->restack_id = g_signal_connect (display, "restacked", G_CALLBACK (cdos_workspace_sync_stacking), self);

    return self;
}

static void
cdos_workspace_monitor_init (CdosWorkspaceMonitor *self)
{

}

static void
cdos_workspace_monitor_dispose (GObject *object)
{
    CdosWorkspaceMonitor *cwm = (CdosWorkspaceMonitor *)object;

    if (cwm->overview_hidden_id)
    {
        CdosOverview *overview = cdos_overview_get ();
        g_signal_handler_disconnect (overview, cwm->overview_hidden_id);
        cwm->overview_hidden_id = 0;
    }

    if (cwm->meta_workspace)
    {
        g_signal_handler_disconnect (cwm->meta_workspace, cwm->window_added_id);
        g_signal_handler_disconnect (cwm->meta_workspace, cwm->window_removed_id);
        cwm->meta_workspace = NULL;
    }

    if (cwm->windows)
    {
        /* 确保CdowWindowClone在CdosWindowOverlay之前释放 */
        for (GList *iter = cwm->windows; iter; iter = iter->next)
        {
            ClutterActor *clone = iter->data;
            clutter_actor_destroy (clone);
        }
        g_list_free (cwm->windows);
        cwm->windows = NULL;
    }
    if (cwm->add_window_ids)
    {
        guint i;
        for (i = 0; i < cwm->add_window_ids->len; i++)
        {
            WindowAddId *id = cwm->add_window_ids->pdata[i];
            g_source_remove (id->id);
        }
        g_ptr_array_free (cwm->add_window_ids, 1);
        cwm->add_window_ids = NULL;
    }
    G_OBJECT_CLASS (cdos_workspace_monitor_parent_class)->dispose (object);
}

static void
cdos_workspace_monitor_class_init (CdosWorkspaceMonitorClass *klass)
{
    GObjectClass *gobject_class = G_OBJECT_CLASS (klass);
    gobject_class->dispose = cdos_workspace_monitor_dispose;
}
