#include <stdlib.h>
#include <meta/meta-workspace-manager.h>
#include "cdos-workspace.h"
#include "cdos-workspaces-view.h"
#include "cdos-overview.h"
//#include "cdos-export-overview.h"
#include "cdos-overview-workspace-thumbnail.h"
#include "cdos-layout-manager.h"
#include "cdos-transition.h"

/**
 * SECTION:cdos-workspaces-view
 * @Title: CdosWorkspacesView
 * @Short_description: 工作区视图
 *
 * CdosWorkspacesView 的成员 actor 保存所有 workspace，cdos 桌面环境默认有四个工作区，不提供增加删除功能，
 * 作为 CdosOverview 和 CdosWorkspace 之间的桥梁， CdosWorkspacesView 在初始化时会先获取当前所有的工作
 * 区，保存到链表中，并隐藏非 active 的工作区。
 */

typedef struct _CdosWorkspacesView
{
    StWidget parent;
    guint width;
    guint height;
    guint x;
    guint y;
    guint workspace_ratio_spacing;
    guint spacing;

    gboolean animating;

    GList *workspaces;

    gulong n_workspaces_changed_id;
    gulong overview_showing_id;
    gulong switch_workspace_notify_id;
} CdosWorkspacesView;

G_DEFINE_TYPE (CdosWorkspacesView, cdos_workspaces_view, ST_TYPE_WIDGET);

static void
cdos_workspaces_view_update_visibility (CdosWorkspacesView *self)
{
    CdosGlobal *global = cdos_global_get ();
    MetaDisplay *display = cdos_global_get_display (global);
    MetaWorkspaceManager *wsm = meta_display_get_workspace_manager (display);
    MetaWorkspace *active = meta_workspace_manager_get_active_workspace (wsm);
    GList *workspaces = self->workspaces;
    GList *iter;

    for (iter = workspaces; iter; iter = iter->next)
    {
        CdosWorkspace *workspace = iter->data;

        if (self->animating)
        {
            clutter_actor_show ((ClutterActor *)workspace);
        }
        else
        {
            if (active == workspace->meta_workspace)
            {
                // cdos_workspace_position_windows (workspace, 0);
                cdos_workspace_zoom_to_overview (workspace, self->height);
                clutter_actor_show ((ClutterActor *)workspace);
            }
            else
                clutter_actor_hide ((ClutterActor *)workspace);
        }
    }
}

static void
on_workspaces_changed (MetaWorkspaceManager *wsm, GParamSpec *spec, gpointer data)
{
    CdosWorkspacesView *self = (CdosWorkspacesView *)data;

    GList *workspaces = self->workspaces;

    while (workspaces != NULL)
    {
        GList *next = workspaces->next;
        CdosWorkspace *workspace = workspaces->data;
        clutter_actor_destroy ((ClutterActor *)workspace);
        workspaces = next;
    }
    g_list_free (self->workspaces);
    self->workspaces = NULL;

    GList *wks = meta_workspace_manager_get_workspaces (wsm);
    GList *iter;

    for (iter = wks; iter; iter = iter->next)
    {
        MetaWorkspace *lastWs = iter->data;
        CdosWorkspace *newWs= cdos_workspace_new (lastWs, self);
        self->workspaces = g_list_append (self->workspaces, newWs);
        clutter_actor_add_child ((ClutterActor *)self, (ClutterActor *)newWs);
    }

    self->animating = FALSE;
    cdos_workspaces_view_update_visibility (self);
}

gboolean
cdos_workspaces_view_key_press_event (CdosWorkspacesView *self, ClutterKeyEvent *ek)
{
    ClutterEvent *event = (ClutterEvent *)ek;
    //ClutterModifierType modifiers = clutter_event_get_state (event);

    guint symbol = clutter_event_get_key_symbol (event);

    if (symbol == CLUTTER_KEY_Escape)
    {
        CdosOverview *overview = cdos_overview_get ();
        clutter_actor_hide ((ClutterActor *)overview);
        return TRUE;
    }

    CdosGlobal *global = cdos_global_get ();
    MetaDisplay *display = cdos_global_get_display (global);
    MetaWorkspaceManager *wsm = meta_display_get_workspace_manager (display);
    guint activeWorkspace_index = meta_workspace_manager_get_active_workspace_index (wsm);
    CdosWorkspace *activeWorkspace = g_list_nth_data (self->workspaces, activeWorkspace_index);
    return cdos_workspace_on_key_press (activeWorkspace, event);
}

static void
cdos_workspaces_view_set_geometry (CdosWorkspacesView *self, guint x, guint y,
                                   guint width, guint height, guint spacing)
{
    self->width = width;
    self->height = height;
    self->x = x;
    self->y = y;
    self->workspace_ratio_spacing = spacing;
}

static void
cdos_workspaces_view_hide (ClutterActor *actor)
{
    CdosWorkspacesView *self = CDOS_WORKSPACES_VIEW (actor);

    CdosGlobal *global = cdos_global_get ();
    MetaDisplay *display = cdos_global_get_display (global);
    MetaWorkspaceManager *wsm = meta_display_get_workspace_manager (display);
    guint activeWorkspace_index = meta_workspace_manager_get_active_workspace_index (wsm);
    CdosWorkspace *activeWorkspace = g_list_nth_data (self->workspaces, activeWorkspace_index);

    clutter_actor_set_child_above_sibling (actor, (ClutterActor *)activeWorkspace, NULL);
    cdos_workspace_zoom_from_overview (activeWorkspace);
    //CLUTTER_ACTOR_CLASS (cdos_workspaces_view_parent_class)->hide (actor);
}

static void
on_workspace_transitions_completed (ClutterTimeline *tml, CdosWorkspacesView *self)
{
    self->animating = FALSE;
    cdos_workspaces_view_update_visibility (self);
    g_debug ("%s", __func__);
}

static void
cdos_workspaces_view_update_workspace_actors (CdosWorkspacesView *self, gboolean animate)
{
    CdosGlobal *global = cdos_global_get ();
    MetaDisplay *display = cdos_global_get_display (global);
    MetaWorkspaceManager *wsm = meta_display_get_workspace_manager (display);
    guint active = meta_workspace_manager_get_active_workspace_index (wsm);

    CdosLayoutManager *layout = cdos_layout_manager_get_default ();
    GPtrArray *monitors = cdos_layout_manager_get_monitors (layout);
    //const MetaRectangle *primary = cdos_layout_manager_get_primary_monitor (layout);
    //gfloat w = primary->width * thumbnail_box_scale / 2.0;
    gboolean animating = self->animating;
    self->animating = (monitors->len < 2) && animate;

    guint i = 0;
    GList *workspaces = self->workspaces;
    GList *iter;

    for (iter = workspaces; iter; iter = iter->next, i++)
    {
        CdosWorkspace *workspace = iter->data;

        gfloat x = (i-active) * (self->width + self->spacing + self->workspace_ratio_spacing);
        //x += w;

        if (!animating && self->animating)
        {
            clutter_actor_show ((ClutterActor *)workspace);
            ClutterTransition *trans = cdos_transition_simple ((ClutterActor *)workspace,
                                                                CLUTTER_EASE_OUT_QUAD,
                                                                300,
                                                                "x", x,
                                                                NULL);
            if (i == 0)
            {
                g_signal_connect (trans, "completed",
                                  G_CALLBACK (on_workspace_transitions_completed), self);
            }
        }
        else
        {
            clutter_actor_set_position ((ClutterActor *)workspace, x, 0);
            if (i == 0)
                cdos_workspaces_view_update_visibility (self);
        }

    }
}

static void
on_switch_workspace (MetaDisplay *display,
                     gint from,
                     gint to,
                     gint direction,
                     CdosWorkspacesView *self)
{
    cdos_workspaces_view_update_workspace_actors (self, TRUE);
}

/*
static void
cdos_workspaces_view_map (ClutterActor *actor)
{
    CLUTTER_ACTOR_CLASS (cdos_workspaces_view_parent_class)->map (actor);

    CdosWorkspacesView *self = (CdosWorkspacesView *)actor;
    CdosWorkspacesViewPrivate *priv = CDOS_WORKSPACES_VIEW_GET_PRIVATE (self);

    gboolean ismapped;
    ismapped = clutter_actor_is_mapped (actor);

    CdosOverview *overview = cdos_overview_get ();
    if (ismapped){
        guint direction = SWIPE_DIRECTION_HORIZONTAL;
        cdos_overview_set_scroll_adjustment (overview, self->scrollAdjustment, direction);

        self->swipeScrollBeginId = g_signal_connect (overview, "swipe-scroll-begin",
                                        G_CALLBACK (on_swipe_scroll_begin), self);

        self->swipeScrollEndId = g_signal_connect (overview, "swipe-scroll-end",
                                        G_CALLBACK (on_swipe_scroll_end), self);
    }else {
        g_signal_handler_disconnect ( overview, self->swipeScrollBeginId);
        g_signal_handler_disconnect ( overview, self->swipeScrollEndId);
    }
}
*/

static void
cdos_workspaces_view_on_destroy (CdosWorkspacesView *self)
{
    if (self->switch_workspace_notify_id)
    {
        CdosGlobal *global = cdos_global_get ();
        CdosBaseWM *window_manager = cdos_global_get_window_manager (global);
        g_signal_handler_disconnect (window_manager, self->switch_workspace_notify_id);
        self->switch_workspace_notify_id = 0L;
    }
    if (self->n_workspaces_changed_id)
    {
        MetaDisplay *display = cdos_global_get_display (cdos_global_get ());
        MetaWorkspaceManager *wsm = meta_display_get_workspace_manager (display);
        g_signal_handler_disconnect (wsm, self->n_workspaces_changed_id);
        self->n_workspaces_changed_id = 0L;
    }
    if (self->overview_showing_id)
    {
        CdosOverview *overview = cdos_overview_get ();
        g_signal_handler_disconnect (overview, self->overview_showing_id);
        self->overview_showing_id = 0L;
    }
}

static void
cdos_workspaces_view_dispose (GObject *gobject)
{
    CdosWorkspacesView *self = CDOS_WORKSPACES_VIEW (gobject);

    cdos_workspaces_view_on_destroy (self);

    G_OBJECT_CLASS (cdos_workspaces_view_parent_class)->dispose (gobject);
}

static void
cdos_workspaces_view_finalize (GObject *gobject)
{
    CdosWorkspacesView *self = CDOS_WORKSPACES_VIEW (gobject);
    g_list_free (self->workspaces);
    self->workspaces = NULL;
    G_OBJECT_CLASS (cdos_workspaces_view_parent_class)->finalize (gobject);
}

static void
on_overview_showing (CdosOverview *overview, CdosWorkspacesView *self)
{
    clutter_actor_show ((ClutterActor *)self);

    CdosGlobal *global = cdos_global_get ();
    MetaDisplay *display = cdos_global_get_display (global);
    MetaWorkspaceManager *wsm = meta_display_get_workspace_manager (display);

    g_signal_handler_disconnect (overview, self->overview_showing_id);
    self->overview_showing_id = 0;
    guint activeWorkspace_index = meta_workspace_manager_get_active_workspace_index (wsm);

    CdosWorkspace *current_workspace = g_list_nth_data (self->workspaces, activeWorkspace_index);
    cdos_workspace_zoom_to_overview (current_workspace, self->height);
}

static void
cdos_workspaces_view_allocate (ClutterActor *actor,
                               const ClutterActorBox *box,
                               ClutterAllocationFlags flags)
{
    /*
    CdosWorkspacesView *self = (CdosWorkspacesView *)actor;
    guint height = box->y2 - box->y1;
    if (height != self->height)
    {
        self->height = height;
    }
    */
    CLUTTER_ACTOR_CLASS (cdos_workspaces_view_parent_class)->allocate (actor, box, flags);
}

static void
cdos_workspaces_view_init (CdosWorkspacesView *self)
{
    CdosGlobal *global = cdos_global_get ();
    MetaDisplay *display = cdos_global_get_display (global);
    MetaWorkspaceManager *wsm = meta_display_get_workspace_manager (display);

    self->workspace_ratio_spacing = 0;
    self->spacing = 0;
    self->animating = FALSE;

    CdosOverview *overview = cdos_overview_get ();
    self->overview_showing_id = g_signal_connect (overview, "showing", G_CALLBACK (on_overview_showing), self);

    CdosBaseWM *wm = cdos_global_get_window_manager (global);
    self->switch_workspace_notify_id = g_signal_connect (wm, "switch-workspace", G_CALLBACK (on_switch_workspace), self);

    self->n_workspaces_changed_id = g_signal_connect (wsm, "notify::n-workspaces",
                                    G_CALLBACK (on_workspaces_changed), self);

    CdosLayoutManager *layout = cdos_layout_manager_get_default ();
    const MetaRectangle *primary = cdos_layout_manager_get_primary_monitor (layout);
    cdos_workspaces_view_set_geometry (self, primary->x, primary->y, primary->width,
                                       primary->height, self->spacing);

    MetaWorkspace *active = meta_workspace_manager_get_active_workspace (wsm);
    GList *workspaces = meta_workspace_manager_get_workspaces (wsm);
    gint active_index = g_list_index (workspaces, active);
    gint i = 0;
    for (GList *iter = workspaces; iter; iter = iter->next, i++)
    {
        MetaWorkspace *workspace = iter->data;
        CdosWorkspace *cworkspace = cdos_workspace_new (workspace, self);
        self->workspaces = g_list_append (self->workspaces, cworkspace);
        clutter_actor_add_child ((ClutterActor *)self, (ClutterActor *)cworkspace);
        gfloat x = (i-active_index) * (self->width + self->spacing + self->workspace_ratio_spacing);
        clutter_actor_set_position ((ClutterActor *)cworkspace, x, 0);
        if (workspace != active)
            clutter_actor_hide ((ClutterActor *)cworkspace);
        else
            clutter_actor_set_child_above_sibling ((ClutterActor *)self, (ClutterActor *)cworkspace, NULL);
    }

    ClutterStage *stage = cdos_global_get_stage (global);
    clutter_stage_set_key_focus (stage, (ClutterActor *)self);
}

static void
cdos_workspaces_view_class_init (CdosWorkspacesViewClass *klass)
{
    GObjectClass *gobject_class = G_OBJECT_CLASS (klass);
    ClutterActorClass *actor_class = CLUTTER_ACTOR_CLASS (klass);

    gobject_class->dispose = cdos_workspaces_view_dispose;
    gobject_class->finalize = cdos_workspaces_view_finalize;

    actor_class->allocate = cdos_workspaces_view_allocate;
    actor_class->hide = cdos_workspaces_view_hide;
}

/**
 * cdos_workspaces_view_new:
 *
 * 新建 #CdosWorkspacesView 对象
 *
 * Return value: 工作区视图对象
 */
CdosWorkspacesView *
cdos_workspaces_view_new (void)
{
    CdosWorkspacesView *self;
    self = g_object_new (CDOS_TYPE_WORKSPACES_VIEW,
                         NULL);

    return self;
}

/**
 * cdos_workspaces_view_set_avail_size:
 * @workspaces_view: #CdosWorkspacesView 对象
 * @height: 去除 thumbnail_box 之后的高度
 *
 * 设置去除 thumbnail_box 之后的高度
 */
void cdos_workspaces_view_set_avail_size (CdosWorkspacesView *self,
                                          gint avail_height)
{
    self->height = avail_height;
}
